option(null: hard);

module open_api {

string get_url option(constant) = "/docs";

module __impl__ {

enum schema_element { object_prop, object_type, array_type_empty, array_type_filled };

class element_stack {
public:
	void push(schema_element e);
	schema_element pop();
	schema_element top();
	logical is_empty();

private:
	vector(schema_element) stack_;
};

void element_stack.push(schema_element e)
{
	push_back(stack_, e);
}

schema_element element_stack.pop()
{
	integer n = v_size(stack_);
	if (n == 0)
		throw (E_UNSPECIFIC, "Stack is empty");
	schema_element e = stack_[n-1];
	if (n == 1)
		stack_ = null<vector(schema_element)>;
	else
		stack_ = stack_[0:n-2];
	return e;
}

schema_element element_stack.top()
{
	integer n = v_size(stack_);
	if (n == 0)
		throw (E_UNSPECIFIC, "Stack is empty");
	return stack_[n-1];
}

logical element_stack.is_empty() = v_size(stack_) == 0;

}

class enum dtype {
	str option(str: "string")
		, num option(str: "number")
		, bool option(str: "boolean")
		, int option(str: "integer")
};

class schema {
public:
	schema(json.builder jb);
	string mk_json();
	json.node mk_json_node();

private:
	string json_;
};

schema.schema(json.builder jb)
: json_(jb.get())
{
}

string schema.mk_json() = json_;

json.node schema.mk_json_node() = json.parse(json_);


class schema_builder {
public:
	schema_builder();

	schema_builder set_type_boolean(logical option(nullable) dflt = null<logical>, logical option(nullable) example = null<logical>);
	schema_builder set_type_integer();
	schema_builder set_type_integer_w_default(integer dflt);
	schema_builder set_type_integer_w_example(integer example);
	schema_builder set_type_integer(integer dflt, integer example);
	schema_builder set_type_number(number option(nullable) dflt = null<number>, number option(nullable) example = null<number>);
	schema_builder set_type_string(string option(nullable) dflt = null<string>, string option(nullable) example = null<string>);
	schema_builder set_type_ref(string schema_builder_name);

	schema_builder start_type_object();
	schema_builder add_property(string name, logical requried);
	schema_builder end_type_object();

	schema_builder start_type_array();
	schema_builder end_type_array();

	schema get_schema();

private:
	json.builder jb_;
	logical finished_;
	logical broken_;
	__impl__.element_stack stack_;

	vector(string) required_;

	schema_builder _set_type(
		dtype t,
		string option(nullable) str_dflt,
		string option(nullable) str_ex,
		number option(nullable) num_dflt,
		number option(nullable) num_ex,
		logical option(nullable) logical_dflt,
		logical option(nullable) logical_ex);

	void _check_modify_allowed();
	void _throw(string msg);

	void _require(string name);

	void _check_property_allowed();
	void _check_type_allowed();

	void _type_ended();
};

schema_builder.schema_builder()
{
	jb_ = new json.builder();
	finished_ = false;
	broken_ = false;
	stack_ = new __impl__.element_stack();
	required_ = [""];
}

void schema_builder._check_modify_allowed()
{
	if (broken_)
		throw (E_UNSPECIFIC, "schema_builder has thrown exception and is not usable");
	if (finished_)
		throw (E_UNSPECIFIC, "schema_builder is finished, and cannot be modified");
}

void schema_builder._throw(string msg)
{
	broken_ = true;
	throw (E_UNSPECIFIC, msg);
}

void schema_builder._check_property_allowed()
{
	_check_modify_allowed();
	if (stack_.is_empty() || stack_.top() != __impl__.object_type)
		_throw("Not in object");
}

void schema_builder._check_type_allowed()
{
	_check_modify_allowed();
	if (stack_.is_empty())
		return;

	if (stack_.top() != __impl__.object_prop && stack_.top() != __impl__.array_type_empty)
		_throw("Cannot set type at this point");
}

void schema_builder._type_ended()
{
	if (stack_.is_empty()) {
		finished_ = true;
	} else if (stack_.top() == __impl__.object_prop) {
		stack_.pop();
	} else if (stack_.top() == __impl__.array_type_empty) {
		stack_.pop();
		stack_.push(__impl__.array_type_filled);
	}
}

void schema_builder._require(string name)
{
	integer k = v_size(required_)-1;
	required_[k] = strcat([required_[k], "|", name]);
}

schema_builder schema_builder._set_type(
	dtype t,
	string option(nullable) str_dflt,
	string option(nullable) str_ex,
	number option(nullable) num_dflt,
	number option(nullable) num_ex,
	logical option(nullable) logical_dflt,
	logical option(nullable) logical_ex)
{
	_check_type_allowed();
	jb_.begin_object();
	jb_.begin_member("type");
	jb_.append(string(t));

	logical has_default =
		(t == dtype.str && !null(str_dflt)) ||
		(t == dtype.num && !null(num_dflt)) ||
		(t == dtype.int && !null(num_dflt)) ||
		(t == dtype.bool && !null(logical_dflt));

	if (has_default) {
		jb_.begin_member("default");
		switch (t) {
		case dtype.str:
			jb_.append(str_dflt);
			break;
		case dtype.num:
		case dtype.int:
			jb_.append(num_dflt);
			break;
		case dtype.bool:
			jb_.append(logical_dflt);
			break;
		}
	}

	jb_.end_object();
	_type_ended();
	return this;
}

schema_builder schema_builder.set_type_string(string option(nullable) dflt, string option(nullable) example)
{
	return _set_type(dtype.str, dflt, example, null<number>, null<number>, null<logical>, null<logical>);
}

schema_builder schema_builder.set_type_boolean(logical option(nullable) dflt, logical option(nullable) example)
{
	return _set_type(dtype.bool, null<string>, null<string>, null<number>, null<number>, dflt, example);
}

schema_builder schema_builder.set_type_integer()
{
	return _set_type(dtype.int, null<string>, null<string>, null<number>, null<number>, null<logical>, null<logical>);
}

schema_builder schema_builder.set_type_integer_w_default(integer dflt)
{
	return _set_type(dtype.int, null<string>, null<string>, dflt, null<number>, null<logical>, null<logical>);
}

schema_builder schema_builder.set_type_integer_w_example(integer example)
{
	return _set_type(dtype.int, null<string>, null<string>, null<number>, example, null<logical>, null<logical>);
}

schema_builder schema_builder.set_type_integer(integer dflt, integer example)
{
	return _set_type(dtype.int, null<string>, null<string>, dflt, example, null<logical>, null<logical>);
}

schema_builder schema_builder.set_type_number(number option(nullable) dflt, number option(nullable) example)
{
	return _set_type(dtype.num, null<string>, null<string>, dflt, example, null<logical>, null<logical>);
}

schema_builder schema_builder.start_type_object()
{
	_check_type_allowed();
	jb_.begin_object();
	jb_.begin_member("properties");
	jb_.begin_object();
	push_back(required_, "");
	stack_.push(__impl__.object_type);
	return this;
}

schema_builder schema_builder.add_property(string name, logical requried)
{
	_check_property_allowed();
	stack_.push(__impl__.object_prop);
	jb_.begin_member(name);
	if (requried)
		_require(name);
	return this;
}

schema_builder schema_builder.end_type_object()
{
	_check_modify_allowed();
	if (stack_.is_empty() || stack_.pop() != __impl__.object_type)
		_throw("Not in object");
	jb_.end_object(); // end "properties"
	integer nr = v_size(required_);
	vector(string) req_here = str_tokenize(required_[nr-1], "|");
	if (v_size(req_here) > 1) {
		jb_.begin_member("required");
		jb_.append(req_here[1:]);
	}
	jb_.end_object();
	required_ = required_[0:nr-2];
	_type_ended();
	return this;
}

schema_builder schema_builder.start_type_array()
{
	_check_type_allowed();
	jb_.begin_object();
	jb_.begin_member("type");
	jb_.append("array");
	jb_.begin_member("items");
	stack_.push(__impl__.array_type_empty);
	return this;
}

schema_builder schema_builder.set_type_ref(string schema_builder_name)
{
	_check_type_allowed();
	jb_.begin_object();
	jb_.begin_member("$ref");
	jb_.append(strcat("#/components/schema_builders/", schema_builder_name));
	jb_.end_object();
	_type_ended();
	return this;
}

schema_builder schema_builder.end_type_array()
{
	_check_modify_allowed();
	if (stack_.is_empty() || stack_.pop() != __impl__.array_type_filled)
		_throw("Cannot end array here");
	jb_.end_object(); // end "items"
	_type_ended();
	return this;
}

schema schema_builder.get_schema()
{
	if (!finished_)
		throw (E_UNSPECIFIC, "Schema is incomplete");

	if (broken_)
		throw (E_UNSPECIFIC, "Schema is broken");

	return new schema(jb_);
}


class response_desc {
public:
	response_desc(integer status_code, string desc);
	response_desc(integer status_code, string desc, schema sch);

	void mk_json(json.builder jb);

	void _init(integer status_code, string desc, schema option(nullable) sch);

	integer status_code;
	string desc;
	schema sch;
};

response_desc.response_desc(integer status_code, string desc)
{
	_init(status_code, desc, null<schema>);
}

response_desc.response_desc(integer status_code, string desc, schema sch)
{
	_init(status_code, desc, sch);
}

void response_desc._init(integer status_code, string desc, schema option(nullable) sch)
{
	http_api.default_status_message(status_code); // check that status code is valid
	this.status_code = status_code;
	this.desc = desc;
	this.sch = sch;
}

void response_desc.mk_json(json.builder jb)
{
	jb.begin_member(str(status_code));
	jb.begin_object();
	jb.begin_member("description");
	jb.append(null(desc) ? http_api.default_status_message(status_code) : desc);
	if (!null(this.sch)) {
		jb.begin_member("content");
		jb.begin_object();
		jb.begin_member("application/json");
		jb.begin_object();
		jb.begin_member("schema");
		jb.append_node(sch.mk_json_node());
		jb.end_object();
		jb.end_object();
	}
	jb.end_object();
}


class method {
public:
	string summary;
	string description;
	vector(response_desc) response_descs;
};

class path {
public:
	method get;
	method put;
	method post;
	method delete;
};

} // module open_api
