option(null: hard);	
/*	
	instrument wrapper functions
	Developer: Algorithmica Research, Magnus Nyström			

	--------------------------------------------------------------------
	fltois
	--------------------------------------------------------------------
*/

//----------------------------------------------------------------------
//  class swap_fltois
//----------------------------------------------------------------------
class swap_fltois: public swap_gen
option(category: "Instrument/Interest Rate Swap/Float vs ONIndex")
//option(allow_undeclared_mfuncs)
{
public:
    //override string	instr_type_s();
	override instrument 	inst();
	override swap_fltois clone() ;
	
	override void         __dbg_print(__dbg_label);
    override void         __dbg_browse(__dbg_split);
	
	/*----------ADD FUNCS [PUBLIC]---------------*/

	void 		add_notional(number );
	void 		add_notional_flt_leg(number );
	void 		add_notional_ois_leg(number );
	void 		add_fx_rate_flt_leg(number );
	void 		add_fx_rate_ois_leg(number );
	void 		add_disc_func(disc_func ); 
	void 		add_disc_func_flt_leg(disc_func );
	void 		add_disc_func_ois_leg(disc_func );
	disc_func 	disc_func_flt_leg(error_info option(nullable) error = null<error_info> ) ;
	disc_func 	disc_func_ois_leg(error_info option(nullable) error = null<error_info> ) ;
	void 		add_surface_flt_leg(tenor_surface option(nullable));
	void 		add_surface_ois_leg(tenor_surface option(nullable));
	void		add_fixing_flt_leg(vector(date) ,vector(number) ,logical option(nullable) fixing_as_fwd = false);
	void		add_fixing_ois_leg(vector(date) ,vector(number) ,logical option(nullable) fixing_as_fwd = false);
	void		add_fixing_flt_leg(number ,logical option(nullable) fixing_as_fwd = false);
	void		add_fixing_ois_leg(number ,logical option(nullable) fixing_as_fwd = false);
	

	/*----------MISC FUNCS [PUBLIC]---------------*/
	override void 		currency(out string , out string ,error_info option(nullable) error = null<error_info> );
	override void  		pmt_bus_day(out bd_convention , out bd_convention ,error_info option(nullable) error = null<error_info> );
	override void 		eom(out logical ,out logical ,error_info option(nullable) error = null<error_info> );
	override void 		calendar(out ..calendar ,out ..calendar ,error_info option(nullable) error = null<error_info> );
	override void 		dc_method(	out day_count_method option(nullable), out day_count_method option(nullable),
										error_info option(nullable) error = null<error_info>);	
	override void 		coupon_freq(out integer,out integer,
										error_info option(nullable) error = null<error_info> );
	override void 		reset_freq(	out integer , out integer  , error_info option(nullable) error = null<error_info>) ;
	
	/*----------CASHFLOW FUNCS [PUBLIC]---------------*/
	override void		cash_flow_dates(logical,out vector(date) ,out vector(date) ,error_info option(nullable) error = null<error_info> ) ;
	void				prin_cash_flow_dates(logical,out vector(date) ,out vector(date) ,error_info option(nullable) error = null<error_info> ) ;
	
	/*----------PV FUNCS [PUBLIC]---------------*/
	override number  	present_value(	disc_func option(nullable),disc_func option(nullable),out number ,
								out number ,number option(nullable)=null<number>,logical option(nullable) trade_date_pv = false,		
								error_info option(nullable) error = null<error_info>);
	
	override number  	present_value(	disc_func option(nullable),disc_func option(nullable),out number ,out number ,
								date option(nullable) ,date option(nullable) ,
								logical,logical  ,error_info option(nullable) error = null<error_info>);
	
	override number  	present_value(	out number ,out number pv_ois_leg,
								logical option(nullable) ,error_info option(nullable) error = null<error_info> );

	override number		solver(	number,swap_solver_code ,logical option(nullable) trade_date_pv = false,
						error_info option(nullable) error = null<error_info>);

	
	

	/*----------LEG1/LEG2 FUNCS [PUBLIC]---------------*/
	override void			accr_dates(logical,out vector(date) ,out vector(date) ,out vector(number),
							   out vector(date) ,out vector(date) ,out vector(number),
								error_info option(nullable) error = null<error_info> ) ;
	
	ir_index 				ir_index_flt_leg(error_info option(nullable) error = null<error_info>);
	ir_index 				ir_index_ois_leg(error_info option(nullable) error = null<error_info> ); 
	tenor_code 				index_tenor_code_flt_leg(error_info option(nullable) error = null<error_info> ) ;
	
	string 					index_tenor_flt_leg(error_info option(nullable) error = null<error_info>) ;

	/*-----external swap specific-----*/
	override ql_fixed_income_swap 	swap_instrument(error_info option(nullable) error = null<error_info>) ;
	override instrument_type 		swap_type(error_info option(nullable) error = null<error_info>);
	vector(ql_swap_leg) 	legs(logical,error_info option(nullable) error = null<error_info>);
	ql_float_leg 			float_leg(logical,error_info option(nullable) error = null<error_info>);
	ql_ois_leg 				ois_leg(logical,error_info option(nullable) error = null<error_info>);

	
	//--------------------------------------------------------------------------
	//----------HIDDEN FUNCS [PUBLIC]--(mostly for backward compat)-------------
	//--------------------------------------------------------------------------
	swap_fltois(__instrument);
	//general-one per leg
	override string 		currency(error_info option(nullable) error = null<error_info> );				/*hidden, visible in root*/
	override logical 		eom(error_info option(nullable) error = null<error_info>);						/*hidden, not visible in root*/
	
	swap_fltois(swap_fltois );

protected:
	
	override swap_fltois	create(__instrument option(nullable),out instr_error option(nullable),error_type type = E_INVALID_ARG);

	void					SWAP_ENTRY_FUNC();
	
	logical ext_swap_;
	logical ccy_swap_;
	logical full_swap_;
};

//------------------------------------------------
// __dbg_print
//------------------------------------------------
void swap_fltois.__dbg_print(__dbg_label l)
{

	error_info ee 	= new error_info(true,false);
    l.set_text(strcat([
						"name: ",
						name(ee),
						",  ext_swap: ",
						null(ext_swap_) ? "" : string(ext_swap_),
                        ", ccy_swap: ",
						null(ccy_swap_) ? "" : string(ccy_swap_),
						", valid: ",
						null(this.is_valid(ee)) ? "" : string(this.is_valid(ee))
					]));
}
//------------------------------------------------
// __dbg_browse
//------------------------------------------------
void swap_fltois.__dbg_browse(__dbg_split s)
{
    s.resize(4);
	error_info ee 	= new error_info(true,false);
    s.set_text(0, "name");
    s.set_value(0, this.name(ee));

    s.set_text(1, "ext_swap");
    s.set_value(1, ext_swap_);

    s.set_text(2, "ccy_swap");
    s.set_value(2, ccy_swap_);

	s.set_text(3, "valid");
    s.set_value(3, is_valid(ee));
}

//------------------------------------------------
// constructor [HIDDEN, cannot be protected]
//------------------------------------------------
swap_fltois.swap_fltois(__instrument i) option(hidden): swap_gen(i)
{
	//try {
		error_info ee 	= new error_info(true,true);
		if(!this.is_valid(ee)) {
			ext_swap_ = ccy_swap_ = full_swap_ = false;
			return;
		}
		ext_swap_ 		= this.is_ext_swap(ee);
		ccy_swap_ 		= this.is_ccy_swap(ee);
		//FIX_ME 
		//QL_REQUIRE(!ccy_swap_,"flt-ois cross currency swap is currently work in progress");
		full_swap_ 		= this.verify_full_swap(ee);
		return;									
	//}
	//catch {		
	//	CORE_INT.write_warning_message(err.message());
	//}
}
//------------------------------------------------
// instr_type_s
//------------------------------------------------
//string swap_fltois.instr_type_s() 	{ return string(..instr_type.SWAP_FLTOIS);}
//------------------------------------------------
// inst
//------------------------------------------------
instrument 	swap_fltois.inst()		{ return this;}
//------------------------------------------------
// copy constructor
//------------------------------------------------
swap_fltois.swap_fltois(swap_fltois c) : swap_gen(c),
						  ext_swap_(c.ext_swap_), ccy_swap_(c.ccy_swap_), full_swap_(c.full_swap_) {}
//------------------------------------------------
// copy constructor <FUNCTION>
//------------------------------------------------
swap_fltois swap_fltois(swap_fltois b)			{ return new swap_fltois(b);}
//------------------------------------------------
// clone
//------------------------------------------------
swap_fltois swap_fltois.clone() 	{ return new swap_fltois(this);}
//-----------------------------------------------
// dynamic cast <FUNCTION>   EXT SWAP: ok
//-----------------------------------------------
swap_fltois swap_fltois(instrument i, error_info option(nullable) error = null<error_info>) 	
option (category: 'Instrument/Interest Rate Swap/Float vs ONIndex')
option(com_name: 'swap_fltois_dyncast')
{ 
	try {
		swap_fltois d = dynamic_cast<swap_fltois>(i); 
		if(null(d))
			CORE_INT.add_error_info(error,ERR_T_INIT,"invalid cast (instrument is not a swap_fltois)","swap_fltois" );

		return d;
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois");
		return null<swap_fltois>;
	}
}
//------------------------------------------------
// create  [PROTECTED]
// create swap_fltois from internal instrument
//------------------------------------------------
swap_fltois swap_fltois.create(	__instrument option(nullable) c,
								out instr_error option(nullable) error,
								error_type type )
{
	QL_FAIL_COND(null(c), "invalid/unknown (null) instrument",type );

	logical valid = c.is_valid(error);
	if( !valid && INSTR_CREATE_NULL_ERR)
		return null<swap_fltois>; 
	
	..instr_type it = c.instr_type(error);
	QL_FAIL_COND(null(it), "invalid instrument type (unknown)",type );
	QL_FAIL_COND(it != ..instr_type.SWAP_FLTOIS, "invalid instrument type",type );//should not happen
	
	return new swap_fltois(c);
}

/*-----------------------------------------------------------------------
  function: create_swap_fltois
  ----------------------------------------------------------------------*/
swap_fltois create_swap_fltois(	__instrument option(nullable) c,
									out instr_error option(nullable) error,
									error_type type = E_INVALID_ARG)
option(com_name: 'INTERNAL_create_swap_fltois')
{
	QL_FAIL_COND(null(c), "invalid/unknown (null) instrument",type );

	logical valid = c.is_valid(error);
	if( !valid && INSTR_CREATE_NULL_ERR)
		return null<swap_fltois>; 
	
	..instr_type it = c.instr_type(error);
	QL_FAIL_COND(null(it), "invalid instrument type (unknown)",type );
	QL_FAIL_COND(it != ..instr_type.SWAP_FLTOIS, "invalid instrument type",type );//should not happen
	
	return new swap_fltois(c);
}
swap_fltois create_swap_fltois(	__instrument option(nullable) 		c,
								error_info option(nullable) 		error,
								error_type 							type = E_INIT)
option(com_name: 'INTERNAL_create_swap_fltois_ei')
{
	instr_error ee 	= instr_error();
	swap_fltois b 	= create_swap_fltois(c,ee, E_INIT);
	if(ee.is_error())
		CORE_INT.add_error_info(error,ee.type(),ee.message(), "create_swap_fltois");
	return b;
}

/*-----------------------------------------------------------------------
  ENTRY_FUNC prot
  ----------------------------------------------------------------------*/
void	swap_fltois.SWAP_ENTRY_FUNC()
{
	if(!null(full_swap_) && full_swap_)
		return;
	error_info ee = new error_info(true,true);
	full_swap_ = check_full_swap(full_swap_, ee) ;	
}

/*-----------------------------------------------------------------------
  swap_fltois: add_funcs
  ----------------------------------------------------------------------*/
void swap_fltois.add_notional(number notional)				{ swap_gen.swap_add_notional(notional);}
void swap_fltois.add_notional_flt_leg(number notional) 		{ swap_gen.swap_add_notional_leg1(notional);}
void swap_fltois.add_notional_ois_leg(number notional)		{ swap_gen.swap_add_notional_leg2(notional);	}
void swap_fltois.add_fx_rate_flt_leg(number fx_rate)		{ swap_gen.swap_add_fx_rate_leg1(fx_rate);}
void swap_fltois.add_fx_rate_ois_leg(number fx_rate)		{ swap_gen.swap_add_fx_rate_leg2(fx_rate);}
void swap_fltois.add_disc_func(disc_func disc_func)			{ swap_gen.swap_add_disc_func(disc_func); }
void swap_fltois.add_disc_func_flt_leg(disc_func disc_func) { swap_gen.swap_add_disc_func_leg1(disc_func); }
void swap_fltois.add_disc_func_ois_leg(disc_func  disc_func) { swap_gen.swap_add_disc_func_leg2(disc_func); }

void swap_fltois.add_surface_flt_leg(tenor_surface option(nullable) tenor_surface)	{ swap_gen.swap_add_surface_leg1(tenor_surface);}
void swap_fltois.add_fixing_flt_leg(	vector(date) fixing_dates,
									vector(number) fixings,
									logical option(nullable) fixing_as_fwd )
{			
	swap_gen.swap_add_fixing_leg1(fixing_dates,fixings,fixing_as_fwd);						
}
void swap_fltois.add_surface_ois_leg(tenor_surface option(nullable) tenor_surface)	{ swap_gen.swap_add_surface_leg2(tenor_surface);}
void swap_fltois.add_fixing_ois_leg(	vector(date) fixing_dates,
									vector(number) fixings,
									logical option(nullable) fixing_as_fwd )
{			
	swap_gen.swap_add_fixing_leg2(fixing_dates,fixings,fixing_as_fwd);						
}
void swap_fltois.add_fixing_flt_leg(	number fixing_proxy,
									logical option(nullable) fixing_as_fwd ) 
{			
	swap_gen.swap_add_fixing_leg1(fixing_proxy, fixing_as_fwd);						
}
void swap_fltois.add_fixing_ois_leg(	number fixing_proxy,
									logical option(nullable) fixing_as_fwd ) 
{			
	swap_gen.swap_add_fixing_leg2(fixing_proxy, fixing_as_fwd);						
}
/*-----------------------------------------------------------------------
  swap_fltois: disc_func_flt_leg / disc_func_ois_leg
  ----------------------------------------------------------------------*/
disc_func swap_fltois.disc_func_flt_leg(error_info option(nullable) error) 	{ return swap_gen.swap_disc_func_leg1(error);}
disc_func swap_fltois.disc_func_ois_leg(error_info option(nullable) error) 	{ return swap_gen.swap_disc_func_leg2(error);}

/*-----------------------------------------------------------------------
  swap_type
  ----------------------------------------------------------------------*/
instrument_type swap_fltois.swap_type(error_info option(nullable) error) 
{	
	return swap_gen.swap_type(error);
}

/*-----------------------------------------------------------------------
  solver
  ----------------------------------------------------------------------*/
number swap_fltois.solver(	number 				pv,
							swap_solver_code 	solver_code,
							logical option(nullable) trade_date_pv,
							error_info option(nullable) error)
{	
	return swap_gen.solver(pv,solver_code,trade_date_pv,error);
}

/*-----------------------------------------------------------------------
  swap_instrument
  ----------------------------------------------------------------------*/
ql_fixed_income_swap swap_fltois.swap_instrument(error_info option(nullable) error)
{
	return swap_gen.swap_instrument(error);

}
/*-----------------------------------------------------------------------
  index_tenor_code_leg1
  ----------------------------------------------------------------------*/
tenor_code swap_fltois.index_tenor_code_flt_leg(error_info option(nullable) error) 
{	
	return swap_gen.swap_index_tenor_code_leg1(error);
}

/*-----------------------------------------------------------------------
  coupon_freq   	EXT SWAP: ok
  ----------------------------------------------------------------------*/
void  swap_fltois.coupon_freq(	out integer  cpn_freq_flt_leg, 
								out integer  cpn_freq_ois_leg, 
								error_info option(nullable) error ) 
{	
	try{	
		instr_error_type t;
   		string 			s;	
		i().__swap_cpn_freq(cpn_freq_flt_leg,cpn_freq_ois_leg,t, s);
		logical e = CORE_INT.add_error_info(error,t,s, "swap_fltois.coupon_freq");
		if(e)
			cpn_freq_flt_leg = cpn_freq_ois_leg = err_freq();
		return ;						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois.coupon_freq");
		cpn_freq_flt_leg = cpn_freq_ois_leg = err_freq();
		return ;
	}
}
/*-----------------------------------------------------------------------
  dc_method   	EXT SWAP: ok
  ----------------------------------------------------------------------*/
void  swap_fltois.dc_method(	out day_count_method option(nullable) dc_method_flt_leg, 
								out day_count_method option(nullable) dc_method_ois_leg, 
								error_info option(nullable) error ) 
{	
	try{	
		instr_error_type t;
   		string 			s;	
		i().__swap_dc_method(dc_method_flt_leg,dc_method_ois_leg,t, s);
		logical e = CORE_INT.add_error_info(error,t,s, "swap_fltois.dc_method");
		if(e)
			dc_method_flt_leg = dc_method_ois_leg = null<day_count_method>;
		return ;					
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois.dc_method");
		dc_method_flt_leg = dc_method_ois_leg = null<day_count_method>;
		return ;
	}
}
/*-----------------------------------------------------------------------
  pmt_bus_day 	EXT SWAP: ok
  ----------------------------------------------------------------------*/
void  swap_fltois.pmt_bus_day(	out bd_convention  pmt_bus_day_flt_leg, 
								out bd_convention  pmt_bus_day_ois_leg, 
								error_info option(nullable) error) 
{	
	try{	
		instr_error_type t;
   		string 			s;	
		i().__swap_bd(pmt_bus_day_flt_leg,pmt_bus_day_ois_leg,t, s);
		logical e = CORE_INT.add_error_info(error,t,s, "swap_fltois.pmt_bus_day");
		if(e)
			pmt_bus_day_flt_leg = pmt_bus_day_ois_leg = null<bd_convention>;
		return ;					
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois.pmt_bus_day");		
		pmt_bus_day_flt_leg = pmt_bus_day_ois_leg = null<bd_convention>;
		return ;
	}
}
/*-----------------------------------------------------------------------
  eom	EXT SWAP: ok
  ----------------------------------------------------------------------*/
logical swap_fltois.eom(error_info option(nullable) error) option(hidden)
{	
	try{	
		instr_error_type t;
   		string 			s;	
		logical c1,c2;
		i().__swap_eom(c1,c2,t, s);
		logical e = CORE_INT.add_error_info(error,t,s, "swap_fltois.eom");	
		return e ? null<logical>: c1;							
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois.eom");		
		return null<logical>;
	}
}
/*-----------------------------------------------------------------------
  eom   	EXT SWAP: ok
  ----------------------------------------------------------------------*/
void  swap_fltois.eom(	out logical   eom_flt_leg, 
						out logical  eom_ois_leg, 
						error_info option(nullable) error) 
{	
	try{	
		instr_error_type t;
   		string 			s;	
		i().__swap_eom(eom_flt_leg,eom_ois_leg,t, s);
		logical e = CORE_INT.add_error_info(error,t,s, "swap_fltois.eom");
		if(e)
			eom_flt_leg = eom_ois_leg = null<logical>;
		return ;						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois.eom");		
		eom_flt_leg = eom_ois_leg = null<logical>;
		return ;
	}
}

/*-----------------------------------------------------------------------
  cal   	EXT SWAP: ok
  ----------------------------------------------------------------------*/
void  swap_fltois.calendar(	out ..calendar   cal_flt_leg, 
							out ..calendar  cal_ois_leg, 
							error_info option(nullable) error) 
{	
	try{	
		instr_error_type t;
   		string 			s;	
		i().__swap_calendar(cal_flt_leg,cal_ois_leg,t, s);
		logical e = CORE_INT.add_error_info(error,t,s, "swap_fltois.calendar");
		if(e)
			cal_flt_leg = cal_ois_leg = null<..calendar>;
		return ;						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois.calendar");		
		cal_flt_leg = cal_ois_leg = null<..calendar>;
		return ;
	}
}

/*-----------------------------------------------------------------------
  currency	EXT SWAP: ok
  ----------------------------------------------------------------------*/
void  swap_fltois.currency(	out string  curr_flt_leg, 
							out string  curr_ois_leg, 
							error_info option(nullable) error) 
{	
	try{	
		instr_error_type t;
   		string 			s;	
		i().__swap_currency(curr_flt_leg,curr_ois_leg,t, s);
		logical e = CORE_INT.add_error_info(error,t,s, "swap_fltois.currency");
		if(e)
			curr_flt_leg = curr_ois_leg = null<string>;
		return ;					
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois.currency");		
		curr_flt_leg = curr_ois_leg = null<string>;
		return ;
	}
}
/*-----------------------------------------------------------------------
  currency(leg1- backward compat)	EXT SWAP: ok
  ----------------------------------------------------------------------*/
string  swap_fltois.currency(error_info option(nullable) error) option(hidden)
{	
	try{
		QL_FAIL_COND(ccy_swap_,"inapplicable function call for currency swaps (use 2-leg version)", this,true);
		
		instr_error_type t;
   		string 			s;	
		string c1,c2;
		i().__swap_currency(c1,c2,t, s);
		logical e = CORE_INT.add_error_info(error,t,s, "swap_fltois.currency");	
		return e ? null<string>: c1;						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois.currency");
		
		return null<string>;
	}
}
/*-----------------------------------------------------------------------
  reset_freq   	EXT SWAP: ok
  ----------------------------------------------------------------------*/
void  swap_fltois.reset_freq(	out integer  reset_freq_flt_leg, 
								out integer  reset_freq_ois_leg, 
								error_info option(nullable) error) 
{	
	try{	
		error_info ee = error_info(true,true);
		ql_float_leg flt = this.float_leg(false,ee);
		QL_FAIL_COND(null(flt),"invalid swap_fltois", this, true);
			
		reset_freq_flt_leg = reset_freq_to_int(flt.float_def().reset_freq());
		reset_freq_ois_leg = 365;
		return ;						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois.reset_freq");		
		reset_freq_flt_leg = reset_freq_ois_leg = err_freq();
		return ;
	}
}

/*-----------------------------------------------------------------------
  swap_leg
  ----------------------------------------------------------------------*/
vector(ql_swap_leg) swap_fltois.legs(logical clone,error_info option(nullable) error)
{	
	return swap_gen.swap_leg(clone,error);
}

/*-----------------------------------------------------------------------
  float_leg
  ----------------------------------------------------------------------*/
ql_float_leg swap_fltois.float_leg(logical clone,error_info option(nullable) error)
{	
	vector(ql_float_leg) l = swap_gen.swap_float_leg(clone,error);
	return !null(l) ? l[0] : null<ql_float_leg>;
}

/*-----------------------------------------------------------------------
  ois_leg
  ----------------------------------------------------------------------*/
ql_ois_leg swap_fltois.ois_leg(logical clone,error_info option(nullable) error)
{	
	vector(ql_ois_leg) l = swap_gen.swap_ois_leg(clone,error);
	return !null(l) ? l[0] : null<ql_ois_leg>;
}
/*-----------------------------------------------------------------------
  swap_fltois: ir_index
  ----------------------------------------------------------------------*/
ir_index swap_fltois.ir_index_flt_leg(error_info option(nullable) error)
{	
	return swap_gen.swap_ir_index_leg1(error);
}

/*-----------------------------------------------------------------------
  index_tenor_flat_leg
  ----------------------------------------------------------------------*/
string swap_fltois.index_tenor_flt_leg(error_info option(nullable) error) 
{	
	..ir_index ir = this.ir_index_flt_leg(error) ;
	if(error.is_error()) return null<string>;
	return ir.tenor();
}

/*-----------------------------------------------------------------------
  swap_fltois: ir_index
  ----------------------------------------------------------------------*/
ir_index swap_fltois.ir_index_ois_leg(error_info option(nullable) error)
{	
	return swap_gen.swap_ir_index_leg2(error);
}

/*-----------------------------------------------------------------------
  present_value
  ----------------------------------------------------------------------*/
number swap_fltois.present_value(	disc_func option(nullable)disc_func_flt_leg,
									disc_func option(nullable) 	disc_func_ois_leg,
									out number pv_flt_leg,
									out number pv_ois_leg,
									number option(nullable) notional,
									logical option(nullable) 	trade_date_pv,																								
									error_info option(nullable) error)
{	
	return swap_gen.present_value(disc_func_flt_leg,disc_func_ois_leg,pv_flt_leg,pv_ois_leg,notional,trade_date_pv,error);
}

number swap_fltois.present_value(	disc_func option(nullable)	disc_func_flt_leg,
									disc_func option(nullable) 	disc_func_ois_leg,
									out number pv_flt_leg,
									out number pv_ois_leg,
									date option(nullable) pv_date,
									date option(nullable) settle_date,
									logical  	incl_issue_cf,
									logical  	incl_mat_cf,	
									error_info option(nullable) error)
{	
	return swap_gen.present_value(disc_func_flt_leg,disc_func_ois_leg,pv_flt_leg,pv_ois_leg,pv_date, settle_date ,incl_issue_cf,incl_mat_cf,error );
}

number swap_fltois.present_value(	out number pv_flt_leg,
									out number pv_ois_leg,
									logical option(nullable) 	trade_date_pv,
									error_info option(nullable) error)
{	
	return swap_gen.present_value(pv_flt_leg,pv_ois_leg,trade_date_pv ,error );
}


/*-----------------------------------------------------------------------
  cash_flow_dates	EXT SWAP: ok   <PUBLIC>   interest payment dates, not principal pmt dates
  ----------------------------------------------------------------------*/
void  swap_fltois.cash_flow_dates(	logical 			post_settle, 
									out vector(date) 	pmt_dates_leg1,									
									out vector(date) 	pmt_dates_leg2,
									error_info option(nullable) error) 
{	
	try{	
		instr_error_type t;
   		string 			s;	
		i().__swap_ir_pmt_dates(post_settle,pmt_dates_leg1,pmt_dates_leg2, t, s);
		logical e = CORE_INT.add_error_info(error,t,s, "swap_fltois.cash_flow_dates");
		if(e) {
			pmt_dates_leg1 = pmt_dates_leg2 = null<vector(date)>;  
		}
		return;						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois.cash_flow_dates");		
		pmt_dates_leg1 = pmt_dates_leg2 = null<vector(date)>; 
		return ;
	}
}
/*-----------------------------------------------------------------------
  prin_cash_flow_dates	EXT SWAP: ok   <PUBLIC>   
  ----------------------------------------------------------------------*/
void  swap_fltois.prin_cash_flow_dates(	logical 			post_settle, 
										out vector(date) 	pmt_dates_leg1,									
										out vector(date) 	pmt_dates_leg2,
										error_info option(nullable) error) 
{	
	try{	
		instr_error_type t;
   		string 			s;	
		i().__swap_prin_pmt_dates(post_settle,pmt_dates_leg1,pmt_dates_leg2, t, s);
		logical e = CORE_INT.add_error_info(error,t,s, "swap_fltois.prin_cash_flow_dates");
		if(e) {
			pmt_dates_leg1 = pmt_dates_leg2 = null<vector(date)>;  
		}
		return;						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois.prin_cash_flow_dates");		
		pmt_dates_leg1 = pmt_dates_leg2 = null<vector(date)>; 
		return ;
	}
}

/*-----------------------------------------------------------------------
  accr_dates	EXT SWAP: ok
  ----------------------------------------------------------------------*/
void  swap_fltois.accr_dates(	logical 			post_settle,									
								out vector(date) 	accr_start_dates_leg1,									
								out vector(date) 	accr_end_dates_leg1,
								out vector(number) 	accr_period_leg1,
								out vector(date) 	accr_start_dates_leg2,									
								out vector(date) 	accr_end_dates_leg2,
								out vector(number) 	accr_period_leg2,
								error_info option(nullable) error) 
{
	try{	
		instr_error_type t;
   		string 			s;
	
		i().__swap_accr_dates(post_settle,accr_start_dates_leg1,accr_end_dates_leg1,accr_period_leg1,
									  accr_start_dates_leg2,accr_end_dates_leg2,accr_period_leg2,t, s);
		logical e = CORE_INT.add_error_info(error,t,s, "swap_fltois.accr_dates");
		if(e) {
			accr_start_dates_leg1 = accr_start_dates_leg2 = null<vector(date)>;
			accr_end_dates_leg1 = accr_end_dates_leg2 = null<vector(date)>;
			accr_period_leg1 = accr_period_leg2 = null<vector(number)>;
		
		}			
		return ;					
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_fltois.accr_dates");		
		accr_start_dates_leg1 = accr_start_dates_leg2 = null<vector(date)>;
		accr_end_dates_leg1 = accr_end_dates_leg2 = null<vector(date)>;
		accr_period_leg1 = accr_period_leg2 = null<vector(number)>; 		 
		return ;
	}
}