option(null: hard);	

/*	
	instrument wrapper functions
	Developer: Algorithmica Research, Magnus Nyström

	--------------------------------------------------------------------
	swap_il_fixyy create funcs
	--------------------------------------------------------------------			
*/

//--------
// from db
//--------
/*-----------------------------------------------------------------------
  swap_il_fixyy
  ----------------------------------------------------------------------*/
swap_il_fixyy swap_il_fixyy(instrument_name 			instrument_name, 
							date option(nullable) 		trade_date 	= null<date>,
							string option(nullable) 	quote_side 	= null<string>,
							error_info option(nullable) error		= null<error_info>)
option (category: "Instrument/Inflation Linked/Fixed vs Y-Y")
{	
	try{			
		__instrument c = __instrument(instrument_name,trade_date,quote_side, null<date> );		
		return create_swap_il_fixyy(c,error, E_INIT);						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_il_fixyy");
		return null<swap_il_fixyy>;
	}
}


/*-----------------------------------------------------------------------
  swap_il_fixyy
  ----------------------------------------------------------------------*/
swap_il_fixyy swap_il_fixyy(instrument_name 			instrument_name, 
							date option(nullable) 		trade_date ,
							number  option(nullable) 	quote ,
							error_info option(nullable) error		= null<error_info>)
option (category: 'Instrument/Inflation Linked/Fixed vs Y-Y')
{	
	try{				
		__instrument c = __instrument(instrument_name,trade_date,quote, null<date> );
		return create_swap_il_fixyy(c,error, E_INIT);							
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_il_fixyy");
		return null<swap_il_fixyy>;
	}
}


/*-----------------------------------------------------------------------
  swap_il_fixyy
  ----------------------------------------------------------------------*/
swap_il_fixyy swap_il_fixyy(instr_def 					instr_def, 
							date option(nullable) 		trade_date 	= null<date>,
							string option(nullable) 	quote_side 	= null<string>,
							error_info option(nullable) error		= null<error_info>)
option (category: 'Instrument/Inflation Linked/Fixed vs Y-Y')
{	
	try{				
		__instrument c = __instrument(instr_def,trade_date,quote_side, null<date> );
		return create_swap_il_fixyy(c,error, E_INIT);						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_il_fixyy");
		return null<swap_il_fixyy>;
	}
}
/*-----------------------------------------------------------------------
  instrument
  ----------------------------------------------------------------------*/
swap_il_fixyy swap_il_fixyy(instr_def 					instr_def, 
							date option(nullable) 		trade_date ,
							number  option(nullable) 	quote ,
							error_info option(nullable) error		= null<error_info>)
option (category: 'Instrument/Inflation Linked/Fixed vs Y-Y')
{	
	try{			
		__instrument c = __instrument(instr_def,trade_date,quote, null<date> );
		return create_swap_il_fixyy(c,error, E_INIT);						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_il_fixyy");
		return null<swap_il_fixyy>;
	}
}

//----------------
// from db or not
//----------------


/*-----------------------------------------------------------------------
  func: swap_il_fixyy 
  arguments: 
	instr_def: 	can be created either from the db or user defined 				 
				instr_def is copied and modified
				no lookup of quotes
	instrument_name: user defined name

  ----------------------------------------------------------------------*/
swap_il_fixyy swap_il_fixyy(instr_def 				instr_def, 
								string 					instrument_name  ,			
								date option(nullable) 	trade_date ,
								date option(nullable) 	settle_date ,							
								date  option(nullable) 	maturity,//maturity has precedence over mat_code
								string option(nullable) maturity_code,																
								number option(nullable) rate ,							
								..cpi_link 				cpi_link,
								//number 					cpi_base,
								..cpi_link	option(nullable)	cpi_link_fcast,
								number 	option(nullable) nominal=100,
								error_info option(nullable) error		= null<error_info>)
option (category: 'Instrument/Inflation Linked/Fixed vs Y-Y')
{	
	try{
		__instrument c;
		if(!null(maturity))
			c = __instrument_fixyy_par_nodb(instr_def,instrument_name,trade_date, settle_date,
													 maturity,rate,cpi_link, cpi_link_fcast,nominal );
		else if(!null(maturity_code))
			c = __instrument_fixyy_par_nodb(instr_def,instrument_name,trade_date, settle_date,
													 maturity_code,rate,cpi_link, cpi_link_fcast,nominal );
		else
			QL_FAIL("invalid maturity/maturity_code");

		return create_swap_il_fixyy( c, error);						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_il_fixyy");
		return null<swap_il_fixyy>;
	}
}

//----------------
// no db, swaplib version
// an external swap is created. Note: an external swap has no internal instr_def
// but can be created from an instr_def
//----------------

/*-----------------------------------------------------------------------
  func: swap_il_fixyy
  ----------------------------------------------------------------------*/
swap_il_fixyy swap_il_fixyy(ql_fixed_income_swap ql_fixed_income_swap,
							error_info option(nullable) error = null<error_info>)
option (category: 'Instrument/Interest Rate Swap/Fixed vs Y-Y')
{	
	try{	
		QL_FAIL_COND(CORE_INT.swap_instr_type(ql_fixed_income_swap) != instr_type.IL_SWAP_FIX_Y_Y, "invalid instrument type", E_INIT);

		__instrument c = __instrument_swap(ql_fixed_income_swap);

		return create_swap_il_fixyy(c, error);				
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_il_fixyy");
		return null<swap_il_fixyy>;
	}
}


/*-----------------------------------------------------------------------
  func: swap_il_fixyy
  ----------------------------------------------------------------------*/
swap_il_fixyy swap_il_fixyy(ql_fix_leg 		ql_fix_leg, 							
							ql_yy_infl_leg 	ql_yy_leg,
							string 			id,
							error_info option(nullable) error = null<error_info>)
option (category: 'Curve and Instrument/Interest Rate Swap/Fixed vs Y-Y')
{	
	try{			
		__instrument c = __instrument_swap(ql_fix_leg,ql_yy_leg,id);

		return create_swap_il_fixyy(c, error);						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_il_fixyy");
		return null<swap_il_fixyy>;
	}
}

module CORE_INT_IL_FIXYY
{
	/*-----------------------------------------------------------------------
	  func: swap_il_fixyy_plain_int
	  minimal constructor for single curr "screen"-swap,
	  ----------------------------------------------------------------------*/
	swap_il_fixyy swap_il_fixyy_plain_int(instr_def 					def,
										string 						instrument_name,	
										date option(nullable) 		trade_date,	
										date option(nullable) 		start_date,
										//string option(nullable) 	fwd_start_code,
										//date 	option(nullable)	maturity,
										integer						maturity_years,
										number option(nullable)		fix_infl_rate ,//if null --> will be set to par 									
										number 	 					notional,																										
										logical 					pay_infl,
										cpi_link 					cpi_l ,
										cpi_link option(nullable)	cpi_l_f,
										..disc_func option(nullable) fix_disc_func,
										..disc_func option(nullable) infl_disc_func,																												
										error_info option(nullable) error )
	{	
		try{
			CORE_INT.reset_single(error);
			QL_FAIL_COND(!def.is_il_swap_fix_yy(), "instr_def not a fix-yy swap", E_INIT);
			logical is_ccy				= def.is_ccy_swap();
			QL_FAIL_COND(is_ccy, "constructor not applicable for currency swap", E_INIT);

			date 	maturity = null<date>;
			string fwd_start_code = "";
				
			number fix_fx_rate = 1;
			number infl_fx_rate = 1;
			number fix_notional = notional;
			number infl_notional = notional;
			//---def---
			error_info ee = error_info(true,true);
			notional_exchg_style ne = def.swap_not_exchange(ee);

			if(null(ne))
				ne 	= is_ccy ? NE_NONE : NE_END;

			date_code settle_code 		= date_code(def.settle_code());		
			coupon_freq  cpn_freq 		= coupon_freq_from_int(1)  ;
			day_count_method dc 			= def.swap_dc_method_leg1(ee);
			calendar pmt_calendar		= def.swap_calendar_leg1(ee);
			bd_convention pmt_bus_day 	= def.swap_pmt_bus_day_leg1(ee);
			logical	end_of_month		= def.swap_eom_leg1(ee);
			number pmt_delay			= def.swap_pmt_lag_leg1(ee);
			number mat_pmt_delay 		= pmt_delay;
			infl_index_method index_method = def.il_index_method(ee);
			//quote_style qs 				= null<quote_style>;	
			round_code pv_round_code	= null<round_code>;
			
			
			ql_swap_fix_def fix_def 	= ql_swap_fix_def("fix_def", def.swap_currency_leg1(ee), settle_code, cpn_freq,
														dc,pmt_bus_day, pmt_calendar,end_of_month,ne);
			
			ql_swap_yy_infl_def infl_def = ql_swap_yy_infl_def("infl_def", def.swap_currency_leg2(ee), settle_code, cpn_freq,
																  dc, pmt_bus_day,pmt_calendar, end_of_month,
																  pv_round_code,pmt_delay,mat_pmt_delay, ne,index_method);


			date  fix_eff_date 		= null<date>;
			date  first_cpn_date 	= null<date>;
			date  last_reg_date 	= null<date>;
			integer	cpn_roll_day 	= 0;
			
			date settle_date;
			CORE_INT_FIXFLT.swap_start_date_helper(settle_code,trade_date,pmt_calendar,fwd_start_code,start_date,settle_date);
			QL_FAIL_COND(null(settle_date),"invalid start date", E_INIT);

			tenor_code mat_code = CORE_INT_IL_ZERO_SWAP.swap_il_zero_maturity_helper(def, maturity_years);
			
			if(null(fix_fx_rate)) 
				fix_fx_rate = 1;
			if(null(infl_notional) && !is_ccy) 
				infl_notional = fix_notional;
			if(null(infl_fx_rate)) 
				infl_fx_rate = 1;

			number init_rate 		= null(fix_infl_rate) ? 0 : fix_infl_rate;
			
			ql_fix_leg fix 			= ql_fix_leg(fix_def,"fix_leg",trade_date,settle_date,start_date,maturity,mat_code,
												!pay_infl,fix_notional,fix_eff_date,first_cpn_date,last_reg_date,init_rate,	
												fix_disc_func,cpn_roll_day,fix_fx_rate);
			
			ql_yy_infl_leg infl 	= ql_yy_infl_leg(infl_def, "infl_leg", trade_date,settle_date,start_date,maturity,mat_code,pay_infl,
													   infl_notional,first_cpn_date,last_reg_date,infl_disc_func,cpi_l,cpi_l_f,
														cpn_roll_day,infl_fx_rate);
			
			ql_fixed_income_swap sw = ql_fixed_income_swap(fix,infl,instrument_name);

			if(null(fix_infl_rate)) {
				number f_r	= sw.solver(0,swap_solver_code.SS_FIX_RATE_LEG1,false);
				QL_FAIL_COND(null(f_r),"error solving for fix rate", E_INIT);

				fix 	= ql_fix_leg(	fix_def,"fix_leg",trade_date,settle_date,start_date,maturity,mat_code,
											!pay_infl,fix_notional,fix_eff_date,first_cpn_date,last_reg_date,f_r,	
											fix_disc_func,cpn_roll_day,fix_fx_rate);
				
				sw 		= ql_fixed_income_swap(fix,infl,instrument_name);

			}
			
			return swap_il_fixyy(sw, error);						
		}
		catch {
			CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_il_fixyy");
			return null<swap_il_fixyy>;
		}
	}
}

/*-----------------------------------------------------------------------
  func: swap_il_fixyy_plain
  minimal constructor for single curr "screen"-swap,
  this version: disc_func  input
  ----------------------------------------------------------------------*/
swap_il_fixyy swap_il_fixyy_plain(instr_def 					def,
								string 						instrument_name,	
								date option(nullable) 		trade_date,	
								date option(nullable) 		start_date,			//start_date has precedence over start_date_code
								integer						maturity_years,																													
								number option(nullable)		fix_rate ,//if null --> will be set to par 									
								number 	 					notional,																										
								logical 					pay_inflation,
								cpi_link 					cpi_l ,
								cpi_link option(nullable)	cpi_l_f,
								..disc_func option(nullable) fix_disc_func,
								..disc_func option(nullable) flt_disc_func,													    
								error_info option(nullable) error = null<error_info>)
option (category: 'Instrument/Inflation Linked/Fixed vs Y-Y')
{
	return CORE_INT_IL_FIXYY.swap_il_fixyy_plain_int(	def,instrument_name,trade_date,start_date,maturity_years,
											fix_rate ,notional, pay_inflation,cpi_l,cpi_l_f,
											fix_disc_func,flt_disc_func,error );
}

/*-----------------------------------------------------------------------
  func: swap_il_fixyy_plain
  minimal constructor for single curr "screen"-swap,
  ----------------------------------------------------------------------*/
swap_il_fixyy swap_il_fixyy_plain(INSTR_TMPL.swap_il_fixyy_def_tmpl	tmpl,
								string 						instrument_name,	
								date option(nullable) 		trade_date,	
								date option(nullable) 		start_date,//if tmpl have a code it is ignored
								integer 					maturity_years,//if tmpl have a code it is ignored																												
								number option(nullable)		fix_rate ,//if null --> will be set to par 									
								number 	 					notional,																										
								logical 					pay_inflation,
								cpi_link 					cpi_l ,
								cpi_link option(nullable)	cpi_l_f,
								..disc_func option(nullable) fix_disc_func,
								..disc_func option(nullable) flt_disc_func,															    
								error_info option(nullable) error = null<error_info>)
option (category: 'Instrument/Inflation Linked/Fixed vs Y-Y')
{
	try{		
		instr_def def = instr_def_swap_fixyy(tmpl,error);		
		if(!CORE_INT.check_instr_def(def))
			return null<swap_il_fixyy>;
		return null<swap_il_fixyy>;
		return swap_il_fixyy_plain(	def,instrument_name,trade_date,start_date, maturity_years,fix_rate ,
									notional,pay_inflation,cpi_l,cpi_l_f,fix_disc_func,flt_disc_func,error);
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "swap_il_fixyy");
		return null<swap_il_fixyy>;
	}
}