option(null: hard);

module __w_calc
{
    class y2pdiff_func
    {
    public:
	number eval(number y, out number der_ret);

    public:
	date		d0;
	number		sum_0;
	vector(number)	amounts;
	vector(date)	dates;
    };

    number y2pdiff_func.eval(number y, out number deriv_ret)
    {
	integer		n_cf	= v_size(dates);
	number		diff	= sum_0;
	number		deriv	= 0;

	for (integer i = 0 ; i < n_cf ; i++) {
	    number	t_i		= (dates[i] - d0) / 365.0;
	    number	e_yti		= exp(-y * t_i);
	    number	amount_i_df	= amounts[i] * e_yti;

	    diff += amount_i_df;
	    deriv += -t_i * amount_i_df;
	}

	deriv_ret = deriv;

	return diff;
    }

    number pvbp(date		d0,
		number		sum_0,
		vector(number)	amounts,
		vector(date)	dates)
    {
	y2pdiff_func	f = new y2pdiff_func;

	f.d0		= d0;
	f.sum_0		= sum_0;
	f.amounts	= amounts;
	f.dates		= dates;

	try {
	    number	y;
	    number	pvbp;

	    y = zero_newton(f, &y2pdiff_func.eval, 0.05, 1e-8);
	    f.eval(y, pvbp);

	    return -pvbp;
	}
	catch (E_CALC) {
	    throw(E_CALC, "Yield calculation failed");
	}
    }

    number calc_w_disc(std_weights	w_scheme,
		       date		d0,
		       number		sum_0,
		       vector(number)	amounts,
		       vector(date)	dates)
    {
	switch (w_scheme) {
	case WS_EQUAL:
	    return 1;
	case WS_PVBP:
	    return 1 / pvbp(d0, sum_0, amounts, dates);
	case WS_PVBP2:
	    return 1 / sqr(pvbp(d0, sum_0, amounts, dates));
	}
    }

    number calc_w_fwd(std_weights	w_scheme,
		      vector(number)	amounts)
    {
	switch (w_scheme) {
	case WS_EQUAL:
	    return 1;
	case WS_PVBP:
	    return 1 / v_sum(amounts);
	case WS_PVBP2:
	    return 1 / sqr(v_sum(amounts));
	}
    }
}

/**/

class cashflows_base : public curve_base
{
public:
    virtual date	trade_date()	= 0;
    virtual integer	n_instruments()	= 0;
    virtual vector(number)	get_all_periods() = 0;	
};

class cashflows : public cashflows_base
{
public:
    virtual logical
    get_cashflows(integer			ix,
		  out vector(date)		dates,
		  out vector(number)	amounts) = 0;

    virtual vector(number)	get_all_periods();
	vector(number) 			get_all_periods_ex0();

private: // curve_base
    virtual date	__d0();
    virtual integer	__n_eqns();
    virtual logical	__init_eqns(__zc.cf_eqns,
				    std_weights option(nullable),
				    vector(number) option(nullable));

	vector(number) 	get_all_per(logical);
};

class cashflows_fwd : public curve_base_fwd
{
public:
    virtual date	trade_date()	= 0;
    virtual integer	n_instruments()	= 0;

    virtual logical
    get_cashflows(integer		ix,
		  out vector(date)	fixed_dates,
		  out vector(number)	fixed_amounts,
		  out vector(date)	float_dates,
		  out vector(date)	float_disc_d,
		  out vector(number)	float_nominals) = 0;

    vector(number)	get_all_periods();

private: // curve_base_fwd
    virtual date	__d0();
    virtual integer	__n_eqns();
    virtual logical	__init_eqns(__zc.cf_eqns, disc_func,
				    std_weights option(nullable),
				    vector(number) option(nullable));
};

date cashflows.__d0()
{
    return trade_date();
}

integer cashflows.__n_eqns()
{
    return n_instruments();
}

logical
cashflows.__init_eqns(__zc.cf_eqns			eqns,
		      std_weights option(nullable)	w_scheme,
		      vector(number) option(nullable)	w)
{
    date		trade	= trade_date();
    integer		n	= n_instruments();

    if (n == 0)
	return false;

    integer		n_ok	= 0;

    for (integer i = 0 ; i < n ; i++) {
	vector(date)	dates;
	vector(number)	amounts;

	if (!get_cashflows(i, dates, amounts))
	    continue;

	++n_ok;

	date	settle	= dates[0]; // We assume no flows before settle

	eqns.set(i, 0, settle, amounts, dates);

	if (!null(w_scheme))
	    w[i] = __w_calc.calc_w_disc(w_scheme, trade, 0, amounts, dates);
    }

    if (n_ok == 0)
	return false;

    if (n_ok != n)
	throw(E_NO_DATA,
	      strcat([ "Cashflows inconsistency: could only create ",
		       string(n_ok), " of ", string(n), " instruments" ]));

    return true;
}

date cashflows_fwd.__d0()
{
    return trade_date();
}

integer cashflows_fwd.__n_eqns()
{
    return n_instruments();
}

logical
cashflows_fwd.__init_eqns(__zc.cf_eqns				eqns,
			  disc_func				df,
			  std_weights option(nullable)		w_scheme,
			  vector(number) option(nullable)	w)
{
    date		trade	= trade_date();
    integer		n	= n_instruments();

    if (n == 0)
	return false;

    integer		n_ok	= 0;

    for (integer i = 0 ; i < n ; i++) {
	vector(date)	fixed_dates;
	vector(number)	fixed_amounts;
	vector(date)	fwd_dates;
	vector(date)	fwd_disc_d;
	vector(number)	fwd_nominals;

	if (!get_cashflows(i, fixed_dates, fixed_amounts,
			   fwd_dates, fwd_disc_d, fwd_nominals))
	    continue;

	++n_ok;

	number		sum_0;

	sum_0 = df.disc((fixed_dates - trade) / 365.0) * fixed_amounts;
	if (null(sum_0))
	    throw(E_INVALID_ARG, "Null discount factor or fixed amount");

	fwd_nominals = fwd_nominals .* df.disc((fwd_disc_d - trade) / 365);

	date		settle	= trade; // Not used in the fwd-case

	eqns.set(i, sum_0, settle, fwd_nominals, fwd_dates);

	if (!null(w_scheme))
	    w[i] = __w_calc.calc_w_fwd(w_scheme, fwd_nominals);
    }

    if (n_ok == 0)
	return false;

    if (n_ok != n)
	throw(E_NO_DATA,
	      strcat([ "Cashflows inconsistency: could only create ",
		       string(n_ok), " of ", string(n), " instruments" ]));

    return true;
}

vector(number) cashflows.get_all_periods()
{
	return this.cashflows.get_all_per(true);
}

vector(number) cashflows.get_all_periods_ex0()
{
	return this.cashflows.get_all_per(false);
}

vector(number) cashflows.get_all_per(logical incl_0)
{
    set_num		periods	= new set_num();
    date		d0	= trade_date();
    integer		n_i	= n_instruments();
    integer		n_ok	= 0;

    for (integer i = 0 ; i < n_i ; i++) {
		vector(date)	dates;
		vector(number)	amounts;
	
		if (!get_cashflows(i, dates, amounts))
		    continue;
	
		++n_ok;
	
		integer		n_cf	= v_size(dates);
	
		for (integer i = 0 ; i < n_cf ; i++) {
		    if (null(dates[i]))
				throw(E_INVALID_ARG, "Null cashflow date");
			if (!incl_0 && dates[i] == d0)
				continue;
		    periods.add((dates[i] - d0) / 365);
		}
    }

    if (n_ok == 0)
		return null<vector(number)>;

    if (n_ok != n_i)
		throw(E_NO_DATA, "Cashflows inconsistency");

	if(incl_0)
    	periods.add(0);

    return periods.get_content();
}

vector(number) cashflows_fwd.get_all_periods()
{
    set_num		periods	= new set_num();
    date		d0	= trade_date();
    integer		n_i	= n_instruments();
    integer		n_ok	= 0;

    for (integer i = 0 ; i < n_i ; i++) {
	vector(date)	fixed_dates;
	vector(number)	fixed_amounts;
	vector(date)	float_dates;
	vector(date)	float_disc_d;
	vector(number)	float_nominals;

	if (!get_cashflows(i, fixed_dates, fixed_amounts,
			   float_dates, float_disc_d, float_nominals))
	    continue;

	++n_ok;

	integer		n_cf	= v_size(float_dates);

	for (integer i = 0 ; i < n_cf ; i++) {
	    if (null(float_dates[i]))
		throw(E_INVALID_ARG, "Null cashflow date");
	    periods.add((float_dates[i] - d0) / 365);
	}
    }

    if (n_ok == 0)
	return null<vector(number)>;

    if (n_ok != n_i)
	throw(E_NO_DATA, "Cashflows inconsistency");

    periods.add(0);

    return periods.get_content();
}

/**/

class cashflows_one : public cashflows
{
public:
    cashflows_one(date			trade_date,
		  vector(date)		dates,
		  vector(number)	amounts);

    virtual date	trade_date();
    virtual integer	n_instruments();

    virtual logical
    get_cashflows(integer		ix,
		  out vector(date)	dates,
		  out vector(number)	amounts);

private:
    date		trade_date_;
    vector(date)	dates_;
    vector(number)	amounts_;
};

cashflows_one.cashflows_one(date		trade_date,
			    vector(date)	dates,
			    vector(number)	amounts)
    : trade_date_(trade_date),
      dates_(clone_vector(dates)),
      amounts_(clone_vector(amounts))
{}

date cashflows_one.trade_date()
{
    return trade_date_;
}

integer cashflows_one.n_instruments()
{
    return 1;
}

logical cashflows_one.get_cashflows(integer		ix,
				    out vector(date)	dates,
				    out vector(number)	amounts)
{
    dates	= dates_;
    amounts	= amounts_;

    return true;
}

cashflows
cashflows(date				trade_date,
	  vector(date)			dates,
	  vector(number)		amounts)
    option(com_name: "cashflows_one")
{
    return new cashflows_one(trade_date, dates, amounts);
}

/**/

class cashflows_sum : public cashflows
{
public:
    cashflows_sum(vector(cashflows) option(fixedref) cf);

    virtual date	trade_date();
    virtual integer	n_instruments();

    virtual logical
    get_cashflows(integer		ix,
		  out vector(date)	dates,
		  out vector(number)	amounts);

private:
    vector(cashflows) option(fixedref)	cf_;
    vector(integer)			ix2cf_;
    vector(integer)			ix2subix_;
};

cashflows_sum.cashflows_sum(vector(cashflows) option(fixedref) cf)
    : cf_(cf)
{
    integer		n_cf	= v_size(cf_);

    if (n_cf == 0)
	throw(E_INVALID_ARG, "Empty vector");

    for (integer i = 0 ; i < n_cf ; i++) {
	if (null(cf_[i]))
	    throw(E_INVALID_ARG, "Null element in vector");

	integer	n_instr	= cf_[i].n_instruments();

	for (integer j = 0 ; j < n_instr ; j++) {
	    push_back(ix2cf_,	i);
	    push_back(ix2subix_,	j);
	}
    }
}

date cashflows_sum.trade_date()
{
    return cf_[0].trade_date();
}

integer cashflows_sum.n_instruments()
{
    return v_size(ix2cf_);
}

logical cashflows_sum.get_cashflows(integer		ix,
				    out vector(date)	dates,
				    out vector(number)	amounts)
{
    return cf_[ix2cf_[ix]].get_cashflows(ix2subix_[ix], dates, amounts);
}

cashflows
cashflows(vector(cashflows)		cashflows)
    option(com_name: "cashflows_sum")
{
    return new cashflows_sum(cashflows);
}

/**/

class cashflows_fwd_one : public cashflows_fwd
{
public:
    cashflows_fwd_one(date		trade_date,
		      vector(date)	fixed_dates,
		      vector(number)	fixed_amounts,
		      vector(date)	float_dates,
		      vector(date)	float_disc_d,
		      vector(number)	float_nominals);

    virtual date	trade_date();
    virtual integer	n_instruments();

    virtual logical
    get_cashflows(integer		ix,
		  out vector(date)	fixed_dates,
		  out vector(number)	fixed_amounts,
		  out vector(date)	float_dates,
		  out vector(date)	float_disc_d,
		  out vector(number)	float_nominals);

private:
    date		trade_date_;
    vector(date)	fixed_dates_;
    vector(number)	fixed_amounts_;
    vector(date)	float_dates_;
    vector(date)	float_disc_d_;
    vector(number)	float_nominals_;
};

cashflows_fwd_one.cashflows_fwd_one(date		trade_date,
				    vector(date)	fixed_dates,
				    vector(number)	fixed_amounts,
				    vector(date)	float_dates,
				    vector(date)	float_disc_d,
				    vector(number)	float_nominals)
    : trade_date_(trade_date),
      fixed_dates_(clone_vector(fixed_dates)),
      fixed_amounts_(clone_vector(fixed_amounts)),
      float_dates_(clone_vector(float_dates)),
      float_disc_d_(clone_vector(float_disc_d)),
      float_nominals_(clone_vector(float_nominals))
{}

date cashflows_fwd_one.trade_date()
{
    return trade_date_;
}

integer cashflows_fwd_one.n_instruments()
{
    return 1;
}

logical cashflows_fwd_one.get_cashflows(integer			ix,
					out vector(date)	fixed_dates,
					out vector(number)	fixed_amounts,
					out vector(date)	float_dates,
					out vector(date)	float_disc_d,
					out vector(number)	float_nominals)
{
    fixed_dates		= fixed_dates_;
    fixed_amounts	= fixed_amounts_;
    float_dates		= float_dates_;
    float_disc_d	= float_disc_d_;
    float_nominals	= float_nominals_;

    return true;
}

cashflows_fwd
cashflows_fwd(date			trade_date,
	      vector(date)		fixed_dates,
	      vector(number)		fixed_amounts,
	      vector(date)		float_dates,
	      vector(date)		float_disc_d,
	      vector(number)		float_nominals)
    option(com_name: "cashflows_fwd_one")
{
    return new cashflows_fwd_one(trade_date,
				 fixed_dates, fixed_amounts,
				 float_dates, float_disc_d,
				 float_nominals);
}

/**/

class cashflows_fwd_sum : public cashflows_fwd
{
public:
    cashflows_fwd_sum(vector(cashflows_fwd) option(fixedref) cf);

    virtual date	trade_date();
    virtual integer	n_instruments();

    virtual logical
    get_cashflows(integer		ix,
		  out vector(date)	fixed_dates,
		  out vector(number)	fixed_amounts,
		  out vector(date)	float_dates,
		  out vector(date)	float_disc_d,
		  out vector(number)	float_nominals);

private:
    vector(cashflows_fwd) option(fixedref)	cf_;
    vector(integer)				ix2cf_;
    vector(integer)				ix2subix_;
};

cashflows_fwd_sum.cashflows_fwd_sum(vector(cashflows_fwd)
				    option(fixedref) cf)
    : cf_(cf)
{
    integer		n_cf	= v_size(cf_);

    if (n_cf == 0)
	throw(E_INVALID_ARG, "Empty vector");

    for (integer i = 0 ; i < n_cf ; i++) {
	if (null(cf_[i]))
	    throw(E_INVALID_ARG, "Null element in vector");

	integer	n_instr	= cf_[i].n_instruments();

	for (integer j = 0 ; j < n_instr ; j++) {
	    push_back(ix2cf_,	i);
	    push_back(ix2subix_,	j);
	}
    }
}

date cashflows_fwd_sum.trade_date()
{
    return cf_[0].trade_date();
}

integer cashflows_fwd_sum.n_instruments()
{
    return v_size(ix2cf_);
}

logical cashflows_fwd_sum.get_cashflows(integer			ix,
					out vector(date)	fixed_dates,
					out vector(number)	fixed_amounts,
					out vector(date)	float_dates,
					out vector(date)	float_disc_d,
					out vector(number)	float_nominals)
{
    return cf_[ix2cf_[ix]].get_cashflows(ix2subix_[ix],
					 fixed_dates, fixed_amounts,
					 float_dates, float_disc_d,
					 float_nominals);
}

cashflows_fwd
cashflows_fwd(vector(cashflows_fwd)	cashflows)
    option(com_name: "cashflows_fwd_sum")
{
    return new cashflows_fwd_sum(cashflows);
}

/*
 * cashflow_q
 */

// Hack for testing
number calc_w_disc_q(std_weights	w_scheme,
		     date		d0,
		     vector(date)	dates)
{
    switch (w_scheme) {
    case WS_EQUAL:
	return 1;
    case WS_PVBP:
	return 1 / (dates[v_size(dates) - 1] - d0 + 1);
    case WS_PVBP2:
	return 1 / sqr(dates[v_size(dates) - 1] - d0 + 1);
    }
}

/**/

class cashflows_q : public cashflows_base
{
public:
    virtual logical
    get_cashflows(integer		ix,
		  out vector(date)	dates,
		  out vector(date)	dates_q,
		  out vector(number)	amounts) = 0;

    virtual vector(number) get_all_periods() ;

private: // curve_base
    virtual date	__d0();
    virtual integer	__n_eqns();
    virtual logical	__init_eqns(__zc.cf_eqns,
				    std_weights option(nullable),
				    vector(number) option(nullable));
};

date cashflows_q.__d0()
{
    return trade_date();
}

integer cashflows_q.__n_eqns()
{
    return n_instruments();
}

logical
cashflows_q.__init_eqns(__zc.cf_eqns			eqns,
			std_weights option(nullable)	w_scheme,
			vector(number) option(nullable)	w)
{
    date		trade	= trade_date();
    integer		n	= n_instruments();

    if (n == 0)
	return false;

    integer		n_ok	= 0;

    for (integer i = 0 ; i < n ; i++) {
	vector(date)	dates;
	vector(date)	dates_q;
	vector(number)	amounts;

	if (!get_cashflows(i, dates, dates_q, amounts))
	    continue;

	++n_ok;

	date	settle	= dates[0]; // We assume no flows before settle

	eqns.set_q(i, 0, settle, amounts, dates, dates_q);

	if (!null(w_scheme))
	    w[i] = calc_w_disc_q(w_scheme, trade, dates);
    }

    if (n_ok == 0)
	return false;

    if (n_ok != n)
	throw(E_NO_DATA,
	      strcat([ "Cashflows inconsistency: could only create ",
		       string(n_ok), " of ", string(n), " instruments" ]));

    return true;
}

vector(number) cashflows_q.get_all_periods()
{
    set_num		periods	= new set_num();
    date		d0	= trade_date();
    integer		n_i	= n_instruments();
    integer		n_ok	= 0;

    for (integer i = 0 ; i < n_i ; i++) {
	vector(date)	dates;
	vector(date)	dates_q;
	vector(number)	amounts;

	if (!get_cashflows(i, dates, dates_q, amounts))
	    continue;

	++n_ok;

	integer	n_cf	= v_size(dates);

	for (integer i = 0 ; i < n_cf ; i++) {
	    if (null(dates[i]))
		throw(E_INVALID_ARG, "Null cashflow date");
	    periods.add((dates[i] - d0) / 365);
	}
    }

    if (n_ok == 0)
	return null<vector(number)>;

    if (n_ok != n_i)
	throw(E_NO_DATA, "Cashflows inconsistency");

    periods.add(0);

    return periods.get_content();
}

/**/

class cashflows_q_one : public cashflows_q
{
public:
    cashflows_q_one(date			trade_date,
		    vector(date)		dates,
		    vector(date)		dates_q,		      
		    vector(number)		amounts);

    virtual date	trade_date();
    virtual integer	n_instruments();

    virtual logical
    get_cashflows(integer			ix,
		  out vector(date)		dates,
		  out vector(date)		dates_q,		      
		  out vector(number)		amounts);

private:
    date		trade_date_;
    vector(date)	dates_;
    vector(date)	dates_q_;
    vector(number)	amounts_;
};

cashflows_q_one.cashflows_q_one(date		trade_date,
				vector(date)	dates,
				vector(date)	dates_q,
				vector(number)	amounts)
    : trade_date_(trade_date),
      dates_(clone_vector(dates)),
      dates_q_(clone_vector(dates_q)),
      amounts_(clone_vector(amounts))
{}

date cashflows_q_one.trade_date()
{
    return trade_date_;
}

integer cashflows_q_one.n_instruments()
{
    return 1;
}

logical cashflows_q_one.get_cashflows(integer			ix,
				      out vector(date)		dates,
				      out vector(date)		dates_q,
				      out vector(number)	amounts)
{
    dates	= dates_;
    dates_q	= dates_q_;
    amounts	= amounts_;

    return true;
}

cashflows_q
cashflows_q(date				trade_date,
	    vector(date)			dates,
	    vector(date)			dates_q,	      
	    vector(number)			amounts)
    option(com_name: "cashflows_q_one")
{
    return new cashflows_q_one(trade_date, dates, dates_q, amounts);
}

cashflows_q
cashflows_q(cashflows cf)
    option(com_name: "cashflows_q_cf")
{
    if(cf.n_instruments() != 1)
	return null<cashflows_q> ;

    vector(date) dates ;
    vector(number) amounts ;

    cf.get_cashflows(0, dates, amounts) ;

    vector(date) dates_q[v_size(dates)] ;

    for(integer i = 0 ; i < v_size(dates_q) ; i++)
	dates_q[i] = null<date> ;

    return new cashflows_q_one(cf.trade_date(), dates, dates_q, amounts) ;
}

/**/

class cashflows_q_sum : public cashflows_q
{
public:
    cashflows_q_sum(vector(cashflows_q) option(fixedref) cf);

    virtual date	trade_date();
    virtual integer	n_instruments();

    virtual logical
    get_cashflows(integer			ix,
		  out vector(date)		dates,
		  out vector(date)		dates_q,		      
		  out vector(number)		amounts);

private:
    vector(cashflows_q) option(fixedref)	cf_;
    vector(integer)				ix2cf_;
    vector(integer)				ix2subix_;
};

cashflows_q_sum.cashflows_q_sum(vector(cashflows_q) option(fixedref) cf)
    : cf_(cf)
{
    integer		n_cf	= v_size(cf_);

    if (n_cf == 0)
	throw(E_INVALID_ARG, "Empty vector");

    for (integer i = 0 ; i < n_cf ; i++) {
	if (null(cf_[i]))
	    throw(E_INVALID_ARG, "Null element in vector");

	integer	n_instr	= cf_[i].n_instruments();

	for (integer j = 0 ; j < n_instr ; j++) {
	    push_back(ix2cf_,	i);
	    push_back(ix2subix_,	j);
	}
    }
}

date cashflows_q_sum.trade_date()
{
    return cf_[0].trade_date();
}

integer cashflows_q_sum.n_instruments()
{
    return v_size(ix2cf_);
}

logical cashflows_q_sum.get_cashflows(integer			ix,
				      out vector(date)		dates,
				      out vector(date)		dates_q,
				      out vector(number)	amounts)
{
    return cf_[ix2cf_[ix]].get_cashflows(ix2subix_[ix], dates, dates_q, amounts);
}

cashflows_q
cashflows_q(vector(cashflows_q)		cashflows)
    option(com_name: "cashflows_q_sum")
{
    return new cashflows_q_sum(cashflows);
}
