/*	
	curve class
	Developer: Algorithmica Research, Kalle, Updated by Magnus N
							
*/

typedef string curve_name option(category: "Yield Curve");
string NO_SELECT_CURVE_ = 'Select curve type...';
string NO_SELECT_INSTR_ = 'Select instrument...';
//enum curve_err_filter option(category: "Yield Curve") {CRVF_NONE,CRVF_ERR,CRVF_NULLQ};

//-----------------------
// null_curve_name
//-----------------------
curve_name null_curve_name(curve_name option(nullable) c)
option (category: 'Yield Curve')
{
	if(null(c) || str_trim(c) == NO_SELECT_CURVE_)
		return null<curve_name>;
	else
		return c;

}
//-----------------------
// null_instr_name
//-----------------------
instrument_name null_instr_name(instrument_name option(nullable) i)
option (category: 'Instrument')
{
	if(null(i) || str_trim(i) == NO_SELECT_INSTR_)
		return null<instrument_name>;
	else
		return i;

}

class enum curve_instr_errs option(category: "Yield Curve")
{
	NONE, DB_ERR, NULL_INSTR, TRADE_DATE_ERR, RIBA_FIXED, NULL_QUOTE, DUP_MAT
};


class curve : public cashflows
    option(category: "Yield Curve")
    option(allow_undeclared_mfuncs)
{
private:
    void	__init(	vector(instrument)		instr,
			       	date option(nullable)	trade_date,
			       	string option(nullable)	name,
					logical 				remove_err,
					logical 				remove_nullq,
					logical 				remove_fixed,	
					logical 				remove_dup_mat);

	curve(curve, string option(nullable));

public:
    curve(	curve_name					name,
	  		date						trade_date,
	  		quote_side option(nullable)	quote_side = null<string>,
			logical 					remove_err = false,		//remove invalid instruments
			logical 					remove_nullq = false,	//remove instruments with null quotes
			logical 					remove_fixed = false,	//currently removing only fixed riba's
			logical 					remove_dup_mat = false,//remove duplicate maturities
			logical						use_prefetch = true);

    curve(	vector(instrument)			instr,
	  		string option(nullable)		name = null<string>,
			logical 					remove_err = false,
			logical 					remove_nullq = false,
			logical 					remove_fixed = false,
			logical 					remove_dup_mat = false,
			logical 					allow_db_name = true);

    curve(	vector(instrument_name)		instruments,
	  		date						trade_date,
	  		quote_side option(nullable)	quote_side	= null<string>,
			string option(nullable)		name = null<string>,			 
			logical 					remove_err = false,
			logical 					remove_nullq = false,
			logical 					remove_fixed = false,
			logical 					remove_dup_mat = false,
			logical 					allow_db_name = true,
			logical 					use_prefetch = true);

	
    vector(instrument)	instruments();
	vector(instrument)	err_instruments();
	vector(string)		err();
	vector(curve_instr_errs) err_e();
    string				name();
    logical				empty();

    curve				set_name(string option(nullable) name,
								logical allow_db_name = true);

    curve				exclude(vector(instrument_name) name_v);
    curve				exclude(instrument_name name);

    curve				chop_short(date d, logical interp = false);
    curve				chop_short_interp(date d);
    curve				chop_long(date d, logical interp = false);
    curve				chop_long_interp(date d);
	
public: // cashflows
    virtual date		trade_date();
    virtual integer		n_instruments();
	integer				n_err_instruments();
    virtual logical		get_cashflows(integer ix,out vector(date) dates,out vector(number) amounts);
	vector(number)		get_all_periods(logical);

private:
    date				trade_date_;
    vector(instrument)	instr_;
	vector(instrument)	instr_e_;
	vector(string)		err_;
	vector(curve_instr_errs) err_e_;
    string				name_;
};

curve.curve(curve c, string option(nullable) name)
			:	trade_date_(c.trade_date_), 
				instr_(!null(c.instr_) ? clone_vector(c.instr_) : null<vector(instrument)>), 
				instr_e_(!null(c.instr_e_) ? clone_vector(c.instr_e_) : null<vector(instrument)>), 
				err_(!null(c.err_) ? clone_vector(c.err_) : null<vector(string)>),
				err_e_(!null(c.err_e_) ? clone_vector(c.err_e_) : null<vector(curve_instr_errs)>),
				name_(null(name) ? c.name_ : name)

{}

vector(number) curve.get_all_periods(logical excl_first)
option(com_name: "get_all_periods_l")
{
	if (excl_first)
		return cashflows.get_all_periods_ex0();
	else
		return cashflows.get_all_periods();
}
/**/

date 				curve.trade_date() 			{ return trade_date_;}
integer 			curve.n_instruments()		{ return v_size(instr_);}
integer 			curve.n_err_instruments()	{ return v_size(instr_e_);}
vector(instrument) 	curve.instruments()			{ return null(instr_) ? null<vector(instrument)> : clone_vector(instr_);}
vector(string) 		curve.instrument_names()	{ return null(instr_) ? null<vector(string)> : 	instr_.name();}
vector(instrument) 	curve.err_instruments() 	{ return null(instr_e_) ? null<vector(instrument)> : clone_vector(instr_e_);}
vector(string) 		curve.err() 				{ return null(err_) ? null<vector(string)> : clone_vector(err_);}
vector(curve_instr_errs) curve.err_e() 			{ return null(err_e_) ? null<vector(curve_instr_errs)> : clone_vector(err_e_);}
string 				curve.name()				{ return name_;}
logical 			curve.empty()				{ return v_size(instr_) == 0;}

/*curve curve.set_name(string option(nullable) name)
{ 
	return new curve(instruments(), name,false,false,false,false);
}*/

curve curve.set_name(string option(nullable) name,
					logical allow_db_name)
{ 
	if(!allow_db_name && !null(name) && __exist_curve(name) )
		throw(E_INVALID_ARG, "Invalid curvename (name already exists in database)");
	name = null_curve_name(name);
	//	throw(E_INVALID_ARG, "Invalid curvename");
	return new curve(this, name);
	
}

logical curve.get_cashflows(integer				ix,
			    			out vector(date)	dates,
			    			out vector(number)	amounts)
{
	/*
    instrument	i	= instr_[ix];
    
	number	price	= i.dirty_price();
    date	settle	= i.settle_date();

    if (null(price))
	return false;

    dates = i.cash_flow_dates();
    amounts = i.cash_flow_amounts();

    if (price != 0)
    	if (v_size(dates) != 0 && settle == dates[0])
    	    amounts[0] -= price;
    	else {
    	    dates = concat([ settle ], dates);
    	    amounts = concat([ -price ], amounts);
    	}
	*/	
	try {
    	return instr_[ix].cash_flows_bootstrap(dates,amounts);		
	}
	catch {
		return false;//ex. we end up here when trying to bootstrap a non applicable instrument type
	}
}

/**/

void curve.__init(	vector(instrument)		instr,
		  			date option(nullable)	trade_date,
		  			string option(nullable)	name,
					logical 				remove_err ,
					logical 				remove_nullq ,
					logical 				remove_fixed,
					logical 				remove_dup_mat )
{
    trade_date_	= trade_date;
    name_		= name;
    integer	n	= v_size(instr);

    for (integer i = 0 ; i < n ; i++) {

		if (null(instr[i])) {//can happen when exclude function is run
			push_back(instr_e_, null<instrument>);
			push_back(err_, "null");
			push_back(err_e_, curve_instr_errs.NULL_INSTR);
		    continue;
		}

		if(!instr[i].is_valid()) {
			if(remove_err) { 
				push_back(instr_e_, instr[i]);	
				push_back(err_, instr[i].err_msg());
				push_back(err_e_, curve_instr_errs.DB_ERR);
				continue;
			}
			else {
				push_back(instr_, instr[i]);		
				continue;//we cannot check the other filter rules for an invalid instrument so we continue
			}
		}					
			
		if(null(instr[i].trade_date())) {
			if(remove_err) { 
				push_back(instr_e_, instr[i]);
				push_back(err_, "invalid trade date");
				push_back(err_e_, curve_instr_errs.TRADE_DATE_ERR);
				continue;
			}
			else {
				push_back(instr_, instr[i]);		
				continue;//we will not check the other filter rules when trade date is missing (static instrument)
			}
		}

		date trade_date_i = instr[i].trade_date();
		//if (null(trade_date_i))
		    //throw(E_INVALID_ARG, strcat(["Invalid trade date [",instr[i].name() ,"]"]));

		if (null(trade_date_))
		    trade_date_ = trade_date_i;
		else if (trade_date_i != trade_date_) {
			if(remove_err) { 
				push_back(instr_e_, instr[i]);	
				push_back(err_, "incompatible trade date");
				push_back(err_e_, curve_instr_errs.TRADE_DATE_ERR);
				continue;
			}		    
			//throw(E_INVALID_ARG, strcat(["Instruments have different trade_date [",instr[i].name() ,"]"]));
		}

		if(remove_nullq && null(instr[i].quote())) {			
			push_back(instr_e_, instr[i]);	
			push_back(err_, "null quote");
			push_back(err_e_, curve_instr_errs.NULL_QUOTE);
			continue;			
		}
			
		if(remove_fixed && instr[i].instr_type() == ..instr_type.RIBA && instr[i].riba_is_fixed()) {
			push_back(instr_e_, instr[i]);
			push_back(err_, "instrument is a riba and is fixed to maturity");
			push_back(err_e_, curve_instr_errs.RIBA_FIXED);
			continue;
		}

		push_back(instr_, instr[i]);				
    }

	//note: we can have invalid instruments in instr_ at this point

	n = v_size(instr_);
	
	if(n <= 1) return;
    
	sort(instr_, strcat(string(instr_.maturity()), instr_.name()));
	
	if(remove_dup_mat) {
		vector(integer) dup_i;
		for (integer i = 1 ; i < n ; i++) {			
			if(null(instr_[i].maturity()) || null(instr_[i-1].maturity()))//maturity == null for invalid instruments
				continue;
			if(instr_[i].maturity() == instr_[i-1].maturity()) {
				push_back(dup_i, i);			
			}
		}
		if(v_size(dup_i) > 0) {
			for (integer i = 0 ; i < v_size(dup_i) ; i++){ 
				push_back(instr_e_, instr_[dup_i[i]]);
				push_back(err_, "duplicate maturity");
				push_back(err_e_, curve_instr_errs.DUP_MAT);
			}
			instr_ = v_exclude(instr_, dup_i);
		}
	}
}

curve.curve(curve_name					name,
	    	date						trade_date,
	    	quote_side option(nullable)	quote_side,
			logical 					remove_err ,
			logical 					remove_nullq,
			logical 					remove_fixed,
			logical 					remove_dup_mat,
			logical 					use_prefetch)
{
    vector(instr_def)	i_defs;
	if(remove_err) {
		vector(instr_def)	i_e_defs;
		name = null_curve_name(name);
		if(null(name))
			throw(E_INVALID_ARG, "Invalid curvename");
		__expn_curve_ext(name, trade_date, i_defs, i_e_defs);
		instr_e_ = instrument(i_e_defs, trade_date, "");
		integer s = v_size(instr_e_);
		resize(err_,s);
		resize(err_e_,s);
		if(s>0){
			err_ 	= "db_err";
			err_e_ 	= curve_instr_errs.DB_ERR;
		}
	}	
	else 
		i_defs	= __expn_curve(name, trade_date);    	

	if(use_prefetch && trade_date == today()) {
		
		for (integer i = 0 ; i < v_size(i_defs) ; i++) {
			if(null(i_defs[i]))
				continue;
			CORE_INT.instr_prefetch(i_defs[i], quote_side);		
		}
	}
	
	__init(instrument(	i_defs, trade_date, quote_side), trade_date, name, remove_err,remove_nullq,
						remove_fixed, remove_dup_mat);
}

curve curve(curve_name					name,
	   	 	date						trade_date,
	    	quote_side option(nullable)	quote_side	= null<string>,
			logical 					remove_err = false,
			logical 					remove_nullq = false,
			logical 					remove_fixed = false,
			logical 					remove_dup_mat = false,
			logical 					use_prefetch = true)
    option(category: "Yield Curve")
    option(com_name: "db_curve")
{
	quote_side	= conv_null_string_com(quote_side) ;
    return new curve(name, trade_date, quote_side, remove_err,remove_nullq,remove_fixed,remove_dup_mat,use_prefetch);
}

curve.curve(vector(instrument)		instruments,
	    	string option(nullable)	name,
			logical 				remove_err,
			logical 				remove_nullq,
			logical 				remove_fixed,
			logical 				remove_dup_mat,
			logical 				allow_db_name)
{
	if(!allow_db_name && !null(name) && __exist_curve(name) )
		throw(E_INVALID_ARG, "Invalid curvename (name already exists in database)");
    __init(instruments, null<date>, name, remove_err,remove_nullq,remove_fixed,remove_dup_mat);
}

curve curve(vector(instrument)		instruments,
	    	string option(nullable)	name = null<string> ,
			logical 				remove_err = false,
			logical 				remove_nullq = false,
			logical 				remove_fixed = false,
			logical 				remove_dup_mat = false,
			logical 				allow_db_name = true)
    option(category: "Yield Curve")
    option(com_name: "curve_iv")
{
	name 	= conv_null_string_com(name ) ;
    return new curve(instruments, name, remove_err,remove_nullq,remove_fixed,remove_dup_mat, allow_db_name );
}

curve.curve(vector(instrument_name)		instruments,
	    	date						trade_date,
	    	quote_side option(nullable)	quote_side,
			string option(nullable)		name ,
			logical 					remove_err ,
			logical 					remove_nullq ,
			logical 					remove_fixed,
			logical 					remove_dup_mat,
			logical 					allow_db_name,
			logical 					use_prefetch)
{
	if(!allow_db_name && !null(name) && __exist_curve(name) )
		throw(E_INVALID_ARG, "Invalid curvename (name already exists in database)");
	
	if(use_prefetch && trade_date == today()) {
		
		error_info e = new error_info(true,true);	
		for (integer i = 0 ; i < v_size(instruments) ; i++) {
			instr_def def = instr_def(instruments[i],e);
			if(null(def) || e.is_error())
				continue;
			CORE_INT.instr_prefetch(def, quote_side);		
		}
	}
	vector(instrument) instr = instrument(	instruments, trade_date, quote_side);
	
    __init(instr,trade_date, name,  remove_err,remove_nullq,remove_fixed, remove_dup_mat);
}

curve curve(vector(instrument_name)		instruments,
	    	date						trade_date,
	    	quote_side option(nullable)	quote_side	= null<string>,
			string option(nullable)	name = null<string> ,
			logical 					remove_err = false,
			logical 					remove_nullq = false,
			logical 					remove_fixed = false,
			logical 					remove_dup_mat = false,
			logical 					allow_db_name = true,
			logical 					use_prefetch = true)
    option(category: "Yield Curve")
    option(com_name: "curve_namev")
{
	quote_side	= conv_null_string_com(quote_side) ;
	name 		= conv_null_string_com(name ) ;
    return new curve(instruments, trade_date, quote_side, name, remove_err,remove_nullq,
					remove_fixed, remove_dup_mat, allow_db_name,use_prefetch);
}

curve curve.exclude(vector(instrument_name) name_v)
    option(com_name: "exclude_v")
{
    set_istr		names	= new set_istr;
    vector(instrument)	i_v	= instruments();
    integer		n_i	= v_size(i_v);

    names.add(name_v);
    for (integer i = 0 ; i < n_i ; i++)
	if (names.includes(i_v[i].name()))
	    i_v[i] = null<instrument>;

    return ..curve(i_v, name());
}

curve curve.exclude(instrument_name name)
{
    return exclude([name]);
}

integer __find_chop_short(vector(instrument) v, date d)
{
    integer	n = v_size(v);

    for (integer i = 0 ; i < n ; i++)
	if (v[i].maturity() >= d)
	    return i;

    return n;
}

curve curve.chop_short(date d, logical interp)
{
    vector(instrument)	v	= instruments();
    integer	 n				= v_size(v);
    integer	 chop			= __find_chop_short(v, d);  
	instrument i_interp;

    if (interp && 0 < chop && chop < n && v[chop].maturity() > d) {
		date	d0	= trade_date();
		number	df	= bootstrap(this, ip_linear()).disc((d - d0) / 365);	
		__instrument i_interp_ 	= __create_disc_instr(d0, d, df);
		if (null(i_interp_))
		    throw(E_CALC, "Interpolation failed");
		i_interp = new bill(i_interp_);
		if (null(i_interp))
		    throw(E_CALC, "Interpolation failed");
    }
	
    v = sub_vector(v, chop, n - chop);
	if (!null(i_interp))
    	push_back(v, i_interp);

    return ..curve(v, name());
}

curve curve.chop_short_interp(date d)
{
    return chop_short(d, true);
}

integer __find_chop_long(vector(instrument) v, date d)
{
    integer	n = v_size(v);

    for (integer i = 0 ; i < n ; i++)
	if (v[i].maturity() > d)
	    return i;

    return n;
}

curve curve.chop_long(date d, logical interp)
{
    vector(instrument)	v	= instruments();
    integer	n				= v_size(v);
    integer	chop			= __find_chop_long(v, d);
    instrument i_interp;

    if (interp && 0 < chop && chop < n && v[chop - 1].maturity() < d) {
		date		d0			= trade_date();
		number		df			= bootstrap(this,ip_linear()).disc((d - d0) / 365);
		__instrument i_interp_ 	= __create_disc_instr(d0, d, df);
		if (null(i_interp_))
	    	throw(E_CALC, "Interpolation failed");
		i_interp = new bill(i_interp_);
		if (null(i_interp))
	    	throw(E_CALC, "Interpolation failed");
    }	

    v = sub_vector(v, 0, chop);
	if (!null(i_interp))
    	push_back(v, i_interp);

    return ..curve(v, name());
}

curve curve.chop_long_interp(date d)
{
    return chop_long(d, true);
}

/**/

class enum curve_status option(category: "Yield Curve")
{
	OK, NAME_NULL, NAME_ERR, NULL_CURVE, EMPTY, EMPTY_NULLS, INSTR_ERR, INSTR_NULL_QUOTE
};

//NONE, DB_ERR, NULL_INSTR, TRADE_DATE_ERR, RIBA_FIXED, NULL_QUOTE, DUP_MAT
//NONE, RIBA_FIXED, NULL_QUOTE --> NO ERR
//DB_ERR, NULL_INSTR, TRADE_DATE_ERR, DUP_MAT --> OK_INSTR_ERR

curve_status status_from_err(vector(curve_instr_errs) option(nullable) err_e)
{
	if(null(err_e) || v_size(err_e) == 0)
		return curve_status.OK;
		
	integer err_size 	= v_size(err_e);		
	vector(integer) null_v = v_find_eq(err_e, curve_instr_errs.NULL_QUOTE);
	vector(integer) none_v = v_find_eq(err_e, curve_instr_errs.NONE);
	vector(integer) fixed_v = v_find_eq(err_e, curve_instr_errs.RIBA_FIXED);
	
	integer real_err_size = err_size - v_size(null_v) - v_size(none_v)- v_size(fixed_v);
	if (real_err_size > 0)
		return curve_status.INSTR_ERR;
	else
		return v_size(null_v) > 0 ? curve_status.INSTR_NULL_QUOTE : curve_status.OK;	
}


void curve_diagnostics(	curve_name  		name,
						date 				trade_date,
						quote_side 	option(nullable) quote_side,
						out curve			orig_curve,
						out curve_status	status,						
						out vector(string) 	err_name,
						out vector(curve_instr_errs) err_e,
						out vector(string) 	err_desc)
	option(category: "Yield Curve")
{		
	if(!__exist_curve(name)) {
		orig_curve = null<curve>;
		status = curve_status.NAME_ERR;
		err_name = ["curve_name unknown"];
		err_desc = ["curve_name unknown"];
		err_e = [curve_instr_errs.NONE];
		return ;
	}
	
	orig_curve = curve(name,trade_date,quote_side, false ,false,false ,false );
	if(null(orig_curve) ){
		status = curve_status.NULL_CURVE;
		err_name = ["null curve"];
		err_desc = ["null curve"];
		err_e = [curve_instr_errs.NONE];
		return ;
	}

	integer null_instr = orig_curve.n_err_instruments();
	
	if(orig_curve.empty()) {
		status = null_instr == 0 ? curve_status.EMPTY : curve_status.EMPTY_NULLS;
		err_name = orig_curve.err_instruments().name();
		err_desc = orig_curve.err();
		err_e = orig_curve.err_e();
		return;
	}
	
	vector(instrument) instr_orig = orig_curve.instruments();
	integer size_orig = v_size(instr_orig);
	
	curve cr = curve(name,trade_date,quote_side,true ,true,true ,true );
	if(null(cr)){//possible?
		status = curve_status.NULL_CURVE;
		err_name = orig_curve.instruments().name();
		resize(err_desc,v_size(err_name));
		resize(err_e,v_size(err_name));
		err_desc = "unknown error";
		err_e = curve_instr_errs.DB_ERR;
		return;
	}
		
	if(null_instr == 0) {
		err_name 			= cr.err_instruments().name();
		err_desc 			= cr.err();
		err_e 				= cr.err_e();
		status 				= status_from_err(err_e);	
		return;
	}

	vector(instrument) instr_ok = cr.instruments();
	integer size_ok 			= v_size(instr_ok);

	if(size_orig == size_ok) {
		status = curve_status.OK;
		return;
	}
	
	err_name = concat(orig_curve.err_instruments().name(),cr.err_instruments().name());
	err_desc = concat(orig_curve.err(),cr.err());
	err_e 	= concat(orig_curve.err_e(),cr.err_e());
	status 	= status_from_err(err_e);
	
	return;
}


logical check_curve(curve option(nullable) c)
{
	if(null(c)) return false;
	return !c.empty();
}

module CORE_INT
{
	//-------------------------------------------------------
	// equal_instr_type
	//-------------------------------------------------------
	logical equal_instr_type(instr_type 			t,
							 vector(instr_type) 	t_v)
	{
		integer size = v_size(t_v);
		for(integer i=0;i<size;i++){
			if(t == t_v[i])
				return true;
		}
		return false;
	}
	//-------------------------------------------------------
	// count_instr_type
	//-------------------------------------------------------
	integer count_instr_type(	vector(instrument) 	v_instr,
								vector(instr_type) 	t,
								out vector(integer) idx)
	{
		error_info err 	= new error_info(true,false);
		integer size 	= v_size(v_instr);
		integer c 		= 0;
		resize(idx,0);
		
		for(integer i=0;i<size;i++){
			instr_type ti = v_instr[i].instr_type(err);
			if(!err.is_error() && equal_instr_type(ti, t)) {				
				c++;
				push_back(idx,i);
			}
			CORE_INT.reset_single(err);
		}
		return c;
	}
	//-------------------------------------------------------
	// extract_instr_type
	//-------------------------------------------------------
	void extract_instr_type(vector(instrument) 		v_instr,
							vector(instr_type) 		t,
							out vector(instrument) 	v_instr_t,
							out vector(instrument) 	v_instr_rest )
	{
		integer size 	= v_size(v_instr);
		vector(integer) idx;
		integer size_t 	= count_instr_type(v_instr, t, idx);
		if(size == size_t) {
			v_instr_t = v_instr;
			v_instr_rest = null<vector(instrument)>;
		}
		else {
			v_instr_t 					= ..sub_vector(v_instr, idx);
			vector(integer) idx_s 		= index_vector(size);
			vector(integer) idx_compl 	= v_complement (idx_s,idx);
			v_instr_rest 				= ..sub_vector(v_instr, idx_compl);						
		}
		
	}
	//-----------------------------------------------------------------------------
	//   add_inst
	//-----------------------------------------------------------------------------
	void add_inst(	vector(date) 		base_m,
					vector(date) 		add_m,
					vector(instrument) 	add_i,
					number 				min_days_chop,
					out vector(instrument) option(nullable) v_i)
	{
		for(integer i = 0; i < v_size(add_m); i++){
			vector(number) n = abs(base_m - add_m[i]);
			if(v_min(n) > 	min_days_chop)
				push_back(v_i,add_i[i]);
		}
	}
}

//-----------------------------------------------------------------------------
//   curve_subset
//-----------------------------------------------------------------------------
curve curve_subset(curve 		curve,
				   vector(instr_type) 	t)
option(category: "Yield Curve")
{
	if(curve.empty())
		return curve;

	vector(instrument) 	v_instr, v_instr_rest ;
	CORE_INT.extract_instr_type(curve.instruments(), t, v_instr, v_instr_rest );
	return ..curve(v_instr, curve.name());
}

//-----------------------------------------------------------------------------
//   merge
//-----------------------------------------------------------------------------
curve merge(curve option(nullable)	curve1,
	    	curve option(nullable)	curve2)
option (category: 'Yield Curve/001 Curve Blending')
{
    if (!check_curve(curve1)){
		return curve2;
	}
    else if (!check_curve(curve2)){
		return curve1;
	}
    else {
		string name = strcat([curve1.name(), "/", curve2.name()]);
		return curve(concat(curve1.instruments(),curve2.instruments()),name);
	}
}

//-----------------------------------------------------------------------------
// merge (added by MN)
// min_days_chop = 0: prevents equal maturities to be merged into curve (prio setting determines which one is kept)
// min_days_chop > 0: prevents maturities within min_days diff to be merged into curve (prio setting determines which ones are kept)
//-----------------------------------------------------------------------------
curve merge(curve 	option(nullable) curve1,
			curve 	option(nullable) curve2,
			logical curve1_prio,
			number 	min_days_chop = 0)
option (category: 'Yield Curve/001 Curve Blending')
option(com_name: "merge_ext")
{
	if (!check_curve(curve1))
		return curve2;
	else if (!check_curve(curve2))
		return curve1;

	string name = strcat([curve1.name(), "/", curve2.name()]);
	vector(instrument) v_c1 = curve1.instruments();
	vector(instrument) v_c2 = curve2.instruments();
	vector(date) c1_m 		= v_c1.maturity();
	vector(date) c2_m 		= v_c2.maturity(); 
	if(min_days_chop < 0) 
		min_days_chop = -min_days_chop ;

	if(curve1_prio){			
		CORE_INT.add_inst(c1_m,c2_m,v_c2,min_days_chop ,v_c1);																								
		return curve(v_c1,name);
	}
	else {
		CORE_INT.add_inst(c2_m,c1_m,v_c1,min_days_chop ,v_c2);																								
		return curve(v_c2,name);
	}	
}
//-----------------------
// curve_filter
// OBSOLETE: use remove_fixed argument in curve constructor instead
//-----------------------
curve curve_filter(curve crv)
{
	try {
		vector(instrument) vi = crv.instruments();

		vector(instrument) v ;

		for(integer i=0; i< v_size(vi); i++){			                                      
			if(vi[i].instr_type() == ..instr_type.RIBA && !vi[i].riba_is_fixed())											
				push_back(v, vi[i]);			
			else
				push_back(v, vi[i]) ;
		}              

		if(v_size(v) != v_size(vi)){
			return curve(v, crv.name());
		}
		else
            return crv;
	}
    catch {
        return null<curve>;
    }
}

/*
//-------------------------------------------------------
// bootstrap_on_extrap_flat
// needed because extrapolation in bootstrap_on does not work as intended FIX_ME
//-------------------------------------------------------
disc_func bootstrap_on_extrap_flat(	curve crv, 
									//interpolator ip,
									calendar option(nullable) cal)
{
	//disc_func df = bootstrap_on(crv, ip, cal);

	vector(instrument) vi 	= crv.instruments();	
	integer size 			= v_size(vi);
	QL_REQUIRE(size>0,"invalid curve");
	
	date max = vi[size-1].maturity();
	date td = crv.trade_date();
	//number t =  (max - crv.trade_date())/365;
	return bootstrap_on_extrap_flat(crv,td,max,cal);
}

//-------------------------------------------------------
// bootstrap_on_ext_extrap_flat
// needed because extrapolation in bootstrap_on does not work as intended FIX_ME
//-------------------------------------------------------
disc_func bootstrap_on_ext_extrap_flat(	disc_func df_pre,
										date cut_off_d,
										curve crv,
										//interpolator ip,
										calendar option(nullable) cal)
{
	//disc_func df = bootstrap_on_ext(df_pre,cut_off_d,crv, ip, cal);

	vector(instrument) vi 	= crv.instruments();	
	integer size 			= v_size(vi);
	QL_REQUIRE(size>0,"invalid curve");
	
	date max = vi[size-1].maturity();
	date td = crv.trade_date();
	//number t =  (max - crv.trade_date())/365;
	return bootstrap_on_ext_extrap_flat(df_pre,cut_off_d, crv,td,max,cal);
}
*/