/*	
	Blending functions
	Developed by Algorithmica Research, Magnus Nyström
			
	ToDo: 

	Comments:	

*/

option(null: hard);	
		
class enum curve_prio {
	//NONE	option(str: "None"),
	SHORT	option(str: "Short"),
	MIDDLE	option(str: "Middle"),
	LONG 	option(str: "Long")
} ;

out vector(string) curve_prio_list()
{
	vector(curve_prio) e;
	expand_enum(e);
	return string(e);
}

curve_prio curve_prio_map(number p)
{
	if(p == 1)
		return curve_prio.SHORT;
	else if (p==2)
		return curve_prio.MIDDLE;
	else if(p == 3)
		return curve_prio.LONG;
	
	QL_FAIL("invalid prio order");
}

curve_prio curve_prio_map(string p)
{
	if (p == "FWD")
		return curve_prio.MIDDLE;
	curve_prio e;
	str_to_enum(p,e);
	return e;
}

class enum curve_merge {
	NONE	option(str: "None"),
	ALL		option(str: "All"),
	MIDDLE	option(str: "Middle")
} ;

curve_merge curve_merge_map(string p)
{
	if (p == "FWD")
		return curve_merge.MIDDLE;
	curve_merge e;
	str_to_enum(p,e);
	return e;
}

out vector(string) curve_merge_list()
{
	vector(curve_merge) e;
	expand_enum(e);
	return string(e);
}

void curve_merge_parm(curve_merge e, out logical merge_crv , out logical merge_middle)

{
	merge_middle 	= false;		 
	merge_crv 		= false;
	if(e == curve_merge.ALL) 
		merge_crv = true;
	else if (e == curve_merge.MIDDLE)
		merge_middle = true;	
}

		
module CORE_INT
{

	//-------------------------------------------------------
	// set_prio_order
	//-------------------------------------------------------
	void set_prio_order(curve option(nullable)			short_crv, 						
						curve option(nullable)			fra_crv,
						curve option(nullable)			swap_crv,  					
						out curve_prio 	prio1,
						out curve_prio 	prio2)
	{
		logical sh_ok 	= ..check_curve(short_crv);
		logical f_ok 	= ..check_curve(fra_crv);
		logical sw_ok 	= ..check_curve(swap_crv);
	
		if(!sh_ok && !f_ok && !sw_ok){
			prio1 = curve_prio.LONG;
			prio2 = curve_prio.MIDDLE;
		}
		else if(!sh_ok && !f_ok){
			prio1 = curve_prio.LONG;
			prio2 = curve_prio.MIDDLE;
		}	
		else if(!sw_ok && !f_ok){
			prio1 = curve_prio.SHORT;
			prio2 = curve_prio.MIDDLE;
		}
		else if(!sh_ok && !sw_ok){
			prio1 = curve_prio.MIDDLE;
			prio2 = curve_prio.LONG;
		}
		else if(!f_ok){ //sh_ok && sw_ok
			if(prio1 == curve_prio.MIDDLE)
				prio1 = prio2 == curve_prio.MIDDLE ? curve_prio.LONG : prio2;
			if(prio2 == curve_prio.MIDDLE)
				prio2 = prio1 == curve_prio.LONG ? curve_prio.SHORT : curve_prio.LONG;
		}
		else if(!sw_ok){ //sh_ok && f_ok
			if(prio1 == curve_prio.LONG)
				prio1 = prio2 == curve_prio.LONG ? curve_prio.MIDDLE : prio2;
			if(prio2 == curve_prio.LONG)
				prio2 = prio1 == curve_prio.MIDDLE ? curve_prio.SHORT : curve_prio.MIDDLE;
		}
		else if(!sh_ok){ //f_ok && sw_ok
			if(prio1 == curve_prio.SHORT)
				prio1 = prio2 == curve_prio.SHORT ? curve_prio.MIDDLE : prio2;
			if(prio2 == curve_prio.SHORT)
				prio2 = prio1 == curve_prio.LONG ? curve_prio.MIDDLE : curve_prio.LONG;
		}
	
		return; // all 3 ok	
	}
	//-----------------------------------------------------------------------------
	//  blend_spot_int
	//-----------------------------------------------------------------------------
	curve blend_spot_int(	curve 					short, 
							curve 					long,							
							logical 				shortprio,
							out date 	knot,
							date option(nullable) 		ext_knot,
							integer 				min_days_chop = 0)
	{
		if(min_days_chop < 0) min_days_chop = 0;
		if(shortprio){
			knot = null(ext_knot) ? v_max(short.instruments().maturity()) : ext_knot;
			long = long.chop_short(knot+1+min_days_chop);	
			return null(long) ? short: merge(short, long);
		}
		else {
			knot = null(ext_knot) ? v_min(long.instruments().maturity()): ext_knot;
			short = short.chop_long(knot-1-min_days_chop);	
			return null(short ) ? long: merge(short, long);
		}
	}
	//-----------------------------------------------------------------------------
	//  blend_fraspot_int
	//-----------------------------------------------------------------------------
	curve blend_fraspot_int(	curve 					fra, 
								curve 					spot,
								logical 				merge_fra,							
								logical					fraprio,								
								logical 				no_overlap_fra,								
								out date 	knot,
								date option(nullable) 		ext_knot,
								integer 				min_days_chop = 0)
	{							
		if(merge_fra)
			return merge(fra, spot, fraprio,min_days_chop);
		
		if(min_days_chop < 0) min_days_chop = 0;

		date framin = no_overlap_fra ? v_min(fra.instruments().settle_date()): v_min(fra.instruments().maturity());		
		date framax = v_max(fra.instruments().maturity());
		date spotmin = v_min(spot.instruments().maturity());
		date spotmax = v_max(spot.instruments().maturity());
		
		if(fraprio){
			if(framin <= spotmin && framax >= spotmax )
				return fra;
			else if(framin <= spotmin){
				knot = null(ext_knot) ? framax : ext_knot;
				spot = spot.chop_short(knot+1+min_days_chop);	
				return null(spot) ? fra: merge(fra, spot);
			}
			else if(framax >= spotmax){
				knot = null(ext_knot) ? framin+1 : ext_knot;
				spot = spot.chop_long(knot-1-min_days_chop);	
				return null(spot) ? fra: merge(fra, spot);
			}
			else {//framin>spotmin && framax < spotmax
				knot = null(ext_knot) ? framin+1 : ext_knot;
				curve front = spot.chop_long(knot-1-min_days_chop);
				fra = null(front ) ? fra: merge(fra, front);
				date knot2 = framax;
				spot = spot.chop_short(knot2+1+min_days_chop);
				return null(spot) ? fra: merge(fra, spot);
			}
		}
		else {
			if(framin>=spotmin && framax <= spotmax)
				return spot;
			else if(framax <= spotmin/*framin <= spotmin*/){
				knot = null(ext_knot) ? spotmin : ext_knot;
				fra = fra.chop_long(knot-1-min_days_chop);	
				return null(fra) ? spot: merge(fra, spot);
			}
			else if(framax >= spotmax){
				knot = null(ext_knot) ? spotmax : ext_knot;
				fra = fra.chop_short(knot+1+min_days_chop);	//remove short fras
				return null(fra) ? spot: merge(fra, spot);
			}
			else {//framin < spotmin && framax > spotmax
				knot = null(ext_knot) ? spotmin : ext_knot;
				curve front = fra.chop_long(knot-1-min_days_chop);
				spot = null(front) ? spot: merge(spot, front);
				date knot2 = spotmax;
				fra = fra.chop_short(knot2+1+min_days_chop);
				return null(fra) ? spot: merge(fra, spot);
			}
		}
	}
	
	//-----------------------------------------------------------------------------
	//   blend_3_int
	// note: c2 curve is assumed to be a fra/fut curve
	//-----------------------------------------------------------------------------
	curve blend_3_int(	curve 		short,
					    curve 		fra,
					    curve 		long,							
						curve_prio	prio1,
						curve_prio	prio2,
						logical 	merge_fra,
						logical 	no_overlap_fwd,
						out  vector(date) option(nullable) knot,
						integer		min_days_chop = 0)
	{
		date k;
		resize(knot,0);		
		if(prio1 == curve_prio.SHORT){
			if(prio2 == curve_prio.MIDDLE){//1,2,3
				short = blend_fraspot_int(fra, short, merge_fra, false,false,k,null<date>,min_days_chop);
				if(!null(k)) push_back(knot,k);	
				short = blend_spot_int(short,long,true,k,null<date>,min_days_chop);				
				if(!null(k)) push_back(knot,k);	
				return short;		
			}
			else {//1,3,2
				short = blend_spot_int(short,long,true,k,null<date>,min_days_chop );					
				if(!null(k)) push_back(knot,k);		
				short = blend_fraspot_int(fra, short, merge_fra, false,false,k,null<date>,min_days_chop);
				if(!null(k)) push_back(knot,k);	
				return short;					
			}
		}
		else if(prio1 == curve_prio.MIDDLE){			
			date framax = v_max(fra.instruments().maturity());
			date framin = v_min(fra.instruments().settle_date());
			date shortmax = v_max(short.instruments().maturity());
			date longmin = v_min(long.instruments().maturity());
			
			if(prio2 == curve_prio.SHORT){ // 2,1,3
				if(framin<=shortmax){
					short = blend_fraspot_int(fra,short,merge_fra,true, no_overlap_fwd,k,null<date>,min_days_chop);
					if(!null(k)) push_back(knot,k);	
				}
				else if(framin>=longmin){
					long = blend_fraspot_int(fra,long, merge_fra,true, true, k,null<date>,min_days_chop);					
					if(!null(k)) push_back(knot,k);	
				}
				else {  //framin > shortmax
					short = blend_fraspot_int(fra,short,merge_fra,true, no_overlap_fwd,k,null<date>,min_days_chop);
					if(!null(k)) push_back(knot,k);	
				}
				short = blend_spot_int(short,long,true,k,null<date>, min_days_chop);
				if(!null(k)) push_back(knot,k);	
				return short;
			}
			else {//2,3,1
				date ext_knot = null<date>;
				if(framax>=longmin){					
					long = blend_fraspot_int(fra,long, merge_fra, true, true, k,null<date>,min_days_chop);
					if(no_overlap_fwd && framin < longmin) ext_knot = framin+1;					
					if(!null(k)) push_back(knot,k);	
				}
				else if(framax<=shortmax){
					short = blend_fraspot_int(fra,short,merge_fra,true, no_overlap_fwd,k,null<date>,min_days_chop);
					if(!null(k)) push_back(knot,k);	
				}
				else {//2,3,1 framin < longmin					
					long = blend_fraspot_int(fra,long, merge_fra,true, true,k,null<date>,min_days_chop);
					if(no_overlap_fwd) ext_knot = framin+1;
					if(!null(k)) push_back(knot,k);	
				}
				long = blend_spot_int(short,long,false,k, ext_knot,min_days_chop);
				if(!null(k)) push_back(knot,k);	
				return long;	
			}
		}
		else /*if prio1 == 3*/{
			if(prio2 == curve_prio.SHORT){//3,1,2
				short = blend_spot_int(short, long,false,k,null<date>,min_days_chop);//ok
				if(!null(k)) push_back(knot,k);	
				short = blend_fraspot_int(fra, short, merge_fra, false,false, k,null<date>,min_days_chop);//ok
				if(!null(k)) push_back(knot,k);	
				return short;	
			}
			else{//3,2,1
				date framax = v_max(fra.instruments().maturity());
				date framin = v_min(fra.instruments().settle_date());
				date shortmax = v_max(short.instruments().maturity());
				date longmin = v_min(long.instruments().maturity());
				number size_long = v_size(long.instruments()) ;				
				date ext_knot = null<date>;

				if(framax>=longmin){					
					long = blend_fraspot_int(fra,long, merge_fra, false, true, k,null<date>,min_days_chop);
					if(no_overlap_fwd && framin < longmin && size_long < v_size(long.instruments())) ext_knot = framin+1;
					if(!null(k)) push_back(knot,k);	
				}
				else if(framax<=shortmax){
					short  = blend_fraspot_int(fra,short,merge_fra, true, no_overlap_fwd,k,null<date>,min_days_chop);
					if(!null(k)) push_back(knot,k);	
				}
				else {					
					short = blend_fraspot_int(fra,short, merge_fra, true, no_overlap_fwd,k,null<date>,min_days_chop);
					if(!null(k)) push_back(knot,k);	
				}
				long = blend_spot_int(short,long,false,k, ext_knot,min_days_chop);
				if(!null(k)) push_back(knot,k);	
				return long;						
			}
		}		
	}


	/*-----------------------------------------------------------------------
	  Name :  	blend_spot  
	  Notes :           
	  Changes:
	  			Date    		Who     Detail
	  			? 				MN     	Original code (moved from fiws) 
				20181208		MN		changed to hidden
	  ----------------------------------------------------------------------*/
	curve blend_spot(	curve option(nullable) 	short, 
						curve option(nullable) 	long,							
						logical					shortprio,	
						out date knot,
						integer 				min_days_chop = 0)
	{
	    knot 				= null<date>;		
		logical short_ok 	= ..check_curve(short);
		logical long_ok 	= ..check_curve(long);
		if (short_ok && long_ok) 
			return CORE_INT.blend_spot_int(short,long,shortprio,knot,null<date>,min_days_chop)	;  
		else if (long_ok)
			return long;
		else if (short_ok)
			return short;
		else 
			return null<curve>;					    
	}

	/*-----------------------------------------------------------------------
	  Name :  	blend_fra  
	  Notes :           
	  Changes:
	  			Date    		Who     Detail
	  			? 				MN     	Original code (moved from fiws) 
				20181208		MN		changed to hidden
	  ----------------------------------------------------------------------*/
	curve blend_fra(curve option(nullable) 	fra, 
					curve option(nullable) 	spot,
					logical					merge_fra,							
					logical					fraprio,
					logical 				no_overlap_fra,	
					out date knot,
					integer					min_days_chop = 0)
	{
	    knot 			= null<date>;
		logical fra_ok 	= ..check_curve(fra);
		logical spot_ok = ..check_curve(spot);
		if (fra_ok && spot_ok) 			
			return CORE_INT.blend_fraspot_int(fra,spot,merge_fra,fraprio,no_overlap_fra,knot,null<date>,min_days_chop)	;		
		else if (spot_ok)
			return spot;
		else if (fra_ok)
			return fra;
		else 
			return null<curve>;				    
	}

	/*-----------------------------------------------------------------------
	  Name :  	blend_curve  
	  Notes :           
	  Changes:
	  			Date    		Who     Detail
	  			? 				MN     	Original code (moved from fiws) 
				20181208		MN		changed to hidden
	  ----------------------------------------------------------------------*/
	curve blend_curve(	curve option(nullable) 	short_curve,
					 	curve option(nullable) 	fwd_curve,
				    	curve option(nullable) 	long_curve,
						curve_prio 				curve_prio_1,
						curve_prio				curve_prio_2,
						logical 				fwd_merge,
						logical 				fwd_no_overlap,
						out vector(date)   knots,
						integer					min_days_chop = 0)
	{
	    resize(knots,0);
		logical c1_ok = ..check_curve(short_curve);
		logical c2_ok = ..check_curve(fwd_curve);
		logical c3_ok = ..check_curve(long_curve);		
	
		if (c1_ok && c2_ok && c3_ok) 	
			return CORE_INT.blend_3_int(short_curve, fwd_curve, long_curve, curve_prio_1, curve_prio_2, fwd_merge, fwd_no_overlap,knots,min_days_chop);   	
		else if (c1_ok && c2_ok){
			date k;		
			curve_prio prio = curve_prio_1 == curve_prio.LONG ? curve_prio_2 : curve_prio_1;			
			curve c = CORE_INT.blend_fraspot_int(fwd_curve,short_curve , fwd_merge, prio==curve_prio.MIDDLE, fwd_no_overlap,k,null<date>,min_days_chop);		
			if(!null(k)) push_back(knots,k);	
			return c;		
		}
		else if (c1_ok && c3_ok){
			date k;		
			curve_prio prio = (curve_prio_1 == curve_prio.MIDDLE) ? curve_prio_2 : curve_prio_1;		
			curve c = CORE_INT.blend_spot_int(short_curve, long_curve, prio==curve_prio.SHORT, k,null<date>,min_days_chop);
			if(!null(k)) push_back(knots,k);					
			return c;
	
		}
		else if (c2_ok && c3_ok){
			date k;		
			curve_prio prio = (curve_prio_1 == curve_prio.SHORT) ? curve_prio_2 : curve_prio_1;
			curve c = CORE_INT.blend_fraspot_int(fwd_curve,long_curve, fwd_merge, prio==curve_prio.MIDDLE, true,k,null<date>,min_days_chop);		
			if(!null(k)) push_back(knots,k);	
			return c;	
		}
		else if (c1_ok)
			return short_curve;	
		else if (c2_ok)
			return fwd_curve;	
		else if (c3_ok)
			return long_curve;		
		else
			return null<curve>;
	    
	}

}
/*-----------------------------------------------------------------------
  Name :  	blend_curve
  ----------------------------------------------------------------------*/
curve blend_curve(	curve option(nullable) 	short_curve, 
					curve option(nullable) 	long_curve,							
					logical					short_curve_prio,	
					out date knot,
					integer 				min_days_chop = 0)
option (category: 'Yield Curve/001 Curve Blending')
option(com_name: "blend_curve_2_ext")
{
	return CORE_INT.blend_spot(	short_curve,long_curve,	short_curve_prio,knot, min_days_chop );
}

curve blend_curve(	curve option(nullable) 	short_curve, 
					curve option(nullable) 	long_curve,							
					logical					short_curve_prio=false)
option (category: 'Yield Curve/001 Curve Blending')
option(com_name: "blend_curve_2")
{
    date knot 				= null<date>;
	integer min_days_chop 	= 0	;	
	return CORE_INT.blend_spot(	short_curve,long_curve,	short_curve_prio,knot, min_days_chop );				    
}

/*-----------------------------------------------------------------------
  Name :  	blend_curve_fwd
  ----------------------------------------------------------------------*/
curve blend_curve_fwd(	curve option(nullable) 	fwd_curve, 
						curve option(nullable) 	spot_curve,
						logical					fwd_curve_prio,
						logical 				fwd_no_overlap,	
						out date knot,
						integer					min_days_chop = 0)
option (category: 'Yield Curve/001 Curve Blending')
option(com_name: "blend_curve_2_fwd_ext")
{
	logical fwd_merge = false;
	return CORE_INT.blend_fra(fwd_curve, spot_curve,fwd_merge,fwd_curve_prio,fwd_no_overlap,knot,min_days_chop );				    
}

curve blend_curve_fwd(	curve option(nullable)  fwd_curve, 
						curve option(nullable) 	spot_curve,												
						logical					fwd_curve_prio = true,
						logical 				fwd_no_overlap = true)
option (category: 'Yield Curve/001 Curve Blending')
option(com_name: "blend_curve_2_fwd")
{
	date knot;
	logical fwd_merge = false;
    return CORE_INT.blend_fra(fwd_curve, spot_curve,fwd_merge,fwd_curve_prio,fwd_no_overlap, knot );			    
}

/*-----------------------------------------------------------------------
  Name :  	blend_curve
  Notes :   fwd curve is assumed to be a fra/fut curve
			default prio order is fwd/long/short
  Changes:
  			Date    		Who     Detail
  			? 				MN     	Original code (moved from fiws) 
			20181208		MN		Added default args
  ----------------------------------------------------------------------*/
curve blend_curve(	curve option(nullable) short_curve,
				 	curve option(nullable) fwd_curve,
			    	curve option(nullable) long_curve,
					curve_prio 	curve_prio_1 = curve_prio.MIDDLE,
					curve_prio	curve_prio_2 = curve_prio.LONG,					
					logical 	fwd_no_overlap=true)
option (category: 'Yield Curve/001 Curve Blending')
option(com_name: "blend_curve_3")
{
	logical 	fwd_merge = false;
	vector(date)  knots;
	integer		min_days_chop = 0;
	return CORE_INT.blend_curve(short_curve,fwd_curve,long_curve,curve_prio_1,curve_prio_2,fwd_merge,
								fwd_no_overlap,knots,min_days_chop );
}

/*-----------------------------------------------------------------------
  Name :  	blend_curve  
  Notes :           
  Changes:
  			Date    		Who     Detail
  			? 				MN     	Original code (moved from fiws) 
			20181208		MN		changed to hidden
  ----------------------------------------------------------------------*/
curve blend_curve(	curve option(nullable) 	short_curve,
				 	curve option(nullable) 	fwd_curve,
			    	curve option(nullable) 	long_curve,
					curve_prio 				curve_prio_1,
					curve_prio				curve_prio_2,					
					logical 				fwd_no_overlap,
					out vector(date)   knots,
					integer					min_days_chop = 0)
option (category: 'Yield Curve/001 Curve Blending')
option(com_name: "blend_curve_3_ext")
{
	logical fwd_merge = false;
	return CORE_INT.blend_curve(short_curve,fwd_curve,long_curve,curve_prio_1,curve_prio_2,fwd_merge,
								fwd_no_overlap,knots,min_days_chop);
}

/*-----------------------------------------------------------------------
  Name :  	blend_curve_merge
  Notes :   fwd_curve is assumed to be a fra/fut curve	
			default prio order is fwd/long/short       
  Changes:
  			Date    		Who     Detail
  			? 				MN     	Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
curve blend_curve_merge(curve option(nullable) 		short_curve,
						curve option(nullable)		fwd_curve, 
						curve option(nullable) 		long_curve,  								
						curve_prio					curve_prio_1 = curve_prio.MIDDLE,
						curve_prio					curve_prio_2 = curve_prio.LONG,																			
						integer						min_days_chop = 0)
option (category: 'Yield Curve/001 Curve Blending')
{
	vector(logical) curve_ok[3];				
	curve_ok[0] 	= check_curve(short_curve);
	curve_ok[1] 	= check_curve(fwd_curve);
	curve_ok[2] 	= check_curve(long_curve);
	number no_crv 	= v_size(v_find(curve_ok, true));  
	curve crv;

	if(no_crv == 0)
		return null<curve>;

	if(curve_ok[0] && curve_ok[1] && curve_ok[2]){
		if(curve_prio_1 == curve_prio.SHORT){
			if(curve_prio_2 == curve_prio.MIDDLE)
				crv = merge(merge(short_curve, fwd_curve, true,min_days_chop), long_curve, true, min_days_chop);	
			else
				crv = merge(merge(short_curve, long_curve, true,min_days_chop), fwd_curve, true,min_days_chop);	
		}
		else if(curve_prio_1 == curve_prio.MIDDLE){
			if(curve_prio_2 == curve_prio.SHORT)
				crv = merge(merge(short_curve, fwd_curve, false,min_days_chop), long_curve, true,min_days_chop);	
			else
				crv = merge(merge(long_curve, fwd_curve, false,min_days_chop), short_curve, true,min_days_chop);	
		}
		else {//prio1=3
			if(curve_prio_2 == curve_prio.SHORT)
				crv = merge(merge(short_curve, long_curve, false,min_days_chop), fwd_curve, true,min_days_chop);	
			else
				crv = merge(merge(fwd_curve, long_curve, false,min_days_chop), short_curve, true,min_days_chop);	
		}											
	}																
	else if(curve_ok[0] && curve_ok[2]){
		curve_prio prio = (curve_prio_1 == curve_prio.MIDDLE)? curve_prio_2 : curve_prio_1 ;
		crv = merge(short_curve, long_curve, prio == curve_prio.SHORT,min_days_chop);	
	}							
	else if(curve_ok[0] && curve_ok[1]){
		curve_prio prio = (curve_prio_1 == curve_prio.LONG)? curve_prio_2 : curve_prio_1 ;
		crv = merge(short_curve, fwd_curve, prio == curve_prio.SHORT,min_days_chop);
	}				
	else if(curve_ok[2] && curve_ok[1]){
		curve_prio prio = (curve_prio_1 == curve_prio.SHORT)? curve_prio_2 : curve_prio_1 ;
		crv = merge(fwd_curve, long_curve, prio==curve_prio.MIDDLE,min_days_chop);
	}
	else if(curve_ok[0])
		crv = short_curve;
	else if(curve_ok[1])
		crv = fwd_curve;
	else /*if(curve_ok[2])*/
		crv = long_curve;

	return crv;
}

module CORE_INT
{
	/*-----------------------------------------------------------------------
	  Name :  	blend_curve  
	  Notes :   middle curve is assumed to be a fra/fut curve	       
	  Changes:
	  			Date    		Who     Detail
	  			? 				MN     	Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	curve blend_curve(	curve option(nullable) 		short_curve,
						curve option(nullable)		fwd_curve, 
						curve option(nullable) 		long_curve,  								
						curve_prio					curve_prio_1,
						curve_prio					curve_prio_2,
						logical 					fwd_merge,
						logical 					fwd_no_overlap,
						logical 					merge_crv,							
						out vector(date)   knots,							
						integer						min_days_chop = 0)
	{
		if(merge_crv){
			resize(knots,0);
			return ..blend_curve_merge(short_curve,fwd_curve,long_curve, curve_prio_1,curve_prio_2,min_days_chop );
		}
		else
			return blend_curve(short_curve, fwd_curve, long_curve,curve_prio_1,curve_prio_2,fwd_merge, fwd_no_overlap,knots,min_days_chop);							
	}
	/*-----------------------------------------------------------------------
	  Name :  	blend_spot_ext
	  Notes :   (neither curve is a fra/fut curve)	       
	  Changes:
	  			Date    		Who     Detail
	  			? 				MN     	Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	void blend_spot_ext(	curve_name 	option(nullable) 	c_short, 
							curve_name 	option(nullable)	c,  
							date 							d,
							quote_side option(nullable)		qs_short,
							quote_side option(nullable)		qs,
							logical							short_prio,
							logical 						merge_crv,
							out curve option(nullable)		crv,									
							out number 	no_crv,
							out date		knot,
							integer							min_days_chop = 0)
	{
		logical f_c_short, f_c;
		curve c1, c2;
		no_crv = 0; 
		knot = null<date>;
	
		try {
			c1 = curve(c_short, d, qs_short);	
			QL_REQUIRE(!null(c1), 'null curve');	
			f_c_short = true;	
			no_crv++;
		}
		catch{
			f_c_short = false;	
		}
	
		try {		
			c2 = curve(c, d, qs);
			QL_REQUIRE(!null(c2), 'null curve');		
			f_c = true;	
			no_crv++;
		}
		catch{
			f_c = false	;
		}
	
		if(f_c_short && f_c){
			if(merge_crv)
				crv = merge(c1, c2, short_prio, min_days_chop);
			else 
				crv = CORE_INT.blend_spot_int(c1, c2,short_prio, knot,null<date>,min_days_chop);			
		}		
		else if(f_c_short)	
			crv = c1;
		else if(f_c)
			crv = c2;
		else 
			crv = null<curve>;	
	}	


	/*-----------------------------------------------------------------------
	  Name :  	blend_curve_ext
	  Notes :   middle curve is assumed to be a fra/fut curve	       
	  Changes:
	  			Date    		Who     Detail
	  			? 				MN     	Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	void blend_curve_ext(	curve_name 	option(nullable) 	c_short,
							curve_name 	option(nullable)	c_middle, 
							curve_name 	option(nullable) 	c_long,  
							date 							d,
							quote_side option(nullable)		qs_short,
							quote_side option(nullable)		qs_middle,
							quote_side option(nullable)		qs_long,
							curve_prio						prio1,
							curve_prio						prio2,
							logical 						merge_fra,
							logical 						no_overlap_fwd,
							logical 						merge_crv,								
							out curve 						crv,								
							out vector(logical)  			curve_ok,
							out number 						no_crv,
							out vector(date)  				knots,
							integer							min_days_chop = 0)
	{
		resize(curve_ok,3);		
		curve c1, c2, c3;  		
		no_crv = 0;
	
		try {
			c1 = curve(c_short, d, qs_short);			
			QL_REQUIRE(..check_curve(c1), 'null curve');		
			curve_ok[0] = true;
			no_crv++;	
		}
		catch{
			curve_ok[0] = false;	
		}
	
		try {		
			c2 = curve(c_middle, d, qs_middle);
			QL_REQUIRE(..check_curve(c2), 'null curve');				
			curve_ok[1] = true;
			no_crv++;
		}
		catch{
			curve_ok[1]= false	;
		}
		try {		
			c3 = curve(c_long, d, qs_long);
			QL_REQUIRE(..check_curve(c3), 'null curve');				
			curve_ok[2] = true;
			no_crv++;	
		}
		catch{
			curve_ok[2] = false	;
		}
	
		if(no_crv == 0)
			return;
	
		crv = CORE_INT.blend_curve(c1,c2,c3,prio1,prio2,merge_fra,no_overlap_fwd,merge_crv,knots,min_days_chop);						
	}	

	/*-----------------------------------------------------------------------
	  Name :  	blend_curve_ext
	  Notes :   	       
	  Changes:
	  			Date    		Who     Detail
	  			? 				MN     	Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	void blend_curve_ext(vector(curve_name) crv_name, 								
						date 				trade_date,
						vector(quote_side) 	qs,	
						curve_prio			prio1,
						curve_prio			prio2,
						logical 			merge_fra,
						logical 			no_overlap_fwd,						
						logical 			merge_crv,
						out curve option(nullable)			crv,								
						out vector(logical) option(nullable) curve_ok,								
						out vector(date)   	knots,
						integer				min_days_chop = 0)
	{
		number no_crv;
		
		QL_REQUIRE(v_size(crv_name) == 3, 'invalid vector size (crv_name)');	
	
		if(v_size(qs) == 1)		
			blend_curve_ext(crv_name[0], crv_name[1], crv_name[2], trade_date, qs[0],qs[0], qs[0], prio1, prio2,
							merge_fra,no_overlap_fwd, merge_crv, crv, curve_ok,no_crv, knots,min_days_chop);
		else if(v_size(qs) == 3)	
			blend_curve_ext(crv_name[0], crv_name[1], crv_name[2], trade_date, qs[0],qs[1], qs[2], prio1, prio2,
							merge_fra,no_overlap_fwd, merge_crv, crv, curve_ok,no_crv, knots,min_days_chop);
		else
			QL_FAIL('invalid vector size (qs)');					
	}

	//-----------------------------------------------------------------------------
	//   fwd_blend_3_curve
	//-----------------------------------------------------------------------------
	/*curve fwd_blend_3_curve(	curve 	option(nullable) depo_c,
								    curve 	option(nullable) fra_c,
								    curve 	option(nullable) swap_c,
									number 	prio1,
									number 	prio2,										
									number 	min_days_chop)
	option (category: 'FIWS/Yield Curve/001 Curve Blending')
	{	    
		logical c1_ok = check_curve(depo_c);
		logical c2_ok = check_curve(fra_c);
		logical c3_ok = check_curve(swap_c);		
	
		if (c1_ok && c2_ok && c3_ok) { 	
			vector(instrument) v_fra = fra_c.instruments();
			vector(instrument) v_depo = depo_c.instruments();
			vector(instrument) v_swap = swap_c.instruments();
			vector(date) fra_m = v_fra.maturity();
			vector(date) depo_m = v_depo.maturity(); 
			vector(date) swap_m = v_swap.maturity(); 
			if(prio1 == 1){
				if(prio2 == 2){//1,2,3
					add_inst(depo_m,fra_m,v_fra,min_days_chop,v_depo);															
					depo_m = v_depo.maturity(); 
					add_inst(depo_m,swap_m,v_swap,min_days_chop,v_depo);						
					return curve(v_depo);
				}
				else {//1,3,2
					add_inst(depo_m,swap_m,v_swap,min_days_chop,v_depo);
					depo_m = v_depo.maturity(); 
					add_inst(depo_m,fra_m,v_fra,min_days_chop,v_depo);										
					return curve(v_depo);
				}				
			}
			else if(prio1 == 2){
				if(prio2 == 1){//2,1,3	
					add_inst(fra_m,depo_m,v_depo,min_days_chop,v_fra);
					fra_m = v_fra.maturity();
					add_inst(fra_m,swap_m,v_swap,min_days_chop,v_fra);
					return curve(v_fra);
				}
				else {//2,3,1
					add_inst(fra_m,swap_m,v_swap,min_days_chop,v_fra);
					fra_m = v_fra.maturity();
					add_inst(fra_m,depo_m,v_depo,min_days_chop,v_fra);
					return curve(v_fra);
				}
			}
			else {				
				if(prio2 == 1){//3,1,2
					add_inst(swap_m,depo_m,v_depo,min_days_chop,v_swap);
					swap_m = v_swap.maturity();
					add_inst(swap_m,fra_m,v_fra,min_days_chop,v_swap);
					return curve(v_swap);	
				}
				else {//3,2,1
					add_inst(swap_m,fra_m,v_fra,min_days_chop,v_swap);
					swap_m = v_swap.maturity();
					add_inst(swap_m,depo_m,v_depo,min_days_chop,v_swap);
					return curve(v_swap);
				}
			}	
		}
		else if (c1_ok && c2_ok){
			vector(instrument) v_fra = fra_c.instruments();
			vector(instrument) v_depo = depo_c.instruments();
			vector(date) fra_m = v_fra.maturity();
			vector(date) depo_m = v_depo.maturity();		
			number prio = prio1 == 3 ? prio2 : prio1;	
					
			if(prio == 1){
				add_inst(depo_m,fra_m,v_fra,min_days_chop,v_depo);												
				return curve(v_depo);				
			}
			else {								
				add_inst(fra_m,depo_m,v_depo,min_days_chop,v_fra);
				return curve(v_fra);
			}					
		}
		else if (c1_ok && c3_ok){
			vector(instrument) v_swap = swap_c.instruments();
			vector(instrument) v_depo = depo_c.instruments();
			vector(date) swap_m = v_swap.maturity();
			vector(date) depo_m = v_depo.maturity();
	
			number prio = (prio1 == 2) ? prio2 : prio1;		
							
			if(prio == 1){
				add_inst(depo_m,swap_m,v_swap,min_days_chop,v_depo);				
				return curve(v_depo);	
			}
			else {
				add_inst(swap_m,depo_m,v_depo,min_days_chop,v_swap);				
				return curve(v_swap);
			}
	
		}
		else if (c2_ok && c3_ok){
			vector(instrument) v_fra = fra_c.instruments();
			vector(instrument) v_swap = swap_c.instruments();
			vector(date) fra_m = v_fra.maturity();
			vector(date) swap_m = v_swap.maturity();		
			number prio = (prio1 == 1) ? prio2 : prio1;
			
			if(prio == 2){
				add_inst(fra_m,swap_m,v_swap,min_days_chop,v_fra);				
				return curve(v_fra);
			}
			else {
				add_inst(swap_m,fra_m,v_fra,min_days_chop,v_swap);				
				return curve(v_swap);
			}	
		}
		else if (c1_ok)
			return depo_c;	
		else if (c2_ok)
			return fra_c;	
		else if (c3_ok)
			return swap_c;		
		else
			return null<curve>;
	    
	}
	*/


	/*-----------------------------------------------------------------------
	  Name :  	adj_long_curve
	  Notes :   when fra in the middle of the swap	(--> shorter swaps moved to short curve)       
	  Changes:
	  			Date    		Who     Detail
	  			? 				MN     	Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	void blend_adj_long(curve option(nullable) 	depo_c,
					    curve option(nullable) 	fra_c,
					    curve option(nullable) 	swap_c,
						curve_prio 				prio1,
						curve_prio 				prio2,										
						integer					min_days_chop,
						out curve option(nullable) 	depo_adj_c,								    
					    out curve option(nullable) 	swap_adj_c)
	{	    		
		//if(prio1==3 || (prio2 == 3 && prio1 == 1))
		//	return;
	
		logical c3_ok = ..check_curve(swap_c);
		logical c2_ok = ..check_curve(fra_c);
		if(!c3_ok || !c2_ok){
			depo_adj_c = depo_c;
			swap_adj_c = swap_c;
			return;
		}
	
		logical c1_ok 				= ..check_curve(depo_c);
		vector(instrument) v_swap 	= swap_c.instruments();
		vector(instrument) v_fra 	= fra_c.instruments();
		//vector(date) swap_m 		= v_swap.maturity();
		//vector(date) fra_m 			= v_fra.maturity();
	
		vector(instrument) move_inst;
		vector(integer) excl;
		date fra0mat 	= v_fra[0].maturity();
		integer i 		= 0;
		integer sw_size = v_size(v_swap);
		
		while (i<sw_size && v_swap[i].maturity() < fra0mat){
			push_back(move_inst,v_swap[i]);
			push_back(excl,i);
			i++;
		}
	
		if(i==0)	{
			depo_adj_c = depo_c;
			swap_adj_c = swap_c;
			return;
		}
	
		v_swap 			= v_exclude (v_swap, excl);
		swap_adj_c 		= curve(v_swap);
		curve move_curve = curve(move_inst);
	
		if (c1_ok) { 				
			//vector(instrument) v_depo 	= depo_c.instruments();						
			//vector(date) depo_m 		= v_depo.maturity(); 
			date knot, ext_knot;
			logical shortprio 	= prio1 == curve_prio.SHORT || (prio1 == curve_prio.MIDDLE && prio2 == curve_prio.SHORT);
			depo_adj_c			= CORE_INT.blend_spot_int(depo_c,move_curve,shortprio,knot,ext_knot,min_days_chop);																			
		}				
		else 
			depo_adj_c	= curve(move_inst);					    
	}
}

/*-----------------------------------------------------------------------
  Name :  	blend_curves_depo_swap  - legacy
  Notes :   prio on long end        
  Changes:
  			Date    		Who     Detail
  			? 				?     	Original code (moved from blend.ql) 
  ----------------------------------------------------------------------*/
curve blend_curves_depo_swap(curve depo_c, curve swap_c)
    option(category: 'Obsolete Functions/Yield Curve/Blending functions')
	option(com_name: "LEGACY_blend_curves_depo_swap")
{
    if (swap_c.empty())
		return depo_c;
	else if (depo_c.empty())
		return swap_c;
    else {
		depo_c = depo_c.chop_long(v_min(swap_c.instruments().maturity()) - 1);
		return merge(depo_c, swap_c);
    }
}
/*-----------------------------------------------------------------------
  Name :  	blend_curves_depo_fra  - legacy
  Notes :           
  Changes:
  			Date    		Who     Detail
  			? 				?     	Original code (moved from blend.ql) 
  ----------------------------------------------------------------------*/
curve blend_curves_depo_fra(curve depo_c, curve fra_c)
    option(category: 'Obsolete Functions/Yield Curve/Blending functions')
	option(com_name: "LEGACY_blend_curves_depo_fra")
{
    if (fra_c.empty())
		return depo_c;
	else if (depo_c.empty())
		return fra_c;
    else {
		depo_c = depo_c.chop_long_interp(v_min(fra_c.instruments().settle_date()));
	
		return merge(depo_c, fra_c);
    }
}
/*-----------------------------------------------------------------------
  Name :  	blend_curves_fra_prio  - legacy
  Notes :          
  Changes:
  			Date    		Who     Detail
  			? 				?     	Original code (moved from blend.ql) 
  ----------------------------------------------------------------------*/
curve blend_curves_fra_prio(curve depo_c,
			    			curve fra_c,
			    			curve swap_c,
							logical incl_interp = true)
    option(category: 'Obsolete Functions/Yield Curve/Blending functions')
	option(com_name: "LEGACY_blend_curves_fra_prio")
{
    if (fra_c.empty())
		return blend_curves_depo_swap(depo_c, swap_c);
    else {
		date	chop_d;
		curve	short_c;
	
		chop_d  = v_min(fra_c.instruments().settle_date());
		if(incl_interp)
			depo_c	= depo_c.chop_long_interp(chop_d);
		else
			depo_c	= depo_c.chop_long(chop_d);
	
		short_c	= merge(depo_c, fra_c);
	
		chop_d  = v_max(short_c.instruments().maturity()) + 1;
		swap_c	= swap_c.chop_short(chop_d);
	
		return merge(short_c, swap_c);
    }
}
/*-----------------------------------------------------------------------
  Name :  	blend_curves_swap_prio  - legacy
  Notes :          
  Changes:
  			Date    		Who     Detail
  			? 				?     	Original code (moved from blend.ql) 
  ----------------------------------------------------------------------*/
curve blend_curves_swap_prio(curve depo_c,
						     curve fra_c,
						     curve swap_c)
    option(category: 'Obsolete Functions/Yield Curve/Blending functions')
	option(com_name: "LEGACY_blend_curves_swap_prio")
{
    if (swap_c.empty())
		return blend_curves_depo_fra(depo_c, fra_c);
    else {
		fra_c = fra_c.chop_long(v_min(swap_c.instruments().maturity()) - 1);
		if (fra_c.empty())
		    return blend_curves_depo_swap(depo_c, swap_c);
		else {
		    date	chop_d;
	
		    chop_d = v_min(fra_c.instruments().settle_date());
		    depo_c = depo_c.chop_long_interp(chop_d);
	
		    return merge(depo_c, merge(fra_c, swap_c));
		}
    }
}
/*-----------------------------------------------------------------------
  Name :  	blend_curves_fra_prio  - legacy
  Notes :          
  Changes:
  			Date    		Who     Detail
  			? 				?     	Original code (moved from blend.ql) 
  ----------------------------------------------------------------------*/
curve blend_curves_depo_fra_prio(curve depo_c,
							     curve fra_c,
							     curve swap_c)
    option(category: 'Obsolete Functions/Yield Curve/Blending functions')
	option(com_name: "LEGACY_blend_curves_depo_fra_prio")
{
    if (depo_c.empty())
		return blend_curves_depo_swap(fra_c, swap_c);
    else {
		if (fra_c.empty())
		    return blend_curves_depo_swap(depo_c, swap_c);
		else {
			fra_c = fra_c.chop_short(v_max(depo_c.instruments().maturity()) + 1);
		
		    date	chop_d;
			curve	short_c;
	
			short_c	= merge(depo_c, fra_c);
	
			chop_d  = v_max(short_c.instruments().maturity()) + 1;
			swap_c	= swap_c.chop_short(chop_d);
	
		    return merge(short_c, swap_c);
		}
    }
}

module CORE_INT
{
	/*-----------------------------------------------------------------------
	  Name :  	blender_adj_long_curve
	  Notes :   long_crv also as short (if firstfra > swap)       
	  Changes:
				Date    		Who     Detail  			
	  ----------------------------------------------------------------------*/
	curve blender_adj_long_curve(curve option(nullable) short_crv,
								curve option(nullable) middle_crv,
								curve option(nullable) long_crv,
								curve_prio 		prio1,
								curve_prio		prio2,
								logical			merge_middle,
								logical			merge_crv,
								logical 		no_overlap_fwd,
								integer			blend_buf_days)
	{
		curve short_adj_c;
		curve long_adj_c;
		if(prio1 == curve_prio.MIDDLE || (prio2 == curve_prio.MIDDLE && prio1 == curve_prio.SHORT))
			blend_adj_long(	short_crv, middle_crv , long_crv, prio1, 
							prio2,blend_buf_days,short_adj_c,long_adj_c);//adjust when fra in the middle of the swap
		else {
			short_adj_c = short_crv;
			long_adj_c 	= long_crv;
		}
		vector(date) knots;
		return blend_curve(	short_adj_c, middle_crv , long_adj_c, prio1, prio2, merge_middle,
							no_overlap_fwd, merge_crv, knots, blend_buf_days);
	}
}

/*-----------------------------------------------------------------------
  Name :    class curve_blend_parm
  Notes :   just a container for curve blending params        
  Changes:
  			Date    		Who     Detail
  			2-Aug-2016 		MN       
  ----------------------------------------------------------------------*/
integer dflt_blend_buf_days = 5;

class curve_blend_parm option (category: 'Yield Curve/001 Curve Blending')
{
	public:	
   	//protected:
		curve_blend_parm(curve_prio,curve_prio,logical,logical,logical,integer);
		curve_blend_parm(curve_prio,curve_prio);		
		curve_blend_parm(curve_blend_parm);
		
					
		curve_prio					prio1_;
		curve_prio					prio2_;
		logical 					no_overlap_fwd_;
		logical 					merge_middle_;
		logical 					merge_crv_;
		integer						blend_buf_days_;									
};

curve_blend_parm.curve_blend_parm(	curve_prio	prio1,
									curve_prio	prio2)
									:prio1_(prio1), prio2_(prio2),  no_overlap_fwd_(false), merge_middle_(false), 
								 	merge_crv_(false), blend_buf_days_(dflt_blend_buf_days)
{
}
curve_blend_parm.curve_blend_parm(	curve_prio	prio1,
									curve_prio	prio2,
									logical 	no_overlap_fwd,
									logical 	merge_fra,
									logical 	merge_crv,
									integer		blend_buf_days)
									:prio1_(prio1), prio2_(prio2),  no_overlap_fwd_(no_overlap_fwd), merge_middle_(merge_fra), 
									 merge_crv_(merge_crv), blend_buf_days_(blend_buf_days < 0 ? dflt_blend_buf_days : blend_buf_days)
{
}
curve_blend_parm.curve_blend_parm(	curve_blend_parm o)
									:prio1_(o.prio1_), prio2_(o.prio2_),  no_overlap_fwd_(o.no_overlap_fwd_), merge_middle_(o.merge_middle_), 
									 merge_crv_(o.merge_crv_), blend_buf_days_(o.blend_buf_days_)
{
}

curve_blend_parm curve_blend_parm(	curve_prio	prio1,
									curve_prio	prio2)									
{
	return new curve_blend_parm(prio1,prio2);
}

curve_blend_parm curve_blend_parm(	curve_prio	prio1,
									curve_prio	prio2,
									logical 	no_overlap_fwd,
									logical 	merge_middle,
									logical 	merge_crv,
									integer		blend_buf_days)
{
	return new curve_blend_parm(prio1,prio2,no_overlap_fwd,merge_middle,merge_crv,blend_buf_days);
}

module WSP_UTIL
{
	class curve_data_disc
	{
	public:
		curve_data_disc(curve_name option(nullable),curve_name option(nullable),curve_name option(nullable));
		
		curve_blend_parm 	blend_parm();
		string 				ccy_crv();
		date 				trade_date_crv();
		void 				update_short_curve(curve option(nullable) , curve option(nullable) );
		void 				update_middle_curve(curve option(nullable), curve option(nullable));
		void 				update_long_curve(curve option(nullable), curve option(nullable));
		void 				update_blend_curve(curve );
		void 				update_df(disc_func);
		void 				update_df_sprd(disc_func_spr );
		
		curve_name	short_name_;
		curve_name 	middle_name_;
		curve_name 	long_name_;

		curve 		short_crv_;
		curve 		middle_crv_;
		curve 		long_crv_;
		curve		crv_;

		curve 		short_orig_crv_;
		curve 		middle_orig_crv_;
		curve 		long_orig_crv_;
		
		logical		no_extrap_;
		curve_prio  prio1_;
		curve_prio  prio2_;
		curve_merge merge_;
		logical		no_overlap_middle_;
		integer 	blend_buf_days_; 
		disc_func	df_;
		disc_func	df_spr_;
	};

	curve_data_disc.curve_data_disc(curve_name option(nullable) sh,
									curve_name option(nullable) mi,
									curve_name option(nullable) lo)
				:short_name_(sh), middle_name_(mi), long_name_(lo),
				 short_crv_(null<curve>), middle_crv_(null<curve>),long_crv_(null<curve>),
				 crv_(null<curve>),short_orig_crv_(null<curve>),
				 middle_orig_crv_(null<curve>),long_orig_crv_(null<curve>),
				 no_extrap_(true),
				 prio1_(curve_prio.SHORT),prio2_(curve_prio.MIDDLE),merge_(curve_merge.NONE),
				 no_overlap_middle_(false),blend_buf_days_(dflt_blend_buf_days),
				 df_(null<disc_func>), df_spr_(null<disc_func_spr>)
	{}

	void curve_data_disc.update_short_curve(curve option(nullable) crv_incl, curve option(nullable) crv_orig)
	{
		short_crv_ 		= crv_incl;
		short_orig_crv_ = crv_orig;
		
	}
	void curve_data_disc.update_middle_curve(curve option(nullable) crv_incl, curve option(nullable) crv_orig)
	{
		middle_crv_ 		= crv_incl;
		middle_orig_crv_ 	= crv_orig;
		
	}
	void curve_data_disc.update_long_curve(curve option(nullable) crv_incl, curve option(nullable) crv_orig)
	{
		long_crv_ 		= crv_incl;
		long_orig_crv_ 	= crv_orig;
		
	}
	void curve_data_disc.update_blend_curve(curve crv)
	{
		crv_ = crv;
	}
	void curve_data_disc.update_df(disc_func df)
	{
		df_ = df;
	}
	void curve_data_disc.update_df_sprd(disc_func_spr df_spr)
	{
		df_spr_ = df_spr;
	}
	
	curve_blend_parm curve_data_disc.blend_parm()
	{
		logical mc,mm;
		curve_merge_parm(merge_, mc,mm);
		return curve_blend_parm(prio1_,prio2_,no_overlap_middle_, mm,mc,blend_buf_days_);
	}

	string curve_data_disc.ccy_crv()
	{
		if(check_curve(short_crv_)){
			return short_crv_.instruments()[0].currency();
		}
		else if(check_curve(middle_crv_)){
			return middle_crv_.instruments()[0].currency();
		}
		else if(check_curve(long_crv_)){
			return long_crv_.instruments()[0].currency();
		}
		else {
			return null<string>;
		}
	}

	date curve_data_disc.trade_date_crv()
	{
		if(check_curve(short_crv_)){
			return short_crv_.instruments()[0].trade_date();
		}
		else if(check_curve(middle_crv_)){
			return middle_crv_.instruments()[0].trade_date();
		}
		else if(check_curve(long_crv_)){
			return long_crv_.instruments()[0].trade_date();
		}
		else {
			return null<date>;
		}
	}
	
	class curve_data_fwd: public curve_data_disc		
	{
		public:
		curve_data_fwd(curve_name option(nullable),curve_name option(nullable),curve_name option(nullable),
					   curve_data_disc option(fixedref) , fwd_func_tenor option(nullable));
		curve_data_disc option(fixedref) disc_;
		fwd_func 			f_fwd_;
		fwd_func_tenor		fwd_tenor_;//if null it will be extracted from the curves (1. fra, 2. swap, 3. short or fixing (if 1 instr in curve))
	};

	curve_data_fwd.curve_data_fwd(curve_name option(nullable) sh,
								  curve_name option(nullable) mi,
								  curve_name option(nullable) lo,
								  curve_data_disc option(fixedref) disc,
								  fwd_func_tenor option(nullable) fwd_tenor)
				:curve_data_disc(sh,mi,lo), disc_(disc),f_fwd_(null<fwd_func>), fwd_tenor_(fwd_tenor)
	{}
}

module WSP_UTIL
{

	//-----------------------------------------------------------------------------
	//  curve_init_b
	//-----------------------------------------------------------------------------
	logical curve_init_b(	date 				trade_date,
							vector(curve_name) 	crv_name, 									
							vector(quote_side)	qs,
							curve_prio			prio1,
							curve_prio			prio2,
							logical				merge_fra,
							logical 			no_overlap_fwd,											
							vector(instr_type) 	allowed_instr_types,
							out curve			crv,
							out vector(instrument) inst,																	
							out vector(logical) curve_ok,														
							out string 			err_msg,
							logical				merge_all)
	{
		try {		
			
			vector(date) knots;	

			CORE_INT.blend_curve_ext(crv_name,trade_date,qs,prio1, prio2,merge_fra,no_overlap_fwd, merge_all,crv, curve_ok, knots);

			if(null(crv))
				throw(E_INVALID_ARG, "invalid curve");

			if(!init_curve(crv,allowed_instr_types, inst, err_msg))
				throw(E_INVALID_ARG, err_msg);
			
			return true;
		}
		catch {
			err_msg = err.message();
			return false;
		}
	}
	
	//-----------------------------------------------------------------------------
	//  curve_init
	// FIX_ME: uses yield and set_yield so need to fix for non-db swaps
	//-----------------------------------------------------------------------------
	logical curve_init(	date 				trade_date,
						vector(curve_name) 	crv_name, 									
						vector(quote_side)	qs,
						curve_prio			prio1,
						curve_prio			prio2,
						logical				merge_fra,
						logical 			no_overlap_fwd,											
						vector(instr_type) 	allowed_instr_types,
						out curve			crv,
						out vector(instrument) inst,																													
						out vector(string) 	crv_i_name,
						out vector(number) 	crv_i_yld,
						out vector(number) 	crv_i_user_yld,									
						out vector(logical) curve_ok,														
						out string 			err_msg,
						logical				merge_all)

	{
		try {	

			CORE_INT.set_to_null(crv_i_yld);

			if(!curve_init_b(trade_date,crv_name,qs,prio1,prio2,merge_fra,no_overlap_fwd,allowed_instr_types,crv,inst,curve_ok,err_msg,merge_all))
				throw(E_INVALID_ARG, err_msg);
			
			integer size = v_size(inst);
		
			if(v_size(crv_i_user_yld) != size){
				resize(crv_i_user_yld, size);
				crv_i_user_yld = null<number>;	
			}
					
			crv_i_name = inst.name();

			error_info ee = new error_info(true,false);
			crv_i_yld = inst.yield(false, ee)*100;
					
			if(!set_curve_yld(crv,crv_i_user_yld,crv))
				throw(E_INVALID_ARG, "invalid curve");
	
			inst = crv.instruments();

			return true;
		}
		catch {
			err_msg = err.message();
			return false;
		}
	}

	//-------------------------------------------------------
	// input_curve
	//-------------------------------------------------------
	logical input_curve(vector(WSP_UTIL.input_curve_data_ext) c_data,
						curve_prio 			prio1,
						curve_prio 			prio2,
						vector(instr_type) 	allowed_instr_types,
						out curve 			blend_crv,
						out string 			err_msg)
	{
		try {
			integer size = v_size(c_data);
			if(size != 3 )
				throw(E_INVALID_ARG, "input_curve_data");

			vector(curve) crv[3];
			
			for(integer i=0; i< size; i++){
				if(null(c_data[i])){
					crv[i] = null<curve>;
					continue;
				}
				
				//c_data[i].init();
				if(!c_data[i].is_init())
					throw(E_INVALID_ARG, c_data[i].err_msg());

				crv[i] = c_data[i].crv_incl_;
			}
			
			vector(date) knots;
			blend_crv = CORE_INT.blend_curve(crv[0],crv[1],crv[2],prio1,prio2,
											false,false,false,knots,dflt_blend_buf_days);

			vector(instrument) i_v;
			logical ok = WSP_UTIL.init_curve(blend_crv,allowed_instr_types, i_v, err_msg);
			if(!ok)
				throw(E_INVALID_ARG, err_msg);
			
			return true;
		}
		catch {
			err_msg = err.message();
			return false;
		}
	}
}