/*	
	Core date functions
	Developer: Algorithmica Research, Magnus Nyström			
*/
option(null: hard);	
/*-----------------------------------------------------------------------
  Name :    is_imm
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
logical is_imm(date d) 
option(category: 'Basic Types and Functions/Date Manipulation')
{
	return d == date_code("IMM0").apply_fwd(d);

}

/*-----------------------------------------------------------------------
  Name :    move_imm
  Notes : 	date_code_method=true: 	This tries to mimic the behaviour of the way the date_code's 
									work when created from "IMM0", "IMM1", etc. One difference is  
									that in this function the offset argument can be positive or 
									negative but date_codes's move dates by applying the 
									apply_back() and apply_fwd() members. Hence, since offset=0 
									does not imply a direction we assume in this that offset = 0
									uses apply_fwd() --> date_code("IMM0").apply_back() cannot be obtained. 
									An imm-date is always returned regardless of the input date
			date_code_method=false:	If offset is a positive (negative) number a forward (backward) imm date is returned. 
									If offset = 0 the input date is returned i.e. normally not an imm date		          
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code
  ----------------------------------------------------------------------*/
date move_imm(	date 		d, 
				integer 	offs , 
				logical date_code_method = false) 
option(category: 'Basic Types and Functions/Date Manipulation')
{
	//number offs = trunc(offset);

	if(date_code_method) {//note offset=0 means we're using apply_fwd 
		logical back;
		if(offs < 0){				
			back = true;
			offs = -offs;						
		}
		else {		
			back = false;		
		}

		date_code dc = date_code(strcat("IMM", str(offs)));	
		return back ? dc.apply_back(d) : dc.apply_fwd(d);
	}
	else { //offset=0 means we return the input i.e. no imm date return
		if(offs == 0) return d;
		
		logical back;
		if(offs < 0){				
			back = true;
			offs = -offs;						
		}
		else {		
			back = false;		
		}
	
		offs--;
		if (is_imm(d))
			offs++;	
				
		date_code dc = date_code(strcat("IMM", str(offs)));	
		return back ? dc.apply_back(d) : dc.apply_fwd(d);
	}
}

/*-----------------------------------------------------------------------
  Name :    next_imm
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
date next_imm(date d) 
option(category: 'Basic Types and Functions/Date Manipulation')
{
	return  move_imm(d, 1, false);
}
/*-----------------------------------------------------------------------
  Name :    prev_imm
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
date prev_imm(date d) 
option(category: 'Basic Types and Functions/Date Manipulation')
{
	return  move_imm(d, -1, false);
}
 	
class enum day_of_week
option(category: 'Basic Types and Functions/Date Manipulation')
{
	MONDAY 		option(str: "Monday") ,
	TUESDAY 	option(str: "Tuesday"),
	WEDNESDAY 	option(str: "Wednesday"),
	THURSDAY 	option(str: "Thursday"),
	FRIDAY 		option(str: "Friday"),
	SATURDAY 	option(str: "Saturday"),
	SUNDAY  	option(str: "Sunday")
};

//-------------------------------------------------------
//  day_of_week_list
//-------------------------------------------------------
vector(string) day_of_week_list() 
option(category: 'Basic Types and Functions/Date Manipulation')
{
	vector(day_of_week) e;
	expand_enum(e) ;
	return string(e);
}

//-------------------------------------------------------
//  day_of_week_map
//-------------------------------------------------------
day_of_week day_of_week_map(string s)
option(category: 'Basic Types and Functions/Date Manipulation')
{
	day_of_week e;
	str_to_enum(s,e);
	return e;
}

//-------------------------------------------------------
//  day_of_week_from_int
//-------------------------------------------------------
day_of_week day_of_week_from_int(integer i)
option (category: 'Basic Types and Functions/Date Manipulation')
{
	day_of_week e;

	switch(i)
	{
		case 1 : e = day_of_week.MONDAY ; 	break;
		case 2 : e = day_of_week.TUESDAY ;	break;
		case 3 : e = day_of_week.WEDNESDAY ;break;
		case 4 : e = day_of_week.THURSDAY ;	break;
		case 5 : e = day_of_week.FRIDAY ;	break;
		case 6 : e = day_of_week.SATURDAY ;	break;
		case 7 : e = day_of_week.SUNDAY  ;	break;		
		default : e = null<day_of_week>;
	}
		
	if (null(e))
		throw(E_INVALID_ARG, strcat(["No day_of_week found matching the integer ", string(i), "."]));
	return e;
}

//-------------------------------------------------------
//  day_of_week_to_int
//-------------------------------------------------------	 
integer day_of_week_to_int(day_of_week e)
option (category: 'Basic Types and Functions/Date Manipulation')
{
	integer i = 0;
	while (day_of_week_from_int(++i) != e) {
	}
	return i;
}
		

/*-----------------------------------------------------------------------
  Name :    weekday
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		?      	Original code (moved from util31.ql)
  ----------------------------------------------------------------------*/
integer weekday(date d)
option(category: 'Basic Types and Functions/Date Manipulation')
{	
	date monday = #1753-01-01;
	return integer(fmod((d - monday),7) + 1);
}

integer weekday(day_of_week d)
option(category: 'Basic Types and Functions/Date Manipulation')
{	
	return day_of_week_to_int(d);
}

day_of_week weekday_e(date d)	
option (category: 'Basic Types and Functions/Date Manipulation')
{	
	return day_of_week_from_int(weekday(d));	
}

day_of_week weekday_e(integer i)	
option (category: 'Basic Types and Functions/Date Manipulation')
{	
	return day_of_week_from_int(i);	
}

string weekday_s(date d)	
option (category: 'Basic Types and Functions/Date Manipulation')
{				 	
	return string(weekday_e(d))	;
}

string weekday_s(day_of_week d)	
option (category: 'Basic Types and Functions/Date Manipulation')
option(com_name: "weekday_s_w")
{			
 	return string(d);	
}

/*-----------------------------------------------------------------------
  Name :    nth_weekday
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code (algorithm from NR)
  ----------------------------------------------------------------------*/
date nth_weekday(	date 		d,
					integer 	n, 
					integer 	weekday_num)
option(category: 'Basic Types and Functions/Date Manipulation')
{
	QL_FAIL_COND((weekday_num <= 0)|| (weekday_num > 7),'invalid weekday');
	QL_FAIL_COND(n<=0, 'invalid number n');

	integer y = year(d);
	integer m = month(d);

	integer first = integer(weekday(date(y, m, 1)));
    integer skip = n - (weekday_num >= first ? 1 : 0);
    return date(y, m, 1 + weekday_num - first + skip * 7);
}
/*-----------------------------------------------------------------------
  Name :    nth_weekday
  Notes :   nth weekday in the month        
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code 
  ----------------------------------------------------------------------*/
date nth_weekday(	date 			d,
					integer 		n, 
					day_of_week 	weekday_e)
option(category: 'Basic Types and Functions/Date Manipulation')
option(com_name: "nth_weekday_e")
{	
	return nth_weekday(	d,n,day_of_week_to_int(weekday_e));
}

/*-----------------------------------------------------------------------
  Name :    third_wed
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code 
  ----------------------------------------------------------------------*/
date third_wed(date d) 
option (category: 'Basic Types and Functions/Date Manipulation')
{	    
    return nth_weekday(d, 3, day_of_week.WEDNESDAY);
}

date third_wed(date d, calendar cal, bd_convention bd)
option (category: 'Basic Types and Functions/Date Manipulation')
 {
	return cal.adjust_date(third_wed(d),bd);
 }

/*-----------------------------------------------------------------------
  Name :    move_third_wed
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code 
  ----------------------------------------------------------------------*/
date  move_third_wed(	date 	d, 
				   		integer	n)
option (category: 'Basic Types and Functions/Date Manipulation')
{
    if(n == 0)
		return (d) ;

    date wed_3 = nth_weekday(d, 3, day_of_week.WEDNESDAY);

    if(n > 0 && d < wed_3) 
		n--;
    else if(n < 0 && d > wed_3)
		n++;

    if (n==0)
		return wed_3;
    else {			 
		wed_3 = add_months (wed_3, n);
		return nth_weekday(wed_3, 3, day_of_week.WEDNESDAY);
    }
}
/*-----------------------------------------------------------------------
  Name :    is_third_wed
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code 
  ----------------------------------------------------------------------*/
logical is_third_wed(date d) 
option(category: 'Basic Types and Functions/Date Manipulation')
{
	return d == third_wed(d);
}

logical is_third_wed(date d, calendar cal, integer tol = 1)
{
	date thw = third_wed(d);
	if(d == thw)
		return true;
	else if(d >= cal.move_bus_days(thw, -tol) &&  d <= cal.move_bus_days(thw, tol))
		return true;

	return false;
}
/*-----------------------------------------------------------------------
  Name :    next_third_wed
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code 
  ----------------------------------------------------------------------*/	
date next_third_wed(date d) 
option (category: 'Basic Types and Functions/Date Manipulation')
{	 
	return move_third_wed(d, 1) ;   
}
/*-----------------------------------------------------------------------
  Name :    prev_third_wed
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code 
  ----------------------------------------------------------------------*/
date prev_third_wed(date d) 
option (category: 'Basic Types and Functions/Date Manipulation')
{	 
	return move_third_wed(d, -1) ;   
}

/*-----------------------------------------------------------------------
  Name :    third_fri
  Notes :           
  Changes:
  			Date    		Who     Detail
  			12-Oct-2019 	Robert       
  ----------------------------------------------------------------------*/
date third_fri(date d) 
option (category: 'Basic Types and Functions/Date Manipulation')
{	    
    return nth_weekday(d, 3, day_of_week.FRIDAY);
}

date third_fri(date d, calendar cal, bd_convention bd)
option (category: 'Basic Types and Functions/Date Manipulation')
 {
	return cal.adjust_date(third_fri(d),bd);
 }

/*-----------------------------------------------------------------------
  Name :    is_third_fri
  Notes :           
  Changes:
  			Date    		Who     Detail
  			12-Oct-2019 	Robert   
  ----------------------------------------------------------------------*/
logical is_third_fri(date d) 
option(category: 'Basic Types and Functions/Date Manipulation')
{
	return d == third_fri(d);
}

logical is_third_fri(date d, calendar cal, integer tol = 1)
{
	date thf = third_fri(d);
	if(d == thf)
		return true;
	else if(d >= cal.move_bus_days(thf, -tol) &&  d <= cal.move_bus_days(thf, tol))
		return true;

	return false;
}

/*-----------------------------------------------------------------------
  Name :    move_bus_days
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		?      	Original code (moved from util31.ql)
			21-Apr-2021		MN		Added start_date_bd argument
  ----------------------------------------------------------------------*/
date move_bus_days(	date 		d,
					integer 	n_days,
					calendar 	cal,
					bd_convention option(nullable) start_date_bd = null<bd_convention>)
option(category: 'Basic Types and Functions/Date Manipulation')
{
	if(null(start_date_bd)) {
		return cal.move_bus_days(d,n_days);
	}
	else if(start_date_bd == BD_NONE){
		if(n_days < 0) {//before: BD_PREVIOUS
			date _d = cal.adjust_date(d, BD_FOLLOWING);
			return cal.move_bus_days(_d,n_days);
		}
		else if(n_days > 0){//NO CHANGE from current
			date _d = cal.adjust_date (d, BD_PREVIOUS);
			return cal.move_bus_days(_d,n_days);
		}
		else {//before: no adjustment
			return d;//cal.adjust_date(d, BD_FOLLOWING);
		}
	}
	else {
		date _d = cal.adjust_date(d, start_date_bd);
		return cal.move_bus_days(_d,n_days);
	}
}

/*-----------------------------------------------------------------------
  Name :    set_to_end_of_month
  Notes :           
  Changes:
  			Date    		Who     Detail
  			13-Jul-2016 		?      	Original code (moved from util31.ql)
  ----------------------------------------------------------------------*/
void set_to_end_of_month(	calendar option(nullable) cal, 
							out date dt)
option (category: 'Basic Types and Functions/Date Manipulation')
{		
	integer d_i_m 	= days_in_month(dt);    	       
	dt 				= date(year(dt), month(dt), d_i_m); 
	if(!null(cal)) cal.adjust_date(dt, BD_PREVIOUS);			
}

/*-----------------------------------------------------------------------
  Name :    is_end_of_month
  Notes :           
  Changes:
  			Date    		Who     Detail
  					 		MN      Original code
  ----------------------------------------------------------------------*/
logical is_end_of_month(	calendar option(nullable) cal, 
							date dt)
option (category: 'Basic Types and Functions/Date Manipulation')
{
	if(!null(cal)) dt = cal.adjust_date(dt, BD_PREVIOUS);	
	date dt_end = dt; 
	set_to_end_of_month(cal, dt_end);
	return dt == dt_end;
}
/*-----------------------------------------------------------------------
  Name :    move_month
  Notes :   bond_style        
  Changes:
  			Date    		Who     Detail
  			13-Jul-2016 	?   	Original code (moved from util31.ql)
			2-Aug-2016		MN		Added bond_style argument and changed handling in the case eom 
									is true (only allow bonds to move to next month when 
									bus-day method is applied and we have eom = true )
  ----------------------------------------------------------------------*/
date move_month(date 		d,
				number 		m,
				logical 	eom,
				calendar 	option(nullable) cal,
				bd_convention bd /*= BD_MOD_FOLLOWING*/,
				logical 	bond_style /*= true*/)
option(category: 'Basic Types and Functions/Date Manipulation')
{
	if(m < 0)
		return null<date>;
	else if(m == 0)
		return d;

	logical is_eom = bond_style ? is_end_of_month(null<calendar>, d) : is_end_of_month(cal, d); 
	eom = eom && is_eom;
	
	string mm = strcat(str(m),'M');
	date t = date_code(mm).apply_fwd(d);
	if(eom) {
		set_to_end_of_month(null<calendar >, t);
		if(bd != BD_NONE) {			
			if(bond_style){ 			
				if(!null(cal)) t = cal.adjust_date (t, bd);	//we can move to next month	
			}		 
			else {
				if(!null(cal)) t = cal.adjust_date (t, BD_PREVIOUS);//we cannot move to next month
			}
		}	
	}
	else {
		if(!null(cal)) t = cal.adjust_date (t, bd);
	}
	return t;
}

//-------------------------------------------------------
// months_between
// note: this function just takes into account the "month value" i.e.
// if we have start = #2021-12-31 and end = #2022-01-01 we get 1
//-------------------------------------------------------
module CORE_INT{
	
	integer months_between(date start, date end)
	{
		QL_REQUIRE( end >= start,"invalid start/end dates");
		integer diff_days   = end - start;
	   
		date start_date   = date(year(end), month(end),1);
		date end_date	  = start_date;
		set_to_end_of_month(null<calendar>,end_date);

		integer diff_months = integer(trunc (diff_days / 31)) ;
		date d    = start;
		d = add_months(d,diff_months);

		while (!(start_date <= d && d <= end_date)) {
			d = add_months(d,1);
			diff_months++;
		}
		return diff_months;
	}

	//----------------------------------------------------------------------
	// dates_1v_to_2v                                                          
	//---------------------------------------------------------------------- 
	void dates_1v_to_2v(vector(date) 		dates,										
						out vector(date)   	beg,
						out vector(date)   	end)
	{    	    
		resize(beg,0);
		resize(end,0);
		for(integer i = 0; i < v_size(dates)-1; i++){
			push_back(beg, dates[i]);
			push_back(end, dates[i+1]);
		}   
	    return;
	}
	//----------------------------------------------------------------------
	// dates_2v_to_1v                                                          
	//---------------------------------------------------------------------- 
	void dates_2v_to_1v(vector(date) 	beg,
						vector(date) 	end,
						out vector(date)   dates)
	{    	    
		dates = ..clone_vector(beg);
		push_back(dates , end[v_size(end)-1]);		   
	    return;
	}
}