/*	
	util
	Developer: Algorithmica Research, Magnus N
	Comment: cf_min: necessary because we need a curve class before instrument classes are loaded and
			the public curve class requires instrument
			Note that the curve class is not available here
							
*/

//-------------------------------------------------------
// extrap_flat_spot 
//-------------------------------------------------------
disc_func extrap_flat_spot(	disc_func 	df,
							number 		t,
							rate_type 	rt)
option(category: "Yield Curve/Discount Function Manipulation")
{
	number df_t 		= df.disc(t);
	disc_func df_xpol 	= disc_func_interp([t ], [df_t ], ip_step(), rt);
	return disc_func_splice(df, df_xpol, t);
}

//-------------------------------------------------------
// extrap_flat_inst_fwd 
//-------------------------------------------------------
disc_func extrap_flat_inst_fwd(disc_func 	df, 
							   number 		t)
	option(category: "Yield CurveDiscount Function Manipulation")
{
	number e 			= 1.0/365.0;
	number df_t 		= df.disc(t);
	number df_te 		= df.disc(t-e);
	disc_func df_xpol 	= disc_func_interp([t-e, t ], [ df_te, df_t ], ip_linear(null<number>,null<number>),RT_LOG_DF);
	return disc_func_splice(df, df_xpol, t);
}

module CORE_INT
{
	//-------------------------------------------------------
	// on_fwd_to_df
	//-------------------------------------------------------
	disc_func on_fwd_to_df(	date				d,
							date				mat,	
							fwd_func			fwd,						
							day_count_method 	dc,
							calendar 			cal,
							logical 			allow_extrap_flat,
							out vector(date) 	on_r_settle,
							out vector(number) 	on_r)
	{
		QL_REQUIRE(d < mat, "invalid maturity");

		date s 	= d;
		date se = cal.move_bus_days(s,1);
		
		vector(number) per_v;
		vector(number) df_v;
		resize(on_r,0);
		resize(on_r_settle,0);
		
		number df_r = 1;
		while (se  <= mat|| (se > mat && s < mat)){
							
			number per_e = period(d,se,DC_ACT_365);	
			push_back(per_v, per_e);

			number per_s = period(d,s,DC_ACT_365);
			push_back(on_r_settle, s);
			number r 	= fwd.fwd(per_s);
			push_back(on_r, r);
			number df_f = 1.0/(1.0 + r * period(s,se,dc));
			df_r 		*= df_f;
			push_back(df_v, df_r);
			s 			= se;
			se 			= cal.move_bus_days(s,1);
		}
		disc_func df = disc_func_interp(per_v,df_v, ip_linear(), RT_LOG_DF);

		if(allow_extrap_flat) {
			number t =  (mat - d)/365;
			return extrap_flat_inst_fwd(df,t);
		}
		else {
			return disc_func_restrict (df, 0, per_v[v_size(per_v)-1]);
		}

	}
	//-------------------------------------------------------
	// on_fwd_to_df
	//-------------------------------------------------------
	disc_func on_fwd_to_df(	date				d,
							date				mat,	
							fwd_func			fwd,						
							day_count_method 	dc,
							calendar 			cal,
							logical 			allow_extrap_flat )
	{
		vector(date) 	on_r_settle;
		vector(number) 	on_r;
		return on_fwd_to_df(d,mat,fwd,dc,cal,allow_extrap_flat,on_r_settle,on_r);
	}
}

//-------------------------------------------------------
// extrap_none 
//-------------------------------------------------------
disc_func extrap_none(disc_func df,
					 number t)
option(category: "Yield Curve/Discount Function Manipulation")
{
	number df_t 		= df.disc(t);
	disc_func df_xpol 	= disc_func_interp([ t ], [ df_t ], ip_step(IP_NO_XPOL_R));
	return disc_func_splice(df, df_xpol, t);
}
//-------------------------------------------------------
// extrap_none 
//-------------------------------------------------------
fwd_func extrap_none(fwd_func fwd,
					 number t)
option(category: "Yield Curve/Discount Function Manipulation")
{
	fwd_func fwd_xpol = fwd_func_interp([ t ], [ 0.01 ], ip_step(IP_NO_XPOL_R));
	return fwd_func_splice(fwd, fwd_xpol, t);
}
//-------------------------------------------------------
// extrap_flat 
//-------------------------------------------------------
fwd_func extrap_flat(	fwd_func fwd,
						number 	t)
option(category: "Yield Curve/Discount Function Manipulation")
{
	number fwd_t 		= fwd.fwd(t);
	fwd_func fwd_xpol 	= fwd_func_interp([ t ], [ fwd_t ], ip_step());
	return fwd_func_splice(fwd, fwd_xpol, t);
}

//-------------------------------------------------------
// extrap_flat_sprd 
//-------------------------------------------------------
fwd_func extrap_flat_sprd(	fwd_func fwd,
							number 	t,
							number sprd)
option(category: "Yield Curve/Discount Function Manipulation")
{
	number fwd_t 		= fwd.fwd(t);
	fwd_func fwd_xpol 	= fwd_func_interp([ t ], [ fwd_t + sprd ], ip_step());
	return fwd_func_splice(fwd, fwd_xpol, t);
}

//-------------------------------------------------------
// bootstrap_on_extrap_flat
// needed because extrapolation in bootstrap_on does not work as intended FIX_ME
//-------------------------------------------------------
disc_func bootstrap_on_extrap_flat(	curve_base crv,
									date d,
									date mat,
									//interpolator ip,
									calendar option(nullable) cal/*,
									number extrap_t*/)
{
	fwd_func ff = bootstrap_on(crv, /*ip,*/ cal);
	disc_func df = CORE_INT.on_fwd_to_df(d,mat,	ff, DC_ACT_365, cal,true);
	return df;
	//return extrap_flat_inst_fwd(df,extrap_t);
}

//-------------------------------------------------------
// bootstrap_on_ext_extrap_flat
// needed because extrapolation in bootstrap_on does not work as intended FIX_ME
//-------------------------------------------------------
disc_func bootstrap_on_ext_extrap_flat(	disc_func df_pre,
										date cut_off_d,
										curve_base crv,
										date d,
										date mat,
										calendar option(nullable) cal/*,
										number extrap_t*/)
{
	//disc_func df = bootstrap_on_ext(df_pre,cut_off_d,crv, ip, cal);
	fwd_func ff = bootstrap_on_ext(null<fwd_func>,df_pre,cut_off_d,crv, cal);
	disc_func df = CORE_INT.on_fwd_to_df(d,mat,	ff, DC_ACT_365, cal,true);
	return df;
	//return extrap_flat_inst_fwd(df,extrap_t);
}


module CORE_INT
{
	class cf_min
	{
	public:
		cf_min(vector(date) dates,vector(number) amounts);
		vector(date) 	dates_;
		vector(number) 	amounts_;
	};

	cf_min.cf_min(vector(date) dates, vector(number) amounts)
				: dates_(dates), amounts_(amounts)
	{}


	class curve_min : public cashflows
	{
	public:
		curve_min(	date d, vector(date) ,vector(number), logical);
		curve_min(	date d, vector(date), vector(date), vector(number), 
					number option(nullable),day_count_method, rate_type);
		curve_min(	date d, vector(date), vector(date), fwd_func, 
					number option(nullable),day_count_method, rate_type);
		override date		trade_date();
	    override integer	n_instruments();
	    override logical	get_cashflows(integer ix,out vector(date) dates,out vector(number) amounts);
	
	private:
		void init_zero(	vector(date), vector(date), vector(number), 
						number option(nullable),day_count_method, rate_type);
		void init_single(vector(date) ,vector(number));
		void init_multi(vector(date) ,vector(number));
		date 			trade_date_;
		vector(cf_min) 	cf_;
		
	};
	
	//----------------------------------
	//constructor (zeros)
	//----------------------------------
	curve_min.curve_min(date 					d, 
						vector(date)			date_start, 
						vector(date) 			date_end, 
						vector(number) 			r,
						number option(nullable) s, 
						day_count_method 		dc,
						rate_type				rt)
						: trade_date_(d)
	{
		init_zero(date_start,date_end,r,s,dc,rt);	
	}
	//----------------------------------
	//constructor (zeros)
	//----------------------------------
	curve_min.curve_min(date 					d, 
						vector(date)			date_start, 
						vector(date) 			date_end, 
						fwd_func 				f,
						number option(nullable) s, 
						day_count_method 		dc,
						rate_type				rt)
						: trade_date_(d)
	{
		integer no_instr = v_size(date_start);
		vector(number)  r;
		resize(r,no_instr);
		for(integer i=0;i<no_instr;i++) 
			r[i] = f.fwd((date_start[i] - trade_date_)/365);
		init_zero(date_start,date_end,r,s,dc,rt);	
	}
	//----------------------------------
	// init_zero 
	//----------------------------------
	void curve_min.init_zero(	vector(date)			date_start, 
								vector(date) 			date_end, 
								vector(number) 			r,
								number option(nullable) s, 
								day_count_method 		dc,
								rate_type				rt)				
	{
		integer no_instr = v_size(date_start);
		resize(cf_,0);
		vector(date) 	_dates[2];
		vector(number) 	_amounts[2];
		if(null(s))
			s = 0;
		
		for(integer i=0;i<no_instr;i++) {
			if(trade_date_>	date_start[i])
				continue;				   	    	
		    _amounts[0] = -100.0;			
		    _amounts[1] = future_value(100,r[i]+s,date_start[i], date_end[i],dc,rt) ; 
			QL_FAIL_COND(null(_amounts[1]), "invalid cashflow");
		    _dates[0] 	= date_start[i];
		    _dates[1] 	= date_end[i];
			push_back(cf_, new cf_min(..clone_vector(_dates),..clone_vector(_amounts)));
		}
		
	}
	//----------------------------------
	//constructor (for par swap fixed leg)
	//----------------------------------
	curve_min.curve_min(date d, 
						vector(date) dates, 
						vector(number) amounts, 
						logical single_instr)
						: trade_date_(d)
	{
		if(single_instr)
			init_single(dates, amounts);
		else
			init_multi(dates, amounts);
	}
	//----------------------------------
	// init_single  
	//----------------------------------
	void curve_min.init_single(vector(date) dates, vector(number) amounts)				
	{
		cf_ = [new cf_min(dates,amounts)];
	}
	//----------------------------------
	// init_multi (for par swap fixed leg)
	//----------------------------------	
	void curve_min.init_multi(vector(date) dates, vector(number) amounts)				
	{
		integer no_instr = v_size(dates)-1;
		resize(cf_,no_instr);
		
		for(integer i=0;i<no_instr;i++) {
			vector(date) _dates[i+2];
			vector(number) _amounts[i+2];
			_amounts[0] = amounts[0];
			_dates[0] = dates[0];
			for(integer j=1;j<i+2;j++) {
				_amounts[j] = amounts[j];
				_dates[j] = dates[j];
			} 
			cf_[i] = new cf_min(..clone_vector(_dates),..clone_vector(_amounts));
		}
		number not = -amounts[0];
		for(integer i=0;i<no_instr-1;i++) {
			integer s = v_size(cf_[i].amounts_);
			cf_[i].amounts_[s-1] += not;
		}
	}
	//----------------------------------
	// get_cashflows - override
	//----------------------------------
	logical curve_min.get_cashflows(	integer				ix,
				    					out vector(date)	dates,
				    					out vector(number)	amounts)
	{
		dates 	= cf_[ix].dates_;
		amounts = cf_[ix].amounts_;
		return true;		
	}
	
	date 		curve_min.trade_date() 			{ return trade_date_;}
	integer 	curve_min.n_instruments()		{ return v_size(cf_);}
	//----------------------------------
	//  bootstrap_min
	//----------------------------------	
	disc_func bootstrap_min(curve_min 		c,
							interpolator	ip 		= ip_linear(),
							rate_type		ip_rt 	= RT_LOG_DF)
	{
		return bootstrap(c,ip, ip_rt);
	}
	
	/*-----------------------------------------------------------------------
	  fwd_parallell_shift
	  ----------------------------------------------------------------------*/
	fwd_func fwd_parallell_shift_f(	fwd_func 	fwd_f,
									number 		delta)
	{
		return fwd_func_add(fwd_f, fwd_func_interp([1], [delta], ip_linear()));
	}
	
	/*-----------------------------------------------------------------------
	  fwd_parallell_shift
	  ----------------------------------------------------------------------*/
	disc_func fwd_parallell_shift_df(	disc_func 	fwd_df,
										number 		delta,
										date 		d,
										vector(date) index_start_dates,
										vector(date) index_end_dates,
										day_count_method dc)
	{
		vector(number) r = fwd_df.simple_rate(d,index_start_dates,index_end_dates, dc);
		CORE_INT.curve_min fwd_crv 	= new CORE_INT.curve_min(d,index_start_dates,index_end_dates, 
															r ,delta, dc,RT_SIMPLE);
		return CORE_INT.bootstrap_min(fwd_crv);
	}

	/*-----------------------------------------------------------------------
	  fwd_disc_spreads
	  Note: index_dates should be in sync with the tenor of the fwd_func
	  ----------------------------------------------------------------------*/
	vector(number) fwd_disc_spreads(fwd_func 			fwd_f,
									disc_func 			disc_r,
									date 				d,
									vector(date) 		index_start_dates,
									vector(date) 		index_end_dates,
									day_count_method 	dc,
									out vector(number) 	fr,
									out vector(number) 	dr)
	{
		integer s = v_size(index_start_dates);
		vector(number) 	spreads[s];
		resize(fr,s);
		resize(dr,s);
		
		for(integer i=0;i<s;i++){
			fr[i] 		= fwd_f.fwd((index_start_dates[i]-d)/365.0);
			dr[i] 		= disc_r.simple_rate(d,index_start_dates[i],index_end_dates[i],dc);
			spreads[i] 	= fr[i]-dr[i];
		}
		return spreads;
	}
	
	/*-----------------------------------------------------------------------
	  fwd_disc_spreads_daily
	  ----------------------------------------------------------------------*/
	vector(number) fwd_disc_spreads_daily(	fwd_func 			fwd_f,
											disc_func 			disc_r,
											date 				d,
											date 				end,
											ir_index 			idx,
											out vector(date) 	start_dates,
											out vector(date) 	end_dates,
											out vector(number) 	fr,
											out vector(number) 	dr)
	{
		
		resize(fr,0);
		resize(dr,0);
		resize(start_dates,0);
		resize(end_dates,0);
		day_count_method dc = idx.day_count_method();
		calendar cal 		= idx.fixing_calendar();
		date fix			= d;
		while(!idx.is_valid_fixing_date(fix))
			fix = cal.adjust_date(d,BD_FOLLOWING);
		date s0 			= idx.value_date(fix);
		date s1 			= idx.maturity(s0);
		date s1_lag			= s1;
		while(s1 <= end) {						
			push_back(fr,fwd_f.fwd((s0-d)/365.0));
			push_back(dr,disc_r.simple_rate(d,s0,s1,dc));
			push_back(start_dates,s0);
			push_back(end_dates,s1);
			fix = cal.move_bus_days(fix,1);
			while(!idx.is_valid_fixing_date(fix))
				fix = cal.adjust_date(d,BD_FOLLOWING);
			s0 	= idx.value_date(fix);
			s1 	= idx.maturity(s0);
			while(s1 != s1_lag){//sometimes more than one valuedate can result in the same maturity 
				fix = cal.move_bus_days(fix,1);
				while(!idx.is_valid_fixing_date(fix))
					fix = cal.adjust_date(d,BD_FOLLOWING);
				s0 	= idx.value_date(fix);
				s1 	= idx.maturity(s0);
			}
			s1_lag	= s1;
		}
		return fr-dr;
	}

	//-------------------------------------------------------
	// disc_func_from_fwd
	//-------------------------------------------------------
	/*disc_func disc_func_from_fwd(	date 			d,			
									vector(date) 	idx_start,
									vector(date) 	idx_end,								
									fwd_func		fwd,
									day_count_method idx_dc,
									interpolator 	ip,
									rate_type		rt)
	{							
		number n = v_size(idx_start);
		if(n == 0 || n != v_size(idx_end) || idx_start[0] < d)
			return null<disc_func>;
	 			
		vector(number) float_dt = period (idx_start, idx_end, idx_dc);
		vector(number) r 	 	= fwd.fwd((idx_start-d)/365);
		vector(number) fra_cf 	= float_dt .* r;
	
		vector(cashflows) cf;
		for (integer i = 0 ; i < n ; i++) {	
			vector(number) fixed_amounts = [-100, (1+fra_cf[i])*100];
			vector(date) fixed_dates = [idx_start[i],idx_end[i]];									
			push_back(cf,cashflows(d, fixed_dates, fixed_amounts));
		}
			
		return bootstrap(cashflows(cf), ip, rt);			
	}*/

	//-------------------------------------------------------
	// fwd_func_from_r
	// rate-vector --> fwd_func
	// use fwd_func_interp instead??
	//-------------------------------------------------------
	fwd_func fwd_func_from_r(	date 			d,			
								vector(date) 	idx_start,
								vector(date) 	idx_end,
								vector(number) 	r,//all r must have same tenor, simple rates
								day_count_method dc,
								disc_func		df_disc,
								interpolator 	fwd_ip = ip_linear())
	{							
		number n = v_size(r);
		if(n == 0 || n != v_size(idx_start) || n != v_size(idx_end))
			return null<fwd_func>;
	 		
		vector(cashflows_fwd) cf;
		vector(number) float_mult 	 = period (idx_start, idx_end, dc)*100;
		vector(number) fixed_amounts = -float_mult .* r;
		fixed_amounts *= 100;
	
		for (integer i = 0 ; i < n ; i++) 										
			push_back(cf,cashflows_fwd(d, [idx_end[i]], [fixed_amounts[i]], [idx_start[i]], [idx_end[i]], [float_mult[i]]));
				
		return bootstrap_fwd(cashflows_fwd(cf), df_disc, fwd_ip);			
	}

	/*-----------------------------------------------------------------------
	  disc_func_from_r
	  // rate-vector --> disc_func
	  ----------------------------------------------------------------------*/
	disc_func disc_func_from_r(	date 			d,			
								vector(date) 	idx_start,
								vector(date) 	idx_end,
								vector(number) 	r,//all r must have same tenor, simple rates
								day_count_method dc,
								interpolator	ip 		= ip_linear(),
								rate_type		ip_rt 	= RT_LOG_DF)
	{
		CORE_INT.curve_min fwd_crv 	= new CORE_INT.curve_min(d,idx_start,idx_end, 
															r ,0, dc,RT_SIMPLE);
		return CORE_INT.bootstrap_min(fwd_crv,ip,ip_rt);
	}

	//-------------------------------------------------------
	// on_to_df
	//-------------------------------------------------------
	disc_func on_to_df(	date 				d,
						//date 				settle, 
						date 				mat, 
						number 				on_r,
						day_count_method 	dc,
						calendar 			cal,
						logical 			allow_extrap_flat )
	{	
		QL_REQUIRE(d < mat, "invalid maturity");

		date s 	= d;
		date se = cal.move_bus_days(s,1);
		
		vector(number) per_v;
		vector(number) df_v;
		number df_r = 1;
		while (se  <= mat || (se > mat && s < mat)){
				
			number per_e = period(d,se,DC_ACT_365);	
			push_back(per_v, per_e);

			number df_f = 1.0/(1.0 + on_r * period(s,se,dc));
			df_r 		*= df_f;
			push_back(df_v, df_r);
			s 			= se;
			se 			= cal.move_bus_days(s,1);
		}
		disc_func df = disc_func_interp(per_v,df_v, ip_linear(), RT_LOG_DF);
		
		if(allow_extrap_flat) {
			number t =  (mat - d)/365;
			return extrap_flat_inst_fwd(df,t);
		}
		else {
			return disc_func_restrict (df, 0, per_v[v_size(per_v)-1]);
		}
	}

	

	//-------------------------------------------------------
	// on_to_df
	//-------------------------------------------------------
	disc_func on_to_df(	date 				d,
						disc_func			df_pre,
						date				cutoff_date,//last date inclusive in df_pre 
						date 				mat, 
						number 				on_r,
						day_count_method 	dc,
						calendar 			cal,
						logical 			allow_extrap_flat )
	{	
		QL_REQUIRE(d < cutoff_date, "invalid cutoff date");

		cutoff_date = cal.adjust_date(cutoff_date,BD_FOLLOWING);
		QL_REQUIRE(cutoff_date < mat, "invalid maturity");
		
		date s 		= cutoff_date;
		date se 	= cal.move_bus_days(s,1);
		
		vector(number) per_v;
		vector(number) df_v;
		number cut_off = (cutoff_date-d)/365;
		number df_r = df_pre.disc(cut_off);

		while (se  <= mat || (se > mat && s < mat)){
					
			push_back(per_v, period(d,se,DC_ACT_365));
			number df_f = 1.0/(1.0 + on_r * period(s,se,dc));
			df_r 		*= df_f;
			push_back(df_v, df_r);
			s 			= se;
			se 			= cal.move_bus_days(s,1);
		}
		disc_func df_post 	= disc_func_interp(per_v,df_v, ip_linear(), RT_LOG_DF);
		disc_func df 		= disc_func_splice(df_pre,df_post,cut_off);

		if(allow_extrap_flat) {
			number t =  (mat - d)/365;
			return extrap_flat_inst_fwd(df,t);
		}
		else {
			return disc_func_restrict (df, 0, per_v[v_size(per_v)-1]);
		}
	}
	//-------------------------------------------------------
	// on_fwd_to_df
	//-------------------------------------------------------
	disc_func on_fwd_to_df(	date 				d,
							disc_func			df_pre,
							date				cutoff_date,//last date inclusive in df_pre 
							date 				mat, 
							fwd_func			fwd,
							day_count_method 	dc,
							calendar 			cal,
							logical 			allow_extrap_flat)
	{	
		QL_REQUIRE(d < cutoff_date, "invalid cutoff date");

		cutoff_date = cal.adjust_date(cutoff_date,BD_FOLLOWING);
		QL_REQUIRE(cutoff_date < mat, "invalid maturity");
		
		date s 		= cutoff_date;
		date se 	= cal.move_bus_days(s,1);
		
		vector(number) per_v;
		vector(number) df_v;
		number cut_off 	= (cutoff_date-d)/365;
		number df_r 	= df_pre.disc(cut_off);

		while (se  <= mat || (se > mat && s < mat)){
							
			number per_e = period(d,se,DC_ACT_365);	
			push_back(per_v, per_e);
			
			number per_s 	= period(d,s,DC_ACT_365);
			number on_r 	= fwd.fwd(per_s);
		
			number df_f 	= 1.0/(1.0 + on_r * period(s,se,dc));
			df_r 			*= df_f;
			push_back(df_v, df_r);
			s 				= se;
			se 				= cal.move_bus_days(s,1);
		}
		disc_func df_post 	= disc_func_interp(per_v,df_v, ip_linear(), RT_LOG_DF);
		disc_func df 		= disc_func_splice(df_pre,df_post,cut_off);
		if(allow_extrap_flat) {
			number t =  (mat - d)/365;
			return extrap_flat_inst_fwd(df,t);
		}
		else {		
			return disc_func_restrict (df, 0, per_v[v_size(per_v)-1]);
		}
	}
}

module CORE_INT
{
	//-------------------------------------------------------
	// cf_fra  (cashflow sign = lending)
	//-------------------------------------------------------
	cashflows cf_fra(	date d0, 
						date beg, 
						date end, 
						number rdt, 
						number nom)
	{
		vector(date)	fixed_dates 	= [ beg, end ];
		vector(number)	fixed_amounts 	= [ -nom, (1 + rdt)*nom];
		return cashflows(d0, fixed_dates, fixed_amounts);
	}

	//-------------------------------------------------------
	// cf_settle_flow  (cashflow sign = lending)
	//-------------------------------------------------------
	cashflows cf_settle_flow(	date 		trade,
								date 		settle,
								disc_func	df_disc)
	{
		vector(number)	fixed_amounts[2];
		vector(date)	fixed_dates = [trade, settle];	
	
		number pv = df_disc.disc((settle - trade) / 365);																 							
		fixed_amounts[1] = 100;
		fixed_amounts[0] = -pv*100;	
		return cashflows(trade, fixed_dates, fixed_amounts); 
	}
}
/*-----------------------------------------------------------------------
  Name :    disc_func_fwd_roll
  Notes :           
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
disc_func disc_func_fwd_roll(	fwd_func			fwd,
							    date				d0,
							    date				d_max,
							    integer				freq,
								logical				eom,
							    calendar			cal,
								bd_convention 		bd,
							    day_count_method	dc)
	option (category: 'Yield Curve/Discount Function Manipulation')
{
	if(freq == 365)
		return CORE_INT.on_fwd_to_df(d0,d_max,fwd,dc,cal,true);

	vector(date)	v_beg;
	vector(date)	v_end;
	vector(number)	v_rdt;

	for (date beg = d0 ;; ++beg){
	    if (!cal.is_holiday(beg)) {
			date end = move_month(beg, 12 / freq ,eom,cal, bd, false);

			if (end > d_max)
			    break;
	
			push_back(v_beg, beg);
			push_back(v_end, end);
			push_back(v_rdt,  fwd.fwd((beg - d0) / 365) * period(beg, end, dc));
	    }
	}
	integer		n_src		= v_size(v_beg);
	integer		src_i		= 0;
	integer		dest_i		= 0;

	while (src_i < n_src) {
	    integer	last_neq	= src_i + 1;

	    while (last_neq < n_src && v_end[last_neq] == v_end[src_i])
			++last_neq;

	    integer	mid_i	= integer(floor((src_i + last_neq - 1) / 2));		
	    v_beg[dest_i] 	= v_beg[mid_i];
	    v_end[dest_i] 	= v_end[mid_i];
	    v_rdt[dest_i] 	= v_rdt[mid_i];

	    ++dest_i;
	    src_i = last_neq;
	}

	resize(v_beg, dest_i);
	resize(v_end, dest_i);
	resize(v_rdt, dest_i);

	return bootstrap(cashflows(CORE_INT.cf_fra(d0, v_beg, v_end, v_rdt,100)));
}
module CORE_INT
{
	/*-----------------------------------------------------------------------
	  Name :    disc_func_fwd_roll
	  Notes :   this version uses fwd_max which is the max date for the fwd_func        
	  Changes:
	  			Date    		Who     Detail
	  			? 				MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	disc_func disc_func_fwd_roll(	fwd_func		fwd,
								    date			d0,
								    date			d_max,
									logical			max_is_beg,
								    integer			freq,
									logical			eom,
								    calendar		cal,
									bd_convention 	bd,
								    day_count_method	dc,
									interpolator	ip,
									rate_type		ip_rt)
	{
		vector(date)	v_beg;
		vector(date)	v_end;
		vector(number)	v_rdt;
	
		for (date beg = d0 ;; ++beg)
		    if (!cal.is_holiday(beg)) {
				date end;
				if(max_is_beg) {
					if (beg > d_max)
					    break;
					end = move_month(beg, 12 / freq ,eom,cal, bd, false);
				}
				else {
					end = move_month(beg, 12 / freq ,eom,cal, bd, false);
					if (end > d_max)//note that there is no guarantee that end == d_max when it breaks i.e. using 
					    break;		//the resulting disc_func up to d_max can result in unintended extrapolation
				}
	
				push_back(v_beg, beg);
				push_back(v_end, end);
				push_back(v_rdt,  fwd.fwd((beg - d0) / 365) * period(beg, end, dc));
		    }
	
		integer		n_src		= v_size(v_beg);
		integer		src_i		= 0;
		integer		dest_i		= 0;
	
		while (src_i < n_src) {
		    integer	last_neq	= src_i + 1;
	
		    while (last_neq < n_src && v_end[last_neq] == v_end[src_i])
				++last_neq;
	
		    integer	mid_i		= integer(floor((src_i + last_neq - 1) / 2));		
	
		    v_beg[dest_i] = v_beg[mid_i];
		    v_end[dest_i] = v_end[mid_i];
		    v_rdt[dest_i] = v_rdt[mid_i];
	
		    ++dest_i;
		    src_i = last_neq;
		}
	
		resize(v_beg, dest_i);
		resize(v_end, dest_i);
		resize(v_rdt, dest_i);
	
		return bootstrap(cashflows(CORE_INT.cf_fra(d0, v_beg, v_end, v_rdt,100)), ip, ip_rt);
	}

	
}

fwd_func fwd_func_add_spread(	fwd_func 	fwd,
								number 		spread)
	option(category: "Yield Curve/Discount Function Manipulation")
{
	return CORE_INT.fwd_parallell_shift_f(fwd,spread);
}