option(null: hard);	// Throw error if a null value is used in a non-nullable context
/*	
	fxcurve
	Developer: Algorithmica Research, Magnus Nyström
			
*/


/*-----------------------------------------------------------------------
  Name :  	class fx_curve
  Notes :   This curve can contain a mix of either fx swaps or fx fwds with one restriction that before spot it must be 
			either fx swaps or fx fwds. The purpose of this class is to produce for example fx implied disc_funcs, synthetic deposits,
			interpolated fwd fx quotes (interpolation is done via disc_funcs i.e. no interpolation directly on fx fwd quotes) , ...       
  Changes:
  			Date    		Who     Detail
  			11-Dec-2013 	MN      Original code (from fiws but redesigned)
			29-Aug-2017		MN		converted to new fx class, OBS old fx_curve can be accessed by CORE_INT_COMPAT.fx_curve


argument all:   if true all instruments in the original curve are returned, 
				sometimes the on or tn quote is not available due to calendar effects and if 
				this flag is false those quotes will be removed
				the flag,keep_on_if_dup_, determines which one to keep (currently hard coded to keep on and discard tn)

Example where on and tn have same settle and maturity: USDJPY, d = #2014-11-11 (US veterans day)
-->
	settle		mat
on	2014-11-12	2014-11-13
tn	2014-11-12	2014-11-13
spot date = 2014-11-13
  ----------------------------------------------------------------------*/

	
class fx_curve option (category: 'Financial/FX')
{
   	public:
		
		fx_curve(vector(instrument),fx_spot, fx_spot, quote_side_code);								
		fx_curve(fx_curve);

		vector(instrument)		orig_instruments(logical=false);
		vector(fx_swap)			orig_fx_swaps(logical=false);
		vector(fx_fwd)			orig_fx_fwds(logical=false);
			
		void 					swaps_pre_spot(	logical,instr_type option(nullable),logical ,logical,
												out vector(date) , out vector(number) , logical =false);

		void 					swaps_post_spot(logical,instr_type option(nullable),logical ,logical,
												out vector(date) , out vector(number) );
		
		void 					swaps(	logical,instr_type option(nullable),logical ,logical,logical,
										out vector(date) , out vector(number) );

		void 					outrights_pre_spot(	logical,instr_type option(nullable),logical ,logical,
													out vector(date) , out vector(number) );

		void 					outrights_post_spot(logical,instr_type option(nullable),logical ,logical,
													out vector(date) , out vector(number) );

		void					outrights(	logical,instr_type option(nullable),logical ,logical,logical,
											out vector(date) , out vector(number) );

		void					quotes_df_pre_spot(disc_func,disc_func,logical,instr_type option(nullable),logical,logical,
													out vector(date) ,out vector(number) );
		void					quotes_df_post_spot(disc_func,disc_func,logical,instr_type option(nullable),logical,logical,
													out vector(date) ,out vector(number) );

		void					quotes_df(disc_func,disc_func,logical,instr_type option(nullable),logical,logical,
											out vector(date) , out vector(number) );
		vector(number)			quotes_df(disc_func,disc_func, logical);
		vector(number)			quotes(instr_type option(nullable) ,logical = false);

		disc_func				disc_base(disc_func,interpolator,rate_type,logical = false);
		disc_func				disc_price(disc_func,interpolator,rate_type,logical = false);			
		curve					synt_depo_base(disc_func,interpolator,rate_type,day_count_method ,
												out vector(date) , logical = false,
											   date option(nullable) = null<date>);
		curve					synt_depo_price(disc_func,interpolator,rate_type,day_count_method ,
												out vector(date) , logical = false,
												date option(nullable) = null<date>);

		number					ip_outright(disc_func,disc_func,date, logical);
		number					ip_swap_post_spot(disc_func,disc_func,date,logical);
		number					ip_swap_post_spot_ff(disc_func option(nullable),disc_func option(nullable),number option(nullable),
													 disc_func option(nullable),disc_func option(nullable),number option(nullable),date,date,logical);

		vector(instrument)		fx_inst_pre_spot(instr_type option(nullable), logical = false, logical = true ) ;
		vector(instrument)		fx_inst_post_spot(instr_type option(nullable), logical = false, logical = true); 
		vector(instrument)		fx_inst(instr_type option(nullable), logical = false, logical = true );
		vector(instrument)		fx_inst(logical = true); 

		vector(date) 			maturity(instr_type option(nullable) type, logical =false);
		vector(date) 			settle_date(instr_type option(nullable) type, logical =false);

		vector(logical)			inst_in_curve(instr_type option(nullable));	
		number					spot_quote(logical,logical,logical =false);
		date 					spot_settle();
		date 					trade_date();
		integer  				pip_decimals();
		integer					scale();

		void					set_spot_quote(number, number);
		//void 					set_quotes(vector(number), instr_type option(nullable) ,logical);

		void 					update_s_pre(vector(instrument) );
		void 					update_swap_post(vector(instrument) );
		vector(logical)			orig_incl();
		vector(logical)			s_pre_l();
		vector(logical)			f_pre_l();
		vector(logical)			post_l();

		vector(logical) 		included_post_spot(instr_type option(nullable));
		vector(logical) 		included_pre_spot(instr_type option(nullable));
		vector(logical) 		included(instr_type option(nullable));
	protected:		
		//void					check_inst_type(instrument);
		//void					set_fx_spot(instrument );
		void					check_fx_spot();
		void					check_fx_instr();
		//void					check_type();	
		void					init_type();
		void					init_sort();
		void 					init_int();
		void 					init_int_fx_swap(date ,vector(date) ,logical ,integer );
		void 					init_int_fx_fwd(date ,vector(date) ,logical ,integer );
		logical 				check(out string );
		void 					check_ccy(vector(instrument), string, string);
		vector(instrument) 		s_pre(logical);
		vector(instrument) 		f_pre(logical);
		vector(instrument) 		post(logical);
		vector(instr_type) 		type_post() ;						

		void 					swaps_pre_spot(	logical,instr_type option(nullable),logical, logical,out vector(date) ,
												out vector(number) , vector(instrument) option(nullable), logical r=false);

		void 					swaps_post_spot(logical,instr_type option(nullable),logical ,logical,out vector(date) ,
												out vector(number) ,vector(instrument) option(nullable));

		void 					outrights_pre_spot(logical,instr_type option(nullable),logical,logical,out vector(date) ,
													out vector(number) , vector(instrument) option(nullable));

		void					outrights_post_spot(logical ,instr_type option(nullable),logical,logical ,
													out vector(date),out vector(number)  ,vector(instrument) option(nullable) );

		disc_func				disc(disc_func,logical,interpolator,rate_type,logical,out vector(date) ,out vector(number) );
		
		curve					synt_depo(	disc_func,logical ,interpolator,rate_type,day_count_method,logical,
											date option(nullable), out vector(date) );
						
	private:					
		fx_spot					fx_spot_bid_;
		fx_spot					fx_spot_ask_;
		vector(instrument)		fx_i_;
		vector(instrument)		s_pre_;
		vector(instrument)		f_pre_;
		vector(instrument)		post_;	
		vector(logical)			s_pre_l_;
		vector(logical)			f_pre_l_;
		vector(logical)			post_l_;			
		vector(instr_type) 		type_;
		logical 				pre_is_swap_;
		quote_side_code			spot_qs_;
		logical 				keep_on_if_dup_;
														
};

//-------------------------------------------------------
// init_sort
//-------------------------------------------------------
void fx_curve.init_sort()
{
	vector(date) mat = fx_i_.maturity();
	sort (fx_i_, mat);
	vector(logical) post_spot 	= fx_i_.fx_is_post_spot(fx_spot_bid_); 
	if(v_size(post_spot) >1 && !post_spot[0] && !post_spot[1] && mat[0] == mat[1]) {//on and tn with same maturity (one will be discarded)
		logical is_on = fx_i_[0].fx_is_on();	
		if(!is_on) {//tn sorted before on
			instrument	t = fx_i_[0].clone();
			fx_i_[0] = fx_i_[1];
			fx_i_[1] = t;
		}
	}
}

//-------------------------------------------------------
// constructor
//-------------------------------------------------------
fx_curve.fx_curve(	vector(instrument) 	fx_instruments,
					fx_spot 			fx_spot_bid,
					fx_spot 			fx_spot_ask,
					quote_side_code 	spot_quote_side)
					: fx_spot_bid_(fx_spot_bid), fx_spot_ask_(fx_spot_ask), spot_qs_(spot_quote_side), keep_on_if_dup_(true)
{
	QL_FAIL_COND(v_size(fx_instruments) == 0,"invalid instrument vector");	
	fx_i_ = clone_vector_deep(fx_instruments);
		
	check_fx_spot();
	check_fx_instr();	
	init_sort();	
	init_type();
	init_int();
}
//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
fx_curve.fx_curve(fx_curve f):  fx_spot_bid_(f.fx_spot_bid_), fx_spot_ask_(f.fx_spot_ask_),
								fx_i_( clone_vector_deep(f.fx_i_)), 
								s_pre_( clone_vector_deep(f.s_pre_)), 
								f_pre_( clone_vector_deep(f.f_pre_)), 
								post_( clone_vector_deep(f.post_)),
								s_pre_l_(!null(f.s_pre_l_) ? clone_vector(f.s_pre_l_) : null<vector(logical)>), 
								f_pre_l_(!null(f.f_pre_l_) ? clone_vector(f.f_pre_l_): null<vector(logical)>), 
								post_l_( !null(f.post_l_) ? clone_vector(f.post_l_): null<vector(logical)>), 
								type_(   !null(f.type_) ? clone_vector(f.type_): null<vector(instr_type)>), 
								pre_is_swap_(f.pre_is_swap_), spot_qs_(f.spot_qs_), keep_on_if_dup_(f.keep_on_if_dup_)

{}
//-------------------------------------------------------
// check_fx_spot
//-------------------------------------------------------
void fx_curve.check_fx_spot()
{
	QL_FAIL_COND(!fx_spot_bid_.is_valid(),"invalid fx spot(bid)");
	QL_FAIL_COND(!fx_spot_ask_.is_valid(),"invalid fx spot(ask)");

	QL_FAIL_COND(fx_spot_bid_.base_ccy()  != fx_spot_ask_.base_ccy(),	strcat(["invalid base currency [",fx_spot_ask_.name(),"]"]));
	QL_FAIL_COND(fx_spot_bid_.price_ccy() != fx_spot_ask_.price_ccy(),	strcat(["invalid price currency [",fx_spot_ask_.name(),"]"]));

	//QL_FAIL_COND(fx_spot_bid_.settle_date()  != fx_spot_ask_.settle_date(),	strcat(["invalid settlement date [",fx_spot_ask_.name(),"]"]));
	QL_FAIL_COND(fx_spot_bid_.trade_date()  != fx_spot_ask_.trade_date(),	strcat(["invalid trade date [",fx_spot_ask_.name(),"]"]));

	//QL_FAIL_COND(fx_spot_bid_.pip_decimals()  != fx_spot_ask_.pip_decimals(),	strcat(["invalid pip decimals [",fx_spot_ask_.name(),"]"]));
	//QL_FAIL_COND(fx_spot_bid_.reverse_ric()  != fx_spot_ask_.reverse_ric(),	strcat(["invalid reverse_ric setting [",fx_spot_ask_.name(),"]"]));

}
//-------------------------------------------------------
// check_fx_instr
//-------------------------------------------------------
void fx_curve.check_fx_instr()
{
	integer size = v_size(fx_i_);
	
	for(integer i = 0; i < size;i++){
		if(!fx_i_[i].is_valid()){
			string s = strcat(["invalid instrument in fx curve [",fx_i_[i].name(),"]"]);
			QL_FAIL(s);		
		}
		if (!fx_i_[i].is_fx_fwd() && !fx_i_[i].is_fx_swap()){			
			string s = strcat(["invalid instrument type in fx curve (must be fx swap or fx fwd) [",fx_i_[i].name(),"]"]);
			QL_FAIL(s);			
		}

		if(fx_spot_bid_.fx_base_ccy()  != fx_i_[i].fx_base_ccy())
			QL_FAIL(strcat(["invalid base currency [",fx_i_[i].name(),"]"]));
		if(fx_spot_bid_.fx_price_ccy() != fx_i_[i].fx_price_ccy())
			QL_FAIL(strcat(["invalid price currency [",fx_i_[i].name(),"]"]));
		if(fx_spot_bid_.trade_date()   != fx_i_[i].trade_date())
			QL_FAIL(strcat(["invalid trade date [",fx_i_[i].name(),"]"]));
	}
}
//-------------------------------------------------------
// init_type
//-------------------------------------------------------
void fx_curve.init_type()
{
	integer size = v_size(fx_i_);
	resize(type_,size);
	for(integer i = 0; i < size;i++) {
		type_[i] = fx_i_[i].is_fx_fwd() ? instr_type.FX_FORWARD : instr_type.FX_SWAP;	//type already checked	
	}
}
//-------------------------------------------------------
// init_int_fx_swap
//-------------------------------------------------------
void fx_curve.init_int_fx_swap(date spot_settle,
							   vector(date) mat,
							   logical post_spot,
							   integer i)
{
	if(mat[i] <= spot_settle)	{
		push_back(s_pre_, fx_i_[i]);
		logical incl = false;
		if(!post_spot) {
			if(i == 0)
				incl = true;
			else if(i==1 && mat[i] > mat[i-1])
				incl = true;
			else if(i==1 && mat[i] == mat[i-1]){
				QL_FAIL_COND(v_size(s_pre_) != 2,"mixing fx_swap and fx fwd instruments before spot is not supported");
				if(!keep_on_if_dup_) {
					incl = true;
					s_pre_l_[0] = false;//i.e. remove o/n
					s_pre_[0] = s_pre_[0].set_quote(null<number>);
				}
			}					
		}
		push_back(s_pre_l_, incl);
		if(!incl){
			integer k = v_size(s_pre_)-1;
			s_pre_[k] = s_pre_[k].set_quote(null<number>);
		}
	}
	else if(!post_spot) {//could be a tn (that will be removed)
		push_back(s_pre_, fx_i_[i]);		
		push_back(s_pre_l_, false);
		integer k = v_size(s_pre_)-1;
		s_pre_[k] = s_pre_[k].set_quote(null<number>);
	}
	else {
		push_back(post_,fx_i_[i]);
		logical incl = post_spot  && (i == 0 || (i > 0 && mat[i] > mat[i-1]));
		push_back(post_l_, incl);
		if(!incl){
			integer k = v_size(post_)-1;
			post_[k] = post_[k].set_quote(null<number>);
		}
	}
}
//-------------------------------------------------------
// init_int_fx_fwd
//-------------------------------------------------------
void fx_curve.init_int_fx_fwd(date spot_settle,
							   vector(date) mat,
							   logical post_spot,
							   integer i)
{
	if(mat[i] < spot_settle) {
		push_back(f_pre_, fx_i_[i]);
		logical incl = false;
		if(!post_spot) {
			if(i == 0)
				incl = true;
			else if(i == 1 && mat[i] > mat[i-1])
				incl = true;
			else if(i == 1 && mat[i] == mat[i-1]){
				QL_FAIL_COND(v_size(f_pre_) != 2,"mixing fx_swap and fx fwd instruments before spot is not supported");
				if(!keep_on_if_dup_) {
					incl = true;
					f_pre_l_[0] = false;//i.e. remove o/n
					f_pre_[0] = f_pre_[0].set_quote(null<number>);
				}
			}
		}
		push_back(f_pre_l_, incl);
		if(!incl){
			integer k = v_size(f_pre_)-1 ;
			f_pre_[k] = f_pre_[k].set_quote(null<number>);
		}
	}
	else if(!post_spot) {//could be a tn (that will be removed)
		push_back(f_pre_, fx_i_[i]);		
		push_back(f_pre_l_, false);
		integer k = v_size(f_pre_)-1;
		f_pre_[k] = f_pre_[k].set_quote(null<number>);
	}
	else if(mat[i] > spot_settle) {
		push_back(post_, fx_i_[i]);
		logical incl = post_spot  && (i == 0 || (i > 0 && mat[i] > mat[i-1]));					
		push_back(post_l_, incl);
		if(!incl){
			integer k = v_size(post_)-1;
			post_[k] = post_[k].set_quote(null<number>);
		}
	}
	else {//spot
		push_back(post_, fx_i_[i]);
		push_back(post_l_, false);
		integer k = v_size(post_)-1;
		post_[k] = post_[k].set_quote(null<number>);
	}
}
//-------------------------------------------------------
// init_int
//-------------------------------------------------------
void fx_curve.init_int()
{
	vector(date) mat 			= fx_i_.maturity(); 
	vector(logical) post_spot 	= fx_i_.fx_is_post_spot(fx_spot_bid_); 
	date spot_settle 			= fx_spot_bid_.settle_date();

	resize(s_pre_,0);
	resize(f_pre_,0);
	resize(post_,0);
	resize(s_pre_l_,0);
	resize(f_pre_l_,0);
	resize(post_l_,0);

	for(integer i = 0; i < v_size(fx_i_);i++){
		
		if(type_[i] == instr_type.FX_FORWARD){	//maturity == settle for fx fwds
			init_int_fx_fwd(spot_settle, mat, post_spot[i], i);
		}
		else {//fx_swap
			init_int_fx_swap(spot_settle, mat, post_spot[i], i);
		}		
	} 
	integer size_p_s = v_size(s_pre_);
	integer size_p_f = v_size(f_pre_);
	QL_FAIL_COND(size_p_s > 0 && size_p_f > 0,"mixing fx_swap and fx fwd instruments before spot is not supported");
	pre_is_swap_ = size_p_f == 0;//note: pre_is_swap_ is true if size_p_s >= 0

	string err_msg;
	logical ok;
	ok = check(err_msg);
	return;
}
//-------------------------------------------------------
// check
//-------------------------------------------------------	
logical fx_curve.check(out string  err_msg)
{
	err_msg = '';
	
	vector(instrument) fx_s_pre_tmp 	= s_pre(false);	
	integer size_p_s_e 					= v_size(fx_s_pre_tmp);

	QL_FAIL_COND(size_p_s_e > 0 && fx_s_pre_tmp[size_p_s_e - 1].maturity() != fx_spot_bid_.settle_date(),"missing fx swap (before spot)");
	
	for(integer i = 1; i < size_p_s_e; i++) {
		if(fx_s_pre_tmp[i].maturity() <= fx_s_pre_tmp[i - 1].maturity())			
			QL_FAIL(strcat(["duplicate (or unordered) maturities (fx swaps pre spot) [",fx_s_pre_tmp[i].name(),"]"]));
		if(fx_s_pre_tmp[i].settle_date() <= fx_s_pre_tmp[i - 1].settle_date())			
			QL_FAIL(strcat(["duplicate (or unordered) settlement dates (fx swaps pre spot) [",fx_s_pre_tmp[i].name(),"]"]));
		if(fx_s_pre_tmp[i].settle_date() != fx_s_pre_tmp[i - 1].maturity())			
			QL_FAIL(strcat(["non consecutive fx swap (fx swaps pre spot) [",fx_s_pre_tmp[i].name(),"]"]));
	}

	vector(instrument) fx_f_pre_tmp 	= f_pre(false);	
	integer size_p_f_e 					= v_size(fx_f_pre_tmp);
	for(integer i = 1; i < size_p_f_e; i++) {
		if(fx_f_pre_tmp[i].maturity() <= fx_f_pre_tmp[i - 1].maturity())			
			QL_FAIL(strcat(["duplicate maturities (fx fwd pre spot) [",fx_f_pre_tmp[i].name(),"]"]));
	}	

	vector(instrument) fx_post_tmp 	= post(false);		
	integer size_e 					= v_size(fx_post_tmp);	
	for(integer i=1; i<size_e; i++) {
		if(fx_post_tmp[i].maturity() <= fx_post_tmp[i - 1].maturity())
			QL_FAIL(strcat(["duplicate (or unordered) maturities (after spot) [",fx_post_tmp[i].name(),"]"]));		
	}

	date settle = null<date>;
	for(integer i=0; i<size_e; i++) {
		if(fx_post_tmp[i].instr_type() != instr_type.FX_SWAP)
			continue;
		if(null(settle))
			settle = fx_post_tmp[i].settle_date();
		else {		
			if(fx_post_tmp[i].settle_date() != settle)
				QL_FAIL(strcat(["invalid settlement date (fx swap after spot) [",fx_post_tmp[i].name(),"]"]));
		}
	}

	/* this check is already done
	string base_c 	= fx_spot_bid_.base_ccy();
	string price_c 	= fx_spot_bid_.price_ccy();
	check_ccy(fx_s_pre_tmp, base_c,price_c);
	check_ccy(fx_f_pre_tmp, base_c,price_c);
	check_ccy(fx_post_tmp, base_c,price_c);
	*/
	return true;
}
//-------------------------------------------------------
// check_ccy
//-------------------------------------------------------
void fx_curve.check_ccy(vector(instrument) inst,
						string base_c,
						string price_c)
{
	for(integer i=0; i < v_size(inst); i++) {
		QL_FAIL_COND(base_c!= inst[i].fx_base_ccy(),strcat(["invalid base currency [",inst[i].name(),"]"]));
		QL_FAIL_COND(price_c!= inst[i].fx_price_ccy(),strcat(["invalid price currency [",inst[i].name(),"]"]));
	}
}
//-------------------------------------------------------
// type_post
//-------------------------------------------------------	
vector(instr_type) fx_curve.type_post()
{
	integer s_pre = pre_is_swap_? v_size(s_pre_) : v_size(f_pre_);
	return sub_vector(type_, s_pre, v_size(type_) - s_pre);
}
//-------------------------------------------------------
// set_spot_quote
//-------------------------------------------------------	
void fx_curve.set_spot_quote(number bid_quote, number ask_quote) 		
{ 
	fx_spot_bid_ = fx_spot_bid_.set_quote(bid_quote,false) ;
	fx_spot_ask_ = fx_spot_ask_.set_quote(ask_quote,false) ;
}
/*
void fx_curve.set_quotes(vector(number) quotes, instr_type option(nullable) type,logical all) 			
{ 
	vector(instrument) fx_i = fx_inst(type, all, true);
	QL_FAIL_COND(v_size(fx_i) != v_size(quotes),"invalid size of quotes vector");
	
	for(integer i=0;i<v_size(fx_i);i++) {

		if(fx_i[i].is_fx_fwd()){					
			fx_i[i] = dynamic_cast<fxfwd>(fx_i[i]).set_quote(quotes[i],false) ;
		}
		else if(fx_i[i].is_fx_swap())
			fx_i[i] = dynamic_cast<fxswap>(fx_i[i]).set_quote(quotes[i],false) ;
		else{			
			string s = strcat(["invalid instrument in fx curve (must be fx swap or fx fwd) [",fx_i[i].name(),"]"]);
			QL_FAIL(s);			
		}
	}
	
}*/

date 			fx_curve.trade_date()				{ return fx_spot_bid_.trade_date();}
integer 		fx_curve.scale()					{ return fx_spot_bid_.scale();}
date 			fx_curve.spot_settle()				{ return fx_spot_bid_.settle_date();}
integer 		fx_curve.pip_decimals()				{ return fx_spot_bid_.pip_decimals();}
vector(logical)	fx_curve.s_pre_l()	option(hidden)	{ return s_pre_l_;}
vector(logical)	fx_curve.f_pre_l()	option(hidden)	{ return f_pre_l_;}
vector(logical)	fx_curve.post_l()	option(hidden)	{ return post_l_;}

//-------------------------------------------------------
// orig_instruments
//-------------------------------------------------------
vector(instrument)	fx_curve.orig_instruments(logical all)
{
	if(all){
		return clone_vector_deep(fx_i_);
	}
	else {
		vector(logical)	in = orig_incl();
		integer size = v_size(in);
		vector(instrument) res[0];
		QL_FAIL_COND(size != v_size(fx_i_),"vector size mismatch");//should not happen

		for(integer i = 0; i < size; i++){
			if(in[i])
				push_back(res, fx_i_[i]);
		}
		return res;
	}
}
//-------------------------------------------------------
// orig_fx_swaps
//-------------------------------------------------------
vector(fx_swap)	fx_curve.orig_fx_swaps(logical all)
{
	vector(fx_swap) res[0];
	if(all) {
		for(integer i = 0; i < v_size(fx_i_); i++){
			if(fx_i_[i].instr_type() == instr_type.FX_SWAP){
				push_back(res, fx_swap(fx_i_[i]));
			}		
		}
	}
	else {
		vector(logical)	in = orig_incl();
		integer size = v_size(in);
		QL_FAIL_COND(size != v_size(fx_i_),"vector size mismatch");//should not happen
		for(integer i = 0; i < size; i++){
			if(fx_i_[i].instr_type() == instr_type.FX_SWAP && in[i]){
				push_back(res, fx_swap(fx_i_[i]));				
			}		
		}
	}
	return res;
}
//-------------------------------------------------------
// orig_fx_fwds
//-------------------------------------------------------
vector(fx_fwd)	fx_curve.orig_fx_fwds(logical all)
{
	vector(fx_fwd) res[0];
	if(all) {
		for(integer i = 0; i < v_size(fx_i_); i++){
			if(fx_i_[i].instr_type() == instr_type.FX_FORWARD){
				push_back(res, fx_fwd(fx_i_[i]));
			}		
		}
	}
	else {
		vector(logical)	in = orig_incl();
		integer size = v_size(in);
		QL_FAIL_COND(size != v_size(fx_i_),"vector size mismatch");//should not happen
		for(integer i = 0; i < size; i++){
			if(fx_i_[i].instr_type() == instr_type.FX_FORWARD && in[i]){
				push_back(res, fx_fwd(fx_i_[i]));				
			}		
		}
	}
	return res;
}
//-------------------------------------------------------
// orig_incl
//-------------------------------------------------------
vector(logical)	fx_curve.orig_incl()
{
	if(pre_is_swap_)
		return concat(s_pre_l_,post_l_);
	else
		return concat(f_pre_l_,post_l_);
}
/*
void fx_curve.set_fx_spot(instrument fx_i)
{
	check_inst_type(fx_i);

	if(fx_i.is_fx_fwd()) {
		fxfwd fx	= fxfwd(fx_i) ;
		fx_spot_	= fx.fx_spot();
	}
	else {
		fxswap fx = fxswap(fx_i);
		fx_spot_	= fx.fx_spot();
	}
	
}*/
//-------------------------------------------------------
// s_pre
//-------------------------------------------------------
vector(instrument) fx_curve.s_pre(logical all)
{
	if(all) 
		return s_pre_;
	vector(integer) i = v_find(s_pre_l_, [true]);
	return sub_vector(s_pre_, i);
}
//-------------------------------------------------------
// f_pre
//-------------------------------------------------------
vector(instrument) fx_curve.f_pre(logical all)
{
	if(all) 
		return f_pre_;
	vector(integer) i = v_find(f_pre_l_, [true]);
	return sub_vector(f_pre_, i);
}
//-------------------------------------------------------
// post
//-------------------------------------------------------
vector(instrument) fx_curve.post(logical all)
{
	if(all) 
		return post_;
	vector(integer) i = v_find(post_l_, [true]);
	return sub_vector(post_, i);
}
//-------------------------------------------------------
// update_s_pre
//-------------------------------------------------------
void fx_curve.update_s_pre(vector(instrument) fx_s_pre) option(hidden)
{
	QL_FAIL_COND(v_size(s_pre_) != v_size(fx_s_pre),"vector size mismatch");
	
	for(integer i=0; i < v_size(s_pre_); i++) {
		QL_FAIL_COND(!fx_s_pre[i].is_fx_swap(),"invalid instrument type");	
		s_pre_[i] = fx_s_pre[i];
	}
}
//-------------------------------------------------------
// update_swap_post
//-------------------------------------------------------
void fx_curve.update_swap_post(vector(instrument) fx_post) option(hidden)
{
	QL_FAIL_COND(v_size(post_) != v_size(fx_post),"vector size mismatch");
	
	for(integer i=0; i < v_size(post_); i++) {
		if(fx_post[i].is_fx_swap())	
			post_[i] = fx_post[i];
	}
}
//-------------------------------------------------------
// fx_inst_pre_spot
//-------------------------------------------------------
vector(instrument) fx_curve.fx_inst_pre_spot(	instr_type option(nullable) t, 
												logical 					all,
												logical 					ref) 	
{ 	
	vector(instrument) fx;

	if(null(t))			
		fx = pre_is_swap_ ? s_pre(all) : f_pre(all);		
	else if(t == instr_type.FX_FORWARD)
		fx = f_pre(all);
	else if(t == instr_type.FX_SWAP)
		fx = s_pre(all);
	else
		return null<vector(instrument)>;

	return ref ? fx : clone_vector_deep(fx);
}
//-------------------------------------------------------
// fx_inst_post_spot
//-------------------------------------------------------
vector(instrument) fx_curve.fx_inst_post_spot(	instr_type option(nullable) t, 
												logical 					all,
												logical 					ref) 	
{ 
	vector(instrument) fx;

	if(null(t))
		fx = post(all);
	else {			
		if(all) {
			vector(integer) i = v_find(type_post(), t);
			fx = sub_vector(post_, i);
		}
		else {
			vector(integer) i 				= v_find(post_l_, [true]);
			vector(instrument) post_sub 	= sub_vector(post_, i);
			vector(instr_type) type_sub 	= sub_vector(type_post(), i);
			i 								= v_find(type_sub, t);
			fx = sub_vector(post_sub, i);
		}			
	}

	return ref ? fx : clone_vector_deep(fx);
}
//-------------------------------------------------------
// fx_inst
//-------------------------------------------------------
vector(instrument) fx_curve.fx_inst(logical ref)
{
	if(ref) {
		vector(instrument) fx 	= pre_is_swap_ ? s_pre(true) : f_pre(true);
		fx 	= !null(fx) ? clone_vector(fx) : null<vector(instrument)>;
		vector(instrument) fxp 	= post(true);
 		fxp = !null(fxp) ? clone_vector(fxp) : null<vector(instrument)>;
		return concat(fx, fxp);
	}
	else {
		vector(instrument) fx = pre_is_swap_ ? clone_vector_deep(s_pre(true)) : clone_vector_deep(f_pre(true));
		return concat(fx, clone_vector_deep(post(true)));
	}
}
//-------------------------------------------------------
// fx_inst
//-------------------------------------------------------
vector(instrument) fx_curve.fx_inst(	instr_type option(nullable) 	type, 
										logical 						all,
										logical 						ref) 	
{ 
	vector(instrument)	fx_pre 	= fx_inst_pre_spot(type, all,ref); 
	vector(instrument)	fx_post = fx_inst_post_spot(type, all,ref);
	return concat(fx_pre,fx_post);
}
//-------------------------------------------------------
// maturity
//-------------------------------------------------------
vector(date)  fx_curve.maturity(instr_type option(nullable) t,
								logical all)
{ 	
	if(!all)
		return fx_inst(t,all,true).maturity();

	//if all we would like to nullify maturity for not included instruments

	vector(instrument) fx = fx_inst(t,true,true);
	vector(logical) fx_l 	= inst_in_curve(t);
	vector(date) m 			= fx.maturity();

	for(integer i = 0; i<v_size(fx) ; i++){
		if(!fx_l[i])
			m[i] = null<date>;		
	} 
	
	return m;

}
//-------------------------------------------------------
// quotes
//-------------------------------------------------------
vector(number) fx_curve.quotes(instr_type option(nullable) type,
							   logical all)
{ 
	return fx_inst(type,all,true).fx_quote() ;
}

//-------------------------------------------------------
// settle_date
//-------------------------------------------------------
vector(date)  fx_curve.settle_date(instr_type option(nullable) t,
								   logical all)
{ 	
	if(!all)
		return fx_inst(t,all,true).settle_date();

	//if all we would like to nullify maturity for not included instruments

	vector(instrument) fx = fx_inst(t,true,true);
	vector(logical) fx_l 	= inst_in_curve(t);
	vector(date) s 			= fx.settle_date();

	for(integer i = 0; i<v_size(fx) ; i++){
		if(!fx_l[i])
			s[i] = null<date>;		
	} 
	
	return s;
}
//-------------------------------------------------------
// inst_in_curve
//-------------------------------------------------------
vector(logical) fx_curve.inst_in_curve(instr_type option(nullable) t)
{
	vector(logical) fx_l = pre_is_swap_ ? s_pre_l_ : f_pre_l_;
	fx_l = concat(fx_l, post_l_);

	if(null(t))
		return fx_l;
	else {	
		vector(integer) i = v_find(type_, t);
		return sub_vector(fx_l, i);	
	}
}
//-------------------------------------------------------
// spot_quote
//-------------------------------------------------------
number fx_curve.spot_quote(logical unit_quote,
						   logical no_rnd,
						   logical rev)
{
	if(spot_qs_ == quote_side_code.QS_BID)
		return rev ? fx_spot_ask_.fx_quote(unit_quote,no_rnd):fx_spot_bid_.fx_quote(unit_quote,no_rnd);	
	else if(spot_qs_ == quote_side_code.QS_ASK)
		return rev ? fx_spot_bid_.fx_quote(unit_quote,no_rnd):fx_spot_ask_.fx_quote(unit_quote,no_rnd);
	else 
		return (fx_spot_bid_.fx_quote(unit_quote,no_rnd) + fx_spot_ask_.fx_quote(unit_quote,no_rnd))/2.0;	
}


//**************************************
//swap quote output
//**************************************
//-------------------------------------------------------
// swaps_pre_spot
//-------------------------------------------------------
void fx_curve.swaps_pre_spot(	logical 					unit_quote,
								instr_type option(nullable) type,
								logical 					no_rnd,
								logical						all,
								out vector(date)  			mat,
								out vector(number)  		quote,
								logical 					rev_bidask)
{	
	swaps_pre_spot(	unit_quote,type,no_rnd,all,mat,quote, null<vector(instrument)>, rev_bidask);
}

/*void fx_curve.swaps_pre_spot(	vector(instrument) fx_f_pre_bid,
								vector(instrument) fx_f_pre_ask,
								logical 			unit_quote,
								instr_type option(nullable) type,
								logical 			no_rnd,
								logical 			calc_bid,
								logical				all,
								logical				flip_bidask_pre_spot,
								logical				flip_sign_neg_bidask_sprd,
								out vector(date)  	mat,
								out vector(number)  quote)
{	
	vector(instrument) fx_f_pre;

	if((null(type) ||  (!null(type) && type == INST_FX_FWD)) && flip_bidask_pre_spot){
		for(integer i=0; i< v_size(fx_f_pre_bid); i++) {

			
			if(calc_bid)
				push_back(fx_f_pre,fx_f_pre_ask[i]);
			else
				push_back(fx_f_pre,fx_f_pre_bid[i]);
		}
	}
	swaps_pre_spot(	unit_quote,type,no_rnd,all,mat,quote, fx_f_pre, flip_bidask_pre_spot);
	
	if(flip_sign_neg_bidask_sprd) {
		for(integer i=0; i< v_size(fx_f_pre_bid); i++) {
			if(quote[i] < 0) quote[i] *= -1;
		}
	}
}*/
//-------------------------------------------------------
// swaps_pre_spot
//-------------------------------------------------------
void fx_curve.swaps_pre_spot(	logical 					unit_quote,
								instr_type option(nullable) type,
								logical 					no_rnd,
								logical						all,
								out vector(date)  			mat,
								out vector(number)   		quote,
								vector(instrument) option(nullable) fx_f_pr,
								logical 					rev_bidask)
{	
	logical incl_fx_swaps = null(type) ||  (!null(type) && type == instr_type.FX_SWAP);
	logical incl_fx_fwds = null(type) ||  (!null(type) && type == instr_type.FX_FORWARD);

	if(incl_fx_fwds) {
		if(null(fx_f_pr))
			fx_f_pr = f_pre(all);
		else if(v_size(f_pre(all)) != v_size(fx_f_pr))
			QL_FAIL("size mismatch (internal error)");	
	}

	resize(mat, 0);
	resize(quote, 0);
	
	integer size_p_f = v_size(fx_f_pr);

	vector(instrument)	fx_s_pr = s_pre(all);
	integer size_p_s = v_size(fx_s_pr);

	if(size_p_f > 0 && incl_fx_fwds) {

						
		number fx_ref = spot_quote(false ,no_rnd, rev_bidask);
		QL_FAIL_COND(null(fx_ref),"invalid fx reference quote");
		// we are matching swap bid with spot bid --> 
		// i.e. a bid quotes vector will have ask side swaps before spot

		for(integer i = size_p_f-1; i >= 0; i--){

			number f_q 	= fx_f_pr[i].fx_quote(false ,no_rnd);
			date f_mat 	= fx_f_pr[i].maturity();					
			number f_s 	= fx_fwd_to_swap(f_mat, spot_settle(), fx_ref, f_q, pip_decimals());										
			QL_FAIL_COND(null(f_s),strcat(["invalid fx swap (from fwd outright) [",fx_f_pr[i].name(),"]"]));			
			fx_ref 		= fx_spot_adjust(fx_ref,-f_s,pip_decimals(), 1, false);
			if(unit_quote) f_s /= scale();

			push_back(quote,f_s);	
			push_back(mat,f_mat);						   					
		}
	}
	else if(size_p_s > 0 && incl_fx_swaps) {
		/*if(fx_s_pre[size_p_s-1].maturity() != spot_settle)
			throw(E_INVALID_ARG, "missing fx swap (before spot)"); -> already checked*/
				
		for(integer i = size_p_s-1; i >= 0; i--){
			
			number f = fx_s_pr[i].fx_quote(unit_quote, no_rnd);
			QL_FAIL_COND(null(f),strcat(["invalid fx swap [",fx_s_pr[i].name(),"]"]));	
						
			push_back(quote,f);	
			push_back(mat,fx_s_pr[i].settle_date());
			//spot_adj = null(spot_adj) ? fx_s_pre_[i].quote(false,no_rnd) : spot_adj + fx_s_pre_[i].quote(false,no_rnd);	//OBS		
		}			
	}

	sort(quote, mat);	
	sort(mat, mat);		
}
//-------------------------------------------------------
// swaps_post_spot
//-------------------------------------------------------
void fx_curve.swaps_post_spot(	logical 					unit_quote,
								instr_type option(nullable) type,
								logical 					no_rnd,
								logical						all,
								out vector(date)  			mat,
								out vector(number)  		quote)
{
	swaps_post_spot(unit_quote,type,no_rnd,all,mat,quote, null<vector(instrument)>);
}

/*void fx_curve.swaps_post_spot(	vector(instrument) fx_post_bid,
								vector(instrument) fx_post_ask,
								logical 			unit_quote,
								instr_type option(nullable) type,
								logical 			no_rnd,
								logical 			calc_bid,
								logical				all,								
								logical				flip_sign_neg_bidask_sprd,
								out vector(date)  	mat,
								out vector(number)  quote)
{	
	
	if(calc_bid)
		swaps_post_spot(unit_quote,type,no_rnd,all,mat,quote, fx_post_bid);
	else
		swaps_post_spot(unit_quote,type,no_rnd,all,mat,quote, fx_post_ask);

	if((null(type) ||  (!null(type) && type == INST_FX_FWD)) && flip_sign_neg_bidask_sprd){		
		for(integer i=0; i< v_size(fx_post_bid); i++) {
			if(fx_post_bid[i].is_fx_fwd()){ 
				if(quote[i] < 0) quote[i] *= -1;
			}
		}
	}

}*/


//-------------------------------------------------------
// swaps_post_spot
//-------------------------------------------------------
void fx_curve.swaps_post_spot(	logical 					unit_quote,
								instr_type option(nullable) type,
								logical 					no_rnd,
								logical						all,
								out vector(date) 			mat,
								out vector(number)  		quote,
								vector(instrument) option(nullable) fx)
{	
	logical incl_fx_swaps = null(type) ||  (!null(type) && type == instr_type.FX_SWAP);
	logical incl_fx_fwds = null(type) ||  (!null(type) && type == instr_type.FX_FORWARD);

	resize(mat, 0);
	resize(quote, 0);
	//vector(instrument) fx = post(all);

	if(null(fx))
		fx = post(all);
	else if(v_size(post(all)) != v_size(fx))
		QL_FAIL("size mismatch (internal error)");	
	
				
	for(integer i = 0; i < v_size(fx); i++){
		
		if(!incl_fx_fwds && fx[i].instr_type() == instr_type.FX_FORWARD)
			continue;
		
		if(!incl_fx_swaps && fx[i].instr_type() == instr_type.FX_SWAP)
			continue;		
 
		number s;
		if(fx[i].instr_type() == instr_type.FX_FORWARD){ 
			number f_q  = fx[i].fx_quote(false ,no_rnd);
								
			if(spot_qs_ == quote_side_code.QS_BID)
				s	= fx_fwd_to_swap(fx[i].maturity(), spot_settle(), fx_spot_bid_.fx_quote(false ,no_rnd), f_q, pip_decimals());
			else if(spot_qs_ == quote_side_code.QS_ASK)
				s	= fx_fwd_to_swap(fx[i].maturity(), spot_settle(), fx_spot_ask_.fx_quote(false ,no_rnd), f_q, pip_decimals());
			else {
				number f_s_b	= fx_fwd_to_swap(fx[i].maturity(), spot_settle(), fx_spot_bid_.fx_quote(false ,no_rnd), f_q, pip_decimals());	
				number f_s_a	= fx_fwd_to_swap(fx[i].maturity(), spot_settle(), fx_spot_ask_.fx_quote(false ,no_rnd), f_q, pip_decimals());
				s	= (f_s_b + f_s_a)/2.0;
			}
			if(unit_quote)
				s /= scale() ;
		}
		else {
			s = fx[i].fx_quote(unit_quote ,no_rnd);
		}
		QL_FAIL_COND(null(s),strcat(["invalid fx swap [",fx[i].name(),"]"]));	
			
		push_back(quote,s);	
		push_back(mat,fx[i].maturity());				
	}		
}
//-------------------------------------------------------
// swaps
//-------------------------------------------------------
void fx_curve.swaps(	logical 			unit_quote,
						instr_type option(nullable) type,
						logical 			no_rnd,					
						logical 			incl_spot,
						logical				all,
						out vector(date)  	mat,
						out vector(number)   quote)
{	
	swaps_pre_spot(	unit_quote,type,no_rnd,all,mat,quote);

	if(incl_spot) {
		push_back(quote, spot_quote(unit_quote, no_rnd));	
		push_back(mat, spot_settle());
	}

	vector(date) 	mat_post;
	vector(number) 	quote_post;
	swaps_post_spot(unit_quote,type,no_rnd,all,mat_post,quote_post);

	mat		= concat(mat, mat_post);
	quote	= concat(quote, quote_post);		
}

//**************************************
//outright quote output
//**************************************
//-------------------------------------------------------
// outrights_pre_spot
//-------------------------------------------------------
void fx_curve.outrights_pre_spot(	logical 					unit_quote,
									instr_type option(nullable) type,
									logical 					no_rnd,
									logical						all,
									out vector(date)  	mat_fwd,
									out vector(number)  quote_fwd)
{	
	outrights_pre_spot(	unit_quote,type,no_rnd,all,mat_fwd,quote_fwd, null<vector(instrument)>);
}

module CORE_INT
{
	/*-----------------------------------------------------------------------
	  Name :  	fx_swap_input_flip
	  Notes :   handles a possible reverse of bid ask sides   
	  Changes:
	  			Date    		Who     Detail
	  			12-Mar-2014 	MN      Original code 
	  ----------------------------------------------------------------------*/	
	void fx_swap_input_flip(fx_swap 	fx_bid_in,
							fx_swap 	fx_ask_in,
							logical 	pre_spot,
							logical option(nullable)	flip_ba_pre_spot, //normally true
							logical	option(nullable)	flip_sign_neg_sprd, //normally true, only applicable if both bid and ask is positive numbers					
							out fx_swap fx_bid, 
							out fx_swap fx_ask)
	{
		if(null(flip_ba_pre_spot))
			flip_ba_pre_spot=true;
		
		if(null(flip_sign_neg_sprd))
			flip_sign_neg_sprd=true;
		
		number b = fx_bid_in.quote_orig();
		number a = fx_ask_in.quote_orig(); 
		
		if(null(a) && null(b)) {
			fx_bid = fx_bid_in;
			fx_ask = fx_ask_in;						
			return;
		}
		
		logical ab_valid = !null(a) && !null(b);
		
		if(ab_valid && (b < 0 || a < 0) && b > a) {
			if(flip_ba_pre_spot && pre_spot) {
				fx_bid = fx_ask_in;
				fx_ask = fx_bid_in;
			}
			else {
				fx_bid = fx_bid_in;
				fx_ask = fx_ask_in;	
			}					
			return;
		}
		
		logical flip_sign = flip_sign_neg_sprd && ab_valid && b > a;
		if(flip_sign) {
			if(pre_spot) {
				if(flip_ba_pre_spot) {
					fx_bid = fx_ask_in.set_quote(-a,false);
					fx_ask = fx_bid_in.set_quote(-b,false);
				}
				else {
					fx_bid = fx_bid_in.set_quote(-b,false);
					fx_ask = fx_ask_in.set_quote(-a,false);
				}
			}
			else {
				fx_bid = fx_bid_in.set_quote(-b,false);
				fx_ask = fx_ask_in.set_quote(-a,false);
			}
		}

		if(pre_spot && flip_ba_pre_spot) {//only O/N and T/N if spot is t+2
			/*
			 ex 1
			 on	-10 +1	(b<a)
			 tn	-20 -18  (b<a)
			 ex 2
			 on	10 8	(b>a)
			 tn	20 18  (b>a)
			 ex 3
			 on	-8	-10	(b>a) 	--> NOT OK		 
			 tn -18	-20 (b>a)	--> NOT OK
			 ex 4
			 on	+1 -10	(b>a)	--> NOT OK
			 tn	-20 -18  (b<a)	--> NOT OK
			 */						
			fx_bid = fx_ask_in;
			fx_ask = fx_bid_in;
														
		}
		else {						
			fx_bid = fx_bid_in;
			fx_ask = fx_ask_in;	
									 			
		}			
	}

	/*-----------------------------------------------------------------------
	  Name :  	fx_swap_output_flip
	  Notes :   handles a possible reverse of bid ask sides   
	  Changes:
	  			Date    		Who     Detail
	  			12-Mar-2014 	MN      Original code 
	  ----------------------------------------------------------------------*/	
	void fx_swap_output_flip(	out number 	fx_s_bid,
								out number 	fx_s_ask,
								logical 	pre_spot,
								logical 	flip_ba_pre_spot, 
								logical		flip_sign_neg_sprd)
	{														
		number b = fx_s_bid;
		number a = fx_s_ask; 
		
		if(null(a) || null(b))  					
			return;
		

		logical flip_sign = flip_sign_neg_sprd && a <= 0 && b <= 0 ;

		if(pre_spot) {//only O/N and T/N if spot is t+2			
			if(flip_sign) {
				if(flip_ba_pre_spot) {
					fx_s_bid = -a;
					fx_s_ask = -b;
				}
				else {
					fx_s_bid = -b;
					fx_s_ask = -a;
				}
			}
			else {
				if(flip_ba_pre_spot) {
					fx_s_bid = a;
					fx_s_ask = b;
				}				
			}
									
		}
		else {			
			if(flip_sign) {
				fx_s_bid = -b;
				fx_s_ask = -a;
			}									 			
		}			
	}

}

/*void fx_curve.outrights_pre_spot(	vector(instrument) 	fx_s_pre_bid,
									vector(instrument) 	fx_s_pre_ask,
									logical 				unit_quote,
									instr_type option(nullable) type,
									logical 				no_rnd,
									logical 				calc_bid,
									logical					all,
									logical					flip_bidask_pre_spot,
									logical					flip_sign_neg_bidask_sprd,
									out vector(date) 	mat_fwd,
									out vector(number) 	quote_fwd)
{	
	vector(instrument) fx_s_pre;

	if((null(type) ||  (!null(type) && type == INST_FX_SWAP)) && (flip_bidask_pre_spot || flip_sign_neg_bidask_sprd)){
		for(integer i=0; i< v_size(fx_s_pre_bid); i++) {
			
			fx_swap  fx_s_bid, fx_s_ask;
			//CORE_INT.fx_swap_flip(	dynamic_cast<fx_swap  >(fx_s_pre_bid[i]), dynamic_cast<fx_swap >(fx_s_pre_ask[i]),
				//					true, flip_bidask_pre_spot, flip_sign_neg_bidask_sprd, fx_s_bid, fx_s_ask);
			CORE_INT.fx_swap_input_flip(	fx_swap(fx_s_pre_bid[i]), fx_swap(fx_s_pre_ask[i]),
									true, flip_bidask_pre_spot, flip_sign_neg_bidask_sprd, fx_s_bid, fx_s_ask);

			if(calc_bid)
				push_back(fx_s_pre,fx_s_bid);
			else
				push_back(fx_s_pre,fx_s_ask);
		}
	}
	outrights_pre_spot(	unit_quote,type,no_rnd,all,mat_fwd,quote_fwd, fx_s_pre);

}
*/
//-------------------------------------------------------
// outrights_pre_spot
//-------------------------------------------------------
void fx_curve.outrights_pre_spot(	logical 					unit_quote,
									instr_type option(nullable) type,
									logical 					no_rnd,
									logical						all,
									out vector(date) 			mat_fwd,
									out vector(number)  		quote_fwd,
									vector(instrument) option(nullable) fx_s_pr)
{	
	logical incl_fx_swaps = null(type) ||  (!null(type) && type == instr_type.FX_SWAP);
	logical incl_fx_fwds = null(type) ||  (!null(type) && type == instr_type.FX_FORWARD);

	if(incl_fx_swaps) {
		if(null(fx_s_pr))
			fx_s_pr = s_pre(all);
		else if(v_size(s_pre(all)) != v_size(fx_s_pr))
			QL_FAIL("size mismatch (internal error)");	
	}

	resize(mat_fwd, 0);
	resize(quote_fwd, 0);
	
	vector(instrument) fx_f_pr 	= f_pre(all);
	integer size_p_f 				= v_size(fx_f_pr);
	integer size_p_s 				= v_size(fx_s_pr);

	if(size_p_f > 0 && incl_fx_fwds) {
		
		for(integer i = size_p_f-1; i >= 0; i--){
			number f = fx_f_pr[i].fx_quote(unit_quote,no_rnd);
			QL_FAIL_COND(null(f),strcat(["invalid outright [",fx_f_pr[i].name(),"]"]));			
			push_back(quote_fwd,f);	
			push_back(mat_fwd,fx_f_pr[i].maturity());				
		}
	}
	else if(size_p_s > 0 && incl_fx_swaps) {
		/*if(fx_s_pre[size_p_s-1].maturity() != spot_settle)
			throw(E_INVALID_ARG, "missing fx swap (before spot)"); -> already checked*/
							
		// we are matching swap bid with spot bid --> the vector of instruments should have flipped bid/ask (if before spot)
		// i.e. a bid curve should have ask side swaps before spot
		number fx_ref = spot_quote(false ,no_rnd);			
		QL_FAIL_COND(null(fx_ref),"invalid fx reference quote");

		for(integer i = size_p_s-1; i >= 0; i--){
						
			date mat 	= fx_s_pr[i].maturity();
			number s_q 	= fx_s_pr[i].fx_quote(false,no_rnd);
			QL_FAIL_COND(null(s_q),"invalid fx swap quote");
			number s_f 	= fx_swap_to_fwd(mat , spot_settle(), fx_ref, s_q, pip_decimals());			
			QL_FAIL_COND(null(s_f),strcat(["invalid outright (from fx swap) [",fx_s_pr[i].name(),"]"]));
			if(unit_quote ) s_f /= scale()	;
			fx_ref 		= fx_spot_adjust(fx_ref,-s_q, pip_decimals(), 1, false);	

			date settle = fx_s_pr[i].settle_date();
			push_back(quote_fwd,s_f);	
			push_back(mat_fwd,settle );							
		}			
	}

	sort(quote_fwd, mat_fwd);	
	sort(mat_fwd, mat_fwd);	
}
//-------------------------------------------------------
// outrights_post_spot
//-------------------------------------------------------
void fx_curve.outrights_post_spot(	logical 					unit_quote,
									instr_type option(nullable) type,
									logical 					no_rnd,
									logical						all,
									out vector(date)  			mat_fwd,
									out vector(number)  		quote_fwd)
{
	outrights_post_spot(unit_quote,type,no_rnd,all,mat_fwd, quote_fwd, null<vector(instrument)> );	
}

/*void fx_curve.outrights_post_spot(	vector(instrument) 	fx_all_bid,
									vector(instrument) 	fx_all_ask,
									logical 				unit_quote,
									instr_type option(nullable) type,
									logical 				no_rnd,
									logical 				calc_bid,
									logical					all,									
									logical					flip_sign_neg_bidask_sprd,
									out vector(date) 	mat_fwd,
									out vector(number) 	quote_fwd)
{
	vector(instrument) fx_all;

	if((null(type) ||  (!null(type) && type == INST_FX_SWAP)) && flip_sign_neg_bidask_sprd){	
		for(integer i=0; i< v_size(fx_all_bid); i++) {
		
			if(fx_all_bid[i].is_fx_swap()) {
				fx_swap  fx_s_bid, fx_s_ask;
				CORE_INT.fx_swap_input_flip(	fx_swap(fx_all_bid[i]), fx_swap(fx_all_ask[i]),
										false, false, flip_sign_neg_bidask_sprd, fx_s_bid, fx_s_ask);
	
				if(calc_bid)
					push_back(fx_all,fx_s_bid);
				else
					push_back(fx_all,fx_s_ask);
			}
			else {
				if(calc_bid)
					push_back(fx_all,fx_all_bid[i]);
				else
					push_back(fx_all,fx_all_ask[i]);
			}
		}
	}

	outrights_post_spot(unit_quote,type,no_rnd,all,mat_fwd, quote_fwd, fx_all );
}
*/
//-------------------------------------------------------
// outrights_post_spot
//-------------------------------------------------------
void fx_curve.outrights_post_spot(	logical 					unit_quote,
									instr_type option(nullable) type,
									logical 					no_rnd,
									logical						all,
									out vector(date)  			mat_fwd,
									out vector(number)  		quote_fwd,
									vector(instrument) option(nullable) fx)
{	
	logical incl_fx_swaps = null(type) ||  (!null(type) && type == instr_type.FX_SWAP);
	logical incl_fx_fwds = null(type) ||  (!null(type) && type == instr_type.FX_FORWARD);

	resize(mat_fwd, 0);
	resize(quote_fwd, 0);
	//vector(instrument)fx 	= post(all);
	
	
	if(null(fx))
		fx = post(all);
	else if(v_size(post(all)) != v_size(fx))
		QL_FAIL("size mismatch (internal error)");	
	

	for(integer i = 0; i < v_size(fx); i++){
		
		if(!incl_fx_fwds && fx[i].instr_type() == instr_type.FX_FORWARD)
			continue;
		
		if(!incl_fx_swaps && fx[i].instr_type() == instr_type.FX_SWAP)
			continue;

		number f ;
		if(fx[i].instr_type() == instr_type.FX_SWAP) {
			number s_q  = fx[i].fx_quote(false ,no_rnd);
			QL_FAIL_COND(null(s_q),"invalid fx swap quote");
			if(spot_qs_ == quote_side_code.QS_BID)
				f 	= fx_swap_to_fwd(fx[i].maturity(), spot_settle(), fx_spot_bid_.fx_quote(false ,no_rnd), s_q, pip_decimals());
			else if(spot_qs_ == quote_side_code.QS_ASK)
				f 	= fx_swap_to_fwd(fx[i].maturity(), spot_settle(), fx_spot_ask_.fx_quote(false ,no_rnd), s_q, pip_decimals());
			else {
				date m 		= fx[i].maturity();
				date s 		= spot_settle();
				integer d 	= pip_decimals();
				number f_a 	= fx_swap_to_fwd(m, s, fx_spot_bid_.fx_quote(false ,no_rnd), s_q, d);
				number f_b 	= fx_swap_to_fwd(m, s, fx_spot_ask_.fx_quote(false ,no_rnd), s_q, d);
				f			= (f_b + f_a)/2.0;
			}
			if(unit_quote) f /= scale();				
		}
		else  
			f = fx[i].fx_quote(unit_quote ,no_rnd);
		

		QL_FAIL_COND(null(f),strcat(["invalid outright [",fx[i].name(),"]"]));
			
		push_back(quote_fwd,f);	
		push_back(mat_fwd,fx[i].maturity());				
	}		
}
//-------------------------------------------------------
// outrights
//-------------------------------------------------------
void fx_curve.outrights(logical 			unit_quote,
						instr_type option(nullable) type,
						logical 			no_rnd,
						logical 			incl_spot,
						logical				all,
						out vector(date)  	mat_fwd,
						out vector(number)  quote_fwd)
{	
	outrights_pre_spot(	unit_quote,type,no_rnd,all,mat_fwd,quote_fwd);

	if(incl_spot) {
		push_back(quote_fwd, spot_quote(unit_quote, no_rnd));	
		push_back(mat_fwd, spot_settle());
	}

	vector(date) 	mat_fwd_post;
	vector(number) 	quote_fwd_post;
	outrights_post_spot(unit_quote,type,no_rnd,all,mat_fwd_post,quote_fwd_post);

	mat_fwd 	= concat(mat_fwd, mat_fwd_post);
	quote_fwd 	= concat(quote_fwd, quote_fwd_post);			
}

//-------------------------------------------------------
// quotes_df_pre_spot
//-------------------------------------------------------
void fx_curve.quotes_df_pre_spot(	disc_func 			df_base,
									disc_func 			df_price,
									logical 			unit_quote,
									instr_type option(nullable) type,
									logical 			no_rnd,
									logical				all,
									out vector(date)  	mat,
									out vector(number)  quote)
{	
	logical incl_fx_swaps = null(type) ||  (!null(type) && type == instr_type.FX_SWAP);
	logical incl_fx_fwds = null(type) ||  (!null(type) && type == instr_type.FX_FORWARD);
	
	resize(mat, 0);
	resize(quote, 0);
	
	vector(instrument) fx_f_pr 	= f_pre(all);
	integer size_p_f 			= v_size(fx_f_pr);
	vector(instrument) fx_s_pr 	= s_pre(all);
	integer size_p_s 			= v_size(fx_s_pr);	

	if(size_p_f > 0 && incl_fx_fwds) {

		number fx_spot = spot_quote(false ,no_rnd);
		QL_FAIL_COND(null(fx_spot),"invalid fx spot quote");

		for(integer i = size_p_f-1; i >= 0; i--){
			number f = fx_df_to_fwd(trade_date(),fx_f_pr[i].maturity(),spot_settle(),fx_spot,df_base,df_price);

			QL_FAIL_COND(null(f),strcat(["invalid outright (from disc_func) [",fx_f_pr[i].name(),"]"]));

			if(unit_quote) f /= scale();
		
			push_back(quote,f);	
			push_back(mat,fx_f_pr[i].maturity());								
		}
	}
	else if(size_p_s > 0 && incl_fx_swaps) {
		
		number fx_ref = spot_quote(false ,no_rnd);

		for(integer i = size_p_s-1; i >= 0; i--){

			number s = fx_df_to_swap(	trade_date(),fx_s_pr[i].settle_date(), fx_s_pr[i].maturity(),spot_settle(),fx_ref,
										df_base,df_price,pip_decimals());			
			QL_FAIL_COND(null(s),strcat(["invalid fx swap (from disc_func) [",fx_s_pr[i].name(),"]"]));

			fx_ref 	= fx_spot_adjust(fx_ref,-s, pip_decimals(), 1, false);			
			if(unit_quote) s /= scale();

			push_back(quote,s);	
			push_back(mat,fx_s_pr[i].maturity());					
		}			
	}

	sort(quote, mat);	
	sort(mat, mat);	
}

//-------------------------------------------------------
// quotes_df_post_spot
//-------------------------------------------------------
void fx_curve.quotes_df_post_spot(	disc_func 			df_base,
									disc_func 			df_price,
									logical 			unit_quote,
									instr_type option(nullable) type,
									logical 			no_rnd,
									logical				all,
									out vector(date)  	mat,
									out vector(number)  quote)
{	
	logical incl_fx_swaps = null(type) ||  (!null(type) && type == instr_type.FX_SWAP);
	logical incl_fx_fwds = null(type) ||  (!null(type) && type == instr_type.FX_FORWARD);

	resize(mat, 0);
	resize(quote, 0);
	vector(instrument)	fx = post(all);

	number fx_spot = spot_quote(false ,no_rnd);
	QL_FAIL_COND(null(fx_spot),"invalid fx spot quote");

	for(integer i = 0; i < v_size(fx); i++){

		try {
			number f;
			if(incl_fx_fwds && fx[i].instr_type() == instr_type.FX_FORWARD) {			
				f = fx_df_to_fwd(trade_date(),fx[i].maturity(),spot_settle(),fx_spot,df_base,df_price);    			
			}				
			else if(incl_fx_swaps && fx[i].instr_type() == instr_type.FX_SWAP) {
				//f = fx_df_to_swap(trade_date(),fx[i].maturity(),spot_settle(),fx_spot,df_base,df_price);  

				f = fx_df_to_swap(	trade_date(),fx[i].settle_date(), fx[i].maturity(),spot_settle(),fx_spot,
											df_base,df_price,pip_decimals());
				//f = fx[i].swap(df_base,df_price,unit_quote,null<number>,no_rnd);
			}	
			if(unit_quote)
				f /= scale();

			push_back(quote,f);	
		
		}
		catch {
			push_back(quote,null<number>);	
		}
					
		push_back(mat,fx[i].maturity());				
	}		
}
//-------------------------------------------------------
// included_post_spot
//-------------------------------------------------------
vector(logical) fx_curve.included_pre_spot(instr_type option(nullable) type)
{
	logical incl_fx_swaps = null(type) ||  (!null(type) && type == instr_type.FX_SWAP);
	logical incl_fx_fwds = null(type) ||  (!null(type) && type == instr_type.FX_FORWARD);

	vector(instrument) fx_f_pr 	= f_pre(true);
	integer size_p_f 			= v_size(fx_f_pr);
	vector(instrument) fx_s_pr 	= s_pre(true);
	integer size_p_s 			= v_size(fx_s_pr);
	vector(logical) res[0];
	
	if(size_p_f > 0 && incl_fx_fwds) {
		resize(res,size_p_f);
		res = false;
		for(integer i = 0; i < size_p_f; i++){
			res[i] = f_pre_l_[i];
		}
		
	}
	else if(size_p_s > 0 && incl_fx_swaps) {
		resize(res,size_p_s);
		res = false;
		for(integer i = 0; i < size_p_s; i++){
			res[i] = s_pre_l_[i];
		}
	}
	return res;
}
//-------------------------------------------------------
// included_post_spot
//-------------------------------------------------------
vector(logical) fx_curve.included_post_spot(instr_type option(nullable) type)
{
	logical incl_fx_swaps 	= null(type) ||  (!null(type) && type == instr_type.FX_SWAP);
	logical incl_fx_fwds	= null(type) ||  (!null(type) && type == instr_type.FX_FORWARD);
	
	vector(instrument)	fx 	= post(true);
	integer size 			= v_size(fx);
	vector(logical) res[size];
	res = false;
	
	for(integer i = 0; i < size; i++){
		if(incl_fx_fwds && fx[i].instr_type() == instr_type.FX_FORWARD) {			
			res[i] = post_l_[i];
		}				
		else if(incl_fx_swaps && fx[i].instr_type() == instr_type.FX_SWAP) {
			res[i] = post_l_[i];
		}
	}
	return res;
}
//-------------------------------------------------------
// included
//-------------------------------------------------------
vector(logical) fx_curve.included(instr_type option(nullable) type)
{
	vector(logical) pre = included_pre_spot(type);
	vector(logical) post = included_post_spot(type);
	return concat(pre, post);
}

//-------------------------------------------------------
// quotes_df
//-------------------------------------------------------
vector(number) fx_curve.quotes_df(	disc_func df_base,
									disc_func df_price,
									logical all)
{	
	vector(date) 	mat;
	vector(number) 	quote;
	fx_curve.quotes_df(df_base,df_price,false,null<instr_type>,false,all,mat,quote);
	return quote;	
}
//-------------------------------------------------------
// quotes_df
//-------------------------------------------------------
void fx_curve.quotes_df(disc_func 			df_base,
						disc_func 			df_price,
						logical 			unit_quote,
						instr_type option(nullable) type,
						logical 			no_rnd,	
						logical				all,					
						out vector(date)  	mat,
						out vector(number)  quote)
{	
	quotes_df_pre_spot(df_base,df_price,unit_quote,type,no_rnd,all,mat,quote);	

	vector(date) 	mat_post;
	vector(number) 	quote_post;
	quotes_df_post_spot(df_base,df_price,unit_quote,type,no_rnd,all,mat_post,quote_post);

	mat		= concat(mat, mat_post);
	quote	= concat(quote, quote_post);		
}

//-------------------------------------------------------
// disc
//-------------------------------------------------------
disc_func fx_curve.disc(	disc_func 			df,
							logical 			df_is_price,
							interpolator 		ip,
							rate_type			ip_rt,
							logical 			no_rnd,							
							out vector(date)  	mat_fwd,
							out vector(number)  quote_fwd)
{		
	logical all = false;				
	outrights(true, null<instr_type>,no_rnd,true, all,mat_fwd, quote_fwd);

	number size = v_size(quote_fwd);
	vector(number)	v_df;
	vector(number)	v_per;
	
	date d = trade_date();
	
	number dff;
	if(mat_fwd[0] > d){
		number p 	= (mat_fwd[0] - d)/365;
		dff 		= df.disc(p);//approx
		push_back(v_df, dff);
		push_back(v_per, p);	
	}
	else 
		dff = 1;

	date sp = spot_settle();
	
	for(integer i = 1; i < size; i++){

		number fx_start = quote_fwd[i-1];
		number fx_end 	= quote_fwd[i];
		
		if(mat_fwd[i] <= spot_settle()) 
			CORE_INT.flip (fx_start, fx_end) ;

		if(df_is_price) 
			dff *= fx_fwd_to_df_base(d,mat_fwd[i-1], mat_fwd[i],sp,fx_start, fx_end,df );
		else
			dff *= fx_fwd_to_df_price(d,mat_fwd[i-1], mat_fwd[i],sp,fx_start, fx_end,df);
		if(!null(dff)){
			push_back(v_df, dff);
			push_back(v_per, (mat_fwd[i] - d)/365);	
		}
	}		
	
	return disc_func_interp(v_per, v_df, ip, ip_rt);	
}
//-------------------------------------------------------
// disc_base
//-------------------------------------------------------
disc_func fx_curve.disc_base(	disc_func 	df_price,										
								interpolator ip,
								rate_type	ip_rt,
								logical 	no_rnd)
{	
	vector(date) 	mat_fwd;
	vector(number) 	quote_fwd;			
	return disc(df_price,true,ip,ip_rt, no_rnd,mat_fwd, quote_fwd)	;	
}
//-------------------------------------------------------
// disc_price
//-------------------------------------------------------
disc_func fx_curve.disc_price(	disc_func 	df_base,										
								interpolator ip,
								rate_type	ip_rt,
								logical 	no_rnd)
{	
	vector(date) 	mat_fwd;
	vector(number) 	quote_fwd;			
	return disc(df_base,false,ip,ip_rt, no_rnd,mat_fwd, quote_fwd)	;	
}
//-------------------------------------------------------
// ip_outright
//-------------------------------------------------------
number fx_curve.ip_outright(disc_func 	df_base,
							disc_func 	df_price,
							date 		mat,
							logical 	unit_quote)
{	
	logical no_rnd = false;
	
	number fx_sp = spot_quote(unit_quote, no_rnd);									
	return fx_df_to_fwd(trade_date(), mat, spot_settle(), fx_sp, df_base, df_price);	
}
//-------------------------------------------------------
// ip_swap_post_spot
//-------------------------------------------------------
number fx_curve.ip_swap_post_spot(	disc_func 	df_base,
									disc_func 	df_price,									
									date 		mat,
									logical 	unit_quote)
{	
	logical no_rnd = false;
	
	number fx_sp = spot_quote(unit_quote, no_rnd);									
	return fx_df_to_swap(trade_date(), spot_settle(),mat, spot_settle(), fx_sp, df_base, df_price, pip_decimals());	
}
//-------------------------------------------------------
// ip_swap_post_spot_ff
// separate disc_funcs for near and far to allow for bid and ask calculation
//-------------------------------------------------------
number fx_curve.ip_swap_post_spot_ff(disc_func option(nullable)	df_base_near,
									 disc_func option(nullable)	df_price_near,
									 number option(nullable) sw_near,
									 disc_func option(nullable)	df_base_far,
									disc_func option(nullable)	df_price_far,
									 number option(nullable) sw_far,
									date 		near,
									date 		far,
									logical 	unit_quote)
{	
	
	date spot = spot_settle();
	QL_REQUIRE(spot < near,"spot settlement must be before near date");
	QL_REQUIRE(near < far,"near date must be before far date");

	if(null(sw_near)) {
		QL_REQUIRE(!null(df_base_near),"df_base_near required when sw_near not provided");
		QL_REQUIRE(!null(df_price_near),"df_price_near required when sw_near not provided");
		sw_near = ip_swap_post_spot(df_base_near, df_price_near,near,unit_quote);
	}
	if(null(sw_far)) {
		QL_REQUIRE(!null(df_base_far),"df_base_near required when sw_far not provided");
		QL_REQUIRE(!null(df_price_far),"df_price_near required when sw_far not provided");
		sw_far = ip_swap_post_spot(df_base_far, df_price_far,far,unit_quote);
	}

	return sw_far - sw_near;
	//number fx_near = ip_outright(df_base, df_price,near,false);
	//return fx_df_to_swap(trade_date(), near,far, spot_settle(), fx_near, df_base, df_price, pip_decimals());	
}
//-------------------------------------------------------
// synt_depo
//-------------------------------------------------------
curve fx_curve.synt_depo(	disc_func 			df,	
							logical 			df_is_price,
							interpolator 		ip,
							rate_type			ip_rt,
							day_count_method 	dc,
							logical				no_rnd,
							date option(nullable) max,
							out vector(date) 	excl_mat)
{				
	resize(excl_mat,0);
	vector(date) 	mat_fwd;
	vector(number) 	quote_fwd;
	disc_func df_calc = disc(df,df_is_price,ip,ip_rt, no_rnd,mat_fwd, quote_fwd);
	
	date d 				= trade_date();
	//date spot_settle 	= spot_settle();
	string synt_ccy 	= df_is_price ? fx_spot_bid_.base_ccy(): fx_spot_bid_.price_ccy();
	vector(instrument) vi_d;

	//before spot	
	vector(instrument) fx_f_pr 	= f_pre(false);
	integer size_p_f 				= v_size(fx_f_pr);
	vector(instrument) fx_s_pr 	= s_pre(false);
	integer size_p_s 				= v_size(fx_s_pr);

	instr_def i_d = create_depo_instr_def(dc, BD_MOD_FOLLOWING, true, "bd2", calendar(),synt_ccy);

	if(size_p_f > 0) {
		//the depos created here should be mkt maturitities eg. if we have [fwdt=t] and a [fwdt=t+1] and we have spot=t+2
		//--> the depos created should be an O/N and a T/N
		// for this to work we need to look at the fwds in reverse order		
		//first fwd used is the last one i.e. the one just before spot 
		date m 			= spot_settle();
		date s 			= fx_f_pr[size_p_f-1].maturity();
		number rate 	= df_calc.simple_rate(d, s, m, dc) ;		
		//instrument depo = create_depo_instr(i_d,fx_f_pr[size_p_f-1].name(),d,s, s, m, rate, rate);
		string name = strcat(fx_f_pr[size_p_f-1].name(),"-ImpliedRate");
		instrument depo = deposit(i_d,name,d,s, s, m,null<string>, rate, rate);

		if(!null(depo))
			push_back(vi_d, depo);
		else
			push_back(excl_mat, s);

		for(integer i = size_p_f-2; i >=0 ; i--){//we loop backwards
			m 				= fx_f_pr[i+1].maturity();
			s 				= fx_f_pr[i].maturity();
			rate 			= df_calc.simple_rate(d, s, m, dc) ;			
			//instrument depo = create_depo_instr(i_d, fx_f_pr[i].name(),d,s, s, m, rate, rate);
			string name = strcat(fx_f_pr[i].name(),"-ImpliedRate");
			instrument depo = deposit(i_d, name,d,s, s, m,null<string>, rate, rate);

			if(!null(depo))
				push_back(vi_d, depo);
			else
				push_back(excl_mat, s);
		}		
	}
	else if(size_p_s > 0) {
		for(integer i = 0; i < size_p_s; i++){
			date m 			= fx_s_pr[i].maturity();
			date s 			= fx_s_pr[i].settle_date();
			number rate 	= df_calc.simple_rate(d, s, m, dc) ;
			//instrument depo = create_depo_instr(i_d,fx_s_pr[i].name(),d,s, s, m, rate, rate);
			string name = strcat(fx_s_pr[i].name(),"-ImpliedRate");
			instrument depo = deposit(i_d,name,d,s, s, m, null<string>,rate, rate);
			if(!null(depo))
				push_back(vi_d, depo);
			else
				push_back(excl_mat, m);
		}
	}

	//after spot
	vector(instrument) fx = post(false);

	for(integer i = 0; i < v_size(fx); i++){
		number rate;
		try {
			rate = df_calc.simple_rate(d, spot_settle(), fx[i].maturity(), dc) ;
		}
		catch {
			rate = null<number>;
		}
		if(!null(max) && fx[i].maturity() > max && null(rate))
			continue;

		//instrument depo = create_depo_instr(i_d,fx[i].name(),d,spot_settle(), spot_settle(), fx[i].maturity(), rate, rate);
		string name = strcat(fx[i].name(),"-ImpliedRate");
		
		instrument depo = deposit(i_d,name,d,spot_settle(), spot_settle(), fx[i].maturity(),null<string>, rate, rate);
		if(!null(depo))
			push_back(vi_d, depo);
		else
			push_back(excl_mat, fx[i].maturity());
	}
	
	return curve(vi_d);	
}
//-------------------------------------------------------
// synt_depo_base
//-------------------------------------------------------
curve fx_curve.synt_depo_base(	disc_func 			df_price,
								interpolator 		ip,
								rate_type			ip_rt,
								day_count_method 	dc,
								out vector(date) 	excl_mat,
								logical				no_rnd,
								date option(nullable) max)
{	
	return synt_depo(df_price,true,ip,ip_rt,dc, no_rnd,max,excl_mat);					
}
//-------------------------------------------------------
// synt_depo_price
//-------------------------------------------------------
curve fx_curve.synt_depo_price(disc_func 			df_base,
								interpolator 		ip,
								rate_type			ip_rt,
								day_count_method 	dc,
								out vector(date) 	excl_mat,
								logical				no_rnd,
								date option(nullable) max)
{	
	return synt_depo(df_base,false,ip,ip_rt,dc,no_rnd,max,excl_mat);					
}

/*-----------------------------------------------------------------------
  Name :  	class fx_curve_bidask
  Notes :   This curve can contain a mix of either fx swaps or fx fwds with one restriction that before spot it must be 
			either fx swaps or fx fwds. The purpose of this class is to produce for example fx implied disc_funcs, synthetic deposits,
			interpolated fwd fx quotes (interpolation is done via disc_funcs i.e. no interpolation directly on fx fwd quotes) , ...       
  Changes:
  			Date    		Who     Detail
  			11-Dec-2013 	MN      Original code (from fiws but redesigned)
			29-Aug-2017		MN		converted to new fx class
  ----------------------------------------------------------------------*/

	
class fx_curve_bidask option (category: 'Financial/FX')
{
   	public:
		
		fx_curve_bidask(fx_curve, fx_curve, logical option(nullable), logical option(nullable),
						logical option(nullable) flip_o=null<logical>);	//vector(instrument
		fx_curve_bidask(vector(instrument), vector(instrument), fx_spot,fx_spot, logical option(nullable),
						logical option(nullable),logical option(nullable) flip_o=null<logical>);	//vector(instrument
		fx_curve_bidask(fx_curve_bidask );
		fx_curve 	fx_curve_bid();
		fx_curve 	fx_curve_ask();

		void 		outrights(	logical ,instr_type option(nullable) ,logical,logical ,out vector(date)  ,						
								out vector(number)  ,out vector(number)  );

		void 		swaps(	logical ,instr_type option(nullable) ,logical,logical ,out vector(date)  ,						
							out vector(number)  ,out vector(number)  );

		void		quotes_df(	disc_func,disc_func,disc_func,disc_func,logical ,instr_type option(nullable) ,logical ,
								logical,out vector(date) ,out vector(number)  ,
								out vector(number)  );

	protected:
		void		init();
		void 		init_pre_spot();
		void 		init_post_spot();

		fx_curve 	fx_curve_bid_;
		fx_curve 	fx_curve_ask_;
		//fx_curve_bid_/fx_curve_ask_ will contain proper bid/ask curves i.e. with pre-spot adjustments and negative bidask-spread adjustments for fx swaps
		logical		flip_ba_pre_spot_;
		logical		flip_neg_ba_sprd_in_;
		logical		flip_neg_ba_sprd_out_;
};

fx_curve_bidask.fx_curve_bidask(fx_curve_bidask f)
							  : fx_curve_bid_(f.fx_curve_bid_), fx_curve_ask_(f.fx_curve_ask_),
								flip_ba_pre_spot_(f.flip_ba_pre_spot_),flip_neg_ba_sprd_in_(f.flip_neg_ba_sprd_in_),
								flip_neg_ba_sprd_out_(f.flip_neg_ba_sprd_out_)

{}

fx_curve_bidask.fx_curve_bidask(fx_curve 	curve_bid, 
								fx_curve 	curve_ask,
								logical	option(nullable) flip_bidask_pre_spot,
								logical	option(nullable) flip_sign_neg_bidask_sprd_in,
								logical	option(nullable) flip_sign_neg_bidask_sprd_out)
								: fx_curve_bid_(curve_bid), fx_curve_ask_(curve_ask), 
								  flip_ba_pre_spot_(flip_bidask_pre_spot), flip_neg_ba_sprd_in_(flip_sign_neg_bidask_sprd_in) , 
								  flip_neg_ba_sprd_out_(flip_sign_neg_bidask_sprd_out) 
{
	if(null(flip_ba_pre_spot_))
		flip_ba_pre_spot_ = true;
	if(null(flip_neg_ba_sprd_in_))
		flip_neg_ba_sprd_in_ = true;
	if(null(flip_neg_ba_sprd_out_))
		flip_neg_ba_sprd_out_ = false;
	init();	
}
fx_curve_bidask.fx_curve_bidask(vector(instrument) 	fx_instruments_bid,
								vector(instrument) 	fx_instruments_ask,
								fx_spot 				fx_spot_bid,
								fx_spot 				fx_spot_ask,
								logical	option(nullable) flip_bidask_pre_spot,
								logical	option(nullable) flip_sign_neg_bidask_sprd_in,
								logical	option(nullable) flip_sign_neg_bidask_sprd_out)
								: flip_ba_pre_spot_(flip_bidask_pre_spot), flip_neg_ba_sprd_in_(flip_sign_neg_bidask_sprd_in) , 
								  flip_neg_ba_sprd_out_(flip_sign_neg_bidask_sprd_out)  
{
	if(null(flip_ba_pre_spot_))
		flip_ba_pre_spot_ = true;
	if(null(flip_neg_ba_sprd_in_))
		flip_neg_ba_sprd_in_ = true;
	if(null(flip_neg_ba_sprd_out_))
		flip_neg_ba_sprd_out_ = false;
	
	fx_curve_bid_ = new fx_curve(fx_instruments_bid,fx_spot_bid,fx_spot_ask, quote_side_code.QS_BID);
	fx_curve_ask_ = new fx_curve(fx_instruments_ask,fx_spot_bid,fx_spot_ask, quote_side_code.QS_ASK);

	init();

}

void fx_curve_bidask.init()
{
	vector(instrument) fx_ib = fx_curve_bid_.fx_inst_pre_spot(null<instr_type>); 
	vector(instrument) fx_ia = fx_curve_ask_.fx_inst_pre_spot(null<instr_type>);
	QL_FAIL_COND(v_size(fx_ib) != v_size(fx_ia),"invalid pre spot instrument vector (bid curve and ask curve not equal size)");

	for(integer i=0;i<v_size(fx_ib);i++) 
		QL_FAIL_COND(fx_ib[i].instr_type() != fx_ia[i].instr_type(),"invalid pre spot instrument vector (instrument mismatch)");	

	fx_ib = fx_curve_bid_.fx_inst_post_spot(null<instr_type>); 
	fx_ia = fx_curve_ask_.fx_inst_post_spot(null<instr_type>);
	QL_FAIL_COND(v_size(fx_ib) != v_size(fx_ia),"invalid instrument vector (bid curve and ask curve not equal size)");

	for(integer i=0;i<v_size(fx_ib);i++) 
		QL_FAIL_COND(fx_ib[i].instr_type() != fx_ia[i].instr_type(),"invalid instrument vector (instrument mismatch)");

	
	QL_FAIL_COND(v_size(fx_curve_bid_.s_pre_l()) != v_size(fx_curve_ask_.s_pre_l()),"invalid pre spot instrument vector (swaps)");
	QL_FAIL_COND(v_size(fx_curve_bid_.f_pre_l()) != v_size(fx_curve_ask_.f_pre_l()),"invalid pre spot instrument vector (forwards)");
	QL_FAIL_COND(v_size(fx_curve_bid_.post_l()) != v_size(fx_curve_ask_.post_l()),"invalid instrument vector");

	init_pre_spot();
	init_post_spot();
}

fx_curve 	fx_curve_bidask.fx_curve_bid() { return fx_curve_bid_;}
fx_curve 	fx_curve_bidask.fx_curve_ask() { return fx_curve_ask_;}

void fx_curve_bidask.init_pre_spot()
{	
	if(!flip_ba_pre_spot_ && !flip_neg_ba_sprd_in_) return;

	vector(instrument) fx_i_bid = fx_curve_bid_.fx_inst_pre_spot(null<instr_type>,true,true);
	vector(instrument) fx_i_ask = fx_curve_ask_.fx_inst_pre_spot(null<instr_type>,true,true);

	/*fx_swap fx_s_ask = fx_swap(fx_i_ask[0]);
	fx_swap fx_s_bid = fx_swap(fx_i_bid[0]);
	fx_i_bid[0] = fx_s_ask;	
	fx_i_ask[0] = fx_s_bid;	*/
	
	logical is_upd = false;

	for(integer i=0; i< v_size(fx_i_bid); i++) {
		
		if(fx_i_bid[i].is_fx_swap()) {
			if(!fx_curve_bid_.s_pre_l()[i])
				continue;
			fx_swap  fx_s_bid, fx_s_ask;	
			CORE_INT.fx_swap_input_flip(fx_swap(fx_i_bid[i]), fx_swap(fx_i_ask[i]),true, 
										flip_ba_pre_spot_, flip_neg_ba_sprd_in_, fx_s_bid, fx_s_ask);

			fx_i_bid[i] = fx_s_bid;
			fx_i_ask[i] = fx_s_ask;
			is_upd 		= true;
		}
	}
	if(is_upd) {		
		fx_curve_bid_.update_s_pre(fx_i_bid );
		fx_curve_ask_.update_s_pre(fx_i_ask);
	}

	//vector(instrument) fx_i_bid2 = fx_curve_bid_.fx_inst_pre_spot(null<instr_type>);
	//vector(instrument) fx_i_ask2 = fx_curve_ask_.fx_inst_pre_spot(null<instr_type>);	
}

void fx_curve_bidask.init_post_spot()
{
	if(!flip_neg_ba_sprd_in_)
		return;
	
	vector(instrument) fx_i_bid = fx_curve_bid_.fx_inst_post_spot(null<instr_type>,true,true);
	vector(instrument) fx_i_ask = fx_curve_ask_.fx_inst_post_spot(null<instr_type>,true,true);

	logical is_upd = false;

	for(integer i=0; i< v_size(fx_i_bid); i++) {
	
		if(fx_i_bid[i].is_fx_swap()) {
			if(!fx_curve_bid_.post_l()[i])
				continue;
			fx_swap  fx_s_bid, fx_s_ask;
			CORE_INT.fx_swap_input_flip(fx_swap(fx_i_bid[i]), fx_swap(fx_i_ask[i]),false,
										false, flip_neg_ba_sprd_in_, fx_s_bid, fx_s_ask);

			fx_i_bid[i] = fx_s_bid;
			fx_i_ask[i] = fx_s_ask;
			is_upd 		= true;
		}
	}
	if(is_upd) {
		fx_curve_bid_.update_swap_post(fx_i_bid );
		fx_curve_ask_.update_swap_post(fx_i_ask);
	}		
}
/*-----------------------------------------------------------------------
  outrights
  ----------------------------------------------------------------------*/
void fx_curve_bidask.outrights(	logical 								unit_quote,
								instr_type option(nullable) 			type,
								logical 								no_rnd,
								logical 								incl_spot,						
								out vector(date)  		mat_fwd,
								out vector(number)  	quote_fwd_bid,
								out vector(number)  	quote_fwd_ask)

{	
	fx_curve_bid().outrights(unit_quote,type,no_rnd,incl_spot,false,mat_fwd,quote_fwd_bid);
	fx_curve_ask().outrights(unit_quote,type,no_rnd,incl_spot,false,mat_fwd,quote_fwd_ask);	
}

/*-----------------------------------------------------------------------
  swaps
  ----------------------------------------------------------------------*/
void fx_curve_bidask.swaps(	logical 					unit_quote,
							instr_type option(nullable) type,
							logical 					no_rnd,
							logical 					incl_spot,					
							out vector(date)  			mat,
							out vector(number)  		quote_bid,
							out vector(number)  		quote_ask)
{
	
	fx_curve_bid().swaps_pre_spot(unit_quote,type,no_rnd,false,  mat,quote_bid, false/*flip_ba_pre_spot_*/);
	fx_curve_ask().swaps_pre_spot(unit_quote,type,no_rnd,false,  mat,quote_ask, false/*flip_ba_pre_spot_*/);

	CORE_INT.fx_swap_output_flip(quote_bid,quote_ask,true,flip_ba_pre_spot_,flip_neg_ba_sprd_out_);	

	if(incl_spot) {		
		push_back(quote_bid, fx_curve_bid().spot_quote(unit_quote, no_rnd));	
		push_back(quote_ask, fx_curve_ask().spot_quote(unit_quote, no_rnd));
		push_back(mat, fx_curve_bid().spot_settle());
	}
	
	vector(date) 	mat_post;
	vector(number) 	quote_bid_post,quote_ask_post;
	fx_curve_bid().swaps_post_spot(unit_quote,type,no_rnd,false,mat_post,quote_bid_post);
	fx_curve_ask().swaps_post_spot(unit_quote,type,no_rnd,false,mat_post,quote_ask_post);

	CORE_INT.fx_swap_output_flip(quote_bid_post,quote_ask_post,false,flip_ba_pre_spot_,flip_neg_ba_sprd_out_);

	mat			= concat(mat, mat_post);
	quote_bid 	= concat(quote_bid, quote_bid_post);
	quote_ask 	= concat(quote_ask, quote_ask_post);
}
/*-----------------------------------------------------------------------
  quotes_df 
  ----------------------------------------------------------------------*/
void fx_curve_bidask.quotes_df(	disc_func			df_base_bid,
								disc_func			df_base_ask,
								disc_func			df_price_bid,
								disc_func			df_price_ask,
								logical 			unit_quote,
								instr_type option(nullable) type,
								logical 			no_rnd,
								logical 			incl_spot,						
								out vector(date)  	mat,
								out vector(number)  quote_bid,
								out vector(number)  quote_ask)
{
	fx_curve_bid().quotes_df_pre_spot(df_base_ask,df_price_bid,unit_quote,type,no_rnd,false,mat,quote_bid);
	fx_curve_ask().quotes_df_pre_spot(df_base_bid,df_price_ask,unit_quote,type,no_rnd,false,mat,quote_ask);
	
	if(incl_spot) {		
		push_back(quote_bid, fx_curve_bid().spot_quote(unit_quote, no_rnd));	
		push_back(quote_ask, fx_curve_ask().spot_quote(unit_quote, no_rnd));
		push_back(mat, fx_curve_bid().spot_settle());
	}
	
	vector(date) 	mat_post;
	vector(number) 	quote_bid_post,quote_ask_post;

	fx_curve_bid().quotes_df_post_spot(df_base_ask,df_price_bid,unit_quote,type,no_rnd,false,mat_post,quote_bid_post);
	fx_curve_ask().quotes_df_post_spot(df_base_bid,df_price_ask,unit_quote,type,no_rnd,false,mat_post,quote_ask_post);
	
	mat			= concat(mat, mat_post);
	quote_bid 	= concat(quote_bid, quote_bid_post);
	quote_ask 	= concat(quote_ask, quote_ask_post);	
}

/*-----------------------------------------------------------------------
  Name :  	fx_curve_swaps
  Notes :         
  Changes:
  			Date    		Who     Detail
  			13-Mar-2014 	MN      Original code 
  ----------------------------------------------------------------------*/
void fx_curve_quotes_df(fx_curve			fx_c_bid,
						fx_curve			fx_c_ask,
						disc_func			df_base_bid,
						disc_func			df_base_ask,
						disc_func			df_price_bid,
						disc_func			df_price_ask,
						logical 			unit_quote,
						instr_type option(nullable) type,
						logical 			no_rnd,
						logical 			incl_spot,						
						out vector(date)  	mat,
						out vector(number)  quote_bid,
						out vector(number)  quote_ask,
						logical				flip_bidask_pre_spot=true,
						logical				flip_sign_neg_bidask_sprd=true)
option (category: 'Financial/FX')
{
	fx_curve_bidask fx_curve_ba = new fx_curve_bidask(fx_c_bid, fx_c_ask,flip_bidask_pre_spot,
													flip_sign_neg_bidask_sprd);	
	fx_curve_ba.quotes_df(	df_base_bid,df_base_ask,df_price_bid,df_price_ask,unit_quote,type,
							no_rnd,false,mat,quote_bid,quote_ask);

}

/*-----------------------------------------------------------------------
  Name :  	fx_curve_swaps
  Notes :         
  Changes:
  			Date    		Who     Detail
  			13-Mar-2014 	MN      Original code 
  ----------------------------------------------------------------------*/
void fx_curve_swaps(fx_curve			fx_c_bid,
					fx_curve			fx_c_ask,
					logical 			unit_quote,
					instr_type option(nullable) type,
					logical 			no_rnd,
					logical 			incl_spot,					
					out vector(date)  	mat,
					out vector(number)  quote_bid,
					out vector(number)  quote_ask,
					logical				flip_bidask_pre_spot=true,
					logical				flip_sign_neg_bidask_sprd_in=true,
					logical				flip_sign_neg_bidask_sprd_out=false)
option (category: 'Financial/FX')
{	
	fx_curve_bidask fx_curve_ba = new fx_curve_bidask(fx_c_bid, fx_c_ask,flip_bidask_pre_spot,
													flip_sign_neg_bidask_sprd_in,flip_sign_neg_bidask_sprd_out);
	fx_curve_ba.swaps(unit_quote,type,no_rnd, incl_spot,mat,quote_bid, quote_ask);
}

/*-----------------------------------------------------------------------
  Name :  	fx_curve_outrights
  Notes :         
  Changes:
  			Date    		Who     Detail
  			13-Mar-2014 	MN      Original code 
			2-Sep-2017		MN		updated with fx_curve_bidask
  ----------------------------------------------------------------------*/
void fx_curve_outrights(fx_curve			fx_c_bid,
						fx_curve			fx_c_ask,
						logical 			unit_quote,
						instr_type option(nullable) type,
						logical 			no_rnd,
						logical 			incl_spot,						
						out vector(date)  	mat_fwd,
						out vector(number)  quote_fwd_bid,
						out vector(number)  quote_fwd_ask,
						logical				flip_bidask_pre_spot=true,
						logical				flip_sign_neg_bidask_sprd=true)
option (category: 'Financial/FX')
{
	fx_curve_bidask fx_curve_ba = new fx_curve_bidask(fx_c_bid, fx_c_ask,flip_bidask_pre_spot,flip_sign_neg_bidask_sprd);
	fx_curve_ba.outrights(unit_quote,type,no_rnd,incl_spot,mat_fwd,quote_fwd_bid,quote_fwd_ask);
}

/*-----------------------------------------------------------------------
  Name :  	fx_curve_disc_base
  Notes :         
  Changes:
  			Date    		Who     Detail
  			17-Mar-2014 	MN      Original code 
			2-Sep-2017		MN		updated with fx_curve_bidask
  ----------------------------------------------------------------------*/
void fx_curve_disc_base(fx_curve		fx_c_bid,
						fx_curve		fx_c_ask,						
						disc_func		df_price_bid,
						disc_func		df_price_ask,
						interpolator 	ip,
						rate_type		ip_rt,
						logical			no_rnd,
						out disc_func  	df_bid,
						out disc_func  	df_ask,
						logical			flip_bidask_pre_spot=true,
						logical			flip_sign_neg_bidask_sprd=true)
option (category: 'Financial/FX')
{	
	fx_curve_bidask fx_curve_ba = new fx_curve_bidask(fx_c_bid, fx_c_ask,flip_bidask_pre_spot,flip_sign_neg_bidask_sprd);	
	df_ask = fx_curve_ba.fx_curve_bid().disc_base(df_price_ask,ip, ip_rt,no_rnd);
	df_bid = fx_curve_ba.fx_curve_ask().disc_base(df_price_bid,ip, ip_rt,no_rnd);		
}
/*-----------------------------------------------------------------------
  Name :  	fx_curve_disc_base
  Notes :         
  Changes:
  			Date    		Who     Detail
  			17-Mar-2014 	MN      Original code 
  ----------------------------------------------------------------------*/
disc_func fx_curve_disc_base(	fx_curve		fx_c,
								disc_func		df_price,
								interpolator 	ip,
								rate_type		ip_rt,
								logical			no_rnd)
option (category: 'Financial/FX')
{		
	return fx_c.disc_base(df_price,ip, ip_rt,no_rnd);	
}
/*-----------------------------------------------------------------------
  Name :  	fx_curve_disc_base
  Notes :         
  Changes:
  			Date    		Who     Detail
  			17-Mar-2014 	MN      Original code
			2-Sep-2017		MN		updated with fx_curve_bidask 
  ----------------------------------------------------------------------*/
void fx_curve_disc_price(	fx_curve		fx_c_bid,
							fx_curve		fx_c_ask,						
							disc_func		df_base_bid,
							disc_func		df_base_ask,
							interpolator 	ip,
							rate_type		ip_rt,
							logical			no_rnd,
							out disc_func 	df_bid,
							out disc_func 	df_ask,
							logical			flip_bidask_pre_spot=true,
							logical			flip_sign_neg_bidask_sprd=true)
option (category: 'Financial/FX')
{	
	fx_curve_bidask fx_curve_ba = new fx_curve_bidask(fx_c_bid, fx_c_ask,flip_bidask_pre_spot,flip_sign_neg_bidask_sprd);		
	df_ask = fx_curve_ba.fx_curve_ask().disc_price(df_base_ask,ip, ip_rt,no_rnd);
	df_bid = fx_curve_ba.fx_curve_bid().disc_price(df_base_bid,ip, ip_rt,no_rnd);		
}
/*-----------------------------------------------------------------------
  Name :  	fx_curve_disc_base
  Notes :         
  Changes:
  			Date    		Who     Detail
  			17-Mar-2014 	MN      Original code 
  ----------------------------------------------------------------------*/
disc_func fx_curve_disc_price(	fx_curve		fx_c,						
								disc_func		df_base,
								interpolator 	ip,
								rate_type		ip_rt,
								logical			no_rnd)
option (category: 'Financial/FX')
{		
	return fx_c.disc_price(df_base,ip, ip_rt,no_rnd);	
}
