/*	
	index
	Developer: Algorithmica Research, Magnus Nyström

	--------------------------------------------------------------------
	
	--------------------------------------------------------------------			
*/

option(null: hard);

//-------------------------------------------------------
// class index_instr
//-------------------------------------------------------
class index_instr 
{
	public:
		index_instr(instrument_name, date option(nullable) = null<date>);
		
		instrument_name			name();
		ir_index				index();
		fixing					fix();
		number 					fixing_rate(date,date,logical,logical = true);
		vector(number) 			add_fixing_quote(date ,vector(date),logical);
		void 					add_fixing(vector(date),vector(number),logical);
		void					set_trade_date(date);
		void					set_proxy_hist(number option(nullable) );
		logical					index_has_pub_delay();
	protected:
		
		instrument_name			name_;
		__instrument 			c_;
		..instr_def 			def_;
		date 					trade_;
		quote_side				qs_;
		ir_index				index_;
		fixing					fix_;		
};

//------------------------------------------------
// constructor
//------------------------------------------------
index_instr.index_instr(instrument_name	name,
						date  option(nullable) trade_date) : name_(name), trade_(trade_date)
{
	logical is_class_name;
	..instr_def def_ 	= __instr_def(name, is_class_name);
	QL_FAIL_COND(null(def_) || is_class_name , "invalid instrument" );
	
	instr_type type = CORE_INT.conv_instr_type(def_.__instr_type());	
	QL_FAIL_COND(type != ..instr_type.DEPOSIT_INDEX  , "invalid instrument type" );
	
	instr_error_type t;
	string s;
	qs_ 				= def_.__fixing_instr_qs(t, s);
	QL_FAIL_COND(t != instr_error_type.ERR_T_NONE , s );

	index_ 	= def_.__ir_index( t, s);
	QL_FAIL_COND(t != instr_error_type.ERR_T_NONE , s );
	QL_FAIL_COND(null(index_) , "invalid ir_index" );
	
	c_ 					= __instrument(name_,trade_,qs_, null<date> );
	QL_FAIL_COND(null(c_) , "null index instrument" );
		
	vector(date) dates[0];
	vector(number) rates[0];
	number proxy_rate_hist 	= null<number>;
	number proxy_rate_fwd 	= null<number>;
	logical allow_fwd 		= false;
	fix_ 					= fixing(dates, rates,proxy_rate_hist,
									 proxy_rate_fwd,allow_fwd);
}

string		index_instr.name() 										{ return name_;}
ir_index	index_instr.index()										{ return index_;}
fixing		index_instr.fix()										{ return fix_;}
void		index_instr.set_proxy_hist(number option(nullable) r)	{ fix_.set_proxy_hist(r);}
logical		index_instr.index_has_pub_delay() 						{ return index_.pub_delay_days() > 0;}

//------------------------------------------------
// set_trade_date
//------------------------------------------------
void index_instr.set_trade_date(date d)
{	
	c_ = c_.__set_date(d, null<date>,true,true );
	QL_FAIL_COND(null(c_), "null index instrument" );
	instr_error_type t;
	string s;
	logical valid = c_.__is_valid(t,s);
	QL_FAIL_COND(!valid, "invalid index instrument" );
}
//------------------------------------------------
// fixing_rate
//------------------------------------------------
number index_instr.fixing_rate(date trade_date,
							   date d,
							   logical is_pub_date,
							   logical add_if_null)
{
	number r;
	
	/*if(fix_.empty()){//could have a proxy
		r = null<number>;
		if(add_if_null && d <= trade_date){
			vector(number) _r = add_fixing_quote([d]);
			r = _r[0];			
		}

		return r;
	}*/
	
	
	if( index_has_pub_delay()) {
		if(is_pub_date) {
			r = fix_.rate(trade_date,d);
			if(add_if_null && null(r) && d <= trade_date){
				vector(number) _r = add_fixing_quote(trade_date,[d],true);
				r = _r[0];			
			}
		}
		else {
			calendar cal = index_.fixing_calendar();
			integer ds = index_.pub_delay_days();
			date pub_date = cal.move_bus_days(d,ds);
			r = fix_.rate(trade_date,pub_date);
			if(add_if_null && null(r) && pub_date <= trade_date){
				vector(number) _r = add_fixing_quote(trade_date,[pub_date],true);
				r = _r[0];			
			}
		}
	}
	else {
		r = fix_.rate(trade_date,d); 		
		if(add_if_null && null(r) && d <= trade_date){
			vector(number) _r = add_fixing_quote(trade_date,[d],true);
			r = _r[0];			
		}
	}

	
	return r;
}

//------------------------------------------------
// add_fixing_quote
//------------------------------------------------
vector(number) index_instr.add_fixing_quote(date trade_date,
											vector(date) publish_dates,//always publish date
											logical overwrite)
{
	instr_error_type t;
	string s;
	date td = c_.__trade_date(t,s);
	
	if(null(td) || td != trade_date)
		set_trade_date(trade_date);

	vector(integer) vu = v_find_unique(publish_dates);
	vector(date) pd = sub_vector(publish_dates, vu) ;
	
	vector(number) rates = c_.__quote(pd, qs_,t,s );
	if(null(rates))
		return null<vector(number)>;
	fix_.add(pd, rates, overwrite);
	return rates;
}

//------------------------------------------------
// add_fixing_quote
//------------------------------------------------
void index_instr.add_fixing(vector(date) publish_dates,
							vector(number) rates,
							logical overwrite)
{
	fix_.add(publish_dates, rates, overwrite);
}
//-----------------------------------------------------------------------
//  class index_map

//----------------------------------------------------------------------
class index_map
{
public:
	index_map();
	index_instr get_index(instrument_name);
	map_istr_obj<index_instr> map_;
};

index_map.index_map()
{
	map_ 				= new map_istr_obj<index_instr>();

	__instr_type t = CORE_INT.conv_rev_instr_type(instr_type.DEPOSIT_INDEX);
	
	vector(string) d_i 	= __instrument_search(t,true,null<date>,null<date>);
	integer size 		= v_size(d_i);
	
	for(integer i=0;i<size;i++){
		try {
			index_instr idx_i = new index_instr(d_i[i],null<date>);
			if(!null(idx_i)) {
				map_.add(d_i[i], idx_i);
			}
		}
		catch {
			continue;
		}
	}
}

index_instr index_map.get_index(instrument_name name)
{
	return map_.find(name);
}


index_map gl_index_map;

index_map get_gl_index_map()
{
	if(null(gl_index_map)) {
		gl_index_map = new index_map();
	}
	return gl_index_map;
}