/*	
	Utility funcs for instrument wrapper functions
	Developer: Algorithmica Research, Magnus Nyström			
*/

option(null: hard);	


/*string cat_curve_depo_instr	= "Curve and Instrument/Deposit";
string cat_curve_bill_instr	= "Curve and Instrument/Bill";
string cat_curve_billfut_instr	= "Curve and Instrument/Bill Future";
string cat_curve_fedfut_instr	= "Curve and Instrument/Fed Fund Future";
string cat_curve_bond_instr	= "Curve and Instrument/Bond";
string cat_curve_bondfut_instr	= "Curve and Instrument/Bond Future";
string cat_curve_frn_instr	= "Curve and Instrument/Frn";
string cat_curve_fra_instr	= "Curve and Instrument/Fra";
string cat_curve_riba_instr	= "Curve and Instrument/Riba";
string cat_curve_mmfut_instr	= "Curve and Instrument/Interest Rate Future";
string cat_curve_cpi_instr	= "Curve and Instrument/Index linked General";
string cat_curve_il_bond_instr	= "Curve and Instrument/Index Linked Bond";
string cat_curve_il_swap_instr	= "Curve and Instrument/Index Linked Swap";
string cat_curve_il_zero_instr	= "Curve and Instrument/Index Linked Zero Bond";
string cat_curve_il_zeroswap_instr = "Curve and Instrument/Index Linked Zero Swap";
string cat_curve_fx_instr	= "Curve and Instrument/FX";
string cat_curve_swap_instr	= "Curve and Instrument/Interest Rate Swap";
string cat_curve_nois_instr	= "Curve and Instrument/Nois";
string cat_curve_ois_instr	= "Curve and Instrument/Interest Rate Swap";*/


/*-----------------------------------------------------------------------
  instr_error  --> Ska byta namn till ngt annat typ err_info (klassen som samlar alla errors heter error_info) 
  Note: type_ = ERR_T_NONE means it is NOT an error  
  ----------------------------------------------------------------------*/
class instr_error  
{
public:

	instr_error();
	
	instr_error(instr_error_type option(nullable),
				string option(nullable),
				string option(nullable) l = null<string>);
	
	instr_error(error_type,
				string ,
				string option(nullable) l = null<string>);
	
    instr_error_type 	type() ;
   	string 				message();
	string 				location();
	logical				is_error();
	void				set(instr_error_type option(nullable),
							string option(nullable),
							string option(nullable) l = null<string>);
	error_type 			ql_error_type() ;
protected:
	void				init();
    instr_error_type 	type_;
   	string 				message_;
	string 				location_;

	error_type			ql_type_;//--> when error is created in a catch statement 
};

error_type instr_error.ql_error_type() 
{
	if(!null(ql_type_))
		return ql_type_;
	
	switch(type_){
	case ERR_T_ARGUMENT :
		return E_INVALID_ARG;
	case  ERR_T_INIT :
		return E_INIT;
	case  ERR_T_CALC :
		return E_CALC;
	default:
		return E_CALC;
	}
}

instr_error.instr_error(): type_(ERR_T_NONE), message_(""), location_(""), ql_type_(null<error_type>)  	{}

instr_error.instr_error(instr_error_type option(nullable) 	type,
						string option(nullable) 			s,
						string option(nullable) 			l)
						: type_(type), message_(s), location_(l), ql_type_(null<error_type>)
{
	init();
}

instr_error.instr_error(error_type  						ql_type,
						string 								s,
						string option(nullable) 			l)
						: message_(s), location_(l), ql_type_(ql_type) 
{
	switch(ql_type_){
	case E_INVALID_ARG:
		type_ = ERR_T_ARGUMENT;
		break;
	case  E_INIT:
		type_ = ERR_T_INIT;
		break;
	case  E_CALC :
	default:
		type_ = ERR_T_CALC;
		break;
	}
	init();
}

void instr_error.init()
{
	if(null(type_)) { 
		type_ = ERR_T_NONE;
		message_ = location_ = "";	
	}
	else if(is_error()) {
		if(null(message_))  
			message_ = "error message not defined";
		if(null(location_))
			location_ = "";
		
	}
	else { 
		message_ = location_ = "";		 
	}
}

void instr_error.set(instr_error_type option(nullable) 	type,
					 string option(nullable) 			s,
					 string option(nullable) 			l)   
{
	type_ 		= type;
	message_ 	= s;
	location_ 	= l;	
	init();
}

instr_error_type instr_error.type()			{ return type_; }
string instr_error.message()				{ return message_; }
string instr_error.location()				{ return location_; }
logical instr_error.is_error()				{ return type() != ERR_T_NONE;}

//----------------------------------------------------------------------
// instr_error create function                                                
//----------------------------------------------------------------------
instr_error instr_error(instr_error_type option(nullable) 	type,
						string option(nullable) 			s,
						string option(nullable) 			l)
//option(com_name: 'instr_error')
{
	return new instr_error(type, s, l);
}
//----------------------------------------------------------------------
// instr_error create function                                               
//----------------------------------------------------------------------
instr_error instr_error()
//option(com_name: 'instr_error_dflt')
{
	return new instr_error();
}

//----------------------------------------------------------------------
// set_error                                                    
//----------------------------------------------------------------------
void set_instr_error(	out instr_error 					error,
						instr_error_type option(nullable)	t,
						string option(nullable)				s)
{
	if (null(error))
		error = new instr_error(t,s);
	else
		error.set(t,s);
}

/*-----------------------------------------------------------------------
  error_info
  NOTE: will only be populated with errors i.e. instr_errors with type == ERR_T_NONE will not be added
  ----------------------------------------------------------------------*/
class error_info option(category: "Instrument")
{
public:
	error_info(logical force_throw = false);
	error_info(logical, logical);
	
	void add(instr_error);
	
	void add(instr_error_type option(nullable),
			 string option(nullable),
			 string option(nullable) l = null<string>);
	
	void add(error_type,
			 string ,
			 string option(nullable) l = null<string>);
	
	logical 			is_error();
	integer 			error_count();
	void 				reset();
	logical 			force_throw();
	logical 			single();
	
	vector(instr_error) info();
	instr_error 		info_last();
	instr_error 		info_first();
	string				message(integer i = -1);
	instr_error_type	type(integer i = -1);
	error_type 			ql_error_type(integer i = -1) ;

	void                __dbg_print(__dbg_label);
    void                __dbg_browse(__dbg_split);
	
protected:		
	vector(instr_error) err_;
	logical				force_throw_;
	logical				single_;
};

void error_info.__dbg_print(__dbg_label l)
{
    l.set_text(strcat([ "err count: ",
						string(error_count()),
                        ", last err message: ",
						message()]));
}

void error_info.__dbg_browse(__dbg_split s)
{
    s.resize(2);

    s.set_text(0, "err count: ");
    s.set_value(0, error_count());

    s.set_text(1, "last err message");
    s.set_value(1, message());

}

error_info.error_info(logical force_throw):force_throw_(force_throw), single_(false)	{ resize(err_,0);}//i.e. err_ is never null

error_info.error_info(logical single, logical force_throw):force_throw_(force_throw), single_(single)	{ resize(err_,0);}//i.e. err_ is never null

error_type error_info.ql_error_type(integer i)
{
	try {
		if(i < 0 || i >= error_count()-1){
			instr_error inl = error_info.info_last();
			return null(inl) ? null<error_type> : inl.ql_error_type();
		}
		else if(i==0){
			instr_error inf = error_info.info_first();
			return null(inf) ? null<error_type> : inf.ql_error_type();
		}
		else
			err_[i].ql_error_type();
	}
	catch {
		return null<error_type>;
	}
}

instr_error_type error_info.type(integer i)
{
	try {
		if(i < 0 || i >= error_count()-1){
			instr_error inl = error_info.info_last();
			return null(inl) ? instr_error_type.ERR_T_NONE : inl.type();
		}
		else if(i==0){
			instr_error inf = error_info.info_first();
			return null(inf) ? instr_error_type.ERR_T_NONE : inf.type();
		}
		else
			err_[i].type();
	}
	catch {
		return instr_error_type.ERR_T_NONE;
	}
}

string error_info.message(integer i)
{
	try {
		if(i < 0 || i >= error_count()-1){
			instr_error inl = error_info.info_last();
			return null(inl) ? "" : inl.message();
		}
		else if(i==0){
			instr_error inf = error_info.info_first();
			return null(inf) ? "" : inf.message();
		}
		else
			err_[i].message();
	}
	catch {
		return "";
	}
}

void error_info.add(instr_error e)
{
	if(!e.is_error()) {
		if(single_)
			reset();
		return;
	}
	
	if(single_){
		if(v_size(err_) == 0)
			push_back(err_,e);
		else
			err_[0] = e;
	}	
	else
		push_back(err_,e);
	
}

void error_info.add(instr_error_type option(nullable)	t,
					string option(nullable)				s,
					string option(nullable)				l)
{
	instr_error e = new instr_error(t,s,l);
	this.add(e);
}

void error_info.add(error_type 							ql_t,
					string 								s,
					string option(nullable)				l)
{
	instr_error e = new instr_error(ql_t,s,l);
	this.add(e);
}

void error_info.reset() 		{ resize(err_,0);}

logical error_info.force_throw() 	{ return force_throw_;}
logical error_info.single() 		{ return single_;}
logical error_info.is_error() 		{ return v_size(err_) > 0;}
integer error_info.error_count() 	{ return v_size(err_);}

vector(instr_error) error_info.info()
{
	return clone_vector(err_);
}

instr_error error_info.info_last()
{
	integer s = v_size(err_);
	if(s==0)
		return null<instr_error>;
	else
		return err_[s-1];
}

instr_error error_info.info_first()
{
	integer s = v_size(err_);
	if(s==0)
		return null<instr_error>;
	else
		return err_[0];
}

error_info error_info(logical force_throw = false)			{ return new error_info(force_throw);}
error_info error_info(logical single, logical force_throw)	{ return new error_info(single,force_throw);}


void QL_FAIL_COND(error_info e)
option(com_name: 'QL_FAIL_COND_ERR')
option (category: 'Core Programming/Error Handling')
{
	if(e.is_error()) {
		string m = e.message();
		error_type type = e.ql_error_type();
		if(null(type))
			type = E_INVALID_ARG;
		throw(type,throw_err(m,"",false));
	}
}

module CORE_INT
{
	/*-----------------------------------------------------------------------
	  reset_single
	  ----------------------------------------------------------------------*/
	void reset_single(error_info option(nullable)  error)
	{
		if(null(error)) return;
		if(error.single()) error.reset();
		return;
	}
	
	/*-----------------------------------------------------------------------
	  add_error_info
	  
	  designed to be used when a call to an instrument dll function returns an error
	  ----------------------------------------------------------------------*/
	logical add_error_info(error_info option(nullable) 		error,
						instr_error_type option(nullable)	t,
						string option(nullable)				err_msg,
						string option(nullable)				location = null<string>)
	{
		if(null(t))
			t = ERR_T_NONE;
		
		if(null(error)) {
			if(t == ERR_T_NONE)	return false;
			if(null(err_msg)) 	err_msg = "error message not defined";			
			CORE_INT.write_warning_message(err_msg);						
		}
		else {
			if(t == ERR_T_NONE) {
				if(error.single())
					error.reset();							
				return false;
			}
			
			if(error.force_throw()) {//force_throw = true --> should not call error.add(...) since it will be called in catch_error_info;
				throw(E_CALC,err_msg);
			}
			else {
				error.add(t,err_msg,location);				
			}
		}	

		return true;
	}
	/*-----------------------------------------------------------------------
	  catch_error_info
	  designed to be used inside a catch statement
	  (--> use error_type instead of instr_error_type)
	  ----------------------------------------------------------------------*/
	void catch_error_info(	error_info option(nullable)	error,
							error_type 					ql_t,
							string 						err_msg,
							string option(nullable)		location = null<string>)
	{
		if(null(error)) 
			CORE_INT.write_warning_message(err_msg);			
		else {
			error.add(ql_t,err_msg,location);
			if(error.force_throw()) 
				throw(ql_t,err_msg);			
		}
		return;
	}
}

//----------------------------------------------------------------------
// error_info_com      <used in vba>                                              
//----------------------------------------------------------------------
error_info error_info_com()	 { return new error_info(true,false);}

//----------------------------------------------------------------------
// set_error_info_com  <used in vba>                                                  
//----------------------------------------------------------------------
error_info set_error_info_com(	instr_error_type 	t,
								string 				s)
{
	error_info error = error_info_com();
	if(t == ERR_T_NONE)
		return error;
	error.add(t,s,"");	
	return error;
}


/*-----*/


logical INSTR_CREATE_NULL_ERR = false;

class enum quote_side_code {
	QS_BID option(str: "Bid") ,
	QS_ASK option(str: "Ask"),
	QS_MID option(str: "Mid")
};

module CORE_INT
{
	
	//----------------------------------------------------------------------
	// swap_instr_type                                                   
	//---------------------------------------------------------------------- 
	instr_type swap_instr_type(	ql_fixed_income_swap sw) 
	{
	
	    switch (sw.type()) {
	    case swap_type.IT_SWAP_FLOAT_FLOAT:
			return instr_type.SWAP_FLTFLT;
	    case swap_type.IT_SWAP_FIX_FLOAT:
			return instr_type.SWAP_FIXFLT;
	    case swap_type.IT_SWAP_FLOAT_OIS:
			return instr_type.SWAP_FLTOIS;
	    case swap_type.IT_SWAP_FIX_OIS:
			return instr_type.SWAP_FIXOIS;
		case swap_type.IT_SWAP_OIS_OIS:
			return instr_type.SWAP_OISOIS;
	    case swap_type.IT_SWAP_FIX_FIX:
			return instr_type.SWAP_FIXFIX;    
	    case swap_type.IT_SWAP_ZERO_FLOAT:
			return instr_type.SWAP_ZEROFLT;
	    case swap_type.IT_SWAP_ZERO_OIS:
			return instr_type.SWAP_ZEROOIS;
			
		case swap_type.IT_SWAP_FIXINFLCI_FLOAT:
			return instr_type.IL_SWAP_REAL_FIXFLT;
		case swap_type.IT_SWAP_FIXINFLCI_OIS:
			return instr_type.IL_SWAP_REAL_FIXOIS;
		case swap_type.IT_SWAP_ZERO_ZEROINFL:
			return instr_type.IL_SWAP_ZERO;
		case swap_type.IT_SWAP_FIX_YYINFL:
			return instr_type.IL_SWAP_FIX_Y_Y;
		case swap_type.IT_SWAP_FLT_YYINFL:
			return instr_type.IL_SWAP_FLT_Y_Y;
		case swap_type.IT_SWAP_OIS_YYINFL:
			return instr_type.IL_SWAP_OIS_Y_Y;
	    case swap_type.IT_SWAP_FIX_ZERO:	       
	    case swap_type.IT_SWAP_ZERO_ZERO:		
		case swap_type.IT_SWAP_ZERO_YYINFL:
		case swap_type.IT_SWAP_ZEROINFL_YYINFL:
	    default:
			QL_FAIL("unsupported swap type");
	    }	
	}
	//----------------------------------------------------------------------
	// swap_instr_type                                                   
	//---------------------------------------------------------------------- 
	instr_type swap_instr_type(	ql_swap_leg ql_swap_leg1, 							
								ql_swap_leg ql_swap_leg2) 
	{
		swap_type t1 = ql_swap_leg1.type();
		swap_type t2 = ql_swap_leg2.type();
	
	    switch (t1) {
	  	case swap_type.IT_SWAPLEG_FLOAT:
			switch (t2) {
			case swap_type.IT_SWAPLEG_FLOAT: 
			    return instr_type.SWAP_FLTFLT;
			case swap_type.IT_SWAPLEG_FIX:
			    return instr_type.SWAP_FIXFLT;
			case swap_type.IT_SWAPLEG_OIS:
			    return instr_type.SWAP_FLTOIS;
			case swap_type.IT_SWAPLEG_FIXZERO:
			    return instr_type.SWAP_ZEROFLT;

			case swap_type.IT_SWAPLEG_FIXINFLCI:
			    return instr_type.IL_SWAP_REAL_FIXFLT;
				
			default:
			    QL_FAIL("unsupported swap type");
			}	
			break;
	    case swap_type.IT_SWAPLEG_FIX:
			switch (t2) {
			case swap_type.IT_SWAPLEG_FLOAT:
			    return instr_type.SWAP_FIXFLT;
			case swap_type.IT_SWAPLEG_FIX:
			    return instr_type.SWAP_FIXFIX;
			case swap_type.IT_SWAPLEG_OIS:
			    return instr_type.SWAP_FIXOIS;
			//case swap_type.IT_SWAPLEG_FIXZERO:
			  //  return INST_ZERO_FIX_SWAP;
			default:
			    QL_FAIL("unsupported swap type");
			}
			break;
	    case swap_type.IT_SWAPLEG_OIS:
			switch (t2) {
			case swap_type.IT_SWAPLEG_FLOAT:
			    return instr_type.SWAP_FLTOIS;
			case swap_type.IT_SWAPLEG_FIX:
			    return instr_type.SWAP_FIXOIS;
			case swap_type.IT_SWAPLEG_OIS:
			    return instr_type.SWAP_OISOIS;
			case swap_type.IT_SWAPLEG_FIXZERO:	   
			    return instr_type.SWAP_ZEROOIS;
				
			case swap_type.IT_SWAPLEG_FIXINFLCI:
			    return instr_type.IL_SWAP_REAL_FIXOIS;
			default:
			    QL_FAIL("unsupported swap type");
			}
			break;
	    case swap_type.IT_SWAPLEG_FIXZERO:
			switch (t2) {
			case swap_type.IT_SWAPLEG_FLOAT:
			    return instr_type.SWAP_ZEROFLT;			
			//case swap_type.IT_SWAPLEG_FIX:	 
			    //return type_code_t::TC_SWAP_FIX_FIXZERO;
			case swap_type.IT_SWAPLEG_OIS:
			    return instr_type.SWAP_ZEROOIS;
			//case swap_type.IT_SWAPLEG_FIXZERO:
			    //return type_code_t::TC_SWAP_FIXZERO_FIXZERO;

			case swap_type.IT_SWAPLEG_ZEROINFL:
			    return instr_type.IL_SWAP_ZERO;
			default:
			    QL_FAIL("unsupported swap type");
			}
			break;
		case swap_type.IT_SWAPLEG_ZEROINFL:
			switch (t2) {
			case swap_type.IT_SWAPLEG_FIXZERO:
			    return instr_type.IL_SWAP_ZERO;			
			default:
			    QL_FAIL("unsupported swap type");
			}
			break;
		case swap_type.IT_SWAPLEG_FIXINFLCI:
			switch (t2) {
			case swap_type.IT_SWAPLEG_FLOAT:
			    return instr_type.IL_SWAP_REAL_FIXFLT;
			case swap_type.IT_SWAPLEG_OIS:
			    return instr_type.IL_SWAP_REAL_FIXOIS;	
			default:
			    QL_FAIL("unsupported swap type");
			}
			break; 
	    default:
			QL_FAIL("unsupported swap type");
	    }
	}
}


/*-----------------------------------------------------------------------
  func: frn_cash_flow_dates
  ----------------------------------------------------------------------*/
vector(date) frn_cash_flow_dates(integer 	cpn_freq, 
								date 		issue_date ,
								date option(nullable) cpn_start_date ,
								date option(nullable) first_cpn_date ,
								date option(nullable) last_reg_cpn_date,
								date  		maturity ,
								string 		roll_method,
								bd_convention bd_convention,
								calendar	calendar,
								logical 	eom,
								integer 	roll_day = 0,
								error_info option(nullable) error = null<error_info>)
option (category: 'Financial/Frn')
{	
	try{			
		vector(date) c = __frn_cash_flow_dates(	cpn_freq, issue_date, cpn_start_date, first_cpn_date, last_reg_cpn_date,
												maturity, roll_method, bd_convention, calendar, eom, roll_day);
		if(null(c)) {
			CORE_INT.add_error_info(error,ERR_T_INIT,"null vector(date)", "frn_cash_flow_dates");			
			return c;
		}
		else
			return c;						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "frn_cash_flow_dates");	
		return null<vector(date) >;
	}
}



/*-----------------------------------------------------------------------
  func: frn_fixing_dates
  ----------------------------------------------------------------------*/
void  frn_fixing_dates(	integer 	cpn_freq, 
						date 		issue_date ,
						date option(nullable) cpn_start_date ,
						date option(nullable) first_cpn_date ,
						date option(nullable) last_reg_cpn_date,
						date  		maturity ,
						string 		roll_method,
						bd_convention bd_convention,
						calendar	pmt_calendar,
						calendar	fix_calendar,
						logical 	eom,
						integer 	roll_day,
						string		fixing_code,
						logical		is_frn_rfr,
						out vector(date)   fixing_dates,
						out vector(date)   pmt_dates,
						error_info option(nullable) error = null<error_info>)
option (category: 'Financial/Frn')
{	
	try{			
		__frn_fixing_dates(	cpn_freq, issue_date, cpn_start_date, first_cpn_date, last_reg_cpn_date,
							maturity, roll_method, bd_convention, pmt_calendar, 
							fix_calendar, eom, roll_day, fixing_code,is_frn_rfr,fixing_dates,pmt_dates);		
		return ;						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "frn_fixing_dates");	
		fixing_dates = pmt_dates = null<vector(date)>;
		return ;
	}
}


/*-----------------------------------------------------------------------
  get_all_repo_data

  ----------------------------------------------------------------------*/
void get_all_repo_data(	out vector(date)  			announce_dates,
						out vector(date)  			from_dates,
						out vector(date)  			to_dates,						
						out vector(number)  		repo_rates,						
						error_info option(nullable) error = null<error_info>)
option (category: 'Instrument/Riba')
{	
	try{	
		instr_error_type t;
   		string 			s;	
		__get_all_repo_data(announce_dates,from_dates,to_dates,repo_rates,t, s);
		logical e = CORE_INT.add_error_info(error,t,s, "get_all_repo_data");		
		if(e){ 
			repo_rates = null<vector(number)>;
			announce_dates = from_dates = to_dates = null<vector(date)>;
		}		
		return ;						
	}
	catch {
		repo_rates = null<vector(number)>;
		announce_dates = from_dates = to_dates = null<vector(date)>;
		CORE_INT.catch_error_info(error,err.type(),err.message(), "get_all_repo_data");		
		return ;
	}
}


void  get_repo_rates(	out vector(date)  		announce_dates,
						out vector(date)  		from_dates,
						out vector(date)  		to_dates,						
						out vector(number)  		repo_rates) 
option(com_name: 'LEGACY_get_repo_rates') 
{
	get_all_repo_data(announce_dates,from_dates,to_dates,repo_rates);
}

//-------------------------------------------------------
//  day_count_method_list  -current
//-------------------------------------------------------
vector(string) day_count_method_list()
option (category: 'Basic Types and Functions/Date Manipulation') 
{
	vector(day_count_method) e;
	expand_enum(e) ;
	return string(e);
	
	/*return [string(DC_ACT_365), string(DC_ACT_360), string(DC_30_360G), string(DC_30_360),string(DC_30E_360),string(DC_30EP_360),
			string(DC_30D_360),string(DC_ACT_365L),string(DC_ACT_365NL),string(DC_ACT_365NL_JPN),string(DC_ACT_ACT_ISDA),string(DC_30_360_EOM),
			string(DC_30E_360_ISDA),string(DC_ACT_ACT), string(DC_ACTBD_252),string(DC_30IT_360),string(DC_ACT_ACT_AFB)];*/
}
//obsolete
vector(string) dc_string_list()
option (category: 'Obsolete Functions/Basic Types and Functions/Date Manipulation') 
{
	return day_count_method_list();
}
/*-----------------------------------------------------------------------
  Name :  	dc_dbtool_compat_map
  Notes :      
  Changes:
  			Date    		Who     Detail
  			2-Apr-2014 		MN      Original code (moved from fiws)  
  ----------------------------------------------------------------------*/
day_count_method dc_dbtool_compat_map(string dc)
{
	try {
		day_count_method e;
		str_to_enum(dc,e);
		if(null(e))
			throw(E_INVALID_ARG,"");
		return e;
	}
	catch {}
	
	switch(str_to_upper(dc)){
	case 'ACT360':
	case 'ACT/360':
		return DC_ACT_360;
	case 'ACT365':
	case 'ACT/365':
		return DC_ACT_365;
	case 'ACT/365L':
		return DC_ACT_365L;
	case 'ACT/365NL':
		return DC_ACT_365NL;
	case 'ACT/365NL-JPN':
		return DC_ACT_365NL_JPN;
	case 'EU30360':
	case '30E/360':
		return DC_30E_360;
	case '30360':
	case '30/360':
		return DC_30_360;
	case '30/360G':
		return DC_30_360G;
	case '30EP/360':
		return DC_30EP_360;
	case '30D/360':
		return DC_30D_360;
	case '30IT/360':
		return DC_30IT_360;
	case 'ACT/ACT-ISDA':
		return DC_ACT_ACT_ISDA;
	case 'ACT/ACT-AFB':
		return DC_ACT_ACT_AFB;
	case '30/360-EOM':
		return DC_30_360_EOM;
	case '30E/360-ISDA':
		return DC_30E_360_ISDA;
	case 'ACT/ACT':
		return DC_ACT_ACT;
	case 'ACTBD/252':
		return DC_ACTBD_252;
	default:
		QL_FAIL("invalid daycount method");
	}
}

day_count_method day_count_method_map(string dc)
option (category: 'Basic Types and Functions/Date Manipulation') 
{
	return dc_dbtool_compat_map(dc);
}

/*-----------------------------------------------------------------------
  Name :  	dc_string_map
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-may-2014 		MN      Original code 
  ----------------------------------------------------------------------*/
day_count_method dc_string_map(string dc)
option (category: 'Obsolete Functions/Basic Types and Functions/Date Manipulation') 
option(com_name: 'LEGACY_dc_string_map')  
{
	return dc_dbtool_compat_map(dc);
}

/*-----------------------------------------------------------------------
  Name :  	rate_type_list / rate_type_map
  Notes :           
  Changes:
  			Date    		Who     Detail
  			13-Nov-2018 	MN      
  ----------------------------------------------------------------------*/
vector(string) rate_type_list()
option (category: 'Financial') 
{
	vector(rate_type) e;
	expand_enum(e) ;
	return string(e);
}
//cont,simple,effective,semi-annual,quarterly,monthly,on,bankdisc,log_df
rate_type rate_type_map(string rt)
option (category: 'Financial') 
{
	try {
		rate_type e;
		str_to_enum(rt,e);
		if(null(e))
			throw(E_INVALID_ARG,"");
		return e;
	}
	catch {}
	
	switch(str_to_upper(rt)){	
	case 'CONT':
	case 'CONTINOUS':
		return rate_type.RT_CONT;
	case 'ON':
	case 'O/N':
		return rate_type.RT_ON;
	case 'SIMPLE':
	case 'SIMP':
		return rate_type.RT_SIMPLE;
	case 'EFFECTIVE':
	case 'EFF':
		return rate_type.RT_EFFECTIVE;
	case 'SEMI_ANNUAL':
	case 'SEMI-ANNUAL':
	case 'SEMI':
		return rate_type.RT_SEMI_ANNUAL;
	case 'QUARTERLY':
	case 'QUART':
		return rate_type.RT_QUARTERLY;
	case 'MONTHLY':
	case 'MONTH':
		return rate_type.RT_MONTHLY;
	case 'BANKDISC':
	case 'DISC':
		return rate_type.RT_BANKDISC;
	case 'LOGDF':
	case 'LOG_DF':
		return rate_type.RT_LOG_DF;
	default:
		QL_FAIL("invalid rate type");
	}
}

//-------------------------------------------------------
//  base_rate_list
//-------------------------------------------------------
vector(string) base_rate_list()
  
{ 
	return  ['Simple', 'Effective', 'O/N','BankDisc','Cont'];
}
//-------------------------------------------------------
//  base_rate_map
//-------------------------------------------------------
rate_type base_rate_map(string rt)

{
	rate_type _rt = rate_type_map(rt);
	
	switch(_rt){
	case  rate_type.RT_SIMPLE:
	case  rate_type.RT_EFFECTIVE:
	case  rate_type.RT_ON:
	case  rate_type.RT_BANKDISC:
	case  rate_type.RT_CONT:
		return _rt;
		break;
	default:
		throw(E_INVALID_ARG, throw_err("invalid rate type"));
	}
}

fx_spot_rule fx_spot_rule_map(string rule)  option (category: 'Instrument/FX')
{
	 fx_spot_rule e;
	 str_to_enum(rule,e);	 
	 return e;
}

//"std","std_usd_adj","la"
vector(string) fx_spot_rule_list()  option (category: 'Instrument/FX')
{
	 vector(fx_spot_rule) e;
	 expand_enum(e);
	 return string(e);
}

bond_fut_pvbp bond_fut_pvbp_map(string type)  option (category: 'Instrument/Bond Future')
{
	 bond_fut_pvbp e;
	 str_to_enum(type,e);	 
	 return e;
}

vector(string) bond_fut_pvbp_list()  option (category: 'Instrument/Bond Future')
{
	 vector(bond_fut_pvbp) e;
	 expand_enum(e);
	 return string(e);
}

bond_fut_style bond_fut_style_map(string style)  option (category: 'Instrument/Bond Future')
{
	 bond_fut_style e;
	 str_to_enum(style,e);	 
	 return e;
}

vector(string) bond_fut_style_list()  option (category: 'Instrument/Bond Future')
{
	 vector(bond_fut_style) e;
	 expand_enum(e);
	 return string(e);
}

bond_fut_pf bond_fut_pf_map(string pf)  option (category: 'Instrument/Bond Future')
{
	 bond_fut_pf e;
	 str_to_enum(pf,e);	 
	 return e;
}

vector(string) bond_fut_pf_list()  option (category: 'Instrument/Bond Future')
{
	 vector(bond_fut_pf) e;
	 expand_enum(e);
	 return string(e);
}

//-----------------------------------------------
// dynamic cast <FUNCTION>
//-----------------------------------------------
//disc_func disc_func(fit_result fr) 	
//option (category: 'Instrument/Bond')
//option(com_name: 'disc_func_dyncast') 
//{ 
//	disc_func d = dynamic_cast<disc_func>(fr); 
//}

/*instr_type instr_type_from_opt(integer i)
{
	switch(i)
	{
	//case 0 : return ..instr_type.IT_UNKNOWN = 0,
	case 1 : return ..instr_type.BILL;	
	case 2 : return ..instr_type.BILL_FUTURE;	
		
	case 3 : return ..instr_type.FIXED_CPN_BOND;	
	case 4 : return ..instr_type.CALLABLE_FIX_CPN_BOND;			
	case 5 : return ..instr_type.BOND_FUTURE;		
	case 6 : return ..instr_type.RATING_ADJ_BOND;		
	case 7 : return ..instr_type.ZERO_CPN_BOND;

	case 8 : return ..instr_type.FRN;
	case 9 : return ..instr_type.CALLABLE_FRN;	

	case 10 : return ..instr_type.DEPOSIT;
	case 11 : return ..instr_type.DEPOSIT_INDEX;
	case 12 : return ..instr_type.FRA;
	case 13 : return ..instr_type.IR_FUTURE;
	case 14 : return ..instr_type.OVERNIGHT_FUTURE;				
	case 15 : return ..instr_type.RIBA;	
	case 16 : return ..instr_type.NOIS;

	case 17 : return ..instr_type.FX_SPOT;
	case 18 : return ..instr_type.FX_SWAP;
	case 19 : return ..instr_type.FX_FORWARD;	

	case 20 : return ..instr_type.SWAP_FIXFLT;
	case 21 : return ..instr_type.SWAP_FIXFIX;
	case 22 : return ..instr_type.SWAP_FLTFLT;
	case 23 : return ..instr_type.SWAP_FLTFLT2S;		
	case 24 : return ..instr_type.SWAP_ZEROFLT;

	case 25 : return ..instr_type.SWAP_FIXOIS;    //--> Overnight Indexed Swap (OIS): fixed/floating irs with the float leg tied to a published overnight rate index
	case 26 : return ..instr_type.SWAP_FLTOIS;
	case 27 : return ..instr_type.SWAP_OISOIS;
	case 28 : return ..instr_type.SWAP_ZEROOIS;
	
	case 29 : return ..instr_type.IL_REAL_BILL;		//y=real rate
	case 30 : return ..instr_type.IL_REAL_FIXED_CPN_BOND;	//y=real rate
	case 31 : return ..instr_type.IL_REAL_ZERO_CPN_BOND;	//y=real rate	
	case 32 : return ..instr_type.IL_SWAP_REAL_FIXFLT;	//y=real rate, fix vs flt
	case 33 : return ..instr_type.IL_SWAP_REAL_FIXOIS;	//y=real rate, fix vs ois
	case 34 : return ..instr_type.IL_SWAP_ZERO;	//y=inflation
	case 35 : return ..instr_type.IL_SWAP_FIX_Y_Y;	//y=inflation
	case 36 : return ..instr_type.IL_SWAP_FLT_Y_Y;	//y=inflation	
	case 37 : return ..instr_type.IL_SWAP_OIS_Y_Y;	//y=inflation
	
	case 38 : return ..instr_type.EQUITY;	
	case 39 : return ..instr_type.GENERIC;
	default : return null<instr_type>;
	}

	instr_type instr_type_from_opt(instr_type i)
{
	switch(i)
	{
	case instr_type.BILL:	
	case instr_type.BILL_FUTURE:	
		
	case instr_type.FIXED_CPN_BOND:	
	case instr_type.CALLABLE_FIX_CPN_BOND:			
	case instr_type.BOND_FUTURE:		
	case instr_type.RATING_ADJ_BOND:		
	case instr_type.ZERO_CPN_BOND:

	case instr_type.FRN:
	case instr_type.CALLABLE_FRN:	

	case instr_type.DEPOSIT:
	case instr_type.DEPOSIT_INDEX:
	case instr_type.FRA:
	case instr_type.IR_FUTURE:
	case instr_type.OVERNIGHT_FUTURE:				
	case instr_type.RIBA:	
	case instr_type.NOIS:

	case instr_type.FX_SPOT:
	case instr_type.FX_SWAP:
	case instr_type.FX_FORWARD:	

	case instr_type.SWAP_FIXFLT:
	case instr_type.SWAP_FIXFIX:
	case instr_type.SWAP_FLTFLT:
	case instr_type.SWAP_FLTFLT2S:		
	case instr_type.SWAP_ZEROFLT:

	case instr_type.SWAP_FIXOIS:    
	case instr_type.SWAP_FLTOIS:
	case instr_type.SWAP_OISOIS:
	case instr_type.SWAP_ZEROOIS:
	
	case instr_type.IL_REAL_BILL:		
	case instr_type.IL_REAL_FIXED_CPN_BOND:
	case instr_type.IL_REAL_ZERO_CPN_BOND:		
	case instr_type.IL_SWAP_REAL_FIXFLT:
	case instr_type.IL_SWAP_REAL_FIXOIS:	
	case instr_type.IL_SWAP_ZERO:	
	case instr_type.IL_SWAP_FIX_Y_Y:	
	case instr_type.IL_SWAP_FLT_Y_Y:		
	case instr_type.IL_SWAP_OIS_Y_Y:	
	
	case instr_type.EQUITY:	
	case instr_type.GENERIC:
	}
}*/



string DEBUG_FX_NAME_PREFIX 	= '';

/*-----------------------------------------------------------------------
  Name :  	fx_spot_db_name
  Notes :   an altname must be checked for base and price currencies         
  Changes:
  			Date    		Who     Detail
			6-Mar-2014		MN		Original code 	
  ----------------------------------------------------------------------*/
instrument_name fx_spot_db_name(string base_ccy, 
								string price_ccy, 
								out instrument_name option(nullable) alt_name)
{
	if(equal_casei(base_ccy , "USD"))
		alt_name = strcat([DEBUG_FX_NAME_PREFIX, str_to_upper(price_ccy)]);
	else if(equal_casei(price_ccy , "USD"))
		alt_name = strcat([DEBUG_FX_NAME_PREFIX, str_to_upper(base_ccy)]);
	else
		alt_name = "";
	return strcat([DEBUG_FX_NAME_PREFIX, str_to_upper(base_ccy), str_to_upper(price_ccy)]);
}

module CORE_INT
{
	//-------------------------------------------------------
	// get_prefix_qs
	//-------------------------------------------------------
	string get_prefix_qs(string dflt, out number p)
	{	
		try {
			p = str_find_casei(dflt, "Mid");
			if(null(p)){
				p = str_find_casei(dflt, "Bid");
				if(null(p)){
					p = str_find_casei(dflt, "Ask");
					if(null(p))
						throw(E_INVALID_ARG,"");
				}
			}
			return (p == 0) ? "" : sub_string(dflt, 0, integer(p));
		}
		catch {
			p = 0;
			return "";
		}
	}
	//-------------------------------------------------------
	// get_suffix_qs
	//-------------------------------------------------------
	string get_suffix_qs(string dflt, number p)
	{	
		try {	
			integer pi = integer(p);		
			return strlen(dflt) > 3+pi ? sub_string(dflt, 3+pi, strlen(dflt)-3-pi) : ""; 
		}
		catch {
			return "";
		}
	}	
	//-------------------------------------------------------
	// quote_side_default_mid
	//-------------------------------------------------------
	string quote_side_default_mid()
	{
		try {
			string dflt = quote_side_default();
			number p;
			string prefix = get_prefix_qs(dflt,p);
			string suffix = get_suffix_qs(dflt,p); 
			return strcat([prefix,"Mid",suffix]);
		}
		catch {
			return "Mid";
		}		
	}	
	//-------------------------------------------------------
	// quote_side_default_bid
	//-------------------------------------------------------
	string quote_side_default_bid()
	{
		try {
			string dflt = quote_side_default();
			number p;
			string prefix = get_prefix_qs(dflt,p);
			string suffix = get_suffix_qs(dflt,p); 
			return strcat([prefix,"Bid",suffix]);
		}
		catch {
			return "Bid";
		}
	}
	//-------------------------------------------------------
	// quote_side_default_ask
	//-------------------------------------------------------
	string quote_side_default_ask()
	{
		try {
			string dflt = quote_side_default();
			number p;
			string prefix = get_prefix_qs(dflt,p);
			string suffix = get_suffix_qs(dflt,p); 
			return strcat([prefix,"Ask",suffix]);
		}
		catch {
			return "Ask";
		}
	}
	//----------------------------------------------------------------------
	// nose_tail                                     
	//----------------------------------------------------------------------
	public number nose_tail(number per_short, 
						number per_second, 
						number per_solve, 
						number option(nullable) days_short, 
						number option(nullable) days_second, 
						number option(nullable) days_solve,
						number rate_short, 
						number rate_second, 
						number rate_solve_code,
						rate_type rt_short, 
						rate_type rt_second, 
						rate_type rt_solve)
	
	{
	    number r;
	    //number sec_val = CORE_INT.fv(1, rate_second, per_second, days_second, rt_second);
		number sec_val = future_value(1, rate_second, per_second, rt_second);

	    //number short_val = CORE_INT.fv(1, rate_short, per_short, days_short, rt_short);
		number short_val = future_value(1, rate_short, per_short, rt_short);
	    
	    if(rate_solve_code == 1 || rate_solve_code == 2) 
			//r = CORE_INT.rate(sec_val, short_val, per_solve, days_solve, rt_solve);	
			r = implied_rate(sec_val, short_val, per_solve, rt_solve);		  
	    else  if(rate_solve_code == 3) 
			//r = CORE_INT.rate(short_val*sec_val, 1,per_solve, days_solve, rt_solve);           
			r = implied_rate(short_val*sec_val, 1,per_solve, rt_solve);
	    else
			return 0;
	    return r;
	}
	//----------------------------------------------------------------------
	// nose_tail                                     
	//----------------------------------------------------------------------
	public number nose_tail(date  settle, 
					    	date  short_mat, 
					    	date  long_mat, 
					    	number rate_short, 
							number rate_second,
					    	number rate_solve_code, 
							day_count_method dc_short, 
							day_count_method dc_second, 
							day_count_method dc_solve, 
							rate_type cm_short, 
							rate_type cm_second, 
							rate_type cm_solve)
	
	{
	    number per_short, per_second, per_solve ;
	    number days_short, days_second, days_solve;
	
	    if(rate_solve_code== 1)  {
			per_short = period (short_mat, long_mat, dc_short); 
			if(cm_short== RT_ON)
			    days_short = n_days(short_mat, long_mat, dc_short) ;
			per_solve = period (settle, short_mat, dc_solve) ;
			if(cm_solve== RT_ON)
			    days_solve = n_days(settle, short_mat, dc_solve) ;
			per_second = period ( settle, long_mat,dc_second) ;
			if(cm_second== RT_ON)
			    days_second = n_days( settle, long_mat, dc_second) ;
	    }	
	    else if(rate_solve_code== 2) {
			per_short = period (settle,short_mat, dc_short) ;
			if(cm_short== RT_ON)
			    days_short = n_days(settle,short_mat, dc_short) ;
			per_solve = period (short_mat,long_mat, dc_solve) ; 
			if(cm_solve== RT_ON)
			    days_solve = n_days(short_mat,long_mat, dc_solve) ;
			per_second = period (settle,long_mat, dc_second) ;	
			if(cm_second== RT_ON)
			    days_second = n_days( settle, long_mat, dc_second) ;
	    }
	    else  if(rate_solve_code== 3) {
			per_short = period (settle,short_mat, dc_short) ;
			if(cm_short== RT_ON)
			    days_short = n_days( settle,short_mat, dc_short) ;
			per_solve = period (settle,long_mat, dc_solve) ; 
			if(cm_solve== RT_ON)
			    days_solve = n_days(settle,long_mat, dc_solve) ;
			per_second = period (short_mat,long_mat, dc_second) ; 
			if(cm_second== RT_ON)
			    days_second = n_days(short_mat,long_mat, dc_second) ;
	    }
	    else
			return 0;
	
		return nose_tail(per_short, per_second, per_solve, days_short, days_second, days_solve,
							rate_short, rate_second, rate_solve_code, cm_short, cm_second, cm_solve);
	}

	number df_to_rate(	date 		d, 
						date 		mat1, 
						date 		mat2, 
						disc_func 	df, 
						day_count_method dc, 
						rate_type rt)
	
	{	
		number df1 = df.disc ((mat1-d)/365);
		number df2 = df.disc ((mat2-d)/365);	
		return disc_to_rate(df2/df1, rt, mat1, mat2, dc);	
	}

	//-----------------------------------------------------------------------------
	//  ilb_fwd_rate_dur                                                           
	//-----------------------------------------------------------------------------
	void ilb_fwd_rate_dur(	number 		y_r_sh,
							number 		y_r_lo,
							number 		dur_sh,
							number 		dur_lo,
							vector(number) nom_dur,
							vector(number) nom_yld,
							rate_type 	rt,
							out number 	ilb_nom_y,
							out number 	ilb_y)
	{ 		
		number alfa 		= dur_lo - dur_sh;
		number ip_n_yld_s 	= ..ip_linear(0, 0).interpolate(nom_dur, nom_yld).eval(dur_sh);//ip_linear(dur_sh, nom_dur, nom_yld);
		number ip_n_yld_l 	= ..ip_linear(0, 0).interpolate(nom_dur, nom_yld).eval(dur_lo);//ip_linear(dur_lo, nom_dur, nom_yld);					
		number sec_val 		= future_value(1, ip_n_yld_l, dur_lo,  RT_EFFECTIVE);
    	number short_val 	= future_value(1, ip_n_yld_s, dur_sh, RT_EFFECTIVE);
		ilb_nom_y 			= implied_rate(sec_val, short_val, alfa, rt)*100;
		sec_val 			= future_value(1, y_r_lo, dur_lo,  RT_EFFECTIVE);
    	short_val 			= future_value(1, y_r_sh, dur_sh,  RT_EFFECTIVE);
		ilb_y 				= implied_rate(sec_val, short_val, alfa, rt)*100;
	}
	//-----------------------------------------------------------------------------
	// ilb_beinfl_sprd                                                          
	//-----------------------------------------------------------------------------
	number ilb_beinfl_sprd(	number nom_yld,
							number real_yld,
							number nom_yld_fwd,
							number real_yld_fwd)	
	{ 							
		number infl 		= (1+nom_yld)/(1+real_yld)-1; 													
		number infl_fwd 	= (1+nom_yld_fwd)/(1+real_yld_fwd)-1; 
		return -(infl_fwd-infl)*10000;
	}

	
}