/*	
	Core instrument functions
	Developer: Algorithmica Research, Magnus Nyström			
*/
option(null: hard);	

/*-----------------------------------------------------------------------
  sub_vector
  ----------------------------------------------------------------------*/
vector(instrument) sub_vector(vector(instrument) v, vector(integer) index)
option (category: 'Basic Types and Functions/Vector Functions')
{
	vector(instrument) v_out[0];

	for(integer i;i<v_size(index);i++) 		
		push_back(v_out, v[index[i]]);				
	
	return v_out;
}
/*-----------------------------------------------------------------------
  clone_vector_deep
  ----------------------------------------------------------------------*/
vector(instrument) clone_vector_deep(vector(instrument) option(nullable) vi)//todo
{
	return clone_vector(vi);
	/*if(null(vi))
		return null<vector(instrument)>;

	integer s = v_size(vi);
	if(s == 0)
		return null<vector(_fx)>;

	vector(_fx) vi_new[s];
	for(integer i = 0; i < s; i++) 
		vi_new[i] = vi[i].clone();
	return vi_new;*/								
}


module CORE_INT
{
	//-------------------------------------------------------
	// check_eq_ext_swap
	//-------------------------------------------------------
	 logical check_eq_ext_swap(	vector(instrument) 	vi, 
								out logical e, 
								out string  name)
	{	
		try {
			name = "";
			integer s = v_size(vi);			
			if(s == 0)
				return true;

			error_info ee = new error_info(true,false);
			error_info ee2 = new error_info(true,false);
			if(null(e)){
				e = vi[0].is_ext_swap(ee);
				QL_FAIL_COND(null(e), ee.message(), vi[0].name(ee2), true);	
							
			}
			
			for(integer i = 0; i < s; i++){
				logical ex = vi[i].is_ext_swap(ee);
				QL_FAIL_COND(null(ex), ee.message(), vi[i].name(ee2), true);
				if( e != ex)
					throw(E_INVALID_ARG,vi[i].name(ee));
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}

	 
	 //-------------------------------------------------------
	// check_eq_class
	//-------------------------------------------------------
	 logical check_eq_class(vector(instrument) 	vi, 
							out string  c_n, 
							out string  name)
	{	
		try {
			name = "";
			
			integer s = v_size(vi);			
			if(s == 0)
				return true;

			error_info ee = new error_info(true,false);
			error_info ee2 = new error_info(true,false);
			//swap_fix_flt sw = swap_fixflt(vi[0],ee);
			
			if(null(c_n)){
				logical e = vi[0].is_ext_swap(ee);
				c_n =  vi[0].class_name(ee);
				QL_FAIL_COND(!e && null(c_n), ee.message(), vi[0].name(ee2), true);							
			}
							
			for(integer i = 0; i < s; i++){
				logical e = vi[i].is_ext_swap(ee);
				if(e)
					continue;
				string cc =  vi[i].class_name(ee);
				QL_FAIL_COND(null(cc), ee.message(), vi[i].name(ee2), true);
				if(null(c_n))
					c_n = cc;
				if(!equal_casei(c_n, cc))
					throw(E_INVALID_ARG,vi[i].name(ee));
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}
	//-------------------------------------------------------
	// check_eq_class
	//-------------------------------------------------------
	 /*logical check_eq_class(vector(instrument) 	vi, 
							out string  c_n, 
							out string  name)
	{	
		try {
			name = "";
			
			integer s = v_size(vi);			
			if(s == 0)
				return true;

			error_info ee = new error_info(true,false);
			
			if(null(c_n)){
				c_n =  vi[0].class_name(ee);
				QL_FAIL_COND(null(c_n), ee.message(), vi[0].name(ee), true);							
			}
							
			for(integer i = 0; i < s; i++){
				string cc =  vi[i].class_name(ee);
				QL_FAIL_COND(null(cc), ee.message(), vi[i].name(ee2), true);
				if(!equal_casei(c_n, cc))
					throw(E_INVALID_ARG,vi[i].name(ee));
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}*/

	//-------------------------------------------------------
	// check_eq_index_tenor_code
	// either for instr with only one index or this is for leg 1
	//-------------------------------------------------------
	
	logical check_eq_index_tenor_code(	vector(instrument) 	vi, 
										out tenor_code  	t, 
										out string 			name)
	{	
		try {
			name = "";

			integer s = v_size(vi);			
			if(s == 0)
				return true;
			
			error_info ee = new error_info(true,false);
			error_info ee2 = new error_info(true,false);
			if(null(t)){
				t =  vi[0].index_tenor_code(ee);
				QL_FAIL_COND(null(t), ee.message(), vi[0].name(ee2), true);
								
			}
			
			for(integer i = 0; i < s; i++){				
				tenor_code tc = vi[i].index_tenor_code(ee);
				QL_FAIL_COND(null(tc), ee.message(), vi[i].name(ee2), true);
				if(t.frequency() != tc.frequency())
					throw(E_INVALID_ARG,vi[i].name(ee));							
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}

	//-------------------------------------------------------
	// check_eq_dc_method
	//-------------------------------------------------------
	 logical check_eq_dc_method(vector(instrument) 		vi, 
								out day_count_method 	dc, 
								out string 				name)
	{	
		try {
			name = "";

			integer s = v_size(vi);			
			if(s == 0)
				return true;
			
			error_info ee = new error_info(true,false);
			error_info ee2 = new error_info(true,false);
			if(null(dc)){
				dc =  vi[0].accr_dc_method(ee);
				QL_FAIL_COND(null(dc), ee.message(), vi[0].name(ee2), true);
								
			}
					
			for(integer i = 0; i < s; i++){
				day_count_method dd = vi[i].accr_dc_method(ee);
				QL_FAIL_COND(null(dd), ee.message(), vi[i].name(ee2), true);
				if(dc != dd)
					throw(E_INVALID_ARG,vi[i].name(ee));
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}

	//-------------------------------------------------------
	// check_eq_currency
	//-------------------------------------------------------
	 logical check_eq_currency(	vector(instrument) 		vi, 
								out string curr, 
								out string  name)
	{	
		try {
			name = "";

			integer s = v_size(vi);			
			if(s == 0)
				return true;
			
			error_info ee = new error_info(true,false);
			error_info ee2 = new error_info(true,false);
			if(null(curr)){
				curr =  vi[0].currency(ee);
				QL_FAIL_COND(null(curr), ee.message(), vi[0].name(ee2), true);							
			}
			
			for(integer i = 0; i < s; i++){
				string cc = vi[i].currency(ee);
				QL_FAIL_COND(null(cc), ee.message(), vi[i].name(ee2), true);
				if(!equal_casei(curr , cc))
					throw(E_INVALID_ARG,vi[i].name(ee));
				
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}

	//-------------------------------------------------------
	// check_eq_issue_date
	//-------------------------------------------------------
	 logical check_eq_issue_date(	vector(instrument) 		vi, 
									out date  				isd, 
									out string  			name)
	{	
		try {
			name = "";
			integer s = v_size(vi);			
			if(s == 0)
				return true;
			
			error_info ee = new error_info(true,false);
			error_info ee2 = new error_info(true,false);
			if(null(isd)){
				isd =  vi[0].issue_date(ee);
				QL_FAIL_COND(null(isd), ee.message(), vi[0].name(ee2), true);
								
			}
			
			for(integer i = 0; i < s; i++){
				date idt =  vi[i].issue_date(ee);
				QL_FAIL_COND(null(idt), ee.message(), vi[i].name(ee2), true);
				if(isd != idt)
					throw(E_INVALID_ARG,vi[i].name(ee));
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}

	 //-------------------------------------------------------
	// check_eq_settle
	//-------------------------------------------------------
	 logical check_eq_settle(	vector(instrument) 	vi, 
								out date  			stl, 
								out string  		name)
	{	
		try {
			name = "";
			integer s = v_size(vi);			
			if(s == 0)
				return true;
			
			error_info ee = new error_info(true,false);
			error_info ee2 = new error_info(true,false);
			if(null(stl)){
				stl =  vi[0].settle_date(ee);
				QL_FAIL_COND(null(stl), ee.message(), vi[0].name(ee2), true);
								
			}	
			for(integer i = 0; i < s; i++){
				date st =  vi[i].settle_date(ee);
				QL_FAIL_COND(null(st), ee.message(), vi[i].name(ee2), true);
				if(stl != st)
					throw(E_INVALID_ARG,vi[i].name(ee));
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}

	//-------------------------------------------------------
	// check_eq_class_flat
	//-------------------------------------------------------
	 logical check_eq_class_flat(	vector(instrument) 	vi, 
									out string  		c_n, 
									out string  		name)
	{	
		try {
			name = "";
			integer s = v_size(vi);			
			if(s == 0)
				return true;
			
			error_info ee = new error_info(true,false);
			error_info ee2 = new error_info(true,false);
			instr_def def;
			if(null(c_n)){
				def = vi[0].instr_def();
				QL_FAIL_COND(null(def), ee.message(), vi[0].name(ee2), true);
								
				c_n =  def.fltflt_swap_class_flat(ee);
				QL_FAIL_COND(null(c_n), ee.message(), vi[0].name(ee2), true);						
			}
			
			for(integer i = 0; i < s; i++){
				def = vi[i].instr_def();
				QL_FAIL_COND(null(def), ee.message(), vi[i].name(ee2), true);

				string cc = def.fltflt_swap_class_flat(ee);
				QL_FAIL_COND(null(cc), ee.message(), vi[i].name(ee2), true);
				if(!equal_casei(c_n, cc))
					throw(E_INVALID_ARG,vi[i].name(ee));
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}
	//-------------------------------------------------------
	// check_eq_class_sprd
	//-------------------------------------------------------
	 logical check_eq_class_sprd(	vector(instrument) 	vi, 
									out string  		c_n, 
									out string  		name)
	{	
		try {
			name = "";
			integer s = v_size(vi);			
			if(s == 0)
				return true;
			
			error_info ee = new error_info(true,false);
			error_info ee2 = new error_info(true,false);
			instr_def def;
			if(null(c_n)){
				def = vi[0].instr_def();
				QL_FAIL_COND(null(def), ee.message(), vi[0].name(ee2), true);
				
				c_n =  def.fltflt_swap_class_sprd(ee);
				QL_FAIL_COND(null(c_n), ee.message(), vi[0].name(ee2), true);
				
			}
			
			for(integer i = 0; i < s; i++){
				def = vi[i].instr_def();
				QL_FAIL_COND(null(def), ee.message(), vi[i].name(ee2), true);

				string cc = def.fltflt_swap_class_sprd(ee);
				QL_FAIL_COND(null(cc), ee.message(), vi[i].name(ee2), true);
				if(!equal_casei(c_n, cc))
					throw(E_INVALID_ARG,vi[i].name(ee));
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}
	//-------------------------------------------------------
	// check_eq_coupon_freq
	//-------------------------------------------------------
	 logical check_eq_coupon_freq(	vector(instrument) 		vi, 
									out integer 			f, 
									out string  			name)
	{	
		try {
			name = "";
			integer s = v_size(vi);			
			if(s == 0)
				return true;
			
			error_info ee = new error_info(true,false);
			error_info ee2 = new error_info(true,false);
			if(f==-1 || f == err_freq()){
				f =  vi[0].coupon_freq(ee);
				QL_FAIL_COND(f==-1 || f == err_freq(), ee.message(), vi[0].name(ee2), true);
			
			}
			
			for(integer i = 0; i < s; i++){
				integer ff = vi[i].coupon_freq(ee);
				QL_FAIL_COND(f==-1 || f == err_freq(), ee.message(), vi[i].name(ee2), true);
				if(f != ff)
					throw(E_INVALID_ARG,vi[i].name(ee));
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}
	
	
	
	//-------------------------------------------------------
	// check_instr_type
	//-------------------------------------------------------
	logical check_instr_type(vector(instrument) vi,
							instr_type 			it,
							out string  name)
	{
		try{
			name = "";
			if(v_size(vi)==0)
				return false;
			for(integer i=0; i < v_size(vi); i++){
				if(!vi[i].is_instr_type(it)) {
					name = vi[i].name();
					return false;
				}
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}
	
	//-------------------------------------------------------
	// check_depo_index
	//-------------------------------------------------------
	/*logical check_depo_index(	vector(instrument) vi,							
								out string name)
	{
		try{
			name = "";
			if(v_size(vi)==0)
				return false;
			
			for(integer i=0; i < v_size(vi); i++){
				if(!vi[i].is_instr_type(instr_type.DEPOSIT_INDEX)) {
					name = vi[i].name();
					return false;
				}
				
			}
			return true;
		}
		catch {		
			return false;
		}
	}*/
	//-------------------------------------------------------
	// check_instr_type
	//-------------------------------------------------------
	logical check_instr_type(vector(instrument) vi,
							vector(instr_type) 	v_it,
							out string  name)
	{
		try{
			name = "";
			if(v_size(vi)==0)
				return false;
			
			integer size = v_size(v_it);
			if(size == 0)
				return false;
	
			for(integer i=0; i < v_size(vi); i++){
				logical is = false;
				for(integer k=0; k < size; k++){
					if(vi[i].is_instr_type(v_it[k])) {
						is = true;
						break;
					}
				}
				if(!is) {
					name = vi[i].name();
					return false;
				}
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}
	//-------------------------------------------------------
	// check_instr_type
	//-------------------------------------------------------
	logical check_instr_type(	vector(instrument) vi, 
								instr_type 			it)
	{
		string name;
		return check_instr_type(vi, it, name);
	
	}
	//-------------------------------------------------------
	// check_instr_type
	//-------------------------------------------------------
	logical check_instr_type(	vector(instrument) vi, 
								vector(instr_type) v_it)
	{
		string name;
		return check_instr_type(vi, v_it, name);
	
	}

	//-------------------------------------------------------
	// check_calendar_name
	//-------------------------------------------------------
	logical check_calendar_name(integer 				cal_i, 
								vector(instrument) 		vi, 
								string option(nullable) cal_name, 
								out string  name)
	{	
		try {
			name = "";
			QL_REQUIRE(cal_i>=1 && cal_i<=4,'invalid calendar index');		
			for(integer i = 0; i < v_size(vi); i++){
	
				string c = vi[i].instr_def().calendar_name(cal_i);
				if(null(cal_name) && null(c)) 
					return true;
				else if(!null(cal_name) && null(c)) { 
					name = vi[i].name();
					return false;
				}
				else if(null(cal_name) && !null(c)) { 
					name = vi[i].name();
					return false;
				}
				else if(!equal_casei(cal_name, c)){
					name = vi[i].name();
					return false;
				}
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}
	//-------------------------------------------------------
	// check_calendar_name
	//-------------------------------------------------------
	logical check_calendar_name(vector(instrument) 	vi, 
								vector(string) 		cal_name, 
								out string  name)
	{	
		try {
			name = "";
			QL_REQUIRE(v_size(cal_name) == 4,'invalid calendar index');		
			for(integer i = 0; i < 4; i++){			
				logical ok = check_calendar_name(i+1, vi, cal_name[i], name);
				if(!ok) return ok;
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}
	
	
	//-------------------------------------------------------
	// swap_currency
	//-------------------------------------------------------
	 string swap_currency(	instrument i,
							logical leg1,
							out string e)
	 {
		 try {
			 error_info ee = error_info(true,true);
			 string res;
			 string c_f1,c_f2;
			 if(i.instr_type() == instr_type.SWAP_FIXFLT) {
				swap_fixflt sw = swap_fixflt(i,ee);
				sw.currency(c_f1,c_f2,ee);
				res = leg1 ? c_f1 : c_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FIXOIS) {
				swap_fixois sw = swap_fixois(i,ee);
				sw.currency(c_f1,c_f2,ee);
				res = leg1 ? c_f1 : c_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTFLT) {
				swap_fltflt sw = swap_fltflt(i,ee);
				sw.currency(c_f1,c_f2,ee);
				res = leg1 ? c_f1 : c_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_OISOIS) {
				swap_oisois sw = swap_oisois(i,ee);
				sw.currency(c_f1,c_f2,ee);
				res = leg1 ? c_f1 : c_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTOIS) {
				swap_fltois sw = swap_fltois(i,ee);
				sw.currency(c_f1,c_f2,ee);
				res = leg1 ? c_f1 : c_f2 ;
			 }
			else if(i.instr_type() == instr_type.SWAP_FLTFLT2S) {				
				swap_fltflt2s sw = swap_fltflt2s(i,ee);
				res = leg1 ? sw.currency_sprd_leg(ee) : sw.currency_flat_leg(ee);
			 }
			 else {
				 e = "invalid instrument type";
				 return null<string>;
			 }

			 QL_FAIL_COND(null(res), ee.message());
			 return res;
		 }
		 catch {
			 e = err.message();
			 return null<string>;
		 }
		 
	 }

	 //-------------------------------------------------------
	// swap_dc
	//-------------------------------------------------------
	 day_count_method swap_dc(instrument i,
							  logical leg1,
							  out string e)
	 {
		 try {
			 error_info ee = error_info(true,true);
			 day_count_method dc_f1,dc_f2;
			 day_count_method res;
			 if(i.instr_type() == instr_type.SWAP_FIXFLT) {
				swap_fixflt sw = swap_fixflt(i,ee);
				sw.dc_method(dc_f1,dc_f2,ee);
				res = leg1 ? dc_f1 : dc_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FIXOIS) {
				swap_fixois sw = swap_fixois(i,ee);
				sw.dc_method(dc_f1,dc_f2,ee);
				res = leg1 ? dc_f1 : dc_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTFLT) {
				swap_fltflt sw = swap_fltflt(i,ee);
				sw.dc_method(dc_f1,dc_f2,ee);
				res = leg1 ? dc_f1 : dc_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_OISOIS) {
				swap_oisois sw = swap_oisois(i,ee);
				sw.dc_method(dc_f1,dc_f2,ee);
				res = leg1 ? dc_f1 : dc_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTOIS) {
				swap_fltois sw = swap_fltois(i,ee);
				sw.dc_method(dc_f1,dc_f2,ee);
				res = leg1 ? dc_f1 : dc_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTFLT2S) {				
				swap_fltflt2s sw = swap_fltflt2s(i,ee);
				res = leg1 ? sw.dc_method_sprd_leg(ee) : sw.dc_method_flat_leg(ee);
			 }
			 else {
				 e = "invalid instrument type";
				 return null<day_count_method>;
			 }
			 QL_FAIL_COND(null(res), ee.message());
			 return res;
		 }
		 catch {
			 e = err.message();
			 return null<day_count_method>;
		 }		 
	 }
	//-------------------------------------------------------
	// swap_cfreq
	//-------------------------------------------------------
	 integer swap_cfreq(instrument i,
						logical leg1,
						out string e)
	 {
		 try {
			 error_info ee = error_info(true,true);
			 integer c_f1,c_f2;
			 integer res;
			 if(i.instr_type() == instr_type.SWAP_FIXFLT) {
				swap_fixflt sw = swap_fixflt(i,ee);
				sw.coupon_freq(c_f1,c_f2,ee);
				res = leg1 ? c_f1 : c_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FIXOIS) {
				swap_fixois sw = swap_fixois(i,ee);
				sw.coupon_freq(c_f1,c_f2,ee);
				res = leg1 ? c_f1 : c_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTFLT) {
				swap_fltflt sw = swap_fltflt(i,ee);
				sw.coupon_freq(c_f1,c_f2,ee);
				res = leg1 ? c_f1 : c_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_OISOIS) {
				swap_oisois sw = swap_oisois(i,ee);
				sw.coupon_freq(c_f1,c_f2,ee);
				res = leg1 ? c_f1 : c_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTOIS) {
				swap_fltois sw = swap_fltois(i,ee);
				sw.coupon_freq(c_f1,c_f2,ee);
				res = leg1 ? c_f1 : c_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTFLT2S) {				
				swap_fltflt2s sw = swap_fltflt2s(i,ee);
				res = leg1 ? sw.coupon_freq_sprd_leg(ee) : sw.coupon_freq_flat_leg(ee);
			 }
			 
			 else {
				 e = "invalid instrument type";
				 return err_freq();
			 }
			 QL_FAIL_COND(res == -1 || res == err_freq(), ee.message());
			 return res;
		 }
		 catch {
			 e = err.message();
			 return err_freq();
		 }		 
	 }
	 //-------------------------------------------------------
	// swap_rfreq
	//-------------------------------------------------------
	 integer swap_rfreq(instrument i,
						logical leg1,
						out string e)
	 {
		 try {
			 error_info ee = error_info(true,true);
			 integer c_f1,c_f2;
			 integer res;
			 if(i.instr_type() == instr_type.SWAP_FIXFLT) {
				swap_fixflt sw = swap_fixflt(i,ee);
				res = sw.reset_freq_flt_leg(ee);
			 }
			 else if(i.instr_type() == instr_type.SWAP_FIXOIS) {				
				res = 365 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTFLT) {
				swap_fltflt sw = swap_fltflt(i,ee);
				sw.reset_freq(c_f1,c_f2,ee);
				res = leg1 ? c_f1 : c_f2 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTOIS) {
				swap_fltois sw = swap_fltois(i,ee);
				sw.reset_freq(c_f1,c_f2,ee);
				res = leg1 ? c_f1 : 365 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_OISOIS) {				
				res = 365 ;
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTFLT2S) {				
				swap_fltflt2s sw = swap_fltflt2s(i,ee);
				res = leg1 ? sw.reset_freq_sprd_leg(ee) : sw.reset_freq_flat_leg(ee);
			 }			
			 else {
				 e = "invalid instrument type";
				 return err_freq();
			 }
			 QL_FAIL_COND(res == -1 || res == err_freq(), ee.message());
			 return res;
		 }
		 catch {
			 e = err.message();
			 return err_freq();
		 }		 
	 }

	 //-------------------------------------------------------
	// index_tenor_code
	//-------------------------------------------------------
	 tenor_code swap_index_tenor_code(	instrument i,
										logical leg1,
										out string e)
	 {
		 try {
			 error_info ee = error_info(true,true);
			 tenor_code res;
			 if(i.instr_type() == instr_type.SWAP_FIXFLT) {
				swap_fixflt sw = swap_fixflt(i,ee);
				res = sw.index_tenor_code_flt_leg(ee);	
			 }
			 else if(i.instr_type() == instr_type.SWAP_FIXOIS) {
				res = tenor_code(365);
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTOIS) {
				swap_fltois sw = swap_fltois(i,ee);
				res = leg1 ? sw.index_tenor_code_flt_leg(ee) : tenor_code(365);
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTFLT) {
				swap_fltflt sw = swap_fltflt(i,ee);
				res = leg1 ? sw.index_tenor_code_leg1(ee) : sw.index_tenor_code_leg2(ee);
			 }
			 else if(i.instr_type() == instr_type.SWAP_FLTFLT2S) {				
				swap_fltflt2s sw = swap_fltflt2s(i,ee);
				res = leg1 ? sw.index_tenor_code_sprd_leg(ee) : sw.index_tenor_code_flat_leg(ee);
			 }			
			 else {
				 e = "invalid instrument type";
				 return null<tenor_code>;
			 }
			 QL_FAIL_COND(null(res), ee.message());
			 return res;
		 }
		 catch {
			 e = err.message();
			 return null<tenor_code>;
		 }		 
	 }
	 //-------------------------------------------------------
	// swap_index_tenor_code_check
	//-------------------------------------------------------
	 logical swap_index_tenor_code_check(	vector(instrument) 	vi, 
											out tenor_code 		tc,
											logical 			leg1,
											out string  		name)
	{	
		try {
			name = "";
			integer s = v_size(vi);			
			if(s == 0)
				return true;

			error_info ee = error_info(true,false);
			string e = "";
			
			if(null(tc)){

				tc = swap_index_tenor_code(vi[0], leg1, e);
				QL_FAIL_COND(null(tc), e, vi[0].name(ee), true);
			}
			
			for(integer i = 0; i < s; i++){

				tenor_code ff = swap_index_tenor_code(vi[i], leg1, e);
				QL_FAIL_COND(null(ff), e, vi[i].name(ee), true);

				if(tc.frequency() != ff.frequency())
					throw(E_INVALID_ARG,vi[i].name(ee));
				
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}


	//-------------------------------------------------------
	// swap_currency_check
	//-------------------------------------------------------
	 logical swap_currency_check(	vector(instrument) 	vi, 
									out string 			f,//input and output if not valid
									logical 			leg1,
									out string  		name)
	{	
		try {
			name = "";
			integer s = v_size(vi);			
			if(s == 0)
				return true;

			error_info ee = error_info(true,false);
			string e = "";
			
			if(null(f)){				
				f = swap_currency(vi[0], leg1, e);
				QL_FAIL_COND(null(f), e, vi[0].name(ee), true);
			}
			
			for(integer i = 0; i < s; i++){
				string ff = swap_currency(vi[i], leg1, e);
				QL_FAIL_COND(null(ff), e, vi[i].name(ee), true);
				
				if(!equal(f ,ff))
					throw(E_INVALID_ARG,vi[i].name(ee));
				
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}

	//-------------------------------------------------------
	// swap_dc_check
	//-------------------------------------------------------
	 logical swap_dc_check(	vector(instrument) 			vi, 
							out day_count_method 		dc,
							logical 					leg1,
							out string  name)
	{	
		try {
			name = "";
			integer s = v_size(vi);			
			if(s == 0)
				return true;

			error_info ee = error_info(true,false);
			string e = "";
			
			if(null(dc)){
				dc = swap_dc(vi[0], leg1, e);
				QL_FAIL_COND(null(dc), e, vi[0].name(ee), true);					
			}
			
			for(integer i = 0; i < s; i++){
				day_count_method ff = swap_dc(vi[i], leg1,e);
				QL_FAIL_COND(null(ff), e,vi[i].name(ee), true);	
				if(dc != ff)
					throw(E_INVALID_ARG,vi[i].name(ee));
				
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}
	//-------------------------------------------------------
	// swap_cfreq_check
	//-------------------------------------------------------
	 logical swap_cfreq_check(	vector(instrument) 	vi, 
								out integer 		f,//input and output if not valid
								logical 			leg1,
								out string  		name)
	{	
		try {
			name = "";
			integer s = v_size(vi);			
			if(s == 0)
				return true;

			error_info ee = error_info(true,false);
			string e = "";
			
			if(f == -1 || f == err_freq()){
				f = swap_cfreq(vi[0], leg1, e);
				QL_FAIL_COND(f == -1 || f == err_freq(), e, vi[0].name(ee), true);
			}
			
			for(integer i = 0; i < s; i++){

				integer ff = swap_cfreq(vi[i], leg1, e);
				QL_FAIL_COND(ff == -1 || ff == err_freq(), e, vi[i].name(ee), true);
				if(f != ff)
					throw(E_INVALID_ARG,vi[i].name(ee));
				
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}

	//-------------------------------------------------------
	// swap_rfreq_check
	//-------------------------------------------------------
	 logical swap_rfreq_check(	vector(instrument) 	vi, 
								out integer 		f,//input and output if not valid
								logical 			leg_1,
								out string  		name)
	{
		try {
			name = "";
			integer s = v_size(vi);			
			if(s == 0)
				return true;

			error_info ee = error_info(true,false);
			string e = "";
			
			if(f == -1 || f == err_freq()){				
			
				f = swap_rfreq(vi[0], leg_1, e);
				QL_FAIL_COND(f == -1 || f == err_freq(), e, vi[0].name(ee), true);
			}
			
			for(integer i = 0; i < s; i++){
				integer ff = swap_rfreq(vi[i], leg_1, e);
				QL_FAIL_COND(ff == -1 || ff == err_freq(), e, vi[i].name(ee), true);
				
				if(f != ff)
					throw(E_INVALID_ARG,vi[i].name(ee));
				
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}

	
	//-------------------------------------------------------
	// check_is_mtm
	//-------------------------------------------------------
	 logical check_is_mtm(	vector(instrument) 	vi, 
							logical 			mtm,  
							out string  		name)
	{	
		try {
			name = "";				
			for(integer i = 0; i < v_size(vi); i++){
				if(mtm != vi[i].ccbs_is_mtm()){
					name = vi[i].name();
					return false;
				}
			}
			return true;
		}
		catch {
			name = err.message();
			return false;
		}
	}
	
		
	

	 //-------------------------------------------------------
	// check_class_base  backward compat (just a name change)
	//-------------------------------------------------------
	/* logical check_class_base(	vector(instrument) 	vi, 
								out string  		c_n, 
								out string  		name)
	{
		return check_class_sprd(vi, c_n, name);
	}*/

	
	//-------------------------------------------------------
	// init_quote 
	//-------------------------------------------------------
	void init_quote(	date 						d, 
						instrument_name 			i_n,
						quote_side option(nullable) q_s, 
						out instrument   			i,
						out number  				quote)
	
	{		
		if(null(quote)){			
			i = null(q_s) ? instrument(i_n, d) : instrument(i_n, d, q_s);			
			quote = null(i) ? null<number> : i.quote();
		}
		else
			i = instrument(i_n, d, quote);
	}
	//-------------------------------------------------------
	// init_quote_bond 
	//-------------------------------------------------------
	void init_quote_bond(	date 						d, 
							instrument_name 			i_n,
							quote_side option(nullable) q_s, 
							out bond   					i,
							out number  				quote)
	
	{		
		if(null(quote)){			
			i = bond(i_n, d, q_s);	
			quote = null(i) ? null<number> : i.quote();
		}
		else
			i = bond(i_n, d, quote);
	}
	//-------------------------------------------------------
	// init_quote_il_bond 
	//-------------------------------------------------------
	void init_quote_il_bond(	date 						d, 
								instrument_name 			i_n,
								quote_side option(nullable) q_s, 
								out il_bond   				i,
								out number  				quote)
	
	{		
		if(null(quote)){			
			i = il_bond(i_n, d, q_s);	
			quote = null(i) ? null<number> : i.quote();
		}
		else
			i = il_bond(i_n, d, quote);
	}
	//-------------------------------------------------------
	// init_quote 
	//-------------------------------------------------------
	void init_quote_bond_fut(	date 						d, 
								instrument_name 			i_n,
								quote_side option(nullable) f_q_s,
								quote_side option(nullable) b_q_s, 
								out bond_future   			i,
								out number  				quote)
	
	{		
		if(null(quote)){			
			i = bond_future(i_n, d, f_q_s, b_q_s);	
			quote = null(i) ? null<number> : i.quote();
		}
		else
			i = bond_future(i_n, d, quote,b_q_s);
	}
	//-------------------------------------------------------
	//  get_t_from_s
	//-------------------------------------------------------
	date  get_t_from_s(	instrument 	inst,
						date 		s)	
	{
		switch(inst.instr_type()){   
		case ..instr_type.FRA:
		case ..instr_type.FX_SPOT:
		case ..instr_type.FX_SWAP:
		case ..instr_type.FX_FORWARD:
		case ..instr_type.IR_FUTURE:
		case ..instr_type.BOND_FUTURE:
		case ..instr_type.BILL_FUTURE:
		case ..instr_type.NOIS:
		case ..instr_type.OVERNIGHT_FUTURE:
			QL_FAIL("not applicable for current instrument <get_t_from_s>");		
		}
			
		date_code dc = date_code(inst.settle_code());
		QL_FAIL_COND(null(dc),"invalid settle code <get_t_from_s>");
		return dc.apply_back(s, inst.calendar(0)) ;		
	}


	
}

