/*	
	Core functions
	Developer: Algorithmica Research, Magnus Nyström			
*/

option(null: hard);	

number INSTR_AUX_VERSION_MAJOR = 1645;
number INSTR_AUX_VERSION_MINOR = 0;


void version_info_ql(out number major, out number minor)	
{ 
	major = INSTR_AUX_VERSION_MAJOR;
	minor = INSTR_AUX_VERSION_MINOR ;
	return ;
}

void version_info_qll(out number major, out number minor)
	option (category: 'Instrument/001 Version')
{ 
	major = __qll_version_major();
	minor = __qll_version_minor() ;
	return ;
}


string null_string_com() {	return "NULL";}
logical is_null_string_com(string option(nullable) s)	
{	
	if (null(s)) 
		return true; 
	else 
		return equal(s,"NULL");
}
string conv_null_string_com(string option(nullable) s) 		
{ 
	if(is_null_string_com(s)) 
		return  null<string>;	
	else
		return s;
}

number null_number_com() {	return -214748364;}
logical is_null_number_com(number option(nullable) n)	
{	
	if (null(n)) 
		return true; 
	else 		
		return n == -214748364;
}
number conv_null_number_com(number option(nullable) n) 		
{ 
	if(is_null_number_com(n)) 
		return  null<number>;	
	else
		return n;
}

integer null_int_com()					{	return -214748364;}
logical is_null_int_com(integer i)		{	return i == -214748364;}

date null_date_com()					{	return date(1900,1,1);}
logical is_null_date_com(date option(nullable) d)		
{ 
	if (null(d)) 
		return true; 
	else 
		return d == date(1900,1,1);
}

date conv_null_date_com(date option(nullable) d) 		
{ 
	if(is_null_date_com(d)) 
		return  null<date>;	
	else
		return d;
}

number disc_func_disc_com(disc_func df, number t) 		
{ 
	return df.disc(t);
}


class enum instr_type option(category:"Instrument") {
	BILL			option(str:  "bill"),
    BILL_FUTURE		option(str:  "bill future"),
    FRN				option(str:  "frn"),
    CALLABLE_FRN	option(str:  "callable frn"),//currently not used
    SWAP_FIXFLT		option(str:  "swap fixflt"),
    SWAP_FIXFIX		option(str:  "swap fixfix"),
    SWAP_FLTFLT		option(str:  "swap fltflt"),
    SWAP_ZEROFLT	option(str:  "swap zeroflt"),
    SWAP_FIXOIS		option(str:  "swap fixois"),
    SWAP_FLTOIS		option(str:  "swap fltois"),
    SWAP_OISOIS		option(str:  "swap oisois"),
    SWAP_ZEROOIS	option(str:  "swap zeroois"),
    SWAP_FLTFLT2S	option(str:  "swap fltflt2s"),
    FRA				option(str:  "fra"),
    IR_FUTURE		option(str:  "ir future"),
    OVERNIGHT_FUTURE option(str:  "overnight future"),
    DEPOSIT			option(str:  "deposit"),
    DEPOSIT_INDEX	option(str:  "deposit index"),
    FIXED_CPN_BOND	option(str:  "fixed cpn bond"),
    CALLABLE_FIX_CPN_BOND option(str:  "callable fix cpn bond"),//currently not used
    ZERO_CPN_BOND	option(str:  "zero cpn bond"),
    RIBA			option(str:  "riba"),    
    NOIS			option(str:  "nois"),
    FX_SPOT			option(str:  "fx spot"),
    FX_SWAP			option(str:  "fx swap"),
    FX_FORWARD		option(str:  "fx forward"),
    IL_REAL_BILL	option(str:  "il real bill"),//currently not used
    IL_REAL_FIXED_CPN_BOND option(str:  "il real fixed cpn bond"),
    IL_REAL_ZERO_CPN_BOND option(str:  "il real zero cpn bond"),
    IL_SWAP_REAL_FIXFLT	option(str:  "il swap real fixflt"),
    IL_SWAP_REAL_FIXOIS	option(str:  "il swap real fixois"),
    IL_SWAP_ZERO	option(str:  "il swap zero"),
    IL_SWAP_FIX_Y_Y	option(str:  "il swap fix y-y"),
    IL_SWAP_FLT_Y_Y	option(str:  "il swap flt y-y"),
    IL_SWAP_OIS_Y_Y	option(str:  "il swap ois y-y"),
    BOND_FUTURE		option(str:  "bond future"),		
    RATING_ADJ_BOND	option(str:  "rating adj bond"),//currently not used
    GENERIC			option(str:  "generic"),
    EQUITY			option(str:  "equity"),
	EQUITY_OPTION_VANILLA option(str:  "equity option vanilla"),
	SWAP_FIX_FLT_ASW option(str:  "swap fixed vs float asw"),
	SWAP_FIX_OIS_ASW option(str:  "swap fixed vs overnight index asw")
};

/*vals		(size = 39)
	[0]	BILL
	[1]	BILL_FUTURE
	[2]	FIXED_CPN_BOND
	[3]	CALLABLE_FIX_CPN_BOND
	[4]	BOND_FUTURE
	[5]	RATING_ADJ_BOND
	[6]	ZERO_CPN_BOND
	[7]	FRN
	[8]	CALLABLE_FRN
	[9]	DEPOSIT
	[10]	DEPOSIT_INDEX
	[11]	FRA
	[12]	IR_FUTURE
	[13]	OVERNIGHT_FUTURE
	[14]	RIBA
	[15]	NOIS
	[16]	FX_SPOT
	[17]	FX_SWAP
	[18]	FX_FORWARD
	[19]	SWAP_FIXFLT
	[20]	SWAP_FIXFIX
	[21]	SWAP_FLTFLT
	[22]	SWAP_FLTFLT2S
	[23]	SWAP_ZEROFLT
	[24]	SWAP_FIXOIS
	[25]	SWAP_FLTOIS
	[26]	SWAP_OISOIS
	[27]	SWAP_ZEROOIS
	[28]	IL_REAL_BILL
	[29]	IL_REAL_FIXED_CPN_BOND
	[30]	IL_REAL_ZERO_CPN_BOND
	[31]	IL_SWAP_REAL_FIXFLT
	[32]	IL_SWAP_REAL_FIXOIS
	[33]	IL_SWAP_ZERO
	[34]	IL_SWAP_FIX_Y_Y
	[35]	IL_SWAP_FLT_Y_Y
	[36]	IL_SWAP_OIS_Y_Y
	[37]	EQUITY
	[38]	GENERIC
*/

vector(instr_type) instr_type_values() option(category:"Instrument")
{
	vector(instr_type) e;
	expand_enum(e) ;
	return e;
}

instr_type instr_type_map(string s) option(category:"Instrument")
{
	instr_type e;
	str_to_enum(s,e);
	return e;
}

vector(string) instr_type_list() option (category: 'Instrument') 
{ 
	vector(instr_type) e;
	expand_enum(e) ;
	return string(e);
}

module CORE_INT
{
	__instr_type __instr_type_map(string s) 
	{
		__instr_type e;
		str_to_enum(s,e);
		return e;
	}

	instr_type conv_instr_type(..__instr_type __type)
	{
		string s = string(__type);	
		return instr_type_map(s) ;
	}

	__instr_type conv_rev_instr_type(..instr_type type)
	{
		//this conversion is not always safe since we have more instr_type than __instr_type
		try {
			string s = string(type);	
			return __instr_type_map(s) ;
		}
		catch {
			return null<__instr_type>;
		}
	}
}


vector(string) instrument_invalid_names(..instr_type it,out vector(string) errs)
option(category: "Instrument/002 Database") 
{
	__instr_type __it = CORE_INT.conv_rev_instr_type(it);

	if(null(__it)) {
		errs =  null<vector(string)>;
		return null<vector(string)>;
	}
	else {
		return __instrument_invalid_names(__it,errs);
	}
}


vector(string) instrument_search(..instr_type option(nullable) it,
								 logical only_valid,
								 date option(nullable) min_maturity,
								 date option(nullable) max_maturity)
option(category: "Instrument/002 Database") 
{
	if(null(it))
		return __instrument_search(null<__instr_type>,only_valid, min_maturity, max_maturity);
	
	__instr_type __it = CORE_INT.conv_rev_instr_type(it);

	if(null(__it)) {
		return null<vector(string)>;
	}
	else {
		return __instrument_search(__it,only_valid, min_maturity, max_maturity);
	}
}


module CORE_INT
{
	//number EPSILON_LOW 	= 1.0e-10;
	number EPSILON 		= 1.0e-10;
	number PI 			= 3.14159265358979;
	number DBL_EPSILON 	= 2.22044604925031E-16;
	number MAX_NUMBER  	= 2.2204460492503131E32;
	number MIN_NUMBER  	= -MAX_NUMBER;

	
	
	/*-----------------------------------------------------------------------
	  Name :    write_warning_message
	  Notes :           
	  Changes:
	  			Date    		Who     Detail
	  			9-Dec-2013 		MN      Original code 
	  ----------------------------------------------------------------------*/
	void write_warning_message(	string 	msg, logical incl_w_text = true) 
	{
		//if(display) {
			string s = incl_w_text ? strcat("WARNING: ", msg) : msg;
			log_message(LOG_WARNING,s);
		//}
	}
		
	void flip (	out number val1, 
				out number val2)
	{
		number tmp = val2;
		val2 = val1;
		val1 = tmp;
	}

	void flip (	out date val1, 
				out date val2)
	{
		date tmp = val2;
		val2 = val1;
		val1 = tmp;
	}

	logical is_unique(vector(string) option(nullable) v) {
		if(null(v)) return true;
		return v_size(v_find_uniquei(v)) == v_size(v);
	}
	
	logical is_equal(number option(nullable) a,
					   number option(nullable) b,
					   number tol = EPSILON)
	{
		if(null(a) || null(b))
			return null<logical>;
		return (abs((a)-(b)) < tol);
	}
	
	logical is_equal_zero(number option(nullable) a,
							number tol = EPSILON )
	{
		if(null(a))
			return null<logical>;
		return (abs((a)-(0.0)) < tol);
	}

	/*-----------------------------------------------------------------------
	  Name :    is_equal_par
	  Notes :           
	  Changes:
	  			Date    		Who     Detail
	  			6-Dec-2013 		MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	logical is_equal_par(number option(nullable) a,
						number option(nullable) b)
	{
		return is_equal(a,b, 0.00001);
	}

	
	logical negative_check(out number n, number tol = EPSILON)//neg --> true
	{
		if(n >= tol)
			return false;
		else if (n <= -tol)
			return true;
		else {
			n = 0;
			return false;
		}
	}
	logical zero_negative_check(out number n, number tol = EPSILON)//zero or neg --> true
	{
		if(n >= tol)
			return false;
		else if (n <= -tol)
			return true;
		else {
			n = 0;
			return true;
		}
	}

	/*-----------------------------------------------------------------------
	  Name :    solve_quadratic
	  Notes :	solves x^2+a2*x+a3 = 0          
	  Changes:
				Date    		Who     Detail
				10-Dec-2013 	MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	vector(number) solve_quadratic(	number a2, 
									number a3)
	//option(category: 'Mathematics/Basic Functions')
	{
		
		number D = sqr(a2)/4-a3;
		if(D<0){
			vector(number) roots[1];
			roots[0] = null<number>;
			return roots;
		}
		else if(D==0){
			vector(number) roots[1];
			roots[0] = -a2/2;
			return roots;
		}
		else {
			vector(number) roots[2];
			roots[0] = -a2/2+sqrt(D);
			roots[1] = -a2/2-sqrt(D);
			return roots;
		}		
	}
		
	/*-----------------------------------------------------------------------
	  Name :    solve_cubic
	  Notes :	solves x^3+a2*x^2+a3*x^3+a4 = 0           
	  Changes:
				Date    		Who     Detail
				10-Dec-2013 	MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	vector(number) solve_cubic(	number a2, 
								number a3, 
								number a4)
	//option(category: 'Mathematics/Basic Functions')
	{
		vector(number) roots;
		number th = 1/3;
		number p = a3-pow(a2,2)*th ;
		number q = a4+(2*pow(a2,3)-9*a2*a3)/27;
		number D = pow(p/3,3)+pow(q/2,2);			

		if(D > 0){
			number u;
			number v;			
			if(-q/2+sqrt(D) >= 0)
				u = pow(-q/2+sqrt(D),th);
			else
				u = -pow(q/2-sqrt(D),th);
			if(-q/2-sqrt(D) > 0)
				v = pow(-q/2-sqrt(D),th);
			else
				v = -pow(q/2+sqrt(D),th);
			push_back(roots, u+v-a2*th);
		
		}
		else if(D == 0){
			number root;
			if(-q/2 >= 0)
				root = pow(-q/2, th );
			else
				root = -pow(q/2, th );
			push_back(roots, 2*root - a2*th );
			push_back(roots, -root - a2*th ); 
		}
		else if( D < 0){ 
			number theta = atan2( sqrt(-D), -q/2);
			number r = sqrt(pow(-q/2,2)-D);

			number reU = pow(r, th ) * cos(theta*th );
			number imU = pow(r, th ) * sin(theta*th );
			push_back(roots, 2*reU  - a2/3);
			push_back(roots, -reU - imU * sqrt(3) - a2*th );
			push_back(roots, -reU + imU * sqrt(3) - a2*th );
		} 
		return roots;
		
	}
}

/*-----------------------------------------------------------------------
  Name :    error handling funcs
  Notes :           
  Changes:
  			Date    		Who     Detail
  			13-Jul-2016 		MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
string throw_err(string option(nullable) e)
{		
	if(null(e) || e == ""){
		return "<unknown error>";		
	}
	if( sub_string(e, 0, 1) == "<")
		return e;
	
	return strcat(["<",e,">"]);		
}

string throw_err(string option(nullable) e,
				 string option(nullable) e2,
				 logical option(nullable) inst_err = null<logical>)
{	
	if(null(e) || e == "")
		e = "unknown error";
	if(null(e2) || e2 == "")
		e2 = "unknown name";

	logical h = sub_string(e, 0, 1) == "<";
	
	if(null(inst_err) || !inst_err)		
		return h ? strcat([	e ,"," ,e2]) : strcat([	"<",e ,"," ,e2,">"]);	
	else
		return h ? strcat([	e ,"[", e2,"]"]) : strcat([	"<",e ,"[", e2,"]", ">"]);			
}

void QL_REQUIRE(logical check,
				string option(nullable) e,
				error_type option(nullable) type = E_INVALID_ARG)
option (category: 'Core Programming/Error Handling')
{	
	if(!check) {
		if(null(type)) type = E_INVALID_ARG;
		throw(type,throw_err(e));
	}
}

void QL_REQUIRE(logical check,
				string option(nullable) e,
				string option(nullable) e2, 
				logical option(nullable) inst_err = null<logical>,
				error_type option(nullable) type = E_INVALID_ARG)
option(com_name: 'QL_REQUIRE_EXT')
option (category: 'Core Programming/Error Handling')
{	
	if(!check) {
		if(null(type)) type = E_INVALID_ARG;
		throw(type,throw_err(e,e2,inst_err));
	}
}

void QL_FAIL(string e,
			 error_type option(nullable) type = E_INVALID_ARG) 
option (category: 'Core Programming/Error Handling')
{
	if(null(type)) type = E_INVALID_ARG;
	throw(type,throw_err(e));
}

void QL_FAIL_COND(logical check,
				  string option(nullable) e,
				  error_type option(nullable) type = E_INVALID_ARG)
option (category: 'Core Programming/Error Handling')
{
	if(check) {
		if(null(type)) type = E_INVALID_ARG;
		throw(type,throw_err(e));
	}
}

void QL_FAIL_COND(	logical check,
					string option(nullable) e,
					string option(nullable) e2,
					logical option(nullable) inst_err = null<logical>,
					error_type option(nullable) type = E_INVALID_ARG)
option(com_name: 'QL_FAIL_COND_EXT')
option (category: 'Core Programming/Error Handling')
{
	if(check) {
		if(null(type)) type = E_INVALID_ARG;
		throw(type,throw_err(e,e2,inst_err));
	}
}

void QL_FAIL(string e,
			 string e2,
			logical option(nullable) inst_err = null<logical>,
			error_type option(nullable) type = E_INVALID_ARG)
option(com_name: 'QL_FAIL_EXT')
option (category: 'Core Programming/Error Handling')
{
	if(null(type)) type = E_INVALID_ARG;
	throw(type,throw_err(e,e2,inst_err));
}

void QL_WARNING(string option(nullable) e)
option (category: 'Core Programming/Error Handling')
{
	if(null(e) || e == "")	e = "unknown warning";	
	CORE_INT.write_warning_message(throw_err(e));
}

/*-----------------------------------------------------------------------
  Name :    sub_vector
  ----------------------------------------------------------------------*/
vector(instr_type) sub_vector(vector(instr_type) v,
							  vector(integer) index)
option (category: 'Basic Types and Functions/Vector Functions')
{
	vector(instr_type) v_out[0];

	for(integer i;i<v_size(index);i++) 		
		push_back(v_out, v[index[i]]);				
	
	return v_out;
}
/*-----------------------------------------------------------------------
  Name :    v_find
  ----------------------------------------------------------------------*/
vector(integer ) v_find(vector(instr_type) v, instr_type v_ref)
	option (category: 'Basic Types and Functions/Vector Functions')
{
	vector(integer) v_out[0];

	for(integer i;i<v_size(v);i++) {		
		if (v[i] == v_ref)	
			push_back(v_out, i);				
	}	
	return v_out;
}
/*-----------------------------------------------------------------------
  Name :    v_find
  ----------------------------------------------------------------------*/
vector(integer ) v_find(vector(tenor_code) v, tenor_code v_ref)
	option (category: 'Basic Types and Functions/Vector Functions')
{
	vector(integer) v_out[0];

	for(integer i;i<v_size(v);i++) {		
		if (v[i] == v_ref)	
			push_back(v_out, i);				
	}	
	return v_out;
}
/*-----------------------------------------------------------------------
  Name :    is_digit
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
logical is_digit(string s, integer i)
option (category: 'Basic Types and Functions/String Manipulation')
{
	switch (sub_string(s, i, 1)) {
   	case "0":
    case "1":
    case "2":
    case "3":
    case "4":
    case "5":
    case "6":
    case "7":
    case "8":
    case "9":
    	return true;
    default:
    	return false;
    }
}


/*-----------------------------------------------------------------------
  Name :    trunc
  Notes :           
  Changes:
  			Date    		Who     Detail
  			6-Dec-2013 		MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
number trunc(number value) 
option(category: 'Mathematics/Basic Functions')
{
	return value < 0 ? ceil(value) : floor (value);
}

/*-----------------------------------------------------------------------
  Name :    z_score
  Notes :   ex assetswap 
		
			We can track this cheapness/richness using a Z-score
			
				Z = (CurrentASW - AverageASW)/ Standard Deviation of ASW  
	
			Typically the Z-score will be measured by using daily observations over a 3-month to 6-month period. A
			positive number denotes that the bond is trading cheap to its historic range, and, conversely, a negative
			number denotes it is trading expensive. For example, assuming asset swap spreads are normally distributed
			then a Z-score of +1 denotes that the bond's asset swap spread is 1 standard deviation cheap to its
			historical range, while a Z-score of -2 denotes that a bond's asset swap is 2 standard deviations expensive
			to its range.        
  Changes:
  			Date    		Who     Detail
  			10-Dec-2013 	MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/	
number z_score(	number		x,								
				number 		avg,
				number 		stdev,
				//number	stdev_dec,
				out number	crit_prob)
option (category: 'Mathematics/Statistics')
{	
	if(stdev == 0){
		crit_prob = null<number>;
		CORE_INT.write_warning_message("stdev == 0");
		return null<number>;
	}
	else {
		number z = (x-avg)/stdev; 
		crit_prob = cum_normal(-abs(z))*100;
		return z;
	}
}	
/*-----------------------------------------------------------------------
  Name :    z_score
  Notes :          
  Changes:
  			Date    		Who     Detail
  			10-Dec-2013 	MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/	
number z_score(	date 				from, 
				date 				to, 
				series<date>(number) s,
				number 				curr_obs,
				out number 			avg,
				out number 			sd,					
				out number 			crit_prob)
option (category: 'Mathematics/Statistics')
option(com_name: "z_score_ext")
{
	try{
		date start 	= s_x0(s);
		date end 	= start + s_dx(s)*s_size(s)-1 ;
		QL_FAIL_COND(from < start || end > to || from > to,"date outside series range");

		series<date>(number) ss = series(t: from, to;  s[t]);		
		avg 					= average(ss);
		sd 						= std_dev(s);
		return z_score(curr_obs, avg, sd, crit_prob);
	}
	catch {
		CORE_INT.write_warning_message(err.message());
		avg 		= null<number>;
		sd 			= null<number>;
		crit_prob 	= null<number>;
		return null<number>; 
	}
}

/*-----------------------------------------------------------------------
  Name :    disc_func_interp_ips
  Notes :   creation of disc-func with interpolation from settledate      
  Changes:
  			Date    		Who     Detail
  			10-Dec-2013 	MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/	
disc_func disc_func_interp_ips(	vector(number)	periods,
								vector(number)	disc_f,
								date 			trade_date,
								date 			settle_date,
								interpolator 	interp,
								rate_type 		rate_t)
option (category: 'Yield Curve/Discount Function Manipulation') 
 {
	try {
		disc_func df_pre 	= disc_func_interp(periods, disc_f, interp, rate_t);
		number per_s 		= (settle_date-trade_date)/365;
		number df_s 		= df_pre.disc(per_s);		
		periods 			-= per_s;
		disc_f 				/= df_s;
		while(periods[0] <= 0) {
			periods = v_exclude (periods , 0);
			disc_f = v_exclude (disc_f, 0);
		}
		disc_func df = disc_func_interp(periods, disc_f, interp, rate_t);
		df = disc_func_shift_fwd(df, per_s, df_s);
		return disc_func_splice(df_pre, df, per_s);
	}
	catch {
		CORE_INT.write_warning_message(err.message());
		return null<disc_func>;
	}
}
/*-----------------------------------------------------------------------
  Name :    str_curr
  Notes :   str_ex(1300.23, 2, ",") ==> "1,300.23"     
  Changes:
  			Date    		Who     Detail
  			2-Aug-2016 		MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
string str_curr(	number option(nullable) val, 
					integer dec, 
					string option(nullable) thousand_sep = null<string> )
option (category: 'Basic Types and Functions/String Conversion') 

{	
	if(null(val))
		return "";

	if(null( thousand_sep))
		 thousand_sep = ",";

	string s 		= str(val, dec) ;	
	integer sz 		= strlen(s) ;	
	integer decs 	= (dec != 0 ? dec+1 : 0);
	 
	for(integer i = sz-3-decs; i > 0; i -= 3){
		if(i>1 || sub_string(s,0,i) != '-')
			s = strcat([sub_string(s,0, i), thousand_sep, sub_string(s, i, sz-i)]) ; 
		sz++;
	} 	                           
	return s ;	
}

/*-----------------------------------------------------------------------
  Name :    format_time
  Notes :       
  Changes:
  			Date    		Who     Detail
  			2-Aug-2016 		MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
void format_time(	out string d, 
					out string t)
option (category: 'Basic Types and Functions/String Conversion')
{

	timestamp n = now() ;
	d = str(date(n));
	vector(string) v[7];
	v[0] = str(hour(n));
	if(strlen(v[0]) == 1)
		v[0] = strcat("0",v[0]);
	v[1] = ':';
	v[2] = str(minute(n));
	if(strlen(v[2]) == 1)
		v[2] = strcat("0",v[2]);
	v[3] = ':';
	v[4] = str(second(n));
	if(strlen(v[4]) == 1)
		v[4] = strcat("0",v[4]);
	v[5] = ':';
	v[6] = str(millisecond(n));
	t = strcat(v);	
	return;		
}
/*-----------------------------------------------------------------------
  Name :    format_time
  Notes :       
  Changes:
  			Date    		Who     Detail
  			2-Aug-2016 		MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
void format_time(	timestamp 					ts,
					out string d,
					out string t,
					out string t_sh)
option (category: 'Basic Types and Functions/String Conversion')
{

	d = str(date(ts));
	vector(string) v[7];
	v[0] = str(hour(ts));
	if(strlen(v[0]) == 1)
		v[0] = strcat("0",v[0]);
	v[1] = ':';
	v[2] = str(minute(ts));
	if(strlen(v[2]) == 1)
		v[2] = strcat("0",v[2]);
	v[3] = ':';
	v[4] = str(second(ts));
	if(strlen(v[4]) == 1)
		v[4] = strcat("0",v[4]);
	v[5] = ':';
	v[6] = str(millisecond(ts));
	t = strcat(v);
	t_sh = strcat([v[0],v[1],v[2],v[3],v[4]]);
	return;		
}

module CORE_INT
{
	//-------------------------------------------------------
	//  rgb
	//------------------------------------------------------- 
	integer rgb(integer r, integer g, integer b){
		return b*65536+g*256+r;
	}
	integer rgb_neg_number(){
		return rgb(255,0,0); 
	}
	integer rgb_pos_number(){
		return rgb(0,0,255); 
	}
	//-------------------------------------------------------
	//  format_rgb_n
	//------------------------------------------------------- 
	text_rgb format_rgb_n(string option(nullable) res, number option(nullable) res_numb, string option(nullable) tooltip)
	{
		if(null(tooltip))
			tooltip = "";
		
		if(null(res_numb) || null(res)) {
			return text_rgb("", rgb(0,0,0),rgb(255,255,255), rgb(0,0,0),rgb(255,255,255),tooltip);
		}
		else {
			if(res_numb < 0)
				return text_rgb(res, rgb_neg_number(),rgb(255,255,255), rgb(0,0,0),rgb(255,255,255),tooltip);
			else
				return text_rgb(res, rgb_pos_number(),rgb(255,255,255), rgb(0,0,0),rgb(255,255,255),tooltip);
		}
	}

	/*-----------------------------------------------------------------------
	  Name :    str_p
	  Notes :       
	  Changes:
	  			Date    		Who     Detail
	  			2-Aug-2016 		MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	string str_p(	logical 	is_curr,
					number 	option(nullable)	val, 
					string option(nullable) thousand_sep = null<string>)
	
	
	{	
		if(null(val))
		return "";

		integer dec = is_curr? 0: 5;
	
		if(is_curr)
			return str_curr(val, dec, thousand_sep);
		else
			return  str(val,dec);			
	}

	string str_p(	number 	option(nullable)	val,
					integer dec)
	
	
	{	
		if(null(val))
			return "";

		return  str(val,dec);			
	}
	
	//-------------------------------------------------------
	//str_rate
	//-------------------------------------------------------
	string str_rate(number option(nullable)  val,
					integer dec)
	{
		if(null(val))
			return "";
		return strcat(str(val, dec)," %");
	}
	//-------------------------------------------------------
	//str_date
	//-------------------------------------------------------
	string str_date(date option(nullable)  val)
	{
		if(null(val))
			return "";
		return str(val);
	}
	//------------------------------------------------------
	//str_rate
	//-------------------------------------------------------
	string str_rate(number option(nullable)  val)
	{
		if(null(val))
			return "";
		return strcat(str(val, 3)," %");
	}
	//-------------------------------------------------------
	//str_bp
	//-------------------------------------------------------
	string str_bp(number option(nullable)  val, integer dec)
	{
		if(null(val))
			return "";
		if(val>=0)
			return strcat(strcat("+",str(val, dec))," bp");
		else
			return strcat(str(val, dec)," bp");
	}
	//-------------------------------------------------------
	//str_bp
	//-------------------------------------------------------
	string str_bp(number option(nullable) val)
	{
		if(null(val))
			return "";
		if(val>=0)
			return strcat(strcat("+",str(val, 2))," bp");
		else
			return strcat(str(val, 2)," bp");
	}

	//-------------------------------------------------------
	// match
	//-------------------------------------------------------
	logical match(	string 			s,
					vector(string) 	vs,						
					out integer		idx)
	{
		idx = -1;

		if(v_size(vs)== 0)
			return false;
			
		vector(integer) v_idx = v_findi(vs, [s]) ;

		if(null(v_idx) || v_size(v_idx) == 0){
			if(s=="12M") {
				v_idx = v_findi(vs, ["1Y"]) ;
				if(null(v_idx) || v_size(v_idx) == 0)
					return false;
						
			}
			else if(s=="1Y") {
				v_idx = v_findi(vs, ["12M"]) ;
				if(null(v_idx) || v_size(v_idx) == 0)
					return false;
			}
			else if(s == "1D") {
				v_idx = v_findi(vs, ["BD1"]) ;
				if(null(v_idx) || v_size(v_idx) == 0){
					v_idx = v_findi(vs, ["CD1"]) ;
					if(null(v_idx) || v_size(v_idx) == 0)
						return false;
				}
			}
			else
				return false;

		}
		
		idx = v_idx[0];

		return true;
	}
	//-------------------------------------------------------
	// match
	//-------------------------------------------------------
	logical match(	tenor_code 		s,
					vector(tenor_code) 	vs,						
					out integer		idx)
	{
		for(integer i=0; i < v_size(vs); i++) {
			if( s == vs[i]) {
				idx = i;
				return true;
			}
		}

		return false;
	}
	
	//-------------------------------------------------------
	// match
	//-------------------------------------------------------
	logical match(	date 			d,
					vector(date) 	vd,						
					out integer 	idx)
	{
		idx = -1;

		if(v_size(vd)== 0)
			return false;
		
		vector(date) ref = [d];
		vector(integer) v_idx = ..v_find(vd, ref) ;

		if(null(v_idx) || v_size(v_idx) == 0)
			return false;
		
		idx = v_idx[0];

		return true;

	}
	
	//-------------------------------------------------------
	// consolidate
	//-------------------------------------------------------
	void consolidate(	vector(date) 		vd,						
						vector(number) 		va,	
						out vector(date) 	vd_c,
						out vector(number) 	va_c)
	{
		integer c = 0;
		vector(logical) incl = [true];		
		for(integer i = 1; i < v_size(va); i++) {
			if(vd[i] == vd[i-1]) {
				push_back(incl,false);
				va[i-1] += va[i];
				c++;
			}
			else
				push_back(incl,true);
		}
		if(c > 0) {
			resize(vd_c,0);
			resize(va_c,0);
			for(integer i = 0; i < v_size(va); i++) {
				if(incl[i]) {
					push_back(vd_c, vd[i]);
					push_back(va_c, va[i]);
				}				
			}
		}
		else {
			vd_c = ..clone_vector(vd);
			va_c = ..clone_vector(va);
		}
	}

		
	/*-----------------------------------------------------------------------
	  Name :    num1der  (central difference)
	  Notes :   ex 1yr simple rate, r = 0.05, delta 0.0001, nom = 1
				dpdy = 1.727675         
	  Changes:
	  			Date    		Who     Detail
	  			6-Dec-2013 		MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	number num1der(	number pv_x_up,
					number pv_x_down,
					number dx) 
	{    
	    number div = 2*dx;
		if(is_equal(div,0, CORE_INT.DBL_EPSILON))  return null<number>;
		
		return (pv_x_up-pv_x_down)/div ;
	}
	/*-----------------------------------------------------------------------
	  Name :    num1der_fd, num1der_bd  (forward difference, backward difference)
	  ----------------------------------------------------------------------*/
	number num1der_fd(	number pv_x,
						number pv_x_up,
						number dx) 
	{    
		if(is_equal(dx,0,CORE_INT.DBL_EPSILON))
			return null<number>;
		return (pv_x_up-pv_x)/dx ;
	}
	
	number num1der_bd(	number pv_x,
						number pv_x_down,
						number dx) 
	{    
		if(is_equal(dx,0, CORE_INT.DBL_EPSILON))
			return null<number>;
		return (pv_x-pv_x_down)/dx ;
	}
	//-----------------------------------------------------------------------------
	//  num1der_rel
	//-----------------------------------------------------------------------------
	number  num1der_rel(number pv,
						number pv_y_up,
						number pv_y_down,
						number dy) 
	{    
		number der = num1der(pv_y_up,pv_y_down,dy ) ;
		if(is_equal_zero(pv, CORE_INT.DBL_EPSILON))
			return null<number>;
		return der/pv;
	}
	/*-----------------------------------------------------------------------
	  Name :    num2der (symmetrical)
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			6-Dec-2013 		MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	number num2der(	number pv_x,
					number pv_x_up,
					number pv_x_down,
					number dx) 
	{    
	    number div = dx*dx;
		if(is_equal(div,0, CORE_INT.DBL_EPSILON))
			return null<number>;
		
		return (pv_x_up+pv_x_down-2*pv_x)/div;
	}
		
		
	//-------------------------------------------------------
	// rate
	// OBS! this function suffers from the "interpolation on rate problem"  
	//-------------------------------------------------------
	number rate(	date 		d, 
					date 		mat1, 
					date 		mat2, 
					disc_func 	df, 
					day_count_method dc, 
					rate_type 	rt)
	{	
		number df1 = df.disc((mat1-d)/365);
		number df2 = df.disc((mat2-d)/365);	
		return disc_to_rate(df2/df1, rt, mat1, mat2, dc);	
	}

	/*-----------------------------------------------------------------------
	  Name :    pv_diff
	  Notes :           
	  Changes:
	  			Date    		Who     Detail
	  			11-Dec-2013 	MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	number pv_diff(	number 		spot_cf, 
			    	number 		rate, 
					number 		rate_shift,
			    	rate_type 	rt, 
			    	date 		settle_date, 
			    	date 		mat, 
			    	day_count_method dc,
					out number end_cf) 
	{									
	    number df 	= present_value(1, rate, settle_date,  mat, dc,  rt);
		end_cf 		= spot_cf/df;

		if(is_equal_zero(rate_shift))
			return 0;
						
		number pv_shift = end_cf * present_value(1, rate + rate_shift, settle_date,  mat, dc,  rt);
		return pv_shift - spot_cf;
	}
	/*-----------------------------------------------------------------------
	  Name :    pv_diff_end
	  Notes :           
	  Changes:
	  			Date    		Who     Detail
	  			11-Dec-2013 	MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	number pv_diff_end(	number 		end_cf, 
				    	number 		rate, 
						number 		rate_shift,
				    	rate_type 	rt, 
				    	date 		settle_date, 
				    	date 		mat, 
				    	day_count_method dc,
						out number spot_cf) 
	{									
	    number df 	= present_value(1, rate, settle_date,  mat, dc,  rt);
		spot_cf = end_cf * df;

		if(is_equal_zero(rate_shift))
			return 0;
				
		number pv_shift = end_cf * present_value(1, rate + rate_shift, settle_date,  mat, dc,  rt);
		return pv_shift - spot_cf;
	}


	class enum ip_type {
		IP_LIN option(str: "Linear"),
		IP_SPLINE option(str: "Spline"),
		IP_HERMITE option(str: "Hermite")
	};
	
	//-------------------------------------------------------
	// ip
	//-------------------------------------------------------
	number ip(	vector(number) 	knot_x,	
				vector(number) 	knot_y, 																						
				number 			x_ip,
				ip_type			ip_t,										
				logical 	 	extrap_ok 		= false,
				logical 	 	extrap_const 	= true)
	{
		try { 
			number y_ip;
			integer size = v_size(knot_x);
			y_ip = null<number>;
			interpolation ip;
			number extrap_der = extrap_const ? 0 : null<number>;
	
			if( knot_x[0] > knot_x[size-1]){
	
				if(!extrap_ok && (x_ip > knot_x[0] || x_ip < knot_x[size-1]))
					QL_FAIL("outside interpolation range");
		
				vector(number) knot_x_r = clone_vector(knot_x);
				vector(number) knot_y_r = clone_vector(knot_y);
	
				reverse(knot_x_r);
				reverse(knot_y_r);
	
				if(ip_t == ip_type.IP_LIN)
					ip = extrap_ok ? 	interpolation(ip_linear(extrap_der,extrap_der),knot_x_r, knot_y_r) : 
										interpolation(ip_linear(null<number>,null<number>, IP_NO_XPOL_L | IP_NO_XPOL_R),knot_x_r, knot_y_r);																		
				else if(ip_t == ip_type.IP_SPLINE)
					ip = extrap_ok ? 	interpolation(ip_spline(extrap_der,extrap_der),knot_x_r, knot_y_r) : 
										interpolation(ip_spline(null<number>,null<number>, IP_NO_XPOL_L | IP_NO_XPOL_R),knot_x_r, knot_y_r);	
					//y_ip = ip_spline(x_ip, knot_x_r, knot_y_r);
				else if(ip_t == ip_type.IP_HERMITE)
					ip = extrap_ok ? 	interpolation(ip_hermite_catmull_rom(extrap_der,extrap_der),knot_x_r, knot_y_r) : 
										interpolation(ip_hermite_catmull_rom(null<number>,null<number>, IP_NO_XPOL_L | IP_NO_XPOL_R),knot_x_r, knot_y_r);	
					//y_ip = ip_hermite(x_ip, knot_x_r, knot_y_r);
				else	
					QL_FAIL("invalid interpolation type");
	
			}
			else {
				if(!extrap_ok && (x_ip < knot_x[0] || x_ip > knot_x[size-1]))
					QL_FAIL("outside interpolation range");
	
				if(ip_t == ip_type.IP_LIN)
					ip = extrap_ok ? 	interpolation(ip_linear(extrap_der,extrap_der),knot_x, knot_y) : 
										interpolation(ip_linear(null<number>,null<number>, IP_NO_XPOL_L | IP_NO_XPOL_R),knot_x, knot_y);
				else if(ip_t == ip_type.IP_SPLINE)
					ip = extrap_ok ? 	interpolation(ip_spline(extrap_der,extrap_der),knot_x, knot_y) : 
										interpolation(ip_spline(null<number>,null<number>, IP_NO_XPOL_L | IP_NO_XPOL_R),knot_x, knot_y);	
				else if(ip_t == ip_type.IP_HERMITE)
					ip = extrap_ok ? 	interpolation(ip_hermite_catmull_rom(extrap_der,extrap_der),knot_x, knot_y) : 
										interpolation(ip_hermite_catmull_rom(null<number>,null<number>, IP_NO_XPOL_L | IP_NO_XPOL_R),knot_x, knot_y);
				else	
					QL_FAIL("invalid interpolation type");
			}
			return ip.eval(x_ip);							
		}	
		catch {
			CORE_INT.write_warning_message(err.message());
			return null<number>;
		}
	}	
	//-------------------------------------------------------
	// ip_lin
	//-------------------------------------------------------
	number ip_lin(	number 	x0, 
					number 	x1, 
					number 	y0, 
					number 	y1, 
					number 	x,
					logical extrap_ok = false)

	{
		vector(number) knot_x = [x0, x1];
		vector(number) knot_y = [y0, y1];
		//number y_ip;
		return ip(knot_x,knot_y,x,ip_type.IP_LIN,extrap_ok);					
	}	
	//-------------------------------------------------------
	// ip_lin
	//-------------------------------------------------------
	number ip_lin(	date 	base,
					date 	x0, 
					date 	x1, 
					number 	y0, 
					number 	y1, 
					date 	x,
					logical extrap_ok = false)

	{		
		vector(number) knot_x = [x0-base, x1-base];
		vector(number) knot_y = [y0, y1];
		//number y_ip;
		return ip(knot_x,knot_y,x-base,ip_type.IP_LIN,extrap_ok);		
	}
	//-------------------------------------------------------
	// ip_lin_2
	//-------------------------------------------------------
	number ip_lin_2(number 	x1_1, 
					number 	x2_1, 
					number 	x1_2, 
					number 	x2_2, 
					number 	y1_1, 
					number 	y2_1, 
					number 	y1_2, 
					number 	y2_2, 
					number 	x1,
					number 	x2)
	{
		try {
			number 	x 		= x2_1 - x1 ;
			number 	xx 		= x1 - x1_1 ;
			number den 		= (x2_1-x1_1)*(x2_2-x1_2);
			QL_REQUIRE(den != 0, "invalid interpolation range");	
			number term1 	= x*(x2_2-x2)*y1_1 + x*(x2-x1_2)*y1_2;
			number term2 	= xx*(x2_2-x2)*y2_1 + xx*(x2-x1_2)*y2_2;
			return term1/den + term2/den;
		}	
		catch {
			CORE_INT.write_warning_message(err.message());
			return null<number>;
		}
	}	

	//-------------------------------------------------------
	// interpol_2d  WORK IN PROGRESS
	//-------------------------------------------------------
	/*class interpol_2d
	{
		interpol_2d(vector(number) 	knot_x,	
					vector(number) 	knot_y,
					matrix(number)	data,
					ip_type			ip_type,										
					logical 	 	allow_extrap = false,
					logical 	 	extrap_const = true);
		
		number max_x();
		number min_x();
		number max_y();
		number min_y();
		integer locate_x(number);
		integer locate_y(number);
		logical in_range(number x, number y);
		number eval(number ,number,logical option(nullable) allow_extrap=null<logical>);
		
	protected:
		logical check_range(number , number , logical);
		
		vector(number) 	knot_x_;
		vector(number) 	knot_y_;
		matrix(number)	data_;
		logical			allow_extrap_;
		interpolation 	i_;
		integer 		size_x_;
		integer 		size_y_;
		
	};
	//-------------------------------------------------------
	// constructor 
	//-------------------------------------------------------
	interpol_2d.interpol_2d(vector(number) 	knot_x,	
							vector(number) 	knot_y,
							matrix(number)	data,
							ip_type			ip_type,										
							logical 	 	allow_extrap ,
							logical 	 	extrap_const )
							:knot_x_(knot_x),knot_y_(knot_y),data_(data),
							 allow_extrap_(allow_extrap),
							 size_x_(v_size(knot_x_)), size_y_(v_size(knot_y_))
	{
		
		QL_REQUIRE(size_x_ >= 2, "not enough elements in x-vector");
        QL_REQUIRE(size_y_ >= 2, "not enough elements in y-vector");
		QL_REQUIRE(size_y_ != n_rows(data_), "");
		QL_REQUIRE(size_x_ != n_cols(data_), "");
		
		for (integer i=0, j=1; j<size_x_; i++) {
			QL_REQUIRE(knot_x_[j] > knot_x_[i], "unsorted x values");
			++j;
		}
		
		for (integer i=0, j=1; j<size_y_; i++) {
			QL_REQUIRE(knot_y_[j] > knot_y_[i], "unsorted y values");
			++j;
		}
		
		
	}

	number interpol_2d.max_x() { return knot_x_[size_x_-1]; }
	number interpol_2d.min_x() { return knot_x_[0]; }
	number interpol_2d.max_y() { return knot_y_[size_y_-1]; }
	number interpol_2d.min_y() { return knot_y_[0]; }

	integer interpol_2d.locate_x(number x)
	{
		if (x <= min_x())
            return 0;
		else if (x >= max_x())
			return size_x_-1;
		else {
			for(integer i=1;i<size_x_;i++)
				if(x < knot_x_[i]) return i-1;			
		}
	}
	integer interpol_2d.locate_y(number y)
	{
		if (y <= min_y())
            return 0;
		else if (y >= max_y())
			return size_y_-1;
		else {
			for(integer i=1;i<size_y_;i++)
				if(y < knot_y_[i]) return i-1;			
		}
	}
	
	logical interpol_2d.in_range(number x, number y)
	{
				
		number x1 = min_x();
		number x2 = max_x();
		
		logical in = (x >= x1 && x <= x2);
		if (!in) return false;
				
		number y1 = min_y();
		number y2 = max_y();
		
		return (y >= y1 && y <= y2);
		
	}
	logical interpol_2d.check_range(number x, number y, logical allow_extrap)
	{
		if(allow_extrap) return true;
		return in_range( x,  y);
	}
	
	number interpol_2d.eval(number x, number y,logical option(nullable) allow_extrap)
	{
		if(null(allow_extrap)) allow_extrap = allow_extrap_;
		   
		integer j = locate_y(y);
		number z1,z2;
		if(x <= min_x()) {
			z1 = data_[j,0];
			z2 = data_[j+1,0];
		}
		else {
			integer i = locate_x(x);
			if(x == knot_x_[i]) {
				z1 = data_[j,i];
				z2 = data_[j+1,i];
			}
			else {
				z1 = data_[j,i];
				z2 = data_[j+1,i+1];
			}
		}

		number u = (y-knot_y_[j])/(knot_y_[j+1]-knot_y_[j]);

		return (1.0-u)*z1 + u*z2;
    } */

	//-------------------------------------------------------
	// clean_mat_code
	//-------------------------------------------------------
	string clean_mat_code(string option(nullable) mc)
	{
		if(null(mc))
			return null<string>;

		number p = str_find(mc, ";") ;
		if(!null(p) && p > 0)
			return null<string>;//compounded code
		
		p = str_find(mc, "_") ;
		if(!null(p) && p > 0)
			mc = sub_string_start (mc, integer(p));

		p = str_find(mc, "eom") ;
		if(!null(p) && p > 0)
			mc = sub_string_start (mc, integer(p));

		integer size = strlen(mc);		
		if(size < 2)
			return null<string>;
		
		string first = str_to_upper(sub_string(mc, 0, 1)) ;
		
		if(is_digit(first, 0))
		   return mc;//ex. "1M"
		
		string first_dig = size >= 2 ? sub_string(mc, 1, 1) : "";
		if(!is_digit(first_dig, 0))
			return mc;//"BD1", "IMM1"

		integer i = 0;
		
		vector(string) digs = [first_dig];
		while(size >= 3+i && is_digit(sub_string(mc, 2+i, 1),0)) {
			push_back(digs,sub_string(mc, 2+i, 1));
			i++;
		}

		string c = strcat(digs);
		
		if(first == "M"){
			
			return strcat(c,"M");
		}
		else if(first == "Y"){
			
			return strcat(c,"Y");			
		}
		else if(first == "W"){
			
			return strcat(c,"W");			
		}
		else{
			return null<string>;
		}
	}

	vector(number) clone_vector(vector(number) option(nullable) v)
	{
		if(null(v))
			return null<vector(number)>;
		else 
			return ..clone_vector(v);
	}

	vector(date) clone_vector(vector(date) option(nullable) v)
	{
		if(null(v))
			return null<vector(date)>;
		else 
			return ..clone_vector(v);
	}
	vector(string) clone_vector(vector(string) option(nullable) v)
	{
		if(null(v))
			return null<vector(string)>;
		else 
			return ..clone_vector(v);
	}
	//returns a consolidated date vector i.e. vector of unique dates
	vector(date) consolidate_date(	vector(date) v1,
									vector(date) v2,
									out vector(integer) v1_i,//index of v1 in the common vector
									out vector(integer) v2_i)//index of v2 in the common vector
{
		QL_REQUIRE(v_size(v1)>0, "invalid vector size (v1)");
		QL_REQUIRE(v_size(v2)>0, "invalid vector size (v2)");
		
		vector(integer) f 	= ..v_find(v1, v2);			//common dates
		integer add 		= v_size(v2) - v_size(f);	//how many missing
		
		if(add>0){
			vector(date) con;
			if(add == v_size(v2)){
				con = concat(v1,v2);
			}
			else {
				vector(integer) _f 	= index_vector(v_size(v2));
				vector(integer) f_c = v_complement (_f, f);		//non-common index vector
				vector(date) _t 	= sub_vector(v2,f_c);		//get non-common dates
				con 				= concat(v1,_t);			//non sorted merged date vector
			}
			sort (con, con);							

			resize(v1_i,v_size(con));
			resize(v2_i,v_size(con));
			v2_i = v1_i = -1;

			vector(integer) f_ir 	= ..v_find(con, v1);
			vector(integer) f_p 	= ..v_find(con, v2);
			
			for(integer i;i<v_size(f_ir);i++)			
				v1_i[f_ir[i]] = i;
			
			for(integer i;i<v_size(f_p);i++)			
				v2_i[f_p[i]] = i;
			
			return con;
		}
		else {
			v1_i = index_vector(v_size(v1));
			resize(v2_i,v_size(v1_i));
			v2_i = -1;
			for(integer i;i<v_size(f);i++){
				v2_i[f[i]] = i;
			}

			return ..clone_vector(v1);
		}

		
	}

	//-------------------------------------------------------
	// set_to_zero
	//-------------------------------------------------------
	void set_to_zero(out vector(number)   v)
	{
		try {
			if(null(v))
				return;
			
			v = 0;

			return;
		}
				
		catch {}						
				
	}
	//-------------------------------------------------------
	// set_to_empty
	//-------------------------------------------------------
	void set_to_empty(out vector(string) v)
	{
		try {
			if(null(v))
				return;

			v = "";

			return;
		}
				
		catch {}						
				
	}
	//-------------------------------------------------------
	// set_to_empty
	//-------------------------------------------------------
	void set_to_empty(out vector(text_rgb)  v)
	{
		try {
			if(null(v))
				return;

			v = text_rgb('', null<number>,null<number>, false);
			
			return;
		}
				
		catch {}						
				
	}
	//-------------------------------------------------------
	// set_to_empty
	//-------------------------------------------------------
	void set_to_empty(	integer start,
						integer end,
						out vector(string)	 v)
	{
		try {
			if(null(v))
				return;

			integer size = v_size(v);
				
			if(size == 0 || start < 0 || start >= size || end < 0 || end >= size || end < start)
				return;

			v = "";
			
			return;
		}
				
		catch {}						
				
	}
	//-------------------------------------------------------
	// set_to_null
	//-------------------------------------------------------
	void set_to_null(out vector(number)  v)
	{
		try {
			if(null(v))
				return;

			v = null<number>;

			return;
		}
				
		catch {}						
				
	}
	//-------------------------------------------------------
	// set_to_null
	//-------------------------------------------------------
	void set_to_null(out vector(text_rgb)  v)
	{
		try {
			if(null(v))
				return;

			v = null<text_rgb>;

			return;
		}
				
		catch {}						
				
	}
	//-------------------------------------------------------
	// set_to_null
	//-------------------------------------------------------
	void set_to_null(	integer start,
						integer end,
						out vector(number)  v)
	{
		try {
			if(null(v))
				return;

			integer size = v_size(v);
				
			if(size == 0 || start < 0 || start >= size || end < 0 || end >= size || end < start)
				return;

			v = null<number>;
			
			return;
		}
				
		catch {}						
				
	}
	//-------------------------------------------------------
	// set_to_null
	//-------------------------------------------------------
	void set_to_null(out vector(date) v)
	{
		try {
			if(null(v))
				return;

			v = null<date>;
			
			return;
		}
				
		catch {}						
				
	}
	//-------------------------------------------------------
	// set_to_null
	//-------------------------------------------------------
	void set_to_null(out matrix(number)  m)
	{
		try {
			if(null(m))
				return;

			integer rows = n_rows(m);
			integer cols = n_cols(m);
	
			if(rows == 0)
				return;
	
			for(integer i = 0; i < rows; i++ )
				for(integer j = 0; j < cols; i++ )
					m[i,j] = null<number>;
			
			return;
		}
				
		catch {}						
				
	}

	//-------------------------------------------------------
	// check_date_order
	//-------------------------------------------------------
	logical check_date_order( vector(date) dates, out string e)
	{
		e = "";
		for(integer i=0; i < v_size(dates); i++){
			if(null(dates[i])){
				e = "null date in date vector"; 
				return false;
			}
			if(i>0){
				if(dates[i] <= dates[i-1]){
					e = "vector not sorted";
					return false;
				}
			}			
		}
		return true	;	
	}
	//-------------------------------------------------------
	// check_order
	//-------------------------------------------------------
	logical check_order( vector(number) d, out string e)
	{
		for(integer i=0; i < v_size(d); i++){
			if(null(d[i])){
				e = "null value in vector"; 
				return false;
			}
			if(i>0){
				if(d[i] <= d[i-1]){
					e = "vector not sorted";
					return false;
				}			
			}			
		}
		return true	;	
	}

	//-------------------------------------------------------
	//  warning_admin
	//-------------------------------------------------------
	void warning_admin(	string 		s,
						out integer k,
						out vector(string) msg)
	//option (category: 'FIWS/Utility')
	{
		try{
			if(v_size(msg) <= k)
				return;
	
			if(k==0){
				msg[k] = "Warning(s)!";
				k++;
			}	
			msg[k] = strcat([str(k),") ", s]);			
			k++;
		}
		catch {
		}
	}
	//-------------------------------------------------------
	//  
	//-------------------------------------------------------
	void info_admin(	string 			s,
						integer 		k_start,
						out integer 	k,
						out integer 	k_disp,
						out vector(string) msg,
						logical 	chg_line)	
	//option (category: 'FIWS/Utility')			
	{		
		try{
			if(k==0){
				if(v_size(msg) <= k_start)
					return;
				msg[k_start] = "Note(s)!";
				k++;
				k_disp++;
			}	
			
			if(v_size(msg) <= k_start+k)
				return;
	
			if(chg_line){
				msg[k_start+k] = strcat([str(k_disp),") ", s]);				
				k_disp++;
			}
			else
				msg[k_start+k] = s;
		
			k++;
		}
		catch {
		}
	}

	//-------------------------------------------------------
	//  msg_err
	//-------------------------------------------------------
	void msg_err(string option(nullable) curr_ver,
						string option(nullable) last_ver,
						string					err_msg,
						out vector(string) 		msg)
	//option (category: 'FIWS/Utility')				
	{
		try {		
			if(!null(curr_ver) || !null(last_ver)){
				if(v_size(msg)<4)
					resize(msg,4);
			}
			else {
				if(v_size(msg)<2)
					resize(msg,2);
			}					

			string algo = !null(last_ver) ? "QUANTLAB DEMO" : "";
			//string algo_tel = "[ Support: +46 (8)4404400 ]";
			integer s = v_size(msg);
	
			msg[0] = "Error! Calculation unsuccessful." ;
			msg[1] = err_msg ;	
	
			/*if (s >= 5){						
				if(!null(curr_ver) && !null(last_ver)){	
					msg[s-3] = algo;
					msg[s-2] = algo_tel;								
					msg[s-1] = strcat(["Version ", curr_ver, " [latest: ", last_ver, "]"]);
				}
				else if(!null(last_ver)){					
					msg[s-3] = algo;
					msg[s-2] = algo_tel;											
					msg[s-1] = strcat("Latest version ", last_ver );
				} 
				else if(!null(curr_ver))													
					msg[s-1] = strcat("Version ", curr_ver );			
			}
			else*/ if (s >= 4){						
				if(!null(curr_ver) && !null(last_ver)){	
					msg[s-2] = algo;								
					msg[s-1] = strcat(["Version ", curr_ver, " [latest: ", last_ver, "]"]);
				}
				else if(!null(last_ver)){					
					msg[s-2] =  algo;											
					msg[s-1] = strcat("Latest version ", last_ver );
				} 
				else if(!null(curr_ver))													
					msg[s-1] = strcat("Version ", curr_ver );			
			}					
		}
		catch {
		}		
	}

	//-------------------------------------------------------
	//  msg_end
	//-------------------------------------------------------
	public void msg_end(string option(nullable) curr_ver,
						string option(nullable) last_ver,
						out vector(string) 		msg)	
	//option (category: 'FIWS/Utility')			
	{
		try {
			integer s = v_size(msg);
			if(s == 0)
				resize(msg,2);

			string algo = !null(last_ver) ? "QUANTLAB DEMO" : "";
			//string algo_tel = "[ Support: +46 (8)4404400 ]";
			string d, t;
			format_time(d, t);
			string dt =  strcat(["[",d,"]  ", t]);

			/*if (s >= 5 && strlen(algo) > 1 && msg[s-5] == ''){
				msg[s-5] = algo;
				msg[s-4] = algo_tel;
				msg[s-3] = 'Time of last update';
				if(!null(curr_ver) && !null(last_ver)){
					msg[s-2] = dt;				
					msg[s-1] = strcat(["Version ", curr_ver, " [latest: ", last_ver, "]"]);
				}				
				else {					
					msg[s-2] = dt;										
					msg[s-1] = strcat("Latest version ", last_ver );
				} 				
			}
			else*/ if (s >= 4 && strlen(algo) > 1 && msg[s-4] == ''){
				msg[s-4] = algo;
				msg[s-3] = 'Time of last update';
				if(!null(curr_ver) && !null(last_ver)){
					msg[s-2] = dt;				
					msg[s-1] = strcat(["Version ", curr_ver, " [latest: ", last_ver, "]"]);
				}				
				else {					
					msg[s-2] = dt;										
					msg[s-1] = strcat("Latest version ", last_ver );
				} 				
			}
			else if (s >= 3){				
				if(!null(curr_ver) && !null(last_ver)){
					msg[s-3] = strcat('Time of last update', algo );
					msg[s-2] = dt;
					msg[s-1] = strcat(["Version ", curr_ver, " [latest: ", last_ver, "]"]);
				}
				else if(!null(curr_ver)){
					msg[s-3] = 'Time of last update';
					msg[s-2] = dt;										
					msg[s-1] = strcat("Version ", curr_ver );
				}
				else if(!null(last_ver)){
					msg[s-3] = strcat('Time of last update', algo );
					msg[s-2] = dt;											
					msg[s-1] = strcat("Latest version ", last_ver );
				} 
				else {
					msg[s-2] = 'Time of last update';
					msg[s-1] = dt;
				}
			}			
			else if( s == 2){
				msg[s-2] = strlen(algo) > 1 ? strcat('Time of last update', algo ) : 'Time of last update';								
				msg[s-1] = dt; 
			}
			else
				msg[s-1] = strcat('Time of last update: ', dt);
			
		}
		catch {
		}
	}
}
//a fwd_func can currently not handle a 1d-tenor and therefore is also missing from the fwd_func_tenor enum 
class enum fwd_func_tenor option (category: 'Yield Curve/003 Curve Building')
{
	FT_1D option(str: "1D"),
	FT_1M option(str: "1M"),
	FT_3M option(str: "3M"),
	FT_6M option(str: "6M"),
	FT_12M option(str: "12M")
};
	
tenor_code fwd_func_tenor_to_tc(fwd_func_tenor e) option (category: 'Yield Curve/003 Curve Building')
{
	switch(e)
	{
		case fwd_func_tenor.FT_1D : return tenor_code(1,tenor_unit.BD);
		case fwd_func_tenor.FT_1M : return tenor_code(1,tenor_unit.M);
		case fwd_func_tenor.FT_3M : return tenor_code(3,tenor_unit.M);
		case fwd_func_tenor.FT_6M : return tenor_code(6,tenor_unit.M);
		case fwd_func_tenor.FT_12M : return tenor_code(12,tenor_unit.M);
		default : return null<tenor_code>;
	}
}

fwd_func_tenor fwd_func_tenor_from_tc(tenor_code e) option (category: 'Yield Curve/003 Curve Building')
{
	
	switch(e.frequency())
	{
		case 1 : return fwd_func_tenor.FT_12M;
		case 2 : return fwd_func_tenor.FT_6M;
		case 4 : return fwd_func_tenor.FT_3M;
		case 12 : return fwd_func_tenor.FT_1M;
		case 365 : return fwd_func_tenor.FT_1D;
		default : return null<fwd_func_tenor>;
	}
}


integer fwd_func_tenor_to_freq(fwd_func_tenor e) option (category: 'Yield Curve/003 Curve Building') 
{
	switch(e) {
	case fwd_func_tenor.FT_1D: return 365;
	case fwd_func_tenor.FT_1M: return 12;
	case fwd_func_tenor.FT_3M: return 4;
	case fwd_func_tenor.FT_6M: return 2;
	case fwd_func_tenor.FT_12M: return 1;
	default:
		QL_FAIL("invalid tenor");	
	}
}

fwd_func_tenor fwd_func_tenor_from_freq(integer freq) option (category: 'Yield Curve/003 Curve Building') 
{
	tenor_code e = tenor_code(freq);
	return fwd_func_tenor_from_tc(e);
}