/*	
	z spread functions
	Developed by Algorithmica Research, Magnus Nyström
			
	ToDo: 
	
	Comments:	

*/
option(null: hard);

module CORE_INT
{
	
	//-------------------------------------------------------
	// check_cf_vect
	//-------------------------------------------------------
	void check_cf_vect(	date 			settle,
						vector(date) 	cf_dates,
						vector(number)	cf)
	{
		integer size = v_size(cf);
		if(size == 0 || size != v_size(cf_dates))
			QL_FAIL("invalid vector size");

		string e;
		if(!check_date_order(cf_dates,e))
			QL_FAIL(e);

		if(cf_dates[0] <= settle)
			QL_FAIL("invalid date in vector");		
	}
	//-------------------------------------------------------
	// check_cf_vect
	//-------------------------------------------------------
	void check_cf_vect(	vector(number) 	cf_per,
						vector(number)	cf)
	{
		integer size = v_size(cf);
		if(size == 0 || size != v_size(cf_per))
			QL_FAIL("invalid vector size");
		for(integer i = 0; i < size; i++ ){
			if(cf_per[i] <= 0)
				QL_FAIL("invalid period (<=0)");	
		}	
	}
}

//-------------------------------------------------------
//class  z_sprd_solve 
//-------------------------------------------------------	
class z_sprd_solve 
{
	public:
	
	vector(number) 	cf100_;
	vector(number) 	per_;
	number 			n_;
	number  		dp100_;
	vector(number) 	disc_r_;
	rate_type		rt_;
	number 			f(number s) ;		
};
	
number z_sprd_solve.f(number s)
{		
	number pv100 = 0;
	for(integer i = 0; i < n_; i++ ) {

		number df = present_value(1,disc_r_[i]+s,per_[i], rt_);
		//number df = 1/pow(1.0+(disc_r_[i]+s), per_[i]);
		pv100 += cf100_[i]*df	;
	}
	return dp100_-pv100;
}	
//-------------------------------------------------------
// gen_z_spread  (yld sprd based on cashflows)
//-------------------------------------------------------
number gen_z_spread(	date 			d,
						disc_func		df,
						date 			settle,
						vector(date) 	cf_dates,
						vector(number)	cf,					
						number 			dp,
						number 			nom,
						number option(nullable)	sprd_guess,
						rate_type		rt,						
						day_count_method dc) option(category: "Financial")	
{
	//try {
		CORE_INT.check_cf_vect(settle,cf_dates,cf);
		z_sprd_solve z = new z_sprd_solve;
						
		z.n_ 		= v_size(cf);	
		z.cf100_ 	= cf/nom * 100;
		z.per_ 		= period(settle, cf_dates, dc);		
		z.dp100_ 	= dp/nom*100;
		z.rt_ 		= rt;
		z.disc_r_	= df.zero_rate(d, settle, cf_dates,rt, dc) ;
	
		number x1 	= null(sprd_guess) ? -z.disc_r_[0]+0.001 : sprd_guess - 0.001;
		number x2 	= null(sprd_guess) ? 0.01: sprd_guess + 0.001;

		number f1 = z.f(x1);//neg
		QL_REQUIRE(!null(f1), "boundary calculation error (f1)");
		number i = 0;
		number max_i = 500;
		while (f1>= 0 && i < max_i){
			x1 -= 0.01;
			f1 = z.f(x1);
			QL_REQUIRE(!null(f1), "boundary calculation error (f1)");
			i++;
		}
		if(i >= max_i)
			QL_FAIL("max iterations reached");

		number f2 = z.f(x2);//neg
		QL_REQUIRE(!null(f2), "boundary calculation error (f2)");
		i = 0;
		while (f2<= 0 && i < max_i){
			x2 += 0.01;
			f2 = z.f(x2);
			QL_REQUIRE(!null(f2), "boundary calculation error (f2)");
			i++;
		}
		if(i >= max_i)
			QL_FAIL("max iterations reached");

		number tol 	= 0.0000001;
		
		zero_bracket(z, &z_sprd_solve.f, x1, x2);
		return zero_brent(z, &z_sprd_solve.f, x1, x2, tol);
	//}
	//catch {
		//warning(err.message());
		//return null<number>;
	//}
}

//-------------------------------------------------------
// gen_i_spread 
//-------------------------------------------------------
vector(number) gen_i_spread(vector(date) mat_v,
							vector(number) y_v,
							vector(date) mat_ip_v,
							vector(number) y_ip_v) option(category: "Financial")	
{
	string e;
	logical ok = CORE_INT.check_date_order(mat_v,e);
	QL_REQUIRE(ok,e);
	integer size = v_size(mat_v);
	if(size == 0 || size != v_size(y_v))
		QL_FAIL("invalid vector size");

	for(integer i=0;i<size;i++){
		QL_REQUIRE(!null(y_v[i]),"invalid yield");
	}

	vector(number) mat_days_v = mat_v - mat_v[0];
	interpolation ip =  ip_linear(mat_days_v,y_v);
		
	integer size_ip = v_size(mat_ip_v);
	if(size_ip == 0 || size_ip != v_size(y_ip_v))
		QL_FAIL("invalid vector size");
	
	vector(number) res[size_ip];

	for(integer i = 0; i < size_ip; i++ ) {
		date mat = mat_ip_v[i];
		if(null(mat))
			res[i] = null<number>;
		else {
			date min = mat_v[0];
			date max = mat_v[size-1];
			
			if(mat < min || mat > max)
				res[i] = null<number>;
			else				
				res[i] = y_ip_v[i] - ip.eval(mat - mat_v[0]);
		}
	}
	return res;
}

//-------------------------------------------------------
// gen_d_spread 
//-------------------------------------------------------
vector(number) gen_d_spread(vector(number) dur_v,
							vector(number) y_v,
							vector(number) dur_ip_v,
							vector(number) y_ip_v)	 option(category: "Financial")	
{
	
	string e;
	logical ok = CORE_INT.check_order(dur_v,e);
	QL_REQUIRE(ok,e);
	
	integer size = v_size(dur_v);
	if(size == 0 || size != v_size(y_v))
		QL_FAIL("invalid vector size");

	for(integer i=0;i<size;i++){
		QL_REQUIRE(!null(y_v[i]),"invalid yield");
	}

	interpolation ip =  ip_linear(dur_v,y_v);
		
	integer size_ip = v_size(dur_ip_v);
	if(size_ip == 0 || size_ip != v_size(y_ip_v))
		QL_FAIL("invalid vector size");
	
	vector(number) res[size_ip];

	for(integer i = 0; i < size_ip; i++ ) {
		number d = dur_ip_v[i];
		if(null(d))
			res[i] = null<number>;
		else {
			number min = dur_v[0];
			number max = dur_v[size-1];
			
			if(d < min || d > max)
				res[i] = null<number>;
			else				
				res[i] = y_ip_v[i] - ip.eval(d);
		}
	}
	return res;
}

//-------------------------------------------------------
// gen_i_spread 
//-------------------------------------------------------
vector(number) gen_bench_spread(vector(date) mat_v,
								vector(number) y_v,
								vector(date) mat_ip_v,
								vector(number) y_ip_v,
								out vector(integer) idx) option(category: "Financial")	
{
	string e;
	logical ok = CORE_INT.check_date_order(mat_v,e);
	QL_REQUIRE(ok,e);
	integer size = v_size(mat_v);
	if(size == 0 || size != v_size(y_v))
		QL_FAIL("invalid vector size");

	for(integer i=0;i<size;i++){
		QL_REQUIRE(!null(y_v[i]),"invalid yield");
	}

	integer size_ip = v_size(mat_ip_v);
	vector(number) res[size_ip];
	resize(idx,size_ip);
	
	for(integer i = 0; i < size_ip; i++ ) {
		
		date mat = mat_ip_v[i];			
		
		if(null(mat))
			res[i] = null<number>;
		else {
			
			date min = mat_v[0];
			date max = mat_v[size-1];
			
			if(mat < min){
				idx[i] = 0;
			}
			else if(mat > max){
				idx[i] = size-1;
			}
			else {
				
				vector(integer) mat_days_v = integer(abs(mat_v - mat));
				integer m 			= integer(v_min(mat_days_v));
				vector(integer) v 	= v_find_eq (mat_days_v, m);
				idx[i] 				= v[0];
			}
			res[i] = y_ip_v[i] - y_v[idx[i]] ;
		}
	}
	return res;
}


//-------------------------------------------------------
//class  yld_solve 
//-------------------------------------------------------	
class yld_solve 
{
	public:

	vector(number) 	cf;
	vector(number) 	per;
	number 			n;
	number  		dp;
	rate_type		rt;
	number 			f(number yld) ;		
};

number yld_solve.f(number yld)
{		
	number pv = 0;
	for(integer i = 0; i < n; i++ )
		pv += cf[i]*rate_to_disc(yld, rt, per[i]);///pow(1.0+yld, per[i]);
	return dp-pv;
}


//-------------------------------------------------------
// gen_ytm (yld based on cashflows)
//-------------------------------------------------------
number gen_ytm(	vector(number) 	periods,
				vector(number)	cash_flows,					
				number 			dirty_price,
				number 			nominal,
				number option(nullable)	yld_guess,
				rate_type		ytm_type) option(category: "Financial")

{
	//try {
		//err_msg = "";
		CORE_INT.check_cf_vect(periods,cash_flows);

		yld_solve y = new yld_solve;
			
		y.n 		= v_size(cash_flows);	
		y.cf 		= cash_flows/nominal * 100;
		y.per 		= clone_vector(periods);		
		y.dp 		= dirty_price/nominal*100;
		y.rt		= ytm_type;

		number x1 	= null(yld_guess)? 0: yld_guess-0.005;
		number x2 	= null(yld_guess)? 1: yld_guess+0.005;

		number f1 = y.f(x1);//neg
		QL_REQUIRE(!null(f1), "boundary calculation error (f1)");
		number i = 0;
		number max_i = 500;
		while (f1>= 0 && i < max_i){
			x1 -= 0.005;
			f1 = y.f(x1);
			QL_REQUIRE(!null(f1), "boundary calculation error (f1)");
			i++;
		}
		if(i >= max_i)
			QL_FAIL("max iterations reached");

		number f2 = y.f(x2);//neg
		QL_REQUIRE(!null(f2), "boundary calculation error (f2)");
		i = 0;
		while (f2<= 0 && i < max_i){
			x2 += 0.005;
			f2 = y.f(x2);
			QL_REQUIRE(!null(f2), "boundary calculation error (f2)");
			i++;
		}
		if(i >= max_i)
			QL_FAIL("max iterations reached");

		number tol 	= 0.0000001;
		
		zero_bracket(y, &yld_solve.f, x1, x2);
		return zero_brent(y, &yld_solve.f, x1, x2, tol);
	//}
	//catch {
		//err_msg = err.message();
		//return null<number>;
	//}
}
//-------------------------------------------------------
// gen_ytm (yld based on cashflows)
//-------------------------------------------------------
number gen_ytm(	date 			settle,
				vector(date) 	cash_flow_dates,
				vector(number)	cash_flows,					
				number 			dirty_price,
				number 			nominal,
				number option(nullable)	yld_guess,
				rate_type		ytm_type,
				day_count_method dc) option(category: "Financial")

{
	//try {
		//err_msg = "";
		CORE_INT.check_cf_vect(settle,cash_flow_dates,cash_flows);
		vector(number) cf_per = period(settle, cash_flow_dates, dc);
		return gen_ytm(cf_per,cash_flows,dirty_price,nominal,yld_guess,ytm_type);
		
	//}
	//catch {
		//err_msg = err.message();
		//return null<number>;
	//}
}
