/*	
	Option funcs bs 
	Developer: Algorithmica Research, Magnus Nyström			
*/
option(null: hard);
//-------------------------------------------------------
// black_scholes_prem
//-------------------------------------------------------
void black_scholes_prem(	CORE_INT_OPT.bs_pre_calc bs,
							opt_type 		type,
							number 			s, 
							number 			k, 
							out number 		fp, 
							out number 		sp)
{
	if(!bs.z_vol_){
		
		if(type == opt_type.OPT_CALL) 
			sp = s*bs.dfq_*bs.Nd1()-k*bs.df_*bs.Nd2() ;			
		else if(type == opt_type.OPT_PUT) 
			sp = -s*bs.dfq_*bs.Nd1_rev()+k*bs.df_*bs.Nd2_rev() ;			
		else 
			QL_FAIL("invalid option type");
		
		fp = sp/bs.df_ ;
	}
	else {		
		number f = s*bs.dfq_/bs.df_;
		
		if(type == opt_type.OPT_CALL) 
			fp = max(0,f-k) ;				
		else if(type == opt_type.OPT_PUT) 	
			fp = max(0,-f+k) ;			 
		else 
			QL_FAIL("invalid option type");

		sp = bs.df_*fp ;
	}	
}

void black_scholes_prem(	opt_type 		type,
							number 			s, 
							number 			k, 
							number 			v, 
							number 			cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 			tau,						
							number 			tau_v, 
							out number 		fp, 
							out number 		sp)
option (category: 'Financial/Options/Vanilla/Functions')
{
	if(CORE_INT.negative_check(tau_v, CORE_INT.DBL_EPSILON) || CORE_INT.negative_check(v, CORE_INT.DBL_EPSILON) ){//exactly 0 returns false
		fp = 0 ;
		sp = 0 ;
		return;
	}		

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);

	black_scholes_prem(	bs,type,s, k,fp, sp);
	
}
//-------------------------------------------------------
// black_scholes_prem
//-------------------------------------------------------
/*void black_scholes_prem(opt_type 		type,
						number 			s, 
						number 			k, 
						number 			v, 
						number 			cont_r,
						number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
						number 			tau,						
						number 			tau_v, 
						out number 		fp, 
						out number 		sp)
option (category: 'Financial/Options/Vanilla/Functions')
{
	if(null(cont_u_r)) cont_u_r = 0;
	number f = CORE_INT_OPT.bl_f_from_s(s, cont_r, tau,	 cont_u_r);
	black76_prem(type,f, k,v,cont_r,tau, tau_v, 0, fp, sp);
}*/

//-------------------------------------------------------
// black_scholes_delta, dV/dS
//-------------------------------------------------------
number black_scholes_delta_spot(opt_type 	type,
								number 		s, 
								number 		k, 
								number 		v, 
								number 		cont_r,
								number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
								number 		tau, 
								number 		tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{        
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		return 0;
	}

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);

	if(type == opt_type.OPT_CALL) 
		return bs.dfq_ * bs.Nd1();					
	else if(type == opt_type.OPT_PUT) 
		return -bs.dfq_ * bs.Nd1_rev();		
	else 
		QL_FAIL("invalid option type");

}

void black_scholes_delta_spot(	number 		s, 
								number 		k, 
								number 		v, 
								number 		cont_r,
								number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
								number 		tau, 
								number 		tau_v,
								out number delta_c,
								out number delta_p)
{        
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		delta_c = delta_p = 0;
		return ;
	}

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);

	delta_c = bs.dfq_ * bs.Nd1();					
	delta_p = -bs.dfq_ * bs.Nd1_rev();		
	return;
}
//-------------------------------------------------------
// black_scholes_delta_spot_c2p
//-------------------------------------------------------
number black_scholes_delta_spot_c2p(number 		delta_call, 									
									number  	cont_u_r,//eg. foreign curr rate, dividend-yld
									number 		tau)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number dfq 	= exp(-cont_u_r*tau);
	return delta_call - dfq;
}
//-------------------------------------------------------
// black_scholes_delta_spot_p2c
//-------------------------------------------------------
number black_scholes_delta_spot_p2c(number 		delta_put, 									
									number  	cont_u_r,//eg. foreign curr rate, dividend-yld
									number 		tau)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number dfq 	= exp(-cont_u_r*tau);
	return delta_put + dfq;
}
//-------------------------------------------------------
// black_scholes_delta, dV/dF
//-------------------------------------------------------
number black_scholes_delta_fut(opt_type 	type,
								number 		s, 
								number 		k, 
								number 		v, 
								number 		cont_r,
								number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
								number 		tau, 
								number 		tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{        
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		return 0;
	}

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);

	if(type == opt_type.OPT_CALL) 
		return bs.df_ * bs.Nd1();					
	else if(type == opt_type.OPT_PUT) 
		return -bs.df_ * bs.Nd1_rev();		
	else 
		QL_FAIL("invalid option type");
}
//-------------------------------------------------------
// black_scholes_delta_fut_c2p
//-------------------------------------------------------
number black_scholes_delta_fut_c2p(number 		delta_call, 									
									number  	cont_r,
									number 		tau)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df 	= exp(-cont_r*tau);
	return delta_call - df;
}
//-------------------------------------------------------
// black_scholes_delta_fut_p2c
//-------------------------------------------------------
number black_scholes_delta_fut_p2c(number 		delta_put, 									
									number  	cont_r,
									number 		tau)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df 	= exp(-cont_r*tau);
	return delta_put + df;
}
//-------------------------------------------------------
// black_scholes_delta_fwd
//-------------------------------------------------------
number black_scholes_delta_fwd(	opt_type 	type,
								number 		s, 
								number 		k, 
								number 		v, 
								number 		cont_r,
								number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
								number 		tau, 
								number 		tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{        
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		return 0;
	}

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
		
	if(type == opt_type.OPT_CALL) 
		return bs.Nd1();					
	else if(type == opt_type.OPT_PUT) 
		return -bs.Nd1_rev();		
	else 
		QL_FAIL("invalid option type");
}

void black_scholes_delta_fwd(	number 		s, 
								number 		k, 
								number 		v, 
								number 		cont_r,
								number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
								number 		tau, 
								number 		tau_v,
								out number 	delta_c,
								out number 	delta_p)
{        
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		delta_c = delta_p = 0;
		return ;
	}

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
		
	delta_c = bs.Nd1();					
	delta_p = -bs.Nd1_rev();		
}

//-------------------------------------------------------
// black_scholes_delta_fwd_c2p
//-------------------------------------------------------
number black_scholes_delta_fwd_c2p(number 		delta_call )
option (category: 'Financial/Options/Vanilla/Functions')
{
	return delta_call - 1.0;
}
//-------------------------------------------------------
// black_scholes_delta_fwd_p2c
//-------------------------------------------------------
number black_scholes_delta_fwd_p2c(number 		delta_put )
option (category: 'Financial/Options/Vanilla/Functions')
{
	return delta_put + 1.0;
}

//----------------------------------------------
// black_scholes_delta_spot_pa
// alt. delta_spot_prem_adj = delta_spot - spot_prem/s
//----------------------------------------------
number black_scholes_delta_spot_pa(	opt_type 	type,
									number 		s, 
									number 		k, 
									number 		v, 
									number 		cont_r,
									number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
									number 		tau, 
									number 		tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		return 0;
	}
	
	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	if(type == ..opt_type.OPT_CALL) 
		return bs.dfq_ * bs.Nd2()*k/bs.f_;					
	else if(type == ..opt_type.OPT_PUT) 
		return -bs.dfq_ * bs.Nd2_rev()*k/bs.f_;		
	else 
		QL_FAIL("invalid option type");
}
void black_scholes_delta_spot_pa(	number 		s, 
									number 		k, 
									number 		v, 
									number 		cont_r,
									number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
									number 		tau, 
									number 		tau_v,
									out number 	delta_c,
									out number 	delta_p)
{
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		delta_c = delta_p = 0;
		return ;
	}
	
	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	delta_c = bs.dfq_ * bs.Nd2()*k/bs.f_;					
	delta_p = -bs.dfq_ * bs.Nd2_rev()*k/bs.f_;		
}
//-------------------------------------------------------
// black_scholes_delta_spot_pa_c2p
//-------------------------------------------------------
number black_scholes_delta_spot_pa_c2p(number 	delta_call,
										number 	s, 
										number 	k, 
										number 	cont_r,
										number  cont_u_r,//eg. foreign curr rate, dividend-yld
										number 	tau)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df 	= exp(-cont_r*tau);
	number dfq 	= exp(-cont_u_r*tau);
	number f	= CORE_INT_OPT.bl_f_from_s(s, df,dfq);	
	return delta_call - dfq*k/f;
}
//-------------------------------------------------------
// black_scholes_delta_spot_pa_p2c
//-------------------------------------------------------
number black_scholes_delta_spot_pa_p2c(	number 	delta_put,
										number 	s, 
										number 	k, 
										number 	cont_r,
										number  cont_u_r,//eg. foreign curr rate, dividend-yld
										number 	tau)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df 	= exp(-cont_r*tau);
	number dfq 	= exp(-cont_u_r*tau);
	number f	= CORE_INT_OPT.bl_f_from_s(s, df,dfq);	
	return delta_put + dfq*k/f;
}
//----------------------------------------------
// black_scholes_delta_fwd_pa
//----------------------------------------------
number black_scholes_delta_fwd_pa(	opt_type 	type,
									number 		s, 
									number 		k, 
									number 		v, 
									number 		cont_r,
									number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
									number 		tau, 
									number 		tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		return 0;
	}
	
	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	if(type == ..opt_type.OPT_CALL) 
		return bs.Nd2()*k/bs.f_;					
	else if(type == ..opt_type.OPT_PUT) 
		return -bs.Nd2_rev()*k/bs.f_;		
	else 
		QL_FAIL("invalid option type");
}

void black_scholes_delta_fwd_pa(	number 		s, 
									number 		k, 
									number 		v, 
									number 		cont_r,
									number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
									number 		tau, 
									number 		tau_v,
									out number 	delta_c,
									out number 	delta_p)
{
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		delta_c = delta_p = 0;
		return ;
	}
	
	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	delta_c = bs.Nd2()*k/bs.f_;					
	delta_p = -bs.Nd2_rev()*k/bs.f_;		
}
//-------------------------------------------------------
// black_scholes_delta_fwd_pa_c2p
//-------------------------------------------------------
number black_scholes_delta_fwd_pa_c2p(number 	delta_call,
										number 	s, 
										number 	k, 
										number 	cont_r,
										number  cont_u_r,//eg. foreign curr rate, dividend-yld
										number 	tau)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df 	= exp(-cont_r*tau);
	number dfq 	= exp(-cont_u_r*tau);
	number f	= CORE_INT_OPT.bl_f_from_s(s, df,dfq);	
	return delta_call - k/f;
}
//-------------------------------------------------------
// black_scholes_delta_fwd_pa_p2c
//-------------------------------------------------------
number black_scholes_delta_fwd_pa_p2c(	number 	delta_put,
										number 	s, 
										number 	k, 
										number 	cont_r,
										number  cont_u_r,//eg. foreign curr rate, dividend-yld
										number 	tau)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df 	= exp(-cont_r*tau);
	number dfq 	= exp(-cont_u_r*tau);
	number f	= CORE_INT_OPT.bl_f_from_s(s, df,dfq);	
	return delta_put + k/f;
}
//-------------------------------------------------------
// black_scholes_gamma
//-------------------------------------------------------
number black_scholes_gamma(	CORE_INT_OPT.bs_pre_calc bs,
							number s)
{     
	return bs.dfq_* bs.pdf_nd1()/(s*bs.vv_);					 
}

number black_scholes_gamma(	number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	return black_scholes_gamma(bs,s);					 
}

//-------------------------------------------------------
// black_scholes_vega, dV/dVol
// Vega is typically expressed as the amount of money per underlying share that the option's value
// will gain or lose as volatility rises or falls by 1%. All options (both calls and puts) will
// gain value with rising volatility.
//-------------------------------------------------------
//-------------------------------------------------------
// black_scholes_vega
//-------------------------------------------------------
number black_scholes_vega(	CORE_INT_OPT.bs_pre_calc bs,
							number s,
							number tau_v)
{     	
	return s * bs.dfq_ * bs.pdf_nd1() * pow(tau_v, 0.5);					 
}

number black_scholes_vega(	number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
		
	return black_scholes_vega(bs,s,tau_v);					 
}

//-------------------------------------------------------
// black_scholes_vanna, dDelta/dVol == dVega/dS
// Vanna can be a useful sensitivity to monitor when maintaining a delta- or vega-hedged portfolio
// as vanna will help the trader to anticipate changes to the effectiveness of a delta-hedge as
// volatility changes or the effectiveness of a vega-hedge against change in the underlying spot price.
//-------------------------------------------------------
number black_scholes_vanna(	number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	number vega = black_scholes_vega(bs,s,tau_v);	

	return vega / s * (1.0 - bs.d1_/bs.vv_)	;				 
}
//-------------------------------------------------------
// black_scholes_volga (vomma, vega convexity), dVega/dVol
// With positive volga, a position will become long vega as implied volatility increases
// and short vega as it decreases, which can be scalped in a way analogous to long gamma.
//-------------------------------------------------------
number black_scholes_volga(	number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	number vega = black_scholes_vega(bs,s,tau_v);
		
	return vega * bs.d1_ * bs.d2_ / v;					 
}
//-------------------------------------------------------
// black_scholes_ultima, dvolga/dVol ( third-order derivative)
//-------------------------------------------------------
number black_scholes_ultima(number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	number vega = black_scholes_vega(bs,s,tau_v);
	number d1d2 = bs.d1_ * bs.d2_;

	return -vega/(v*v) * (d1d2 *(1.0-d1d2) + bs.d1_*bs.d1_ + bs.d2_*bs.d2_);					 
}


//-------------------------------------------------------
// black_scholes_dual_delta
//-------------------------------------------------------
number black_scholes_dual_delta(opt_type type,
								number 	s, 
								number 	k, 
								number 	v, 
								number 	cont_r,
								number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
								number 	tau, 
								number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);

	if(type == opt_type.OPT_CALL) 
		return -bs.df_ * bs.Nd2();					
	else if(type == opt_type.OPT_PUT) 
		return bs.df_ * bs.Nd2_rev();		
	else 
		QL_FAIL("invalid option type");				 
}
//-------------------------------------------------------
// black_scholes_dual_gamma
//-------------------------------------------------------
number black_scholes_dual_gamma(number 	s, 
								number 	k, 
								number 	v, 
								number 	cont_r,
								number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
								number 	tau, 
								number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	return bs.df_ * bs.pdf_nd2() / (k * bs.vv_);					 
}
//-------------------------------------------------------
// black_scholes_lambda (omega, elasticity), dV/dS*S/V
//  percentage change in option value per percentage change in the underlying
// price, a measure of leverage, sometimes called gearing.
//-------------------------------------------------------
number black_scholes_lambda(opt_type type,
							number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number 	fp,sp;
	number sd = black_scholes_delta_spot(type,s, k,	v,cont_r,cont_u_r,tau,tau_v);
	if(CORE_INT.is_equal_zero(sd, CORE_INT.DBL_EPSILON ))
		return 0;
	black_scholes_prem(	type,s, k, v, cont_r,cont_u_r,tau,tau_v,fp,sp);
	if(CORE_INT.is_equal_zero(sp, CORE_INT.DBL_EPSILON ))
		return null<number>;
	return sd * s / sp;
}

//-------------------------------------------------------
// black_scholes_delta_to_k  WORK-IN-PROGRESS
//-------------------------------------------------------
number black_scholes_delta_spot_to_strike(	number delta_spot,
											opt_type type,
											number 	s, 								
											number 	v, 
											number 	cont_r,
											number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
											number 	tau, 
											number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	if(null(cont_u_r)) cont_u_r = 0;
	number vv 		= v * pow(tau_v, 0.5) ;	
	number ph 		= type == opt_type.OPT_CALL ? 1 :-1;
	number t2		= v * v * 0.5 * tau_v;
	
	number d1		= ph * exp(cont_u_r*tau_v) * delta_spot;
	number cfd		= normal_inv(d1);
	number e1		= exp(-ph *cfd * vv + t2);
	number f		= CORE_INT_OPT.bl_f_from_s(	 s, cont_r,	 tau, cont_u_r);
	return f * e1;	
}
//-------------------------------------------------------
// black_scholes_delta_to_k  WORK-IN-PROGRESS
//-------------------------------------------------------
number black_scholes_delta_fwd_to_strike(number delta_fwd,
										opt_type type,
										number 	f, 								
										number 	v, 									
										number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number vv 		= v * pow(tau_v, 0.5) ;	
	number ph 		= type == opt_type.OPT_CALL ? 1 :-1;
	number t2		= v * v * 0.5 * tau_v;
	
	number d1		= ph * delta_fwd;
	number cfd		= normal_inv(d1);
	number e1		= exp(-ph *cfd * vv + t2);
	return f * e1;	
}

//-------------------------------------------------------
// black_scholes_gamma_fwd
//-------------------------------------------------------
number black_scholes_gamma_fwd(	number 	s, 
								number 	k, 
								number 	v, 
								number 	cont_r,
								number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
								number 	tau, 
								number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	if(null(cont_u_r)) cont_u_r = 0;

	number f = CORE_INT_OPT.bl_f_from_s(s, cont_r, tau,cont_u_r);

	return /*bs.df_ **/ bs.pdf_nd1()/(f*v*pow(tau_v, 0.5));				 
}
//-------------------------------------------------------
// black_scholes_speed, dGamma/dS (3rd order greek)
// Speed can be important to monitor when delta-hedging or gamma-hedging a portfolio.
//-------------------------------------------------------
number black_scholes_speed(	number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	number gamma = black_scholes_gamma(bs, s);

	return -gamma/s*(bs.d1_/bs.vv_+1.0)	;				 
}
//-------------------------------------------------------
// black_scholes_zomma dGamma/dVol == dVanna/dS(3rd order greek)
// Zomma can be a useful sensitivity to monitor when maintaining a gamma-hedged portfolio
// as zomma will help the trader to anticipate changes to the effectiveness of the hedge as volatility changes.
//-------------------------------------------------------
number black_scholes_zomma(	number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	number gamma = black_scholes_gamma(bs, s);

	return gamma*((bs.d1_*bs.d2_-1.0)/v);					 
}
//-------------------------------------------------------
// black_scholes_rho, dV/dr
// Rho is typically expressed as the amount of money, per share of the underlying, that the value
// of the option will gain or lose as the risk free interest rate rises or falls by 1.0% per annum (100 basis points).
//-------------------------------------------------------
number black_scholes_rho(	opt_type type,
							number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;
	
	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);

	if(type == opt_type.OPT_CALL) 
		return k * bs.df_ * tau * bs.Nd2();					
	else if(type == opt_type.OPT_PUT) 
		return -k * bs.df_ * tau * bs.Nd2_rev();		
	else 
		QL_FAIL("invalid option type");
}
//-------------------------------------------------------
// black_scholes_psi, dV/dq
//-------------------------------------------------------
number black_scholes_psi(	opt_type type,
							number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;
	
	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);

	if(type == opt_type.OPT_CALL) 
		return -bs.dfq_ * s * tau * bs.Nd1();					
	else if(type == opt_type.OPT_PUT) 
		return bs.dfq_ * s * tau * bs.Nd1_rev();		
	else 
		QL_FAIL("invalid option type");
}
//-------------------------------------------------------
// black_scholes_theta, -dV/dTau
// The mathematical result of the formula for theta is expressed in value per year.
// By convention, it is usual to divide the result by the number of days in a year,
// to arrive at the amount an option's price will drop, in relation to the underlying stock's price.
// Theta is almost always negative for long calls and puts, and positive for short (or written)
// calls and puts. An exception is a deep in-the-money European put.
//-------------------------------------------------------
number black_scholes_theta(	opt_type type,
							number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;
	
	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	if(null(cont_u_r)) cont_u_r = 0;

	number term1 = -bs.dfq_ * s * bs.pdf_nd1() * v / (2.0*pow(tau_v,0.5));
	
	if(type == opt_type.OPT_CALL){
		number term2 = cont_r * k * bs.df_ * bs.Nd2();
		number term3 = cont_u_r * s * bs.dfq_ * bs.Nd1(); 
		return term1 - term2 + term3;
	}
	else if(type == opt_type.OPT_PUT) {
		number term2 = cont_r * k * bs.df_ * bs.Nd2_rev();
		number term3 = cont_u_r * s * bs.dfq_ * bs.Nd1_rev(); 
		return term1 + term2 - term3;
	}
	else 
		QL_FAIL("invalid option type");
}
//-------------------------------------------------------
// black_scholes_charm (delta decay), -dDelta/dTau
// Charm can be an important Greek to measure/monitor when delta-hedging
// a position over a weekend.
// The mathematical result of the formula for charm (see below) is expressed in delta/year.
// It is often useful to divide this by the number of days per year to arrive at the
// delta decay per day. This use is fairly accurate when the number of days remaining until
// option expiration is large. When an option nears expiration, charm itself may change quickly,
// rendering full day estimates of delta decay inaccurate.
//-------------------------------------------------------
number black_scholes_charm(	opt_type type,
							number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
    if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;
	
	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	if(null(cont_u_r)) cont_u_r = 0;

	number term2 = bs.dfq_ * bs.pdf_nd1() * (2.0*(cont_r-cont_u_r)*tau-bs.d2_*bs.vv_)/(2.0*tau_v*bs.vv_);
	
	if(type == opt_type.OPT_CALL) 
		return -cont_u_r*bs.dfq_ * bs.Nd1()-term2;					
	else if(type == opt_type.OPT_PUT) 
		return cont_u_r*bs.dfq_ * bs.Nd1_rev()-term2;	
	else 
		QL_FAIL("invalid option type");
}
//-------------------------------------------------------
// black_scholes_veta, dVega/dTau
// It is common practice to divide the mathematical result of veta by 100 times the number of
// days per year to reduce the value to the percentage change in vega per one day.
//-------------------------------------------------------
number black_scholes_veta(	number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
    if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;
	
	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);

	if(null(cont_u_r)) cont_u_r = 0;
	
	number term1 = -s*bs.dfq_*bs.pdf_nd1()*pow(tau_v, 0.5) ;
	number term2 = cont_u_r + ((cont_r-cont_u_r)*bs.d1_)/bs.vv_ - (1.0+bs.d1_*bs.d2_)/(2.0*tau);
	return term1*term2;					

}
//-------------------------------------------------------
// black_scholes_color (gamma decay) (third-order derivative), dGamma/dTau
// Color can be an important sensitivity to monitor when maintaining a gamma-hedged portfolio as it
// can help the trader to anticipate the effectiveness of the hedge as time passes.

//The mathematical result of the formula for color (see below) is expressed in gamma/year. It is
// often useful to divide this by the number of days per year to arrive at the change in gamma
// per day. This use is fairly accurate when the number of days remaining until option
// expiration is large. When an option nears expiration, color itself may change quickly,
// rendering full day estimates of gamma change inaccurate.
//-------------------------------------------------------
number black_scholes_color(	number 	s, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
							number 	tau, 
							number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;
	
	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	if(null(cont_u_r)) cont_u_r = 0;

	number term1 = -bs.dfq_*bs.pdf_nd1()/(2.0*s*tau_v*bs.vv_);
	number term2 = 2.0*cont_u_r*tau + 1.0 + (2.0*(cont_r-cont_u_r)*tau-bs.d2_*bs.vv_)/bs.vv_*bs.d1_;
	return term1*term2;					

}
//-------------------------------------------------------
// black_scholes_imp_vol
//-------------------------------------------------------
number black_scholes_imp_vol(	opt_type type,
								number 	s, 
								number 	k, 								
								number 	cont_r,
								number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
								number 	tau, 
								number 	tau_v,
								number 	prem)
option (category: 'Financial/Options/Vanilla/Functions')
{
	if(tau_v <= CORE_INT.DBL_EPSILON )
		return 0;
		
	
	number f 	= CORE_INT_OPT.bl_f_from_s(s, cont_r, tau,	 cont_u_r);
	number df 	= exp(-cont_r*tau);
		
	if (!CORE_INT_OPT.bl_check_bounds(type,f, k, prem/df))
		return -1;
		
	number sv = CORE_INT_OPT.bl_initial_sigma(type,f,k,tau_v,df,prem) ;
	
	if(sv <= CORE_INT.DBL_EPSILON)
		sv = 0.01 ;

	number  fp, sp, value, prime ;
	number sol_break 	= 0.000001 ;
	number nosol_break 	= 0.001 ; 
	number max_i 		= 50 ;

	do {
		CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k,sv,cont_r,cont_u_r,tau,tau_v);
		
		black_scholes_prem(bs,type,s,k,fp,sp);
	
		value = sp - prem;

		if(abs(value) <= sol_break)
			return sv ;

		prime = black_scholes_vega(bs,s,tau_v) ;
							
		if(CORE_INT.is_equal_zero(prime, CORE_INT.DBL_EPSILON ))
			return -1 ;
		
		sv -= value/prime ;
	
	} while(abs(value) > sol_break && sv > nosol_break && max_i-- > 0) ;
	
	if(abs(value) <= sol_break)
		return sv ;
	else
		return -1 ; 
}


//-------------------------------------------------------
//class  k_solve 
//-------------------------------------------------------	
class k_solve 
{
public:
	number d_;
	logical fwd_;
	opt_type cp_;
	number s_;
	number v_;
	number cont_r_;
	number cont_u_r_;
	number tau_;						
	number tau_v_;
	number f_;
	number 	f(number) ;
	
};

number k_solve.f(number k)
{
	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(	s_, k, v_,  cont_r_, cont_u_r_, tau_, tau_v_);
	number dfq = fwd_ ? 1 : bs.dfq_;
	if(cp_ == ..opt_type.OPT_CALL) 
		return d_ - (dfq * bs.Nd2()*k/f_);					
	else 
		return d_ - (-dfq * bs.Nd2_rev()*k/f_);
}
//-------------------------------------------------------
// delta_to_k_solver
// the premium adjusted call delta is
// not monotonic in strike, such that two solutions
// might occur. The one right to the max of the delta is
// considered to be the correct strike.
//-------------------------------------------------------
number delta_to_k_solver(	number delta,
							logical fwd,
							..opt_type cp,
							number s,
							number f,
							number v,
							number cont_r ,
							number cont_u_r,
							number tau,						
							number tau_v)
{								
	k_solve k = new k_solve;

	k.d_ 		= delta;
	k.fwd_ 		= fwd;
	k.cp_		= cp;
	k.s_ 		= s;
	k.v_		= v;
	k.cont_r_ 	= cont_r;
	k.cont_u_r_ = cont_u_r;
	k.tau_		= tau;						
	k.tau_v_	= tau_v;
	k.f_ 		= f;

	/* 	the premium-adjusted delta for a strike K is always smaller than 
		the non-adjusted delta corresponding to the same strike.
	*/
	number x2;
	if (!fwd) 
		x2 = black_scholes_delta_spot_to_strike(delta, k.cp_, k.s_, k.v_, k.cont_r_, 
											k.cont_u_r_, k.tau_, k.tau_v_);
	else 
		x2 = black_scholes_delta_fwd_to_strike(delta, k.cp_, k.f_, k.v_, k.tau_v_);
				 
	
	number tol 	= 0.000001;
	
	if(k.cp_ == ..opt_type.OPT_PUT) {
		number x1 = 0;
		zero_bracket(k, &k_solve.f, x1, x2);
		return zero_brent(k, &k_solve.f, x1, x2, tol);
	}

	number x1 = CORE_INT_OPT.k_max_solve(k.s_,k.v_,k.tau_,k.tau_v_,k.cont_r_,k.cont_u_r_,					
							0,x2,false);

	//zero_bracket(k, &k_solve.f, x1, x2);
	return zero_brent(k, &k_solve.f, x1, x2, tol);		
}

//-------------------------------------------------------
// black_scholes_delta_fwd_pa_to_strike
//-------------------------------------------------------
number black_scholes_delta_fwd_pa_to_strike(number delta,							
											..opt_type cp,
											number s,
											number f,
											number v,
											number cont_r ,
											number cont_u_r,
											number tau,						
											number tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	return delta_to_k_solver(delta,true, cp, s, f, v, cont_r , cont_u_r, tau, tau_v);
}
//-------------------------------------------------------
// black_scholes_delta_spot_pa_to_strike
//-------------------------------------------------------
number black_scholes_delta_spot_pa_to_strike(number delta,							
											..opt_type cp,
											number s,
											number f,
											number v,
											number cont_r ,
											number cont_u_r,
											number tau,						
											number tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	return delta_to_k_solver(delta,false, cp, s, f, v, cont_r , cont_u_r, tau, tau_v);
}
//-------------------------------------------------------
// enum delta_type
//-------------------------------------------------------
class enum delta_type option (category: 'Financial/Options/Vanilla')
{
	UNADJ_SPOT 			option(str:"unadj spot"),
	UNADJ_FWD 			option(str:"unadj fwd"),
	PREMADJ_SPOT 		option(str:"prem-adj spot"),
	PREMADJ_FWD 		option(str:"prem-adj fwd")		
};
//-------------------------------------------------------
// enum atm_type
//-------------------------------------------------------
class enum atm_type option (category: 'Financial/Options/Vanilla')
{
	SPOT 						option(str:"spot"),
	FWD 						option(str:"fwd"),
	NEUTRAL 					option(str:"neutral"),
	MAX_GAMMA 					option(str:"max gamma"),
	MAX_VEGA 					option(str:"max vega"),
	DELTA_NEUTRAL_UNADJ_SPOT 	option(str:"delta neutral spot"),//ex. --> zero-delta straddle strike placement when prem is paid in domestic curr
	DELTA_NEUTRAL_UNADJ_FWD 	option(str:"delta neutral fwd"),
	DELTA_NEUTRAL_PREMADJ_SPOT 	option(str:"delta neutral prem-adj spot"),//ex. --> zero-delta straddle strike placement when prem is paid in foreign curr
	DELTA_NEUTRAL_PREMADJ_FWD 	option(str:"delta neutral prem-adj fwd")
};

//----------------------------------------------
//black_scholes_delta_to_k_cp
//----------------------------------------------
void black_scholes_delta_to_k_cp(	delta_type 		delta_t,
									number 			delta,
									number 			s,
									number 			f,
									number 			vol_c,
									number 			vol_p,
									number 			cont_r,
									number 			cont_u_r,
									number 			tau,					  
									out number 		k_dd_c,
									out number 		k_dd_p)
{
	switch(delta_t){
	case delta_type.UNADJ_SPOT:
		QL_REQUIRE(abs(delta) <= exp(-cont_u_r*tau), "Spot delta out of range.");
		k_dd_c 	= black_scholes_delta_spot_to_strike( delta, opt_type.OPT_CALL, s, vol_c, cont_r, cont_u_r, tau,  tau);			
		k_dd_p 	= black_scholes_delta_spot_to_strike(-delta, opt_type.OPT_PUT,  s, vol_p, cont_r, cont_u_r, tau,  tau);			
		break;
	case delta_type.UNADJ_FWD:
		QL_REQUIRE(abs(delta) <= 1.0, "Forward delta out of range.");
		k_dd_c 	= black_scholes_delta_fwd_to_strike( delta, opt_type.OPT_CALL, f, vol_c, tau);			
		k_dd_p 	= black_scholes_delta_fwd_to_strike(-delta, opt_type.OPT_PUT,  f, vol_p, tau);			
		break;		
	case delta_type.PREMADJ_SPOT:
		k_dd_c 	= black_scholes_delta_spot_pa_to_strike( delta, opt_type.OPT_CALL, s, f,vol_c, cont_r, cont_u_r, tau,  tau);			
		k_dd_p 	= black_scholes_delta_spot_pa_to_strike(-delta, opt_type.OPT_PUT,  s, f,vol_p, cont_r, cont_u_r, tau,  tau);			
		break;
	case delta_type.PREMADJ_FWD:	
		k_dd_c 	= black_scholes_delta_fwd_pa_to_strike( delta, opt_type.OPT_CALL, s, f,vol_c, cont_r, cont_u_r, tau,  tau);			
		k_dd_p 	= black_scholes_delta_fwd_pa_to_strike(-delta, opt_type.OPT_PUT,  s, f,vol_p, cont_r, cont_u_r, tau,  tau);			
		break;
	default:	
		QL_FAIL("invalid delta type (not applicable in this function)");
	}
}

//----------------------------------------------
// black_scholes_delta
//----------------------------------------------
void black_scholes_delta( 	delta_type 		delta_t,
							 number 		s ,
							 number			k,
							 number 		v,
							 number 		tau,							 							 
							 number 		cont_r,
							 number 		cont_u_r,
							 out number 	delta_call,							 
							 out number 	delta_put)
option (category: 'Financial/Options/Vanilla/Functions')
{
	switch(delta_t){
	case delta_type.UNADJ_SPOT:			
		black_scholes_delta_spot(s, k,v, cont_r, cont_u_r, tau, tau,delta_call,delta_put);	
		break;
	case delta_type.UNADJ_FWD:			
		black_scholes_delta_fwd(s, k,v, cont_r, cont_u_r, tau, tau,delta_call,delta_put);
		break;
	case delta_type.PREMADJ_SPOT:
		black_scholes_delta_spot_pa(s, k,v, cont_r, cont_u_r, tau, tau,delta_call,delta_put);
		break;
	case delta_type.PREMADJ_FWD:	
		black_scholes_delta_fwd_pa(s, k,v, cont_r, cont_u_r, tau, tau,delta_call,delta_put);
		break;
	default:	
		QL_FAIL("invalid delta type (not applicable in this function)");
	}
}

//----------------------------------------------
// black_scholes_delta_c2p
//----------------------------------------------
number black_scholes_delta_c2p(delta_type 	delta_t,
								number 		delta_call,
								number 		s ,
								number		k,
								number 		v,
								number 		tau,							 							 
								number 		cont_r,
								number 		cont_u_r)
option (category: 'Financial/Options/Vanilla/Functions')
{
	switch(delta_t){
	case delta_type.UNADJ_SPOT:			
		return black_scholes_delta_spot_c2p(delta_call, cont_u_r,tau);	
		break;
	case delta_type.UNADJ_FWD:			
		return black_scholes_delta_fwd_c2p(delta_call);
		break;
	case delta_type.PREMADJ_SPOT:
		return black_scholes_delta_spot_pa_c2p(delta_call,s, k,cont_r, cont_u_r, tau);
		break;
	case delta_type.PREMADJ_FWD:	
		return black_scholes_delta_fwd_pa_c2p(delta_call,s, k,cont_r, cont_u_r, tau);
		break;
	default:	
		QL_FAIL("invalid delta type (not applicable in this function)");
	}
}

//----------------------------------------------
// black_scholes_delta_p2c
//----------------------------------------------
number black_scholes_delta_p2c(delta_type 	delta_t,
								number 		delta_put,
								number 		s ,
								number		k,
								number 		v,
								number 		tau,							 							 
								number 		cont_r,
								number 		cont_u_r)
option (category: 'Financial/Options/Vanilla/Functions')
{
	switch(delta_t){
	case delta_type.UNADJ_SPOT:			
		return black_scholes_delta_spot_p2c(delta_put, cont_u_r,tau);	
		break;
	case delta_type.UNADJ_FWD:			
		return black_scholes_delta_fwd_p2c(delta_put);
		break;
	case delta_type.PREMADJ_SPOT:
		return black_scholes_delta_spot_pa_p2c(delta_put,s, k,cont_r, cont_u_r, tau);
		break;
	case delta_type.PREMADJ_FWD:	
		return black_scholes_delta_fwd_pa_p2c(delta_put,s, k,cont_r, cont_u_r, tau);
		break;
	default:	
		QL_FAIL("invalid delta type (not applicable in this function)");
	}
}

//----------------------------------------------
// black_scholes_atm_strike
//----------------------------------------------
number black_scholes_atm_strike( atm_type 	atm_t,
								 number 	vol_atm ,
								 number 	s ,
								 number 	tau,							 							 
								 number 	df ,
								 number 	dfq)
option (category: 'Financial/Options/Vanilla/Functions')
{
	switch(atm_t){
	case atm_type.SPOT:
		return s;
		break;
	case atm_type.FWD:
	case atm_type.NEUTRAL://equiv to atm_fwd(); k such that call value = put value
		return s * dfq / df;
		break;
	case atm_type.DELTA_NEUTRAL_UNADJ_SPOT:
	case atm_type.DELTA_NEUTRAL_UNADJ_FWD:
	case atm_type.MAX_GAMMA:
	case atm_type.MAX_VEGA:
		return s * dfq / df * exp(0.5*vol_atm *vol_atm * tau) ;
		break;
	case atm_type.DELTA_NEUTRAL_PREMADJ_SPOT:
	case atm_type.DELTA_NEUTRAL_PREMADJ_FWD:
		return s * dfq / df * exp(-0.5* vol_atm *vol_atm * tau) ;
	default:	
		QL_FAIL("atm type not supported");
	}		
}
