/*	
	Option model classes
	Developer: Algorithmica Research, Magnus Nyström

	option_model [abstract-root]
		vanilla_option_model [abstract]
			b76_option_model
			bs_option_model
			bnorm_option_model
			
	option_model_args [abstract-root]
		vanilla_option_model_args [abstract]
			b76_model_args
			bs_model_args
			bnorm_model_args

	Note: 	the args-classes should ideally be a nested class within option_model hierarchy but this
			is currently not possible in qlang
*/
option(null: hard);
class enum option_model_e  option (category: 'Financial/Options') { B76, BS, BNORM, CRR, BAW};
class enum option_qs_e  option (category: 'Financial/Options') { VOL, PREM_S, PREM_F};
//**********************************************
//**********************************************
// class  option_model_args  <abstract>
// class-hierarchy parallell to class-hierarchy for option_model  
// (would have preferred to have this class inside the option_model class (nested)
// but currently not allowed)
//**********************************************
//**********************************************
class option_model_args 
{
public:
	virtual option_model_e get_model() = 0;
};

//**********************************************
//**********************************************
// class vanilla_option_model_args  <abstract>
// parent class to all vanilla models (B76, BS, CRR,...)
//**********************************************
//**********************************************
class vanilla_option_model_args : public option_model_args
{
public:
	vanilla_option_model_args();
	void set_args(exercise_type,opt_type,number option(nullable) ,number option(nullable),
				  number,number option(nullable),option_qs_e,number,number option(nullable),number, number,
				  vector(number) option(nullable) pay = null<vector(number)>,
				  vector(number) option(nullable) ex = null<vector(number)>,
				  vector(number) option(nullable) div = null<vector(number)>)  ;
	//void set_args_f(number);
//protected:
	exercise_type 	et_;
	opt_type 		cp_;
	number			f_;
	number			s_;
	number			k_;
	number 			v_p_;
	option_qs_e		qs_;
	number 			cont_r_;
	number 			tau_;
	number			cont_u_r_;
	number 			tau_v_;
	vector(number) pay_;
	vector(number) ex_;
	vector(number) div_;
};


//----------------------------------------------
// vanilla_option_model_args constructor
//----------------------------------------------
vanilla_option_model_args.vanilla_option_model_args(){}
//----------------------------------------------
// vanilla_option_model_args.set_args
//----------------------------------------------
void vanilla_option_model_args.set_args(exercise_type 	et,
										opt_type 		cp,
										number option(nullable) f,//either f or s must be input depending on the model
										number option(nullable) s,
										number 			k,
										number option(nullable) v_p,
										option_qs_e		qs,
										number 			cont_r,
										number option(nullable) cont_u_r,
										number			tau,
										number 			tau_v,
										vector(number) option(nullable) pay,
										vector(number) option(nullable) ex,
										vector(number) option(nullable) div)
{
	et_ 	= et;
	cp_ 	= cp;
	f_ 		= f;
	s_ 		= s;
	k_ 		= k;
	v_p_ 	= v_p;
	qs_		= qs;
	cont_r_ = cont_r;
	cont_u_r_ = null(cont_u_r) ? 0 : cont_u_r;
	tau_	= tau;
	tau_v_ 	= tau_v;
	pay_	= null(pay) ? null<vector(number)> : clone_vector(pay);
	ex_		= null(ex) ? null<vector(number)> : clone_vector(ex);
	div_	= null(div) ? null<vector(number)> : clone_vector(div);
	CORE_INT_OPT.filter_div(tau_v,pay_,  ex_, div_);
	//df_		= exp(-cont_r*tau);
	//u_df_	= null(cont_u_r) ? null<number>: exp(-cont_u_r*tau);
}
/*void vanilla_option_model_args.set_args_f(number f)
{
	f_ = f;
}*/

//**********************************************
//**********************************************
// class b76_model_args
//**********************************************
//**********************************************
class b76_model_args : public vanilla_option_model_args
//option (category: 'Financial/Options/Vanilla')
{
public:
	
	b76_model_args();	
	override option_model_e get_model() ;
protected:
};
//----------------------------------------------
// b76_model_args constructor
//----------------------------------------------
b76_model_args.b76_model_args()
{}
//----------------------------------------------
// b76_model_args.get_model
//----------------------------------------------
option_model_e b76_model_args.get_model()
{
	return option_model_e.B76;
}

//**********************************************
//**********************************************
// class bs_model_args
//**********************************************
//**********************************************
class bs_model_args : public vanilla_option_model_args
//option (category: 'Financial/Options/Vanilla')
{
public:
	
	bs_model_args();	
	override option_model_e get_model() ;
protected:
};
//----------------------------------------------
// bs_model_args constructor
//----------------------------------------------
bs_model_args.bs_model_args()
{}
//----------------------------------------------
// bs_model_args.get_model
//----------------------------------------------
option_model_e bs_model_args.get_model()
{
	return option_model_e.B76;
}

//**********************************************
//**********************************************
// class bnorm_model_args
//**********************************************
//**********************************************
class bnorm_model_args : public vanilla_option_model_args
//option (category: 'Financial/Options/Vanilla')
{
public:
	
	bnorm_model_args();	
	override option_model_e get_model() ;
protected:
};
//----------------------------------------------
// bnorm_model_args constructor
//----------------------------------------------
bnorm_model_args.bnorm_model_args()
{}
//----------------------------------------------
// bnorm_model_args.get_model
//----------------------------------------------
option_model_e bnorm_model_args.get_model()
{
	return option_model_e.BNORM;
}


//**********************************************
//**********************************************
// class crr_model_args
//**********************************************
//**********************************************
class crr_model_args : public vanilla_option_model_args
//option (category: 'Financial/Options/Vanilla')
{
public:
	
	crr_model_args();	
	override option_model_e get_model() ;
protected:
};
//----------------------------------------------
// crr_model_args constructor
//----------------------------------------------
crr_model_args.crr_model_args()
{}
//----------------------------------------------
// crr_model_args.get_model
//----------------------------------------------
option_model_e crr_model_args.get_model()
{
	return option_model_e.CRR;
}

//**********************************************
//**********************************************
// class baw_model_args
//**********************************************
//**********************************************
class baw_model_args : public vanilla_option_model_args
//option (category: 'Financial/Options/Vanilla')
{
public:
	
	baw_model_args();	
	override option_model_e get_model() ;
protected:
};
//----------------------------------------------
// baw_model_args constructor
//----------------------------------------------
baw_model_args.baw_model_args()
{}
//----------------------------------------------
// baw_model_args.get_model
//----------------------------------------------
option_model_e baw_model_args.get_model()
{
	return option_model_e.BAW;
}

//**********************************************
//**********************************************
// class option_model  <abstract>
//**********************************************
//**********************************************
class option_model
{
public:
	//virtual option_model clone() = 0;
	virtual option_model_args get_args() 						= 0;

	virtual void add_u_sprd(number d) 							= 0;

	virtual number vol() 										= 0;
	virtual void prem(out number , out number ) 				= 0;
	virtual void prem_u(number d, out number , out number ) 	= 0;
	virtual void prem_u_spot(number d, out number , out number) = 0;
	virtual void prem_u_fwd(number d, out number , out number ) = 0;
	virtual void prem_v(number d, out number , out number ) 	= 0;
	virtual void prem_r(number d, out number , out number ) 	= 0;
	virtual void prem_u_r(number d, out number , out number) 	= 0;
	virtual void prem_tau(number d, out number , out number ) 	= 0;
	virtual number imp_vol(number) 								= 0;
	
	virtual option_model_e get_model() 							= 0;
	virtual logical check_exercise(exercise_type) 				= 0;
	virtual logical check_asset(integer) 						= 0;
	virtual number delta_spot(number  ,number , logical) ;
	virtual number delta_fut(number  ,number , logical) ;
	virtual number delta_fwd(number  ,number , logical) ;
	virtual number vega(number,number , logical ) ;
	virtual number vanna(number,number , logical ) ;
	virtual number volga(number,number , logical ) ;
	virtual number gamma_spot(number,number , logical ) ;
	virtual number gamma_fwd(number,number , logical ) ;
	virtual number theta(number,number, out number  ) ;
	virtual number rho(number,number , logical  ) ;
	virtual number psi(number,number , logical  ) ;//derivative of the option value with respect to the foreign interest rate (or dividend yield)
	
protected:
	//virtual number vega(number ,number, number, logical ) ;
	
	virtual number delta_anal_spot(number) 					= 0;
	virtual number delta_anal_fut(number) 					= 0;
	virtual number delta_anal_fwd(number) 					= 0;
	virtual number gamma_anal_spot(number ) 				= 0;
	virtual number gamma_anal_fwd(number ) 					= 0;
	virtual number vega_anal(number  ) 						= 0;
	
};


//----------------------------------------------
// option_model.delta_spot  (override if necessary to use cont_u_r argument in prem_u_spot(...)  eg. for B76)
//----------------------------------------------
number option_model.delta_spot(number mult,
							   number d,
							   logical centered)
{
	if(CORE_INT.is_equal_zero(d,CORE_INT.DBL_EPSILON)) {
		return delta_anal_spot(mult);
	}
	   
	number t, spd_u;
	prem_u_spot(d, t, spd_u);
	
	number sd;	
	if(centered) {
		number spd_d;
		prem_u_spot(-d, t, spd_d);
		sd = CORE_INT.num1der(spd_u,spd_d,d);
	}
	else {
		number sp;
		prem(t, sp);
		sd = CORE_INT.num1der_fd(sp,spd_u,d);
						
	}
	return sd*mult;
		
}
//----------------------------------------------
// option_model.delta_fut
//----------------------------------------------
number option_model.delta_fut(	number mult,
								number d,								
								logical centered)
{
	if(CORE_INT.is_equal_zero(d,CORE_INT.DBL_EPSILON)) {		
		return delta_anal_fut(mult);
	}
	
	number t, spd_u;
	prem_u_fwd(d, t, spd_u);
	
	number sd;	
	if(centered) {
		number spd_d;
		prem_u_fwd(-d, t, spd_d);
		sd = CORE_INT.num1der(spd_u,spd_d,d);
	}
	else {
		number sp;
		prem(t, sp);
		sd = CORE_INT.num1der_fd(sp,spd_u,d);
	}
	return sd*mult;
}
//----------------------------------------------
// option_model.delta_fwd
//----------------------------------------------
number option_model.delta_fwd(	number mult,
								number d,								
								logical centered)
{
	if(CORE_INT.is_equal_zero(d,CORE_INT.DBL_EPSILON)) {
		return delta_anal_fwd(mult);
	}
	number t, fpd_u;
	prem_u_fwd(d, fpd_u, t);
	
	number fd;	
	if(centered) {
		number fpd_d;
		prem_u_fwd(-d, fpd_d, t);
		fd = CORE_INT.num1der(fpd_u,fpd_d,d);
	}
	else {
		number fp;
		prem(fp, t);
		fd = CORE_INT.num1der_fd(fp,fpd_u,d);
	}
	return fd*mult;

}
//----------------------------------------------
// option_model.vega
//----------------------------------------------
number option_model.vega(number mult,
						 number d,						 
						 logical centered)
{
	if(CORE_INT.is_equal_zero(d,CORE_INT.DBL_EPSILON)) {		
		return vega_anal(mult);
	}
	
	number t, spd_u;
	prem_v(d, t, spd_u);
	
	number sd;	
	if(centered) {
		number spd_d;
		prem_v(-d, t, spd_d);
		sd = CORE_INT.num1der(spd_u,spd_d,d);
	}
	else {
		number sp;
		prem(t, sp);
		sd = CORE_INT.num1der_fd(sp,spd_u,d);
	}
	return sd * mult;
}
//----------------------------------------------
// option_model.vanna
//----------------------------------------------
number option_model.vanna(number mult,
						 number d,						 
						 logical not_used)
{
	return null<number>;

	if(CORE_INT.is_equal_zero(d,CORE_INT.DBL_EPSILON)) {		
		d = CORE_INT.EPSILON;
	}
	
	number spd_u = this.vega(1, d,false);	
	number spd_d = this.vega(1, -d,false);		
	number sd = CORE_INT.num1der(spd_u,spd_d,d);
	
	return sd * mult;
}
//----------------------------------------------
// option_model.volga
//----------------------------------------------
number option_model.volga(number mult,
						 number d,						 
						 logical not_used)
{
	if(CORE_INT.is_equal_zero(d,CORE_INT.DBL_EPSILON)) {		
		d = CORE_INT.EPSILON;
	}

	number sp = this.vega(1, d, true);

	add_u_sprd(d);
	number spd_u = this.vega(1, d, true);	
	add_u_sprd(-d);
	
	number sd = CORE_INT.num1der_fd(sp,spd_u,d);
	return sd * mult;	
}
//----------------------------------------------
// option_model.rho
//----------------------------------------------
number option_model.rho(number mult,
						 number d,						 
						 logical centered)
{
	if(CORE_INT.is_equal_zero(d,CORE_INT.DBL_EPSILON)) {		
		//return rho_anal(mult);
		d = CORE_INT.EPSILON;
	}
	
	number t, spd_u;
	prem_r(d, t, spd_u);
	
	number sd;	
	if(centered) {
		number spd_d;
		prem_r(-d, t, spd_d);
		sd = CORE_INT.num1der(spd_u,spd_d,d);
	}
	else {
		number sp;
		prem(t, sp);
		sd = CORE_INT.num1der_fd(sp,spd_u,d);
	}
	return sd * mult;
}
//----------------------------------------------
// option_model.psi
//----------------------------------------------
number option_model.psi(number mult,
							number d,						 
							logical centered)
{
	if(CORE_INT.is_equal_zero(d,CORE_INT.DBL_EPSILON)) {		
		//return rho_anal(mult);
		d = CORE_INT.EPSILON;
	}
	
	number t, spd_u;
	prem_u_r(d, t, spd_u);
	
	number sd;	
	if(centered) {
		number spd_d;
		prem_u_r(-d, t, spd_d);
		sd = CORE_INT.num1der(spd_u,spd_d,d);
	}
	else {
		number sp;
		prem(t, sp);
		sd = CORE_INT.num1der_fd(sp,spd_u,d);
	}
	return sd * mult;
}
//----------------------------------------------
// option_model.theta
//----------------------------------------------
number option_model.theta(	number mult,
							number d,
							out number diff)
{
	if(CORE_INT.is_equal_zero(d,CORE_INT.DBL_EPSILON)) {
		d = CORE_INT.EPSILON;
		//return theta_anal(mult);
	}
	
	number t, spd_d;
	prem_tau(d, t, spd_d);

	number sp;
	prem(t, sp);
	diff = (spd_d - sp) ;
	number sd = CORE_INT.num1der_fd(sp,spd_d,d);	
	return sd * mult;
}
//----------------------------------------------
// option_model.gamma_spot  (override if necessary to use cont_u_r argument in prem_u_spot(...)  eg. for B76)
//----------------------------------------------
number option_model.gamma_spot(	number mult,
								number d,						 
								logical centered)
{
	if(CORE_INT.is_equal_zero(d,CORE_INT.DBL_EPSILON)) {		
		return gamma_anal_spot(mult);
	}
	
	number t, sp,spd_u;
	prem(t, sp);
	prem_u_spot(d, t, spd_u);
	
	number sd;	
	if(centered) {
		number spd_d;
		prem_u_spot(-d, t, spd_d);
		sd = CORE_INT.num2der(sp,spd_u,spd_d,d);
	}
	else {
		number spd_u2;
		prem_u_spot(d*2, t, spd_u2);
		sd = CORE_INT.num2der(spd_u,spd_u2,sp,d);
	}
	return sd * mult;
}

//----------------------------------------------
// option_model.gamma_fwd
//----------------------------------------------
number option_model.gamma_fwd(	number mult,
								number d,						 
								logical centered)
{
	if(CORE_INT.is_equal_zero(d,CORE_INT.DBL_EPSILON)) {		
		return gamma_anal_fwd(mult);
	}
	
	number t, fp,fpd_u;
	prem(fp, t);
	prem_u_fwd(d, fpd_u, t);
	
	number fd;	
	if(centered) {
		number fpd_d;
		prem_u_fwd(-d, fpd_d, t);
		fd = CORE_INT.num2der(fp,fpd_u,fpd_d,d);
	}
	else {
		number fpd_u2;
		prem_u_fwd(d*2, fpd_u2, t);
		fd = CORE_INT.num2der(fpd_u,fpd_u2,fp,d);
	}
	return fd * mult;
}
//**********************************************
//**********************************************
// class vanilla_option_model  <abstract>
// parent class to all vanilla models (B76, BS, CRR,...)
//**********************************************
//**********************************************
class vanilla_option_model : public option_model
{
public:
	override option_model_args 	get_args() ;
	override number 			vol();
	
protected:
	vanilla_option_model_args args_;
};

//----------------------------------------------
// vanilla_option_model.vol
//----------------------------------------------
number vanilla_option_model.vol()  
{
	number v;
	switch(args_.qs_){
	case option_qs_e.VOL:
		v = args_.v_p_;
		break;	
	case option_qs_e.PREM_S:
		{
			v = imp_vol(args_.v_p_);		
			break;
		}
	case option_qs_e.PREM_F:
		{
			number df 	= exp(-args_.cont_r_*args_.tau_);					
			v = imp_vol(args_.v_p_*df);
			break;
		}
	}
	return v;
}

//----------------------------------------------
// vanilla_option_model.get_args
//----------------------------------------------
option_model_args vanilla_option_model.get_args()
{
	return args_;
}


//class enum option_pricing_model
	//option (category: 'Financial/Options/Vanilla') {B76,BS,BNORM,CRR};


//**********************************************
//**********************************************
// class b76_option_model
//**********************************************
//**********************************************
class b76_option_model : public vanilla_option_model
option (category: 'Financial/Options/Vanilla')
{
public:
	b76_option_model(number option(nullable) d = null<number>);

	number 					displacement();
	override option_model_e get_model();
	override void 			prem(out number, out number) ;
	override logical 		check_exercise(exercise_type) ;
	override logical 		check_asset(integer) ;

	override void add_u_sprd(number d) ;
	
	override void 	prem_u(number d, out number , out number ) ;
	override void 	prem_u_fwd(number d, out number , out number ) ;
	override void 	prem_u_spot(number d, out number , out number) ;
	override void 	prem_v(number d, out number , out number ) ;
	override void 	prem_r(number d, out number , out number ) ;
	override void 	prem_u_r(number d, out number , out number) ;
	override void 	prem_tau(number d, out number , out number ) ;
	//override number vega(number mult = 1,number d = 0, logical centered = true) ;
	override number imp_vol(number);
		
protected:
	override number delta_anal_spot(number) ;
	override number delta_anal_fut(number) ;
	override number delta_anal_fwd(number) ;
	override number gamma_anal_spot(number) ;
	override number gamma_anal_fwd(number) ;
	override number vega_anal(number) ;
	
	number 		displacement_;
};
//----------------------------------------------
// b76_option_model constructor
//----------------------------------------------
b76_option_model.b76_option_model(number option(nullable) d): displacement_(d)
{
	args_ = new b76_model_args();
}
//----------------------------------------------
// b76_option_model.get_model
//----------------------------------------------
option_model_e b76_option_model.get_model()
{
	return option_model_e.B76;
}
//----------------------------------------------
// b76_option_model.displacement
//----------------------------------------------
number b76_option_model.displacement()
{
	return displacement_;
}
//----------------------------------------------
// b76_option_model.check_exercise
//----------------------------------------------
logical b76_option_model.check_exercise(exercise_type et)
{
	return et == exercise_type.EURO;
}
//----------------------------------------------
// b76_option_model.check_asset
//----------------------------------------------
logical b76_option_model.check_asset(integer i)
{
	return i == 1;
}
//----------------------------------------------
// b76_option_model.prem
//----------------------------------------------
void b76_option_model.prem(out number fp, out number sp)  
{
	switch(args_.qs_){
	case option_qs_e.VOL:
		black76_prem(args_.cp_, args_.f_,   args_.k_,  args_.v_p_,  args_.cont_r_,
				 args_.tau_, args_.tau_v_,  displacement_, fp, sp);
		break;
	
	case option_qs_e.PREM_S:
		{
			number df 	= exp(-args_.cont_r_*args_.tau_);
			sp = args_.v_p_;
			fp = sp/df;
			break;
		}
	case option_qs_e.PREM_F:
		{
			number df 	= exp(-args_.cont_r_*args_.tau_);
			fp = args_.v_p_;
			sp = fp*df;
			break;
		}
	}
}

//----------------------------------------------
// b76_option_model.prem_u
//----------------------------------------------
void b76_option_model.prem_u(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}

	black76_prem(args_.cp_, args_.f_+ d,   args_.k_,  vol(),  args_.cont_r_,
				 args_.tau_, args_.tau_v_,  displacement_, fp, sp);
	
}
//----------------------------------------------
// b76_option_model.add_u_sprd
//----------------------------------------------
void b76_option_model.add_u_sprd(number d)
{
	args_.f_ = args_.f_+d;
}
//----------------------------------------------
// b76_option_model.prem_u_spot
//----------------------------------------------
void b76_option_model.prem_u_spot(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	number cont_u_r = null(args_.cont_u_r_) ? 0 : args_.cont_u_r_;
	number s = CORE_INT_OPT.bl_s_from_f(args_.f_, args_.cont_r_,args_.tau_, cont_u_r );
	number f = CORE_INT_OPT.bl_f_from_s(s+d,      args_.cont_r_,args_.tau_,	cont_u_r);
	black76_prem(args_.cp_, f,  args_.k_,  vol(),  args_.cont_r_,  args_.tau_,
				 args_.tau_v_,  displacement_, fp, sp);
}
//----------------------------------------------
// b76_option_model.prem_u_fwd
//----------------------------------------------
void b76_option_model.prem_u_fwd(number d, out number fp, out number sp)
{
	this.prem_u( d, fp, sp);
}

//----------------------------------------------
// b76_option_model.prem_v
//----------------------------------------------
void b76_option_model.prem_v(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	black76_prem(args_.cp_, args_.f_,   args_.k_,  vol()+ d,  args_.cont_r_,
				 args_.tau_, args_.tau_v_,  displacement_, fp, sp);
}

//----------------------------------------------
// b76_option_model.prem_r
//----------------------------------------------
void b76_option_model.prem_r(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	black76_prem(args_.cp_, args_.f_,   args_.k_,  vol(),  args_.cont_r_+ d,
				 args_.tau_, args_.tau_v_,  displacement_, fp, sp);
}
//----------------------------------------------
// b76_option_model.prem_u_r  
//----------------------------------------------
void b76_option_model.prem_u_r(number d, out number fp, out number sp )
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	number cont_u_r = null(args_.cont_u_r_) ? 0 : args_.cont_u_r_;
	number s = CORE_INT_OPT.bl_s_from_f(args_.f_, args_.cont_r_,args_.tau_, cont_u_r );
	number f = CORE_INT_OPT.bl_f_from_s(s, args_.cont_r_,args_.tau_,cont_u_r+d);
	black76_prem(args_.cp_, f,  args_.k_,  vol(),  args_.cont_r_,  args_.tau_,
				 args_.tau_v_,  displacement_, fp, sp);
}
//----------------------------------------------
// b76_option_model.prem_tau
//----------------------------------------------
void b76_option_model.prem_tau(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	black76_prem(args_.cp_, args_.f_,   args_.k_,  vol(),  args_.cont_r_,
				 args_.tau_-d, args_.tau_v_-d,  displacement_, fp, sp);
}
//----------------------------------------------
// b76_option_model.imp_vol
//----------------------------------------------
number b76_option_model.imp_vol(number sp)  
{
	return black_76_imp_vol(args_.cp_, args_.f_,   args_.k_,  args_.cont_r_,
							args_.tau_, args_.tau_v_,  displacement_, sp);
}
//----------------------------------------------
// b76_option_model.vega
//----------------------------------------------
/*number b76_option_model.vega(	logical div_spot,
								number mult,
								number d,
								logical centered )
{
	number sf;
	if(div_spot) 	
		sf = null(args_.s_) ? CORE_INT_OPT.bl_s_from_f(args_.f_, args_.cont_r_,args_.tau_,0 ) : args_.s_;	
	else
		sf = 1;

	return option_model.vega(sf, mult, d,centered );
}*/

//----------------------------------------------
// b76_option_model.delta_anal_spot
//----------------------------------------------
number b76_option_model.delta_anal_spot(number mult)
{
	number sd = black_76_delta_spot(args_.cp_, args_.f_,  args_.k_, vol(), args_.cont_u_r_,
									args_.tau_, args_.tau_v_, displacement_);
	return sd * mult;
}
//----------------------------------------------
// b76_option_model.delta_anal_fut
//----------------------------------------------
number b76_option_model.delta_anal_fut(number mult)
{
	number fd = black_76_delta_fut(args_.cp_, args_.f_,  args_.k_, vol(),  args_.cont_r_,
									args_.tau_, args_.tau_v_, displacement_);
	return fd * mult;
}
//----------------------------------------------
// b76_option_model.delta_anal_fwd
//----------------------------------------------
number b76_option_model.delta_anal_fwd(number mult)
{
	number fd = black_76_delta_fwd(args_.cp_, args_.f_,  args_.k_, vol(),
										args_.tau_v_, displacement_);
	return fd * mult;
}
//----------------------------------------------
// b76_option_model.gamma_anal_fwd
//----------------------------------------------
number b76_option_model.gamma_anal_fwd( number mult)
{
	number g = black_76_gamma_fwd(args_.f_,  args_.k_, vol(),
								  args_.tau_v_, displacement_);//FIX_ME (future gamma inte forward)
	return g*mult;
}
//----------------------------------------------
// b76_option_model.gamma_anal_spot
//----------------------------------------------
number b76_option_model.gamma_anal_spot( number mult)
{
	number g = black_76_gamma_spot(args_.f_,  args_.k_, vol(),  args_.cont_r_,
								   args_.cont_u_r_,  args_.tau_, args_.tau_v_, displacement_);
	return g*mult;
}
//----------------------------------------------
// b76_option_model.vega_anal
//----------------------------------------------
number b76_option_model.vega_anal(number mult )
{
	number v = black_76_vega(args_.f_,  args_.k_, vol(),  args_.cont_r_,
							 args_.tau_, args_.tau_v_, displacement_);
	return v*mult;
}

//**********************************************
//**********************************************
// class bs_option_model
//**********************************************
//**********************************************
class bs_option_model : public vanilla_option_model
option (category: 'Financial/Options/Vanilla')
{
public:
	bs_option_model();

	override option_model_e get_model();
	override void 			prem(out number, out number) ;
	override logical 		check_exercise(exercise_type) ;
	override logical 		check_asset(integer) ;

	override void add_u_sprd(number d) ;
	
	override void 	prem_u(number d, out number , out number ) ;
	override void 	prem_u_fwd(number d, out number , out number ) ;
	override void 	prem_u_spot(number d, out number , out number) ;
	override void 	prem_v(number d, out number , out number ) ;
	override void 	prem_r(number d, out number , out number ) ;
	override void 	prem_u_r(number d, out number , out number) ;
	override void 	prem_tau(number d, out number , out number ) ;
	//override number vega(logical div_spot = false,number mult = 1,number d = 0, logical centered = true ) ;
	override number imp_vol(number);

protected:
	override number delta_anal_spot(number) ;
	override number delta_anal_fut(number) ;
	override number delta_anal_fwd(number) ;
	override number gamma_anal_spot(number ) ;
	override number gamma_anal_fwd(number ) ;
	override number vega_anal(number ) ;
};
//----------------------------------------------
// bsoption_model constructor
//----------------------------------------------
bs_option_model.bs_option_model()
{
	args_ = new bs_model_args();
}
//----------------------------------------------
// bs_option_model.get_model
//----------------------------------------------
option_model_e bs_option_model.get_model()
{
	return option_model_e.BS;
}

//----------------------------------------------
// bs_option_model.check_exercise
//----------------------------------------------
logical bs_option_model.check_exercise(exercise_type et)
{
	return et == exercise_type.EURO;
}
//----------------------------------------------
// bs_option_model.check_asset
//----------------------------------------------
logical bs_option_model.check_asset(integer i)
{
	return i == 1;
}
//----------------------------------------------
// bs_option_model.prem
//----------------------------------------------
void bs_option_model.prem(out number fp, out number sp)  
{
	
	switch(args_.qs_){
	case option_qs_e.VOL:
		black_scholes_prem( args_.cp_, args_.s_,   args_.k_,  args_.v_p_,  args_.cont_r_,
						args_.cont_u_r_, args_.tau_, args_.tau_v_,  fp, sp);
		break;
	
	case option_qs_e.PREM_S:
		{
			number df 	= exp(-args_.cont_r_*args_.tau_);
			sp = args_.v_p_;
			fp = sp/df;
			break;
		}
	case option_qs_e.PREM_F:
		{
			number df 	= exp(-args_.cont_r_*args_.tau_);
			fp = args_.v_p_;
			sp = fp*df;
			break;
		}
	}
}

//----------------------------------------------
// bs_option_model.prem_u
//----------------------------------------------
void bs_option_model.prem_u(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	
	black_scholes_prem( args_.cp_, args_.s_+d,   args_.k_,  vol(),  args_.cont_r_,
						args_.cont_u_r_, args_.tau_, args_.tau_v_,  fp, sp);
}
//----------------------------------------------
// bs_option_model.add_u_sprd
//----------------------------------------------
void bs_option_model.add_u_sprd(number d)
{
	args_.s_ = args_.s_+d;
}
//----------------------------------------------
// bs_option_model.prem_v
//----------------------------------------------
void bs_option_model.prem_v(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	black_scholes_prem( args_.cp_, args_.s_,   args_.k_,  vol()+d,  args_.cont_r_,
						args_.cont_u_r_, args_.tau_, args_.tau_v_,  fp, sp);
}

//----------------------------------------------
// bs_option_model.prem_r
//----------------------------------------------
void bs_option_model.prem_r(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	black_scholes_prem( args_.cp_, args_.s_,   args_.k_,  vol(),  args_.cont_r_+d,
						args_.cont_u_r_, args_.tau_, args_.tau_v_,  fp, sp);
}
//----------------------------------------------
// bs_option_model.prem_u_r
//----------------------------------------------
void bs_option_model.prem_u_r(number d, out number fp, out number sp )
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	black_scholes_prem( args_.cp_, args_.s_,   args_.k_,  vol(),  args_.cont_r_,
						args_.cont_u_r_+d, args_.tau_, args_.tau_v_,  fp, sp);
}
//----------------------------------------------
// bs_option_model.prem_tau
//----------------------------------------------
void bs_option_model.prem_tau(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	black_scholes_prem( args_.cp_, args_.s_,   args_.k_,  vol(),  args_.cont_r_,
						args_.cont_u_r_, args_.tau_-d, args_.tau_v_-d,  fp, sp);
}
//----------------------------------------------
// bs_option_model.prem_u_fwd
//----------------------------------------------
void bs_option_model.prem_u_fwd(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	number f = CORE_INT_OPT.bl_f_from_s(args_.s_, args_.cont_r_,args_.tau_,	args_.cont_u_r_);
	number s = CORE_INT_OPT.bl_s_from_f(f+d,      args_.cont_r_,args_.tau_,args_.cont_u_r_ );	
	black_scholes_prem( args_.cp_, s,  args_.k_,  vol(),  args_.cont_r_,
						args_.cont_u_r_, args_.tau_, args_.tau_v_,  fp, sp);
}
//----------------------------------------------
// bs_option_model.prem_u_spot
//----------------------------------------------
void bs_option_model.prem_u_spot(number d, out number fp, out number sp )
{
	this.prem_u( d, fp, sp);
}
//----------------------------------------------
// bs_option_model.imp_vol
//----------------------------------------------
number bs_option_model.imp_vol(number sp)  
{
	return black_scholes_imp_vol(args_.cp_, args_.s_,   args_.k_,  args_.cont_r_,
								 args_.cont_u_r_, args_.tau_, args_.tau_v_,  sp);
}
//----------------------------------------------
// bs_option_model.vega
//----------------------------------------------
/*number bs_option_model.vega(	logical div_spot,
								number mult,
								number d,								
								logical centered )
{
	number sf = div_spot ? args_.s_: 1;	
	
	return option_model.vega(sf, mult, d, centered );
}*/
//----------------------------------------------
// bs_option_model.delta_anal_spot
//----------------------------------------------
number bs_option_model.delta_anal_spot(number mult)
{
	number sd = black_scholes_delta_spot(args_.cp_, args_.s_, args_.k_, vol(),  args_.cont_r_,
										 args_.cont_u_r_, args_.tau_, args_.tau_v_);
	return sd * mult;
}
//----------------------------------------------
// bs_option_model.delta_anal_fut
//----------------------------------------------
number bs_option_model.delta_anal_fut(number mult)
{
	number fd = black_scholes_delta_fut(args_.cp_, args_.s_, args_.k_, vol(),  args_.cont_r_,
										 args_.cont_u_r_, args_.tau_, args_.tau_v_);
	return fd * mult;
}
//----------------------------------------------
// bs_option_model.delta_anal_fwd
//----------------------------------------------
number bs_option_model.delta_anal_fwd(number mult)
{
	number fd = black_scholes_delta_fwd(args_.cp_, args_.s_, args_.k_, vol(),  args_.cont_r_,
										args_.cont_u_r_, args_.tau_, args_.tau_v_);
	return fd * mult;
}
//----------------------------------------------
// bs_option_model.gamma_anal_fwd
//----------------------------------------------
number bs_option_model.gamma_anal_fwd(number mult )
{
	number g = black_scholes_gamma_fwd(args_.s_,  args_.k_, vol(),  args_.cont_r_,
									   args_.cont_u_r_, args_.tau_, args_.tau_v_);
	return g*mult;
}
//----------------------------------------------
// bs_option_model.gamma_anal_spot
//----------------------------------------------
number bs_option_model.gamma_anal_spot(number mult )
{
	number g = black_scholes_gamma(args_.s_,  args_.k_, vol(),  args_.cont_r_,
								   args_.cont_u_r_, args_.tau_, args_.tau_v_);
	return g*mult;
}
//----------------------------------------------
// bs_option_model.vega_anal
//----------------------------------------------
number bs_option_model.vega_anal(number mult )
{
	number v = black_scholes_vega(args_.s_,  args_.k_, vol(),  args_.cont_r_,
								  args_.cont_u_r_, args_.tau_, args_.tau_v_);
	return v*mult;
}

//**********************************************
//**********************************************
// class bnorm_option_model
//**********************************************
//**********************************************
class bnorm_option_model : public vanilla_option_model
option (category: 'Financial/Options/Vanilla')
{
public:
	bnorm_option_model();

	override option_model_e get_model();
	override void 			prem(out number, out number) ;
	override logical 		check_exercise(exercise_type) ;
	override logical 		check_asset(integer) ;

	override void add_u_sprd(number d) ;
	
	override void 	prem_u(number d, out number , out number ) ;
	override void 	prem_u_fwd(number d, out number , out number ) ;
	override void 	prem_u_spot(number d, out number , out number) ;
	override void 	prem_v(number d, out number , out number ) ;
	override void 	prem_r(number d, out number , out number ) ;
	override void 	prem_u_r(number d, out number , out number) ;
	override void 	prem_tau(number d, out number , out number ) ;
	//override number vega(logical div_spot = false,number mult = 1,number d = 0, logical centered = true ) ;
	override number imp_vol(number);
	
protected:
	override number delta_anal_spot(number) ;
	override number delta_anal_fut(number) ;
	override number delta_anal_fwd(number) ;
	override number gamma_anal_spot(number ) ;
	override number gamma_anal_fwd(number ) ;
	override number vega_anal(number ) ;

};
//----------------------------------------------
// bnorm_option_model constructor
//----------------------------------------------
bnorm_option_model.bnorm_option_model()
{
	args_ = new bs_model_args();
}
//----------------------------------------------
// bnorm_option_model.get_model
//----------------------------------------------
option_model_e bnorm_option_model.get_model()
{
	return option_model_e.BNORM;
}

//----------------------------------------------
// bnorm_option_model.check_exercise
//----------------------------------------------
logical bnorm_option_model.check_exercise(exercise_type et)
{
	return et == exercise_type.EURO;
}
//----------------------------------------------
// bnorm_option_model.check_asset
//----------------------------------------------
logical bnorm_option_model.check_asset(integer i)
{
	return i == 1;
}
//----------------------------------------------
// bnorm_option_model.prem
//----------------------------------------------
void bnorm_option_model.prem(out number fp, out number sp)  
{
	
	switch(args_.qs_){
	case option_qs_e.VOL:
		black_norm_prem( args_.cp_, args_.f_,   args_.k_,  args_.v_p_,  args_.cont_r_,
					 args_.tau_, args_.tau_v_,  fp, sp);
		break;
	
	case option_qs_e.PREM_S:
		{
			number df 	= exp(-args_.cont_r_*args_.tau_);
			sp = args_.v_p_;
			fp = sp/df;
			break;
		}
	case option_qs_e.PREM_F:
		{
			number df 	= exp(-args_.cont_r_*args_.tau_);
			fp = args_.v_p_;
			sp = fp*df;
			break;
		}
	}
}

//----------------------------------------------
// bnorm_option_model.prem_u
//----------------------------------------------
void bnorm_option_model.prem_u(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	black_norm_prem( args_.cp_, args_.f_+d,   args_.k_,  vol(),  args_.cont_r_,
					 args_.tau_, args_.tau_v_,  fp, sp);
}
//----------------------------------------------
// bnorm_option_model.add_u_sprd
//----------------------------------------------
void bnorm_option_model.add_u_sprd(number d)
{
	args_.f_ = args_.f_+d;
}
//----------------------------------------------
// bnorm_option_model.prem_v
//----------------------------------------------
void bnorm_option_model.prem_v(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	black_norm_prem( args_.cp_, args_.f_,   args_.k_,  vol()+d,  args_.cont_r_,
					 args_.tau_, args_.tau_v_,  fp, sp);
}

//----------------------------------------------
// bnorm_option_model.prem_r
//----------------------------------------------
void bnorm_option_model.prem_r(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	black_norm_prem( args_.cp_, args_.f_,   args_.k_,  vol(),  args_.cont_r_+d,
					 args_.tau_, args_.tau_v_,  fp, sp);
}
//----------------------------------------------
// bnorm_option_model.prem_u_r  (u_r not available here)
//----------------------------------------------
void bnorm_option_model.prem_u_r(number d, out number fp, out number sp )
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	number cont_u_r = null(args_.cont_u_r_) ? 0 : args_.cont_u_r_;
	number s = CORE_INT_OPT.bl_s_from_f(args_.f_, args_.cont_r_,args_.tau_, cont_u_r );
	number f = CORE_INT_OPT.bl_f_from_s(s, args_.cont_r_,args_.tau_,cont_u_r+d);
	black_norm_prem(args_.cp_, f,  args_.k_,  vol(),  args_.cont_r_,  args_.tau_,
					args_.tau_v_,  fp, sp);
}
//----------------------------------------------
// bnorm_option_model.prem_tau
//----------------------------------------------
void bnorm_option_model.prem_tau(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	black_norm_prem( args_.cp_, args_.f_,   args_.k_,  vol(),  args_.cont_r_,
					 args_.tau_-d, args_.tau_v_-d,  fp, sp);
}

//----------------------------------------------
// bnorm_option_model.prem_u_spot
//----------------------------------------------
void bnorm_option_model.prem_u_spot(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	number cont_u_r = null(args_.cont_u_r_) ? 0 : args_.cont_u_r_;
	number s = CORE_INT_OPT.bl_s_from_f(args_.f_, args_.cont_r_,args_.tau_,cont_u_r );
	number f = CORE_INT_OPT.bl_f_from_s(s+d, args_.cont_r_,args_.tau_,	 cont_u_r);
	black_norm_prem( args_.cp_, f,   args_.k_,  vol(),  args_.cont_r_,
					 args_.tau_, args_.tau_v_,  fp, sp);
}
//----------------------------------------------
// bnorm_option_model.prem_u_fwd
//----------------------------------------------
void bnorm_option_model.prem_u_fwd(number d, out number fp, out number sp)
{
	this.prem_u( d, fp, sp);
}

//----------------------------------------------
// bnorm_option_model.imp_vol
//----------------------------------------------
number bnorm_option_model.imp_vol(number sp)  
{
	return black_norm_imp_vol(args_.cp_, args_.f_,   args_.k_,  args_.cont_r_,
							  args_.tau_, args_.tau_v_,  sp);
}

//----------------------------------------------
// bnorm_option_model.vega
//----------------------------------------------
/*number bnorm_option_model.vega(	logical div_spot,
								number mult,
								number d,
								logical centered )
{
	number sf;
	if(div_spot) 	
		sf = null(args_.s_) ? CORE_INT_OPT.bl_s_from_f(args_.f_, args_.cont_r_,args_.tau_,0 ) : args_.s_;	
	else
		sf = 1;
	return option_model.vega(sf, mult, d,centered );
}*/
//----------------------------------------------
// bnorm_option_model.delta_anal_spot  CHECK THIS
//----------------------------------------------
number bnorm_option_model.delta_anal_spot(number mult)
{
	number sd = black_norm_delta(args_.cp_, args_.f_, args_.k_, vol(),
								 args_.cont_r_,  args_.tau_, args_.tau_v_);
	return sd*mult;
}
//----------------------------------------------
// bnorm_option_model.delta_anal_fut  CHECK THIS (same as delta_anal_spot)
//----------------------------------------------
number bnorm_option_model.delta_anal_fut(number mult)
{
	number sd = black_norm_delta(args_.cp_, args_.f_, args_.k_, vol(),
								 args_.cont_r_,  args_.tau_, args_.tau_v_);
	return sd*mult;
}
//----------------------------------------------
// bnorm_option_model.delta_anal_fwd
//----------------------------------------------
number bnorm_option_model.delta_anal_fwd(number mult)
{
	number fd = black_norm_delta_fwd(args_.cp_, args_.f_, args_.k_, vol(),
									 args_.cont_r_,  args_.tau_,args_.tau_v_);
	return fd*mult;
}

//----------------------------------------------
// bnorm_option_model.gamma_anal_fwd
//----------------------------------------------
number bnorm_option_model.gamma_anal_fwd( number mult)
{
	number g = black_norm_gamma(args_.f_,  args_.k_, vol(),  args_.cont_r_,
								args_.tau_, args_.tau_v_);
	return g*mult;
}
//----------------------------------------------
// bnorm_option_model.gamma_anal_spot
//----------------------------------------------
number bnorm_option_model.gamma_anal_spot( number mult)
{
	//number g = black_norm_gamma(args_.f_,  args_.k_, args_.v_,  args_.cont_r_,  args_.tau_, args_.tau_v_);
	//return g*mult;
	return null<number>;
}
//----------------------------------------------
// bnorm_option_model.vega_anal
//----------------------------------------------
number bnorm_option_model.vega_anal( number mult)
{
	number v = black_norm_vega(args_.f_,  args_.k_, vol(),  args_.cont_r_,
							   args_.tau_, args_.tau_v_);
	return v*mult;
}

//**********************************************
//**********************************************
// class crr_option_model
//**********************************************
//**********************************************
class crr_option_model : public vanilla_option_model
option (category: 'Financial/Options/Vanilla')
{
public:
	crr_option_model(integer);

	override option_model_e get_model();	
	override void 			prem(out number, out number) ;
	override logical 		check_exercise(exercise_type) ;
	override logical 		check_asset(integer) ;

	override void add_u_sprd(number d) ;
	
	override void 	prem_u(number d, out number , out number ) ;
	override void 	prem_u_fwd(number d, out number , out number ) ;
	override void 	prem_u_spot(number d, out number , out number) ;
	override void 	prem_v(number d, out number , out number ) ;
	override void 	prem_r(number d, out number , out number ) ;
	override void 	prem_u_r(number d, out number , out number ) ;
	override void 	prem_tau(number d, out number , out number ) ;
	//override number vega(logical div_spot = false,number mult = 1,number d = 0, logical centered = true ) ;
	override number imp_vol(number);

protected:
	override number delta_anal_spot(number) ;
	override number delta_anal_fut(number) ;
	override number delta_anal_fwd(number) ;
	override number gamma_anal_spot(number ) ;
	override number gamma_anal_fwd(number ) ;
	override number vega_anal(number) ;
	integer steps_;
};
//----------------------------------------------
// crr_option_model constructor
//----------------------------------------------
crr_option_model.crr_option_model(integer steps): steps_(steps)
{
	QL_FAIL_COND(steps_== err_int() || steps_ <= 0,"invalid number of steps in crr model");
	args_ = new crr_model_args();
}
//----------------------------------------------
// crr_option_model.get_model
//----------------------------------------------
option_model_e crr_option_model.get_model()
{
	return option_model_e.CRR;
}

//----------------------------------------------
// crr_option_model.check_exercise
//----------------------------------------------
logical crr_option_model.check_exercise(exercise_type et)
{
	return et == exercise_type.EURO || et == exercise_type.AMER;
}
//----------------------------------------------
// crr_option_model.check_asset
//----------------------------------------------
logical crr_option_model.check_asset(integer i)
{
	return i == 1;
}
//----------------------------------------------
// crr_option_model.prem
//----------------------------------------------
void crr_option_model.prem(out number fp, out number sp)  
{
	switch(args_.qs_){
	case option_qs_e.VOL:
		crr_prem(	args_.cp_, args_.et_, args_.s_,   args_.k_,  args_.v_p_,  args_.cont_r_,
				args_.pay_, args_.ex_, args_.div_,args_.tau_,args_.tau_v_,steps_, fp, sp);
		break;
	
	case option_qs_e.PREM_S:
		{
			number df 	= exp(-args_.cont_r_*args_.tau_);
			sp = args_.v_p_;
			fp = sp/df;
			break;
		}
	case option_qs_e.PREM_F:
		{
			number df 	= exp(-args_.cont_r_*args_.tau_);
			fp = args_.v_p_;
			sp = fp*df;
			break;
		}
	}
}

//----------------------------------------------
// crr_option_model.prem_u
//----------------------------------------------
void crr_option_model.prem_u(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	crr_prem(	args_.cp_, args_.et_, args_.s_+d,   args_.k_, vol(),  args_.cont_r_,
				args_.pay_, args_.ex_, args_.div_,args_.tau_,args_.tau_v_,steps_, fp, sp);
}
//----------------------------------------------
// crr_option_model.add_u_sprd
//----------------------------------------------
void crr_option_model.add_u_sprd(number d)
{
	args_.s_ = args_.s_+d;
}
//----------------------------------------------
// crr_option_model.prem_v
//----------------------------------------------
void crr_option_model.prem_v(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	crr_prem(	args_.cp_, args_.et_, args_.s_,   args_.k_,  vol()+d,  args_.cont_r_,
				args_.pay_, args_.ex_, args_.div_,args_.tau_,args_.tau_v_,steps_, fp, sp);
}

//----------------------------------------------
// crr_option_model.prem_r
//----------------------------------------------
void crr_option_model.prem_r(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	crr_prem(	args_.cp_, args_.et_, args_.s_,   args_.k_,  vol(),  args_.cont_r_+d,
				args_.pay_, args_.ex_, args_.div_,args_.tau_,args_.tau_v_,steps_, fp, sp);
}
//----------------------------------------------
// crr_option_model.prem_u_r  (u_r not available here)
//----------------------------------------------
void crr_option_model.prem_u_r(number d, out number fp, out number sp )
{
	fp = sp = null<number>;
}
//----------------------------------------------
// crr_option_model.prem_tau  FIX_ME --> dividend data must be updated
//----------------------------------------------
void crr_option_model.prem_tau(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	crr_prem(	args_.cp_, args_.et_, args_.s_,   args_.k_,  vol(),  args_.cont_r_,
				args_.pay_, args_.ex_, args_.div_,args_.tau_-d,args_.tau_v_-d,steps_, fp, sp);
}

//----------------------------------------------
// crr_option_model.prem_u_fwd
//----------------------------------------------
void crr_option_model.prem_u_fwd(number d, out number fp, out number sp)
{
	if(CORE_INT.is_equal_zero(d, CORE_INT.DBL_EPSILON)) {
		this.prem(fp, sp);
		return;		
	}
	
	number pv_div = CORE_INT_OPT.pv_div(args_.tau_v_,0,
									args_.pay_,
									args_.ex_,
									args_.div_,
									args_.cont_r_);//check div w.r.t. trade vs settle issue, pv is as of tradedate
	
	number cont_u_r = CORE_INT_OPT.div_to_cont_div_yld(args_.s_,pv_div, args_.tau_);
	
	number f = CORE_INT_OPT.bl_f_from_s(args_.s_, args_.cont_r_,args_.tau_,	cont_u_r);
	number s = CORE_INT_OPT.bl_s_from_f(f+d, args_.cont_r_,args_.tau_, cont_u_r);	
	crr_prem(	args_.cp_, args_.et_, s,   args_.k_,  vol(),  args_.cont_r_,
				args_.pay_, args_.ex_, args_.div_,args_.tau_,args_.tau_v_,steps_, fp, sp);
}
//----------------------------------------------
// crr_option_model.prem_u_spot
//----------------------------------------------
void crr_option_model.prem_u_spot(number d, out number fp, out number sp)
{
	this.prem_u( d, fp, sp);
}
//----------------------------------------------
// crr_option_model.imp_vol
//----------------------------------------------
number crr_option_model.imp_vol(number p)  
{
	return -1;//TODO
}
//----------------------------------------------
// crr_option_model.vega
//----------------------------------------------
/*number crr_option_model.vega(logical div_spot,
							 number mult,
							 number d,							 							 
							 logical centered )
{
	number sf = div_spot ? args_.s_: 1;	
	return option_model.vega(sf, mult, d,centered );
}*/
//----------------------------------------------
// crr_option_model.delta_anal_spot
//----------------------------------------------
number crr_option_model.delta_anal_spot(number mult)
{
	return null<number>;
}
//----------------------------------------------
// crr_option_model.delta_anal_fut
//----------------------------------------------
number crr_option_model.delta_anal_fut(number mult)
{
	return null<number>;
}
//----------------------------------------------
// crr_option_model.delta_anal_fwd
//----------------------------------------------
number crr_option_model.delta_anal_fwd(number mult)
{
	return null<number>;
}
//----------------------------------------------
// crr_option_model.gamma_anal_spot
//----------------------------------------------
number crr_option_model.gamma_anal_spot(number mult)
{
	return null<number>;
}
//----------------------------------------------
// crr_option_model.gamma_anal_fwd
//----------------------------------------------
number crr_option_model.gamma_anal_fwd(number mult)
{
	return null<number>;
}
//----------------------------------------------
// crr_option_model.vega_anal
//----------------------------------------------
number crr_option_model.vega_anal(number mult)
{
	return null<number>;
}

//**********************************************
//**********************************************
// class baw_option_model
//**********************************************
//**********************************************
// TO BE DONE

module CORE_INT
{
	//----------------------------------------------
	// get_vanilla_model
	//----------------------------------------------
	vanilla_option_model get_vanilla_model(option_model_e m,
										   number option(nullable) d = null<number>,
										   integer steps = err_int())
	{
		switch(m) {
		case option_model_e.B76:
			return new b76_option_model(d);
		case option_model_e.BS:
			return new bs_option_model();
		case option_model_e.CRR:
			return new crr_option_model(steps);
		case option_model_e.BNORM:
			return new bnorm_option_model();
		}

		return null<vanilla_option_model>;
	}
}