/*	
	Option sabr funcs
	Developer: Algorithmica Research, Magnus Nyström			
*/

module CORE_INT_VOL
{
	//-------------------------------------------------------
	// sabr_alpha
	//-------------------------------------------------------
	number sabr_alpha(	number f,
						number beta, 
						number rho, 
						number volvol,
						number t_exp,
						number sigma_atm)
	{
		try{			
			vector(number) alphas;
			if(beta==1){
				number a2 	= (1+(2-3*sqr(rho))/24*sqr(volvol)*t_exp)/(rho*volvol*t_exp/4);
				number a3 	= -sigma_atm/(rho*volvol*t_exp/4);
				alphas 		= CORE_INT.solve_quadratic(a2, a3);
			}
			else{
				number h 	= sqr(1-beta)*t_exp/(24*pow(f,2-2*beta));				
				number a2 	= ( (rho*beta*volvol*t_exp)/(4*pow(f,1-beta)) )/h;
				number a3 	= ( 1+ (2-3*sqr(rho))*sqr(volvol)*t_exp/24 )/h;
				number a4 	= -(pow(f,1-beta)*sigma_atm)/h;
				alphas 		= CORE_INT.solve_cubic(a2, a3, a4);	
			}
			sort(alphas);
			number alpha = null<number>;		
			for( integer i=0; i < v_size(alphas); i++){
				if( alphas[i] >= 0 ){
					alpha = alphas[i];
					break;
				}
			}
			return alpha;	
		}
		catch {
			return  null<number>;
		}	
	}

	//-------------------------------------------------------
	// sabr_blackvol_obloj
	// Jan Obloj. Fine-tune your smile : Correction to Hagan, 2008
	// http://arxiv.org/abs/0708.0998
	//-------------------------------------------------------
	number sabr_blackvol_obloj(	number f,
								number k, 
								number t_exp, 
								number alpha, 
								number beta, 
								number rho, 
								number volvol) 
	
	{
		number t1 = sqr(1-beta)/24.0;
		number t2 = alpha*alpha/pow(f*k,1.0-beta);
		number t3 = 0.25*rho*beta*alpha*volvol/pow(f*k, (1.0-beta)/2.0);
		number t4 = (2.0-3.0*rho*rho)/24.0*volvol*volvol;
		number s1 = t1*t2+t3+t4;

		number r1 = volvol/alpha*(pow(f,1.0-beta) - pow(k,1.0-beta))/(1.0-beta);
		number t5 = (sqrt(1.0-2.0*rho*r1+r1*r1)+r1-rho)/(1.0-rho);
		number s0 = volvol * log(f/k)/log(t5);
		number val = s0*(1.0+s1*t_exp);
		return val;

	}
	//-------------------------------------------------------
	// sabr_blackvol_obloj
	//-------------------------------------------------------
	number sabr_blackvol_obloj(	number			f,
							number 			k, 
							number 			t_exp, 
							vector(number) 	param)
	{
		return sabr_blackvol_obloj(f, k , t_exp, param[0], param[1], param[2], param[3]);
	}
}