/*	
	Option utility funcs 
	Developer: Algorithmica Research, Magnus Nyström			
*/

option(null: hard);	

module CORE_INT_OPT
{
	//-------------------------------------------------------
	//class  k_max_solve 
	//-------------------------------------------------------	
	class k_max_solve 
	{
	public:
		k_max_solve(number, number,number, number,number,number option(nullable));
		k_max_solve(number, number,number);
		number 	f(number) ;	
	protected:
		number vv_;
		logical z_vol_;
		number t1_;
		number s_;
		number f_;				
	};

	k_max_solve.k_max_solve(number s,
							number v,
							number tau,
							number tau_v,
							number cont_r,
							number option(nullable) cont_u_r): s_(s)
	{
		vv_ 		= v * pow(tau_v, 0.5) ;
		QL_FAIL_COND(CORE_INT.negative_check(vv_, CORE_INT.DBL_EPSILON),"invalid volatility (negative)");			
		z_vol_		= CORE_INT.is_equal_zero(vv_,CORE_INT.DBL_EPSILON );		
		if(null(cont_u_r)) cont_u_r = 0;
		t1_ 		= (cont_r-cont_u_r+v*v*0.5)*tau_v;
		number df 	= exp(-cont_r*tau);	
		number dfq 	= exp(-cont_u_r*tau);
		f_ 			= bl_f_from_s(s, df,dfq);	
	}

	k_max_solve.k_max_solve(number f,
							number v,							
							number tau_v): t1_(null<number>),f_(f)
	{
		vv_ 	= v * pow(tau_v, 0.5) ;
		QL_FAIL_COND(CORE_INT.negative_check(vv_, CORE_INT.DBL_EPSILON),"invalid volatility (negative)");			
		z_vol_	= CORE_INT.is_equal_zero(vv_,CORE_INT.DBL_EPSILON );		
		t1_ 	= null<number>;		
	}
	
	number k_max_solve.f(number k)
	{
		QL_FAIL_COND(CORE_INT.negative_check(k, CORE_INT.DBL_EPSILON),"invalid strike (negative)");
		
		number d2;
		
		if(z_vol_) {
			if(f_ == k)
				d2 = 0 ;			
			else if(f_> k)
				d2 = CORE_INT.MAX_NUMBER;
			else
				d2 = CORE_INT.MIN_NUMBER;
		}
		else {
			logical z_k	= CORE_INT.is_equal_zero(k,CORE_INT.DBL_EPSILON );
			if(z_k) 
				d2 = CORE_INT.MAX_NUMBER;			
			else {
				number d1 	= null(t1_) ? log(f_/k)/vv_ + vv_ / 2.0 : (log(s_/k)+t1_)/vv_  ;				
				d2 			= d1-vv_ ;
			}
		}
		return vv_ * normal_cdf(d2) - normal_pdf(d2);		
	}

	//-------------------------------------------------------
	// k_max_solve
	//-------------------------------------------------------
	number k_max_solve(	number s,
						number v,
						number tau,
						number tau_v,
						number cont_r,
						number option(nullable) cont_u_r,
						number left_limit,
						number right_limit,
						logical fix_limits 	= false,
						number tol			= 0.000001)
	{
							
			
		CORE_INT_OPT.k_max_solve k = new CORE_INT_OPT.k_max_solve(s,v,tau,tau_v,cont_r,cont_u_r);

		number x1 	= left_limit;
		number x2	= right_limit;		

		if(x1==0) x1 = tol;
		
		if(!fix_limits)
			zero_bracket(k, &CORE_INT_OPT.k_max_solve.f, x1, x2);
		return zero_brent(k, &CORE_INT_OPT.k_max_solve.f, x1, x2, tol);
		
	}
	//-------------------------------------------------------
	// k_max_solve
	//-------------------------------------------------------
	number k_max_solve(	number f,
						number v,							
						number tau_v,
						number left_limit,
						number right_limit,
						logical fix_limits 	= false,
						number tol			= 0.000001)
	{
							
			
		CORE_INT_OPT.k_max_solve k = new CORE_INT_OPT.k_max_solve(f,v,tau_v);

		number x1 	= left_limit;
		number x2	= right_limit;		

		if(x1==0) x1 = tol;
		
		if(!fix_limits)
			zero_bracket(k, &CORE_INT_OPT.k_max_solve.f, x1, x2);
		return zero_brent(k, &CORE_INT_OPT.k_max_solve.f, x1, x2, tol);
		
	}
}