/*	
	Option utility funcs 
	Developer: Algorithmica Research, Magnus Nyström			
*/

option(null: hard);	
//-------------------------------------------------------
// enum delivery_type
//-------------------------------------------------------
class enum delivery_type
	option (category: 'Financial/Options/Vanilla')
{
	CASH option(str:"cash"),
	PHYS option(str:"physical")
};
//-------------------------------------------------------
// enum opt_exp
//-------------------------------------------------------
class enum opt_exp
	option (category: 'Financial/Options/Vanilla')
{
	OPTE_IN
		option(str:"in")
		option(desc:"< expiry"),
	OPTE_EXP
		option(str:"exp")
		option(desc:"= expiry"),
	OPTE_DEL
		option(str:"del")
		option(desc:"expired: < delivery"),
	OPTE_OUT
		option(str:"out")
		option(desc:"expired: >= delivery")
};
//-------------------------------------------------------
// enum opt_type
//-------------------------------------------------------
class enum opt_type
	option (category: 'Financial/Options/Vanilla')
{
	OPT_CALL option(str:"call"),
	OPT_PUT option(str:"put")
};

out vector(string) opt_types()
	option (category: 'Financial/Options/Vanilla')
{
	vector(opt_type) e;
	expand_enum(e);
	return string(e);
}

opt_type opt_type_map(string et)
	option (category: 'Financial/Options/Vanilla')
{
	//try {
		opt_type e;
		str_to_enum(et,e);
		return e;
	//}
	//catch {
	//	return null<opt_type>;
	//}
}
class enum opt_prem_type  option (category: 'Financial/Options/Vanilla')
{
	UNIT option(str:"unit"),
	PCNT option(str:"pcnt"),
	AMNT option(str:"amnt")
};
//-------------------------------------------------------
// enum exercise_type
//-------------------------------------------------------
class enum exercise_type
	option (category: 'Financial/Options/Vanilla')
{
	AMER option(str:"american"),
	/*BERM option(str:"bermudan"),*/
	EURO option(str:"european")
};

out vector(string) exercise_types()
	option (category: 'Financial/Options/Vanilla')
{
	vector(exercise_type) e;
	expand_enum(e);
	return string(e);
}

exercise_type exercise_type_map(string et)
	option (category: 'Financial/Options/Vanilla')
{
	//try{
		exercise_type e;
		str_to_enum(et,e);
		return e;
	//}
	//catch {
	//	return null<exercise_type>;
	//}
}




module CORE_INT_OPT
{
	//-------------------------------------------------------
	// bl_s_from_f
	//-------------------------------------------------------
	number bl_s_from_f(	number f, 
						number df,
						number u_df = 1.0)
	{
		return f*df/u_df;
	}
	//-------------------------------------------------------
	// bl_s_from_f
	//-------------------------------------------------------
	number bl_s_from_f(	number f, 
						number cont_r,			
						number tau,				//basedate=settlement
						number cont_u_r )		
	{
		number df 	= exp(-cont_r*tau);
		number u_df = exp(-cont_u_r*tau);
		return bl_s_from_f(	 f, df,u_df );
	}
	
	//-------------------------------------------------------
	// bl_f_from_s
	//-------------------------------------------------------
	number bl_f_from_s(	number s, 
						number cont_r,			
						number tau,				//basedate=settlement
						number cont_u_r)	
	{
		number df 	= exp(-cont_r*tau);
		number u_df = exp(-cont_u_r*tau);
		return s*u_df/df;
	}
	//-------------------------------------------------------
	// bl_f_from_s
	//-------------------------------------------------------
	number bl_f_from_s(	number s, 
						number df,
						number u_df =1.0)	
	{
		return s*u_df/df;
	}
	//-------------------------------------------------------
	// bl_cont_u_r_from_sf
	//-------------------------------------------------------
	number bl_cont_u_r_from_sf(	number s,
								number f,
								number cont_r,			
								number tau )		
	{
		number df = exp(-cont_r*tau);
		return -log(f/s*df)/tau;
	}
	//-------------------------------------------------------
	// bl_cont_r_from_sf
	//-------------------------------------------------------
	number bl_cont_r_from_sf(	number s,
								number f,
								number cont_u_r,			
								number tau )		
	{
		number u_df = exp(-cont_u_r*tau);
		return -log(s/f*u_df)/tau;
	}
	//-------------------------------------------------------
	// bs_pre_calc
	//-------------------------------------------------------
	class bs_pre_calc
	{
	public:
		bs_pre_calc(number 			s,
					number 			k, 
					number 			v, 
					number 			cont_r,
					number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
					number 			tau,						
					number 			tau_v);
		
		bs_pre_calc(number 			f, 
					number 			k, 
					number 			v, 
					number 			cont_r,						
					number 			tau,						
					number 			tau_v);
		
		bs_pre_calc(number 			f, 
					number 			k, 
					number 			v, 
					number 			df,						
					number 			tau_v);
	
		
		number Nd1();
		number Nd2();
		number Nd1_rev();
		number Nd2_rev();
		number pdf_nd1();
		number pdf_nd2();
		
		
		number f_;
		number k_;
		number vv_;
		number d1_;
		number d2_;
		number df_;
		number dfq_;
		logical z_vol_;
		logical z_k_;
		
	protected:
		void init(number, number);
	};

	number bs_pre_calc.Nd1()
	{
		if(z_vol_)
			return f_ < k_ ? 0.0 : normal_cdf(d1_);
		else {
			return z_k_ ? 1.0 : normal_cdf(d1_);			
		}
	}
	number bs_pre_calc.Nd1_rev() //put
	{
		if(z_vol_)
			return f_ < k_ ? 1.0 : normal_cdf(-d1_);
		else {
			return z_k_ ? 0.0 : normal_cdf(-d1_);			
		}
	}	

	number bs_pre_calc.Nd2() 
	{
		if(z_vol_)
			return f_ < k_ ? 0.0 : normal_cdf(d2_);
		else {
			return z_k_ ? 1.0 : normal_cdf(d2_);			
		}
	}
	number bs_pre_calc.Nd2_rev() {//put
		if(z_vol_)
			return f_ < k_ ? 1.0 : normal_cdf(-d2_);
		else {
			return z_k_ ? 0.0 : normal_cdf(-d2_);			
		}
	}

	number bs_pre_calc.pdf_nd1() 		{ return z_vol_ || z_k_ ? 0: normal_pdf(d1_);}
	number bs_pre_calc.pdf_nd2() 		{ return z_vol_ || z_k_ ? 0: normal_pdf(d2_);}

	//-------------------------------------------------------
	// init
	//-------------------------------------------------------
	void bs_pre_calc.init(	number v,
							number tau_v)
	{
		vv_ 		= v * pow(tau_v, 0.5) ;

		QL_FAIL_COND(CORE_INT.negative_check(vv_, CORE_INT.DBL_EPSILON),"invalid volatility (negative)");			
		z_vol_	= CORE_INT.is_equal_zero(vv_,CORE_INT.DBL_EPSILON );
		QL_FAIL_COND(CORE_INT.negative_check(k_, CORE_INT.DBL_EPSILON),"invalid strike (negative)");
		z_k_ 	= CORE_INT.is_equal_zero(k_,CORE_INT.DBL_EPSILON );
			
		if(z_vol_) {
			if(f_ == k_){
				d1_ = 0 ;
				d2_ = 0 ;
			}
			else if(f_ > k_) {
				d1_ = CORE_INT.MAX_NUMBER ;
				d2_ = CORE_INT.MAX_NUMBER ;
			}
			else {
				d1_ = CORE_INT.MIN_NUMBER;
				d2_ = CORE_INT.MIN_NUMBER ;
			}
		}
	}
	
	//-------------------------------------------------------
	// constructor
	//-------------------------------------------------------
	bs_pre_calc.bs_pre_calc(	number 			s,
								number 			k, 
								number 			v, 
								number 			cont_r,
								number option(nullable) cont_u_r,//eg. foreign curr rate, dividend-yld
								number 			tau,						
								number 			tau_v): k_(k)
	{
		if(null(cont_u_r)) cont_u_r = 0;
		df_ 	= exp(-cont_r*tau);	
		dfq_ 	= exp(-cont_u_r*tau);
		f_ 		= bl_f_from_s(s, df_,dfq_);			
		init( v, tau_v);
		
		if(!z_vol_) {
			
			if(z_k_) {
				d1_ = CORE_INT.MAX_NUMBER ;
				d2_ = CORE_INT.MAX_NUMBER;
			}
			else {
				number t1 = (cont_r-cont_u_r+v*v*0.5)*tau_v;
				d1_ = (log(s/k)+t1)/vv_  ;
				d2_ = d1_-vv_ ;
			}
		}		
	}
	//-------------------------------------------------------
	// constructor
	//-------------------------------------------------------
	bs_pre_calc.bs_pre_calc(	number 			f, 
								number 			k, 
								number 			v, 
								number 			cont_r,						
								number 			tau,						
								number 			tau_v): f_(f),k_(k),dfq_(null<number>)
	{
		df_ = exp(-cont_r*tau);	
		init( v, tau_v);
		if(!z_vol_) {
			if(z_k_) {  
				d1_ = CORE_INT.MAX_NUMBER ;
				d2_ = CORE_INT.MAX_NUMBER;
			}
			else {
				d1_ = log(f_/k_)/vv_ + vv_/2.0 ;
				d2_ = d1_-vv_ ;
			}
		}		
	}

	//-------------------------------------------------------
	// constructor
	//-------------------------------------------------------
	bs_pre_calc.bs_pre_calc(	number 			f, 
								number 			k, 
								number 			v, 
								number 			df,																			
								number 			tau_v): f_(f), k_(k),df_(df), dfq_(null<number>)
	{
		init( v, tau_v);
		if(!z_vol_) {
			if(z_k_) {
				d1_ = CORE_INT.MAX_NUMBER ;
				d2_ = CORE_INT.MAX_NUMBER;
			}
			else {
				d1_ = log(f_/k_)/vv_ + vv_/2.0 ;
				d2_ = d1_-vv_ ;
			}
		}
	}
	
	//-------------------------------------------------------
	// bs_w
	// LambertW function ( from DESY)
	// W(x) * exp(W(x)) = x
	//-------------------------------------------------------
	number lambert_w(number x ) 
	{
		if (x > 376.0 ) 
  			return log(x - 4.0) - (1.0 - 1.0 / log(x)) * log(log(x));
 		else
			return 0.665 * (1.0 + 0.0195 * log(x + 1.0)) * log(x + 1.0) + 0.04;
	}

	//-------------------------------------------------------
	// bs_normed_prem
	//-------------------------------------------------------
	number bs_normed_prem(	number absmu, 
							number s)
	{
		if (s <= CORE_INT.DBL_EPSILON) 
	  		return 0;
  	
		number eta = absmu / s;
		number h = exp(absmu);
	
		return (1.0-normal_cdf(eta - s)) / h - (1.0-normal_cdf(eta + s)) * h;
	}
	
	//-------------------------------------------------------
	// bl_ln2normal_vol
	//-------------------------------------------------------
	number 	bl_ln2normal_vol(	number lnvol,
								number f,
								number k) 	
	{ 
		return lnvol*pow(f*k,0.5);
	}
	//-------------------------------------------------------
	// bl_ln2normal_vol_alt
	// Hagan's variant
	//-------------------------------------------------------
	number 	bl_ln2normal_vol_alt(	number lnvol,
									number f,
									number k,
									number tau) 	
	{
		if(CORE_INT.is_equal(f, k)) 
			return k * lnvol / (1.0 + lnvol * lnvol * tau /24.0);
		return ((f-k)/log(f/k)) * lnvol / (1.0 + lnvol * lnvol * tau /24.0);
	}

	
	//-------------------------------------------------------
	// bl_ln2cir_vol
	//-------------------------------------------------------
	number 	bl_ln2cir_vol(	number lnvol,
							number f,
							number k) 	
	{ 
		return lnvol*pow(f*k,0.25);
	}
	//-------------------------------------------------------
	// bl_normal2ln_vol
	//-------------------------------------------------------
	number 	bl_normal2ln_vol(	number nvol,
								number f,
								number k) 	
	{ 
		return nvol/pow(f*k,0.5);
	}
	//-------------------------------------------------------
	// bl_normal2ln_vol
	//-------------------------------------------------------
	number 	bl_normal2ln_vol_alt(	number nvol,
									number f,
									number k,
									number tau) 	
	{ 
		number a2;

		if(CORE_INT.is_equal(f, k)) 
            a2 = - k / nvol * (24.0 / tau);
        else
          	a2 =   -((f - k) / log(f / k)) / nvol * (24.0 / tau);
		
        number a3 = 24.0 / tau;		
		vector(number) res_v = CORE_INT.solve_quadratic(a2,a3);
		sort(res_v);			
		
		if( res_v[0] >= 0 )
			return res_v[0];
		else if( res_v[1] >= 0 )
			return res_v[1];
		else
			return null<number>;
		
	}
	//-------------------------------------------------------
	// bl_cir2ln_vol
	//-------------------------------------------------------
	number 	bl_cir2ln_vol(	number nvol,
							number f,
							number k) 	
	{ 
		return nvol/pow(f*k,0.25);
	}
	
	//-------------------------------------------------------
	// bs_expiry
	//-------------------------------------------------------
	opt_exp bs_expiry(	date t,							
						date e,
						date d)
	{
		if(t < e)
			return opt_exp.OPTE_IN;
		else if(t == e)
			return opt_exp.OPTE_EXP;
		else if(t > e && t < d)
			return opt_exp.OPTE_DEL;
		else
			return opt_exp.OPTE_OUT;
	}
	
	//-------------------------------------------------------
	// bs_intrinsic
	//-------------------------------------------------------
	number bs_intrinsic(	opt_type 	type,
							number 		s, 
							number 		k, 									 
							number 		df)
	{       
	    if(type == opt_type.OPT_CALL) 
		    return max(0.0,s-k*df) ;		    
		else if (type == opt_type.OPT_PUT)	
		    return max(0.0,-s+k*df) ;
		else
			return null<number>;
	}
	//-------------------------------------------------------
	// bl_fwd_intrinsic
	//-------------------------------------------------------
	number bl_fwd_intrinsic(	opt_type 	type,
								number 		f, 
								number 		k)
	{       
		if(type == opt_type.OPT_CALL) 
		    return max(0.0,f-k) ;		    
		else if (type == opt_type.OPT_PUT)		
		    return max(0.0,-f+k) ;
		else
			return null<number>;
	}	
	//-------------------------------------------------------
	// bl_check_bounds
	//-------------------------------------------------------
	logical bl_check_bounds(opt_type 	type,
							number 		f, 
							number 		k,
							number 		fwd_prem)
	{       				
		number min_price = bl_fwd_intrinsic(type,f,k);
		if(fwd_prem < min_price) 
			return false;

		number max_price = f * (1.0 - 8.0 * CORE_INT.DBL_EPSILON);		
		if (fwd_prem > max_price)
			return false;
		
		return true;
	}
	
	//-------------------------------------------------------
	// bl_initial_sigma_mk
	//-------------------------------------------------------
	number bl_initial_sigma_mk(	number 		f, 
								number 		k,
								number 		tau_v) 
	{
		return pow(abs(log(f/k))*2.0/tau_v, 0.5);
	}

	
	integer IMPVOL_GUESS_TYPE = 1;
	
	//-------------------------------------------------------
	// bl_initial_sigma
	//-------------------------------------------------------
	number bl_initial_sigma(opt_type 	type,
							number 		f, 
							number 		k,
							number 		tau_v,
							number 		df,								
							number 		prem) 
	{
		if(IMPVOL_GUESS_TYPE == 0)
			return CORE_INT_OPT.bl_initial_sigma_mk(f,k,tau_v) ;			
		else {
			number s 	= CORE_INT_OPT.bl_s_from_f(f, df);
			prem 		-= bl_fwd_intrinsic(type,f,k)*df;
			if(prem < CORE_INT.DBL_EPSILON)
				prem = CORE_INT.DBL_EPSILON;
			number rt 	= -log(df);
			prem 		*= exp(rt / 2.0) / sqrt(k * s);			
			number absmu = abs((rt + log(s / k)) / 2.0);
			number m 	= (absmu <= 1E-20)? 1E-20 : absmu;			  		
			number tmp 	= (m * CORE_INT.PI / prem * sqrt(2)) ^ (2.0 / 3.0);
			number w 	= CORE_INT_OPT.lambert_w(tmp / CORE_INT.PI / 3.0);
			number v0 	= m / sqrt(3 * w);
			number v1 	= sqrt(m) + sqrt(CORE_INT.PI / 2.0) * exp(m) * (prem + exp(m) * (1.0-normal_cdf(2.0 * sqrt(m))) - exp(-m) / 2.0);			
			number p0 	= CORE_INT_OPT.bs_normed_prem(absmu, v0);
			number p1 	= CORE_INT_OPT.bs_normed_prem(absmu, v1);
		
			if (abs(p0 - prem ) <= abs(p1 - prem )) 
		 	 	return 2.0 * v0 / sqrt(tau_v);  
			else
		  		return 2.0 * v1 / sqrt(tau_v);
		}
	}
}


