/*	
	Equity option utility funcs 
	Developer: Algorithmica Research, Magnus Nyström			
*/

option(null: hard);

number EQ_DFLT_VEGA_MULT 		= 0.01;
number EQ_DFLT_VEGA_DIFF 		= 0.01;
logical EQ_DFLT_VEGA_CENT 		= true;

number EQ_DFLT_DELTA_MULT 		= 1;
number EQ_DFLT_DELTA_DIFF 		= 0;
logical EQ_DFLT_DELTA_CENT 		= true;

number EQ_DFLT_GAMMA_MULT 		= 1;
number EQ_DFLT_GAMMA_DIFF 		= 0;
logical EQ_DFLT_GAMMA_CENT 		= true;

number EQ_DFLT_RHO_MULT 		= 0.01;
number EQ_DFLT_RHO_DIFF 		= 0.01;
logical EQ_DFLT_RHO_CENT 		= true;

number EQ_DFLT_THETA_MULT		= 1;
number EQ_DFLT_THETA_DIFF		= 1;//days

/*number EQ_DFLT_VANNA_MULT 		= 0.01;
number EQ_DFLT_VANNA_DIFF 		= 0.01;
logical EQ_DFLT_VANNA_CENT 		= true;

number EQ_DFLT_VOLGA_MULT 		= 0.01;
number EQ_DFLT_VOLGA_DIFF 		= 0.01;
logical EQ_DFLT_VOLGA_CENT 		= true;*/

//-------------------------------------------------------
// enum eq_delta_type
//-------------------------------------------------------
class enum eq_delta_type
	option (category: 'Financial/Options/Vanilla/Equity')
{	
	SPOT 				option(str:"spot"),			
	FWD 				option(str:"fwd"),			
	SPOT_AMNT 			option(str:"spot amount"),	
	FWD_AMNT 			option(str:"fwd amount")	
};


class enum eq_gamma_type  option (category: 'Financial/Options/Vanilla/Equity')
{
	SPOT 		option(str:"spot"),
	FWD 		option(str:"fwd"),	
	SPOT_AMNT option(str:"spot amnt"),
	FWD_AMNT 	option(str:"fwd amnt")	
};

class enum eq_vega_type  option (category: 'Financial/Options/Vanilla/Equity')
{
	UNIT 		option(str:"unit"),
	AMNT 		option(str:"amnt")
};

class enum eq_theta_type  option (category: 'Financial/Options/Vanilla/Equity')
{
	CAL_DAY 		option(str:"calendar day"),
	BUS_DAY 		option(str:"business day") 
};

class enum eq_rho_type  option (category: 'Financial/Options/Vanilla/Equity')
{
	UNIT 		option(str:"unit"),
	AMNT 		option(str:"amnt")
};

/*class enum eq_vanna_type  option (category: 'Financial/Options/Vanilla/Equity')//work in progress
{
	UNIT 		option(str:"unit"),
	AMNT 		option(str:"amnt")
};*/

/*class enum eq_volga_type  option (category: 'Financial/Options/Vanilla/Equity')//work in progress
{
	UNIT 		option(str:"unit"),
	AMNT 		option(str:"amnt")
};*/
module CORE_INT_OPT
{
	//-------------------------------------------------------
	// filter_div  
	//-------------------------------------------------------
	void filter_div(	date 	trade,
						date 	settle,
						date 	exp,
						out vector(date)  	pay,
						out vector(date)  	ex,
						out vector(number) 	div)					 
	{
		if(null(pay) || null(ex) || null(div)){
			pay = ex = null<vector(date)>;
			div = null<vector(number)>;
			return;
		}
		integer size = v_size(pay);
		QL_FAIL_COND(size != v_size(ex) || size != v_size(div), "invalid vector size");

		for(integer i=0; i<size; i++){
			QL_FAIL_COND(null(pay[i]), "invalid payment date");
			QL_FAIL_COND(null(ex[i]), "invalid ex-dividend date");
			QL_FAIL_COND(ex[i]>pay[i], "invalid relationship between payment date and ex-dividend date");
			if( ex[i] <= trade || ex[i] > exp)
				div[i] = 0;
			else 
				QL_FAIL_COND(null(div[i]), "invalid dividend");			
		}		
		return;
	}

	//-------------------------------------------------------
	// filter_div  
	//-------------------------------------------------------
	void filter_div(	number 	time_trd_exp,
						out vector(number)  pay,
						out vector(number)  ex,
						out vector(number) 	div)					 
	{
		if(null(pay) || null(ex) || null(div)){
			pay = ex = div = null<vector(number)>;
			return;
		}
		integer size = v_size(pay);
		QL_FAIL_COND(size != v_size(ex) || size != v_size(div), "invalid vector size");

		for(integer i=0; i<size; i++){
			QL_FAIL_COND(null(pay[i]), "invalid payment period");
			QL_FAIL_COND(null(ex[i]), "invalid ex-dividend period");
			QL_FAIL_COND(ex[i]>pay[i], "invalid relationship between payment date and ex-dividend date");
			if( ex[i] <= 0 || ex[i] > time_trd_exp)
				div[i] = 0;
			else 
				QL_FAIL_COND(null(div[i]), "invalid dividend");			
		}		
		return;
	}
	//-------------------------------------------------------
	// convert_div  
	//-------------------------------------------------------
	void convert_div(	date 	trade,
						date 	settle,
						date 	exp,
						vector(date) option(nullable) 	pay,
						vector(date) option(nullable) 	ex,
						vector(number) 	option(nullable) div,
						out vector(number)  	pay_conv,
						out vector(number)  	ex_conv,
						out vector(number)  	div_conv)
	{
		filter_div(	trade,settle,exp,pay,ex,div);
		
		if(null(div)){
			pay_conv = ex_conv = div_conv = null<vector(number)>;
			return;
		}

		resize(pay_conv,0);
		resize(ex_conv,0);
		resize(div_conv,0);
		
		integer size = v_size(div);
		for(integer i=0; i<size; i++){
			if(!CORE_INT.is_equal_zero(div[i])) {
				push_back(pay_conv,(pay[i]-trade)/365.0);
				push_back(ex_conv,(ex[i]-trade)/365.0);
				push_back(div_conv,div[i]);
			}
		}
		return;
	}
	//-------------------------------------------------------
	// pv_div   pv as of settlement
	//-------------------------------------------------------
	number pv_div(	date 	trade,
					date 	settle,
					date 	exp,
					vector(date) option(nullable) 	pay,
					vector(date) option(nullable) 	ex,
					vector(number) option(nullable)	div,
					disc_func	df)//base=trade_date					 
	{
		if(null(pay) || null(ex) || null(div))
			return 0;

		number n = v_size(pay);		
		number pv = 0;
		for(integer i=0; i<n; i++){			
			number dfs 	= df.disc((settle-trade)/365.0);
			if( ex[i] > trade && ex[i] <= exp){
				number per 	= (pay[i]-trade)/365.0;
				number dfv 	= df.disc(per)/dfs;
				pv 			+= div[i]*dfv;
			}
		}
		return pv;
	}
	//-------------------------------------------------------
	// pv_div   pv as of settlement
	//-------------------------------------------------------
	number pv_div(	number 	time_trd_exp,
					number 	time_trd_settle,
					vector(number) option(nullable) 	pay,
					vector(number) option(nullable) 	ex,
					vector(number) option(nullable)	div,
					disc_func	df)//base=trade_date					 
	{
		if(null(pay) || null(ex) || null(div))
			return 0;

		number n = v_size(pay);		
		number pv = 0;
		for(integer i=0; i<n; i++){			
			number dfs 	= df.disc(time_trd_settle);
			if( ex[i] > 0 && ex[i] <= time_trd_exp){				
				number dfv 	= df.disc(pay[i])/dfs;
				pv 			+= div[i]*dfv;
			}
		}
		return pv;
	}
	//-------------------------------------------------------
	// pv_div   pv as of settlement
	//-------------------------------------------------------
	number pv_div(	number 	time_trd_exp,
					number 	time_trd_settle,
					vector(number) option(nullable) 	pay,
					vector(number) option(nullable) 	ex,
					vector(number) option(nullable)	div,
					number	cont_r)//base=trade_date					 
	{
		if(null(pay) || null(ex) || null(div))
			return 0;

		number n = v_size(pay);		
		number pv = 0;
		for(integer i=0; i<n; i++){			
			number dfs 	= exp(-cont_r*time_trd_settle);
			if( ex[i] > 0 && ex[i] <= time_trd_exp){				
				number dfv 	= exp(-cont_r*pay[i])/dfs;
				pv 			+= div[i]*dfv;
			}
		}
		return pv;
	}
	//-------------------------------------------------------
	// div_to_udf
	//-------------------------------------------------------
	number div_to_udf(number s,
					  number pv_div)
	{
		return (s-pv_div)/s;
	}
	//-------------------------------------------------------
	// div_to_cont_div_yld
	//-------------------------------------------------------
	number div_to_cont_div_yld(	number s,
								number pv_div,
								number tau_disc)
	{
		number udf = div_to_udf( s, pv_div);
		return -1.0/tau_disc*log(udf);
	}
}