/*	
	Option utilities fx
	Developer: Algorithmica Research, Magnus Nyström			
*/

option(null: hard);

number FX_DFLT_VEGA_MULT 		= 1;
number FX_DFLT_VEGA_DIFF 		= 0.01;
logical FX_DFLT_VEGA_CENT 		= true;

number FX_DFLT_DELTA_MULT 		= 100;
number FX_DFLT_DELTA_DIFF 		= 0;
logical FX_DFLT_DELTA_CENT 	= true;

number FX_DFLT_GAMMA_MULT 		= 100;
number FX_DFLT_GAMMA_DIFF 		= 0;
logical FX_DFLT_GAMMA_CENT 	= true;

number FX_DFLT_RHO_MULT 		= 0.01;
number FX_DFLT_RHO_DIFF 		= 0.01;
logical FX_DFLT_RHO_CENT 		= true;

number FX_DFLT_THETA_MULT		= 1;
number FX_DFLT_THETA_DIFF		= 1;//days

number FX_DFLT_VANNA_MULT 		= 0.01;
number FX_DFLT_VANNA_DIFF 		= 0.01;
logical FX_DFLT_VANNA_CENT 	= true;

number FX_DFLT_VOLGA_MULT 		= 0.01;
number FX_DFLT_VOLGA_DIFF 		= 0.01;
logical FX_DFLT_VOLGA_CENT 	= true;

//-------------------------------------------------------
// enum fx_delta_type
//-------------------------------------------------------
class enum fx_delta_type
	option (category: 'Financial/Options/Vanilla/FX')
{
	UNADJ_SPOT 			option(str:"unadj spot"),
	UNADJ_FWD 			option(str:"unadj fwd"),
	PREMADJ_SPOT 		option(str:"prem-adj spot"),
	PREMADJ_FWD 		option(str:"prem-adj fwd"),
	
	UNADJ_SPOT_REV		option(str:"unadj spot reverse"),
	UNADJ_FWD_REV 		option(str:"unadj fwd reverse"),
	PREMADJ_SPOT_REV 	option(str:"prem-adj spot reverse"),
	PREMADJ_FWD_REV 	option(str:"prem-adj fwd reverse"),

	UNADJ_SPOT_AMNT 		option(str:"unadj spot amount"),
	UNADJ_FWD_AMNT 			option(str:"unadj fwd amount"),
	PREMADJ_SPOT_AMNT 		option(str:"prem-adj spot amnt"),
	PREMADJ_FWD_AMNT 		option(str:"prem-adj fwd amnt"),
	
	UNADJ_SPOT_REV_AMNT		option(str:"unadj spot reverse amnt"),
	UNADJ_FWD_REV_AMNT 		option(str:"unadj fwd reverse amount"),
	PREMADJ_SPOT_REV_AMNT 	option(str:"prem-adj spot reverse amnt"),
	PREMADJ_FWD_REV_AMNT	option(str:"prem-adj fwd reverse amnt"),

	SPOT 				option(str:"spot"),			//mapped via prem_adj logical
	FWD 				option(str:"fwd"),			//mapped via prem_adj logical
	SPOT_AMNT 			option(str:"spot amount"),	//mapped via prem_adj logical
	FWD_AMNT 			option(str:"fwd amount")	//mapped via prem_adj logical
};

out vector(string) fx_delta_types()
	option (category: 'Financial/Options/Vanilla/FX')
{
	 vector(fx_delta_type) e;
	 expand_enum(e);
	 return string(e);
}

fx_delta_type fx_delta_type_map(string delta_type)
	option (category: 'Financial/Options/Vanilla/FX')
{
	//try {
		fx_delta_type e;
		str_to_enum(delta_type,e);	 
		return e;
	//}
	//catch {
	//	return null<fx_delta_type>;
	//}
}

fx_delta_type fx_delta_type_map(string delta_type, logical prem_adj)
	option (category: 'Financial/Options/Vanilla/FX')
{
	
	fx_delta_type e;
	str_to_enum(delta_type,e);
	if(e == fx_delta_type.SPOT)
		return prem_adj ? fx_delta_type.PREMADJ_SPOT : fx_delta_type.UNADJ_SPOT;	 
	else if(e == fx_delta_type.FWD) 
		return prem_adj ? fx_delta_type.PREMADJ_FWD : fx_delta_type.UNADJ_FWD;
	else if(e == fx_delta_type.SPOT_AMNT)
		return prem_adj ? fx_delta_type.PREMADJ_SPOT_AMNT : fx_delta_type.UNADJ_SPOT_AMNT;	 
	else if(e == fx_delta_type.FWD_AMNT) 
		return prem_adj ? fx_delta_type.PREMADJ_FWD_AMNT : fx_delta_type.UNADJ_FWD_AMNT;
	return e;
	
}

//-------------------------------------------------------
// enum fx_atm_type
//-------------------------------------------------------
class enum fx_atm_type
	option (category: 'Financial/Options/Vanilla/FX')
{
	SPOT 						option(str:"spot"),
	FWD 						option(str:"fwd"),
	NEUTRAL 					option(str:"neutral"),
	MAX_GAMMA 					option(str:"max gamma"),
	MAX_VEGA 					option(str:"max vega"),
	DELTA_NEUTRAL_UNADJ_SPOT 	option(str:"delta neutral spot"),//ex. --> zero-delta straddle strike placement when prem is paid in domestic curr
	DELTA_NEUTRAL_UNADJ_FWD 	option(str:"delta neutral fwd"),
	DELTA_NEUTRAL_PREMADJ_SPOT 	option(str:"delta neutral prem-adj spot"),//ex. --> zero-delta straddle strike placement when prem is paid in foreign curr
	DELTA_NEUTRAL_PREMADJ_FWD 	option(str:"delta neutral prem-adj fwd"),
	DELTA_NEUTRAL_STRADDLE 		option(str:"delta neutral straddle")//mapped via delta_type i.e. requires delta_type
};

out vector(string) fx_atm_types()
	option (category: 'Financial/Options/Vanilla/FX')
{
	vector(fx_atm_type) e;
	expand_enum(e);
	return string(e);
}

fx_atm_type fx_atm_type_map(string atm_type)
	option (category: 'Financial/Options/Vanilla/FX')
{
	
	fx_atm_type e;
	str_to_enum(atm_type,e);
	return e;
	
}

fx_atm_type fx_atm_type_map(string atm_type, fx_delta_type delta_type)
	option (category: 'Financial/Options/Vanilla/FX')
{
	
	fx_atm_type e;
	str_to_enum(atm_type,e);
	if(e == fx_atm_type.DELTA_NEUTRAL_STRADDLE) {
		switch(delta_type){
		case fx_delta_type.UNADJ_SPOT:			
			return fx_atm_type.DELTA_NEUTRAL_UNADJ_SPOT;	
			break;
		case fx_delta_type.UNADJ_FWD:			
			return fx_atm_type.DELTA_NEUTRAL_UNADJ_FWD;
			break;
		case fx_delta_type.PREMADJ_SPOT:
			return fx_atm_type.DELTA_NEUTRAL_PREMADJ_SPOT;
			break;
		case fx_delta_type.PREMADJ_FWD:	
			return fx_atm_type.DELTA_NEUTRAL_PREMADJ_FWD;
			break;
		default:	
			QL_FAIL("invalid delta type (no corresponding fx_atm_type)");
		}
	}
	return e;
	
}


class enum fx_gamma_type  option (category: 'Financial/Options/Vanilla/FX')
{
	UNADJ_SPOT 		option(str:"spot"),
	UNADJ_FWD 		option(str:"fwd"),	
	UNADJ_SPOT_AMNT option(str:"spot amnt"),
	UNADJ_FWD_AMNT 	option(str:"fwd amnt")	
};

class enum fx_vega_type  option (category: 'Financial/Options/Vanilla/FX')
{
	UNADJ_DOM 		option(str:"dom-ccy"),
	UNADJ_FOR 		option(str:"for-ccy"),
	UNADJ_DOM_AMNT 	option(str:"dom-ccy amnt"),
	UNADJ_FOR_AMNT 	option(str:"for-ccy amnt")	
};

class enum fx_theta_type  option (category: 'Financial/Options/Vanilla/FX')
{
	CAL_DAY 		option(str:"calendar day"),
	BUS_DAY 		option(str:"business day"),
	CAL_DAY_FOR 	option(str:"calendar day for-ccy"),
	BUS_DAY_FOR 	option(str:"business day for-ccy")
};

class enum fx_rho_type  option (category: 'Financial/Options/Vanilla/FX')
{
	UNADJ_DOM 		option(str:"dom-ccy"),
	UNADJ_FOR 		option(str:"for-ccy"),
	UNADJ_DOM_AMNT 	option(str:"dom-ccy amnt"),
	UNADJ_FOR_AMNT 	option(str:"for-ccy amnt")
};

class enum fx_vanna_type  option (category: 'Financial/Options/Vanilla/FX')//work in progress
{
	UNADJ_DOM 		option(str:"dom-ccy"),
	UNADJ_FOR 		option(str:"for-ccy"),
	UNADJ_DOM_AMNT 	option(str:"dom-ccy amnt"),
	UNADJ_FOR_AMNT 	option(str:"for-ccy amnt")
};

class enum fx_volga_type  option (category: 'Financial/Options/Vanilla/FX')//work in progress
{
	UNADJ_DOM 		option(str:"dom-ccy"),
	UNADJ_FOR 		option(str:"for-ccy"),
	UNADJ_DOM_AMNT 	option(str:"dom-ccy amnt"),
	UNADJ_FOR_AMNT 	option(str:"for-ccy amnt")
};

class enum fx_prem_type  option (category: 'Financial/Options/Vanilla/FX')
{
	DOM_UNIT_FOR option(str:"dom_unit"), DOM_PIPS_FOR option(str:"dom_pips"),
	// unit vs pips is just a scale factor of 10000 (assumes 4 decimals type of pips)
	// 1. prem in price-ccy, foreign-domestic price, price in ccy2 for unit ccy1, standard way for some vanilla pairs
	// 2. CASTAGNA: numeraire curr units
	// 3. FENICS: "Price" if face-curr = for-curr 
	
	DOM_CASH option(str:"dom_cash"),
	// 1. prem in price-ccy, = DOM_UNIT_FOR * notional in base-ccy
	// 2. CASTAGNA:
	// 3. FENICS = "Premium"
		
	DOM_UNIT_DOM option(str:"dom_pcnt_u"), DOM_PCNT_DOM option(str:"dom_pcnt"),
	// unit vs pcnt is just a scale factor of 100
	// units: prem in price-ccy units, price in ccy2 for unit ccy2, = DOM_UNIT_FOR/strike, --> pmt in price-ccy = DOM_UNIT_DOM * notional in price-ccy
	
	// 1. prem in price-ccy %, = DOM_UNIT_FOR/strike*100, standard for some exotics
	// 2. CASTAGNA: numeraire curr %
	// 3. FENICS: "PctPrice" if face-curr = dom-curr
	
	FOR_UNIT_FOR option(str:"for_pcnt_u"), FOR_PCNT_FOR option(str:"for_pcnt"),
	// unit vs pcnt is just a scale factor of 100
	// units: prem in base-ccy, price in ccy1 for unit ccy1, = DOM_UNIT_FOR/spot  [VÄXLING]
	
	// 1. prem in base-ccy %, = DOM_UNIT_FOR/spot*100, standard for some exotics (barriers) and for some vanilla pairs	[VÄXLING]
	// 2. CASTAGNA: base curr %
	// 3. FENICS: "PctPrice" if face-curr = for-curr

	FOR_CASH option(str:"for_cash"),
	// 1. prem in base-ccy, foreign-domestic price converted to foreign currency, = DOM_CASH_FOR/spot [VÄXLING]
	// 2. FENICS = "CtrPremium"
	
	FOR_UNIT_DOM option(str:"for_unit"),FOR_PIPS_DOM option(str:"for_pips"),
	// unit vs pips is just a scale factor of 10000 (assumes 4 decimals type of pips)
	// 1. prem in base-ccy, domestic-foreign price, = DOM_UNIT_FOR/(spot*strike), --> pmt in base_ccy = FOR_UNIT_DOM * notional in price-ccy
	// 2. CASTAGNA: base curr units
	// 3. FENICS: "Price" if face-curr = dom-curr

	UNIT_FACE option(str:"face_unit"),
	// either DOM_UNIT_FOR or FOR_UNIT_DOM (see above)
	// FENICS: "Price"
	//p=face==forcurr ? DOM_UNIT_FOR: FOR_UNIT_DOM
	PCNT_FACE option(str:"face_pcnt")
	//p=face==forcurr ? FOR_PCNT_FOR: DOM_PCNT_DOM
	// either DOM_PCNT_DOM or FOR_PCNT_FOR (see above)
	// FENICS: "PctPrice"

		
	/* note:
	 1. % prices have same notional and premium currency
	 2. pips prices have different notional and premium currencies
	 (pairs where premium is paid in base ccy --> left-hand-side (LHS) pairs)
	 (pairs where premium is paid in price ccy --> right-hand-side (RHS) pairs)
	 */
};



module CORE_INT_FX
{
	//-------------------------------------------------------
	// fx_currency_pair_check
	//-------------------------------------------------------
	void fx_currency_pair_check(out string currency_pair, out string for_ccy, out string dom_ccy)
	{
		QL_FAIL_COND(strlen(currency_pair) != 6,"invalid currency pair (must have 6 characters)");
		currency_pair 	= str_to_upper(currency_pair);
		for_ccy 		= sub_string(currency_pair, 0, 3);
		dom_ccy 		= sub_string(currency_pair, 3, 3);

		QL_FAIL_COND(strcmp(dom_ccy, for_ccy) == 0,"invalid currency pair (must contain two currency codes)");
	}
	//-------------------------------------------------------
	// fx_prem_ccy_default
	// based on Wystrup, FX Volatility Smile Construction, CPQF Working Paper Series No. 20
	// needs to be checked
	//-------------------------------------------------------
	string fx_prem_ccy_default(string for_ccy, string dom_ccy, out logical reg)
	{
		string prem_ccy;
		if(     strcmp(for_ccy, "USD") == 0 || strcmp(dom_ccy, "USD") == 0)
			prem_ccy = "USD";
		else if(strcmp(for_ccy, "EUR") == 0 || strcmp(dom_ccy, "EUR") == 0)
			prem_ccy = "EUR";
		else if(strcmp(for_ccy, "GBP") == 0 || strcmp(dom_ccy, "GBP") == 0)
			prem_ccy = "GBP";
		else if(strcmp(for_ccy, "AUD") == 0 || strcmp(dom_ccy, "AUD") == 0)
			prem_ccy = "AUD";
		else if(strcmp(for_ccy, "NZD") == 0 || strcmp(dom_ccy, "NZD") == 0)
			prem_ccy = "NZD";
		else if(strcmp(for_ccy, "CAD") == 0 || strcmp(dom_ccy, "CAD") == 0)
			prem_ccy = "CAD";
		else if(strcmp(for_ccy, "CHF") == 0 || strcmp(dom_ccy, "CHF") == 0)
			prem_ccy = "CHF";
		else {
			logical is_nok = strcmp(for_ccy, "NOK") == 0 || strcmp(dom_ccy, "NOK") == 0;
			logical is_sek = strcmp(for_ccy, "SEK") == 0 || strcmp(dom_ccy, "SEK") == 0;
			logical is_dkk = strcmp(for_ccy, "DKK") == 0 || strcmp(dom_ccy, "DKK") == 0;
			
			if((is_nok && is_sek) || (is_nok && is_dkk) || (is_sek && is_dkk))
				prem_ccy = dom_ccy;
			else if(is_nok)
				prem_ccy = "NOK";
			else if(is_sek)
				prem_ccy = "SEK";
			else if(is_dkk)
				prem_ccy = "DKK";
			else {
				logical is_czk = strcmp(for_ccy, "CZK") == 0 || strcmp(dom_ccy, "CZK") == 0;
				logical is_pln = strcmp(for_ccy, "PLN") == 0 || strcmp(dom_ccy, "PLN") == 0;
				logical is_try = strcmp(for_ccy, "TRY") == 0 || strcmp(dom_ccy, "TRY") == 0;
				logical is_mxn = strcmp(for_ccy, "MXN") == 0 || strcmp(dom_ccy, "MXN") == 0;

				if(    (is_czk && is_pln) || (is_czk && is_try) || (is_czk && is_mxn)
					|| (is_pln && is_try) || (is_pln && is_mxn)
					|| (is_try && is_mxn))
					prem_ccy = dom_ccy;

				else if(is_czk)
					prem_ccy = "CZK";
				else if(is_pln)
					prem_ccy = "PLN";
				else if(is_try)
					prem_ccy = "TRY";
				else if(is_mxn)
					prem_ccy = "MXN";			
				else if(strcmp(for_ccy, "JPY") == 0 || strcmp(dom_ccy, "JPY") == 0)
					prem_ccy = "JPY";			
				else
					prem_ccy = dom_ccy;
			}
		}

		reg = strcmp(dom_ccy, prem_ccy) == 0;
		return prem_ccy;
	}
	//-------------------------------------------------------
	// fx_delta_default
	// based on Wystrup, FX Volatility Smile Construction, CPQF Working Paper Series No. 20
	// needs to be checked
	//-------------------------------------------------------
	fx_delta_type fx_delta_default(string for_ccy, string dom_ccy, string prem_ccy, logical use_longterm_dflts)
	{
		logical prem_adj = strcmp(dom_ccy, prem_ccy) != 0 ;

		if(use_longterm_dflts)
			return prem_adj ? fx_delta_type.PREMADJ_FWD : fx_delta_type.UNADJ_FWD;
		else if(   strcmp(for_ccy, "ISK") == 0 || strcmp(dom_ccy, "ISK") == 0
				|| strcmp(for_ccy, "TRY") == 0 || strcmp(dom_ccy, "TRY") == 0
				|| strcmp(for_ccy, "MXN") == 0 || strcmp(dom_ccy, "MXN") == 0
				|| strcmp(for_ccy, "CZK") == 0 || strcmp(dom_ccy, "CZK") == 0
				|| strcmp(for_ccy, "KRW") == 0 || strcmp(dom_ccy, "KRW") == 0
				|| strcmp(for_ccy, "HUF") == 0 || strcmp(dom_ccy, "HUF") == 0
				|| strcmp(for_ccy, "PLN") == 0 || strcmp(dom_ccy, "PLN") == 0
				|| strcmp(for_ccy, "ZAR") == 0 || strcmp(dom_ccy, "ZAR") == 0
				|| strcmp(for_ccy, "SGD") == 0 || strcmp(dom_ccy, "SGD") == 0)		
			return prem_adj ? fx_delta_type.PREMADJ_SPOT : fx_delta_type.UNADJ_SPOT;
		else
			return prem_adj ? fx_delta_type.PREMADJ_FWD : fx_delta_type.UNADJ_FWD;

	}
	//-------------------------------------------------------
	// fx_atm_default
	// loosely based on Wystrup, FX Volatility Smile Construction, CPQF Working Paper Series No. 20
	// needs to be checked
	//-------------------------------------------------------
	fx_atm_type fx_atm_default(string for_ccy,
							   string dom_ccy,
							   string prem_ccy,
							   logical use_longterm_dflts,
							   fx_delta_type 	option(nullable) delta_type = null<fx_delta_type>)
	{
		if(!null(delta_type)){
			switch(delta_type){
			case fx_delta_type.UNADJ_SPOT:			
				return fx_atm_type.DELTA_NEUTRAL_UNADJ_SPOT;	
				break;
			case fx_delta_type.UNADJ_FWD:			
				return fx_atm_type.DELTA_NEUTRAL_UNADJ_FWD;
				break;
			case fx_delta_type.PREMADJ_SPOT:
				return fx_atm_type.DELTA_NEUTRAL_PREMADJ_SPOT;
				break;
			case fx_delta_type.PREMADJ_FWD:	
				return fx_atm_type.DELTA_NEUTRAL_PREMADJ_FWD;
				break;			
			}
		}
		
		if(!use_longterm_dflts) {
			logical prem_adj = strcmp(dom_ccy, prem_ccy) != 0 ;
			return prem_adj ? fx_atm_type.DELTA_NEUTRAL_PREMADJ_SPOT : fx_atm_type.DELTA_NEUTRAL_UNADJ_SPOT;
		}
		else
			return fx_atm_type.FWD ;
	}
}

module CORE_INT_OPT
{
	//-------------------------------------------------------
	// 'Vanna-Volga smile approximation of Proposition 4.4.3 chapter 4, Castagna
	//-------------------------------------------------------
	number vannavolga_smile_interp(	number s ,
									number tau,
									number k , 
									number k_put ,
									number v_put ,
									number k_atm ,
									number v_atm ,
									number k_call ,
									number v_call ,
									number df , 
									number dfq)

	{	
		number f 	= s * dfq / df;			
		number x1 	= (log(k_atm / k) * log(k_call / k)) / (log(k_atm / k_put) * log(k_call / k_put));
		number x2 	= (log(k / k_put) * log(k_call / k)) / (log(k_atm / k_put) * log(k_call / k_atm));
		number x3 	= (log(k / k_put) * log(k / k_atm)) / (log(k_call / k_put) * log(k_call / k_atm));
		number d1 	= (log(f / k) + 0.5 * v_atm * v_atm * tau) / (v_atm * sqrt(tau));
		number d2 	= d1 - v_atm * sqrt(tau);
		number d11 	= (log(f / k_put) + 0.5 * v_atm*v_atm * tau) / (v_atm * sqrt(tau));
		number d21 	= d11 - v_atm * sqrt(tau);
		//number d12 	= (log(f / k_atm) + 0.5 * v_atm*v_atm * tau) / (v_atm * sqrt(tau));
		//number d22 	= d12 - v_atm * sqrt(tau);
		number d13 	= (log(f / k_call) + 0.5 * v_atm*v_atm * tau) / (v_atm * sqrt(tau));
		number d23 	= d13 - v_atm * sqrt(tau);
		number dk1 	= x1 * v_put + x2 * v_atm + x3 * v_call - v_atm;
		number t1	= pow(v_put - v_atm,2.0) ;
		number t2 	= pow(v_call - v_atm,2.0) ;
		number dk2 	= x1 * d11 * d21 * t1 + x3 * d13 * d23 * t2;

		number vol;
		if ((v_atm*v_atm + d1 * d2 * (2.0 * v_atm * dk1 + dk2)) < 0 ) {

			number cont_u_r = -log(dfq)/tau;
			number cont_r 	= -log(df)/tau;
			
			if ( k > k_atm ) 				
				vol = black_scholes_imp_vol(opt_type.OPT_CALL, s, k, cont_r, cont_u_r, tau,  tau, 0.00001 * s);							
			else 				
				vol = black_scholes_imp_vol(opt_type.OPT_PUT, s, k, cont_r, cont_u_r, tau,  tau, 0.00001 * s);
		}
		else
			vol = v_atm + (-v_atm + sqrt(v_atm*v_atm + d1 * d2 * (2.0 * v_atm * dk1 + dk2))) / (d1 * d2);

		return vol;	
	}
}

module CORE_INT_FX_OPT
{
	//----------------------------------------------
	// swap_unit_q
	//----------------------------------------------
	number swap_unit_q(number q, integer scale, integer decimals)
	{
		number adj_prec = floor(log(scale)/log(10.0)+0.5)	;		
		return q / pow(10.0, decimals + adj_prec);
	}
		
	//----------------------------------------------
	// delta_type_conv
	//----------------------------------------------
	delta_type delta_type_conv( fx_delta_type 	delta_t)
	{
		switch(delta_t){
		case fx_delta_type.UNADJ_SPOT:			
			return delta_type.UNADJ_SPOT;			
		case fx_delta_type.UNADJ_FWD:			
			return delta_type.UNADJ_FWD;
		case fx_delta_type.PREMADJ_SPOT:
			return delta_type.PREMADJ_SPOT;
		case fx_delta_type.PREMADJ_FWD:	
			return delta_type.PREMADJ_FWD;
		default:	
			QL_FAIL("invalid delta type (not applicable in this function)");
		}
	}
	//----------------------------------------------
	// delta_type_conv
	//----------------------------------------------
	fx_delta_type delta_type_conv( delta_type 	delta_t)
	{
		switch(delta_t){
		case delta_type.UNADJ_SPOT:			
			return fx_delta_type.UNADJ_SPOT;			
		case delta_type.UNADJ_FWD:			
			return fx_delta_type.UNADJ_FWD;
		case delta_type.PREMADJ_SPOT:
			return fx_delta_type.PREMADJ_SPOT;
		case delta_type.PREMADJ_FWD:	
			return fx_delta_type.PREMADJ_FWD;
		default:	
			QL_FAIL("invalid delta type (not applicable in this function)");
		}
	}
	//----------------------------------------------
	// delta
	//----------------------------------------------
	void delta( fx_delta_type 	delta_t,
				 number 		s ,
				 number			k,
				 number 		v,
				 number 		tau,							 							 
				 number 		cont_r,
				 number 		cont_u_r,
				 out number 	delta_call,							 
				 out number 	delta_put)
	{
		delta_type dt = delta_type_conv(delta_t);
		black_scholes_delta( dt, s , k, v, tau, cont_r, cont_u_r, delta_call, delta_put);					
	}

	//----------------------------------------------
	// delta_c2p
	//----------------------------------------------
	number delta_c2p( 	fx_delta_type 	delta_t,
						number 			delta_call,
						number 		s ,
						number			k,
						number 		v,
						number 		tau,							 							 
						number 		cont_r,
						number 		cont_u_r)
	{
		delta_type dt = delta_type_conv(delta_t);
		return black_scholes_delta_c2p(dt,delta_call,s ,k,v,tau,cont_r,cont_u_r);
	}

	//----------------------------------------------
	// delta_p2c
	//----------------------------------------------
	number delta_p2c( 	fx_delta_type 	delta_t,
						number 			delta_put,
						number 		s ,
						number			k,
						number 		v,
						number 		tau,							 							 
						number 		cont_r,
						number 		cont_u_r)
	{
		delta_type dt = delta_type_conv(delta_t);
		return black_scholes_delta_p2c(dt,delta_put,s ,k,v,tau,cont_r,cont_u_r);
	}

	//----------------------------------------------
	// atm_strike
	//----------------------------------------------
	atm_type atm_strike_conv(	fx_atm_type atm_t,
								fx_delta_type 	option(nullable) delta_type = null<fx_delta_type>)
	{
		if(atm_t== fx_atm_type.DELTA_NEUTRAL_STRADDLE) {
			QL_REQUIRE(!null(delta_type), "delta neutral straddle requires input of delta_type");
			switch(delta_type){
			case fx_delta_type.UNADJ_SPOT:			
				return atm_type.DELTA_NEUTRAL_UNADJ_SPOT;	
			case fx_delta_type.UNADJ_FWD:			
				return atm_type.DELTA_NEUTRAL_UNADJ_FWD;
			case fx_delta_type.PREMADJ_SPOT:
				return atm_type.DELTA_NEUTRAL_PREMADJ_SPOT;
			case fx_delta_type.PREMADJ_FWD:	
				return atm_type.DELTA_NEUTRAL_PREMADJ_FWD;
			default:	
				QL_FAIL("invalid delta type (not applicable in this function)");
			}
		}

		switch(atm_t){
		case fx_atm_type.SPOT:
			return atm_type.SPOT;
		case fx_atm_type.FWD:
			return atm_type.FWD;
		case fx_atm_type.NEUTRAL://equiv to atm_fwd(); k such that call value = put value
			return atm_type.NEUTRAL;
		case fx_atm_type.DELTA_NEUTRAL_UNADJ_SPOT:
			return atm_type.DELTA_NEUTRAL_UNADJ_SPOT;
		case fx_atm_type.DELTA_NEUTRAL_UNADJ_FWD:
			return atm_type.DELTA_NEUTRAL_UNADJ_FWD;
		case fx_atm_type.MAX_GAMMA:
			return atm_type.MAX_GAMMA;
		case fx_atm_type.MAX_VEGA:
			return atm_type.MAX_VEGA;
		case fx_atm_type.DELTA_NEUTRAL_PREMADJ_SPOT:
			return atm_type.DELTA_NEUTRAL_PREMADJ_SPOT;
		case fx_atm_type.DELTA_NEUTRAL_PREMADJ_FWD:
			return atm_type.DELTA_NEUTRAL_PREMADJ_FWD;
		default:	
			QL_FAIL("atm type not supported");
		}		
	}
	
	//----------------------------------------------
	// atm_strike
	//----------------------------------------------
	number atm_strike(	 fx_atm_type 	atm_t,
						 number 		vol_atm ,
						 number 		s ,
						 number 		tau,							 							 
						 number 		df ,
						 number 		dfq,
						 fx_delta_type 	option(nullable) delta_type = null<fx_delta_type>)
	{
		atm_type at = atm_strike_conv(atm_t,delta_type );
		return black_scholes_atm_strike( at, vol_atm , s , tau,	 df , dfq);		
	}

	//----------------------------------------------
	//delta_to_k_cp
	//----------------------------------------------
	void delta_to_k_cp(fx_delta_type 	delta_t,
						number 			delta,
						number 			s,
						number 			f,
						number 			vol_c,
						number 			vol_p,
						number 			cont_r,
						number 			cont_u_r,
						number 			tau,					  
						out number 		k_dd_c,
						out number 		k_dd_p)
	{
		delta_type dt = delta_type_conv(delta_t);
		black_scholes_delta_to_k_cp(dt, delta, s, f, vol_c, vol_p, cont_r, cont_u_r, tau, k_dd_c, k_dd_p);	
	}
	/*risk_reversal
	25-delta risk reversals show the difference in volatility,
	 and therefore price, between puts and calls on the most
	 liquid out-of-the-money (OTM) options quoted on the OTC market.
	 Positive values indicate calls being more expensive than puts
	 (upside protection on the underlying forex spot is relatively
	 more expensive), while negative values indicate puts are more
	 expensive than calls (downside protection is relatively more
	 expensive).
	 ex. RR25 = 25 Delta Call – 25 Delta Put*/

	/*
	 Butterfly is the difference between the avarage volatility of the call price and put price with the
	 same moneyness level and at the money volatility level. In other words for example for 25 delta
	 level, butterfly defines how far the average volatility of 25 delta call and 25 delta put is away
	 from the at the money volatiltiy level.

	BF25 = (25 Delta Call + 25 Delta Put ) /2 – ATM
	 */
	
	//----------------------------------------------
	// fx_malz_smile_interp (should be forward delta according to "A Guide to FX Options Quoting Conventions"
	// by DIMITRI REISWICH AND UWEWYSTUP, and not prem-adjusted
	//- the formula is only valid for the forward delta neutral at-the-money quotation,
	//- the formula is only valid for risk reversal and strangle quotes associated with a delta of 0.25,
	//- the matching of the market strangle restriction (25) is guaranteed for small risk reversals only.
	//----------------------------------------------
	number fx_malz_smile_interp(number 		delta_fwd,			//delta is positive here, 0 - 100, 75 delta call = 25 delta put
								//number 		atm_delta,
								number		atm_vol,			//forward delta-neutral at-the-money quotation
								number 		rr,					//risk reversal calls-over vol
								number 		sf,					//smile-fly vol (NOT broker fly vol)
								out number 	call_vol,
								out number 	put_vol)
	{	
		return atm_vol - 2.0 * rr * (delta_fwd - 0.5) + 16.0 * sf * pow(delta_fwd - 0.5,2.0);//wystrup
		//return atm_vol + 2.0 * rr * (delta_fwd - 0.5) + 16.0 * sf * pow(delta_fwd - 0.5,2.0); Jewitt have pos sign
		
	}

	number fx_wystr_smile_interp(	number 		a,
									number 		delta,	
									number		delta_atm,	//call delta
									number 		atm_vol,
									number 		rr,				//risk reversal calls-over vol
									number 		sf,				
									out number 	call_vol,
									out number 	put_vol)
	{
		number t1 = 2.0 * (sf +rr);
		number d1 = delta-delta_atm;
		number t2 = delta*delta*rr+4*sf*delta*delta_atm+rr*delta_atm*delta_atm;
		number t3 = 2.0*(2.0*delta-a)*d1*(delta-a+delta_atm);
		number c1 = (a*a*t1 - 2.0*a*t1*(delta+delta_atm)+2.0*t2)/t3;
		number t4 = 4.0*sf*delta-a*t1+2.0*rr*delta_atm;		
		number c2 = t4/t3;		
		return atm_vol + c1 * d1 + c2 * d1*d1;
	}
	//----------------------------------------------
	// fx_rr_ss_vols (simplified and not consistent with how butterflies are quoted interbank)
	// vol smile quote from
	// 1) x-delta risk reversal and
	// 2) x-delta smile-fly (NOT broker fly == market butterfly)  
	//		[selling ATM straddle and buying symmetric delta strangle]
	// market butterfly: the 25d wing strikes are calculated from a single vol disregarding the rr price
	//----------------------------------------------
	void fx_rr_ss_vols(	number		atm_vol,
						number 		rr_vol_calls_over,	//risk reversal calls-over vol i.e. quote is callVol25d - putVol25d
						number 		sf_vol,				//smile-fly vol i.e. quote is 0.5 * (callVol25d + putVol25d) - atmVol
						out number 	call_vol,
						out number 	put_vol)
	{	
		call_vol 	= atm_vol + 0.5 * rr_vol_calls_over + sf_vol;
		put_vol 	= atm_vol - 0.5 * rr_vol_calls_over + sf_vol;
		return;
	}
	

	//----------------------------------------------
	//
	//
	//smile_quote
	//----------------------------------------------
	/*void smile_quote(	number 	atm_vol,
					   fx_atm_type 	atm_type,
					   number	rr_25_delta,
					   number	ss_25_delta,
					   fx_delta_type delta_type,
					   out number vol_25_call,
					   out number vol_25_put,
					   out number atm_k,
					   out number call_25_k,
					   out number put_25_k)
	{
		vol_25_call = atm_vol + 0.5 * rr_25_delta + ss_25_delta;
		vol_25_put 	= atm_vol - 0.5 * rr_25_delta + ss_25_delta;
		atm_k 		= atm(atm_type);
		call_25_k 	= delta_to_k(delta_type, vol_25_call);
		put_25_k 	= delta_to_k(delta_type, vol_25_put);
	}
	*/
	

	/* Castagna example:
	 s=102.65, df=0.9949, dfq=0.9835, vatm=11.95%, rr=-4.7%, vwb=0.12%
	 rr:
	 --> vol25call=9.72%, vol25put=14.42% --> k25put=94.87, k25call=106.19
	 --> C=1.056, P=1.485 --> 25d strangle=2.54 (the sum)
	vwb: volvwb=12.59%  --> k25put=95.91, k25call=107.38 --> C=1.32, P=1.25 --> sum=2.565


	 */
	/*
	void fx_rr_ss_vols(	number		atm_vol,
						number 		rr_vol_calls_over,	//risk reversal calls-over vol i.e. quote is callVol25d - putVol25d
						number 		wwb_vol,				
						out number 	call_vol,
						out number 	put_vol)
	{
		//step1
		number volvwb = atm_vol+wwb_vol;
		number katm = atm_delta_neutral();
		//step2
		number wwb_vol_i = wwb_vol;
		number call_vol,put_vol;
		fx_rr_ss_vols(	atm_vol,rr_vol_calls_over,wwb_vol_i,call_vol,put_vol);
		//step3
		number kcalli = black_scholes_delta_spot_to_k(0.25, call, s, call_vol,  cont_r, cont_u_r, tau,  tau_v);
		number kputi = black_scholes_delta_spot_to_k(0.25, put, s, put_vol,  cont_r, cont_u_r, tau,  tau_v);
		//step4
		number kvwbcalli = black_scholes_delta_spot_to_k(0.25, call, s, wwb_vol,  cont_r, cont_u_r, tau,  tau_v);
		number kvwbputi = black_scholes_delta_spot_to_k(0.25, put, s, wwb_vol,  cont_r, cont_u_r, tau,  tau_v);
		//step5
		number si = (call(kvwbcalli,volkvwbcalli)+put(kvwbputi,volkvwbputi)) -
					(call(kvwbcalli,wwb_vol)+put(kvwbputi,wwb_vol));
		
		return;
	}*/

	

	//----------------------------------------------
	//----------------------------------------------
	//----------------------------------------------
	// fx_smile_interp	WORK IN PROGRESS
	//----------------------------------------------
	//----------------------------------------------
	//----------------------------------------------
	/*class fx_smile_interp	
	{
	public:
		fx_smile_interp();
		virtual number 			interpol_strike(number ) = 0;
		virtual number 			interpol_delta(opt_type,number) = 0;
		virtual void 			symm_delta_to_vol(number ,out number,out number,out number	,out number) = 0;
		virtual void 			update_wing(logical,vector(number) , vector(number)) = 0;
		virtual number 			atm_vol( ) = 0;
		virtual	fx_delta_type 	delta_type( ) = 0;
	};
	
	fx_smile_interp.fx_smile_interp(){}*/

	//----------------------------------------------
	//----------------------------------------------
	//----------------------------------------------
	// fx_vannavolga_interp	WORK IN PROGRESS
	//----------------------------------------------
	//----------------------------------------------
	//----------------------------------------------
	/*class fx_vannavolga_interp: public fx_smile_interp
	{
	public:
		fx_vannavolga_interp( 	vector(number) ,								
								vector(number) ,
								number ,
								number,							
								number,
								number,fx_delta_type,number,number option(nullable));

		override number 		atm_vol( );
		override fx_delta_type 	delta_type( );
		override void 			update_wing(logical,vector(number) ,vector(number));
		
		override number 		interpol_strike(number );
		override number 		interpol_delta(opt_type,number);
		override void 			symm_delta_to_vol(number ,out number,out number ,
												out number,out number );

		number 			s_ ;
		number 			tau_ ;
		number 			df_ ;
		number 			dfq_;
		
		vector(number) 	strike_;
		vector(number) 	vol_;
		logical			_3point_;
		number 			cont_u_r_;
		number 			cont_r_;
		fx_delta_type  	delta_type_;
		//number			delta_atm_;
		number 			delta_l1_;
		number 			delta_l2_;
	};
	//----------------------------------------------
	//constructor
	//----------------------------------------------
	fx_vannavolga_interp.fx_vannavolga_interp( 	vector(number) 	strike,
												vector(number) 	vol,
												number s ,
												number tau,										
												number df ,
												number dfq,
												fx_delta_type delta_type,
												//number delta_atm,
												number delta1,
												number option(nullable) delta2)
												: fx_smile_interp(), s_(s), tau_(tau), df_(df), dfq_(dfq),
												  strike_(strike), vol_(vol),delta_type_(delta_type),
												  delta_l1_(delta1), delta_l2_(delta2)
	{
		integer size = v_size(vol_);
		QL_REQUIRE(size == 3 || size == 5, "invalid vector size");
		QL_REQUIRE(size == v_size(strike_), "vector sizes not equal (vol)");
		QL_REQUIRE(delta1 > 0 && delta1 < 1, "invalid level1-delta (must be >0 and <1)");
		_3point_ = v_size(vol_) == 3;
		if(!_3point_) {
			QL_REQUIRE(!null(delta2), "missing level2-delta");
			QL_REQUIRE(delta2 > 0 && delta2 < 1, "invalid level2-delta (must be >0 and <1)");
			QL_REQUIRE(delta2 < delta1, "invalid level2-delta (>= delta1)");
		}
		
		cont_u_r_ 	= -log(dfq_)/tau_;
		cont_r_ 	= -log(df_)/tau_;
	}
	//----------------------------------------------
	//delta_type
	//----------------------------------------------
	fx_delta_type fx_vannavolga_interp.delta_type()
	{
		return delta_type_ ;
	}
	//----------------------------------------------
	//atm_vol
	//----------------------------------------------
	number fx_vannavolga_interp.atm_vol()
	{
		integer i_atm = _3point_? 1 : 2;
		return vol_[i_atm] ;
	}
	//----------------------------------------------
	//update_near_wing
	//----------------------------------------------
	void fx_vannavolga_interp.update_wing(	logical near,
											vector(number) strike,
											vector(number) vol)
	{
		integer i_atm 		= _3point_? 1 : 2;
		if(near || _3point_){
			strike_[i_atm-1] 	= strike[0];		
			strike_[i_atm+1] 	= strike[1];
			vol_[i_atm-1] 		= vol[0];		
			vol_[i_atm+1] 		= vol[1];
		}
		else {			
			strike_[i_atm-2] 	= strike[0];		
			strike_[i_atm+2] 	= strike[1];
			vol_[i_atm-2] 		= vol[0];		
			vol_[i_atm+2] 		= vol[1];
		}
	}
	//----------------------------------------------
	//interpol_strike
	// (fx_vv_smile_interp only defined for 3 strikes, solution below for 5 strikes
	//  is a "pragmatic" solution to avoid non-smoothness!?!
	//----------------------------------------------
	number fx_vannavolga_interp.interpol_strike(number k)
	{
		QL_REQUIRE(k > 0 , "invalid strike");
		
		integer i_atm = _3point_? 1 : 2;

		if(_3point_ || (k >= strike_[i_atm-1] && k <= strike_[i_atm+1]))//near wing ip
			return CORE_INT_OPT.vannavolga_smile_interp(	s_ , tau_, k ,
														strike_[i_atm-1], vol_[i_atm-1] ,
														strike_[i_atm] ,  vol_[i_atm] ,
														strike_[i_atm+1], vol_[i_atm+1] ,
														df_ ,  dfq_);

		
		else if(k <= strike_[i_atm-2] || k >= strike_[i_atm+2])//far wing ip
			return CORE_INT_OPT.vannavolga_smile_interp(	s_ , tau_, k ,
														strike_[i_atm-2], vol_[i_atm-2] ,
														strike_[i_atm] ,  vol_[i_atm] ,
														strike_[i_atm+2], vol_[i_atm+2] ,
														df_ ,  dfq_);
		else {//weighted ip between far and near
			number v_near = CORE_INT_OPT.vannavolga_smile_interp(	s_ , tau_, k ,
														strike_[i_atm-1], vol_[i_atm-1] ,
														strike_[i_atm] ,  vol_[i_atm] ,
														strike_[i_atm+1], vol_[i_atm+1] ,
														df_ ,  dfq_);
			number v_far = CORE_INT_OPT.vannavolga_smile_interp(	s_ , tau_, k ,
														strike_[i_atm-2], vol_[i_atm-2] ,
														strike_[i_atm] ,  vol_[i_atm] ,
														strike_[i_atm+2], vol_[i_atm+2] ,
														df_ ,  dfq_);
			
			if(k >= strike_[i_atm-2] && k <= strike_[i_atm-1]) {
				number prop = (k-strike_[i_atm-2])/(strike_[i_atm-1]-strike_[i_atm-2]);			
				return prop * v_near + (1.0-prop)*v_far;
			}
			else {
				number prop = (k-strike_[i_atm+1])/(strike_[i_atm+2]-strike_[i_atm+1]);
				return prop * v_far + (1.0-prop)*v_near;
			}
		}
	}
	//----------------------------------------------
	//interpol_delta
	//----------------------------------------------
	number fx_vannavolga_interp.interpol_delta(opt_type type,number delta)
	{
		if(type == opt_type.OPT_PUT && delta < 0)
			delta = -delta;
		
		number 	vol_c,k_c,vol_p,k_p;
		symm_delta_to_vol(delta,vol_c,k_c,vol_p,k_p);
		return type == opt_type.OPT_PUT ? vol_p : vol_c;
	}
	//----------------------------------------------
	//symm_delta_to_vol  (delta --> k --> vol) 
	//----------------------------------------------
	void fx_vannavolga_interp.symm_delta_to_vol(number  		delta,											
												out number 		vol_c,
												out number 		k_c,
												out number		vol_p,
												out number 		k_p)
	{		
		QL_REQUIRE(delta > 0 && delta < 1, "invalid delta");
		
		integer i_atm 	= _3point_? 1 : 2;
		number f 		= s_ * dfq_ / df_;
		number d;
		if(!_3point_ && (delta <= delta_l2_ || delta >= delta_l2_)){
			vol_c 	= vol_[i_atm+2];//initial guess = 10delta call vol
			vol_p 	= vol_[i_atm-2];//initial guess = 10delta put vol
		}
		else {
			vol_c 	= vol_[i_atm+1];//initial guess = 25delta call vol
			vol_p 	= vol_[i_atm-1];//initial guess = 25delta put vol
		}
		try{
			for(integer i=0;i<5;i++){			
				delta_to_k_cp(delta_type_,delta,s_,f,vol_c,vol_p,cont_r_,cont_u_r_, tau_,k_c,k_p);			
				vol_c = interpol_strike(k_c);
				vol_p = interpol_strike(k_p);
			}
		}
		catch {
			vol_c = vol_p = null<number>; 
		}
	}*/

	
	//----------------------------------------------
	//equivalent fly procedure 4.9.1 chapter 4, Castagna, s. 118
	//broker_fly --> smile_fly
	// PROBLEM IF near_wing==false combined with vannavolga interpol (vv interpol only ip with near wing???)
	//this func only defined for rr and bfly att anchor points i.e. 25d or 10d
	//----------------------------------------------
	/*number smile_fly_equiv(	 number s ,
							 number f,
							 number tau,
							 number rr ,
							 number bfly ,//broker_fly
							 number df ,
							 number dfq ,
							 number  delta,
							 fx_smile_interp sm,//must have proper data to start with and --> near wing will be updated
							 logical near_wing,
							 out number KC,
							 out number KP,
							 out number VP,
							 out number VC)
	 
	{
		//****************
		//castagna STEP 1: calc KA (here an input) 
		//****************
		number cont_u_r = -log(dfq)/tau;
		number cont_r 	= -log(df)/tau;
		//****************
		//castagna STEP 2: calc as if bfly == smilefly, consistent vols
		//****************
		//number KA = sm.atm_strike();
		number VA = sm.atm_vol();
		//number VC, VP;
		fx_rr_ss_vols(VA,rr,bfly,VC, VP);

		//****************
		//castagna STEP 3: calc implied strikes, Formula 2.46 chapter 2
		//****************
		//strikes using consistent vols for call and put
		delta_to_k_cp(sm.delta_type(),delta,s,f,VC,VP,cont_r,cont_u_r, tau,KC,KP);
		
		//****************
		//castagna STEP 4: calculate strikes for the fly with unique vol
		//****************
		//vol for call and put in brokerfly
		number vol_bfly = VA + bfly; //market covention for the fly volatility		
				
		//strikes using bfly vol i.e. same vol for call and put
		number k_dd_bc,k_dd_bp;
		delta_to_k_cp(sm.delta_type(),delta,s,f,vol_bfly,vol_bfly,cont_r,cont_u_r,tau,k_dd_bc,k_dd_bp);		
		sm.update_wing(near_wing,[KP,KC],[VP,VC]);
		//****************
		//castagna STEP 5: 
		//****************
		//interpolate vol from strikes implied by bflyvol
		number vol_ip_bc = sm.interpol_strike(k_dd_bc);
		number vol_ip_bp = sm.interpol_strike(k_dd_bp);
		
		//calc prices 
		number fp;
		number call25c, call25b, put25p, put25b;
		//calc prices for new strike and interp vols
		black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_ip_bc,  cont_r, cont_u_r, tau,tau,  fp, call25c);		
		black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_ip_bp,  cont_r, cont_u_r, tau,tau,  fp, put25p);
		number stgl_mkt = call25c + put25p;
		
		//calc prices for new strike and bfly vol
		black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, call25b);
		black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, put25b);
		number stgl_conv = call25b + put25b;
		
		//diff between strangle at market vols and the same strangle with conventional unique vol		
		number f0 = stgl_mkt - stgl_conv;

		if (abs(f0) < 0.0000001 * s) 
			return bfly;
		

		bfly = bfly + 0.0001;
		number dfly = 0.0001;

		//while (abs(f0) > 0.00000001 * s){
		while (abs(f0) > 0.0000001 * s){
			
			fx_rr_ss_vols(VA,rr,bfly,VC, VP);
			delta_to_k_cp(sm.delta_type(),delta, s,f,VC,VP, cont_r,cont_u_r,tau, KC,KP);
			sm.update_wing(near_wing,[KP,KC],[VP,VC]);
			
			delta_to_k_cp(sm.delta_type(),delta,s,f,vol_bfly,vol_bfly,cont_r,cont_u_r,tau, k_dd_bc,k_dd_bp);

			vol_ip_bc = sm.interpol_strike(k_dd_bc);
			vol_ip_bp = sm.interpol_strike(k_dd_bp);
		
			black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_ip_bc,  cont_r, cont_u_r, tau,tau,  fp, call25c);			
			black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_ip_bp,  cont_r, cont_u_r, tau,tau,  fp, put25p);
			
			black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, call25b);	
			black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, put25b);
		
			number fu		= (call25c + put25p) - (call25b + put25b);
			number df_der 	= (fu - f0) / dfly;		
			bfly 			= bfly - fu / df_der;
			dfly 			= -fu / df_der;
			f0 				= fu;

		}
		return bfly;
	}*/

	//----------------------------------------------
	//equivalent fly (delvis baserad på procedure 4.9.1 chapter 4, Castagna, s. 118)
	//broker_fly --> smile_fly
	// Note: Castagna proceduren är bara definierad för en vinge --> proceduren är tweakad för att hantera 2 vingar
	//this func only defined for rr and bfly att anchor points i.e. 25d or 10d
	//----------------------------------------------
	/*number smile_fly_equiv_near( number s ,
								 number f,
								 number tau,
								 number rr ,
								 number bfly ,//broker_fly
								 number rr2 ,
								 number bfly2 ,//broker_fly
								 number df ,
								 number dfq ,
								 number  delta,
								 number  delta2,
								 fx_smile_interp sm,//must have proper data but near wing will be updated
								 //logical near_wing,
								 out number KC,
								 out number KP,
								 out number VP,
								 out number VC)
	 
	{
		//****************
		//castagna STEP 1: calc KA (here an input) 
		//****************
		number cont_u_r = -log(dfq)/tau;
		number cont_r 	= -log(df)/tau;
		//****************
		//castagna STEP 2: calc as if bfly == smilefly, consistent vols
		//****************

		number VA = sm.atm_vol();
		fx_rr_ss_vols(VA,rr,bfly,VC, VP);

		//****************
		//castagna STEP 3: calc implied strikes, Formula 2.46 chapter 2
		//****************
		//strikes using consistent vols for call and put
		
		delta_to_k_cp(sm.delta_type(),delta,s,f,VC,VP,cont_r,cont_u_r, tau,KC,KP);

		//****************
		//castagna STEP 4: calculate strikes for the fly with unique vol
		//****************
		//vol for call and put in brokerfly
		number vol_bfly = VA + bfly; //market covention for the fly volatility		
				
		//strikes using bfly vol i.e. same vol for call and put
		number k_dd_bc,k_dd_bp;
		delta_to_k_cp(sm.delta_type(),delta,s,f,vol_bfly,vol_bfly,cont_r,cont_u_r,tau,k_dd_bc,k_dd_bp);
		
		sm.update_wing(true,[KP,KC],[VP,VC]);
		if(k_dd_bc > KC || k_dd_bp < KP) {//extrapolation --> need to update far wing
			number KC2,KP2;
			number VC2,VP2;
			smile_fly_equiv(s ,f,tau,rr2 ,bfly2 ,df ,dfq , delta2,sm, false, KC2, KP2, VP2, VC2);//sm is updated!!
		}
		//****************
		//castagna STEP 5: 
		//****************
		//interpolate vol from strikes implied by bflyvol
		number vol_ip_bc = sm.interpol_strike(k_dd_bc);
		number vol_ip_bp = sm.interpol_strike(k_dd_bp);
		
		//calc prices 
		number fp;
		number call25c, call25b, put25p, put25b;
		//calc prices for new strike and interp vols
		black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_ip_bc,  cont_r, cont_u_r, tau,tau,  fp, call25c);		
		black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_ip_bp,  cont_r, cont_u_r, tau,tau,  fp, put25p);
		number stgl_mkt = call25c + put25p;
		
		//calc prices for new strike and bfly vol
		black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, call25b);
		black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, put25b);
		number stgl_conv = call25b + put25b;
		
		//diff between strangle at market vols and the same strangle with conventional unique vol		
		number f0 = stgl_mkt - stgl_conv;

		if (abs(f0) < 0.0000001 * s) 
			return bfly;
		

		bfly = bfly + 0.0001;
		number dfly = 0.0001;

		while (abs(f0) > 0.0000001 * s){
			
			fx_rr_ss_vols(VA,rr,bfly,VC, VP);
			delta_to_k_cp(sm.delta_type(),delta, s,f,VC,VP, cont_r,cont_u_r,tau, KC,KP);
			sm.update_wing(true,[KP,KC],[VP,VC]);

			if(k_dd_bc > KC || k_dd_bp < KP) {//extrapolation
				number KC2,KP2;
				number VC2,VP2;
				smile_fly_equiv(s ,f,tau,rr2 ,bfly2 ,df ,dfq , delta2,sm, false, KC2, KP2, VP2, VC2);
			}
	
			delta_to_k_cp(sm.delta_type(),delta,s,f,vol_bfly,vol_bfly,cont_r,cont_u_r,tau, k_dd_bc,k_dd_bp);

			vol_ip_bc = sm.interpol_strike(k_dd_bc);
			vol_ip_bp = sm.interpol_strike(k_dd_bp);
		
			black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_ip_bc,  cont_r, cont_u_r, tau,tau,  fp, call25c);			
			black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_ip_bp,  cont_r, cont_u_r, tau,tau,  fp, put25p);
			
			black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, call25b);	
			black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, put25b);
		
			number fu		= (call25c + put25p) - (call25b + put25b);
			number df_der 	= (fu - f0) / dfly;		
			bfly 			= bfly - fu / df_der;
			dfly 			= -fu / df_der;
			f0 				= fu;

		}
		return bfly;
	}*/

}



class enum fx_term_struct_calc_pref 
{
	SWAP		option(str:"swap"),	
	DOM_R 		option(str:"domestic rate"),
	FOR_R 		option(str:"foreign rate") 	
};

vector(string) fx_term_struct_calc_prefs() 
	{
		vector(fx_term_struct_calc_pref) e;
		expand_enum(e);
		return string(e);
	}
	
fx_term_struct_calc_pref fx_term_struct_calc_pref_map(string calc) 
{
	fx_term_struct_calc_pref e;
	str_to_enum(calc,e);	 
	return e;	
}

//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
// class fx_term_struct
// a simple helper class for fx when we have domestic and foreign rates, spot and forward
//----------------------------------------------
//----------------------------------------------
//----------------------------------------------

class fx_term_struct
{
public:
	fx_term_struct(fx_term_struct );
	logical			inited();
	virtual number 	r_dom(date,day_count_method dc = DC_ACT_365, rate_type rt = RT_CONT);
	virtual number 	cont_r_dom(number);
	virtual number 	r_for(date,day_count_method dc = DC_ACT_365, rate_type rt = RT_CONT);
	virtual number 	cont_r_for(number);
	virtual number 	df_dom(date);
	virtual number 	df_dom(number);
	virtual number 	df_for(date);
	virtual number 	df_for(number);
	virtual date 	ref_date() ;
	vector(date) 	exp_dates();
	virtual number 	fwd_outr(date) ;
	virtual number 	fwd_outr(number) ;
	virtual number 	swap(date) ;
	virtual number 	swap(number) ;

	virtual void set_spot( number option(nullable)) = 0;
	virtual void set_df_dom( disc_func option(nullable));
	virtual void set_df_for( disc_func option(nullable));
	virtual void set_fwd(vector(number) option(nullable));
	virtual void set_swap(vector(number) option(nullable));
	
	virtual number spot() 		= 0;
	virtual disc_func df_dom() 	= 0;
	virtual disc_func df_for() 	= 0;
	virtual	fx_term_struct_calc_pref type() = 0;
protected:
	fx_term_struct(date ,vector(date));	

	date 			ref_date_;
	vector(date) 	exp_;
	vector(number) 	tau_;
	logical			inited_;
};

fx_term_struct.fx_term_struct(	date 			ref_date,
								vector(date) 	expiry_dates)
								:ref_date_(ref_date),exp_(clone_vector(expiry_dates)),inited_(false)					
{
	integer size = v_size(exp_); 
	QL_REQUIRE(size > 0, "invalid vector size (expiry dates)");
	resize(tau_,size);
	
	for(integer i=0;i<size;i++){
		QL_REQUIRE(exp_[i] > ref_date, "invalid expiry date (< ref date)");
		tau_[i] = (exp_[i]-ref_date_)/365.0;
	}
}
	
//----------------------------------------------
// copy constructor
//----------------------------------------------
fx_term_struct.fx_term_struct(fx_term_struct o)
						:ref_date_(o.ref_date_), exp_(clone_vector(o.exp_)),tau_(clone_vector(o.tau_)), inited_(o.inited_)

{}

logical			fx_term_struct.inited() 		{ return inited_; }
date 			fx_term_struct.ref_date() 		{ return ref_date_;}
vector(date) 	fx_term_struct.exp_dates() 		{ return exp_;}

void fx_term_struct.set_fwd(vector(number) option(nullable)) 	{ QL_FAIL("invalid function call");}
void fx_term_struct.set_swap(vector(number) option(nullable)) 	{ QL_FAIL("invalid function call");}
void fx_term_struct.set_df_dom( disc_func option(nullable))		{ QL_FAIL("invalid function call");}
void fx_term_struct.set_df_for( disc_func option(nullable))		{ QL_FAIL("invalid function call");}

number fx_term_struct.r_dom(date 				expiry,
							day_count_method 	dc,
							rate_type 			rt)
{
	QL_REQUIRE(inited_, "fx term structure not initialized");
	number df = df_dom(expiry);	
	return implied_rate(1,df,ref_date_,expiry,dc,rt);
}

number fx_term_struct.cont_r_dom(number tau)
{
	QL_REQUIRE(inited_, "fx term structure not initialized");
	QL_REQUIRE(tau > 0, "invalid time to maturity");
	return df_dom().cont_rate_spot(tau);
}

number fx_term_struct.r_for(date 				expiry,
							day_count_method 	dc,
							rate_type 			rt)
{
	QL_REQUIRE(inited_, "fx term structure not initialized");
	number df = df_for(expiry);	
	return implied_rate(1,df,ref_date_,expiry,dc,rt);
}

number fx_term_struct.cont_r_for(number tau)
{
	QL_REQUIRE(inited_, "fx term structure not initialized");
	QL_REQUIRE(tau > 0, "invalid time to maturity");
	return df_for().cont_rate_spot(tau);
}


number fx_term_struct.df_dom(date expiry)
{
	QL_REQUIRE(inited_, "fx term structure not initialized");
	QL_REQUIRE(expiry > ref_date_, "invalid expiry date (< ref date)");
	number tau = (expiry-ref_date_)/365.0;
	return df_dom().disc(tau);
}

number fx_term_struct.df_dom(number tau)
{
	QL_REQUIRE(inited_, "fx term structure not initialized");
	QL_REQUIRE(tau > 0, "invalid time to maturity");
	return df_dom().disc(tau);
}
number fx_term_struct.df_for(date expiry)
{
	QL_REQUIRE(inited_, "fx term structure not initialized");
	QL_REQUIRE(expiry > ref_date_, "invalid expiry date (< ref date)");
	number tau = (expiry-ref_date_)/365.0;
	return df_for().disc(tau);
}

number fx_term_struct.df_for(number tau)
{
	QL_REQUIRE(inited_, "fx term structure not initialized");
	QL_REQUIRE(tau > 0, "invalid time to maturity");
	return df_for().disc(tau);
}

number fx_term_struct.fwd_outr(date expiry)
{
	QL_REQUIRE(inited_, "fx term structure not initialized");
	QL_REQUIRE(expiry >= ref_date_, "invalid expiry date (< ref date)");
	number tau = (expiry-ref_date_)/365.0;
	if(tau==0) return spot();
	return spot()*df_for().disc(tau)/df_dom().disc(tau);
}

number fx_term_struct.fwd_outr(number tau)
{
	QL_REQUIRE(inited_, "fx term structure not initialized");
	QL_REQUIRE(tau >= 0, "invalid time to maturity");
	if(tau==0) return spot();
	return spot()*df_for().disc(tau)/df_dom().disc(tau);
}

number fx_term_struct.swap(date expiry)
{
	return fwd_outr(expiry)-spot();
}

number fx_term_struct.swap(number tau)
{
	return fwd_outr(tau)-spot();
}

//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
// class fx_term_struct_f_solve
//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
class fx_term_struct_f_solve : public fx_term_struct
{
public:
	fx_term_struct_f_solve(fx_term_struct_f_solve );
	fx_term_struct_f_solve(date ,vector(date), number, disc_func, disc_func);
	fx_term_struct_f_solve(date ,vector(date));

	override void set_spot( number option(nullable) );
	override void set_df_dom( disc_func option(nullable));
	override void set_df_for( disc_func option(nullable));
	
	vector(number) 	fwd_outr();
	vector(number) 	swap();

	override disc_func df_dom() ;
	override disc_func df_for() ;
	override number spot();
	override fx_term_struct_calc_pref type() ;
protected:
	void 			recalc();
	
	number 			s_;
	disc_func 		df_dom_;
	disc_func 		df_for_;
	vector(number) 	f_;
};

fx_term_struct_f_solve.fx_term_struct_f_solve(date 			ref_date,
											  vector(date) 	expiry_dates,
											  number 		spot,
											  disc_func 	df_dom,
											  disc_func 	df_for)
											  :fx_term_struct(ref_date,expiry_dates),s_(spot),df_dom_(df_dom), df_for_(df_for)
{
	integer size = v_size(exp_);	
	resize(f_, size);
	
	try {
		for(integer i=0;i<size;i++) {
			f_[i] = s_*df_for_.disc(tau_[i])/df_dom_.disc(tau_[i]);
			QL_REQUIRE(!null(f_[i]), "invalid fwd calculation");			
		}
		inited_ = true;
	}
	catch {
		f_ = null<vector(number)>;
	}
}

fx_term_struct_f_solve.fx_term_struct_f_solve(date 			ref_date,
											  vector(date) 	expiry_dates)
											  :fx_term_struct(ref_date,expiry_dates),s_(null<number>),
											   df_dom_(null<disc_func>), df_for_(null<disc_func>),
											   f_(null<vector(number)>)
{}

//----------------------------------------------
// copy constructor
//----------------------------------------------
fx_term_struct_f_solve.fx_term_struct_f_solve(fx_term_struct_f_solve o)
						:fx_term_struct(o),s_(o.s_), df_dom_(o.df_dom_),df_for_(o.df_for_),f_(clone_vector(o.f_))

{}

void fx_term_struct_f_solve.recalc()
{
	if(!null(s_) && !null(df_dom_) && !null(df_for_)) {
		integer size = v_size(exp_);	
		resize(f_, size);
		try {
			for(integer i=0;i<size;i++) {
				f_[i] = s_*df_for_.disc(tau_[i])/df_dom_.disc(tau_[i]);
				QL_REQUIRE(!null(f_[i]), "invalid fwd calculation");	
			}
			inited_ = true;
		}
		catch {
			f_ = null<vector(number)>;
		}
	}
}

void fx_term_struct_f_solve.set_spot( number option(nullable) s)
{
	if(null(s)) return;
	s_ = s;
	recalc();
}
void fx_term_struct_f_solve.set_df_dom( disc_func option(nullable) df_dom)
{
	if(null(df_dom)) return;
	df_dom_ = df_dom;
	recalc();
}
void fx_term_struct_f_solve.set_df_for( disc_func option(nullable) df_for)
{
	if(null(df_for)) return;
	df_for_ = df_for;
	recalc();
}

	
fx_term_struct_calc_pref fx_term_struct_f_solve.type()  { return fx_term_struct_calc_pref.SWAP;}
disc_func fx_term_struct_f_solve.df_dom() 				{ return df_dom_;}
disc_func fx_term_struct_f_solve.df_for() 				{ return df_for_;}
vector(number) fx_term_struct_f_solve.fwd_outr()		{ QL_REQUIRE(inited_, "fx term structure not initialized"); return clone_vector(f_);}
vector(number) fx_term_struct_f_solve.swap()			{ QL_REQUIRE(inited_, "fx term structure not initialized"); return f_-s_;}
number fx_term_struct_f_solve.spot() 					{ return s_;}

//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
// class fx_term_struct_df_dom_solve
//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
class fx_term_struct_df_dom_solve : public fx_term_struct
{
public:
	fx_term_struct_df_dom_solve(fx_term_struct_df_dom_solve );
	fx_term_struct_df_dom_solve(date,vector(date),number,vector(number)	,disc_func );
	fx_term_struct_df_dom_solve(date,vector(date) );
	
	override disc_func 	df_dom() ;
	override disc_func 	df_for() ;
	override number spot();

	override void set_spot( number option(nullable) );
	override void set_fwd(vector(number) option(nullable));
	override void set_swap(vector(number) option(nullable));
	override void set_df_for( disc_func option(nullable));
	
	override fx_term_struct_calc_pref type() ;
protected:
	void 				recalc_f();
	void 				recalc_sw();
	void 				recalc();
	vector(number) 		f_;
	vector(number) 		sw_;
	disc_func 			df_dom_;
	disc_func 			df_for_;	
	number				s_;
};

//----------------------------------------------
// copy constructor
//----------------------------------------------
fx_term_struct_df_dom_solve.fx_term_struct_df_dom_solve(fx_term_struct_df_dom_solve o)
						:fx_term_struct(o),f_(clone_vector(o.f_)), sw_(clone_vector(o.sw_)), df_dom_(o.df_dom_),df_for_(o.df_for_),s_(o.s_)

{}

fx_term_struct_df_dom_solve.fx_term_struct_df_dom_solve(date 			ref_date,
														vector(date) 	expiry_dates,
														number 			spot,
														vector(number)	fwd,											  
														disc_func 		df_for)
														:fx_term_struct(ref_date,expiry_dates),f_(clone_vector(fwd)),
														 sw_(null<vector(number)>),df_for_(df_for), s_(spot)
{
	integer size = v_size(exp_);
	QL_REQUIRE(size == v_size(fwd), "invalid vector size (fwd)");
	
	vector(number) df_vals[size];

	try {
		for(integer i=0;i<size;i++) {
			QL_REQUIRE(!null(f_[i]), "invalid fwd");
			df_vals[i] = s_/f_[i]*df_for_.disc(tau_[i]);
			QL_REQUIRE(!null(df_vals[i]), "invalid fwd calculation");			
		}
		df_dom_ = disc_func_interp(tau_,df_vals,ip_linear(),RT_CONT);
		inited_ = !null(df_dom_);
	}
	catch {
		df_dom_ = null<disc_func>;
	}
}

fx_term_struct_df_dom_solve.fx_term_struct_df_dom_solve(date 			ref_date,
														vector(date) 	expiry_dates)
														:fx_term_struct(ref_date,expiry_dates),f_(null<vector(number)>),
														 sw_(null<vector(number)>),df_for_(null<disc_func>), s_(null<number>)
{}

void fx_term_struct_df_dom_solve.set_fwd( vector(number) option(nullable)  fwd)
{
	if(null(fwd)) return;
	integer size = v_size(exp_);
	QL_REQUIRE(size == v_size(fwd), "invalid vector size (fwd)");
	f_ = clone_vector(fwd);
	sw_ = null<vector(number)>;
	recalc_f();
}

void fx_term_struct_df_dom_solve.set_swap( vector(number) option(nullable)  swap)
{
	if(null(swap)) return;
	integer size = v_size(exp_);
	QL_REQUIRE(size == v_size(swap), "invalid vector size (swaps)");
	sw_ = clone_vector(swap);
	f_ = null<vector(number)>;
	recalc_sw();
}

void fx_term_struct_df_dom_solve.set_spot( number option(nullable)  s)
{
	if(null(s)) return;
	s_ = s;
	recalc();
}

void fx_term_struct_df_dom_solve.set_df_for( disc_func option(nullable)  df_for)
{
	if(null(df_for)) return;
	df_for_ = df_for;
	recalc();
}

void fx_term_struct_df_dom_solve.recalc()
{
	if(null(f_))
		recalc_sw();
	else
		recalc_f();
}


void fx_term_struct_df_dom_solve.recalc_f()
{
	if(!null(s_) && !null(f_) && !null(df_for_)) {
		integer size = v_size(exp_);	
			
		vector(number) df_vals[size];
		try{
			for(integer i=0;i<size;i++) {
				QL_REQUIRE(!null(f_[i]), "invalid fwd");
				df_vals[i] = s_/f_[i]*df_for_.disc(tau_[i]);
				QL_REQUIRE(!null(df_vals[i]), "invalid fwd calculation");
			}
			df_dom_ = disc_func_interp(tau_,df_vals,ip_linear(),RT_CONT);			
			inited_ = !null(df_dom_);
		}
		catch {
			df_dom_ = null<disc_func>;
		}
	}
	if(!null(s_) && !null(f_))
	   sw_ = f_ - s_;
}

void fx_term_struct_df_dom_solve.recalc_sw()
{
	if(!null(s_) && !null(sw_))
	   f_ = s_ + sw_;

	recalc_f();		
}


fx_term_struct_calc_pref fx_term_struct_df_dom_solve.type() { return fx_term_struct_calc_pref.DOM_R;}
disc_func fx_term_struct_df_dom_solve.df_dom() 				{ return df_dom_;}
disc_func fx_term_struct_df_dom_solve.df_for() 				{ return df_for_;}
number fx_term_struct_df_dom_solve.spot() 					{ return s_;}

//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
// class fx_term_struct_df_for_solve
//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
class fx_term_struct_df_for_solve : public fx_term_struct
{
public:
	fx_term_struct_df_for_solve(fx_term_struct_df_for_solve );
	fx_term_struct_df_for_solve(date,vector(date),number,vector(number)	,disc_func );
	fx_term_struct_df_for_solve(date,vector(date) );
	
	override disc_func df_dom() ;
	override disc_func df_for() ;
	override number spot();
	
	override void set_spot( number option(nullable) );
	override void set_fwd(vector(number) option(nullable) );
	override void set_swap(vector(number) option(nullable) );
	override void set_df_dom( disc_func option(nullable) );
	override fx_term_struct_calc_pref type() ;
protected:
	void 				recalc_f();
	void 				recalc_sw();
	void 				recalc();
	vector(number) 		f_;
	vector(number) 		sw_;
	disc_func 			df_dom_;
	disc_func 			df_for_;	
	number				s_;
};

//----------------------------------------------
// copy constructor
//----------------------------------------------
fx_term_struct_df_for_solve.fx_term_struct_df_for_solve(fx_term_struct_df_for_solve o)
						:fx_term_struct(o),f_(clone_vector(o.f_)), sw_(clone_vector(o.sw_)), df_dom_(o.df_dom_),df_for_(o.df_for_),s_(o.s_)

{}

fx_term_struct_df_for_solve.fx_term_struct_df_for_solve(date 			ref_date,
														vector(date) 	expiry_dates,
														number 			spot,
														vector(number)	fwd,											  
														disc_func 		df_dom)
														:fx_term_struct(ref_date,expiry_dates),f_(clone_vector(fwd)),
														 sw_(null<vector(number)>),df_dom_(df_dom), s_(spot)
{
	integer size = v_size(exp_);
	QL_REQUIRE(size == v_size(fwd), "invalid vector size (fwd)");
	vector(number) df_vals[size];
	
	try {
		for(integer i=0;i<size;i++) {
			QL_REQUIRE(!null(f_[i]), "invalid fwd");
			df_vals[i] = f_[i]/s_*df_dom_.disc(tau_[i]);
			QL_REQUIRE(!null(df_vals[i]), "invalid fwd calculation");			
		}
		df_for_ = disc_func_interp(tau_,df_vals,ip_linear(),RT_CONT);
		inited_ = !null(df_for_);
	}
	catch {
		df_for_ = null<disc_func>;
	}
}

fx_term_struct_df_for_solve.fx_term_struct_df_for_solve(date 			ref_date,
														vector(date) 	expiry_dates )
														:fx_term_struct(ref_date,expiry_dates),f_(null<vector(number)>),
														 sw_(null<vector(number)>),df_dom_(null<disc_func>), s_(null<number>)
{}

void fx_term_struct_df_for_solve.set_fwd( vector(number) option(nullable) fwd)
{
	if(null(fwd)) return;
	integer size = v_size(exp_);
	QL_REQUIRE(size == v_size(fwd), "invalid vector size (fwd)");
	f_ = clone_vector(fwd);
	sw_ = null<vector(number)>;
	recalc_f();
}

void fx_term_struct_df_for_solve.set_swap( vector(number) option(nullable) swap)
{
	if(null(swap)) return;
	integer size = v_size(exp_);
	QL_REQUIRE(size == v_size(swap), "invalid vector size (swap)");
	sw_ = clone_vector(swap);
	f_ = null<vector(number)>;
	recalc_sw();
}

void fx_term_struct_df_for_solve.set_spot( number option(nullable)  s)
{
	if(null(s)) return;
	s_ = s;
	recalc();
}

void fx_term_struct_df_for_solve.set_df_dom( disc_func option(nullable) df_dom)
{
	if(null(df_dom)) return;
	df_dom_ = df_dom;
	recalc();
}

void fx_term_struct_df_for_solve.recalc_f()
{
	if(!null(s_) && !null(f_) && !null(df_dom_)) {
		integer size = v_size(exp_);	
			
		vector(number) df_vals[size];
		try {
			for(integer i=0;i<size;i++) {
				QL_REQUIRE(!null(f_[i]), "invalid fwd");
				df_vals[i] = f_[i]/s_*df_dom_.disc(tau_[i]);
				QL_REQUIRE(!null(df_vals[i]), "invalid fwd calculation");
			}
			df_for_ = disc_func_interp(tau_,df_vals,ip_linear(),RT_CONT);
			inited_ = !null(df_for_);
		}
		catch {
			df_for_ = null<disc_func>;
		}
	}
	if(!null(s_) && !null(f_))
	   sw_ = f_ - s_;
}

void fx_term_struct_df_for_solve.recalc_sw()
{
	if(!null(s_) && !null(sw_))
	   f_ = s_ + sw_;

	recalc_f();		
}

void fx_term_struct_df_for_solve.recalc()
{
	if(null(f_))
		recalc_sw();
	else
		recalc_f();
}

fx_term_struct_calc_pref fx_term_struct_df_for_solve.type() { return fx_term_struct_calc_pref.FOR_R;}
disc_func fx_term_struct_df_for_solve.df_dom() 				{ return df_dom_;}
disc_func fx_term_struct_df_for_solve.df_for() 				{ return df_for_;}
number fx_term_struct_df_for_solve.spot() 					{ return s_;}



fx_term_struct create_fx_term_struct(	date 			ref_date,
										vector(date) 	expiry_dates,
										fx_term_struct_calc_pref type)
{
	if(type == fx_term_struct_calc_pref.SWAP)
		return new fx_term_struct_f_solve(ref_date,expiry_dates);
	else if(type == fx_term_struct_calc_pref.DOM_R)
		return new fx_term_struct_df_dom_solve(ref_date,expiry_dates);
	else 
		return new fx_term_struct_df_for_solve(ref_date,expiry_dates);
}