/*	
	Option vanilla class
	Developer: Algorithmica Research, Magnus Nyström			
*/

option(null: hard);

/*				option parameter
underlying		spot price	vol				passage_of_time		div_yield
----------		----------	-----			---------------		---------
value			delta		vega			theta				epsilon(psi)
delta			gamma		vanna			charm
vega			vanna		vomma(volga)	veta
theta			charm		veta
rho							vera(rhova)
gamma			speed		zomma			color
vomma(volga)				ultima
https://www.daytrading.com/greeks-options-pricing
 */
//**********************************************
// class  vanilla_option
//**********************************************
class vanilla_option 
{
public: 

	//vanilla_option();
	vanilla_option(	exercise_type ,opt_type,number option(nullable) );//static
	vanilla_option(	string ,string,number option(nullable) );//static
	
	void			set_model(option_model_e);
	void			set_model(option_model_e, number option(nullable));
	void			set_model(option_model_e, integer );
	
	/*void			set_args(number  option(nullable),number  option(nullable),
							 number ,number,number, number option(nullable),number,number);
	
	void			set_args(number,number,number,number,
							vector(number) option(nullable) ,vector(number) option(nullable) ,
							vector(number) option(nullable),number,number);*/

	//set_mkt_args is updating the instrument with market data, static data must be set
	void			set_mkt_args(number  option(nullable),number  option(nullable),number option(nullable),
								number,option_qs_e,number, number option(nullable),number,number);

	void			set_mkt_args(number,number option(nullable), number,option_qs_e,number,
								vector(number) option(nullable) ,vector(number) option(nullable) ,
								vector(number) option(nullable),number,number);
	
	virtual void	prem(out number fp,out number sp) ;  
	virtual number 	delta_spot(number mult = 1,number d = 0, logical centered = true ) ;
	virtual number 	delta_fwd(number mult = 1,number d = 0, logical centered = true ) 	;
	virtual number 	vega(number mult = 1,number d = 0, logical centered = true) 	;
	virtual number 	gamma_spot(number mult = 1,number d = 0, logical centered = true) 	;
	virtual number 	gamma_fwd(number mult = 1,number d = 0, logical centered = true) 	;
	virtual number 	rho(number mult = 1,number d = 0, logical centered = true ) ;
	virtual number 	psi(number mult = 1,number d = 0, logical centered = true ) ;
	virtual number 	theta(number,number, out number  ) ;
	virtual number 	vanna(	number mult = 1,number d = 0, logical centered = true);
	virtual number 	volga(	number mult = 1,number d = 0, logical centered = true);
	
	virtual number 	imp_vol(number p);

	vanilla_option_model		pmodel();
	vanilla_option_model_args 	args();
	exercise_type 				exercise_type();
	opt_type 					type();
	number						strike();
	logical						static();
protected:

	void			set_fwd_price(number);
	logical			is_fwd_model();
	
private:
	
//protected:
	vanilla_option_model		pmodel_;
	vanilla_option_model_args 	args_;

	exercise_type 	exercise_type_;
	opt_type 		type_;
	number			strike_;
	logical			static_;
};

vanilla_option_model		vanilla_option.pmodel() { return pmodel_;}
vanilla_option_model_args 	vanilla_option.args() 	{ return args_;}
exercise_type 				vanilla_option.exercise_type() { return exercise_type_;}
opt_type 					vanilla_option.type() 	{ return type_;}
number						vanilla_option.strike() { return strike_;}
logical						vanilla_option.static() { return static_;}
	
//----------------------------------------------
// vanilla_option constructor
//----------------------------------------------
/*vanilla_option.vanilla_option() : pmodel_(null<vanilla_option_model>),
								  args_(null<vanilla_option_model_args>),
								  static_(true)
{}*/
//----------------------------------------------
// vanilla_option constructor
//----------------------------------------------
vanilla_option.vanilla_option(	..exercise_type 			exercise_type,
								..opt_type 				type,	
								number	option(nullable) strike)
								: pmodel_(null<vanilla_option_model>),
								  args_(null<vanilla_option_model_args>),
								  exercise_type_(exercise_type),
								  type_(type),	
								  strike_(strike),
								  static_(true)
{}
//----------------------------------------------
// vanilla_option constructor
//----------------------------------------------
vanilla_option.vanilla_option(	string					exercise_type,
								string					type,	
								number	option(nullable) strike)
								: pmodel_(null<vanilla_option_model>),
								  args_(null<vanilla_option_model_args>),								  
								  strike_(strike),
								  static_(true)
{
	type_			= opt_type_map(type);
	QL_FAIL_COND(null(type_),"invalid option type");
	exercise_type_ 	= exercise_type_map(exercise_type);
	QL_FAIL_COND(null(exercise_type_),"invalid exercise type");
}
//----------------------------------------------
// vanilla_option set_model (BS, BN)
//----------------------------------------------
void vanilla_option.set_model(option_model_e m)
{
	pmodel_ = CORE_INT.get_vanilla_model(m);
}
//----------------------------------------------
// vanilla_option set_model (B76)
//----------------------------------------------
void vanilla_option.set_model(	option_model_e 	m,
								number  option(nullable) d  )
{
	if(null(d)) d = 0;
	pmodel_ = CORE_INT.get_vanilla_model(m, d);
}
//----------------------------------------------
// vanilla_option set_model (CRR)
//----------------------------------------------
void vanilla_option.set_model(	option_model_e 	m,
								integer 				steps )
{
	pmodel_ = CORE_INT.get_vanilla_model(m, null<number>, steps);
}
	
//----------------------------------------------
// vanilla_option.set_args
// set_model must be executed first
//----------------------------------------------
/*void vanilla_option.set_args(	//exercise_type 	exercise_type,
								//opt_type 		type,
								number 	option(nullable) spot_price,
								number 	option(nullable) fwd_price,
								number 			strike,
								number 			vol, 
								number 			rate_cont,
								number option(nullable) rate_u_cont,
								number 			ttm,							
								number 			ttm_vol)
{
	QL_FAIL_COND(null(pmodel_), "model not set");
	if( pmodel_.get_model() == option_model_e.B76 || pmodel_.get_model() == option_model_e.BNORM)	
		QL_FAIL_COND(null(fwd_price), "invalid fwd price (required for current model)");
	else if(pmodel_.get_model() == option_model_e.BS)
		QL_FAIL_COND(null(spot_price), "invalid spot price (required for current model)");
	else
		QL_FAIL("invalid model for current function");
	
	QL_FAIL_COND(!pmodel_.check_asset(1), "invalid model (number of underlyings)");
	QL_FAIL_COND(!pmodel_.check_exercise(exercise_type), "invalid model (exercise type)");
	args_ = dynamic_cast<vanilla_option_model_args>(pmodel_.get_args());
	QL_FAIL_COND(null(args_), "invalid model");
	args_.set_args(exercise_type, type,fwd_price,spot_price,strike,vol,rate_cont,rate_u_cont,ttm,ttm_vol);

	exercise_type_ 	= exercise_type;
	type_ 			= type;	
	strike_ 		= strike;	
	static_ 		= false;
}
*/
//----------------------------------------------
// vanilla_option.set_mkt_args
// set_model must be executed first
//----------------------------------------------
void vanilla_option.set_mkt_args(number option(nullable) spot_price,
								number 	option(nullable) fwd_price,
								number option(nullable) strike,
								number 			vol_prem,
								option_qs_e		qs,
								number 			rate_cont,
								number option(nullable) rate_u_cont,
								number 			ttm,							
								number 			ttm_vol)
{
	QL_FAIL_COND(null(exercise_type_),"static data not set");
	if(!null(strike)) {
		strike_ = strike;
	}
	else
		QL_FAIL_COND(null(strike_),"strike not set");
	
	QL_FAIL_COND(null(pmodel_), "model not set");
	if( pmodel_.get_model() == option_model_e.B76 || pmodel_.get_model() == option_model_e.BNORM)	
		QL_FAIL_COND(null(fwd_price), "invalid fwd price (required for current model)");
	else if(pmodel_.get_model() == option_model_e.BS)
		QL_FAIL_COND(null(spot_price), "invalid spot price (required for current model)");
	else
		QL_FAIL("invalid model for current function");
	
	QL_FAIL_COND(!pmodel_.check_asset(1), "invalid model (number of underlyings)");
	QL_FAIL_COND(!pmodel_.check_exercise(exercise_type_), "invalid model (exercise type)");
	args_ = dynamic_cast<vanilla_option_model_args>(pmodel_.get_args());
	QL_FAIL_COND(null(args_), "invalid model");
	args_.set_args(exercise_type_, type_,fwd_price,spot_price,strike_,vol_prem,qs,rate_cont,rate_u_cont,ttm,ttm_vol);
	
	static_ = false;
}

//----------------------------------------------
// vanilla_option.set_args  FIX_ME should support BS and B76
//----------------------------------------------
/*void vanilla_option.set_args(	exercise_type 	exercise_type,
								opt_type 		type,
								number 			spot_price,								
								number 			strike,
								number 			vol, 
								number 			rate_cont,
								vector(number) option(nullable) pay,
								vector(number) option(nullable) ex,
								vector(number) option(nullable) div,
								number 			ttm,							
								number 			ttm_vol)
{
	QL_FAIL_COND(null(pmodel_), "model not set");
	logical ok = pmodel_.get_model() == option_model_e.CRR;
	QL_FAIL_COND(!ok, "function not applicable for current model");
	
	QL_FAIL_COND(!pmodel_.check_asset(1), "invalid model (number of underlyings)");
	QL_FAIL_COND(!pmodel_.check_exercise(exercise_type), "invalid model (exercise type)");
	args_ = dynamic_cast<vanilla_option_model_args>(pmodel_.get_args());
	QL_FAIL_COND(null(args_), "invalid model");
	args_.set_args(exercise_type, type,null<number>,spot_price,strike,vol,rate_cont,null<number>,ttm,ttm_vol,pay,ex,div);

	exercise_type_ 	= exercise_type;
	type_ 			= type;	
	strike_ 		= strike;
	static_ 		= false;
}*/

//----------------------------------------------
// vanilla_option.set_mkt_args  FIX_ME should support BS and B76
//----------------------------------------------
void vanilla_option.set_mkt_args(number 		spot_price,
								 number option(nullable) strike,
								number 			vol_prem,
								 option_qs_e	qs,
								number 			rate_cont,
								vector(number) option(nullable) pay,
								vector(number) option(nullable) ex,
								vector(number) option(nullable) div,
								number 			ttm,							
								number 			ttm_vol)
{
	QL_FAIL_COND(null(exercise_type_),"static data not set");
	if(!null(strike)) {
		strike_ = strike;
	}
	else
		QL_FAIL_COND(null(strike_),"strike not set");
	QL_FAIL_COND(null(pmodel_), "model not set");
	logical ok = pmodel_.get_model() == option_model_e.CRR;
	QL_FAIL_COND(!ok, "function not applicable for current model");
	
	QL_FAIL_COND(!pmodel_.check_asset(1), "invalid model (number of underlyings)");
	QL_FAIL_COND(!pmodel_.check_exercise(exercise_type_), "invalid model (exercise type)");
	args_ = dynamic_cast<vanilla_option_model_args>(pmodel_.get_args());
	QL_FAIL_COND(null(args_), "invalid model");
	args_.set_args(exercise_type_, type_,null<number>,spot_price,strike_,vol_prem,qs,rate_cont,null<number>,ttm,ttm_vol,pay,ex,div);

	static_ = false;
}
//----------------------------------------------
// vanilla_option.is_fwd_model
//----------------------------------------------
logical vanilla_option.is_fwd_model()
{
	QL_FAIL_COND(null(pmodel_),"pricing model not set");
	return pmodel_.get_model() == option_model_e.B76 || pmodel_.get_model() == option_model_e.BNORM;
}
//----------------------------------------------
// vanilla_option.set_fwd_price
//----------------------------------------------
void vanilla_option.set_fwd_price(number f)
{
	args_.f_ = f;
}
//----------------------------------------------
// vanilla_option.prem
//----------------------------------------------
void vanilla_option.prem(out number fp,out number sp)
{
	QL_FAIL_COND(static_,"function not applicable for static instrument");
	pmodel_.prem(fp,  sp);
}

//----------------------------------------------
// vanilla_option.delta_spot
//----------------------------------------------
number vanilla_option.delta_spot(number mult,
								 number d,								
								 logical centered)
{
	QL_FAIL_COND(static_,"function not applicable for static instrument");
	return pmodel_.delta_spot(mult,d,centered);
}

//----------------------------------------------
// vanilla_option.delta_fwd
//----------------------------------------------
number vanilla_option.delta_fwd(number mult,
								number d,	
								logical centered)
{
	QL_FAIL_COND(static_,"function not applicable for static instrument");
	return pmodel_.delta_fwd(mult,d,centered);
}
//----------------------------------------------
// vanilla_option.vega
//----------------------------------------------
number vanilla_option.vega(	number mult,
							number d,	
							logical centered)
{
	QL_FAIL_COND(static_,"function not applicable for static instrument");
	return pmodel_.vega(mult,d,centered);
}
//----------------------------------------------
// vanilla_option.vanna
//----------------------------------------------
number vanilla_option.vanna(number mult,
							number d,	
							logical centered)
{
	QL_FAIL_COND(static_,"function not applicable for static instrument");
	return pmodel_.vanna( mult,d,centered);
}
//----------------------------------------------
// vanilla_option.volga
//----------------------------------------------
number vanilla_option.volga(number mult,
							number d,	
							logical centered)
{
	QL_FAIL_COND(static_,"function not applicable for static instrument");
	return pmodel_.volga( mult,d,centered);
}
//----------------------------------------------
// vanilla_option.gamma_spot
//----------------------------------------------
number vanilla_option.gamma_spot(	number mult,
									number d,	
									logical centered)
{
	QL_FAIL_COND(static_,"function not applicable for static instrument");
	return pmodel_.gamma_spot(mult,d,centered);
}
//----------------------------------------------
// vanilla_option.gamma_fwd
//----------------------------------------------
number vanilla_option.gamma_fwd(	number mult,
									number d,	
									logical centered)
{
	QL_FAIL_COND(static_,"function not applicable for static instrument");
	return pmodel_.gamma_fwd(mult,d,centered);
}
//----------------------------------------------
// vanilla_option.rho
//----------------------------------------------
number vanilla_option.rho(	number mult,
							number d,								
							logical centered)
{
	QL_FAIL_COND(static_,"function not applicable for static instrument");
	return pmodel_.rho( mult, d,centered);
}
//----------------------------------------------
// vanilla_option.theta
//----------------------------------------------
number vanilla_option.theta(number mult,
							number d,
							out number diff)
{
	QL_FAIL_COND(static_,"function not applicable for static instrument");
	return pmodel_.theta( mult, d, diff);
}
//----------------------------------------------
// vanilla_option.psi
//----------------------------------------------
number vanilla_option.psi(	number mult,
							number d,								
							logical centered)
{
	QL_FAIL_COND(static_,"function not applicable for static instrument");
	return pmodel_.psi( mult, d,centered);
}

//----------------------------------------------
// vanilla_option.imp_vol
//----------------------------------------------
number vanilla_option.imp_vol(number p)
{
	QL_FAIL_COND(static_,"function not applicable for static instrument");
	return pmodel_.imp_vol(p);
}

