/*	
	Option vanilla equity class
	Developer: Algorithmica Research, Magnus Nyström			
*/

option(null: hard);


//**********************************************
// class  vanilla_option_eq
//**********************************************
class vanilla_option_eq : public vanilla_option //option (category: 'Financial/Options/Vanilla Equity')
{
public: 

	vanilla_option_eq(..exercise_type,..opt_type,number	option(nullable), number);

	override void	prem(out number fp,out number sp) ;
	
	void			set_mkt_args(number option(nullable),number option(nullable),
								number option(nullable),number ,option_qs_e,number, number option(nullable),number,number);
	
	void			set_mkt_args(number,number option(nullable),number,option_qs_e,number,number ,number,
								vector(number) option(nullable) ,vector(number) option(nullable) ,
								vector(number) option(nullable));
	
protected:	
	number		ul_quantity_;
};

//----------------------------------------------
// vanilla_option_eq constructor
//----------------------------------------------
vanilla_option_eq.vanilla_option_eq(..exercise_type 	exercise_type,
									..opt_type 		type,	
									number	option(nullable) strike,
									number			ul_quantity)
									:vanilla_option(exercise_type,type,strike), ul_quantity_(ul_quantity){}

//----------------------------------------------
// set_mkt_args
//----------------------------------------------
void vanilla_option_eq.set_mkt_args(number 			spot_price,
										number 	option(nullable) strike,
										number 			vol_prem,
										option_qs_e		qs,
										number 			rate_cont,								
										number 			ttm,
										number 			ttm_vol,
										vector(number) option(nullable) pay,
										vector(number) option(nullable) ex,
										vector(number) option(nullable) div )
{
	
	vanilla_option.set_mkt_args(spot_price,strike,vol_prem,qs,rate_cont,pay,ex,div,ttm,ttm_vol);
}

//----------------------------------------------
// set_mkt_args
//----------------------------------------------
void vanilla_option_eq.set_mkt_args(number 	option(nullable) spot_price,
									number 	option(nullable) fwd_price,
									number 	option(nullable) strike,
									number 			vol_prem,
									option_qs_e		qs,
									number 			rate_cont,
									number option(nullable) div_yld_cont,
									number 			ttm,
									number 			ttm_vol )
{
	if(is_fwd_model() && null(fwd_price)){
		QL_FAIL_COND(null(spot_price), "missing price of underlying");
		QL_FAIL_COND(null(div_yld_cont), "dividend yield required for current operation");
		fwd_price = CORE_INT_OPT.bl_f_from_s(spot_price, rate_cont, ttm,div_yld_cont);
	}
	else if(null(spot_price)) {
		QL_FAIL_COND(null(fwd_price), "missing price of underlying");
		QL_FAIL_COND(null(div_yld_cont), "dividend yield required for current operation");
		spot_price = CORE_INT_OPT.bl_s_from_f(fwd_price, rate_cont, ttm,div_yld_cont);
	}
	
	vanilla_option.set_mkt_args(spot_price,fwd_price,strike,vol_prem,qs,rate_cont,div_yld_cont,ttm,ttm_vol);
}

//----------------------------------------------
// prem
//----------------------------------------------
void vanilla_option_eq.prem(out number fp,out number sp)
{
	vanilla_option.prem(fp,  sp);
	//fp *= ul_quantity_;
	//sp *= ul_quantity_;
}
