/*	
	Option vanilla fx class
	Developer: Algorithmica Research, Magnus Nyström			
*/
option(null: hard);


//**********************************************
// class  vanilla_option_fx
//**********************************************
/*
 FOR-DOM, represents the number of units of domestic currency needed to buy one
unit of foreign currency at time t. For example, EURUSD = 1.3900 means that one EUR is worth 1.3900
USD. In this case, EUR is the foreign currency and USD is the domestic one. “Domestic” does not
refer to any geographical region, but merely represents a particular side of the deal.
The domestic currency is also referred to as the numeraire currency (price currency) and the foreign one
as the base currency
example:
EUR-USD Spot = 1.39 (EUR=foreign, USD = domestic)
Strike = 1.35
Nominal = EUR 1000000
Price = USD 0.1024
--> Total Price = 0.1024 * 1000000 = USD 102400 (= EUR 73669) 
call buyer: receives EUR = 1 000 000, pays USD = 1 350 000
 */
/*
volatility smile is quoted using
1) ATM volatility, and
2) 25-Delta Risk Reversal and
3) 25-Delta Strangle.

the ATM strike depends on the delta conventions
Risk Reversal: diff the volatility of a call option with a delta of 0.25 and the volatility of a put option with delta of -0.25,
again the delta conventions vary.
The strangle(fly) refers to the market strangle (assume a smile strangle) which is the difference between the average
of the 25-delta call and 25-delta put volatilities and the ATM volatility.
 */

class vanilla_option_fx : public vanilla_option 
{
public: 

	vanilla_option_fx(vanilla_option_fx_def	,date 	option(nullable) ,								
						number 	option(nullable) ,number); //will be a static instrument

	void			set_mkt_args(date,number, number option(nullable),number,fx_delta_type,number,
								 number, number option(nullable));

	void			set_mkt_args(date,number, number option(nullable),fx_atm_type,number,number,
								 number option(nullable));

	void			set_mkt_args(date ,number  ,number option(nullable) ,number option(nullable),
									number,number,number option(nullable))	;

	/*WORK IN PROGRESS*/
	void set_mkt_args(fx_vol_surf_smile ,number option(nullable) ,number option(nullable) );
	/*----*/
	number			prem(fx_prem_type, out number fwd_prem);
	
	number 			delta(fx_delta_type,
						  number mult 			= FX_DFLT_DELTA_MULT,
						  number d 				= FX_DFLT_DELTA_DIFF,
						  logical centered 		= FX_DFLT_DELTA_CENT);
		
	number 			vega(fx_vega_type,
						 number mult 			= FX_DFLT_VEGA_MULT,
						 number d 				= FX_DFLT_VEGA_DIFF,
						 logical centered 		= FX_DFLT_VEGA_CENT) ;
	
	number 			gamma( 	fx_gamma_type,
							number mult 		= FX_DFLT_GAMMA_MULT,
							number d 			= FX_DFLT_GAMMA_DIFF,
							logical centered 	= FX_DFLT_GAMMA_CENT) ;

	number 			rho_dom(fx_rho_type,
							number mult 		= FX_DFLT_RHO_MULT,
							number d 			= FX_DFLT_RHO_DIFF,
							logical centered 	= FX_DFLT_RHO_CENT ) ;
	
	number 			rho_for(fx_rho_type,
							number mult 		= FX_DFLT_RHO_MULT,
							number d 			= FX_DFLT_RHO_DIFF,
							logical centered 	= FX_DFLT_RHO_CENT ) ;

	
	number 			theta(	fx_theta_type,
							number mult,
							number d ,
							out number diff) ;

	number 			vanna(	fx_vanna_type,
							number mult 		= FX_DFLT_VANNA_MULT,
							number d 			= FX_DFLT_VANNA_DIFF,
							logical centered 	= FX_DFLT_VANNA_CENT ) ;
	
	
	number 			volga(	fx_volga_type,
							number mult 		= FX_DFLT_VOLGA_MULT,
							number d 			= FX_DFLT_VOLGA_DIFF,
							logical centered 	= FX_DFLT_VOLGA_CENT ) ;
	

	number			atm(fx_atm_type);
	number			delta_to_k(fx_delta_type,number);
	
	number 			imp_vol(number,fx_prem_type);
	number 			strike_for_dom();
	number 			vol();
	number 			ttm();
	
	number			rate_cont_dom();
	number			rate_cont_for();
	exercise_type 	exercise_type();
	opt_type 		opt_type();
	number			nominal_for_ccy();
	number			nominal_dom_ccy();
	date 			delivery();
	date 			spot_settlement();
	date 			expiry();
	date 			trade();

protected:
	
	number 	delta_to_k_solver(number,logical);

	number 	ttm_disc();
	number 	ttm_vol();
	
	virtual number 	imp_vol(number);
	
	void	init_mkt_dates(date);
	void 	init_mkt_prices(number,out number, number ,out number  ,number	,out number );
	void 	init_nom(number option(nullable) , number);
	void 	init_nom(number );
	void 	init_fx_spot(date option(nullable),number option(nullable));
	
	override void	prem(out number fp,out number sp) ; //returns a value in domestic currency;
														// also called foreign-domestic price
	
	override number delta_spot(number mult = FX_DFLT_DELTA_MULT,number d = FX_DFLT_DELTA_DIFF, logical centered = FX_DFLT_DELTA_CENT ) ;
	override number delta_fwd(number mult = FX_DFLT_DELTA_MULT,number d = FX_DFLT_DELTA_DIFF, logical centered = FX_DFLT_DELTA_CENT ) 	;
	override number vega(number mult = FX_DFLT_VEGA_MULT,number d = FX_DFLT_VEGA_DIFF, logical centered = FX_DFLT_VEGA_CENT) 	;
	override number gamma_spot(number mult = FX_DFLT_GAMMA_MULT,number d = FX_DFLT_GAMMA_DIFF, logical centered = FX_DFLT_GAMMA_CENT) 	;
	override number gamma_fwd(number mult = FX_DFLT_GAMMA_MULT,number d = FX_DFLT_GAMMA_DIFF, logical centered = FX_DFLT_GAMMA_CENT) 	;
	override number rho(number,number , logical  ) ;
	override number psi(number,number , logical  ) ;	
	override number theta(number, number, out number) ;
	override number vanna(number mult = FX_DFLT_VANNA_MULT,number d = FX_DFLT_VANNA_DIFF,logical centered = FX_DFLT_VANNA_CENT ) ;	
	override number volga(number mult = FX_DFLT_VOLGA_MULT,number d = FX_DFLT_VOLGA_DIFF,logical centered = FX_DFLT_VOLGA_CENT ) ;
	
	number 	delta_spot_pa(number /*, number ,logical */);	//dspa = ds-vfd/s (vfdpcnt=vfd/s)
	number 	delta_fwd_pa(number  );							//dfpa
	number	atm_delta_neutral();							// k such that call delta = − put delta, specify delta type
	number	atm_delta_neutral_pa();
	number	outr_for_dom();
	number	swap_for_dom();
	number	spot_for_dom();
	number	forward_value();

	//member variables
	vanilla_option_fx_def def_;

	fx_spot	fx_spot_;
	number	nominal_for_ccy_;//foreign currency == base currency
	number	nominal_dom_ccy_;
		
	date 	trade_;
	date 	spot_settle_;
	date 	expiry_;
	date 	delivery_ ;

	fx_vol_surface vol_surf_;
};

//----------------------------------------------
// vanilla_option_fx constructor (instrument is STATIC after this constructor)
//----------------------------------------------
vanilla_option_fx.vanilla_option_fx(vanilla_option_fx_def		def,
									date 	option(nullable) 	expiry,//(=maturity)									
									number 	option(nullable) 	strike_for_dom,
									number						nominal)
									:vanilla_option(def.exercise_type(),def.opt_type(),strike_for_dom),									 
									 def_(def),expiry_(expiry), vol_surf_(null<fx_vol_surface>)
{
	init_fx_spot(null<date>,null<number>);
	
	if(!null(expiry_)) {
		logical adj_td 	= false;		
		delivery_ 		= fx_spot_date(expiry_, fx_spot_.base_ccy_calendar(), def_.for_ccy(),
										fx_spot_.price_ccy_calendar(),def_.dom_ccy(),
										fx_spot_.spot_rule(), fx_spot_.spot_days(),fx_spot_.usd_calendar(),adj_td);
	}
	init_nom(strike_for_dom, nominal);
			
}

//----------------------------------------------
// vanilla_option_fx.set_mkt_args  WORK IN PROGRESS
//----------------------------------------------
void vanilla_option_fx.set_mkt_args(fx_vol_surf_smile 		vol_surf,
									number 	option(nullable) swap_for_dom,
									number 	option(nullable) strike_for_dom)//no handling of null strike yet	
{
	QL_FAIL_COND(null(pmodel()), "model not set");
	
	vol_surf_ 				= vol_surf;
	date trade 				= vol_surf.ref_date();
	number spot_for_dom 	= vol_surf.spot();
	
	init_fx_spot(trade,spot_for_dom);
		
	init_mkt_dates(trade);

	number tau = ttm();
	number r_dom,r_for;		
	vol_surf.int_rates(tau, r_dom,r_for);
		
	number outr_for_dom;
	init_mkt_prices(spot_for_dom, swap_for_dom,r_dom,r_for,tau,outr_for_dom);

	number vol = vol_surf_.vol_strike(expiry_,strike_for_dom);
	option_qs_e		qs = option_qs_e.VOL;
	vanilla_option.set_mkt_args(spot_for_dom, outr_for_dom, strike_for_dom, vol,qs,
								r_dom,r_for,tau,tau);
	
	init_nom(strike());
}

//----------------------------------------------
// vanilla_option_fx.set_mkt_args
//----------------------------------------------
void vanilla_option_fx.set_mkt_args(date 			trade,	//=horizon date				
									number 			spot_for_dom,	
									number 	option(nullable) swap_for_dom,
									number 	option(nullable) strike_for_dom,
									number 			vol, 					
									number 			rate_cont_dom,			
									number option(nullable) rate_cont_for)	
{
	if(null(strike_for_dom) && null(strike())){
		if(def_.is_atm_strike()) {
			set_mkt_args(trade,spot_for_dom,swap_for_dom,
						def_.atm_type(),vol, rate_cont_dom,rate_cont_for);
			return;
		}
		number delta;
		if(def_.is_delta_strike(delta)){
			set_mkt_args(trade,spot_for_dom,swap_for_dom,delta,
						def_.delta_type(),	vol, rate_cont_dom,rate_cont_for);
			return;
		}
	}
	
	QL_FAIL_COND(null(pmodel()), "model not set");
	
	vol_surf_ = fx_vol_surf_const(vol);
	
	init_fx_spot(trade,spot_for_dom);
		
	init_mkt_dates(trade);

	number tau = ttm();
	number outr_for_dom;
	init_mkt_prices(spot_for_dom,swap_for_dom,rate_cont_dom,rate_cont_for,tau,outr_for_dom);

	option_qs_e		qs = option_qs_e.VOL;
	vanilla_option.set_mkt_args(spot_for_dom,outr_for_dom,strike_for_dom,vol,qs,
								rate_cont_dom,rate_cont_for,tau,tau);
	
	init_nom(strike());

}
//----------------------------------------------
// vanilla_option_fx.set_mkt_args
//----------------------------------------------
void vanilla_option_fx.set_mkt_args(date 			trade,//=horizon date
									number 			spot_for_dom,
									number 	option(nullable) swap_for_dom,									
									number 			delta,
									fx_delta_type 	delta_type,										
									number 			vol, 
									number 			rate_cont_dom,
									number option(nullable) rate_cont_for)
{
	QL_FAIL_COND(null(pmodel()), "model not set");	

	vol_surf_ = fx_vol_surf_const(vol);
	
	init_fx_spot(trade,spot_for_dom);
		
	init_mkt_dates(trade);

	number tau = ttm();
	number outr_for_dom;
	init_mkt_prices(spot_for_dom,swap_for_dom,rate_cont_dom,rate_cont_for,tau,outr_for_dom);

	number strike_for_dom;
	
	switch(delta_type){
	case fx_delta_type.UNADJ_SPOT:
		if(null(rate_cont_for)) rate_cont_for = 0;
		strike_for_dom = black_scholes_delta_spot_to_strike(delta, type(), spot_for_dom, vol, rate_cont_dom, 
														rate_cont_for, tau, tau);
		break;
	case fx_delta_type.UNADJ_FWD:
		strike_for_dom = black_scholes_delta_fwd_to_strike(delta, type(), outr_for_dom, vol, tau);
		break;
	case fx_delta_type.PREMADJ_SPOT:
		strike_for_dom = ..delta_to_k_solver(delta,false,type(),spot_for_dom,outr_for_dom,
											vol,rate_cont_dom,  rate_cont_for, tau, tau);
		break;
	case fx_delta_type.PREMADJ_FWD:	
		strike_for_dom = ..delta_to_k_solver(delta,true,type(),spot_for_dom,outr_for_dom,
											vol,rate_cont_dom,  rate_cont_for, tau, tau);
		break;
	default:	
		QL_FAIL("invalid delta type (not applicable in this function)");
	}
	
	QL_FAIL_COND(null(strike_for_dom),"delta to strike calculation failed");

	option_qs_e		qs = option_qs_e.VOL;
	vanilla_option.set_mkt_args(spot_for_dom,outr_for_dom,strike_for_dom,vol,qs,
							rate_cont_dom,rate_cont_for,tau,tau);
	init_nom(strike());

	
}

//----------------------------------------------
// vanilla_option_fx.set_mkt_args
//----------------------------------------------
void vanilla_option_fx.set_mkt_args(date 			trade,//=horizon date
									number 			spot_for_dom,
									number 	option(nullable) swap_for_dom,
									fx_atm_type 	atm_type,										
									number 			vol, 
									number 			rate_cont_dom,
									number option(nullable) rate_cont_for )
{
	QL_FAIL_COND(null(pmodel()), "model not set");	

	vol_surf_ = fx_vol_surf_const(vol);
	
	init_fx_spot(trade,spot_for_dom);
	
	init_mkt_dates(trade);

	number tau = ttm();
	number outr_for_dom;
	init_mkt_prices(spot_for_dom,swap_for_dom,rate_cont_dom,rate_cont_for,tau,outr_for_dom);

	number strike_for_dom;
	
	switch(atm_type){
	case fx_atm_type.SPOT:
		strike_for_dom = spot_for_dom;
		break;
	case fx_atm_type.FWD:
	case fx_atm_type.NEUTRAL:
		strike_for_dom =  outr_for_dom;
		break;
	case fx_atm_type.DELTA_NEUTRAL_UNADJ_SPOT:
	case fx_atm_type.DELTA_NEUTRAL_UNADJ_FWD:
		strike_for_dom = outr_for_dom * exp(0.5*vol*vol*tau);
		break;
	case fx_atm_type.DELTA_NEUTRAL_PREMADJ_SPOT:
	case fx_atm_type.DELTA_NEUTRAL_PREMADJ_FWD:
		strike_for_dom 	= outr_for_dom * exp(-0.5*vol*vol*tau) ;
		break;
	default:	
		QL_FAIL("invalid atm type (not applicable in this function)");
	}

	QL_FAIL_COND(null(strike_for_dom),"atm strike calculation failed");

	option_qs_e		qs = option_qs_e.VOL;
	vanilla_option.set_mkt_args(spot_for_dom,outr_for_dom,strike_for_dom,vol,qs,
							rate_cont_dom,rate_cont_for,tau,tau);
	init_nom(strike_for_dom);

}
//----------------------------------------------
// init_nom
//----------------------------------------------
void vanilla_option_fx.init_nom(number option(nullable) strike_for_dom, number nominal)
{
	
	if(def_.face_ccy_is_for()) {
		nominal_for_ccy_ = nominal;
		if(!null(strike_for_dom))
			nominal_dom_ccy_ = strike_for_dom * nominal;
	}
	else{
		nominal_dom_ccy_ = nominal;
		if(!null(strike_for_dom))
			nominal_for_ccy_ = nominal/strike_for_dom;
	}	
}
//----------------------------------------------
// init_nom
//----------------------------------------------
void vanilla_option_fx.init_nom(number strike_for_dom)
{
	
	if(def_.face_ccy_is_for()) {
		nominal_dom_ccy_ = strike_for_dom * nominal_for_ccy_;
	}
	else{
		nominal_for_ccy_ = nominal_dom_ccy_/strike_for_dom;
	}	
}
//----------------------------------------------
// init_fx_spot
//----------------------------------------------
void vanilla_option_fx.init_fx_spot(date option(nullable) trade,
									number option(nullable) spot)
{
	
	error_info error = error_info();
	fx_spot_ = fx_spot(def_.for_ccy(),def_.dom_ccy(), trade, null<string>, error);
	QL_FAIL_COND(error.is_error(), error.message());
	QL_FAIL_COND(null(fx_spot_), "error creating fx-spot instrument");
	if(!null(spot))
		fx_spot_.add_quote(spot,true);
	
}
//----------------------------------------------
// vanilla_option_fx.init_mkt_dates
//----------------------------------------------
void vanilla_option_fx.init_mkt_dates(	date trade)
{
	trade_			= trade;	
	logical adj_td 	= false;		
	spot_settle_ 	= fx_spot_date(	trade, fx_spot_.base_ccy_calendar(), def_.for_ccy(),
									fx_spot_.price_ccy_calendar(),def_.dom_ccy(),
									fx_spot_.spot_rule(), fx_spot_.spot_days(),
									fx_spot_.usd_calendar(),adj_td);
	
	if(null(expiry_)) {
		QL_FAIL_COND(null(def_.expiry_code()), "invalid expiry date/expiry code");
		
		fx_option_expiry(trade, def_.expiry_code(),
						def_.expiry_code_my(), def_.expiry_cal(), fx_spot_.instr_def(),
						spot_settle_,expiry_,delivery_);
		
		
	}
	
}
//----------------------------------------------
// vanilla_option_fx.init_mkt_prices
//----------------------------------------------
void vanilla_option_fx.init_mkt_prices(	number 			s,
										out number   	sw,														
										number 			cont_r,
										out number 		cont_u_r,
										number			tau,
										out number 		f)
{

	if(null(sw)){
		QL_FAIL_COND(null(cont_u_r), "foreign discount rate required to initialize fwd price");
		f = CORE_INT_OPT.bl_f_from_s(s, cont_r, tau,cont_u_r);
		sw = f-s;
	}
	else {
		f = s + sw;
		if(null(cont_u_r))
			cont_u_r = CORE_INT_OPT.bl_cont_u_r_from_sf( s,f,cont_r,tau );
	}
	
}
//----------------------------------------------
// vanilla_option_fx.ttm_disc
//----------------------------------------------
number vanilla_option_fx.ttm_disc()
{
	return (delivery_ - spot_settle_)/365.0;
}
//----------------------------------------------
// vanilla_option_fx.ttm_vol
//----------------------------------------------
number vanilla_option_fx.ttm_vol()
{
	return (expiry_ - trade_)/365.0;
}
//----------------------------------------------
// vanilla_option_fx.ttm
//----------------------------------------------
number vanilla_option_fx.ttm()
{
	return ttm_vol();
}
//----------------------------------------------
// vanilla_option_fx.outr_for_dom
//----------------------------------------------
number	vanilla_option_fx.outr_for_dom()
{
	QL_FAIL_COND(static(),"function not applicable for static instrument");
	
	if(null(args().f_))
		return CORE_INT_OPT.bl_f_from_s(args().s_, args().cont_r_, args().tau_,args().cont_u_r_);		
	else
		return args().f_;
}
//----------------------------------------------
// vanilla_option_fx.swap_for_dom
//----------------------------------------------
number	vanilla_option_fx.swap_for_dom()
{
	return outr_for_dom() - spot_for_dom();
}
//----------------------------------------------
// vanilla_option_fx.forward_value
//----------------------------------------------
number	vanilla_option_fx.forward_value()
{
	number s = spot_for_dom();
	return s * exp(-args().cont_u_r_*args().tau_) - args().k_ * exp(-args().cont_r_)*args().tau_;	
}
//----------------------------------------------
// vanilla_option_fx.outr_for_dom
//----------------------------------------------
number	vanilla_option_fx.spot_for_dom()
{
	QL_FAIL_COND(static(),"function not applicable for static instrument");
	if(null(args().s_)) {
		QL_FAIL_COND(null(args().cont_u_r_), "foreign discount rate required for current operation");		
		return CORE_INT_OPT.bl_s_from_f(args().f_, args().cont_r_, args().tau_,args().cont_u_r_);
	}
	else
		return args().s_;
}

//----------------------------------------------
// vanilla_option_fx.prem
//----------------------------------------------
number vanilla_option_fx.prem(fx_prem_type type, out number fwd_prem)
{
	
	number sp,fp;
	vanilla_option.prem(fp,  sp);
	integer dec = 4;
	
	switch(type){
	case fx_prem_type.DOM_UNIT_FOR:
		fwd_prem = fp;
		return sp;
	case fx_prem_type.DOM_CASH:
		fwd_prem = fp * nominal_for_ccy_;
		return sp*nominal_for_ccy_;
	case fx_prem_type.DOM_PIPS_FOR:
		fwd_prem = fp * pow(10.0, dec );
		return sp * pow(10.0, dec );
	case fx_prem_type.DOM_UNIT_DOM:
		fwd_prem = fp / args().k_;
		return sp/args().k_;
	case fx_prem_type.DOM_PCNT_DOM:
		fwd_prem = fp/args().k_*100.0;
		return sp/args().k_*100.0;
	case fx_prem_type.FOR_CASH:
		fwd_prem = fp / this.spot_for_dom()*nominal_for_ccy_;
		return sp / this.spot_for_dom()*nominal_for_ccy_;		
	case fx_prem_type.FOR_UNIT_FOR:
		fwd_prem = fp / this.spot_for_dom();
		return sp / this.spot_for_dom();
	case fx_prem_type.FOR_PCNT_FOR:
		fwd_prem = fp / this.spot_for_dom()*100.0;
		return sp / this.spot_for_dom()*100.0;		
	case fx_prem_type.FOR_UNIT_DOM:
		fwd_prem = (1.0/args().k_)*(fp/this.spot_for_dom());
		return (1.0/args().k_)*(sp/this.spot_for_dom());
	case fx_prem_type.FOR_PIPS_DOM:
		fwd_prem = (1.0/args().k_)*(fp/this.spot_for_dom()) * pow(10.0, dec );
		return (1.0/args().k_)*(sp/this.spot_for_dom()) * pow(10.0, dec );

	case fx_prem_type.UNIT_FACE:
		if(def_.face_ccy_is_for()) {
			fwd_prem = fp;
			return sp;
		}
		else {
			fwd_prem = (1.0/args().k_)*(fp/this.spot_for_dom());
			return (1.0/args().k_)*(sp/this.spot_for_dom());
		}
		
	case fx_prem_type.PCNT_FACE:
		if(def_.face_ccy_is_for()) {
			fwd_prem = fp / this.spot_for_dom()*100.0;
			return sp / this.spot_for_dom()*100.0;
		}
		else {
			fwd_prem = fp/args().k_*100.0;
			return sp/args().k_*100.0;
		}
	default:	
		QL_FAIL("premium type not supported");
	}
		
}	
//----------------------------------------------
// vanilla_option_fx.prem
//----------------------------------------------
void vanilla_option_fx.prem(out number fp,out number sp)
{
	vanilla_option.prem(fp,  sp);
	fp *= nominal_for_ccy_;
	sp *= nominal_for_ccy_;
}

//----------------------------------------------
// vanilla_option.imp_vol
//----------------------------------------------
number vanilla_option_fx.imp_vol(number p, fx_prem_type type)
{
	QL_FAIL_COND(static(),"function not applicable for static instrument");
	
	integer dec = 4;
	number sp;
	
	switch(type){
	case fx_prem_type.DOM_UNIT_FOR:		
		sp = p;
		break;
	case fx_prem_type.DOM_CASH:		
		sp = p/nominal_for_ccy_;
		break;
	case fx_prem_type.DOM_PIPS_FOR:
		sp = p / pow(10.0, dec );		
		break;
	case fx_prem_type.DOM_UNIT_DOM:
		sp = p*args().k_;	
		break;
	case fx_prem_type.DOM_PCNT_DOM:
		sp = p *args().k_/100.0;		
		break;
	case fx_prem_type.FOR_CASH:
		sp = p * this.spot_for_dom() /nominal_for_ccy_;		
		break;
	case fx_prem_type.FOR_UNIT_FOR:
		sp = p*this.spot_for_dom();		
		break;
	case fx_prem_type.FOR_PCNT_FOR:
		sp = p*this.spot_for_dom()/100.0	;	
		break;
	case fx_prem_type.FOR_UNIT_DOM:
		sp = p*args().k_*this.spot_for_dom();		
		break;
	case fx_prem_type.FOR_PIPS_DOM:
		sp = p*args().k_*this.spot_for_dom()/pow(10.0, dec );		
		break;
	case fx_prem_type.UNIT_FACE:
		if(def_.face_ccy_is_for()) 
			sp = p;			
		else 
			sp = p*args().k_*this.spot_for_dom();			
		break;
	case fx_prem_type.PCNT_FACE:
		if(def_.face_ccy_is_for()) 			
			sp = p*this.spot_for_dom()/100.0	;	
		else 			
			sp = p *args().k_/100.0;		
		break;
	default:	
		QL_FAIL("premium type not supported");
	}
	
	return vanilla_option.imp_vol(sp);
}
//----------------------------------------------
// vanilla_option.imp_vol
//----------------------------------------------
number vanilla_option_fx.imp_vol(number p)
{
	return vanilla_option.imp_vol(p);
}

//----------------------------------------------
// vanilla_option_fx.atm
//----------------------------------------------
number 	vanilla_option_fx.atm(fx_atm_type type)
{
	switch(type){
	case fx_atm_type.SPOT:
		return this.spot_for_dom();
	case fx_atm_type.FWD:
	case fx_atm_type.NEUTRAL://equiv to atm_fwd(); k such that call value = put value
		return this.outr_for_dom();		
	case fx_atm_type.DELTA_NEUTRAL_UNADJ_SPOT:
	case fx_atm_type.DELTA_NEUTRAL_UNADJ_FWD:
	case fx_atm_type.MAX_GAMMA:
	case fx_atm_type.MAX_VEGA:
		return atm_delta_neutral();
	case fx_atm_type.DELTA_NEUTRAL_PREMADJ_SPOT:
	case fx_atm_type.DELTA_NEUTRAL_PREMADJ_FWD:
		return atm_delta_neutral_pa();
	default:	
		QL_FAIL("atm type not supported");
	}
}

//----------------------------------------------
// vanilla_option_fx.atm_delta_neutral
// Unadjusted Spot Delta convention == Unadjusted Forward Delta convention
// --> zero-delta straddle strike placement when prem is paid in domestic curr
//----------------------------------------------
number vanilla_option_fx.atm_delta_neutral()
{
	number f 	= this.outr_for_dom();
	number v	= vol();
	number d1 	= f * exp(0.5*v*v*args().tau_v_) ;
	return d1;
}
//----------------------------------------------
// vanilla_option_fx.atm_delta_neutral_pa
// Prem Adj Spot Delta convention == Prem Adj Forward Delta convention
// --> zero-delta straddle strike placement when prem is paid in foreign curr
//----------------------------------------------
number vanilla_option_fx.atm_delta_neutral_pa()
{
	number f 	= this.outr_for_dom();
	number v	= vol();
	number d1 	= f * exp(-0.5*v*v*args().tau_v_) ;
	return d1;
}
//----------------------------------------------
// vanilla_option_fx.delta_to_k
//----------------------------------------------
number	vanilla_option_fx.delta_to_k(fx_delta_type type,
									 number delta)
{
	//require quote_style=VOL?
	QL_FAIL_COND(static(),"function not applicable for static instrument");
	switch(type){
	case fx_delta_type.UNADJ_SPOT:
		return black_scholes_delta_spot_to_strike(delta, args().cp_, this.spot_for_dom(), vol(), 
											args().cont_r_, args().cont_u_r_, args().tau_, args().tau_v_);
	case fx_delta_type.UNADJ_FWD:
		return black_scholes_delta_fwd_to_strike(delta, args().cp_, this.outr_for_dom(), vol(), 
											 args().tau_v_);
	case fx_delta_type.PREMADJ_SPOT:
		return delta_to_k_solver(delta,false);
	case fx_delta_type.PREMADJ_FWD:	
		return delta_to_k_solver(delta,true);
	default:	
		QL_FAIL("delta to strike calculation for this delta type is not supported");
	}
	
}


number vanilla_option_fx.delta_to_k_solver(number delta,logical fwd)
{
	//require quote_style=VOL?
	return ..delta_to_k_solver(delta,fwd,args().cp_,this.spot_for_dom(),
							this.outr_for_dom(),vol(),
							args().cont_r_ ,args().cont_u_r_,
							args(). tau_,args().tau_v_);
}


//----------------------------------------------
// vanilla_option_fx.delta
//----------------------------------------------
number 	vanilla_option_fx.delta(fx_delta_type 	type,
								number 			mult,
								number 			d,								
								logical 		centered)
{
	switch(type){
	case fx_delta_type.UNADJ_SPOT:
		return vanilla_option.delta_spot(mult,d,centered);
	case fx_delta_type.UNADJ_SPOT_AMNT:
		return vanilla_option.delta_spot(mult,d,centered)/mult*nominal_dom_ccy();
		
	case fx_delta_type.UNADJ_FWD:
		return vanilla_option.delta_fwd(mult,d,centered);
	case fx_delta_type.UNADJ_FWD_AMNT:
		return vanilla_option.delta_fwd(mult,d,centered)/mult*nominal_dom_ccy();
		
	case fx_delta_type.PREMADJ_SPOT:
		return delta_spot_pa( mult/*,d,centered*/);
	case fx_delta_type.PREMADJ_SPOT_AMNT:
		return delta_spot_pa( mult/*,d,centered*/)/mult*nominal_dom_ccy();
		
	case fx_delta_type.PREMADJ_FWD:
		return delta_fwd_pa( mult);
	case fx_delta_type.PREMADJ_FWD_AMNT:
		return delta_fwd_pa( mult)/mult*nominal_dom_ccy();
		
	case fx_delta_type.UNADJ_SPOT_REV:
		return -vanilla_option.delta_spot(mult,d,centered)*this.spot_for_dom()/args().k_;
	case fx_delta_type.UNADJ_SPOT_REV_AMNT:
		return -vanilla_option.delta_spot(mult,d,centered)*this.spot_for_dom()/mult*nominal_dom_ccy();

		
	case fx_delta_type.PREMADJ_SPOT_REV:
		return -delta_spot_pa( mult)*this.spot_for_dom()/args().k_;
	case fx_delta_type.PREMADJ_SPOT_REV_AMNT:
		return -delta_spot_pa( mult)*this.spot_for_dom()/mult*nominal_dom_ccy();

		
	case fx_delta_type.UNADJ_FWD_REV:
		return -vanilla_option.delta_fwd(mult,d,centered)*this.outr_for_dom()/args().k_;
	case fx_delta_type.UNADJ_FWD_REV_AMNT:
		return -vanilla_option.delta_fwd(mult,d,centered)*this.outr_for_dom()/mult*nominal_dom_ccy();
		
	case fx_delta_type.PREMADJ_FWD_REV:
		return -delta_fwd_pa( mult)*this.outr_for_dom()/args().k_;
	case fx_delta_type.PREMADJ_FWD_REV_AMNT:
		return -delta_fwd_pa( mult)*this.outr_for_dom()/mult*nominal_dom_ccy();
	default:	
		QL_FAIL("delta type not supported");
	}
	//unadjusted spot delta ("raw" spot delta) multiplied by nom_for gives the amount of for-ccy that should be bought to hedge a short option
	//unadjusted spot delta reverse --> reverse delta is the delta in the other currency i.e.
	// a for-call is a for-put  --> [buy for-ccy] <--> [sell dom-ccy multipled by spot]
	// negative sign because buy becomes a sell
	// applying the delta hedge to 1 unit of dom-ccy instead of k units of dom requires dividing by k

	/* FENICS
	 *PctDelta			Percent Delta in Face Currency						Yes, If PremCcy is the same as FaceCcy		.25
	 
	*Delta				Points Delta in Counter Face Currency				Yes, If PremCcy is different from FaceCcy	.25
	
	PctVolAdjDelta		Delta for pcnt of the  adj. for vol curve 			No											.25							
						
	VolAdjDelta			Delta for counter currency points 
						per currency adjusted for vol curve					No											.25
						
	PctFwdDelta			Percent Fwd Delta in Face Currency					No											.03030033
					
	PctFwdDeltaAmt		Percent Fwd Delta Amount in Fore  Ccy				No											1000								
						
	FwdDelta			Points Fwd Delta in Counter Face Ccy				No											.03030033
											
	FwdDeltaAmt			Points Fwd Delta Amount in Fore Ccy					No											1000
											
	CtrPctDeltaAmt		Percent Delta Amount in Counter Currency			No											1000
												
	CtrDeltaAmt			Points Delta Amount in Counter Currency				No											1000

	CtrPctFwdDeltaAmt	Percent Fwd Delta Amount in Counter Currency		No											1000
											
	CtrFwdDeltaAmt		Points Fwd Delta Amount in Counter Currency			No											1000						
	 */
}

//----------------------------------------------
// vanilla_option_fx.delta_spot
//----------------------------------------------
number vanilla_option_fx.delta_spot(number mult,
									number d,								
									logical centered)
{
	return vanilla_option.delta_spot( mult, d,centered);
}

//----------------------------------------------
// vanilla_option_fx.delta_fwd
//----------------------------------------------
number vanilla_option_fx.delta_fwd(	number mult,
									number d,	
									logical centered)
{
	return vanilla_option.delta_fwd( mult, d,centered);
}
//----------------------------------------------
// vanilla_option_fx.delta_spot_pa
// alt. delta_spot_pa = delta_spot - spot_prem/s
//----------------------------------------------
number vanilla_option_fx.delta_spot_pa(number mult)
											 //number d,											 
											 //logical centered)
{
	return black_scholes_delta_spot_pa(	args().cp_,
										this.spot_for_dom(), 
										args().k_, 
										vol(), 
										args().cont_r_,
										args().cont_u_r_,
										args().tau_,						
										args().tau_v_) * mult;
}

//----------------------------------------------
// vanilla_option_fx.delta_fwd_pa
//----------------------------------------------
number vanilla_option_fx.delta_fwd_pa(number mult)
{
	return black_scholes_delta_fwd_pa(	args().cp_,
										this.spot_for_dom(), 
										args().k_, 
										vol(), 
										args().cont_r_,
										args().cont_u_r_,
										args().tau_,						
										args().tau_v_) * mult;
		
}

//----------------------------------------------
// vanilla_option_fx.gamma_spot
//----------------------------------------------
number vanilla_option_fx.gamma_spot(number mult,
									number d,	
									logical centered)
{
	return vanilla_option.gamma_spot(mult,d,centered);
}
//----------------------------------------------
// vanilla_option_fx.gamma_fwd
//----------------------------------------------
number vanilla_option_fx.gamma_fwd(	number mult,
									number d,	
									logical centered)
{
	return vanilla_option.gamma_fwd(mult,d,centered);
}
//----------------------------------------------
// vanilla_option_fx.gamma
//----------------------------------------------
number 	vanilla_option_fx.gamma(fx_gamma_type 	type,
								number 			mult,
								number 			d,								
								logical 		centered)
{
	switch(type){
	case fx_gamma_type.UNADJ_SPOT:
		return this.gamma_spot(mult,d,centered);
	case fx_gamma_type.UNADJ_FWD:
		return this.gamma_fwd(mult,d,centered);
	case fx_gamma_type.UNADJ_SPOT_AMNT:
		return this.gamma_spot(mult,d,centered)/mult*nominal_dom_ccy();
	case fx_gamma_type.UNADJ_FWD_AMNT:
		return this.gamma_fwd(mult,d,centered)/mult*nominal_dom_ccy();
	default:	
		QL_FAIL("gamma type not supported");
	}
}
/*
 Fenics:
 Gamma			Gamma in percent of the currency 
				amount for a one percent move in spot	No			.023754
GammaAmt		Gamma of the currency amount for a 
				one percent move in spot				No			1000
 */

//----------------------------------------------
// vanilla_option_fx.vega
//----------------------------------------------
number vanilla_option_fx.vega(	number mult,
								number d,	
								logical centered)
{
	return vanilla_option.vega(mult,d,centered);
}
//----------------------------------------------
// vanilla_option_fx.vega
//----------------------------------------------
number 	vanilla_option_fx.vega(fx_vega_type 	type,
								number 			mult,
								number 			d,								
								logical 		centered)
{
	switch(type){
	case fx_vega_type.UNADJ_DOM:
		return this.vega(mult,d,centered);
	case fx_vega_type.UNADJ_FOR:
		return this.vega(mult,d,centered)/this.spot_for_dom();
	case fx_vega_type.UNADJ_DOM_AMNT:
		return this.vega(mult,d,centered)/mult*nominal_dom_ccy();
	case fx_vega_type.UNADJ_FOR_AMNT:
		return this.vega(mult,d,centered)/this.spot_for_dom()/mult*nominal_dom_ccy();
	default:	
		QL_FAIL("vega type not supported");
	}
}
/*
 Fenics:
 PctVega		Percent Vega in Face Currency			No				.0003544
 PctVegaAmt		Percent Vega Amount in Fore Currency	No				1000
 Vega			Points Vega in Counter Face Currency	No				.0004607
 VegaAmt		Points Vega Amount in Fore Currency		No				1000
 CtrPctVegaAmt	Percent Vega Amount in Counter Currency	No				1000
 CtrVegaAmt		Points Vega Amount in Counter Currency	No				1000
 */
//----------------------------------------------
// vanilla_option_fx.vanna (dVega dSpot)
//----------------------------------------------
number vanilla_option_fx.vanna(	number 			mult,
								number 			d,								
								logical 		centered)//centered currently not used
{
	return vanilla_option.vanna(mult,d,centered);
}
//----------------------------------------------
// vanilla_option_fx.vanna
//----------------------------------------------
number vanilla_option_fx.vanna(	fx_vanna_type type,
								number mult,
								number d,								
								logical centered)
{
	switch(type){
	case fx_vanna_type.UNADJ_DOM:
		return this.vanna( mult, d,centered);
	case fx_vanna_type.UNADJ_FOR:
		return this.vanna( mult, d,centered)/this.spot_for_dom();
	case fx_vanna_type.UNADJ_DOM_AMNT:
		return this.vanna( mult, d,centered)*nominal_for_ccy();
	case fx_vanna_type.UNADJ_FOR_AMNT:
		return this.vanna( mult, d,centered)/this.spot_for_dom()*nominal_for_ccy();
	default:	
		QL_FAIL("vanna type not supported");
	}
	
}
//Fenics: dVega dSpot
	//1. %of for-curr for 1% rise
	//2. amount for-curr for 1% rise
//----------------------------------------------
// vanilla_option_fx.volga (dVega dVol)
//----------------------------------------------
number vanilla_option_fx.volga(	number 			mult,
								number 			d,								
								logical 		centered)//centered currently not used
{
	return vanilla_option.volga(mult,d,centered);
}
//Fenics: dVega dVol:
	//1. in percent of the for-curr for a 1% rise,
	//2. amount of the for-curr for a 1% rise
	//3. in dom-curr points per for-curr for a 1% rise,
	//4. amount in dom-curr points per for-curr for a 1% rise

//----------------------------------------------
// vanilla_option_fx.volga
//----------------------------------------------
number vanilla_option_fx.volga(	fx_volga_type type,
								number mult,
								number d,								
								logical centered)
{
	switch(type){
	case fx_volga_type.UNADJ_DOM:
		return this.volga( mult, d,centered);
	case fx_volga_type.UNADJ_FOR:
		return this.volga( mult, d,centered)/this.spot_for_dom();
	case fx_volga_type.UNADJ_DOM_AMNT:
		return this.volga( mult, d,centered)*nominal_for_ccy();
	case fx_volga_type.UNADJ_FOR_AMNT:
		return this.volga( mult, d,centered)/this.spot_for_dom()*nominal_for_ccy();
	default:	
		QL_FAIL("volga type not supported");
	}
	
}
//----------------------------------------------
// vanilla_option_fx.rho_dom
//----------------------------------------------
number vanilla_option_fx.rho_dom(	fx_rho_type type,
									number mult,
									number d,								
									logical centered)
{
	switch(type){
	case fx_rho_type.UNADJ_DOM:
		return this.rho( mult, d,centered);
	case fx_rho_type.UNADJ_FOR:
		return this.rho( mult, d,centered)/this.spot_for_dom();
	case fx_rho_type.UNADJ_DOM_AMNT:
		return this.rho( mult, d,centered)*nominal_for_ccy();
	case fx_rho_type.UNADJ_FOR_AMNT:
		return this.rho( mult, d,centered)/this.spot_for_dom()*nominal_for_ccy();
	default:	
		QL_FAIL("rho type not supported");
	}
	
}
/*
 Fenics:
 PctRho			Rho in percent of the currency for a
				one percent rise						No			.00004277
 Rho			Rho in counter currency points per 
				currency for a one percent rise			No			.00005560
*/
//----------------------------------------------
// vanilla_option_fx.rho
//----------------------------------------------
number vanilla_option_fx.rho(	number mult,
								number d,								
								logical centered)
{
	return vanilla_option.rho( mult, d,centered);
}

//----------------------------------------------
// vanilla_option_fx.rho_for (also called psi and phi)
//----------------------------------------------
number vanilla_option_fx.rho_for(	fx_rho_type type,
									number mult,
									number d,								
									logical centered)
{
	switch(type){
	case fx_rho_type.UNADJ_DOM:
		return this.psi( mult, d,centered);
	case fx_rho_type.UNADJ_FOR:
		return this.psi( mult, d,centered)/this.spot_for_dom();
	case fx_rho_type.UNADJ_DOM_AMNT:
		return this.psi( mult, d,centered)*nominal_for_ccy();
	case fx_rho_type.UNADJ_FOR_AMNT:
		return this.psi( mult, d,centered)/this.spot_for_dom()*nominal_for_ccy();
	default:	
		QL_FAIL("rho type not supported");
	}
}
/*
 Fenics:
 PctPhi		Phi in percent of the currency for
			a one percent rise						No			.00004301
 Phi		Phi in counter currency points per 
			currency for a one percent rise			No			.00005592
*/
//----------------------------------------------
// vanilla_option_fx.psi
//----------------------------------------------
number vanilla_option_fx.psi(number mult,
							number d,								
							logical centered)
{
	return vanilla_option.psi( mult, d,centered);
}
//----------------------------------------------
// vanilla_option_fx.theta
//----------------------------------------------
number vanilla_option_fx.theta(	fx_theta_type 	type,
								number 			mult,
								number 			d,
								out number 		diff)
{
	QL_FAIL_COND(static(),"function not applicable for static instrument");
	if(d < 0)
		d = -d;
	
	switch(type){
	case fx_theta_type.BUS_DAY:
		{
			date next = def_.expiry_cal().move_bus_days(trade_,integer(d));
			return this.theta( mult, (next-trade_)/365.0,diff);
		}
		break;
	case fx_theta_type.CAL_DAY:
		return this.theta( mult, d/365.0,diff);
		break;
	case fx_theta_type.BUS_DAY_FOR:
		{
			date next = def_.expiry_cal().move_bus_days(trade_,integer(d));
			return this.theta( mult, (next-trade_)/365.0,diff)/this.spot_for_dom();
		}
		break;
	case fx_theta_type.CAL_DAY_FOR:
		return this.theta( mult, d/365.0,diff)/this.spot_for_dom();
		break;
	default:	
		QL_FAIL("theta type not supported");
	}
}
/*
 Fenics:
 PctDecay		one-business-day decay in percent 
				of the for-curr							No		.00002912
 Decay			one-business-day decay in dom-curr
				points per for-curr						No		.00003708
 */
//----------------------------------------------
// vanilla_option.theta
//----------------------------------------------
number vanilla_option_fx.theta(	number mult,
								number d,
								out number diff)
{	
	return vanilla_option.theta( mult, d/365.0,diff);
}

//----------------------------------------------
// vanilla_option.strike
//----------------------------------------------
number vanilla_option_fx.strike_for_dom()
{	
	return vanilla_option.strike();
}
//----------------------------------------------
// vanilla_option.vol
//----------------------------------------------
number vanilla_option_fx.vol()
{
	QL_FAIL_COND(static(),"function not applicable for static instrument");
	return pmodel().vol();
}
//----------------------------------------------
// vanilla_option.rate_cont_dom
//----------------------------------------------
number vanilla_option_fx.rate_cont_dom()
{
	QL_FAIL_COND(static(),"function not applicable for static instrument");
	return args().cont_r_;
}
//----------------------------------------------
// vanilla_option.rate_cont_for
//----------------------------------------------
number vanilla_option_fx.rate_cont_for()
{
	QL_FAIL_COND(static(),"function not applicable for static instrument");
	return args().cont_u_r_;
}
//----------------------------------------------
// vanilla_option.exercise_type
//----------------------------------------------
exercise_type vanilla_option_fx.exercise_type()
{
	return vanilla_option.exercise_type();
}
//----------------------------------------------
// vanilla_option.opt_type
//----------------------------------------------
opt_type vanilla_option_fx.opt_type()
{
	return vanilla_option.type();
}
//----------------------------------------------
// vanilla_option.nominal_for_ccy
//----------------------------------------------
number vanilla_option_fx.nominal_for_ccy()
{
	return nominal_for_ccy_;
}
//----------------------------------------------
// vanilla_option.nominal_dom_ccy
//----------------------------------------------
number vanilla_option_fx.nominal_dom_ccy()
{
	return nominal_dom_ccy_;
}
//----------------------------------------------
// dates
//----------------------------------------------
date vanilla_option_fx.delivery()
{
	return delivery_;
}
date vanilla_option_fx.spot_settlement()
{
	QL_FAIL_COND(static(),"function not applicable for static instrument");
	return spot_settle_;
}
date vanilla_option_fx.expiry()
{
	return expiry_;
}
date vanilla_option_fx.trade()
{
	QL_FAIL_COND(static(),"function not applicable for static instrument");
	return trade_;
}



vanilla_option_fx_def instr_def_fx_option(	INSTR_TMPL.vanilla_option_fx_def_tmpl	tmpl,
											error_info option(nullable) error = null<error_info>)
//option (category: 'Financial/Options/Vanilla/FX')
{	
	try{		
		return vanilla_option_fx_def(tmpl.currency_pair(), tmpl.exercise_type(), tmpl.opt_type(), tmpl.cut_off(),  tmpl.bd(),  tmpl.eom(),									
									tmpl.expiry_code(), tmpl.expiry_cal(), tmpl.strike_code(), tmpl.face_ccy(), tmpl.prem_ccy(), tmpl.prem_at_spot(),
									tmpl.delta_type(),tmpl.atm_type(), tmpl.del_type(),error);						
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "vanilla_option_fx_def.instr_def_fx_option");
		return null<vanilla_option_fx_def>;
	}
}