/*	
	Option vanilla fx vol stuff
	Developer: Algorithmica Research, Magnus Nyström			
*/
option(null: hard);

class enum fx_atm_curve_type option (category: 'Financial/Options/Vanilla/FX')
{
	ATM option(str:"atm"),
	ATM_SMILE_1_WING option(str:"atm smile one wing"),
	ATM_SMILE_2_WINGS option(str:"atm smile two wings")
};

class enum fx_atm_curve_vol_ip_type option (category: 'Financial/Options/Vanilla/FX')
{
	VOL_LIN option(str:"linear vol"),
	VAR_LIN option(str:"linear variance") 
};

//----------------------------------------------
// class fx_quote_curve_atm
/* 	common ex.
	atm_type = delta neutral
	atm_long_cutoff = 2
	atm_type_long = fx_atm_type.FWD
*/
//----------------------------------------------
class fx_quote_curve_atm	
{
public:
	fx_quote_curve_atm(	date ,vector(date) , vector(number) ,vector(fx_atm_type),
					vector(fx_delta_type) option(nullable),fx_atm_curve_vol_ip_type ,logical );
	
	virtual 		fx_atm_curve_type type();
	date 			expiry(integer);
	vector(date) 	expiry();
	date 			max_expiry();
	date 			min_expiry();
	vector(number) 	ttm();
	date 			ref_date();
	number 			atm_vol(date , logical option(nullable)  );
	number 			atm_strike(date ,number,number,number ,number);

	fx_atm_type		atm_type(date);
	fx_delta_type	delta_type(date);
	logical			allow_extrap();
	virtual logical	atm_smile_vol(	date,logical option(nullable) ,out vector(number) 	res);
	
protected:
	
	logical 	check_range(date , logical );
	logical  	in_range(date );
	integer 	locate_expiry(date );
	number 		ip_vol(vector(number) , number  );
	number 		ip_lin(vector(number) ,vector(number), number  );
	number 		ip_lin_sqrt(vector(number) ,vector(number), number  );
	
	date 					ref_date_;
	vector(date) 			exp_;
	vector(number) 			vol_;	
	vector(fx_atm_type)		atm_type_;
	logical 				allow_extrap_;
	fx_atm_curve_vol_ip_type 	vol_ip_type_;
	vector(fx_delta_type)	delta_type_;
	
};

//----------------------------------------------
// constructor
//----------------------------------------------
fx_quote_curve_atm.fx_quote_curve_atm(	date 					ref_date,
										vector(date) 			expiry,
										vector(number) 			vol,						  
										vector(fx_atm_type)		atm_type,
										vector(fx_delta_type) option(nullable)	delta_type,
										fx_atm_curve_vol_ip_type 	vol_ip_type,						 					
										logical 				allow_extrap)
										//number option(nullable) atm_long_cutoff,
										//fx_atm_type option(nullable) atm_type_long)
										: ref_date_(ref_date),exp_(expiry), vol_(vol),
										allow_extrap_(allow_extrap), vol_ip_type_(vol_ip_type)
{
	integer size = v_size(exp_);
	QL_REQUIRE(size > 0, "invalid vector size (expiry)");
	QL_REQUIRE(size == v_size(vol_), "vector sizes not equal");	

	for (integer i=0; i<size; i++) {
		QL_REQUIRE(!null(exp_[i]), "invalid expiry date (null)");
		QL_REQUIRE(!null(vol_[i]), "invalid volatility (null)");		
	}
	
	for (integer i=0, j=1; j<size; i++) {
		QL_REQUIRE(expiry[j] > expiry[i], "unsorted or equal expiry dates");
		++j;
	}
	
	QL_REQUIRE(expiry[0] > ref_date, "invalid ref_date");

	QL_REQUIRE(size == v_size(atm_type) || v_size(atm_type) == 1, "invalid vector size (atm_type)");
	if(v_size(atm_type) == 1) {
		resize(atm_type_, size);
		atm_type_ = atm_type[0];
		QL_REQUIRE(!null(atm_type[0]), "invalid atm type (null)");	
	}
	else {
		atm_type_ = atm_type;
		for (integer i=0; i<size; i++) {
			QL_REQUIRE(!null(atm_type_[i]), "invalid atm type (null)");	
		}
		
	}
	if(null(delta_type)) {
		resize(delta_type_, size);
		delta_type_ = null<fx_delta_type>;
	}
	else {
		QL_REQUIRE(size == v_size(delta_type) || v_size(delta_type) == 1, "invalid vector size (delta_type)");
		if(v_size(delta_type) == 1) {
			resize(delta_type_, size);
			delta_type_ = delta_type[0];		
		}
		else
			delta_type_ = delta_type;
	}
}

fx_atm_curve_type 	fx_quote_curve_atm.type() 		{ return fx_atm_curve_type.ATM; }
logical 		fx_quote_curve_atm.allow_extrap() 	{ return allow_extrap_;}
vector(number) 	fx_quote_curve_atm.ttm() 			{ return (exp_-ref_date_)/365.0;}
date 			fx_quote_curve_atm.ref_date() 		{ return ref_date_;}
vector(date) 	fx_quote_curve_atm.expiry()			{ return clone_vector(exp_);}
date 			fx_quote_curve_atm.max_expiry() 	{ return exp_[v_size(exp_)-1]; }
date 			fx_quote_curve_atm.min_expiry() 	{ return exp_[0]; }

//----------------------------------------------
// expiry
//----------------------------------------------
date fx_quote_curve_atm.expiry(integer i)
{
	QL_REQUIRE(i >= 0 && i < v_size(exp_), "invalid index");
	return exp_[i];
}

//----------------------------------------------
// ip_vol WORK IN PROGRESS requires a lot more stuff!!
// checks must have been done before calling this func
//----------------------------------------------
number fx_quote_curve_atm.ip_vol(vector(number) ttm, number t)
{
	if(vol_ip_type_ == fx_atm_curve_vol_ip_type.VAR_LIN){
		vector(number) var = vol_.* vol_ .* ttm;
		number var_t = ip_lin(var, ttm,  t);
		return sqrt(var_t/t);
	}
	else
		return ip_lin(vol_, ttm,  t);
}
//----------------------------------------------
// ip_lin
// checks must have been done before calling this func
//----------------------------------------------
number fx_quote_curve_atm.ip_lin(vector(number) v, vector(number) ttm, number t)
{	
	return ip_linear(0,0).interpolate(ttm,v).eval(t);
}
//----------------------------------------------
// ip_lin_sqrt
// checks must have been done before calling this func
//----------------------------------------------
number fx_quote_curve_atm.ip_lin_sqrt(vector(number) v, vector(number) ttm, number t)
{
	ttm = sqrt(ttm);
	t 	= sqrt(t);
	return ip_linear(0,0).interpolate(ttm,v).eval(t);
}
//----------------------------------------------
// locate_expiry
//----------------------------------------------
integer fx_quote_curve_atm.locate_expiry(date e)
{
	if (e <= min_expiry())
		return 0;
	else if (e >= max_expiry())
		return v_size(exp_)-1;
	else {
		for(integer i=1;i<v_size(exp_);i++) {
			if(e < exp_[i]) return i-1;
		}
	}
}
//----------------------------------------------
// in_range
//----------------------------------------------
logical fx_quote_curve_atm.in_range(date exp)
{
	if(exp <= ref_date_) return false;		
	date x1 = min_expiry();
	date x2 = max_expiry();		
	return (exp >= x1 && exp <= x2);
}
//----------------------------------------------
// check_range
//----------------------------------------------
logical fx_quote_curve_atm.check_range(date exp, logical allow_extrap)
{
	if(exp <= ref_date_)
		return false;
	else if(allow_extrap)
		return true;
	else
		return in_range(exp);
}
//----------------------------------------------
// atm_type
//----------------------------------------------
fx_atm_type	fx_quote_curve_atm.atm_type(date expiry)
{
	integer i = locate_expiry(expiry);
	return atm_type_[i];
}
//----------------------------------------------
// delta_type
//----------------------------------------------
fx_delta_type fx_quote_curve_atm.delta_type(date expiry)
{
	integer i = locate_expiry(expiry);
	return delta_type_[i];
}
//----------------------------------------------
// atm_strike
//----------------------------------------------
number fx_quote_curve_atm.atm_strike(	date 		expiry,
										number 		vol_atm,
										number 		s ,
										number 		cont_r,
										number 		cont_u_r )
{
	number tau 				= (expiry-ref_date_)/365.0;
	number df 				= exp(-cont_r*tau);
	number dfq 				= exp(-cont_u_r*tau);	
	integer i 				= locate_expiry(expiry);
	fx_atm_type atm_t 		= atm_type_[i];
	fx_delta_type delta_t	= delta_type_[i];
	return CORE_INT_FX_OPT.atm_strike(atm_t, vol_atm , s , tau, df , dfq, delta_t);	
}
//----------------------------------------------
// atm_vol
//----------------------------------------------
number fx_quote_curve_atm.atm_vol(	date expiry,
									logical option(nullable) allow_extrap )
{
	if(null(allow_extrap)) allow_extrap = allow_extrap_;
	
	QL_REQUIRE(check_range(expiry,allow_extrap), "expiry is outside of allowed expiry dates");
			   
	integer i = locate_expiry(expiry);
		
	if(expiry <= min_expiry()) 
		return vol_[0];
	else if(expiry >= max_expiry()) 
		return vol_[v_size(vol_)-1];	
	else  if(expiry == exp_[i]) 
		return vol_[i];
	else 
		return ip_vol(ttm(),(expiry-ref_date_)/365.0);	
}

//----------------------------------------------
// atm_smile_vol
//----------------------------------------------
logical fx_quote_curve_atm.atm_smile_vol(	date 				expiry,
											logical option(nullable) allow_extrap,
											out vector(number) 	res)
{
	resize(res,2);		
	res[0] = (expiry-ref_date_)/365.0;
	res[1] = atm_vol(expiry,allow_extrap );
	return true;
}

//----------------------------------------------
// class fx_quote_curve_smile
//----------------------------------------------
class fx_quote_curve_smile : public fx_quote_curve_atm	
{
public:
	fx_quote_curve_smile(date, vector(date), vector(number), vector(number) ,
						 vector(number) ,number ,vector(number) ,
						 vector(number) ,number,
						 vector(fx_delta_type) ,vector(fx_atm_type)	,
						 fx_atm_curve_vol_ip_type, fx_term_struct ,
						 vector(logical) is_mkt_fly = [true],logical allow_extrap = false );

	fx_quote_curve_smile(date, vector(date), vector(number), vector(number) ,
						 vector(number) ,number ,vector(fx_delta_type) ,vector(fx_atm_type),
						 fx_atm_curve_vol_ip_type, fx_term_struct ,
						 vector(logical) is_mkt_fly = [true],logical allow_extrap = false);

	override fx_atm_curve_type type();		
	override logical	atm_smile_vol(	date,logical option(nullable) ,out vector(number) );
	
	void 				atm_delta(	date ,out number ,out number );	
	void 				delta(out number,out number);
	void  				rr_ss_vols(out vector(number) ,out vector(number) ,out vector(number) ,
									out vector(number) ,out vector(number) );
	void				rr_ss_vols(	date,logical option(nullable) ,out number ,out number ,out number ,
									out number ,out number );

	void  				rr_ss_strikes(out vector(number) ,out vector(number) ,out vector(number) ,
									out vector(number) ,out vector(number) );
	
	number				spot();
	disc_func			df_dom();
	disc_func			df_for();
protected:
	logical 			use_equiv_proc1(integer);
	logical 			use_equiv_proc2(integer);
	void				init_fly();
	
	vector(number)  	rr_;
	vector(number)  	fly_;	
	number				delta_;
	
	vector(number)  	rr2_;
	vector(number)  	fly2_;
	number				delta2_;

	vector(logical) 	is_broker_fly_;
	vector(logical)		_2wing_ok_;
	vector(number)  	sm_fly_;
	vector(number)  	sm_fly2_;

	fx_term_struct		r_ts_;	
};


//----------------------------------------------
// constructor  (input of curve with only one delta level eg. 25delta)
//----------------------------------------------
fx_quote_curve_smile.fx_quote_curve_smile(date 					ref_date,
										  vector(date) 			expiry,
										  vector(number) 		vol,
										  vector(number) 		rr,
										  vector(number) 		fly,
										  number 				delta,
										  vector(fx_delta_type) delta_type,
										  vector(fx_atm_type)	atm_type,
										  fx_atm_curve_vol_ip_type vol_ip_type,
										  fx_term_struct 		r_ts,
										  vector(logical) 		is_broker_fly,						  
										  logical 				allow_extrap )
										 : fx_quote_curve_atm(ref_date,expiry,vol,atm_type,delta_type,vol_ip_type,allow_extrap),
										   rr_(rr), fly_(fly), delta_(delta),
										   rr2_(null<vector(number)>), fly2_(null<vector(number)>),delta2_(null<number>),
										   sm_fly2_(null<vector(number)>),
										   r_ts_(r_ts)
{
	integer size = v_size(exp_);
	QL_REQUIRE(size == v_size(rr_), "rr vector size not equal to number of expiry dates");
	QL_REQUIRE(size == v_size(fly_), "fly vector size not equal to number of expiry dates");
	QL_REQUIRE(delta > 0 && delta < 1, "delta must be between 0 and 1");	
	QL_REQUIRE(size == v_size(is_broker_fly) || v_size(is_broker_fly) == 1, "invalid vector size for is_broker_fly");
	if(v_size(is_broker_fly) == 1){
		resize(is_broker_fly_,size);
		is_broker_fly_	= is_broker_fly[0];
	}
	else {
		is_broker_fly_ = is_broker_fly;
	}

	for (integer i=0; i<size; i++) {
		QL_REQUIRE(!null(rr_[i]), "invalid rr quote (null)");
		QL_REQUIRE(!null(fly_[i]), "invalid fly quote (null)");
		QL_REQUIRE(!null(is_broker_fly_[i]), "invalid fly convention (null)");
	}
	resize(_2wing_ok_,size);
	_2wing_ok_ = false;
	
	resize(sm_fly_,size);
	init_fly();
		
}
//----------------------------------------------
// constructor  (input of curve with two delta levels eg. 25delta + 10delta)
// it is allowed to have missing 10delta quotes
//----------------------------------------------
fx_quote_curve_smile.fx_quote_curve_smile(date 				ref_date,
										  vector(date) 		expiry,
										  vector(number) 	vol,
										  vector(number) 	rr,
										  vector(number) 	fly,
										  number 			delta,
										  vector(number) 	rr2,
										  vector(number) 	fly2,
										  number 			delta2,
										  vector(fx_delta_type) delta_type,
										  vector(fx_atm_type)	atm_type,
										  fx_atm_curve_vol_ip_type vol_ip_type,
										  fx_term_struct 	r_ts,
										  vector(logical) 	is_broker_fly,						  
										  logical 			allow_extrap)
										 : fx_quote_curve_atm(ref_date,expiry,vol,atm_type,delta_type,vol_ip_type,allow_extrap),
										   rr_(rr), fly_(fly), delta_(delta),
										   rr2_(rr2), fly2_(fly2),delta2_(delta2),r_ts_(r_ts)
{
	integer size = v_size(exp_);	
	QL_REQUIRE(size == v_size(rr_), "rr vector size not equal to number of expiry dates");
	QL_REQUIRE(size == v_size(fly_), "fly vector size not equal to number of expiry dates");
	QL_REQUIRE(size == v_size(rr2_), "rr2 vector size not equal to number of expiry dates");
	QL_REQUIRE(size == v_size(fly2_), "fly2 vector size not equal to number of expiry dates");
	QL_REQUIRE(delta > 0 && delta < 1, "delta must be between 0 and 1");
	QL_REQUIRE(delta2 > 0 && delta < 1, "delta2 must be between 0 and 1");
	QL_REQUIRE(delta > delta2, "delta must be strictly larger than delta2");
	QL_REQUIRE(size == v_size(is_broker_fly) || v_size(is_broker_fly) == 1, "invalid vector size for is_broker_fly");
	if(v_size(is_broker_fly) == 1){
		resize(is_broker_fly_,size);
		is_broker_fly_	= is_broker_fly[0];
	}
	else {
		is_broker_fly_ = is_broker_fly;
	}

	resize(_2wing_ok_,size);
	for (integer i=0; i<size; i++) {
		QL_REQUIRE(!null(rr_[i]), "invalid rr quote (null)");
		QL_REQUIRE(!null(fly_[i]), "invalid fly quote (null)");
		_2wing_ok_[i] = !(null(rr2_[i]) || null(fly2_[i]));					
		//QL_REQUIRE(!null(rr2_[i]), "invalid rr2 quote (null)");
		//QL_REQUIRE(!null(fly2_[i]), "invalid fly2 quote (null)");
		QL_REQUIRE(!null(is_broker_fly_[i]), "invalid fly convention (null)");
	}

	resize(sm_fly_,size);
	resize(sm_fly2_,size);
	init_fly();
}

fx_atm_curve_type 	fx_quote_curve_smile.type() 	{ return null(delta2_) ? fx_atm_curve_type.ATM_SMILE_1_WING : fx_atm_curve_type.ATM_SMILE_2_WINGS; }
number 			fx_quote_curve_smile.spot() 	{ return r_ts_.spot(); }
disc_func 		fx_quote_curve_smile.df_dom() 	{ return r_ts_.df_dom(); }
disc_func 		fx_quote_curve_smile.df_for() 	{ return r_ts_.df_for(); }
//----------------------------------------------
// use_equiv_proc
//----------------------------------------------
logical fx_quote_curve_smile.use_equiv_proc1(integer i)
{
	return /*true;//*/is_broker_fly_[i] && abs(rr_[i]) >= 0.1 * vol_[i];
}
logical fx_quote_curve_smile.use_equiv_proc2(integer i)
{
	return /*true;//*/is_broker_fly_[i] && abs(rr2_[i]) >= 0.1 * vol_[i];
}

//----------------------------------------------
// init_fly
//----------------------------------------------
void fx_quote_curve_smile.init_fly()
{
	logical is_1_wing 	= type() == fx_atm_curve_type.ATM_SMILE_1_WING;	
	integer size 		= v_size(exp_);	
	number s 			= r_ts_.spot();
		
	for (integer i=0; i<size; i++) {
					
		if(!is_broker_fly_[i]) {
			sm_fly_[i] 	= fly_[i];
			if(!is_1_wing)
				sm_fly2_[i] = _2wing_ok_[i] ? fly2_[i] : null<number>;
			continue;
		}
		number tau 		= (exp_[i]-ref_date_)/365.0;
		number df		= r_ts_.df_dom(tau);
		number dfq		= r_ts_.df_for(tau);
		number f 		= s * dfq / df;		
		number k_atm 	= CORE_INT_FX_OPT.atm_strike(atm_type_[i], vol_[i] , s , tau, df , dfq, delta_type_[i]);		
		..delta_type dt = CORE_INT_FX_OPT.delta_type_conv( delta_type_[i]);

		if(is_1_wing) {
			
			if(!use_equiv_proc1(i)) {
				sm_fly_[i] = fly_[i] ;
				continue;
			}

			//tmp_kl1,tmp_kh1:  set to something that will pass the checks in constructor (will be updated and is not used)
			number tmp_kl1 = k_atm-0.0001;
			number tmp_kh1 = k_atm+0.0001;
			CORE_INT_VOL.vol_smile_model_vannavolga_ip sm = new CORE_INT_VOL.vol_smile_model_vannavolga_ip( [tmp_kl1,k_atm,tmp_kh1],[0,vol_[i],0],
																											s, tau, df , dfq, dt, delta_,
																											null<number>,true);
			number k_c,k_p,vol_p,vol_c;
			sm_fly_[i] = CORE_INT_FX_OPT.bfly_to_sfly(s ,f, tau ,rr_[i] ,fly_[i] ,df ,dfq ,
													  delta_, sm, true, k_c,k_p,vol_p, vol_c);

			QL_REQUIRE(!null(sm_fly_[i]), "invalid smile-fly (conversion from broker-fly failed)");
			
		}
		else {
			if(!_2wing_ok_[i]) {
				sm_fly2_[i] = null<number>;
				continue;
			}
			number k_c,k_p,vol_p,vol_c;			
			if(!use_equiv_proc1(i)) {
				sm_fly_[i] = fly_[i] ;				
				CORE_INT_FX_OPT.fx_rr_ss_vols(vol_[i],rr_[i],sm_fly_[i],vol_c, vol_p);
				number cont_u_r = -log(dfq)/tau;
				number cont_r 	= -log(df)/tau;
				CORE_INT_FX_OPT.delta_to_k_cp(delta_type_[i],delta_,s,f,vol_c,vol_p,cont_r,cont_u_r,tau,k_c,k_p);
			}
			else {
				//tmp_kl1,tmp_kh1,tmp_kl2,tmp_kh2:  set to something that will pass the checks in constructor (will be updated and is not used)			
				number tmp_kl1 = k_atm-0.0001;
				number tmp_kh1 = k_atm+0.0001;
				number tmp_kl2 = k_atm-0.0002;
				number tmp_kh2 = k_atm+0.0002;
				CORE_INT_VOL.vol_smile_model_vannavolga_ip sm = new CORE_INT_VOL.vol_smile_model_vannavolga_ip( [tmp_kl2,tmp_kl1,k_atm,tmp_kh1,tmp_kh2],[0,0,vol_[i],0,0],
																											s, tau, df , dfq, dt, delta_, delta2_,true);			
				sm_fly_[i] = CORE_INT_FX_OPT.bfly_to_sfly_near(	s ,f, tau ,rr_[i] ,fly_[i] , rr2_[i] ,fly2_[i],
																df ,dfq ,delta_, delta2_,sm, k_c,k_p,vol_p, vol_c);
				QL_REQUIRE(!null(sm_fly_[i]), "invalid smile-fly1 (conversion from broker-fly failed)");
			}
			if(!use_equiv_proc2(i)) {
				sm_fly2_[i] = fly2_[i] ;
				continue;
			}
			number tmp_kl2 = k_p-0.0001;
			number tmp_kh2 = k_c+0.0001;
			number k2_c,k2_p,vol2_p,vol2_c;
			CORE_INT_VOL.vol_smile_model_vannavolga_ip sm = new CORE_INT_VOL.vol_smile_model_vannavolga_ip( [tmp_kl2,k_p,k_atm,k_c,tmp_kh2],[0,vol_p,vol_[i],vol_c,0],
																											s , tau, df , dfq, dt ,delta_,delta2_,true);
			sm_fly2_[i] = CORE_INT_FX_OPT.bfly_to_sfly(	s ,f, tau, rr2_[i] ,fly2_[i] , df ,dfq ,delta2_,sm, false,k2_c,k2_p,vol2_p,vol2_c);
			
			QL_REQUIRE(!null(sm_fly2_[i]), "invalid smile-fly2 (conversion from broker-fly failed)");
		
		}		
	}
}

//----------------------------------------------
// atm_smile_vol  -use if it is required or desired to not interpolate locally such as when there is a
//					mix of brokerflys and smileflys
// OBS! what about a mix of deltatypes?
//----------------------------------------------
/*logical fx_quote_curve_smile.atm_smile_vol(	date 						expiry,
											logical option(nullable) 	allow_extrap,
											logical 					allow_fly_ip,
											out logical 				fly_ip_fail,
											out vector(number) 			res1,
											out vector(number) 			res2)		
{
	if(null(allow_extrap)) allow_extrap = allow_extrap_;
	
	QL_REQUIRE(check_range(expiry,allow_extrap), "outside range of allowed expiry dates");
	
	integer i 	= locate_expiry(expiry);

	if(expiry <= min_expiry() || expiry >= max_expiry() || expiry == exp_[i]) {
		fly_ip_fail = false;
		res2 = null<vector(number)>;
		return atm_smile_vol(expiry,allow_extrap,res1);
	}

	if(allow_fly_ip && is_broker_fly_[i] == is_broker_fly_[i+1]) {
		//here we say it is ok to interpolate linear on either smileflys or brokerflys as long as they are the same
		//(alt. we always interpolate on smileflys i.e. convert brokerflys before any interpolation
		// or vice versa)
		fly_ip_fail = false;
		res2 = null<vector(number)>;
		return atm_smile_vol(expiry,allow_extrap,res1);
	}
	
	fly_ip_fail 		= true;	
	logical is_1_wing 	= type() == atm_curve_type.ATM_SMILE_1_WING;
	
	if(is_1_wing){
		resize(res1,6);
		resize(res2,6);
		res1[0] = (exp_[i]-ref_date_)/365.0;		
		res1[1] = vol_[i];
		res1[2] = rr_[i];
		res1[3] = fly_[i];
		res1[4] = res2[4] = delta_;
		res1[5] = is_broker_fly_[i] ? 1:0;

		res2[0] = (exp_[i+1]-ref_date_)/365.0;
		res2[1] = vol_[i+1];
		res2[2] = rr_[i+1];
		res2[3] = fly_[i+1];
		res2[5] = is_broker_fly_[i+1] ? 1:0;
	}
	else {
		resize(res1,9);
		resize(res2,9);
		res1[0] = (exp_[i]-ref_date_)/365.0;	
		res1[1] = vol_[i];
		res1[2] = rr_[i];
		res1[3] = fly_[i];
		res1[4] = res2[4] = delta_;
		res1[5] = rr2_[i];
		res1[6] = fly2_[i];
		res1[7] = res2[7] = delta2_;
		res1[8] = is_broker_fly_[i] ? 1:0;

		res2[0] = (exp_[i+1]-ref_date_)/365.0;
		res2[1] = vol_[i+1];
		res2[2] = rr_[i+1];
		res2[3] = fly_[i+1];
		res2[5] = rr2_[i+1];
		res2[6] = fly2_[i+1];
		res2[8] = is_broker_fly_[i+1] ? 1:0;
	}

	return true;
	
}*/

//----------------------------------------------
// atm_vol
//----------------------------------------------
logical fx_quote_curve_smile.atm_smile_vol(	date 						expiry,
											logical option(nullable) 	allow_extrap,
											out vector(number) 			res) 
{
	if(null(allow_extrap)) allow_extrap = allow_extrap_;
	
	QL_REQUIRE(check_range(expiry,allow_extrap), "outside range of allowed expiry dates");

	logical is_2_wing 	= type() == fx_atm_curve_type.ATM_SMILE_2_WINGS;
	number tau 			= (expiry-ref_date_)/365.0;
	integer i 			= locate_expiry(expiry);
	
	number atm_vol,rr,smfly,delta,rr2,smfly2,delta2;
	atm_vol=rr=smfly=delta=rr2=smfly2=delta2=null<number>;
	logical _2wing_i;
	
	if(expiry <= min_expiry()) {
		atm_vol = vol_[0];
		rr 		= rr_[0];
		smfly 	= sm_fly_[0];
		_2wing_i	= is_2_wing && _2wing_ok_[0];
		if(_2wing_i) {
			rr2 	= rr2_[0];
			smfly2 	= sm_fly2_[0];
		}
	}
	else if(expiry >= max_expiry()) {
		integer s 	= v_size(vol_);
		atm_vol 	= vol_[s-1];
		rr 			= rr_[s-1];
		smfly 		= sm_fly_[s-1];
		_2wing_i		= is_2_wing && _2wing_ok_[s-1];
		if(_2wing_i) {
			rr2 	= rr2_[s-1];
			smfly2 	= sm_fly2_[s-1];
		}		
	}
	else  if(expiry == exp_[i]) {
		atm_vol = vol_[i];
		rr 		= rr_[i];
		smfly 	= sm_fly_[i];
		_2wing_i	= is_2_wing && _2wing_ok_[1];
		if(_2wing_i) {
			rr2 	= rr2_[i];
			smfly2 	= sm_fly2_[i];
		}
	}
	else {//ip
		vector(number) t_v = ttm() ;		
		atm_vol = ip_vol(t_v,tau);		
		rr 		= ip_lin_sqrt(rr_,t_v,tau);
		smfly	= ip_lin_sqrt(sm_fly_,t_v,tau) ;
		if(null(atm_vol) || null(rr) || null(smfly)) return false;
		_2wing_i	= is_2_wing && _2wing_ok_[i] && _2wing_ok_[i+1];
		if(_2wing_i) {
			rr2 	= ip_lin_sqrt(rr2_,t_v,tau);
			smfly2	= ip_lin_sqrt(sm_fly2_,t_v,tau) ;
			if(null(rr2) || null(smfly2)) return false;
		}
	}

	if(!is_2_wing){
		resize(res,6);
		res[0] = tau;
		res[1] = atm_vol;
		res[2] = rr;
		res[3] = smfly;
		res[4] = delta_;
		res[5] = 0;
	}
	else {
		resize(res,9);
		res[0] = tau;
		res[1] = atm_vol;
		res[2] = rr;
		res[3] = smfly;
		res[4] = delta_;
		res[5] = rr2;
		res[6] = smfly2;
		res[7] = delta2_;
		res[8] = _2wing_i ? 1 : 0;
	}
	return true;
}

//----------------------------------------------
// rr_ss_vols
//----------------------------------------------
void fx_quote_curve_smile.rr_ss_vols(	date 	expiry,
										logical option(nullable) allow_extrap,
										out number atm_vol,
										out number c_vol,
										out number p_vol,
										out number c_vol2,
										out number p_vol2)
{
	if(null(allow_extrap)) allow_extrap = allow_extrap_;
	
	QL_REQUIRE(check_range(expiry,allow_extrap), "outside range of allowed expiry dates");

	logical is_2_wing = type() == fx_atm_curve_type.ATM_SMILE_2_WINGS;
	
	//integer i 	= locate_expiry(expiry);

	vector(number) res;	
	logical ok = atm_smile_vol(expiry,allow_extrap,res);
	QL_REQUIRE(ok, "vol interpolation failed");
	
	atm_vol = res[1];
	CORE_INT_FX_OPT.fx_rr_ss_vols(res[1],res[2],res[3],c_vol, p_vol);
	if(is_2_wing && res[8] == 1)		
		CORE_INT_FX_OPT.fx_rr_ss_vols(res[1],res[5],res[6],c_vol2, p_vol2);
	else
		c_vol2 = p_vol2 = null<number>;
	
}
//----------------------------------------------
// rr_ss_vols
//----------------------------------------------
void fx_quote_curve_smile.rr_ss_vols(	out vector(number) atm_vol,
										out vector(number) c_vol,
										out vector(number) p_vol,
										out vector(number) c_vol2,
										out vector(number) p_vol2)
{
	logical is_2_wing = type() == fx_atm_curve_type.ATM_SMILE_2_WINGS;
	
	integer size = v_size(exp_);
	resize(atm_vol,size);
	resize(c_vol,size);
	resize(p_vol,size);
	resize(c_vol2,size);
	resize(p_vol2,size);
	
	
	for(integer i=0;i<size;i++){
		atm_vol[i] = vol_[i];			
		if(!null(sm_fly_[i]))
			CORE_INT_FX_OPT.fx_rr_ss_vols(vol_[i],rr_[i],sm_fly_[i],c_vol[i], p_vol[i]);
		else
			c_vol[i] = p_vol[i] = null<number>;

		if(is_2_wing && !null(sm_fly2_[i]))
			CORE_INT_FX_OPT.fx_rr_ss_vols(vol_[i],rr2_[i],sm_fly2_[i],c_vol2[i], p_vol2[i]);
		else
			c_vol2[i] = p_vol2[i] = null<number>;		
	}		
}
//----------------------------------------------
// rr_ss_vols
//----------------------------------------------
void fx_quote_curve_smile.rr_ss_strikes(out vector(number) atm_k,
										out vector(number) c_k,
										out vector(number) p_k,
										out vector(number) c_k2,
										out vector(number) p_k2)
{
	vector(number) atm_vol,c_vol, p_vol, c_vol2, p_vol2;
	rr_ss_vols(	atm_vol,c_vol,p_vol,c_vol2,p_vol2);
	
	logical is_2_wing = type() == fx_atm_curve_type.ATM_SMILE_2_WINGS;
	
	integer size = v_size(exp_);
	resize(atm_k,size);
	resize(c_k,size);
	resize(p_k,size);
	resize(c_k2,size);
	resize(p_k2,size);

	number s = r_ts_.spot();
	
	for(integer i=0;i<size;i++){
		number tau 		= (exp_[i]-ref_date_)/365.0;
		number cont_r	= r_ts_.cont_r_dom(tau);
		number cont_u_r	= r_ts_.cont_r_for(tau);
		number df		= r_ts_.df_dom(tau);
		number dfq		= r_ts_.df_for(tau);
		number f 		= r_ts_.fwd_outr(tau);

		atm_k[i] 		= CORE_INT_FX_OPT.atm_strike(atm_type_[i], vol_[i] , s , tau, df , dfq, delta_type_[i]);
		
		if(!null(c_vol) && !null(p_vol))
			CORE_INT_FX_OPT.delta_to_k_cp(	delta_type_[i], delta_,
											s,f,c_vol,p_vol,cont_r,cont_u_r,tau,					  
											c_k[i],p_k[i]);
		else
			c_k[i] = p_k[i] = null<number>;
		
		if(is_2_wing && !null(c_vol2) && !null(p_vol2))
			CORE_INT_FX_OPT.delta_to_k_cp(	delta_type_[i], delta2_,
											s,f,c_vol2,p_vol2,cont_r,cont_u_r,tau,					  
											c_k2[i],p_k2[i]);
		else
			c_k2[i] = p_k2[i] = null<number>;
		
	}		
}
//----------------------------------------------
// atm_delta
//----------------------------------------------
void fx_quote_curve_smile.atm_delta(date 		expiry,
									//number 		vol_atm,
									//number 		s ,
									//number 		cont_r,
									//number 		cont_u_r,
									//number 		k_atm,
									out number delta_c_atm,
									out number delta_p_atm)
{
	number tau 		= (expiry-ref_date_)/365.0;
	number s		= r_ts_.spot();
	number cont_r	= r_ts_.cont_r_dom(tau);
	number cont_u_r	= r_ts_.cont_r_for(tau);
	number df		= r_ts_.df_dom(tau);
	number dfq		= r_ts_.df_for(tau);
	number vol_atm	= atm_vol(expiry,null<logical>);
	number k_atm 	= CORE_INT_FX_OPT.atm_strike(atm_type(expiry), vol_atm , s , tau, df , dfq, delta_type(expiry));
	CORE_INT_FX_OPT.delta( delta_type(expiry), s , k_atm, vol_atm, tau, cont_r, cont_u_r,delta_c_atm,delta_p_atm);
}
//----------------------------------------------
// delta
//----------------------------------------------
void fx_quote_curve_smile.delta(out number delta1, out number delta2)
{
	delta1 = delta_;
	delta2 = delta2_;
	return ;
}


module CORE_INT_FX_OPT
{
	
	//----------------------------------------------
	// fx_risk_reversal
	//----------------------------------------------
	void fx_risk_reversal(	date 	trade_date,
							//date 	spot_date,						
							date 	expiry_date,
							//date 	delivery_date,
							string 	cut,
							number 	delta,
							fx_delta_type 	delta_type,	
							number 	nominal,
							number	spot,
							number 	swap,
							number 	rate_dom,
							number 	rate_for,
							number	call_vol,
							number 	rr_vol_calls_over)//spread
							//string 	model,
							//out vanilla_option_fx call,
							//out vanilla_option_fx put)
	{
		
		//number f 		= spot+swap;
		//number put_vol 	= call_vol - rr_vol_calls_over;
		//calendar cal 	= calendar();

		
		/*call 			= new vanilla_option_fx("USD/SEK","european","call",
												expiry_date,cal,null<number>,nominal) ;
		put 			= new vanilla_option_fx("USD/SEK","european","put",
												expiry_date,cal,null<number>,nominal) ;*/
					
		//call.set_model(option_pricing_model.BS);	
		//put.set_model(option_pricing_model.BS);

		//call.set_mkt_args(trade_date,spot,f,delta,delta_type,call_vol,rate_dom,rate_for);
		//put.set_mkt_args( trade_date,spot,f,delta,delta_type,put_vol,rate_dom,rate_for);
		
		return;
	}

	


	number day_weight(date e, number w_per_cal, calendar cal1, calendar cal2, vector(date) spec, vector(number) spec_w)
	{
		vector(integer) idx = v_find(spec, [e]);
		if(v_size(idx)==1) {
			return spec_w[idx[0]];
		}
		
		number w=0;
		if(!cal1.is_holiday(e))
			w += w_per_cal;
		if(!cal2.is_holiday(e))
			w += w_per_cal;
		return w;
	}

	void vol_accum(date horizon, number base_vol, vector(date) expiry,
				   calendar cal1, calendar cal2, vector(date) spec, vector(number) spec_w,
				   out vector(number) vol, out vector(number) d_vol)
	{
		integer max = expiry[v_size(expiry)-1]-horizon;
		vector(date) e[max];
		vector(number) cal_time[max];
		vector(number) w[max];
		vector(number) var[max];
		resize(d_vol,max);
		
		resize(vol,v_size(expiry));
		number sum = 0;
		for(integer i=0, j=0;i<max;i++){
			e[i] = horizon+i+1;
			cal_time[i] = (i+1)/365.0;
			number dw = day_weight(e[i],0.5,cal1,cal2,spec,spec_w);
			sum += dw; 
			w[i] = sum;
			
			var[i] = w[i]/365.0*base_vol*base_vol;
			if(e[i] == expiry[j]){
				vol[j] = sqrt(var[i]/cal_time[i]);
				++j;
			}
			if(i==0)
				d_vol[i] = var[i];
			else
				d_vol[i] = var[i] - var[i-1] ;
			d_vol[i] = sqrt(d_vol[i]/365.0); 
		}
	}
}
