/*	
	Option vanilla fx vol surface stuff
	Developer: Algorithmica Research, Magnus Nyström			
*/

class enum fx_vol_surface_type option (category: 'Financial/Options/Vanilla/FX')
{
	SURF_CONST 	option(str:"constant vol"),
	SURF_FLAT 	option(str:"flat vol"),
	SURF_SMILE 	option(str:"smile vol") 
};

//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
// class fx_vol_surface  abstract class
//----------------------------------------------
//----------------------------------------------
//----------------------------------------------

class fx_vol_surface	
{
public:	
	virtual number vol_delta(date option(nullable) = null<date>,
						   number option(nullable) = null<number>) = 0;
	virtual number vol_strike(date option(nullable) = null<date>,
						   number option(nullable) = null<number>) = 0;
	virtual fx_vol_surface_type type() = 0;

	virtual void rr_smile_fly(date  option(nullable),out number rr1, out number rr2,
								  out number sf1, out number sf2) = 0;
	virtual void broker_fly(date  option(nullable),out number bf1, out number bf2) = 0;
protected:
	fx_vol_surface();
};

fx_vol_surface.fx_vol_surface() {}


//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
// class fx_vol_surf_const
//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
class fx_vol_surf_const : public fx_vol_surface	option (category: 'Financial/Options/Vanilla/FX')
{
public:
	fx_vol_surf_const(number );
	override number vol_delta(date option(nullable) = null<date>,
						   number option(nullable) = null<number>) ;
	override number vol_strike(date option(nullable) = null<date>,
						   number option(nullable) = null<number>) ;
	override fx_vol_surface_type type() ;
	override void rr_smile_fly(date  option(nullable),out number rr1, out number rr2,
								  out number sf1, out number sf2);
	override void broker_fly(date  option(nullable),out number bf1, out number bf2);
protected:
	number vol_;
};

void fx_vol_surf_const.rr_smile_fly(date  option(nullable),out number rr1, out number rr2,
									out number sf1, out number sf2) { rr1 = rr2 = sf1 = sf2 = 0;}

void fx_vol_surf_const.broker_fly(date  option(nullable),out number bf1, out number bf2) { bf1 = bf1 = 0;}		
//----------------------------------------------
// constructor
//----------------------------------------------
fx_vol_surf_const.fx_vol_surf_const(number vol)
									: fx_vol_surface(), vol_(vol)
{}
//----------------------------------------------
// type
//----------------------------------------------
fx_vol_surface_type fx_vol_surf_const.type()
{
	return fx_vol_surface_type.SURF_CONST;
}
//----------------------------------------------
// vol_delta
//----------------------------------------------
number fx_vol_surf_const.vol_delta(date option(nullable), number option(nullable))
{
	return vol_;
}
//----------------------------------------------
// vol_strike
//----------------------------------------------
number fx_vol_surf_const.vol_strike(date option(nullable), number option(nullable))
{
	return vol_;
}
//----------------------------------------------
// create function
//----------------------------------------------
fx_vol_surf_const fx_vol_surf_const(number option(nullable) vol)
option (category: 'Financial/Options/Vanilla/FX')	
{
	return new fx_vol_surf_const(vol);
}

//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
// class fx_vol_surf_flat
//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
class fx_vol_surf_flat: public fx_vol_surface	option (category: 'Financial/Options/Vanilla/FX')
{
public:
	fx_vol_surf_flat(fx_quote_curve_atm);
	override number 	vol_delta(date option(nullable) = null<date>,
									number option(nullable) = null<number>) ;
	override number 	vol_strike(date option(nullable) = null<date>,
									number option(nullable) = null<number>) ;
	override 			fx_vol_surface_type type() ;
	override void 		rr_smile_fly(date option(nullable),out number rr1, out number rr2,
									out number sf1, out number sf2);
	override void 		broker_fly(date  option(nullable),out number bf1, out number bf2);
	virtual date		ref_date();
protected:
	fx_quote_curve_atm 	quote_curve_;
};

void fx_vol_surf_flat.rr_smile_fly(date  option(nullable),out number rr1, out number rr2,
									out number sf1, out number sf2) { rr1 = rr2 = sf1 = sf2 = 0;}
void fx_vol_surf_flat.broker_fly(date  option(nullable),out number bf1, out number bf2) { bf1 = bf1 = 0;}
//----------------------------------------------
// constructor
//----------------------------------------------
fx_vol_surf_flat.fx_vol_surf_flat(fx_quote_curve_atm quote_curve )
							: fx_vol_surface(), quote_curve_(quote_curve)
{}
//----------------------------------------------
// ref_date
//----------------------------------------------
date fx_vol_surf_flat.ref_date()
{
	return quote_curve_.ref_date();
}
//----------------------------------------------
// type
//----------------------------------------------
fx_vol_surface_type fx_vol_surf_flat.type()
{
	return fx_vol_surface_type.SURF_FLAT;
}
//----------------------------------------------
// vol_delta
//----------------------------------------------
number fx_vol_surf_flat.vol_delta(date option(nullable) expiry, number option(nullable))
{
	quote_curve_.atm_vol(expiry,null<logical> );
}
//----------------------------------------------
// vol_strike
//----------------------------------------------
number fx_vol_surf_flat.vol_strike(date option(nullable) expiry, number option(nullable))
{
	quote_curve_.atm_vol(expiry,null<logical> );
}
//----------------------------------------------
// create function
//----------------------------------------------
fx_vol_surf_flat fx_vol_surf_flat(fx_quote_curve_atm quote_curve)
option (category: 'Financial/Options/Vanilla/FX')
{
	return new fx_vol_surf_flat(quote_curve);
}

class enum fx_vol_model_type  { VANNAVOLGA };

//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
// class fx_vol_surf_smile
//----------------------------------------------
//----------------------------------------------
//----------------------------------------------
class fx_vol_surf_smile : public  fx_vol_surf_flat option (category: 'Financial/Options/Vanilla/FX')
{
public:
	fx_vol_surf_smile(fx_quote_curve_smile, fx_vol_model_type	);
	
	override fx_vol_surface_type 	type() ;
	override number 				vol_delta(date option(nullable), number option(nullable));
	override number 				vol_strike(date option(nullable), number option(nullable));	
	override void 					rr_smile_fly(date option(nullable),out number, out number,out number, out number);
	void							rr_smile_fly(date option(nullable) ,number,out number ,out number);
	
	override void 					broker_fly(date option(nullable),out number , out number );
	number 							broker_fly(date option(nullable),number);

	void							data(date option(nullable),logical, out vector(number),out vector(number) ,out vector(number) );
	number							spot();
	void							int_rates(number option(nullable), out number, out number);
	void							int_rates(date option(nullable), out number, out number);

	number 							delta_put2call(date option(nullable),number);//put delta is positive
	number  						delta_call2put(date option(nullable),number);//put delta is positive
	vector(number)  				delta_put2call(vector(number));//put delta is positive
	vector(number)  				delta_call2put(vector(number));//put delta is positive
protected:

	void 		update_smile(date);
	logical		do_update(date);
	void		init_1wing(	date,number,number,number,number ,number );
	void		init_2wings(date ,number,number,number ,number ,number ,number,number,number );
	logical		use_equiv_proc(number,number);
	logical 	is_1_wing();
	
	number 				s_;
	disc_func 			df_dom_;
	disc_func 			df_for_;
	fx_vol_model_type	vol_mod_t_;
	vector(number) 		quotes_;
	date 				last_expiry_;
	CORE_INT_VOL.vol_smile_model_delta_wing sm_mod_;
	fx_quote_curve_smile 	quote_crv_sm_;
};
//----------------------------------------------
// constructor
//----------------------------------------------
fx_vol_surf_smile.fx_vol_surf_smile(fx_quote_curve_smile 	quote_crv_sm,									
									fx_vol_model_type		vol_mod_t)
									: fx_vol_surf_flat(quote_crv_sm),
									  s_(quote_crv_sm.spot()),
									  df_dom_(quote_crv_sm.df_dom()),
									  df_for_(quote_crv_sm.df_for()),
									  vol_mod_t_(vol_mod_t),									  
									  quotes_(null<vector(number)>), 
									  last_expiry_(null<date>),
									  sm_mod_(null<CORE_INT_VOL.vol_smile_model_delta_wing>),
									  quote_crv_sm_(quote_crv_sm)
{}
//----------------------------------------------
// do_update
//----------------------------------------------
logical	fx_vol_surf_smile.do_update(date expiry)
{
	return null(last_expiry_) || expiry != last_expiry_;
}
//----------------------------------------------
// update_smile
//----------------------------------------------
void fx_vol_surf_smile.update_smile(date expiry)
{	
	vector(number) res;
	logical ok = quote_crv_sm_.atm_smile_vol(expiry,null<logical>,res);
	QL_REQUIRE(ok, "smile quotes from curve not valid");
	
	logical is_q_1_wing = quote_crv_sm_.type() == fx_atm_curve_type.ATM_SMILE_1_WING;
	if(is_q_1_wing) 
		init_1wing(	expiry, res[0],res[1],res[2],res[3],res[4]);	
	else {
		if(res[8] == 1)
			init_2wings(expiry,res[0],res[1],res[2],res[3],res[5],res[6],res[4],res[7]);
		else
			init_1wing(	expiry, res[0],res[1],res[2],res[3],res[4]);//we have only one wing because 2nd wing quotes are missing
	}
	last_expiry_ = expiry;
}
//----------------------------------------------
// init_1wing
//----------------------------------------------
logical fx_vol_surf_smile.is_1_wing()
{
	QL_REQUIRE(!null(sm_mod_), "smile object not initialized");
	return sm_mod_.type() == CORE_INT_VOL.sm_type.DELTA_1WING;
}
//----------------------------------------------
// init_1wing
//----------------------------------------------
void fx_vol_surf_smile.init_1wing(	date  					expiry,
									number 					tau,
									number 					atm_vol,
									number 					rr,
									number 					smfly,									
									number 					delta)
{
	logical	allow_extrap = quote_crv_sm_.allow_extrap();
	number vol_p,vol_c;		
	CORE_INT_FX_OPT.fx_rr_ss_vols(atm_vol,rr,smfly,vol_c, vol_p);

	fx_delta_type fdt 	= quote_crv_sm_.delta_type(expiry);
	atm_type at 		= CORE_INT_FX_OPT.atm_strike_conv(quote_crv_sm_.atm_type(expiry),fdt);
	delta_type dt 		= CORE_INT_FX_OPT.delta_type_conv(fdt);
	
	sm_mod_ 			= new CORE_INT_VOL.vol_smile_model_vannavolga_ip([vol_p, atm_vol, vol_c], s_,tau,
																		df_dom_.disc(tau) ,df_for_.disc(tau),
																		at,dt,
																		delta,null<number>, allow_extrap);
	resize(quotes_,2);
	quotes_[0] 	= rr;
	quotes_[1] 	= smfly;	

}

//----------------------------------------------
// init_2wings
//----------------------------------------------
void fx_vol_surf_smile.init_2wings(	date  					expiry,
									number 					tau,
									number 					atm_vol,
									number 					rr,
									number 					smfly,
									number 					rr2,
									number 					smfly2,									
									number 					delta,
									number 					delta2)
{
	logical	allow_extrap = quote_crv_sm_.allow_extrap();
	number vol_p,vol_c,vol2_c, vol2_p;		
	CORE_INT_FX_OPT.fx_rr_ss_vols(atm_vol,rr,smfly,vol_c, vol_p);
	CORE_INT_FX_OPT.fx_rr_ss_vols(atm_vol,rr2,smfly2,vol2_c, vol2_p);
	
	fx_delta_type fdt 	= quote_crv_sm_.delta_type(expiry);
	atm_type at 		= CORE_INT_FX_OPT.atm_strike_conv(quote_crv_sm_.atm_type(expiry),fdt);
	delta_type dt 		= CORE_INT_FX_OPT.delta_type_conv(fdt);
	
	sm_mod_ 			= new CORE_INT_VOL.vol_smile_model_vannavolga_ip([vol2_p,vol_p, atm_vol, vol_c,vol2_c], s_,tau,
																		df_dom_.disc(tau) ,df_for_.disc(tau),
																		at,dt,
																		delta,delta2, allow_extrap);
	resize(quotes_,4);
	quotes_[0] 	= rr;
	quotes_[1] 	= smfly;
	quotes_[2] 	= rr2;
	quotes_[3] 	= smfly2;	
}

//----------------------------------------------
// spot
//----------------------------------------------
number fx_vol_surf_smile.spot() { return s_;}
//----------------------------------------------
// int_rates
//----------------------------------------------
void fx_vol_surf_smile.int_rates(number option(nullable) tau,
								 out number dom_r,
								 out number for_r)
{
	if(null(tau)) {
		QL_REQUIRE(!null(last_expiry_), "tau not available");
		tau	= (last_expiry_-quote_curve_.ref_date())/365.0; 
	}
	dom_r = df_dom_.cont_rate_spot(tau);
	for_r = df_for_.cont_rate_spot(tau);
}
//----------------------------------------------
// int_rates
//----------------------------------------------
void fx_vol_surf_smile.int_rates(date option(nullable) expiry,
								 out number dom_r,
								 out number for_r)
{
	if(null(expiry)) {
		QL_REQUIRE(!null(last_expiry_), "expiry not available");
		expiry	= last_expiry_;	
	}
	
	number tau	= (expiry-quote_curve_.ref_date())/365.0;
	dom_r 		= df_dom_.cont_rate_spot(tau);
	for_r 		= df_for_.cont_rate_spot(tau);
}

//----------------------------------------------
// delta_put2call
//----------------------------------------------
number fx_vol_surf_smile.delta_put2call(date option(nullable) expiry,
										number delta_put)//delta_put with pos sign
{
	if(null(expiry)) {
		QL_REQUIRE(!null(last_expiry_), "expiry not available");
		expiry	= last_expiry_;
	}
		
	if(do_update(expiry))
		update_smile(expiry);
	
	vector(number) deltas 	= sm_mod_.deltas_put();
	integer i_atm 			= is_1_wing() ? 1 : 2;	
	number sum_atm_delta 	= deltas[i_atm+1]+deltas[i_atm-1] ;	
	return -delta_put+sum_atm_delta;	
}

//----------------------------------------------
// delta_put2call
//----------------------------------------------
vector(number) fx_vol_surf_smile.delta_put2call(vector(number) delta_put)//delta_put with pos sign
{
	vector(date) exp_v = quote_crv_sm_.expiry();
	integer size = v_size(exp_v);
	QL_REQUIRE(size == v_size(delta_put), "invalid vector size");
	
	vector(number) res;
	resize(res,size);
	
	for(integer i=0; i<size;i++) {
		
		if(do_update(exp_v[i]))
			update_smile(exp_v[i]);
		
		vector(number) deltas 	= sm_mod_.deltas_put();
		integer i_atm 			= is_1_wing() ? 1 : 2;	
		number sum_atm_delta 	= deltas[i_atm+1]+deltas[i_atm-1] ;	
		res[i] = -delta_put[i]+sum_atm_delta;
	}
}
//----------------------------------------------
// delta_call2put
//----------------------------------------------
number fx_vol_surf_smile.delta_call2put(date option(nullable) expiry,
										number delta_call)//result = delta_put with pos sign
{
	if(null(expiry)) {
		QL_REQUIRE(!null(last_expiry_), "expiry not available");
		expiry	= last_expiry_;		
	}
	
	if(do_update(expiry))
		update_smile(expiry);
	
	vector(number) deltas 	= sm_mod_.deltas_put();
	integer i_atm 			= is_1_wing() ? 1 : 2;		
	number sum_atm_delta 	= deltas[i_atm+1]+deltas[i_atm-1] ;	
	return -(delta_call-sum_atm_delta);
	
}
//----------------------------------------------
// delta_call2put
//----------------------------------------------
vector(number) fx_vol_surf_smile.delta_call2put(vector(number) delta_call)//result = delta_put with pos sign
{
	vector(date) exp_v = quote_crv_sm_.expiry();
	integer size = v_size(exp_v);
	QL_REQUIRE(size == v_size(delta_call), "invalid vector size");
	
	vector(number) res;
	resize(res,size);
	
	for(integer i=0; i<size;i++) {
		if(do_update(exp_v[i]))
			update_smile(exp_v[i]);
				
		vector(number) deltas 	= sm_mod_.deltas_put();
		integer i_atm 			= is_1_wing() ? 1 : 2;		
		number sum_atm_delta 	= deltas[i_atm+1]+deltas[i_atm-1] ;	
		res[i] = -(delta_call[i]-sum_atm_delta);
	}
}

//----------------------------------------------
// data
//----------------------------------------------
void fx_vol_surf_smile.data(	date option(nullable) expiry,
								logical			 	put_delta,
								out vector(number) 	deltas,										
								out vector(number) 	strikes,										
								out vector(number) 	vols)
{
	if(null(expiry)) {
		QL_REQUIRE(!null(last_expiry_), "expiry not available");
		expiry	= last_expiry_;		
	}
	
	if(do_update(expiry))
		update_smile(expiry);
	
	deltas 	= sm_mod_.deltas_put();
	strikes = sm_mod_.strikes();
	vols 	= sm_mod_.vols();
	if(!put_delta)
		deltas = sm_mod_.delta_put2call(deltas);
}
//----------------------------------------------
// use_equiv_proc
//----------------------------------------------
logical fx_vol_surf_smile.use_equiv_proc(number 	rr,
										 number		atm_vol)
{
	return abs(rr) >= 0.1 * atm_vol;
}


//----------------------------------------------
// rr_smile_fly
//----------------------------------------------
void fx_vol_surf_smile.rr_smile_fly(date option(nullable) expiry,
									out number rr1, out number rr2,
									out number sf1, out number sf2)
{
	if(null(expiry)) {
		QL_REQUIRE(!null(last_expiry_), "expiry not available");
		expiry	= last_expiry_;	
	}
	
	if(do_update(expiry))
		update_smile(expiry);

	vector(number) vols = sm_mod_.vols();
	
	if(is_1_wing()) {
		rr1 = vols[2] - vols[0];
		rr2 = null<number>;
		sf1 = (vols[2] + vols[0])/2.0-vols[1];
		sf2 = null<number>;
	}
	else{
		rr1 = vols[3] - vols[1];
		rr2 = vols[4] - vols[0];
		sf1 = (vols[3] + vols[1])/2.0-vols[2];
		sf2 = (vols[4] + vols[0])/2.0-vols[2];
	}
}

//----------------------------------------------
// rr_smile_fly
//----------------------------------------------
void fx_vol_surf_smile.rr_smile_fly(date option(nullable) expiry,
									number	delta,
									out number rr,
									out number sf)
{
	if(null(expiry)) {
		QL_REQUIRE(!null(last_expiry_), "expiry not available");
		expiry	= last_expiry_;	
	}
	
	if(do_update(expiry))
		update_smile(expiry);

	number sigdc,sigdp,kdc,kdp ;	
	sm_mod_.symm_delta_to_vol(delta,sigdc,kdc, sigdp,kdp);

	//vector(number) vols = sm_mod_.vols();
	number atm_vol 		= sm_mod_.atm_vol();
	rr 					= sigdc-sigdp;
	sf 					= (sigdc + sigdp)/2.0-atm_vol;
}
//----------------------------------------------
// broker_fly
//----------------------------------------------
void fx_vol_surf_smile.broker_fly(	date option(nullable) expiry,
									out number bf1,
									out number bf2)
{
	if(null(expiry)) {
		QL_REQUIRE(!null(last_expiry_), "expiry not available");
		expiry	= last_expiry_;	
	}
	
	if(do_update(expiry))
		update_smile(expiry);

	vector(number) deltas_p = sm_mod_.deltas_put();
	if(is_1_wing()) {		
		bf1 = broker_fly(expiry,deltas_p[0]);
		bf2 = null<number>;
	}
	else {		
		bf1 = broker_fly(expiry,deltas_p[1]);
		bf2 = broker_fly(expiry,deltas_p[0]);
	}
}
//----------------------------------------------
// broker_fly
//----------------------------------------------
number fx_vol_surf_smile.broker_fly(date option(nullable) expiry,
									number	delta)
{
	if(null(expiry)) {
		QL_REQUIRE(!null(last_expiry_), "expiry not available");
		expiry	= last_expiry_;	
	}
	
	if(do_update(expiry))
		update_smile(expiry);

	number tau			= (expiry-quote_crv_sm_.ref_date())/365.0;
	number df 			= df_dom_.disc(tau);
	number dfq 			= df_for_.disc(tau);
	//number cont_u_r 	= -log(dfq)/tau;
	//number cont_r 		= -log(df)/tau;		
	number f 			= s_ * dfq / df;	

	//vector(number) vols = sm_mod_.vols();
	
	number rr,sf;	
	rr_smile_fly(expiry,delta, rr,sf);
	number atm_vol 	= sm_mod_.atm_vol();
	logical use_equiv = use_equiv_proc(rr,atm_vol);
	if(!use_equiv)
		return sf;
	
	number bfly = CORE_INT_FX_OPT.broker_fly(s_ ,f, tau,df ,dfq ,delta, sm_mod_);

	return bfly;
}

//----------------------------------------------
// type
//----------------------------------------------
fx_vol_surface_type fx_vol_surf_smile.type()
{
	return fx_vol_surface_type.SURF_SMILE;
}	
//----------------------------------------------
// vol_delta
//----------------------------------------------
number fx_vol_surf_smile.vol_delta(date option(nullable) expiry,
								   number option(nullable) delta)//nullable because of virtual in root
{
	QL_REQUIRE(!null(delta), "invalid delta");

	if(null(expiry)) {
		QL_REQUIRE(!null(last_expiry_), "expiry not available");
		expiry	= last_expiry_;	
	}
	
	if(do_update(expiry))
		update_smile(expiry);

	number sigdc,sigdp,kdc,kdp ;	
	sm_mod_.symm_delta_to_vol(delta,sigdc,kdc, sigdp,kdp);

	return sigdp;
}
//----------------------------------------------
// vol_strike
//----------------------------------------------
number fx_vol_surf_smile.vol_strike(date option(nullable) expiry,
									number option(nullable)  strike) //nullable because of virtual in root
{
	QL_REQUIRE(!null(strike), "invalid strike");

	if(null(expiry)) {
		QL_REQUIRE(!null(last_expiry_), "expiry not available");
		expiry	= last_expiry_;	
	}
	
	if(do_update(expiry))
		update_smile(expiry);

	return sm_mod_.interpol_strike(strike);
}

//----------------------------------------------
// create function
//----------------------------------------------
fx_vol_surf_smile fx_vol_surf_smile(fx_quote_curve_smile 	atm_curve,									
									fx_vol_model_type		vol_mod_t)
option (category: 'Financial/Options/Vanilla/FX')
{
	return new fx_vol_surf_smile(atm_curve,vol_mod_t);
}