/*	
	Option vol model
	Developer: Algorithmica Research, Magnus Nyström			
*/

module CORE_INT_VOL
{
	
	class enum sm_ip_type  { K,MN_KF,MN_LNKF };
	class enum sm_type  { GENERIC_STRIKE,GENERIC_DELTA, DELTA_1WING,DELTA_2WING };
	//class enum sm_model { SM_IP_LIN,SM_IP_SPL, SM_IP_HER, SM_SABR_0, SM_SABR_05, SM_SABR_1};


	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	// class vol_smile_model
	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	class vol_smile_model 
	{
	public:
		vol_smile_model(vector(number) ,number);
		virtual void 	add_data_point(number,number) = 0;
		virtual number 	interpol_strike(number) ;
		virtual number 	interpol_delta(opt_type,number) ;
		virtual sm_type type() = 0;
		vector(number)	vols();	
	protected:
		void			check_strike_vector(vector(number));
		void			check_vol_vector(vector(number));
		void			check_delta_vector(vector(number));
		void 			add_data_point(out vector(number),number,number) ;
		void			check_data();
				
		vector(number)	v_;		
		number			ttm_;		
	};

	void vol_smile_model.vol_smile_model(	vector(number) 	v, 
											number			ttm)
											: v_(clone_vector(v)), ttm_(ttm)
	{
		check_data();
	}

	vector(number)	vol_smile_model.vols() { return clone_vector(v_);}
	//-------------------------------------------------------
	// check_data	
	//-------------------------------------------------------
	void vol_smile_model.check_data()
	{
		QL_REQUIRE(ttm_ > 0, "invalid time to maturity");
		check_vol_vector( v_);
	}
	//-------------------------------------------------------
	// check_vol_vector	
	//-------------------------------------------------------
	void vol_smile_model.check_vol_vector(vector(number) v)
	{
		QL_REQUIRE(v_size(v) > 0, "invalid vector (vol)");
		for (integer i = 0 ; i < v_size(v) ; i++) {			
			QL_FAIL_COND(null(v_[i]),"invalid smile data (null in volatility)");
			QL_FAIL_COND(v_[i] < 0,"invalid smile data (negative volatility)");
		}
	}
	//-------------------------------------------------------
	// check_strike_vector	
	//-------------------------------------------------------
	void vol_smile_model.check_strike_vector(vector(number) k)
	{
		QL_REQUIRE(v_size(v_) == v_size(k), "vector sizes not equal (strike/vol)");
		for (integer i = 0 ; i < v_size(k) ; i++) {			
			QL_FAIL_COND(null(k[i]),"invalid smile data (strike)"); 
			QL_FAIL_COND(k[i] <= 0,"invalid smile data (negative or zero strike)");					
			QL_FAIL_COND(i>0 && k[i] <= k[i-1],"invalid smile data (strike order)"); 
		}
	}
	//-------------------------------------------------------
	// check_delta_vector	
	//-------------------------------------------------------
	void vol_smile_model.check_delta_vector(vector(number) d)
	{
		QL_REQUIRE(v_size(v_) == v_size(d), "vector sizes not equal (delta/vol)");
		for (integer i = 0 ; i < v_size(d) ; i++) {			
			QL_FAIL_COND(null(d[i]),"invalid smile data (delta vector)"); 
			QL_FAIL_COND(d[i] < 0 || d[i] > 1,"invalid smile data (outside valid range)");					
			QL_FAIL_COND(i>0 && d[i] <= d[i-1],"invalid smile data (delta order)"); 
		}
	}

	//-------------------------------------------------------
	// interpol_strike
	//-------------------------------------------------------
	number vol_smile_model.interpol_strike(number)
	{
		QL_FAIL("function not applicable");
	}
	//-------------------------------------------------------
	// interpol_delta
	//-------------------------------------------------------
	number vol_smile_model.interpol_delta(opt_type,number)
	{
		QL_FAIL("function not applicable");
	}
	//-------------------------------------------------------
	// add_data_point	
	//-------------------------------------------------------
	void vol_smile_model.add_data_point(	out vector(number) x_v,
											number	x,
											number 	v)
	{	
		integer	i = binary_search(round(x,10), round(x_v,10));
		if(i < 0){
			push_back(x_v, x);
			push_back(v_, v);
			sort(v_, x_v);
			sort(x_v, x_v);
		}
		else {
			v_[i] = v;
		}
	}
	
	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	// class vol_smile_model_strike_ip
	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	class vol_smile_model_strike_ip: public vol_smile_model 
	{
	public:
		vol_smile_model_strike_ip(	vector(number) ,vector(number) ,number ,number,
									CORE_INT.ip_type, sm_ip_type, logical);
		override number 	interpol_strike(number) ;
		number 				interpol_strike_shift2atm_(number,number,number) ;
		override void 		add_data_point(number,number );
		override sm_type 	type();
	protected:
		void 				check_data();
		
		vector(number)		k_;
		number				f_;	
		CORE_INT.ip_type	ip_type_;
		logical				allow_extrap_;
		logical				extrap_const_;
		sm_ip_type			sm_ip_type_;
	};

	//-------------------------------------------------------
	// constructor
	//-------------------------------------------------------
	void vol_smile_model_strike_ip.vol_smile_model_strike_ip(vector(number) 	k,
															vector(number) 		v, 
															number 				f,
															number				ttm,
															CORE_INT.ip_type	ip_t,
															sm_ip_type			sm_ip_t,
															logical				allow_extrap)
															: vol_smile_model(v,ttm),
															  k_(clone_vector(k)),f_(f),ip_type_(ip_t), allow_extrap_(allow_extrap),
															  extrap_const_(true), sm_ip_type_(sm_ip_t)
	{
		
		check_data();
	}
	//-------------------------------------------------------
	// check_data	
	//-------------------------------------------------------
	void vol_smile_model_strike_ip.check_data()
	{
		QL_REQUIRE(sm_ip_type_ == sm_ip_type.K
				   || sm_ip_type_ == sm_ip_type.MN_KF
				   || sm_ip_type_ == sm_ip_type.MN_LNKF,"smile ip type not applicable");
		
		check_strike_vector(k_);
		QL_REQUIRE(f_ > 0, "invalid forward price");
	}
	//-------------------------------------------------------
	// type
	//-------------------------------------------------------
	sm_type vol_smile_model_strike_ip.type()
	{
		return sm_type.GENERIC_STRIKE;
	}
	//-------------------------------------------------------
	// interpol_strike
	//-------------------------------------------------------
	number vol_smile_model_strike_ip.interpol_strike(number k)
	{
		number v_ip_o;
		if(sm_ip_type_ == sm_ip_type.K)
			v_ip_o = CORE_INT.ip(k_,v_,k,ip_type_,allow_extrap_,extrap_const_);
		else if(sm_ip_type_ == sm_ip_type.MN_KF) {
			v_ip_o = CORE_INT.ip(k_/f_,v_,k/f_,ip_type_,allow_extrap_,extrap_const_);
			v_ip_o *= f_; 
		}
		else /*if(sm_ip_type_ == sm_ip_type.MN_LNKF) */{
			v_ip_o = CORE_INT.ip(log(k_/f_),v_,log(k/f_),ip_type_,allow_extrap_,extrap_const_);
			v_ip_o = exp(v_ip_o)*f_; 
		}
		
		return v_ip_o;
	}
	
	//-------------------------------------------------------
	// interpol_strike_shift2atm  CHECK THIS
	// alt. for atm consistency is to add atm_strike to k-vector
	//-------------------------------------------------------
	number vol_smile_model_strike_ip.interpol_strike_shift2atm_(number k,
															   number atm_vol,
															   number atm_strike)
	{
		
		if(CORE_INT.is_equal(k,atm_strike))
		   return atm_vol;
		
		number v_ip_o 	= interpol_strike(k);
		number v_ip_atm = interpol_strike(atm_strike);
		number adj 		= atm_vol / v_ip_atm;		
		return v_ip_o * adj; //CHECK THIS
	}
	//-------------------------------------------------------
	// add_data_point	
	//-------------------------------------------------------
	void vol_smile_model_strike_ip.add_data_point(	number	k,
													number 	v)
	{	
		vol_smile_model.add_data_point(	k_,k,v);
	}

	
	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	// class vol_smile_model_delta_wing  -abstract
	// assume input is a fx_style symmetric delta input
	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	class vol_smile_model_delta_wing: public vol_smile_model
	{
	public:
		vol_smile_model_delta_wing(	vector(number) ,vector(number) ,number ,number,number,number,
										..delta_type,number,number option(nullable),logical allow_extrap=true);

		vol_smile_model_delta_wing(	vector(number) ,number ,number,number,number ,atm_type,
										..delta_type,number,number option(nullable) ,logical allow_extrap = true);
		
		override number 	interpol_strike(number) = 0;
		override number 	interpol_delta(opt_type,number) ;
		override void 		add_data_point(number,number );
		override sm_type 	type();
		
		virtual void 		symm_delta_to_vol(number ,out number,out number ,
												out number,out number );
		void 				update_wing(logical,vector(number) ,vector(number));
		number 				atm_vol();
		
		delta_type 			delta_type();
		number 				delta_put2call(number);//put delta is positive
		number 				delta_call2put(number);//put delta is positive
		vector(number)		deltas_put();
		vector(number)		strikes();
		
	protected:
		void				init_delta_v();			
		void 				check_data();
		integer 			index_atm();
		void				index_near(out integer,out integer);
		void				index_far(out integer,out integer);
		
		number 				s_ ;
		number				f_;
		number 				df_ ;
		number 				dfq_;
		number 				cont_u_r_;
		number 				cont_r_;
		vector(number)		k_;
		vector(number)		delta_put_;
		logical				allow_extrap_;
		delta_type  		delta_type_;		
		number 				delta_w1_;
		number 				delta_w2_;
		number 				delta_max_;
	};

	

	//-------------------------------------------------------
	// constructor
	//-------------------------------------------------------
	void vol_smile_model_delta_wing.vol_smile_model_delta_wing(	vector(number) 		k,	//assume middle node is always atm strike
																vector(number) 		v, 	//assume middle node is always atm vol
																number 				s,
																number				ttm,
																number 				df ,
																number 				dfq,
																..delta_type		delta_t,
																number 				delta1w,
																number option(nullable) delta2w,
																logical				allow_extrap)
																: vol_smile_model(v,ttm),
																  s_(s), df_(df), dfq_(dfq),k_(clone_vector(k)),allow_extrap_(allow_extrap),
																  delta_type_(delta_t), delta_w1_(delta1w),delta_w2_(delta2w)
	{
		check_data();		
		check_strike_vector(k_);			
		
		cont_u_r_ 	= -log(dfq_)/ttm_;
		cont_r_ 	= -log(df_)/ttm_;
		f_			= s_*dfq_/df;
		init_delta_v();
	}

	//-------------------------------------------------------
	// constructor k-vector will be calculated (assumes we have valid vol vector)
	//-------------------------------------------------------
	void vol_smile_model_delta_wing.vol_smile_model_delta_wing(	vector(number) 		v, 	//assume middle node is always atm vol and wings are putvol and callvol
																number 				s,
																number				ttm,
																number 				df ,
																number 				dfq,
																atm_type			atm_t,
																..delta_type		delta_t,
																number 				delta1w,
																number option(nullable) delta2w,
																logical				allow_extrap)
																: vol_smile_model(v,ttm),
																  s_(s), df_(df), dfq_(dfq),allow_extrap_(allow_extrap),
																  delta_type_(delta_t), delta_w1_(delta1w),delta_w2_(delta2w)
	{
		check_data();		
		resize(k_,v_size(v_));					
		
		cont_u_r_ 		= -log(dfq_)/ttm_;
		cont_r_ 		= -log(df_)/ttm_;
		f_				= s_*dfq_/df;		
		integer i_atm 	= index_atm();
		k_[i_atm] 		= black_scholes_atm_strike(atm_t, v_[i_atm] , s_ , ttm_, df_ , dfq_);
			
		black_scholes_delta_to_k_cp(delta_type_,delta_w1_,s_,f_,v_[i_atm+1],v_[i_atm-1],cont_r_,cont_u_r_,ttm_,k_[i_atm+1],k_[i_atm-1]);
		
		if(type() == sm_type.DELTA_2WING) 
			black_scholes_delta_to_k_cp(delta_type_,delta_w2_,s_,f_,v_[i_atm+2],v_[i_atm-2],cont_r_,cont_u_r_,ttm_,k_[i_atm+2],k_[i_atm-2]);					

		check_strike_vector(k_);
		init_delta_v();
	}
	
	//-------------------------------------------------------
	// check_data	
	//-------------------------------------------------------
	void vol_smile_model_delta_wing.check_data()
	{
	
		integer size = v_size(v_);
		QL_REQUIRE(size == 3 || size == 5, "invalid vector size");		
		QL_REQUIRE(delta_w1_ > 0 && delta_w1_ < 1, "invalid wing1-delta (must be >0 and <1)");
				
		if(type() == sm_type.DELTA_2WING){
			QL_REQUIRE(!null(delta_w2_), "missing wing2-delta");
			QL_REQUIRE(delta_w2_ > 0 && delta_w2_ < 1, "invalid wing2-delta (must be >0 and <1)");
			QL_REQUIRE(delta_w2_ < delta_w1_, "invalid wing2-delta (>= wing1-delta)");
		}
		QL_REQUIRE(s_ > 0, "invalid spot price");
		QL_REQUIRE(df_ > 0, "invalid discount factor (df)");
		QL_REQUIRE(dfq_ > 0, "invalid discount factor (dfq)");
	}
	//-------------------------------------------------------
	// init_delta_v	(requires atm strime and atm vol)
	//-------------------------------------------------------
	void vol_smile_model_delta_wing.init_delta_v()
	{
		integer size = v_size(v_);
		resize(delta_put_,size);
		integer i_atm 	= index_atm();
		
		number delta_p_atm,delta_c_atm;
		black_scholes_delta( delta_type_, s_ , k_[i_atm], v_[i_atm], ttm_, cont_r_, cont_u_r_,delta_c_atm, delta_p_atm);	
		delta_max_ 			= delta_c_atm - delta_p_atm;				
		delta_put_[i_atm-1] = delta_w1_;
		delta_put_[i_atm] 	= -delta_p_atm;
		delta_put_[i_atm+1] = delta_max_-delta_w1_;
		if(type() == sm_type.DELTA_2WING) {
			delta_put_[i_atm-2] = delta_w2_;
			delta_put_[i_atm+2] = delta_max_-delta_w2_;
		}
		check_delta_vector(delta_put_);		
	}
	//-------------------------------------------------------
	// deltas_put
	//-------------------------------------------------------
	vector(number)	vol_smile_model_delta_wing.deltas_put() { return clone_vector(delta_put_);}
	//-------------------------------------------------------
	// strikes
	//-------------------------------------------------------
	vector(number)	vol_smile_model_delta_wing.strikes() { return clone_vector(k_);}
	
	//----------------------------------------------
	// delta_put2call
	//----------------------------------------------
	number vol_smile_model_delta_wing.delta_put2call(number delta_put)//delta_put with pos sign
	{
		logical is1wing 	= type() == sm_type.DELTA_1WING ;
		integer i_atm 		= is1wing ? 1 : 2;	
		number sum_atm_delta = delta_put_[i_atm+1]+delta_put_[i_atm-1] ;	
		return -delta_put+sum_atm_delta;
	}
	//----------------------------------------------
	// delta_call2put
	//----------------------------------------------
	number vol_smile_model_delta_wing.delta_call2put(number delta_call)//result = delta_put with pos sign
	{
		logical is1wing 	= type() == sm_type.DELTA_1WING ;
		integer i_atm 		= is1wing ? 1 : 2;		
		number sum_atm_delta = delta_put_[i_atm+1]+delta_put_[i_atm-1] ;	
		return -(delta_call-sum_atm_delta);
	}
	//-------------------------------------------------------
	// type
	//-------------------------------------------------------
	sm_type vol_smile_model_delta_wing.type()
	{
		return v_size(v_) == 3 ? sm_type.DELTA_1WING : sm_type.DELTA_2WING;
	}
	//-------------------------------------------------------
	// add_data_point	
	//-------------------------------------------------------
	void vol_smile_model_delta_wing.add_data_point(	number	k,
													number 	v)
	{	
		integer	i = binary_search(round(k,10), round(k_,10));
		QL_REQUIRE(i>=0, "adding data points not permitted for vannavolga");		
		v_[i] = v;		
	}
	//----------------------------------------------
	//update_near_wing
	//----------------------------------------------
	void vol_smile_model_delta_wing.update_wing(	logical 		near,
													vector(number) 	strike,
													vector(number) 	vol)
	{
		logical delta_1w = type() == sm_type.DELTA_1WING;
		
		integer i_atm = index_atm();
		if(near || delta_1w){
			k_[i_atm-1] = strike[0];		
			k_[i_atm+1] = strike[1];
			v_[i_atm-1] = vol[0];		
			v_[i_atm+1] = vol[1];
		}
		else {			
			k_[i_atm-2] = strike[0];		
			k_[i_atm+2] = strike[1];
			v_[i_atm-2] = vol[0];		
			v_[i_atm+2] = vol[1];
		}
	}
	//----------------------------------------------
	//atm_vol
	//----------------------------------------------
	number vol_smile_model_delta_wing.atm_vol()
	{
		return v_[index_atm()] ;
	}
	//-------------------------------------------------------
	// index_atm	
	//-------------------------------------------------------
	integer vol_smile_model_delta_wing.index_atm()
	{
		return type() == sm_type.DELTA_1WING ? 1 : 2;
	}
	//----------------------------------------------
	//delta_type
	//----------------------------------------------
	delta_type vol_smile_model_delta_wing.delta_type()
	{
		return delta_type_ ;
	}
	//-------------------------------------------------------
	// index_near	
	//-------------------------------------------------------
	void vol_smile_model_delta_wing.index_near(out integer i1,out integer i2)
	{
		if(type() == sm_type.DELTA_1WING) {
			i1 = 0;
			i1 = 2;
		}
		else {
			i1 = 1;
			i1 = 3;
		}
	}
	//-------------------------------------------------------
	// index_far	
	//-------------------------------------------------------
	void vol_smile_model_delta_wing.index_far(out integer i1,out integer i2)
	{
		if(type() == sm_type.DELTA_1WING) {
			i1 = 0;
			i1 = 2;
		}
		else {
			i1 = 0;
			i1 = 4;
		}
	}
	
	//----------------------------------------------
	//symm_delta_to_vol  (delta --> k --> vol)  (--> override for delta interp functions)
	//----------------------------------------------
	void vol_smile_model_delta_wing.symm_delta_to_vol(	number  		delta,											
														out number 		vol_c,
														out number 		k_c,
														out number		vol_p,
														out number 		k_p)
	{		
		QL_REQUIRE(delta > 0 && delta < delta_max_, "invalid delta");
		
		integer i_atm = index_atm();
		//number d;
		if(type() == sm_type.DELTA_2WING && (delta <= delta_w2_ || delta >= delta_w2_)){
			vol_c 	= v_[i_atm+2];//initial guess = 10delta call vol
			vol_p 	= v_[i_atm-2];//initial guess = 10delta put vol
		}
		else {
			vol_c 	= v_[i_atm+1];//initial guess = 25delta call vol
			vol_p 	= v_[i_atm-1];//initial guess = 25delta put vol
		}
		try{
			for(integer i=0;i<5;i++){//should be ok			
				black_scholes_delta_to_k_cp(delta_type_,delta,s_,f_,vol_c,vol_p,cont_r_,cont_u_r_, ttm_,k_c,k_p);			
				vol_c = interpol_strike(k_c);
				vol_p = interpol_strike(k_p);
			}
		}
		catch {
			vol_c = vol_p = null<number>; 
		}
	}
	//----------------------------------------------
	//interpol_delta
	//----------------------------------------------
	number vol_smile_model_delta_wing.interpol_delta(opt_type type,number delta)
	{
		if(type == opt_type.OPT_PUT && delta < 0)
			delta = -delta;
		
		number 	vol_c,k_c,vol_p,k_p;
		symm_delta_to_vol(delta,vol_c,k_c,vol_p,k_p);
		return type == opt_type.OPT_PUT ? vol_p : vol_c;
	}

	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	// class vol_smile_model_vannavolga_ip
	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	class vol_smile_model_vannavolga_ip: public vol_smile_model_delta_wing  
	{
	public:
		vol_smile_model_vannavolga_ip(	vector(number) ,vector(number) ,number ,number,number,number,
										..delta_type,number,number option(nullable),logical allow_extrap=true);

		vol_smile_model_vannavolga_ip(	vector(number) ,number ,number,number,number ,atm_type,
										..delta_type,number,number option(nullable) ,logical allow_extrap = true);
		override number interpol_strike(number);
	};

	//-------------------------------------------------------
	// constructor
	//-------------------------------------------------------
	void vol_smile_model_vannavolga_ip.vol_smile_model_vannavolga_ip(vector(number) 	k,	//assume middle node is always atm strike
																	vector(number) 		v, 	//assume middle node is always atm vol
																	number 				s,
																	number				ttm,
																	number 				df ,
																	number 				dfq,
																	..delta_type		delta_t,
																	number 				delta1w,
																	number option(nullable) delta2w,
																	logical				allow_extrap)
																	: vol_smile_model_delta_wing(k,v,s,ttm,df,dfq,delta_t,delta1w, delta2w, allow_extrap)
	{}

	//-------------------------------------------------------
	// constructor k-vector will be calculated (assumes we have valid vol vector)
	//-------------------------------------------------------
	void vol_smile_model_vannavolga_ip.vol_smile_model_vannavolga_ip(	vector(number) 		v, 	//assume middle node is always atm vol and wings are putvol and callvol
																		number 				s,
																		number				ttm,
																		number 				df ,
																		number 				dfq,
																		atm_type			atm_t,
																		..delta_type		delta_t,
																		number 				delta1w,
																		number option(nullable) delta2w,
																		logical				allow_extrap)
																		: vol_smile_model_delta_wing(v,s,ttm,df,dfq,atm_t,delta_t,delta1w, delta2w, allow_extrap)
	{}
	
	//----------------------------------------------
	//interpol_strike
	// (fx_vv_smile_interp only defined for 3 strikes, solution below for 5 strikes
	//  is a "pragmatic" solution to avoid non-smoothness!?!
	//----------------------------------------------
	number vol_smile_model_vannavolga_ip.interpol_strike(number k)
	{
		QL_REQUIRE(k > 0 , "invalid strike");
		
		integer i_atm = index_atm();

		if(!allow_extrap_){
			if(type() == sm_type.DELTA_1WING)
				QL_REQUIRE(k >= k_[i_atm-1] && k <= k_[i_atm+1] , "invalid strike (extrapolation setting is set to off )");
			else 
				QL_REQUIRE(k >= k_[i_atm-2] && k <= k_[i_atm+2] , "invalid strike (extrapolation setting is set to off )");
		}
		
		if(type() == sm_type.DELTA_1WING || (k >= k_[i_atm-1] && k <= k_[i_atm+1]))//near wing ip
			return CORE_INT_OPT.vannavolga_smile_interp(s_ , ttm_, k ,
														k_[i_atm-1], v_[i_atm-1] ,
														k_[i_atm] ,  v_[i_atm] ,
														k_[i_atm+1], v_[i_atm+1] ,
														df_ ,  dfq_);

		
		else if(k <= k_[i_atm-2] || k >= k_[i_atm+2])//far wing ip
			return CORE_INT_OPT.vannavolga_smile_interp(s_ , ttm_, k ,
														k_[i_atm-2], v_[i_atm-2] ,
														k_[i_atm] ,  v_[i_atm] ,
														k_[i_atm+2], v_[i_atm+2] ,
														df_ ,  dfq_);
		else {//weighted ip between far and near
			number v_near = CORE_INT_OPT.vannavolga_smile_interp(s_ , ttm_, k ,
																k_[i_atm-1], v_[i_atm-1] ,
																k_[i_atm] ,  v_[i_atm] ,
																k_[i_atm+1], v_[i_atm+1] ,
																df_ ,  dfq_);
			number v_far = CORE_INT_OPT.vannavolga_smile_interp(s_ , ttm_, k ,
																k_[i_atm-2], v_[i_atm-2] ,
																k_[i_atm] ,  v_[i_atm] ,
																k_[i_atm+2], v_[i_atm+2] ,
																df_ ,  dfq_);
			
			if(k >= k_[i_atm-2] && k <= k_[i_atm-1]) {
				number prop = (k-k_[i_atm-2])/(k_[i_atm-1]-k_[i_atm-2]);			
				return prop * v_near + (1.0-prop)*v_far;
			}
			else {
				number prop = (k-k_[i_atm+1])/(k_[i_atm+2]-k_[i_atm+1]);
				return prop * v_far + (1.0-prop)*v_near;
			}
		}
	}
	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	// class vol_smile_model_delta_ip
	// OBS! currently it is assumed that deltatype = delta_type.UNADJ_FWD
	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	class vol_smile_model_delta_ip: public vol_smile_model 
	{
	public:
		vol_smile_model_delta_ip(	vector(number) ,vector(number) ,number,
									CORE_INT.ip_type, logical);
		override void 		add_data_point(number,number);
		override number 	interpol_delta(opt_type,number) ;
		override sm_type 	type();
		delta_type 			delta_type();
		
	protected:
		vector(number) 		delta_put_;		
		CORE_INT.ip_type	ip_type_;
		logical				allow_extrap_;
		logical				extrap_const_;

	};

	void vol_smile_model_delta_ip.vol_smile_model_delta_ip(	vector(number) 		delta_put,//no sign
															vector(number) 		v, 
															//number 				f,
															number				ttm,
															CORE_INT.ip_type	ip_t,
															logical				allow_extrap)
															: vol_smile_model(v,ttm),
															  delta_put_(clone_vector(delta_put)), ip_type_(ip_t), allow_extrap_(allow_extrap),
															  extrap_const_(true)
	{
		check_delta_vector(delta_put_);
	}
	//-------------------------------------------------------
	// delta_type
	//-------------------------------------------------------
	delta_type vol_smile_model_delta_ip.delta_type()
	{
		return ..delta_type.UNADJ_FWD;
	}
	//-------------------------------------------------------
	// type
	//-------------------------------------------------------
	sm_type vol_smile_model_delta_ip.type()
	{
		return sm_type.GENERIC_DELTA;
	}
	//-------------------------------------------------------
	// interpol_delta
	//-------------------------------------------------------
	number vol_smile_model_delta_ip.interpol_delta(opt_type ot,number d)
	{
		number v_ip_put;
		v_ip_put = CORE_INT.ip(delta_put_,v_,d,ip_type_,allow_extrap_,extrap_const_);
		if(ot == opt_type.OPT_PUT)
			return v_ip_put;
		else
			return black_scholes_delta_fwd_p2c(v_ip_put);//assume deltatype = delta_type.UNADJ_FWD
	}

	//-------------------------------------------------------
	// add_data_point	
	//-------------------------------------------------------
	void vol_smile_model_delta_ip.add_data_point(	number	delta,
													number 	v)
	{	
		vol_smile_model.add_data_point(	delta_put_,delta,v);
	}	



	
	vector(number) rho_bounds = [-0.99,0.99]; 
	vector(number) volvol_bounds  = [0,100]; 	
	vector(number) alpha_bounds  = [0.001, 10]; 
	class enum min_method {MM_L1,MM_L2, MM_DHSA};

	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	// class vol_smile_model_sabr  
	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	class vol_smile_model_sabr: public vol_smile_model 
	{
	public:
		vol_smile_model_sabr(	vector(number) ,vector(number) ,number ,number	,number	,logical, 
								number	option(nullable) ,vector(number) option(nullable),min_method ,logical );
		void 				set_guess(vol_smile_model_sabr);
		override void 		add_data_point(number,number );
		override number 	interpol_strike(number);
		override sm_type 	type();
	protected:
		void			init();
		void			check_data();
		number 			alpha(number, number);

		vector(number) 	b76vol_free(vector(number) param );
		vector(number) 	b76vol_fix_atm(vector(number) param) ;
		number 			b76vol_free_dhsa(vector(number) param );
		number 			b76vol_fix_atm_dhsa(vector(number) param) ;
		
		logical 		param_cal(out string ) ;
		vector(number) 	param_cal_l1l2();
		vector(number) 	param_cal_dhsa();

		vector(number)	k_;
		number			f_;
		logical			fit_atm_vol_;
		number 			atm_vol_;
		vector(number) 	weights_;
		
		number 			beta_;
		number			alpha_;
		number 			rho_;
		number 			volvol_;	

		min_method		min_method_;
		number 			alpha_guess_;
		number 			rho_guess_;
		number 			volvol_guess_;	
		
		number 			min_norm_;
		logical			calibrated_;
	};

	void vol_smile_model_sabr.vol_smile_model_sabr(	vector(number) 			k,
													vector(number) 			v, 
													number 					f,
													number					ttm,
													number					beta,
													logical					fit_atm_vol,
													number	option(nullable) 		atm_vol,
													vector(number) option(nullable) weights,
													min_method 				min_method,
													logical 				delay_cal)
													:vol_smile_model(v, ttm),
													 k_(clone_vector(k)),f_(f),fit_atm_vol_(fit_atm_vol),atm_vol_(atm_vol),beta_(beta),
													 min_method_(min_method), calibrated_(false) 						
	{
		if(!null(weights))
			weights_ = clone_vector(weights);
		else {
			integer size = v_size(v_);
			resize(weights_,size);
			for (integer i = 0 ; i < size; i++)
				weights_[i] = 1.0;
		}
		check_data();
		init();
		if(!delay_cal){
			string err_msg;
			QL_FAIL_COND(!vol_smile_model_sabr.param_cal(err_msg), err_msg);
		}
	}
	//-------------------------------------------------------
	// check_data	
	//-------------------------------------------------------
	void vol_smile_model_sabr.check_data()
	{
		integer size = v_size(v_);
		QL_REQUIRE(size == v_size(weights_),"invalid weights (vector size)");
		for (integer i = 0 ; i < size ; i++) {						
			QL_FAIL_COND(null(weights_[i]) || weights_[i] < 0,"invalid weights (null or < 0)"); 
		}

		check_strike_vector(k_);
		
		QL_FAIL_COND(fit_atm_vol_ && null(atm_vol_),"invalid atm_vol");
		QL_REQUIRE(f_ > 0, "invalid forward price");
	}
	//-------------------------------------------------------
	// type
	//-------------------------------------------------------
	sm_type vol_smile_model_sabr.type()
	{
		return sm_type.GENERIC_STRIKE;
	}
	//-------------------------------------------------------
	// add_data_point	
	//-------------------------------------------------------
	void vol_smile_model_sabr.add_data_point(number	k,
											number 	v)
	{	
		vol_smile_model.add_data_point(	k_,k,v);
	}
	//-------------------------------------------------------
	// init
	//-------------------------------------------------------
	void vol_smile_model_sabr.init()
	{
		rho_guess_ 		= -0.5;
		volvol_guess_ 	= 0.5;
		alpha_guess_ 	= alpha(rho_guess_, volvol_guess_);		
	}
	//-------------------------------------------------------
	// alpha
	//-------------------------------------------------------
	number vol_smile_model_sabr.alpha(	number rho, 
									number volvol)
	{
		return sabr_alpha(f_, beta_,rho,volvol,ttm_, atm_vol_);
	}
	//-------------------------------------------------------
	// alpha
	//-------------------------------------------------------
	void vol_smile_model_sabr.set_guess(vol_smile_model_sabr other)
	{
		if(other.calibrated_){
			rho_guess_ = other.rho_;
			volvol_guess_ = other.volvol_;
			alpha_guess_ = other.alpha_;
		}			
	}
	//-------------------------------------------------------
	// sabr_param_cal
	//-------------------------------------------------------
	logical vol_smile_model_sabr.param_cal(out string err_msg) 
	{
		try{
			vector(number) param;
			if(min_method_ == min_method.MM_DHSA)
				param = param_cal_dhsa();
			else
				param = param_cal_l1l2();
	
			rho_ = param[0];
			volvol_ = param[1];
			if(!fit_atm_vol_)					
				alpha_ = param[2];
			
			calibrated_ = !null(min_norm_);
			return calibrated_;
		}
		catch {
			min_norm_ = null<number>;
			err_msg = err.message();
			return false;
		}
	}	
	//-------------------------------------------------------
	// sabr_param_cal
	//-------------------------------------------------------
	vector(number) vol_smile_model_sabr.param_cal_dhsa() 
	{		
		vector(number) param;
		min_norm_ = null<number>;				
		integer n_coolings = 100; 
		integer max_per_cooling = 10; 
		number T0 = 0.01; 
		number dT = 0.9;
		
		if(fit_atm_vol_){
			vector(number) lb 	= [rho_bounds[0],volvol_bounds[0]];
			vector(number) ub 	= [rho_bounds[1],volvol_bounds[1]];
			param 				= [ rho_guess_, volvol_guess_];							
			min_dhsa(&b76vol_fix_atm_dhsa, this, param, lb, ub, n_coolings, max_per_cooling, T0, dT) ;				
			vector(number) v 	= b76vol_fix_atm(param);
			min_norm_ 			= v_sum(sqr(v - v_)); 											
		}
		else{
			vector(number) lb 	= [rho_bounds[0],volvol_bounds[0], alpha_bounds[0]];
			vector(number) ub 	= [rho_bounds[1],volvol_bounds[1], alpha_bounds[1]];						
			param 				= [ rho_guess_, volvol_guess_, alpha_guess_];				
			min_dhsa(&b76vol_free_dhsa, this, param, lb, ub, n_coolings, max_per_cooling, T0, dT) ;							
			vector(number) v 	= b76vol_free(param);				
			min_norm_ 			= v_sum(sqr(v - v_)); 
		}
		if(!null(param[1]) && param[1] < 0 ){
			param[0] = -param[0];
			param[1] = -param[1];
		}
		return param;	
	}

	//-------------------------------------------------------
	// sabr_param_cal
	//-------------------------------------------------------
	vector(number) vol_smile_model_sabr.param_cal_l1l2() 
	{		
		vector(number) param;		
		min_norm_ = null<number>;
		
		if(fit_atm_vol_){
			vector(number) lb 	= [rho_bounds[0],volvol_bounds[0]];
			vector(number) ub 	= [rho_bounds[1],volvol_bounds[1]];
			param 				= [ rho_guess_, volvol_guess_];
			
			if(min_method_ == min_method.MM_L2)
				min_l2(&b76vol_fix_atm, this, param, v_, weights_, lb, ub);
			else if(min_method_ == min_method.MM_L1)
				min_l1(&b76vol_fix_atm, this, param, v_, weights_, lb, ub);
			else
				QL_FAIL("internal error (sabr)");
				
			vector(number) v 	= b76vol_fix_atm(param);
			min_norm_ 			= v_sum(sqr(v - v_)); 					
		}
		else{
			vector(number) lb 	= [rho_bounds[0],volvol_bounds[0], alpha_bounds[0]];
			vector(number) ub 	= [rho_bounds[1],volvol_bounds[1], alpha_bounds[1]];						
			param 				= [ rho_guess_, volvol_guess_, alpha_guess_];	
			
			if(min_method_ == min_method.MM_L2)
				min_l2(&b76vol_free, this, param, v_, weights_, lb, ub);
			else if(min_method_ == min_method.MM_L1)
				min_l1(&b76vol_free, this, param, v_, weights_, lb, ub);
			else
				QL_FAIL("internal error (sabr)");				
			vector(number) v 	= b76vol_free(param);				
			min_norm_ 			= v_sum(sqr(v - v_)); 
		}
		if(!null(param[1]) && param[1] < 0 ){
			param[0] = -param[0];
			param[1] = -param[1];
		}
		return param;	
	}

	//-------------------------------------------------------
	// b76vol_free
	//-------------------------------------------------------
	vector(number) vol_smile_model_sabr.b76vol_free(vector(number) param) 
	{				
		return sabr_blackvol_obloj(f_, k_, ttm_, [param[2], beta_, param[0], param[1]]);
	}
	//-------------------------------------------------------
	// b76vol_fix_atm
	//-------------------------------------------------------
	vector(number) vol_smile_model_sabr.b76vol_fix_atm(vector(number) param) 
	{									
		alpha_ = alpha(param[0],param[1]);
		QL_FAIL_COND(null(alpha_),"invalid alpha (sabr)");						
						
		return sabr_blackvol_obloj(f_, k_, ttm_, [alpha_, beta_, param[0], param[1]]);
	}
	//-------------------------------------------------------
	// b76vol_free_dhsa
	//-------------------------------------------------------
	number vol_smile_model_sabr.b76vol_free_dhsa(vector(number) param) 
	{				
		vector(number) v = sabr_blackvol_obloj(f_, k_, ttm_, [param[2], beta_, param[0], param[1]]);					
		return sqr((v - v_)*weights_);
	}
	//-------------------------------------------------------
	// b76vol_fix_atm_dhsa
	//-------------------------------------------------------
	number vol_smile_model_sabr.b76vol_fix_atm_dhsa(vector(number) param) 
	{									
		alpha_ = alpha(param[0],param[1]);
		QL_FAIL_COND(null(alpha_),"invalid alpha (sabr)");						
						
		vector(number) v = sabr_blackvol_obloj(f_, k_, ttm_, [alpha_, beta_, param[0], param[1]]);
		return sqr((v - v_)*weights_);
	}

	//-------------------------------------------------------
	// interpol_strike
	//-------------------------------------------------------
	number vol_smile_model_sabr.interpol_strike(number k)
	{
		if(!calibrated_){
			string err_msg;
			if(!param_cal(err_msg))
				return null<number>;			
		}

		number alpha 	= CORE_INT_VOL.sabr_alpha(f_, beta_,rho_,volvol_,ttm_, atm_vol_);
		number vol 		= sabr_blackvol_obloj(f_, k, ttm_, alpha, beta_, rho_, volvol_);
		return vol;
	}
}

module CORE_INT_FX_OPT
{
	//----------------------------------------------
	//equivalent fly procedure 4.9.1 chapter 4, Castagna, s. 118
	//broker_fly --> smile_fly
	// PROBLEM IF near_wing==false combined with vannavolga interpol (vv interpol only ip with near wing???)
	//this func only defined for rr and bfly att anchor points i.e. 25d or 10d
	//----------------------------------------------
	number bfly_to_sfly( number s ,
						 number f,
						 number tau,
						 number rr ,
						 number bfly ,//broker_fly
						 number df ,
						 number dfq ,
						 number  delta,
						 CORE_INT_VOL.vol_smile_model_delta_wing sm,//must have proper data to start with and --> near wing will be updated
						 logical near_wing,
						 out number KC,
						 out number KP,
						 out number VP,
						 out number VC)
	 
	{
		fx_delta_type dt = delta_type_conv( sm.delta_type());
		//****************
		//castagna STEP 1: calc KA (here an input) 
		//****************
		number cont_u_r = -log(dfq)/tau;
		number cont_r 	= -log(df)/tau;
		//****************
		//castagna STEP 2: calc as if bfly == smilefly, consistent vols
		//****************
		//number KA = sm.atm_strike();
		number VA = sm.atm_vol();
		//number VC, VP;
		fx_rr_ss_vols(VA,rr,bfly,VC, VP);

		//****************
		//castagna STEP 3: calc implied strikes, Formula 2.46 chapter 2
		//****************
		//strikes using consistent vols for call and put
		delta_to_k_cp(dt,delta,s,f,VC,VP,cont_r,cont_u_r, tau,KC,KP);
		
		//****************
		//castagna STEP 4: calculate strikes for the fly with unique vol
		//****************
		//vol for call and put in brokerfly
		number vol_bfly = VA + bfly; //market covention for the fly volatility		
				
		//strikes using bfly vol i.e. same vol for call and put
		number k_dd_bc,k_dd_bp;
		delta_to_k_cp(dt,delta,s,f,vol_bfly,vol_bfly,cont_r,cont_u_r,tau,k_dd_bc,k_dd_bp);		
		sm.update_wing(near_wing,[KP,KC],[VP,VC]);
		//****************
		//castagna STEP 5: 
		//****************
		//interpolate vol from strikes implied by bflyvol
		number vol_ip_bc = sm.interpol_strike(k_dd_bc);
		number vol_ip_bp = sm.interpol_strike(k_dd_bp);
		
		//calc prices 
		number fp;
		number call25c, call25b, put25p, put25b;
		//calc prices for new strike and interp vols
		black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_ip_bc,  cont_r, cont_u_r, tau,tau,  fp, call25c);		
		black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_ip_bp,  cont_r, cont_u_r, tau,tau,  fp, put25p);
		number stgl_mkt = call25c + put25p;
		
		//calc prices for new strike and bfly vol
		black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, call25b);
		black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, put25b);
		number stgl_conv = call25b + put25b;
		
		//diff between strangle at market vols and the same strangle with conventional unique vol		
		number f0 = stgl_mkt - stgl_conv;

		if (abs(f0) < 0.0000001 * s) 
			return bfly;
		

		bfly = bfly + 0.0001;
		number dfly = 0.0001;

		//while (abs(f0) > 0.00000001 * s){
		while (abs(f0) > 0.0000001 * s){/*orig*/
			
			fx_rr_ss_vols(VA,rr,bfly,VC, VP);
			delta_to_k_cp(dt,delta, s,f,VC,VP, cont_r,cont_u_r,tau, KC,KP);
			sm.update_wing(near_wing,[KP,KC],[VP,VC]);
			
			delta_to_k_cp(dt,delta,s,f,vol_bfly,vol_bfly,cont_r,cont_u_r,tau, k_dd_bc,k_dd_bp);

			vol_ip_bc = sm.interpol_strike(k_dd_bc);
			vol_ip_bp = sm.interpol_strike(k_dd_bp);
		
			black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_ip_bc,  cont_r, cont_u_r, tau,tau,  fp, call25c);			
			black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_ip_bp,  cont_r, cont_u_r, tau,tau,  fp, put25p);
			
			black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, call25b);	
			black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, put25b);
		
			number fu		= (call25c + put25p) - (call25b + put25b);
			number df_der 	= (fu - f0) / dfly;		
			bfly 			= bfly - fu / df_der;
			dfly 			= -fu / df_der;
			f0 				= fu;

		}
		return bfly;
	}

	//----------------------------------------------
	//equivalent fly (delvis baserad på procedure 4.9.1 chapter 4, Castagna, s. 118)
	//broker_fly --> smile_fly
	// Note: Castagna proceduren är bara definierad för en vinge --> proceduren är tweakad för att hantera 2 vingar
	//this func only defined for rr and bfly att anchor points i.e. 25d or 10d
	//----------------------------------------------
	number bfly_to_sfly_near(number s ,
							 number f,
							 number tau,
							 number rr ,
							 number bfly ,//broker_fly
							 number rr2 ,
							 number bfly2 ,//broker_fly
							 number df ,
							 number dfq ,
							 number  delta,
							 number  delta2,
							 CORE_INT_VOL.vol_smile_model_delta_wing sm,//must have proper data but near wing will be updated
							 //logical near_wing,
							 out number KC,
							 out number KP,
							 out number VP,
							 out number VC)
	 
	{
		fx_delta_type dt = delta_type_conv( sm.delta_type());
		
		//****************
		//castagna STEP 1: calc KA (here an input) 
		//****************
		number cont_u_r = -log(dfq)/tau;
		number cont_r 	= -log(df)/tau;
		//****************
		//castagna STEP 2: calc as if bfly == smilefly, consistent vols
		//****************

		number VA = sm.atm_vol();
		fx_rr_ss_vols(VA,rr,bfly,VC, VP);

		//****************
		//castagna STEP 3: calc implied strikes, Formula 2.46 chapter 2
		//****************
		//strikes using consistent vols for call and put
		
		delta_to_k_cp(dt,delta,s,f,VC,VP,cont_r,cont_u_r, tau,KC,KP);

		//****************
		//castagna STEP 4: calculate strikes for the fly with unique vol
		//****************
		//vol for call and put in brokerfly
		number vol_bfly = VA + bfly; //market covention for the fly volatility		
				
		//strikes using bfly vol i.e. same vol for call and put
		number k_dd_bc,k_dd_bp;
		delta_to_k_cp(dt,delta,s,f,vol_bfly,vol_bfly,cont_r,cont_u_r,tau,k_dd_bc,k_dd_bp);
		
		sm.update_wing(true,[KP,KC],[VP,VC]);
		if(k_dd_bc > KC || k_dd_bp < KP) {//extrapolation --> need to update far wing
			number KC2,KP2;
			number VC2,VP2;
			bfly_to_sfly(s ,f,tau,rr2 ,bfly2 ,df ,dfq , delta2,sm, false, KC2, KP2, VP2, VC2);//sm is updated!!
		}
		//****************
		//castagna STEP 5: 
		//****************
		//interpolate vol from strikes implied by bflyvol
		number vol_ip_bc = sm.interpol_strike(k_dd_bc);
		number vol_ip_bp = sm.interpol_strike(k_dd_bp);
		
		//calc prices 
		number fp;
		number call25c, call25b, put25p, put25b;
		//calc prices for new strike and interp vols
		black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_ip_bc,  cont_r, cont_u_r, tau,tau,  fp, call25c);		
		black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_ip_bp,  cont_r, cont_u_r, tau,tau,  fp, put25p);
		number stgl_mkt = call25c + put25p;
		
		//calc prices for new strike and bfly vol
		black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, call25b);
		black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, put25b);
		number stgl_conv = call25b + put25b;
		
		//diff between strangle at market vols and the same strangle with conventional unique vol		
		number f0 = stgl_mkt - stgl_conv;

		if (abs(f0) < 0.0000001 * s) 
			return bfly;
		

		bfly = bfly + 0.0001;
		number dfly = 0.0001;

		while (abs(f0) > 0.0000001 * s){
			
			fx_rr_ss_vols(VA,rr,bfly,VC, VP);
			delta_to_k_cp(dt,delta, s,f,VC,VP, cont_r,cont_u_r,tau, KC,KP);
			sm.update_wing(true,[KP,KC],[VP,VC]);

			if(k_dd_bc > KC || k_dd_bp < KP) {//extrapolation
				number KC2,KP2;
				number VC2,VP2;
				bfly_to_sfly(s ,f,tau,rr2 ,bfly2 ,df ,dfq , delta2,sm, false, KC2, KP2, VP2, VC2);
			}
	
			delta_to_k_cp(dt,delta,s,f,vol_bfly,vol_bfly,cont_r,cont_u_r,tau, k_dd_bc,k_dd_bp);

			vol_ip_bc = sm.interpol_strike(k_dd_bc);
			vol_ip_bp = sm.interpol_strike(k_dd_bp);
		
			black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_ip_bc,  cont_r, cont_u_r, tau,tau,  fp, call25c);			
			black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_ip_bp,  cont_r, cont_u_r, tau,tau,  fp, put25p);
			
			black_scholes_prem(	opt_type.OPT_CALL,s, k_dd_bc, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, call25b);	
			black_scholes_prem(	opt_type.OPT_PUT, s, k_dd_bp, vol_bfly,   cont_r, cont_u_r, tau,tau,  fp, put25b);
		
			number fu		= (call25c + put25p) - (call25b + put25b);
			number df_der 	= (fu - f0) / dfly;		
			bfly 			= bfly - fu / df_der;
			dfly 			= -fu / df_der;
			f0 				= fu;

		}
		return bfly;
	}

	//----------------------------------------------
	//broker_fly  Castagna, s. 137
	//smile_fly --> broker_fly  
	//----------------------------------------------
	number broker_fly(	 number s ,
						 number f,
						 number tau,
						 //number rr ,
						 //number sfly ,//smile_fly
						 number df ,
						 number dfq ,
						 number  delta,						
						 CORE_INT_VOL.vol_smile_model_delta_wing sm)
	 
	{
		fx_delta_type dt = delta_type_conv( sm.delta_type());
		
		//****************
		//castagna STEP 1: calc KA (here an input) 
		//****************		
		number cont_u_r = -log(dfq)/tau;
		number cont_r 	= -log(df)/tau;

		//****************
		//castagna STEP 2: calc as if bfly == smilefly, consistent vols
		//****************
		//number VP,VC;
		//fx_rr_ss_vols(VA,rr,sfly,VC, VP);

		//****************
		//castagna STEP 3: calc implied wing strikes, Formula 2.46 chapter 2
		//****************
		//strikes using consistent vols for call and put
		//number KP,KC;
		//delta_to_k_cp(delta_type,delta,s,f,VC,VP,cont_r,cont_u_r, tau,KC,KP);//make input

		number sigdc,sigdp,kdc,kdp ;
		sm.symm_delta_to_vol(delta,sigdc,kdc, sigdp,kdp);
		
		number sigdb = (sigdc + sigdp) / 2;
		number kdbc,kdbp;
		delta_to_k_cp(dt,delta,s,f,sigdb,sigdb,cont_r,cont_u_r, tau,kdbc,kdbp);

		number sigdbc = sm.interpol_strike(kdbc);
		number sigdbp = sm.interpol_strike(kdbp);
		
		//calc prices 
		number fp;
		number calldc, calldb, putdp, putdb;
		//calc prices for new strike and interp vols
		black_scholes_prem(	opt_type.OPT_CALL,s, kdbc, sigdc,  cont_r, cont_u_r, tau,tau,  fp, calldc);		
		black_scholes_prem(	opt_type.OPT_PUT, s, kdbp, sigdp,  cont_r, cont_u_r, tau,tau,  fp, putdp);
		number stgl_mkt = calldc + putdp;
		
		//calc prices for new strike and bfly vol
		black_scholes_prem(	opt_type.OPT_CALL,s, kdbc, sigdb,   cont_r, cont_u_r, tau,tau,  fp, calldb);
		black_scholes_prem(	opt_type.OPT_PUT, s, kdbp, sigdb,   cont_r, cont_u_r, tau,tau,  fp, putdb);
		number stgl_conv = calldb + putdb;
		
		//diff between strangle at market vols and the same strangle with conventional unique vol
		number VA 	= sm.atm_vol();		
		number f0 	= stgl_mkt - stgl_conv;		
		number dfly = sigdb - VA;

		if( abs(f0) < 0.0000001 * s ) {
		 return dfly;
		 
		}

		sigdb 		= sigdb + 0.00005;
		integer i 	= 0;	
		number dsig = 0.00005;

		//while( abs(f0) > 0.000000001 * s) {//test
		while( abs(f0) > 0.00000001 * s) {//orig

			i = i + 1;
			delta_to_k_cp(dt,delta,s,f,sigdb,sigdb,cont_r,cont_u_r, tau,kdbc,kdbp);

			sigdbc = sm.interpol_strike(kdbc);
			sigdbp = sm.interpol_strike(kdbp);

			black_scholes_prem(	opt_type.OPT_CALL,s, kdbc, sigdbc,  cont_r, cont_u_r, tau,tau,  fp, calldc);		
			black_scholes_prem(	opt_type.OPT_PUT, s, kdbp, sigdbp,  cont_r, cont_u_r, tau,tau,  fp, putdp);
			stgl_mkt = calldc + putdp;
			
			black_scholes_prem(	opt_type.OPT_CALL,s, kdbc, sigdb,   cont_r, cont_u_r, tau,tau,  fp, calldb);
			black_scholes_prem(	opt_type.OPT_PUT, s, kdbp, sigdb,   cont_r, cont_u_r, tau,tau,  fp, putdb);
			stgl_conv = calldb + putdb;
			
			number fu 		= stgl_mkt - stgl_conv;		
			number df_der 	= (fu - f0) / dsig;			
			sigdb 			= sigdb - fu / df_der;
			dsig 			= -fu / df_der;
			f0 				= fu;
			
			if(i * 0.00001 >= dfly) {
				return sigdb - VA;//error?
			}
			
		}

		return sigdb - VA;
	}
}