/*	
	Bond future functions
	Developed by Algorithmica Research, Magnus Nyström
			
	ToDo: 
	
	Comments:	

*/
option(null: hard);

//-------------------------------------------------------
// fut_scen_type
//-------------------------------------------------------
class enum fut_scen_type
{
	SST_USER 		option(str: "UserInput") ,
	SST_BE_REPO 	option(str: "BreakEvenRepo") ,
	SST_BE_FUT_NR 	option(str: "BreakEvenFut(flat repo)") ,
	SST_BE_FUT_R 	option(str: "BreakEvenFut(incl repo)") ,
	SST_BE_CTD_NR 	option(str: "BreakEvenCTD(flat repo)") ,
	SST_BE_CTD_R 	option(str: "BreakEvenCTD(incl repo)") ,
	SST_PARALLELL 	option(str: "Parallell") 
};

vector(string) fut_scen_type_list() 
{
	vector(fut_scen_type) e;
	expand_enum(e) ;
	return string(e);
}

fut_scen_type fut_scen_map(string s)
{
	fut_scen_type e;
	str_to_enum(s,e);
	return e;
}

//-------------------------------------------------------
// y_shift_scen
//-------------------------------------------------------
void y_shift_scen(	bond_future 				fut,
					integer 					ctd_idx,
					fut_scen_type 				st,
					logical option(nullable) 	disable_rnd,
					out number   				ctd_y_shift,
					out number   				fut_y_shift,
					out number   				repo_shift)
{
	try{
		error_info ee = new error_info(true,true);
		
		switch (st) {
		case fut_scen_type.SST_USER://"UserInput"
			QL_FAIL_COND(null(ctd_y_shift) || null(fut_y_shift) || null(repo_shift),"missing yield shift (all are required for selected scenario)");
			break;	
		case fut_scen_type.SST_PARALLELL: //"Parallell"
			QL_FAIL_COND(null(ctd_y_shift), "missing yield shift - ctd");
			fut_y_shift	= ctd_y_shift;
			repo_shift = ctd_y_shift;
			break;
		case fut_scen_type.SST_BE_FUT_NR://"BreakEvenFut(flat repo)"
			QL_FAIL_COND(null(ctd_y_shift), "missing yield shift - ctd");
			repo_shift = 0;
			fut_y_shift = fut.fut_y_sens(ctd_idx, repo_shift, ctd_y_shift, disable_rnd, ee );
			
			break;
		case fut_scen_type.SST_BE_CTD_NR://"BreakEvenCTD(flat repo)"
			QL_FAIL_COND(null(fut_y_shift), "missing yield shift - bond future");				
			repo_shift = 0;
			ctd_y_shift = fut.ctd_y_sens(ctd_idx, repo_shift,fut_y_shift, disable_rnd, ee );
			
			break;
		case fut_scen_type.SST_BE_REPO:	//"BreakEvenRepo"			
			QL_FAIL_COND(null(ctd_y_shift), "missing yield shift - ctd");
			if(null(fut_y_shift))
				fut_y_shift = ctd_y_shift;
			repo_shift = fut.repo_sens(ctd_idx,  fut_y_shift, ctd_y_shift, disable_rnd, ee );
			break;
		case fut_scen_type.SST_BE_FUT_R://"BreakEvenFut(incl repo)"
			QL_FAIL_COND(null(ctd_y_shift), "missing yield shift - ctd");
			if(null(repo_shift))
				repo_shift = ctd_y_shift; 							
			fut_y_shift = fut.fut_y_sens(ctd_idx,repo_shift, ctd_y_shift, disable_rnd, ee );
			break;
		case fut_scen_type.SST_BE_CTD_R: //"BreakEvenCTD(incl repo)"
			QL_FAIL_COND(null(fut_y_shift), "missing yield shift - bond future");
			if(null(repo_shift))
				repo_shift = fut_y_shift; 				
			ctd_y_shift = fut.ctd_y_sens(ctd_idx,repo_shift,fut_y_shift, disable_rnd );
			break;
		default:
			fut_y_shift = null<number>;
			ctd_y_shift = null<number>;
			repo_shift = null<number>;
		}
	}
	catch {
		fut_y_shift = null<number>;
		ctd_y_shift = null<number>;
		repo_shift = null<number>;
	}
}
//-------------------------------------------------------
// futroll_scen_type
//-------------------------------------------------------
class enum futroll_scen_type 
{
	FRST_USER			option(str: "UserInput"),
	FRST_BE_REPO		option(str: "BreakEvenRepo"),
	FRST_BE_FUT_SH_NR	option(str: "BreakEvenFutFr(flat repo)"),
	FRST_BE_FUT_SH_R	option(str: "BreakEvenFutFr(incl repo)"),
	FRST_BE_FUT_LO_NR	option(str: "BreakEvenFutBa(flat repo)"),
	FRST_BE_FUT_LO_R	option(str: "BreakEvenFutBa(incl repo)"),
	FRST_PARALLELL		option(str: "Parallell")
};

vector(string) futroll_scen_type_list() 
{
	vector(futroll_scen_type) e;
	expand_enum(e) ;
	return string(e);
}

futroll_scen_type futroll_scen_map(string s)

{
	futroll_scen_type e;
	str_to_enum(s,e);
	return e;
}



class bondfut_data
{
public:
	bondfut_data(	date,instrument_name,quote_side ,number option(nullable),day_count_method , 
					rate_type,quote_side ,vector(number) option(nullable) ,vector(number) option(nullable) );

	bondfut_data(	date ,instrument_name,quote_side,number option(nullable),day_count_method , 
					rate_type,instrument_name,quote_side,number option(nullable),number option(nullable));
	
	bond_future	 	future();
	bond	 		ctd();
	number	 		ctd_repo();
	number	 		ctd_pf();
	date			settle();	
	date			settle_fwd();	
	logical	 		init_ok();
	string	 		err_msg();
		
	vector(number) 		pf_;
	day_count_method 	dc_; 
	rate_type 			rt_;	
	bond_future 		fut_;

	number 				ctd_pf_;
	integer 			ctd_idx_;
	bond 				ctd_;
	
	string 				err_msg_;
	logical				init_ok_;		
};

bondfut_data.bondfut_data(	date d,
							instrument_name 					f_n,
							quote_side 							f_q_s,
							number option(nullable) 			f_quote,							
							day_count_method 					dc, 
							rate_type 							rt, 
							quote_side  						b_q_s,						
							vector(number) option(nullable) 	b_quote,
							vector(number) option(nullable) 	repo_r )
							:dc_(dc), rt_(rt), err_msg_(""), init_ok_(false) 
{		
	try {
		error_info ee = error_info(true,true);

		if(null(f_quote)){			
			fut_ = bond_future(f_n, d, f_q_s, b_q_s,ee);	
		}
		else {
			fut_ = bond_future(f_n, d, f_quote,b_q_s,ee);
		}
		
		if(!null(b_quote)) {
			fut_.add_quote_deliverables(b_quote, ee);
		}
		
		fut_.add_repo_quotes(repo_r, dc, false, ee);//init_ctd is run a second time!!
		
		logical is_val = fut_.is_valid(ee);
		QL_REQUIRE(is_val, ee.message());
		
		ctd_ = fut_.ctd( ctd_pf_, ctd_idx_, ee);
		QL_REQUIRE(ctd_idx_ != err_int(), ee.message());
		
		init_ok_ = true;
	}
	catch {
		init_ok_ = false;
		err_msg_ = err.message();
	}
}
//for single deliverable input --> will be assumed to be ctd (but must be part of deliv basket)
bondfut_data.bondfut_data(	date d,
							instrument_name 					f_n,
							quote_side 							f_q_s,
							number option(nullable) 			f_quote,							
							day_count_method 					dc, 
							rate_type 							rt,
							instrument_name						b_n,
							quote_side  						b_q_s,						
							number option(nullable) 			b_quote,
							number option(nullable) 			repo_r )
							:dc_(dc), rt_(rt), err_msg_(""), init_ok_(false) 
{		
	try {
		error_info ee = error_info(true,true);

		if(null(f_quote)){			
			fut_ = bond_future(f_n, d, f_q_s, b_q_s,ee);	
		}
		else {
			fut_ = bond_future(f_n, d, f_quote,b_q_s,ee);
		}

		vector(string) del 	= fut_.deliverable_v_s();
		vector(integer) v_o = v_findi(del, [b_n]);
		QL_REQUIRE(v_size(v_o) == 1,strcat(b_n, " is not a deliverable bond")) ;

		ctd_idx_ 			= v_o[0];
		
		if(!null(b_quote)) {
			fut_.add_quote_deliverable(ctd_idx_, b_quote, ee);
		}
		
		fut_.add_repo_quote(ctd_idx_, repo_r, dc, true, ee);//init_ctd is run a second time!!
		

		logical is_val = fut_.is_valid(ee);
		QL_REQUIRE(is_val, ee.message());
		
		ctd_ = fut_.deliv( ctd_idx_, ee);
		
		init_ok_ = true;
	}
	catch {
		init_ok_ = false;
		err_msg_ = err.message();
	}
}

bond_future 	bondfut_data.future() 			{ QL_REQUIRE(init_ok_,err_msg_); return fut_;}
bond			bondfut_data.ctd() 				{ QL_REQUIRE(init_ok_,err_msg_); return ctd_;}	
//bond 			bondfut_data.deliv(integer i) 	{ QL_REQUIRE(init_ok_,err_msg_); return fut_.deliv(i);}
logical 		bondfut_data.init_ok() 			{ return init_ok_;}												
string 			bondfut_data.err_msg() 			{ return err_msg_;}
date			bondfut_data.settle()			{ QL_REQUIRE(init_ok_,err_msg_); return ctd_.settle_date();}	
date			bondfut_data.settle_fwd()		{ QL_REQUIRE(init_ok_,err_msg_); return fut_.expiry_settle_date();}
number 			bondfut_data.ctd_repo()			{ QL_REQUIRE(init_ok_,err_msg_); return fut_.repo(ctd_idx_);}
number 			bondfut_data.ctd_pf()			{ QL_REQUIRE(init_ok_,err_msg_); return fut_.price_factor(ctd_idx_);}


//-------------------------------------------------------
// class bondfut_data
//-------------------------------------------------------
class bond_future_roll 
{
	public:
	
	bond_future_roll(	bond_future,bond_future,bond,number option(nullable) ,bond,number option(nullable),
						number option(nullable),number option(nullable),number option(nullable),number option(nullable),
						logical);

	void  	init(logical);
	void 	init_calc();
	void 	hdg_ratio_ctd_ba(logical option(nullable),  logical option(nullable),out number,out number);
	void	ctd_shift_scen(	fut_scen_type,number,logical option(nullable),out number,out number,out number,out number,out number);
	void	fut_shift_scen(	futroll_scen_type,logical option(nullable),out number,out number ,out number );
	void 	fut_roll_shift_scen(futroll_scen_type,logical option(nullable) ,out number,out number ,out number);
	
	date 		settle_fwd_sh() ;	
	date 		settle_fwd_lo() ;	
	bond_future fut_sh(); 			
	bond_future fut_lo();			
	bond_future fut_fwd(error_info option(nullable) );
	number 		imp_repo_sh(error_info option(nullable));
	number 		imp_repo_lo(error_info option(nullable));
	number 		imp_repo_ff(error_info option(nullable));
	number 		repo_sh();
	number 		repo_lo();
	number 		repo_ff(); 
	bond		ctd(error_info option(nullable));
	bond		ctd_fwd(error_info option(nullable));
	number		fut_lo_y(error_info option(nullable) );
	number		fut_sh_y(error_info option(nullable) );
	number		deliv_y_fwd(error_info option(nullable) );
	number		deliv_y(error_info option(nullable) );
	number		deliv_dp_fwd(error_info option(nullable) );
	number		deliv_dp(error_info option(nullable) );
	number 		repo_ff_sens(number ,logical,number option(nullable) ,logical option(nullable),error_info option(nullable) );
	number		fut_y_sprd(error_info option(nullable) );
	number 		imp_fut_y_sprd(error_info option(nullable));

	logical		init_ok();
	logical		init_calc_ok();
	string		err_msg();

	protected:
					
	bond_future 		fut_sh_;
	bond_future 		fut_lo_;	
	bond				ctd_sh_;
	bond				ctd_lo_;
	number  			pf_sh_;
	number  			pf_lo_;
	number				repo_sh_;
	number				repo_lo_;
	number				repo_fwd_;
	number 				ctd_sprd_;
	bond_future 		fut_fwd_;
	number				imp_repo_fwd_;
	integer 			ctd_idx_sh_;
	integer 			ctd_idx_lo_;
	string				err_msg_;
	logical				init_ok_;		
	logical				init_calc_ok_;

};

bond_future_roll.bond_future_roll(	bond_future 			fut_sh,
									bond_future 			fut_lo,
									bond 					ctd_sh,
									number option(nullable) pf_sh,
									bond 					ctd_lo,
									number option(nullable) pf_lo,
									number option(nullable) ctd_sprd,	
									number option(nullable) repo_sh,
									number option(nullable) repo_lo,
									number option(nullable) repo_fwd,//repo_fwd has priority
									logical 				dflt_repo_is_imp)
									:fut_sh_(fut_sh), fut_lo_(fut_lo), ctd_sh_(ctd_sh),ctd_lo_(ctd_lo),
									 pf_sh_(pf_sh), pf_lo_(pf_lo),repo_sh_(repo_sh),repo_lo_(repo_lo),repo_fwd_(repo_fwd),
									 ctd_sprd_(ctd_sprd), fut_fwd_(null<bond_future>), imp_repo_fwd_(null<number>),
									 ctd_idx_sh_(err_int()), ctd_idx_lo_(err_int()), err_msg_(""), init_ok_(false), init_calc_ok_(false)
{										
	init(dflt_repo_is_imp);
	return;
	
}

//-------------------------------------------------------
// init
//-------------------------------------------------------
void bond_future_roll.init(logical dflt_repo_is_imp)									
{
	try {
		error_info ee = new error_info(true,true);
		
		if(!ctd_sh_.is_valid(ee)) {
			err_msg_ = ctd_sh_.err_msg();	
			return ;
		}

		if(!ctd_lo_.is_valid(ee)) {
			err_msg_ = ctd_lo_.err_msg();	
			return ;
		}
		
		if(!fut_sh_.is_valid(ee)) {
			err_msg_ = fut_sh_.err_msg();	
			return ;
		}

		if(!fut_lo_.is_valid(ee)) {
			err_msg_ = fut_lo_.err_msg();	
			return ;
		}

		vector(string) del 	= fut_sh_.deliverable_v_s();
		vector(integer) v_o = v_findi(del, [ctd_sh_.name(ee)]);
		QL_REQUIRE(v_size(v_o) == 1,strcat(ctd_sh_.name(ee), " is not a deliverable bond")) ;
		ctd_idx_sh_ 		= v_o[0];

		del 				= fut_lo_.deliverable_v_s();
		v_o 				= v_findi(del, [ctd_lo_.name(ee)]);
		QL_REQUIRE(v_size(v_o) == 1,strcat(ctd_lo_.name(ee), " is not a deliverable bond")) ;
		ctd_idx_lo_ 		= v_o[0];
		
		date settle 		= ctd_sh_.settle_date(ee);
		date settle_fwd1 	= fut_sh_.expiry_settle_date(ee);
		date settle_fwd2 	= fut_lo_.expiry_settle_date(ee);
		
		if(fut_sh_.style(ee) != fut_lo_.style(ee)){
			err_msg_ = "different types of futures not supported (style)";	
			return ;
		}	

		if(fut_sh_.trade_date(ee) != fut_lo_.trade_date(ee)){
			err_msg_ = "invalid trade date (bond future)";	
			return ;
		}
		if(ctd_sh_.trade_date(ee) != ctd_lo_.trade_date(ee)){
			err_msg_ = "invalid trade date (ctd)";	
			return ;
		}
		
		if(settle_fwd1 >= settle_fwd2){
			err_msg_ = "invalid expiry settlement date (bond future)";	
			return ;
		}		
		
		if(settle  != ctd_lo_.settle_date(ee)){
			err_msg_ = "invalid settlement date (ctd)";	
			return ;
		}

		day_count_method dc = fut_sh_.repo_dc_method(ee);
		if(dc != fut_lo_.repo_dc_method(ee)){
			err_msg_ = "invalid day_count_method (mismatch between futures)";	
			return ;
		}

		string msh = fut_sh_.maturity_code(ee);
		string mlo = fut_lo_.maturity_code(ee);			
		if(!null(msh) && !null(mlo)){
			if(msh != mlo){
				err_msg_ = "different types of futures not supported (underlying)";	
				return ;
			}
		}
		else {
			date matsh = fut_sh_.synt_maturity(ee);
			date matlo = fut_lo_.synt_maturity(ee);
			integer mbsh = CORE_INT.months_between(settle_fwd1, matsh);
			integer mblo = CORE_INT.months_between(settle_fwd2, matlo);
			if(mbsh != mblo){
				err_msg_ = "different types of futures not supported (underlying)";	
				return ;
			}
		}

		if(ctd_sh_.name(ee) != ctd_lo_.name(ee) && null(ctd_sprd_)){
			err_msg_ = "spread is required when ctd bonds are different";		
			return ;
		}
				
		if(null(ctd_sprd_) || ctd_sh_.name(ee) == ctd_lo_.name(ee))
			ctd_sprd_ = 0;

		if(!null(repo_fwd_)) {

			if(null(repo_sh_)) 
				repo_sh_ = dflt_repo_is_imp ? fut_sh_.imp_repo(ctd_idx_sh_,ee) : fut_sh_.repo(ctd_idx_sh_,ee);
			
			QL_REQUIRE(!null(repo_sh_),"invalid repo rate - front future (failed calculation)");
			repo_lo_ 	= CORE_INT.nose_tail(settle , settle_fwd1, settle_fwd2, repo_sh_, repo_fwd_, 3,dc,dc,
											 dc,rate_type.RT_SIMPLE,rate_type.RT_SIMPLE,rate_type.RT_SIMPLE);
			QL_REQUIRE(!null(repo_lo_),"invalid repo rate - back future (failed calculation)");	
		}
		else {

			if(null(repo_sh_)) {
				repo_sh_ = dflt_repo_is_imp ? fut_sh_.imp_repo(ctd_idx_sh_,ee):fut_sh_.repo(ctd_idx_sh_,ee);
			}
			if(null(repo_lo_))
				repo_lo_ = dflt_repo_is_imp ? fut_lo_.imp_repo(ctd_idx_lo_,ee):fut_lo_.repo(ctd_idx_lo_,ee);
							
			QL_REQUIRE(!null(repo_sh_),"invalid repo rate - front future");
			QL_REQUIRE(!null(repo_lo_),"invalid repo rate - back future");
			repo_fwd_ = CORE_INT.nose_tail(	settle , settle_fwd1, settle_fwd2, repo_sh_, repo_lo_, 2,dc,dc,
											dc,rate_type.RT_SIMPLE,rate_type.RT_SIMPLE,rate_type.RT_SIMPLE);
			QL_REQUIRE(!null(repo_fwd_),"invalid fwd repo rate (failed calculation)");
		}

		if(null(fut_sh_.repo(ctd_idx_sh_,ee)))
			fut_sh_.add_repo_quote(ctd_idx_sh_,repo_sh_,null<day_count_method>, false,ee);

		if(null(fut_lo_.repo(ctd_idx_lo_,ee)))
			fut_lo_.add_repo_quote(ctd_idx_lo_,repo_lo_,null<day_count_method>, false,ee);
			
		err_msg_ = "";
		init_ok_ = true;

		return;
	}
	catch {
		err_msg_ = err.message();
		init_ok_ = false;
	}	
}

//-------------------------------------------------------
// init_calc
//-------------------------------------------------------
void bond_future_roll.init_calc()
{
	try {
		
		QL_REQUIRE(init_ok_,err_msg_);
		error_info ee 	= new error_info(true,true);
		
		bond ctd 		= ctd_lo_.set_settle_date(fut_sh_.expiry_settle_date(ee));				
		number y_fwd 	= ctd_sh_.yield(true,ee) + ctd_sprd_; 
		ctd 			= ctd.set_yield(y_fwd);

		string name		= strcat(fut_lo_.name(),"fwd");
		fut_fwd_ 		= bond_future(fut_lo_.instr_def(), name,fut_lo_.trade_date(ee) ,
									  fut_lo_.expiry_settle_date(ee),fut_lo_.quote(ee),100,ctd,pf_lo_ );
				
		imp_repo_fwd_ = fut_fwd_.imp_repo(ctd_idx_lo_,ee);
		
		init_calc_ok_ = true;
		return;
	}
	catch {
		err_msg_ = err.message();
		init_calc_ok_ = false;
	}
}

date 		bond_future_roll.settle_fwd_sh() 	{ return fut_sh_.expiry_settle_date();}
date 		bond_future_roll.settle_fwd_lo() 	{ return fut_lo_.expiry_settle_date();}
bond_future bond_future_roll.fut_sh() 			{ return fut_sh_;}
bond_future bond_future_roll.fut_lo()			{ return fut_lo_;}
logical		bond_future_roll.init_calc_ok()		{ return init_calc_ok_;}
logical		bond_future_roll.init_ok()			{ return init_ok_;}
string		bond_future_roll.err_msg()			{ return err_msg_;}
number 		bond_future_roll.repo_sh()			{ return repo_sh_;}
number 		bond_future_roll.repo_lo()			{ return repo_lo_;}
number 		bond_future_roll.repo_ff()			{ return repo_fwd_;}

//-------------------------------------------------------
// imp_repo_sh
//-------------------------------------------------------
number bond_future_roll.imp_repo_sh(error_info option(nullable) error)
{
	try {
		error_info ee = new error_info(true,true);
		return fut_sh_.imp_repo(ctd_idx_sh_,ee);
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future.imp_repo_sh");
		return null<number>;
	}
}
//-------------------------------------------------------
// imp_repo_lo
//-------------------------------------------------------
number bond_future_roll.imp_repo_lo(error_info option(nullable) error)
{
	try {
		error_info ee = new error_info(true,true);
		return fut_lo_.imp_repo(ctd_idx_lo_,ee);
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future.imp_repo_lo");
		return null<number>;
	}
}
	
//-------------------------------------------------------
// fut_sh_y
//-------------------------------------------------------
number bond_future_roll.fut_sh_y(error_info option(nullable) error)
{	
	try {
		error_info ee = new error_info(true,true);
		return fut_sh_.yield(true,ee);
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future.fut_sh_y");
		return null<number>;
	}
}

//-------------------------------------------------------
// fut_lo_y
//-------------------------------------------------------
number bond_future_roll.fut_lo_y(error_info option(nullable) error)
{	
	try {
		error_info ee = new error_info(true,true);
		return fut_lo_.yield(true,ee);
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future.fut_lo_y");
		return null<number>;
	}
}

//-------------------------------------------------------
// imp_repo_ff
//-------------------------------------------------------
number bond_future_roll.imp_repo_ff(error_info option(nullable) error)
{	
	try {
		QL_REQUIRE(init_calc_ok_,err_msg_);
		return imp_repo_fwd_;
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future.imp_repo_ff");
		return null<number>;
	}
}

//-------------------------------------------------------
// fut_fwd
//-------------------------------------------------------
bond_future bond_future_roll.fut_fwd(error_info option(nullable) error)
{
	try {
		QL_REQUIRE(init_calc_ok_,err_msg_);
		return fut_fwd_;
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future.fut_fwd");
		return null<bond_future>;
	}
}
//-------------------------------------------------------
// deliv_y
//-------------------------------------------------------
number bond_future_roll.deliv_y(error_info option(nullable) error)
{
	try {
		error_info ee = new error_info(true,true);
		return this.ctd(ee).yield(true,ee);
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future.deliv_y");
		return null<number>;
	}
}
//-------------------------------------------------------
// deliv_dp
//-------------------------------------------------------
number bond_future_roll.deliv_dp(error_info option(nullable) error)
{
	try {
		error_info ee = new error_info(true,true);
		return this.ctd(ee).dirty_price(null<number>,false,true,ee);
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future.deliv_dp");
		return null<number>;
	}
}
//-------------------------------------------------------
// deliv_dp_fwd
//-------------------------------------------------------
number bond_future_roll.deliv_dp_fwd(error_info option(nullable) error)
{
	try {
		error_info ee = new error_info(true,true);
		return this.ctd_fwd(ee).dirty_price(null<number>,false,true,ee);
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future.deliv_dp_fwd");
		return null<number>;
	}
}
//-------------------------------------------------------
// deliv_y_fwd
//-------------------------------------------------------
number bond_future_roll.deliv_y_fwd(error_info option(nullable) error)
{
	try {
		error_info ee = new error_info(true,true);
		return this.ctd_fwd(ee).yield(true,ee) ;
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future.deliv_y_fwd");
		return null<number>;
	}
}
//-------------------------------------------------------
// ctd
//-------------------------------------------------------
bond bond_future_roll.ctd(error_info option(nullable) error)
{
	try {
		QL_REQUIRE(init_calc_ok_,err_msg_);
		error_info ee = new error_info(true,true);
		return this.fut_fwd(ee).deliv(ctd_idx_lo_,ee);
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future.ctd");
		return null<bond>;
	}
}

//-------------------------------------------------------
// ctd_fwd
//-------------------------------------------------------
bond bond_future_roll.ctd_fwd(error_info option(nullable) error)
{
	try {
		QL_REQUIRE(init_calc_ok_,err_msg_);
		error_info ee = new error_info(true,true);
		return this.fut_fwd(ee).deliv_fwd(ctd_idx_lo_,ee);
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future_roll.ctd_fwd");
		return null<bond>;
	}
}
//-------------------------------------------------------
// fut_y_sprd
//-------------------------------------------------------
number bond_future_roll.fut_y_sprd(error_info option(nullable) error)
{
	try {
		error_info ee 	= new error_info(true,true);
		bond bond_fwd 	= fut_lo_.deliv_fwd(ctd_idx_lo_,ee);
		number ylo 		= bond_fwd.yield(false,ee);

		bond_fwd 		= fut_sh_.deliv_fwd(ctd_idx_sh_,ee);
		number ysh 		= bond_fwd.yield(false,ee);
		
		return ylo - ysh;
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future_roll.fut_y_sprd");
		return null<number>;
	}
}
//-------------------------------------------------------
// imp_fut_y_sprd
//-------------------------------------------------------
number bond_future_roll.imp_fut_y_sprd(error_info option(nullable) error)
{
	try {
		error_info ee = new error_info(true,true);
		number	cp, ylo;
		fut_lo_.imp_fut(ctd_idx_lo_,repo_lo_, cp, ylo,ee);

		number	ysh;
		fut_sh_.imp_fut(ctd_idx_sh_,repo_sh_, cp, ysh,ee);
		
		return ylo - ysh;
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future_roll.imp_fut_y_sprd");
		return null<number>;
	}
}

//-------------------------------------------------------
// repo_ff_sens
//-------------------------------------------------------
number bond_future_roll.repo_ff_sens(	number y_fut_lo_shift,
										logical incl_fut_sh,
										number option(nullable) y_fut_sh_shift,
										logical option(nullable) disable_rnd,
										error_info option(nullable) error)
{
	try {
		QL_REQUIRE(init_calc_ok_,err_msg_);
		error_info ee = new error_info(true,true);

		if(!incl_fut_sh)
			y_fut_sh_shift = 0;
		else
			QL_REQUIRE(!null(y_fut_sh_shift),"invalid y_fut_sh_shift");
		
		return this.fut_fwd(ee).repo_sens(ctd_idx_lo_, y_fut_lo_shift, y_fut_sh_shift, disable_rnd, ee);
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bond_future_roll.repo_ff_sens");
		return null<number>;
	}
}

//-------------------------------------------------------
// hdg_ratio_ctd
//-------------------------------------------------------
void bond_future_roll.hdg_ratio_ctd_ba(	logical option(nullable) 	incl_tail,		
										//number option(nullable)		delta,	
										//logical option(nullable) 	pvbp_centered,
										logical option(nullable)	disable_round,
										out number 					front_hr,
										out number 					back_hr) 	
{		
	try {
		if(!init_calc_ok_){
			front_hr = null<number>;
			back_hr = null<number>;
			return;
		}
		error_info ee = new error_info(true,true);

		if(null(incl_tail)){
			front_hr = null<number>;
			back_hr = null<number>;
			return;
		}
		
		if(!fut_sh_.is_price_factor_style(ee)){
			back_hr 		= fut_lo_.hdg_ratio_ctd(ctd_idx_lo_,incl_tail,disable_round,ee );			
			number pvbp 	= this.ctd(ee).pvbp(100, 0.0001, false, disable_round,ee);
			number pvbp_fsh	= fut_sh_.pvbp(ctd_idx_sh_, 100, 0.0001, disable_round,ee);
			front_hr 		= pvbp/pvbp_fsh;	
			return;			
		}
		else {			
			
			/*
			if(incl_tail)
				return pf_[idx_ctd_] * disc_fact_ctd() ;
			else
				return pf_[idx_ctd_];
			*/
			QL_FAIL("work in progress");
			front_hr = null<number>;//for now
			back_hr = null<number>;
			return;							
		}
	}
	catch {
		front_hr = null<number>;//for now
		back_hr = null<number>;
	}		
}

//-------------------------------------------------------
// ctd_shift_scen
//-------------------------------------------------------
void bond_future_roll.ctd_shift_scen(	fut_scen_type 		st,
										number 				ctd_y_shift,
										logical option(nullable)	disable_rnd,
										out number 			fut_y_shift_sh,
										out number 			fut_y_shift_lo,
										out number 			repo_shift_sh,
										out number 			repo_shift_lo,
										out number 			fwd_repo_shift)
{
	try{

		QL_FAIL_COND(!init_calc_ok_,"roll not initialized");
		QL_FAIL_COND(st == fut_scen_type.SST_BE_CTD_NR || st == fut_scen_type.SST_BE_CTD_R,"invalid scenario type");//"BreakEvenCTD(flat repo)" and "BreakEvenCTD(incl repo)"
			
		/*	SST_USER 		option(str: "UserInput") ,
			SST_BE_REPO 	option(str: "BreakEvenRepo") ,
			SST_BE_FUT_NR 	option(str: "BreakEvenFut(flat repo)") ,
			SST_BE_FUT_R 	option(str: "BreakEvenFut(incl repo)") ,
			SST_BE_CTD_NR 	option(str: "BreakEvenCTD(flat repo)") ,
			SST_BE_CTD_R 	option(str: "BreakEvenCTD(incl repo)") ,
			SST_PARALLELL 	option(str: "Parallell") */
		
		y_shift_scen(fut_sh_,ctd_idx_sh_,st,disable_rnd,ctd_y_shift,fut_y_shift_sh,repo_shift_sh);
		y_shift_scen(fut_lo_,ctd_idx_lo_,st,disable_rnd,ctd_y_shift,fut_y_shift_lo,repo_shift_lo);

		error_info ee 		= new error_info(true,true);
		number repo_sh 		= fut_sh_.repo(ctd_idx_sh_,ee) + repo_shift_sh;
		number repo_lo 		= fut_lo_.repo(ctd_idx_lo_,ee) + repo_shift_lo;

		date settle 		= ctd_sh_.settle_date(ee);
		date settle_fwd1 	= fut_sh_.expiry_settle_date(ee);
		date settle_fwd2 	= fut_lo_.expiry_settle_date(ee);
				
		number rf 			= CORE_INT.nose_tail(settle , settle_fwd1, settle_fwd2, repo_sh, repo_lo, 2,
												fut_sh_.repo_dc_method(),fut_lo_.repo_dc_method(),fut_lo_.repo_dc_method(),
												rate_type.RT_SIMPLE,rate_type.RT_SIMPLE,rate_type.RT_SIMPLE);
		fwd_repo_shift 		= rf - repo_fwd_;
		return;
	}
	catch {
		fut_y_shift_sh = null<number>;
		fut_y_shift_lo = null<number>;
		fwd_repo_shift = null<number>;
	}
}

//-------------------------------------------------------
// fut_shift_scen
//-------------------------------------------------------
void bond_future_roll.fut_shift_scen(	futroll_scen_type 	st,												
										logical option(nullable) disable_rnd,
										out number 			fut_y_shift_sh,
										out number 			fut_y_shift_lo,
										out number 			fwd_repo_shift)
{
	try{	
		
		QL_FAIL_COND(!init_calc_ok_, err_msg_);
		
		error_info ee 		= new error_info(true,true);

		switch(st) {
		case  futroll_scen_type.FRST_USER:
			QL_FAIL_COND(null(fut_y_shift_sh) || null(fut_y_shift_lo) || null(fwd_repo_shift),"missing yield shift (all are required for selected scenario)");
			break;		
		case  futroll_scen_type.FRST_PARALLELL:
			QL_FAIL_COND(null(fut_y_shift_sh), "missing yield shift for front future");
			fut_y_shift_lo	= fut_y_shift_sh;
			fwd_repo_shift = fut_y_shift_sh;
			break;
		case  futroll_scen_type.FRST_BE_FUT_LO_NR:
			QL_FAIL_COND(null(fut_y_shift_sh), "missing yield shift for front future");				
			
			y_shift_scen(fut_fwd(ee),ctd_idx_lo_,fut_scen_type.SST_BE_FUT_NR,disable_rnd,fut_y_shift_sh,fut_y_shift_lo,fwd_repo_shift);				
			break;
		case  futroll_scen_type.FRST_BE_FUT_SH_NR:
			QL_FAIL_COND(null(fut_y_shift_lo), "missing yield shift for back future");
			
			fut_y_shift_sh = fut_fwd(ee).ctd_y_sens(ctd_idx_lo_,0, fut_y_shift_lo,disable_rnd );
			fwd_repo_shift = 0;
			break;
		case  futroll_scen_type.FRST_BE_REPO:
			QL_FAIL_COND(null(fut_y_shift_sh), "missing yield shift for front future");
			if(null(fut_y_shift_lo))
				fut_y_shift_lo = fut_y_shift_sh;
			fwd_repo_shift = repo_ff_sens(fut_y_shift_lo, true, fut_y_shift_sh, disable_rnd, ee );
			break;
		case  futroll_scen_type.FRST_BE_FUT_LO_R:
			QL_FAIL_COND(null(fut_y_shift_sh), "missing yield shift for front future");
			
			if(null(fwd_repo_shift))
				fwd_repo_shift = fut_y_shift_sh; 				
						
			y_shift_scen(fut_fwd(ee),ctd_idx_lo_,fut_scen_type.SST_BE_FUT_R,disable_rnd,fut_y_shift_sh,fut_y_shift_lo,fwd_repo_shift);
			break;
		case  futroll_scen_type.FRST_BE_FUT_SH_R:
			QL_FAIL_COND(null(fut_y_shift_lo), "missing yield shift for back future");	
			if(null(fwd_repo_shift))
				fwd_repo_shift = fut_y_shift_lo; 						
			fut_y_shift_sh = fut_fwd(ee).ctd_y_sens(ctd_idx_lo_,fwd_repo_shift, fut_y_shift_lo, disable_rnd );
			break;
		default:
			fut_y_shift_sh = null<number>;
			fut_y_shift_lo = null<number>;
			fwd_repo_shift = null<number>;
		}
	}
	catch {
		fut_y_shift_sh = null<number>;
		fut_y_shift_lo = null<number>;
		fwd_repo_shift = null<number>;
	}

		
		/*
		fut_sh_.y_shift_scen(st,disable_rnd,ctd_y_shift,fut_y_shift_sh,repo_shift_sh);
		fut_lo_.y_shift_scen(st,disable_rnd,ctd_y_shift,fut_y_shift_lo,repo_shift_lo);
		
		number repo_sh 		= fut_sh_.get_repo_ctd() + repo_shift_sh;
		number repo_lo 		= fut_lo_.get_repo_ctd() + repo_shift_lo;
		date settle 		= fut_sh_.get_settle();
		date settle_fwd1 	= fut_sh_.get_settle_fwd();
		date settle_fwd2 	= fut_lo_.get_settle_fwd();
		number rf 			= nose_tail(	settle , settle_fwd1, settle_fwd2, repo_sh, repo_lo, 2,
											fut_sh_.get_dc(),fut_sh_.get_dc(),fut_sh_.get_dc(),
											fut_sh_.get_rt(),fut_sh_.get_rt(),fut_sh_.get_rt());
		fwd_repo_shift = rf - repo_fwd_;
		*/
	
}

//-------------------------------------------------------
// fut_roll_shift_scen (before: fut_shift_scen)
//-------------------------------------------------------
void bond_future_roll.fut_roll_shift_scen(	futroll_scen_type 	st,												
											logical option(nullable) disable_rnd,
											out number 			fut_y_shift_sh,
											out number 			fut_y_shift_lo,
											out number 			fwd_repo_shift)
{
	try{			
		QL_FAIL_COND(!init_calc_ok_, err_msg_);

		error_info ee = new error_info(true,true);

		switch(st) {
		case futroll_scen_type.FRST_USER://"UserInput"
			QL_FAIL_COND(null(fut_y_shift_sh) || null(fut_y_shift_lo) || null(fwd_repo_shift),"missing yield shift (all are required for selected scenario)");			
			break;
		case futroll_scen_type.FRST_PARALLELL: //"Parallell"
			QL_FAIL_COND(null(fut_y_shift_sh), "missing yield shift for front future");
			fut_y_shift_lo	= fut_y_shift_sh;
			fwd_repo_shift = fut_y_shift_sh;
			break;
		case futroll_scen_type.FRST_BE_FUT_LO_NR://"BreakEvenFutBa(flat repo)"
			QL_FAIL_COND(null(fut_y_shift_sh), "missing yield shift for front future");				
			y_shift_scen(fut_fwd(ee),ctd_idx_lo_,fut_scen_type.SST_BE_FUT_NR,disable_rnd,fut_y_shift_sh,fut_y_shift_lo,fwd_repo_shift);				
			break;
		case futroll_scen_type.FRST_BE_FUT_SH_NR:	//"BreakEvenFutFr(flat repo)"
			QL_FAIL_COND(null(fut_y_shift_lo), "missing yield shift for back future");		
			fut_y_shift_sh = fut_fwd(ee).ctd_y_sens(ctd_idx_lo_,0,fut_y_shift_lo, disable_rnd );
			fwd_repo_shift = 0;
			break;
		case futroll_scen_type.FRST_BE_REPO:	//"BreakEvenRepo"			
			QL_FAIL_COND(null(fut_y_shift_sh), "missing yield shift for front future");
			if(null(fut_y_shift_lo))
				fut_y_shift_lo = fut_y_shift_sh;
			fwd_repo_shift = repo_ff_sens(fut_y_shift_lo, true, fut_y_shift_sh, disable_rnd,ee );
			break;
		case futroll_scen_type.FRST_BE_FUT_LO_R://"BreakEvenFutBa(incl repo)"
			QL_FAIL_COND(null(fut_y_shift_sh), "missing yield shift for front future");
			if(null(fwd_repo_shift))
				fwd_repo_shift = fut_y_shift_sh; 							
			y_shift_scen(fut_fwd(ee),ctd_idx_lo_,fut_scen_type.SST_BE_FUT_R,disable_rnd,fut_y_shift_sh,fut_y_shift_lo,fwd_repo_shift);
			break;
		case futroll_scen_type.FRST_BE_FUT_SH_R: //"BreakEvenFutFr(incl repo)"
			QL_FAIL_COND(null(fut_y_shift_lo), "missing yield shift for back future");
			if(null(fwd_repo_shift))
				fwd_repo_shift = fut_y_shift_lo; 				
			fut_y_shift_sh = fut_fwd(ee).ctd_y_sens(ctd_idx_lo_,fwd_repo_shift,fut_y_shift_lo, disable_rnd );
			break;
		default:
			fut_y_shift_sh = null<number>;
			fut_y_shift_lo = null<number>;
			fwd_repo_shift = null<number>;
		}
	}
	catch {
		fut_y_shift_sh = null<number>;
		fut_y_shift_lo = null<number>;
		fwd_repo_shift = null<number>;
	}

		
		/*
		fut_sh_.y_shift_scen(st,disable_rnd,ctd_y_shift,fut_y_shift_sh,repo_shift_sh);
		fut_lo_.y_shift_scen(st,disable_rnd,ctd_y_shift,fut_y_shift_lo,repo_shift_lo);
		
		number repo_sh 		= fut_sh_.get_repo_ctd() + repo_shift_sh;
		number repo_lo 		= fut_lo_.get_repo_ctd() + repo_shift_lo;
		date settle 		= fut_sh_.get_settle();
		date settle_fwd1 	= fut_sh_.get_settle_fwd();
		date settle_fwd2 	= fut_lo_.get_settle_fwd();
		number rf 			= nose_tail(	settle , settle_fwd1, settle_fwd2, repo_sh, repo_lo, 2,
											fut_sh_.get_dc(),fut_sh_.get_dc(),fut_sh_.get_dc(),
											fut_sh_.get_rt(),fut_sh_.get_rt(),fut_sh_.get_rt());
		fwd_repo_shift = rf - repo_fwd_;
		*/
	
}