/*	
	Bond spreads functions
	Developed by Algorithmica Research, Magnus Nyström
			
	ToDo: 
	
	Comments:	

*/
option(null: hard);	
module CORE_INT
{

	//-------------------------------------------------------
	//  barbell_spread_wing_w
	//-------------------------------------------------------
	 
	void barbell_spread_wing_w(	number dp_bu,
								number dp_sh,
								number dp_lo,
								number pvbp_bu,
								number pvbp_sh,
								number pvbp_lo,
								out number nom_w_sh, 
								out number nom_w_lo)
								//option (category: 'Financial/Bond Spreads')
	{	   		
		nom_w_sh = pvbp_bu/(2.0*pvbp_sh) ;
		nom_w_lo = pvbp_bu/(2.0*pvbp_lo) ;
		/*
		nom_w_sh = (dp_bu*mod_dur_bu)/(2.0*mod_dur_sh*dp_sh) ;
		nom_w_lo = (dp_bu*mod_dur_bu)/(2.0*mod_dur_lo*dp_lo) ;	
		*/	    	     
	}	
	//-------------------------------------------------------
	//  barbell_spread_w
	//-------------------------------------------------------
	 
	void barbell_spread_w(	number dp_bu,
							number dp_sh,
							number dp_lo,
							number pvbp_bu,
							number pvbp_sh,
							number pvbp_lo,
							out number nom_w_sh, 
							out number nom_w_lo)
							//option (category: 'Financial/Bond Spreads')
	{
		nom_w_sh = (dp_bu*pvbp_lo-dp_lo*pvbp_bu)/(dp_sh*pvbp_lo-dp_lo*pvbp_sh) ;
		nom_w_lo = (dp_sh*pvbp_bu-dp_bu*pvbp_sh)/(dp_sh*pvbp_lo-dp_lo*pvbp_sh)  ;
		/*
		nom_w_sh = (dp_bu*(mod_dur_bu-mod_dur_lo))/(dp_sh*mod_dur_sh-dp_sh*mod_dur_lo) ;
		nom_w_lo = (dp_bu*(mod_dur_bu-mod_dur_sh))/(dp_lo*mod_dur_lo-dp_lo*mod_dur_sh) ;
		*/	
	}
	//-------------------------------------------------------
	//  barbell_cf_y
	//-------------------------------------------------------
	 number barbell_cf_y(	number y_sh,
							number y_lo,																
							number pvbp_sh,
							number pvbp_lo,
							number nom_w_sh, 
							number nom_w_lo)
							//option (category: 'Financial/Bond Spreads')
	{
		return CORE_INT.portf_yld_cf_approx(	[y_sh,y_lo],[pvbp_sh,pvbp_lo], [nom_w_sh,nom_w_lo]); 
		/*return  ((y_sh*((nom_w_sh*dp_sh)/dp_bu)*mod_dur_sh)
				+(y_lo*((nom_w_lo*dp_lo)/dp_bu)*mod_dur_lo))
				/mod_dur_bu; */
	}
	//-------------------------------------------------------
	//  barbell_cf_w
	//-------------------------------------------------------
	
	 void  barbell_cf_w(	number dp_sh,
							number dp_lo,
							number nom_w_sh, 
							number nom_w_lo,
							out number cf_w_sh, 
							out number cf_w_lo, 
							out number tot_cf)
							//option (category: 'Financial/Bond Spreads') 
	{
		tot_cf = nom_w_sh*dp_sh + nom_w_lo*dp_lo;
		cf_w_sh =(nom_w_sh*dp_sh)/tot_cf;
		cf_w_lo =(nom_w_lo*dp_lo)/tot_cf;
	}
	//-------------------------------------------------------
	//  barbell_avg_y
	//-------------------------------------------------------
	
	 number barbell_avg_y(	number y_sh,
							number y_lo,						
							number dp_sh,
							number dp_lo,
							number nom_w_sh, 
							number nom_w_lo)
							//option (category: 'Financial/Bond Spreads')
	{
		return CORE_INT.portf_yld_avg([y_sh,y_lo],[dp_sh,dp_lo], [nom_w_sh,nom_w_lo]); 
	}
	
	//----------------------------------------------------------------------
	//  barbell_solve                                                             
	//----------------------------------------------------------------------
	 number barbell_solve(	number option(nullable) y_bu,
							number option(nullable) y_sh,
							number option(nullable) y_lo,
							number option(nullable) cf_y_spread,
							number dp_bu,
							number dp_sh,
							number dp_lo,
							number pvbp_bu,
							number pvbp_sh,
							number pvbp_lo,
							number solve_code)
							//option (category: 'Financial/Bond Spreads')
	{	
		number nom_w_sh,nom_w_lo;
	    barbell_spread_w(dp_bu,dp_sh,dp_lo,pvbp_bu,pvbp_sh,pvbp_lo,nom_w_sh,nom_w_lo);//FIX_ME weighting style
		
		number mod_dur_bu = pvbp_bu/dp_bu;
		number mod_dur_sh = pvbp_sh/dp_sh;
		number mod_dur_lo = pvbp_lo/dp_lo;
	  
	    if (solve_code == 1){//cf_y sprd		
			number cf_y = barbell_cf_y( y_sh,y_lo,pvbp_sh,pvbp_lo,nom_w_sh,nom_w_lo);
			return cf_y - y_bu;
	    }
	    else if (solve_code == 2){//y_bu
			number cf_y = barbell_cf_y( y_sh,y_lo,pvbp_sh,pvbp_lo,nom_w_sh,nom_w_lo);
			return cf_y - cf_y_spread;	
	    }
	    else if (solve_code == 3){//y_sh
			number x = ((nom_w_sh*dp_sh)/dp_bu)*mod_dur_sh;
			number y = ((nom_w_lo*dp_lo)/dp_bu)*mod_dur_lo;
			return (((cf_y_spread + y_bu) * mod_dur_bu - y_lo * y) / x)  ; 
	    }
	    else if (solve_code == 4){//y_lo
			number x = ((nom_w_sh*dp_sh)/dp_bu)*mod_dur_sh;
			number y = ((nom_w_lo*dp_lo)/dp_bu)*mod_dur_lo;
			return ((cf_y_spread + y_bu) * mod_dur_bu - y_sh * x )/y ; 
	    }
	    else 
			return null<number>;						
	}
}


class enum  barbell_w_style  option (category: 'Financial/Bond Spreads')
{
	CASH_RISK	option(str: "Cash/RiskNeutral"),
	WING		option(str: "EqualWing"),
	CASH_RISK_F	option(str: "Cash/RiskNeutral-Fwd"),
	WING_F		option(str: "EqualWing-Fwd")
};

//-------------------------------------------------------
//  barbell_w_list
//-------------------------------------------------------
vector(string) barbell_w_list() 
option (category: 'Financial/Bond Spreads') 
{
	vector(barbell_w_style) e;
	expand_enum(e) ;
	return string(e);
}

//-------------------------------------------------------
//  barbell_w_map
//-------------------------------------------------------
barbell_w_style barbell_w_map(string w)
option (category: 'Financial/Bond Spreads')
{
	barbell_w_style e;
	str_to_enum(w,e);
	return e;
}

module CORE_INT
{
	//-------------------------------------------------------
	//  barbell_cf_spread
	//-------------------------------------------------------
	 number barbell_cf_spread(	barbell_w_style w,
								instrument		i_sh,
								instrument		i_bu,
								instrument		i_lo)
								//option (category: 'Financial/Bond Spreads')
	{		
		try {
			number	y_sh = i_sh.yield();
			number	y_bu = i_bu.yield();
			number	y_lo = i_lo.yield();
			number	dp_sh = i_sh.dirty_price();
			number	dp_bu = i_bu.dirty_price();
			number	dp_lo = i_lo.dirty_price();
			number	pvbp_sh = i_sh.pvbp();
			number	pvbp_bu = i_bu.pvbp();
			number	pvbp_lo = i_lo.pvbp();
			
			number nom_w_sh,nom_w_lo;
			if (w == barbell_w_style.CASH_RISK) 
				CORE_INT.barbell_spread_w(dp_bu,dp_sh,dp_lo,pvbp_bu,pvbp_sh,pvbp_lo,nom_w_sh,nom_w_lo);			
			else /*BB_WING*/
				CORE_INT.barbell_spread_wing_w(dp_bu,dp_sh,dp_lo,pvbp_bu,pvbp_sh,pvbp_lo,nom_w_sh,nom_w_lo);		
			
			number cf_y = CORE_INT.barbell_cf_y(	y_sh,y_lo,pvbp_sh,pvbp_lo,nom_w_sh,nom_w_lo);
			return cf_y - y_bu;
		}
		catch {
			return null<number>;
		}
	}
	
	//-------------------------------------------------------
	//  barbell_cf_spread
	//-------------------------------------------------------
	 number barbell_bf_spread(	instrument		i_sh,
								instrument		i_bu,
								instrument		i_lo)
								//option (category: 'Financial/Bond Spreads') 
	{		
		try {
			number	y_sh = i_sh.yield();
			number	y_bu = i_bu.yield();
			number	y_lo = i_lo.yield();							
			return -y_sh + 2* y_bu - y_lo;
		}
		catch {
			return null<number>;
		}
	}
	//-------------------------------------------------------
	//  barbell_cf_spread
	//-------------------------------------------------------
	 number barbell_cf_spread(	date d,
								barbell_w_style w,
								instrument_name 		i_sh,
								instrument_name 		i_bu,
								instrument_name 		i_lo,						
								quote_side 				qs_sh, 					
								quote_side 				qs_bu, 											 					
								quote_side 				qs_lo)
								//option (category: 'Financial/Bond Spreads')
	{
		instrument i1, i2, i3;
		number q1,q2,q3;
		CORE_INT.init_quote( d, i_sh,qs_sh,i1,q1);
		CORE_INT.init_quote( d, i_bu,qs_bu,i2,q2);
		CORE_INT.init_quote( d, i_lo,qs_lo,i3,q3);		
		return barbell_cf_spread(w,i1,i2,i3);
	}
	
	//-------------------------------------------------------
	//  barbell_cf_spread
	//-------------------------------------------------------
	 number barbell_bf_spread(	date d,
								instrument_name 		i_sh,
								instrument_name 		i_bu,
								instrument_name 		i_lo,						
								quote_side 			qs_sh, 					
								quote_side 			qs_bu, 											 					
								quote_side 			qs_lo)
								//option (category: 'Financial/Bond Spreads')
	{
		instrument i1, i2, i3;
		number q1,q2,q3;
		CORE_INT.init_quote( d, i_sh,qs_sh,i1,q1);
		CORE_INT.init_quote( d, i_bu,qs_bu,i2,q2);
		CORE_INT.init_quote( d, i_lo,qs_lo,i3,q3);		
		return barbell_bf_spread(i1,i2,i3);
	}
	//-------------------------------------------------------
	//  box_spread
	//-------------------------------------------------------
	void box_spread(	instrument		leg1_sh,							
						instrument		leg1_lo,
						instrument 		leg2_sh,
						instrument		leg2_lo,
						out number		leg1_sprd,
						out number		leg2_sprd,
						out number		box_sprd) 
	
	{		
		try {
			number	y_leg1_sh = leg1_sh.yield();							
			number	y_leg1_lo = leg1_lo.yield();
			number	y_leg2_sh = leg2_sh.yield();
			number	y_leg2_lo = leg2_lo.yield();
			leg1_sprd = y_leg1_lo -y_leg1_sh;
			leg2_sprd = y_leg2_lo -y_leg2_sh; 
			box_sprd = leg2_sprd-leg1_sprd;
		}
		catch {
			leg1_sprd = null<number>;
			leg2_sprd = null<number>;
			box_sprd = null<number>;
		}
	}
	
	//-------------------------------------------------------
	//  box_spread
	//-------------------------------------------------------
	number box_spread(	date 				d,
						instrument_name 	leg1_sh,							
						instrument_name 	leg1_lo,
						instrument_name 	leg2_sh,
						instrument_name 	leg2_lo, 
						quote_side 			qs_leg1_sh,							
						quote_side 			qs_leg1_lo,
						quote_side 			qs_leg2_sh,
						quote_side 			qs_leg2_lo) 
	
	{		
		try {
			instrument i1, i2, i3, i4;
			number q1,q2,q3,q4;
			CORE_INT.init_quote( d, leg1_sh,qs_leg1_sh,i1,q1);
			CORE_INT.init_quote( d, leg1_lo,qs_leg1_lo,i2,q2);
			CORE_INT.init_quote( d, leg2_sh,qs_leg2_sh,i3,q3);
			CORE_INT.init_quote( d, leg2_lo,qs_leg2_lo,i4,q4);
			number leg1_sprd,leg2_sprd,box_sprd;
			box_spread(	i1, i2, i3, i4,leg1_sprd,leg2_sprd,box_sprd);
			return box_sprd; 
		}
		catch {
			return null<number>;
		}
	}
	
	
	//-------------------------------------------------------
	//  box_spread_w (spread neutral, risk neutral)
	// pvbp = moddur * dp / 10000 ($ per 1 bp change in yld) 
	
	/*The DV01 is analogous to the delta in derivative pricing (The Greeks)- it is the ratio of a price 
	change in output (dollars) to unit change in input (a basis point of yield). Dollar duration or DV01 is 
	the change in price in dollars, not in percentage. It gives the dollar variation in a bond's value per 
	unit change in the yield. It is often measured per 1 basis point - DV01 is short for "dollar value of 
	an 01" (or 1 basis point)*/ 
	//-------------------------------------------------------
	void box_spread_w(	number v,
						number 	leg1_sh_pvbp,							
						number  leg1_lo_pvbp,
						number  leg2_sh_pvbp,
						number  leg2_lo_pvbp,
						number 	leg1_sh_dp,							
						number  leg1_lo_dp,
						number  leg2_sh_dp,
						number  leg2_lo_dp,
						out number 	leg1_sh_nom,							
						out number  	leg1_lo_nom,
						out number  	leg2_sh_nom,
						out number  	leg2_lo_nom) 
	{	
		number d1 = leg2_sh_pvbp/leg2_sh_dp;	
		number d2 = leg1_lo_pvbp/leg1_lo_dp;
		number d3 = leg1_sh_pvbp/leg1_sh_dp;
		number d4 = leg2_lo_pvbp/leg2_lo_dp;
		
		number v1 = v * d4 * (d2-d3)/(d2*d4-d1*d3);	
		number v2 = v - v1;
		number v3 = v2 * d2/d3;
		number v4 = v - v3;
			
		leg2_sh_nom = v1 / leg2_sh_dp * 100;						
		leg1_lo_nom = v2 / leg1_lo_dp * 100;
		leg1_sh_nom = v3 / leg1_sh_dp * 100;
		leg2_lo_nom = v4 / leg2_lo_dp * 100;
	}
	//-------------------------------------------------------
	//  box_spread_anch_w
	//-------------------------------------------------------
	void box_spread_anch_w(	number 	anch_nom,							
							number 	anch_pvbp,
							number 	leg1_sh_pvbp,							
							number  leg1_lo_pvbp,
							number  leg2_sh_pvbp,
							number  leg2_lo_pvbp,							
							out number 	leg1_sh_nom,							
							out number  leg1_lo_nom,
							out number  leg2_sh_nom,
							out number  leg2_lo_nom) 
	{	
		leg1_sh_nom = anch_pvbp/leg1_sh_pvbp * anch_nom;
		leg1_lo_nom = anch_pvbp/leg1_lo_pvbp * anch_nom;
		leg2_sh_nom = anch_pvbp/leg2_sh_pvbp * anch_nom;
		leg2_lo_nom = anch_pvbp/leg2_lo_pvbp * anch_nom;		
	}
}



class enum  box_spread_w_style  option (category: 'Financial/Bond Spreads') 
{
	BSW_SPREAD_NEUTRAL option(str: "SprdNeutral"),
	BSW_LEG1_SH_PVBP option(str: "PVBPLeg1Short"),
	BSW_LEG1_LO_PVBP option(str: "PVBPLeg1Long"),
	BSW_LEG2_SH_PVBP option(str: "PVBPLeg2Short"),
	BSW_LEG2_LO_PVBP option(str: "PVBPLeg2Long"),
	BSW_SPREAD_NEUTRAL_F option(str: "SprdNeutral-Fwd"),
	BSW_LEG1_SH_PVBP_F option(str: "PVBPLeg1Short-Fwd"),
	BSW_LEG1_LO_PVBP_F option(str: "PVBPLeg1Long-Fwd"),
	BSW_LEG2_SH_PVBP_F option(str: "PVBPLeg2Short-Fwd"),
	BSW_LEG2_LO_PVBP_F option(str: "PVBPLeg2Long-Fwd") 
};

//-------------------------------------------------------
//  box_spread_w_list
//-------------------------------------------------------
vector(string) box_spread_w_list() 
option (category: 'Financial/Bond Spreads')
{
	vector(box_spread_w_style) e;
	expand_enum(e) ;
	return string(e);
}

//-------------------------------------------------------
//  box_spread_w_map
//-------------------------------------------------------
 box_spread_w_style box_spread_w_map(string w)
option (category: 'Financial/Bond Spreads') 
{
	box_spread_w_style e;
	str_to_enum(w,e);
	return e;
}
//-------------------------------------------------------
// bond_spread
//-------------------------------------------------------
class bond_spread option (category: 'Financial/Bond Spreads')
{
   	public:
		bond_spread(date, vector(instrument), vector(instrument), logical);
		bond_spread(date, vector(instrument), vector(instrument), vector(number),logical);								
		bond_spread(bond_spread);
		
		virtual void 	set_nom(vector(number));
		vector(number) 	get_nom(logical);
		number 			net_cash(logical fwd = false);			
		number 			net_cash_fund(number,day_count_method,rate_type,out number);
		number 			pl(	vector(number),day_count_method,rate_type,cpn_reinv_type,out vector(number)  );
		number 			coc(vector(number),day_count_method,rate_type,cpn_reinv_type,out vector(number)  );
		virtual number 	pvbp(logical,out vector(number)  ,out number,out number);
		void			cashflow_w(logical,out vector(number)  ,out vector(number)  );
		virtual number	dol_convexity(logical,out vector(number)  ,out number,out number);
		void			cashflow_yld(logical,out number,out number);
		void			average_yld(logical,out number,out number);

	protected:
					
		void 			init();
		
	public:
		date				d_;	
		vector(instrument)	inst_; 				
		vector(instrument)	inst_fwd_;	
		vector(number)		nom_;				
		logical				buy_;
		
		vector(number) 		pvbp_;							
		vector(number)		pvbp_fwd_;			
		vector(number)		dp_;							
		vector(number)		dp_fwd_;			
																	
};
//-------------------------------------------------------
// constructor
//-------------------------------------------------------
bond_spread.bond_spread(date		d,
						vector(instrument) inst, 				
						vector(instrument) inst_fwd,						
						logical		buy) 
						: d_(d), inst_(clone_vector(inst)), inst_fwd_(clone_vector(inst_fwd)), buy_(buy)

{
	init();								
}
//-------------------------------------------------------
// constructor
//-------------------------------------------------------
bond_spread.bond_spread(date		d,
						vector(instrument) inst, 				
						vector(instrument) inst_fwd,
						vector(number)		nom,						
						logical		buy) 
						: d_(d), inst_(clone_vector(inst)), inst_fwd_(clone_vector(inst_fwd)), nom_(clone_vector(nom)), buy_(buy)

{
	init();	
	integer size = v_size(inst_);
	QL_FAIL_COND(size != v_size(nom_),"invalid size of nominal vector");	

	for(integer i = 0; i < size ; i++)
		QL_FAIL_COND(null(nom_[i]) || nom_[i] <= 0, strcat(["invalid nominal (bond ", str(i),")"]));						
}	
//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
bond_spread.bond_spread(bond_spread f) 
						:d_(f.d_), inst_(clone_vector(f.inst_)), inst_fwd_(clone_vector(f.inst_fwd_)), buy_(f.buy_), pvbp_(clone_vector(f.pvbp_)), 
						pvbp_fwd_(clone_vector(f.pvbp_fwd_)), dp_(clone_vector(f.dp_)), dp_fwd_(clone_vector(f.dp_fwd_))
{}
//-------------------------------------------------------
// init
//-------------------------------------------------------
void bond_spread.init()
{
	integer size = v_size(inst_);
	QL_FAIL_COND(size != v_size(inst_fwd_), "invalid size of instrument vector");

	string err_name;
	QL_REQUIRE(CORE_INT.check_instr_type(inst_,[instr_type.FIXED_CPN_BOND,instr_type.BILL,instr_type.ZERO_CPN_BOND],err_name), 
				"invalid instrument type (only bills, bonds and zeros supported)", err_name, true);

	resize(pvbp_, size );							
	resize(pvbp_fwd_, size );			
	resize(dp_, size );							
	resize(dp_fwd_, size );

	date settle 	= inst_[0].settle_date();
	date settle_fwd = inst_fwd_[0].settle_date();

	for(integer i = 0; i < size ; i++){

		QL_REQUIRE(inst_.instr_type() == inst_fwd_.instr_type(),"invalid instrument type (fwd instrument is a different instrument type)");

		pvbp_[i] 		= inst_[i].pvbp();
		pvbp_fwd_[i] 	= inst_fwd_[i].pvbp();	
		dp_[i] 			= inst_[i].dirty_price();	
		dp_fwd_[i] 		= inst_fwd_[i].dirty_price();

		QL_FAIL_COND(null(pvbp_[i]) || null(pvbp_fwd_[i] ) || null(dp_[i]) || null(dp_fwd_[i] ),"invalid bond", inst_[i].name(), true);			
		QL_FAIL_COND(settle != inst_[i].settle_date(), "invalid settlement date", inst_[i].name(), true);
		QL_FAIL_COND(settle_fwd != inst_fwd_[i].settle_date(), "invalid forward settlement date", inst_fwd_[i].name(), true);
	}
}
//-------------------------------------------------------
// set_nom
//-------------------------------------------------------
void bond_spread.set_nom(vector(number) nom) 

{
	integer size = v_size(inst_);
	QL_FAIL_COND(v_size(nom) != size ,"invalid size of nominal vector");	

	resize(nom_, size);
	for(integer i = 0; i < size ; i++){
		QL_FAIL_COND(nom[i] <= 0,"invalid nominal", inst_[i].name(), true);		
		nom_[i] = nom[i];
	}					
}
//-------------------------------------------------------
// get_nom
//-------------------------------------------------------
vector(number) bond_spread.get_nom(logical signs) 
{
	integer size = v_size(inst_);
	if(null(nom_) || v_size(nom_) != size)
		return null<vector(number)>;

	if(signs)
		return clone_vector(nom_);
	else {
		vector(number) t = clone_vector(nom_);
		for(integer i = 0; i < v_size(t) ; i++)
			t[i] = abs(t[i]);
		return t;
	}
}
//-------------------------------------------------------
// net_cash
//-------------------------------------------------------	
number bond_spread.net_cash(logical fwd)
{
	integer size = v_size(inst_);
	if(null(nom_) || v_size(nom_) != size)
		return null<number>;
	
	number sum = 0;
	if(fwd) {
		for(integer i = 0; i < size ; i++)
			sum += (-nom_[i]*dp_fwd_[i]);	
	}
	else {
		for(integer i = 0; i < size ; i++)
			sum += (-nom_[i]*dp_[i]);
	}
	return sum/100;
}	
//-------------------------------------------------------
// net_cash_fund
//-------------------------------------------------------	
number bond_spread.net_cash_fund(	number 				take_out_rate,
									day_count_method 	dc, 
									rate_type 			rt,		
									out number 			cash_diff_int)
{
	date settle = inst_[0].settle_date();
	date settle_fwd = inst_fwd_[0].settle_date();
	number cash_diff = net_cash(false);
	number cash_diff_fund;				
	if(cash_diff != 0){					
		//cash_diff_fund= CORE_INT.fv(cash_diff, take_out_rate, rt, settle, settle_fwd, dc);
		cash_diff_fund= future_value(cash_diff, take_out_rate, settle, settle_fwd, dc, rt);

		cash_diff_int = cash_diff_fund - cash_diff;
	}
	else 
		cash_diff_fund = cash_diff_int = 0;	
	
	return cash_diff_fund;
}

//-------------------------------------------------------
// coc
//-------------------------------------------------------	
number bond_spread.coc(	vector(number) 		r,
						day_count_method 	dc, 
						rate_type 			rt,
						cpn_reinv_type crt,
						out vector(number)   	coc)
{
	integer size = v_size(inst_);

	if(null(nom_) || v_size(nom_) != size || v_size(r) != size)
		return null<number>;

	resize(coc, size);
	number reinv_cpns = 0, no_reinv_cpns = 0;
	number dp, acc, acc_fwd, dp_fwd, fund, net_fund;

	for(integer i = 0; i < size ; i++){
		CORE_INT.hor_y_bond(inst_[i],inst_fwd_[i],d_,r[i],null<number>, null<disc_func>, dc, rt, abs(nom_[i]), crt, dp, dp_fwd, 
							acc,acc_fwd,reinv_cpns,no_reinv_cpns,fund,net_fund,coc[i]) ;
		coc[i] *= nom_[i];
	}
	return v_sum(coc);	
}			
//-------------------------------------------------------
// pl
//-------------------------------------------------------	
number bond_spread.pl(	vector(number) 		r,
						day_count_method 	dc, 
						rate_type 			rt,
						cpn_reinv_type 	crt,
						out vector(number)    pl)
{
	integer size = v_size(inst_);

	if(null(nom_) || v_size(nom_) != size || v_size(r) != size)
		return null<number>;
	
	resize(pl, size);
	number 	cash_inst,cash_inst_fwd,fund_cost;
	
	for(integer i = 0; i < size ; i++)
		pl[i] = CORE_INT.pl_bond(	inst_[i],inst_fwd_[i],d_,r[i],null<number>,null<disc_func>,dc, rt, 	
									abs(nom_[i]),crt, nom_[i] > 0 ,cash_inst,cash_inst_fwd,fund_cost) ;
	
	return v_sum(pl);		
}
//-------------------------------------------------------
// pvbp
//-------------------------------------------------------	
number bond_spread.pvbp(logical 			fwd,
						out vector(number)   	all_pvbp,
						out number  option(nullable)			long_pvbp,
						out number  option(nullable)			short_pvbp)
{
	integer size = v_size(inst_);
	if(null(nom_) || v_size(nom_) != size)
		return null<number>;

	vector(number) pvbp = fwd ? pvbp_fwd_ : pvbp_;
	long_pvbp = short_pvbp = 0;
	resize(all_pvbp,size);

	for(integer i = 0; i < size; i++) {
		all_pvbp[i] = nom_[i]*pvbp[i]/100; 
		if(nom_[i] >= 0)
			long_pvbp += all_pvbp[i];
		else
			short_pvbp += all_pvbp[i];
	}
	
	return long_pvbp + short_pvbp ;	
}
//-------------------------------------------------------
// cashflow_yld
//-------------------------------------------------------	
void bond_spread.cashflow_yld(	logical 	fwd,
								out number 	long_y,
								out number 	short_y)
{
	integer size = v_size(inst_);
	if(null(nom_) || v_size(nom_) != size){
		long_y = short_y = null<number>;
		return;
	}

	vector(number) pvbp = fwd ? pvbp_fwd_ : pvbp_;
	vector(instrument) inst = fwd ? inst_fwd_ : inst_;

	vector(number) long_pvbp, short_pvbp, l_y, s_y, long_nom, short_nom;
			
	for(integer i = 0; i < size; i++) {
		if(nom_[i] >= 0){
			push_back(long_pvbp,pvbp[i]);
			push_back(l_y,inst[i].yield());
			push_back(long_nom,nom_[i]);
		}
		else {
			push_back(short_pvbp,pvbp[i]);
			push_back(s_y,inst[i].yield());
			push_back(short_nom,-nom_[i]);
		}
	}		
	long_y 	= CORE_INT.portf_yld_cf_approx(l_y, long_pvbp, long_nom);
	short_y = CORE_INT.portf_yld_cf_approx(s_y, short_pvbp, short_nom);
}

//-------------------------------------------------------
// average_yld
//-------------------------------------------------------	
void bond_spread.average_yld(	logical 	fwd,
								out number		long_y,
								out number 	short_y)
{
	integer size = v_size(inst_);
	if(null(nom_) || v_size(nom_) != size){
		long_y = short_y = null<number>;
		return;
	}

	vector(number) dp = fwd ? dp_fwd_ : dp_;
	vector(instrument) inst = fwd ? inst_fwd_ : inst_;

	vector(number) long_dp, short_dp, l_y, s_y, long_nom, short_nom;
			
	for(integer i = 0; i < size; i++) {
		if(nom_[i] >= 0){
			push_back(long_dp,dp[i]);
			push_back(l_y,inst[i].yield());
			push_back(long_nom,nom_[i]);
		}
		else {
			push_back(short_dp,dp[i]);
			push_back(s_y,inst[i].yield());
			push_back(short_nom,-nom_[i]);
		}
	}		
	long_y 	= CORE_INT.portf_yld_avg(l_y, long_dp, long_nom);
	short_y = CORE_INT.portf_yld_avg(s_y, short_dp, short_nom);
}
//-------------------------------------------------------
// cashflow_w
//-------------------------------------------------------	
void bond_spread.cashflow_w(logical 			fwd,
							out vector(number)   	long_w,
							out vector(number)   	short_w)
{
	integer size = v_size(inst_);
	if(null(nom_) || v_size(nom_) != size){
		long_w = short_w = null<vector(number)>;
		return;
	}
	vector(number) dp = fwd ? dp_fwd_ : dp_;
	number long_cf = 0;
	number short_cf = 0;
	for(integer i = 0; i < size; i++) {
		if(nom_[i] >= 0)
			long_cf += nom_[i]*dp[i];
		else 
			short_cf += nom_[i]*dp[i];
	}
	resize(long_w, 0);
	resize(short_w, 0);

	for(integer i = 0; i < size; i++) {
		if(nom_[i] >= 0)
			push_back(long_w, (nom_[i]*dp[i])/long_cf);
		else
			push_back(short_w, (nom_[i]*dp[i])/short_cf);
	}
}
//-------------------------------------------------------
// convexity
//-------------------------------------------------------	
number bond_spread.dol_convexity(	logical fwd,
									out vector(number)   	all_dol_con,
									out number long_con,
									out number short_con)
{
	integer size = v_size(inst_);
	if(null(nom_) || v_size(nom_) != size)
		return null<number>;
	
	//vector(number) long_w, short_w;
	//cashflow_w(fwd,long_w,short_w);

	vector(instrument) inst = fwd ? inst_fwd_ : inst_;
	long_con = short_con = 0;
	
	resize(all_dol_con, size);

	number l = 0;
	number s = 0;
	for(integer i = 0; i < size; i++) {
		all_dol_con[i] = nom_[i]*inst[i].dol_convexity()/100;

		if(nom_[i] >= 0){
			long_con += all_dol_con[i];
			l++;
		}
		else{
			short_con += all_dol_con[i];
			s++;
		}
	}		
	return long_con + short_con ;	
}

//-------------------------------------------------------
// box_spread
//-------------------------------------------------------
 class box_spread  : public  bond_spread option (category: 'Financial/Bond Spreads')
{
   	public:
		box_spread(date, vector(instrument), vector(instrument), logical);
		box_spread(date, vector(instrument), vector(instrument), vector(number)	,logical);								
		box_spread(box_spread);
		void 			spread(logical, out number,out number,out number);			
		virtual void 	nominals(number ,box_spread_w_style,logical, out number,out number,out number,out number);
		virtual void 	buysell(out logical option(nullable),out logical option(nullable),out logical option(nullable),out logical option(nullable));
		virtual void 	set_nom(vector(number));
					
		number 			pl(	number ,number,number,number,day_count_method,rate_type,cpn_reinv_type,
							out number,out number,out number,out number);
		number 			coc(number ,number,number,number,day_count_method,rate_type,cpn_reinv_type,
							out number,out number,out number,out number);
		void			yield_spread(logical,out number,out number,out number,out number,
									out number,out number);
	protected:
					
		
		void set_nom_signs(out vector(number)  );							
																	
};
//-------------------------------------------------------
// constructor
//-------------------------------------------------------
box_spread.box_spread(	date				d,
						vector(instrument) 	inst, 				
						vector(instrument) 	inst_fwd,						
						logical				buy) 
						: bond_spread(d, inst, inst_fwd, buy)

{
	QL_REQUIRE(v_size(inst_)== 4,"invalid size of instrument vector");							
}
//-------------------------------------------------------
// constructor
//-------------------------------------------------------
box_spread.box_spread(	date		d,
						vector(instrument) inst, 				
						vector(instrument) inst_fwd,
						vector(number)		nom,						
						logical		buy) 
						: bond_spread(d, inst, inst_fwd, nom, buy)

{
	QL_REQUIRE(v_size(inst_)== 4,"invalid size of instrument vector");			
	set_nom_signs(nom_)	;				
}	
//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
box_spread.box_spread(box_spread f) 
					 :bond_spread(f)
{}
//-------------------------------------------------------
// set_nom_signs
//-------------------------------------------------------
void box_spread.set_nom_signs(out vector(number)   nom)
{
	if (buy_) {
		nom_[0] = -abs(nom_[0]);
		nom_[1] = abs(nom_[1]);
		nom_[2] = abs(nom_[2]);
		nom_[3] = -abs(nom_[3]);
	}
	else {
		nom_[0] = abs(nom_[0]);
		nom_[1] = -abs(nom_[1]);		
		nom_[2] = -abs(nom_[2]);
		nom_[3] = abs(nom_[3]);		
	}
}
//-------------------------------------------------------
// set_nom
//-------------------------------------------------------
void box_spread.set_nom(vector(number) nom) 
{
	bond_spread.set_nom(nom); 
	set_nom_signs(nom_)	;		
}		
//-------------------------------------------------------
// coc
//-------------------------------------------------------	
number box_spread.coc(	number r1,
						number r2,
						number r3,
						number r4,
						day_count_method dc, 
						rate_type rt,
						cpn_reinv_type crt,
						out number coc1,
						out number coc2,
						out number coc3,
						out number coc4)
{
	vector(number) r = [r1,r2,r3,r4];
	vector(number) coc ;
	number c = bond_spread.coc(r,dc,rt,crt,coc);
	coc1 = coc[0];
	coc2 = coc[1];
	coc3 = coc[2];
	coc4 = coc[3];
	return c;	
}			
//-------------------------------------------------------
// pl
//-------------------------------------------------------	
number box_spread.pl(	number r1,
						number r2,
						number r3,
						number r4,
						day_count_method dc, 
						rate_type rt,
						cpn_reinv_type crt,
						out number pl1,
						out number pl2,
						out number pl3,
						out number pl4)
{
	vector(number) r = [r1,r2,r3,r4];
	vector(number) pl ;
	number p = bond_spread.pl(r,dc,rt,crt,pl);
	pl1 = pl[0];
	pl2 = pl[1];
	pl3 = pl[2];
	pl4 = pl[3];
	return p;	
}		
//-------------------------------------------------------
// nominals
//-------------------------------------------------------
void box_spread.nominals(	number scale,
							box_spread_w_style 	w,
							logical signs,
							out number leg1_sh_nom,
							out number leg1_lo_nom,
							out number leg2_sh_nom,
							out number leg2_lo_nom)
{
	leg1_sh_nom = leg1_lo_nom = leg2_sh_nom = leg2_lo_nom = null<number>;
	switch (w){
	case box_spread_w_style.BSW_SPREAD_NEUTRAL_F: 
		CORE_INT.box_spread_w(scale,pvbp_fwd_[0],pvbp_fwd_[1],pvbp_fwd_[2],pvbp_fwd_[3],dp_[0],dp_[1],dp_[2],dp_[3],
					leg1_sh_nom,leg1_lo_nom,leg2_sh_nom,leg2_lo_nom); 
		break;
	case box_spread_w_style.BSW_LEG1_SH_PVBP_F:
		CORE_INT.box_spread_anch_w(	scale,pvbp_fwd_[0],pvbp_fwd_[0],pvbp_fwd_[1],pvbp_fwd_[2],pvbp_fwd_[3],							
						leg1_sh_nom,leg1_lo_nom,leg2_sh_nom,leg2_lo_nom) ;
		break; 
	case box_spread_w_style.BSW_LEG1_LO_PVBP_F:
		CORE_INT.box_spread_anch_w(	scale,pvbp_fwd_[1],pvbp_fwd_[0],pvbp_fwd_[1],pvbp_fwd_[2],pvbp_fwd_[3],							
						leg1_sh_nom,leg1_lo_nom,leg2_sh_nom,leg2_lo_nom) ;
		break; 
	case box_spread_w_style.BSW_LEG2_SH_PVBP_F:
		CORE_INT.box_spread_anch_w(	scale,pvbp_fwd_[2],pvbp_fwd_[0],pvbp_fwd_[1],pvbp_fwd_[2],pvbp_fwd_[3],							
						leg1_sh_nom,leg1_lo_nom,leg2_sh_nom,leg2_lo_nom) ;
		break; 
	case box_spread_w_style.BSW_LEG2_LO_PVBP_F:
		CORE_INT.box_spread_anch_w(	scale,pvbp_fwd_[3],pvbp_fwd_[0],pvbp_fwd_[1],pvbp_fwd_[2],pvbp_fwd_[3],							
						leg1_sh_nom,leg1_lo_nom,leg2_sh_nom,leg2_lo_nom) ;
		break;
	case box_spread_w_style.BSW_SPREAD_NEUTRAL: 
		CORE_INT.box_spread_w(scale,pvbp_[0],pvbp_[1],pvbp_[2],pvbp_[3],dp_[0],dp_[1],dp_[2],dp_[3],
					leg1_sh_nom,leg1_lo_nom,leg2_sh_nom,leg2_lo_nom); 
		break;
	case box_spread_w_style.BSW_LEG1_SH_PVBP:
		CORE_INT.box_spread_anch_w(	scale,pvbp_[0],pvbp_[0],pvbp_[1],pvbp_[2],pvbp_[3],							
						leg1_sh_nom,leg1_lo_nom,leg2_sh_nom,leg2_lo_nom) ;
		break; 
	case box_spread_w_style.BSW_LEG1_LO_PVBP:
		CORE_INT.box_spread_anch_w(	scale,pvbp_[1],pvbp_[0],pvbp_[1],pvbp_[2],pvbp_[3],							
						leg1_sh_nom,leg1_lo_nom,leg2_sh_nom,leg2_lo_nom) ;
		break; 
	case box_spread_w_style.BSW_LEG2_SH_PVBP:
		CORE_INT.box_spread_anch_w(	scale,pvbp_[2],pvbp_[0],pvbp_[1],pvbp_[2],pvbp_[3],							
						leg1_sh_nom,leg1_lo_nom,leg2_sh_nom,leg2_lo_nom) ;
		break; 
	case box_spread_w_style.BSW_LEG2_LO_PVBP:
		CORE_INT.box_spread_anch_w(	scale,pvbp_[3],pvbp_[0],pvbp_[1],pvbp_[2],pvbp_[3],							
						leg1_sh_nom,leg1_lo_nom,leg2_sh_nom,leg2_lo_nom) ;
		break;

	default:
		QL_FAIL("invalid box spread weighting style");
	
	}
	if(signs) {
		if(buy_) {
			leg1_sh_nom *= -1;		
			leg2_lo_nom *= -1; 		
		}
		else {
			leg1_lo_nom *= -1;		
			leg1_sh_nom *= -1; 
		}
	}
}

//-------------------------------------------------------
// buysell
//-------------------------------------------------------
void box_spread.buysell(out logical option(nullable) leg1_sh_buy,
						out logical option(nullable) leg1_lo_buy,
						out logical option(nullable) leg2_sh_buy,
						out logical option(nullable) leg2_lo_buy)
{				
	leg1_sh_buy = !buy_;	
	leg1_lo_buy = buy_;		
	leg2_lo_buy = !buy_;
	leg2_sh_buy = buy_;			
}
//-------------------------------------------------------
// spread
//-------------------------------------------------------
void box_spread.spread(	logical fwd,
						out number leg1_sprd,
						out number leg2_sprd,
						out number box_sprd)
{
	if(fwd)
		CORE_INT.box_spread(inst_fwd_[0],inst_fwd_[1],inst_fwd_[2],inst_fwd_[3],leg1_sprd,leg2_sprd, box_sprd) ;
	else
		CORE_INT.box_spread(inst_[0],inst_[1],inst_[2],inst_[3],leg1_sprd,leg2_sprd, box_sprd) ;	
}
//-------------------------------------------------------
// spread
//-------------------------------------------------------
void box_spread.yield_spread(	logical fwd,
								out number cf_long_y,
								out number cf_short_y,
								out number cf_y_sprd,
								out number avg_long_y,
								out number avg_short_y,
								out number avg_y_sprd)
{
	cashflow_yld(fwd,cf_long_y,cf_short_y);
	cf_y_sprd = !buy_ ? cf_long_y -  cf_short_y : cf_short_y - cf_long_y;

	average_yld(fwd,avg_long_y,avg_short_y);
	avg_y_sprd = buy_ ? avg_long_y -  avg_short_y : avg_short_y -  avg_long_y;	
}
//-------------------------------------------------------
// barbell_spread
//-------------------------------------------------------
 class barbell_spread  : public  bond_spread option (category: 'Financial/Bond Spreads')
{
   	public:
		barbell_spread(date, vector(instrument), vector(instrument), logical);
		barbell_spread(date, vector(instrument), vector(instrument), vector(number)	,logical);								
		barbell_spread(barbell_spread);
		virtual void 	yield_spread(logical ,out number,out number,out number,out number);	
		number 			bfly_spread(logical);			
		virtual void 	nominals(number ,barbell_w_style,logical, out number,out number,out number);
		virtual void 	buysell(out logical option(nullable),out logical option(nullable),out logical option(nullable));
		virtual void 	set_nom(vector(number));
					
		number 			pl(	number ,number,number,day_count_method,rate_type,cpn_reinv_type,
							out number,out number,out number);
		number 			coc(number ,number,number,day_count_method,rate_type,cpn_reinv_type,
							out number,out number,out number);
		virtual number 	pvbp(logical,out vector(number)  , out number  option(nullable),out number  option(nullable));
		virtual number 	dol_convexity(logical,out vector(number)  , out number ,out number );
	protected:
					
		
		void set_nom_signs(out vector(number)  );							
		void nominal_w(	barbell_w_style,out number,out number);															
};

//-------------------------------------------------------
// constructor
//-------------------------------------------------------
barbell_spread.barbell_spread(	date		d,
								vector(instrument) inst, 				
								vector(instrument) inst_fwd,						
								logical		buy) 
								: bond_spread(d, inst, inst_fwd, buy)

{
	QL_REQUIRE(v_size(inst_)== 3,"invalid size of instrument vector");								
}
//-------------------------------------------------------
// constructor
//-------------------------------------------------------
barbell_spread.barbell_spread(	date		d,
								vector(instrument) inst, 				
								vector(instrument) inst_fwd,
								vector(number)		nom,						
								logical		buy) 
								: bond_spread(d, inst, inst_fwd, nom, buy)

{
	QL_REQUIRE(v_size(inst_)== 3,"invalid size of instrument vector");			
	set_nom_signs(nom_);				
}	
//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
barbell_spread.barbell_spread(barbell_spread f) 
					 		:bond_spread(f)
{}

//-------------------------------------------------------
// set_nom_signs
//-------------------------------------------------------
void barbell_spread.set_nom_signs(out vector(number) nom)
{
	if (buy_) {
		nom_[0] = abs(nom_[0]);
		nom_[1] = -abs(nom_[1]);
		nom_[2] = abs(nom_[2]);
	}
	else {
		nom_[0] = -abs(nom_[0]);
		nom_[1] = abs(nom_[1]);		
		nom_[2] = -abs(nom_[2]);		
	}
}
//-------------------------------------------------------
// set_nom
//-------------------------------------------------------
void barbell_spread.set_nom(vector(number) nom) 
{
	bond_spread.set_nom(nom);
	set_nom_signs(nom_);			
}		
//-------------------------------------------------------
// coc
//-------------------------------------------------------	
number barbell_spread.coc(	number r1,
							number r2,
							number r3,								
							day_count_method dc, 
							rate_type rt,
							cpn_reinv_type crt,
							out number coc1,
							out number coc2,
							out number coc3)
{
	vector(number) r = [r1,r2,r3];
	vector(number) coc ;
	number c = bond_spread.coc(r,dc,rt,crt,coc);
	coc1 = coc[0];
	coc2 = coc[1];
	coc3 = coc[2];
	return c;	
}			
//-------------------------------------------------------
// pl
//-------------------------------------------------------	
number barbell_spread.pl(	number r1,
							number r2,
							number r3,
							day_count_method dc, 
							rate_type rt,
							cpn_reinv_type crt,
							out number pl1,
							out number pl2,
							out number pl3)
{
	vector(number) r = [r1,r2,r3];
	vector(number) pl ;
	number p = bond_spread.pl(r,dc,rt,crt,pl);
	pl1 = pl[0];
	pl2 = pl[1];
	pl3 = pl[2];
	return p;	
}
//-------------------------------------------------------
// pvbp
//-------------------------------------------------------	
number barbell_spread.pvbp(	logical fwd,
							out vector(number)   all_pvbp,
							out number bullet_pvbp,
							out number barbell_pvbp)
{
	number long_pvbp,short_pvbp;
	number net = bond_spread.pvbp(fwd,all_pvbp,long_pvbp,short_pvbp);

	bullet_pvbp = buy_ ? short_pvbp : long_pvbp;
	barbell_pvbp = !buy_ ? short_pvbp : long_pvbp;

	return net ;	
}
//-------------------------------------------------------
// pvbp
//-------------------------------------------------------	
number barbell_spread.dol_convexity(logical fwd,
									out vector(number)   all_con,
									out number bullet_con,
									out number barbell_con)
{
	number long_con,short_con;		
	number net = bond_spread.dol_convexity(fwd,all_con,long_con,short_con);

	bullet_con= buy_ ? short_con: long_con;
	barbell_con= !buy_ ? short_con: long_con;

	return net ;	
}
//-------------------------------------------------------
// nominal_w
//-------------------------------------------------------
void barbell_spread.nominal_w(	barbell_w_style w,									
								out number nom_w_sh,
								out number nom_w_lo)
{
	switch (w){
	case barbell_w_style.CASH_RISK: 
		CORE_INT.barbell_spread_w(dp_[1],dp_[0],dp_[2],pvbp_[1],pvbp_[0],pvbp_[2],nom_w_sh,nom_w_lo);			
		break;
	case barbell_w_style.WING:
		CORE_INT.barbell_spread_wing_w(dp_[1],dp_[0],dp_[2],pvbp_[1],pvbp_[0],pvbp_[2],nom_w_sh,nom_w_lo);		
		break; 
	case barbell_w_style.CASH_RISK_F: 
		CORE_INT.barbell_spread_w(dp_[1],dp_[0],dp_[2],pvbp_fwd_[1],pvbp_fwd_[0],pvbp_fwd_[2],nom_w_sh,nom_w_lo);			
		break;
	case barbell_w_style.WING_F:
		CORE_INT.barbell_spread_wing_w(dp_[1],dp_[0],dp_[2],pvbp_fwd_[1],pvbp_fwd_[0],pvbp_fwd_[2],nom_w_sh,nom_w_lo);		
		break; 		
	default:
		QL_FAIL("invalid barbell spread weighting style");		
	}		
}
//-------------------------------------------------------
// nominals
//-------------------------------------------------------
void barbell_spread.nominals(	number 			scale,
								barbell_w_style w,
								logical 		signs,
								out number  option(nullable)		sh_nom,	
								out number  option(nullable)		bu_nom,									
								out number  option(nullable)		lo_nom)
{
	sh_nom = bu_nom = lo_nom = null<number>;

	number nom_w_sh,nom_w_lo;
	nominal_w(w, nom_w_sh,nom_w_lo);

	bu_nom = abs(scale);
	sh_nom = bu_nom * nom_w_sh;
	lo_nom = bu_nom * nom_w_lo;

	if(signs) {
		if(buy_) 
			bu_nom *= -1;						
		else {
			lo_nom *= -1;		
			sh_nom *= -1; 
		}
	}
}
//-------------------------------------------------------
// buysell
//-------------------------------------------------------
void barbell_spread.buysell(out logical option(nullable) sh_buy,
							out logical option(nullable) bu_buy,								
							out logical option(nullable) lo_buy)
{				
	sh_buy = buy_;	
	bu_buy = !buy_;		
	lo_buy = buy_;		
}
//-------------------------------------------------------
// spread 
//-------------------------------------------------------
void barbell_spread.yield_spread(logical fwd,
										out number cfy,
										out number cfy_sprd,
										out number avgy,
										out number avg_sprd)
{
	number long_y,short_y;
	cashflow_yld(fwd,long_y,short_y);
	cfy = buy_ ? long_y : short_y;
	cfy_sprd = buy_ ? long_y -  short_y : short_y - long_y;

	average_yld(fwd,long_y,short_y);
	avgy = buy_ ? long_y : short_y;
	avg_sprd = buy_ ? long_y -  short_y : short_y -  long_y;
}
//-------------------------------------------------------
// spread 
//-------------------------------------------------------
number barbell_spread.bfly_spread(logical fwd)
{
	return fwd ? CORE_INT.barbell_bf_spread(inst_fwd_[0],inst_fwd_[1],inst_fwd_[2]) : CORE_INT.barbell_bf_spread(inst_[0],inst_[1],inst_[2]);
}	
