/*	
	Disc model functions
	Developed by Algorithmica Research, Magnus Nyström
	
	
	Comments: 
	I denna ql fil finns funktionerna tension_spline_fwd_exact(...) och tension_spline_fwd_spread(...). Dessa kräver 
	minst Quantlab version 3.1315

	ToDo:
	1. pvbp calc from cashflows_q MUST BE CHECKED [calc_pvbp_cfq]
	2. support for chop_short(cashflows) [disc_z_model_step_boot.df(cashflows_base crv)]
	3. unsupported disc model _spr for max_smooth and max_smooth_spreads[max_smooth.df]
	4. unsupported disc model _spr for bootstrap_on
	5. unsupported disc model _spr for tanggaard
	6. 
	7. unsupported fwd model: _ext for tension_spreads [fwd_z_model_tension.fwd]

*/
option(null: hard);

class enum disc_z_model_type option (category: 'Yield Curve/003 Curve Building') 
{
	//DZ_BOOT_ON			option(str: "Boot_on"),	//COMPAT.__bootstrap_on --> removed because interpolator argument no longer supported (therefore this model == DZ_ON_STEP) 
	DZ_BOOT					option(str: "Boot"),
	DZ_ON_STEP				option(str: "On_Step"),	//-->COMPAT.__bootstrap_on
	DZ_MAX_SMOOTH			option(str: "Max_Smooth"),
	DZ_MAX_SMOOTH_SPREAD	option(str: "Max_Smooth_Spread"),
	DZ_ON_STEP_BOOT 		option(str: "On_Step_Boot"),//-->COMPAT.__bootstrap_on
	DZ_TANGGAARD			option(str: "Tanggaard"),
	DZ_TENSION_W_EXACT		option(str: "Tension_w_Exact"),
	DZ_TENSION_W_SPREAD		option(str: "Tension_w_Spread"),
	DZ_TENSION_Y_EXACT		option(str: "Tension_y_Exact"),
	DZ_TENSION_Y_SPREAD		option(str: "Tension_y_Spread"),
	DZ_TENSION_Z_EXACT		option(str: "Tension_z_Exact"),
	DZ_TENSION_Z_SPREAD		option(str: "Tension_z_Spread"),
	DZ_NS					option(str: "Nelson_Siegel")
};

vector(string) disc_z_model_list()  
{
	vector(disc_z_model_type) e;
	expand_enum(e) ;
	return string(e);
}

class enum disc_ip_type option (category: 'Yield Curve/003 Curve Building')
{ 
	DIPT_LINEAR			option(str: "Linear"),
	DIPT_SPLINE_NAT		option(str: "Spline_Nat"),
	DIPT_SPLINE_FIN		option(str: "Spline_Fin"),
	DIPT_STEP			option(str: "Step"),
	DIPT_HAGAN_WEST		option(str: "Hagan_West"),
	DIPT_HERMITE_AKIMA	option(str: "Hermite_Akima"),
	DIPT_HERMITE_CR		option(str: "Hermite_cr"),
	DIPT_HERMITE_FINDIFF option(str: "Hermite_Findiff"),
	DIPT_HERMITE_FB		option(str: "Hermite_fb"),
	DIPT_HERMITE_KRUGER	option(str: "Hermite_Kruger"),
	DIPT_HERMITE_MONO	option(str: "Hermite_Mono"),
	DIPT_HERMITE_P		option(str: "Hermite_p")
};

vector(string) disc_ip_list()  
{
	vector(disc_ip_type) e;
	expand_enum(e) ;
	return string(e);
}

class enum disc_extrap_type option (category: 'Yield Curve/003 Curve Building') 
{ 
	DXP_MODEL 		option(str: "Model"),
	DXP_MODEL_FLAT 	option(str: "Model_Flat"),
	DXP_ON_FLAT 	option(str: "ON_Flat")
};

vector(string) disc_extrap_list()  
{
	vector(disc_extrap_type) e;
	expand_enum(e) ;
	return string(e);
}

class enum z_mod_spread_type option (category: 'Yield Curve/003 Curve Building')
{ 
	ZMS_BP	option(str: "BasisPoint"),
	ZMS_P	option(str: "Price")
};

/*-----------------------------------------------------------------------
  Name :    disc_z_map
  Notes :         
  Changes:
  			Date    		Who     Detail
  			? 				MN   	Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
disc_z_model_type disc_z_map(string s) 
option (category: 'Yield Curve/003 Curve Building')
{
	disc_z_model_type e;
	str_to_enum(s,e);
	return e;
}

/*-----------------------------------------------------------------------
  Name :    disc_ip_type_map
  Notes :         
  Changes:
  			Date    		Who     Detail
  			? 				MN   	Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
disc_ip_type disc_ip_type_map(string s) 
option (category: 'Yield Curve/003 Curve Building')
{
	disc_ip_type e;
	str_to_enum(s,e);
	return e;
}
/*-----------------------------------------------------------------------
  Name :    disc_ip_type_map
  Notes :         
  Changes:
  			Date    		Who     Detail
  			? 				MN   	Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
disc_ip_type disc_ip_type_map(	disc_z_model_type 	disc_model_t, 
								string 				ip_str) 
option (category: 'Yield Curve/003 Curve Building')
{
	switch (disc_model_t) 
	{		
	case disc_z_model_type.DZ_ON_STEP:	
	case disc_z_model_type.DZ_MAX_SMOOTH:
	case disc_z_model_type.DZ_MAX_SMOOTH_SPREAD:		
	case disc_z_model_type.DZ_TANGGAARD:
	case disc_z_model_type.DZ_TENSION_W_EXACT: 
	case disc_z_model_type.DZ_TENSION_W_SPREAD:
	case disc_z_model_type.DZ_TENSION_Y_EXACT:
	case disc_z_model_type.DZ_TENSION_Y_SPREAD: 
	case disc_z_model_type.DZ_TENSION_Z_EXACT:
	case disc_z_model_type.DZ_TENSION_Z_SPREAD:
	case disc_z_model_type.DZ_NS:
		return null<disc_ip_type>;
	}

	return disc_ip_type_map(ip_str);
}
/*-----------------------------------------------------------------------
  Name :    disc_ip_map
  Notes :         
  Changes:
  			Date    		Who     Detail
  			? 				MN   	Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
interpolator disc_ip_map(disc_ip_type				bs_ip,
						logical option(nullable) 	mc_force_pos,
						logical option(nullable)	apply_hyman,						
						number option(nullable) 	left_der = null<number>,
						number option(nullable) 	hermite_right_der = null<number>) 
option (category: 'Yield Curve/003 Curve Building')
{
	logical  allow_x_pol = true;
	integer flag;

	if(bs_ip != disc_ip_type.DIPT_LINEAR && bs_ip != disc_ip_type.DIPT_STEP) {
		QL_REQUIRE(!null(apply_hyman),"invalid hyman flag (disc model)");
		if(apply_hyman )					
			flag = !allow_x_pol ? IP_HYMAN | IP_NO_XPOL_R :IP_HYMAN ;
		else
			flag = !allow_x_pol ? IP_NO_XPOL_R :0 ;
	}
	switch (bs_ip)
	{ 	
	case disc_ip_type.DIPT_LINEAR:
		return !allow_x_pol ? ip_linear(left_der, null<number>, IP_NO_XPOL_R) : ip_linear(left_der, null<number>) ;
	case disc_ip_type.DIPT_SPLINE_NAT:
		return  ip_spline(left_der, null<number>, flag) ;
	case disc_ip_type.DIPT_SPLINE_FIN:
		return  ip_spline(left_der, 0, flag) ;
	case disc_ip_type.DIPT_STEP:
		return !allow_x_pol ? ip_step(IP_NO_XPOL_R) : ip_step() ;
	case disc_ip_type.DIPT_HAGAN_WEST:
		QL_REQUIRE(!null(mc_force_pos), "invalid force positive flag (Hagan-West)");
		return  !allow_x_pol ? ip_hagan_west(mc_force_pos, IP_NO_XPOL_R) : ip_hagan_west(mc_force_pos);
	case disc_ip_type.DIPT_HERMITE_AKIMA:
		return  ip_hermite_akima(left_der, hermite_right_der, flag);
	case disc_ip_type.DIPT_HERMITE_CR:
		return  ip_hermite_catmull_rom(left_der, hermite_right_der, flag);
	case disc_ip_type.DIPT_HERMITE_FINDIFF:
		return  ip_hermite_finite_diff(left_der, hermite_right_der, flag );
	case disc_ip_type.DIPT_HERMITE_FB:
		return  ip_hermite_fritsch_butland(left_der, hermite_right_der, flag );
	case disc_ip_type.DIPT_HERMITE_KRUGER:
		return  ip_hermite_kruger(left_der, hermite_right_der, flag);
	case disc_ip_type.DIPT_HERMITE_MONO:
		return  ip_hermite_monotone(left_der, hermite_right_der, flag);
	case disc_ip_type.DIPT_HERMITE_P:
		return  ip_hermite_parabolic(left_der, hermite_right_der, flag );
	default:
		QL_FAIL("invalid disc model");
	}
}

/*-----------------------------------------------------------------------
  Name :    disc_extrap_map
  Notes :         
  Changes:
  			Date    		Who     Detail
  			? 				MN   	Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
disc_extrap_type disc_extrap_map(string s) 
option (category: 'Yield Curve/003 Curve Building')
{
	disc_extrap_type e;
	str_to_enum(s,e);
	return e;

}
/*-----------------------------------------------------------------------
  Name :    disc_rt_map
  Notes :  cont,simple,effective,semi-annual,quarterly,monthly,on,bankdisc,log_df       
  Changes:
  			Date    		Who     Detail
  			? 				MN   	Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
/**/
rate_type disc_rt_map(string s) 
option (category: 'Yield Curve/003 Curve Building')
{
	rate_type e;
	str_to_enum(s,e);
	return e;
}

vector(string) disc_rt_list()
option (category: 'Yield Curve/003 Curve Building')
{
	vector(rate_type) e;
	expand_enum(e) ;
	return string(e);
}
/*-----------------------------------------------------------------------
  Name :    disc_rt_map
  Notes :         
  Changes:
  			Date    		Who     Detail
  			? 				MN   	Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
rate_type disc_rt_map(	disc_z_model_type  				disc_model_t, 
						disc_ip_type option(nullable)	disc_ip_t,
						string 	option(nullable)		bs_rt_str) 
option (category: 'Yield Curve/003 Curve Building')
{
	if(null(disc_ip_t))
		return null<rate_type>;
	
	switch (disc_model_t) 
	{		
	case disc_z_model_type.DZ_ON_STEP:	
	case disc_z_model_type.DZ_MAX_SMOOTH:
	case disc_z_model_type.DZ_MAX_SMOOTH_SPREAD:		
	case disc_z_model_type.DZ_TANGGAARD:
	case disc_z_model_type.DZ_TENSION_W_EXACT: 
	case disc_z_model_type.DZ_TENSION_W_SPREAD:
	case disc_z_model_type.DZ_TENSION_Y_EXACT:
	case disc_z_model_type.DZ_TENSION_Y_SPREAD: 
	case disc_z_model_type.DZ_TENSION_Z_EXACT:
	case disc_z_model_type.DZ_TENSION_Z_SPREAD:
	case disc_z_model_type.DZ_NS:
		return null<rate_type>;
	}

	if(disc_ip_t == disc_ip_type.DIPT_HAGAN_WEST)
		return RT_CONT;

	QL_REQUIRE(!null(bs_rt_str), "bs_rt_str is a required input for this disc_model");
	
	return disc_rt_map(bs_rt_str);
}



module CORE_INT
{
	//-------------------------------------------------------
	// str2stdw
	//------------------------------------------------------- 
	std_weights str2stdw(string w)
	{		
		switch(w){
		case "equal": 
			return WS_EQUAL;
		case "pvbp":		
			return WS_PVBP; 
		case "macdur":		
			return WS_MACDUR; 
		case "modur":		
			return WS_MODDUR; 
		default:
			throw(E_INVALID_ARG, throw_err("ouch"));
		}

	}
	
	/*-----------------------------------------------------------------------
	  Name :    flatten_ois_curve
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			? 				Kalle   Original code (moved from fiws)				
	  ----------------------------------------------------------------------*/
	disc_func flatten_ois_curve(disc_func			df,
								date				d0,
								vector(date)		d,
								calendar option(nullable)	cal)
								
	{
		integer		n_d	= v_size(d);
		if(n_d == 0)
			return df;
		
		vector(date) dd = ..v_exclude (d, ..v_find(d, [null<date>]));
	    n_d	= v_size(dd);
		if(n_d == 0)
			return df;
	
	    vector(cashflows)	cf[0];
		logical cut = false;
		date cutoff = d0;
		
	    for (integer i = 0 ; i < n_d ; i++) {
			if(dd[i] > d0){
				if(!cut) {
					cut = true;
					cutoff = dd[i]-1;
				}
				dd[i] 		= cal.adjust_date (dd[i], BD_FOLLOWING);
				number df_i	= df.disc((dd[i] - d0) / 365.0);	
				push_back(cf, cashflows(d0, [ d0, dd[i] ], [ -1, 1 / df_i ]));
			}
	    }
	
		if(v_size(cf) == 0)
			return  df;
	
	    
	    //df_pre = bootstrap_on_ext(df, cutoff, cashflows(cf), ip_step(), cal);//???
		fwd_func ff_pre = bootstrap_on_ext(null<fwd_func>,df, d0, cashflows(cf), cal);
		disc_func df_pre = CORE_INT.on_fwd_to_df(d0,v_max(d),ff_pre, DC_ACT_365, cal,true);
		
	    disc_func dff = disc_func_splice(df_pre, df, (v_max(d) - d0) / 365.0);	
	
		return dff; 
	}
}

module COMPAT
{
	disc_func __bootstrap_on(	cashflows_base crv,
								calendar option(nullable) cal )
	{
		fwd_func ff = bootstrap_on(crv,  cal);//OBS! always returns act365-rates
		date d = crv.trade_date();
		date max = CORE_INT.max_curve_date(crv);
		disc_func df = CORE_INT.on_fwd_to_df(d,max,	ff, DC_ACT_365, cal,true);
		return df;
	}
}
/*-----------------------------------------------------------------------
  Name :    disc_z_model - abstract root
  Notes :           
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws)
			10-Aug-2016		MN		Added new tension _spr models and args for l-der / r-der (from ql_zc 1319 and 1320)
  ----------------------------------------------------------------------*/
class disc_z_model
{
	public:
	
	virtual 		disc_func df(curve, vector(number) option(nullable)) 					= 0;
	virtual 		disc_func df(cashflows_base, vector(number) option(nullable)) 			= 0;

	virtual 		disc_func df(disc_func, curve, vector(number) option(nullable)) 		= 0;
	virtual 		disc_func df(disc_func,cashflows_base, vector(number) option(nullable)) = 0;

	virtual 		disc_z_model_type type() 												= 0;
	logical 		check_cb_dates(vector(date) option(nullable));

	protected:

	disc_z_model(disc_extrap_type, number option(nullable));
	logical 			calc_pvbp(curve , out vector(number)  ,out string );
	logical 			calc_pvbp_cfb(cashflows_base ,  out vector(number)  , out string );
	logical 			calc_pvbp_cf(cashflows , out vector(number)  , out string );
	logical 			calc_pvbp_cfq(cashflows_q , out vector(number)  , out string );
	disc_func			extrapol_model(disc_func, number);
	virtual disc_func	extrapol_df(disc_func,number) 										= 0;

	disc_extrap_type 	xpol_type_;
	number 				xpol_cut_; //-1->no cutoff,  0 or null --> cutoff at curve max
};

logical disc_z_model.calc_pvbp_cfb(	cashflows_base 	crvb, 									
									out vector(number)   pvbp, 
									out string 	option(nullable)	err_msg)
{
	err_msg = "";	
	try {		
		cashflows crv = dynamic_cast<cashflows>(crvb);
		if(null(crv))
			throw(E_NO_DATA,"");
		logical ok = calc_pvbp_cf(crv, pvbp, err_msg);
		return ok;
	}
	catch {}
	
	try {
		cashflows_q crv = dynamic_cast<cashflows_q>(crvb);
		if(null(crv))
			throw(E_NO_DATA,"");
		logical ok = calc_pvbp_cfq(crv, pvbp, err_msg);
		return ok;
	}
	catch {
		err_msg = 'invalid pvbp vector (invalid cast)';
		return false;		
	}
	return false;	
}

logical disc_z_model.calc_pvbp_cf(	cashflows 	crv, 
									out vector(number)   pvbp, 
									out string	err_msg)
{
	err_msg 	= "";
	date trade	= crv.trade_date();
    integer n	= crv.n_instruments();
	
    if (n == 0) {
		err_msg = "invalid pvbp vector (no instruments)";
		return false;
	}

	resize(pvbp,n);

    integer		n_ok	= 0;

    for (integer i = 0 ; i < n ; i++) {
		vector(date)	dates;
		vector(number)	amounts;

		if (!crv.get_cashflows(i, dates, amounts))
	    	continue;

		++n_ok;

		//date	settle	= dates[0]; // We assume no flows before settle

		pvbp[i] = __w_calc.pvbp(trade, 0, amounts, dates)/10000;
	}
	if (n_ok == 0) {
		err_msg = "invalid pvbp vector (no instrument cashflows)";
		return false;
	}

    if (n_ok != n) {
		err_msg = strcat([ "invalid pvbp vector (cashflows inconsistency: could only calculate pvbp for ",
		       			string(n_ok), " of ", string(n), " instruments)" ]);
		return false;
	}

    return true;
}

logical disc_z_model.calc_pvbp_cfq(	cashflows_q 	crv, 
									out vector(number)   pvbp, 
									out string 	option(nullable)	err_msg)
{
	err_msg = "invalid pvbp vector (pvbp from cashflows_q (mtm swaps) not supported)";
	return false;

	err_msg 	= "";
	date trade	= crv.trade_date();
    integer n	= crv.n_instruments();
	
    if (n == 0) {
		err_msg = "invalid pvbp vector (no instruments)";
		return false;
	}

	resize(pvbp,n);

    integer		n_ok	= 0;

    for (integer i = 0 ; i < n ; i++) {
		vector(date)	dates, dates_q;
		vector(number)	amounts;

		if (!crv.get_cashflows(i, dates, dates_q, amounts))
	    	continue;

		++n_ok;

		//date	settle	= dates[0]; // We assume no flows before settle
		//pvbp calc from cashflows_q MUST BE CHECKED
		pvbp[i] = __w_calc.pvbp(trade, 0, amounts, dates)/10000;
	}

	if (n_ok == 0) {
		err_msg = "invalid pvbp vector (no instrument cashflows)";
		return false;
	}

    if (n_ok != n) {
		err_msg = strcat([ "invalid pvbp vector (cashflows inconsistency: could only calculate pvbp for ",
		       			string(n_ok), " of ", string(n), " instruments)" ]);
		return false;
	}

    return true;
}

logical disc_z_model.calc_pvbp(	curve 			crv, 
								out vector(number)   pvbp,
								out string  option(nullable)		err_msg)
{	
	err_msg 	= "";
    integer n	= crv.n_instruments();
	
    if (n == 0) {
		err_msg = "invalid pvbp vector (no instruments)";
		return false;
	}

	resize(pvbp,n);
    integer		n_ok	= 0;

    for (integer i = 0 ; i < n ; i++) {
		
		pvbp[i] = crv.instruments()[i].pvbp();		

		if(null(pvbp[i])) {
			vector(date)	dates;
			vector(number)	amounts;
	
			if (crv.get_cashflows(i, dates, amounts))		    								
				pvbp[i] = __w_calc.pvbp(crv.trade_date(), 0, amounts, dates)/10000;
		}

		if(!null(pvbp[i]))
			++n_ok;

	}
	if (n_ok == 0) {
		err_msg = "invalid pvbp vector (no instrument cashflows)";
		return false;
	}
    
	if (n_ok != n) {
		err_msg = strcat([ "invalid pvbp vector (cashflows inconsistency: could only calculate pvbp for ",
		       			string(n_ok), " of ", string(n), " instruments)" ]);
		return false;
	}

    return true;
}

disc_z_model.disc_z_model(	disc_extrap_type xpol_type, 
							number option(nullable) xpol_cut) 
							: xpol_type_(xpol_type), xpol_cut_(xpol_cut)
{
	if(null(xpol_cut_))
		xpol_cut_ = 0;
}

logical disc_z_model.check_cb_dates(vector(date) option(nullable) cb)
{
	if(null(cb) || v_size(cb) == 0 || v_size(v_find(cb, [null<date>])) == v_size(cb))
		return false;
	else
		return true;
}

disc_func disc_z_model.extrapol_model(	disc_func 	df, 
										number 		crv_max)
{
	if(xpol_cut_ == -1) {
		df =  extrapol_df(df, crv_max);
		return df;// //-1->no cutoff
	}		
	else if(xpol_cut_ == 0) // 0 --> cutoff at curve max
		return  extrap_none(df, crv_max ) ;
	else if(xpol_cut_ <= crv_max) // 0 or less than max --> cutoff at curve max
		return  extrap_none(df, xpol_cut_) ;
	else {//xpol_cut_ > crv_max		
		df =  extrapol_df(df, crv_max);
		return  extrap_none(df, xpol_cut_) ;
	}	
}

//-------------------------------------------------------
// class disc_z_model_ns
//-------------------------------------------------------
class disc_z_model_ns : public disc_z_model option (category: 'Yield Curve/003 Curve Building')
{
   	public:

	disc_z_model_ns(std_weights,vector(date) option(nullable), calendar option(nullable), 
					disc_extrap_type, number option(nullable));

	virtual disc_func 			df(curve, vector(number) option(nullable));
	virtual disc_func 			df(cashflows_base, vector(number) option(nullable) );
	virtual disc_func 			df(disc_func, curve, vector(number) option(nullable) );
	virtual disc_func 			df(disc_func, cashflows_base, vector(number) option(nullable) );

	virtual disc_z_model_type 	type();
	
	protected:

	virtual disc_func			extrapol_df(disc_func,number);

	vector(date) 		cb_dates_;
	calendar  			cal_;
	logical				use_cb_dates_;
	std_weights			w_;
};

disc_z_model_ns.disc_z_model_ns(std_weights w,
								vector(date) option(nullable) 	cb_dates,
								calendar option(nullable) 		cal,
								disc_extrap_type 				xpol_type, 
								number option(nullable) 		xpol_cut )
								:disc_z_model(xpol_type, xpol_cut),
								 cb_dates_(CORE_INT.clone_vector(cb_dates)), 
								 cal_(cal), w_(w)
{
	use_cb_dates_ = check_cb_dates(cb_dates);	
	
}

disc_z_model_type disc_z_model_ns.type() {return disc_z_model_type.DZ_NS;}

disc_func disc_z_model_ns.extrapol_df(	disc_func 	df, 
										number 		crv_max)
{
	switch (xpol_type_) 
	{
	case disc_extrap_type.DXP_MODEL:
		break;	
	case disc_extrap_type.DXP_MODEL_FLAT:
	case disc_extrap_type.DXP_ON_FLAT:
		df = extrap_flat_inst_fwd(df, crv_max );
		break;
	default:
		QL_FAIL("invalid disc extrapolation model [nelson_siegel]");
	}
	return df;
}

disc_func disc_z_model_ns.df(curve crv,vector(number) option(nullable) )
{
	QL_REQUIRE(check_curve(crv),"invalid curve [nelson_siegel]");
	//ns().bound_params()
	//ns().fix_param();
	//vector(string) pn = ns().parameter_names();
	disc_func df = fit(crv,ns(), w_, 2);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [nelson_siegel]');

	if(use_cb_dates_) {
		date d = crv.instruments()[0].trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [nelson_siegel]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [nelson_siegel]');
	return df;
}

disc_func disc_z_model_ns.df(cashflows_base crv, vector(number) option(nullable) )
{
	disc_func df = fit(crv,ns(), w_, 2);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [nelson_siegel]');

	if(use_cb_dates_) {
		date d = crv.trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [nelson_siegel]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [nelson_siegel]');
	return df;
}

disc_func disc_z_model_ns.df(disc_func dfb, curve crv, vector(number) option(nullable) )
{
	QL_REQUIRE(check_curve(crv),"invalid curve [nelson_siegel]");

	disc_func df = fit_spr(dfb, crv, ns(), w_, 2);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [nelson_siegel]');

	if(use_cb_dates_) {
		date d = crv.instruments()[0].trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [nelson_siegel]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [nelson_siegel]');
	return df;
}

disc_func disc_z_model_ns.df(disc_func dfb, cashflows_base crv, vector(number) option(nullable) )
{
	
	disc_func df = fit_spr(dfb, crv, ns(), w_, 2);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [nelson_siegel]');

	if(use_cb_dates_) {
		date d = crv.trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [nelson_siegel]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [nelson_siegel]');
	return df;
}

//-------------------------------------------------------
// disc_z_ns
//-------------------------------------------------------
disc_z_model_ns disc_z_ns(	std_weights w,
							vector(date) option(nullable) 	cb_dates,
							calendar option(nullable) 		cal, 									
							disc_extrap_type 				xpol_type, 
							number option(nullable) 		xpol_cut)
option (category: 'Yield Curve/003 Curve Building')
{				
	return new disc_z_model_ns(w, cb_dates,cal, xpol_type, xpol_cut);	
}

//-------------------------------------------------------
// class disc_z_model_max_smooth
//-------------------------------------------------------
class disc_z_model_max_smooth : public disc_z_model option (category: 'Yield Curve/003 Curve Building')
{
   	public:

	disc_z_model_max_smooth(disc_z_model_type, vector(date) option(nullable), calendar option(nullable), 
							disc_extrap_type, number option(nullable), smoothness_measure sm = SMOOTH_C2);
	disc_z_model_max_smooth(vector(date) option(nullable), calendar option(nullable), vector(number) option(nullable), z_mod_spread_type, 
							disc_extrap_type, number option(nullable), smoothness_measure sm = SMOOTH_C2); 

	virtual disc_func 			df(curve, vector(number) option(nullable));
	virtual disc_func 			df(cashflows_base, vector(number) option(nullable) );
	virtual disc_func 			df(disc_func, curve, vector(number) option(nullable) );
	virtual disc_func 			df(disc_func, cashflows_base, vector(number) option(nullable) );

	virtual disc_z_model_type 	type();
	void 						update_spreads(vector(number));
	vector(number)				calc_spreads(curve,vector(number) option(nullable) );
	vector(number)				calc_spreads(cashflows_base,vector(number) option(nullable));

	protected:

	virtual disc_func			extrapol_df(disc_func,number);

	disc_z_model_type	type_;
	vector(date) 		cb_dates_;
	calendar  			cal_;
	logical				use_cb_dates_;
	smoothness_measure 	smooth_;
	vector(number)		spreads_;
	z_mod_spread_type	spread_t_;
};

disc_z_model_type disc_z_model_max_smooth.type() {return type_;}

disc_z_model_max_smooth.disc_z_model_max_smooth(disc_z_model_type				type,
												vector(date) option(nullable) 	cb_dates, 
												calendar option(nullable) 		cal,
												disc_extrap_type 				xpol_type, 
												number option(nullable) 		xpol_cut,				 
												smoothness_measure 				smooth)
												:disc_z_model(xpol_type, xpol_cut), type_(type), 
												 cb_dates_(CORE_INT.clone_vector(cb_dates)),  
												 cal_(cal), smooth_(smooth), spread_t_(z_mod_spread_type.ZMS_BP)
{
	QL_REQUIRE(type_ == disc_z_model_type.DZ_MAX_SMOOTH_SPREAD || type_ == disc_z_model_type.DZ_MAX_SMOOTH,'invalid disc model type (should be a max_smooth model)');	
	use_cb_dates_ = check_cb_dates(cb_dates);	
	resize(spreads_,0);	
}

disc_z_model_max_smooth.disc_z_model_max_smooth(vector(date) option(nullable) 	cb_dates, 
												calendar option(nullable) 		cal,
												vector(number) option(nullable) spreads,
												z_mod_spread_type				spread_t,
												disc_extrap_type 				xpol_type, 
												number option(nullable) 		xpol_cut,			 
												smoothness_measure 				smooth)
												:disc_z_model(xpol_type, xpol_cut), type_(disc_z_model_type.DZ_MAX_SMOOTH_SPREAD), 
												 cb_dates_(CORE_INT.clone_vector(cb_dates)),  cal_(cal), 
												 smooth_(smooth), 
												 spreads_(CORE_INT.clone_vector(spreads)), 
												 spread_t_(spread_t)
{
	use_cb_dates_ = check_cb_dates(cb_dates);

	integer size = v_size(spreads_);
	for(integer i = 0; i < size; i++) 		
		QL_REQUIRE(!null(spreads_[i]),'invalid value in spreads vector [max_smooth]');	
}

vector(number) disc_z_model_max_smooth.calc_spreads(curve 							crv,
													vector(number) option(nullable) pvbp)
{	
	vector(number) res;

	if(v_size(spreads_) == 1) {
		if(spread_t_ == z_mod_spread_type.ZMS_BP) {
			
			if(null(pvbp)) {				
				string err_msg;						
				logical ok = calc_pvbp(crv, pvbp, err_msg);
				QL_REQUIRE(ok,strcat(err_msg," [max_smooth]"));
			}
			else {
				integer size = v_size(crv.instruments());
				QL_REQUIRE(size == v_size(pvbp) ,"invalid pvbp vector (size)[max_smooth]");				
			}
			res = pvbp*spreads_[0];
			//no div with 100 for disc_func because cashflows comes 
			//from instrument member and here nom=100
		}
		else {
			integer size = v_size(crv.instruments());
			resize(res, size);
			res = spreads_[0];				
		}
		return res;
	}
	else {
		integer size = v_size(crv.instruments());
		if(size  != v_size(spreads_))
			return res;
		if(spread_t_ == z_mod_spread_type.ZMS_BP) {
			if(null(pvbp)) {
				string err_msg;
				logical ok = calc_pvbp(crv, pvbp, err_msg);
				QL_REQUIRE(ok,strcat(err_msg," [max_smooth]"));
			}
			else
				QL_REQUIRE(size == v_size(pvbp) ,"invalid pvbp vector (size)[max_smooth]");

			resize(res, size);
			for(integer i=0;i<size;i++)
				res[i] = pvbp[i]*spreads_[i];
				//no div with 100 for disc_func because cashflows comes 
				//from instrument member and here nom=100
			return res;
		}
		else 
			return CORE_INT.clone_vector(spreads_);
	}
}


vector(number) disc_z_model_max_smooth.calc_spreads(cashflows_base 					crvb,
													vector(number) option(nullable) pvbp)
{	
	/*integer size = crv.n_instruments();
	QL_REQUIRE(v_size(spreads_) == size,'invalid size of spread vector (must match number of instruments) [max_smooth_spread, cf interface]');
	QL_REQUIRE(spread_t_ != ZMS_BP,'invalid spread type (pvbp weights not supported)[max_smooth_spread, cf interface]');			
	return clone_vector(spreads_);*/

	vector(number) res;

	if(v_size(spreads_) == 1) {
		if(spread_t_ == z_mod_spread_type.ZMS_BP) {
			if(null(pvbp)) {				
				string err_msg;						
				logical ok = calc_pvbp_cfb(crvb, pvbp, err_msg);
				QL_REQUIRE(ok,strcat(err_msg," [max_smooth]"));
			}
			else 
				QL_REQUIRE(crvb.n_instruments() == v_size(pvbp) ,"invalid pvbp vector (size)[max_smooth]");				
			
			res = pvbp*spreads_[0];
			//no div with 100 for disc_func because nom=100
		}
		else {
			integer size = crvb.n_instruments();
			resize(res, size);
			res = spreads_[0];				
		}
		return res;
	}
	else {
		integer size = crvb.n_instruments();
		if(size  != v_size(spreads_))
			return res;
		if(spread_t_ == z_mod_spread_type.ZMS_BP) {
			if(null(pvbp)) {
				string err_msg;
				logical ok = calc_pvbp_cfb(crvb,  pvbp, err_msg);
				QL_REQUIRE(ok,strcat(err_msg," [max_smooth]"));
			}
			else
				QL_REQUIRE(size == v_size(pvbp) ,"invalid pvbp vector (size)[max_smooth]");
	
			resize(res, size);
			for(integer i=0;i<size;i++)
				res[i] = pvbp[i]*spreads_[i];
				//no div with 100 for disc_func because nom=100
			return res;
		}
		else 
			return CORE_INT.clone_vector(spreads_);
	}
}

void disc_z_model_max_smooth.update_spreads(vector(number) spreads)
{
	spreads_= CORE_INT.clone_vector(spreads);	
}

disc_func disc_z_model_max_smooth.extrapol_df(	disc_func 	df, 
												number 		crv_max)
{
	switch (xpol_type_) 
	{
	case disc_extrap_type.DXP_MODEL:
		break;	
	case disc_extrap_type.DXP_MODEL_FLAT:	
	case disc_extrap_type.DXP_ON_FLAT:
		df = extrap_flat_inst_fwd(df, crv_max );
		break;
	default:
		QL_FAIL("invalid disc extrapolation model [max_smooth]");
	}
	return df;
}

disc_func disc_z_model_max_smooth.df(curve crv, vector(number) option(nullable) pvbp)
{
	QL_REQUIRE(check_curve(crv),"invalid curve [max_smooth]");

	disc_func df;
	if (type_ == disc_z_model_type.DZ_MAX_SMOOTH_SPREAD ) {
		vector(number) s = calc_spreads(crv, pvbp);
		QL_REQUIRE(!null(s) && v_size(s) > 0,'invalid spread vector (not initialized) [max_smooth_spread]');		
		QL_REQUIRE(v_size(s) == v_size(crv.instruments()),'invalid size of spread vector (must match number of instruments) [max_smooth_spread]');
		df = max_smooth_spreads(crv, s, smooth_) ;
	}	
	else
		df = max_smooth(crv, smooth_);
	
	QL_REQUIRE(!null(df),'failed to create basic disc_func [max_smooth]');

	if(use_cb_dates_) {
		date d = crv.instruments()[0].trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [max_smooth]');
	}
		
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));	
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [max_smooth]');

	return df;
}

disc_func disc_z_model_max_smooth.df(cashflows_base crv, vector(number) option(nullable) pvbp)
{
	//QL_REQUIRE(fiws.check_curve(crv),"invalid curve [max_smooth]");

	disc_func df;
	if (type_ == disc_z_model_type.DZ_MAX_SMOOTH_SPREAD ) {
		vector(number) s = calc_spreads(crv, pvbp);
		QL_REQUIRE(!null(s) && v_size(s) > 0,'invalid spread vector (not initialized) [max_smooth_spread]');		
		df = max_smooth_spreads(crv, s, smooth_) ;
	}	
	else
		df = max_smooth(crv, smooth_);
	
	QL_REQUIRE(!null(df),'failed to create basic disc_func [max_smooth]');

	if(use_cb_dates_) {
		date d = crv.trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [max_smooth]');
	}
		
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));	
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [max_smooth]');

	return df;
}

disc_func disc_z_model_max_smooth.df(disc_func df, curve crv, vector(number) option(nullable) pvbp)
{
	QL_FAIL('unsupported model for fitting spread [max_smooth]');
}

disc_func disc_z_model_max_smooth.df(disc_func df, cashflows_base crv, vector(number) option(nullable) pvbp)
{
	QL_FAIL('unsupported model for fitting spread [max_smooth]');
}
//-------------------------------------------------------
// disc_z_max_smooth
//-------------------------------------------------------
disc_z_model_max_smooth disc_z_max_smooth(	disc_z_model_type 				type,											
											vector(date) option(nullable) 	cb_dates, 
											calendar option(nullable) 		cal,
											disc_extrap_type 				xpol_type, 
											number option(nullable) 		xpol_cut, 
											smoothness_measure  option(nullable)  smooth = SMOOTH_C2)
option (category: 'Yield Curve/003 Curve Building')
{	
	if(null(smooth)) smooth = SMOOTH_C2;		
	return new disc_z_model_max_smooth(type, cb_dates,cal, xpol_type, xpol_cut, smooth);	
}	

//-------------------------------------------------------
// disc_z_max_smooth
//-------------------------------------------------------
disc_z_model_max_smooth disc_z_max_smooth(	vector(date) option(nullable) 	cb_dates, 
											calendar option(nullable) 		cal, 
											vector(number) option(nullable) spreads,
											z_mod_spread_type				spread_t,
											disc_extrap_type 				xpol_type, 
											number option(nullable) 		xpol_cut,
											smoothness_measure option(nullable) smooth = SMOOTH_C2)
option (category: 'Yield Curve/003 Curve Building')
{
	if(null(smooth)) smooth = SMOOTH_C2;				
	return new disc_z_model_max_smooth(cb_dates,cal,spreads, spread_t, xpol_type, xpol_cut, smooth);	
}

//-------------------------------------------------------
// class disc_z_model_boot
//-------------------------------------------------------
class disc_z_model_boot : public disc_z_model option (category: 'Yield Curve/003 Curve Building')
{
   	public:

	disc_z_model_boot(	vector(date) option(nullable), calendar option(nullable), disc_ip_type, logical option(nullable), logical option(nullable), 
						number option(nullable),number option(nullable),rate_type, disc_extrap_type, number option(nullable) ); 
	virtual disc_func df(curve,vector(number) option(nullable)  );
	virtual disc_func df(cashflows_base, vector(number) option(nullable)  );
	virtual disc_func df(disc_func,curve,vector(number) option(nullable) );
	virtual disc_func df(disc_func, cashflows_base, vector(number) option(nullable)  );
	virtual disc_z_model_type type();

	protected:
		
	virtual disc_func	extrapol_df(disc_func,number );

	vector(date) 		cb_dates_;
	calendar  			cal_;
	logical				use_cb_dates_;
	interpolator  		ip_;
	rate_type 			ip_rt_;
};
disc_z_model_type disc_z_model_boot.type() {return disc_z_model_type.DZ_BOOT;}

disc_z_model_boot .disc_z_model_boot (	vector(date) option(nullable) 	cb_dates, 
										calendar option(nullable) 		cal, 
										disc_ip_type  					ip,
										logical option(nullable) 		mc_force_pos,								
										logical option(nullable) 		apply_hyman,																				
										number option(nullable) 		left_der,
										number option(nullable) 		hermite_right_der, 
										rate_type						ip_rt,
										disc_extrap_type 				xpol_type, 
										number option(nullable) 		xpol_cut)
										:disc_z_model(xpol_type, xpol_cut), 
										 cb_dates_(CORE_INT.clone_vector(cb_dates)),  cal_(cal), ip_rt_(ip_rt)
{
	ip_ = disc_ip_map(ip,mc_force_pos,apply_hyman,left_der,hermite_right_der);
	QL_REQUIRE(!null(ip_), "failed to create the interpolator [boot disc model]");
	use_cb_dates_ = check_cb_dates(cb_dates);		
}

disc_func disc_z_model_boot.extrapol_df(	disc_func 	df, 
											number 		crv_max)
{
	switch (xpol_type_) 
	{
	case disc_extrap_type.DXP_MODEL:
		break;	
	case disc_extrap_type.DXP_MODEL_FLAT:
		df = extrap_flat_spot(df, crv_max, ip_rt_);		
		break;
	case disc_extrap_type.DXP_ON_FLAT:
		df = extrap_flat_inst_fwd(df, crv_max );
		break;
	default:
		QL_FAIL("invalid disc extrapolation model [boot]");
	}
	return df;
}

disc_func disc_z_model_boot.df(curve crv,vector(number) option(nullable) )
{
	QL_REQUIRE(check_curve(crv),"invalid curve [boot]");

	disc_func df = bootstrap(crv, ip_, ip_rt_);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [boot]');

	if(use_cb_dates_) {
		date d = crv.instruments()[0].trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [boot]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [boot]');
	return df;
}

disc_func disc_z_model_boot.df(cashflows_base crv, vector(number) option(nullable) )
{
	//QL_REQUIRE(fiws.check_curve(crv),"invalid curve [boot]");

	disc_func df = bootstrap(crv, ip_, ip_rt_);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [boot]');

	if(use_cb_dates_) {
		date d = crv.trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [boot]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [boot]');
	return df;
}

disc_func disc_z_model_boot.df(disc_func dfb, curve crv, vector(number) option(nullable) )
{
	QL_REQUIRE(check_curve(crv),"invalid curve [boot]");

	disc_func df = bootstrap_spr(dfb, crv, ip_, ip_rt_);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [boot]');

	if(use_cb_dates_) {
		date d = crv.instruments()[0].trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [boot]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [boot]');
	return df;
}

disc_func disc_z_model_boot.df(disc_func dfb, cashflows_base crv, vector(number) option(nullable) )
{
	//QL_REQUIRE(fiws.check_curve(crv),"invalid curve [boot]");

	disc_func df = bootstrap_spr(dfb, crv, ip_, ip_rt_);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [boot]');

	if(use_cb_dates_) {
		date d = crv.trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [boot]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [boot]');
	return df;
}
//-------------------------------------------------------
// disc_z_boot
//-------------------------------------------------------
disc_z_model_boot disc_z_boot(	vector(date) option(nullable) 	cb_dates, 
								calendar option(nullable) 		cal, 							
								disc_ip_type  					ip,
								logical option(nullable) 		mc_force_pos,								
								logical option(nullable) 		apply_hyman,																				
								number option(nullable) 		left_der,
								number option(nullable) 		hermite_right_der,  
								rate_type 						ip_rt,
								disc_extrap_type 				xpol_type, 
								number option(nullable) 		xpol_cut)
option (category: 'Yield Curve/003 Curve Building')
{	
	return new disc_z_model_boot (cb_dates,cal,ip,mc_force_pos,apply_hyman,left_der,hermite_right_der,ip_rt, xpol_type, xpol_cut);		
}


//-------------------------------------------------------
// class disc_z_model_boot_on
//-------------------------------------------------------
/*class disc_z_model_boot_on : public disc_z_model option (category: 'Yield Curve/003 Curve Building')
{
   	public:

	disc_z_model_boot_on(vector(date) option(nullable), calendar option(nullable), disc_ip_type,  logical option(nullable), 
						number option(nullable),number option(nullable),number option(nullable) ); //removed rate_type, disc_extrap_type
	virtual disc_func df(curve,vector(number) option(nullable)  );
	virtual disc_func df(cashflows_base, vector(number) option(nullable)  );
	virtual disc_func df(disc_func,curve,vector(number) option(nullable) );
	virtual disc_func df(disc_func, cashflows_base, vector(number) option(nullable)  );
	virtual disc_z_model_type type();

	protected:
		
	virtual disc_func	extrapol_df(disc_func,number );

	vector(date) 		cb_dates_;
	calendar  			cal_;
	logical				use_cb_dates_;
	interpolator  		ip_;
	//rate_type 			ip_rt_;
};
disc_z_model_type disc_z_model_boot_on.type() {return disc_z_model_type.DZ_BOOT_ON;}

disc_z_model_boot_on .disc_z_model_boot_on (vector(date) option(nullable) 	cb_dates, 
											calendar option(nullable) 		cal, 
											disc_ip_type  					ip,
											//logical option(nullable) 		mc_force_pos,								
											logical option(nullable) 		apply_hyman,																				
											number option(nullable) 		left_der,
											number option(nullable) 		hermite_right_der, 
											//rate_type						ip_rt,
											//disc_extrap_type 				xpol_type, 
											number option(nullable) 		xpol_cut)
											:disc_z_model(disc_extrap_type.DXP_ON_FLAT, xpol_cut), 
											 cb_dates_(CORE_INT.clone_vector(cb_dates)),  cal_(cal)//, ip_rt_(ip_rt)
{
	QL_REQUIRE(ip != disc_ip_type.DIPT_HAGAN_WEST, "Hagan-West model not applicable for bootstrap_on");
			
	ip_ = disc_ip_map(ip,false,apply_hyman,left_der,hermite_right_der);
	QL_REQUIRE(!null(ip_), "failed to create the interpolator [boot_on disc model]");
	use_cb_dates_ = check_cb_dates(cb_dates);		
}

disc_func disc_z_model_boot_on.extrapol_df(	disc_func 	df, 
											number 		crv_max)
{
	switch (xpol_type_) 
	{
	case disc_extrap_type.DXP_MODEL:
		QL_FAIL("unsupported function call for bootstrap_on");
		break;	
	case disc_extrap_type.DXP_MODEL_FLAT:
		QL_FAIL("unsupported function call for bootstrap_on");
		df = extrap_flat_spot(df, crv_max, rate_type.RT_ON);		
		break;
	case disc_extrap_type.DXP_ON_FLAT:
		df = extrap_flat_inst_fwd(df, crv_max );
		break;
	default:
		QL_FAIL("invalid disc extrapolation model [boot_on]");
	}
	return df;
}

disc_func disc_z_model_boot_on.df(curve crv,vector(number) option(nullable) pvbp)
{
	QL_REQUIRE(check_curve(crv),"invalid curve [boot_on]");

	disc_func df = COMPAT.__bootstrap_on(crv, cal_);
	//old: disc_func df = bootstrap_on(crv, ip_, cal_);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [boot_on]');

	if(use_cb_dates_) {
		date d = crv.instruments()[0].trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [boot_on]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [boot_on]');
	return df;
}

disc_func disc_z_model_boot_on.df(cashflows_base crv, vector(number) option(nullable) pvbp)
{
	//QL_REQUIRE(fiws.check_curve(crv),"invalid curve [boot]");
	disc_func df = COMPAT.__bootstrap_on(crv, cal_);
	
	QL_REQUIRE(!null(df),'failed to create basic disc_func [boot_on]');

	if(use_cb_dates_) {
		date d = crv.trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [boot_on]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [boot_on]');
	return df;
}

disc_func disc_z_model_boot_on.df(disc_func dfb, curve crv, vector(number) option(nullable) pvbp)
{
	QL_FAIL("unsupported function call for bootstrap_on");
	
	QL_REQUIRE(check_curve(crv),"invalid curve [boot_on]");

	disc_func df = bootstrap_spr(dfb, crv, ip_, rate_type.RT_ON);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [boot_on]');

	if(use_cb_dates_) {
		date d = crv.instruments()[0].trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [boot_on]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [boot_on]');
	return df;
}

disc_func disc_z_model_boot_on.df(disc_func dfb, cashflows_base crv, vector(number) option(nullable) pvbp)
{
	QL_FAIL("unsupported function call for bootstrap_on");
	
	//QL_REQUIRE(fiws.check_curve(crv),"invalid curve [boot]");

	disc_func df = bootstrap_spr(dfb, crv, ip_, rate_type.RT_ON);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [boot_on]');

	if(use_cb_dates_) {
		date d = crv.trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [boot_on]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [boot_on]');
	return df;
}
//-------------------------------------------------------
// disc_z_boot
//-------------------------------------------------------
disc_z_model_boot_on disc_z_boot_on(vector(date) option(nullable) 	cb_dates, 
									calendar option(nullable) 		cal, 							
									disc_ip_type  					ip,
									//logical option(nullable) 		mc_force_pos,								
									logical option(nullable) 		apply_hyman,																				
									number option(nullable) 		left_der,
									number option(nullable) 		hermite_right_der,  
									//rate_type 						ip_rt,
									//disc_extrap_type 				xpol_type, 
									number option(nullable) 		xpol_cut)
option (category: 'Yield Curve/003 Curve Building')
{	
	return new disc_z_model_boot_on(cb_dates,cal,ip,apply_hyman,left_der,hermite_right_der,xpol_cut);		
}
*/
//-------------------------------------------------------
// class disc_z_model_step
//-------------------------------------------------------
class disc_z_model_step: public disc_z_model option (category: 'Yield Curve/003 Curve Building')
{
   	public:

	disc_z_model_step(calendar option(nullable) ,disc_extrap_type, number option(nullable) ); 
	virtual disc_func df(curve,vector(number) option(nullable));
	virtual disc_func df(cashflows_base, vector(number) option(nullable) );
	virtual disc_func df(disc_func,curve,vector(number) option(nullable) );
	virtual disc_func df(disc_func, cashflows_base, vector(number) option(nullable));
	virtual disc_z_model_type type();

	protected:
	
	virtual disc_func	extrapol_df(disc_func,number );

	calendar  cal_;
};

disc_z_model_type disc_z_model_step.type() {return disc_z_model_type.DZ_ON_STEP;}

disc_z_model_step.disc_z_model_step(calendar option(nullable) 	cal,
									disc_extrap_type 			xpol_type, 
									number option(nullable) 	xpol_cut)
									:disc_z_model(xpol_type, xpol_cut), cal_(cal)
{}

disc_func disc_z_model_step.extrapol_df(	disc_func 	df, 
											number 		crv_max)
{
	switch (xpol_type_) 
	{
	case disc_extrap_type.DXP_MODEL:
	case disc_extrap_type.DXP_MODEL_FLAT:
		break;		
	case disc_extrap_type.DXP_ON_FLAT:
		df = extrap_flat_inst_fwd(df, crv_max );
		break;
	default:
		QL_FAIL("invalid disc extrapolation model [on_step]");
	}
	return df;
}


disc_func disc_z_model_step.df(curve crv, vector(number) option(nullable) )
{
	//if(use_cb_dates_)
	//	fiws.QL_WARNING();
	
	QL_REQUIRE(check_curve(crv),"invalid curve [on_step]");
	disc_func df = COMPAT.__bootstrap_on(crv, cal_);
	//old: disc_func df = bootstrap_on(crv, ip_step(),cal_);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [on_step]');
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [on_step]');
	return df;
}
disc_func disc_z_model_step.df(cashflows_base crv, vector(number) option(nullable) )
{
	//if(use_cb_dates_)
	//	fiws.QL_WARNING();
	
	//QL_REQUIRE(fiws.check_curve(crv),"invalid curve [on_step]");
	disc_func df = COMPAT.__bootstrap_on(crv, cal_);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [on_step]');
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [on_step]');
	return df;
}

disc_func disc_z_model_step.df(disc_func df, curve crv, vector(number) option(nullable) )
{
	QL_FAIL('unsupported model for fitting spread [on_step]');
}
disc_func disc_z_model_step.df(disc_func df, cashflows_base crv, vector(number) option(nullable) )
{
	QL_FAIL('unsupported model for fitting spread [on_step]');
}
//-------------------------------------------------------
// disc_z_step
//-------------------------------------------------------
disc_z_model_step disc_z_step(	calendar option(nullable) 	cal,
								disc_extrap_type 			xpol_type, 
								number option(nullable) 	xpol_cut)
option (category: 'Yield Curve/003 Curve Building')
{	
	return new disc_z_model_step(cal,xpol_type,xpol_cut);	
}

//-------------------------------------------------------
// class disc_z_model_step_boot
//-------------------------------------------------------
class disc_z_model_step_boot: public disc_z_model option (category: 'Yield Curve/003 Curve Building')
{
   	public:

	disc_z_model_step_boot(	calendar option(nullable), disc_ip_type, logical option(nullable),logical option(nullable), number option(nullable),number option(nullable),
							rate_type, integer ,disc_extrap_type, number option(nullable) ); 
	virtual disc_func df(curve, vector(number) option(nullable)  );
	virtual disc_func df(cashflows_base, vector(number) option(nullable));
	virtual disc_func df(disc_func,curve, vector(number) option(nullable) );
	virtual disc_func df(disc_func, cashflows_base, vector(number) option(nullable) );
	virtual disc_z_model_type type();

	protected:

	virtual disc_func	extrapol_df(disc_func,number );

	calendar  			cal_;
	interpolator  		ip_;
	rate_type 			ip_rt_;
	integer 			step_end_;
	
	number 				step_end_t_;//calculated
};

disc_z_model_type disc_z_model_step_boot.type() {return disc_z_model_type.DZ_ON_STEP_BOOT;}

disc_z_model_step_boot.disc_z_model_step_boot(	calendar option(nullable) 	cal, 
												disc_ip_type  				ip,
												logical option(nullable) 	mc_force_pos,	
												logical option(nullable) 	apply_hyman,
												number option(nullable) 	left_der,
												number option(nullable) 	hermite_right_der,
												rate_type					ip_rt,
												integer 					step_end,
												disc_extrap_type 			xpol_type, 
												number option(nullable) 	xpol_cut)
												:disc_z_model(xpol_type, xpol_cut), cal_(cal), ip_rt_(ip_rt), step_end_(step_end)
{
	ip_ = disc_ip_map(ip,mc_force_pos,apply_hyman,left_der,hermite_right_der);
	QL_REQUIRE(!null(ip_), "failed to create the interpolator [on_step_boot disc model]");
}

disc_func disc_z_model_step_boot.extrapol_df(	disc_func 	df, 
												number 		crv_max)
{
	switch (xpol_type_) 
	{
	case disc_extrap_type.DXP_MODEL:
		break;
	case disc_extrap_type.DXP_MODEL_FLAT:			
		if(step_end_t_ < crv_max)			
			df = extrap_flat_spot(df, crv_max, ip_rt_);	
		break;			
	case disc_extrap_type.DXP_ON_FLAT:
		df = extrap_flat_inst_fwd(df, crv_max );
		break;
	default:
		QL_FAIL("invalid disc extrapolation model [on_step_boot]");
	}
	return df;
}

disc_func disc_z_model_step_boot.df(curve crv, vector(number) option(nullable) )
{
	QL_REQUIRE(check_curve(crv),"invalid curve [on_step_boot]");
	disc_func df = COMPAT.__bootstrap_on(crv, cal_);
	//old: disc_func df = bootstrap_on(crv, ip_step(), cal_);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [on_step_boot]');

	date d 				= crv.instruments()[0].trade_date();	
	date step_end_date 	= d + step_end_;
	step_end_t_ 		= (step_end_date-d)/365;

	if(step_end_date < CORE_INT.max_curve_date(crv)) {		
		curve cut_curve = crv.chop_short(step_end_date);			
		df = bootstrap_ext(df, step_end_date, cut_curve, ip_, ip_rt_, false);//sista argumentet ????????????????????
		QL_REQUIRE(!null(df),'failed to create long end disc_func [on_step_boot]');
	}
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [on_step_boot]');
	return df;
}

disc_func disc_z_model_step_boot.df(cashflows_base crv, vector(number) option(nullable) )
{
	//QL_REQUIRE(fiws.check_curve(crv),"invalid curve [on_step_boot]");
	disc_func df 		= COMPAT.__bootstrap_on(crv, cal_);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [on_step_boot]');

	date d 				= crv.trade_date();	
	date step_end_date 	= d + step_end_;
	step_end_t_ 		= (step_end_date-d)/365;

	if(step_end_date < CORE_INT.max_curve_date(crv)) {
		QL_FAIL('support for creating separate short and long end disc_func not available [on_step_boot, cf interface]');		
		//curve cut_curve = crv.chop_short(step_end_date);			
		//df = bootstrap_ext(df, step_end_date, cut_curve, ip_, ip_rt_, false);//sista argumentet ????????????????????
		//QL_REQUIRE(!null(df),'failed to create long end disc_func [on_step_boot]');
	}
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [on_step_boot]');
	return df;
}

disc_func disc_z_model_step_boot.df(disc_func df, curve crv, vector(number) option(nullable) )
{
	QL_FAIL('unsupported model for fitting spread [on_step_boot]');
}
disc_func disc_z_model_step_boot.df(disc_func df, cashflows_base crv, vector(number) option(nullable) )
{
	QL_FAIL('unsupported model for fitting spread [on_step_boot]');
}
//-------------------------------------------------------
//
//-------------------------------------------------------
disc_z_model_step_boot disc_z_step_boot(calendar option(nullable) 	cal, 
										disc_ip_type  				ip,
										logical option(nullable) 	mc_force_pos,	
										logical option(nullable) 	apply_hyman,
										number option(nullable) 	left_der,
										number option(nullable) 	hermite_right_der,
										rate_type 					ip_rt,
										integer 					step_end,
										disc_extrap_type 			xpol_type, 
										number option(nullable) 	xpol_cut)
option (category: 'Yield Curve/003 Curve Building')
{	
	return new disc_z_model_step_boot(cal,ip,mc_force_pos, apply_hyman, left_der, hermite_right_der, ip_rt,step_end,xpol_type,xpol_cut);	
}

//-------------------------------------------------------
// class disc_z_model_tanggaard
//-------------------------------------------------------
class disc_z_model_tanggaard : public disc_z_model option (category: 'Yield Curve/003 Curve Building')
{
   	public:

	disc_z_model_tanggaard(	vector(date) option(nullable), calendar option(nullable), disc_extrap_type, number option(nullable),
							number, std_weights w = WS_PVBP, smoothness_measure sm = SMOOTH_C2); 
	virtual disc_func 			df(curve, vector(number) option(nullable)  );
	virtual disc_func 			df(cashflows_base, vector(number) option(nullable) );
	virtual disc_func			df(disc_func,curve, vector(number) option(nullable) );
	virtual disc_func 			df(disc_func, cashflows_base, vector(number) option(nullable) );
	virtual disc_z_model_type 	type();	

	protected:
	
	virtual disc_func	extrapol_df(disc_func,number );

	disc_z_model_type	type_;
	vector(date) 		cb_dates_;
	calendar  			cal_;
	logical				use_cb_dates_;
	number				w_factor_;
	std_weights			w_;
	smoothness_measure 	smooth_;
	
};
disc_z_model_type disc_z_model_tanggaard.type() {return disc_z_model_type.DZ_TANGGAARD;}

disc_z_model_tanggaard.disc_z_model_tanggaard(	vector(date) option(nullable) 	cb_dates, 
												calendar option(nullable) 		cal,
												disc_extrap_type 				xpol_type, 
												number option(nullable) 		xpol_cut,
												number							w_factor, 
												std_weights						w, 
												smoothness_measure 				smooth)
												:disc_z_model(xpol_type, xpol_cut), 
												 cb_dates_(CORE_INT.clone_vector(cb_dates)),  
												 cal_(cal), w_factor_(w_factor), w_(w), smooth_(smooth)
{
	use_cb_dates_ = check_cb_dates(cb_dates);	
}

disc_func disc_z_model_tanggaard.extrapol_df(	disc_func 	df, 
												number 		crv_max)
{
	switch (xpol_type_) 
	{
	case disc_extrap_type.DXP_MODEL:
		break;
	case disc_extrap_type.DXP_MODEL_FLAT:							
	case disc_extrap_type.DXP_ON_FLAT:
		df = extrap_flat_inst_fwd(df, crv_max );
		break;
	default:
		QL_FAIL("invalid disc extrapolation model [tanggaard]");
	}
	return df;
}

disc_func disc_z_model_tanggaard.df(curve crv, vector(number) option(nullable) )
{
	QL_REQUIRE(check_curve(crv),"invalid curve [tanggaard]");
	disc_func df = tanggaard(crv, smooth_, w_factor_, w_);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [tanggaard]');

	if(use_cb_dates_){
		date d = crv.instruments()[0].trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [tanggaard]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [tanggaard]');
	return df;
}
disc_func disc_z_model_tanggaard.df(cashflows_base crv, vector(number) option(nullable) )
{
	//QL_REQUIRE(fiws.check_curve(crv),"invalid curve [tanggaard]");
	disc_func df = tanggaard(crv, smooth_, w_factor_, w_);
	QL_REQUIRE(!null(df),'failed to create basic disc_func [tanggaard]');

	if(use_cb_dates_){
		date d = crv.trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [tanggaard]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [tanggaard]');
	return df;
}

disc_func disc_z_model_tanggaard.df(disc_func df, curve crv, vector(number) option(nullable) )
{
	QL_FAIL('unsupported model for fitting spread [tanggaard]');
}

disc_func disc_z_model_tanggaard.df(disc_func df, cashflows_base crv, vector(number) option(nullable) )
{
	QL_FAIL('unsupported model for fitting spread [tanggaard]');
}
//-------------------------------------------------------
// disc_z_tanggaard
//-------------------------------------------------------
disc_z_model_tanggaard disc_z_tanggaard (	vector(date) option(nullable) 	cb_dates, 
											calendar option(nullable) 		cal,
											disc_extrap_type 				xpol_type, 
											number option(nullable) 		xpol_cut,
											number							w_factor, 
											std_weights	option(nullable) w = WS_PVBP,  
											smoothness_measure option(nullable)	smooth = SMOOTH_C2)
option (category: 'Yield Curve/003 Curve Building')
{
	if(null(smooth)) smooth = SMOOTH_C2;
	if(null(w)) w = WS_PVBP;			
	return new disc_z_model_tanggaard ( cb_dates,cal,xpol_type,xpol_cut,w_factor, w, smooth);	
}

//-------------------------------------------------------
// class disc_z_model_tension
//-------------------------------------------------------
class disc_z_model_tension : public disc_z_model option (category: 'Yield Curve/003 Curve Building')
{
   	public:

	disc_z_model_tension(	disc_z_model_type,  vector(date) option(nullable), calendar option(nullable), interpolation, number,disc_extrap_type, 
							number option(nullable),number option(nullable),number option(nullable));
	disc_z_model_tension(	vector(date) option(nullable), calendar option(nullable), vector(number) option(nullable) , z_mod_spread_type,interpolation, number,
							disc_extrap_type, number option(nullable),number option(nullable),number option(nullable)); 
	disc_z_model_tension(	disc_z_model_type,  vector(date) option(nullable), calendar option(nullable), interpolation,disc_extrap_type, 
							number option(nullable),number option(nullable),number option(nullable));
	disc_z_model_tension(	disc_z_model_type,  vector(date) option(nullable), calendar option(nullable), vector(number) option(nullable) , z_mod_spread_type, 
							interpolation, disc_extrap_type, number option(nullable),number option(nullable),number option(nullable)); 	
	virtual disc_func 			df(curve, vector(number) option(nullable) );
	virtual disc_func 			df(cashflows_base, vector(number) option(nullable) );
	virtual disc_func 			df(disc_func,curve, vector(number) option(nullable) );
	virtual disc_func 			df(disc_func, cashflows_base, vector(number) option(nullable) );
	virtual disc_z_model_type 	type();
	void 						update_spreads(vector(number));
	vector(number)				calc_spreads(curve, vector(number) option(nullable) );
	vector(number)				calc_spreads(cashflows_base , vector(number) option(nullable) );

	protected:

	virtual disc_func	extrapol_df(disc_func,number );

	disc_z_model_type	type_;
	vector(date) 		cb_dates_;
	calendar  			cal_;
	logical				use_cb_dates_;
	interpolation		sigma_;
	number 				epsilon_;
	vector(number)		spreads_;
	z_mod_spread_type	spread_t_;
	logical				is_spread_;
	number 				l_der_; 
	number 				r_der_; 
};
disc_z_model_type disc_z_model_tension.type() {return type_;}

disc_z_model_tension.disc_z_model_tension(	disc_z_model_type				type,
											vector(date) option(nullable) 	cb_dates, 
											calendar option(nullable) 		cal,
											interpolation					sigma,				 
											number 							epsilon,
											disc_extrap_type 				xpol_type, 
											number option(nullable) 		xpol_cut,
											number option(nullable)			l_der, 
											number  option(nullable)		r_der)
											:disc_z_model(xpol_type, xpol_cut), type_(type), 
											 cb_dates_(CORE_INT.clone_vector(cb_dates)),  cal_(cal), 
											 sigma_(sigma), epsilon_(epsilon), spread_t_(z_mod_spread_type.ZMS_BP), is_spread_(false),l_der_(l_der),
											 r_der_(r_der)
{	 
	QL_REQUIRE(type_ == disc_z_model_type.DZ_TENSION_W_EXACT || type_ == disc_z_model_type.DZ_TENSION_W_SPREAD,'invalid disc model type (should be a tension_w model)');	
	use_cb_dates_ = check_cb_dates(cb_dates);	
	resize(spreads_,0);	
}

disc_z_model_tension.disc_z_model_tension(	vector(date) option(nullable) 	cb_dates, 
											calendar option(nullable) 		cal,
											vector(number) option(nullable) spreads,
											z_mod_spread_type				spread_t,				 
											interpolation					sigma,				 
											number 							epsilon,
											disc_extrap_type 				xpol_type, 
											number option(nullable) 		xpol_cut,
											number option(nullable)			l_der, 
											number  option(nullable)		r_der)
											:disc_z_model(xpol_type, xpol_cut), type_(disc_z_model_type.DZ_TENSION_W_SPREAD), 
											 cb_dates_(CORE_INT.clone_vector(cb_dates)), 
											 cal_(cal), 
											 sigma_(sigma), epsilon_(epsilon), 
											 spreads_(CORE_INT.clone_vector(spreads)),
											 spread_t_(spread_t),is_spread_(true),l_der_(l_der),
											 r_der_(r_der)
{		
	use_cb_dates_ = check_cb_dates(cb_dates);

	integer size = v_size(spreads_);
	for(integer i = 0; i < size; i++) 
		QL_REQUIRE(!null(spreads_[i]),'invalid value in spreads vector [tension]');	
}

disc_z_model_tension.disc_z_model_tension(	disc_z_model_type				type,
											vector(date) option(nullable) 	cb_dates, 
											calendar option(nullable) 		cal,
											interpolation					sigma,
											disc_extrap_type 				xpol_type, 
											number option(nullable) 		xpol_cut,
											number option(nullable)			l_der, 
											number  option(nullable)		r_der)
											:disc_z_model(xpol_type, xpol_cut), type_(type), 
											 cb_dates_(CORE_INT.clone_vector(cb_dates)),  cal_(cal), 
											 sigma_(sigma), epsilon_(null<number>), spread_t_(z_mod_spread_type.ZMS_BP),is_spread_(false),l_der_(l_der),
											 r_der_(r_der)
{
	switch (type_)
	{
	case disc_z_model_type.DZ_TENSION_Y_EXACT:
	case disc_z_model_type.DZ_TENSION_Y_SPREAD:
	case disc_z_model_type.DZ_TENSION_Z_EXACT:
	case disc_z_model_type.DZ_TENSION_Z_SPREAD:
		break;
	default: 
		QL_FAIL('invalid disc model type (should be a tension_y or tension_z model) [tension]');
	}
	use_cb_dates_ = check_cb_dates(cb_dates);
	
	resize(spreads_,0);	
}

disc_z_model_tension.disc_z_model_tension(	disc_z_model_type				type,
											vector(date) option(nullable) 	cb_dates, 
											calendar option(nullable) 		cal,
											vector(number) option(nullable) spreads,
											z_mod_spread_type				spread_t,				 
											interpolation					sigma,
											disc_extrap_type 				xpol_type, 
											number option(nullable) 		xpol_cut,
											number option(nullable)			l_der, 
											number  option(nullable)		r_der)
											:disc_z_model(xpol_type, xpol_cut), type_(type), 
											  cb_dates_(CORE_INT.clone_vector(cb_dates)),  cal_(cal), 
											  sigma_(sigma), epsilon_(null<number>), 
											  spreads_(CORE_INT.clone_vector(spreads)),
											  spread_t_(spread_t),is_spread_(true),l_der_(l_der), r_der_(r_der)
{	
	QL_REQUIRE(type_ == disc_z_model_type.DZ_TENSION_Y_SPREAD || type_ == disc_z_model_type.DZ_TENSION_Z_SPREAD,'invalid disc model type (should be a tension_y_spread or tension_z_spread model)');		
	use_cb_dates_ = check_cb_dates(cb_dates);

	integer size = v_size(spreads_);
	for(integer i = 0; i < size; i++) 
		QL_REQUIRE(!null(spreads_[i]),'invalid value in spreads vector [tension_spread]');
	
}

vector(number) disc_z_model_tension.calc_spreads(curve crv, vector(number) option(nullable) pvbp)
{	
	vector(number) res;

	if(v_size(spreads_) == 1) {
		if(spread_t_ == z_mod_spread_type.ZMS_BP) {			
			if(null(pvbp)) {
				string err_msg;
				logical ok = calc_pvbp(crv, pvbp,err_msg);
				QL_REQUIRE(ok,strcat(err_msg," [tension_spread]"));
			}
			else {
				integer size = v_size(crv.instruments());
				QL_REQUIRE(size == v_size(pvbp) ,"invalid pvbp vector (size)[tension_spread]");				
			}
			res = pvbp*spreads_[0];
			//no div with 100 for disc_func because cashflows comes 
			//from instrument member and here nom=100
		}
		else {
			integer size = v_size(crv.instruments());
			resize(res, size);
			res = spreads_[0];				
		}
		return res;
	}
	else {
		integer size = v_size(crv.instruments());
		if(size  != v_size(spreads_))
			return res;
		if(spread_t_ == z_mod_spread_type.ZMS_BP) {			
			if(null(pvbp)) {
				string err_msg;
				logical ok = calc_pvbp(crv, pvbp,err_msg);
				QL_REQUIRE(ok,strcat(err_msg," [tension_spread]"));
			}
			else
				QL_REQUIRE(size == v_size(pvbp) ,"invalid pvbp vector (size)[tension_spread]");
			resize(res, size);
			for(integer i=0;i<size;i++)
				res[i] = pvbp[i]*spreads_[i];
				//no div with 100 for disc_func because cashflows comes 
				//from instrument member and here nom=100
			return res;
		}
		else 
			return clone_vector(spreads_);
	}
}

vector(number) disc_z_model_tension.calc_spreads(cashflows_base crvb, vector(number) option(nullable) pvbp)
{	
	/*integer size = crv.n_instruments();
	QL_REQUIRE(v_size(spreads_) == size,'invalid size of spread vector (must match number of instruments) [tension_spread, cf interface]');
	QL_REQUIRE(spread_t_ != ZMS_BP,'invalid spread type (pvbp weights not supported)[tension_spread, cf interface]');			
	return clone_vector(spreads_);*/
	vector(number) res;

	if(v_size(spreads_) == 1) {
		if(spread_t_ == z_mod_spread_type.ZMS_BP) {
			if(null(pvbp)) {
				string err_msg;
				logical ok = calc_pvbp_cfb(crvb, pvbp, err_msg);
				QL_REQUIRE(ok,strcat(err_msg," [tension_spread]"));
			}
			else 
				QL_REQUIRE(crvb.n_instruments() == v_size(pvbp) ,"invalid pvbp vector (size)[tension_spread]");

			res = pvbp*spreads_[0];
			//no div with 100 for disc_func because nom=100
		}
		else {
			integer size = crvb.n_instruments();
			resize(res, size);
			res = spreads_[0];				
		}
		return res;
	}
	else {
		integer size = crvb.n_instruments();
		if(size  != v_size(spreads_))
			return res;
		if(spread_t_ == z_mod_spread_type.ZMS_BP) {
			if(null(pvbp)) {
				string err_msg;
				logical ok = calc_pvbp_cfb(crvb, pvbp, err_msg);
				QL_REQUIRE(ok,strcat(err_msg," [tension_spread]"));
			}
			else
				QL_REQUIRE(size == v_size(pvbp) ,"invalid pvbp vector (size)[tension_spread]");
			resize(res, size);
			for(integer i=0;i<size;i++)
				res[i] = pvbp[i]*spreads_[i];
				//no div with 100 for disc_func because nom=100
			return res;
		}
		else 
			return CORE_INT.clone_vector(spreads_);
	}
}

void disc_z_model_tension.update_spreads(vector(number) spreads)
{	
	QL_REQUIRE(v_size(spreads) > 0,'invalid spread vector [tension_spread]');
	spreads_= CORE_INT.clone_vector(spreads);
	is_spread_ = true;	
}

disc_func disc_z_model_tension.extrapol_df(	disc_func 	df, 
											number 		crv_max)
{
	if(xpol_type_ == disc_extrap_type.DXP_MODEL)
		return df;

	switch(type_)
	{
	case disc_z_model_type.DZ_TENSION_Y_SPREAD:		
	case disc_z_model_type.DZ_TENSION_Y_EXACT:
	{
		switch (xpol_type_) 
		{
		case disc_extrap_type.DXP_MODEL_FLAT:
			df = extrap_flat_spot(df, crv_max, RT_CONT);			
			break;							
		case disc_extrap_type.DXP_ON_FLAT:
			df = extrap_flat_inst_fwd(df, crv_max );			
			break;
		default:
			QL_FAIL("invalid disc extrapolation model [tension_y]");
		} 
		break;
	}
	case disc_z_model_type.DZ_TENSION_Z_SPREAD:		
	case disc_z_model_type.DZ_TENSION_Z_EXACT:
	case disc_z_model_type.DZ_TENSION_W_SPREAD://FIX_ME		
	case disc_z_model_type.DZ_TENSION_W_EXACT://FIX_ME
	{
		switch (xpol_type_) 
		{
		case disc_extrap_type.DXP_MODEL_FLAT:							
		case disc_extrap_type.DXP_ON_FLAT:
			df = extrap_flat_inst_fwd(df, crv_max );			
			break;
		default:
			QL_FAIL("invalid disc extrapolation model [tension]");
		} 
		break;
	}						
	default:
		QL_FAIL('invalid disc model type [tension]');
	}

	
	return df;
}

disc_func disc_z_model_tension.df(curve crv, vector(number) option(nullable) pvbp)
{
	QL_REQUIRE(check_curve(crv),"invalid curve [tension]");

	disc_func df;		

	if(is_spread_) {
		vector(number) s = calc_spreads(crv, pvbp);
		QL_REQUIRE(!null(s) && v_size(s) > 0,'invalid spread vector (not initialized) [tension_spread]');
		QL_REQUIRE(v_size(s) == v_size(crv.instruments()),'invalid size of spread vector (must match number of instruments) [tension]');

		switch(type_)
		{
		case disc_z_model_type.DZ_TENSION_W_SPREAD:			
			df = tension_spline_w_spreads(crv, s, sigma_, epsilon_, l_der_);
			break;
		case disc_z_model_type.DZ_TENSION_Y_SPREAD:
			df = tension_spline_y_spreads(crv, s, sigma_, l_der_, r_der_);
			break; 
		case disc_z_model_type.DZ_TENSION_Z_SPREAD:
			df = tension_spline_z_spreads(crv, s, sigma_);
			break;			
		default:
			QL_FAIL('invalid disc model type [tension_spread]');
		}

	}
	else {
		switch(type_)
		{
		case disc_z_model_type.DZ_TENSION_W_EXACT:
			df = tension_spline_w(crv, sigma_, epsilon_, l_der_);
			break;
		case disc_z_model_type.DZ_TENSION_Y_EXACT:
			df = tension_spline_y(crv, sigma_, l_der_, r_der_);
			break; 
		case disc_z_model_type.DZ_TENSION_Z_EXACT:
			df = tension_spline_z(crv, sigma_);
			break;			
		default:
			QL_FAIL('invalid disc model type [tension]');
		}
	}

	QL_REQUIRE(!null(df),'failed to create basic disc_func [tension]');

	if(use_cb_dates_) {
		date d = crv.instruments()[0].trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [tension]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [tension]');

	return df;
}


disc_func disc_z_model_tension.df(cashflows_base crv, vector(number) option(nullable) pvbp)
{
	//QL_REQUIRE(fiws.check_curve(crv),"invalid curve [tension]");

	disc_func df;

	if(is_spread_) {
		vector(number) s = calc_spreads(crv, pvbp);
		QL_REQUIRE(!null(s) && v_size(s) > 0,'invalid spread vector (not initialized) [tension_spread]');
		
		switch(type_)
		{
		case disc_z_model_type.DZ_TENSION_W_SPREAD:
			df = tension_spline_w_spreads(crv, s, sigma_, epsilon_, l_der_);
			break;
		case disc_z_model_type.DZ_TENSION_Y_SPREAD:
			df = tension_spline_y_spreads(crv, s, sigma_, l_der_, r_der_);
			break; 
		case disc_z_model_type.DZ_TENSION_Z_SPREAD:
			df = tension_spline_z_spreads(crv, s, sigma_);
			break;			
		default:
			QL_FAIL('invalid disc model type [tension_spread]');
		}

	}
	else {
		switch(type_)
		{
		case disc_z_model_type.DZ_TENSION_W_EXACT:
			df = tension_spline_w(crv, sigma_, epsilon_, l_der_);
			break;
		case disc_z_model_type.DZ_TENSION_Y_EXACT:
			df = tension_spline_y(crv, sigma_, l_der_, r_der_);
			break; 
		case disc_z_model_type.DZ_TENSION_Z_EXACT:
			df = tension_spline_z(crv, sigma_);
			break;			
		default:
			QL_FAIL('invalid disc model type [tension]');
		}
	}

	QL_REQUIRE(!null(df),'failed to create basic disc_func [tension]');

	if(use_cb_dates_) {		
		date d = crv.trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [tension]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [tension]');

	return df;
}

disc_func disc_z_model_tension.df(disc_func dfb, curve crv, vector(number) option(nullable) pvbp)
{
	
	QL_REQUIRE(check_curve(crv),"invalid curve [tension]");

	disc_func df;

	if(is_spread_) {
		vector(number) s = calc_spreads(crv, pvbp);
		QL_REQUIRE(!null(s) && v_size(s) > 0,'invalid spread vector (not initialized) [tension_spread]');
		
		switch(type_)
		{
		case disc_z_model_type.DZ_TENSION_W_SPREAD:
			df = tension_spline_w_spreads_spr(dfb,crv, s, sigma_, epsilon_, l_der_);
			break;
		case disc_z_model_type.DZ_TENSION_Y_SPREAD:
			df = tension_spline_y_spreads_spr(dfb,crv, s, sigma_, l_der_, r_der_);
			break; 
		case disc_z_model_type.DZ_TENSION_Z_SPREAD:
			df = tension_spline_z_spreads_spr(dfb,crv, s, sigma_);
			break;			
		default:
			QL_FAIL('invalid disc model type [tension_spread]');
		}

		//QL_FAIL('unsupported model for fitting spread (only exact y-model available) [tension_spread]');		
	}
	else {
		switch(type_)
		{
		case disc_z_model_type.DZ_TENSION_W_EXACT:
			df = tension_spline_w_spr(dfb,crv, sigma_, epsilon_, l_der_);
			//QL_FAIL('unsupported model for fitting spread (only exact y-model available) [tension w]');
			break;
		case disc_z_model_type.DZ_TENSION_Y_EXACT:
			df = tension_spline_y_spr(dfb,crv, sigma_, l_der_, r_der_);
			break; 
		case disc_z_model_type.DZ_TENSION_Z_EXACT:
			df = tension_spline_z_spr(dfb,crv, sigma_);
			//QL_FAIL('unsupported model for fitting spread (only exact y-model available) [tension z]');
			break;			
		default:
			QL_FAIL('invalid disc model type [tension]');
		}
	}

	QL_REQUIRE(!null(df),'failed to create basic disc_func [tension]');

	if(use_cb_dates_) {
		date d = crv.instruments()[0].trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [tension]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [tension]');

	return df;

}


disc_func disc_z_model_tension.df(disc_func dfb, cashflows_base crv, vector(number) option(nullable) pvbp)
{
	
	//QL_REQUIRE(fiws.check_curve(crv),"invalid curve [tension]");

	disc_func df;

	if(is_spread_) {
		vector(number) s = calc_spreads(crv, pvbp);
		QL_REQUIRE(!null(s) && v_size(s) > 0,'invalid spread vector (not initialized) [tension_spread]');
		
		switch(type_)
		{
		case disc_z_model_type.DZ_TENSION_W_SPREAD:
			df = tension_spline_w_spreads_spr(dfb,crv, s, sigma_, epsilon_, l_der_);
			break;
		case disc_z_model_type.DZ_TENSION_Y_SPREAD:
			df = tension_spline_y_spreads_spr(dfb,crv, s, sigma_, l_der_, r_der_);
			break; 
		case disc_z_model_type.DZ_TENSION_Z_SPREAD:
			df = tension_spline_z_spreads_spr(dfb,crv, s, sigma_);
			break;			
		default:
			QL_FAIL('invalid disc model type [tension_spread]');
		}

		//QL_FAIL('unsupported model for fitting spread (only exact y-model available) [tension_spread]');		
	}
	else {
		switch(type_)
		{
		case disc_z_model_type.DZ_TENSION_W_EXACT:
			df = tension_spline_w_spr(dfb,crv, sigma_, epsilon_, l_der_);
			//QL_FAIL('unsupported model for fitting spread (only exact y-model available) [tension w]');
			break;
		case disc_z_model_type.DZ_TENSION_Y_EXACT:
			df = tension_spline_y_spr(dfb,crv, sigma_, l_der_, r_der_);
			break; 
		case disc_z_model_type.DZ_TENSION_Z_EXACT:
			df = tension_spline_z_spr(dfb,crv, sigma_);
			//QL_FAIL('unsupported model for fitting spread (only exact y-model available) [tension z]');
			break;			
		default:
			QL_FAIL('invalid disc model type [tension]');
		}
	}

	QL_REQUIRE(!null(df),'failed to create basic disc_func [tension]');

	if(use_cb_dates_) {
		date d = crv.trade_date();
		df = CORE_INT.flatten_ois_curve(df, d, cb_dates_, cal_);
		QL_REQUIRE(!null(df),'failed to create central bank date adjusted disc_func [tension]');
	}
	
	df = extrapol_model(df, CORE_INT.max_curve_t(crv));
	QL_REQUIRE(!null(df),'failed to create extrapolation disc_func [tension]');

	return df;

}

//-------------------------------------------------------
// disc_z_tension
//-------------------------------------------------------
disc_z_model_tension disc_z_tension(	disc_z_model_type				type,
										vector(date) option(nullable) 	cb_dates, 
										calendar option(nullable) 		cal,
										interpolation					sigma,				 
										number 							epsilon,
										number option(nullable)			l_der, 
										number  option(nullable)		r_der,
										disc_extrap_type 				xpol_type, 
										number option(nullable) 		xpol_cut)
option (category: 'Yield Curve/003 Curve Building')
{			
	return new disc_z_model_tension(type,cb_dates,cal,sigma, epsilon,xpol_type,xpol_cut,l_der,r_der);	
}	

//-------------------------------------------------------
// disc_z_tension
//-------------------------------------------------------
disc_z_model_tension disc_z_tension(	vector(date) option(nullable) 	cb_dates, 
										calendar option(nullable) 		cal,
										vector(number) option(nullable) spreads,
										z_mod_spread_type				spread_t,				 
										interpolation					sigma,
										number 							epsilon,
										number option(nullable)			l_der, 
										number  option(nullable)		r_der,
										disc_extrap_type 				xpol_type, 
										number option(nullable) 		xpol_cut)
option (category: 'Yield Curve/003 Curve Building')
{				
	return new disc_z_model_tension(cb_dates,cal,spreads, spread_t,sigma, epsilon,xpol_type,xpol_cut,l_der,r_der);	
}

//-------------------------------------------------------
// disc_z_tension
//-------------------------------------------------------
disc_z_model_tension disc_z_tension(	disc_z_model_type				type,
										vector(date) option(nullable) 	cb_dates, 
										calendar option(nullable) 		cal,
										interpolation					sigma,
										number option(nullable)			l_der, 
										number  option(nullable)		r_der,
										disc_extrap_type 				xpol_type, 
										number option(nullable) 		xpol_cut)
option (category: 'Yield Curve/003 Curve Building')
{			
	return new disc_z_model_tension(type,cb_dates,cal,sigma,xpol_type,xpol_cut,l_der,r_der);	
}	

//-------------------------------------------------------
// disc_z_tension
//-------------------------------------------------------
disc_z_model_tension disc_z_tension(	disc_z_model_type				type,
										vector(date) option(nullable) 	cb_dates, 
										calendar option(nullable) 		cal,
										vector(number) option(nullable) spreads,
										z_mod_spread_type				spread_t,				 
										interpolation					sigma,
										number option(nullable)			l_der, 
										number  option(nullable)		r_der,
										disc_extrap_type 				xpol_type, 
										number option(nullable) 		xpol_cut)
option (category: 'Yield Curve/003 Curve Building')
{				
	return new disc_z_model_tension(type, cb_dates,cal,spreads, spread_t,sigma,xpol_type,xpol_cut,l_der,r_der);	
}


//-------------------------------------------------------
// disc_df
// FIX_ME for swap_first_fix_
//-------------------------------------------------------
/*
disc_func calc_disc_df(	date 							d,
						curve							crv,
						calendar option(nullable)		cal,
						disc_z_model_type 				disc_model,
						interpolator  option(nullable) 	boot_ip,
						rate_type option(nullable)		boot_ip_rt,									
						vector(date) option(nullable)	cb_dates,
						date option(nullable)			step_end)
							   
{			
	disc_func df;
	if(disc_model == DZ_BOOT){			
		df = bootstrap(crv, boot_ip, boot_ip_rt);
		if(null(cb_dates) || v_size(cb_dates) == 0)
			return df;
		else 
			return CORE_INT.flatten_ois_curve(df, d, cb_dates, cal);			
	}
	else if(disc_model ==  DZ_MAX_SMOOTH){
		df = max_smooth(crv, SMOOTH_C2);
		if(null(cb_dates) || v_size(cb_dates) == 0 || v_size(v_find(cb_dates, [null<date>])) == v_size(cb_dates))
			return df;
		else 
			return CORE_INT.flatten_ois_curve(df, d, cb_dates, cal);
	}
	else if(disc_model ==  DZ_ON_STEP){			
		return bootstrap_on(crv, ip_step(),cal);
	}
	else {//DZ__STEP_BOOT		
		df = bootstrap_on(crv, ip_step(), cal);
		curve cut_curve = crv.chop_short(step_end);			
		return bootstrap_ext(df, step_end, cut_curve, boot_ip, boot_ip_rt, false);		
	}													
}
*/