option(null: hard);	

/*	
	Bootstrap cashflow functions
	Developed by Algorithmica Research, Magnus Nyström
	
	
	Comments: 


*/

module CORE_INT
{ 
	//-------------------------------------------------------
	// cf_fwd_on_multi
	// cashflows_fwd with o/n freq i.e. fixois
	//-------------------------------------------------------
	cashflows_fwd 	cf_fwd_on(	date d,
								date settle,
								calendar cal,
								day_count_method on_dc,
								vector(number) fixed_amounts,
								vector(date)	 fixed_dates)
	{
		vector(date) float_pmt;
		vector(date) float_idx0 ;
		date s 			= settle;
		date maturity 	= fixed_dates[v_size(fixed_dates)-1];
		maturity 		= cal.adjust_date(maturity,bd_convention.BD_FOLLOWING);//just in case
		
		while (s  < maturity){
			push_back(float_idx0, s );
			s = cal.move_bus_days(s,1);
			push_back(float_pmt, s);						
		}	

		vector(number) float_dt = period(float_idx0, float_pmt, on_dc)*100;
		return cashflows_fwd(d, fixed_dates, -fixed_amounts, float_idx0, float_pmt, float_dt);
	}
	//-------------------------------------------------------
	// cf_fwd_on_simple_flow
	// single flow as a fixois
	//-------------------------------------------------------
	cashflows_fwd 	cf_fwd_on_simple_flow(	date d,
											date settle,
											date maturity,
											calendar cal,
											bd_convention bd,
											day_count_method dc,
											number rate,
											day_count_method on_dc)
	{
		maturity						= cal.adjust_date(maturity,bd);
		vector(date) fixed_dates 		= [maturity];
		number irflow100				= period(settle,maturity,dc)*rate*100.0;
		vector(number) fixed_amounts 	= [irflow100] ;
		return cf_fwd_on(d,settle, cal,on_dc,fixed_amounts,fixed_dates);	
	}

	//-------------------------------------------------------
	// cf_fwd_fra_on
	// fra as a fixois
	//-------------------------------------------------------
	cashflows_fwd 	cf_fwd_fra_on(	instrument f,
									day_count_method on_dc,
									number r_sprd)
	{
		fra finst 			= fra(f);
		QL_FAIL_COND(null(finst),"invalid fra/ir-future");
		
		date d 				= finst.trade_date(); 	
		date settle 		= finst.settle_date();
		bd_convention bd 	= finst.pmt_bus_day();
		if(null(bd))
			bd = BD_FOLLOWING;

		number dp 			= finst.dirty_price();
		QL_FAIL_COND(null(dp) || !CORE_INT.is_equal_par(dp,0),"invalid fra/ir-future (not par)");

		day_count_method dc = finst.dc_method();
		date mat			= finst.index_maturity();
		//irflow based on index_mat (which is not the real cashflow) because we want the on-rates to compound to indexmat
		//number irflow100	= period(settle,mat,dc)*finst.yield()*100.0;
																		
		calendar cal 		= finst.calendar();
		number r 			= finst.yield()+r_sprd;
		return cf_fwd_on_simple_flow(d,settle,mat,cal,bd,dc,r,on_dc);

	}
	//-------------------------------------------------------
	// cf_fwd_depo_on
	// depo as a fix_ois
	//-------------------------------------------------------
	cashflows_fwd 	cf_fwd_depo_on(instrument dep,
								   day_count_method on_dc,
									number r_sprd)
	{
		
		date d 				= dep.trade_date(); 	
		date settle 		= dep.settle_date();
		bd_convention bd 	= dep.pmt_bus_day();
		if(null(bd))
			bd = BD_FOLLOWING;

		QL_FAIL_COND(settle != dep.issue_date(),"invalid deposit (start date <> settlement date)");	
		
		number y = dep.yield();
		number c = dep.coupon_rate();
		QL_FAIL_COND(!CORE_INT.is_equal_par(y,c),"invalid index deposit (not par)");
		

		calendar cal 		= dep.calendar();
		day_count_method dc = dep.dc_method();
		number r 			= dep.coupon_rate()+r_sprd;
		return cf_fwd_on_simple_flow(d,settle,dep.maturity(),cal,bd,dc,r,on_dc);

	}
	//-------------------------------------------------------
	// cf_fwd_fixois_on
	//-------------------------------------------------------
	cashflows_fwd 	cf_fwd_fixois_on(	instrument instr,
										day_count_method on_dc,
										number r_sprd)
	{
		
		date d 				= instr.trade_date();
		integer freq 		= instr.coupon_freq();			
		date settle 		= instr.settle_date();
		date issue_date		= settle;
		QL_FAIL_COND(settle != issue_date,"invalid fixois swap (start date <> settlement date)");
		date mat			= instr.maturity(); 
		bd_convention bd 	= instr.pmt_bus_day_d(BD_MOD_FOLLOWING);
		logical eom 		= instr.eom_d(false);
		calendar cal		= instr.calendar();
		day_count_method dc	= instr.accr_dc_method();
		number r			= instr.yield()+r_sprd;
		vector(date) d_pmt;
		swap_pmt_dates(issue_date,null<date>,null<date>, mat,0,freq,bd,cal,eom,d_pmt);

		integer size = v_size(d_pmt)-1;
		vector(number) fixed_amounts[size];
		vector(date) fixed_dates[size];
		for (integer i = 0 ; i < size; i++) {
			number per 	= period(d_pmt[i] ,d_pmt[i+1],dc);
			number cf 	= r * per * 100.0;
			fixed_amounts[i] = cf;
			fixed_dates[i] = d_pmt[i+1];
		}

		return cf_fwd_on(d, settle,cal,on_dc, fixed_amounts, fixed_dates);
	}
	
	//-------------------------------------------------------
	// cf_fwd_parswap
	//-------------------------------------------------------
	cashflows_fwd cf_fwd_parswap(	instrument		inst_in,									
									number			y_adj,
									number option(nullable) first_fix)
	{	
		QL_FAIL_COND(!inst_in.is_swap_fixflt(),"invalid instrument type");		
		QL_FAIL_COND(inst_in.settle_date() != inst_in.issue_date(),"invalid par swap (start date <> settlement date)");	

		instrument inst;
		if(!is_equal_zero(y_adj)) {
			swap_fixflt c 	= swap_fixflt(inst_in);
			c 				= c.set_rate_fix_leg(inst_in.yield() + y_adj );
			inst 			= c;
			//inst = inst_in.set_quote_from_yield(inst_in.yield() + y_adj);	//not for external swap		
		}
		else {
			inst = inst_in;
		}

		/*if(!inst.is_ext_swap()){
			error_info error = new error_info(true,false);
			number y = inst.yield(false,error);
			QL_FAIL_COND(null(y),strcat("invalid par swap - ", error.message()));			
			QL_FAIL_COND(!CORE_INT.is_equal_par(y,inst.coupon()),"invalid par swap (not par)");	
		}*/
		
		vector(date)	fixed_dates, float_idx0, float_pmt;
		vector(number)	fixed_amounts, float_dt;
		
		error_info error = new error_info(true,true);
		logical ok = inst.cash_flows_fwd_bootstrap(fixed_dates,fixed_amounts, first_fix,float_idx0,float_pmt,float_dt,error) ;

		return ok ? cashflows_fwd(inst.trade_date(), fixed_dates, fixed_amounts, float_idx0, float_pmt, float_dt) : null<cashflows_fwd>;	
	}

	//-------------------------------------------------------
	// cf_parswap
	//-------------------------------------------------------
	cashflows cf_parswap(	instrument		inst_in,																
							number			y_adj)
	{		
		QL_FAIL_COND(!inst_in.is_swap_fixflt(),"invalid instrument type");		
		QL_FAIL_COND(inst_in.settle_date() != inst_in.issue_date(),"invalid par swap (start date <> settlement date)");	

		instrument inst;
		if(!is_equal_zero(y_adj)) {
			
			swap_fixflt c 	= swap_fixflt(inst_in);
			c 				= c.set_rate_fix_leg(inst_in.yield() + y_adj );
			inst = c;
			//inst = inst_in.set_quote_from_yield(inst_in.yield() + y_adj);
			
		}
		else {
			inst = inst_in;
		}

		/*if(!inst.is_ext_swap()){
			error_info error = new error_info(true,false);
			number y = inst.yield(false,error);
			QL_FAIL_COND(null(y),strcat("invalid par swap - ", error.message()));			
			QL_FAIL_COND(!CORE_INT.is_equal_par(y,inst.coupon()),"invalid par swap (not par)");	
		}*/
				
		vector(date)	fixed_dates;
		vector(number)	fixed_amounts;
		error_info error = error_info(true,true);
		logical ok = inst.cash_flows_bootstrap(fixed_dates,fixed_amounts,error) ;
		

		return ok ? cashflows(inst.trade_date(), fixed_dates, fixed_amounts): null<cashflows>;
	}

	//-------------------------------------------------------
	// cf_parswap_sprd
	// assumption: settlement date == issue date
	// this is a swap plus a float spread
	//-------------------------------------------------------
	cashflows cf_parswap_sprd(	instrument		inst_in,
								number 			sprd,									
								//number			roll_day,															
								//integer			tenor_freq,
								//day_count_method dc,
								//bd_convention 	bd,
								//calendar		cal,
								//logical			eom,								
								number			y_adj)
	{
		
		QL_FAIL_COND(!inst_in.is_swap_fixflt(),"invalid instrument type");	
		QL_FAIL_COND(inst_in.settle_date() != inst_in.issue_date(),"invalid par swap (start date <> settlement date)");	
			
		//QL_FAIL_COND(!c.is_par_flat(),"invalid par swap (not par)");

		/*instrument inst;
		if(!is_equal_zero(y_adj)) {
			number cpn1,cpn2;
			c.coupon(cpn1,cpn2);
			//c.add_quote(cpn1 + y_adj,..quote_style.YIELD);
			//inst = c;//CORE_INT.create_par_swap(inst, inst.yield() + y_adj, null<string>);
			inst = inst_in.set_quote_from_yield(cpn1 + y_adj);			
		}
		else {
			inst = inst_in;
		}*/
		
		swap_fixflt c 	= swap_fixflt(inst_in);

		error_info ee 	= error_info(true,true);
		number			roll_day = c.roll_day(ee);
		
		integer  cpn_freq_fix_leg, tenor_freq;
		c.coupon_freq(cpn_freq_fix_leg, tenor_freq,ee);
		
		day_count_method dc_fix, dc;	
		c.dc_method(dc_fix,dc,ee);
		
		bd_convention bd_fix,bd;
		c.pmt_bus_day(bd_fix,bd,ee);

		logical eom_fix, eom;
		c.eom(eom_fix,eom,ee);//eom_d(true);

		calendar cal_fix,cal;
		c.calendar(cal_fix,cal,ee);//calendar();
	
		instrument inst;
		if(!is_equal_zero(y_adj)) {
			
			
			c 				= c.set_rate_fix_leg(inst_in.yield() + y_adj );
			inst 			= c;
		}
		else {
			inst = inst_in;
		}
		
		
		vector(date)	fixed_dates_o;
		vector(number)	fixed_amounts_o;			 			
		inst.cash_flows(fixed_dates_o, fixed_amounts_o);
	
		QL_FAIL_COND(null(fixed_amounts_o) || v_size(fixed_amounts_o) == 0,"invalid par swap (cash flows)");

		number inst_nom = inst.nominal();
		fixed_amounts_o = fixed_amounts_o/inst_nom*100;
		
		bd_convention	idx_bd = bd;
		calendar		idx_cal = cal;
		logical 		idx_eom = eom;

		logical ip_idx_stub = false;
		date out_first,out_last_reg;
		number out_roll_day;
		logical is_reg_first,is_reg_last,eom_roll;
		vector(date) float_idx0, float_pmt, float_idx1;
		CORE_INT.swap_flt_dates(inst.issue_date(),null<date>,null<date>, inst.maturity(),roll_day,tenor_freq,bd,cal,eom,idx_bd,idx_cal,idx_eom,
							ip_idx_stub,out_first, out_last_reg,out_roll_day,is_reg_first,is_reg_last,eom_roll,
							float_idx0,float_pmt,float_idx1);
	
	
		vector(number) float_dt = period(float_idx0, float_pmt, dc);
		vector(number) cf_sprd = float_dt * sprd * 100;
		
		vector(date)	fixed_dates 	= concat(float_pmt, fixed_dates_o);
		vector(number)	fixed_amounts	= concat(cf_sprd,fixed_amounts_o);

		sort (fixed_amounts, fixed_dates);
		sort (fixed_dates, fixed_dates);
		
		fixed_dates 	= concat([inst.issue_date()], fixed_dates);
		fixed_amounts	= concat([-100],fixed_amounts);


		vector(date) fixed_dates_c;
		vector(number) fixed_amounts_c;
		CORE_INT.consolidate(fixed_dates,fixed_amounts, fixed_dates_c,fixed_amounts_c);

		return cashflows(inst.trade_date(), fixed_dates_c, fixed_amounts_c);
	}
	//-------------------------------------------------------
	// cf_fwd_fra
	//-------------------------------------------------------
	cashflows_fwd cf_fwd_fra(	instrument		inst_in,														
								number			rate_adj)
	
	{	
		QL_FAIL_COND(!inst_in.is_fra(),"invalid instrument type");	

		instrument inst;
		if(!is_equal_zero(rate_adj)) {
			//fra c = fra(inst_in);
			inst = inst_in.set_quote_from_yield(inst_in.yield() + rate_adj);
			//inst = c;
		}
		else {
			inst = inst_in;
		}

		//number y = inst.yield();
		//QL_FAIL_COND(null(y) || !CORE_INT.is_equal_par(y,inst.coupon()),"invalid fra (not par)");	
				
		vector(date)	fixed_dates, float_idx0, float_pmt;
		vector(number)	fixed_amounts, float_dt;
		error_info error = error_info(true,true);	
		logical ok = inst.cash_flows_fwd_bootstrap(fixed_dates,fixed_amounts, null<number>,float_idx0,float_pmt,float_dt,error) ;
		
		return ok ? cashflows_fwd(inst.trade_date(), fixed_dates, fixed_amounts, float_idx0, float_pmt, float_dt) : null<cashflows_fwd>;

	}

	//-------------------------------------------------------
	// cf_fra  (cashflow sign = lending)
	//-------------------------------------------------------
	/*cashflows cf_fra(	date d0, 
						date beg, 
						date end, 
						number rdt, 
						number nom)
	{
		vector(date)	fixed_dates 	= [ beg, end ];
		vector(number)	fixed_amounts 	= [ -nom, (1 + rdt)*nom];
		return cashflows(d0, fixed_dates, fixed_amounts);
	}*/	
	//-------------------------------------------------------
	// cf_fra
	//-------------------------------------------------------
	cashflows cf_fra(	instrument		inst_in,
						number			rate_adj)
	
	{		
		QL_FAIL_COND(!inst_in.is_fra(),"invalid instrument type");	

		instrument inst;
		if(!is_equal_zero(rate_adj)) {
			//fra c = fra(inst_in);
			//c.add_quote(inst.yield() + rate_adj,true);
			inst = inst_in.set_quote_from_yield(inst_in.yield() + rate_adj);
			//inst = c;
		}
		else {
			inst = inst_in;
		}

		//number y = inst.yield();
		//QL_FAIL_COND(null(y) || !CORE_INT.is_equal_par(y,inst.coupon()),"invalid fra (not par)");	

		vector(date)	fixed_dates;
		vector(number)	fixed_amounts;
		error_info error = error_info(true,true);
		logical ok = inst.cash_flows_bootstrap(fixed_dates,fixed_amounts,error) ;
		

		return ok ? cashflows(inst.trade_date(), fixed_dates, fixed_amounts): null<cashflows>;				
	}
		
	
	//-------------------------------------------------------
	// cf_fwd_mmfut
	//-------------------------------------------------------
	cashflows_fwd cf_fwd_mmfut(	instrument		inst_in,									
								number			rate_adj)//negative sign for convexadjust						
	{	
		QL_FAIL_COND(!inst_in.is_ir_future(),"invalid instrument type");	

		instrument inst;
		if(!is_equal_zero(rate_adj)) {
			//ir_future c = ir_future(inst_in);
			inst = inst_in.set_quote_from_yield(inst_in.yield() + rate_adj);
			//inst = c;
		}
		else {
			inst = inst_in;
		}

		//number y = inst.yield();
		//QL_FAIL_COND(null(y) || !CORE_INT.is_equal_par(y,inst.coupon()),"invalid ir future (not par)");	
				
		vector(date)	fixed_dates, float_idx0, float_pmt;
		vector(number)	fixed_amounts, float_dt;
		error_info error = error_info(true,true);	
		logical ok = inst.cash_flows_fwd_bootstrap(fixed_dates,fixed_amounts, null<number>,float_idx0,float_pmt,float_dt,error) ;
		
		return ok ? cashflows_fwd(inst.trade_date(), fixed_dates, fixed_amounts, float_idx0, float_pmt, float_dt) : null<cashflows_fwd>;

		/*			
		number dp = inst.dirty_price();
		QL_FAIL_COND(null(dp) || !CORE_INT.is_equal_par(dp,0),"invalid mm future (not par)");
	
		vector(date) fixed_dates 		= [move_month(inst.settle_date(),12/tenor_freq,eom,cal, bd, false)];
		number r 						= inst.yield() + rate_adj;
		vector(number) fixed_amounts 	= [-period (inst.settle_date(), fixed_dates[0], dc) * r * 100];				
		//if(rev_sign)
			//fixed_amounts = -fixed_amounts ;
			
		vector(date) float_idx0 = [inst.settle_date()];
		vector(date) float_pmt = [fixed_dates[0]];	
		vector(number) float_dt = period (float_idx0, float_pmt, dc) * 100;
		return cashflows_fwd(inst.trade_date(), fixed_dates, fixed_amounts, float_idx0, float_pmt, float_dt);*/
	}
	//-------------------------------------------------------
	// cf_mmfut  (cashflow sign = lending)
	//-------------------------------------------------------
	cashflows cf_mmfut(	instrument		inst,
						number			rate_adj)
	
	{
		return cf_fra(inst, rate_adj);
	}	
	//-------------------------------------------------------
	// cf_fwd_depo
	//-------------------------------------------------------
	cashflows_fwd cf_fwd_depo(	instrument		inst_in,
								number			rate_adj)
	{	
		QL_FAIL_COND(!inst_in.is_deposit(),"invalid instrument type");	
		QL_FAIL_COND(inst_in.settle_date() != inst_in.issue_date(),"invalid deposit (issue date <> settlement date)");

		instrument inst;
		if(!is_equal_zero(rate_adj)) {
			//deposit c = deposit(inst);
			inst = inst_in.set_quote_from_yield(inst_in.yield() + rate_adj);			
		}
		else {
			inst = inst_in;
		}

		//number y = inst.yield();
		//QL_FAIL_COND(null(y) || !CORE_INT.is_equal_par(y,inst.coupon()),"invalid deposit (not par)");	
				
		vector(date)	fixed_dates, float_idx0, float_pmt;
		vector(number)	fixed_amounts, float_dt;
		error_info error = error_info(true,true);	
		logical ok = inst.cash_flows_fwd_bootstrap(fixed_dates,fixed_amounts, null<number>,float_idx0,float_pmt,float_dt,error) ;
		
		return ok ? cashflows_fwd(inst.trade_date(), fixed_dates, fixed_amounts, float_idx0, float_pmt, float_dt) : null<cashflows_fwd>;
	}
	//-------------------------------------------------------
	// cf_depo
	//-------------------------------------------------------
	cashflows cf_depo(	instrument		inst_in,
						number			rate_adj)
	
	{				
		QL_FAIL_COND(!inst_in.is_deposit(),"invalid instrument type");	
		QL_FAIL_COND(inst_in.settle_date() != inst_in.issue_date(),"invalid deposit (issue date <> settlement date)");	
	
		instrument inst;
		if(!is_equal_zero(rate_adj)) {
			//deposit c = deposit(inst);
			inst = inst_in.set_quote_from_yield(inst_in.yield() + rate_adj);			
		}
		else {
			inst = inst_in;
		}

		//number y = inst.yield();
		//QL_FAIL_COND(null(y) || !CORE_INT.is_equal_par(y,inst.coupon()),"invalid deposit (not par)");	

		vector(date)	fixed_dates;
		vector(number)	fixed_amounts;
		error_info error = error_info(true,true);
		logical ok = inst.cash_flows_bootstrap(fixed_dates,fixed_amounts,error) ;
			

		return ok ? cashflows(inst.trade_date(), fixed_dates, fixed_amounts): null<cashflows>;	
	}
	

	class on_helper
	{
	public:
		on_helper(date,day_count_method,calendar, date);

		date 			end();
		void 			init(date);
		void 			calc_df(date , date , number );
		vector(number) 	calc_per(date start, date end);
		
		date 				d_;
		day_count_method 	dc_;
		calendar 			cal_;
		vector(date) 		accr_start_;
		vector(date) 		accr_end_;
		vector(number) 		accr_per_;
	};

	on_helper.on_helper(	date 		d,
						  day_count_method 	dc,
						  calendar 	cal,
						  date 		max): d_(d), dc_(dc), cal_(cal)
	{
		resize(accr_start_,0);
		resize(accr_end_,0);
		resize(accr_per_,0);
		init(max);
	}

	void on_helper.init(date max)
	{		
		date s 	= d_;
		date se = cal_.move_bus_days(s,1);
		
		while (se  <= max || (se > max && s < max)){
			push_back(accr_start_,s);
			push_back(accr_end_,se);	
			push_back(accr_per_, period(s,se,dc_));
			s 	= se;
			se 	= cal_.move_bus_days(s,1);
		}
	}

	date on_helper.end()
	{
		return accr_end_[v_size(accr_end_)-1];
	}
	
	void on_helper.calc_df(date start, date end, number r)
	{
		vector(integer) vs = v_find_eq (accr_start_, start) ;
		vector(integer) ve = v_find_eq (accr_end_, end) ;
		number df = 1;
		for(integer i=vs[0]; i <= ve[0];i++) {
			df /= (1+r*accr_per_[i]);
		}
	}

	vector(number) on_helper.calc_per(date start, date end)
	{
		vector(integer) vs = v_find_eq (accr_start_, start) ;
		vector(integer) ve = v_find_eq (accr_end_, end) ;

		return sub_vector(accr_per_, vs[0], ve[0]-vs[0]+1); 
	}
	//-------------------------------------------------------
	// cf_fwd_on
	//-------------------------------------------------------
	/*cashflows_fwd cf_fwd_on(	date d,
								date settle,
								date maturity,
								number r,
								day_count_method dc,
								calendar cal)
	{
		on_helper on 	= new on_helper(settle,dc,cal,maturity);
		number per 		= period(settle,on.end(),dc);
		number ir 		= future_interest(100,r,per, RT_SIMPLE);		
		return cashflows_fwd(d, [maturity], [-ir], on.accr_start_, on.accr_end_, on.accr_per_*100);
	}
	*/
	//-------------------------------------------------------
	// fwd_on_step
	//-------------------------------------------------------
	/*fwd_func fwd_on_step_(	date 		d,
							date 		settle,
							date 		maturity,
							number 		r,
							day_count_method dc,
							calendar 	cal,
							out disc_func df_disc)
	{
		//cashflows_fwd cf = cf_fwd_on(d,settle,maturity, r,dc,cal);
		instr_def def 	= create_depo_instr_def(dc, BD_MOD_FOLLOWING,true ,"BD2" , cal, "ZZZ");
		instrument depo = deposit(def,"",d,settle,settle,maturity,null<string>,r,null<number>,100);
		curve crv 		= curve([depo]);
		//df_disc 		= bootstrap_on_extrap_flat(crv, d,max,cal);
		//OBS dc
		return bootstrap_on(crv,  cal);
		//return bootstrap_fwd(cf, df_disc, ip_step());
	}*/
	//-------------------------------------------------------
	// fwd_on_step
	//-------------------------------------------------------
	/*fwd_func fwd_on_step_(	date 		d,
							date 		settle,
							date 		maturity,
							number 		r,
							day_count_method dc,
							calendar 	cal,
							disc_func	df_pre,
							fwd_func	fwd_pre,
							date 		cut_off_d,
							out disc_func df_disc)
	{
		cashflows_fwd cf 	= cf_fwd_on(d,settle,maturity, r,dc,cal);
		instr_def def 		= create_depo_instr_def(dc, BD_MOD_FOLLOWING,true ,"BD2" , cal, "ZZZ");
		instrument depo 	= deposit(def,"",d,settle,settle,maturity,null<string>,r,null<number>,100);
		curve crv 			= curve([depo]);
		//df_disc 			= bootstrap_on_ext_extrap_flat(df_pre,cut_off_d,crv, ip_step(),cal);		
		//return bootstrap_fwd_ext(fwd_pre,cut_off_d,cf, df_disc, ip_step());
		return bootstrap_on_ext(null<fwd_func>,df_pre,cut_off_d,crv, cal);
		//-->doesn't work because of
		//1) the "backward filling" in step() in combination with
		//2)
		//fixed but remains to be tested
	}*/
	//-------------------------------------------------------
	// fwd_on_step_alt
	//-------------------------------------------------------
	/*fwd_func fwd_on_step_alt(	date 		d,
							date 		settle,
							date 		maturity,
							number 		r,
							day_count_method dc,
							calendar 	cal,
							disc_func	df_pre,
							fwd_func	fwd_pre,
							date 		cut_off_d,
							out disc_func df_disc)
	{
		cashflows_fwd cf 	= cf_fwd_on(d,settle,maturity, r,dc,cal);
		instr_def def 		= create_depo_instr_def(dc, BD_MOD_FOLLOWING,true ,"BD2" , cal, "ZZZ");
		instrument depo 	= deposit(def,"",d,settle,settle,maturity,null<string>,r,null<number>,100);
		curve crv 			= curve([depo]);
		//df_disc 			= bootstrap_on_ext_extrap_flat(df_pre,cut_off_d,crv, ip_step(),cal);
		//return fwd_func(df_disc,365,d,dc,cal,BD_FOLLOWING,false);

		return bootstrap_on_ext(null<fwd_func>,df_pre,cut_off_d,crv, cal);
		//return bootstrap_fwd_ext(fwd_pre,cut_off_d,cf, df_disc, ip_step());
	}*/
	//-------------------------------------------------------
	// cf_fwd_oisswap
	// assumption: settlement date == issue date
	// also assume floating index has the same cal as float leg
	//-------------------------------------------------------
	cashflows_fwd cf_fwd_oisswap(	instrument		inst_in,								
									//day_count_method dc,								
									//calendar		cal,																
									number			y_adj,
									number option(nullable) first_fix)
	{		
		QL_FAIL_COND(!inst_in.is_swap_fixois(),"invalid instrument type");		
		QL_FAIL_COND(inst_in.settle_date() != inst_in.issue_date(),"invalid ois (start date <> settlement date)");	

		instrument inst;
		if(!is_equal_zero(y_adj)) {
			
			inst = inst_in.set_quote_from_yield(inst_in.yield() + y_adj); //FIX_ME for external swaps
			
			//swap_fixois c 	= swap_fixois(inst_in);
			//c 				= c.set_rate_fix_leg(inst_in.yield() + y_adj );
			//inst = c;
		}
		else {
			inst = inst_in;
		}

		//number y = inst.yield();
		//QL_FAIL_COND(null(y) || !CORE_INT.is_equal_par(y,inst.coupon()),"invalid ois (not par)");	
		
			
		vector(date)	fixed_dates, float_idx0, float_pmt;
		vector(number)	fixed_amounts, float_dt;
		error_info error = error_info(true,true);
		logical ok = inst.cash_flows_fwd_bootstrap(fixed_dates,fixed_amounts, null<number>,float_idx0,float_pmt,float_dt,error) ;
		
		return ok ? cashflows_fwd(inst.trade_date(), fixed_dates, fixed_amounts, float_idx0, float_pmt, float_dt) : null<cashflows_fwd>;
		
	}

	//-------------------------------------------------------
	// cf_oisswap
	// assumption: settlement date == issue date
	//-------------------------------------------------------
	cashflows cf_oisswap(	instrument		inst_in,																
							number			rate_adj)
	{
		QL_FAIL_COND(!inst_in.is_swap_fixois(),"invalid instrument type");	
		QL_FAIL_COND(inst_in.settle_date() != inst_in.issue_date(),"invalid ois (issue date <> settlement date)");	

		instrument inst;
		if(!is_equal_zero(rate_adj)) {
			
			inst = inst_in.set_quote_from_yield(inst_in.yield() + rate_adj);
			//swap_fixois c 	= swap_fixois(inst_in);
			//c 				= c.set_rate_fix_leg(inst_in.yield() + y_adj );
			//inst = c;
		}
		else {
			inst = inst_in;
		}

		//number y = inst.yield();
		//QL_FAIL_COND(null(y) || !CORE_INT.is_equal_par(y,inst.coupon()),"invalid ois (not par)");	

		vector(date)	fixed_dates;
		vector(number)	fixed_amounts;
		error_info error = error_info(true,true);
		logical ok = inst.cash_flows_bootstrap(fixed_dates,fixed_amounts,error) ;
		

		return ok ? cashflows(inst.trade_date(), fixed_dates, fixed_amounts): null<cashflows>;	
			
	}

	
	
	//-------------------------------------------------------
	// cf_float_bswap (NOT for currency basis swaps)
	// (basis swap quoted as 1 swap)
	// ex EUR 3M vs 6M, 1yr: +31bp
	// Party A pays: 3-month Euribor + spread (bps), quarterly, act/360 for both Euribor and spread
	// Party B pays: 6-month Euribor, semi-annually, act/360
	//-------------------------------------------------------
	cashflows_fwd cf_fwd_bswap(	instrument		inst,										
								logical			solve_flat, 
								fwd_func		ff,//for the leg we don't solve for
								number			q_adj_bp,
								number option(nullable) first_fix_flat,
								number option(nullable) first_fix)
	{
	    QL_FAIL_COND(!inst.is_swap_fltflt(),"invalid instrument type");	
		QL_FAIL_COND(inst.settle_date() != inst.issue_date(),"invalid float-float swap (issue date <> settlement date)");
		QL_FAIL_COND(inst.trade_date() > inst.issue_date(),"invalid basis swap (start date)");
					
		swap_fltflt sw 				= swap_fltflt(inst);
		logical ip_idx_stub 		= false;

		//---spread leg date calc---
		integer freq1,freq2;
		sw.reset_freq(freq1,freq2 );		

		integer cfreq1,cfreq2;//currently not used 
		sw.coupon_freq(cfreq1,cfreq2 );
		
		day_count_method dc1,dc2;
		sw.dc_method(dc1,dc2); 
		
		bd_convention bd1, bd2;
		sw.pmt_bus_day(bd1, bd2 );

		calendar cal1, cal2;
		sw.calendar(cal1, cal2 );

		logical eom1, eom2;
		sw.eom(eom1, eom2 );

		integer roll_day1,roll_day2;
		sw.roll_day(roll_day1,roll_day2 );	

		bd_convention	idx_bd1 	= sw.ir_index_leg1().bd_convention();
		calendar		idx_cal1 	= sw.ir_index_leg1().fixing_calendar();
		logical 		idx_eom1 	= sw.ir_index_leg1().eom();
		
		date out_first,out_last_reg;
		number out_roll_day;
		logical is_reg_first,is_reg_last,eom_roll;
		vector(date) float_idx0_sprd, float_pmt_sprd, float_idx1_sprd;
		CORE_INT.swap_flt_dates(inst.issue_date(),null<date>,null<date>, inst.maturity(),roll_day1,freq1,bd1,cal1,eom1,
								idx_bd1,idx_cal1,idx_eom1,ip_idx_stub,out_first, out_last_reg,out_roll_day,is_reg_first,
								is_reg_last,eom_roll,float_idx0_sprd,float_pmt_sprd,float_idx1_sprd);
		
		integer n_sprd 				= v_size(float_idx0_sprd);		

		//---flat leg date calc---
		bd_convention	idx_bd2 	= sw.ir_index_leg2().bd_convention();
		calendar		idx_cal2 	= sw.ir_index_leg2().fixing_calendar();
		logical 		idx_eom2 	= sw.ir_index_leg2().eom();

		vector(date) float_idx0_flat, float_pmt_flat, float_idx1_flat;
		CORE_INT.swap_flt_dates(inst.issue_date(),null<date>,null<date>, inst.maturity(),roll_day2,freq2,bd2,cal2,eom2,
								idx_bd2,idx_cal2,idx_eom2,ip_idx_stub,out_first, out_last_reg,out_roll_day,is_reg_first,
								is_reg_last,eom_roll,float_idx0_flat,float_pmt_flat,float_idx1_flat);
		
		number n_flat = v_size(float_idx0_flat);
	
		//-----------------------

		number spr 	= !null(q_adj_bp) ? (inst.quote()+q_adj_bp)/10000 : inst.quote()/10000;

		vector(number)	fixed_amounts[0];
		vector(date)	fixed_dates[0];		
	    				
		if(solve_flat) {				
	
			vector(number) per_pay_sprd = period (float_idx0_sprd, float_pmt_sprd, dc1);
			vector(number) per 			= (float_idx0_sprd - inst.trade_date()) / 365;
			vector(number) r_fwd		= ff.fwd(per);
			if(!null(first_fix)) r_fwd[0] = first_fix;

			//FIX_ME do adjustment of the known rates to account for cpnfreq<>resetfreq
			
			for (integer i = 0 ; i < n_sprd ; i++) {																		
				push_back(fixed_dates, float_pmt_sprd[i]); 
				push_back(fixed_amounts, (r_fwd[i] + spr) * per_pay_sprd[i]);										
		    }
	
			vector(number) float_dt_flat = period(float_idx0_flat , float_pmt_flat , dc2);
			
			if(!null(first_fix_flat)){
				if(v_size(float_idx0_flat) == 1) return null<cashflows_fwd>;			
				number cf 		= period (inst.issue_date(), float_pmt_flat[0], dc2) * first_fix_flat; 
				push_back(fixed_dates, float_pmt_flat[0]);				
				push_back(fixed_amounts, -cf);		
				float_idx0_flat = v_exclude(float_idx0_flat, 0);			
				float_pmt_flat 	= v_exclude(float_pmt_flat, 0);
				float_dt_flat 	= v_exclude(float_dt_flat, 0);								
			}
			fixed_amounts *= 100;
			float_dt_flat *= 100;
			return cashflows_fwd(inst.trade_date(), fixed_dates, -fixed_amounts, float_idx0_flat , float_pmt_flat , float_dt_flat);
			
		}
		else { 
			vector(number) per_pay_sprd = period (float_idx0_sprd, float_pmt_sprd, dc1);
			vector(number) per_pay_flat = period (float_idx0_flat, float_pmt_flat, dc2);			
			vector(number) per 			= (float_idx0_flat - inst.trade_date()) / 365;
			vector(number) r_fwd		= ff.fwd(per);
			if(!null(first_fix_flat)) r_fwd[0] = first_fix_flat;
	
			for (integer i = 0 ; i < n_flat; i++) {																
				push_back(fixed_dates, float_pmt_flat[i]); 
				push_back(fixed_amounts, r_fwd[i] * per_pay_flat[i]);			
		    }
	
			for (integer i = 0 ; i < n_sprd ; i++) {																		
				push_back(fixed_dates, float_pmt_sprd[i]); 
				push_back(fixed_amounts, -spr * per_pay_sprd[i]);											
		    }
			vector(number) float_dt_sprd = period(float_idx0_sprd, float_pmt_sprd, dc1);
			
	
			if(!null(first_fix)){	
				if(v_size(float_idx0_sprd) == 1) return null<cashflows_fwd>;			
				number cf 		= period (inst.issue_date(), float_pmt_sprd[0], dc1) * first_fix; 
				push_back(fixed_dates, float_pmt_sprd[0]);				
				push_back(fixed_amounts, -cf);		
				float_idx0_sprd = v_exclude(float_idx0_sprd, 0);			
				float_pmt_sprd 	= v_exclude(float_pmt_sprd, 0);
				float_dt_sprd 	= v_exclude(float_dt_sprd, 0);							
			}
			fixed_amounts *= 100;
			float_dt_sprd *= 100;
			return cashflows_fwd(inst.trade_date(), fixed_dates, -fixed_amounts, float_idx0_sprd, float_pmt_sprd, float_dt_sprd);				
		}
	}

	//-------------------------------------------------------
	// cf_fwd_parswap2s
	// assumption: settlement date == issue date
	// also assume floating index has the same eom, cal, bd as float leg
	//-------------------------------------------------------
	cashflows_fwd cf_fwd_parswap2s(	instrument		inst_in,									
									number option(nullable)	roll_day,															
									integer			tenor_freq,
									day_count_method  dc,
									bd_convention 	option(nullable) bd,
									calendar		option(nullable) cal,
									logical			option(nullable) eom,
									number			y_adj,
									number option(nullable) first_fix)
	{			
		QL_FAIL_COND(inst_in.settle_date() != inst_in.issue_date(),"invalid par swap (start date <> settlement date)");	
		
		swap_fixflt c = swap_fixflt(inst_in);	
		QL_FAIL_COND(!c.is_par_flat(),"invalid par swap (not par)");

		if(null(roll_day))
			roll_day = day(inst_in.issue_date());

		if(null(bd))
			bd = inst_in.pmt_bus_day();

		if(null(cal))
			cal = inst_in.calendar();

		if(null(eom))
			eom = inst_in.eom();
				
		instrument inst;
		if(!is_equal_zero(y_adj)) {			
			number cpn1,cpn2;
			c.coupon(cpn1,cpn2);
			inst = inst_in.set_quote_from_yield(cpn1 + y_adj);
			
		}
		else {
			inst = inst_in;
		}
		
		vector(date)	fixed_dates;
		vector(number)	fixed_amounts;			 			
		inst.cash_flows(fixed_dates, fixed_amounts);

		number inst_nom = inst.nominal(); 
		QL_FAIL_COND(null(fixed_amounts) || v_size(fixed_amounts) == 0,"invalid par swap (cash flows)");
	
		fixed_amounts[v_size(fixed_amounts) - 1] -= inst_nom;
		fixed_amounts = -fixed_amounts/inst_nom ;
		
		bd_convention	idx_bd = bd;
		calendar		idx_cal = cal;
		logical 		idx_eom = eom;

		logical ip_idx_stub = false;
		date out_first,out_last_reg;
		number out_roll_day;
		logical is_reg_first,is_reg_last,eom_roll;
		vector(date) float_idx0, float_pmt, float_idx1;
		CORE_INT.swap_flt_dates(inst.issue_date(),null<date>,null<date>, inst.maturity(),roll_day,tenor_freq,bd,cal,eom,idx_bd,idx_cal,idx_eom,
							ip_idx_stub,out_first, out_last_reg,out_roll_day,is_reg_first,is_reg_last,eom_roll,
							float_idx0,float_pmt,float_idx1);
	
	
		vector(number) float_dt = period(float_idx0, float_pmt, dc);
	
		if(!null(first_fix)){	
			if(v_size(float_idx0) == 1)
				return null<cashflows_fwd>;
			number cf = period (inst.issue_date(), float_pmt[0], dc) * first_fix; 
			push_back(fixed_dates, float_pmt[0]);
			push_back(fixed_amounts, cf);		
			float_idx0 = v_exclude(float_idx0, 0);		
			float_pmt = v_exclude(float_pmt, 0);
			float_dt = v_exclude(float_dt, 0);				
		}
		fixed_amounts 	*= 100;
		float_dt		*= 100;
		return cashflows_fwd(inst.trade_date(), fixed_dates, fixed_amounts, float_idx0, float_pmt, float_dt);	
	}
	//-------------------------------------------------------
	// vector_adj
	//-------------------------------------------------------
	void vector_adj(integer size,
					number option(nullable)  dflt,
					out vector(number)	v)
	{
		if(null(v)) {
			resize(v,size);
			v = dflt;
		}
		else if(v_size(v) == 1 && size > 1) {
			number a = v[0];
			resize(v,size);
			v = a;
		}
	}

	//-------------------------------------------------------
	// cf_b1
	//-------------------------------------------------------
	cashflows cf_b1(vector(instrument) src,
					vector(instrument) adj_lib_swap,
					vector(number) basis_sprd,
					vector(number) option(nullable)	y_adj)
	{		
		integer size = v_size(src);
		
		vector_adj(size,0,y_adj);
		QL_REQUIRE(v_size(y_adj) == size, "invalid size of y-shift vector");
		
		vector(cashflows)	dest[0];			
		vector(string) adj_lib_swap_name = adj_lib_swap.name();
		error_info ee = new error_info(true,true);
		
		for(integer i = 0; i < size ; i++){
			
			cashflows c;

			instr_type it = src[i].instr_type(ee);
			
			switch(it){
			case instr_type.DEPOSIT:
			case instr_type.DEPOSIT_INDEX:
				c = CORE_INT.cf_depo(src[i], y_adj[i]);
				break;
			case instr_type.SWAP_FIXOIS:
				c = CORE_INT.cf_oisswap(src[i], y_adj[i]);
				break;
			case instr_type.FRA:
				c = CORE_INT.cf_fra(src[i], y_adj[i]);
				break;
			case instr_type.IR_FUTURE:
				c = CORE_INT.cf_mmfut(src[i], y_adj[i]);
				break;
				
			case instr_type.SWAP_FIXFLT:
			{			
				vector(integer) v_f = v_findi(adj_lib_swap_name, [src[i].name()]);
				integer v_f_size = v_size(v_f);				
				QL_REQUIRE(v_f_size < 2,'invalid swap index (internal error)'); 
				if(v_f_size == 1) 					 
					c = CORE_INT.cf_parswap_sprd(src[i], -basis_sprd[v_f[0]], y_adj[i]);	
				else 
					c = CORE_INT.cf_parswap(src[i], y_adj[i]);
				break;
				
			}										
			default:
				QL_FAIL("invalid instrument type in curve (b1 disc)");
			}
			
			if(null(c)) {
				string msg = strcat([	"<","invalid cashflow (b1 disc)" ,"[", src[i].name(),"]", ">"]);	;
				CORE_INT.write_warning_message( msg);
			}
			else
				push_back(dest, c);
		}

		return cashflows(dest);
	}
	//-------------------------------------------------------
	// cf_fwd
	//-------------------------------------------------------
	cashflows_fwd cf_fwd(vector(instrument) src,
						 vector(number) option(nullable) first_fix,
						 vector(number) option(nullable)	y_adj,
						 logical require_all_cf)
	{					
		integer size = v_size(src);
		vector_adj(size,null<number>,first_fix);		
		QL_REQUIRE(v_size(first_fix) == size, "invalid size of first fix vector");
		
		vector_adj(size,0,y_adj);
		QL_REQUIRE(v_size(y_adj) == size, "invalid size of y-shift vector");
		
		
		vector(cashflows_fwd)	dest[0];
		error_info ee = new error_info(true,true);
		
		for(integer i = 0; i < size ; i++){
		
			cashflows_fwd c;
			
			instr_type it = src[i].instr_type(ee);
			
			switch(it){
			case instr_type.DEPOSIT:
			case instr_type.DEPOSIT_INDEX:
				c = CORE_INT.cf_fwd_depo(src[i], y_adj[i]);
				break;
			case instr_type.SWAP_FIXFLT:
				c = CORE_INT.cf_fwd_parswap(src[i], y_adj[i], first_fix[i]);
				break;
			case instr_type.SWAP_FIXOIS:
				c = CORE_INT.cf_fwd_oisswap(src[i],y_adj[i], first_fix[i]);
				break;
			case instr_type.FRA:
				c = CORE_INT.cf_fwd_fra(src[i], y_adj[i]);
				break;
			case instr_type.IR_FUTURE:
				c = CORE_INT.cf_fwd_mmfut(src[i], y_adj[i]);
				break;			
			default:
				QL_FAIL("invalid instrument type in curve (fwd)");
			}
			
			if(require_all_cf) {
				QL_REQUIRE(!null(c),strcat([	"<","invalid cashflow (fwd)" ,"[", src[i].name(),"]", ">"]));
				push_back(dest, c);
			}
			else {
				if(null(c)) {
					string msg = strcat([	"<","invalid cashflow (fwd)" ,"[", src[i].name(),"]", ">"]);	;
					CORE_INT.write_warning_message( msg);
				}
				else {
					push_back(dest, c);
				}
			}
		}

		return cashflows_fwd(dest);
	}

	//-------------------------------------------------------
	// cf_fwd_b1
	//-------------------------------------------------------
	cashflows_fwd cf_fwd_b1( 	vector(instrument) src,
								logical solve_flat,
								fwd_func ff,
								vector(number) option(nullable) first_fix,
								vector(number) option(nullable) first_fix_b1_flat,
								vector(number) option(nullable) 	y_adj,
								logical require_all_cf)
	{		
		integer size = v_size(src);
		vector_adj(size,null<number>,first_fix);		
		QL_REQUIRE(v_size(first_fix) == size, "invalid size of first_fix vector");

		vector_adj(size,null<number>,first_fix_b1_flat);		
		QL_REQUIRE(v_size(first_fix_b1_flat) == size, "invalid size of first_fix_b1_flat vector");
		
		vector_adj(size,0,y_adj);
		QL_REQUIRE(v_size(y_adj) == size, "invalid size of y-shift vector");


		vector(cashflows_fwd)	dest[0];			
		error_info ee = new error_info(true,true);	
		

		for(integer i = 0; i < size ; i++){
			cashflows_fwd c;

			instr_type it = src[i].instr_type(ee);
			
			switch(it){
			case instr_type.DEPOSIT:
			case instr_type.DEPOSIT_INDEX:
				c = CORE_INT.cf_fwd_depo(src[i], y_adj[i]);
				break;
			case instr_type.SWAP_FIXFLT:	
				//FIX_ME roll day not ok, must be parswap with no broken date (src[i].maturity() is an adjusted date)				
				c = CORE_INT.cf_fwd_parswap(src[i],y_adj[i], first_fix[i]);
				break;
			case instr_type.FRA:
				c = CORE_INT.cf_fwd_fra(src[i],y_adj[i]);
				break;
			case instr_type.IR_FUTURE:
				c = CORE_INT.cf_fwd_mmfut(src[i], y_adj[i]);
				break;
			case instr_type.SWAP_FLTFLT:
				c = CORE_INT.cf_fwd_bswap(	src[i], solve_flat, ff, y_adj[i], first_fix_b1_flat[i], first_fix[i]);
				break;
			default:
				QL_FAIL("invalid instrument type in curve (fwd)");
			}

			if(require_all_cf) {
				QL_REQUIRE(!null(c),strcat([	"<","invalid cashflow (b1 fwd)" ,"[", src[i].name(),"]", ">"]));
				push_back(dest, c);
			}
			else {
				if(null(c)) {
					string msg = strcat([	"<","invalid cashflow (b1 fwd)" ,"[", src[i].name(),"]", ">"]);	
					CORE_INT.write_warning_message( msg);
				}
				else
					push_back(dest, c);
			}

		}	
		return cashflows_fwd(dest);
	}

	//-------------------------------------------------------
	// cf_fwd_b2
	//-------------------------------------------------------
	cashflows_fwd cf_fwd_b2(	vector(instrument) src,
								vector(number) option(nullable) first_fix,
								vector(number) option(nullable)	y_adj,
								integer flt_freq,
								day_count_method dc,
								logical require_all_cf)
	{		
	
		integer size = v_size(src);
		vector_adj(size,null<number>,first_fix);		
		QL_REQUIRE(v_size(first_fix) == size, "invalid size of first fix vector");
		
		vector_adj(size,0,y_adj);
		QL_REQUIRE(v_size(y_adj) == size, "invalid size of y-shift vector");
		
		
		vector(cashflows_fwd)	dest[0];
		error_info ee = new error_info(true,true);

							
		for(integer i = 0; i < size ; i++){
			cashflows_fwd c;

			instr_type it = src[i].instr_type(ee);
			
			switch(it){
			case instr_type.DEPOSIT:
			case instr_type.DEPOSIT_INDEX:
				c = CORE_INT.cf_fwd_depo(src[i], y_adj[i]);
				break;				
			case instr_type.SWAP_FIXFLT:
				c = CORE_INT.cf_fwd_parswap(src[i], y_adj[i], first_fix[i]);
				break;
			case instr_type.SWAP_FLTFLT2S:
				c = CORE_INT.cf_fwd_parswap2s(src[i],null<number>, flt_freq,dc,null<bd_convention>,null<calendar>,
											  null<logical>, y_adj[i], first_fix[i]);
				break;
			case instr_type.FRA:
				c = CORE_INT.cf_fwd_fra(src[i], y_adj[i]);
				break;
			case instr_type.IR_FUTURE:
				c = CORE_INT.cf_fwd_mmfut(src[i], y_adj[i]);
				break;			
			default:
				QL_FAIL("invalid instrument type");
			}

			
			if(require_all_cf) {
				QL_REQUIRE(!null(c),strcat([	"<","invalid cashflow (b2 fwd)" ,"[", src[i].name(),"]", ">"]));
				push_back(dest, c);
			}
			else {
				if(null(c)) {
					string msg = strcat([	"<","invalid cashflow (b2 fwd)" ,"[", src[i].name(),"]", ">"]);	;
					CORE_INT.write_warning_message( msg);
				}
				else				
					push_back(dest, c);
			}
		}

		return cashflows_fwd(dest);
	}
}