/*	
	Dual bootstrap utility functions
	Developed by Algorithmica Research, Magnus Nyström
	
	
	Comments: 
	FRA: 	we assume the theoretical convexity correction is small enough to be ignored. 
			Fabio Mercurio derives an approximation for the correction (assuming shifted-lognormal processes) in 
			the paper "LIBOR Market Models with Stochastic Basis", OTC Derivatives and Structured Notes, Bloomberg LP (2010)			
			He argues that under "reasonable" conditions the convexity is small, even for long maturities. 

	ToDo: FRA convexity correction?	

*/
option(null: hard);	

module CORE_INT
{ 
	
	
	
	
	//-------------------------------------------------------
	// cf_fwd_parswap
	// assumption: settlement date == issue date
	// also assume floating index has the same eom, cal, bd as float leg
	//-------------------------------------------------------
	/*cashflows_fwd cf_fwd_parswap(	instrument		inst,									
									number			roll_day,															
									number 			tenor_freq,
									day_count_method dc,
									bd_convention 	bd,
									calendar		cal,
									logical			eom,
									//logical 		fwd_as_pmt,
									number			y_adj,
									number option(nullable) first_fix)
	option (category: 'Yield Curve/003 Curve Building')
	{	
		vector(date)	fixed_dates;
		vector(number)	fixed_amounts;
		
		if(inst.settle_date() != inst.issue_date()) 				
			throw(E_INVALID_ARG, throw_err("invalid par swap (issue date <> settlement date)"));	
			
		number dp = inst.dirty_price();
		if(null(dp) || !fiws.is_equal_par(dp,100)) 	//check that yld == cpn			
			throw(E_INVALID_ARG, throw_err("invalid par swap (not par)"));	
			
		if(y_adj != 0)
			inst = ..create_par_swap(inst, inst.yield() + y_adj, null<string>);
			 			
		inst.cash_flows(fixed_dates, fixed_amounts);
	
		if (null(fixed_amounts) || v_size(fixed_amounts) == 0)
			throw(E_INVALID_ARG, throw_err("invalid par swap (cash flows)"));
	
		fixed_amounts[v_size(fixed_amounts) - 1] -= 100;
		fixed_amounts = -fixed_amounts ;
		//vector(number)	fixed_amounts2 = -i.cash_flow_cpn();
			
		logical ip_idx_stub = false;
		date out_first,out_last_reg;
		number out_roll_day;
		logical is_reg_first,is_reg_last,eom_roll;
		vector(date) float_idx0, float_pmt, float_idx1;
		fiws.swap_flt_dates(inst.issue_date(),null<date>,null<date>, inst.maturity(),roll_day,tenor_freq,bd,cal,eom,bd,cal,eom,
							ip_idx_stub,out_first, out_last_reg,out_roll_day,is_reg_first,is_reg_last,eom_roll,
							float_idx0,float_pmt,float_idx1);
	
	
		vector(number) float_dt = period(float_idx0, float_pmt, dc);
	
		if(!null(first_fix)){	
			if(v_size(float_idx0) == 1)
				return null<cashflows_fwd>;
			number cf = period (inst.issue_date(), float_pmt[0], dc) * first_fix; 
			push_back(fixed_dates, float_pmt[0]);
			push_back(fixed_amounts, cf);		
			float_idx0 = v_exclude(float_idx0, 0);		
			float_pmt = v_exclude(float_pmt, 0);
			float_dt = v_exclude(float_dt, 0);				
		}
		cashflows_fwd c = cashflows_fwd(inst.trade_date(), fixed_dates, fixed_amounts, float_idx0, float_pmt, float_dt);
		QL_REQUIRE(!null(c),"invalid cashflow_fwd", inst.name(), true);
		return c;
	}
	*/
	
	//-------------------------------------------------------
	// adj_ip_spr_basis2
	//-------------------------------------------------------	
	number adj_ip_spr_basis2(	date 		d,
								disc_func	df,
								instrument 	swp,
								instrument 	bswp,
								number 		rate_s,
								number 		spr) //spr always of type fixleg in basis2swap
	{
		integer freq_s 			= integer(swp.coupon_freq()); 
		integer freq_b 			= integer(bswp.coupon_freq());
		day_count_method dc_s 	= swp.accr_dc_method();
		day_count_method dc_b 	= bswp.accr_dc_method();
		
					
		if((freq_s != freq_b || dc_s != dc_b)){
			date issue_date	= bswp.issue_date();
			date mat		= bswp.maturity(); 
			
			vector(date) dates_in, dates_out;
			logical is_reg_first,is_reg_last,eom_roll;
			integer out_roll_day;
			date out_first,out_last_reg;
			..swap_pmt_dates(issue_date,null<date>,null<date>, mat, day(issue_date), freq_s, swp.pmt_bus_day(),swp.calendar(),swp.eom(),
							out_first,out_last_reg,out_roll_day,is_reg_first,is_reg_last, eom_roll,dates_in);
	
			..swap_pmt_dates(issue_date,null<date>,null<date>, mat, day(issue_date), freq_b, bswp.pmt_bus_day(),bswp.calendar(),bswp.eom(),
							out_first,out_last_reg,out_roll_day,is_reg_first,is_reg_last, eom_roll,dates_out);
						
			number rate_in_adj 	= swap_fix_rate_conv(d, issue_date,df,rate_s, dates_in,dc_s, dates_out, dc_b) ;//conv to basis
			return rate_in_adj 	+ spr;			
		}
		else 				
			return rate_s + spr;
	}
	//-------------------------------------------------------
	// hit_vector
	//-------------------------------------------------------
	void hit_vector(vector(number) base, vector(number) ref, number buf, out vector(logical) option(nullable) hit)
	{
		resize(hit,v_size(base));
		hit = false;
		integer k_start = 0;
		for(integer i = 0; i < v_size(base); i++){
			for(integer k = k_start; k < v_size(ref); k++){
				if(ref[k] - base[i] > buf)  
					break;
				else if(abs(ref[k] - base[i]) <= buf){
					hit[i] = true;
					k_start = k;
					break;
				}
				 
			}
		}
	}

	//-------------------------------------------------------
	// create_swap_from_basis
	//-------------------------------------------------------
	instr_def create_swap_def_from_basis(	instrument 	bswp,
											logical 	add_basis)
	{		
		string class_name ;
		integer cfreq_flt_b;
		if(add_basis){
			class_name = bswp.instr_def().fltflt_swap_class_flat() ;
			cfreq_flt_b = bswp.instr_def().swap_coupon_freq_leg2();
		}
		else {
			class_name = bswp.instr_def().fltflt_swap_class_sprd() ;
			cfreq_flt_b = bswp.instr_def().swap_coupon_freq_leg1();
		}

		string err_name = bswp.name();
		QL_REQUIRE(!null(class_name),'invalid swap class name', err_name, true);
	
		swap_fixflt class_swap = CORE_INT_FIXFLT.create_class_fixflt(class_name);
		
		QL_REQUIRE(!null(class_swap),'invalid class swap', err_name, true);
		
		error_info error = new error_info(true,false);
		integer cfreq_fix, cfreq_flt;
		class_swap.coupon_freq(cfreq_fix, cfreq_flt, error);
		QL_REQUIRE(!error.is_error(), error.message());
					
		instr_def def;
		logical cfreq_conv;
		if(cfreq_flt != cfreq_flt_b) {
			def 		= class_swap.instr_def();
			string n 	= strcat(def.name(),"-cfreq-conv");
			def 		= instr_def(def,n,"1Y",100,null<number>,cfreq_flt_b,null<number>,true,error);
			QL_REQUIRE(!error.is_error(), error.message());
			cfreq_conv 	= true;
		}
		else {
			def = class_swap.instr_def();
			cfreq_conv = false;
		}
		return def;
	

	}

	//-------------------------------------------------------
	// create_swap_name_from_basis
	//-------------------------------------------------------
	string create_swap_name_from_basis(instr_def 	fixflt_def,
									   instrument 	inst)
	{		
		integer rf = fixflt_def.swap_reset_freq_leg2();
		string t;
		if(rf == 0 || rf == -1)
			t = "X";
		else
			t = strcat(string(12/rf),"M");
		string name = strcat(inst.name(),strcat("-implied fix-flt-",t));

		//if(cfreq_conv)
				//name = strcat(name,"(cfreq-conv)");
		return name;
	}
	//-------------------------------------------------------
	// create_swap_from_basis
	//-------------------------------------------------------
	swap_fixflt create_swap_from_basis(instr_def 	fixflt_def,
									   instrument 	inst,
									   number 		r_sw,
									   number option(nullable) first_fix,
									   string 		name)
	{
		try {			
			error_info error = new error_info(true,false);								
			swap_fixflt swp = swap_fixflt_plain(fixflt_def, name, inst.trade_date(),inst.settle_date(),
												null<string>,inst.maturity(),null<string>,r_sw ,
												100,true,null<disc_func>,null<fwd_func>,first_fix, -1,error);
			return swp;
		}
		catch {
			return null<swap_fixflt>;
		}
	}
	//-------------------------------------------------------
	// create_swap_from_basis_interp
	//-------------------------------------------------------
	/*swap_fixflt create_swap_from_basis_interp(	instr_def 	fixflt_def,
												number 		r_sw,
												string name)
	{
		try {			
			error_info error = new error_info(true,false);
			
			number first_fix 	= null<number>;
			date d 				= inst_basis.trade_date();
						
			swap_fixflt swp = swap_fixflt_plain(fixflt_def, name, d,inst_basis.settle_date(),inst_basis.maturity(),r_sw ,									
												100,true,null<disc_func>,null<fwd_func>,first_fix, error);
			return swp;
		}
		catch {
			return null<swap_fixflt>;
		}
	}*/
	//-------------------------------------------------------
	// calc_swap_from_basis2
	//-------------------------------------------------------
	curve calc_swap_from_basis2(date 		d,
								disc_func	df,
								curve 		swap_crv,
								curve 		basis_crv,
								logical 	add_basis,
								number option(nullable) first_fix,
								string option(nullable)	name,
								logical		x_pol_basis)
	{
		string err_name;
		QL_REQUIRE(CORE_INT.check_instr_type(swap_crv, instr_type.SWAP_FIXFLT, err_name),'invalid swap curve (mixed curve)', err_name, true);
	
		vector(instrument) inst_swap 	= swap_crv.instruments();
		integer size_swap 				= v_size(inst_swap );
		QL_REQUIRE(size_swap > 0,'invalid swap curve');
	
		vector(instrument) inst_basis 	= basis_crv.instruments();
		integer size_basis 				= v_size(inst_basis );
		QL_REQUIRE(size_basis > 0,'invalid basis curve');
		
		instrument swp 					= inst_swap[0];
		instrument bswp 				= inst_basis[0];
				
		vector(number) ttm_swap 		= inst_swap.maturity() - d; 
		vector(number) y_swap 			= inst_swap.yield();	
		number ttm_max_swap 			= ttm_swap[size_swap-1]; 	
		number ttm_min_swap 			= ttm_swap[0]; 
			
		vector(number) ttm_basis		= inst_basis.maturity() - d; 
		vector(number) spr_basis		= add_basis ? inst_basis.quote()/10000 : -inst_basis.quote()/10000; 	
		number ttm_max_basis 			= ttm_basis[size_basis-1]; 
		number ttm_min_basis 			= ttm_basis[0]; 
				
		vector(instrument) sw[0];
	
		/*string i_c_n ;
		integer cfreq_flt_b;
		if(add_basis){
			i_c_n = bswp.instr_def().fltflt_swap_class_flat() ;
			cfreq_flt_b = bswp.instr_def().swap_coupon_freq_leg2();
		}
		else {
			i_c_n = bswp.instr_def().fltflt_swap_class_sprd() ;
			cfreq_flt_b = bswp.instr_def().swap_coupon_freq_leg1();
		}
		
		QL_REQUIRE(!null(i_c_n),'invalid swap class name', err_name, true);

		logical cfreq_conv;*/	
					
		instr_def def 			= create_swap_def_from_basis(bswp, add_basis);			
		interpolation swap_ip 	= interpolation(ip_linear(0,0), ttm_swap, y_swap);
		
		for(integer i = 0; i < size_basis; i++){
			if(ttm_basis[i] > ttm_max_swap + 5)	
				break;
	
			if(ttm_basis[i] < ttm_min_swap - 5)	
				continue;
	
			number y_swap_ip = swap_ip.eval(ttm_basis[i]);	 						
			if(null(y_swap_ip)){
				CORE_INT.write_warning_message('invalid interpolation of swap yield (at least one instrument)');
				continue;
			}
			number r_sw = adj_ip_spr_basis2(d, df, swp, bswp, y_swap_ip, spr_basis[i]);
			if(null(r_sw)){
				CORE_INT.write_warning_message('invalid conversion of swap yield (at least one instrument)');
				continue;
			}
			/*string name;
			try {
				error_info ee = new error_info(true,false);
				instr_def i_d 	= instr_def(i_c_n , ee);
				integer rf = i_d.swap_reset_freq_leg2();
				QL_REQUIRE(rf != 0 && rf != -1,"");
				string t = strcat(string(12/rf),"M");
				name = strcat(inst_basis[i].name(),strcat("-implied fix-flt-",t));
			}
			catch {
				name = strcat(inst_basis[i].name(),"-implied fix-flt");
			}
			
			instrument ii = CORE_INT.create_fixflt_par(i_c_n, name,d,inst_basis[i].settle_date(),
													inst_basis[i].maturity(),r_sw);*/
			string name 	= create_swap_name_from_basis(def,inst_basis[i]);
			swap_fixflt ii 	= create_swap_from_basis(def,inst_basis[i], r_sw,first_fix,name);
				
			if(null(ii)){
				CORE_INT.write_warning_message('failed to create interpolated basis-2 swap (at least one instrument)');
				continue;
			}
			push_back(sw, ii);
							
		}
		
		if(!x_pol_basis)
			return curve(sw, name);	
		
		number cut_off = 15;
		vector(logical) hit;
		hit_vector(ttm_swap, ttm_basis, cut_off, hit);
		
		interpolation basis_ip = interpolation(ip_linear(0,0), ttm_basis,spr_basis);
	
		for(integer i = 0; i < size_swap; i++){
			if(hit[i]) continue;
	
			number b_sprd_ip = basis_ip.eval(ttm_swap[i]);	 						
			if(null(b_sprd_ip)){
				CORE_INT.write_warning_message('invalid interpolation of basis-2 spread (at least one instrument)');
				continue;
			}
			number r_sw = adj_ip_spr_basis2(d, df, swp, bswp, y_swap[i], b_sprd_ip);
			if(null(r_sw)){
				CORE_INT.write_warning_message('invalid conversion of swap yield (at least one instrument)');
				continue;
			}
	
			string sw_name;
			if(ttm_swap[i] < ttm_min_basis || ttm_swap[i] > ttm_max_basis)
				sw_name = strcat(inst_swap[i].name(),"-xpol");
			else
				sw_name = strcat(inst_swap[i].name(),"-ipol");
			
			//instrument ii = CORE_INT.create_fixflt_par(	i_c_n, sw_name,d,inst_swap[i].settle_date(),
				//									inst_swap[i].maturity(),r_sw);
			
			swap_fixflt ii = create_swap_from_basis(def,inst_swap[i], r_sw,first_fix,sw_name);
			
			if(null(ii)){
				CORE_INT.write_warning_message('failed to create interpolated basis-2 swap (at least one instrument)');
				continue;
			}
			push_back(sw, ii);
			
		}
	
		return curve(sw, name);	
	}
	
	//-------------------------------------------------------
	// create_adj_basis1_curve
	//-------------------------------------------------------
	curve create_adj_basis1_curve(	date 		d,						
									curve 		basis_crv,
									vector(date) option(nullable) swap_mat,						
									string option(nullable)	name)
	
	{		
		if(null(swap_mat))
			return basis_crv;
		integer size_mat 	= v_size(swap_mat );
		if(size_mat == 0)
			return basis_crv;
	
		vector(instrument) inst_basis 	= basis_crv.instruments();		
		instrument bswp 				= inst_basis[0];
		integer size_basis 				= v_size(inst_basis );
		QL_REQUIRE(size_basis > 0,'invalid basis curve');
	
		vector(number) ttm_swap 		= swap_mat - d;  
		vector(number) ttm_basis		= inst_basis.maturity() - d; 
		vector(number) spr_basis		= inst_basis.quote(); 	
		number ttm_max_basis 			= ttm_basis[size_basis-1]; 
		number ttm_min_basis 			= ttm_basis[0]; 			
		vector(instrument) sw 			= basis_crv.instruments();
		string i_c_n 					= bswp.class_name();	
				
		number cut_off = 15;
		vector(logical) hit;
		hit_vector(ttm_swap, ttm_basis, cut_off, hit);
		
		interpolation basis_ip = interpolation(ip_linear(0,0), ttm_basis, spr_basis);
	
		for(integer i = 0; i < size_mat; i++){
			if(hit[i])
				continue;
	
			number b_sprd_ip = basis_ip.eval(ttm_swap[i]);	 						
			if(null(b_sprd_ip)){
				CORE_INT.write_warning_message('invalid interpolation of basis-1 spread (at least one instrument)');
				continue;
			}		
	
			string sw_name;
			if(ttm_swap[i] < ttm_min_basis || ttm_swap[i] > ttm_max_basis)
				sw_name = "basisswap-xpol";
			else
				sw_name = "basisswap-ipol";
			
			instrument ii = CORE_INT.create_par_fltflt(i_c_n, sw_name,d,bswp.settle_date(),swap_mat[i],b_sprd_ip/10000);
			if(null(ii)){
				CORE_INT.write_warning_message('failed to create interpolated basis-1 swap (at least one instrument)');
				continue;
			}
			push_back(sw, ii);
			
		}
	
		return curve(sw, name);	
	}
	
	//-------------------------------------------------------
	// calc_swap_from_basis1
	//-------------------------------------------------------
	/*curve	calc_swap_from_basis1(	date 		d,
									disc_func	df,
									string 		irs_class_name,
									curve 		basis_crv,
									fwd_func 	f_flt_idx,
									integer		flt_reset_days,
									number option(nullable) first_fix,
									string option(nullable)	name)
	{ 
		vector(instrument) vi = basis_crv.instruments();
		integer size = v_size(vi);	
			
		swap_fixflt class_swap = CORE_INT.create_class_fixflt_par(irs_class_name);
		if(null(class_swap)){
			CORE_INT.write_warning_message('invalid swap (creating from class_name)');	
			return null<curve>;
		}
		calendar cal 		= class_swap.calendar();
		date_code dc 		= date_code(class_swap.settle_code());
		date settle 		= dc.apply_fwd(d, cal) ;
		bd_convention bd	= class_swap.pmt_bus_day_d(BD_MOD_FOLLOWING);

		day_count_method fix_dc,flt_dc;
		class_swap.dc_method(fix_dc,flt_dc);
	
		integer fix_freq,flt_freq;	
		class_swap.coupon_freq(fix_freq,flt_freq);
				
		logical eom 		= class_swap.eom(); 		
	
		vector(instrument) sw[size];	
		for(integer i = 0; i < size ; i++){
			number y 	= CORE_INT.par_fixflt_swap_rate(	d,cal,settle,bd,settle,vi[i].maturity(),eom,flt_freq,fix_freq,
											flt_dc,fix_dc,df,f_flt_idx, flt_reset_days, first_fix);	
			sw[i] 		= CORE_INT.create_fixflt_par(irs_class_name,strcat([vi[i].name(),"-irs"]),d,settle,vi[i].maturity(),y);
			if(null(sw[i])){
				CORE_INT.write_warning_message('failed to create swap from basis1-swap (at least one instrument)');
				continue;
			}		
		}			
		return curve(sw,name);
	}*/
}