/*	
	Cross currency bootstrap utility functions
	Developed by Algorithmica Research, Magnus Nyström
	
	
	Comments: 
	

*/
option(null: hard);	

module CORE_INT
{
	//-------------------------------------------------------
	// cf_currbswap_ois 
	// (crosscurr basis swap quoted as 1 swap)
	// ex USDSEK fx=6, USD = flat, freq =4, i = SEK
	// Todo: different bd,cal,eom for the index of the legs not handled 
	//-------------------------------------------------------
	cashflows cf_currbswap_fltois(	date 			trade,
								date			issue_date,//== settle_date,
								date			maturity,
								integer			freq_sprd,
								integer			freq_flat, 
								day_count_method dc_sprd,
								day_count_method dc_flat,
								vector(date) 	idx0_sprd,
								vector(date) 	idx1_sprd,
								vector(date) 	pmt_sprd,
								vector(date) 	prin_pmt_sprd,
								vector(date) 	idx0_flat,
								vector(date) 	idx1_flat,
								vector(date) 	pmt_flat,
								vector(date) 	prin_pmt_flat,
								logical			solve_flat,//true-->solve for USD fximplied disc_func
								fwd_func		f_fwd_sprd,//SEK
								number option(nullable)	first_fix_sprd,	
								disc_func		df_fwd_flat,//USD
								disc_func		df_disc,//solve_flat=true --> df_disc = SEK disc
								number			spread_sprd,//spread on the sprd leg
								number option(nullable) spread_flat = null<number> )
	{
		number nom 					= 100;   				
		integer n_sprd = v_size(idx0_sprd);
		integer n_flat = v_size(idx0_flat);
		integer np_sprd = v_size(prin_pmt_sprd);
		integer np_flat = v_size(prin_pmt_flat);
		
		if(null(spread_flat))
			spread_flat = 0;

		vector(number)	fixed_amounts[0];
		vector(date)	fixed_dates[0];		
	    
		push_back(fixed_dates, issue_date); 
		push_back(fixed_amounts, 0);
	
		vector(number) per_pay_sprd = period (idx0_sprd, pmt_sprd, dc_sprd);
		vector(number) per_pay_flat	= period (idx0_flat, pmt_flat, dc_flat);
		vector(number) per_idx0_sprd = (idx0_sprd - trade) / 365;
		vector(number) r_fwd_sprd 	= f_fwd_sprd.fwd(per_idx0_sprd);
		if(!null(first_fix_sprd))
			r_fwd_sprd[0] = first_fix_sprd;
		
		vector(number) r_fwd_flat 	= df_fwd_flat.simple_rate(trade,idx0_flat,idx1_flat,dc_flat);	
		vector(number) cf_sprd 		= (r_fwd_sprd + spread_sprd) .* per_pay_sprd;
		vector(number) cf_flat 		= (r_fwd_flat + spread_flat) .* per_pay_flat ;
				
		//------------------------
		//-------fix for non-convergence in bootstrap
		//-------(principal pmt different from last ir pmt)
		//------------------------
		if(prin_pmt_flat[np_flat-1] != pmt_flat[n_flat-1]){
			pmt_flat[n_flat-1] = prin_pmt_flat[np_flat-1];
		}
		if(prin_pmt_sprd[np_sprd-1] != pmt_sprd[n_sprd-1]){
			pmt_sprd[n_sprd-1] = prin_pmt_sprd[np_sprd-1];
		}
		//------------------------
		//------------------------

		if(solve_flat) {//USD			
			number dfs			= df_disc.disc((issue_date - trade) / 365);
			vector(number) df	= df_disc.disc((pmt_sprd - trade) / 365)/dfs;	
			number pv_sprd 		= cf_sprd * df;
		    number prin_df		= df_disc.disc((prin_pmt_sprd[np_sprd-1] - trade) / 365)/dfs;
			pv_sprd 			+= prin_df;					
				
			for (integer i = 0 ; i < n_flat; i++) {									
				push_back(fixed_dates, pmt_flat[i]); 
				push_back(fixed_amounts, cf_flat[i]);
		    }
			
			if(prin_pmt_flat[np_flat-1] != pmt_flat[n_flat-1]){//with the non-convergence fix above we will not enter here
				push_back(fixed_dates, prin_pmt_flat[np_flat-1]); 
				push_back(fixed_amounts, 1);
				sort(fixed_amounts,fixed_dates);
				sort(fixed_dates,fixed_dates);
			}
			else {
				fixed_amounts[n_flat] += 1;
				fixed_amounts[0] = -pv_sprd;
			}
		}
		else { //SEK			
			number dfs			= df_disc.disc((issue_date - trade) / 365);//df_disc is df_disc_flat
			vector(number) df	= df_disc.disc((pmt_flat - trade) / 365)/dfs;																 							
			number pv_flat 		= cf_flat * df;			
			//pv 					+= df[n_flat-1] ;				
			number prin_df		= df_disc.disc((prin_pmt_flat[np_flat-1] - trade) / 365)/dfs;
			pv_flat 			+= prin_df;
			 
			for (integer i = 0 ; i < n_sprd; i++) {																	
				push_back(fixed_dates, pmt_sprd[i]); 
				push_back(fixed_amounts, cf_sprd[i]);										
		    }
			
			if(prin_pmt_sprd[np_sprd-1] != pmt_sprd[n_sprd-1]){//with the non-convergence fix above we will not enter here but keep for now
				push_back(fixed_dates, prin_pmt_sprd[np_sprd-1]); 
				push_back(fixed_amounts, 1);
				sort(fixed_amounts,fixed_dates);
				sort(fixed_dates,fixed_dates);
			}
			else {
				fixed_amounts[n_sprd] += 1;
				fixed_amounts[0] = -pv_flat;
			}
		}
		fixed_amounts = fixed_amounts * nom;
		return cashflows(trade, fixed_dates, fixed_amounts);
	}
	
	//-------------------------------------------------------
	// cf_currbswap_fltois
	//-------------------------------------------------------
	cashflows cf_currbswap_fltois(	instrument		inst,
								logical			solve_flat,																	
								fwd_func		f_fwd_sprd,
								number option(nullable)	first_fix_sprd,	
								disc_func		df_fwd_flat,
								disc_func		df_disc,
								number			q_adj)
	{	    
		number spr = q_adj != 0 ? (inst.quote() + q_adj)/10000 : inst.quote()/10000;

		QL_REQUIRE(inst.is_swap_fltois(), "invalid instrument type");
		error_info ee = error_info(true,false);
		error_info ee2 = error_info(true,false);
		swap_fltois sw = swap_fltois(inst,ee);
		QL_FAIL_COND(null(sw), ee.message(), inst.name(ee2), true);		

		integer cf1,cf2;
		sw.coupon_freq(cf1,cf2,ee);

		day_count_method dc1,dc2;
		sw.dc_method(dc1,dc2,ee);

		vector(date) 	accr_start_dates_leg1, accr_end_dates_leg1, accr_start_dates_leg2,	 accr_end_dates_leg2;
		vector(date) 	pmt_dates1,pmt_dates2,p_pmt_dates1,p_pmt_dates2;
		vector(number) 	accr_period_leg1,accr_period_leg2;

		sw.accr_dates(	true, accr_start_dates_leg1, accr_end_dates_leg1, accr_period_leg1, accr_start_dates_leg2,									
						accr_end_dates_leg2, accr_period_leg2, ee);
		sw.cash_flow_dates(true,  pmt_dates1,	 pmt_dates2,ee) ;
		sw.prin_cash_flow_dates(true, p_pmt_dates1, p_pmt_dates2,ee) ;
		return cf_currbswap_fltois(sw.trade_date(),sw.issue_date(),sw.maturity(), cf1,cf2, dc1,dc2,
								accr_start_dates_leg1,accr_end_dates_leg1,pmt_dates1,p_pmt_dates1,
								accr_start_dates_leg2,accr_end_dates_leg2,pmt_dates2,p_pmt_dates2,
								solve_flat,f_fwd_sprd,first_fix_sprd,df_fwd_flat,df_disc,spr);
	
	}
	
	//-------------------------------------------------------
	// cf_currbswap_ois 
	// (crosscurr basis swap quoted as 1 swap)
	// ex USDSEK fx=6, USD = flat, freq =4, i = SEK
	// Todo: different bd,cal,eom for the index of the legs not handled 
	//-------------------------------------------------------
	cashflows cf_currbswap_ois(	date 			trade,
								date			issue_date,//== settle_date,
								date			maturity,
								integer			freq_sprd,
								integer			freq_flat, 
								day_count_method dc_sprd,
								day_count_method dc_flat,
								vector(date) 	idx0_sprd,
								vector(date) 	idx1_sprd,
								vector(date) 	pmt_sprd,
								vector(date) 	prin_pmt_sprd,
								vector(date) 	idx0_flat,
								vector(date) 	idx1_flat,
								vector(date) 	pmt_flat,
								vector(date) 	prin_pmt_flat,
								logical			solve_flat,//true-->solve for USD fximplied disc_func
								disc_func		df_fwd_sprd,//SEK
								disc_func		df_fwd_flat,//USD
								disc_func		df_disc,//solve_flat=true --> df_disc = SEK disc
								number			spread_sprd,//spread on the sprd leg
								number option(nullable) spread_flat = null<number> )
	{
		number nom 					= 100;   				
		integer n_sprd = v_size(idx0_sprd);
		integer n_flat = v_size(idx0_flat);
		integer np_sprd = v_size(prin_pmt_sprd);
		integer np_flat = v_size(prin_pmt_flat);
		
		if(null(spread_flat))
			spread_flat = 0;

		vector(number)	fixed_amounts[0];
		vector(date)	fixed_dates[0];		
	    
		push_back(fixed_dates, issue_date); 
		push_back(fixed_amounts, 0);
	
		vector(number) per_pay_sprd = period (idx0_sprd, pmt_sprd, dc_sprd);
		vector(number) per_pay_flat	= period (idx0_flat, pmt_flat, dc_flat);
		vector(number) r_fwd_sprd 	= df_fwd_sprd.simple_rate(trade,idx0_sprd,idx1_sprd,dc_sprd); 	
		vector(number) r_fwd_flat 	= df_fwd_flat.simple_rate(trade,idx0_flat,idx1_flat,dc_flat);	
		vector(number) cf_sprd 		= (r_fwd_sprd + spread_sprd) .* per_pay_sprd;
		vector(number) cf_flat 		= (r_fwd_flat + spread_flat) .* per_pay_flat ;

		//------------------------
		//-------fix for non-convergence in bootstrap
		//-------(principal pmt different from last ir pmt)
		//------------------------
		if(prin_pmt_flat[np_flat-1] != pmt_flat[n_flat-1]){
			pmt_flat[n_flat-1] = prin_pmt_flat[np_flat-1];
		}
		if(prin_pmt_sprd[np_sprd-1] != pmt_sprd[n_sprd-1]){
			pmt_sprd[n_sprd-1] = prin_pmt_sprd[np_sprd-1];
		}
		//------------------------
		//------------------------

		if(solve_flat) {//USD			
			number dfs			= df_disc.disc((issue_date - trade) / 365);
			vector(number) df	= df_disc.disc((pmt_sprd - trade) / 365)/dfs;	
			number pv_sprd 		= cf_sprd * df;
		    number prin_df		= df_disc.disc((prin_pmt_sprd[np_sprd-1] - trade) / 365)/dfs;
			pv_sprd 			+= prin_df;					
				
			for (integer i = 0 ; i < n_flat; i++) {									
				push_back(fixed_dates, pmt_flat[i]); 
				push_back(fixed_amounts, cf_flat[i]);
		    }
			
			if(prin_pmt_flat[np_flat-1] != pmt_flat[n_flat-1]){//with the non-convergence fix above we will not enter here
				push_back(fixed_dates, prin_pmt_flat[np_flat-1]); 
				push_back(fixed_amounts, 1);
				sort(fixed_amounts,fixed_dates);
				sort(fixed_dates,fixed_dates);
			}
			else {
				fixed_amounts[n_flat] += 1;
				fixed_amounts[0] = -pv_sprd;
			}
		}
		else { //SEK			
			number dfs			= df_disc.disc((issue_date - trade) / 365);//df_disc is df_disc_flat
			vector(number) df	= df_disc.disc((pmt_flat - trade) / 365)/dfs;																 							
			number pv_flat 		= cf_flat * df;			
			//pv 					+= df[n_flat-1] ;				
			number prin_df		= df_disc.disc((prin_pmt_flat[np_flat-1] - trade) / 365)/dfs;
			pv_flat 			+= prin_df;
			 
			for (integer i = 0 ; i < n_sprd; i++) {																	
				push_back(fixed_dates, pmt_sprd[i]); 
				push_back(fixed_amounts, cf_sprd[i]);										
		    }
			
			if(prin_pmt_sprd[np_sprd-1] != pmt_sprd[n_sprd-1]){//with the non-convergence fix above we will not enter here but keep for now
				push_back(fixed_dates, prin_pmt_sprd[np_sprd-1]); 
				push_back(fixed_amounts, 1);
				sort(fixed_amounts,fixed_dates);
				sort(fixed_dates,fixed_dates);
			}
			else {
				fixed_amounts[n_sprd] += 1;
				fixed_amounts[0] = -pv_flat;
			}
		}
		fixed_amounts = fixed_amounts * nom;
		return cashflows(trade, fixed_dates, fixed_amounts);
	}

	//-------------------------------------------------------
	// cf_currbswap_ois
	//-------------------------------------------------------
	cashflows cf_currbswap_ois(	instrument		inst,
								logical			solve_flat,																	
								disc_func		df_fwd_sprd,
								disc_func		df_fwd_flat,
								disc_func		df_disc,
								number			q_adj)
	{	    
		number spr = q_adj != 0 ? (inst.quote() + q_adj)/10000 : inst.quote()/10000;

		QL_REQUIRE(inst.is_swap_oisois(), "invalid instrument type");
		error_info ee = error_info(true,false);
		error_info ee2 = error_info(true,false);
		swap_oisois sw = swap_oisois(inst,ee);
		QL_FAIL_COND(null(sw), ee.message(), inst.name(ee2), true);		

		integer cf1,cf2;
		sw.coupon_freq(cf1,cf2,ee);

		day_count_method dc1,dc2;
		sw.dc_method(dc1,dc2,ee);

		vector(date) 	accr_start_dates_leg1, accr_end_dates_leg1, accr_start_dates_leg2,	 accr_end_dates_leg2;
		vector(date) 	pmt_dates1,pmt_dates2,p_pmt_dates1,p_pmt_dates2;
		vector(number) 	accr_period_leg1,accr_period_leg2;

		sw.accr_dates(	true, accr_start_dates_leg1, accr_end_dates_leg1, accr_period_leg1, accr_start_dates_leg2,									
						accr_end_dates_leg2, accr_period_leg2, ee);
		sw.cash_flow_dates(true,  pmt_dates1,	 pmt_dates2,ee) ;
		sw.prin_cash_flow_dates(true, p_pmt_dates1, p_pmt_dates2,ee) ;
		return cf_currbswap_ois(sw.trade_date(),sw.issue_date(),sw.maturity(), cf1,cf2, dc1,dc2,
								accr_start_dates_leg1,accr_end_dates_leg1,pmt_dates1,p_pmt_dates1,
								accr_start_dates_leg2,accr_end_dates_leg2,pmt_dates2,p_pmt_dates2,
								solve_flat,df_fwd_sprd,df_fwd_flat,df_disc,spr);
	
	}
	//-------------------------------------------------------
	// cf_currbswap_ois_mtm 
	// (crosscurr basis swap quoted as 1 swap)
	// ex USDSEK fx=6, USD = flat, freq =4, i = SEK
	// Todo: different bd,cal,eom for the index of the legs not handled 
	//-------------------------------------------------------
	cashflows_q cf_currbswap_ois_mtm(	date 			trade,
										date			issue_date,//== settle_date,
										date			maturity,									
										integer 		freq_sprd,
										integer 		freq_flat, 
										day_count_method dc_sprd,
										day_count_method dc_flat,
										vector(date) 	idx0_sprd,
										vector(date) 	idx1_sprd,
										vector(date) 	pmt_sprd,
										vector(date) 	prin_pmt_sprd,
										vector(date) 	idx0_flat,
										vector(date) 	idx1_flat,
										vector(date) 	pmt_flat,
										vector(date) 	prin_pmt_flat,											 	
										logical			solve_flat,//true-->solve for USD fximplied disc_func
										disc_func		df_fwd_sprd,//SEK
										disc_func		df_fwd_flat,//USD
										disc_func		df_disc,
										number			spread_sprd,//spread on the sprd leg
										number option(nullable) spread_flat = null<number>)									
	{	 
		number fx_rate_settle 	= 1;
	   	number nom_sprd 		= 100;
		number nom_base 		= 100;						
		integer n_sprd 			= v_size(idx0_sprd);
		integer n_flat 			= v_size(idx0_flat);
		integer np_sprd 		= v_size(prin_pmt_sprd);
		integer np_flat 		= v_size(prin_pmt_flat);
		
		if(null(spread_flat))
			spread_flat = 0;
	
		vector(number)	fixed_amounts[0];
		vector(date)	fixed_dates[0];		
	    vector(date)	fixed_dates_q[0];
	
		push_back(fixed_dates, trade) ;
		push_back(fixed_dates_q, null<date>) ;
		push_back(fixed_amounts, 0) ;	

		vector(number) per_pay_sprd = period (idx0_sprd, pmt_sprd, dc_sprd);
		vector(number) per_pay_flat	= period (idx0_flat, pmt_flat, dc_flat);
		vector(number) r_fwd_sprd 	= df_fwd_sprd.simple_rate(trade,idx0_sprd,idx1_sprd,dc_sprd); 	
		vector(number) r_fwd_flat 	= df_fwd_flat.simple_rate(trade,idx0_flat,idx1_flat,dc_flat);	
		vector(number) cf_sprd 		= (r_fwd_sprd + spread_sprd) .* per_pay_sprd;
		vector(number) cf_flat 		= (r_fwd_flat + spread_flat) .* per_pay_flat ;

		//------------------------
		//-------fix for non-convergence in bootstrap
		//-------(principal pmt different from last ir pmt)
		//------------------------
		if(prin_pmt_flat[np_flat-1] != pmt_flat[n_flat-1]){
			pmt_flat[n_flat-1] = prin_pmt_flat[np_flat-1];
		}
		if(prin_pmt_sprd[np_sprd-1] != pmt_sprd[n_sprd-1]){
			pmt_sprd[n_sprd-1] = prin_pmt_sprd[np_sprd-1];
		}

		
		if(solve_flat) {//USD
			
			number dfs			= df_disc.disc((issue_date - trade) / 365.0);
			number rel_df		= 1.0/dfs;
			vector(number) df	= df_disc.disc((pmt_sprd - trade) / 365.0)*rel_df;	
			number pv 			= -(cf_sprd * df) ;	
			pv 					= pv + 1 - df[n_sprd-1];		    							
			pv					= pv * nom_sprd - nom_base;
	
			push_back(fixed_dates_q, issue_date);
	
			for (integer i = 0 ; i < n_flat; i++) {
		
				push_back(fixed_dates, pmt_flat[i]); 
				if(i>0) 
					push_back(fixed_dates_q, pmt_flat[i-1]);			
	
				number p_df_imp_f	= i==0 ? df_disc.disc((issue_date-trade)/365.0) : df_disc.disc((pmt_flat[i-1]-trade)/365.0);
		    	number df_imp_f		= df_disc.disc((pmt_flat[i]-trade)/365.0);	
				number k 			= nom_sprd*df_imp_f*rel_df/fx_rate_settle ;
				number p_k 			= i==0? nom_base : nom_sprd*p_df_imp_f*rel_df/fx_rate_settle ;
				push_back(fixed_amounts, p_k+cf_flat[i]* p_k) ;					
										
				if(i < n_flat-1) 			
					pv -= k ;
		    }
	
			fixed_amounts[0] 	= pv;
	
		}
		else { //SEK
			// Add constant flow
			//number dfs			= df_fwd_flat.disc((issue_date - trade) / 365);
			//number rel_df		= 1/dfs;
			number k = -nom_base * df_disc.disc((issue_date-trade)/365.0) / df_disc.disc((issue_date-trade)/365.0) ;
	
			for(integer i = 0 ;i < v_size(pmt_flat) ; i++) {  
				number p_df_base_f	= df_disc.disc(((i == 0 ? issue_date : pmt_flat[i-1])-trade)/365.0);
				number df_base_f	= df_disc.disc((pmt_flat[i]-trade)/365.0);
	
				number p_k ;
	
				if(i == 0) {
					p_k = nom_base * df_base_f/p_df_base_f ;
				}
				else
					p_k = df_base_f/p_df_base_f*nom_sprd/fx_rate_settle ;
	
				number koff = p_k + (r_fwd_flat[i] + spread_flat)*per_pay_flat[i]* p_k ;
	
				if(i > 0) {
					koff -= nom_sprd/fx_rate_settle ;
					koff -= (r_fwd_sprd[i-1] + spread_sprd)*per_pay_sprd[i-1]*nom_sprd ;
				}
				else
					koff += nom_sprd ;
			
				date _fixed_dates = (i == 0 ? issue_date : pmt_flat[i-1]) ;
	
				if(_fixed_dates == issue_date) {
					// if fixed_dates == fixed_dates_q -> add to k (constant term)
					k += koff ;
				}
				else {
					push_back(fixed_dates, _fixed_dates) ;
					push_back(fixed_dates_q, issue_date) ;
					push_back(fixed_amounts, koff) ;
				}
		    }
	
			integer i = v_size(pmt_flat)-1 ;
	
			number term = -(r_fwd_sprd[i] + spread_sprd)*per_pay_sprd[i]*nom_sprd-nom_sprd ;
	
			if(pmt_flat[i] == issue_date) {
				// if fixed_dates == fixed_dates_q -> add to k (constant term)
				k += term ;
			}
			else {
				push_back(fixed_dates, pmt_flat[i]) ;
				push_back(fixed_dates_q, issue_date) ;
				push_back(fixed_amounts, term) ;
			}
	
			fixed_amounts[0] = k ;
		}
	
		return cashflows_q(trade, fixed_dates, fixed_dates_q, fixed_amounts);
	}
	//-------------------------------------------------------
	// cf_currbswap_ois_mtm
	//-------------------------------------------------------
	cashflows_q cf_currbswap_ois_mtm(	instrument		inst,
										logical			solve_flat,		
										disc_func		df_fwd_sprd,
										disc_func		df_fwd_flat,
										disc_func		df_disc,
										number			q_adj)
	{	    
		number spr = q_adj != 0 ? (inst.quote() + q_adj)/10000 : inst.quote()/10000;

		QL_REQUIRE(inst.is_swap_oisois(), "invalid instrument type");
		error_info ee = error_info(true,false);
		error_info ee2 = error_info(true,false);
		swap_oisois sw = swap_oisois(inst,ee);
		QL_FAIL_COND(null(sw), ee.message(), inst.name(ee2), true);		

		integer cf1,cf2;
		sw.coupon_freq(cf1,cf2,ee);

		day_count_method dc1,dc2;
		sw.dc_method(dc1,dc2,ee);

		vector(date) 	accr_start_dates_leg1, accr_end_dates_leg1, accr_start_dates_leg2,	 accr_end_dates_leg2;
		vector(date) 	pmt_dates1,pmt_dates2,p_pmt_dates1,p_pmt_dates2;
		vector(number) 	accr_period_leg1,accr_period_leg2;

		sw.accr_dates(	true, accr_start_dates_leg1, accr_end_dates_leg1, accr_period_leg1, accr_start_dates_leg2,									
						accr_end_dates_leg2, accr_period_leg2, ee);
		sw.cash_flow_dates(true,  pmt_dates1,	 pmt_dates2,ee) ;
		sw.prin_cash_flow_dates(true, p_pmt_dates1, p_pmt_dates2,ee) ;
		
		return cf_currbswap_ois_mtm(sw.trade_date(),sw.issue_date(),sw.maturity(),cf1,cf2, dc1,dc2,
								accr_start_dates_leg1,accr_end_dates_leg1,pmt_dates1,p_pmt_dates1,
								accr_start_dates_leg2,accr_end_dates_leg2,pmt_dates2,p_pmt_dates2,
								solve_flat,df_fwd_sprd,df_fwd_flat,df_disc,spr);
	
	}
	//-------------------------------------------------------
	// cf_currbswap 
	// (crosscurr basis swap quoted as 1 swap)
	// ex USDSEK fx=6, USD = flat, freq =4, i = SEK
	// Todo: different bd,cal,eom for the index of the legs not handled 
	//-------------------------------------------------------
	cashflows cf_currbswap(	date 			trade,
							date			issue_date,//== settle_date,
							date			maturity,
							number			roll_day,
							integer			freq_base,
							integer			freq_flat, 
							day_count_method dc_base,
							day_count_method dc_flat,
							bd_convention 	bd,
							calendar		cal,
							logical			eom,											 	
							logical			solve_flat,//true-->solve for USD fximplied disc_func
			      			fwd_func		f_fwd_base,//SEK
							number option(nullable)	first_fix_base,	
							fwd_func		f_fwd_flat,//USD
							number option(nullable)	first_fix_flat,	
							disc_func		df_disc,//solve_flat=true --> df_disc = SEK disc
							number			fx_bswap_sprd,//fx libor spread on the base leg
							number			fx_ois_adj_sprd,//use if fx_bswap_sprd is a ois basis swap spread)
							number option(nullable) fx_flat_sprd = null<number>)
	{	 
		number nom 					= 100;   				
		bd_convention idx_bd_base 	= bd;
		calendar idx_cal_base 		= cal;
		logical	idx_eom_base 		= eom;			
		logical ip_idx_stub 		= false;
		
		vector(date) float_idx0_base, float_pmt_base, float_idx1_base;
		CORE_INT.swap_flt_dates(issue_date,null<date>,null<date>, maturity,roll_day,freq_base,bd,cal,eom,idx_bd_base,
							idx_cal_base,idx_eom_base,ip_idx_stub, float_idx0_base,float_pmt_base,float_idx1_base);
	
		integer n_base = v_size(float_idx0_base);
		
		bd_convention idx_bd_flat	= bd;
		calendar idx_cal_flat		= cal;
		logical idx_eom_flat		= eom;		
		vector(date) float_idx0_flat, float_pmt_flat, float_idx1_flat;
		CORE_INT.swap_flt_dates(issue_date,null<date>,null<date>, maturity,roll_day,freq_flat,bd,cal,eom,idx_bd_flat,
							idx_cal_flat,idx_eom_flat,ip_idx_stub,float_idx0_flat,float_pmt_flat,float_idx1_flat);
	
		integer n_flat = v_size(float_idx0_flat);
	
		number fx_lib_sprd = fx_bswap_sprd + fx_ois_adj_sprd;
		if(null(fx_flat_sprd))
			fx_flat_sprd = 0;
	
		vector(number)	fixed_amounts[0];
		vector(date)	fixed_dates[0];		
	    
		push_back(fixed_dates, issue_date); 
		push_back(fixed_amounts, 0);
	
		vector(number) per_pay_base 	= period (float_idx0_base, float_pmt_base, dc_base);
		vector(number) per_pay_flat		= period (float_idx0_flat, float_pmt_flat, dc_flat);
		vector(number) per_idx0_base 	= (float_idx0_base - trade) / 365;
		vector(number) r_fwd_base 		= f_fwd_base.fwd(per_idx0_base);
		if(!null(first_fix_base))
			r_fwd_base[0] = first_fix_base;
		vector(number) per_idx0_flat	= (float_idx0_flat - trade) / 365;					
		vector(number) r_fwd_flat 		= f_fwd_flat.fwd(per_idx0_flat);
		if(!null(first_fix_flat))
			r_fwd_flat[0] = first_fix_flat;
	
		vector(number) cf_base 	= (r_fwd_base + fx_lib_sprd) .* per_pay_base;
		vector(number) cf_flat 	= (r_fwd_flat + fx_flat_sprd) .* per_pay_flat ;
		number dfs				= df_disc.disc((issue_date - trade) / 365);
	
		if(solve_flat) {//USD															 			
			vector(number) df	= df_disc.disc((float_pmt_base - trade) / 365)/dfs;	
			number pv 			= cf_base * df;			    
			pv 					+= df[n_base-1];					
				
			for (integer i = 0 ; i < n_flat; i++) {									
				push_back(fixed_dates, float_pmt_flat[i]); 
				push_back(fixed_amounts, cf_flat[i]);
		    }
			fixed_amounts[n_flat] += 1;
			fixed_amounts[0] = -pv;
		}
		else { //SEK
			
			vector(number) df	= df_disc.disc((float_pmt_flat - trade) / 365)/dfs;																 							
			number pv 			= cf_flat * df;			
			pv 					+= df[n_flat-1] ;				
			
			for (integer i = 0 ; i < n_base; i++) {																	
				push_back(fixed_dates, float_pmt_base[i]); 
				push_back(fixed_amounts, cf_base[i]);										
		    }
			fixed_amounts[n_base] += 1;
			fixed_amounts[0] = -pv;
		}
		fixed_amounts = fixed_amounts * nom;
		return cashflows(trade, fixed_dates, fixed_amounts);
	}
	
	//-------------------------------------------------------
	// cf_currbswap_mtm 
	// (crosscurr basis swap quoted as 1 swap)
	// ex USDSEK fx=6, USD = flat, freq =4, i = SEK
	// Todo: different bd,cal,eom for the index of the legs not handled 
	//-------------------------------------------------------
	/*cashflows_q cf_currbswap_mtm_(date 			trade,
								date			issue_date,//== settle_date,
								date			maturity,
								number			roll_day,
								integer 		freq_sprd,
								integer 		freq_flat, 
								day_count_method dc_sprd,
								day_count_method dc_flat,
								bd_convention 	bd,
								calendar		cal,
								logical			eom,											 	
								logical			solve_flat,//true-->solve for USD fximplied disc_func
				      			fwd_func		f_fwd_sprd,//SEK
								number option(nullable)	first_fix_sprd,	
								fwd_func		f_fwd_flat,//USD
								number option(nullable)	first_fix_flat,	
								disc_func		df_disc,//solve_flat=true --> df_disc = SEK disc
								number			fx_bswap_sprd,//fx libor spread on the base leg
								number			fx_ois_adj_sprd,//use if fx_bswap_sprd is a ois basis swap spread)
								number option(nullable) fx_flat_sprd = null<number>)
	{	 
		number fx_rate_settle 		= 1;
	   	number nom_sprd 			= 100;
		number nom_base 			= 100;
			
		bd_convention idx_bd_sprd 	= bd;
		calendar idx_cal_sprd 		= cal;
		logical	idx_eom_sprd 		= eom;			
		logical ip_idx_stub 		= false;
		
		vector(date) float_idx0_sprd, float_pmt_sprd, float_idx1_sprd;
		CORE_INT.swap_flt_dates(issue_date,null<date>,null<date>, maturity,roll_day,freq_sprd,bd,cal,eom,idx_bd_sprd,
							idx_cal_sprd,idx_eom_sprd,ip_idx_stub, float_idx0_sprd,float_pmt_sprd,float_idx1_sprd);
	
		integer n_sprd = v_size(float_idx0_sprd);
		
		bd_convention idx_bd_flat	= bd;
		calendar idx_cal_flat		= cal;
		logical idx_eom_flat		= eom;		
		vector(date) float_idx0_flat, float_pmt_flat, float_idx1_flat;
		CORE_INT.swap_flt_dates(issue_date,null<date>,null<date>, maturity,roll_day,freq_flat,bd,cal,eom,idx_bd_flat,
							idx_cal_flat,idx_eom_flat,ip_idx_stub,float_idx0_flat,float_pmt_flat,float_idx1_flat);
	
		integer n_flat = v_size(float_idx0_flat);
	
		number fx_lib_sprd = fx_bswap_sprd + fx_ois_adj_sprd;
		if(null(fx_flat_sprd))
			fx_flat_sprd = 0;
	
		vector(number)	fixed_amounts[0];
		vector(date)	fixed_dates[0];		
	    vector(date)	fixed_dates_q[0];
	
		push_back(fixed_dates, trade) ;
		push_back(fixed_dates_q, null<date>) ;
		push_back(fixed_amounts, 0) ;	
	
		vector(number) per_pay_sprd 	= period (float_idx0_sprd, float_pmt_sprd, dc_sprd);
		vector(number) per_pay_flat		= period (float_idx0_flat, float_pmt_flat, dc_flat);
		vector(number) per_idx0_sprd 	= (float_idx0_sprd - trade) / 365;
		vector(number) r_fwd_sprd 		= f_fwd_sprd.fwd(per_idx0_sprd);
		if(!null(first_fix_sprd))
			r_fwd_sprd[0] = first_fix_sprd;
		vector(number) per_idx0_flat	= (float_idx0_flat - trade) / 365;					
		vector(number) r_fwd_flat 		= f_fwd_flat.fwd(per_idx0_flat);
		if(!null(first_fix_flat))
			r_fwd_flat[0] = first_fix_flat;
	
		vector(number) cf_sprd 	= (r_fwd_sprd + fx_lib_sprd) .* per_pay_sprd;
		vector(number) cf_flat 	= (r_fwd_flat + fx_flat_sprd) .* per_pay_flat ;
		number dfs				= df_disc.disc((issue_date - trade) / 365);
		number rel_df			= 1/dfs;
	
		if(solve_flat) {//USD
			
																			 			
			vector(number) df	= df_disc.disc((float_pmt_sprd - trade) / 365)*rel_df;	
			number pv 			= -(cf_sprd * df) ;	
			pv 					= pv + 1 - df[n_sprd-1];		    							
			pv					= pv * nom_sprd - nom_base;
	
			push_back(fixed_dates_q, issue_date);
			push_back(fixed_dates_q, issue_date);
	
			for (integer i = 0 ; i < n_flat; i++) {
		
				push_back(fixed_dates, float_pmt_flat[i]); 
				push_back(fixed_dates, float_pmt_flat[i]); 
				if(i>0) {
					push_back(fixed_dates_q, float_pmt_flat[i-1]);
					push_back(fixed_dates_q, float_pmt_flat[i-1]);
				}
	
				number p_df_imp_f	= i==0 ? df_disc.disc((issue_date-trade)/365.0) : df_disc.disc((float_pmt_flat[i-1]-trade)/365.0);
		    	number df_imp_f		= df_disc.disc((float_pmt_flat[i]-trade)/365.0);	
				number k 			= nom_sprd*df_imp_f*rel_df/fx_rate_settle ;
				number p_k 			= i==0? nom_base : nom_sprd*p_df_imp_f*rel_df/fx_rate_settle ;
	
				push_back(fixed_amounts, p_k);
				push_back(fixed_amounts, cf_flat[i]* p_k) ;					
										
				if(i < n_flat-1) 			
					pv -= k ;
		    }
	
			fixed_amounts[0] = pv;
	
		}
		else { //SEK
			//to be done, no mtm
			QL_FAIL("err");
	
			//vector(number) df	= df_disc.disc((float_pmt_flat - trade) / 365)/dfs;																 							
			//number pv 			= cf_flat * df;			
			//pv 					+= df[n_flat-1] ;				
			
			//for (integer i = 0 ; i < n_sprd; i++) {																	
			//	push_back(fixed_dates, float_pmt_sprd[i]); 
			//	push_back(fixed_amounts, cf_sprd[i]);										
		    //}
			//fixed_amounts[n_sprd] += 1;
			//fixed_amounts[0] = -pv;
		}
	
		return cashflows_q(trade, fixed_dates, fixed_dates_q, fixed_amounts);
	}
	*/
	//-------------------------------------------------------
	// cf_currbswap_mtm 
	// (crosscurr basis swap quoted as 1 swap)
	// ex USDSEK fx=6, USD = flat, freq =4, i = SEK
	// Todo: different bd,cal,eom for the index of the legs not handled 
	//-------------------------------------------------------
	cashflows_q cf_currbswap_mtm(	date 			trade,
									date			issue_date,//== settle_date,
									date			maturity,
									number			roll_day,
									integer 		freq_sprd,
									integer 		freq_flat, 
									day_count_method dc_sprd,
									day_count_method dc_flat,
									bd_convention 	bd,
									calendar		cal,
									logical			eom,											 	
									logical			solve_flat,//true-->solve for USD fximplied disc_func
					      			fwd_func		f_fwd_sprd,//SEK
									number option(nullable)	first_fix_sprd,	
									fwd_func		f_fwd_flat,//USD
									number option(nullable)	first_fix_flat,	
									disc_func		df_disc,//solve_flat=true --> df_disc = SEK disc
									number			fx_bswap_sprd,//fx libor spread on the base leg
									number			fx_ois_adj_sprd,//use if fx_bswap_sprd is a ois basis swap spread)
									number option(nullable) fx_flat_sprd = null<number>)
	{	 
		number fx_rate_settle 		= 1;
	   	number nom_sprd 			= 100;
		number nom_base 			= 100;
			
		bd_convention idx_bd_sprd 	= bd;
		calendar idx_cal_sprd 		= cal;
		logical	idx_eom_sprd 		= eom;			
		logical ip_idx_stub 		= false;
		
		vector(date) float_idx0_sprd, float_pmt_sprd, float_idx1_sprd;
		CORE_INT.swap_flt_dates(issue_date,null<date>,null<date>, maturity,roll_day,freq_sprd,bd,cal,eom,idx_bd_sprd,
							idx_cal_sprd,idx_eom_sprd,ip_idx_stub, float_idx0_sprd,float_pmt_sprd,float_idx1_sprd);
	
		integer n_sprd = v_size(float_idx0_sprd);
		
		bd_convention idx_bd_flat	= bd;
		calendar idx_cal_flat		= cal;
		logical idx_eom_flat		= eom;		
		vector(date) float_idx0_flat, float_pmt_flat, float_idx1_flat;
		CORE_INT.swap_flt_dates(issue_date,null<date>,null<date>, maturity,roll_day,freq_flat,bd,cal,eom,idx_bd_flat,
							idx_cal_flat,idx_eom_flat,ip_idx_stub,float_idx0_flat,float_pmt_flat,float_idx1_flat);
	
		integer n_flat = v_size(float_idx0_flat);
	
		number fx_lib_sprd = fx_bswap_sprd + fx_ois_adj_sprd;
		if(null(fx_flat_sprd))
			fx_flat_sprd = 0;
	
		vector(number)	fixed_amounts[0];
		vector(date)	fixed_dates[0];		
	    vector(date)	fixed_dates_q[0];
	
		push_back(fixed_dates, trade) ;
		push_back(fixed_dates_q, null<date>) ;
	//	push_back(fixed_dates_q, trade) ;
		push_back(fixed_amounts, 0) ;	
	
		vector(number) per_pay_sprd 	= period (float_idx0_sprd, float_pmt_sprd, dc_sprd);
		vector(number) per_pay_flat		= period (float_idx0_flat, float_pmt_flat, dc_flat);
		vector(number) per_idx0_sprd 	= (float_idx0_sprd - trade) / 365.0;
		vector(number) r_fwd_sprd 		= f_fwd_sprd.fwd(per_idx0_sprd);
		if(!null(first_fix_sprd))
			r_fwd_sprd[0] = first_fix_sprd;
		vector(number) per_idx0_flat	= (float_idx0_flat - trade) / 365.0;					
		vector(number) r_fwd_flat 		= f_fwd_flat.fwd(per_idx0_flat);
		if(!null(first_fix_flat))
			r_fwd_flat[0] = first_fix_flat;
	
		vector(number) cf_sprd 	= (r_fwd_sprd + fx_lib_sprd) .* per_pay_sprd;
		vector(number) cf_flat 	= (r_fwd_flat + fx_flat_sprd) .* per_pay_flat ;
		
	
		if(solve_flat) {//USD
			
			number dfs			= df_disc.disc((issue_date - trade) / 365.0);
			number rel_df		= 1.0/dfs;															 			
			vector(number) df	= df_disc.disc((float_pmt_sprd - trade) / 365.0)*rel_df;	
			number pv 			= -(cf_sprd * df) ;	
			pv 					= pv + 1 - df[n_sprd-1];		    							
			pv					= pv * nom_sprd - nom_base;
	
			push_back(fixed_dates_q, issue_date);
	
			for (integer i = 0 ; i < n_flat; i++) {
		
				push_back(fixed_dates, float_pmt_flat[i]); 
				if(i>0) 
					push_back(fixed_dates_q, float_pmt_flat[i-1]);			
	
				number p_df_imp_f	= i==0 ? df_disc.disc((issue_date-trade)/365.0) : df_disc.disc((float_pmt_flat[i-1]-trade)/365.0);
		    	number df_imp_f		= df_disc.disc((float_pmt_flat[i]-trade)/365.0);	
				number k 			= nom_sprd*df_imp_f*rel_df/fx_rate_settle ;
				number p_k 			= i==0? nom_base : nom_sprd*p_df_imp_f*rel_df/fx_rate_settle ;
				push_back(fixed_amounts, p_k+cf_flat[i]* p_k) ;					
										
				if(i < n_flat-1) 			
					pv -= k ;
		    }
	
			fixed_amounts[0] 	= pv;
	
		}
		else { //SEK
			// Add constant flow
			number k = -nom_base * df_disc.disc((issue_date-trade)/365.0) / df_disc.disc((issue_date-trade)/365.0) ;
	
			for(integer i = 0 ;i < v_size(float_pmt_flat) ; i++) {  
				number p_df_base_f	= df_disc.disc(((i == 0 ? issue_date : float_pmt_flat[i-1])-trade)/365.0);
				number df_base_f	= df_disc.disc((float_pmt_flat[i]-trade)/365.0);
	
				number p_k ;
	
				if(i == 0) {
					p_k = nom_base * df_base_f/p_df_base_f ;
				}
				else
					p_k = df_base_f/p_df_base_f*nom_sprd/fx_rate_settle ;
	
				number koff = p_k + (r_fwd_flat[i] + fx_flat_sprd)*per_pay_flat[i]* p_k ;
	
				if(i > 0) {
					koff -= nom_sprd/fx_rate_settle ;
					koff -= (r_fwd_sprd[i-1] + fx_lib_sprd)*per_pay_sprd[i-1]*nom_sprd ;
				}
				else
					koff += nom_sprd ;
			
				date _fixed_dates = (i == 0 ? issue_date : float_pmt_flat[i-1]) ;
	
				if(_fixed_dates == issue_date) {
					// if fixed_dates == fixed_dates_q -> add to k (constant term)
					k += koff ;
				}
				else {
					push_back(fixed_dates, _fixed_dates) ;
					push_back(fixed_dates_q, issue_date) ;
					push_back(fixed_amounts, koff) ;
				}
		    }
	
			integer i = v_size(float_pmt_flat)-1 ;
	
			number term = -(r_fwd_sprd[i] + fx_lib_sprd)*per_pay_sprd[i]*nom_sprd-nom_sprd ;
	
			if(float_pmt_flat[i] == issue_date) {
				// if fixed_dates == fixed_dates_q -> add to k (constant term)
				k += term ;
			}
			else {
				push_back(fixed_dates, float_pmt_flat[i]) ;
				push_back(fixed_dates_q, issue_date) ;
				push_back(fixed_amounts, term) ;
			}
	
			fixed_amounts[0] = k ;
		}
	
		return cashflows_q(trade, fixed_dates, fixed_dates_q, fixed_amounts);
	}
	//-------------------------------------------------------
	// cf_currbswap
	//-------------------------------------------------------
	cashflows cf_currbswap(	instrument		inst,
							number 			fx_ois_adj_sprd,											 	
							logical			solve_flat,																	
			      			fwd_func		f_fwd_sprd,
							number option(nullable)	first_fix_sprd,											
							fwd_func		f_fwd_flat,
							number option(nullable)	first_fix_flat,	
							disc_func		df_disc,										
							number			q_adj)
	{	    
		number spr = q_adj != 0 ? (inst.quote() + q_adj)/10000 : inst.quote()/10000;

		QL_REQUIRE(inst.is_swap_fltflt(), "invalid instrument type");
		error_info ee = error_info(true,false);
		error_info ee2 = error_info(true,false);
		swap_fltflt sw = swap_fltflt(inst,ee);
		QL_FAIL_COND(null(sw), ee.message(), inst.name(ee2), true);		

		integer rf1,rf2;
		sw.reset_freq(rf1,rf2,ee);

		day_count_method dc1,dc2;
		sw.dc_method(dc1,dc2,ee);

		bd_convention bd1,bd2;
		sw.pmt_bus_day(bd1,bd2,ee);

		calendar c1,c2;
		sw.calendar(c1,c2,ee);

		logical e1,e2;
		sw.eom(e1,e2,ee);
		
		return cf_currbswap(sw.trade_date(),sw.issue_date(),sw.maturity(),day(sw.issue_date()),
							rf1,rf2, dc1,dc2,bd1, c1,e1,											 	
							solve_flat,f_fwd_sprd,first_fix_sprd,f_fwd_flat,first_fix_flat,df_disc,spr,fx_ois_adj_sprd);
	
	}
	
	//-------------------------------------------------------
	// cf_currbswap
	//-------------------------------------------------------
	cashflows_q cf_currbswap_mtm(	instrument		inst,
								number 			fx_ois_adj_sprd,
								logical			solve_flat,		
				      			fwd_func		f_fwd_sprd,
								number option(nullable)	first_fix_sprd,	
								fwd_func		f_fwd_flat,
								number option(nullable)	first_fix_flat,	
								disc_func		df_disc,
								number			q_adj)
	{	    
		number spr = q_adj != 0 ? (inst.quote() + q_adj)/10000 : inst.quote()/10000;

		QL_REQUIRE(inst.is_swap_fltflt(), "invalid instrument type");
		error_info ee = error_info(true,false);
		error_info ee2 = error_info(true,false);
		swap_fltflt sw = swap_fltflt(inst,ee);
		QL_FAIL_COND(null(sw), ee.message(), inst.name(ee2), true);

		integer rf1,rf2;
		sw.reset_freq(rf1,rf2,ee);

		day_count_method dc1,dc2;
		sw.dc_method(dc1,dc2,ee);

		bd_convention bd1,bd2;
		sw.pmt_bus_day(bd1,bd2,ee);

		calendar c1,c2;
		sw.calendar(c1,c2,ee);

		logical e1,e2;
		sw.eom(e1,e2,ee);
		
		return cf_currbswap_mtm(sw.trade_date(),sw.issue_date(),sw.maturity(),day(sw.issue_date()),
								rf1,rf2, dc1,dc2,bd1, c1,e1,											 	
								solve_flat,f_fwd_sprd,first_fix_sprd,f_fwd_flat,first_fix_flat,df_disc,spr,fx_ois_adj_sprd);
	
	}
	//-------------------------------------------------------
	// cf_currffswap 
	// (crosscurr fixflt swap)
	// ex USDKRW fx=1000, USD = flt, freq =4, i = KRW
	// Todo: different bd,cal,eom for the legs not handled 
	//-------------------------------------------------------
	cashflows cf_currffswap(instrument		inst,																							 	
							logical			solve_flt,//true-->solve for USD fximplied disc_func		      			
							fwd_func		flt_fwd,//USD
							number option(nullable)	first_fix,	
							disc_func		df_disc,//solve_flt=true --> df_disc = KRW disc
							number			y_adj)
	{
		swap_fixflt sw 			= swap_fixflt(inst);
		
		number nom				= 100;
		number roll_day 		= day(inst.issue_date());
		date trade				= inst.trade_date();
		date issue_date 		= inst.issue_date();
		date maturity 			= inst.maturity();
		integer flt_freq 		= inst.swap_coupon_freq_leg2();						
		day_count_method flt_dc = inst.swap_dc_method_leg2();		
	
	 	bd_convention bd 		= inst.pmt_bus_day_d(BD_MOD_FOLLOWING);
		calendar cal,cal2;
		sw.calendar(cal,cal2);
		logical eom 			= inst.eom_d(true);//FIX_ME

		//if(q_adj != 0)
			//inst = CORE_INT.create_par_swap(inst, inst.yield() + q_adj, null<string>);
			
		if(y_adj != 0) {
			swap_fixflt c = swap_fixflt(inst);
			c.add_quote(inst.yield() + y_adj,..quote_style.YIELD);
			inst = c;//CORE_INT.create_par_swap(inst, inst.yield() + y_adj, null<string>);
		}
		
		QL_FAIL_COND(inst.settle_date() != issue_date,"invalid par swap (issue date <> settlement date)");	

		error_info ee = new error_info(true,true);
		number dp = inst.dirty_price(null<number>,false,true,ee);
		QL_FAIL_COND(null(dp) || !CORE_INT.is_equal_par(dp,100),"invalid par swap (not par)");	
			
		vector(date)	fix_dates;
		vector(number)	fix_amounts;			 			
		inst.cash_flows(fix_dates, fix_amounts);
		QL_FAIL_COND(null(fix_amounts) || v_size(fix_amounts) == 0,"invalid par swap (fix leg cash flows)");

		
		number inst_nom, inst_nom_flt;
		sw.notional(inst_nom, inst_nom_flt);
		//number inst_nom = inst.nominal();
		integer n_fix = v_size(fix_amounts);
		fix_amounts[n_fix - 1] -= inst_nom;
		fix_amounts /= inst_nom;
		
					
		bd_convention idx_bd 	= bd;
		calendar idx_cal 		= cal;
		logical	idx_eom 		= eom;			
		logical ip_idx_stub 	= false;
		
		vector(date) float_idx0, float_pmt, float_idx1;
		CORE_INT.swap_flt_dates(issue_date,null<date>,null<date>, maturity,roll_day,flt_freq,bd,cal,eom,idx_bd,
							idx_cal,idx_eom,ip_idx_stub, float_idx0,float_pmt,float_idx1);
	
		integer n_flt = v_size(float_idx0);
	
		vector(number) per_pay		= period (float_idx0, float_pmt, flt_dc);
		vector(number) per_idx0		= (float_idx0- trade) / 365;					
		vector(number) r_fwd		= flt_fwd.fwd(per_idx0);
		if(!null(first_fix))
			r_fwd[0] = first_fix;
		vector(number) flt_cf 		= r_fwd.* per_pay;
		number dfs					= df_disc.disc((issue_date - trade) / 365);
			
	
		if(solve_flt) {//USD															 			
			vector(number) df	= df_disc.disc((fix_dates - trade) / 365)/dfs;	
			number pv 			= fix_amounts * df;			    
			pv 					+= df[n_fix-1];					
				
			vector(number)	fixed_amounts = [0];
			vector(date)	fixed_dates = [issue_date];			    		
	
			for (integer i = 0 ; i < n_flt; i++) {									
				push_back(fixed_dates, float_pmt[i]); 
				push_back(fixed_amounts, flt_cf[i]);
		    }
			fixed_amounts[n_flt] += 1;
			fixed_amounts[0] = -pv;
	
			fixed_amounts *= nom;
			return cashflows(trade, fixed_dates, fixed_amounts);
	
		}
		else { //KRW
			
			vector(number) df	= df_disc.disc((float_pmt - trade) / 365)/dfs;																 							
			number pv 			= flt_cf * df;			
			pv 					+= df[n_flt-1] ;
	
			fix_amounts[n_fix-1] 	+= 1;
			fix_amounts = concat ([-pv], fix_amounts);
	    	fix_dates = concat ([inst.issue_date()], fix_dates);
	
			fix_amounts *= nom;	
			return cashflows(trade, fix_dates, fix_amounts);													
		}
		
	}
}
