option(null: hard);	
/*	
	dual boot risk functions
	Developer: Algorithmica Research, Magnus Nyström			


*/


/*-----------------------------------------------------------------------
  risk_ladder_disc
  ----------------------------------------------------------------------*/
vector(number) risk_ladder_disc(swap_fixflt 		sw,
								swap_curve_disc_ext disc_curve,
								disc_z_model 		disc_model,
								fwd_func 			fwd_base,						
								number 				y_shift,
								vector(integer) 	incl,
								out vector(instrument) instr_bump,
								out number			sw_risk_all) 
{	
	disc_func df_base,df_all;
	
	vector(disc_func) df_all_v;
	vector(disc_func) df_incl_v;
	df_incl_v = disc_curve.instr_bumper_df_vect(disc_model,y_shift,incl,df_base, df_all, instr_bump,df_all_v);
	
	vector(number) sw_risk_incl;
	sw.risk_ladder_disc(fwd_base,df_incl_v,df_base,df_all,y_shift,sw_risk_incl,sw_risk_all);

	integer size = v_size(df_all_v);
	integer size_incl = v_size(df_incl_v);
	if(size != size_incl) {
		vector(number) sw_risk[size];
		sw_risk = 0;
		for(integer i=0;i<size_incl;i++) {
			sw_risk[incl[i]] = sw_risk_incl[i];
		}
		return sw_risk;
	}
	else {		
		return sw_risk_incl;
	}
}


/*-----------------------------------------------------------------------
  risk_ladder_disc
  ----------------------------------------------------------------------*/
vector(number) risk_ladder_disc(swap_fixois 		sw,
								swap_curve_disc_ext disc_curve,
								disc_z_model 		disc_model,
								//fwd_func 			fwd_base,						
								number 				y_shift,
								vector(integer) 	incl,
								out vector(instrument) instr_bump,
								out number			sw_risk_all) 
{	
	disc_func df_base,df_all;
	
	vector(disc_func) df_all_v;
	vector(disc_func) df_incl_v;
	df_incl_v = disc_curve.instr_bumper_df_vect(disc_model,y_shift,incl,df_base, df_all, instr_bump,df_all_v);
	
	vector(number) sw_risk_incl;
	sw.risk_ladder_disc(/*fwd_base,*/df_incl_v,df_base,df_all,y_shift,sw_risk_incl,sw_risk_all);

	integer size = v_size(df_all_v);
	integer size_incl = v_size(df_incl_v);
	if(size != size_incl) {
		vector(number) sw_risk[size];
		sw_risk = 0;
		for(integer i=0;i<size_incl;i++) {
			sw_risk[incl[i]] = sw_risk_incl[i];
		}
		return sw_risk;
	}
	else {		
		return sw_risk_incl;
	}
}

/*-----------------------------------------------------------------------
  risk_ladder_fwd
  ----------------------------------------------------------------------*/
vector(number)  risk_ladder_fwd(swap_fixflt 		sw,
								swap_curve_fwd_ext 	fwd_curve,
								fwd_z_model			fwd_model,					
								number 				y_shift,								
								vector(integer) 	incl,
								out vector(instrument) instr_bump,
								out number			sw_risk_all ) 
{	
	fwd_func fwd_base, fwd_all;
	vector(fwd_func) f_all_v;
	vector(fwd_func) f_incl_v;
	vector(instrument) all_instr;
	f_incl_v = fwd_curve.instr_bumper_ff_vect(fwd_model,y_shift,incl,fwd_base, fwd_all,all_instr,instr_bump,f_all_v);
	vector(number) sw_risk_incl;
	sw.risk_ladder_fwd(f_incl_v,fwd_base,fwd_all,y_shift,sw_risk_incl,sw_risk_all);

	integer size = v_size(f_all_v);
	integer size_incl = v_size(f_incl_v);
	if(size != size_incl) {
		vector(number) sw_risk[size];
		sw_risk = 0;
		for(integer i=0;i<size_incl;i++) {
			sw_risk[incl[i]] = sw_risk_incl[i];
		}
		return sw_risk;
	}
	else {		
		return sw_risk_incl;
	}
}
/*-----------------------------------------------------------------------
  risk_ladder_fwd_wave
  ----------------------------------------------------------------------*/
vector(number)  risk_ladder_fwd_wave(swap_fixflt 		sw,
									swap_curve_fwd_ext 	fwd_curve,
									fwd_z_model			fwd_model,					
									number 				y_shift,
									vector(integer) 	incl,
									out vector(instrument) 	instr_bump,
									out number			sw_risk_all ) 
{	
	fwd_func fwd_base, fwd_all;
	vector(instrument) all_instr;
	vector(fwd_func) f_incl_v = fwd_curve.wave_bumper_ff_vect(fwd_model,y_shift,incl,fwd_base, fwd_all,all_instr,instr_bump);
	vector(number) sw_risk_incl;
	sw.risk_ladder_fwd(f_incl_v,fwd_base,fwd_all,y_shift,sw_risk_incl,sw_risk_all);

	integer size = v_size(all_instr);
	integer size_incl = v_size(f_incl_v);
	if(size != size_incl) {
		vector(number) sw_risk[size];
		sw_risk = 0;
		for(integer i=0;i<size_incl;i++) {
			sw_risk[incl[i]] = sw_risk_incl[i];
		}
		return sw_risk;
	}
	else {		
		return sw_risk_incl;
	}
}

/*-----------------------------------------------------------------------
  risk_ladder_fwd
  ----------------------------------------------------------------------*/
vector(number)  risk_ladder_fwd(swap_fixois 		sw,
								swap_curve_fwd_ext 	fwd_curve,
								fwd_z_model			fwd_model,					
								number 				y_shift,								
								vector(integer) 	incl,
								out vector(instrument) instr_bump,
								out number			sw_risk_all ) 
{	
	fwd_func fwd_base, fwd_all;
	vector(fwd_func) f_all_v;
	vector(fwd_func) f_incl_v;
	vector(instrument) all_instr;
	f_incl_v = fwd_curve.instr_bumper_ff_vect(fwd_model,y_shift,incl,fwd_base, fwd_all,all_instr,instr_bump,f_all_v);
	vector(number) sw_risk_incl;
	sw.risk_ladder_fwd(f_incl_v,fwd_base,fwd_all,y_shift,sw_risk_incl,sw_risk_all);

	integer size = v_size(f_all_v);
	integer size_incl = v_size(f_incl_v);
	if(size != size_incl) {
		vector(number) sw_risk[size];
		sw_risk = 0;
		for(integer i=0;i<size_incl;i++) {
			sw_risk[incl[i]] = sw_risk_incl[i];
		}
		return sw_risk;
	}
	else {		
		return sw_risk_incl;
	}
}
/*-----------------------------------------------------------------------
  risk_ladder_fwd_wave
  ----------------------------------------------------------------------*/
vector(number)  risk_ladder_fwd_wave(swap_fixois 		sw,
									swap_curve_fwd_ext 	fwd_curve,
									fwd_z_model			fwd_model,					
									number 				y_shift,
									vector(integer) 	incl,
									out vector(instrument) 	instr_bump,
									out number			sw_risk_all ) 
{	
	fwd_func fwd_base, fwd_all;
	vector(instrument) all_instr;
	vector(fwd_func) f_incl_v = fwd_curve.wave_bumper_ff_vect(fwd_model,y_shift,incl,fwd_base, fwd_all,all_instr,instr_bump);
	vector(number) sw_risk_incl;
	sw.risk_ladder_fwd(f_incl_v,fwd_base,fwd_all,y_shift,sw_risk_incl,sw_risk_all);

	integer size = v_size(all_instr);
	integer size_incl = v_size(f_incl_v);
	if(size != size_incl) {
		vector(number) sw_risk[size];
		sw_risk = 0;
		for(integer i=0;i<size_incl;i++) {
			sw_risk[incl[i]] = sw_risk_incl[i];
		}
		return sw_risk;
	}
	else {		
		return sw_risk_incl;
	}
}

/*-----------------------------------------------------------------------
  risk_ladder_disc
  ----------------------------------------------------------------------*/
void risk_ladder_disc(	date 				d,
						vector(date)		cash_flow_date,
						vector(number) 		fixed_cash_flow,						
						disc_func 			df_base,
						disc_func option(nullable) df_all,
						vector(disc_func) 	df_v_pert,						
						number 				y_shift,
						out number 			pv_base,
						out vector(number) 	pv_diff_pert,
						out number 			pv_diff_all) 
{	
	integer size 		= v_size(df_v_pert);
	resize(pv_diff_pert,size);
	pv_diff_pert 		= 0;
	number mult 		= 0.0001 / y_shift;

	date pv_d 		= null<date>;
	date cutoff_d 	= null<date>;
	pv_base 		= CORE_INT.cf_present_value(cash_flow_date,fixed_cash_flow,d,pv_d,cutoff_d,df_base);

	for(integer i=0;i<size;i++){
					
		number pvi 		= CORE_INT.cf_present_value(cash_flow_date,fixed_cash_flow,d,pv_d,cutoff_d,df_v_pert[i]);		
		pv_diff_pert[i] = (pvi - pv_base)*mult;			
	
	}
	if(!null(df_all)){
		
		number pv_all 	= CORE_INT.cf_present_value(cash_flow_date,fixed_cash_flow,d,pv_d,cutoff_d,df_all);		
		pv_diff_all 	= (pv_all - pv_base)*mult;
	}
	else {
		pv_diff_all 	= null<number>;
	}
}

/*-----------------------------------------------------------------------
  risk_ladder_fwd
  ----------------------------------------------------------------------*/
void risk_ladder_fwd(	date 				d,
						vector(date)		cash_flow_reset_date,
						vector(date)		cash_flow_start_date,
						vector(date)		cash_flow_mat_date,								
						vector(number) 		float_ir_cash_flow,//must be flat i.e. without a fixed spread
						vector(number) 		float_nom_cash_flow,
						day_count_method 	dc,
						disc_func 			df_base,
						fwd_func 			f_base,
						fwd_func option(nullable) f_all,
						vector(fwd_func) 	f_v_pert,						
						number 				y_shift,
						out number 			pv_base,
						out vector(number) 	pv_diff_pert,
						out number 			pv_diff_all) 
{	
	integer size 		= v_size(f_v_pert);
	resize(pv_diff_pert,size);
	pv_diff_pert 		= 0;
	number mult 		= 0.0001 / y_shift;

	date pv_d 		= null<date>;
	date cutoff_d 	= null<date>;
	pv_base 		= CORE_INT.cf_flt_present_value(cash_flow_reset_date,cash_flow_start_date,cash_flow_mat_date,
													float_ir_cash_flow,float_nom_cash_flow,dc, d,pv_d,cutoff_d,df_base,f_base);

	for(integer i=0;i<size;i++){
					
		number pvi 		= CORE_INT.cf_flt_present_value(cash_flow_reset_date,cash_flow_start_date,cash_flow_mat_date,
													float_ir_cash_flow,float_nom_cash_flow,dc, d,pv_d,cutoff_d,df_base,f_v_pert[i]);		
		pv_diff_pert[i] = (pvi - pv_base)*mult;			
	
	}
	if(!null(f_all)){
		
		number pv_all 	= CORE_INT.cf_flt_present_value(cash_flow_reset_date,cash_flow_start_date,cash_flow_mat_date,
													float_ir_cash_flow,float_nom_cash_flow,dc, d,pv_d,cutoff_d,df_base,f_all);		
		pv_diff_all 	= (pv_all - pv_base)*mult;
	}
	else {
		pv_diff_all 	= null<number>;
	}
}


