/*	
	Dual bootstrap utility functions
	Developed by Algorithmica Research, Magnus Nyström
	
	
	Comments: 
	I denna ql fil finns funktionerna tension_spline_fwd_exact(...) och tension_spline_fwd_spread(...). Dessa kräver 
	minst Quantlab version 3.1315

	ToDo:
	1. pvbp calc from cashflows_q MUST BE CHECKED [calc_pvbp_cfq]
	2. support for chop_short(cashflows) [disc_z_model_step_boot.df(cashflows_base crv)]
	3. unsupported disc model _spr for max_smooth and max_smooth_spreads[max_smooth.df]
	4. unsupported disc model _spr for bootstrap_on
	5. unsupported disc model _spr for tanggaard
	6. 
	7. unsupported fwd model: _ext for tension_spreads [fwd_z_model_tension.fwd]

*/
option(null: hard);	

	
module CORE_INT
{
	/*-----------------------------------------------------------------------
	  Name :    cashflows_fwd_periods
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			? 				MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	vector(number) cashflows_fwd_periods(	cashflows_fwd 	cf,
											integer 		i)
	
	{	
		vector(number) fixed_amounts,nominals; 
		vector(date) fwd_dates, fixed_dates,disc_dates;
		logical f = cf.get_cashflows(i, fixed_dates, fixed_amounts, fwd_dates, disc_dates,nominals);
		if(!f) return null<vector(number)>;
		return (fwd_dates-cf.trade_date())/365;	
	}
	/*-----------------------------------------------------------------------
	  Name :    cashflows_fwd_periods
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			? 				MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	logical cashflows_fwd_periods(	cashflows_fwd cf,
									out vector(number)   periods)	
	{	
		try {
			periods = cf.get_all_periods();
			return true;
		}
		catch {
			QL_WARNING( err.message());
			return false;
		}

		/*set_date s = new set_date();
		vector(date) fwd_dates;
	
		for(integer i = 0; i< cf.n_instruments(); i++){
			vector(number) fixed_amounts,nominals; 
			vector(date) fixed_dates,disc_dates;
			logical f = cf.get_cashflows(i, fixed_dates, fixed_amounts, fwd_dates, disc_dates,nominals);
			if(!f) return false;
			for(integer j = 0; j < v_size(fwd_dates); j++)
				s.add(fwd_dates[j]);
		}
		   	
		fwd_dates = s.get_content();
		periods = (fwd_dates-cf.trade_date())/365;	
		return true;*/
	}
	/*-----------------------------------------------------------------------
	  Name :    cashflows_fwd_pmt_periods (single)
	  Notes :   the nominals argument contains 100 * pmt_period       
	  Changes:
	  			Date    		Who     Detail
	  			2016-08-30 		MN  
	  ----------------------------------------------------------------------*/
	vector(number) cashflows_fwd_pmt_periods(	cashflows_fwd 	cf,
												integer 		i)	
	{			
		
		vector(number) fixed_amounts,nominals; 
		vector(date) fixed_dates,fwd_dates,disc_dates;
		logical f = cf.get_cashflows(i, fixed_dates, fixed_amounts, fwd_dates, disc_dates, nominals);
		if(!f) return null<vector(number)>;			
		return nominals/100;
	}
	/*-----------------------------------------------------------------------
	  Name :    cashflows_fwd_pvbp
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			2016-08-30 		MN      
	  ----------------------------------------------------------------------*/
	number cashflows_fwd_pvbp(	cashflows_fwd 	cf,
								integer 		i,								
								disc_func 		df)
	{
		try {
			vector(number) fixed_amounts,nominals; 
			vector(date) fixed_dates,fwd_dates,disc_dates;
			logical f = cf.get_cashflows(i, fixed_dates, fixed_amounts, fwd_dates, disc_dates, nominals);
			QL_REQUIRE(!null(f),"invalid cashflows_fwd data");
			vector(number) float_periods = cashflows_fwd_pmt_periods(cf,i);

			vector(number) df_v = df.disc((disc_dates - cf.trade_date())/365);
			number sum_a_df 	= float_periods * df_v;
			return 0.01 * sum_a_df;
		}
		catch {	
			QL_WARNING( err.message());		
			return null<number>;
		}
			
	}
	/*-----------------------------------------------------------------------
	  Name :    cashflows_fwd_max_per
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			? 				MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	number cashflows_fwd_max_per(cashflows_fwd cf)
	//option (category: 'Yield Curve/003 Curve Building')
	{	
		date m ;
		
		for(integer i = 0; i< cf.n_instruments(); i++){
			vector(number) fixed_amounts,nominals; 
			vector(date) fixed_dates,disc_dates, fwd_dates;
			logical f = cf.get_cashflows(i, fixed_dates, fixed_amounts, fwd_dates, disc_dates,nominals);
			if(!f) return null<number>;
			integer size = v_size(fwd_dates);
			if (i==0)
				m = fwd_dates[size-1] ;
			else if(fwd_dates[size-1] > m)
				m = fwd_dates[size-1] ;		
		}
		return (m - cf.trade_date())/365;	
	}

	//-------------------------------------------------------
	// chop_short
	//-------------------------------------------------------
	cashflows chop_short(cashflows cf, date d)
	{
	    integer	n = cf.n_instruments();
		vector(cashflows) c;
	
		for (integer i = 0 ; i < n ; i++) {
			vector(date)	dates;
			vector(number)	amounts;
			logical ok 	= cf.get_cashflows(i, dates, amounts);
			QL_REQUIRE(ok,"chop_short fail");
			date mat 	= dates[v_size(dates)-1];
			if(mat >= d)
				push_back(c, cashflows(cf.trade_date(), dates, amounts) );
		}
	    
	    return cashflows(c);
	}			
}

class enum fwd_z_model_type option (category: 'Yield Curve/003 Curve Building')
{
	FZM_LINEAR			option(str: "Linear"),
	FZM_SPLINE_NAT		option(str: "Spline_Nat"),
	FZM_SPLINE_FIN		option(str: "Spline_Fin"),
	FZM_STEP			option(str: "Step"),
	FZM_STEP_LINEAR		option(str: "Step_Linear"),
	FZM_HERMITE_AKIMA	option(str: "Hermite_Akima"),
	FZM_HERMITE_CR		option(str: "Hermite_cr"),
	FZM_HERMITE_FINDIFF	option(str: "Hermite_Findiff"),
	FZM_HERMITE_FB		option(str: "Hermite_fb"),
	FZM_HERMITE_KRUGER	option(str: "Hermite_Kruger"),
	FZM_HERMITE_MONO	option(str: "Hermite_Mono"),
	FZM_HERMITE_P		option(str: "Hermite_p"),
	FZM_TENSION_EXACT	option(str: "Tension_Exact"),
	FZM_TENSION_SPREAD	option(str: "Tension_Spread")
};

vector(string) fwd_z_model_list()  
{
	vector(fwd_z_model_type) e;
	expand_enum(e) ;
	return string(e);
}

class enum fwd_extrap_type option (category: 'Yield Curve/003 Curve Building') 
{
	FXP_MODEL 		option(str: "Model"),
	FXP_TENOR_FLAT 	option(str: "Tenor_Flat")
};

vector(string) fwd_extrap_list()  
{
	vector(fwd_extrap_type) e;
	expand_enum(e) ;
	return string(e);
}

/*-----------------------------------------------------------------------
  Name :    fwd_z_map
  Notes :         
  Changes:
  			Date    		Who     Detail
  			? 				MN   	Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
fwd_z_model_type fwd_z_map(string s) 
option (category: 'Yield Curve/003 Curve Building')
{
	fwd_z_model_type e;
	str_to_enum(s,e);
	return e;

}

/*-----------------------------------------------------------------------
  Name :    fwd_ip_map
  Notes :         
  Changes:
  			Date    		Who     Detail
  			? 				MN   	Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
interpolator fwd_ip_map(fwd_z_model_type 			fwd_model_t,
						logical option(nullable)	apply_hyman,						
						number option(nullable) 	left_der = null<number>,
						number option(nullable) 	hermite_right_der = null<number>) 
option (category: 'Yield Curve/003 Curve Building')
{
	logical allow_x_pol = true;
	integer flag;

	if(fwd_model_t != fwd_z_model_type.FZM_LINEAR && fwd_model_t != fwd_z_model_type.FZM_STEP) {
		QL_REQUIRE(!null(apply_hyman),"invalid hyman flag (fwd model)");
		if(apply_hyman )		
			flag = !allow_x_pol ? IP_HYMAN | IP_NO_XPOL_R :IP_HYMAN ;
		else
			flag = !allow_x_pol ? IP_NO_XPOL_R :0 ;
	}

	switch (fwd_model_t)
	{ 	
	case fwd_z_model_type.FZM_LINEAR:
		return !allow_x_pol ? ip_linear(left_der, null<number>, IP_NO_XPOL_R) : ip_linear(left_der, null<number>) ;
	case fwd_z_model_type.FZM_SPLINE_NAT:
		return  ip_spline(left_der, null<number>, flag  );
	case fwd_z_model_type.FZM_SPLINE_FIN:
		return  ip_spline(left_der, 0, flag  );
	case fwd_z_model_type.FZM_STEP:
		return  !allow_x_pol ? ip_step(IP_NO_XPOL_R) : ip_step();	
	case fwd_z_model_type.FZM_HERMITE_AKIMA:
		return  ip_hermite_akima(left_der, hermite_right_der, flag );
	case fwd_z_model_type.FZM_HERMITE_CR:
		return  ip_hermite_catmull_rom(left_der, hermite_right_der, flag);
	case fwd_z_model_type.FZM_HERMITE_FINDIFF:
		return  ip_hermite_finite_diff(left_der, hermite_right_der, flag );
	case fwd_z_model_type.FZM_HERMITE_FB:
		return  ip_hermite_fritsch_butland(left_der, hermite_right_der, flag );
	case fwd_z_model_type.FZM_HERMITE_KRUGER:
		return  ip_hermite_kruger(left_der, hermite_right_der, flag);
	case fwd_z_model_type.FZM_HERMITE_MONO:
		return  ip_hermite_monotone(left_der, hermite_right_der, flag );
	case fwd_z_model_type.FZM_HERMITE_P:
		return  ip_hermite_parabolic(left_der, hermite_right_der, flag );
	default:
		QL_FAIL("invalid fwd model");
	}
}
/*-----------------------------------------------------------------------
  Name :    fwd_extrap_map
  Notes :         
  Changes:
  			Date    		Who     Detail
  			? 				MN   	Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
fwd_extrap_type fwd_extrap_map(string s) 
option (category: 'Yield Curve/003 Curve Building')
{
	fwd_extrap_type e;
	str_to_enum(s,e);
	return e;

}

/*-----------------------------------------------------------------------
  Name :    fwd_z_model - abstract root
  Notes :           
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws)
			10-Aug-2016		MN		Added new tension _spr models and args for l-der / r-der (from ql_zc 1319 and 1320)
  ----------------------------------------------------------------------*/
class fwd_z_model
{
	public:

	virtual fwd_func 	fwd(cashflows_fwd, disc_func, curve option(nullable) crv = null<curve>) 					= 0;
	virtual fwd_func 	fwd(fwd_func, date, cashflows_fwd, disc_func,  curve option(nullable) crv = null<curve>) 	= 0;
	virtual fwd_func 	fwd(fwd_func, date, fwd_func, cashflows_fwd, disc_func,  curve option(nullable) crv = null<curve>) 	= 0;
	virtual fwd_func 	fwd(fwd_func, cashflows_fwd, disc_func,  curve option(nullable) crv = null<curve> ) 		= 0 ;


	virtual fwd_z_model_type 	type() 							= 0;
	virtual interpolator 		fwd_ip() 						= 0;
	virtual fwd_func			extrapol_fwd(fwd_func,number) 	= 0;

	protected:

	fwd_z_model( fwd_z_model_type, fwd_extrap_type, number option(nullable) );
	logical 					calc_pvbp(cashflows_fwd ,  disc_func, out vector(number)  , out string);
	logical 					calc_pvbp(cashflows_fwd , curve ,  disc_func, out vector(number)  , out string);
	fwd_func 					extrapol_model(fwd_func,number);

	fwd_z_model_type	type_;
	fwd_extrap_type 	xpol_type_;
	number 				xpol_cut_; //-1->no cutoff,  0 or null --> cutoff at curve max
};

fwd_z_model.fwd_z_model(fwd_z_model_type 		type,
						fwd_extrap_type 		xpol_type,
						number option(nullable) xpol_cut ) 
						: type_(type), xpol_type_(xpol_type), xpol_cut_(xpol_cut) 
{
	if(null(xpol_cut_))
		xpol_cut_ = 0;
}



logical fwd_z_model.calc_pvbp(	cashflows_fwd 	cf,
								disc_func 		df, 
								out vector(number)   pvbp, 
								out string err_msg)
{
	err_msg = "";
	//date trade	= cf.trade_date();
    integer n	= cf.n_instruments();
	
    if (n == 0) {
		err_msg = "invalid pvbp vector (no instruments)";
		return false;
	}

	resize(pvbp,n);

    integer		n_ok	= 0;

    for (integer i = 0 ; i < n ; i++) {

		pvbp[i] = CORE_INT.cashflows_fwd_pvbp(cf,i, df);
		if(null(pvbp[i]))
			continue;

		++n_ok;

		/*vector(date)	fixed_dates;
		vector(number)	fixed_amounts;
		vector(date)	float_dates;
		vector(date)	float_disc_d;
		vector(number)	float_nominals;
		

		if (!cf.get_cashflows(i, fixed_dates,fixed_amounts,float_dates,float_disc_d,float_nominals))
	    	continue;

		++n_ok;

		//date	settle	= dates[0]; // We assume no flows before settle
		vector(date)	fixed_dates2;
		vector(number)	fixed_amounts2;
		integer s = v_size(fixed_dates)+1;
		resize(fixed_dates2, s);
		resize(fixed_amounts2, s);
		fixed_dates2[0] = trade;
		fixed_amounts2[0] = -100;

		for (integer k = 1 ; k < s ; k++) {
			fixed_amounts2[k] = -fixed_amounts[k-1];
			fixed_dates2[k] = fixed_dates[k-1];
		}	
		fixed_amounts2[s - 1] += 100;

		pvbp[i] = __w_calc.pvbp(trade, 0, fixed_amounts2, fixed_dates2)/10000;*/
	}
	
	if (n_ok == 0) {
		err_msg = "invalid pvbp vector (no instrument cashflows)";
		return false;
	}

    if (n_ok != n) {
		err_msg = strcat([ "invalid pvbp vector (cashflows inconsistency: could only calculate pvbp for ",
		       			string(n_ok), " of ", string(n), " instruments)" ]);
		return false;
	}

    return true;
}


logical fwd_z_model.calc_pvbp(	cashflows_fwd 	cf, 
								curve 			crv, 
								disc_func		df,
								out vector(number)   pvbp, 
								out string err_msg)
{	
	err_msg = "";
    integer n = crv.n_instruments();
	
    if (n == 0) {
		err_msg = "invalid pvbp vector (no instruments)";
		return false;
	}

	resize(pvbp,n);
    integer		n_ok	= 0;

    for (integer i = 0 ; i < n ; i++) {
		instrument inst = crv.instruments()[i];
		if(inst.is_swap_fltflt() || inst.is_swap_oisois()) {
			pvbp[i] = CORE_INT.cashflows_fwd_pvbp(cf,i, df);
		}
		else {
			pvbp[i] = CORE_INT.cashflows_fwd_pvbp(cf,i, df);
			pvbp[i] = inst.pvbp();
		}

		/*if(null(pvbp[i])) {
			vector(date)	fixed_dates;
			vector(number)	fixed_amounts;
			vector(date)	float_dates;
			vector(date)	float_disc_d;
			vector(number)	float_nominals;
	
			if (crv.get_cashflows(i, fixed_dates,fixed_amounts,float_dates,float_disc_d,float_nominals))		    								
				pvbp[i] = __w_calc.pvbp(crv.trade_date(), 0, fixed_amounts, fixed_dates)/10000;
		}*/

		if(!null(pvbp[i]))
			++n_ok;

	}
	if (n_ok == 0) {
		err_msg = "invalid pvbp vector (pvbp calculation failed)";
		return false;
	}

    if (n_ok != n) {
		err_msg = strcat([ "invalid pvbp vector (cashflows inconsistency: could only calculate pvbp for ",
		       			string(n_ok), " of ", string(n), " instruments)" ]);
		return false;
	}

    return true;
}


fwd_func fwd_z_model.extrapol_model(fwd_func 	f, 
									number 		crv_max)
{
	if(xpol_cut_ == -1) {
		f =  extrapol_fwd(f, crv_max);
		return f;// //-1->no cutoff
	}		
	else if(xpol_cut_ == 0) // 0 --> cutoff at curve max
		return  extrap_none(f, crv_max ) ;
	else if(xpol_cut_ <= crv_max) // 0 or less than max --> cutoff at curve max
		return  extrap_none(f, xpol_cut_) ;
	else {//xpol_cut_ > crv_max		
		f =  extrapol_fwd(f, crv_max);
		return  extrap_none(f, xpol_cut_) ;
	}	
}

//-------------------------------------------------------
// class fwd_z_model_ip
//-------------------------------------------------------
class fwd_z_model_ip : public fwd_z_model option (category: 'Yield Curve/003 Curve Building')
{
   	public:

	fwd_z_model_ip(fwd_z_model_type, logical option(nullable), number option(nullable),number option(nullable),fwd_extrap_type, number option(nullable));
 
	virtual fwd_func 			fwd(cashflows_fwd, disc_func, curve option(nullable) crv = null<curve>);
	virtual fwd_func 			fwd(fwd_func, date, cashflows_fwd, disc_func,  curve option(nullable) crv = null<curve>);
	virtual fwd_func 			fwd(fwd_func, date, fwd_func, cashflows_fwd, disc_func,  curve option(nullable) crv = null<curve>);
	virtual fwd_func 			fwd(fwd_func, cashflows_fwd, disc_func,  curve option(nullable) crv = null<curve>);

	virtual fwd_z_model_type 	type();
	virtual interpolator 		fwd_ip();

	protected:

	virtual fwd_func			extrapol_fwd(fwd_func,number );	

	interpolator 	fwd_ip_;
	
};

fwd_z_model_type fwd_z_model_ip.type() 	{ return type_;}
interpolator  fwd_z_model_ip.fwd_ip() 	{ return fwd_ip_;}

fwd_z_model_ip.fwd_z_model_ip( 	fwd_z_model_type 			type, 
								logical option(nullable) 	apply_hyman,							
								number option(nullable) 	left_der,
								number option(nullable) 	hermite_right_der,
								fwd_extrap_type 			xpol_type,
								number option(nullable) 	xpol_cut )
							    :fwd_z_model(type,xpol_type,xpol_cut)
{
	fwd_ip_ = fwd_ip_map(type_,apply_hyman,left_der,hermite_right_der);
	QL_REQUIRE(!null(fwd_ip_), "failed to create the interpolator [fwd ip model]");
}

fwd_func fwd_z_model_ip.extrapol_fwd(	fwd_func 	f, 
										number 		crv_max)
{
	switch (xpol_type_) 
	{
	case fwd_extrap_type.FXP_MODEL:
		break;			
	case fwd_extrap_type.FXP_TENOR_FLAT:
		f = extrap_flat(f, crv_max );
		break;
	default:
		QL_FAIL("invalid fwd extrapolation model [fwd ip model]");
	}
	return f;
}

fwd_func fwd_z_model_ip.fwd(cashflows_fwd cf, disc_func df_disc,  curve option(nullable) )
{
	QL_REQUIRE(cf.n_instruments()>0,"invalid curve [fwd ip model]");
	fwd_func f = bootstrap_fwd(cf, df_disc, fwd_ip_);
	QL_REQUIRE(!null(f),'failed to create basic fwd_func [fwd ip model]');
	number crv_max = CORE_INT.cashflows_fwd_max_per(cf);
	QL_REQUIRE(!null(crv_max),"invalid fwd cashflow [fwd ip model]");	
	f = extrapol_model(f, crv_max);
	QL_REQUIRE(!null(f),'failed to create extrapolation fwd_func [fwd ip model]');
	return f;
}

fwd_func fwd_z_model_ip.fwd(fwd_func fwd_pre, date cutoff_d, cashflows_fwd cf, disc_func df_disc,  curve option(nullable) )
{
	QL_REQUIRE(cf.n_instruments()>0,"invalid curve [fwd ip model]");
	fwd_func f = bootstrap_fwd_ext(fwd_pre, cutoff_d, cf, df_disc, fwd_ip_);
	QL_REQUIRE(!null(f),'failed to create basic fwd_func [fwd ip model]');
	number crv_max = CORE_INT.cashflows_fwd_max_per(cf);
	QL_REQUIRE(!null(crv_max),"invalid fwd cashflow [fwd ip model]");	
	f = extrapol_model(f, crv_max);
	QL_REQUIRE(!null(f),'failed to create extrapolation fwd_func [fwd ip model]');
	return f;
}

fwd_func fwd_z_model_ip.fwd(fwd_func fwd_pre, date cutoff_d, fwd_func fwd_base, cashflows_fwd cf, disc_func df_disc,  curve option(nullable) )
{
	QL_REQUIRE(cf.n_instruments()>0,"invalid curve [fwd ip model]");
	fwd_func f = bootstrap_fwd_spr_ext(fwd_pre, cutoff_d, fwd_base, cf, df_disc, fwd_ip_);
	QL_REQUIRE(!null(f),'failed to create basic fwd_func [fwd ip model]');
	number crv_max = CORE_INT.cashflows_fwd_max_per(cf);
	QL_REQUIRE(!null(crv_max),"invalid fwd cashflow [fwd ip model]");	
	f = extrapol_model(f, crv_max);
	QL_REQUIRE(!null(f),'failed to create extrapolation fwd_func [fwd ip model]');
	return f;
}

fwd_func fwd_z_model_ip.fwd(fwd_func fwd_base, cashflows_fwd cf, disc_func df_disc,  curve option(nullable) )
{
	QL_REQUIRE(cf.n_instruments()>0,"invalid curve [fwd ip model]");
	fwd_func f = bootstrap_fwd_spr(fwd_base, cf, df_disc, fwd_ip_);
	QL_REQUIRE(!null(f),'failed to create basic fwd_func [fwd ip model]');
	number crv_max = CORE_INT.cashflows_fwd_max_per(cf);
	QL_REQUIRE(!null(crv_max),"invalid fwd cashflow [fwd ip model]");	
	f = extrapol_model(f, crv_max);
	QL_REQUIRE(!null(f),'failed to create extrapolation fwd_func [fwd ip model]');
	return f;
}
//-------------------------------------------------------
// class fwd_z_model_tension
//-------------------------------------------------------
class fwd_z_model_tension: public fwd_z_model option (category: 'Yield Curve/003 Curve Building')
{
   	public:

	fwd_z_model_tension(fwd_z_model_type,interpolation,fwd_extrap_type, number option(nullable),number option(nullable),number option(nullable)); 
	fwd_z_model_tension(interpolation, vector(number) option(nullable), z_mod_spread_type,fwd_extrap_type, number option(nullable),
						number option(nullable),number option(nullable));

	virtual fwd_func 			fwd(cashflows_fwd, disc_func, curve option(nullable) crv = null<curve>);
	virtual fwd_func 			fwd(fwd_func, date, cashflows_fwd, disc_func,  curve option(nullable) crv = null<curve>);
	virtual fwd_func 			fwd(fwd_func, date, fwd_func, cashflows_fwd, disc_func,  curve option(nullable) crv = null<curve>);
	virtual fwd_func 			fwd(fwd_func, cashflows_fwd, disc_func,  curve option(nullable) crv = null<curve>);

	vector(number)				calc_spreads(cashflows_fwd, curve option(nullable), disc_func);
	void 						update_spreads(vector(number));
	virtual fwd_z_model_type 	type();
	virtual interpolator 		fwd_ip();

	protected:
	
	virtual fwd_func			extrapol_fwd(fwd_func,number );	

	interpolation		sigma_;
	vector(number) 		spreads_;
	z_mod_spread_type	spread_t_;
	number				left_der_;
	number				right_der_;
};

fwd_z_model_type fwd_z_model_tension.type() {return type_;}
interpolator  fwd_z_model_tension.fwd_ip() 	{return null<interpolator>;}

fwd_z_model_tension.fwd_z_model_tension( 	fwd_z_model_type 		type,
											interpolation			sigma,
											fwd_extrap_type 		xpol_type,
											number option(nullable) xpol_cut,
											number option(nullable)	left_der,
											number option(nullable)	right_der)
							 				:fwd_z_model(type,xpol_type,xpol_cut), sigma_(sigma), spread_t_(z_mod_spread_type.ZMS_BP), left_der_(left_der),right_der_(right_der)
{
	QL_REQUIRE(type_ == fwd_z_model_type.FZM_TENSION_SPREAD || type_ == fwd_z_model_type.FZM_TENSION_EXACT,'invalid fwd model type');
	resize(spreads_,0);
}

fwd_z_model_tension.fwd_z_model_tension(interpolation			sigma,
										vector(number)	option(nullable) spreads,
										z_mod_spread_type		spread_t,
										fwd_extrap_type 		xpol_type,
										number option(nullable) xpol_cut,
										number option(nullable)	left_der,
										number option(nullable)	right_der)
							 			:fwd_z_model(fwd_z_model_type.FZM_TENSION_SPREAD,xpol_type,xpol_cut), sigma_(sigma), 
										 spreads_(!null(spreads) ? clone_vector(spreads) : null<vector(number)>), 
										 spread_t_(spread_t), left_der_(left_der),right_der_(right_der)
{
	integer size = v_size(spreads_);
	for(integer i = 0; i < size; i++) 
		QL_REQUIRE(!null(spreads_[i]),'invalid value in spreads vector [tension_spread model]');	
}

fwd_func fwd_z_model_tension.extrapol_fwd(	fwd_func 	f, 
											number 		crv_max)
{
	switch (xpol_type_) 
	{
	case fwd_extrap_type.FXP_MODEL:
		break;		
	case fwd_extrap_type.FXP_TENOR_FLAT:
		f = extrap_flat(f, crv_max );
		break;
	default:
		QL_FAIL("invalid fwd extrapolation model [tension model]");
	}
	return f;
}

vector(number) fwd_z_model_tension.calc_spreads(cashflows_fwd cf, curve option(nullable) crv, disc_func df_disc)
{	
	vector(number) res;

	if(v_size(spreads_) == 1) {
		if(spread_t_ == z_mod_spread_type.ZMS_BP) {
			vector(number) pvbp;
			string err_msg;
			logical ok = null(crv) ? calc_pvbp(cf, df_disc,pvbp, err_msg) : calc_pvbp(cf, crv, df_disc, pvbp, err_msg);				
			QL_REQUIRE(ok,strcat(err_msg," [tension_spread model]"));
			res = pvbp*spreads_[0];//div with 100 because cashflows_fwd has nom=1
		}
		else {			
			resize(res, cf.n_instruments());
			res = spreads_[0];				
		}
		return res;
	}
	else {
		integer size = cf.n_instruments();
		QL_REQUIRE(v_size(spreads_) == size ,'invalid size of spread vector (must match number of instruments) [tension_spread model]');
		
		/*if(size  != v_size(spreads_)) 
			return res;
		*/
		if(spread_t_ == z_mod_spread_type.ZMS_BP) {
			vector(number) pvbp;
			string err_msg;
			logical ok = null(crv) ? calc_pvbp(cf, df_disc, pvbp, err_msg) : calc_pvbp(cf, crv, df_disc, pvbp, err_msg);				
			QL_REQUIRE(ok,strcat(err_msg," [tension_spread model]"));
			resize(res, size);
			for(integer i=0;i<size;i++)
				res[i] = pvbp[i]*spreads_[i];
			return res;
		}
		else 
			return clone_vector(spreads_);
	}	
}

void fwd_z_model_tension.update_spreads(vector(number) spreads)
{
	spreads_= clone_vector(spreads);	
}

fwd_func fwd_z_model_tension.fwd(cashflows_fwd cf, disc_func df_disc, curve option(nullable) crv)
{
	QL_REQUIRE(cf.n_instruments()>0,"invalid curve [tension model]");
	fwd_func f;
	if (type_ == fwd_z_model_type.FZM_TENSION_SPREAD ) {
		vector(number) s = calc_spreads(cf, crv, df_disc);		
		QL_REQUIRE(!null(s) && v_size(s) > 0,'invalid spread vector (not initialized) [tension_spread model]');		
		QL_REQUIRE(v_size(s) == cf.n_instruments(),'invalid size of spread vector (must match number of instruments) [tension_spread model]');
		f = tension_spline_fwd_spreads(cf, s, df_disc, sigma_) ;
	}	
	else {
		f = tension_spline_fwd(cf, df_disc, sigma_) ;
	}
	QL_REQUIRE(!null(f),'failed to create basic fwd_func [tension model]');

	number crv_max = CORE_INT.cashflows_fwd_max_per(cf);	
	QL_REQUIRE(!null(crv_max),"invalid fwd cashflow [tension model]");

	f = extrapol_model(f, crv_max);
	QL_REQUIRE(!null(f),'failed to create extrapolation fwd_func [tension model]');
	return f;
}

fwd_func fwd_z_model_tension.fwd(fwd_func fwd_pre, date cutoff_d, cashflows_fwd cf, disc_func df_disc,  curve option(nullable) )
{	
	if (type_ == fwd_z_model_type.FZM_TENSION_SPREAD )  
		QL_FAIL('unsupported model for extension of fwd_func [tension_spread model]');	

	QL_REQUIRE(cf.n_instruments()>0,"invalid curve [tension model]");
	
	fwd_func f = tension_spline_fwd_ext(fwd_pre, cutoff_d, cf, df_disc, sigma_,left_der_,right_der_) ;
	
	QL_REQUIRE(!null(f),'failed to create basic fwd_func [tension model]');

	number crv_max = CORE_INT.cashflows_fwd_max_per(cf);	
	QL_REQUIRE(!null(crv_max),"invalid fwd cashflow [tension model]");

	f = extrapol_model(f, crv_max);
	QL_REQUIRE(!null(f),'failed to create extrapolation fwd_func [tension model]');
	return f;	
}

fwd_func fwd_z_model_tension.fwd(fwd_func fwd_pre, date cutoff_d, fwd_func fwd_base, cashflows_fwd cf, disc_func df_disc,  curve option(nullable) )
{	
	if (type_ == fwd_z_model_type.FZM_TENSION_SPREAD )  
		QL_FAIL('unsupported model for extension of fwd_func [tension_spread model]');	

	QL_REQUIRE(cf.n_instruments()>0,"invalid curve [tension model]");
	
	fwd_func f = tension_spline_fwd_spr_ext(fwd_pre, cutoff_d, fwd_base, cf, df_disc, sigma_,left_der_,right_der_) ;
	
	QL_REQUIRE(!null(f),'failed to create basic fwd_func [tension model]');

	number crv_max = CORE_INT.cashflows_fwd_max_per(cf);	
	QL_REQUIRE(!null(crv_max),"invalid fwd cashflow [tension model]");

	f = extrapol_model(f, crv_max);
	QL_REQUIRE(!null(f),'failed to create extrapolation fwd_func [tension model]');
	return f;	
}

fwd_func fwd_z_model_tension.fwd(fwd_func fwd_base, cashflows_fwd cf, disc_func df_disc,  curve option(nullable) crv)
{
	QL_REQUIRE(cf.n_instruments()>0,"invalid curve [tension model]");
	fwd_func f;
	if (type_ == fwd_z_model_type.FZM_TENSION_SPREAD ) {
		vector(number) s = calc_spreads(cf, crv, df_disc);		
		QL_REQUIRE(!null(s) && v_size(s) > 0,'invalid spread vector (not initialized) [tension_spread model]');		
		QL_REQUIRE(v_size(s) == cf.n_instruments(),'invalid size of spread vector (must match number of instruments) [tension_spread model]');
		f = tension_spline_fwd_spreads_spr(fwd_base, cf, s, df_disc, sigma_,left_der_,right_der_) ;
		//QL_FAIL('unsupported model for spread on fwd_func [tension_spread model]');
	}
	else {
		f = tension_spline_fwd_spr(fwd_base, cf, df_disc, sigma_,left_der_,right_der_) ;
	}

	QL_REQUIRE(!null(f),'failed to create basic fwd_func [tension model]');

	number crv_max = CORE_INT.cashflows_fwd_max_per(cf);	
	QL_REQUIRE(!null(crv_max),"invalid fwd cashflow [tension model]");

	f = extrapol_model(f, crv_max);
	QL_REQUIRE(!null(f),'failed to create extrapolation fwd_func [tension model]');
	return f;	
}
//-------------------------------------------------------
// class fwd_z_model_step_linear
//-------------------------------------------------------
class fwd_z_model_step_linear : public fwd_z_model option (category: 'Yield Curve/003 Curve Building')
{
   	public:

	fwd_z_model_step_linear(number option(nullable),fwd_extrap_type, number option(nullable)); 

	virtual fwd_func 			fwd(cashflows_fwd, disc_func, curve option(nullable) crv = null<curve>);
	virtual fwd_func 			fwd(fwd_func, date, cashflows_fwd, disc_func,  curve option(nullable) crv = null<curve>);
	virtual fwd_func 			fwd(fwd_func, date, fwd_func, cashflows_fwd, disc_func,  curve option(nullable) crv = null<curve>);
	virtual fwd_func 			fwd(fwd_func, cashflows_fwd, disc_func,  curve option(nullable) crv = null<curve>);

	virtual fwd_z_model_type 	type();
	virtual interpolator 		fwd_ip();

	protected:
	
	virtual fwd_func			extrapol_fwd(fwd_func,number );	
	fwd_func					fwd_help(fwd_func, cashflows_fwd);

	number 			left_der_;
	interpolator 	fwd_ip_;
};

fwd_z_model_type 	fwd_z_model_step_linear.type() 		{ return type_;}
interpolator  		fwd_z_model_step_linear.fwd_ip()	{ return null<interpolator>;}

fwd_z_model_step_linear.fwd_z_model_step_linear(number option(nullable) left_der,
												fwd_extrap_type 		xpol_type,
												number option(nullable) xpol_cut)
												: fwd_z_model(fwd_z_model_type.FZM_STEP_LINEAR,xpol_type,xpol_cut), left_der_(left_der)
{
	fwd_ip_ = fwd_ip_map(fwd_z_model_type.FZM_STEP, false, 0,0);
	QL_REQUIRE(!null(fwd_ip_), "failed to create the step interpolator [step_linear model]");
}

fwd_func fwd_z_model_step_linear.extrapol_fwd(	fwd_func 	f, 
												number 		crv_max)
{
	switch (xpol_type_) 
	{
	case fwd_extrap_type.FXP_MODEL:
		break;		
	case fwd_extrap_type.FXP_TENOR_FLAT:
		f = extrap_flat(f, crv_max );
		break;
	default:
		QL_FAIL("invalid fwd extrapolation model [step_linear model]");
	}
	return f;
}

fwd_func fwd_z_model_step_linear.fwd_help(fwd_func f_step, cashflows_fwd cf)
{
	interpolator fwd_ip = fwd_ip_map(fwd_z_model_type.FZM_LINEAR, false, left_der_, null<number>);
	QL_REQUIRE(!null(fwd_ip), "failed to create the linear interpolator [step_linear model]");

	vector(number) periods;
	if(!CORE_INT.cashflows_fwd_periods(cf,periods))
		QL_FAIL("failed to get periods from cashflows_fwd [step_linear model]");

	vector(number) rates = f_step.fwd(periods);
	
	fwd_func f = fwd_func_interp(periods,rates,fwd_ip);
	QL_REQUIRE(!null(f),'failed to create basic fwd_func [step_linear model]');

	number crv_max = CORE_INT.cashflows_fwd_max_per(cf);
	QL_REQUIRE(!null(crv_max),"invalid fwd cashflow [step_linear model]");

	f = extrapol_model(f, crv_max);
	QL_REQUIRE(!null(f),'failed to create extrapolation fwd_func [step_linear model]');
	return f;
}

fwd_func fwd_z_model_step_linear.fwd(cashflows_fwd cf, disc_func df_disc, curve option(nullable) )
{
	QL_REQUIRE(cf.n_instruments()>0,"invalid curve [step_linear model]");
	fwd_func f_step = bootstrap_fwd(cf, df_disc, fwd_ip_);
	QL_REQUIRE(!null(f_step), "failed to create the step fwd_func [step_linear model]");

	return fwd_help(f_step, cf);
}

fwd_func fwd_z_model_step_linear.fwd(fwd_func fwd_pre, date cutoff_d, cashflows_fwd cf, disc_func df_disc,  curve option(nullable) )
{
	QL_REQUIRE(cf.n_instruments()>0,"invalid curve [step_linear model]");
	fwd_func f_step = bootstrap_fwd_ext(fwd_pre, cutoff_d, cf, df_disc, fwd_ip_);
	QL_REQUIRE(!null(f_step), "failed to create the step fwd_func [step_linear model]");

	return fwd_help(f_step, cf);
}
fwd_func fwd_z_model_step_linear.fwd(fwd_func fwd_pre, date cutoff_d, fwd_func fwd_base, cashflows_fwd cf, disc_func df_disc,  curve option(nullable) )
{
	QL_REQUIRE(cf.n_instruments()>0,"invalid curve [step_linear model]");
	fwd_func f_step = bootstrap_fwd_spr_ext(fwd_pre, cutoff_d, fwd_base, cf, df_disc, fwd_ip_);
	QL_REQUIRE(!null(f_step), "failed to create the step fwd_func [step_linear model]");

	return fwd_help(f_step, cf);
}

fwd_func fwd_z_model_step_linear.fwd(fwd_func fwd_base, cashflows_fwd cf, disc_func df_disc,  curve option(nullable) )
{
	QL_REQUIRE(cf.n_instruments()>0,"invalid curve [step_linear model]");
	fwd_func f_step = bootstrap_fwd_spr(fwd_base, cf, df_disc, fwd_ip_);
	QL_REQUIRE(!null(f_step), "failed to create the step fwd_func [step_linear model]");

	return fwd_help(f_step, cf);
}
//-------------------------------------------------------
// fwd_z_ip
//-------------------------------------------------------
fwd_z_model_ip fwd_z_ip(fwd_z_model_type 		fwd_z,
						logical option(nullable) apply_hyman,							
						number option(nullable) left_der,
						number option(nullable) hermite_right_der,
						fwd_extrap_type 		xpol_type,
						number option(nullable) xpol_cut) 
option (category: 'Yield Curve/003 Curve Building')
{	
	return new fwd_z_model_ip(fwd_z,apply_hyman,left_der,hermite_right_der,xpol_type,xpol_cut);		
}

//-------------------------------------------------------
// fwd_z_tension
//-------------------------------------------------------
fwd_z_model_tension fwd_z_tension(	fwd_z_model_type 		type, 
									interpolation 			sigma,
									number option(nullable) left_der,
									number option(nullable) right_der,
									fwd_extrap_type 		xpol_type,
									number option(nullable) xpol_cut) 
option (category: 'Yield Curve/003 Curve Building')
{			
	return new fwd_z_model_tension(type, sigma,xpol_type,xpol_cut,left_der,right_der);		
}

//-------------------------------------------------------
// fwd_z_tension
//-------------------------------------------------------
fwd_z_model_tension fwd_z_tension(	interpolation 	sigma,
									vector(number) option(nullable) spreads,
									z_mod_spread_type	spread_t,
									number option(nullable) left_der,
									number option(nullable) right_der,
									fwd_extrap_type 		xpol_type,
									number option(nullable) xpol_cut) 
option (category: 'Yield Curve/003 Curve Building')
{			
	return new fwd_z_model_tension(sigma, spreads, spread_t,xpol_type,xpol_cut,left_der,right_der);		
}

//-------------------------------------------------------
// fwd_z_step_linear
//-------------------------------------------------------
fwd_z_model_step_linear fwd_z_step_linear(	number option(nullable) left_der, 											
											fwd_extrap_type 		xpol_type,
											number option(nullable) xpol_cut) 
option (category: 'Yield Curve/003 Curve Building')
{			
	return new fwd_z_model_step_linear(left_der, xpol_type,xpol_cut);		
}

//-------------------------------------------------------
// fwd_func_create
//-------------------------------------------------------
fwd_func fwd_func_create(fwd_z_model m, cashflows_fwd cf, disc_func df_disc)
option (category: 'Yield Curve/003 Curve Building')
{
	return m.fwd(cf, df_disc);
}




