option(null: hard);	

/*	
	Basic swap functions for "dual" bootstrap
	Parameter classes
	Developed by Algorithmica Research, Magnus Nyström
			
	Comments:

	ToDo: 
	
*/


class enum swap_curve_type 
{	
	SCE_DISC, SCE_DISC_B1,
	SCE_FWD, SCE_FWD_F, 
	SCE_FWD_B1, SCE_FWD_B1_F, 
	SCE_FWD_B2, SCE_FWD_B2_F, 
	SCE_CURR_B1, SCE_CURR_B1_F,SCE_CURR_B1_O, 
	SCE_CURR_FF, SCE_CURR_FF_F,
	SCE_FIX
};
	

/*-----------------------------------------------------------------------
  Name :    class swap_curve_ext_parm - abstract root
  Notes :   just a container for params , note: swap curve could be a moneymarket swap curve       
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
class swap_curve_ext_parm : public  curve_ext_parm  
{
	public:
		virtual swap_curve_type	type() = 0;
		curve  					fra_crv(); 
		curve  					swap_crv();					
		logical 				merge_fra();
		number  				swap_first_fix();
		virtual void 			set_fra_crv(curve option(nullable));
		void 					set_swap_crv(curve option(nullable));
		void 					extend_fra_crv(curve option(nullable) );
		
	protected:
		swap_curve_ext_parm(curve option(nullable), curve option(nullable), curve option(nullable), 
							number option(nullable) ,logical,logical,logical,curve_prio,curve_prio, 
							integer,swap_curve_type, vector(number) option(nullable),logical);
		swap_curve_ext_parm(curve option(nullable), curve option(nullable), curve option(nullable), 
							number option(nullable) ,logical,logical,logical,curve_prio,curve_prio, 
							integer,swap_curve_type, vector(number) option(nullable),vector(number) option(nullable),
							vector(number) option(nullable),vector(number) option(nullable), logical);					
		swap_curve_ext_parm(swap_curve_ext_parm);
		
		number  		swap_first_fix_;					
		swap_curve_type type_;										
};

swap_curve_ext_parm.swap_curve_ext_parm(curve option(nullable) 		short_crv,											
										curve option(nullable) 		fra_crv,
										curve option(nullable)		swap_crv,
										number option(nullable) 	swap_first_fix,										
										logical						no_overlap_fwd,
										logical						merge_fra,
										logical						merge_crv,
										curve_prio					prio1,
										curve_prio					prio2,
										integer						blend_buf_days,
										swap_curve_type 			type,
										vector(number) option(nullable) edfut_cvx_adj,
										logical 					run_prio) 
										:curve_ext_parm(short_crv,fra_crv,edfut_cvx_adj,swap_crv,no_overlap_fwd,
														merge_fra,merge_crv,prio1,prio2,blend_buf_days,run_prio ),
										swap_first_fix_(swap_first_fix), type_(type)
{}	
swap_curve_ext_parm.swap_curve_ext_parm(curve option(nullable) 		short_crv,											
										curve option(nullable) 		fra_crv,
										curve option(nullable)		swap_crv,
										number option(nullable) 	swap_first_fix,										
										logical						no_overlap_fwd,
										logical						merge_fra,
										logical						merge_crv,
										curve_prio					prio1,
										curve_prio					prio2,
										integer						blend_buf_days,
										swap_curve_type 			type,
										vector(number) option(nullable) short_crv_y_spr,
										vector(number) option(nullable) fra_crv_y_spr,
										vector(number) option(nullable) swap_crv_y_spr,
										vector(number) option(nullable) edfut_cvx_adj,
										logical 					run_prio) 
										:curve_ext_parm(short_crv,fra_crv,edfut_cvx_adj,swap_crv,no_overlap_fwd,
														merge_fra,merge_crv,prio1,prio2,blend_buf_days,
														short_crv_y_spr,swap_crv_y_spr,swap_crv_y_spr,run_prio ) ,
										swap_first_fix_(swap_first_fix), type_(type)

{}

swap_curve_ext_parm.swap_curve_ext_parm(swap_curve_ext_parm s) : curve_ext_parm(s),
										swap_first_fix_(s.swap_first_fix_),type_(s.type_)
{}

curve  		swap_curve_ext_parm.fra_crv()			{ return middle_crv_;} 
curve  		swap_curve_ext_parm.swap_crv()			{ return long_crv_;}					
logical 	swap_curve_ext_parm.merge_fra()			{ return merge_middle_;}
number  	swap_curve_ext_parm.swap_first_fix()	{ return swap_first_fix_;}
void 		swap_curve_ext_parm.set_swap_crv(curve option(nullable) s)	{ set_long_crv(s);}
void 		swap_curve_ext_parm.set_fra_crv(curve option(nullable) f) 	{ set_middle_crv(f);}

//-------------------------------------------------------
// extend_fra_crv
//-------------------------------------------------------
void swap_curve_ext_parm.extend_fra_crv(curve option(nullable) fra_ext_crv)
{
	if(!..check_curve(fra_ext_crv))
		return;

	string cname = (null(middle_crv_) || null(middle_crv_.name())) ? fra_ext_crv.name() : strcat(middle_crv_.name(),"-ext");
	if(null(cname ))
		cname  = "";

	orig_middle_crv_ = merge(fra_ext_crv, orig_middle_crv_) ;
	orig_middle_crv_.set_name(cname);
	
	middle_crv_ = merge(fra_ext_crv, middle_crv_) ;					
	middle_crv_.set_name(cname);
		
	if(null(irfut_cvx_adj_))					
		return;
			
	//integer size_orig 		= v_size(edfut_cvx_adj_);		
	vector(instrument) v_i 	= middle_crv_.instruments();
	integer size_new 		= v_size(v_i);		
	vector(number) cvx[size_new] ;
	vector(string) cvx_n[size_new] ;
	
	for(integer i = 0; i < size_new; i++){
		if(!v_i[i].is_ir_future()) {
			cvx[i] = null<number>;
			cvx_n[i] = "";
			continue;
		}			
		cvx_n[i] 			= str_to_upper(v_i[i].name());
		vector(integer) v_f = v_findi(irfut_names_,[cvx_n[i]]);
		if(v_size(v_f) == 1) {
			cvx[i] = irfut_cvx_adj_[v_f[0]];
		}
		else {
			cvx[i] = 0;
		}
	}

	irfut_cvx_adj_ = cvx;
	irfut_names_ = cvx_n;
	set_prio_order(true);
}

