option(null: hard);	

/*	
	Basic swap functions for "dual" bootstrap
	Parameter classes
	Developed by Algorithmica Research, Magnus Nyström
			
	Comments:

	ToDo: 
	
*/


/*-----------------------------------------------------------------------
  Name :    class swap_curve_b1_f_parm
  Notes :	ex when we solve for 3M:
				fra_crv=3M, basis_crv=3M6M

	ex template
		CURVE_TMPL.fwd_basis_curve_tmpl(euribor6m,"EURIBOR", "EUR3MFRA", "EUR3M6MSWAP[1SWP]",
										fwd_func_tenor.FT_3M, "Ask","Mid","Mid", true, true, "EURIBORFIX3M");
																					
  Changes:
  			Date    		Who     Detail
  			4-Aug-2016 		MN      
  ----------------------------------------------------------------------*/
class swap_curve_b1_f_parm : public  swap_curve_fwd_f_parm option (category: 'Yield Curve/003 Curve Building')
{   	
	public:
		swap_curve_b1_f_parm(	fixing_curve option(nullable), curve option(nullable), curve, curve_blend_parm,
								vector(number) option(nullable), vector(number) option(nullable),
								vector(number) option(nullable));
		swap_curve_b1_f_parm(	curve option(nullable), curve option(nullable), curve, curve_blend_parm,
								vector(number) option(nullable), vector(number) option(nullable),
								vector(number) option(nullable),number option(nullable),number option(nullable));	
	
		swap_curve_b1_f_parm(swap_curve_b1_f_parm);
		
		override swap_curve_type	type();
		override void 				set_prio_order(logical);
		curve 						basis_crv();
		number						basis_first_fix_flat();
		number						basis_first_fix_sprd();
		override string 			ccy();
		override calendar 			cal();
			
	protected:
		curve 			basis_crv_;	
		//vector(number)	basis_crv_q_spr_;
		number			basis_first_fix_flat_;
		number			basis_first_fix_sprd_;
};

swap_curve_type swap_curve_b1_f_parm.type() 		{ return type_;}
curve 			swap_curve_b1_f_parm.basis_crv() 	{ return basis_crv_;}

swap_curve_b1_f_parm.swap_curve_b1_f_parm(	fixing_curve  option(nullable) fixing_crv,
											curve option(nullable) 		fra_crv,												
											curve 						basis_crv,												
											curve_blend_parm			crv_blend_parm,																					
											vector(number) option(nullable) fra_crv_y_spr,
											vector(number) option(nullable) basis_crv_q_spr,
											vector(number) option(nullable) edfut_cvx_adj) 
											:swap_curve_fwd_f_parm(fixing_crv, fra_crv, null<curve>, crv_blend_parm, swap_curve_type.SCE_FWD_B1_F, 
																	fra_crv_y_spr, null<vector(number)>,edfut_cvx_adj,false),
										 	basis_crv_(basis_crv),/*, basis_crv_q_spr_(clone_vector(basis_crv_q_spr))*/
											 basis_first_fix_flat_(null<number>), basis_first_fix_sprd_(null<number>)
{
	init_adj(basis_crv_, basis_crv_q_spr, false);
	set_prio_order(false);
}	

swap_curve_b1_f_parm.swap_curve_b1_f_parm(	curve  option(nullable) 	short_crv,
											curve option(nullable) 		fra_crv,												
											curve 						basis_crv,												
											curve_blend_parm			crv_blend_parm,																					
											vector(number) option(nullable) fra_crv_y_spr,
											vector(number) option(nullable) basis_crv_q_spr,
											vector(number) option(nullable) edfut_cvx_adj,
											number option(nullable)		basis_first_fix_flat,
											number option(nullable)		basis_first_fix_sprd)
											:swap_curve_fwd_f_parm(short_crv, fra_crv, null<curve>, crv_blend_parm, swap_curve_type.SCE_FWD_B1_F, 
																	fra_crv_y_spr, null<vector(number)>,edfut_cvx_adj,null<number>,false),
										 	basis_crv_(basis_crv),/*, basis_crv_q_spr_(clone_vector(basis_crv_q_spr))*/
											basis_first_fix_flat_(basis_first_fix_flat), basis_first_fix_sprd_(basis_first_fix_sprd)
{
	init_adj(basis_crv_, basis_crv_q_spr, false);
	set_prio_order(false);
}

swap_curve_b1_f_parm.swap_curve_b1_f_parm(swap_curve_b1_f_parm s) 
										:swap_curve_fwd_f_parm(s), basis_crv_(s.basis_crv_)/*, basis_crv_q_spr_(clone_vector(s.basis_crv_q_spr_))*/
{}

number	swap_curve_b1_f_parm.basis_first_fix_flat() { return basis_first_fix_flat_;}
number	swap_curve_b1_f_parm.basis_first_fix_sprd()	{ return basis_first_fix_sprd_;}
		
void swap_curve_b1_f_parm.set_prio_order(logical use_orig)
{
	if(use_orig){
		prio1_ = orig_prio1_;
		prio2_ = orig_prio2_;
	}
	CORE_INT.set_prio_order(short_crv_,middle_crv_,basis_crv_,prio1_,prio2_);
}
//-------------------------------------------------------
// ccy
//-------------------------------------------------------
string swap_curve_b1_f_parm.ccy()
{
	string ccy = curve_ext_parm.ccy();
	if(!null(ccy)) 
		return ccy;
		
	if(check_curve(basis_crv_)){		
		return basis_crv_.instruments()[0].currency();					
	}	
	else {
		return null<string>;
	}
}
//-------------------------------------------------------
// cal
//-------------------------------------------------------
calendar swap_curve_b1_f_parm.cal()
{
	calendar cal = curve_ext_parm.cal();
	if(!null(cal)) 
		return cal;
		
	if(check_curve(basis_crv_)){		
		return basis_crv_.instruments()[0].calendar();					
	}	
	else {
		return null<calendar>;
	}
}
//-------------------------------------------------------
// create 
//-------------------------------------------------------
swap_curve_b1_f_parm swap_curve_b1_f_parm(	fixing_curve option(nullable)	fixing_crv,
											curve option(nullable) 		fra_crv,													
											curve 						basis_crv,
											curve_blend_parm			crv_blend_parm,																					
											vector(number) option(nullable) fra_crv_y_spr = null<vector(number)>,
											vector(number) option(nullable) basis_crv_q_spr = null<vector(number)>,
											vector(number) option(nullable) edfut_cvx_adj= null<vector(number)>) 
										
{					
	return new swap_curve_b1_f_parm(fixing_crv, fra_crv, basis_crv,  crv_blend_parm, fra_crv_y_spr, basis_crv_q_spr,edfut_cvx_adj);
}

//-------------------------------------------------------
// create 
//-------------------------------------------------------
swap_curve_b1_f_parm swap_curve_b1_f_parm(	curve option(nullable)		short_crv,
											curve option(nullable) 		fra_crv,													
											curve 						basis_crv,
											curve_blend_parm			crv_blend_parm,																					
											vector(number) option(nullable) fra_crv_y_spr 		= null<vector(number)>,
											vector(number) option(nullable) basis_crv_q_spr 	= null<vector(number)>,
											vector(number) option(nullable) edfut_cvx_adj		= null<vector(number)>,
											number option(nullable)		basis_first_fix_flat 	= null<number>,
											number option(nullable)		basis_first_fix_sprd 	= null<number>)
										
{					
	return new swap_curve_b1_f_parm(short_crv, fra_crv, basis_crv,  crv_blend_parm, fra_crv_y_spr,
									basis_crv_q_spr,edfut_cvx_adj, basis_first_fix_flat, basis_first_fix_sprd);
}

//-------------------------------------------------------
// create 
//-------------------------------------------------------
/*swap_curve_b1_f_parm swap_curve_b1_f_parm(	swap_curve_b1_f_parm			parm,
											fixing_curve  option(nullable)	fixing_crv) 
										
{
	swap_curve_b1_f_parm p = new swap_curve_b1_f_parm(parm);
	p.set_fixing_crv(fixing_crv);
	return p;	
}*/

//-------------------------------------------------------
// create 
//-------------------------------------------------------
/*swap_curve_b1_f_parm swap_curve_b1_f_parm(	swap_curve_b1_f_parm		parm,
											curve  option(nullable)		short_crv) 
										
{
	swap_curve_b1_f_parm p = new swap_curve_b1_f_parm(parm);
	p.set_short_crv(short_crv);
	return p;	
}*/