/*	
	Basic swap utility functions
	Developed by Algorithmica Research, Magnus Nyström
	
	
	Comments:

	ToDo: 	

*/

option(null: hard);	// Throw error if a null value is used in a non-nullable context

//-------------------------------------------------------
// class swap_curve_surface
//-------------------------------------------------------
class swap_curve_surface option (category: 'Yield Curve/003 Curve Building')
{
   	public:
		swap_curve_surface(date, disc_func, vector(fwd_func), vector(number), string);	
		/*day_count_method 	flt_dc();
		bd_convention 		flt_bd();
		calendar			flt_cal();
		logical				flt_eom();
		rate_type			flt_rate_type();*/
		string				currency();
		date				trade_date();
		disc_func			disc();
		number				calc_rate(date,number, date option(nullable) d = null<date> );
	protected:							
				
	date				d_;
	disc_func			df_ois_;
	vector(fwd_func)	fwd_f_;		
	vector(number)		tenor_;
	string				currency_;
	logical				extrap_ex_ois_;
		
};
//-------------------------------------------------------
// constructor
//-------------------------------------------------------
swap_curve_surface.swap_curve_surface(	date 				d,
										disc_func			df_ois,
										vector(fwd_func)	fwd_f,									
										vector(number)		tenor,
										string				ccy)
										: d_(d), df_ois_(df_ois), fwd_f_(clone_vector(fwd_f)), tenor_(clone_vector(tenor)), currency_(ccy), extrap_ex_ois_(false)
{
	sort(fwd_f_, tenor_);
	sort(tenor_, tenor_);	
}
//-------------------------------------------------------
// get
//-------------------------------------------------------
disc_func 			swap_curve_surface.disc()			{ return df_ois_;}
date 				swap_curve_surface.trade_date()		{ return d_;}
string				swap_curve_surface.currency()		{ return currency_;}
/*
day_count_method 	swap_curve_surface.flt_dc()			{ return base_crv_.flt_dc_;}
bd_convention 		swap_curve_surface.flt_bd()			{ return base_crv_.flt_bd_;}
calendar			swap_curve_surface.flt_cal()		{ return base_crv_.flt_cal_;}
logical				swap_curve_surface.flt_eom()		{ return base_crv_.flt_eom_;}
rate_type			swap_curve_surface.flt_rate_type()	{ return RT_SIMPLE;}
*/
//-------------------------------------------------------
// calc_rate
// exposing d useful when calculating for example roll down
//-------------------------------------------------------
number 	swap_curve_surface.calc_rate(	date 					per_beg, 
										number 					tenor,//if odd tenor the input should be (end-beg)/360*12
										date option(nullable) 	d)
{			
	//date mat = move_cpn_months(settle,0,tenor,BD_MOD_FOLLOWING,calendar());
	if(null(d))
		d = d_;

	logical extrap_ok = false;
	if(tenor < 0 )  
		return null<number>;

	integer s_t = v_size(tenor_);
	if( tenor > tenor_[s_t-1]){//extrap ok!!		
		if(tenor >= 12 || s_t == 1)
			return fwd_f_[s_t-1].fwd((per_beg-d)/365);//beyond 1y it is constant
		else {
			vector(number) r = [fwd_f_[s_t-2].fwd((per_beg-d)/365),fwd_f_[s_t-1].fwd((per_beg-d)/365)];				
			vector(number) ttm = [tenor_[s_t-2],tenor_[s_t-1]];
			reg_result lr = lin_reg(r, ttm, false);				
			return lr.alpha() + lr.coeffs(0) * tenor;
		}
	}
	else if(tenor < tenor_[0]){
		if(!extrap_ex_ois_){
			vector(number) t = [0, tenor_[0]];
			number r_ois = df_ois_.simple_rate((per_beg-d)/365, (per_beg-d)/365+tenor/12); 
			vector(number) r = [r_ois, fwd_f_[0].fwd((per_beg-d)/365)];			
			return CORE_INT.ip(t, r, tenor,CORE_INT.ip_type.IP_LIN,extrap_ok);			
		}
		else {			
			if(s_t == 1) 			
				return fwd_f_[0].fwd((per_beg-d)/365);	
			else {
				vector(number) r = [fwd_f_[0].fwd((per_beg-d)/365),fwd_f_[1].fwd((per_beg-d)/365)];				
				vector(number) ttm = [tenor_[0],tenor_[1]];
				reg_result lr = lin_reg(r, ttm, false);				
				return lr.alpha() + lr.coeffs(0) * tenor;
			}
		}
	}	
	else {
		vector(integer) v_f = v_find(tenor_, [tenor]);
		if(v_size(v_f) == 1) 			
			return fwd_f_[v_f[0]].fwd((per_beg-d)/365);	
		else {
			vector(number) r = fwd_f_.fwd((per_beg-d)/365);
			vector(date) tenor_end = move_month(per_beg,tenor_,true, calendar(), BD_MOD_FOLLOWING, false);//approx --> need more arguments
			vector(number) per = (tenor_end - per_beg)/360*12; 				
			return CORE_INT.ip(per, r, tenor,CORE_INT.ip_type.IP_LIN,extrap_ok);
			
		}
	}
}