/*	
	Basic swap functions for "dual" bootstrap
	Developed by Algorithmica Research, Magnus Nyström
	
	
	
	Comments:

	ToDo: 
	1) forward settlement??	
	2) rollday (maturity from instrument is adjusted --> day(issue_date)), not ok sometimes
	3) convexadj for disc function
	4) 
	5) model the spread
*/
option(null: hard);	

/*-----------------------------------------------------------------------
  Name :    class swap_curve_ext - abstract root
  Notes :           
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws)
  ----------------------------------------------------------------------*/
class swap_curve_ext : public  curve_ext
{   	
	public:									

		virtual curve			fra_curve(logical=false);
		virtual curve			swap_curve();
		virtual swap_curve_type	type() = 0;

		override curve_ext_parm  param();
	protected:

		swap_curve_ext(swap_curve_ext_parm, date, string, swap_curve_type );			
		swap_curve_ext(swap_curve_ext) ;
		
	protected:
		swap_curve_type			type_;
		swap_curve_ext_parm 	param_;
		curve 					synt_crv_;
					
};

curve_ext_parm  swap_curve_ext.param() { return param_;}

//-------------------------------------------------------
// constructor 
//-------------------------------------------------------
swap_curve_ext.swap_curve_ext(	swap_curve_ext_parm		p,
								date					d,																									
								string 					name,
								swap_curve_type			type) 
								:curve_ext(p,d,name,false), 
								type_(type),param_(p),synt_crv_(null<curve>)
{}
//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
swap_curve_ext.swap_curve_ext(swap_curve_ext s) 
							:curve_ext(s), 
							 type_(s.type_),param_(s.param_),synt_crv_(s.synt_crv_)
{}


//-------------------------------------------------------
// fra_curve
//-------------------------------------------------------
curve	swap_curve_ext.fra_curve(logical incl_synt)
{
	if(incl_synt || !check_curve(synt_crv_))
		return middle_curve();
	
	vector(instrument) vi 	= middle_curve().instruments();
	vector(instrument) vsi 	= synt_crv_.instruments();
	vector(integer) vf 		= v_findi(vi.name(),vsi.name());
	vi 						= v_exclude (vi, vf);		
		
	return curve(vi,middle_curve().name());
}

curve	swap_curve_ext.swap_curve() 	{ return long_curve();}

/*integer				swap_curve_ext.flt_freq()		option(hidden)	{ QL_FAIL("inapplicable function call");}
date 				swap_curve_ext.flt_reset_date() option(hidden)	{ QL_FAIL("inapplicable function call");}
day_count_method 	swap_curve_ext.flt_dc() 		option(hidden)	{ QL_FAIL("inapplicable function call");}
bd_convention		swap_curve_ext.flt_bd() 		option(hidden)	{ QL_FAIL("inapplicable function call");}
logical				swap_curve_ext.flt_eom() 		option(hidden)	{ QL_FAIL("inapplicable function call");}
calendar			swap_curve_ext.flt_cal() 		option(hidden)	{ QL_FAIL("inapplicable function call");}
string 				swap_curve_ext.settle_code() 	option(hidden)	{ QL_FAIL("inapplicable function call");}	
integer	 			swap_curve_ext.flt_reset_days()	option(hidden)	{ QL_FAIL("inapplicable function call");} */






