/*	
	Basic swap functions for "dual" bootstrap
	Developed by Algorithmica Research, Magnus Nyström
	
	
	
	Comments:

	ToDo: 
	
*/

option(null: hard);	

//-------------------------------------------------------
// class swap_curve_b1_ext
// class used for basis swaps quoted as 1 swap i.e. 2 float legs
//-------------------------------------------------------
class swap_curve_b1_ext: public swap_curve_fwd_ext option (category: 'Yield Curve/003 Curve Building')
{
   	public:
		//---LEGACY------
		swap_curve_b1_ext(swap_curve_ext_parm, date, vector(date) option(nullable) , integer, string);
		swap_curve_b1_ext(swap_curve_ext_parm, date, disc_func option(nullable), fwd_func ,
						  vector(date) option(nullable) ,integer, string);

		swap_curve_b1_ext(swap_curve_ext_parm, date, vector(date) option(nullable) , integer, string, fwd_func, date);
		swap_curve_b1_ext(swap_curve_ext_parm, date, disc_func option(nullable), fwd_func ,
						  vector(date) option(nullable) ,integer, string, fwd_func, date);
		//----------------

		swap_curve_b1_ext(fwd_func_tenor,swap_curve_b1_f_parm, date, disc_func option(nullable), fwd_func ,
						  vector(date) option(nullable) , fwd_func_tenor, string,
						  fwd_func option(nullable), date option(nullable));

		swap_curve_b1_ext(	fwd_func_tenor,swap_curve_b1_f_parm, date, disc_func option(nullable), fwd_func ,
							vector(date) option(nullable) , fwd_func_tenor,
							disc_func option(nullable),date,fwd_z_model ,
							string ,fwd_func  option(nullable) ,date option(nullable) );

		swap_curve_b1_ext(swap_curve_b1_ext) ;
					
	protected:
		
		void					check_b1_curve(curve);
		void 					check_underl_swaps(curve ,out string ,out string,
											out instrument option(nullable),out instrument option(nullable));
		void					init_ir_index(instrument);	
		//override cashflows_fwd 	cf_fwd(curve,vector(number) option(nullable));	
		override cashflows_fwd 	cf_fwd(vector(instrument),vector(number) option(nullable),logical);
	public:
		override swap_curve_type type();
		curve					swap_curve(fwd_z_model);
		override curve			swap_curve();
		override curve			imp_swap_curve(fwd_func);
		curve					basis_curve();
		fwd_func 				fwd_spr(fwd_z_model);
		override fwd_func 		fwd_spr(fwd_func,fwd_z_model);
		override integer		flt_freq();//i.e. the fwd_freq we are solving for	
		override integer		fix_freq();		
		override number 		first_flt_fix(date option(nullable));			
		void 					set_disc_fwd_func(disc_func ,fwd_func ) ;

	private:
		
		void			init_legacy();
		void			init(fwd_func_tenor, fwd_func_tenor,
							 disc_func option(nullable),date option(nullable),fwd_z_model option(nullable) );
		void			init_parm(curve);
		void 			init_underl_class_names(curve);
		void			init_idx(fwd_func_tenor, fwd_func_tenor);
		void			init_common(curve, fixing_curve option(nullable) );
		void			blend(curve, fixing_curve option(nullable));
		number 			first_flt_fix_other(date option(nullable));
			
		fwd_func		f_fwd_;
		number			f_fwd_freq_;//i.e. not the fwd_freq we are solving for	
		integer			flt_freq_sprd_;		
		integer			flt_freq_flat_;
		logical 		add_basis_;
		string			class_name_sprd_;
		string			class_name_flat_;
		vector(date)	synt_mat_;				
};	

swap_curve_type	swap_curve_b1_ext.type() 			{ return type_;}
curve			swap_curve_b1_ext.swap_curve()		{ QL_FAIL('invalid function call (a fwd-model is required, use other function) (b1 fwd)');}
integer 		swap_curve_b1_ext.flt_freq()		{ return f_fwd_freq_ == flt_freq_sprd_ ? flt_freq_flat_ : flt_freq_sprd_;}	
//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
swap_curve_b1_ext.swap_curve_b1_ext(swap_curve_b1_ext s) 
									:swap_curve_fwd_ext(s), f_fwd_(s.f_fwd_),f_fwd_freq_(s.f_fwd_freq_), 
								  	 flt_freq_sprd_(s.flt_freq_sprd_),flt_freq_flat_(s.flt_freq_flat_), add_basis_(add_basis_),
									 class_name_sprd_(s.class_name_sprd_), class_name_flat_(s.class_name_flat_), 
									 synt_mat_(!null(s.synt_mat_)? clone_vector(s.synt_mat_): null<vector(date)>) 
{}

//-------------------------------------------------------
// constructor
// ex basis_crv = 3M vs 6M
// [LEGACY version]
//-------------------------------------------------------
swap_curve_b1_ext.swap_curve_b1_ext(swap_curve_ext_parm		p,
									date					d,	
									disc_func option(nullable) df_disc,									
									fwd_func 				f_fwd,
									vector(date) option(nullable) synt_mat,
									integer					f_fwd_freq,
									string 					name) 
									:swap_curve_fwd_ext(p, d, df_disc, name, swap_curve_type.SCE_FWD_B1, null<fwd_func>, null<date>), //must call base-constructor without checking/blending/etc functions 
									 f_fwd_(f_fwd),f_fwd_freq_(f_fwd_freq), synt_mat_(!null(synt_mat)? clone_vector(synt_mat): null<vector(date)>) 
{
	init_legacy();			
}
//-------------------------------------------------------
// constructor
// ex basis_crv = 3M vs 6M
// [LEGACY version]
//-------------------------------------------------------
swap_curve_b1_ext.swap_curve_b1_ext(swap_curve_ext_parm		p,
									date					d,	
									disc_func option(nullable) df_disc,									
									fwd_func 				f_fwd,
									vector(date) option(nullable) synt_mat,
									integer					f_fwd_freq,
									string 					name,
									fwd_func  				ff_pre, 
									date  					cut_off) 
									:swap_curve_fwd_ext(p, d, df_disc, name, swap_curve_type.SCE_FWD_B1, ff_pre, cut_off), //must call base-constructor without checking/blending/etc functions 
									 f_fwd_(f_fwd),f_fwd_freq_(f_fwd_freq), synt_mat_(!null(synt_mat)? clone_vector(synt_mat): null<vector(date)>) 
{
	init_legacy();			
}
//-------------------------------------------------------
// constructor (no disc_func or fwd_func in this constructor), use to get info on blending for example
// [LEGACY version]
//-------------------------------------------------------
swap_curve_b1_ext.swap_curve_b1_ext(swap_curve_ext_parm		p,
									date					d,										
									vector(date) option(nullable) synt_mat,
									integer					f_fwd_freq,
									string 					name) 
									:swap_curve_fwd_ext(p, d, null<disc_func>, name, swap_curve_type.SCE_FWD_B1, null<fwd_func>, null<date>), //must call base-constructor without checking/blending/etc functions 
									 f_fwd_(null<fwd_func>),f_fwd_freq_(f_fwd_freq), synt_mat_(!null(synt_mat)? clone_vector(synt_mat): null<vector(date)>) 
{
	init_legacy();					
}
//-------------------------------------------------------
// constructor (no disc_func or fwd_func in this constructor), use to get info on blending for example
// [LEGACY version]
//-------------------------------------------------------
swap_curve_b1_ext.swap_curve_b1_ext(swap_curve_ext_parm		p,
									date					d,										
									vector(date) option(nullable) synt_mat,
									integer					f_fwd_freq,
									string 					name,
									fwd_func  				ff_pre, 
									date  					cut_off) 
									:swap_curve_fwd_ext(p, d, null<disc_func>, name, swap_curve_type.SCE_FWD_B1, ff_pre, cut_off), //must call base-constructor without checking/blending/etc functions 
									 f_fwd_(null<fwd_func>),f_fwd_freq_(f_fwd_freq), synt_mat_(!null(synt_mat)? clone_vector(synt_mat): null<vector(date)>) 
{
	init_legacy();					
}
//-------------------------------------------------------
// constructor
// ex basis_crv = 3M vs 6M
//-------------------------------------------------------
/*swap_curve_b1_ext.swap_curve_b1_ext(fwd_func_tenor 	fwd_tenor_out,//the index we're solving for
									swap_curve_b1_f_parm			p,
									date							d,	
									disc_func option(nullable) 		df_disc,									
									fwd_func 						f_fwd,
									vector(date) option(nullable) 	synt_mat,
									fwd_func_tenor 	fwd_tenor_in,//the tenor of the input fwd_func
									string 							name) 
									:swap_curve_fwd_ext(p, d, df_disc, name, SCE_FWD_B1_F, null<fwd_func>, null<date>), //must call base-constructor without checking/blending/etc functions 
									 f_fwd_(f_fwd),f_fwd_freq_(fwd_func_tenor.to_freq(fwd_tenor_in)), synt_mat_(!null(synt_mat)? clone_vector(synt_mat): null<vector(date)>) 
{
	init(fwd_tenor_out, fwd_tenor_in);			
}*/
//-------------------------------------------------------
// constructor
// ex basis_crv = 3M vs 6M
//-------------------------------------------------------
swap_curve_b1_ext.swap_curve_b1_ext(fwd_func_tenor 	fwd_tenor_out,
									swap_curve_b1_f_parm			p,
									date							d,	
									disc_func option(nullable) 		df_disc,									
									fwd_func 						f_fwd,
									vector(date) option(nullable) 	synt_mat,
									fwd_func_tenor 	fwd_tenor_in,
									string 							name,
									fwd_func  option(nullable)		ff_pre, 
									date  	option(nullable)		cut_off) 
									:swap_curve_fwd_ext(p, d, df_disc, name, swap_curve_type.SCE_FWD_B1_F, ff_pre, cut_off), //must call base-constructor without checking/blending/etc functions 
									 f_fwd_(f_fwd),f_fwd_freq_(fwd_func_tenor_to_freq(fwd_tenor_in)),
									 synt_mat_(!null(synt_mat)? clone_vector(synt_mat): null<vector(date)>)
{
	init(fwd_tenor_out, fwd_tenor_in, null<disc_func>,null<date>,null<fwd_z_model>);			
}

//-------------------------------------------------------
// constructor
// ex basis_crv = 3M vs 6M
//-------------------------------------------------------
swap_curve_b1_ext.swap_curve_b1_ext(fwd_func_tenor 	fwd_tenor_out,
									swap_curve_b1_f_parm			p,
									date							d,	
									disc_func option(nullable) 		df_disc,									
									fwd_func 						f_fwd,
									vector(date) option(nullable) 	synt_mat,
									fwd_func_tenor 	fwd_tenor_in,
									disc_func option(nullable)		df_synt_base,
									date 							synt_end,
									fwd_z_model  					fwd_model_synt,
									string 							name,
									fwd_func  option(nullable)		ff_pre, 
									date  	option(nullable)		cut_off) 
									:swap_curve_fwd_ext(p, d, df_disc, name, swap_curve_type.SCE_FWD_B1_F, ff_pre, cut_off), //must call base-constructor without checking/blending/etc functions 
									 f_fwd_(f_fwd),f_fwd_freq_(fwd_func_tenor_to_freq(fwd_tenor_in)),
									 synt_mat_(!null(synt_mat)? clone_vector(synt_mat): null<vector(date)>)
{
	init(fwd_tenor_out, fwd_tenor_in,null(df_synt_base) ? df_disc : df_synt_base, synt_end,fwd_model_synt);			
}
//-------------------------------------------------------
// init_underl_class_names
//-------------------------------------------------------
void swap_curve_b1_ext.init_underl_class_names(curve basis_crv)
{
	try {
		instrument  sw_flat, sw_base;
		check_underl_swaps(basis_crv, class_name_flat_, class_name_sprd_, sw_flat, sw_base);
	}
	catch {
		class_name_flat_ = class_name_sprd_ = null<string>;
	}
}
//-------------------------------------------------------
// init_common
//-------------------------------------------------------
void swap_curve_b1_ext.init_common(curve basis_crv, fixing_curve option(nullable) f)
{

	blend(basis_crv, f);				
	
	if(!null(ff_pre_)){	
		QL_REQUIRE(!null(cut_off_) && cut_off_ > d_,'invalid cut off date (b1 fwd)');
		vector(number) y_adj;
		logical ok = chop_short(cut_off_, curve_orig_, null<vector(number)>, curve_, y_adj);	
		QL_REQUIRE(ok && check_curve(curve_),'invalid blended fwd curve (chopped curve) (b1 fwd)');					
	}
	
	cf_ = cf_fwd(curve_.instruments(),null<vector(number)>,false);
	QL_REQUIRE(!null(cf_),'failed to create cashflows_fwd (fwd)');	
}
//-------------------------------------------------------
// init   [LEGACY]
//-------------------------------------------------------
void swap_curve_b1_ext.init_legacy()
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_FWD_B1,'invalid type of param object (should be a b1 fwd param)');	
	swap_curve_b1_parm param = dynamic_cast<swap_curve_b1_parm>(param_);

	QL_REQUIRE(check_curve(param.basis_crv_),'invalid basis curve (b1 fwd)');	
	QL_REQUIRE(!check_curve(param_.swap_crv()),'swap curve not applicable (only short, fra/fut/sps and basis curve should be defined) (b1 fwd)');

	init_ir_index(param.basis_crv_.instruments()[0]);

	curve_ext.check();
	check_short_curve_types();
	check_fra_curve_types();
	check_fra_instr(flt_freq(), flt_dc());
	check_b1_curve(param.basis_crv_);
	init_underl_class_names(param.basis_crv_);
	add_basis_ = f_fwd_freq_ == flt_freq_sprd_;
	init_common(param.basis_crv_, null<fixing_curve>);
}
//-------------------------------------------------------
// init_idx
//-------------------------------------------------------
void swap_curve_b1_ext.init(fwd_func_tenor 		fwd_tenor_out, 
								fwd_func_tenor 	fwd_tenor_in,
								disc_func option(nullable)		df_synt_base,
								date 	option(nullable)		synt_end,
								fwd_z_model option(nullable) 	fwd_model_synt)
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_FWD_B1_F,'invalid type of param object (should be a b1 fwd param)');
	swap_curve_b1_f_parm param = dynamic_cast<swap_curve_b1_f_parm>(param_);
	
	QL_REQUIRE(check_curve(param.basis_crv()),'invalid basis curve (b1 fwd)');	
	QL_REQUIRE(!check_curve(param_.swap_crv()),'swap curve defined but not applicable (only short, fra/fut/sps and basis curve should be defined) (b1 fwd)');
	curve_ext.check();
	check_short_curve_types();//depo, depoindex,fixflt
	check_fra_curve_types();//irfut,fra or fixflt
	
	init_idx(fwd_tenor_out,   fwd_tenor_in);
	
	check_fra_instr(flt_freq(), flt_dc());
	check_b1_curve(param.basis_crv());
	init_underl_class_names(param.basis_crv());
	add_basis_ = f_fwd_freq_ == flt_freq_sprd_;
	
	logical incl_synt = !null(df_synt_base) && !null(synt_end);
	if(incl_synt) {
		QL_REQUIRE(!null(fwd_model_synt), "invalid fwd_model_synt (required when creating synthetic instruments)");

		//swap_curve_b1_ext s_t 	= new swap_curve_b1_ext(fwd_tenor_out,param, d_, df_disc_, f_fwd_,
			//											synt_mat_, fwd_tenor_in, "", null<fwd_func>, null<date>);
		//curve swap_out_crv 		= s_t.swap_curve(fwd_model_synt);
		
		blend(param.basis_crv(), param.fixing_crv());				
		cf_ 				= cf_fwd(curve_.instruments(),null<vector(number)>,false);
		QL_REQUIRE(!null(cf_),'failed to create cashflows_fwd (fwd)');
		fwd_func f 			= fwd_model_synt.fwd(cf_, df_disc_, curve_);
		curve swap_out_crv 	= imp_swap_curve(f);
		curve df_basis_crv;								
		disc_func_spr df_basis = CORE_INT.df_tenor_spread(	d_,df_synt_base, synt_end, param.short_crv(), 
															param.fra_crv(), swap_out_crv,param.prio1(),param.prio2(),
															param.merge_fra(),param.blend_buf_days(), df_basis_crv);
		QL_REQUIRE(!null(df_basis), "failed to create basis spread function");
		
		synt_crv_ = CORE_INT.synt_curve_create(d_,param.fra_crv(), fwd_idx_, df_basis, df_basis_crv);

		param.extend_fra_crv(synt_crv_);
	}
	
	
	init_common(param.basis_crv(), param.fixing_crv());
}

//-------------------------------------------------------
// init_idx
//-------------------------------------------------------
void swap_curve_b1_ext.init_idx(fwd_func_tenor fwd_tenor_out, 
								fwd_func_tenor fwd_tenor_in)
{
		
	swap_curve_b1_f_parm param = dynamic_cast<swap_curve_b1_f_parm>(param_);

	
	if(!null(param.fixing_crv())) {
		string fwd_tenor_out_s	= string(fwd_tenor_out);
		fixing_curve new_fixing_crv;
		logical upd = CORE_INT.b1_init(d_, param.fixing_crv(),fwd_tenor_in,fwd_tenor_out,fwd_tenor_out_s,fwd_idx_, new_fixing_crv); 	
		if(upd) {
			param.set_fixing_crv(new_fixing_crv); 	
		}
			
	}
	else {
		fwd_idx_ = null<ir_index>;
		curve new_short_crv;
		logical upd = CORE_INT.b1_init_depo( param.short_crv(),fwd_tenor_in,fwd_tenor_out, new_short_crv); 	
		if(upd) {
			param.set_short_crv(new_short_crv); 		
		}
	}
	
	init_parm(param.basis_crv());
	
}
//-------------------------------------------------------
// init_parm
//-------------------------------------------------------
void swap_curve_b1_ext.init_parm(curve basis_crv )
{
	instrument basis_inst = basis_crv.instruments()[0];
		
	flt_freq_sprd_ 	= basis_inst.swap_reset_freq_leg1();
	QL_REQUIRE(flt_freq_sprd_>0 && flt_freq_sprd_ != 365,'invalid float freq (spread leg) in swap curve (b1 fwd)',basis_inst.name(),true);
	
	flt_freq_flat_ 	= basis_inst.swap_reset_freq_leg2();
	QL_REQUIRE(flt_freq_flat_ > 0 && flt_freq_flat_ != 365,'invalid float freq (flat leg) in swap curve (b1 fwd)',basis_inst.name(),true);

	QL_REQUIRE(flt_freq_sprd_ != flt_freq_flat_,'invalid basis swap (same float freq in both legs) (b1 fwd)'); //i.e. for same currency only
	QL_REQUIRE(f_fwd_freq_ == flt_freq_sprd_ || f_fwd_freq_ == flt_freq_flat_, "invalid float freq (basis swap curve not consistent with fwd_tenor) (basis)");

	if(null(fwd_idx_)) {
		if(flt_freq() == flt_freq_flat_)
			fwd_idx_ =  basis_inst.swap_ir_index_leg2() ;
		else
			fwd_idx_ =  basis_inst.swap_ir_index_leg1();
	}
}
//-------------------------------------------------------
// check_b1_curve
//-------------------------------------------------------
void swap_curve_b1_ext.check_b1_curve(curve basis_crv)
{
	QL_REQUIRE(check_curve(basis_crv),'invalid basis curve (b1 fwd)');
	string err_name;
	
	//check type	
	QL_REQUIRE(CORE_INT.check_instr_type(basis_crv, instr_type.SWAP_FLTFLT,err_name),'invalid basis swap curve (wrong instrument type) (b1 fwd)', err_name, true);

	
	vector(instrument) vi 	= basis_crv.instruments();
	instrument bi 			= vi[0];
	//start
	date settle  			= date_code(bi.settle_code()).apply_fwd(d_, bi.calendar());		
	QL_REQUIRE(CORE_INT.check_bswap_issue(vi, settle, err_name ),'invalid basis swap (mixed start dates) (b1 fwd)', err_name, true);				

	//ccy
	string curr = param_.ccy();
	QL_REQUIRE(CORE_INT.swap_currency_check(vi, curr, true, err_name),'invalid basis swap curve (mixed currencies, sprd leg) (b1 fwd)', err_name, true);			
	QL_REQUIRE(CORE_INT.swap_currency_check(vi, curr, false,err_name),'invalid basis swap curve (mixed currencies, flat leg) (b1 fwd)', err_name, true);			
	
	//reset_freq
	integer flt_freq_sprd = -1;		
	QL_REQUIRE(CORE_INT.swap_rfreq_check(vi,flt_freq_sprd,true,err_name),'invalid basis swap (mixed reset frequencies, spread leg) (b1 fwd)', err_name, true);
	QL_REQUIRE(flt_freq_sprd > 0 && flt_freq_sprd != 365,'invalid float reset freq (sprd leg) in basis swap (b1 fwd)',bi.name(),true);
	
	integer flt_freq_flat = -1;	
	QL_REQUIRE(CORE_INT.swap_rfreq_check(vi,flt_freq_flat,false,err_name),'invalid basis swap (mixed reset frequencies, flat leg) (b1 fwd)', err_name, true);
	QL_REQUIRE(flt_freq_flat > 0 && flt_freq_flat != 365,'invalid float reset freq (flat leg) in basis swap (b1 fwd)',bi.name(),true);
	
	//tenor_code	
	tenor_code t = tenor_code(flt_freq_sprd);	
	QL_REQUIRE(CORE_INT.swap_index_tenor_code_check(vi, t, true,err_name),'invalid float tenor (spread leg) in basis swap (b1 fwd)', err_name, true);				
	t = tenor_code(flt_freq_flat);	
	QL_REQUIRE(CORE_INT.swap_index_tenor_code_check(vi, t, false,err_name),'invalid float tenor (flat leg) in basis swap (b1 fwd)', err_name, true);			
		
	//QL_REQUIRE(CORE_INT.check_class_flat(vi,vi[0].instr_def().fltflt_swap_class_flat(),err_name),'invalid basis swap curve (class name flat) (b1 fwd)', err_name, true);	
	//QL_REQUIRE(CORE_INT.check_class_base(vi,vi[0].instr_def().fltflt_swap_class_sprd(),err_name),'invalid basis swap curve (class name base) (b1 fwd)', err_name, true);

	//cpn freq
	integer cfreq = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,cfreq,true,err_name),'invalid basis swap (mixed coupon frequencies - sprd leg) (b1 fwd)',err_name,true);
	cfreq = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,cfreq,false,err_name),'invalid basis swap (mixed coupon frequencies - flat leg) (b1 fwd)',err_name,true);
	
	//quotes
	QL_REQUIRE(check_quote(basis_crv,err_name),'invalid quote(s) in basis curve (b1 fwd)',err_name,true);
}
//-------------------------------------------------------
// check_underl_swap
//-------------------------------------------------------
void swap_curve_b1_ext.check_underl_swaps(curve basis_crv, 
										out string class_name_flat, 
										out string class_name_sprd,
										out instrument option(nullable) sw_flat,
										out instrument option(nullable) sw_sprd)
{	
	
	vector(instrument) vi = basis_crv.instruments();
	instrument inst 	= vi[0];
	instr_def def		= inst.instr_def();
	QL_REQUIRE(!null(def),"invalid instr_def");
	
	class_name_flat 	= def.fltflt_swap_class_flat();
	QL_REQUIRE(!null(class_name_flat),"invalid class name (flat-leg)");
	
	class_name_sprd 	= def.fltflt_swap_class_sprd();
	QL_REQUIRE(!null(class_name_sprd),"invalid class name (sprd-leg)");
	
	string err_name;		
	QL_REQUIRE(CORE_INT.check_eq_class_flat(vi,class_name_flat,err_name),'incomplete basis swap curve (class name flat-leg) (b1 fwd)', err_name, true);	
	QL_REQUIRE(CORE_INT.check_eq_class_sprd(vi,class_name_sprd,err_name),'incomplete basis swap curve (class name sprd-leg) (b1 fwd)', err_name, true);
	
	if(!null(class_name_flat)) {
		sw_flat = CORE_INT_FIXFLT.create_class_fixflt(class_name_flat);
		QL_REQUIRE(!null(sw_flat),'invalid swap creation from class - flat leg (b1 fwd)',class_name_flat,true);

		integer flt_freq_sw_flat = sw_flat.swap_reset_freq_leg2();
		//QL_REQUIRE(flt_freq_sw_flat>0 && flt_freq_sw_flat!= 365,'invalid float freq in swap curve (b1 fwd)',inst.name(),true);
		QL_REQUIRE(inst.swap_reset_freq_leg2() == flt_freq_sw_flat,'invalid float freq in swap curve (inconsistency between instrument and swap class (flat)) (b1 fwd)',class_name_flat,true);
	}
	
	if(!null(class_name_sprd)) {
		sw_sprd = CORE_INT_FIXFLT.create_class_fixflt(class_name_sprd);
		QL_REQUIRE(!null(sw_sprd),'invalid swap creation from class - spread leg (b1 fwd)',class_name_sprd,true);

		integer flt_freq_sw_sprd = sw_sprd.swap_reset_freq_leg2();
		//QL_REQUIRE(flt_freq_sw_sprd>0 && flt_freq_sw_sprd != 365,'invalid float freq in swap curve (b1 fwd)',inst.name(),true);
		QL_REQUIRE(inst.swap_reset_freq_leg1() == flt_freq_sw_sprd,'invalid float freq in swap curve (inconsistency between instrument and swap class (spread)) (b1 fwd)',class_name_sprd,true);	
		
	}
	
}
/*
//-------------------------------------------------------
// init_underl_swaps
//-------------------------------------------------------
void swap_curve_b1_ext.init_underl_swaps(	curve 			basis_crv,
											out instrument option(nullable) sw_flat,
											out instrument option(nullable) sw_base)
{
	instrument inst 	= basis_crv.instruments()[0];
	class_name_flat_ 	= inst.instr_def().fltflt_swap_class_flat();
	class_name_base_ 	= inst.instr_def().fltflt_swap_class_sprd();
	
	if(!null(class_name_flat_)) {
		sw_flat = CORE_INT.create_class_fixflt_par(class_name_flat_);
		QL_REQUIRE(!null(sw_flat),'invalid swap creation from class - flat leg (b1 fwd)',class_name_flat_,true);

		integer flt_freq_sw_flat = sw_flat.swap_coupon_freq_leg2();
		QL_REQUIRE(flt_freq_sw_flat>0 && flt_freq_sw_flat!= 365,'invalid float freq in swap curve (b1 fwd)',inst.name(),true);
		QL_REQUIRE(inst.swap_coupon_freq_leg2() == flt_freq_sw_flat,'invalid float freq in swap curve (inconsistency between instrument and swap class (flat)) (b1 fwd)',class_name_flat_,true);
	}
	
	if(!null(class_name_base_)) {
		sw_base = CORE_INT.create_class_fixflt_par(class_name_base_);
		QL_REQUIRE(!null(sw_base),'invalid swap creation from class - base leg (b1 fwd)',class_name_base_,true);

		integer flt_freq_sw_base = sw_base.swap_coupon_freq_leg2();
		QL_REQUIRE(flt_freq_sw_base>0 && flt_freq_sw_base != 365,'invalid float freq in swap curve (b1 fwd)',inst.name(),true);
		QL_REQUIRE(inst.swap_coupon_freq_leg1() == flt_freq_sw_base,'invalid float freq in swap curve (inconsistency between instrument and swap class (spread)) (b1 fwd)',class_name_base_,true);	
		
	}		
}*/
//-------------------------------------------------------
// init_ir_index 
// [[LEGACY version]]
//-------------------------------------------------------
void swap_curve_b1_ext.init_ir_index(instrument basis_inst)
{	
	flt_freq_sprd_ 	= basis_inst.swap_reset_freq_leg1();
	QL_REQUIRE(flt_freq_sprd_>0 && flt_freq_sprd_ != 365,'invalid flt_freq_sprd in swap curve (b1 fwd)',basis_inst.name(),true);
	
	flt_freq_flat_ 	= basis_inst.swap_reset_freq_leg2();
	QL_REQUIRE(flt_freq_flat_ > 0 && flt_freq_flat_ != 365,'invalid flt_freq_flat in swap curve (b1 fwd)',basis_inst.name(),true);

	if(f_fwd_freq_!= flt_freq_sprd_ && f_fwd_freq_!= flt_freq_flat_)
		QL_FAIL('invalid f_fwd_freq (b1 fwd)');
				
	QL_REQUIRE(flt_freq_sprd_ != flt_freq_flat_,'invalid basis swap (same float freq in both legs) (b1 fwd)'); //i.e. for same currency only

	integer freq		= f_fwd_freq_ == flt_freq_sprd_ ? flt_freq_flat_ : flt_freq_sprd_;
	//set_flt_freq(freq);

	day_count_method dc = basis_inst.accr_dc_method();  
	QL_REQUIRE(!null(dc),'invalid daycount method (b1 fwd)',basis_inst.name(),true);
	//set_flt_dc(dc);	
 
	bd_convention bd	= basis_inst.pmt_bus_day_d(BD_MOD_FOLLOWING);
	QL_REQUIRE(!null(bd),'invalid business day convention (b1 fwd)',basis_inst.name(),true);	
	//set_flt_bd(bd);
			
	logical eom			= basis_inst.eom_d(true);
	//set_flt_eom(eom);

	calendar cal 		= basis_inst.calendar();
	string cal_name 	= basis_inst.calendar_name();
	QL_REQUIRE(!null(cal),'invalid calendar (b1 fwd)',basis_inst.name(),true);
	//set_flt_cal(cal);

	integer rdays		= integer(basis_inst.flt_reset_days_d(2));
	QL_REQUIRE(rdays >= 0,'invalid flt_reset_days in swap curve (b1 fwd)',basis_inst.name(),true);	
		
	string ccy 			= basis_inst.currency();
	//set_ccy(ccy);
	integer spot_days 	= rdays;
	integer value_days 	= spot_days;
	string tenor 		= strcat(str(12/freq),"M");

	fwd_idx_ = ir_index_ibor(	strcat("Ibor-",ccy),tenor,value_days,rdays,
							ccy, cal,cal_name,dc,bd,eom,spot_days);			
	QL_REQUIRE(!null(fwd_idx_),'failed to create ir_index (b1 fwd)');	
}		
//-------------------------------------------------------
// blend
//-------------------------------------------------------
void swap_curve_b1_ext.blend( curve basis_crv, fixing_curve option(nullable) fix_crv)
{		
	curve short_adj_c ;
	
	if(type_ == swap_curve_type.SCE_FWD_B1) {//legacy
		if(!null(param_.short_crv()))	
			short_adj_c = CORE_INT_COMPAT.extract_tenor_depo(param_.short_crv(),12/flt_freq(),flt_cal(),flt_eom(),flt_bd());
		else
			short_adj_c = param_.short_crv();
	}
	else {
		if(!null(fix_crv)){//fixing curve has 2 instruments for b1
			instrument fix_inst	= fix_crv.find_inst(fwd_idx_.tenor(), d_);
			short_adj_c 		= curve([fix_inst]);
		}
		else {
			short_adj_c	= CORE_INT.extract_tenor_depo(param_.short_crv(),fwd_idx_.tenor());
		}
	}

	string name = basis_crv.name();
	if(null(name))
		name = "basis_noname";
	curve swap_adj_c = CORE_INT.create_adj_basis1_curve(d_,basis_crv,synt_mat_,strcat(name,"-adj"));
	
	if(!null(first_flt_fix(null<date>)))		
		swap_adj_c = swap_curve_fix_adj(swap_adj_c);	
		
	/*vector(date) knots;	
	if(param_.prio1()==curve_prio.MIDDLE || (param_.prio2()== curve_prio.MIDDLE && param_.prio1()== curve_prio.SHORT))
		CORE_INT.blend_adj_long(short_adj_c, param_.fra_crv() , swap_adj_c, param_.prio1(), param_.prio2(),param_.blend_buf_days(),short_adj_c,swap_adj_c);
	
	curve_ = curve_orig_ = CORE_INT.blend_curve(short_adj_c, param_.fra_crv() , swap_adj_c, param_.prio1(), param_.prio2(), param_.merge_fra(),
								param_.no_overlap_fwd(), param_.merge_crv(), knots, param_.blend_buf_days());*/

	curve_ = curve_orig_ = CORE_INT.blender_adj_long_curve(	short_adj_c, param_.fra_crv() , swap_adj_c, param_.prio1(), 
													param_.prio2(),param_.merge_fra(), param_.merge_crv(), param_.no_overlap_fwd(),
													param_.blend_buf_days());

	
	if(null(curve_)) {
		curve_msg_ 	= "null curve (swap_curve_b1_ext-blend)";		
		return;
	}

	curve_ = curve_orig_ = curve_.set_name(strcat(name_,":blended curve"));
	
	string n;
	logical nq = check_quote(curve_,n);
	
	if(!nq) {
		curve_msg_ = strcat(["null quotes in curve (swap_curve_b1_ext-blend) [",n,"]"]);
		//curve_msg_ 	= "null quotes in curve (swap_curve_b1_ext-blend)";		
		return;
	}
		
	curve_msg_ 	= check_curve(curve_) ? "" : "empty curve (swap_curve_b1_ext-blend)";	
}
//-------------------------------------------------------
// set_disc_func
//-------------------------------------------------------
void swap_curve_b1_ext.set_disc_fwd_func(	disc_func df,
											fwd_func f) 									
{						
	df_disc_ = df;
	f_fwd_ = f;
	QL_REQUIRE(curve_ok(),'failed to initialize calculations (no curve is initialized in object) (fwd)');		
	cf_ = cf_fwd(curve_.instruments(),null<vector(number)>,false);
	QL_REQUIRE(!null(cf_),'failed to create cashflows_fwd (fwd)');			
}
	
//-------------------------------------------------------
// swap_curve
//-------------------------------------------------------
curve swap_curve_b1_ext.swap_curve(fwd_z_model fwd_model)
{ 
	fwd_func f = fwd(fwd_model);
	return imp_swap_curve(f);
}
//-------------------------------------------------------
// basis_curve
//-------------------------------------------------------
curve swap_curve_b1_ext.basis_curve()
{ 
	if(param_.type() == swap_curve_type.SCE_FWD_B1_F) {
		swap_curve_b1_f_parm param = dynamic_cast<swap_curve_b1_f_parm>(param_);
		return param.basis_crv();
	}
	else if(param_.type() == swap_curve_type.SCE_FWD_B1) {
		swap_curve_b1_parm param = dynamic_cast<swap_curve_b1_parm>(param_);
		return param.basis_crv_;
	}
	else
		QL_FAIL("invalid function call");
}
//-------------------------------------------------------
// imp_swap_curve
//-------------------------------------------------------
curve swap_curve_b1_ext.imp_swap_curve(fwd_func f_flt_idx)
{ 	
	string class_name;
	if(f_fwd_freq_ == flt_freq_sprd_){//solve for flat
		QL_REQUIRE(!null(class_name_flat_),'missing irs class name tag - flat leg (b1 fwd)');
		class_name = class_name_flat_;			
	}
	else {
		QL_REQUIRE(!null(class_name_sprd_),'missing irs class name tag - spread leg (b1 fwd)');
		class_name = class_name_sprd_;
	}
	if(null(class_name))
		return null<curve>;

	swap_fixflt class_swap = CORE_INT_FIXFLT.create_class_fixflt(class_name);
	//Note: the class-swap may have a non-matching cpn freq (but reset_freq is always a match)
	if(null(class_swap)){
		CORE_INT.write_warning_message('invalid swap (creating from class_name)');	
		return null<curve>;
	}
	
	curve b_curve 			= basis_curve();
	vector(instrument) vi 	= b_curve.instruments();
	integer size 			= v_size(vi);
	vector(instrument) sw[size];
	
	
	error_info error 		= new error_info(true,false);

	/*integer cfreq_fix, cfreq_flt;
	class_swap.coupon_freq(cfreq_fix, cfreq_flt, error);
	QL_REQUIRE(!error.is_error(), error.message());
	integer cfreq_flt_b = f_fwd_freq_ == flt_freq_sprd_ ? vi[0].swap_coupon_freq_leg2() : vi[0].swap_coupon_freq_leg1() ;

	instr_def def;
	logical cfreq_conv;
	if(cfreq_flt != cfreq_flt_b) {
		def = class_swap.instr_def();
		string n = strcat(def.name(),"-cfreq-conv");
		def = instr_def(def,n,"1Y",100,null<number>,cfreq_flt_b,null<number>,false,error);
		QL_REQUIRE(!error.is_error(), error.message());
		cfreq_conv = true;
	}
	else {
		def = class_swap.instr_def();
		cfreq_conv = false;
	}*/

	instr_def def = class_swap.instr_def();
	
	for (integer i = 0 ; i < size ; i++) {

		string mat_code = vi[i].maturity_code(error);
		string name;
		string curr = vi[i].currency();
		if(!null(mat_code))
			name = strcat([curr,CORE_INT.clean_mat_code(mat_code)," - implied"]);
		else	
			name = strcat([curr,string(year(vi[i].maturity()))," - implied"]);

		//if(cfreq_conv)
			//name = strcat(name,"(cfreq-conv)");
		
		CORE_INT.reset_single(error);
		number first_fix = first_flt_fix(vi[i].settle_date());
		swap_fixflt swp = swap_fixflt_plain(def, name, d_,vi[i].settle_date(),null<string>,
											vi[i].maturity(),null<string>,null<number> ,
											100,true,df_disc_,	f_flt_idx,first_fix, -1,error);
	
		if(error.is_error())
			sw[i] = null<swap_fixflt>;
		else
			sw[i] = swp;
	}
	
	return curve(sw,"created fix-flt curve (b1)");
	//return CORE_INT.calc_swap_from_basis1(d_,df_disc_, class_name, basis_curve(),f_flt_idx,flt_reset_days(),first_flt_fix(),"created irs curve (b1)");
}

//-------------------------------------------------------
// first_flt_fix (for the index we solve for)
//-------------------------------------------------------
number swap_curve_b1_ext.first_flt_fix(date option(nullable) settle)
{
	date rd = null(settle) ? flt_reset_date() : flt_reset_date(settle);
	
	if(null(rd) || rd > d_ )
		return null<number>;
	
	if(type_ == swap_curve_type.SCE_FWD_B1_F){
		swap_curve_b1_f_parm param = dynamic_cast<swap_curve_b1_f_parm>(param_);
		
		if(null(param.fixing_crv())){
			return f_fwd_freq_ == flt_freq_sprd_ ? param.basis_first_fix_flat() : param.basis_first_fix_sprd();	
		}
		else {
			string 	err_msg;
			return param.fixing_crv().fixing_rate(d_, fwd_idx_.tenor(), err_msg);
		}
	}
	else if(type_ == swap_curve_type.SCE_FWD_B1){
		swap_curve_b1_parm param = dynamic_cast<swap_curve_b1_parm>(param_);
		return f_fwd_freq_ == flt_freq_sprd_ ? param.basis_first_fix_flat_ : param.basis_first_fix_base_;		
	}
	else
		QL_FAIL("invalid function call");
}


//-------------------------------------------------------
// first_flt_fix_other
//-------------------------------------------------------
number swap_curve_b1_ext.first_flt_fix_other(date option(nullable) settle)
{
	date rd = null(settle) ? flt_reset_date() : flt_reset_date(settle);
	
	if(null(rd) || rd > d_ )
		return null<number>;
	
	if(type_ == swap_curve_type.SCE_FWD_B1_F){
		swap_curve_b1_f_parm param = dynamic_cast<swap_curve_b1_f_parm>(param_);
		if(null(param.fixing_crv())){
			return f_fwd_freq_ == flt_freq_sprd_ ? param.basis_first_fix_sprd() : param.basis_first_fix_flat();
		}
		else {
			string 	err_msg;
			string tenor = strcat(str(12/f_fwd_freq_),"M");
			return param.fixing_crv().fixing_rate(d_, tenor, err_msg);
		}
	}
	else if(type_ == swap_curve_type.SCE_FWD_B1){
		swap_curve_b1_parm param = dynamic_cast<swap_curve_b1_parm>(param_);		
		return f_fwd_freq_ == flt_freq_sprd_ ? param.basis_first_fix_base_ : param.basis_first_fix_flat_;		
	}
	else
		QL_FAIL("invalid function call");
}
//-------------------------------------------------------
// fix_freq
// for the fixed leg in the swap we solve for
//-------------------------------------------------------
integer	swap_curve_b1_ext.fix_freq()
{ 	
	//string class_name_flat, class_name_base;
	//instrument sw_flat,sw_base;
	//check_underl_swaps(basis_curve(),class_name_flat,class_name_base,sw_flat,sw_base);

	string class_name; 
	if(f_fwd_freq_ == flt_freq_sprd_) {//solve for flat
		QL_REQUIRE(!null(class_name_flat_),'invalid class name tag - flat leg (b1 fwd)');
		class_name = class_name_flat_ ; 
	}
	else {
		QL_REQUIRE(!null(class_name_sprd_),'invalid class name tag - spread leg (b1 fwd)');
		class_name = class_name_sprd_;
	}
	integer f1,f2;
	CORE_INT_FIXFLT.create_class_fixflt(class_name).coupon_freq(f1,f2);
	return integer(f1);	
}

//-------------------------------------------------------
// cf_fwd
//-------------------------------------------------------
cashflows_fwd swap_curve_b1_ext.cf_fwd( vector(instrument) src,										
										vector(number) option(nullable) 	y_adj,
										logical require_all_cf)
{
	integer size = v_size(src);
	vector(number) first_fix[size];
	vector(number) first_fix_b1_flat[size];
	
	for(integer i = 0; i < size ; i++){
		if(src[i].is_swap_fixflt()){
			first_fix[i] = 	first_flt_fix(src[i].settle_date());
			first_fix_b1_flat[i] = null<number>;
		}
		else if(src[i].is_swap_fltflt()) {
			number fix 			= first_flt_fix(src[i].settle_date());
			number fix_other 	= first_flt_fix_other(src[i].settle_date());
			first_fix_b1_flat[i] = flt_freq_flat_ == f_fwd_freq_ ? fix_other : fix;
			first_fix[i] 		= flt_freq_flat_ == f_fwd_freq_ ? fix : fix_other;
		}
		else {
			first_fix[i] = 	first_fix_b1_flat[i] = null<number>;
		}
	}

	logical solve_flat = add_basis_;
	return CORE_INT.cf_fwd_b1(src,solve_flat,f_fwd_,first_fix,first_fix_b1_flat,y_adj,require_all_cf);

}
//-------------------------------------------------------
// fwd_spr
//-------------------------------------------------------
fwd_func swap_curve_b1_ext.fwd_spr(fwd_z_model fwd_model)
{
	return	swap_curve_fwd_ext.fwd_spr(f_fwd_,fwd_model);						   
}
//-------------------------------------------------------
// fwd_spr
//-------------------------------------------------------
fwd_func swap_curve_b1_ext.fwd_spr(	fwd_func 	fwd_func_base, 
									fwd_z_model fwd_model)
{
	return	swap_curve_fwd_ext.fwd_spr(fwd_func_base,fwd_model);						   
}
//-------------------------------------------------------
// create  LEGACY
//-------------------------------------------------------
swap_curve_b1_ext swap_curve_b1_ext(swap_curve_ext_parm		p,
									date					d,	
									disc_func 				df_disc,									
									fwd_func 				f_fwd,
									vector(date) option(nullable) synt_mat,
									integer					f_fwd_freq,
									string 					name,
									fwd_func option(nullable) ff_pre = null<fwd_func>, 
									date option(nullable)	cut_off = null<date>) 
{
	if(null(ff_pre))
		return new swap_curve_b1_ext(p,d,df_disc,f_fwd, synt_mat, f_fwd_freq, name)	;	
	else
		return new swap_curve_b1_ext(p,d,df_disc,f_fwd, synt_mat, f_fwd_freq, name, ff_pre, cut_off)	;
}

//-------------------------------------------------------
// create
//-------------------------------------------------------
swap_curve_b1_ext swap_curve_b1_ext(fwd_func_tenor 	fwd_tenor,//fwd_tenor_out
									swap_curve_b1_f_parm			p,
									date							d,	
									disc_func 						df_disc,									
									fwd_func 						f_fwd,
									vector(date) option(nullable) 	synt_mat,
									fwd_func_tenor 	f_fwd_freq,
									string 							name,
									fwd_func option(nullable) 		ff_pre = null<fwd_func>, 
									date option(nullable)			cut_off = null<date>) 
{
	if(null(ff_pre))
		return new swap_curve_b1_ext(fwd_tenor,p,d,df_disc,f_fwd, synt_mat, f_fwd_freq, name,null<fwd_func>,null<date>)	;	
	else
		return new swap_curve_b1_ext(fwd_tenor,p,d,df_disc,f_fwd, synt_mat, f_fwd_freq, name, ff_pre, cut_off)	;
}

//-------------------------------------------------------
// create
//-------------------------------------------------------
swap_curve_b1_ext swap_curve_b1_ext(fwd_func_tenor 	fwd_tenor,//fwd_tenor_out
									swap_curve_b1_f_parm			p,
									date							d,	
									disc_func 						df_disc,									
									fwd_func 						f_fwd,
									vector(date) option(nullable) 	synt_mat,
									fwd_func_tenor 	f_fwd_freq,
									disc_func option(nullable)		df_synt_base,
									date 							synt_end,
									fwd_z_model  					fwd_model_synt,
									string 							name,
									fwd_func option(nullable) 		ff_pre = null<fwd_func>, 
									date option(nullable)			cut_off = null<date>) 
{
	if(null(ff_pre))
		return new swap_curve_b1_ext(fwd_tenor,p,d,df_disc,f_fwd, synt_mat, f_fwd_freq,
									 df_synt_base,synt_end,fwd_model_synt,name,null<fwd_func>,null<date>)	;
	else
		return new swap_curve_b1_ext(fwd_tenor,p,d,df_disc,f_fwd, synt_mat, f_fwd_freq,
									 df_synt_base,synt_end,fwd_model_synt,name, ff_pre, cut_off)	;
}