/*	
	Basic swap functions for "dual" bootstrap
	Developed by Algorithmica Research, Magnus Nyström
	
	
	
	Comments:

	ToDo: 
	
*/

option(null: hard);	
//-------------------------------------------------------
// class swap_curve_b2_ext
// class used for basis swaps quoted as 2 swaps i.e. the basis swap spread is the spread on the common fixed leg 
// if fixed legs differ in freq and/or dc the rates will be converted 
//-------------------------------------------------------
class swap_curve_b2_ext: public swap_curve_fwd_ext option (category: 'Yield Curve/003 Curve Building')
{
   	public:
		//----legacy------
		swap_curve_b2_ext(swap_curve_ext_parm, date, logical, string);
		swap_curve_b2_ext(swap_curve_ext_parm, date, logical, string,fwd_func ,date);

		swap_curve_b2_ext(swap_curve_ext_parm, date, disc_func option(nullable), logical, string);		
		swap_curve_b2_ext(swap_curve_ext_parm, date, disc_func option(nullable), logical, string,fwd_func ,date );
		//----------------
		
		swap_curve_b2_ext(fwd_func_tenor, swap_curve_b2_f_parm, date, disc_func option(nullable),
						  logical, string,fwd_func option(nullable),date option(nullable));
		
		swap_curve_b2_ext(fwd_func_tenor, swap_curve_b2_f_parm, date, disc_func option(nullable),
						  disc_func option(nullable),date,logical, string,fwd_func option(nullable),date option(nullable));

		swap_curve_b2_ext(swap_curve_b2_ext) ;
		
	protected:		
		//override cashflows_fwd 		cf_fwd(curve,vector(number) option(nullable));
		override cashflows_fwd 		cf_fwd(vector(instrument),vector(number) option(nullable),logical);
		
		void						check_b2_curve(curve);		
		void						init_underl_swaps(curve,out instrument option(nullable),out instrument option(nullable));
		void						init_bswap(curve);					
		void						check_fixlegs(curve,instrument,instrument);
		integer	 					flt_freq_other();

	public:	
		override swap_curve_type	type();		
		override curve				swap_curve();
		override curve 				imp_swap_curve(fwd_func);
		override number				first_flt_fix(date option(nullable));
		override integer	 		fix_freq();
										
	private:
		void			init_legacy();
		void			init(fwd_func_tenor,disc_func option(nullable) ,date option(nullable) );
		void			init_idx(fwd_func_tenor);
		void 			init_common(curve);
		void			blend(curve);
		void 			conv_idx_freq(integer);
		void 			check_swap_curve_b2(fwd_func_tenor option(nullable));
		
		curve  			b_swap_crv_;	
		integer			flt_freq_sprd_;		
		integer			flt_freq_flat_;
		logical 		add_basis_;
		string			class_name_sprd_;
		string			class_name_flat_;
		logical 		x_pol_basis_;				
};

swap_curve_type		swap_curve_b2_ext.type() 			{ return type_; }
curve 				swap_curve_b2_ext.swap_curve() 		{ return b_swap_crv_; }
integer				swap_curve_b2_ext.fix_freq()		{ return integer(b_swap_crv_.instruments()[0].coupon_freq());}
//integer 			swap_curve_b2_ext.flt_freq() 		{ return flt_freq_ == flt_freq_sprd_ ? flt_freq_flat_ : flt_freq_sprd_;}//underlying swap freq matches the base freq --> we're solving the flat tenor
integer	 			swap_curve_b2_ext.flt_freq_other()	{ return flt_freq() == flt_freq_sprd_ ? flt_freq_flat_ : flt_freq_sprd_;}
//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
swap_curve_b2_ext.swap_curve_b2_ext(swap_curve_b2_ext s) 
									: swap_curve_fwd_ext(s), b_swap_crv_(s.b_swap_crv_), flt_freq_sprd_(s.flt_freq_sprd_), 
								  	flt_freq_flat_(s.flt_freq_flat_), add_basis_(add_basis_),class_name_sprd_(s.class_name_sprd_), 
									class_name_flat_(s.class_name_flat_), x_pol_basis_(s.x_pol_basis_) 
{}
//-------------------------------------------------------
// constructor (no disc_func in this constructor), use to get info on blending for example
// [LEGACY version]
//-------------------------------------------------------
swap_curve_b2_ext.swap_curve_b2_ext(swap_curve_ext_parm 	p,
									date					d,																			
									logical 				x_pol_basis,
									string 					name) 
									:swap_curve_fwd_ext(p,d, null<disc_func>, name, swap_curve_type.SCE_FWD_B2, null<fwd_func>, null<date>),
									//must call base-constructor without checking/blending/etc functions
									 b_swap_crv_(null<curve>), class_name_sprd_(null<string>), class_name_flat_(null<string>), x_pol_basis_(x_pol_basis)
{		
	init_legacy();	
}
//-------------------------------------------------------
// constructor (no disc_func in this constructor), use to get info on blending for example
// [LEGACY version]
//-------------------------------------------------------
swap_curve_b2_ext.swap_curve_b2_ext(swap_curve_ext_parm 	p,
									date					d,																			
									logical 				x_pol_basis,
									string 					name,
									fwd_func  				ff_pre, 
									date  					cut_off) 
									:swap_curve_fwd_ext(p,d, null<disc_func>, name, swap_curve_type.SCE_FWD_B2, ff_pre, cut_off),
									//must call base-constructor without checking/blending/etc functions
									 b_swap_crv_(null<curve>), class_name_sprd_(null<string>), class_name_flat_(null<string>), 
									 x_pol_basis_(x_pol_basis)
{		
	init_legacy();	
}
//-------------------------------------------------------
// constructor
// ex. basis_crv = 3M vs 6M
// --> swap_crv is the base curve and must be either 3M or 6M
// --> if swap_crv is 6M then we will create a 3M curve (fra and depo will be blended with the created 3M curve)
// [ONLY for LEGACY version]
//-------------------------------------------------------
swap_curve_b2_ext.swap_curve_b2_ext(swap_curve_ext_parm 	p,
									date					d,										
									disc_func option(nullable) df_disc,									
									logical 				x_pol_basis,
									string 					name) 
									:swap_curve_fwd_ext(p,d, df_disc, name, swap_curve_type.SCE_FWD_B2, null<fwd_func>, null<date>),
									//must call base-constructor without checking/blending/etc functions
									 b_swap_crv_(null<curve>), class_name_sprd_(null<string>), class_name_flat_(null<string>), 
									 x_pol_basis_(x_pol_basis)
{		
	init_legacy();	
}
//-------------------------------------------------------
// constructor
// ex. basis_crv = 3M vs 6M
// --> swap_crv is the base curve and must be either 3M or 6M
// --> if swap_crv is 6M then we will create a 3M curve (fra and depo will be blended with the created 3M curve)
// [ONLY for LEGACY version]
//-------------------------------------------------------
swap_curve_b2_ext.swap_curve_b2_ext(swap_curve_ext_parm 	p,
									date					d,										
									disc_func option(nullable) df_disc,									
									logical 				x_pol_basis,
									string 					name,
									fwd_func  				ff_pre, 
									date  					cut_off) 
									:swap_curve_fwd_ext(p,d, df_disc, name, swap_curve_type.SCE_FWD_B2, ff_pre, cut_off),//must call base-constructor without checking/blending/etc functions
									 b_swap_crv_(null<curve>), class_name_sprd_(null<string>), class_name_flat_(null<string>), 
									 x_pol_basis_(x_pol_basis)
{		
	init_legacy();	
}

//-------------------------------------------------------
// constructor
// ex. basis_crv = 3M vs 6M
// --> swap_crv is the base curve and must be either 3M or 6M
// --> if swap_crv is 6M then we will create a 3M curve (fra and depo will be blended with the created 3M curve)
//-------------------------------------------------------
swap_curve_b2_ext.swap_curve_b2_ext(fwd_func_tenor			fwd_tenor,
									swap_curve_b2_f_parm 	p,
									date					d,										
									disc_func option(nullable) df_disc,									
									logical 				x_pol_basis,
									string 					name,
									fwd_func option(nullable) ff_pre, 
									date option(nullable) cut_off) 
									:swap_curve_fwd_ext(p,d, df_disc, name, swap_curve_type.SCE_FWD_B2_F, ff_pre, cut_off),//must call base-constructor without checking/blending/etc functions
									 b_swap_crv_(null<curve>), class_name_sprd_(null<string>), class_name_flat_(null<string>), 
									 x_pol_basis_(x_pol_basis)
{		
	init(fwd_tenor,null<disc_func>, null<date>);	
}

//-------------------------------------------------------
// constructor
// ex. basis_crv = 3M vs 6M
// --> swap_crv is the base curve and must be either 3M or 6M
// --> if swap_crv is 6M then we will create a 3M curve (fra and depo will be blended with the created 3M curve)
//-------------------------------------------------------
swap_curve_b2_ext.swap_curve_b2_ext(fwd_func_tenor			fwd_tenor,
									swap_curve_b2_f_parm 	p,
									date					d,										
									disc_func option(nullable) df_disc,
									disc_func option(nullable)	df_synt_base,
									date 						synt_end,
									logical 				x_pol_basis,
									string 					name,
									fwd_func option(nullable) ff_pre, 
									date option(nullable) cut_off) 
									:swap_curve_fwd_ext(p,d, df_disc, name, swap_curve_type.SCE_FWD_B2_F, ff_pre, cut_off),//must call base-constructor without checking/blending/etc functions
									 b_swap_crv_(null<curve>), class_name_sprd_(null<string>), class_name_flat_(null<string>), 
									 x_pol_basis_(x_pol_basis)
{		
	init(fwd_tenor,null(df_synt_base)? df_disc : df_synt_base, synt_end);	
}
//-------------------------------------------------------
// init_common
//-------------------------------------------------------
void swap_curve_b2_ext.init_common(curve basis_crv)
{	
	blend(basis_crv);				
	
	if(!null(ff_pre_)){	
		QL_REQUIRE(!null(cut_off_) && cut_off_ > d_,'invalid cut off date (b2 fwd)');
		vector(number) y_adj;
		logical ok = chop_short(cut_off_, curve_orig_, null<vector(number)>, curve_, y_adj);	
		QL_REQUIRE(ok && check_curve(curve_),'invalid blended fwd curve (chopped curve) (b2 fwd)');					
	}

	cf_ = cf_fwd(curve_.instruments(),null<vector(number)>,false);
	QL_REQUIRE(!null(cf_),'failed to create cashflows_fwd (b2 fwd)');
}
//-------------------------------------------------------
// init
//-------------------------------------------------------
void swap_curve_b2_ext.init(	fwd_func_tenor 	fwd_tenor,
								disc_func option(nullable)		df_synt_base,
								date 	option(nullable)		synt_end)
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_FWD_B2_F,'invalid type of param object (should be a swap_curve_b2_f_parm ) ');
	swap_curve_b2_f_parm param = dynamic_cast<swap_curve_b2_f_parm>(param_);

	check_swap_curve_types();
	check_swap_curve_b2(fwd_tenor);	//the base swapcurve from which we create a new swapcurve i.e. must have fltresetfreq <> fwd_tenor
	check_fra_curve_types();	
		
	check_b2_curve(param.basis_crv());

	init_idx(fwd_tenor);

	instrument sw_flat,sw_sprd;
	init_underl_swaps(param.basis_crv(),sw_flat,sw_sprd);//flt_freq_sprd_ and flt_freq_flat_ is initialized here
		
	check_fixlegs(param.basis_crv(), sw_flat,sw_sprd);		
	
	string cn = flt_freq() == flt_freq_sprd_ ? class_name_flat_ : class_name_sprd_ ;
	QL_REQUIRE(equal_casei(cn, param_.swap_crv().instruments()[0].class_name()),'invalid basis swap curve (incompatible with swap curve) (b2 fwd)');
	
	check_swap_instr(flt_freq_other(), flt_dc());//base curve check i.e. other freq
	check_fra_instr(flt_freq(), flt_dc());
	
	init_bswap(param.basis_crv());//create b_swap_crv_ which is the converted swap curve
	
	logical incl_synt = !null(df_synt_base) && !null(synt_end);
	if(incl_synt) {
		swap_curve_b2_f_parm param = dynamic_cast<swap_curve_b2_f_parm>(param_);

		curve swap_out_crv 	= swap_curve();//FIX_ME flt_tenor of this curve vs the tenor we solve for
					
		curve	df_basis_crv;		
		disc_func_spr df_basis = CORE_INT.df_tenor_spread(d_,df_synt_base, synt_end, param.short_crv(), 
														param.fra_crv(), swap_out_crv,param.prio1(),param.prio2(),
														param.merge_fra(),param.blend_buf_days(), df_basis_crv);
		QL_REQUIRE(!null(df_basis), "failed to create basis spread function");
	
		synt_crv_ = CORE_INT.synt_curve_create(d_,param.fra_crv(), fwd_idx_, df_basis, df_basis_crv);

		param.extend_fra_crv(synt_crv_);
	}

	init_common(param.basis_crv());
}
//-------------------------------------------------------
// init_idx
//-------------------------------------------------------
void swap_curve_b2_ext.init_idx(fwd_func_tenor	fwd_tenor)
{		
	swap_curve_b2_f_parm param = dynamic_cast<swap_curve_b2_f_parm>(param_);
	
	if(!param.short_is_fix_crv()) {
		//deposit_index, deposit or fix-flt
		curve new_short_crv;
		fwd_idx_ 			= null<ir_index>;
		logical upd = CORE_INT.fwd_init_depo(param.short_crv(), fwd_tenor, new_short_crv);
		if(upd) 		
			param.set_short_crv(new_short_crv);
		
	}
	else {
		//only deposit_index
		string fwd_tenor_s		= string(fwd_tenor);
		fixing_curve 	new_fixing_crv;
		logical upd = CORE_INT.fwd_init(d_, param.fixing_crv(), fwd_tenor, fwd_tenor_s,fwd_idx_,new_fixing_crv);
		if(upd) 
			param.set_fixing_crv(new_fixing_crv);					
	}
	
	if(null(fwd_idx_)){
		if(check_curve(param.fra_crv())) {
			if(param.fra_crv().instruments()[0].instr_type() == instr_type.SWAP_FIXFLT)
				fwd_idx_ = param.swap_crv().instruments()[0].swap_ir_index_leg2();
			else
				fwd_idx_ = param.fra_crv().instruments()[0].ir_index();
		}
		else if(check_curve(param.swap_crv())) {
			//tenor_code tc = fwd_func_tenor.to_tenor_code(fwd_tenor);
			/*ir_index leg1 = param.basis_crv_.instruments()[0].swap_ir_index_leg1();
			fwd_func_tenor t1 = fwd_func_tenor.to_fwd_func_tenor(leg1.tenor_code());
			
			ir_index leg2 = param.basis_crv_.instruments()[0].swap_ir_index_leg2();
			fwd_func_tenor t2 = fwd_func_tenor.to_fwd_func_tenor(leg2.tenor_code());
			*/
			fwd_idx_ = param.basis_crv().instruments()[0].swap_ir_index_leg1();
			fwd_func_tenor t = fwd_func_tenor_from_tc(fwd_idx_.tenor_code());
			if(fwd_tenor != t)
				fwd_idx_ = param.basis_crv().instruments()[0].swap_ir_index_leg2();
		}
	}	
}
//-------------------------------------------------------
// init_legacy
// LEGACY version
//-------------------------------------------------------
void swap_curve_b2_ext.init_legacy()
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_FWD_B2,'invalid type of param object (should be a swap_curve_b2_parm) ');		
	swap_curve_b2_parm param = dynamic_cast<swap_curve_b2_parm>(param_);

	check_swap_curve_types();
	check_swap_curve_b2(null<fwd_func_tenor>);
	
	init_ir_index(); //OBS! in swap_curve_fwd_ext and not virtual, flt_freq_ comes from swap_crv which is not the one we're solving for
	
	check_short_curve_types();
	check_fra_curve_types();	
		
	check_b2_curve(param.basis_crv_);

	instrument sw_flat,sw_sprd;
	init_underl_swaps(param.basis_crv_,sw_flat,sw_sprd);//flt_freq_sprd_ and flt_freq_flat_ is initialized here
	
	conv_idx_freq(flt_freq() == flt_freq_sprd_ ? flt_freq_flat_ : flt_freq_sprd_);

	check_fixlegs(param.basis_crv_, sw_flat,sw_sprd);		
	
	string cn = flt_freq() == flt_freq_sprd_ ? class_name_flat_ : class_name_sprd_ ;
	QL_REQUIRE(equal_casei(cn, param_.swap_crv().instruments()[0].class_name()),'invalid basis swap curve (incompatible with swap curve) (b2 fwd)');
	
	check_swap_instr(flt_freq_other(), flt_dc());//base curve check i.e. other freq
	check_fra_instr(flt_freq(), flt_dc());

	init_bswap(param.basis_crv_);//create b_swap_crv_ which is the converted swap curve 
	init_common(param.basis_crv_);	
}
//-------------------------------------------------------
// conv_idx_freq 
//-------------------------------------------------------
void swap_curve_b2_ext.conv_idx_freq(integer freq)
{
	if(integer(floor(12/fwd_idx_.tenor_months())) == freq)
		return;

	integer spot_days 	= flt_reset_days();
	integer value_days 	= spot_days;
	string tenor 		= strcat(str(12/freq),"M");
	string curr 		= param_.ccy();
	fwd_idx_ = ir_index_ibor(	strcat("Ibor-",curr),tenor,value_days,flt_reset_days(), curr,
							flt_cal(),"na", flt_dc(),flt_bd(),flt_eom(),spot_days);			
	QL_REQUIRE(!null(fwd_idx_),'failed to create ir_index (fwd)');
}
//------------------------------------------------------
// check_swap_curve_b2
// fwd_func_tenor is nullable because we don't have a fwd_func_tenor in the legacy version
//-------------------------------------------------------
void swap_curve_b2_ext.check_swap_curve_b2(fwd_func_tenor option(nullable) t)
{
	QL_REQUIRE(check_curve(param_.swap_crv()),'invalid swap curve (b2 fwd)');

	vector(instrument) sw_in= param_.swap_crv().instruments();
	QL_REQUIRE(v_size(sw_in) > 0, "invalid swap curve (input)");
	integer freq_sw_crv 	= sw_in[0].swap_reset_freq_leg2();
	QL_REQUIRE(freq_sw_crv > 0 && freq_sw_crv != 365, "invalid float reset frequency in swap curve", sw_in[0].name(), true);
	
	if(!null(t)) {//if legacy we will not enter here
		integer freq_out = fwd_func_tenor_to_freq(t);
		QL_REQUIRE(freq_out != freq_sw_crv , "invalid float reset frequency in swap curve (cannot be equal to fwd_tenor)", sw_in[0].name(), true);
	}
}
//-------------------------------------------------------
// check_b2_curve
//-------------------------------------------------------
void swap_curve_b2_ext.check_b2_curve(curve basis_crv)
{				
	QL_REQUIRE(check_curve(basis_crv),'invalid basis curve (b2 fwd)');	
	string err_name;
	
	//check type
	QL_REQUIRE(CORE_INT.check_instr_type(basis_crv, instr_type.SWAP_FLTFLT2S, err_name),'invalid basis swap curve (wrong instrument type) (b2 fwd)',err_name,true);

	
	vector(instrument) vi 	= basis_crv.instruments();
	instrument bi 			= vi[0];

	//start
	date settle  			= date_code(bi.settle_code()).apply_fwd(d_, bi.calendar());	
	QL_REQUIRE(CORE_INT.check_bswap_issue(vi, settle, err_name),'invalid basis swap (mixed start dates) (b2 fwd)',err_name,true);		
	//ccy
	string curr = param_.ccy();
	QL_REQUIRE(!null(curr),"invalid currency");
	QL_REQUIRE(CORE_INT.check_eq_currency(vi, curr, err_name),'invalid basis swap curve (mixed currencies, fixed leg) (b2 fwd)',err_name,true);
	QL_REQUIRE(CORE_INT.swap_currency_check(vi, curr, true,err_name),'invalid basis swap curve (mixed currencies, sprd leg) (b2 fwd)', err_name, true);			
	QL_REQUIRE(CORE_INT.swap_currency_check(vi, curr, false,err_name),'invalid basis swap curve (mixed currencies, flat leg) (b2 fwd)', err_name, true);			

	//reset_freq
	integer flt_freq_sprd = -1;	
	QL_REQUIRE(CORE_INT.swap_rfreq_check(vi,flt_freq_sprd,true,err_name),'invalid basis swap (mixed reset frequencies - sprd leg) (b2 fwd)',err_name,true);
	QL_REQUIRE(flt_freq_sprd > 0 && flt_freq_sprd != 365,'invalid float reset freq (sprd leg) in basis swap (b2 fwd)',bi.name(),true);
	
	integer flt_freq_flat = -1;	
	QL_REQUIRE(CORE_INT.swap_rfreq_check(vi,flt_freq_flat,false,err_name),'invalid basis swap (mixed reset frequencies - flat leg) (b2 fwd)',err_name,true);	
	QL_REQUIRE(flt_freq_flat > 0 && flt_freq_flat != 365,'invalid float reset freq (flat leg) in basis swap (b2 fwd)',bi.name(),true);
	
	//tenor_code	
	tenor_code t = tenor_code(flt_freq_sprd);	
	QL_REQUIRE(CORE_INT.swap_index_tenor_code_check(vi, t, true,err_name),'invalid float tenor (spread leg) in basis swap (b2 fwd)', err_name, true);				
	t = tenor_code(flt_freq_flat);	
	QL_REQUIRE(CORE_INT.swap_index_tenor_code_check(vi, t, false,err_name),'invalid float tenor (flat leg) in basis swap (b2 fwd)', err_name, true);			
	
	//class names
	instr_def def		= bi.instr_def();
	QL_REQUIRE(!null(def),"invalid instr_def");
	string class_name_flat 	= def.fltflt_swap_class_flat();
	string class_name_sprd 	= def.fltflt_swap_class_sprd();

	QL_REQUIRE(!null(class_name_flat),"invalid class name (flat-leg)");
	QL_REQUIRE(!null(class_name_sprd),"invalid class name (sprd-leg)");
	
	QL_REQUIRE(CORE_INT.check_eq_class_flat(vi,class_name_flat,err_name),'invalid basis swap curve (class name flat) (b2 fwd)',err_name,true);	
	QL_REQUIRE(CORE_INT.check_eq_class_sprd(vi,class_name_sprd,err_name),'invalid basis swap curve (class name sprd) (b2 fwd)',err_name,true);

	//cpn freq
	integer cfreq = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,cfreq,true,err_name),'invalid basis swap (mixed coupon frequencies - sprd leg) (b2 fwd)',err_name,true);
	cfreq = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,cfreq,false,err_name),'invalid basis swap (mixed coupon frequencies - flat leg) (b2 fwd)',err_name,true);
	
	//quotes
	QL_REQUIRE(check_quote(basis_crv,err_name),'invalid quote(s) in basis curve (b2 fwd)',err_name,true);
}
//-------------------------------------------------------
// init_underl_swaps
//-------------------------------------------------------
void swap_curve_b2_ext.init_underl_swaps(	curve 			basis_crv,
											out instrument option(nullable) 	sw_flat,
											out instrument option(nullable) 	sw_base)
{
	instrument inst = basis_crv.instruments()[0];

	class_name_flat_ = inst.instr_def().fltflt_swap_class_flat();
	QL_REQUIRE(!null(class_name_flat_),'invalid class name tag - flat leg (b2 fwd)',inst.name(), true);

	class_name_sprd_ = inst.instr_def().fltflt_swap_class_sprd();
	QL_REQUIRE(!null(class_name_sprd_),'invalid class name tag - sprd leg (b2 fwd)',inst.name(), true);	

	sw_flat = CORE_INT_FIXFLT.create_class_fixflt(class_name_flat_);
	QL_REQUIRE(!null(sw_flat),'invalid swap creation from class - flat leg (b2 fwd)',class_name_flat_,true);
	
	sw_base = CORE_INT_FIXFLT.create_class_fixflt(class_name_sprd_);
	QL_REQUIRE(!null(sw_base),'invalid swap creation from class - sprd leg (b2 fwd)',class_name_sprd_,true);

	flt_freq_sprd_	= sw_base.swap_reset_freq_leg2();
	QL_REQUIRE(flt_freq_sprd_>0 && flt_freq_sprd_ != 365,'invalid flt_freq_sprd in swap curve (b2 fwd)',class_name_sprd_,true);	
	flt_freq_flat_	= sw_flat.swap_reset_freq_leg2();
	QL_REQUIRE(flt_freq_flat_ > 0 && flt_freq_flat_ != 365,'invalid flt_freq_flat in swap curve (b2 fwd)',class_name_flat_,true);
			
	QL_REQUIRE(flt_freq_sprd_  != flt_freq_flat_ ,'invalid basis swap (same float freq in both swaps) (b2 fwd)'); //for same currency only
	
	if(flt_freq() != flt_freq_sprd_  && flt_freq() != flt_freq_flat_ )			
		QL_FAIL('incompatible basis swap (float coupon frequency not compatible with swap curve) (b2 fwd)');	
}
//-------------------------------------------------------
// blend
//-------------------------------------------------------
void swap_curve_b2_ext.blend(curve basis_crv)
{	
	

	curve short_adj_c = param_.short_crv();
	if(type() == swap_curve_type.SCE_FWD_B2 ){
		if(!null(param_.short_crv()))	
			short_adj_c = CORE_INT_COMPAT.extract_tenor_depo(param_.short_crv(),12/flt_freq(),flt_cal(),flt_eom(),flt_bd());		
	}

	vector(date) knots;	
	curve swap_adj_c = b_swap_crv_;
	
	if(param_.prio1() == curve_prio.MIDDLE || (param_.prio2() == curve_prio.MIDDLE && param_.prio1() == curve_prio.SHORT))
		CORE_INT.blend_adj_long(	short_adj_c, param_.fra_crv() , swap_adj_c, param_.prio1(), param_.prio2(),
						param_.blend_buf_days(),short_adj_c,swap_adj_c);//when fra in the middle of the swap
	
	curve_ = curve_orig_ = CORE_INT.blend_curve(short_adj_c, param_.fra_crv() , swap_adj_c, param_.prio1(), param_.prio2(), param_.merge_fra(),
												param_.no_overlap_fwd(), param_.merge_crv(),knots, param_.blend_buf_days());


	if(null(curve_)) {
		curve_msg_ 	= "null blended curve (b2 fwd)";		
		return;
	}

	curve_ = curve_orig_ = curve_.set_name(strcat(name_,":blended curve"));
	
	string n;
	logical nq = check_quote(curve_,n);
	
	if(!nq) {
		curve_msg_ = strcat(["null quotes in curve (swap_curve_b2_ext-blend) [",n,"]"]);
		//curve_msg_ 	= "null quotes in in blended curve (b2 fwd)";		
		return;
	}

	curve_msg_ 	= check_curve(curve_) ? "" : "empty curve";		
}	
//-------------------------------------------------------
// init_bswap
//-------------------------------------------------------
void swap_curve_b2_ext.init_bswap(curve basis_crv)
{	
	string n; 
	QL_REQUIRE(check_quote(param_.swap_crv(),n),'invalid quote(s) in swap curve (b2 fwd)',n,true);	
	QL_REQUIRE(check_quote(basis_crv,n),'invalid quote(s) in basis swap curve (b2 fwd)',n,true);
		
	add_basis_ 			= (flt_freq() == flt_freq_flat_); //solve for flat --> spread is "as is"
	number first_fix 	= first_flt_fix(null<date>);//FIX_ME assuming spot start for all basises here
	
	b_swap_crv_ = CORE_INT.calc_swap_from_basis2(d_,df_disc_,param_.swap_crv(),basis_crv,add_basis_,
												 first_fix,"BasisIrs(b2)", x_pol_basis_);
	QL_REQUIRE(check_curve(b_swap_crv_),'invalid swap curve (created from basis curve) (b2 fwd)');
	
	if(!null(first_fix))	
		b_swap_crv_ = swap_curve_fix_adj(b_swap_crv_);			
}
//-------------------------------------------------------
// check
// check only first swap in each curve assuming the rest of the swaps are the same!!
//-------------------------------------------------------
void swap_curve_b2_ext.check_fixlegs(	curve basis_crv,
										instrument sw_f,
										instrument sw_b)
{	
	instrument bs 	= basis_crv.instruments()[0]; 
	integer ff 		= integer(bs.coupon_freq());	
	if(ff != integer(sw_b.coupon_freq()) )			
		QL_FAIL('incompatible basis swap (fixed coupon frequency not compatible with sprd leg swap) (b2 fwd)');
	if(ff != integer(sw_f.coupon_freq()) )			
		QL_FAIL('incompatible basis swap (fixed coupon frequency not compatible with flat leg swap) (b2 fwd)');
	
	//cannot have a check on fixleg dc's ex. 1m eur has fixlegdc = act360		
	/*day_count_method dc = bs.accr_dc_method();	
	if(dc != b.accr_dc_method())			
		QL_FAIL('incompatible basis swap (fixed daycount method not compatible with base leg swap) (b2 fwd)');
	if(dc != f.accr_dc_method())			
		QL_FAIL('incompatible basis swap (fixed daycount method not compatible with flat leg swap) (b2 fwd)');
	*/

	//if(s.daycount_method() != b.daycount_method())
	//	throw(E_INVALID_ARG, throw_err('incompatible basis swap (different fixed leg daycount methods)'));
	
	//if(s.coupon_freq() != b.coupon_freq())
	//	throw(E_INVALID_ARG, throw_err('incompatible basis swap (different fixed leg coupon frequency)'));
	
	
}	
//-------------------------------------------------------
// imp_swap_curve //UNTESTED
// swap curve refers to the one we have created a fwd_func for
//-------------------------------------------------------
curve swap_curve_b2_ext.imp_swap_curve(fwd_func f_flt_idx)
{	

	

	vector(instrument) vi = b_swap_crv_.instruments();
	number size = v_size(vi);

	error_info error = new error_info(true,false);
	
	for (integer i = 0 ; i < size ; i++) {
		if(!vi[i].is_swap_fixflt())
			return null<curve>	;

		string mat_code = vi[i].maturity_code(error);
		string name;
		if(!null(mat_code))
			name = strcat(CORE_INT.clean_mat_code(mat_code)," - implied");
		else	
			name = strcat(string(year(vi[i].maturity()))," - implied");

		CORE_INT.reset_single(error);
		
		swap_fixflt sw = null<swap_fixflt>;

		number first_fix = 	first_flt_fix(vi[i].settle_date());
		
		if(vi[i].is_ext_swap(error)){
			swap_fixflt sw_in = swap_fixflt(vi[i]); 
			sw = swap_fixflt_plain(	sw_in, name, d_,vi[i].settle_date(),vi[i].settle_date(),null<string>,
									vi[i].maturity(),null<string>,null<number> ,
									100,true,df_disc_,	f_flt_idx,first_fix, -1,error);
		}
		else {
			sw = swap_fixflt_plain(	vi[i].instr_def(), name, d_,vi[i].settle_date(),null<string>,
									vi[i].maturity(),null<string>,null<number> ,
									100,true,df_disc_,	f_flt_idx,first_fix, -1,error);
		}
		
		if(error.is_error())
			vi[i] = null<swap_fixflt>;
		else
			vi[i] = sw;			
	}
	
	return curve(vi,b_swap_crv_.name());
	
	//return CORE_INT.par_fixflt_swap_rate(d_,b_swap_crv_, df_disc_, f_flt_idx, flt_reset_days(), first_flt_fix());
}
//-------------------------------------------------------
// first_flt_fix  (only for spot starting swaps)
//-------------------------------------------------------
number swap_curve_b2_ext.first_flt_fix(date option(nullable) settle)
{
	date rd = null(settle) ? flt_reset_date() : flt_reset_date(settle);
	
	if(null(rd) || rd > d_ )
		return null<number>;
	
	if(type() == swap_curve_type.SCE_FWD_B2_F){
		swap_curve_b2_f_parm param = dynamic_cast<swap_curve_b2_f_parm>(param_);
		string 	err_msg;
		if(null(param.fixing_crv()))
			return check_curve(param_.swap_crv()) ? param_.swap_first_fix() : null<number>;
		else
			
		return param.fixing_crv().fixing_rate(d_, fwd_idx_.tenor(), err_msg);
	}
	else if(type_ == swap_curve_type.SCE_FWD_B2){
		return check_curve(param_.swap_crv()) ? param_.swap_first_fix() : null<number>;
	}
	else
		QL_FAIL("invalid function call");
}

//-------------------------------------------------------
// cf_fwd
//-------------------------------------------------------
cashflows_fwd swap_curve_b2_ext.cf_fwd(	vector(instrument) src,										
										vector(number) option(nullable)	y_adj,
										logical require_all_cf)
{
	integer size = v_size(src);
	vector(number) first_fix[size];
	
	for(integer i = 0; i < size ; i++){
		if(src[i].is_swap_fixflt()){
			first_fix[i] = 	first_flt_fix(src[i].settle_date());
		}
		else if(src[i].is_swap_fltflt2s()) {
			first_fix[i] = 	first_flt_fix(src[i].settle_date());
		}
		else {
			first_fix[i] = 	null<number>;
		}
	}
	
	return CORE_INT.cf_fwd_b2(src, first_fix, y_adj, flt_freq(), flt_dc(), require_all_cf);
	
}
//-------------------------------------------------------
// create 
// [legacy version]
//-------------------------------------------------------
swap_curve_b2_ext swap_curve_b2_ext(swap_curve_ext_parm 		p,
									date						d,										
									disc_func  					df_disc,									
									logical 					x_pol_basis,
									string 						name,
									fwd_func option(nullable) 	ff_pre = null<fwd_func>, 
									date option(nullable)		cut_off = null<date>) 
{		
	if(null(ff_pre))
		return new swap_curve_b2_ext(p, d,df_disc,x_pol_basis, name);
	else
		return new swap_curve_b2_ext(p, d,df_disc,x_pol_basis, name,ff_pre, cut_off);
}
//-------------------------------------------------------
// create
//-------------------------------------------------------
swap_curve_b2_ext swap_curve_b2_ext(fwd_func_tenor fwd_tenor,
									swap_curve_b2_f_parm 		p,
									date						d,										
									disc_func  					df_disc,								
									logical 					x_pol_basis,
									string 						name,
									fwd_func option(nullable) 	ff_pre = null<fwd_func>, 
									date option(nullable)		cut_off = null<date>) 
{		
	if(null(ff_pre))
		return new swap_curve_b2_ext(fwd_tenor,p, d,df_disc,x_pol_basis, name, null<fwd_func>, null<date>);
	else
		return new swap_curve_b2_ext(fwd_tenor,p, d,df_disc,x_pol_basis, name,ff_pre, cut_off);
}

//-------------------------------------------------------
// create
//-------------------------------------------------------
swap_curve_b2_ext swap_curve_b2_ext(fwd_func_tenor fwd_tenor,
									swap_curve_b2_f_parm 		p,
									date						d,										
									disc_func  					df_disc,
									disc_func option(nullable)	df_synt_base,
									date 						synt_end,
									logical 					x_pol_basis,
									string 						name,
									fwd_func option(nullable) 	ff_pre = null<fwd_func>, 
									date option(nullable)		cut_off = null<date>) 
{		
	if(null(ff_pre))
		return new swap_curve_b2_ext(fwd_tenor,p, d,df_disc,df_synt_base,synt_end,x_pol_basis, name, null<fwd_func>, null<date>);
	else
		return new swap_curve_b2_ext(fwd_tenor,p, d,df_disc,df_synt_base,synt_end,x_pol_basis, name,ff_pre, cut_off);
}