/*	
	Basic swap curve functions for cross currency bootstrap
	Developed by Algorithmica Research, Magnus Nyström
	
	
	
	Comments:

	
*/
option(null: hard);	
//-------------------------------------------------------
// class swap_curve_bc_ext
// class used for cross currency basis swaps (quoted as 1 swap i.e. 2 float legs)
//-------------------------------------------------------
class swap_curve_bc_ext option (category: 'Yield Curve/003 Curve Building Cross Currency')
{
   	public: 

		//---legacy--------------
		swap_curve_bc_ext(	date,swap_curve_cb1_parm,disc_func option(nullable),disc_func option(nullable),fwd_func,
							fwd_func,string option(nullable) name = null<string>) ;
		swap_curve_bc_ext(	date,swap_curve_cb1_parm,disc_func option(nullable),disc_func option(nullable),disc_func,
							disc_func,string option(nullable) name = null<string>) ;
		//-----------------------
		swap_curve_bc_ext(	date,swap_curve_cb1_f_parm,disc_func option(nullable),disc_func option(nullable),fwd_func,fwd_func_tenor,
							fwd_func,fwd_func_tenor,string option(nullable) name = null<string>) ;
		
		swap_curve_bc_ext(	swap_curve_bc_ext) ;		

		string 			name();
		curve 			basis_curve()	;
		curve 			blended_curve();
		fx_curve		fx_swaps();                                                                               		
		string 			settle_code() ;
		date			trade_date();
		logical			basis_is_mtm();

		
		
		
		
		//sprd leg
		ir_index		fwd_idx_sprd();
		string 			df_sprd_crv_name();
		string 			adj_df_sprd_crv_name();
		disc_func		adj_disc_df_sprd(interpolator,rate_type);		
		disc_func		adj_disc_df_sprd(disc_z_model);
		disc_func		adj_disc_df_sprd(disc_z_model, disc_func option(nullable));
		integer 		flt_reset_days_sprd(); 
		date 			flt_reset_date_sprd();
		number			first_flt_sprd();
		curve 			blended_sprd_curve();
		disc_func		disc_df_sprd();
		fwd_func        f_fwd_sprd() ;
		integer 		flt_freq_sprd()	;
		day_count_method flt_dc_sprd()	;  
		string 			ccy_sprd() 	;
		bd_convention 	flt_bd_sprd() ;
		calendar 		flt_cal_sprd() ;
		logical 		flt_eom_sprd() ;
		vector(number) 	solve_spread_sprd_leg(vector(number) option(nullable) s = null<vector(number)>);
		vector(number) 	solve_ois_ois_spread_sprd_leg();
		vector(number) 	solve_fx_ois_spread_sprd_leg();
		
		//flat leg
		ir_index		fwd_idx_flat();
		string 			df_flat_crv_name();	
		string 			adj_df_flat_crv_name();		
		disc_func		adj_disc_df_flat(interpolator,rate_type);
		disc_func		adj_disc_df_flat(disc_z_model);
		disc_func		adj_disc_df_flat(disc_z_model, disc_func option(nullable));
		integer 		flt_reset_days_flat(); 
		date 			flt_reset_date_flat();
		curve 			blended_flat_curve();
		number			first_flt_flat();
		disc_func		disc_df_flat();
		fwd_func 		f_fwd_flat() ;
		integer 		flt_freq_flat()	;
		day_count_method flt_dc_flat() ;
		string 			ccy_flat() ;
		bd_convention 	flt_bd_flat() ;
		calendar 		flt_cal_flat() ;
		logical 		flt_eom_flat() ;
		vector(number) 	solve_spread_flat_leg(vector(number) option(nullable) s = null<vector(number)>);
		vector(number) 	solve_ois_ois_spread_flat_leg();
		vector(number) 	solve_fx_ois_spread_flat_leg();
		
		//---backward compat only-----------
		integer 		flt_reset_days(); 
		date 			flt_reset_date();
		bd_convention 	flt_bd() ;
		calendar 		flt_cal() ;
		logical 		flt_eom() ;
		disc_func		adj_disc_df_base(interpolator,rate_type);
		vector(number) 	solve_spread_base_leg(vector(number) option(nullable) s = null<vector(number)>);
		vector(number) 	solve_ois_ois_spread_base_leg();
		vector(number) 	solve_fx_ois_spread_base_leg();
		curve 			blended_base_curve(); 	
		number 			flt_freq_base()	;		
		day_count_method flt_dc_base()	;  		
		string 			ccy_base() 	;   		
		disc_func		disc_df_base();			
		fwd_func        f_fwd_base() ; 			
		//--------------------------------------------

	protected:

		swap_curve_bc_ext() ;
		void				init_legacy(disc_func option(nullable) a=null<disc_func>,disc_func option(nullable) b=null<disc_func>);
		//void				init_common(curve option(nullable), fx_spot option(nullable),disc_func option(nullable),disc_func option(nullable) );
		void				init(logical,fwd_func_tenor,fwd_func_tenor);
		void				init_idx(fwd_func_tenor,fwd_func_tenor);
		void				check_b1_curve(curve option(nullable));
		void				init_check_b1_curve(curve option(nullable),fwd_func_tenor, fwd_func_tenor);
			
		void				post_check_b1_curve(curve option(nullable));
		void				init_fx_swap(fx_spot option(nullable) );
		
		void 				init_ir_index_legacy(instrument );

		logical 			create_blend_curve(	out curve option(nullable), out curve option(nullable), out curve option(nullable),
												out curve option(nullable),out vector(date)   ,out string);
		void				blend_curve();
		virtual cashflows 	cf(curve, disc_func,logical);	
		virtual cashflows_q cf_mtm(curve, disc_func,logical);
		disc_func			adj_disc_df_sprd_int(disc_z_model, disc_func option(nullable), logical);		
		disc_func			adj_disc_df_flat_int(disc_z_model, disc_func option(nullable), logical );		 
		
	//private:
		date				d_;
		swap_curve_ext_parm param_;
		fx_curve			fx_swaps_;
				
		fwd_func			f_fwd_sprd_;			
		ir_index			fwd_idx_sprd_;
	
		fwd_func			f_fwd_flat_;									
		ir_index			fwd_idx_flat_;

		disc_func			df_disc_flat_;		
		disc_func			df_disc_sprd_;
									
		string				name_;
		vector(number) 		spread_;
		curve				curve_sprd_;
		curve				curve_flat_;
		logical				fx_flat_eq_base_;
		logical				basis_is_mtm_;
		logical 			curve_ok_;
			
};

ir_index		swap_curve_bc_ext.fwd_idx_flat()	{ return fwd_idx_flat_;}
ir_index		swap_curve_bc_ext.fwd_idx_sprd()	{ return fwd_idx_sprd_;}
date 			swap_curve_bc_ext.trade_date()		{ return d_;}
string 			swap_curve_bc_ext.name()			{ return name_;}        
fwd_func 		swap_curve_bc_ext.f_fwd_flat() 		{ return f_fwd_flat_;}
fwd_func  		swap_curve_bc_ext.f_fwd_sprd()      { return f_fwd_sprd_;}  
disc_func 		swap_curve_bc_ext.disc_df_sprd()	{ return df_disc_sprd_;}
disc_func 		swap_curve_bc_ext.disc_df_flat()	{ return df_disc_flat_;}
fx_curve		swap_curve_bc_ext.fx_swaps()		{ return fx_swaps_;}
logical			swap_curve_bc_ext.basis_is_mtm()	{ return basis_is_mtm_;}

integer			swap_curve_bc_ext.flt_freq_flat()	{ return null(fwd_idx_flat_) ? -1: integer(floor(12/fwd_idx_flat_.tenor_months()));}
integer			swap_curve_bc_ext.flt_freq_sprd()	{ return null(fwd_idx_sprd_) ? -1: integer(floor(12/fwd_idx_sprd_.tenor_months()));}
day_count_method swap_curve_bc_ext.flt_dc_sprd() 	{ return null(fwd_idx_sprd_) ? null<day_count_method> : fwd_idx_sprd_.day_count_method();}
day_count_method swap_curve_bc_ext.flt_dc_flat() 	{ return null(fwd_idx_flat_) ? null<day_count_method> : fwd_idx_flat_.day_count_method();}
bd_convention	swap_curve_bc_ext.flt_bd_flat() 	{ return null(fwd_idx_flat_) ? null<bd_convention>: fwd_idx_flat_.bd_convention();}
logical			swap_curve_bc_ext.flt_eom_flat() 	{ return null(fwd_idx_flat_) ? null<logical>: fwd_idx_flat_.eom();}
calendar		swap_curve_bc_ext.flt_cal_flat() 	{ return null(fwd_idx_flat_) ? null<calendar>: fwd_idx_flat_.fixing_calendar();}
bd_convention	swap_curve_bc_ext.flt_bd_sprd() 	{ return null(fwd_idx_sprd_) ? null<bd_convention>: fwd_idx_sprd_.bd_convention();}
logical			swap_curve_bc_ext.flt_eom_sprd() 	{ return null(fwd_idx_sprd_) ? null<logical>: fwd_idx_sprd_.eom();}
calendar		swap_curve_bc_ext.flt_cal_sprd() 	{ return null(fwd_idx_sprd_) ? null<calendar>: fwd_idx_sprd_.fixing_calendar();}
string 			swap_curve_bc_ext.ccy_flat() 		{ return null(fwd_idx_flat_) ? null<string>: fwd_idx_flat_.currency();}	
string 			swap_curve_bc_ext.ccy_sprd() 		{ return null(fwd_idx_sprd_) ? null<string>: fwd_idx_sprd_.currency();}	
string 			swap_curve_bc_ext.settle_code()
{
	if(!null(fwd_idx_flat_))
		return strcat("BD",str(fwd_idx_flat_.value_days()));
	else if(!null(fwd_idx_sprd_))
		return strcat("BD",str(fwd_idx_sprd_.value_days()));
	else
		return  null<string>;
}

string 			swap_curve_bc_ext.df_flat_crv_name() 		{ return strcat([ccy_flat(),":",ccy_flat()]);}	
string 			swap_curve_bc_ext.adj_df_flat_crv_name()	{ return strcat([ccy_flat(),":",ccy_sprd()]);}
string 			swap_curve_bc_ext.df_sprd_crv_name()		{ return strcat([ccy_sprd(),":",ccy_sprd()]);}		
string 			swap_curve_bc_ext.adj_df_sprd_crv_name()	{ return strcat([ccy_sprd(),":",ccy_flat()]);}


		
//--- [LEGACY]--------
bd_convention swap_curve_bc_ext.flt_bd() 			{ return fwd_idx_flat_.bd_convention();}
logical		swap_curve_bc_ext.flt_eom() 			{ return fwd_idx_flat_.eom();}
calendar	swap_curve_bc_ext.flt_cal() 			{ return fwd_idx_flat_.fixing_calendar();}
number 		swap_curve_bc_ext.flt_freq_base()		{ return flt_freq_sprd();}
day_count_method  swap_curve_bc_ext.flt_dc_base()	{ return flt_dc_sprd();}   
string 		swap_curve_bc_ext.ccy_base() 			{ return ccy_sprd();} 
fwd_func  	swap_curve_bc_ext.f_fwd_base()        	{ return f_fwd_sprd_;}
disc_func 	swap_curve_bc_ext.disc_df_base()		{ return df_disc_sprd_;}
//-------------------------------

//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
swap_curve_bc_ext.swap_curve_bc_ext(swap_curve_bc_ext s) 
								:d_(s.d_), param_(s.param_), fx_swaps_(s.fx_swaps_), 
								f_fwd_sprd_(s.f_fwd_sprd_),fwd_idx_sprd_(s.fwd_idx_sprd_),
								f_fwd_flat_(s.f_fwd_flat_),fwd_idx_flat_(s.fwd_idx_flat_), 
								df_disc_flat_(s.df_disc_flat_) , df_disc_sprd_(s.df_disc_sprd_),
								name_(s.name_), spread_(s.spread_), curve_sprd_(s.curve_sprd_), curve_flat_(s.curve_flat_), 
								fx_flat_eq_base_(s.fx_flat_eq_base_), basis_is_mtm_(s.basis_is_mtm_),curve_ok_(s.curve_ok_)
{}
		
//-------------------------------------------------------
// constructor
//-------------------------------------------------------
swap_curve_bc_ext.swap_curve_bc_ext()
{}

//-------------------------------------------------------
// constructor [LEGACY]
// ex basis_crv = 3M SEK + sprd vs 3M USD Flat
//-------------------------------------------------------
swap_curve_bc_ext.swap_curve_bc_ext(date						d,									
									swap_curve_cb1_parm			p,									
									disc_func option(nullable)	df_disc_flat,
									disc_func option(nullable)	df_disc_sprd,
									fwd_func					f_fwd_flat,																	
									fwd_func 					f_fwd_sprd,									
									string 	option(nullable) 	name) 
									:d_(d), param_(p),
									 f_fwd_sprd_(f_fwd_sprd),fwd_idx_sprd_(null<ir_index>),
									 f_fwd_flat_(f_fwd_flat),fwd_idx_flat_(null<ir_index>),
									 df_disc_flat_(df_disc_flat), df_disc_sprd_(df_disc_sprd), name_(name),
									 basis_is_mtm_(false),curve_ok_(false)
{
	init_legacy();	
}
//-------------------------------------------------------
// constructor [LEGACY]
// ex basis_crv = 3M SEK + sprd vs 3M USD Flat
//-------------------------------------------------------
swap_curve_bc_ext.swap_curve_bc_ext(date						d,									
									swap_curve_cb1_parm			p,									
									disc_func option(nullable)	df_disc_flat,
									disc_func option(nullable)	df_disc_sprd,
									disc_func					f_fwd_flat,	
									disc_func 					f_fwd_sprd,									
									string option(nullable)		name) 
									:d_(d), param_(p),
									 f_fwd_sprd_(null<fwd_func>),fwd_idx_sprd_(null<ir_index>),
									 f_fwd_flat_(null<fwd_func>),fwd_idx_flat_(null<ir_index>),
									 df_disc_flat_(df_disc_flat), df_disc_sprd_(df_disc_sprd), name_(name),
									 basis_is_mtm_(false),curve_ok_(false)
{
	init_legacy(f_fwd_flat,f_fwd_sprd);
}

//-------------------------------------------------------
// constructor (with a basis)
// ex basis_crv = 3M SEK + sprd vs 3M USD Flat
//-------------------------------------------------------
swap_curve_bc_ext.swap_curve_bc_ext(date						d,									
									swap_curve_cb1_f_parm		p,									
									disc_func option(nullable)	df_disc_flat,
									disc_func option(nullable)	df_disc_sprd,
									fwd_func					f_fwd_flat,
									fwd_func_tenor 				fwd_tenor_flat,																	
									fwd_func 					f_fwd_sprd,	
									fwd_func_tenor 				fwd_tenor_sprd,								
									string 	option(nullable) 	name) 
									:d_(d), param_(p),
									 f_fwd_sprd_(f_fwd_sprd),fwd_idx_sprd_(null<ir_index>),
									 f_fwd_flat_(f_fwd_flat),fwd_idx_flat_(null<ir_index>),
									 df_disc_flat_(df_disc_flat), df_disc_sprd_(df_disc_sprd), name_(name),
									 basis_is_mtm_(null<logical>),curve_ok_(false)
{
	init(true,fwd_tenor_flat,fwd_tenor_sprd);	
}

//-------------------------------------------------------
// init
//-------------------------------------------------------
void swap_curve_bc_ext.init(logical 						require_basis,
							fwd_func_tenor	fwd_tenor_flat,
							fwd_func_tenor	fwd_tenor_sprd)
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_CURR_B1_F,'invalid type of param object (should be a swap_curve_cb1_f_parm ) ');		
	swap_curve_cb1_f_parm param = dynamic_cast<swap_curve_cb1_f_parm>(param_);
	if(null(name_))
		name_ = "";

	curve ccb_crv = param.basis_crv();
	QL_FAIL_COND(require_basis && !check_curve(ccb_crv),"invalid currency basis curve");	
	QL_REQUIRE(!null(df_disc_flat_) || !null(df_disc_sprd_),'invalid disc_func (either flat leg or spread (base) leg disc_func must be input)');

	init_check_b1_curve(ccb_crv,fwd_tenor_flat,fwd_tenor_sprd);
	init_idx(fwd_tenor_flat,fwd_tenor_sprd);
	post_check_b1_curve(ccb_crv);	
	
	init_fx_swap(param.fx_spot());
	blend_curve();	
}
//-------------------------------------------------------
// init_idx		(curr) (if only fx curve we have no ir_index)
//-------------------------------------------------------
void swap_curve_bc_ext.init_idx(fwd_func_tenor	fwd_tenor_flat,
								fwd_func_tenor	fwd_tenor_sprd)
{
	swap_curve_cb1_f_parm param = dynamic_cast<swap_curve_cb1_f_parm>(param_);
	if(param.short_is_fix_crv_flat()) {
	
		string fwd_tenor_flat_s		= string(fwd_tenor_flat);
		fixing_curve new_fixing_crv;
		logical upd_flat = CORE_INT.fwd_init(d_, param.fixing_crv_flat(), fwd_tenor_flat, fwd_tenor_flat_s,fwd_idx_flat_,new_fixing_crv);
		if(upd_flat) 
			param.set_fixing_crv_flat(new_fixing_crv); 
	}
	/*else {
		curve new_short_crv;
		fwd_idx_flat_ = null<ir_index>;		
		logical upd = CORE_INT.fwd_init_depo(param.depo_crv_flat(), fwd_tenor_flat, new_short_crv);		
		if(upd)
			param.set_depo_crv_flat(new_short_crv);
	}*/
	if(param.short_is_fix_crv_sprd()) {
		string fwd_tenor_sprd_s		= string(fwd_tenor_sprd);
		fixing_curve new_fixing_crv;
		logical upd_sprd = CORE_INT.fwd_init(d_, param.fixing_crv_sprd(), fwd_tenor_sprd, fwd_tenor_sprd_s,fwd_idx_sprd_,new_fixing_crv);
		if(upd_sprd) 
			param.set_fixing_crv_sprd(new_fixing_crv); 
	}
	/*else {
		curve new_short_crv;
		fwd_idx_sprd_ = null<ir_index>;		
		logical upd = CORE_INT.fwd_init_depo(param.depo_crv_sprd(), fwd_tenor_sprd, new_short_crv);		
		if(upd)
			param.set_depo_crv_sprd(new_short_crv);
	}*/

	if(check_curve(param.basis_crv())) {
		instrument instr = param.basis_crv().instruments()[0];
		swap_fltflt sw = swap_fltflt(instr);
		
		if(null(fwd_idx_flat_)){
			fwd_idx_flat_ = sw.ir_index_leg2();
		}

		if(null(fwd_idx_sprd_)){
			fwd_idx_sprd_ = sw.ir_index_leg1();
		}
	}
}
//-------------------------------------------------------
// init_common
//-------------------------------------------------------
/*void swap_curve_bc_ext.init_common(	curve 	option(nullable)	basis_crv,
									fx_spot option(nullable) 	fxs, 
									disc_func option(nullable) 	f_fwd_flat,	
									disc_func option(nullable) 	f_fwd_sprd 	)
{
	if(!null(basis_crv)) {
		post_check_b1_curve(basis_crv);
			
		spread_ = one_vector(v_size(basis_crv.instruments()), 0);		
		QL_REQUIRE(!null(df_disc_flat_) || !null(df_disc_sprd_),'invalid disc_func (either flat leg or spread (base) leg disc_func must be input)');	

		if(!null(f_fwd_sprd))
			f_fwd_sprd_ = fwd_func(f_fwd_sprd, flt_freq_sprd(), d_, flt_dc_sprd(), flt_cal(), flt_bd(), flt_eom());
		if(!null(f_fwd_flat)) 
			f_fwd_flat_ = fwd_func(f_fwd_flat, flt_freq_flat(), d_ , flt_dc_flat(), flt_cal(), flt_bd(), flt_eom() );
	}
	
	init_fx_swap(fxs);
	blend_curve();	
}*/
//-------------------------------------------------------
// init_legacy   [LEGACY]
//-------------------------------------------------------
void swap_curve_bc_ext.init_legacy(	disc_func option(nullable) f_fwd_flat,	
									disc_func option(nullable) f_fwd_sprd 	)
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_CURR_B1,'invalid type of param object (should be a swap_curve_cb1_parm ) ');		
	swap_curve_cb1_parm param = dynamic_cast<swap_curve_cb1_parm>(param_);
	if(null(name_)) name_ = "";

	QL_REQUIRE(check_curve(param.basis_crv_),"invalid currency basis curve");

	check_b1_curve(param.basis_crv_);

	init_ir_index_legacy(param.basis_crv_.instruments()[0]);
	//init_common(param.basis_crv_, param.fx_spot_, f_fwd_flat, f_fwd_sprd);

	//if(!null(basis_crv)) {
		post_check_b1_curve(param.basis_crv_);
			
		spread_ = one_vector(v_size(param.basis_crv_.instruments()), 0);		
		QL_REQUIRE(!null(df_disc_flat_) || !null(df_disc_sprd_),'invalid disc_func (either flat leg or spread (base) leg disc_func must be input)');	

		if(!null(f_fwd_sprd))
			f_fwd_sprd_ = fwd_func(f_fwd_sprd, flt_freq_sprd(), d_, flt_dc_sprd(), flt_cal(), flt_bd(), flt_eom());
		if(!null(f_fwd_flat)) 
			f_fwd_flat_ = fwd_func(f_fwd_flat, flt_freq_flat(), d_ , flt_dc_flat(), flt_cal(), flt_bd(), flt_eom() );
	//}
	
	init_fx_swap(param.fx_spot_);
	blend_curve();	
}
//-------------------------------------------------------
// init_ir_index_legacy  [LEGACY]
//-------------------------------------------------------
void swap_curve_bc_ext.init_ir_index_legacy(instrument bi)
{
	integer freq_flat = bi.swap_coupon_freq_leg2();
	QL_REQUIRE(freq_flat > 0 && freq_flat != 365,'invalid float coupon freq (flat leg) in currency basis swap',bi.name(),true);

	integer freq_sprd = bi.swap_coupon_freq_leg1();
	QL_REQUIRE(freq_sprd > 0 && freq_sprd != 365,'invalid float coupon freq (spread leg) in currency basis swap',bi.name(),true);

	swap_fltflt sw = swap_fltflt(bi);
		
	calendar cal = sw.calendar();
	//string cal_name = sw.calendar_name();
	string cal_name,cal_name2; 
	sw.calendar_name(cal_name,cal_name2);
	//calendar cal = bi.calendar();
	//string cal_name = bi.calendar_name();
	QL_REQUIRE(!null(cal),'invalid calendar in currency basis swap',bi.name(),true);	 

	day_count_method  dc_sprd = bi.swap_dc_method_leg1(); 
	QL_REQUIRE(!null(dc_sprd),'invalid float daycount method (spread leg) in currency basis swap',bi.name(),true);

	day_count_method  dc_flat = bi.swap_dc_method_leg2() ;
	QL_REQUIRE(!null(dc_flat),'invalid float daycount method (flat leg) in currency basis swap',bi.name(),true);

	bd_convention bd	= bi.pmt_bus_day_d(BD_MOD_FOLLOWING);	
	QL_REQUIRE(!null(bd),'invalid pmt_bus_day in currency basis swap',bi.name(),true);	 

	logical eom	= bi.eom_d(true);

	string ccy_sprd = bi.swap_currency_leg1();
	QL_REQUIRE(!null(ccy_sprd),'invalid currency (spread leg) in currency basis swap',bi.name(),true);	

	string ccy_flat = bi.swap_currency_leg2();
	QL_REQUIRE(!null(ccy_flat),'invalid currency (flat leg) in currency basis swap',bi.name(),true);	

	QL_REQUIRE(!equal_casei(ccy_sprd,ccy_flat),'invalid currency basis swap (same currency in both legs)'); 

	integer rdays = integer(bi.flt_reset_days_d(2));
	QL_REQUIRE(rdays >= 0,'invalid flt_reset_days in basis curve ',bi.name(),true);	

	//string sd 		= strcat("BD",str(rdays));	
	integer spot_days 	= rdays;
	integer value_days 	= spot_days;
	string tenor_flat 	= strcat(str(12/freq_flat),"M");
	string tenor_sprd 	= strcat(str(12/freq_sprd),"M");

	if(null(cal_name))
		cal_name = "unknown";
	
	fwd_idx_flat_ = ir_index_ibor(	strcat("Ibor-",ccy_flat),tenor_flat,value_days,rdays, ccy_flat,
							cal,cal_name, dc_flat,bd,eom,spot_days);

	QL_REQUIRE(!null(fwd_idx_flat_),'failed to create ir_index (flat)');

	fwd_idx_sprd_ = ir_index_ibor(	strcat("Ibor-",ccy_sprd),tenor_sprd,value_days,rdays, ccy_sprd,
							cal,cal_name,dc_sprd,bd,eom,spot_days);

	QL_REQUIRE(!null(fwd_idx_sprd_),'failed to create ir_index (sprd)');

	basis_is_mtm_ = bi.ccbs_is_mtm();				
}



//-------------------------------------------------------
// check_b1_curve()
//-------------------------------------------------------
void swap_curve_bc_ext.check_b1_curve(curve option(nullable) basis_crv)
{
	if(!check_curve(basis_crv))
		return;

	string err_name;
	
	//check type	
	QL_REQUIRE(CORE_INT.check_instr_type(basis_crv, instr_type.SWAP_FLTFLT,err_name),'invalid basis swap curve (wrong instrument type)', err_name, true);

	
	vector(instrument) vi 	= basis_crv.instruments();
	instrument bi 			= vi[0];
	//start
	date settle  			= date_code(bi.settle_code()).apply_fwd(d_, bi.calendar());		
	QL_REQUIRE(CORE_INT.check_bswap_issue(vi, settle, err_name ),'invalid currency basis swap (mixed start dates)', err_name, true);				

	//reset_freq
	/*integer flt_freq_sprd = -1;		
	QL_REQUIRE(CORE_INT.swap_fltflt_rfreq_check(vi,flt_freq_sprd,true,err_name),'invalid currency basis swap (mixed reset frequencies, spread leg)', err_name, true);
	QL_REQUIRE(flt_freq_sprd > 0 && flt_freq_sprd != 365,'invalid float reset freq (sprd leg) in currency basis swap',bi.name(),true);
	
	integer flt_freq_flat = -1;	
	QL_REQUIRE(CORE_INT.swap_fltflt_rfreq_check(vi,flt_freq_flat,false,err_name),'invalid currency basis swap (mixed reset frequencies, flat leg)', err_name, true);
	QL_REQUIRE(flt_freq_flat > 0 && flt_freq_flat != 365,'invalid float reset freq (flat leg) in currency basis swap',bi.name(),true);
	*/
	//cpn freq
	integer cfreq = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,cfreq,true,err_name),'invalid currency basis swap (mixed coupon frequencies - sprd leg)',err_name,true);
	cfreq = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,cfreq,false,err_name),'invalid currency basis swap (mixed coupon frequencies - flat leg)',err_name,true);
	
	//quotes
	QL_REQUIRE(check_quote(basis_crv,err_name),'invalid quote(s) in currency basis curve',err_name,true);

}

//-------------------------------------------------------
// init_check_b1_curve()
//-------------------------------------------------------
void swap_curve_bc_ext.init_check_b1_curve(	curve option(nullable) basis_crv,
											fwd_func_tenor	fwd_tenor_flat,
											fwd_func_tenor	fwd_tenor_sprd)
{
	if(!check_curve(basis_crv))
		return;

	check_b1_curve(basis_crv);//type,start,cfreq,quotes

	integer freq_flat 		= fwd_func_tenor_to_freq(fwd_tenor_flat);
	integer freq_sprd 		= fwd_func_tenor_to_freq(fwd_tenor_sprd);

	vector(instrument) vi	= basis_crv.instruments();
	instrument bi			= vi[0];
	string err_name;
	QL_REQUIRE(CORE_INT.swap_rfreq_check(vi,freq_sprd,true,err_name),'invalid currency basis swap (mixed reset frequencies, spread leg)', err_name, true);
	QL_REQUIRE(freq_sprd > 0 && freq_sprd != 365,'invalid float reset freq (sprd leg) in currency basis swap',bi.name(),true);
	
	QL_REQUIRE(CORE_INT.swap_rfreq_check(vi,freq_flat,false,err_name),'invalid currency basis swap (mixed reset frequencies, flat leg)', err_name, true);
	QL_REQUIRE(freq_flat > 0 && freq_flat != 365,'invalid float reset freq (flat leg) in currency basis swap',bi.name(),true);

	basis_is_mtm_ 			= vi[0].ccbs_is_mtm();
	spread_ 				= one_vector(v_size(vi), 0);
}
//-------------------------------------------------------
// post_check_b1_curve
//-------------------------------------------------------
void swap_curve_bc_ext.post_check_b1_curve(curve option(nullable) basis_crv)
{
	if(!check_curve(basis_crv))
		return;
	
	vector(instrument) v_i 	= basis_crv.instruments();
	 	
	string err_name;
	tenor_code t = tenor_code(flt_freq_flat());
	QL_REQUIRE(CORE_INT.swap_index_tenor_code_check(v_i, t, false,err_name),'invalid float leg tenor (flat leg) in currency basis swap', err_name, true);				
	t = tenor_code(flt_freq_sprd());
	QL_REQUIRE(CORE_INT.swap_index_tenor_code_check(v_i, t, true, err_name),'invalid float leg tenor (spread leg) in currency basis swap', err_name, true);			

	day_count_method dc_flat = flt_dc_flat();
	QL_REQUIRE(CORE_INT.swap_dc_check(v_i,dc_flat,false,err_name),'invalid float leg daycount method (flat leg) in currency basis swap', err_name, true);
	day_count_method dc_sprd = flt_dc_sprd();
	QL_REQUIRE(CORE_INT.swap_dc_check(v_i,dc_sprd,true,err_name),'invalid float leg daycount method (spread leg) in currency basis swap', err_name, true);	

	string curr = ccy_flat();
	QL_REQUIRE(CORE_INT.swap_currency_check(v_i, curr, false,err_name),'invalid float leg currency (flat leg) in currency basis swap', err_name, true);
	curr = ccy_sprd();
	QL_REQUIRE(CORE_INT.swap_currency_check(v_i, curr, true,err_name),'invalid float leg currency (spread leg) in currency basis swap', err_name, true);	
	QL_REQUIRE(CORE_INT.check_is_mtm(v_i, basis_is_mtm_, err_name),'invalid mtm flag in currency basis swap', err_name, true);	
}
//-------------------------------------------------------
// init_fx_swap
//-------------------------------------------------------
void swap_curve_bc_ext.init_fx_swap(fx_spot option(nullable) fxs)
{
	if(check_curve(param_.short_crv())) {
		string err_name;
		QL_REQUIRE(check_quote(param_.short_crv(),err_name),'invalid quote(s) in fx curve',err_name,true);
		fx_swaps_ = new fx_curve(param_.short_crv().instruments(),fxs, fxs, quote_side_code.QS_MID) ;
		QL_REQUIRE(!null(fx_swaps_),'invalid fx curve');
		if(!equal_casei(ccy_sprd(),fxs.base_ccy()) && !equal_casei(ccy_sprd(),fxs.price_ccy()))			
			QL_FAIL('invalid fx swap (incompatible with basis swap)'); 
		if(!equal_casei(ccy_flat(),fxs.base_ccy()) && !equal_casei(ccy_flat(),fxs.price_ccy()))			
			QL_FAIL('invalid fx swap (incompatible with basis swap)'); 
		fx_flat_eq_base_ = equal_casei(ccy_flat(),fxs.base_ccy());
	}
}

//-------------------------------------------------------
// blend_curve
//-------------------------------------------------------
void swap_curve_bc_ext.blend_curve()
{	
	curve synt_depo_flat,synt_depo_sprd;
	string err_msg;
	vector(date) excl_fx_mat;
	curve_ok_ = create_blend_curve(curve_flat_, curve_sprd_, synt_depo_flat,synt_depo_sprd,excl_fx_mat,err_msg);	
	QL_REQUIRE(curve_ok_,err_msg);

	string n;
	if(!null(curve_flat_) && !..check_quote(curve_flat_,n)) {
		string s = strcat(['invalid quote(s) in blended curve (flat leg) [',n,']']);
		QL_FAIL(s);
	}	
	if(!null(curve_sprd_) &&!..check_quote(curve_sprd_,n)) {
		string s = strcat(['invalid quote(s) in blended curve (spread leg) [',n,']']);
		QL_FAIL(s);
	}
}
//-------------------------------------------------------
// basis_curve
//-------------------------------------------------------
curve swap_curve_bc_ext.basis_curve()
{ 	
	if(param_.type() == swap_curve_type.SCE_CURR_B1_F) {
		swap_curve_cb1_f_parm param = dynamic_cast<swap_curve_cb1_f_parm>(param_);
		return param.basis_crv();
	}
	else if(param_.type() == swap_curve_type.SCE_CURR_B1) {
		swap_curve_cb1_parm param = dynamic_cast<swap_curve_cb1_parm>(param_);
		return param.basis_crv_;
	}
	else
		QL_FAIL("invalid function call");
}
//-------------------------------------------------------
// create_blend_curve
//-------------------------------------------------------
logical swap_curve_bc_ext.create_blend_curve(	out curve option(nullable) curve_flat, 
												out curve option(nullable) curve_sprd, 
												out curve option(nullable) synt_depo_flat, 
												out curve option(nullable) synt_depo_sprd,
												out vector(date)   excl_fx_mat, 
												out string err_msg)
{
	if(null(fx_swaps_)){
		curve_flat = curve_sprd = basis_curve();
		synt_depo_flat = synt_depo_sprd = null<curve>;
		err_msg = "";
		return true;
	}
	interpolator ip	= ip_spline(null<number>, 0,IP_NO_XPOL_R) ;
	rate_type ip_rt 	= RT_CONT;

	vector(date) knots;			
	curve_prio	prio1 = param_.prio1() == curve_prio.SHORT ? curve_prio.SHORT : curve_prio.LONG; 
	curve_prio	prio2 = param_.prio1() == curve_prio.SHORT ? curve_prio.LONG : curve_prio.SHORT; 

	if(!null(df_disc_flat_)){//we can create depos from df_disc_flat		
						
		if(fx_flat_eq_base_) {
			synt_depo_sprd = fx_swaps_.synt_depo_price(df_disc_flat_,ip,ip_rt, flt_dc_sprd(), excl_fx_mat,true);//SEK depo			
		}
		else {
			synt_depo_sprd = fx_swaps_.synt_depo_base(df_disc_flat_,ip,ip_rt, flt_dc_sprd(),excl_fx_mat,true);//SEK depo
		}
		if(null(synt_depo_sprd)) {
			err_msg = 'invalid creation of deposit curve (spread leg)';
			return false;
		}		
		curve_sprd = CORE_INT.blend_curve(	synt_depo_sprd, null<curve>, basis_curve(), prio1, prio2, false, false,
											param_.merge_crv(), knots, param_.blend_buf_days());
		if(!..check_curve(curve_sprd)) {
			err_msg = 'invalid blended curve (spread leg)';
			return false;
		}				
		curve_sprd = curve_sprd.set_name("blended curve (spread leg)");	
		
	}
	if(!null(df_disc_sprd_)){		
		if(fx_flat_eq_base_) {
			synt_depo_flat = fx_swaps_.synt_depo_base(df_disc_sprd_,ip,ip_rt, flt_dc_flat(),excl_fx_mat,true);//USD depo
		}
		else {
			synt_depo_flat = fx_swaps_.synt_depo_price(df_disc_sprd_,ip,ip_rt, flt_dc_flat(),excl_fx_mat,true);//USD depo
		}
		if(null(synt_depo_flat)) {
			err_msg = 'invalid creation of deposit curve (flat leg)';
			return false;
		}
		
		curve_flat = CORE_INT.blend_curve(synt_depo_flat,null<curve>, basis_curve(),prio1,prio2,false,false,
										  param_.merge_crv(),knots,param_.blend_buf_days());

		if(!..check_curve(curve_flat)) {
			err_msg = 'invalid blended curve (flat leg)';
			return false;
		}			

		curve_flat = curve_flat.set_name("blended curve (flat leg)");		
	}

	err_msg = "";
	return true;	
}
//-------------------------------------------------------
// blended_curve
//-------------------------------------------------------
curve swap_curve_bc_ext.blended_curve()
{ 
	if(!curve_ok_)
		return null<curve>;

	if(null(fx_swaps_))
		return basis_curve();		

	vector(date) knots;			
	curve_prio	prio1 	= param_.prio1() == curve_prio.SHORT ? curve_prio.SHORT : curve_prio.LONG; 
	curve_prio	prio2 	= param_.prio1() == curve_prio.SHORT ? curve_prio.LONG : curve_prio.SHORT; 
	curve curve_all 	= CORE_INT.blend_curve(param_.short_crv(), null<curve>, basis_curve(), prio1, prio2, false, false, 
												param_.merge_crv(), knots, param_.blend_buf_days());
		
	curve curve_flat, curve_sprd, synt_depo_flat,synt_depo_sprd;
	vector(date) excl_fx_mat;
	string err_msg;
	logical curve_ok = create_blend_curve(curve_flat, curve_sprd, synt_depo_flat, synt_depo_sprd, excl_fx_mat, err_msg);	
	if(!curve_ok)
		return null<curve>;
	
	if(v_size(excl_fx_mat) == 0)
		return curve_all;
	else {//here some fxswaps have been filtered out
		vector(instrument) vi_all 	= curve_all.instruments();
		vector(date) vi_all_m 		= vi_all.maturity();
		vector(integer) index_vector = v_find(vi_all_m, excl_fx_mat) ;		
		vector(instrument) vi 		= v_exclude(vi_all, index_vector) ;
		return curve(vi);
	} 
}
//-------------------------------------------------------
// blended_flat_curve
//-------------------------------------------------------
curve swap_curve_bc_ext.blended_flat_curve()
{ 
	if(!curve_ok_ || null(df_disc_sprd_))
		return null<curve>;
	return curve_flat_;
}
//-------------------------------------------------------
// blended_sprd_curve
//-------------------------------------------------------
curve swap_curve_bc_ext.blended_sprd_curve()
{ 
	if(!curve_ok_ || null(df_disc_flat_))
		return null<curve>;
	return curve_sprd_;
}

//  [LEGACY]
curve swap_curve_bc_ext.blended_base_curve() { 	return blended_sprd_curve(); }

//-------------------------------------------------------
// cf
//-------------------------------------------------------
cashflows swap_curve_bc_ext.cf(	curve		crv,
								disc_func 	df_disc,
								logical 	solve_flat)
{		
	vector(instrument) src	= crv.instruments();
	integer size 			= v_size(src);
	vector(cashflows) dest[0];	
	number fx_ois_adj_sprd 	= 0;
	integer b 				= 0;
		
	for(integer i = 0; i < size ; i++){
		cashflows c;					
		if(src[i].is_swap_fltflt()) {
			c = CORE_INT.cf_currbswap(	src[i],fx_ois_adj_sprd,solve_flat,f_fwd_sprd_,first_flt_sprd(), 
									f_fwd_flat_, first_flt_flat(), df_disc, spread_[b]);//nom=1
			b++;			
		}
		else if(src[i].is_deposit() || src[i].is_deposit_index()) 
			c = CORE_INT.cf_depo(src[i], 0);	//nom=1			
		else
			QL_FAIL("invalid instrument type (currency basis)");

		QL_REQUIRE(!null(c),"invalid cashflow (currency basis)", src[i].name(), true);
		push_back(	dest, c);	
	}
	return cashflows(dest);
}
//-------------------------------------------------------
// cf_mtm
//-------------------------------------------------------
cashflows_q swap_curve_bc_ext.cf_mtm(curve		crv,
									disc_func 	df_disc,
									logical 	solve_flat)
{		
	vector(instrument) src	= crv.instruments();
	integer size 			= v_size(src);
	vector(cashflows_q) dest[0];	
	number fx_ois_adj_sprd 	= 0;
	integer b 				= 0;

	for(integer i = 0; i < size ; i++){
		cashflows_q c;					
		if(src[i].is_swap_fltflt()) {
			c = CORE_INT.cf_currbswap_mtm(	src[i],fx_ois_adj_sprd,solve_flat,f_fwd_sprd_,first_flt_sprd(), 
											f_fwd_flat_, first_flt_flat(), df_disc, spread_[b]);//nom=1
			b++;			
		}
		else if(src[i].is_deposit() || src[i].is_deposit_index()) 
			c = cashflows_q(CORE_INT.cf_depo(src[i], 0)) ;	//nom=1			
		else
			QL_FAIL("invalid instrument type (currency basis)");

		QL_REQUIRE(!null(c),"invalid cashflow (currency basis)", src[i].name(), true);
		push_back(	dest, c);	
	}
	return cashflows_q(dest);
}
//-------------------------------------------------------
// adj_disc_df_sprd
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ext.adj_disc_df_sprd(	interpolator 	disc_ip,
												rate_type		disc_ip_rt)
							   
{	
	QL_REQUIRE(!null(df_disc_flat_),"invalid df_disc_flat");

	if(basis_is_mtm_ ) {
		// QL_FAIL("mtm swaps not supported");

		cashflows_q cflow = cf_mtm(curve_sprd_, df_disc_flat_, false);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (spread leg)");
		if(null(df_disc_sprd_))
			return bootstrap(cflow, disc_ip, disc_ip_rt);		
		else
			return bootstrap_spr(df_disc_sprd_, cflow, disc_ip, disc_ip_rt);
	}
	else {
		cashflows cflow = cf(curve_sprd_, df_disc_flat_, false);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (spread leg)");
		if(null(df_disc_sprd_))
			return bootstrap(cflow, disc_ip, disc_ip_rt);		
		else
			return bootstrap_spr(df_disc_sprd_, cflow, disc_ip, disc_ip_rt);
	}
}

// [LEGACY]
disc_func swap_curve_bc_ext.adj_disc_df_base(interpolator disc_ip, rate_type disc_ip_rt) { return adj_disc_df_sprd(disc_ip,disc_ip_rt);}

//-------------------------------------------------------
// adj_disc_df_flat
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ext.adj_disc_df_flat(	interpolator 	disc_ip,
												rate_type		disc_ip_rt)
							   
{	
	QL_REQUIRE(!null(df_disc_sprd_),"invalid df_disc_sprd");
	if(basis_is_mtm_) {
//		QL_FAIL("mtm swaps not supported");
		cashflows_q cflow = cf_mtm(curve_flat_, df_disc_sprd_, true);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (flat leg)");
		if(null(df_disc_flat_))
			return bootstrap(cflow, disc_ip, disc_ip_rt);		
		else
			return bootstrap_spr(df_disc_flat_, cflow, disc_ip, disc_ip_rt);
	}
	else {
		cashflows cflow = cf(curve_flat_, df_disc_sprd_, true);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (flat leg)");
		if(null(df_disc_flat_))
			return bootstrap(cflow, disc_ip, disc_ip_rt);		
		else
			return bootstrap_spr(df_disc_flat_, cflow, disc_ip, disc_ip_rt);
	}
}

//-------------------------------------------------------
// adj_disc_df_sprd_int
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ext.adj_disc_df_sprd_int(disc_z_model disc_model, 
												disc_func option(nullable) df_disc_sprd_base,
												logical use_base_df)
							   
{	
	QL_REQUIRE(!null(df_disc_flat_),"invalid df_disc_flat");
	if(null(df_disc_sprd_base)) df_disc_sprd_base = df_disc_sprd_;

	if(basis_is_mtm_ ) {
		cashflows_q cflow = cf_mtm(curve_sprd_, df_disc_flat_, false);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (spread leg)");
		if(!use_base_df)  
			return disc_model.df(cflow, curve_sprd_.instruments().pvbp());	
		else 			
			return disc_model.df(df_disc_sprd_base, cflow, curve_sprd_.instruments().pvbp());		
	}
	else {
		cashflows cflow = cf(curve_sprd_, df_disc_flat_, false);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (spread leg)");
		if(!use_base_df) 
			return disc_model.df(cflow, curve_sprd_.instruments().pvbp());
		else 
			return disc_model.df(df_disc_sprd_base,cflow, curve_sprd_.instruments().pvbp());
	}
}

//-------------------------------------------------------
// adj_disc_df_sprd
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ext.adj_disc_df_sprd(disc_z_model disc_model)
							   
{	
	return adj_disc_df_sprd_int(disc_model, df_disc_sprd_, !null(df_disc_sprd_));
}
//-------------------------------------------------------
// adj_disc_df_sprd
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ext.adj_disc_df_sprd(disc_z_model disc_model, disc_func option(nullable) df_disc_sprd_base)
							   
{	
	return adj_disc_df_sprd_int(disc_model, df_disc_sprd_base, !null(df_disc_sprd_base));
}
//-------------------------------------------------------
// adj_disc_df_flat_int
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ext.adj_disc_df_flat_int(	disc_z_model disc_model, 
													disc_func option(nullable) df_disc_flat_base,
													logical use_base_df)
							   
{	
	QL_REQUIRE(!null(df_disc_sprd_),"invalid df_disc_sprd");
	if(basis_is_mtm_) {
		cashflows_q cflow = cf_mtm(curve_flat_, df_disc_sprd_, true);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (flat leg)");
		if(!use_base_df) 
			return disc_model.df(cflow, curve_flat_.instruments().pvbp());			
		else 
			return disc_model.df(df_disc_flat_base,cflow, curve_flat_.instruments().pvbp());			
	}
	else {
		cashflows cflow = cf(curve_flat_, df_disc_sprd_, true);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (flat leg)");
		if(!use_base_df)
			return disc_model.df(cflow, curve_flat_.instruments().pvbp());		
		else 
			return disc_model.df(df_disc_flat_base,cflow, curve_flat_.instruments().pvbp());			
	}
}
//-------------------------------------------------------
// adj_disc_df_flat
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ext.adj_disc_df_flat(disc_z_model disc_model)
							   
{	
	return adj_disc_df_flat_int(disc_model, df_disc_flat_, !null(df_disc_flat_));
}
//-------------------------------------------------------
// adj_disc_df_flat
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ext.adj_disc_df_flat(disc_z_model disc_model, disc_func option(nullable) df_disc_flat_base)
							   
{	
	return adj_disc_df_flat_int(disc_model, df_disc_flat_base, !null(df_disc_flat_base));
}
//-------------------------------------------------------
// solve_spread_flat_leg
//-------------------------------------------------------
vector(number) swap_curve_bc_ext.solve_spread_flat_leg(vector(number) option(nullable) fx_lib_sprd)
{
	vector(instrument) v_i = basis_curve().instruments();
	integer size = v_size(v_i) ;
	vector(number) sprd;

	//if fx_lib_sprd is 0 then we're switching the basis spread to the flat leg i.e. base->flat and flat->base 
	//if fx_lib_sprd is equal to the quoted basis swap then the spread is zero (i.e. flat) since one disc_func is an fx implied
	if(null(fx_lib_sprd)){
		resize(fx_lib_sprd, size);
		fx_lib_sprd = 0;
	}
	else 
		QL_REQUIRE(v_size(fx_lib_sprd) == size,"invalid vector size (basis spread vector)");

	interpolator 	disc_ip 	= ip_linear();
	rate_type		disc_ip_rt 	= RT_LOG_DF;
			
	integer freq_sprd 			= flt_freq_sprd();
	integer freq_solve 			= flt_freq_flat(); 
	day_count_method dc_sprd 	= flt_dc_sprd();
	day_count_method dc_solve 	= flt_dc_flat();
	fwd_func f_fwd_sprd			= f_fwd_sprd_;											
	fwd_func f_fwd_solve		= f_fwd_flat_;

	disc_func df_fx_sprd, df_fx_solve;
	if(null(df_disc_flat_)){
		QL_REQUIRE(!null(df_disc_sprd_),"invalid disc_func (either flat or spread (base) leg disc_func must be input)");
		df_fx_sprd  = df_disc_sprd_;
		df_fx_solve	= adj_disc_df_flat(disc_ip,disc_ip_rt);
	}
	else {
		
		df_fx_sprd  = adj_disc_df_sprd(disc_ip,disc_ip_rt);
		df_fx_solve	=  df_disc_flat_;
	}

	for(integer i = 0; i < size ; i++){
			
		number s = ..flt_sprd_fx_lib(d_,v_i[i].issue_date(),v_i[i].maturity(),day(v_i[i].issue_date()),
									freq_sprd,freq_solve,dc_sprd,dc_solve,
									flt_bd(),flt_cal(),flt_eom(),						
									f_fwd_sprd,f_fwd_solve,
									df_fx_sprd,df_fx_solve,
									fx_lib_sprd[i]);
		push_back(sprd,s);
	}
	return sprd;
}
//-------------------------------------------------------
// solve_spread_sprd_leg
//-------------------------------------------------------
vector(number) swap_curve_bc_ext.solve_spread_sprd_leg(vector(number) option(nullable) flat_sprd)
{
	if(basis_is_mtm_)
		QL_FAIL("mtm swaps not supported for this function");

	vector(instrument) v_i = basis_curve().instruments();
	integer size = v_size(v_i) ;
	vector(number) sprd;

	//if flat_sprd is 0 then we get the quoted spread 	
	if(null(flat_sprd)){
		resize(flat_sprd, size);
		flat_sprd = 0;
	}
	else 
		QL_REQUIRE(v_size(flat_sprd) == size,"invalid vector size (flat spread vector)");

	interpolator 	disc_ip 	= ip_linear();
	rate_type		disc_ip_rt 	= RT_LOG_DF;
					
	integer freq_sprd 			= flt_freq_flat();
	integer freq_solve 			= flt_freq_sprd(); 
	day_count_method dc_sprd 	= flt_dc_flat();
	day_count_method dc_solve 	= flt_dc_sprd();
	fwd_func f_fwd_sprd			= f_fwd_flat_;											
	fwd_func f_fwd_solve		= f_fwd_sprd_;

	disc_func df_fx_sprd, df_fx_solve;
	if(null(df_disc_flat_)){
		QL_REQUIRE(!null(df_disc_sprd_),"invalid disc_func (either flat or spread (base) leg disc_func must be input)");
		df_fx_sprd  = adj_disc_df_flat(disc_ip,disc_ip_rt) ;
		df_fx_solve	= df_disc_sprd_;
	}
	else {
		df_fx_sprd  = df_disc_flat_;
		df_fx_solve	= adj_disc_df_sprd(disc_ip,disc_ip_rt) ;
	}

	for(integer i = 0; i < size ; i++){
	
		number s = ..flt_sprd_fx_lib(d_,v_i[i].issue_date(),v_i[i].maturity(),day(v_i[i].issue_date()),
									freq_sprd,freq_solve,dc_sprd,dc_solve,
									flt_bd(),flt_cal(),flt_eom(),						
									f_fwd_sprd,f_fwd_solve,
									df_fx_sprd,df_fx_solve,
									flat_sprd[i]);
		push_back(sprd,s);
	}
	return sprd;
}

// [LEGACY]
vector(number) swap_curve_bc_ext.solve_spread_base_leg(vector(number) option(nullable) flat_sprd) { return solve_spread_sprd_leg(flat_sprd);}

//-------------------------------------------------------
// solve_ois_ois_spread_sprd_leg
//-------------------------------------------------------
vector(number) swap_curve_bc_ext.solve_ois_ois_spread_sprd_leg()
{
	if(basis_is_mtm_)
		QL_FAIL("mtm swaps not supported for this function");

	vector(instrument) v_i = basis_curve().instruments();
	integer size = v_size(v_i) ;
	vector(number) sprd;
	
	//interpolator 	disc_ip 	= ip_linear();
	//rate_type		disc_ip_rt 	= RT_CONT;

	integer freq_sprd 			= flt_freq_flat();
	integer freq_solve 			= flt_freq_sprd(); 
	day_count_method dc_sprd 	= flt_dc_flat();
	day_count_method dc_solve 	= flt_dc_sprd();
	fwd_func f_fwd_sprd			= f_fwd_flat_;											
	fwd_func f_fwd_solve		= f_fwd_sprd_;		

	QL_REQUIRE(!null(df_disc_sprd_),"invalid spread (base) leg disc_func");
	QL_REQUIRE(!null(df_disc_flat_),"invalid flat leg disc_func");

	disc_func df_ois_solve		= df_disc_sprd_;
	disc_func df_ois_sprd  		= df_disc_flat_;

	for(integer i = 0; i < size ; i++){
						
		number s = ..flt_sprd_ois_ois(d_,v_i[i].issue_date(),v_i[i].maturity(),day(v_i[i].issue_date()),
										freq_sprd,freq_solve,dc_sprd,dc_solve,
										flt_bd(),flt_cal(),flt_eom(),						
										f_fwd_sprd,f_fwd_solve,
										df_ois_sprd,df_ois_solve);
		push_back(sprd,s);
	}
	return sprd;
}

// [LEGACY]
vector(number) swap_curve_bc_ext.solve_ois_ois_spread_base_leg(){ return solve_ois_ois_spread_sprd_leg();}
//-------------------------------------------------------
// solve_ois_ois_spread_flat_leg
//-------------------------------------------------------
vector(number) swap_curve_bc_ext.solve_ois_ois_spread_flat_leg()
{
	if(basis_is_mtm_)
		QL_FAIL("mtm swaps not supported for this function");

	vector(instrument) v_i = basis_curve().instruments();
	integer size = v_size(v_i) ;
	vector(number) sprd;
	
	//interpolator 	disc_ip 	= ip_linear();
	//rate_type		disc_ip_rt 	= RT_CONT;

	integer freq_sprd 			= flt_freq_sprd();
	integer freq_solve 			= flt_freq_flat(); 
	day_count_method dc_sprd 	= flt_dc_sprd();
	day_count_method dc_solve 	= flt_dc_flat();
	fwd_func f_fwd_sprd			= f_fwd_sprd_;											
	fwd_func f_fwd_solve		= f_fwd_flat_;		

	QL_REQUIRE(!null(df_disc_sprd_),"invalid spread (base) leg disc_func");
	QL_REQUIRE(!null(df_disc_flat_),"invalid flat leg disc_func");

	disc_func df_ois_solve	= df_disc_flat_;
	disc_func df_ois_sprd  = df_disc_sprd_;

	for(integer i = 0; i < size ; i++){
						
		number s = ..flt_sprd_ois_ois(d_,v_i[i].issue_date(),v_i[i].maturity(),day(v_i[i].issue_date()),
										freq_sprd,freq_solve,dc_sprd,dc_solve,
										flt_bd(),flt_cal(),flt_eom(),						
										f_fwd_sprd,f_fwd_solve,
										df_ois_sprd,df_ois_solve);
		push_back(sprd,s);
	}
	return sprd;
}
//-------------------------------------------------------
// solve_fx_ois_spread_sprd_leg
//-------------------------------------------------------
vector(number) swap_curve_bc_ext.solve_fx_ois_spread_sprd_leg()
{
	if(basis_is_mtm_)
		QL_FAIL("mtm swaps not supported for this function");

	vector(instrument) v_i = basis_curve().instruments();
	integer size = v_size(v_i) ;
	vector(number) fx_lib_bswap = v_i.quote()/10000;
	vector(number) sprd;
	
	//interpolator 	disc_ip 	= ip_linear();
	//rate_type		disc_ip_rt 	= RT_CONT;

	integer freq_sprd 			= flt_freq_flat();
	integer freq_solve 			= flt_freq_sprd(); 
	day_count_method dc_sprd 	= flt_dc_flat();
	day_count_method dc_solve 	= flt_dc_sprd();
	fwd_func f_fwd_sprd			= f_fwd_flat_;											
	fwd_func f_fwd_solve		= f_fwd_sprd_;		

	QL_REQUIRE(!null(df_disc_sprd_),"invalid spread (base) leg disc_func");
	QL_REQUIRE(!null(df_disc_flat_),"invalid flat leg disc_func");

	disc_func df_ois_sprd  = df_disc_flat_;		
	disc_func df_ois_solve	= df_disc_sprd_;

	for(integer i = 0; i < size ; i++){
						
		number s = ..flt_sprd_fx_ois_imp(d_,v_i[i].issue_date(),v_i[i].maturity(),day(v_i[i].issue_date()),
											freq_sprd,freq_solve,dc_sprd,dc_solve,
											flt_bd(),flt_cal(),flt_eom(),							
											f_fwd_sprd,f_fwd_solve,								
											df_ois_sprd,df_ois_solve,//domestic ois							
											fx_lib_bswap[i]);
		push_back(sprd,s);
	}
	return sprd;
}
vector(number) swap_curve_bc_ext.solve_fx_ois_spread_base_leg() {return solve_fx_ois_spread_sprd_leg();}

//-------------------------------------------------------
// solve_fx_ois_spread_flat_leg
//-------------------------------------------------------
vector(number) swap_curve_bc_ext.solve_fx_ois_spread_flat_leg()
{
	if(basis_is_mtm_)
		QL_FAIL("mtm swaps not supported for this function");

	vector(instrument) v_i = basis_curve().instruments();
	integer size = v_size(v_i) ;
	vector(number) fx_lib_bswap = solve_spread_flat_leg();
	vector(number) sprd;
	
	//interpolator 	disc_ip 	= ip_linear();
	//rate_type		disc_ip_rt 	= RT_CONT;
	
	integer freq_sprd 			= flt_freq_sprd();
	integer freq_solve 			= flt_freq_flat(); 
	day_count_method dc_sprd 	= flt_dc_sprd();
	day_count_method dc_solve 	= flt_dc_flat();
	fwd_func f_fwd_sprd			= f_fwd_sprd_;											
	fwd_func f_fwd_solve		= f_fwd_flat_;		

	QL_REQUIRE(!null(df_disc_sprd_),"invalid spread (base) leg disc_func");
	QL_REQUIRE(!null(df_disc_flat_),"invalid flat leg disc_func");
	
	disc_func df_ois_sprd  = df_disc_sprd_;		
	disc_func df_ois_solve	= df_disc_flat_;

	for(integer i = 0; i < size ; i++){
								
		number s = ..flt_sprd_fx_ois_imp(d_,v_i[i].issue_date(),v_i[i].maturity(),day(v_i[i].issue_date()),
											freq_sprd,freq_solve,dc_sprd,dc_solve,
											flt_bd(),flt_cal(),flt_eom(),							
											f_fwd_sprd,f_fwd_solve,								
											df_ois_sprd,df_ois_solve,//domestic ois							
											fx_lib_bswap[i]);
		push_back(sprd,s);
	}
	return sprd;
}
//-------------------------------------------------------
// flt_reset_days_sprd
//-------------------------------------------------------
integer swap_curve_bc_ext.flt_reset_days_sprd() 
{ 
	integer fd = fwd_idx_sprd_.fixing_days();
	QL_REQUIRE(fd >= 0,'invalid flt_reset_days in ir_index (negative days not supported) (fwd)');
	return fd;
}
//-------------------------------------------------------
// flt_reset_days_flat
//-------------------------------------------------------
integer swap_curve_bc_ext.flt_reset_days_flat() 
{ 
	integer fd = fwd_idx_flat_.fixing_days();
	QL_REQUIRE(fd >= 0,'invalid flt_reset_days in ir_index (negative days not supported) (fwd)');
	return fd;
}
//-------------------------------------------------------
// flt_reset_days
//-------------------------------------------------------
integer swap_curve_bc_ext.flt_reset_days() 
{ 
	return flt_reset_days_flat();
}
//-------------------------------------------------------
// flt_reset_date_flat
//-------------------------------------------------------
date swap_curve_bc_ext.flt_reset_date_flat() 
{ 
	date r = d_;
	while(!fwd_idx_flat_.is_valid_fixing_date(r)) 
		r = flt_cal_flat().move_bus_days(r, 1);
	return r;
}
//-------------------------------------------------------
// flt_reset_date_sprd
//-------------------------------------------------------
date swap_curve_bc_ext.flt_reset_date_sprd() 
{ 
	date r = d_;
	while(!fwd_idx_sprd_.is_valid_fixing_date(r)) 
		r = flt_cal_sprd().move_bus_days(r, 1);
	return r;
}
//-------------------------------------------------------
// flt_reset_date
//-------------------------------------------------------
date swap_curve_bc_ext.flt_reset_date() 
{ 
	return flt_reset_date_flat();
}
//-------------------------------------------------------
// first_flt_flat 
//-------------------------------------------------------
number swap_curve_bc_ext.first_flt_flat()
{	
	if(param_.type() == swap_curve_type.SCE_CURR_B1_F){
		swap_curve_cb1_f_parm param = dynamic_cast<swap_curve_cb1_f_parm>(param_);
		string 	err_msg;
		return null(param.fixing_crv_flat()) ? null<number>: param.fixing_crv_flat().fixing_rate(d_, fwd_idx_flat_.tenor(), err_msg);
	}
	else if(param_.type() == swap_curve_type.SCE_CURR_B1){
		swap_curve_cb1_parm param = dynamic_cast<swap_curve_cb1_parm>(param_);
		return param.basis_first_fix_flat_;
	}
	else
		QL_FAIL("invalid function call");
}
//-------------------------------------------------------
// first_flt_sprd 
//-------------------------------------------------------
number swap_curve_bc_ext.first_flt_sprd()
{	
	if(param_.type() == swap_curve_type.SCE_CURR_B1_F){
		swap_curve_cb1_f_parm param = dynamic_cast<swap_curve_cb1_f_parm>(param_);
		string 	err_msg;
		return null(param.fixing_crv_sprd()) ? null<number>: param.fixing_crv_sprd().fixing_rate(d_, fwd_idx_sprd_.tenor(), err_msg);
	}
	else if(param_.type() == swap_curve_type.SCE_CURR_B1){
		swap_curve_cb1_parm param = dynamic_cast<swap_curve_cb1_parm>(param_);
		return param.basis_first_fix_sprd_;
	}
	else
		QL_FAIL("invalid function call");
}
//-------------------------------------------------------
// create
//-------------------------------------------------------
swap_curve_bc_ext swap_curve_bc_ext(date						d,									
									swap_curve_cb1_parm			p,									
									disc_func option(nullable)	df_disc_flat,
									disc_func option(nullable)	df_disc_sprd,
									fwd_func					f_fwd_flat,																	
									fwd_func 					f_fwd_sprd,									
									string option(nullable)		name)
{
	return new swap_curve_bc_ext(d, p,df_disc_flat,df_disc_sprd, f_fwd_flat,f_fwd_sprd,name)	;	
}

//-------------------------------------------------------
// create
//-------------------------------------------------------
swap_curve_bc_ext swap_curve_bc_ext(date						d,									
									swap_curve_cb1_parm			p,									
									disc_func option(nullable)	df_disc_flat,
									disc_func option(nullable)	df_disc_sprd,
									disc_func					f_fwd_flat,																	
									disc_func 					f_fwd_sprd,									
									string option(nullable)		name)
{
	return new swap_curve_bc_ext(d, p,df_disc_flat,df_disc_sprd, f_fwd_flat,f_fwd_sprd,name)	;	
}

//-------------------------------------------------------
// create
//-------------------------------------------------------
swap_curve_bc_ext swap_curve_bc_ext(date						d,									
									swap_curve_cb1_f_parm		p,									
									disc_func option(nullable)	df_disc_flat,
									disc_func option(nullable)	df_disc_sprd,
									fwd_func					f_fwd_flat,
									fwd_func_tenor fwd_tenor_flat,																		
									fwd_func 					f_fwd_sprd,
									fwd_func_tenor fwd_tenor_sprd,										
									string option(nullable)		name)
{
	return new swap_curve_bc_ext(d, p,df_disc_flat,df_disc_sprd, f_fwd_flat,fwd_tenor_flat,f_fwd_sprd,fwd_tenor_sprd,name)	;	
}



