/*	
	Basic swap curve functions for cross currency bootstrap
	Developed by Algorithmica Research, Magnus Nyström
	
	
	
	Comments:

	
*/
option(null: hard);	


//-------------------------------------------------------
// class swap_curve_cross_bc_ext
// class used for creating cross currency basis swaps
//-------------------------------------------------------
class swap_curve_cross_bc_ext : public swap_curve_bc_ext option (category: 'Yield Curve/003 Curve Building Cross Currency')
{
   	public: 
		swap_curve_cross_bc_ext(	swap_curve_bc_ext option(fixedref), swap_curve_bc_ext option(fixedref)) ;
		//swap_curve_cross_bc_ext(	swap_curve_cross_bc_ext) ;
		disc_func			adj_disc_df_sprd(interpolator,rate_type);
		disc_func			adj_disc_df_sprd(disc_z_model);
		disc_func			adj_disc_df_flat(interpolator,rate_type);		
		disc_func			adj_disc_df_flat(disc_z_model);	
						
		vector(number) 		solve_spread_sprd_leg(vector(number) option(nullable) s = null<vector(number)>);
		vector(number) 		solve_spread_flat_leg(vector(number) option(nullable) s = null<vector(number)>);

		vector(string)		names();

		//---here for backward compat only-----------
		vector(number) 		solve_spread_base_leg(vector(number) option(nullable) s = null<vector(number)>);
		disc_func			adj_disc_df_base(interpolator,rate_type);		
		//-------------------------------------------
	protected:
		
		void				init_cross_bc_curve();		
		cashflows 			implied_sprd_cashflows(	);
		vector(number)		flat_basis_swap_ip();			
		vector(number) 		solve_spread_sprd_leg(disc_func, vector(number) option(nullable) s = null<vector(number)> );

	private:
		swap_curve_bc_ext option(fixedref)	bc_crv_flat_;
		swap_curve_bc_ext option(fixedref) 	bc_crv_sprd_;		
		number								case_;
		vector(number) 						sprd_flat_;
		vector(number) 						sprd_sprd_;	
		vector(string)						names_;
		//logical 							curve_ok_;

		integer 			flt_freq_sprd_;
		day_count_method 	flt_dc_sprd_;	
		string				ccy_sprd_;

		integer 			flt_freq_flat_;
		day_count_method 	flt_dc_flat_;			
		string				ccy_flat_;
		
		bd_convention 		flt_bd_;
		calendar			flt_cal_;
		logical				flt_eom_;
		string				settle_code_;				
};

//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
/*swap_curve_cross_bc_ext.swap_curve_cross_bc_ext(swap_curve_cross_bc_ext s) 
									:crv_flat_(s.crv_flat_), crv_base_(s.crv_base_),curve_ok_(s.curve_ok_)
{}
*/

//-------------------------------------------------------
// constructor
//-------------------------------------------------------
swap_curve_cross_bc_ext.swap_curve_cross_bc_ext(swap_curve_bc_ext option(fixedref)		c_flat,									
												swap_curve_bc_ext option(fixedref)		c_sprd) 
												:bc_crv_flat_(c_flat), bc_crv_sprd_(c_sprd)
{
	init_cross_bc_curve();			
}		
//-------------------------------------------------------
// init_cross_bc_curve
//-------------------------------------------------------
void swap_curve_cross_bc_ext.init_cross_bc_curve() 												
{
	/*	swap_curve_cb1_parm param_;
		fx_curve			fx_swaps_;						
		vector(number) 		spread_;
		curve				curve_sprd_;
		curve				curve_flat_;
		logical				fx_flat_eq_base_;
		logical 			curve_ok_;*/

	string crv_f_sprd_ccy 	= bc_crv_flat_.ccy_sprd() 	;  
	string crv_f_flat_ccy 	= bc_crv_flat_.ccy_flat() 	; 
    string crv_b_sprd_ccy 	= bc_crv_sprd_.ccy_sprd() 	;  
	string crv_b_flat_ccy 	= bc_crv_sprd_.ccy_flat() 	;  			
	
	d_ 						= bc_crv_flat_.trade_date();
	flt_bd_ 				= bc_crv_flat_.flt_bd();
	flt_cal_ 				= calendar(bc_crv_flat_.flt_cal(), bc_crv_sprd_.flt_cal());
	flt_eom_ 				= bc_crv_flat_.flt_eom();
	settle_code_ 			= bc_crv_flat_.settle_code();
	
	//number size 			= v_size(bc_crv_sprd_.basis_curve().instruments());
	names_ 					= bc_crv_sprd_.basis_curve().instruments().maturity_code();
	
	if(	crv_f_sprd_ccy == crv_b_sprd_ccy) {//case 3

		QL_REQUIRE( crv_f_flat_ccy !=  crv_b_flat_ccy,"incompatible basis currencies");

		/*	ex. 	basis swap #1 - flat: 	[ EUR Flat vs USD + x ]
					basis swap #2 - base: 	[ CHF Flat vs USD + y ]
					--> we create new base	[ CHF + b  vs USD + x ]
		
					--> we have				[ EUR Flat vs CHF + b ]

			1. match: find basis quotes in crv_flat_ for every swap in crv_base_
			2. solve: convert quotes in crv_base_ so that spread on base leg matches spread on base leg in crv_flat_ (using implied disc_func for crv_b_base_ccy)
						--> [ CHF + b  vs USD + x ] 
			3. now we have [crv_f_flat_ccy-FLAT/crv_b_flat_ccy+SPRD] basis swap
						--> [ EUR Flat vs CHF + b ] 
		*/		
		
		name_			= strcat([crv_f_flat_ccy, crv_b_flat_ccy,"-BS "]);
		names_ 			= strcat(name_, names_);
		case_ 			= 3;
		
		f_fwd_sprd_ 	= bc_crv_sprd_.f_fwd_flat();
		flt_freq_sprd_	= bc_crv_sprd_.flt_freq_flat();
		flt_dc_sprd_	= bc_crv_sprd_.flt_dc_flat();	
		ccy_sprd_		= crv_b_flat_ccy;
		df_disc_sprd_	= bc_crv_sprd_.disc_df_flat();

		f_fwd_flat_ 	= bc_crv_flat_.f_fwd_flat();
		flt_freq_flat_	= bc_crv_flat_.flt_freq_flat();
		flt_dc_flat_	= bc_crv_flat_.flt_dc_flat();	
		ccy_flat_		= crv_f_flat_ccy;
		df_disc_flat_	= bc_crv_flat_.disc_df_flat();

		vector(number) sprd_flat_b 	= flat_basis_swap_ip()/10000;
		sprd_sprd_ 					= bc_crv_sprd_.solve_spread_flat_leg(sprd_flat_b);//b	
		resize(sprd_flat_,v_size(sprd_sprd_));
		sprd_flat_					= 0;

		//--> [ EUR Flat vs CHF + b ]
	}
	else if(crv_f_sprd_ccy == crv_b_flat_ccy) {//case 1
		QL_REQUIRE( crv_f_flat_ccy !=  crv_b_sprd_ccy,"incompatible basis currencies");

		/*	ex. 	basis swap #1 - flat: 	[ USD Flat vs EUR + x ]
					basis swap #2 - base: 	[ EUR Flat vs HUF + y ]
					--> we create new base	[ EUR + x  vs HUF + b ]
		
					--> we have				[ USD Flat vs HUF + b ]
		
			1. match: find basis quotes in crv_flat_ for every swap in crv_base_
			2. solve: convert quotes in crv_base_ so that spread on flat leg matches spread on base leg in crv_flat_ (using implied disc_func for crv_b_base_ccy)
						--> [ EUR + x vs HUF + b ] 
			3. now we have [crv_f_flat_ccy-FLAT/crv_b_base_ccy+SPRD] basis swap
						--> [ USD Flat vs HUF + b ] 
			
		*/
		name_			= strcat([crv_f_flat_ccy, crv_b_sprd_ccy,"-BS "]);
		names_ 			= strcat(name_, names_);
		case_ 			= 1;
		
		f_fwd_sprd_ 	= bc_crv_sprd_.f_fwd_sprd();
		flt_freq_sprd_	= bc_crv_sprd_.flt_freq_sprd();
		flt_dc_sprd_	= bc_crv_sprd_.flt_dc_sprd();	
		ccy_sprd_		= crv_b_sprd_ccy;
		df_disc_sprd_	= bc_crv_sprd_.disc_df_sprd();

		f_fwd_flat_ 	= bc_crv_flat_.f_fwd_flat();
		flt_freq_flat_	= bc_crv_flat_.flt_freq_flat();
		flt_dc_flat_	= bc_crv_flat_.flt_dc_flat();	
		ccy_flat_		= crv_f_flat_ccy;
		df_disc_flat_	= bc_crv_flat_.disc_df_flat();		

		vector(number) sprd_flat_b 	= flat_basis_swap_ip()/10000;
		sprd_sprd_ 					= bc_crv_sprd_.solve_spread_sprd_leg(sprd_flat_b);//b	
		resize(sprd_flat_,v_size(sprd_sprd_));
		sprd_flat_					= 0;
	
		//--> [ USD Flat vs HUF + b ] 

	}
	else if(crv_f_flat_ccy == crv_b_sprd_ccy) {//case 4
		QL_REQUIRE( crv_f_sprd_ccy !=  crv_b_flat_ccy,"incompatible basis currencies");

		
		/*	ex. 	basis swap #1 - flat: 	[ USD Flat vs CHF + x ]
					basis swap #2 - base: 	[ EUR Flat vs USD + y ]
					--> we create new base	[ EUR + z  vs USD Flat ]
					--> we have				[ CHF + x vs EUR + z ]
		
					--> we can create final	[ CHF Flat vs EUR + b ]
		
			1. solve: convert quotes in crv_base_ so that spread on base leg is flat (using implied disc_func for crv_b_base_ccy)
						--> [ EUR + z vs USD Flat ]  
			2. now we have [crv_f_base_ccy + SPRD / crv_b_flat_ccy + SPRD] basis swap
						--> [ CHF + x vs EUR + z ] 
			3. solve: convert quotes in created basis swap so that spread on flat leg becomes flat (using implied disc_func for crv_b_base_ccy)
						--> [ CHF Flat vs EUR + b ] 		
		*/ 

		name_			= strcat([crv_f_sprd_ccy, crv_b_flat_ccy,"-BS "]);				
		names_ 			= strcat(name_, names_);		

		case_ 			= 4; 
		
		f_fwd_sprd_ 	= bc_crv_sprd_.f_fwd_flat();
		flt_freq_sprd_	= bc_crv_sprd_.flt_freq_flat();
		flt_dc_sprd_	= bc_crv_sprd_.flt_dc_flat();	
		ccy_sprd_		= crv_b_flat_ccy;
		df_disc_sprd_	= bc_crv_sprd_.disc_df_flat();

		f_fwd_flat_ 	= bc_crv_flat_.f_fwd_sprd();
		flt_freq_flat_	= bc_crv_flat_.flt_freq_sprd();
		flt_dc_flat_	= bc_crv_flat_.flt_dc_sprd();	
		ccy_flat_		= crv_f_sprd_ccy;
		df_disc_flat_	= bc_crv_flat_.disc_df_sprd();

		sprd_flat_		= flat_basis_swap_ip()/10000;//x
		sprd_sprd_ 		= bc_crv_sprd_.solve_spread_flat_leg();//z, we're switching the basis spread to the flat leg i.e. base->flat and flat->base	
		
		//--> [ CHF + x vs EUR + z ] 

	}
	else if(crv_f_flat_ccy == crv_b_flat_ccy) {//case 2
		QL_REQUIRE( crv_f_sprd_ccy !=  crv_b_sprd_ccy,"incompatible basis currencies");

		/*	ex. 	basis swap #1 - flat: 	[ USD Flat vs EUR + x ]
					basis swap #2 - base: 	[ USD Flat vs CHF + y ]
					--> we have 			[ EUR + x  vs CHF + y ]
		
					--> we can create final [ EUR Flat vs CHF + b ]
		
			1. we have [crv_f_base_ccy + SPRD / crv_b_base_ccy + SPRD] basis swap
						--> [ EUR + x vs CHF + y ] 
		
			2. solve: convert quotes in created basis swap so that spread on flat leg becomes flat (using implied disc_func for crv_f_base_ccy)
						--> [ EUR Flat vs CHF + b ] 
		*/
		
		name_			= strcat([crv_f_sprd_ccy, crv_b_sprd_ccy,"-BS "]);
		names_ 			= strcat(name_, names_);
		case_ 			= 2;
		
		f_fwd_sprd_ 	= bc_crv_sprd_.f_fwd_sprd();
		flt_freq_sprd_	= bc_crv_sprd_.flt_freq_sprd();
		flt_dc_sprd_	= bc_crv_sprd_.flt_dc_sprd();	
		ccy_sprd_		= crv_b_sprd_ccy;
		df_disc_sprd_	= bc_crv_sprd_.disc_df_sprd();

		f_fwd_flat_ 	= bc_crv_flat_.f_fwd_sprd();
		flt_freq_flat_	= bc_crv_flat_.flt_freq_sprd();
		flt_dc_flat_	= bc_crv_flat_.flt_dc_sprd();	
		ccy_flat_		= crv_f_sprd_ccy;
		df_disc_flat_	= bc_crv_flat_.disc_df_sprd();
		
		sprd_flat_ 		= flat_basis_swap_ip()/10000;//x
		sprd_sprd_ 		= bc_crv_sprd_.basis_curve().instruments().quote()/10000;//y

		//--> [ EUR + x vs CHF + y ]

	}
	else
		QL_FAIL("incompatible basis currencies");		

}

vector(string) swap_curve_cross_bc_ext.names() {return names_;}

//-------------------------------------------------------
// implied_sprd_cashflows
//-------------------------------------------------------
cashflows swap_curve_cross_bc_ext.implied_sprd_cashflows()
{	
	QL_REQUIRE(!null(df_disc_flat_),"invalid df_disc_flat");
	QL_REQUIRE(v_size( sprd_sprd_) == v_size( sprd_flat_),"invalid vector size");

	vector(instrument) v_i_sprd = bc_crv_sprd_.basis_curve().instruments();		
	//number fx_ois_adj_sprd 		= 0;
	number first_fix_sprd		= null<number>;
	number first_fix_flat		= null<number>;
	logical solve_flat 			= false;	

	vector(cashflows) cflow[0];				
	for(integer i = 0; i < v_size(v_i_sprd) ; i++){
		QL_REQUIRE(!null(sprd_sprd_[i]),"invalid spread (base) leg spread");
		QL_REQUIRE(!null(sprd_flat_[i]),"invalid flat leg spread");						
		cashflows c = CORE_INT.cf_currbswap(d_,v_i_sprd[i].issue_date(),v_i_sprd[i].maturity(),day(v_i_sprd[i].issue_date()),						
											flt_freq_sprd_,flt_freq_flat_,flt_dc_sprd_,flt_dc_flat_,
											flt_bd_,flt_cal_,flt_eom_,											 	
											solve_flat,
							      			f_fwd_sprd_,first_fix_sprd,f_fwd_flat_,first_fix_flat,	
											df_disc_flat_,
											sprd_sprd_[i],
											0,
											sprd_flat_[i]);					
		
		QL_REQUIRE(!null(c),"invalid cashflow (currency basis)", v_i_sprd[i].name(), true);
		push_back(	cflow, c);	
	}
		
	return cashflows(cflow);
}
//-------------------------------------------------------
// adj_disc_df_flat
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_cross_bc_ext.adj_disc_df_flat(	interpolator 	disc_ip,
													rate_type		disc_ip_rt)
							   
{	
	QL_FAIL("work in progress");
}
//-------------------------------------------------------
// adj_disc_df_flat
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_cross_bc_ext.adj_disc_df_flat(	disc_z_model disc_model)
							   
{	
	QL_FAIL("work in progress");
}
//-------------------------------------------------------
// adj_disc_df_sprd
// fx-implied disc_func
// here for backward compat
//-------------------------------------------------------
disc_func swap_curve_cross_bc_ext.adj_disc_df_sprd(	interpolator 	disc_ip,
													rate_type		disc_ip_rt)
							   
{	
	QL_REQUIRE(!null(df_disc_flat_),"invalid df_disc_flat");

	cashflows cflow = implied_sprd_cashflows();
	QL_REQUIRE(!null(cflow ),"error creating cashflows (base)");
	
	if(null(df_disc_sprd_))
		return bootstrap(cflow, disc_ip, disc_ip_rt);		
	else
		return bootstrap_spr(df_disc_sprd_, cflow, disc_ip, disc_ip_rt);
}

//-------------------------------------------------------
// adj_disc_df_sprd
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_cross_bc_ext.adj_disc_df_sprd(	disc_z_model disc_model)
							   
{	
	QL_REQUIRE(!null(df_disc_flat_),"invalid df_disc_flat");

	cashflows cflow = implied_sprd_cashflows();
	QL_REQUIRE(!null(cflow ),"error creating cashflows (base)");
	
	if(null(df_disc_sprd_))
		return disc_model.df(cflow, null<vector(number)>);		
	else
		return disc_model.df(df_disc_sprd_, cflow, null<vector(number)>);
}

disc_func swap_curve_cross_bc_ext.adj_disc_df_base(	interpolator disc_ip,rate_type disc_ip_rt) { return adj_disc_df_sprd(disc_ip,disc_ip_rt); }

//-------------------------------------------------------
// flat_basis_swap_ip
//-------------------------------------------------------
vector(number) swap_curve_cross_bc_ext.flat_basis_swap_ip()
{
	vector(instrument) v_i_flat = bc_crv_flat_.basis_curve().instruments();
	vector(instrument) v_i_sprd = bc_crv_sprd_.basis_curve().instruments();
	vector(number) sprd_flat 	= v_i_flat.quote();
	vector(number) mat_flat 	= v_i_flat.maturity()-d_;
	vector(number) mat_sprd 	= v_i_sprd.maturity()-d_;
	
	interpolation ip = interpolation(ip_linear(),mat_flat,sprd_flat);	
	return ip.eval(mat_sprd);	
}
//-------------------------------------------------------
// solve_spread_flat_leg
//-------------------------------------------------------

vector(number) swap_curve_cross_bc_ext.solve_spread_flat_leg(vector(number) option(nullable) sprd)
{
	QL_FAIL("work in progress");	
}
//-------------------------------------------------------
// solve_spread_sprd_leg
//-------------------------------------------------------

vector(number) swap_curve_cross_bc_ext.solve_spread_sprd_leg(vector(number) option(nullable) flat_sprd)
{
	interpolator 	disc_ip 	= ip_linear();
	rate_type		disc_ip_rt 	= RT_LOG_DF;

	disc_func df_imp_sprd = adj_disc_df_sprd(disc_ip,disc_ip_rt);
	return solve_spread_sprd_leg(df_imp_sprd, flat_sprd);
}

vector(number) swap_curve_cross_bc_ext.solve_spread_base_leg(vector(number) option(nullable) flat_sprd) { return solve_spread_sprd_leg(flat_sprd);}

//-------------------------------------------------------
// solve_spread_sprd_leg
//-------------------------------------------------------
vector(number) swap_curve_cross_bc_ext.solve_spread_sprd_leg(	disc_func 						df_imp_sprd,
																vector(number) option(nullable) flat_sprd)
{
	
	vector(instrument) v_i_sprd = bc_crv_sprd_.basis_curve().instruments();

	if(null(flat_sprd)) {
		resize(flat_sprd,v_size(v_i_sprd));
		flat_sprd = 0;
	}
	else
		QL_REQUIRE(v_size(flat_sprd) == v_size(v_i_sprd),"invalid vector size");

	vector(number) sprd;

	integer freq_sprd 			= flt_freq_flat_;
	integer freq_solve 			= flt_freq_sprd_; 
	day_count_method dc_sprd 	= flt_dc_flat_;
	day_count_method dc_solve 	= flt_dc_sprd_;
	fwd_func f_fwd_sprd			= f_fwd_flat_;											
	fwd_func f_fwd_solve		= f_fwd_sprd_;
	disc_func df_fx_sprd		= df_disc_flat_;		
	disc_func df_fx_solve		= df_imp_sprd;

	for(integer i = 0; i < v_size(v_i_sprd) ; i++){
		QL_REQUIRE(!null(flat_sprd[i]),"invalid spread");						
		number s = ..flt_sprd_fx_lib(d_,v_i_sprd[i].issue_date(),v_i_sprd[i].maturity(),day(v_i_sprd[i].issue_date()),
									freq_sprd,freq_solve,dc_sprd,dc_solve,
									flt_bd_,flt_cal_,flt_eom_,						
									f_fwd_sprd,f_fwd_solve,
									df_fx_sprd,df_fx_solve,
									flat_sprd[i]);
		push_back(sprd,s);
	}
	return sprd;
}


