/*	
	Basic swap curve functions for cross currency bootstrap
	Developed by Algorithmica Research, Magnus Nyström
	
	
	
	Comments:

	
*/
option(null: hard);	

//-------------------------------------------------------
// class swap_curve_cff_ext
// class used for currency fix-flt swaps 
//-------------------------------------------------------
class swap_curve_cff_ext option (category: 'Yield Curve/003 Curve Building Cross Currency')
{
   	public:
		//----legacy------------- 
		swap_curve_cff_ext(	date,swap_curve_cff_parm,disc_func option(nullable),disc_func option(nullable),fwd_func,string option(nullable) name = null<string>) ;
		swap_curve_cff_ext(	date,swap_curve_cff_parm,disc_func option(nullable),disc_func option(nullable),disc_func,string option(nullable) name = null<string>) ;
		//-----------------------

		swap_curve_cff_ext(	date,swap_curve_cff_f_parm,disc_func option(nullable),disc_func option(nullable),fwd_func,fwd_func_tenor,string option(nullable) name = null<string>) ;
		swap_curve_cff_ext(	swap_curve_cff_ext) ;

		disc_func			fix_adj_disc_df(interpolator,rate_type);
		disc_func			fix_adj_disc_df(disc_z_model );
		disc_func			fix_adj_disc_df(disc_z_model,disc_func option(nullable) );
		disc_func			flt_adj_disc_df(interpolator,rate_type);
		disc_func			flt_adj_disc_df(disc_z_model);
		disc_func			flt_adj_disc_df(disc_z_model,disc_func option(nullable));
		string 				name();
		curve 				fixflt_curve()	;
		integer 			fix_freq()	;
		day_count_method  	fix_dc()	;   
		string 				fix_ccy() 	; 
		integer 			flt_freq()	;
		day_count_method  	flt_dc()	;   
		string 				flt_ccy() 	;        						
		           
		disc_func			flt_disc_df();
		disc_func			fix_disc_df();
		fwd_func 			flt_fwd() ;
		                                                                              
		bd_convention 		flt_bd() ;
		calendar 			flt_cal() ;
		logical 			flt_eom() ;
		string 				settle_code() ;
		curve 				blended_curve();
		number				first_flt();

	protected:		
		
		void				init(fwd_func_tenor,disc_func option(nullable));
		void				init_common(curve, fx_spot option(nullable), disc_func option(nullable));
		void				init_parm(instrument);
		void				init_ir_index(curve);
		
		void				check_cff_curve(curve);
		void				init_fx_swap(fx_spot option(nullable));
		logical 			create_blend_curve(	out curve option(nullable), out curve option(nullable), out curve option(nullable),
												out curve option(nullable),out vector(date)   ,out string);
		void				blend_curve();
		cashflows 			cf(curve, disc_func,logical);	

		disc_func			fix_adj_disc_df_int(disc_z_model,disc_func option(nullable),logical );
		disc_func			flt_adj_disc_df_int(disc_z_model,disc_func option(nullable),logical );

		//legacy
		void				init_legacy(disc_func option(nullable));
		void				init_ir_index_legacy(instrument);
		void				check_tags_legacy(curve );
		
	private:
		date				d_;
		swap_curve_ext_parm param_;
		fx_curve			fx_swaps_;
				
		fwd_func			flt_fwd_;
		ir_index			fwd_idx_;
							
		integer 			fix_freq_;
		day_count_method 	fix_dc_;			
		string				fix_ccy_;	

		disc_func			flt_df_disc_;		
		disc_func			fix_df_disc_;
								
		string				name_;
		vector(number) 		spread_;
		curve				curve_fix_;
		curve				curve_flt_;
		logical				fx_flt_eq_base_;
		logical 			curve_ok_;			
};

string 				swap_curve_cff_ext.name()			{ return name_;}        
fwd_func 			swap_curve_cff_ext.flt_fwd() 		{ return flt_fwd_;}
integer 			swap_curve_cff_ext.fix_freq() 		{ return fix_freq_;}
day_count_method swap_curve_cff_ext.fix_dc() 			{ return fix_dc_;}
string 				swap_curve_cff_ext.fix_ccy() 		{ return fix_ccy_;}
disc_func 			swap_curve_cff_ext.flt_disc_df()	{ return flt_df_disc_;}
disc_func 			swap_curve_cff_ext.fix_disc_df()	{ return fix_df_disc_;}

integer				swap_curve_cff_ext.flt_freq()		{ return integer(floor(12/fwd_idx_.tenor_months()));}
day_count_method	swap_curve_cff_ext.flt_dc() 		{ return fwd_idx_.day_count_method();}
bd_convention		swap_curve_cff_ext.flt_bd() 		{ return fwd_idx_.bd_convention();}
logical				swap_curve_cff_ext.flt_eom() 		{ return fwd_idx_.eom();}
calendar			swap_curve_cff_ext.flt_cal() 		{ return fwd_idx_.fixing_calendar();}
string 				swap_curve_cff_ext.flt_ccy() 		{ return fwd_idx_.currency();}	
string 				swap_curve_cff_ext.settle_code()	{ return strcat("BD",str(fwd_idx_.value_days()));}

//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
swap_curve_cff_ext.swap_curve_cff_ext(swap_curve_cff_ext s) 
									:d_(s.d_), param_(s.param_), fx_swaps_(s.fx_swaps_), 
									flt_fwd_(s.flt_fwd_),
									fix_freq_(s.fix_freq_), fix_dc_(s.fix_dc_), fix_ccy_(s.fix_ccy_), 
									flt_df_disc_(s.flt_df_disc_) , fix_df_disc_(s.fix_df_disc_) , 
									name_(s.name_), spread_(s.spread_), curve_fix_(s.curve_fix_), curve_flt_(s.curve_flt_), 
									fx_flt_eq_base_(s.fx_flt_eq_base_), curve_ok_(s.curve_ok_)
{}
//-------------------------------------------------------
// constructor
// legacy version
//-------------------------------------------------------
swap_curve_cff_ext.swap_curve_cff_ext(	date						d,									
										swap_curve_cff_parm			p,									
										disc_func option(nullable)	fix_df_disc,
										disc_func option(nullable)	flt_df_disc,
										fwd_func					flt_fwd,																																		
										string 	option(nullable) 	name) 
										:d_(d), param_(p), flt_fwd_(flt_fwd),  
										flt_df_disc_(flt_df_disc), fix_df_disc_(fix_df_disc), name_(name)
{	
	init_legacy(null<disc_func>);	
}
//-------------------------------------------------------
// constructor
// legacy version
//-------------------------------------------------------
swap_curve_cff_ext.swap_curve_cff_ext(	date						d,									
										swap_curve_cff_parm			p,									
										disc_func option(nullable)	fix_df_disc,
										disc_func option(nullable)	flt_df_disc,
										disc_func					flt_fwd,																																			
										string option(nullable)		name) 
										:d_(d), param_(p), flt_df_disc_(flt_df_disc), fix_df_disc_(fix_df_disc), name_(name)
{			
	init_legacy(flt_fwd);	
}
//-------------------------------------------------------
// constructor
//-------------------------------------------------------
swap_curve_cff_ext.swap_curve_cff_ext(	date						d,									
										swap_curve_cff_f_parm		p,									
										disc_func option(nullable)	fix_df_disc,
										disc_func option(nullable)	flt_df_disc,
										fwd_func					flt_fwd,
										fwd_func_tenor	fwd_tenor,
										string 	option(nullable) 	name) 
										:d_(d), param_(p), flt_fwd_(flt_fwd),  
										flt_df_disc_(flt_df_disc), fix_df_disc_(fix_df_disc), name_(null(name) ? "" : name)
{	
	init(fwd_tenor,null<disc_func>);	
}

//-------------------------------------------------------
// init
//-------------------------------------------------------
void swap_curve_cff_ext.init(fwd_func_tenor fwd_tenor,
								 disc_func option(nullable) flt_fwd)
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_CURR_FF_F,'invalid type of param object (should be a swap_curve_cff_f_parm ) ');		
	swap_curve_cff_f_parm param = dynamic_cast<swap_curve_cff_f_parm>(param_);

	if(!null(param.fixing_crv())){
		string fwd_tenor_s = string(fwd_tenor);
		fixing_curve new_fixing_crv;
		logical upd = CORE_INT.fwd_init(d_, param.fixing_crv(), fwd_tenor, fwd_tenor_s,fwd_idx_,new_fixing_crv);
		if(upd) {
			param.set_fixing_crv(new_fixing_crv); 
		}
	}
	
	QL_REQUIRE(check_curve(param.fixflt_crv()),'invalid fixflt swap curve');
	init_ir_index(param.fixflt_crv());
	//integer freq = fwd_func_tenor.to_freq(fwd_tenor);
	check_cff_curve(param.fixflt_crv());	
	init_parm(param.fixflt_crv().instruments()[0]);
	init_common(param.fixflt_crv(), param.fx_spot(), flt_fwd);
}
//-------------------------------------------------------
// init_ir_index
//-------------------------------------------------------
void swap_curve_cff_ext.init_ir_index(curve fixflt_crv)
{
	if(null(fwd_idx_)) {
		instrument i 	= fixflt_crv.instruments()[0];
		swap_fixflt sw 	= swap_fixflt(i);
		fwd_idx_ 		= sw.ir_index_flt_leg();
	}
}
//-------------------------------------------------------
// check_cff_curve()
//-------------------------------------------------------
void swap_curve_cff_ext.check_cff_curve(curve fixflt_crv)
{	
	
	
	string err_name;
	QL_REQUIRE(CORE_INT.check_instr_type(fixflt_crv, instr_type.SWAP_FIXFLT, err_name),'invalid instrument type in swap curve', err_name, true);
	
	vector(instrument) vi = fixflt_crv.instruments();
	date settle  = date_code(settle_code()).apply_fwd(d_, flt_cal());	
	QL_REQUIRE(CORE_INT.check_swap_issue(vi, settle, err_name),'invalid issue date in swap curve (issue_date < spot settle)', err_name, true);

	day_count_method dc = null<day_count_method>;
	QL_REQUIRE(CORE_INT.swap_dc_check(vi,dc,true,err_name),'invalid fixed leg daycount_method in swap curve (mixed)', err_name, true);
	dc = flt_dc();
	QL_REQUIRE(CORE_INT.swap_dc_check(vi, dc, false,err_name),'invalid float leg daycount_method in swap curve (mixed)',err_name,true);
	
	integer f = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,f,true,err_name),'invalid fixed leg coupon frequency in swap curve (mixed)', err_name, true);

	f = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,f,false,err_name),'invalid float leg coupon frequency in swap curve (mixed)', err_name, true);
	f = flt_freq();
	QL_REQUIRE(CORE_INT.swap_rfreq_check(vi,f,false,err_name),'invalid float leg reset frequency in swap curve (mixed)', err_name, true);
	
	string ccy = null<string>;
	QL_REQUIRE(CORE_INT.swap_currency_check(vi, ccy,true,err_name),'invalid fixed leg currency in swap curve (mixed)', err_name, true);
	ccy = null<string>;
	QL_REQUIRE(CORE_INT.swap_currency_check(vi, ccy,false,err_name),'invalid float leg currency in swap curve (mixed)', err_name, true);

	tenor_code t = tenor_code(flt_freq());
	QL_REQUIRE(!null(t),'invalid float leg tenor');
	QL_REQUIRE(CORE_INT.check_eq_index_tenor_code(vi, t, err_name),'invalid tenor in swap curve (mixed)', err_name, true);
}

//-------------------------------------------------------
// init_parm
//-------------------------------------------------------
void swap_curve_cff_ext.init_parm(instrument i)
{	
	fix_dc_  = i.accr_dc_method(); 
	QL_REQUIRE(!null(fix_dc_),'invalid fix_dc in currency fixflt swap',i.name(),true);	

	swap_fixflt sw = swap_fixflt(i);
	string flt_leg_ccy;
	sw.currency(fix_ccy_,flt_leg_ccy);
	
	QL_REQUIRE(!null(fix_ccy_),'invalid fix_ccy in currency fixflt swap',i.name(),true);	
	QL_REQUIRE(!equal_casei(flt_ccy(),fix_ccy_),'invalid fixed leg currency in swap curve (fix_leg currency same as index currency)'); 		
	QL_REQUIRE(equal_casei(flt_ccy(),flt_leg_ccy),'invalid float leg currency in swap curve (differs from fixing curve currency)');
	
	fix_freq_ = i.coupon_freq();
	QL_REQUIRE(fix_freq_ > 0 && fix_freq_ != 365,'invalid fix_freq in currency fixflt swap',i.name(),true);			
}
//-------------------------------------------------------
// init_common
//-------------------------------------------------------
void swap_curve_cff_ext.init_common(curve 						fixflt_crv,
									fx_spot option(nullable) 	fxs,
									disc_func option(nullable) 	flt_fwd)
{	
	spread_ = one_vector(v_size(fixflt_crv.instruments()), 0);
	QL_REQUIRE(!null(fix_df_disc_) || !null(flt_df_disc_),'invalid disc_func (either fix leg or float leg disc_func must be input)');
	if(!null(flt_fwd))
		flt_fwd_ = fwd_func(flt_fwd, flt_freq(), d_, flt_dc(), flt_cal(), flt_bd(), flt_eom()); 
	init_fx_swap(fxs);	
	blend_curve();
}

//-------------------------------------------------------
// init_fx_swap
//-------------------------------------------------------
void swap_curve_cff_ext.init_fx_swap(fx_spot option(nullable) fxs)
{
	if(..check_curve(param_.short_crv())) {
		fx_swaps_ = new fx_curve(param_.short_crv().instruments(),fxs,fxs,quote_side_code.QS_MID) ;
		QL_REQUIRE(!null(fx_swaps_),'invalid fx curve');
		if(!equal_casei(fix_ccy_,fxs.base_ccy()) && !equal_casei(fix_ccy_,fxs.price_ccy()))			
			QL_FAIL('invalid fx swap (incompatible with currency swap)'); 
		if(!equal_casei(flt_ccy(),fxs.base_ccy()) && !equal_casei(flt_ccy(),fxs.price_ccy()))			
			QL_FAIL('invalid fx swap (incompatible with currency swap)'); 
		fx_flt_eq_base_ = equal_casei(flt_ccy(),fxs.base_ccy());
	}
}

//-------------------------------------------------------
// blend_curve
//-------------------------------------------------------
void swap_curve_cff_ext.blend_curve()
{
	/*if(null(fx_swaps_)){
		curve_fix_ = curve_flt_ = param_.fixflt_crv_;		
		string n;
		if(!..check_quote(param_.fixflt_crv_,n)) {
			string s = strcat(['invalid quote(s) in fixflt curve [',n,"]"]);
			QL_FAIL(s);
		}		
		curve_ok_ 	= true;
		return;
	}
	interpolator ip	= ip_spline(null<number>, 0,IP_NO_XPOL_R) ;
	rate_type ip_rt = RT_CONT;

	vector(date) knots;			
	..curve_prio	prio1 = param_.prio1_ == ..CP_SHORT ? ..CP_SHORT : ..CP_LONG; 
	..curve_prio	prio2 = param_.prio1_ == ..CP_SHORT ? ..CP_LONG : ..CP_SHORT; 

	vector(date) excl_fx_mat;

	if(!null(flt_df_disc_)){	//USD	
				
		curve synt_depo_fix;//KRW
		if(fx_flt_eq_base_) {
			//QL_REQUIRE(!null(param_.depo_class_fxprice_),'invalid instrument class (deposit for price currency)');
			day_count_method dc = null(param_.depo_class_fxprice_) ? DC_ACT_360 : instr_def(param_.depo_class_fxprice_).accr_dc_method(); 
			//synt_depo_fix = fx_swaps_.synt_depo_price(param_.depo_class_fxprice_,flt_df_disc_,ip,ip_rt,dc, excl_fx_mat);//KRW depo	
			synt_depo_fix = fx_swaps_.synt_depo_price(flt_df_disc_,ip,ip_rt,dc, excl_fx_mat);//KRW depo		
		}
		else {
			//QL_REQUIRE(!null(param_.depo_class_fxbase_),'invalid instrument class (deposit for base currency)');
			day_count_method dc = null(param_.depo_class_fxbase_) ? DC_ACT_360 : instr_def(param_.depo_class_fxbase_).accr_dc_method(); 
			//synt_depo_fix = fx_swaps_.synt_depo_base(param_.depo_class_fxbase_,flt_df_disc_,ip,ip_rt,dc,excl_fx_mat);//KRW depo
			synt_depo_fix = fx_swaps_.synt_depo_base(flt_df_disc_,ip,ip_rt,dc,excl_fx_mat);//KRW depo
		}
		QL_REQUIRE(!null(synt_depo_fix),'invalid creation of deposit curve (fix)');

		curve_fix_ = ..blend_curve(synt_depo_fix, null<curve>, param_.fixflt_crv_, prio1, prio2, false, false, param_.merge_crv_, knots, param_.blend_buf_days_);		
		QL_REQUIRE(!null(curve_fix_),'invalid blended curve (fix)');

		string n;
		if(!..check_quote(curve_fix_,n)) {
			string s = strcat(['invalid quote(s) in blended curve (fix) [',n,"]"]);
			QL_FAIL(s);
		}	
		curve_fix_ = curve_fix_.set_name("blended curve (fix)");
		QL_REQUIRE(..check_curve(curve_fix_),'invalid blended curve (fix)');					
		
	}
	if(!null(fix_df_disc_)){ //KRW
		curve synt_depo_flt;
		if(fx_flt_eq_base_) {
			//QL_REQUIRE(!null(param_.depo_class_fxbase_),'invalid instrument class (deposit for base currency)');
			day_count_method dc = null(param_.depo_class_fxbase_) ? DC_ACT_360 : instr_def(param_.depo_class_fxbase_).accr_dc_method(); 
			//synt_depo_flt = fx_swaps_.synt_depo_base(param_.depo_class_fxbase_,fix_df_disc_,ip,ip_rt,dc,excl_fx_mat);//USD depo
			synt_depo_flt = fx_swaps_.synt_depo_base(fix_df_disc_,ip,ip_rt,dc,excl_fx_mat);//USD depo
		}
		else {
			//QL_REQUIRE(!null(param_.depo_class_fxprice_),'invalid instrument class (deposit for price currency)');
			day_count_method dc = null(param_.depo_class_fxprice_) ? DC_ACT_360 : instr_def(param_.depo_class_fxprice_).accr_dc_method(); 
			//synt_depo_flt = fx_swaps_.synt_depo_price(param_.depo_class_fxprice_,fix_df_disc_,ip,ip_rt,dc,excl_fx_mat);//USD depo
			synt_depo_flt = fx_swaps_.synt_depo_price(fix_df_disc_,ip,ip_rt,dc,excl_fx_mat);//USD depo
		}
		QL_REQUIRE(!null(synt_depo_flt),'invalid creation of deposit curve (flt)');

		curve_flt_ = ..blend_curve(synt_depo_flt,null<curve>, param_.fixflt_crv_,prio1,prio2,false,false,param_.merge_crv_,knots,param_.blend_buf_days_);
		QL_REQUIRE(!null(curve_flt_),'invalid blended curve (flt)');

		string n;
		if(!..check_quote(curve_flt_,n)) {
			string s = strcat(['invalid quote(s) in blended curve (flt) [',n,"]"]);
			QL_FAIL(s);
		}

		curve_flt_ = curve_flt_.set_name("blended curve (flt)");
		QL_REQUIRE(..check_curve(curve_flt_),'invalid blended curve (flt)');		
	}
				
	curve_ok_ 	= true;	
*/
	curve synt_depo_fix,synt_depo_flt;
	string err_msg;
	vector(date) excl_fx_mat;
	curve_ok_ = create_blend_curve(curve_fix_, curve_flt_, synt_depo_fix, synt_depo_flt,excl_fx_mat,err_msg);	
	QL_REQUIRE(curve_ok_,err_msg);

	string n;
	if(!null(curve_fix_) && !check_quote(curve_fix_,n)) {
		string s = strcat(['invalid quote(s) in blended curve (fix leg) [',n,']']);
		QL_FAIL(s);
	}	
	if(!null(curve_flt_) &&!check_quote(curve_flt_,n)) {
		string s = strcat(['invalid quote(s) in blended curve (float leg) [',n,']']);
		QL_FAIL(s);
	}	
}

//-------------------------------------------------------
// create_blend_curve
//-------------------------------------------------------
logical swap_curve_cff_ext.create_blend_curve(	out curve option(nullable) curve_fix, 
												out curve option(nullable) curve_flt, 
												out curve option(nullable) synt_depo_fix, 
												out curve option(nullable) synt_depo_flt,
												out vector(date)   excl_fx_mat, 
												out string err_msg)
{
	if(null(fx_swaps_)){
		curve_fix = curve_flt = fixflt_curve();	
		synt_depo_fix = synt_depo_flt = null<curve>;
		err_msg = "";
		return true;
	}
	interpolator ip	= ip_spline(null<number>, 0,IP_NO_XPOL_R) ;
	rate_type ip_rt = RT_CONT;

	vector(date) knots;			
	curve_prio	prio1 = param_.prio1() == curve_prio.SHORT ? curve_prio.SHORT : curve_prio.LONG; 
	curve_prio	prio2 = param_.prio1() == curve_prio.SHORT ? curve_prio.LONG : curve_prio.SHORT; 

	if(!null(flt_df_disc_)){	//USD	
						
		if(fx_flt_eq_base_) {
			synt_depo_fix = fx_swaps_.synt_depo_price(flt_df_disc_,ip,ip_rt,fix_dc_, excl_fx_mat);//KRW depo			
		}
		else {
			synt_depo_fix = fx_swaps_.synt_depo_base(flt_df_disc_,ip,ip_rt,fix_dc_,excl_fx_mat);//KRW depo
		}

		if(null(synt_depo_fix)) {
			err_msg = 'invalid creation of deposit curve (fix leg)';
			return false;
		}	
	
		curve_fix = CORE_INT.blend_curve(synt_depo_fix, null<curve>, fixflt_curve(), prio1, prio2, false, false,
										 param_.merge_crv(), knots, param_.blend_buf_days());
				
		if(!..check_curve(curve_fix)) {
			err_msg = 'invalid blended curve (fix leg)';
			return false;
		}				
		curve_fix = curve_fix.set_name("blended curve (fix leg)");	
		
	}
	if(!null(fix_df_disc_)){ //KRW
		
		if(fx_flt_eq_base_) {
			synt_depo_flt = fx_swaps_.synt_depo_base(fix_df_disc_,ip,ip_rt,flt_dc(),excl_fx_mat);//USD depo
		}
		else {
			synt_depo_flt = fx_swaps_.synt_depo_price(fix_df_disc_,ip,ip_rt,flt_dc(),excl_fx_mat);//USD depo
		}

		if(null(synt_depo_flt)) {
			err_msg = 'invalid creation of deposit curve (float leg)';
			return false;
		}
		
		curve_flt = CORE_INT.blend_curve(synt_depo_flt,null<curve>, fixflt_curve(),prio1,prio2,false,false,
										 param_.merge_crv(),knots,param_.blend_buf_days());

		if(!..check_curve(curve_flt )) {
			err_msg = 'invalid blended curve (float leg)';
			return false;
		}			

		curve_flt = curve_flt.set_name("blended curve (float leg)");		
	}

	err_msg = "";
	return true;	
}

//-------------------------------------------------------
// blended_curve
//-------------------------------------------------------
curve swap_curve_cff_ext.blended_curve()
{ 
	/*if(!curve_ok_)
		return null<curve>;

	if(null(fx_swaps_))
		return param_.fixflt_crv_;		

	vector(date) knots;			
	..curve_prio	prio1 = param_.prio1_ == ..CP_SHORT ? ..CP_SHORT : ..CP_LONG; 
	..curve_prio	prio2 = param_.prio1_ == ..CP_SHORT ? ..CP_LONG : ..CP_SHORT; 

	return ..blend_curve(param_.short_crv_, null<curve>, param_.fixflt_crv_, prio1, prio2, false, false, param_.merge_crv_, knots, param_.blend_buf_days_);
*/
	if(!curve_ok_)
		return null<curve>;

	if(null(fx_swaps_))
		return fixflt_curve();		

	vector(date) knots;			
	curve_prio	prio1 	= param_.prio1() == curve_prio.SHORT ? curve_prio.SHORT : curve_prio.LONG; 
	curve_prio	prio2 	= param_.prio1() == curve_prio.SHORT ? curve_prio.LONG : curve_prio.SHORT; 
	curve curve_all 		= CORE_INT.blend_curve(param_.short_crv(), null<curve>, fixflt_curve(), prio1, prio2, false, false, 
												param_.merge_crv(), knots, param_.blend_buf_days());
		
	curve curve_fix, curve_flt, synt_depo_fix,synt_depo_flt;
	vector(date) excl_fx_mat;
	string err_msg;
	logical curve_ok = create_blend_curve(curve_fix, curve_flt, synt_depo_fix, synt_depo_flt, excl_fx_mat, err_msg);	
	if(!curve_ok)
		return null<curve>;
	
	if(v_size(excl_fx_mat) == 0)
		return curve_all;
	else {//here some fxswaps have been filtered out
		vector(instrument) vi_all 	= curve_all.instruments();
		vector(date) vi_all_m 		= vi_all.maturity();
		vector(integer) index_vector = v_find(vi_all_m, excl_fx_mat) ;		
		vector(instrument) vi 		= v_exclude(vi_all, index_vector) ;
		return curve(vi);
	} 		
}
//-------------------------------------------------------
// cf
//-------------------------------------------------------
cashflows swap_curve_cff_ext.cf(curve		crv,
								disc_func 	df_disc,
								logical 	solve_flt)
{		
	vector(instrument) src	= crv.instruments();
	integer size 			= v_size(src);
	vector(cashflows) dest[0];	
	
	integer b = 0;
	number first_fix = /*param_.type_ == SCE_CURR_FF_F ? null<number> :*/ first_flt();
				
	for(integer i = 0; i < size ; i++){
		cashflows c;					
		if(src[i].is_swap_fixflt()) {
			c = CORE_INT.cf_currffswap(src[i],solve_flt,flt_fwd_,first_fix, df_disc, spread_[b]);
			b++;			
		}
		else if(src[i].is_deposit()) 
			c = CORE_INT.cf_depo(src[i], 0);				
		else
			QL_FAIL("invalid instrument type (currency fixflt)");

		QL_REQUIRE(!null(c),"invalid cashflow (currency fixflt)", src[i].name(), true);
		push_back(	dest, c);	
	}
	return cashflows(dest);
}
//-------------------------------------------------------
// fix_adj_disc_df
// fx-implied disc_func
// here for backward compat
//-------------------------------------------------------
disc_func swap_curve_cff_ext.fix_adj_disc_df(	interpolator 	disc_ip,
												rate_type		disc_ip_rt)
							   
{	
	QL_REQUIRE(!null(flt_df_disc_),"invalid flt_df_disc");

	cashflows cflow = cf(curve_fix_, flt_df_disc_, false);
	QL_REQUIRE(!null(cflow ),"error creating cashflows (fix)");
	if(null(fix_df_disc_))
		return bootstrap(cflow, disc_ip, disc_ip_rt);		
	else
		return bootstrap_spr(fix_df_disc_, cflow, disc_ip, disc_ip_rt);
}
//-------------------------------------------------------
// fix_adj_disc_df_int
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_cff_ext.fix_adj_disc_df_int(disc_z_model disc_model,
												disc_func option(nullable) fix_df_disc_base,
												logical use_base_df)
							   
{	
	QL_REQUIRE(!null(flt_df_disc_),"invalid flt_df_disc");

	cashflows cflow = cf(curve_fix_, flt_df_disc_, false);
	QL_REQUIRE(!null(cflow ),"error creating cashflows (fix)");
	if(!use_base_df)
		return disc_model.df(cflow, curve_fix_.instruments().pvbp());		
	else
		return disc_model.df(fix_df_disc_base, cflow, curve_fix_.instruments().pvbp());
}	
//-------------------------------------------------------
// fix_adj_disc_df
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_cff_ext.fix_adj_disc_df(disc_z_model disc_model)							   
{	
	return fix_adj_disc_df_int(disc_model, fix_df_disc_, !null(fix_df_disc_));
}
//-------------------------------------------------------
// fix_adj_disc_df
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_cff_ext.fix_adj_disc_df(	disc_z_model disc_model, 
												disc_func option(nullable) fix_df_disc_base)
							   
{	
	return fix_adj_disc_df_int(disc_model, fix_df_disc_base, !null(fix_df_disc_base));
}	
//-------------------------------------------------------
// flt_adj_disc_df
// fx-implied disc_func
// here for backward compat
//-------------------------------------------------------
disc_func swap_curve_cff_ext.flt_adj_disc_df(	interpolator 	disc_ip,
												rate_type		disc_ip_rt)
							   
{	
	QL_REQUIRE(!null(fix_df_disc_),"invalid fix_df_disc");

	cashflows cflow = cf(curve_flt_, fix_df_disc_, true);
	QL_REQUIRE(!null(cflow ),"error creating cashflows (flt)");
	if(null(flt_df_disc_))
		return bootstrap(cflow, disc_ip, disc_ip_rt);		
	else
		return bootstrap_spr(flt_df_disc_, cflow, disc_ip, disc_ip_rt);
}
//-------------------------------------------------------
// flt_adj_disc_df
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_cff_ext.flt_adj_disc_df_int(disc_z_model disc_model,
												disc_func option(nullable) flt_df_disc_base,
												logical use_base_df)
							   
{	
	QL_REQUIRE(!null(fix_df_disc_),"invalid fix_df_disc");

	cashflows cflow = cf(curve_flt_, fix_df_disc_, true);
	QL_REQUIRE(!null(cflow ),"error creating cashflows (flt)");
	if(!use_base_df)
		return disc_model.df(cflow, curve_flt_.instruments().pvbp());		
	else
		return disc_model.df(flt_df_disc_base, cflow, curve_flt_.instruments().pvbp());
}

//-------------------------------------------------------
// flt_adj_disc_df
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_cff_ext.flt_adj_disc_df(disc_z_model disc_model)
							   
{	
	return flt_adj_disc_df_int(disc_model, flt_df_disc_, !null(flt_df_disc_));
}
//-------------------------------------------------------
// flt_adj_disc_df
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_cff_ext.flt_adj_disc_df(	disc_z_model disc_model, 
												disc_func option(nullable) flt_df_disc_base)
							   
{	
	return flt_adj_disc_df_int(disc_model, flt_df_disc_base, !null(flt_df_disc_base));
}
//-------------------------------------------------------
// fixflt_curve
//-------------------------------------------------------
curve swap_curve_cff_ext.fixflt_curve()
{ 	
	if(param_.type() == swap_curve_type.SCE_CURR_FF_F) {
		swap_curve_cff_f_parm param = dynamic_cast<swap_curve_cff_f_parm>(param_);
		return param.fixflt_crv();
	}
	else if(param_.type() == swap_curve_type.SCE_CURR_FF) {
		swap_curve_cff_parm param = dynamic_cast<swap_curve_cff_parm>(param_);
		return param.fixflt_crv_;
	}
	else
		QL_FAIL("invalid function call");
}
//-------------------------------------------------------
// first_flt 
//-------------------------------------------------------
number swap_curve_cff_ext.first_flt()
{	
	if(param_.type() == swap_curve_type.SCE_CURR_FF_F){
		swap_curve_cff_f_parm param = dynamic_cast<swap_curve_cff_f_parm>(param_);
		string 	err_msg;
		return null(param.fixing_crv()) ? null<number>: param.fixing_crv().fixing_rate(d_, fwd_idx_.tenor(), err_msg);
	}
	else if(param_.type() == swap_curve_type.SCE_CURR_FF){
		swap_curve_cff_parm param = dynamic_cast<swap_curve_cff_parm>(param_);
		return param.flt_first_fix_;
	}
	else
		QL_FAIL("invalid function call");
}

//-------------------------------------------------------
// init   [LEGACY]
//-------------------------------------------------------
void swap_curve_cff_ext.init_legacy(disc_func option(nullable) flt_fwd)
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_CURR_FF,'invalid type of param object (should be a swap_curve_cff_parm ) ');		
	swap_curve_cff_parm param = dynamic_cast<swap_curve_cff_parm>(param_);

	if(null(name_)) name_ = "";
	QL_REQUIRE(check_curve(param.fixflt_crv_),'invalid currency fixflt swap curve');	
	check_cff_curve(param.fixflt_crv_);	

	init_ir_index_legacy(param.fixflt_crv_.instruments()[0]);
	check_tags_legacy(param.fixflt_crv_);
	init_common(param.fixflt_crv_,param.fx_spot_,flt_fwd);
}
//-------------------------------------------------------
// init_ir_index   [LEGACY]
//-------------------------------------------------------
void swap_curve_cff_ext.init_ir_index_legacy(instrument i)
{	
	integer freq = i.swap_coupon_freq_leg2();
	QL_REQUIRE(freq > 0 && freq != 365,'invalid float coupon freq in currency fixflt swap',i.name(),true);
		
	calendar cal = i.calendar();
	string cal_name = i.calendar_name();
	QL_REQUIRE(!null(cal),'invalid calendar in currency fixflt swap',i.name(),true);	 
	
	day_count_method  dc = i.swap_dc_method_leg2() ;
	QL_REQUIRE(!null(dc),'invalid float daycount method in currency fixflt swap',i.name(),true);

	bd_convention bd	= i.pmt_bus_day_d(BD_MOD_FOLLOWING);	
	QL_REQUIRE(!null(bd),'invalid pmt_bus_day in currency fixflt swap',i.name(),true);	 

	logical eom	= i.eom_d(true);

	string ccy = i.swap_currency_leg2();
	QL_REQUIRE(!null(ccy),'invalid float currency in currency fixflt swap',i.name(),true);	
	
	integer rdays = integer(i.flt_reset_days_d(2));
	QL_REQUIRE(rdays >= 0,'invalid flt_reset_days in fixflt curve ',i.name(),true);	

	//string sd 			= strcat("BD",str(rdays));	
	integer spot_days 	= rdays;
	integer value_days 	= spot_days;
	string tenor 		= strcat(str(12/freq),"M");

	fwd_idx_ 			= ir_index_ibor(	strcat("Ibor-",ccy),tenor,value_days,rdays, ccy,
										cal,cal_name, dc,bd,eom,spot_days);
	QL_REQUIRE(!null(fwd_idx_),'failed to create ir_index');

	fix_dc_ 			= i.swap_dc_method_leg1(); 
	QL_REQUIRE(!null(fix_dc_),'invalid fix_dc in currency fixflt swap',i.name(),true);	
	
	fix_ccy_ 			= i.swap_currency_leg1();
	QL_REQUIRE(!null(fix_ccy_),'invalid fix_ccy in currency fixflt swap',i.name(),true);	
	QL_REQUIRE(!equal_casei(ccy,fix_ccy_),'invalid currency fixflt swap (same currency in both legs)'); 		
	
	fix_freq_			= i.swap_coupon_freq_leg1();
	QL_REQUIRE(fix_freq_ > 0 && fix_freq_ != 365,'invalid fix_freq in currency fixflt swap',i.name(),true);		
}

//-------------------------------------------------------
// check_tags_legacy
//-------------------------------------------------------
void swap_curve_cff_ext.check_tags_legacy(curve fixflt_crv)
{				 								
	vector(instrument) v_i 	= fixflt_crv.instruments();
	 	
	string err_name;
	tenor_code t = tenor_code(flt_freq());
	QL_REQUIRE(!null(t),'invalid float leg tenor');
	QL_REQUIRE(CORE_INT.check_eq_index_tenor_code(v_i, t, err_name),'invalid float leg tenor in currency fixflt swap', err_name, true);

	integer f = flt_freq();
	QL_REQUIRE(CORE_INT.swap_cfreq_check(v_i, f, false,err_name),'invalid float leg coupon freq in currency fixflt swap', err_name, true);
	QL_REQUIRE(CORE_INT.swap_cfreq_check(v_i, fix_freq_, true,err_name),'invalid fix leg coupon freq in currency fixflt swap', err_name, true);			

	day_count_method dc = flt_dc();
	QL_REQUIRE(CORE_INT.swap_dc_check(v_i,dc,false,err_name),'invalid float leg daycount method in currency fixflt swap', err_name, true);
	QL_REQUIRE(CORE_INT.swap_dc_check(v_i,fix_dc_,true,err_name),'invalid fix leg daycount method in currency fixflt swap', err_name, true);	

	string curr = flt_ccy();
	QL_REQUIRE(CORE_INT.swap_currency_check(v_i, curr, false,err_name),'invalid float leg currency in currency fixflt swap', err_name, true);
	QL_REQUIRE(CORE_INT.swap_currency_check(v_i, fix_ccy_, true,err_name),'invalid fix leg currency in currency fixflt swap', err_name, true);	
		
}

//-------------------------------------------------------
// create
//-------------------------------------------------------
swap_curve_cff_ext swap_curve_cff_ext(date						d,									
									swap_curve_cff_parm			p,									
									disc_func option(nullable)	fix_df_disc,
									disc_func option(nullable)	flt_df_disc,
									fwd_func					flt_fwd,																																		
									string option(nullable)		name)
{
	return new swap_curve_cff_ext(d, p,fix_df_disc,flt_df_disc, flt_fwd,name)	;	
}

//-------------------------------------------------------
// create
//-------------------------------------------------------
swap_curve_cff_ext swap_curve_cff_ext(date						d,									
									swap_curve_cff_parm			p,									
									disc_func option(nullable)	fix_df_disc,
									disc_func option(nullable)	flt_df_disc,																									
									disc_func 					flt_fwd,									
									string option(nullable)		name)
{
	return new swap_curve_cff_ext(d, p,fix_df_disc,flt_df_disc, flt_fwd, name)	;	
}
//-------------------------------------------------------
// create
//-------------------------------------------------------
swap_curve_cff_ext swap_curve_cff_ext(date						d,									
									swap_curve_cff_f_parm		p,									
									disc_func option(nullable)	fix_df_disc,
									disc_func option(nullable)	flt_df_disc,
									fwd_func					flt_fwd,
									fwd_func_tenor				fwd_tenor,
									string option(nullable)		name)
{
	return new swap_curve_cff_ext(d, p,fix_df_disc,flt_df_disc, flt_fwd,fwd_tenor,name)	;	
}

