/*	
	Basic swap curve functions for cross currency bootstrap
	Developed by Algorithmica Research, Magnus Nyström
	
	
	
	Comments:

	
*/
option(null: hard);

//-------------------------------------------------------
// class swap_curve_bc_fltois
// class used for cross currency basis swaps (ois) (quoted as 1 swap i.e. 2 ois legs)
//-------------------------------------------------------
class swap_curve_bc_fltois option (category: 'Yield Curve/003 Curve Building Cross Currency')
{
   	public: 
	
		swap_curve_bc_fltois(	date,swap_curve_cbois1_f_parm,disc_func option(nullable),disc_func option(nullable),disc_func ,fwd_func,
								fwd_func_tenor,logical, string option(nullable) = null<string>) ;
		
		swap_curve_bc_fltois(	swap_curve_bc_fltois) ;

		string 			name();
		curve 			basis_curve()	;
		curve 			blended_curve();
		fx_curve		fx_swaps();                                                                               		
		string 			settle_code() ;
		date			trade_date();
		logical			basis_is_mtm();
		
		//sprd leg
		ir_index		fwd_idx_sprd();
		string 			df_sprd_crv_name();	
		string 			adj_df_sprd_crv_name();
		disc_func		adj_disc_df_sprd(interpolator,rate_type);		
		disc_func		adj_disc_df_sprd(disc_z_model);
		disc_func		adj_disc_df_sprd(disc_z_model, disc_func option(nullable));
		integer 		flt_reset_days_sprd(); 
		date 			flt_reset_date_sprd();
		number			first_flt_sprd();		
		curve 			blended_sprd_curve();
		disc_func		disc_df_sprd();
		fwd_func        f_fwd_sprd() ;
		string 			ccy_sprd() 	;
		integer 		freq_sprd()	;
		day_count_method dc_sprd()	;
		bd_convention 	bd_sprd() ;
		calendar 		cal_sprd() ;
		logical 		eom_sprd() ;
		//vector(number) 	solve_spread_sprd_leg(vector(number) option(nullable) s = null<vector(number)>);

		
		//flat leg		
		string 			df_flat_crv_name();	
		string 			adj_df_flat_crv_name();
		disc_func		adj_disc_df_flat(interpolator,rate_type);
		disc_func		adj_disc_df_flat(disc_z_model);
		disc_func		adj_disc_df_flat(disc_z_model, disc_func option(nullable));
		curve 			blended_flat_curve();
		disc_func		disc_df_flat();
		disc_func 		fwd_df_flat();
		string 			ccy_flat() 	;
		day_count_method dc_flat()	;
		integer 		freq_flat()	;
		day_count_method dc_flat()	;
		bd_convention 	bd_flat() ;
		calendar 		cal_flat() ;
		logical 		eom_flat() ;
		//vector(number) 	solve_spread_flat_leg(vector(number) option(nullable) s = null<vector(number)>);

	protected:
		swap_curve_bc_fltois() ;
		void				init(fwd_func_tenor);
		void				init_idx(fwd_func_tenor);
		void				check_b1_curve(curve option(nullable));
		void				init_fx_swap(fx_spot option(nullable) );
		
		logical 			create_blend_curve(	out curve option(nullable), out curve option(nullable), out curve option(nullable),
												out curve option(nullable),out vector(date)   ,out string);
		void				blend_curve();
		virtual cashflows 	cf(curve, disc_func,logical);
		virtual cashflows_q cf_mtm(curve, disc_func,logical);
		disc_func			adj_disc_df_sprd_int(disc_z_model, disc_func option(nullable), logical);		
		disc_func			adj_disc_df_flat_int(disc_z_model, disc_func option(nullable), logical );
		
		date				d_;
		swap_curve_ext_parm param_;
		fx_curve			fx_swaps_;						

		disc_func			df_disc_flat_;		
		disc_func			df_disc_sprd_;

		fwd_func			f_fwd_sprd_;
		ir_index			fwd_idx_sprd_;
		
		disc_func			df_fwd_flat_;		
		logical				ois_leg_is_flat_;
		string				name_;
		vector(number) 		spread_;
		curve				curve_sprd_;
		curve				curve_flat_;
		logical				fx_flat_eq_base_;
		logical				basis_is_mtm_;
		logical 			curve_ok_;
		swap_fltois			first_b_;
};

//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
swap_curve_bc_fltois.swap_curve_bc_fltois(swap_curve_bc_fltois s) 
								:d_(s.d_), param_(s.param_), fx_swaps_(s.fx_swaps_), 								
								df_disc_flat_(s.df_disc_flat_) , df_disc_sprd_(s.df_disc_sprd_),
								 f_fwd_sprd_(s.f_fwd_sprd_) , df_fwd_flat_(s.df_fwd_flat_),
								 ois_leg_is_flat_(s.ois_leg_is_flat_),
								name_(s.name_), spread_(s.spread_), curve_sprd_(s.curve_sprd_), curve_flat_(s.curve_flat_), 
								fx_flat_eq_base_(s.fx_flat_eq_base_), basis_is_mtm_(s.basis_is_mtm_),curve_ok_(s.curve_ok_),
								first_b_(s.first_b_)
{}
		
//-------------------------------------------------------
// constructor
//-------------------------------------------------------
swap_curve_bc_fltois.swap_curve_bc_fltois()
{}

//-------------------------------------------------------
// constructor (with a basis)
// ex basis_crv = 3M SEK + sprd vs 3M USD Flat
//-------------------------------------------------------
swap_curve_bc_fltois.swap_curve_bc_fltois(date						d,									
									swap_curve_cbois1_f_parm	p,									
									disc_func option(nullable)	df_disc_flat,
									disc_func option(nullable)	df_disc_sprd,
									disc_func					df_fwd,																	
									fwd_func 					f_fwd,
									fwd_func_tenor 				fwd_tenor,	
									logical						ois_leg_is_flat,
									string 	option(nullable) 	name) 
									:d_(d), param_(p),									 
									 df_disc_flat_(df_disc_flat), df_disc_sprd_(df_disc_sprd),f_fwd_sprd_(f_fwd), 
									 df_fwd_flat_(df_fwd),
									 ois_leg_is_flat_(ois_leg_is_flat),
									 name_(name),
									 basis_is_mtm_(null<logical>),curve_ok_(false),first_b_(null<swap_fltois>)
{
	init(fwd_tenor);	
}


//-------------------------------------------------------
// init
//-------------------------------------------------------
void swap_curve_bc_fltois.init(fwd_func_tenor fwd_tenor_sprd)
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_CURR_B1_O,'invalid type of param object (should be a swap_curve_cbois1_f_parm ) ');		
	swap_curve_cbois1_f_parm param = dynamic_cast<swap_curve_cbois1_f_parm>(param_);
	if(null(name_))
		name_ = "";

	curve ccb_crv = param.basis_crv();
	QL_FAIL_COND(!check_curve(ccb_crv),"invalid currency basis curve");
	
	QL_REQUIRE(!null(df_disc_flat_) || !null(df_disc_sprd_),'invalid disc_func (either flat leg or spread (base) leg disc_func must be input)');

	check_b1_curve(ccb_crv);
	init_idx(fwd_tenor_sprd);
	init_fx_swap(param.fx_spot());
	blend_curve();	
}

//-------------------------------------------------------
// init_idx		(curr) (if only fx curve we have no ir_index)
//-------------------------------------------------------
void swap_curve_bc_fltois.init_idx(fwd_func_tenor	fwd_tenor_sprd)
{
	swap_curve_cbois1_f_parm param = dynamic_cast<swap_curve_cbois1_f_parm>(param_);
	
	if(param.short_is_fix_crv()) {
		string fwd_tenor_sprd_s		= string(fwd_tenor_sprd);
		fixing_curve new_fixing_crv;
		logical upd_sprd = CORE_INT.fwd_init(d_, param.fixing_crv(), fwd_tenor_sprd, fwd_tenor_sprd_s,fwd_idx_sprd_,new_fixing_crv);
		if(upd_sprd) 
			param.set_fixing_crv(new_fixing_crv); 
	}
	/*else {
		curve new_short_crv;
		fwd_idx_sprd_ = null<ir_index>;		
		logical upd = CORE_INT.fwd_init_depo(param.depo_crv_sprd(), fwd_tenor_sprd, new_short_crv);		
		if(upd)
			param.set_depo_crv_sprd(new_short_crv);
	}*/

	if(check_curve(param.basis_crv())) {
		instrument instr = param.basis_crv().instruments()[0];
		swap_fltois sw = swap_fltois(instr);			

		if(null(fwd_idx_sprd_)){
			fwd_idx_sprd_ = sw.ir_index_flt_leg();
		}
	}
}
//-------------------------------------------------------
// init_fx_swap
//-------------------------------------------------------
void swap_curve_bc_fltois.init_fx_swap(fx_spot option(nullable) fxs)
{
	if(check_curve(param_.short_crv())) {
		string err_name;
		QL_REQUIRE(check_quote(param_.short_crv(),err_name),'invalid quote(s) in fx curve',err_name,true);
		fx_swaps_ = new fx_curve(param_.short_crv().instruments(),fxs, fxs, quote_side_code.QS_MID) ;
		QL_REQUIRE(!null(fx_swaps_),'invalid fx curve');
		if(!equal_casei(ccy_sprd(),fxs.base_ccy()) && !equal_casei(ccy_sprd(),fxs.price_ccy()))			
			QL_FAIL('invalid fx swap (incompatible with basis swap)'); 
		if(!equal_casei(ccy_flat(),fxs.base_ccy()) && !equal_casei(ccy_flat(),fxs.price_ccy()))			
			QL_FAIL('invalid fx swap (incompatible with basis swap)'); 
		fx_flat_eq_base_ = equal_casei(ccy_flat(),fxs.base_ccy());
	}
}

ir_index		swap_curve_bc_fltois.fwd_idx_sprd()		{ return fwd_idx_sprd_;}
date 			swap_curve_bc_fltois.trade_date()		{ return d_;}
string 			swap_curve_bc_fltois.name()				{ return name_;}         
disc_func 		swap_curve_bc_fltois.disc_df_sprd()		{ return df_disc_sprd_;}
disc_func 		swap_curve_bc_fltois.disc_df_flat()		{ return df_disc_flat_;}
disc_func 		swap_curve_bc_fltois.fwd_df_flat()		{ return df_fwd_flat_;}
fwd_func 		swap_curve_bc_fltois.f_fwd_sprd()		{ return f_fwd_sprd_;}
fx_curve		swap_curve_bc_fltois.fx_swaps()			{ return fx_swaps_;}
logical			swap_curve_bc_fltois.basis_is_mtm()		{ return basis_is_mtm_;}

string 			swap_curve_bc_fltois.df_flat_crv_name() 		{ return strcat([ccy_flat(),":",ccy_flat()]);}	
string 			swap_curve_bc_fltois.adj_df_flat_crv_name()		{ return strcat([ccy_flat(),":",ccy_sprd()]);}
string 			swap_curve_bc_fltois.df_sprd_crv_name()			{ return strcat([ccy_sprd(),":",ccy_sprd()]);}		
string 			swap_curve_bc_fltois.adj_df_sprd_crv_name()		{ return strcat([ccy_sprd(),":",ccy_flat()]);}
string 			swap_curve_bc_fltois.settle_code()				{ return first_b_.settle_code();}

//-------------------------------------------------------
// first_flt_sprd 
//-------------------------------------------------------
number swap_curve_bc_fltois.first_flt_sprd()
{	
	if(param_.type() == swap_curve_type.SCE_CURR_B1_O){
		swap_curve_cbois1_f_parm param = dynamic_cast<swap_curve_cbois1_f_parm>(param_);
		string 	err_msg;
		return null(param.fixing_crv()) ? null<number>: param.fixing_crv().fixing_rate(d_, fwd_idx_sprd_.tenor(), err_msg);
	}
	//else if(param_.type() == swap_curve_type.SCE_CURR_B1){
	//	swap_curve_cb1_parm param = dynamic_cast<swap_curve_cb1_parm>(param_);
	//	return param.basis_first_fix_sprd_;
	//}
	else
		QL_FAIL("invalid function call");
}

//-------------------------------------------------------
// flt_reset_days_sprd
//-------------------------------------------------------
integer swap_curve_bc_fltois.flt_reset_days_sprd() 
{ 
	integer fd = fwd_idx_sprd_.fixing_days();
	QL_REQUIRE(fd >= 0,'invalid flt_reset_days in ir_index (negative days not supported) (fwd)');
	return fd;
}
//-------------------------------------------------------
// flt_reset_date_sprd
//-------------------------------------------------------
date swap_curve_bc_fltois.flt_reset_date_sprd() 
{ 
	date r = d_;
	while(!fwd_idx_sprd_.is_valid_fixing_date(r)) 
		r = cal_sprd().move_bus_days(r, 1);
	return r;
}

string swap_curve_bc_fltois.ccy_flat()
{
	string c1,c2;
	first_b_.currency(c1,c2);
	return c2;
}

bd_convention swap_curve_bc_fltois.bd_flat()
{
	bd_convention c1,c2;
	first_b_.pmt_bus_day(c1,c2);
	return c2;
}
logical	 swap_curve_bc_fltois.eom_flat()
{
	logical c1,c2;
	first_b_.eom(c1,c2);
	return c2;
}
calendar swap_curve_bc_fltois.cal_flat()
{
	calendar c1,c2;
	first_b_.calendar(c1,c2);
	return c2;
}

day_count_method swap_curve_bc_fltois.dc_flat()
{
	day_count_method c1,c2;
	first_b_.dc_method(c1,c2);
	return c2;
}

integer	swap_curve_bc_fltois.freq_flat()
{
	integer c1,c2;
	first_b_.coupon_freq(c1,c2);
	return c2;
}

string swap_curve_bc_fltois.ccy_sprd()
{
	string c1,c2;
	first_b_.currency(c1,c2);
	return c1;
}
bd_convention	swap_curve_bc_fltois.bd_sprd()
{
	bd_convention c1,c2;
	first_b_.pmt_bus_day(c1,c2);
	return c1;
}
day_count_method swap_curve_bc_fltois.dc_sprd()
{
	day_count_method c1,c2;
	first_b_.dc_method(c1,c2);
	return c1;
}

integer	swap_curve_bc_fltois.freq_sprd()
{
	integer c1,c2;
	first_b_.coupon_freq(c1,c2);
	return c1;
}

logical swap_curve_bc_fltois.eom_sprd()
{
	logical c1,c2;
	first_b_.eom(c1,c2);
	return c1;
}
	
calendar swap_curve_bc_fltois.cal_sprd()
{
	calendar c1,c2;
	first_b_.calendar(c1,c2);
	return c1;
}

//-------------------------------------------------------
// basis_curve
//-------------------------------------------------------
curve swap_curve_bc_fltois.basis_curve()
{ 	
	if(param_.type() == swap_curve_type.SCE_CURR_B1_O) {
		swap_curve_cbois1_f_parm param = dynamic_cast<swap_curve_cbois1_f_parm>(param_);
		return param.basis_crv();
	}
	else
		QL_FAIL("invalid function call");
}

//-------------------------------------------------------
// check_b1_curve()
//-------------------------------------------------------
void swap_curve_bc_fltois.check_b1_curve(curve option(nullable) basis_crv)
{
	if(!check_curve(basis_crv))
		return;

	string err_name;
	
	//check type	
	QL_REQUIRE(CORE_INT.check_instr_type(basis_crv, instr_type.SWAP_FLTOIS,err_name),'invalid basis swap curve (wrong instrument type)', err_name, true);

	
	vector(instrument) vi 	= basis_crv.instruments();
	instrument bi 			= vi[0];
	first_b_				= swap_fltois(bi);
	QL_FAIL_COND(!first_b_.is_valid(),"invalid currency basis swap (oisois)");
	
	//start
	date settle  			= date_code(bi.settle_code()).apply_fwd(d_, bi.calendar());		
	QL_REQUIRE(CORE_INT.check_bswap_issue(vi, settle, err_name ),'invalid currency basis swap (mixed start dates)', err_name, true);				


	//cpn freq
	integer cfreq = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,cfreq,true,err_name),'invalid currency basis swap (mixed coupon frequencies - sprd leg)',err_name,true);
	cfreq = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,cfreq,false,err_name),'invalid currency basis swap (mixed coupon frequencies - flat leg)',err_name,true);
	
	//quotes
	QL_REQUIRE(check_quote(basis_crv,err_name),'invalid quote(s) in currency basis curve',err_name,true);

	basis_is_mtm_ 			= vi[0].ccbs_is_mtm();
	spread_ 				= one_vector(v_size(vi), 0);

	day_count_method dc_flat ;//= ois_dc_flat();
	QL_REQUIRE(CORE_INT.swap_dc_check(vi,dc_flat,false,err_name),'invalid ois leg daycount method (flat leg) in currency basis swap', err_name, true);
	day_count_method dc_sprd ;//= ois_dc_sprd();
	QL_REQUIRE(CORE_INT.swap_dc_check(vi,dc_sprd,true,err_name),'invalid ois leg daycount method (spread leg) in currency basis swap', err_name, true);	

	string curr = ccy_flat();
	QL_REQUIRE(CORE_INT.swap_currency_check(vi, curr, false,err_name),'invalid ois leg currency (flat leg) in currency basis swap', err_name, true);
	curr = ccy_sprd();
	QL_REQUIRE(CORE_INT.swap_currency_check(vi, curr, true,err_name),'invalid ois leg currency (spread leg) in currency basis swap', err_name, true);

	QL_REQUIRE(CORE_INT.check_is_mtm(vi, basis_is_mtm_, err_name),'invalid mtm flag in currency basis swap', err_name, true);	
}

//-------------------------------------------------------
// create_blend_curve
//-------------------------------------------------------
logical swap_curve_bc_fltois.create_blend_curve(	out curve option(nullable) curve_flat, 
												out curve option(nullable) curve_sprd, 
												out curve option(nullable) synt_depo_flat, 
												out curve option(nullable) synt_depo_sprd,
												out vector(date)   excl_fx_mat, 
												out string err_msg)
{
	if(null(fx_swaps_)){
		curve_flat = curve_sprd = basis_curve();
		synt_depo_flat = synt_depo_sprd = null<curve>;
		err_msg = "";
		return true;
	}
	interpolator ip	= ip_linear(null<number>, 0,IP_NO_XPOL_R) ;
	rate_type ip_rt 	= RT_LOG_DF;

	vector(date) knots;			
	curve_prio	prio1 = param_.prio1() == curve_prio.SHORT ? curve_prio.SHORT : curve_prio.LONG; 
	curve_prio	prio2 = param_.prio1() == curve_prio.SHORT ? curve_prio.LONG : curve_prio.SHORT; 

	if(!null(df_disc_flat_)){//we can create depos from df_disc_flat		
						
		if(fx_flat_eq_base_) {
			synt_depo_sprd = fx_swaps_.synt_depo_price(df_disc_flat_,ip,ip_rt, dc_sprd(), excl_fx_mat,true);//SEK depo			
		}
		else {
			synt_depo_sprd = fx_swaps_.synt_depo_base(df_disc_flat_,ip,ip_rt, dc_sprd(),excl_fx_mat,true);//SEK depo
		}
		if(null(synt_depo_sprd)) {
			err_msg = 'invalid creation of deposit curve (spread leg)';
			return false;
		}		
		curve_sprd = CORE_INT.blend_curve(	synt_depo_sprd, null<curve>, basis_curve(), prio1, prio2, false, false,
											param_.merge_crv(), knots, param_.blend_buf_days());
		if(!..check_curve(curve_sprd)) {
			err_msg = 'invalid blended curve (spread leg)';
			return false;
		}				
		curve_sprd = curve_sprd.set_name("blended curve (spread leg)");	
		
	}
	if(!null(df_disc_sprd_)){		
		if(fx_flat_eq_base_) {
			synt_depo_flat = fx_swaps_.synt_depo_base(df_disc_sprd_,ip,ip_rt, dc_flat(),excl_fx_mat,true);//USD depo
		}
		else {
			synt_depo_flat = fx_swaps_.synt_depo_price(df_disc_sprd_,ip,ip_rt, dc_flat(),excl_fx_mat,true);//USD depo
		}
		if(null(synt_depo_flat)) {
			err_msg = 'invalid creation of deposit curve (flat leg)';
			return false;
		}
		
		curve_flat = CORE_INT.blend_curve(synt_depo_flat,null<curve>, basis_curve(),prio1,prio2,false,false,
										  param_.merge_crv(),knots,param_.blend_buf_days());

		if(!..check_curve(curve_flat)) {
			err_msg = 'invalid blended curve (flat leg)';
			return false;
		}			

		curve_flat = curve_flat.set_name("blended curve (flat leg)");		
	}

	err_msg = "";
	return true;	
}
//-------------------------------------------------------
// blend_curve
//-------------------------------------------------------
void swap_curve_bc_fltois.blend_curve()
{	
	curve synt_depo_flat,synt_depo_sprd;
	string err_msg;
	vector(date) excl_fx_mat;
	curve_ok_ = create_blend_curve(curve_flat_, curve_sprd_, synt_depo_flat,synt_depo_sprd,excl_fx_mat,err_msg);	
	QL_REQUIRE(curve_ok_,err_msg);

	string n;
	if(!null(curve_flat_) && !..check_quote(curve_flat_,n)) {
		string s = strcat(['invalid quote(s) in blended curve (flat leg) [',n,']']);
		QL_FAIL(s);
	}	
	if(!null(curve_sprd_) &&!..check_quote(curve_sprd_,n)) {
		string s = strcat(['invalid quote(s) in blended curve (spread leg) [',n,']']);
		QL_FAIL(s);
	}
}
//-------------------------------------------------------
// blended_curve
//-------------------------------------------------------
curve swap_curve_bc_fltois.blended_curve()
{ 
	if(!curve_ok_)
		return null<curve>;

	if(null(fx_swaps_))
		return basis_curve();		

	vector(date) knots;			
	curve_prio	prio1 	= param_.prio1() == curve_prio.SHORT ? curve_prio.SHORT : curve_prio.LONG; 
	curve_prio	prio2 	= param_.prio1() == curve_prio.SHORT ? curve_prio.LONG : curve_prio.SHORT; 
	curve curve_all 	= CORE_INT.blend_curve(param_.short_crv(), null<curve>, basis_curve(), prio1, prio2, false, false, 
												param_.merge_crv(), knots, param_.blend_buf_days());
		
	curve curve_flat, curve_sprd, synt_depo_flat,synt_depo_sprd;
	vector(date) excl_fx_mat;
	string err_msg;
	logical curve_ok = create_blend_curve(curve_flat, curve_sprd, synt_depo_flat, synt_depo_sprd, excl_fx_mat, err_msg);	
	if(!curve_ok)
		return null<curve>;
	
	if(v_size(excl_fx_mat) == 0)
		return curve_all;
	else {//here some fxswaps have been filtered out
		vector(instrument) vi_all 	= curve_all.instruments();
		vector(date) vi_all_m 		= vi_all.maturity();
		vector(integer) index_vector = v_find(vi_all_m, excl_fx_mat) ;		
		vector(instrument) vi 		= v_exclude(vi_all, index_vector) ;
		return curve(vi);
	} 
}
//-------------------------------------------------------
// blended_flat_curve
//-------------------------------------------------------
curve swap_curve_bc_fltois.blended_flat_curve()
{ 
	if(!curve_ok_ || null(df_disc_sprd_))
		return null<curve>;
	return curve_flat_;
}
//-------------------------------------------------------
// blended_sprd_curve
//-------------------------------------------------------
curve swap_curve_bc_fltois.blended_sprd_curve()
{ 
	if(!curve_ok_ || null(df_disc_flat_))
		return null<curve>;
	return curve_sprd_;
}
//-------------------------------------------------------
// cf
//-------------------------------------------------------
cashflows swap_curve_bc_fltois.cf(	curve		crv,
								disc_func 	df_disc,
								logical 	solve_flat)
{		
	vector(instrument) src	= crv.instruments();
	integer size 			= v_size(src);
	vector(cashflows) dest[0];	

	integer b 				= 0;
		
	for(integer i = 0; i < size ; i++){
		cashflows c;					
		if(src[i].is_swap_fltois()) {
			c = CORE_INT.cf_currbswap_fltois(src[i],solve_flat,f_fwd_sprd_,first_flt_sprd(), df_fwd_flat_, df_disc, spread_[b]);//nom=1
			b++;			
		}
		else if(src[i].is_deposit() || src[i].is_deposit_index()) 
			c = CORE_INT.cf_depo(src[i], 0);	//nom=1
		
		else
			QL_FAIL("invalid instrument type (currency basis)");

		QL_REQUIRE(!null(c),"invalid cashflow (currency basis)", src[i].name(), true);
		push_back(	dest, c);	
	}
	return cashflows(dest);
}
//-------------------------------------------------------
// cf_mtm
//-------------------------------------------------------
cashflows_q swap_curve_bc_fltois.cf_mtm(curve		crv,
									disc_func 	df_disc,
									logical 	solve_flat)
{		
	vector(instrument) src	= crv.instruments();
	integer size 			= v_size(src);
	vector(cashflows_q) dest[0];	
	integer b 				= 0;

	for(integer i = 0; i < size ; i++){
		cashflows_q c;					
		if(src[i].is_swap_oisois()) {
			//c = CORE_INT.cf_currbswap_fltois_mtm(	src[i],solve_flat,f_fwd_sprd_, first_flt_sprd(),df_fwd_flat_, df_disc, spread_[b]);//nom=1
			b++;			
		}
		else if(src[i].is_deposit() || src[i].is_deposit_index()) 
			c = cashflows_q(CORE_INT.cf_depo(src[i], 0)) ;	//nom=1			
		else
			QL_FAIL("invalid instrument type (currency basis)");

		QL_REQUIRE(!null(c),"invalid cashflow (currency basis)", src[i].name(), true);
		push_back(	dest, c);	
	}
	return cashflows_q(dest);
}
//-------------------------------------------------------
// adj_disc_df_sprd
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_fltois.adj_disc_df_sprd(	interpolator 	disc_ip,
												rate_type		disc_ip_rt)
							   
{	
	QL_REQUIRE(!null(df_disc_flat_),"invalid df_disc_flat");

	//with the assumption that we have self discounting the mtm style has no effect on the resulting disc_func
	
	/*if(basis_is_mtm_ ) {
		QL_FAIL("mtm swaps not supported");

		cashflows_q cflow = cf_mtm(curve_sprd_, df_disc_flat_, false);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (spread leg)");
		if(null(df_disc_sprd_))
			return bootstrap(cflow, disc_ip, disc_ip_rt);		
		else
			return bootstrap_spr(df_disc_sprd_, cflow, disc_ip, disc_ip_rt);
	}
	else {*/
		cashflows cflow = cf(curve_sprd_, df_disc_flat_, false);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (spread leg)");
		if(null(df_disc_sprd_))
			return bootstrap(cflow, disc_ip, disc_ip_rt);		
		else
			return bootstrap_spr(df_disc_sprd_, cflow, disc_ip, disc_ip_rt);
	//}
}
//-------------------------------------------------------
// adj_disc_df_sprd
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_fltois.adj_disc_df_sprd(disc_z_model disc_model)
							   
{	
	return adj_disc_df_sprd_int(disc_model, df_disc_sprd_, !null(df_disc_sprd_));
}
//-------------------------------------------------------
// adj_disc_df_sprd
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_fltois.adj_disc_df_sprd(disc_z_model disc_model, disc_func option(nullable) df_disc_sprd_base)
							   
{	
	return adj_disc_df_sprd_int(disc_model, df_disc_sprd_base, !null(df_disc_sprd_base));
}


//-------------------------------------------------------
// adj_disc_df_sprd_int
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_fltois.adj_disc_df_sprd_int(disc_z_model disc_model, 
												disc_func option(nullable) df_disc_sprd_base,
												logical use_base_df)
							   
{	
	QL_REQUIRE(!null(df_disc_flat_),"invalid df_disc_flat");
	if(null(df_disc_sprd_base)) df_disc_sprd_base = df_disc_sprd_;

	//with the assumption that we have self discounting the mtm style has no effect on the resulting disc_func
	
	/*if(basis_is_mtm_ ) {
		QL_FAIL("mtm swaps not supported");
		cashflows_q cflow = cf_mtm(curve_sprd_, df_disc_flat_, false);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (spread leg)");
		if(!use_base_df)  
			return disc_model.df(cflow, curve_sprd_.instruments().pvbp());	
		else 			
			return disc_model.df(df_disc_sprd_base, cflow, curve_sprd_.instruments().pvbp());	
	}
	else {*/
		cashflows cflow = cf(curve_sprd_, df_disc_flat_, false);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (spread leg)");
		vector(number) pvbp;// = curve_sprd_.instruments().pvbp();
		if(!use_base_df) 
			return disc_model.df(cflow, pvbp);
		else 
			return disc_model.df(df_disc_sprd_base,cflow, pvbp);
	//}
}

//-------------------------------------------------------
// adj_disc_df_flat
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_fltois.adj_disc_df_flat(	interpolator 	disc_ip,
												rate_type		disc_ip_rt)
							   
{	
	QL_REQUIRE(!null(df_disc_sprd_),"invalid df_disc_sprd");

	//with the assumption that we have self discounting the mtm style has no effect on the resulting disc_func
	
	/*if(basis_is_mtm_) {
		QL_FAIL("mtm swaps not supported");
		cashflows_q cflow = cf_mtm(curve_flat_, df_disc_sprd_, true);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (flat leg)");
		if(null(df_disc_flat_))
			return bootstrap(cflow, disc_ip, disc_ip_rt);		
		else
			return bootstrap_spr(df_disc_flat_, cflow, disc_ip, disc_ip_rt);
	}
	else {*/
		cashflows cflow = cf(curve_flat_, df_disc_sprd_, true);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (flat leg)");
		if(null(df_disc_flat_))
			return bootstrap(cflow, disc_ip, disc_ip_rt);		
		else
			return bootstrap_spr(df_disc_flat_, cflow, disc_ip, disc_ip_rt);
	//}
}
//-------------------------------------------------------
// adj_disc_df_flat
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_fltois.adj_disc_df_flat(disc_z_model disc_model)
							   
{	
	return adj_disc_df_flat_int(disc_model, df_disc_flat_, !null(df_disc_flat_));
}
//-------------------------------------------------------
// adj_disc_df_flat
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_fltois.adj_disc_df_flat(disc_z_model disc_model, disc_func option(nullable) df_disc_flat_base)
							   
{	
	return adj_disc_df_flat_int(disc_model, df_disc_flat_base, !null(df_disc_flat_base));
}

//-------------------------------------------------------
// adj_disc_df_flat_int
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_fltois.adj_disc_df_flat_int(	disc_z_model disc_model, 
													disc_func option(nullable) df_disc_flat_base,
													logical use_base_df)
							   
{	
	QL_REQUIRE(!null(df_disc_sprd_),"invalid df_disc_sprd");

	//with the assumption that we have self discounting the mtm style has no effect on the resulting disc_func
	
	/*if(basis_is_mtm_) {
		QL_FAIL("mtm swaps not supported");
		cashflows_q cflow = cf_mtm(curve_flat_, df_disc_sprd_, true);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (flat leg)");
		if(!use_base_df) 
			return disc_model.df(cflow, curve_flat_.instruments().pvbp());			
		else 
			return disc_model.df(df_disc_flat_base,cflow, curve_flat_.instruments().pvbp());		
	}
	else {*/
		cashflows cflow = cf(curve_flat_, df_disc_sprd_, true);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (flat leg)");
		vector(number) pvbp;// = curve_flat_.instruments().pvbp();
		if(!use_base_df)
			return disc_model.df(cflow, pvbp);		
		else 
			return disc_model.df(df_disc_flat_base,cflow, pvbp);			
	//}
}
//-------------------------------------------------------
// create
//-------------------------------------------------------
swap_curve_bc_fltois swap_curve_bc_fltois(	date						d,									
											swap_curve_cbois1_f_parm	p,									
											disc_func option(nullable)	df_disc_flat,
											disc_func option(nullable)	df_disc_sprd,
											disc_func					df_fwd,																	
											fwd_func 					f_fwd,
											fwd_func_tenor				fwd_tenor,
											logical						ois_leg_is_flat,
											string 	option(nullable) 	name)
{
	return new swap_curve_bc_fltois(d, p,df_disc_flat,df_disc_sprd,df_fwd,f_fwd,fwd_tenor,ois_leg_is_flat,name)	;	
}

//-------------------------------------------------------
//-------------------------------------------------------
//-------------------------------------------------------
//-------------------------------------------------------
//-------------------------------------------------------



//-------------------------------------------------------
// class swap_curve_bc_ois
// class used for cross currency basis swaps (ois) (quoted as 1 swap i.e. 2 ois legs)
//-------------------------------------------------------
class swap_curve_bc_ois option (category: 'Yield Curve/003 Curve Building Cross Currency')
{
   	public: 
	
		swap_curve_bc_ois(	date,swap_curve_cbois1_f_parm,disc_func option(nullable),disc_func option(nullable),disc_func ,disc_func,string option(nullable) = null<string>) ;
		
		swap_curve_bc_ois(	swap_curve_bc_ois) ;		

		string 			name();
		curve 			basis_curve()	;
		curve 			blended_curve();
		fx_curve		fx_swaps();                                                                               		
		string 			settle_code() ;
		date			trade_date();
		logical			basis_is_mtm();

		//sprd leg
		string 			df_sprd_crv_name();	
		string 			adj_df_sprd_crv_name();
		disc_func		adj_disc_df_sprd(interpolator,rate_type);		
		disc_func		adj_disc_df_sprd(disc_z_model);
		disc_func		adj_disc_df_sprd(disc_z_model, disc_func option(nullable));
		curve 			blended_sprd_curve();
		disc_func		disc_df_sprd();
		disc_func		fwd_df_sprd();
		string 			ccy_sprd() 	;
		integer 		ois_freq_sprd()	;
		day_count_method ois_dc_sprd()	;
		bd_convention 	ois_bd_sprd() ;
		calendar 		ois_cal_sprd() ;
		logical 		ois_eom_sprd() ;
		vector(number) 	solve_spread_sprd_leg(vector(number) option(nullable) s = null<vector(number)>);
		
		//flat leg
		string 			df_flat_crv_name();	
		string 			adj_df_flat_crv_name();
		disc_func		adj_disc_df_flat(interpolator,rate_type);
		disc_func		adj_disc_df_flat(disc_z_model);
		disc_func		adj_disc_df_flat(disc_z_model, disc_func option(nullable));
		curve 			blended_flat_curve();
		disc_func		disc_df_flat();
		disc_func		fwd_df_flat();
		string 			ccy_flat() 	;
		day_count_method ois_dc_flat()	;
		integer 		ois_freq_flat()	;
		day_count_method ois_dc_flat()	;
		bd_convention 	ois_bd_flat() ;
		calendar 		ois_cal_flat() ;
		logical 		ois_eom_flat() ;
		vector(number) 	solve_spread_flat_leg(vector(number) option(nullable) s = null<vector(number)>);
		
	protected:

		swap_curve_bc_ois() ;
		void				init();
		void				check_b1_curve(curve option(nullable));
		void				init_fx_swap(fx_spot option(nullable) );
		
		logical 			create_blend_curve(	out curve option(nullable), out curve option(nullable), out curve option(nullable),
												out curve option(nullable),out vector(date)   ,out string);
		void				blend_curve();
		virtual cashflows 	cf(curve, disc_func,logical);
		virtual cashflows_q cf_mtm(curve, disc_func,logical);
		disc_func			adj_disc_df_sprd_int(disc_z_model, disc_func option(nullable), logical);		
		disc_func			adj_disc_df_flat_int(disc_z_model, disc_func option(nullable), logical );
	//private:
		date				d_;
		swap_curve_ext_parm param_;
		fx_curve			fx_swaps_;						

		disc_func			df_disc_flat_;		
		disc_func			df_disc_sprd_;

		disc_func			df_fwd_flat_;		
		disc_func			df_fwd_sprd_;
		
		string				name_;
		vector(number) 		spread_;
		curve				curve_sprd_;
		curve				curve_flat_;
		logical				fx_flat_eq_base_;
		logical				basis_is_mtm_;
		logical 			curve_ok_;
		swap_oisois			first_b_;
};

date 			swap_curve_bc_ois.trade_date()		{ return d_;}
string 			swap_curve_bc_ois.name()			{ return name_;}         
disc_func 		swap_curve_bc_ois.disc_df_sprd()	{ return df_disc_sprd_;}
disc_func 		swap_curve_bc_ois.disc_df_flat()	{ return df_disc_flat_;}
disc_func 		swap_curve_bc_ois.fwd_df_sprd()		{ return df_fwd_sprd_;}
disc_func 		swap_curve_bc_ois.fwd_df_flat()		{ return df_fwd_flat_;}
fx_curve		swap_curve_bc_ois.fx_swaps()		{ return fx_swaps_;}
logical			swap_curve_bc_ois.basis_is_mtm()	{ return basis_is_mtm_;}

string 			swap_curve_bc_ois.df_flat_crv_name() 		{ return strcat([ccy_flat(),":",ccy_flat()]);}	
string 			swap_curve_bc_ois.adj_df_flat_crv_name()	{ return strcat([ccy_flat(),":",ccy_sprd()]);}
string 			swap_curve_bc_ois.df_sprd_crv_name()		{ return strcat([ccy_sprd(),":",ccy_sprd()]);}		
string 			swap_curve_bc_ois.adj_df_sprd_crv_name()	{ return strcat([ccy_sprd(),":",ccy_flat()]);}
string 			swap_curve_bc_ois.settle_code()				{ return first_b_.settle_code();}

string swap_curve_bc_ois.ccy_flat()
{
	string c1,c2;
	first_b_.currency(c1,c2);
	return c2;
}

integer	 swap_curve_bc_ois.ois_freq_flat()
{
	integer c1,c2;
	first_b_.coupon_freq(c1,c2);
	return c2;
}

day_count_method swap_curve_bc_ois.ois_dc_flat()
{
	day_count_method c1,c2;
	first_b_.dc_method(c1,c2);
	return c2;
}
bd_convention swap_curve_bc_ois.ois_bd_flat()
{
	bd_convention c1,c2;
	first_b_.pmt_bus_day(c1,c2);
	return c2;
}
logical	 swap_curve_bc_ois.ois_eom_flat()
{
	logical c1,c2;
	first_b_.eom(c1,c2);
	return c2;
}
calendar swap_curve_bc_ois.ois_cal_flat()
{
	calendar c1,c2;
	first_b_.calendar(c1,c2);
	return c2;
}

integer	swap_curve_bc_ois.ois_freq_sprd()
{
	integer c1,c2;
	first_b_.coupon_freq(c1,c2);
	return c1;
}
day_count_method swap_curve_bc_ois.ois_dc_sprd()
{
	day_count_method c1,c2;
	first_b_.dc_method(c1,c2);
	return c1;
}
bd_convention	swap_curve_bc_ois.ois_bd_sprd()
{
	bd_convention c1,c2;
	first_b_.pmt_bus_day(c1,c2);
	return c1;
}
logical swap_curve_bc_ois.ois_eom_sprd()
{
	logical c1,c2;
	first_b_.eom(c1,c2);
	return c1;
}
	
calendar swap_curve_bc_ois.ois_cal_sprd()
{
	calendar c1,c2;
	first_b_.calendar(c1,c2);
	return c1;
}

string swap_curve_bc_ois.ccy_sprd()
{
	string c1,c2;
	first_b_.currency(c1,c2);
	return c1;
}

//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
swap_curve_bc_ois.swap_curve_bc_ois(swap_curve_bc_ois s) 
								:d_(s.d_), param_(s.param_), fx_swaps_(s.fx_swaps_), 								
								df_disc_flat_(s.df_disc_flat_) , df_disc_sprd_(s.df_disc_sprd_),
								 df_fwd_flat_(s.df_fwd_flat_) , df_fwd_sprd_(s.df_fwd_sprd_),
								name_(s.name_), spread_(s.spread_), curve_sprd_(s.curve_sprd_), curve_flat_(s.curve_flat_), 
								fx_flat_eq_base_(s.fx_flat_eq_base_), basis_is_mtm_(s.basis_is_mtm_),curve_ok_(s.curve_ok_),
								first_b_(s.first_b_)
{}
		
//-------------------------------------------------------
// constructor
//-------------------------------------------------------
swap_curve_bc_ois.swap_curve_bc_ois()
{}

//-------------------------------------------------------
// constructor (with a basis)
// ex basis_crv = 3M SEK + sprd vs 3M USD Flat
//-------------------------------------------------------
swap_curve_bc_ois.swap_curve_bc_ois(date						d,									
									swap_curve_cbois1_f_parm	p,									
									disc_func option(nullable)	df_disc_flat,
									disc_func option(nullable)	df_disc_sprd,
									disc_func					df_fwd_flat,																	
									disc_func 					df_fwd_sprd,	
									string 	option(nullable) 	name) 
									:d_(d), param_(p),									 
									 df_disc_flat_(df_disc_flat), df_disc_sprd_(df_disc_sprd),df_fwd_flat_(df_fwd_flat), df_fwd_sprd_(df_fwd_sprd), 
									 name_(name),
									 basis_is_mtm_(null<logical>),curve_ok_(false),first_b_(null<swap_oisois>)
{
	init();	
}

//-------------------------------------------------------
// init
//-------------------------------------------------------
void swap_curve_bc_ois.init()
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_CURR_B1_O,'invalid type of param object (should be a swap_curve_cbois1_f_parm ) ');		
	swap_curve_cbois1_f_parm param = dynamic_cast<swap_curve_cbois1_f_parm>(param_);
	if(null(name_))
		name_ = "";

	curve ccb_crv = param.basis_crv();
	QL_FAIL_COND(!check_curve(ccb_crv),"invalid currency basis curve");
	
	QL_REQUIRE(!null(df_disc_flat_) || !null(df_disc_sprd_),'invalid disc_func (either flat leg or spread (base) leg disc_func must be input)');

	check_b1_curve(ccb_crv);	
	init_fx_swap(param.fx_spot());
	blend_curve();	
}

//-------------------------------------------------------
// check_b1_curve()
//-------------------------------------------------------
void swap_curve_bc_ois.check_b1_curve(curve option(nullable) basis_crv)
{
	if(!check_curve(basis_crv))
		return;

	string err_name;
	
	//check type	
	QL_REQUIRE(CORE_INT.check_instr_type(basis_crv, instr_type.SWAP_OISOIS,err_name),'invalid basis swap curve (wrong instrument type)', err_name, true);

	
	vector(instrument) vi 	= basis_crv.instruments();
	instrument bi 			= vi[0];
	first_b_				= swap_oisois(bi);
	QL_FAIL_COND(!first_b_.is_valid(),"invalid currency basis swap (oisois)");
	
	//start
	date settle  			= date_code(bi.settle_code()).apply_fwd(d_, bi.calendar());		
	QL_REQUIRE(CORE_INT.check_bswap_issue(vi, settle, err_name ),'invalid currency basis swap (mixed start dates)', err_name, true);				


	//cpn freq
	integer cfreq = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,cfreq,true,err_name),'invalid currency basis swap (mixed coupon frequencies - sprd leg)',err_name,true);
	cfreq = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,cfreq,false,err_name),'invalid currency basis swap (mixed coupon frequencies - flat leg)',err_name,true);
	
	//quotes
	QL_REQUIRE(check_quote(basis_crv,err_name),'invalid quote(s) in currency basis curve',err_name,true);

	basis_is_mtm_ 			= vi[0].ccbs_is_mtm();
	spread_ 				= one_vector(v_size(vi), 0);

	day_count_method dc_flat = ois_dc_flat();
	QL_REQUIRE(CORE_INT.swap_dc_check(vi,dc_flat,false,err_name),'invalid ois leg daycount method (flat leg) in currency basis swap', err_name, true);
	day_count_method dc_sprd = ois_dc_sprd();
	QL_REQUIRE(CORE_INT.swap_dc_check(vi,dc_sprd,true,err_name),'invalid ois leg daycount method (spread leg) in currency basis swap', err_name, true);	

	string curr = ccy_flat();
	QL_REQUIRE(CORE_INT.swap_currency_check(vi, curr, false,err_name),'invalid ois leg currency (flat leg) in currency basis swap', err_name, true);
	curr = ccy_sprd();
	QL_REQUIRE(CORE_INT.swap_currency_check(vi, curr, true,err_name),'invalid ois leg currency (spread leg) in currency basis swap', err_name, true);

	//with the assumption that we have self discounting the mtm style has no effect on the resulting disc_func
	//QL_REQUIRE(CORE_INT.check_is_mtm(vi, basis_is_mtm_, err_name),'invalid mtm flag in currency basis swap', err_name, true);	
}

//-------------------------------------------------------
// init_fx_swap
//-------------------------------------------------------
void swap_curve_bc_ois.init_fx_swap(fx_spot option(nullable) fxs)
{
	if(check_curve(param_.short_crv())) {
		string err_name;
		QL_REQUIRE(check_quote(param_.short_crv(),err_name),'invalid quote(s) in fx curve',err_name,true);
		fx_swaps_ = new fx_curve(param_.short_crv().instruments(),fxs, fxs, quote_side_code.QS_MID) ;
		QL_REQUIRE(!null(fx_swaps_),'invalid fx curve');
		if(!equal_casei(ccy_sprd(),fxs.base_ccy()) && !equal_casei(ccy_sprd(),fxs.price_ccy()))			
			QL_FAIL('invalid fx swap (incompatible with basis swap)'); 
		if(!equal_casei(ccy_flat(),fxs.base_ccy()) && !equal_casei(ccy_flat(),fxs.price_ccy()))			
			QL_FAIL('invalid fx swap (incompatible with basis swap)'); 
		fx_flat_eq_base_ = equal_casei(ccy_flat(),fxs.base_ccy());
	}
}
//-------------------------------------------------------
// basis_curve
//-------------------------------------------------------
curve swap_curve_bc_ois.basis_curve()
{ 	
	if(param_.type() == swap_curve_type.SCE_CURR_B1_O) {
		swap_curve_cbois1_f_parm param = dynamic_cast<swap_curve_cbois1_f_parm>(param_);
		return param.basis_crv();
	}
	else
		QL_FAIL("invalid function call");
}

//-------------------------------------------------------
// create_blend_curve
//-------------------------------------------------------
logical swap_curve_bc_ois.create_blend_curve(	out curve option(nullable) curve_flat, 
												out curve option(nullable) curve_sprd, 
												out curve option(nullable) synt_depo_flat, 
												out curve option(nullable) synt_depo_sprd,
												out vector(date)   excl_fx_mat, 
												out string err_msg)
{
	if(null(fx_swaps_)){
		curve_flat = curve_sprd = basis_curve();
		synt_depo_flat = synt_depo_sprd = null<curve>;
		err_msg = "";
		return true;
	}
	interpolator ip	= ip_linear(null<number>, 0,IP_NO_XPOL_R) ;
	rate_type ip_rt 	= RT_LOG_DF;

	vector(date) knots;			
	curve_prio	prio1 = param_.prio1() == curve_prio.SHORT ? curve_prio.SHORT : curve_prio.LONG; 
	curve_prio	prio2 = param_.prio1() == curve_prio.SHORT ? curve_prio.LONG : curve_prio.SHORT; 

	if(!null(df_disc_flat_)){//we can create depos from df_disc_flat		
						
		if(fx_flat_eq_base_) {
			synt_depo_sprd = fx_swaps_.synt_depo_price(df_disc_flat_,ip,ip_rt, ois_dc_sprd(), excl_fx_mat,true);//SEK depo			
		}
		else {
			synt_depo_sprd = fx_swaps_.synt_depo_base(df_disc_flat_,ip,ip_rt, ois_dc_sprd(),excl_fx_mat,true);//SEK depo
		}
		if(null(synt_depo_sprd)) {
			err_msg = 'invalid creation of deposit curve (spread leg)';
			return false;
		}		
		curve_sprd = CORE_INT.blend_curve(	synt_depo_sprd, null<curve>, basis_curve(), prio1, prio2, false, false,
											param_.merge_crv(), knots, param_.blend_buf_days());
		if(!..check_curve(curve_sprd)) {
			err_msg = 'invalid blended curve (spread leg)';
			return false;
		}				
		curve_sprd = curve_sprd.set_name("blended curve (spread leg)");	
		
	}
	if(!null(df_disc_sprd_)){		
		if(fx_flat_eq_base_) {
			synt_depo_flat = fx_swaps_.synt_depo_base(df_disc_sprd_,ip,ip_rt, ois_dc_flat(),excl_fx_mat,true);//USD depo
		}
		else {
			synt_depo_flat = fx_swaps_.synt_depo_price(df_disc_sprd_,ip,ip_rt, ois_dc_flat(),excl_fx_mat,true);//USD depo
		}
		if(null(synt_depo_flat)) {
			err_msg = 'invalid creation of deposit curve (flat leg)';
			return false;
		}
		
		curve_flat = CORE_INT.blend_curve(synt_depo_flat,null<curve>, basis_curve(),prio1,prio2,false,false,
										  param_.merge_crv(),knots,param_.blend_buf_days());

		if(!..check_curve(curve_flat)) {
			err_msg = 'invalid blended curve (flat leg)';
			return false;
		}			

		curve_flat = curve_flat.set_name("blended curve (flat leg)");		
	}

	err_msg = "";
	return true;	
}
//-------------------------------------------------------
// blend_curve
//-------------------------------------------------------
void swap_curve_bc_ois.blend_curve()
{	
	curve synt_depo_flat,synt_depo_sprd;
	string err_msg;
	vector(date) excl_fx_mat;
	curve_ok_ = create_blend_curve(curve_flat_, curve_sprd_, synt_depo_flat,synt_depo_sprd,excl_fx_mat,err_msg);	
	QL_REQUIRE(curve_ok_,err_msg);

	string n;
	if(!null(curve_flat_) && !..check_quote(curve_flat_,n)) {
		string s = strcat(['invalid quote(s) in blended curve (flat leg) [',n,']']);
		QL_FAIL(s);
	}	
	if(!null(curve_sprd_) &&!..check_quote(curve_sprd_,n)) {
		string s = strcat(['invalid quote(s) in blended curve (spread leg) [',n,']']);
		QL_FAIL(s);
	}
}
//-------------------------------------------------------
// cf
//-------------------------------------------------------
cashflows swap_curve_bc_ois.cf(	curve		crv,
								disc_func 	df_disc,
								logical 	solve_flat)
{		
	vector(instrument) src	= crv.instruments();
	integer size 			= v_size(src);
	vector(cashflows) dest[0];	

	integer b 				= 0;
		
	for(integer i = 0; i < size ; i++){
		cashflows c;					
		if(src[i].is_swap_oisois()) {
			c = CORE_INT.cf_currbswap_ois(src[i],solve_flat,df_fwd_sprd_, df_fwd_flat_, df_disc, spread_[b]);//nom=1
			b++;			
		}
		else if(src[i].is_deposit() || src[i].is_deposit_index()) 
			c = CORE_INT.cf_depo(src[i], 0);	//nom=1
		
		else
			QL_FAIL("invalid instrument type (currency basis)");

		QL_REQUIRE(!null(c),"invalid cashflow (currency basis)", src[i].name(), true);
		push_back(	dest, c);	
	}
	return cashflows(dest);
}
//-------------------------------------------------------
// cf_mtm
//-------------------------------------------------------
cashflows_q swap_curve_bc_ois.cf_mtm(curve		crv,
									disc_func 	df_disc,
									logical 	solve_flat)
{		
	vector(instrument) src	= crv.instruments();
	integer size 			= v_size(src);
	vector(cashflows_q) dest[0];	
	integer b 				= 0;

	for(integer i = 0; i < size ; i++){
		cashflows_q c;					
		if(src[i].is_swap_oisois()) {
			c = CORE_INT.cf_currbswap_ois_mtm(	src[i],solve_flat,df_fwd_sprd_, df_fwd_flat_, df_disc, spread_[b]);//nom=1
			b++;			
		}
		else if(src[i].is_deposit() || src[i].is_deposit_index()) 
			c = cashflows_q(CORE_INT.cf_depo(src[i], 0)) ;	//nom=1			
		else
			QL_FAIL("invalid instrument type (currency basis)");

		QL_REQUIRE(!null(c),"invalid cashflow (currency basis)", src[i].name(), true);
		push_back(	dest, c);	
	}
	return cashflows_q(dest);
}
//-------------------------------------------------------
// adj_disc_df_sprd
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ois.adj_disc_df_sprd(	interpolator 	disc_ip,
												rate_type		disc_ip_rt)
							   
{	
	QL_REQUIRE(!null(df_disc_flat_),"invalid df_disc_flat");

	//with the assumption that we have self discounting the mtm style has no effect on the resulting disc_func
	
	/*if(basis_is_mtm_ ) {
		QL_FAIL("mtm swaps not supported");

		cashflows_q cflow = cf_mtm(curve_sprd_, df_disc_flat_, false);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (spread leg)");
		if(null(df_disc_sprd_))
			return bootstrap(cflow, disc_ip, disc_ip_rt);		
		else
			return bootstrap_spr(df_disc_sprd_, cflow, disc_ip, disc_ip_rt);
	}
	else {*/
		cashflows cflow = cf(curve_sprd_, df_disc_flat_, false);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (spread leg)");
		if(null(df_disc_sprd_))
			return bootstrap(cflow, disc_ip, disc_ip_rt);		
		else
			return bootstrap_spr(df_disc_sprd_, cflow, disc_ip, disc_ip_rt);
	//}
}
//-------------------------------------------------------
// adj_disc_df_sprd
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ois.adj_disc_df_sprd(disc_z_model disc_model)
							   
{	
	return adj_disc_df_sprd_int(disc_model, df_disc_sprd_, !null(df_disc_sprd_));
}
//-------------------------------------------------------
// adj_disc_df_sprd
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ois.adj_disc_df_sprd(disc_z_model disc_model, disc_func option(nullable) df_disc_sprd_base)
							   
{	
	return adj_disc_df_sprd_int(disc_model, df_disc_sprd_base, !null(df_disc_sprd_base));
}


//-------------------------------------------------------
// adj_disc_df_sprd_int
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ois.adj_disc_df_sprd_int(disc_z_model disc_model, 
												disc_func option(nullable) df_disc_sprd_base,
												logical use_base_df)
							   
{	
	QL_REQUIRE(!null(df_disc_flat_),"invalid df_disc_flat");
	if(null(df_disc_sprd_base)) df_disc_sprd_base = df_disc_sprd_;

	//with the assumption that we have self discounting the mtm style has no effect on the resulting disc_func
	
	/*if(basis_is_mtm_ ) {
		QL_FAIL("mtm swaps not supported");
		cashflows_q cflow = cf_mtm(curve_sprd_, df_disc_flat_, false);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (spread leg)");
		if(!use_base_df)  
			return disc_model.df(cflow, curve_sprd_.instruments().pvbp());	
		else 			
			return disc_model.df(df_disc_sprd_base, cflow, curve_sprd_.instruments().pvbp());	
	}
	else {*/
		cashflows cflow = cf(curve_sprd_, df_disc_flat_, false);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (spread leg)");
		if(!use_base_df) 
			return disc_model.df(cflow, curve_sprd_.instruments().pvbp());
		else 
			return disc_model.df(df_disc_sprd_base,cflow, curve_sprd_.instruments().pvbp());
	//}
}

//-------------------------------------------------------
// adj_disc_df_flat
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ois.adj_disc_df_flat(	interpolator 	disc_ip,
												rate_type		disc_ip_rt)
							   
{	
	QL_REQUIRE(!null(df_disc_sprd_),"invalid df_disc_sprd");

	//with the assumption that we have self discounting the mtm style has no effect on the resulting disc_func
	
	/*if(basis_is_mtm_) {
		QL_FAIL("mtm swaps not supported");
		cashflows_q cflow = cf_mtm(curve_flat_, df_disc_sprd_, true);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (flat leg)");
		if(null(df_disc_flat_))
			return bootstrap(cflow, disc_ip, disc_ip_rt);		
		else
			return bootstrap_spr(df_disc_flat_, cflow, disc_ip, disc_ip_rt);
	}
	else {*/
		cashflows cflow = cf(curve_flat_, df_disc_sprd_, true);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (flat leg)");
		if(null(df_disc_flat_))
			return bootstrap(cflow, disc_ip, disc_ip_rt);		
		else
			return bootstrap_spr(df_disc_flat_, cflow, disc_ip, disc_ip_rt);
	//}
}
//-------------------------------------------------------
// adj_disc_df_flat
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ois.adj_disc_df_flat(disc_z_model disc_model)
							   
{	
	return adj_disc_df_flat_int(disc_model, df_disc_flat_, !null(df_disc_flat_));
}
//-------------------------------------------------------
// adj_disc_df_flat
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ois.adj_disc_df_flat(disc_z_model disc_model, disc_func option(nullable) df_disc_flat_base)
							   
{	
	return adj_disc_df_flat_int(disc_model, df_disc_flat_base, !null(df_disc_flat_base));
}

//-------------------------------------------------------
// adj_disc_df_flat_int
// fx-implied disc_func
//-------------------------------------------------------
disc_func swap_curve_bc_ois.adj_disc_df_flat_int(	disc_z_model disc_model, 
													disc_func option(nullable) df_disc_flat_base,
													logical use_base_df)
							   
{	
	QL_REQUIRE(!null(df_disc_sprd_),"invalid df_disc_sprd");

	//with the assumption that we have self discounting the mtm style has no effect on the resulting disc_func
	
	/*if(basis_is_mtm_) {
		QL_FAIL("mtm swaps not supported");
		cashflows_q cflow = cf_mtm(curve_flat_, df_disc_sprd_, true);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (flat leg)");
		if(!use_base_df) 
			return disc_model.df(cflow, curve_flat_.instruments().pvbp());			
		else 
			return disc_model.df(df_disc_flat_base,cflow, curve_flat_.instruments().pvbp());		
	}
	else {*/
		cashflows cflow = cf(curve_flat_, df_disc_sprd_, true);
		QL_REQUIRE(!null(cflow ),"error creating cashflows (flat leg)");
		if(!use_base_df)
			return disc_model.df(cflow, curve_flat_.instruments().pvbp());		
		else 
			return disc_model.df(df_disc_flat_base,cflow, curve_flat_.instruments().pvbp());			
	//}
}

//-------------------------------------------------------
// blended_curve
//-------------------------------------------------------
curve swap_curve_bc_ois.blended_curve()
{ 
	if(!curve_ok_)
		return null<curve>;

	if(null(fx_swaps_))
		return basis_curve();		

	vector(date) knots;			
	curve_prio	prio1 	= param_.prio1() == curve_prio.SHORT ? curve_prio.SHORT : curve_prio.LONG; 
	curve_prio	prio2 	= param_.prio1() == curve_prio.SHORT ? curve_prio.LONG : curve_prio.SHORT; 
	curve curve_all 	= CORE_INT.blend_curve(param_.short_crv(), null<curve>, basis_curve(), prio1, prio2, false, false, 
												param_.merge_crv(), knots, param_.blend_buf_days());
		
	curve curve_flat, curve_sprd, synt_depo_flat,synt_depo_sprd;
	vector(date) excl_fx_mat;
	string err_msg;
	logical curve_ok = create_blend_curve(curve_flat, curve_sprd, synt_depo_flat, synt_depo_sprd, excl_fx_mat, err_msg);	
	if(!curve_ok)
		return null<curve>;
	
	if(v_size(excl_fx_mat) == 0)
		return curve_all;
	else {//here some fxswaps have been filtered out
		vector(instrument) vi_all 	= curve_all.instruments();
		vector(date) vi_all_m 		= vi_all.maturity();
		vector(integer) index_vector = v_find(vi_all_m, excl_fx_mat) ;		
		vector(instrument) vi 		= v_exclude(vi_all, index_vector) ;
		return curve(vi);
	} 
}

//-------------------------------------------------------
// blended_flat_curve
//-------------------------------------------------------
curve swap_curve_bc_ois.blended_flat_curve()
{ 
	if(!curve_ok_ || null(df_disc_sprd_))
		return null<curve>;
	return curve_flat_;
}
//-------------------------------------------------------
// blended_sprd_curve
//-------------------------------------------------------
curve swap_curve_bc_ois.blended_sprd_curve()
{ 
	if(!curve_ok_ || null(df_disc_flat_))
		return null<curve>;
	return curve_sprd_;
}
//-------------------------------------------------------
// solve_spread_sprd_leg
//-------------------------------------------------------
vector(number) swap_curve_bc_ois.solve_spread_sprd_leg(vector(number) option(nullable) flat_sprd)
{
	QL_FAIL("work in progress");
	
	//with the assumption that we have self discounting the mtm style has no effect on the resulting disc_func
	//if(basis_is_mtm_)
		//QL_FAIL("mtm swaps not supported for this function");

	vector(instrument) v_i = basis_curve().instruments();
	integer size = v_size(v_i) ;
	vector(number) sprd;

	//if flat_sprd is 0 then we get the quoted spread 	
	if(null(flat_sprd)){
		resize(flat_sprd, size);
		flat_sprd = 0;
	}
	else 
		QL_REQUIRE(v_size(flat_sprd) == size,"invalid vector size (flat spread vector)");

	interpolator 	disc_ip 	= ip_linear();
	rate_type		disc_ip_rt 	= RT_LOG_DF;
					
	//integer freq_sprd 			= ois_freq_flat();
	//integer freq_solve 			= ois_freq_sprd(); 
	//day_count_method dc_sprd 	= ois_dc_flat();
	//day_count_method dc_solve 	= ois_dc_sprd();

	disc_func df_fx_sprd, df_fx_solve;
	if(null(df_disc_flat_)){
		QL_REQUIRE(!null(df_disc_sprd_),"invalid disc_func (either flat or spread (base) leg disc_func must be input)");
		df_fx_sprd  = adj_disc_df_flat(disc_ip,disc_ip_rt) ;
		df_fx_solve	= df_disc_sprd_;
	}
	else {
		df_fx_sprd  = df_disc_flat_;
		df_fx_solve	= adj_disc_df_sprd(disc_ip,disc_ip_rt) ;
	}

	for(integer i = 0; i < size ; i++){
	
		number s ;/*= ..flt_sprd_fx_lib(d_,v_i[i].issue_date(),v_i[i].maturity(),day(v_i[i].issue_date()),
									freq_sprd,freq_solve,dc_sprd,dc_solve,
									flt_bd(),flt_cal(),flt_eom(),						
									f_fwd_sprd,f_fwd_solve,
									df_fx_sprd,df_fx_solve,
									flat_sprd[i]);*/
		push_back(sprd,s);
	}
	return sprd;
}

//-------------------------------------------------------
// solve_spread_flat_leg
//-------------------------------------------------------
vector(number) swap_curve_bc_ois.solve_spread_flat_leg(vector(number) option(nullable) fx_lib_sprd)
{
	QL_FAIL("work in progress");
	
	vector(instrument) v_i = basis_curve().instruments();
	integer size = v_size(v_i) ;
	vector(number) sprd;

	//if fx_lib_sprd is 0 then we're switching the basis spread to the flat leg i.e. base->flat and flat->base 
	//if fx_lib_sprd is equal to the quoted basis swap then the spread is zero (i.e. flat) since one disc_func is an fx implied
	if(null(fx_lib_sprd)){
		resize(fx_lib_sprd, size);
		fx_lib_sprd = 0;
	}
	else 
		QL_REQUIRE(v_size(fx_lib_sprd) == size,"invalid vector size (basis spread vector)");

	interpolator 	disc_ip 	= ip_linear();
	rate_type		disc_ip_rt 	= RT_LOG_DF;
			
	integer freq_sprd 			= ois_freq_sprd();
	integer freq_solve 			= ois_freq_flat(); 
	day_count_method dc_sprd 	= ois_dc_sprd();
	day_count_method dc_solve 	= ois_dc_flat();


	disc_func df_fx_sprd, df_fx_solve;
	if(null(df_disc_flat_)){
		QL_REQUIRE(!null(df_disc_sprd_),"invalid disc_func (either flat or spread (base) leg disc_func must be input)");
		df_fx_sprd  = df_disc_sprd_;
		df_fx_solve	= adj_disc_df_flat(disc_ip,disc_ip_rt);
	}
	else {
		
		df_fx_sprd  = adj_disc_df_sprd(disc_ip,disc_ip_rt);
		df_fx_solve	=  df_disc_flat_;
	}

	for(integer i = 0; i < size ; i++){
			
		number s;/* = ..flt_sprd_fx_lib(d_,v_i[i].issue_date(),v_i[i].maturity(),day(v_i[i].issue_date()),
									freq_sprd,freq_solve,dc_sprd,dc_solve,
									flt_bd(),flt_cal(),flt_eom(),						
									f_fwd_sprd,f_fwd_solve,
									df_fx_sprd,df_fx_solve,
									fx_lib_sprd[i]);*/
		push_back(sprd,s);
	}
	return sprd;
}

//-------------------------------------------------------
// create
//-------------------------------------------------------
swap_curve_bc_ois swap_curve_bc_ois(date						d,									
									swap_curve_cbois1_f_parm	p,									
									disc_func option(nullable)	df_disc_flat,
									disc_func option(nullable)	df_disc_sprd,
									disc_func 					df_fwd_flat,
									disc_func 					df_fwd_sprd,
									string option(nullable)		name)
{
	return new swap_curve_bc_ois(d, p,df_disc_flat,df_disc_sprd,df_fwd_flat,df_fwd_sprd,name)	;	
}