/*	
	Dual bootstrap utility functions
	Developed by Algorithmica Research, Magnus Nyström
	
	
	Comments: 
	

*/

option(null: hard);	// Throw error if a null value is used in a non-nullable context

/*-----------------------------------------------------------------------
  Name :    flt_sprd_ois_ois
  Notes :   OBS! only for issue == settle  
			relative inter currency liquidity risk [Fx Ois/Ois Basis swap]		
			sprd leg = flat currency leg (1)
			solve leg = spreaded currency leg (2)
			b2(ois-ois) approx equal to (r1_lib - r1_ois) - (r2_lib - r2_ois) //r = constructed equiv fixed rates
	
			Even if all cross-currency basis spreads were zero, the PVs of OIS-discounted LIBOR legs in different 
			currencies are likely to be different.
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/	
number flt_sprd_ois_ois(date 			trade,
						date			issue_date,//== settle_date
						date			maturity,
						number			roll_day,
						integer 		freq_sprd,
						integer 		freq_solve, 
						day_count_method dc_sprd,
						day_count_method dc_solve,
						bd_convention 	bd,
						calendar		cal,
						logical			eom,						
						fwd_func		f_fwd_sprd,											
						fwd_func		f_fwd_solve,
						disc_func		df_ois_sprd,//domestic ois sprd curr
						disc_func		df_ois_solve)//domestic ois solve curr
option (category: 'Yield Curve/003 Curve Building')
{	    
	number	spread = 0;//consider spread as an input argument		
	//return flt_sprd_b_int(	trade,issue_date,maturity,roll_day,freq_sprd,freq_solve, dc_sprd,dc_solve,bd,cal,eom,f_fwd_sprd,
	//						f_fwd_solve,df_ois_sprd,df_ois_solve,spread);
	vector(date) flt_dates_sprd;
	CORE_INT.swap_pmt_dates(issue_date ,null<date>,null<date>, maturity, roll_day, freq_sprd, bd, cal, eom,flt_dates_sprd);

	vector(date) flt_dates_solve;
	CORE_INT.swap_pmt_dates(issue_date ,null<date>,null<date>, maturity, roll_day, freq_solve, bd, cal, eom,flt_dates_solve);


	number flt_curr_fix_sprd= null<number>;
	number flt_curr_fix_solve= null<number>;				
	return swap_imp_flt_sprd(trade,issue_date,	df_ois_sprd, f_fwd_sprd, /*freq_sprd,*/ spread, flt_curr_fix_sprd, flt_dates_sprd, dc_sprd,
												df_ois_solve,f_fwd_solve,/*freq_solve,*/      	flt_curr_fix_solve,flt_dates_solve,dc_solve);
}

/*-----------------------------------------------------------------------
  Name :    flt_sprd_fx_lib
  Notes :   OBS! only for issue == settle  
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
number flt_sprd_fx_lib(	date 			trade,
						date			issue_date,//== settle_date
						date			maturity,
						number			roll_day,
						integer 		freq_sprd,
						integer 		freq_solve, 
						day_count_method dc_sprd,
						day_count_method dc_solve,
						bd_convention 	bd,
						calendar		cal,
						logical			eom,						
						fwd_func		f_fwd_sprd,											
						fwd_func		f_fwd_solve,
						disc_func		df_fx_sprd,//fxadjusted discfunc or ois df (only one of the discfuncs are fx adjusted)
						disc_func		df_fx_solve,//fxadjusted discfunc or ois df (only one of the discfuncs are fx adjusted)
						number			fx_lib_sprd)//spread on the leg we're not solving for
option (category: 'Yield Curve/003 Curve Building Cross Currency')
{	    		
	//return flt_sprd_b_int(	trade,issue_date,maturity,roll_day,freq_sprd,freq_solve, dc_sprd,dc_solve,bd,cal,eom,f_fwd_sprd,
	//						f_fwd_solve,df_fx_sprd,df_fx_solve,fx_lib_sprd);				
	vector(date) flt_dates_sprd;
	CORE_INT.swap_pmt_dates(issue_date ,null<date>,null<date>, maturity, roll_day, freq_sprd, bd, cal, eom,flt_dates_sprd);

	vector(date) flt_dates_solve;
	CORE_INT.swap_pmt_dates(issue_date ,null<date>,null<date>, maturity, roll_day, freq_solve, bd, cal, eom,flt_dates_solve);

	number flt_curr_fix_sprd= null<number>;
	number flt_curr_fix_solve= null<number>;				
	return swap_imp_flt_sprd(trade,issue_date,	df_fx_sprd, f_fwd_sprd, /*freq_sprd,*/ fx_lib_sprd, flt_curr_fix_sprd, flt_dates_sprd, dc_sprd,
												df_fx_solve,f_fwd_solve,/*freq_solve, */     		flt_curr_fix_solve,flt_dates_solve,dc_solve);
}
/*-----------------------------------------------------------------------
  Name :    flt_sprd_fx_lib_imps
  Notes :   OBS! only for issue == settle 
			b(fx/libor) = b(fx/ois) + b(ois/ois)  
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
number flt_sprd_fx_lib_imp(	date 			trade,
							date			issue_date,//== settle_date
							date			maturity,
							number			roll_day,
							integer 		freq_sprd,
							integer 		freq_solve, 
							day_count_method dc_sprd,
							day_count_method dc_solve,
							bd_convention 	bd,
							calendar		cal,
							logical			eom,						
							fwd_func		f_fwd_sprd,											
							fwd_func		f_fwd_solve,								
							disc_func		df_ois_sprd,//domestic ois
							disc_func		df_ois_solve,//domestic ois	
							number 			fx_ois_bswap)//basis swap fx-ois of the solve leg
option (category: 'Yield Curve/003 Curve Building Cross Currency')
{	    
	number bs_ois_ois = flt_sprd_ois_ois(	trade,issue_date,maturity,roll_day,freq_sprd,freq_solve,dc_sprd,dc_solve,
											bd,cal,eom,f_fwd_sprd,f_fwd_solve,df_ois_sprd,df_ois_solve);
	return fx_ois_bswap + bs_ois_ois;				
}

/*-----------------------------------------------------------------------
  Name :    flt_sprd_fx_ois_imp
  Notes :   OBS! only for issue == settle 
			the OIS FX basis spreads can be approximated by
			b(fx/ois) = b(fx/libor) - b(ois/ois)   
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/	
number flt_sprd_fx_ois_imp(	date 			trade,
							date			issue_date,//== settle_date
							date			maturity,
							number			roll_day,
							integer 		freq_sprd,
							integer 		freq_solve, 
							day_count_method dc_sprd,
							day_count_method dc_solve,
							bd_convention 	bd,
							calendar		cal,
							logical			eom,						
							fwd_func		f_fwd_sprd,											
							fwd_func		f_fwd_solve,								
							disc_func		df_ois_sprd,//domestic ois
							disc_func		df_ois_solve,//domestic ois							
							number 			fx_lib_bswap)//basis swap fx-lib of the solve leg
option (category: 'Yield Curve/003 Curve Building Cross Currency')
{	    
	number bs_ois_ois = flt_sprd_ois_ois(	trade,issue_date,maturity,roll_day,freq_sprd,freq_solve,dc_sprd,dc_solve,
											bd,cal,eom,f_fwd_sprd,f_fwd_solve,df_ois_sprd,df_ois_solve);
	return fx_lib_bswap - bs_ois_ois;				
}

/*-----------------------------------------------------------------------
  Name :    flt_sprd_lib_ois
  Notes :   OBS! only for issue == settle 
			same currency: 	libor discounting vs ois discounting [Ois/Libor Basis swap]
			sprd leg = flat leg
			solve leg = spreaded leg
			b(lib-ois) = - (r_lib - r_ois) (r=equiv swaprates)      
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/	
number flt_sprd_lib_ois(date 			trade,
						date			issue_date,//== settle_date
						date			maturity,
						number			roll_day,
						integer 		freq,				
						day_count_method dc,
						bd_convention 	bd,
						calendar		cal,
						logical			eom,						
						fwd_func		f_fwd,											
						disc_func		df_ois)
option (category: 'Yield Curve/003 Curve Building')
{	 
	vector(date) flt_dates;
	CORE_INT.swap_pmt_dates(issue_date ,null<date>,null<date>, maturity, roll_day, freq, bd, cal, eom,flt_dates);
   
	number pv 				= 1;
	number flt_curr_fix 	= null<number>;
	logical incl_not_end 	= true;
	logical incl_not_beg 	= false;				
	return swap_imp_flt_sprd(trade,issue_date,	df_ois, f_fwd, /*freq,*/ flt_curr_fix, flt_dates, dc,pv,incl_not_beg, incl_not_end);		
}	

/*-----------------------------------------------------------------------
  Name :    flt_sprd_conv
  Notes :   OBS! only for issue == settle       
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
number flt_sprd_conv(	date 			trade,
						date			issue_date,//== settle date
						date			maturity,
						number			roll_day,
						integer 		freq_sprd,
						integer 		freq_solve, 
						day_count_method dc_sprd,
						day_count_method dc_solve,
						bd_convention 	bd,
						calendar		cal,
						logical			eom,							
						disc_func		df_disc_sprd,
						disc_func		df_disc_solve,
						number			spread)
option (category: 'Yield Curve/003 Curve Building')
{	    
	if(spread == 0) return 0;

	vector(date) flt_dates_sprd;
	CORE_INT.swap_pmt_dates(issue_date ,null<date>,null<date>, maturity, roll_day, freq_sprd, bd, cal, eom,flt_dates_sprd);
	integer n_sprd = v_size(flt_dates_sprd);

	vector(date) flt_dates_solve;
	CORE_INT.swap_pmt_dates(issue_date ,null<date>,null<date>, maturity, roll_day, freq_solve, bd, cal, eom,flt_dates_solve);
	integer n_solve = v_size(flt_dates_solve);
	
	vector(number) per_pay_sprd;
	resize(per_pay_sprd, n_sprd-1);	
	number df_s_sprd = df_disc_sprd.disc((issue_date - trade) / 365);
	number pv = 0;
	for(integer i = 0; i<n_sprd-1; i++) {
		per_pay_sprd[i] = period (flt_dates_sprd[i] , flt_dates_sprd[i+1] , dc_sprd );
		number cf 		= spread * per_pay_sprd[i];
		number df 		= df_disc_sprd.disc((flt_dates_sprd[i+1] - trade) / 365)/df_s_sprd;
		pv 				+= cf * df;
	}

	vector(number) per_pay_solve;
	resize(per_pay_solve, n_solve-1);
	number df_s_solve = df_disc_solve.disc((issue_date - trade) / 365);
	number adf = 0;
	for(integer i = 0; i<n_solve-1; i++) {
		per_pay_solve[i] = period (flt_dates_solve[i] , flt_dates_solve[i+1] , dc_solve );	
		number df 		= df_disc_solve.disc((flt_dates_solve[i+1] - trade) / 365)/df_s_solve;
		adf 			+= per_pay_solve[i] * df;
	}

	return -pv/adf;	

	/*bd_convention 	idx_bd_base = bd;
	calendar		idx_cal_base = cal;
	logical			idx_eom_base = eom;		
	logical ip_idx_stub = false;
	date out_first,out_last_reg;
	number out_roll_day;
	logical is_reg_first,is_reg_last,eom_roll;
	vector(date) float_idx0_sprd, float_pmt_sprd, float_idx1_sprd;
	fiws.swap_flt_dates(issue_date,null<date>,null<date>, maturity,roll_day,freq_sprd,bd,cal,eom,idx_bd_base,
						idx_cal_base,idx_eom_base,ip_idx_stub,out_first, out_last_reg,out_roll_day,is_reg_first,is_reg_last,eom_roll,
						float_idx0_sprd,float_pmt_sprd,float_idx1_sprd);

	integer n_sprd= v_size(float_idx0_sprd);
	
	bd_convention 	idx_bd_solve = bd;
	calendar		idx_cal_solve= cal;
	logical			idx_eom_solve= eom;
	vector(date) float_idx0_solve, float_pmt_solve, float_idx1_solve;
	fiws.swap_flt_dates(issue_date,null<date>,null<date>, maturity,roll_day,freq_solve,bd,cal,eom,idx_bd_solve,
						idx_cal_solve,idx_eom_solve,ip_idx_stub,out_first, out_last_reg,out_roll_day,is_reg_first,is_reg_last,eom_roll,
						float_idx0_solve,float_pmt_solve,float_idx1_solve);
		
	vector(number) per_pay_sprd = period (float_idx0_sprd , float_pmt_sprd , dc_sprd );			
	number df_s_sprd 			= df_disc_sprd.disc((issue_date - trade) / 365);	
	vector(number) cf 			= spread .* per_pay_sprd;
	vector(number) df 			= df_disc_sprd.disc((float_pmt_sprd - trade) / 365)/df_s_sprd;
	number pv 					= cf * df;
	
	vector(number) per_pay_solve = period (float_idx0_solve , float_pmt_solve , dc_solve );	
	number df_s_solve 			= df_disc_solve.disc((issue_date - trade) / 365);
	df 							= df_disc_solve.disc((float_pmt_solve - trade) / 365)/df_s_solve;
	number adf 					= per_pay_solve * df;
		
    return -pv/adf;	*/			
}
 



module CORE_INT
{	
	//-------------------------------------------------------
	// date_code_str_to_fwd_func_tenor
	//-------------------------------------------------------
	fwd_func_tenor date_code_str_to_fwd_func_tenor(string option(nullable) mc)
	{
		if(null(mc))
			return null<fwd_func_tenor>;

		if(equali(mc,"BD1") || equali(mc,"1D"))
			return fwd_func_tenor.FT_1D;
		
		number p = str_find(mc, "_") ;
		if(!null(p) && p > 0)
			mc = sub_string_start (mc, integer(p));

		p = str_find(mc, "eom") ;
		if(!null(p) && p > 0)
			mc = sub_string_start (mc, integer(p));

		integer size = strlen(mc);
		
		if(size < 2 || size > 3 )
			return null<fwd_func_tenor>;
		
		string first = str_to_upper(sub_string(mc, 0, 1)) ;
		
		if(is_digit(first, 0)){

			string second = str_to_upper(sub_string(mc, 1, 1)) ;			
			if(is_digit(second, 0)  && size == 3) {
				string third = str_to_upper(sub_string(mc, 2, 1)) ;			
				return first == "1" && second == "2" && third == "M"  ? fwd_func_tenor.FT_12M : null<fwd_func_tenor>;				
			}
			else if(size != 2) {
				return null<fwd_func_tenor>;
			}
			else if(second == "M"){
				if (first == "1"){
					return fwd_func_tenor.FT_1M;
				}
				else if (first == "3"){
					return fwd_func_tenor.FT_3M;
				}
				else if (first == "6"){
					return fwd_func_tenor.FT_6M;
				}
			}
			else if(second == "Y"){
				return first == "1" ? fwd_func_tenor.FT_12M : null<fwd_func_tenor>;				
			}
		}
		
		string first_dig = sub_string(mc, 1, 1);
		
		if(!is_digit(first_dig, 0))
			return null<fwd_func_tenor>;

		string second_dig = size > 2 ? sub_string(mc, 2, 1) : "";
		
		if(first == "M"){
			
			if(second_dig == "2" && first_dig == "1"){
				return fwd_func_tenor.FT_12M;
			}
			else if(size >= 3){
				return null<fwd_func_tenor>;
			}
			else if (first_dig == "1"){
				return fwd_func_tenor.FT_1M;
			}
			else if (first_dig == "3"){
				return fwd_func_tenor.FT_3M;
			}
			else if (first_dig == "6"){
				return fwd_func_tenor.FT_6M;
			}
		}
		else if(first == "Y" && first_dig == "1" && size == 2) {
			return fwd_func_tenor.FT_12M;	
		}
		
		return null<fwd_func_tenor>;
	}
}


module CORE_INT
{		
	
	//-------------------------------------------------------
	// df_tenor_spread
	//-------------------------------------------------------
	disc_func_spr df_tenor_spread(	date 			d,						
									disc_func 		df_tenor_base,
									date			synt_end,
									curve option(nullable) short_crv,
									curve option(nullable) fra_crv,
									curve option(nullable) swap_crv,
									curve_prio 		prio1,
									curve_prio		prio2,
									logical			merge_fra, 
									integer			blend_buf_days,
									out curve option(nullable)		crv)
	
	{	    	
		//vector(date) knot;	
		//curve short_adj_c = short_crv;
		//curve swap_adj_c = swap_crv;
		//if(prio1==curve_prio.MIDDLE || (prio2 == curve_prio.MIDDLE && prio1 == curve_prio.SHORT))
			//blend_adj_long(short_crv,fra_crv,swap_crv,prio1,prio2,blend_buf_days,short_adj_c,swap_adj_c);
		
		//crv = blend_curve(short_adj_c,fra_crv,swap_adj_c,prio1,prio2,merge_fra,false,knot,blend_buf_days);
		//swap_crv also as short (if firstfra > swap)
		logical merge_crv = false;
		logical no_overlap_fwd = false;
		crv = CORE_INT.blender_adj_long_curve(short_crv,fra_crv,swap_crv, prio1,prio2,merge_fra, merge_crv, no_overlap_fwd,blend_buf_days);		
		crv = crv.chop_long(synt_end);
		return bootstrap_spr(df_tenor_base, crv, ip_linear(0,0), RT_CONT);	
	}
}

module CORE_INT_COMPAT
{	
	//-------------------------------------------------------
	// create_synt_depo_tenor_crv
	// create synthetic depos for tenor
	//-------------------------------------------------------
	curve create_synt_depo_tenor_crv(	string				class_name,
										date 				d,
										number				tenor,
										date				cut_off,
										disc_func_spr		df)
	{	
		instr_def i 		= instr_def(class_name);
		calendar cal		= i.calendar();
		day_count_method dc = i.accr_dc_method();
		date_code sc		= date_code(i.settle_code());
		date settle 		= sc.apply_fwd(d, cal);
		logical eom			= true;
		bd_convention bd 	= BD_MOD_FOLLOWING;
	
		vector(instrument) depo;
		string prefix_name = strcat(class_name,'-');
		
		cut_off -= 5;
			
		date mat_depo 	= move_month(settle,tenor,eom,cal, bd, false);
		if(mat_depo >= cut_off)
			return null<curve>;  	
		number r_depo 	= df.simple_rate(d, settle, mat_depo, dc);
		string name		= strcat([prefix_name,"synt",str(tenor),"M"]);
	
		instr_def i_d 	= instr_def(class_name);
		instrument depo_i = deposit(i_d, name, d,settle, settle, mat_depo, null<string>,r_depo, r_depo);

		push_back(depo, depo_i);//create_depo_instr(i_d, name, d,settle, settle, mat_depo, r_depo, r_depo));
	
		//push_back(depo, create_depo(class_name, name,d, settle , mat_depo ,r_depo, settle ));
			
		return curve(depo, strcat("syntdepo-",class_name));
	}
	
	//-------------------------------------------------------
	// create_synt_depo_tenor_crv
	// create synthetic depos for tenor
	// TODO: consider extrapolating the basis spread to shorter maturities
	//-------------------------------------------------------
	//seems not used --> will comment out for now
	/*curve create_synt_depo_tenor_crv(	string				class_name,
										date 				d,
										number				tenor,
										disc_func			df_base,
										fwd_func			f_basis)
	{	
		instr_def i 		= instr_def(class_name);
		calendar cal		= i.calendar();
		day_count_method dc = i.accr_dc_method();
		date_code sc		= date_code(i.settle_code());
		date settle 		= sc.apply_fwd(d, cal);
		logical eom			= true;
		bd_convention bd 	= BD_MOD_FOLLOWING;
	
		vector(instrument) depo;
		string prefix_name = strcat(class_name,'-');
			
		date mat_depo 	= move_month(settle,tenor,eom,cal, bd);
		number basis	= f_basis.fwd((settle - d)/365);
		number r_depo 	= df_base.simple_rate(d, settle, mat_depo, dc) + basis;
		string name		= strcat([prefix_name,"synt",str(tenor),"M"]);
	
		instr_def i_d = instr_def(class_name);
		push_back(depo, create_depo_instr(i_d, name, d,settle, settle, mat_depo, r_depo, r_depo));
		//push_back(depo, create_depo(class_name, name,d, settle , mat_depo ,r_depo, settle ));
			
		return curve(depo, strcat("syntdepo-",class_name));
	}*/
	//-------------------------------------------------------
	// create_synt_depo_crv 
	// create synthetic depos for all tenors up to tenor
	// TODO: consider extrapolating the basis spread to shorter maturities
	//-------------------------------------------------------
	//seems not used --> will comment out for now
	/*curve create_synt_depo_crv(	string				class_name,
								date 				d,
								number				tenor,
								date				cut_off,
								disc_func_spr		df)
	{	
		instr_def id 		= instr_def(class_name);
		calendar cal		= id.calendar();
		day_count_method dc = id.accr_dc_method();
		date_code sc		= date_code(id.settle_code());
		date settle 		= sc.apply_fwd(d, cal);
		logical eom			= true;
		bd_convention bd 	= BD_MOD_FOLLOWING;
	
		vector(instrument) depo;
		string prefix_name = strcat(class_name,'-');		
	
		date mat_on 		= cal.move_bus_days(d,1);
		if(mat_on >= cut_off) return null<curve>; 
		number per			= 0;
		number r_depo 		= df.simple_rate(d, d, mat_on, dc);
	
		instr_def i_d 		= instr_def(class_name);
		instrument i 		= create_depo_instr(i_d, strcat(prefix_name,"syntO/N"), d,d, d, mat_on, r_depo, r_depo);
	
		//instrument i 		= create_depo(class_name, strcat(prefix_name,"syntO/N"),d, d, mat_on ,r_depo, d );
		//i 					= i.set_settle_date(d);
		push_back(depo, i);
	
		date mat_tn 		= cal.move_bus_days(mat_on,1);
		if(mat_on >= cut_off) return curve(depo); 
		r_depo 				= df.simple_rate(d, mat_on, mat_tn, dc) ;
		i 					= create_depo_instr(i_d, strcat(prefix_name,"syntT/N"), d,mat_on, mat_on, mat_tn, r_depo, r_depo);
		//i 					= create_depo(class_name, strcat(prefix_name,"syntT/N"),d, mat_on, mat_tn ,r_depo, mat_on );
		//i 					= i.set_settle_date(mat_on);
		push_back(depo, i);	
		
		date mat_depo 		= cal.adjust_date (settle+7, BD_FOLLOWING);
		if(mat_depo >= cut_off) return curve(depo); 
		r_depo 				= df.simple_rate(d, settle, mat_depo, dc);
	
		push_back(depo, create_depo_instr(i_d, strcat(prefix_name,"synt1W"), d,settle , settle , mat_depo, r_depo, r_depo));
		//push_back(depo, create_depo(class_name, strcat(prefix_name,"synt1W"),d, settle , mat_depo ,r_depo, settle ));
		
		mat_depo 	= cal.adjust_date (settle+14, BD_FOLLOWING);
		if(mat_depo >= cut_off) return curve(depo); 
		r_depo 		= df.simple_rate(d, settle, mat_depo, dc) ;
		push_back(depo, create_depo_instr(i_d, strcat(prefix_name,"synt2W"), d,settle , settle , mat_depo, r_depo, r_depo));
		//push_back(depo, create_depo(class_name, strcat(prefix_name,"synt2W"),d, settle , mat_depo ,r_depo, settle ));
		
		cut_off -= 5;
		for(integer i = 1; i < tenor+1; i++){
			mat_depo 	= move_month(settle,i,eom,cal, bd);
			if(mat_depo >= cut_off) return curve(depo); 
			r_depo 		= df.simple_rate(d, settle, mat_depo, dc);
			string name	= strcat([prefix_name,"synt",str(i),"M"]);
			push_back(depo, create_depo_instr(i_d, name, d,settle , settle , mat_depo, r_depo, r_depo));
			//push_back(depo, create_depo(class_name, name,d, settle , mat_depo ,r_depo, settle ));
		}
		
		return curve(depo, strcat("syntdepo-",class_name));
	}
	*/
	//-------------------------------------------------------
	// create_synt_depo_crv 
	// create synthetic depos for all tenors up to tenor
	// TODO: consider extrapolating the basis spread to shorter maturities
	//-------------------------------------------------------
	//seems not used --> will comment out for now
	/*curve create_synt_depo_crv(	string				class_name,
								date 				d,
								number				tenor,
								disc_func			df_base,
								fwd_func			f_basis)
	{	
		instr_def id 		= instr_def(class_name);
		calendar cal		= id.calendar();
		day_count_method dc = id.accr_dc_method();
		date_code sc		= date_code(id.settle_code());
		date settle 		= sc.apply_fwd(d, cal);
		logical eom			= true;
		bd_convention bd 	= BD_MOD_FOLLOWING;
	
		vector(instrument) depo;
		string prefix_name = strcat(class_name,'-');
		
		date mat_on 		= cal.move_bus_days(d,1);
		number per			= 0;
		number basis		= f_basis.fwd(per);
		number r_depo 		= df_base.simple_rate(d, d, mat_on, dc) + basis;
		instr_def i_d 		= instr_def(class_name);
		instrument i 		= create_depo_instr(i_d, strcat(prefix_name,"syntO/N"), d,d, d, mat_on, r_depo, r_depo);
		//instrument i		= create_depo(class_name, strcat(prefix_name,"syntO/N"),d, d, mat_on ,r_depo, d );
		//i 					= i.set_settle_date(d);
		push_back(depo, i);
	
		date mat_tn 		= cal.move_bus_days(mat_on,1);
		per					= (mat_on - d)/365;
		basis				= f_basis.fwd(per);
		r_depo 				= df_base.simple_rate(d, mat_on, mat_tn, dc) + basis;
		i 					= create_depo_instr(i_d, strcat(prefix_name,"syntT/N"), d,mat_on, mat_on, mat_tn, r_depo, r_depo);
		//i 					= create_depo(class_name, strcat(prefix_name,"syntT/N"),d, mat_on, mat_tn ,r_depo, mat_on );
		//i 					= i.set_settle_date(mat_on);
		push_back(depo, i);	
		
		date mat_depo 		= cal.adjust_date (settle+7, BD_FOLLOWING);
		per					= (settle - d)/365;
		basis				= f_basis.fwd(per);
		r_depo 				= df_base.simple_rate(d, settle, mat_depo, dc) + basis;
	
		push_back(depo, create_depo_instr(i_d, strcat(prefix_name,"synt1W"), d,settle , settle , mat_depo, r_depo, r_depo));
		//push_back(depo, create_depo(class_name, strcat(prefix_name,"synt1W"),d, settle , mat_depo ,r_depo, settle ));
		
		mat_depo 	= cal.adjust_date (settle+14, BD_FOLLOWING);
		basis		= f_basis.fwd(per);
		r_depo 		= df_base.simple_rate(d, settle, mat_depo, dc) + basis;
		push_back(depo, create_depo_instr(i_d, strcat(prefix_name,"synt2W"), d,settle , settle , mat_depo, r_depo, r_depo));
		//push_back(depo, create_depo(class_name, strcat(prefix_name,"synt2W"),d, settle , mat_depo ,r_depo, settle ));
		
		for(integer i = 1; i < tenor+1; i++){
			mat_depo 	= move_month(settle,i,eom,cal, bd);
			basis		= f_basis.fwd(per);
			r_depo 		= df_base.simple_rate(d, settle, mat_depo, dc) + basis;
			string name	= strcat([prefix_name,"synt",str(i),"M"]);
			push_back(depo, create_depo_instr(i_d, name, d,settle , settle , mat_depo, r_depo, r_depo));
			//push_back(depo, create_depo(class_name, name,d, settle , mat_depo ,r_depo, settle ));
		}
		
		return curve(depo, strcat("syntdepo-",class_name));
	}
	*/
	//-------------------------------------------------------
	// create_synt_fra_crv 
	// create synthetic fras for all tenors up to first inst
	// TODO: consider extrapolating the basis spread to shorter maturities
	//-------------------------------------------------------
	curve create_synt_fra_crv(	date 				d,
								number				tenor,
								date				cut_off,
								string				fra_class_name,
								string				spot_settle_code,
								disc_func_spr		df)
	{	
		instr_def class_fra = instr_def(fra_class_name);
		calendar cal		= class_fra.calendar();
		day_count_method dc = class_fra.accr_dc_method();
		date_code sc		= date_code(spot_settle_code);
		date spot_settle 	= sc.apply_fwd(d, cal);
		logical eom			= true;
		bd_convention bd 	= BD_MOD_FOLLOWING;
		vector(instrument) fra;
		
		string prefix_name = strcat(fra_class_name,'-');
		string suffix_name ;
	
		cut_off -= 5;
		date fra_mat;
		number r_fra;
	
		date fra_settle = cal.adjust_date (spot_settle+7, BD_FOLLOWING);	
		if(fra_settle < cut_off){
			fra_mat 	= move_month(fra_settle,tenor,eom,cal, bd, false);
			r_fra 		= df.simple_rate(d, fra_settle, fra_mat, dc) ;
			suffix_name = strcat([" x ", "1W+",str(tenor),"M"])	;	
			//push_back(fra, fiws.create_fra(fra_class_name, strcat([prefix_name,"synt1W", suffix_name]),d, fra_settle , fra_mat ,r_fra));
			instrument f = ..fra(instr_def(fra_class_name), strcat([prefix_name,"synt1W", suffix_name]), d, fra_settle , fra_mat,r_fra,null<number>,null<number>);
			push_back(fra, f);//create_fra_instr(instr_def(fra_class_name), strcat([prefix_name,"synt1W", suffix_name]), d, fra_settle , fra_mat,r_fra,r_fra));
		}
		else
			return !null(fra) ? curve(fra) : null<curve>;
	
		fra_settle = cal.adjust_date (spot_settle+14, BD_FOLLOWING);	
		if(fra_settle < cut_off){
			fra_mat 	= move_month(fra_settle,tenor,eom,cal, bd, false);
			r_fra 		= df.simple_rate(d, fra_settle, fra_mat, dc) ;	
			suffix_name = strcat([" x ", "2W+",str(tenor),"M"])	;	
			//push_back(fra, fiws.create_fra(fra_class_name, strcat([prefix_name,"synt2W", suffix_name]),d, fra_settle , fra_mat ,r_fra));
			instrument f = ..fra(instr_def(fra_class_name), strcat([prefix_name,"synt2W", suffix_name]), d, fra_settle , fra_mat,r_fra,null<number>,null<number>);
			push_back(fra, f);//create_fra_instr(instr_def(fra_class_name), strcat([prefix_name,"synt2W", suffix_name]), d, fra_settle , fra_mat,r_fra,r_fra));
		}
		else
			return curve(fra);
	
		number j = 1;
		fra_settle = move_month(spot_settle,j,eom,cal, bd, false);	
		while(fra_settle < cut_off) {
			fra_mat 	= move_month(fra_settle,tenor,eom,cal, bd, false);
			r_fra 		= df.simple_rate(d, fra_settle, fra_mat, dc) ;	
			suffix_name = strcat([" x ", str(j+tenor),"M"]);	
			//push_back(fra, fiws.create_fra(fra_class_name, strcat([prefix_name,"synt", str(j),"M", suffix_name]),d, fra_settle , fra_mat ,r_fra));
			instrument f = ..fra(instr_def(fra_class_name), strcat([prefix_name,"synt", str(j),"M", suffix_name]), d, fra_settle , fra_mat,r_fra,null<number>,null<number>);
			push_back(fra, f);//create_fra_instr(instr_def(fra_class_name), strcat([prefix_name,"synt", str(j),"M", suffix_name]), d, fra_settle , fra_mat,r_fra,r_fra));
			j++;
			fra_settle = move_month(spot_settle,j,eom,cal, bd, false);	
		}
		return curve(fra, strcat("syntfra-",fra_class_name));
	}
	//-------------------------------------------------------
	// create_synt_fra_crv 
	// create synthetic fras for all tenors up to first fra
	// TODO: consider extrapolating the basis spread to shorter maturities
	//-------------------------------------------------------
	//seems not used --> will comment out for now
	/*curve create_synt_fra_crv(	date 				d,
								number				tenor,
								date				cut_off,
								string				fra_class_name,
								string				spot_settle_code,
								disc_func			df_base,
								fwd_func			f_basis)
	{		
		instr_def class_fra = instr_def(fra_class_name);
		calendar cal		= class_fra.calendar();
		day_count_method dc = class_fra.accr_dc_method();
		date_code sc		= date_code(spot_settle_code);
		date spot_settle 	= sc.apply_fwd(d, cal);
		logical eom			= true;
		bd_convention bd 	= BD_MOD_FOLLOWING;
		vector(instrument) fra;	
		string prefix_name = strcat(fra_class_name,'-');
		string suffix_name ;
	
		date fra_mat;
		number per, basis, r_fra;
	
		//date fra_settle = cal.move_bus_days(d,1);	
		//if(fra_settle < cut_off){
		//	fra_mat 	= fiws.move_month(fra_settle,tenor,eom,cal, bd);
		//	per			= (fra_settle - d)/365;		
		//	basis		= f_basis.fwd(per);
		//	r_fra 		= df_base.simple_rate(d, fra_settle, fra_mat, dc) + basis;
		//	suffix_name = strcat([" x ", "T/N+",str(tenor),"M"])	;	
		//	push_back(fra, fiws.create_fra(class_name, strcat([prefix_name,"syntT/N", suffix_name]),d, fra_settle , fra_mat ,r_fra));
		//}
	
		date fra_settle = cal.adjust_date (spot_settle+7, BD_FOLLOWING);	
		if(fra_settle < cut_off){
			fra_mat 	= move_month(fra_settle,tenor,eom,cal, bd);
			per			= (fra_settle - d)/365;
			basis		= f_basis.fwd(per);
			r_fra 		= df_base.simple_rate(d, fra_settle, fra_mat, dc) + basis;
			suffix_name = strcat([" x ", "1W+",str(tenor),"M"])	;	
			//push_back(fra, fiws.create_fra(fra_class_name, strcat([prefix_name,"synt1W", suffix_name]),d, fra_settle , fra_mat ,r_fra));
			push_back(fra, create_fra_instr(instr_def(fra_class_name), strcat([prefix_name,"synt1W", suffix_name]), d, fra_settle , fra_mat,r_fra,r_fra));
		}
		else
			return curve(fra);
	
		fra_settle = cal.adjust_date (spot_settle+14, BD_FOLLOWING);	
		if(fra_settle < cut_off){
			fra_mat 	= move_month(fra_settle,tenor,eom,cal, bd);
			per			= (fra_settle - d)/365;
			basis		= f_basis.fwd(per);
			r_fra 		= df_base.simple_rate(d, fra_settle, fra_mat, dc) + basis;	
			suffix_name = strcat([" x ", "2W+",str(tenor),"M"])	;	
			//push_back(fra, fiws.create_fra(fra_class_name, strcat([prefix_name,"synt2W", suffix_name]),d, fra_settle , fra_mat ,r_fra));
			push_back(fra, create_fra_instr(instr_def(fra_class_name), strcat([prefix_name,"synt2W", suffix_name]), d, fra_settle , fra_mat,r_fra,r_fra));
		}
		else
			return curve(fra);
	
		number j = 1;
		fra_settle = move_month(spot_settle,j,eom,cal, bd);	
		while(fra_settle < cut_off) {
			fra_mat 	= move_month(fra_settle,tenor,eom,cal, bd);
			per			= (fra_settle - d)/365;
			basis		= f_basis.fwd(per);
			r_fra 		= df_base.simple_rate(d, fra_settle, fra_mat, dc) + basis;	
			suffix_name = strcat([" x ", str(j+tenor),"M"]);	
			//push_back(fra, fiws.create_fra(fra_class_name, strcat([prefix_name,"synt", str(j),"M", suffix_name]),d, fra_settle , fra_mat ,r_fra));
			push_back(fra, create_fra_instr(instr_def(fra_class_name), strcat([prefix_name,"synt", str(j),"M", suffix_name]), d, fra_settle , fra_mat,r_fra,r_fra));
			j++;
			fra_settle = move_month(spot_settle,j,eom,cal, bd);	
		}
		return curve(fra, strcat("syntfra-",fra_class_name));
	}
	*/
	//-------------------------------------------------------
	// extend_ois
	//-------------------------------------------------------
	vector(instrument) extend_ois( 	date 						d,
									curve						ois,
									curve 						base,							
									string 						class_name,
									disc_z_model				base_model)
	{	
		string err_msg;
		QL_REQUIRE(CORE_INT.check_curve(ois, true, err_msg), err_msg);	
		QL_REQUIRE(CORE_INT.check_curve(base, true, err_msg),err_msg);
	    
		vector(instrument) v_ois = ois.instruments();
		vector(instrument) v_base = base.instruments();	
		
		date max_ois = v_max(v_ois.maturity());
		date max_base = v_max(v_base.maturity());
		if(max_ois >= max_base) return null<vector(instrument)>;
	
		//disc_func df_base = bootstrap(base, base_ip, base_ip_rt);
		QL_REQUIRE(base_model.type() != disc_z_model_type.DZ_ON_STEP && base_model.type() != disc_z_model_type.DZ_ON_STEP_BOOT,'inapplicable base model (step models are used only for the ois curve)');
		
		disc_func df_base = base_model.df(base, null<vector(number)>);  //calc_disc_df(d,base,null<calendar>,base_model,base_ip,base_ip_rt,null<vector(date)>, null<date>);
		QL_REQUIRE(!null(df_base),'error extending discounting curve (base curve)');	
		disc_func_spr df_basis = bootstrap_spr(df_base, ois, ip_linear(0, 0), RT_CONT);
		QL_REQUIRE(!null(df_basis),'error extending discounting curve (spread curve)');	
	
		vector(instrument) inst;
		instrument i;
		try {
			i = CORE_INT.create_class_depo(class_name,d);//fake inst
		}
		catch {
			i = null<instrument>;
		}
	
		if(null(i))
			i = CORE_INT.create_class_fra(class_name, d);//fake inst
		
		QL_REQUIRE(!null(i),'error extending discounting curve (creation of synthetic ois instrument)');
		
		if(i.is_deposit()) {
			number tenor 		= 12;
			calendar cal		= i.calendar();
			day_count_method dc = i.accr_dc_method();	
			logical eom			= true;
			bd_convention bd 	= BD_MOD_FOLLOWING;
			date_code sc		= date_code(i.settle_code());
			date settle 		= sc.apply_fwd(d, cal);
				
			string prefix_name 	= strcat(class_name,'-');
			date mat 			= move_month(max_ois,tenor,eom,cal, bd, false);
			number j 			= 0;
			instr_def i_d 		= instr_def(class_name);
					
			while(mat <= max_base) {								
				number r_ois	= df_basis.simple_rate(d, settle, mat, dc);
				//number basis	= (r_ois - df_base.simple_rate(d, settle, mat, dc))*10000;	
				if(!null(r_ois)) {
					
					instrument ii = deposit(i_d, strcat([prefix_name,"synt",str(j+1)]), d,settle, settle, mat, null<string>,r_ois, r_ois);
					//instrument ii = create_depo_instr(i_d, strcat([prefix_name,"synt",str(j+1)]), d,settle, settle, mat, r_ois, r_ois);	
					//instrument ii = create_depo(class_name, strcat([prefix_name,"synt",str(j+1)]),d, settle , mat ,r_ois);
					if(!null(ii)) 
						push_back(inst, ii);				
				}
				j++;
				mat = move_month(max_ois,(j+1)*tenor,eom,cal, bd, false);			
			}
		}
		else {
			tenor_code	t 		= i.index_tenor_code();
			number freq_fra		= t.frequency();			
			QL_FAIL_COND(freq_fra < 0 || freq_fra == 365,'invalid fra tenor');
			number tenor 		= 12/freq_fra;
			calendar cal		= i.calendar();
			day_count_method dc = i.accr_dc_method();	
			logical eom			= true;
			bd_convention bd 	= BD_MOD_FOLLOWING;
		
			string prefix_name 	= strcat(class_name,'-');
			
			date fra_settle 	= max_ois;
			date fra_mat 		= move_month(fra_settle,tenor,eom,cal, bd,false);
			instr_def i_d 		= instr_def(class_name);
			number j 			= 0;		
			while(fra_mat <= max_base) {							
				number r_ois_fra 	= df_basis.simple_rate(d, fra_settle, fra_mat, dc);
				//number basis		= (r_ois_fra - df_base.simple_rate(d, fra_settle, fra_mat, dc))*10000;
				if(!null(r_ois_fra)){			
					//instrument ii = fiws.create_fra(class_name, strcat([prefix_name,"synt",str(j+1)]),d, fra_settle , fra_mat ,r_ois_fra);
					instrument ii = ..fra(i_d, strcat([prefix_name,"synt",str(j+1)]), d, fra_settle , fra_mat,r_ois_fra,null<number>,null<number>);
					//instrument ii = create_fra_instr(i_d, strcat([prefix_name,"synt",str(j+1)]), d, fra_settle , fra_mat,r_ois_fra,r_ois_fra);
					if(!null(ii)) 
						push_back(inst, ii);
				}
				j++;	
				fra_settle 	= move_month(max_ois,j*tenor,eom,cal, bd, false);
				fra_mat 	= move_month(fra_settle,tenor,eom,cal, bd, false);						
			}
			
			//test
			/*
			date fra_mat 	= max_ois;
			date fra_settle = add_months(fra_mat, -tenor);//fiws.move_month(fra_mat,-tenor,eom,cal, bd);
			number r_ois_fra = df_basis.simple_rate(d, fra_settle, fra_mat, dc);
			number r_b_fra = df_base.simple_rate(d, fra_settle, fra_mat, dc);
			number basis = r_b_fra-r_ois_fra;
	
			fra_settle 	= max_ois;
			fra_mat 	= fiws.move_month(fra_settle,tenor,eom,cal, bd);
			number j = 0;		
			while(fra_mat <= max_base) {							
				r_b_fra 	= df_base.simple_rate(d, fra_settle, fra_mat, dc);
				r_ois_fra	= r_b_fra + basis ;
				//number basis		= (r_ois_fra - df_base.simple_rate(d, fra_settle, fra_mat, dc))*10000;
				if(!null(r_ois_fra)){			
					instrument ii = fiws.create_fra(class_name, strcat([prefix_name,"synt",str(j+1)]),d, fra_settle , fra_mat ,r_ois_fra);
					if(!null(ii)) 
						push_back(inst, ii);
				}
				j++;	
				fra_settle 	= fiws.move_month(max_ois,j*tenor,eom,cal, bd);
				fra_mat 	= fiws.move_month(fra_settle,tenor,eom,cal, bd);						
			}
			*/
		}
		
		return inst;		
	}
	//-------------------------------------------------------
	// extract_depo
	//-------------------------------------------------------
	curve extract_depo(	curve sh_crv)
	{
		vector(instrument) vi;
		vector(instrument) vi_orig = sh_crv.instruments();
		for(integer i = 0; i < v_size(vi_orig); i++) {
			if(vi_orig[i].is_deposit())
				push_back(vi, vi_orig[i] );
		}
		return !null(vi) ? curve(vi) : null<curve>;	
	}
	
	//-------------------------------------------------------
	// extract_tenor_depo
	//-------------------------------------------------------
	curve extract_tenor_depo(	curve option(nullable) sh_crv,
								number tenor_m,
								calendar cal = calendar(),
								logical eom = true,
								bd_convention bd = BD_MOD_FOLLOWING)
	{
		if(null(sh_crv))
			return null<curve>;
	
		number max_day_diff = 5;
	
		sh_crv = extract_depo(sh_crv);
		vector(instrument) vi = sh_crv.instruments();
		if(v_size(vi) == 0)
			return null<curve>;
	
		date settle = vi[v_size(vi)-1].settle_date();
		date mat_approx = move_month(settle, tenor_m, eom , cal, bd, false );	
		vector(number) diff = abs(vi.maturity() - mat_approx);
		number min = v_min(diff);
		if(min > max_day_diff)
			return null<curve>;
		vector(integer) idx = ..v_find(diff, [min]);
		return curve([vi[idx[0]]]);
	}
}

module CORE_INT
{	
	
		
	//-------------------------------------------------------
	// create_synt_fra_crv 
	// create synthetic fras for all tenors up to first inst
	// TODO: consider extrapolating the basis spread to shorter maturities
	//-------------------------------------------------------
	curve create_synt_fra_crv(	date 				d,								
								date				cut_off,								
								ir_index			fra_idx,
								disc_func_spr		df)
	{	
		
		calendar cal		= fra_idx.fixing_calendar();
		day_count_method dc = fra_idx.day_count_method();
		integer spot_days	= fra_idx.spot_days();
		string settle_code	= strcat("BD" , str(spot_days));				
		date_code sc		= date_code(settle_code);
		date spot_settle 	= sc.apply_fwd(d, cal);
		logical eom			= fra_idx.eom();
		bd_convention bd 	= fra_idx.bd_convention();
		string currency 	= fra_idx.currency();
		number _tenor_m		= fra_idx.tenor_months();		
		QL_REQUIRE(!null(_tenor_m),"invalid ir_index");
		integer tenor_m		= integer(_tenor_m);
		string prefix_name 	= "FRAsynt-";
		//instr_def fra_def 	= create_fra_def(dc,bd,eom,fra_idx.tenor(),integer(spot_days),cal,currency);
		instr_def fra_def 	= create_fra_instr_def(dc,bd,eom,null<date>,null<string>,null<date>,null<string>,
													null<logical>,cal,currency,fra_idx.tenor(),integer(spot_days));

		QL_FAIL_COND(null(fra_def),'invalid fra_def (error in create_synt_fra_crv)');
		
		cut_off -= 5;

		vector(instrument) fra;
		date fra_settle = cal.adjust_date (spot_settle+7, BD_FOLLOWING);	
		if(fra_settle < cut_off){
			date fra_mat 		= move_month(fra_settle,tenor_m,eom,cal, bd, false);
			number r_fra 		= df.simple_rate(d, fra_settle, fra_mat, dc) ;
			string suffix_name 	= strcat([" x ", "1W+",str(tenor_m),"M"])	;	
			string name 		= strcat([prefix_name,"1W", suffix_name]);
			instrument i 		= create_par_fra(fra_def,name,d, fra_settle , fra_mat,r_fra);
			push_back(fra, i);
		}
		else
			return !null(fra) ? curve(fra) : null<curve>;
	
		fra_settle = cal.adjust_date (spot_settle+14, BD_FOLLOWING);	
		if(fra_settle < cut_off){
			date fra_mat 		= move_month(fra_settle,tenor_m,eom,cal, bd, false);
			number r_fra 		= df.simple_rate(d, fra_settle, fra_mat, dc) ;	
			string suffix_name 	= strcat([" x ", "2W+",str(tenor_m),"M"])	;
			string name 		= strcat([prefix_name,"2W", suffix_name]);	
			instrument i 		= create_par_fra(fra_def,name,d, fra_settle , fra_mat,r_fra);
			push_back(fra, i);
		}
		else
			return curve(fra);
	
		number j = 1;
		fra_settle = move_month(spot_settle,j,eom,cal, bd, false);	
		while(fra_settle < cut_off) {
			date fra_mat 		= move_month(fra_settle,tenor_m,eom,cal, bd, false);
			number r_fra 		= df.simple_rate(d, fra_settle, fra_mat, dc) ;	
			string suffix_name 	= strcat([" x ", str(j+tenor_m),"M"]);
			string name 		= strcat([prefix_name, str(j),"M", suffix_name]);	
			instrument i 		= create_par_fra(fra_def,name,d, fra_settle , fra_mat,r_fra);
			push_back(fra, i);
			j++;
			fra_settle 			= move_month(spot_settle,j,eom,cal, bd, false);	
		}
		return curve(fra, strcat("syntfra-",currency));
	}		
	//-------------------------------------------------------
	// extend_disc_curve
	//-------------------------------------------------------
	vector(instrument) extend_disc_curve( 	date 						d,
											curve						disc,
											curve 						base,											
											disc_z_model				base_model)
	{	
		string err_msg;
		QL_REQUIRE(CORE_INT.check_curve(disc, true, err_msg), err_msg);	
		QL_REQUIRE(CORE_INT.check_curve(base, true, err_msg),err_msg);
	    
		vector(instrument) v_disc 	= disc.instruments();
		vector(instrument) v_base 	= base.instruments();			
		date max_disc 				= v_max(v_disc.maturity());
		date max_base 				= v_max(v_base.maturity());
		if(max_disc >= max_base) return null<vector(instrument)>;
			
		QL_REQUIRE(base_model.type() != disc_z_model_type.DZ_ON_STEP && base_model.type() != disc_z_model_type.DZ_ON_STEP_BOOT,
					'inapplicable base model (step models are used only for the ois curve)');
		
		disc_func df_base 		= base_model.df(base, null<vector(number)>);  
		QL_REQUIRE(!null(df_base),'error extending discounting curve (base curve)');	
		disc_func_spr df_basis 	= bootstrap_spr(df_base, disc, ip_linear(0, 0), RT_CONT);
		QL_REQUIRE(!null(df_basis),'error extending discounting curve (spread curve)');	
	
		instrument i = disc.instruments()[0];
		vector(instrument) inst;
				
		number tenor 		= 12;
		calendar cal		= i.calendar();
		string currency		= i.currency();
		day_count_method dc = i.accr_dc_method();	
		logical eom			= true;
		bd_convention bd 	= BD_MOD_FOLLOWING;
		string sc_str		= "bd2";
		date_code sc		= date_code(sc_str);
		date settle 		= sc.apply_fwd(d, cal);				
		date mat 			= move_month(max_disc,tenor,eom,cal, bd, false);
		number j 			= 0;
						
		while(mat <= max_base) {								
			number r_ois	= df_basis.simple_rate(d, settle, mat, dc);			
			if(!null(r_ois)) {
				string name = strcat(["synt",str(j+1)]);
				instrument ii = create_par_depo(dc,bd,eom, sc_str, cal, currency, name,d,settle,mat,r_ois);
				if(!null(ii)) push_back(inst, ii);				
			}
			j++;
			mat = move_month(max_disc,(j+1)*tenor,eom,cal, bd, false);			
		}
				
		return inst;		
	}	
	//-------------------------------------------------------
	// check_swap_issue
	// verify spot_settle <= issue_date
	//-------------------------------------------------------
	
	logical check_swap_issue(vector(instrument) vi,
							date 				spot_settle,
							out string name)
	{
		for(integer i = 0; i < v_size(vi); i++){
			if(vi[i].issue_date() < spot_settle){
				name = vi[i].name();
				return false;
			}
		}	
		return true;
		//return check_equal_issue_date(vi, spot_settle, name);
	}
	//-------------------------------------------------------
	// check_bswap_settle
	// verify settle == issue_date and all settle dates are equal
	//-------------------------------------------------------
	
	logical check_bswap_issue(	vector(instrument) 	vi,
								date 				settle,
								out string 			name)
	{
		logical s_e = CORE_INT.check_eq_settle(vi, settle, name);
		if(!s_e)
			return false;
	
		return CORE_INT.check_eq_issue_date(vi, settle, name);
	}
	//-------------------------------------------------------
	// check_settle_vs_issue
	// verify settle == issue_date
	//-------------------------------------------------------
	logical check_settle_vs_issue(	vector(instrument) 	vi,
									date 				spot_settle,
									out string   		name)
	{
		name = "";
		vector(instrument) id[0];
	
		for(integer i = 0; i < v_size(vi); i++){
			if(vi[i].issue_date() != vi[i].settle_date()){
				name = vi[i].name();
				return false;
			}
			if(vi[i].maturity() > spot_settle)//false for on and tn
				push_back(id,vi[i]); 	
			
		}
		if(v_size(id) <= 1)
			return true;

		date idt = null<date>;
		return CORE_INT.check_eq_issue_date(id, idt, name);
	
	}
	
	//-------------------------------------------------------
	// first_non_fd
	//-------------------------------------------------------
	instrument first_non_fd(curve crv)
	{	
		vector(instrument) v_i = crv.instruments();	
		for(integer i=0; i < v_size(v_i); i++){
			if(!v_i[i].is_fra() && !v_i[i].is_ir_future() && !v_i[i].is_deposit())	
				return v_i[i];		
		}
		return null<instrument>;
	}
	
	//-------------------------------------------------------
	// cut_off_fra
	//-------------------------------------------------------
	date cut_off_fra(	date 					d,
						curve option(nullable)	fra_crv, 
						curve 					df_basis_crv,
						integer 				freq,
						integer 				dflt_days)
	{ 
		if(null(fra_crv)) {			
			instrument i = first_non_fd(df_basis_crv);
			if(!null(i))
				return  i.maturity() - integer(floor(365/freq));//approx ok			
			else
				return  d + dflt_days; 	
		}
		else  
			return fra_crv.instruments()[0].settle_date();
	}
	//-------------------------------------------------------
	// cut_off_depo
	//-------------------------------------------------------
	date cut_off_depo(	date 					d,
						curve option(nullable) 	short_crv, 
						curve 					df_basis_crv,
						integer 				dflt_days)
	{	
		if(null(short_crv)) {			
			instrument i = first_non_fd(df_basis_crv);
			if(!null(i))
				return i.maturity();		
			else
				return d + dflt_days ;	
		}
		else  
			return short_crv.instruments()[0].maturity();
	}
	//-------------------------------------------------------
	// check_b2_inst_freq
	//-------------------------------------------------------
	/*void check_b2_inst_freq(curve 	basis, 
							out integer freq_sw_base, 
							out integer freq_sw_flat )
	{
		vector(instrument) bas_in 	= basis.instruments();
		QL_REQUIRE(v_size(bas_in) > 0, "invalid basis swap curve (in)");
		swap_fltflt2s basis_i 		= swap_fltflt2s(bas_in[0]);
	
		string class_name_base 	= basis_i.irs_class_name_base();
		QL_REQUIRE(!null(class_name_base),'invalid class name tag - base leg (b2 fwd)', basis_i.name(), true);	
		swap_fixflt sw_base 		= CORE_INT.create_class_fixflt_par(class_name_base);
		QL_REQUIRE(!null(sw_base),'invalid swap creation - base leg (b2 fwd)',class_name_base,true);

		//integer freq_fix;
		freq_sw_base = sw_base.reset_freq_flt_leg();
		
		QL_REQUIRE(freq_sw_base>0 && freq_sw_base != 365, "invalid float reset frequency (basis-base)");
		QL_REQUIRE(freq_sw_base == basis_i.reset_freq_sprd_leg(), "flt_reset_freq not compatible with swap created from class name");
	
		string class_name_flat = basis_i.irs_class_name_flat();
		QL_REQUIRE(!null(class_name_flat),'invalid class name tag - flat leg (b2 fwd)', basis_i.name(), true);
		swap_fixflt sw_flat = CORE_INT.create_class_fixflt_par(class_name_flat);
		QL_REQUIRE(!null(sw_flat),'invalid swap creation - flat leg (b2 fwd)',class_name_flat,true);		
		freq_sw_flat = sw_flat.reset_freq_flt_leg();
		QL_REQUIRE(freq_sw_flat>0 && freq_sw_flat != 365, "invalid float reset frequency (basis-flat)");
		QL_REQUIRE(freq_sw_flat == basis_i.reset_freq_flat_leg(), "flt_reset_freq_flat not compatible with swap created from class name");	
	}*/
	//-------------------------------------------------------
	// synt_curve_create
	//-------------------------------------------------------
	curve synt_curve_create(date 			d,
							curve option(nullable)	fra_crv,
							ir_index		fra_idx,						
							disc_func_spr 	df_basis,
							curve			df_basis_crv)
	{
		integer null_cut_days 	= 400;		
		number _tenor_m 		= fra_idx.tenor_months();	
		QL_REQUIRE(!null(_tenor_m),"invalid ir_index");
		integer tenor_m			= integer(_tenor_m);
		integer freq			= integer(floor(12/tenor_m));
		date cut_off_f 			= CORE_INT.cut_off_fra(d,fra_crv,df_basis_crv,freq,null_cut_days);
	 		
		//curve fra_ext_crv = CORE_INT.create_synt_fra_crv(d,12/freq, cut_off_f, settle_code, df_basis);	
		return CORE_INT.create_synt_fra_crv(d,cut_off_f, fra_idx, df_basis);			
	}
	//-------------------------------------------------------
	// synt_curve_create
	//-------------------------------------------------------
	curve synt_curve_create(date 			d,
							curve option(nullable)	fra_crv,
							ir_index		fra_idx,						
							disc_func_spr 	df_basis,
							curve			df_basis_crv,
							out curve		synt_curve)
	{	
		synt_curve = synt_curve_create(d,fra_crv,fra_idx,df_basis,df_basis_crv);
					
		if(..check_curve(synt_curve)) {
			curve fra_ext_crv = merge(synt_curve, fra_crv) ;
			string cname = (null(fra_crv)||null(fra_crv.name())) ? synt_curve.name() : strcat(fra_crv.name(),"-ext");
			if(null(cname )) cname  = "";
			return fra_ext_crv.set_name(cname); 			
		}
		else 
			return fra_crv;				
	}
	
	//-------------------------------------------------------
	// extract_tenor_depo
	//-------------------------------------------------------
	curve extract_tenor_depo(	curve option(nullable)	short_crv,
								string 	tenor)
	{
		if(null(short_crv) || short_crv.empty())
			return null<curve>;

		//integer size_all 					= short_crv.n_instruments();		
		vector(instrument) v_instr_all		= short_crv.instruments();

		vector(instrument) 	v_instr, v_instr_rest ;
		extract_instr_type(v_instr_all, [instr_type.DEPOSIT,instr_type.DEPOSIT_INDEX], v_instr, v_instr_rest );
		integer size = v_size(v_instr);
		integer size_rest = null(v_instr_rest) ? 0 : v_size(v_instr_rest);
		
		if(size == 0) {//no depos in curve 
			return size_rest > 0 ? curve(v_instr_rest) : null<curve>;
		}
		
		vector(fwd_func_tenor) mat_tenor = CORE_INT.date_code_str_to_fwd_func_tenor(v_instr.maturity_code());
		fwd_func_tenor t;
		str_to_enum(tenor,t);
		
		for(integer i=0; i<size;i++){
			if(!null(mat_tenor[i]) && t == mat_tenor[i]) {
				if(size_rest > 0){
					push_back(v_instr_rest,v_instr[i]);
					return curve(v_instr_rest);
				}
				else {
					return curve([v_instr[i]]);
				}
			}
		}
		
		return size_rest > 0 ? curve(v_instr_rest) : null<curve>;
	}
	
	//-------------------------------------------------------
	// fwd_init_depo
	//-------------------------------------------------------
	logical fwd_init_depo(	curve option(nullable)	short_crv,
							fwd_func_tenor 			fwd_tenor,
							out curve  				new_short_crv)
	{
		if(null(short_crv) || short_crv.empty()) {
			new_short_crv = null<curve>;
			return false;
		}

		//integer size_all 				= short_crv.n_instruments();		
		vector(instrument) v_instr_all	= short_crv.instruments();

		vector(instrument) 	v_instr, v_instr_rest ;
		extract_instr_type(v_instr_all, [instr_type.DEPOSIT,instr_type.DEPOSIT_INDEX], v_instr, v_instr_rest );
		integer size = v_size(v_instr);
		if(size == 0) {//no depos in curve --> keep as is i.e. no update
			new_short_crv = null<curve>;
			return false;
		}
		
		integer size_rest = null(v_instr_rest) ? 0 : v_size(v_instr_rest);		
		vector(fwd_func_tenor) mat_tenor = CORE_INT.date_code_str_to_fwd_func_tenor(v_instr.maturity_code());
		
		if(size == 1) {
			logical match = !null(mat_tenor[0]) && fwd_tenor == mat_tenor[0];
			if(match) {//exactly 1 depo in curve and it is a match --> keep as is i.e. no update
				new_short_crv = null<curve>;
				return false;
			}
			else {
				if(size_rest > 0){
					new_short_crv = curve(v_instr_rest);
					return true;
				}
				else {
					new_short_crv = null<curve>;
					return true;
				}
			}
			
		}			
		
		for(integer i=0; i<size;i++){
			if(!null(mat_tenor[i]) && fwd_tenor == mat_tenor[i]) {
				if(size_rest > 0){
					push_back(v_instr_rest,v_instr[i]);
					new_short_crv = curve(v_instr_rest);
					return true;
				}
				else {
					new_short_crv = curve([v_instr[i]]);
					return true;
				}
			}
		}

		new_short_crv = null<curve>;
		return true;				

	}
	//-------------------------------------------------------
	// fwd_init
	//-------------------------------------------------------
	logical fwd_init(	date 							d,
					fixing_curve option(nullable)	f_crv,
					fwd_func_tenor 	fwd_tenor,
					string 							fwd_tenor_s,
					out ir_index 					fwd_idx,
					out fixing_curve  				new_f_crv)
	{
		if(null(f_crv) || f_crv.n_instruments() == 0) {
			fwd_idx = null<ir_index>;
			new_f_crv = null<fixing_curve>;
			return false;
		}
		
		integer size 		= f_crv.n_instruments();		
		fwd_idx 			= f_crv.find_ir_index(fwd_tenor_s);
		instrument fix_inst	= f_crv.find_inst(fwd_tenor_s, d);
		QL_REQUIRE(!null(fix_inst), "invalid fixing instrument (not present in fixing curve)");//not allowing nullifying the curve
		if(size != 1){
			new_f_crv 		= fixing_curve([fix_inst], f_crv);
			return true;
		}
		else {//no need to update
			new_f_crv 	= null<fixing_curve> ;
			return false;
		}
	}
	//-------------------------------------------------------
	// b1_init_depo
	// here we should have 2 fixing instruments in the curve
	//-------------------------------------------------------
	logical b1_init_depo(	curve option(nullable)	short_crv,
							fwd_func_tenor 	fwd_tenor_in,
							fwd_func_tenor 	fwd_tenor_out,
							out curve  new_short_crv)
	{

		if(null(short_crv)|| short_crv.empty()) {			
			new_short_crv 	= null<curve> ;
			return false;
		}
			
		//integer size_all 				= short_crv.n_instruments();
		vector(instrument) v_instr_all	= short_crv.instruments();

		vector(instrument) 	v_instr, v_instr_rest ;
		extract_instr_type(v_instr_all, [instr_type.DEPOSIT,instr_type.DEPOSIT_INDEX], v_instr, v_instr_rest );
		integer size = v_size(v_instr);
		if(size == 0) {//no depos in curve --> keep as is i.e. no update
			new_short_crv = null<curve>;
			return false;
		}

		integer size_rest 	= null(v_instr_rest) ? 0 : v_size(v_instr_rest);		
		vector(fwd_func_tenor) mat_tenor = CORE_INT.date_code_str_to_fwd_func_tenor(v_instr.maturity_code());

		integer i_in = -1;
		for(integer i=0; i<size;i++){
			if(!null(mat_tenor[i]) && fwd_tenor_in == mat_tenor[i]) {				
				i_in = i;
				break;
			}
		}
		integer i_out = -1;
		for(integer i=0; i<size;i++){
			if(!null(mat_tenor[i]) && fwd_tenor_out == mat_tenor[i]) {				
				i_out = i;
				break;
			}
		}

		if(i_in != -1 && i_out != -1) {//both found
			if(size == 2){//exactly 2 depos in curve and it is a match --> keep as is i.e. no update
				new_short_crv = null<curve>;
				return false;//no update
			}
			else{
				if(size_rest > 0){
					push_back(v_instr_rest,v_instr[i_in]);
					push_back(v_instr_rest,v_instr[i_out]);
					new_short_crv = curve(v_instr_rest);
					return true;
				}
				else {
					new_short_crv = curve([v_instr[i_in], v_instr[i_out]]);
					return true;
				}
			}		
		}
		else if(i_in != -1 ) {//found fwd_tenor_in only
			if(size == 1){
				new_short_crv = null<curve>;
				return false;//no update
			}
			else {
				if(size_rest > 0){
					push_back(v_instr_rest,v_instr[i_in]);					
					new_short_crv = curve(v_instr_rest);
					return true;
				}
				else {
					new_short_crv = curve([v_instr[i_in]]);
					return true;
				}
			}			
		}
		else if(i_out != -1 ) {//found fwd_tenor_out only
			if(size == 1){
				new_short_crv = null<curve>;
				return false;//no update
			}
			else{
				if(size_rest > 0){
					push_back(v_instr_rest,v_instr[i_out]);					
					new_short_crv = curve(v_instr_rest);
					return true;
				}
				else {
					new_short_crv = curve([v_instr[i_out]]);
					return true;
				}
			}			
		}
		else {//not found, allow nullifying the curve
			if(size_rest > 0){
				new_short_crv = curve(v_instr_rest);
				return true;
			}
			else {
				new_short_crv = null<curve>;
				return true;
			}
		}
	}
	//-------------------------------------------------------
	// b1_init
	// here we should have 2 fixing instruments in the curve
	//-------------------------------------------------------
	logical b1_init(date 			d,
					fixing_curve option(nullable)	f_crv,
					fwd_func_tenor 	fwd_tenor_in,
					fwd_func_tenor 	fwd_tenor_out,
					string  		fwd_tenor_out_s,
					out ir_index  	fwd_idx_out,
					out fixing_curve  new_f_crv)
	{
		//freq_in 			= fwd_func_tenor.to_freq(fwd_tenor_in);	
		//freq_out 			= fwd_func_tenor.to_freq(fwd_tenor_out);
		//fwd_tenor_out_s		= string(fwd_tenor_out);

		if(null(f_crv)|| f_crv.n_instruments() == 0) {
			fwd_idx_out = null<ir_index>;
			new_f_crv 	= null<fixing_curve> ;
			return false;
		}
		
		integer size 			= f_crv.n_instruments();
		string fwd_tenor_in_s	= string(fwd_tenor_in);
		
		fwd_idx_out			= f_crv.find_ir_index(fwd_tenor_out_s);
		instrument fix_inst_in	= f_crv.find_inst(fwd_tenor_in_s, d);
		QL_REQUIRE(!null(fix_inst_in), "invalid fixing instrument (not present in fixing curve)");
		instrument fix_inst_out	= f_crv.find_inst(fwd_tenor_out_s, d);
		QL_REQUIRE(!null(fix_inst_out), "invalid fixing instrument (not present in fixing curve)");
		if(size != 2) {
			new_f_crv = fixing_curve([fix_inst_in, fix_inst_out], f_crv);
			return true;					
		}
		else { 
			new_f_crv = null<fixing_curve> ;
			return false;
		}
	}

	fwd_func bucket_spread_old(fwd_func f,
						   date d,
						   date s,
						   date e,
						   number sprd,
						   logical is_end)
	{
		QL_REQUIRE(s >= d, "invalid wave start date");
		QL_REQUIRE(e >= d, "invalid wave end date");
		QL_REQUIRE(e >= s, "invalid date order");
		
		number psl 	= (s-d)/365.0;		
		number ps 	= (s+1-d)/365.0;
		
		vector(number) p,r ;
		if(d == s && s == e) {
			p = [0,ps];
			r = [sprd,0];
		}		
		else if(is_end) {
			if(psl==0){
				p = [0,ps];
				r = [0, sprd];
			}
			else {
				p = [0,psl,ps];
				r = [0, 0,sprd];
			}
		}
		else {
			QL_REQUIRE(e > s, "invalid date order");
			number pe = (e-d)/365.0;
			number pel = (e+1-d)/365.0;
			if(psl==0){
				if(s+1 == e){
					p = [0,ps, pel];
					r = [0,sprd, 0];
				}
				else {
					p = [0,ps, pe, pel];
					r = [0,sprd, sprd, 0];
				}
			}
			/*else if(s == e){
				p = [0,psl,ps];
				r = [0,0, sprd, 0];
			}*/
			else if(s+1 == e){
				p = [0,psl,ps, pel];
				r = [0, 0,sprd, 0];
			}
			else {
				p = [0,psl,ps, pe, pel];
				r = [0, 0,sprd, sprd, 0];
			}
		}
		
		fwd_func fa = fwd_func_interp(p,r,ip_linear(0,0));
		fwd_func fn = fwd_func_add(f,fa);
		return fn;
	}

	/*
	d 		#2021-02-23
 f_start	
	[0]		#2021-02-25 depo
	[1]		#2021-03-04 synt1
	[2]		#2021-03-11 synt2
	[3]		#2021-03-17 imm1
	[4]		#2021-06-16
	[5]		#2021-09-15
	[6]		#2021-12-15
	[7]		#2022-03-16
	[8]		#2022-06-15
	[9]		#2022-09-21
	[10]	#2022-12-21
	[11]	#2023-03-15
	[12]	#2023-06-21
	[13]	#2023-09-20
	[14]	#2023-12-20 imm12
	[15]	#2024-11-25 swap4y
	[16]	#2025-11-25
	[17]	#2026-11-25
	[18]	#2027-11-25
	[19]	#2028-11-27
	[20]	#2029-11-26
	[21]	#2030-11-25
	[22]	#2032-11-25 swap12y
	[23]	#2035-11-26
	[24]	#2040-11-26 swap20y
 */
	
	void bucket_spread(fwd_func	fwd_base,
					   date d,
					   vector(date) f_first,
					   vector(date) f_end,
					   number sprd,
					   out vector(fwd_func) f)
	{
		integer size 	= v_size(f_end);
		QL_REQUIRE(size > 0, "invalid start date vector");
		QL_REQUIRE(size == v_size(f_first), "invalid first date vector");
		QL_REQUIRE(f_end[0] >= d, "invalid wave start date");
		QL_REQUIRE(f_first[0] >= d, "invalid wave first date");
		QL_REQUIRE(f_first[0] <= f_end[0], "invalid wave first date vector");
		for(integer i=1;i<size;i++){
			QL_REQUIRE(f_end[i] > f_end[i-1], "invalid wave start date vector");
			QL_REQUIRE(f_first[i] > f_first[i-1], "invalid wave first date vector");
			QL_REQUIRE(f_first[i] <= f_end[i], "invalid wave first date vector");
		}
		
		resize(f,size);
		vector(number) p,r ;		

		if(size == 1) {
			p = [0];
			r = [sprd];
			fwd_func fa = fwd_func_interp(p,r,ip_linear(0,0));
			f[0] 		= fwd_func_add(fwd_base,fa);
			return;
		}
		
		//first
		if(d == f_end[0]) {			
			number k1 	= (f_first[1]-d)/365.0;
			p = [0,k1];
			r = [sprd,0];			
		}
		else {						
			number k1 	= (f_end[0]-d)/365.0;
			number k2 	= (f_first[1]-d)/365.0;
			p = [0,k1,k2];
			r = [sprd,sprd,0];			
		}
		
		fwd_func fa = fwd_func_interp(p,r,ip_linear(0,0));
		f[0] 		= fwd_func_add(fwd_base,fa);

		for(integer i=1;i<size-1;i++){
			
			if(f_first[i] < f_end[i]) {
				number k1 	= (f_end[i-1]-d)/365.0;
				number k2 	= (f_first[i]-d)/365.0;
				number k3 	= (f_end[i]-d)/365.0;
				number k4 	= (f_first[i+1]-d)/365.0;
				p = [k1,k2,k3,k4];
				r = [0,sprd,sprd,0];
			}
			else {
				number k1 	= (f_end[i-1]-d)/365.0;
				number k2 	= (f_end[i]-d)/365.0;
				number k3 	= (f_first[i+1]-d)/365.0;
			
				p = [k1,k2,k3];
				r = [0,sprd,0];
			}
			fa 		= fwd_func_interp(p,r,ip_linear(0,0));
			f[i] 	= fwd_func_add(fwd_base,fa);
		}

		
		number k1 	= (f_end[size-2]-d)/365.0;
		number k2;
		if(f_first[size-1] < f_end[size-1]) {
			
			k2 	= (f_first[size-1]-d)/365.0;
			//number k3 	= (f_end[size-1]-d)/365.0;			
			//p = [k1,k2, k3];
			//r = [0,sprd, sprd];
		}
		else {				
			k2 	= (f_end[size-1]-d)/365.0;						
		}
		p = [k1,k2];
		r = [0,sprd];
			
		fa 			= fwd_func_interp(p,r,ip_linear(0,0));
		f[size-1] 	= fwd_func_add(fwd_base,fa);	
	}

	//-------------------------------------------------------
	// wave_bumper_ff_vect
	//-------------------------------------------------------
	vector(fwd_func) wave_bumper_ff_vect(date 			d,
										cashflows_fwd	cf,
										fwd_z_model		fwd_model,
										fwd_func		fwd_base,
										number 			y_shift,										
										out fwd_func	fwd_all)
	{
		fwd_all = null<fwd_func>;
		integer size 	= cf.n_instruments();
		vector(fwd_func) f;	

		bucket_spread(fwd_base, d, [d],[d], y_shift, f);
		fwd_all 		= f[0];
	
		vector(date) f_end[size];
		vector(date) f_first[size];
		date run = d;
		
		for(integer i=0;i<size;i++){
			vector(number) fixed_amounts,nominals; 
			vector(date) fixed_dates,disc_dates, fwd_dates;
			logical ok = cf.get_cashflows(i, fixed_dates, fixed_amounts, fwd_dates, disc_dates,nominals);
			QL_REQUIRE(ok, "invalid cashflow");

			integer size_d 	= v_size(fwd_dates);//size==1 for single period instruments
			f_end[i] = fwd_dates[size_d-1];
			if(size_d == 1) {
				f_first[i] = f_end[i];
				run = f_end[i];
			}
			else {
				for(integer j=0;j<size_d;j++){
					if(fwd_dates[j] > run) {
						f_first[i] = fwd_dates[j];
						run = f_end[i];
						break;
					}
				}
			}
		}

		bucket_spread(fwd_base, d, f_first,f_end, y_shift, f);
	
		return f;
	}
	
	//-------------------------------------------------------
	// imp_depo
	//-------------------------------------------------------
	deposit create_depo_from_df(instrument instr,
								 disc_func df)
	{
		error_info error = new error_info(true,false);
		date d 		= instr.trade_date();
		number y 	= instr.yield(df);
		string name = strcat(instr.name()," - implied");
		deposit f 	= deposit(	instr.instr_def(),name,d,instr.settle_date(),
								instr.settle_date(),instr.maturity(),null<string>,y,y,100, error);

		if(error.is_error())
			return null<deposit>;
		else
			return f;
	}

	//-------------------------------------------------------
	// imp_fra
	//-------------------------------------------------------
	fra create_fra_from_df(instrument instr,
							disc_func df)
	{
		error_info error = new error_info(true,false);
		date d 			= instr.trade_date();
		number r 		= instr.yield(df);
		string name 	= strcat(instr.name()," - implied");
		fra f 			= fra(instr.instr_def(),name,d,instr.settle_date(),
							  instr.maturity(),r,r,100, error);
				
		if(error.is_error())
			return null<fra>;
		else
			return f;
	}

	//-------------------------------------------------------
	// imp_ir_future
	//-------------------------------------------------------
	ir_future create_irfut_from_df(instrument instr,
									number option(nullable) adj,
									disc_func df)
	{
		error_info error = new error_info(true,false);
		
		if(null(adj))
			adj = 0; 

		date d = instr.trade_date();
		number r 	= instr.yield(df) + adj;
		r 			*= 100;	
		string name = strcat(instr.name()," - implied");
		ir_future f = ir_future(instr.instr_def(),name,d,instr.settle_date(),100-r,
								null<number>,null<number>, error);
				
		if(error.is_error())
			return null<ir_future>;
		else
			return f;
	}
	//-------------------------------------------------------
	// imp_fixflt
	// Note: fwds calc from df and first_fix is null
	// fix cpn is implied
	//-------------------------------------------------------
	swap_fixflt create_fixflt_from_df(	instrument instr,
										disc_func df)
	{
		error_info error = new error_info(true,false);
		string name 	= strcat(instr.name()," - implied");

		swap_fixflt sw = null<swap_fixflt>;
		date d = instr.trade_date();
		
		if(instr.is_ext_swap(error)){
			swap_fixflt sw_in = swap_fixflt(instr);
			sw 	= swap_fixflt_plain(sw_in, name, d,instr.settle_date(),instr.settle_date(),null<string>,
										instr.maturity(),null<string>,null<number> ,
										100,true,df, df,null<number>, -1,error);
		}
		else {
			sw 	= swap_fixflt_plain(instr.instr_def(), name, d,instr.settle_date(),null<string>,
											instr.maturity(),null<string>,null<number> ,
											100,true,df, df,null<number>, -1,error);
		}
				
		if(error.is_error())
			return null<swap_fixflt>;
		else
			return sw;
	}
	//-------------------------------------------------------
	// imp_fixois
	// fix cpn is implied
	//-------------------------------------------------------
	swap_fixois create_fixois_from_df(instrument instr,
										disc_func df)
	{
		error_info error = new error_info(true,false);
		string name 	= strcat(instr.name()," - implied");
		date d 			= instr.trade_date();
		swap_fixois sw 	= null<swap_fixois>;
			
		if(instr.is_ext_swap(error)){
			swap_fixois sw_in = swap_fixois(instr);			
			sw = swap_fixois_plain(	sw_in, name, d,instr.settle_date(),null<string>,
									instr.maturity(),null<string>,null<number> ,
									100,true,df, null<fwd_func>,null<number>, -1,error);
		}
		else {
			sw = swap_fixois_plain(	instr.instr_def(), name, d,instr.settle_date(),null<string>,
									instr.maturity(),null<string>,null<number> ,
									100,true,df, null<fwd_func>,null<number>, -1,error);
		}
		
		if(error.is_error())
			return null<swap_fixois>;
		else
			return sw;
	}

	//-------------------------------------------------------
	// create_instr_from_df
	//-------------------------------------------------------
	logical create_instr_from_df( 	out vector(instrument) vi,
								disc_func df,
								vector(number) option(nullable) ir_fut_adj = null<vector(number)>)
	{
		integer size = v_size(vi);
		error_info ee = new error_info(true,true);

		vector_adj(size,0,ir_fut_adj);
		QL_REQUIRE(v_size(ir_fut_adj) == size, "invalid size of ir future convexity adjustment vector");
	
		logical ok = true;
		for (integer i = 0 ; i < size ; i++) {
			instr_type it = vi[i].instr_type(ee);
				
			switch(it){
			case instr_type.DEPOSIT:
			case instr_type.DEPOSIT_INDEX:
				vi[i] = CORE_INT.create_depo_from_df(vi[i], df);
				break;
			case instr_type.SWAP_FIXOIS:
				vi[i] = CORE_INT.create_fixois_from_df(vi[i], df);
				break;
			case instr_type.FRA:
				vi[i] = CORE_INT.create_fra_from_df(vi[i], df);
				break;
			case instr_type.IR_FUTURE:			
				vi[i] = CORE_INT.create_irfut_from_df(vi[i],ir_fut_adj[i], df);	
				break;			
			case instr_type.SWAP_FIXFLT:
				vi[i] = CORE_INT.create_fixflt_from_df(vi[i], df);
				break;
			default:
				ok = false;
				continue;
			}			
		}
		return ok;
	}

	
	//-------------------------------------------------------
	// create_fixflt_from_fwd
	// fix cpn is implied
	//-------------------------------------------------------
	swap_fixflt create_fixflt_from_fwd(	instrument i,
										disc_func df_disc,
										fwd_func f,
										number option(nullable) first_fix)
	{
		string name = strcat(i.name()," - implied");

		swap_fixflt sw = null<swap_fixflt>;

		error_info error = new error_info(true,false);
		date d = i.trade_date();
		
		if(i.is_ext_swap(error)){
			swap_fixflt sw_in = swap_fixflt(i);
			sw = swap_fixflt_plain(	sw_in, name, d,i.settle_date(),i.settle_date(),null<string>,
									i.maturity(),null<string>,null<number> ,
									100,true,df_disc, f,first_fix, -1,error);
		}
		else {
			
			sw = swap_fixflt_plain(	i.instr_def(), name, d,i.settle_date(),
									null<string>,i.maturity(),null<string>,null<number> ,
									100,true,df_disc, f,first_fix, -1,error);
		}
		
		if(error.is_error())
			return null<swap_fixflt>;
		else
			return sw;
	}

	//-------------------------------------------------------
	// create_fixois_from_fwd
	// fix cpn is implied
	//-------------------------------------------------------
	swap_fixois create_fixois_from_fwd(	instrument i,
										disc_func df_disc,
										fwd_func f ,
										number option(nullable) first_fix)
	{
		string name = strcat(i.name()," - implied");

		swap_fixois sw = null<swap_fixois>;

		error_info error = new error_info(true,false);
		date d = i.trade_date();
		
		if(i.is_ext_swap(error)){
			swap_fixois sw_in = swap_fixois(i);
			sw = swap_fixois_plain(	sw_in, name, d,i.settle_date(),null<string>,
									i.maturity(),null<string>,null<number> ,
									100,true,df_disc, f,first_fix, -1,error);
		}
		else {
			
			sw = swap_fixois_plain(	i.instr_def(), name, d,i.settle_date(),
									null<string>,i.maturity(),null<string>,null<number> ,
									100,true,df_disc, f,first_fix, -1,error);
		}
		
		if(error.is_error())
			return null<swap_fixois>;
		else
			return sw;
	}

	//-------------------------------------------------------
	// create_depo_from_fwd
	//-------------------------------------------------------
	deposit create_depo_from_fwd(instrument i,
								 fwd_func f)
	{
		error_info error = new error_info(true,false);
		date d 		= i.trade_date();
		number r 	= f.fwd((i.settle_date()-d)/365.0);
		
		string name = strcat(i.name()," - implied");
		deposit depo = deposit(i.instr_def(),name,d,i.settle_date(),
							   i.settle_date(),i.maturity(),null<string>,r,r,100, error);
		if(error.is_error())
			return null<deposit>;
		else
			return depo;
	}

	//-------------------------------------------------------
	// create_fra_from_fwd
	//-------------------------------------------------------
	fra create_fra_from_fwd(instrument i,
							fwd_func f)
	{
		error_info error = new error_info(true,false);
		date d = i.trade_date();
		number r = f.fwd((i.settle_date()-d)/365.0) ;

		string name = strcat(i.name()," - implied");
		fra fr = fra(i.instr_def(),name,d,i.settle_date(),i.maturity(),r,r,100, error);

		if(error.is_error())
			return null<fra>;
		else
			return fr;
	}

	//-------------------------------------------------------
	// create_irfut_from_fwd
	//-------------------------------------------------------
	ir_future create_irfut_from_fwd(instrument i,
									fwd_func f,
									number option(nullable) adj)
	{
		error_info error = new error_info(true,false);
		date d = i.trade_date();
		if(null(adj))
			adj = 0; 
		
		number r = f.fwd((i.settle_date()-d)/365.0) + adj;
		r *= 100;
		
		string name = strcat(i.name()," - implied");
		ir_future fu = ir_future(i.instr_def(),name,d,i.settle_date(),100-r,
								 null<number>,null<number>, error);

		if(error.is_error())
			return null<ir_future>;
		else
			return fu;
	}

	//-------------------------------------------------------
	// create_instr_from_fwd
	//-------------------------------------------------------
	logical create_instr_from_fwd( out vector(instrument) vi,																
								fwd_func 	f_flt_idx,
								disc_func 	df_disc,
								vector(number) option(nullable) first_fix = null<vector(number)>,
								vector(number) option(nullable) ir_fut_adj = null<vector(number)>)
	{
		integer size = v_size(vi);
		vector_adj(size,null<number>,first_fix);		
		QL_REQUIRE(v_size(first_fix) == size, "invalid size of first fix vector");

		vector_adj(size,0,ir_fut_adj);
		QL_REQUIRE(v_size(ir_fut_adj) == size, "invalid size of ir future convexity adjustment vector");
		
		error_info ee = new error_info(true,true);
		logical ok = true;
		for (integer i = 0 ; i < size ; i++) {		

			instr_type it = vi[i].instr_type(ee);
				
			switch(it){
			case instr_type.DEPOSIT:
			case instr_type.DEPOSIT_INDEX:
				vi[i] = CORE_INT.create_depo_from_fwd(vi[i], f_flt_idx);
				break;				
			case instr_type.FRA:
				vi[i] = CORE_INT.create_fra_from_fwd(vi[i], f_flt_idx);
				break;
			case instr_type.IR_FUTURE:				
				vi[i] = CORE_INT.create_irfut_from_fwd(vi[i],f_flt_idx,ir_fut_adj[i]);	
				break;			
			case instr_type.SWAP_FIXFLT:
				vi[i] = CORE_INT.create_fixflt_from_fwd(vi[i],df_disc,f_flt_idx, first_fix[i]);	
				break;
			case instr_type.SWAP_FIXOIS:
				vi[i] = CORE_INT.create_fixois_from_fwd(vi[i],df_disc,f_flt_idx, first_fix[i]);	
				break;
			default:
				ok = false;
				continue;
			}
		}
		return ok;
	}

	//-------------------------------------------------------
	// fwd_risk_ois
	// ois with an endogenous disc_func
	//-------------------------------------------------------
	fwd_func fwd_risk_ois(	date 			d,
							curve 			crv,
							fwd_z_model		fwd_model,
							number 			r_shift,
							integer 		freq,
							day_count_method dc,
							calendar		cal,
							out disc_func 	df_disc)					   
	{
		fwd_func f;
		vector(instrument) v_i 	= crv.instruments();

		vector(number) first_fix = null<vector(number)>;
		
		cashflows_fwd cf = CORE_INT.cf_fwd(v_i,first_fix,[r_shift],false);//cf_fwd(v_i,[r_shift],false);
		QL_REQUIRE(!null(cf),'invalid cashflows_fwd (risk) (fwd)');

		/*logical apply_hyman = false;
		interpolator ip;
		try {
			ip = fwd_ip_map(fwd_model.type(), apply_hyman,	0,0);
		}
		catch {
			ip = ip_linear();
		}*/
		date max = max_curve_date(crv);
		df_disc = bootstrap_on_extrap_flat(crv, d,max, cal);
		f = fwd_model.fwd(cf, df_disc);	
	}

	//-------------------------------------------------------
	// fwd_risk_const_basis_ois
	// if const_basis --> approximation
	// ois but with an exogenous disc_func
	//-------------------------------------------------------
	fwd_func fwd_risk_const_basis_ois(	date 			d,
										curve 			crv,
										fwd_z_model		fwd_model,
										disc_func		df_orig,
										number 			r_shift,
										day_count_method dc,
										calendar		cal,
										out disc_func 	df_disc)					   
	{
		vector(instrument) v_i 	= crv.instruments();
		vector(number) first_fix = null<vector(number)>;
		
		/*logical apply_hyman = false;
		interpolator ip;
		try {
			ip = fwd_ip_map(fwd_model.type(), apply_hyman,	0,0);
		}
		catch {
			ip = ip_linear();
		}
		*/
		//disc_func df_fwd = bootstrap_on_extrap_flat(crv, ip, cal);
		date max = max_curve_date(crv);
		disc_func df_fwd = bootstrap_on_extrap_flat(crv, d,max, cal);
		QL_REQUIRE(!null(df_fwd),'invalid disc_func (risk-1) (fwd)');

		date max_date 	= v_max(v_i.maturity());
		date min_date 	= v_min(v_i.issue_date());
		integer days 	= max_date-min_date-1;
		vector(date) float_idx0[days];
		vector(date) float_idx1[days];
		for(integer i=0;i<days;i++) {
			float_idx0[i] = d+i;
			float_idx1[i] = d+i+1;
		}
		
		vector(number) flt_df_orig_r = df_orig.simple_rate(d, float_idx0, float_idx1, dc);	
		vector(number) flt_r 		= df_fwd.simple_rate(d, float_idx0, float_idx1, dc);				
		vector(number) curr_basis 	= flt_r - flt_df_orig_r;

		curve crv_p;
		vector(number) user_y 	= v_i.yield() + r_shift;
		logical ok 				= WSP_UTIL.set_curve_yld(crv,user_y*100,crv_p);
		QL_REQUIRE(ok,"set_curve_yld fail (fwd)");

		disc_func df_fwd_p 		= bootstrap_on_extrap_flat(crv_p, d,max, cal);
		QL_REQUIRE(!null(df_fwd_p),'invalid disc_func (risk-2) (fwd)');
					
		vector(number) periods 	= period(float_idx0, float_idx1, dc);
		df_disc 				= disc_func_add_spreads(df_fwd_p, periods, -curr_basis, ip_linear(), RT_ON);
		QL_REQUIRE(!null(df_disc),'invalid disc_func (risk-3) (fwd)');
	
		cashflows_fwd cf = CORE_INT.cf_fwd(v_i,first_fix,[r_shift],false);//cf_fwd(v_i,[r_shift],false);
		QL_REQUIRE(!null(cf),'invalid cashflows_fwd (risk) (fwd)');
		return fwd_model.fwd(cf, df_disc);
	}
	//-------------------------------------------------------
	// fwd_risk
	// if const_basis --> approximation
	//-------------------------------------------------------
	fwd_func fwd_risk_const_basis(	date 			d,
									curve 			crv,
									fwd_z_model		fwd_model,
									disc_func		df_orig,
									number 			r_shift,
									integer 		freq,
									day_count_method dc,									
									vector(number) first_fix,
									out disc_func 	df_disc)					   
	{
		vector(instrument) v_i 	= crv.instruments();
		//step 1: 	calculate current basis
		//			a. calc disc_func from current fwd instruments (df_fwd)
		//			b. calc rates from current disc_func
		//			c. calc rates from df_fwd
		//			d. calc diff --> = spread between fwd curve and disc curve
		disc_func df_fwd 		= bootstrap(crv, ip_linear(), RT_CONT);
		QL_REQUIRE(!null(df_fwd),'invalid disc_func (risk-1) (fwd)');
		
		date max_date 			= v_max(v_i.maturity());
		date min_date 			= v_min(v_i.issue_date());
		logical ip_idx_stub 	= false;		
		vector(date) float_idx0, float_pmt, float_idx1;

		CORE_INT.swap_flt_dates(min_date,null<date>,null<date>, max_date,null<number>,freq,
								BD_MOD_FOLLOWING,calendar(),false,BD_MOD_FOLLOWING,calendar(),false,
								ip_idx_stub,float_idx0, float_pmt, float_idx1);

		vector(number) flt_df_orig_r = df_orig.simple_rate(d, float_idx0, float_idx1, dc);	
		vector(number) flt_r 		= df_fwd.simple_rate(d, float_idx0, float_idx1, dc);				
		vector(number) curr_basis 	= flt_r - flt_df_orig_r;

		//step 2. shift the fwd curve
		//			a. calc new disc_func from shifted fwd instruments (df_fwd_p)
		//			b. adjust df_fwd_p with the spread
		//			--> we have a new disc_func with a maintained spread between fwd rates and disc rates
		curve crv_p;
		vector(number) user_y 	= v_i.yield() + r_shift;
		logical ok 				= WSP_UTIL.set_curve_yld(crv,user_y*100,crv_p);
		QL_REQUIRE(ok,"set_curve_yld fail (fwd)");

		disc_func df_fwd_p 		= bootstrap(crv_p, ip_linear(), RT_CONT);
		QL_REQUIRE(!null(df_fwd_p),'invalid disc_func (risk-2) (fwd)');
					
		vector(number) periods 	= period(float_idx0, float_idx1, dc);
		df_disc 				= disc_func_add_spreads(df_fwd_p, periods, -curr_basis, ip_linear(), RT_CONT);
		QL_REQUIRE(!null(df_disc),'invalid disc_func (risk-3) (fwd)');
	
		cashflows_fwd cf = CORE_INT.cf_fwd(v_i,first_fix,[r_shift],false);//cf_fwd(v_i,[r_shift],false);
		QL_REQUIRE(!null(cf),'invalid cashflows_fwd (risk) (fwd)');
		return fwd_model.fwd(cf, df_disc);
	}
	//-------------------------------------------------------
	// fwd_risk
	// if const_basis --> approximation
	//-------------------------------------------------------
	fwd_func fwd_risk(	date 			d,
						curve 			crv,
						fwd_z_model		fwd_model,
						disc_func		df_orig,
						number 			r_shift,
						logical 		const_basis,
						integer 		freq,
						day_count_method dc,
						calendar		cal,
						vector(number) first_fix,
						out disc_func 	df_disc)					   
	{
		
		fwd_func f;
					
	
		if(const_basis){
			if(freq == 365)
				f = fwd_risk_const_basis_ois(d,crv,fwd_model,df_orig,r_shift,dc,cal,df_disc);
			else
				f = fwd_risk_const_basis(d,crv,fwd_model,df_orig,r_shift,freq,dc,first_fix,df_disc);
		}
		else {
			vector(instrument) v_i 	= crv.instruments();
			//shift instruments in curve but keep disc_func the same
			cashflows_fwd cf = CORE_INT.cf_fwd(v_i,first_fix,[r_shift],false);//cf_fwd(v_i,[r_shift],false);
			QL_REQUIRE(!null(cf),'invalid cashflows_fwd (risk) (fwd)');	
			f = fwd_model.fwd(cf, df_orig);	
			df_disc = df_orig;
		}
		return f;			
	}

	//-------------------------------------------------------
	// instr_bumper_ff
	//-------------------------------------------------------
	fwd_func instr_bumper_ff(	curve 			crv,
								fwd_z_model		fwd_model,
								number 			y_shift,
								vector(number) first_fix,
								cashflows_fwd option(nullable) cf_base,
								disc_func 		df,
								fwd_func option(nullable) ff_pre,
								date option(nullable) ff_pre_cut_off,
								out fwd_func	fwd_base)
	{
		QL_FAIL_COND(!null(ff_pre) && null(ff_pre_cut_off),"ff_pre_cut_off missing");
		
		fwd_base = null<fwd_func>;
		
		if(null(cf_base))
			cf_base = CORE_INT.cf_fwd(crv.instruments(),first_fix,null<vector(number)>,false);
					
		fwd_base 			= !null(ff_pre) ? fwd_model.fwd(ff_pre, ff_pre_cut_off, cf_base, df,crv): fwd_model.fwd(cf_base, df, crv);
		cashflows_fwd cf 	= CORE_INT.cf_fwd(crv.instruments(),first_fix,[y_shift],false); 
		fwd_func fwd_all_shift = !null(ff_pre) ? fwd_model.fwd(ff_pre, ff_pre_cut_off, cf, df,crv) : fwd_model.fwd(cf, df,crv);
		return fwd_all_shift;
	}
	//-------------------------------------------------------
	// instr_bumper_ff_vect
	//-------------------------------------------------------
	vector(fwd_func) instr_bumper_ff_vect(	curve 			crv,
											fwd_z_model		fwd_model,
											number 			y_shift,
											vector(number) first_fix,
											cashflows_fwd option(nullable) cf_base,
											disc_func 		df,
											fwd_func option(nullable) ff_pre,
											date option(nullable) ff_pre_cut_off,
											out vector(integer) incl,//input and output, size is equal to number of included instruments
											out fwd_func	fwd_base,
											out fwd_func	fwd_all_shift,															
											out vector(instrument) bump_instr,//size is equal to number of included instruments 
											out vector(fwd_func) f_v)//size is equal to number of instruments in the curve (=curve_)
	{
		QL_FAIL_COND(!null(ff_pre) && null(ff_pre_cut_off),"ff_pre_cut_off missing");
		
		fwd_base = fwd_all_shift = null<fwd_func>;
		resize(bump_instr,0);
		
		vector(instrument) v_i 	= crv.instruments();
		integer size_all 		= v_size(v_i);
		integer size_incl ;

		//check and maybe update the incl vector
		if(null(incl) || v_size(incl) == 0) {
			incl 		= index_vector(size_all);
			size_incl 	= size_all;
		}
		else {
			size_incl 	= v_size(incl);
			QL_REQUIRE(size_incl <= size_all,"invalid size of include vector");
			
			for(integer i=0;i<size_incl;i++){
				QL_REQUIRE(incl[i] >= 0 && incl[i] < size_all,"invalid index in include vector");
				if(i>0){
					QL_REQUIRE(incl[i] > incl[i-1],"invalid index order in include vector");
				}
			}
		}

		vector(number) y_adj[size_all];
		y_adj 				= y_shift;
		vector(fwd_func) f_incl_v[0];
		resize(f_v,size_all);
		
		if(null(cf_base))
			cf_base = CORE_INT.cf_fwd(crv.instruments(),first_fix,null<vector(number)>,false);
					
		fwd_base 			= !null(ff_pre) ? fwd_model.fwd(ff_pre, ff_pre_cut_off, cf_base, df,crv): fwd_model.fwd(cf_base, df, crv);
		cashflows_fwd cf 	= CORE_INT.cf_fwd(crv.instruments(),first_fix,[y_shift],false); 
		fwd_all_shift 		= !null(ff_pre) ? fwd_model.fwd(ff_pre, ff_pre_cut_off, cf, df,crv) : fwd_model.fwd(cf, df,crv);
		
		vector(date) mat_v 	=  v_i.maturity();
		
		for(integer i=0;i<size_all;i++){
			logical is_shifted = CORE_INT.bump_spread(y_shift, mat_v, y_adj, incl, i);
			if(is_shifted) {
				push_back(bump_instr,v_i[i]);		
				cf 			= CORE_INT.cf_fwd(crv.instruments(),first_fix,y_adj,false);
				f_v[i]  	= !null(ff_pre) ? fwd_model.fwd(ff_pre, ff_pre_cut_off, cf, df,crv) : fwd_model.fwd(cf, df,crv);			
				push_back(f_incl_v,f_v[i]);
			}
			else {
				f_v[i] 		= fwd_base;
			}
		}
		QL_REQUIRE(v_size(bump_instr) == size_incl,"fatal inconsistency in calculations");
		return f_incl_v;
	}
}