option(null: hard);	// Throw error if a null value is used in a non-nullable context


vector(string) cpn_reinv_type_list() option (category: 'Instrument') 
{
	vector(cpn_reinv_type) e;
	expand_enum(e) ;
	return string(e);
}

cpn_reinv_type cpn_reinv_type_map(string s)
option (category: 'Instrument') 
{
	cpn_reinv_type e;
	str_to_enum(s,e);
	return e;
}


module CORE_INT
{

	
	//-------------------------------------------------------
	// get_crt_rate 
	//-------------------------------------------------------
	number get_crt_rate(cpn_reinv_type				cr_type,
						number option(nullable) 	r, 
						disc_func option(nullable) 	fr,
						date						trade_d,
						date						start,
						date						end,
						day_count_method 			dc, 
						rate_type 					rt)
	
	{ 	
		if(cr_type == cpn_reinv_type.FIXED || cr_type == cpn_reinv_type.FREE)
			return r;
		else if(start == end)
			return 0;
		else if(cr_type == cpn_reinv_type.FWD){
			number fv = fr.disc((start-trade_d)/365.0);
			number pv = fr.disc((end-trade_d)/365.0);
			//return CORE_INT.rate(fv, pv, rt, start, end, dc); 
			return implied_rate(fv, pv, start, end, dc, rt); 
		}
		else if(cr_type == cpn_reinv_type.SPOT){		
			number pv = fr.disc((end-start)/365.0);
			//return CORE_INT.rate(1, pv, rt, start, end, dc); 
			return implied_rate(1, pv, start, end, dc, rt); 
		}
		else
			QL_FAIL("invalid cpn reinvestment type");	
	}
	//-------------------------------------------------------
	// reinv_cpn_bond 
	//-------------------------------------------------------
	void reinv_cpn_bond(	instrument 			inst,
							instrument 			inst_fwd,
							date 				trade_d,					      	
					      	number option(nullable) 	rate, 
							disc_func option(nullable) 	fr,							
							day_count_method 	dc, 
							rate_type 			rt,
							logical 			is_ilb,
							cpn_reinv_type		cr_type,
					      	out number reinv_cpns, 
					      	out number no_reinv_cpns) 
	{    
	    reinv_cpns = 0 ;
	    no_reinv_cpns = 0 ;	   
	    	
		vector(number) vc 	= inst.cash_flow_amounts ();
		vector(date) vd 	= inst.cash_flow_dates_noadj ();	
		vector(date) vd_adj = inst.cash_flow_dates ();			
		date fwd_date 		= inst_fwd.settle_date();
	
		QL_REQUIRE(!null(vd),"invalid instrument");

		//string exc = i.excpn_code ();
		//date next_ex_date_fwd = date_code(exc).apply_back(vd_adj_fwd[0], i.calendar ());	
			
		number index = is_ilb ? inst.index_factor() : 1;		
	
		QL_REQUIRE(!null(index),"internal error");

		number flow_diff;
		if(fwd_date < inst.maturity())
			flow_diff = v_size(vd) - v_size(inst_fwd.cash_flow_dates_noadj ());
		else if(fwd_date == inst.maturity())
			flow_diff = v_size(vd);
		else
			QL_FAIL("invalid forward settlement date");
	
		if(flow_diff == 0 || inst.coupon() == 0)
			return;		
	
		number r, flow;
		date cpn_date_adj,cpn_date;
	    for( integer i = 1; i <= flow_diff; i++ )  {	
			cpn_date 		= vd[i-1];
			cpn_date_adj 	= vd_adj[i-1];			
			flow 			= is_ilb ? vc[i-1]/index * inst.index_factor(cpn_date)/100: vc[i-1]/100;									
			QL_REQUIRE(!null(flow),"internal error");
		
			if (i < flow_diff) {//intermediate flow to reinvest			    		    		   		    	    	    
			    no_reinv_cpns 	+= flow ;	
				r 				= get_crt_rate(cr_type, rate, fr, trade_d, cpn_date_adj, fwd_date, dc, rt) ;
				//reinv_cpns 		+= fv(flow, r, rt, cpn_date_adj, fwd_date,  dc);
				reinv_cpns 		+= future_value(flow, r, cpn_date_adj, fwd_date,  dc, rt);				
			}
			else {	//last flow	    		    		        
			    if(cpn_date <= fwd_date) {
					no_reinv_cpns += flow ;
					if(cpn_date_adj <= fwd_date){ //move coupon forward	
						r 			= get_crt_rate(cr_type, rate, fr, trade_d, cpn_date_adj, fwd_date, dc, rt) ;
						//reinv_cpns 	+= fv(flow, r, rt, cpn_date_adj, fwd_date,  dc);
						reinv_cpns 	+= future_value(flow, r, cpn_date_adj, fwd_date,  dc, rt);

					}
					else {//discount coupon
						r 			= get_crt_rate(cr_type, rate, fr, trade_d, fwd_date, cpn_date_adj, dc, rt) ;
						//reinv_cpns 	+= pv(flow, r, rt, fwd_date, cpn_date_adj,  dc);		    
						reinv_cpns 	+= present_value(flow, r, fwd_date, cpn_date_adj,  dc, rt);
					}
			    }
			    //else if( next_ex_date_fwd > fwd_date) {}//we will not get the coupon
			    else {//we're in the exdiv period
					no_reinv_cpns 	+= flow ;
					r 				= get_crt_rate(cr_type, rate, fr, trade_d, fwd_date, cpn_date_adj, dc, rt) ;
					//reinv_cpns 		+= pv(flow, r, rt, fwd_date, cpn_date_adj,  dc) ;
					reinv_cpns 		+= present_value(flow, r, fwd_date, cpn_date_adj,  dc, rt) ;
				
			    }	    
			}
	    }	                                        
	}
	//-------------------------------------------------------
	//  hor_ret_bond
	//-------------------------------------------------------
	public void hor_ret_bond(	instrument 					inst,
								instrument 					inst_fwd,
								date 						trade_d,					      	
						      	number option(nullable) 	cpn_inv_r,							 
								disc_func option(nullable) 	fr,							
								day_count_method 			dc, 
								rate_type 					rt,
								number 						nominal,
								cpn_reinv_type				cr_type,						
								out number dp, 
								out number dp_fwd,
								out number ai, 
								out number	ai_fwd, 
						      	out number reinv_cpns, 
						      	out number no_reinv_cpns,
								out number fund_cost,
								out number net_fund_cost,
								out number coc,
								out number hr) 

	{		
		logical is_ilb = false;
		..instr_type t = inst.instr_type(); 
		if( t == ..instr_type.IL_REAL_FIXED_CPN_BOND || t == ..instr_type.IL_REAL_ZERO_CPN_BOND)
			is_ilb = true;
		else if( t != ..instr_type.FIXED_CPN_BOND && t != ..instr_type.BILL && t != ..instr_type.ZERO_CPN_BOND)
			QL_FAIL('invalid instrument type');
	
		QL_FAIL_COND(inst.instr_type() != inst_fwd.instr_type(),'invalid instrument type (fwd instrument is a different instrument type)');

		date settle_fwd = inst_fwd.settle_date();
		date mat 		= inst.maturity();
		if(nominal>100){
			dp 		= round(inst.dirty_price()*nominal/100, 2);
			dp_fwd 	= settle_fwd < mat ? round(inst_fwd.dirty_price()*nominal/100, 2): 0;
			ai 		= round(inst.accrued()*nominal/100, 2);	
			ai_fwd 	= settle_fwd < mat ? round(inst_fwd.accrued()*nominal/100, 2): 0;
		}
		else{
			dp 		= inst.dirty_price()*nominal/100;
			dp_fwd 	= settle_fwd < mat ? inst_fwd.dirty_price()*nominal/100: 0;
			ai 		= inst.accrued()*nominal/100;
			ai_fwd 	= settle_fwd < mat ? inst_fwd.accrued()*nominal/100: 0;
		}	
	
		if(cr_type == cpn_reinv_type.FREE)		
			cpn_inv_r = hz_rate_free (inst, get_t_from_s(inst,settle_fwd), inst_fwd.yield(), rt, dc);
		else if(cr_type == cpn_reinv_type.FIXED)
			QL_REQUIRE(!null(cpn_inv_r),"invalid cpn reinvest rate");
		else if((cr_type == cpn_reinv_type.FWD || cr_type == cpn_reinv_type.SPOT) && null(fr))
			QL_FAIL("invalid disc_func");		
	
		reinv_cpn_bond(inst, inst_fwd, trade_d,  cpn_inv_r, fr,dc,rt, is_ilb, cr_type, reinv_cpns, no_reinv_cpns) ;	
		reinv_cpns 		*= nominal;
		no_reinv_cpns 	*= nominal;			
		//hr 				= rate(dp_fwd+reinv_cpns, dp, rt, inst.settle_date(), inst_fwd.settle_date(),dc);
		hr 				= implied_rate(dp_fwd+reinv_cpns, dp, inst.settle_date(), inst_fwd.settle_date(),dc, rt);

		//neg sign is a cost					
		fund_cost		= -(dp_fwd + reinv_cpns - dp);
		net_fund_cost 	= fund_cost + (reinv_cpns-no_reinv_cpns);
		number accr_chg = ai_fwd - ai;
		coc 			= accr_chg + fund_cost + reinv_cpns;
	}

	//-------------------------------------------------------
	//  hor_y_bond
	//-------------------------------------------------------
	void hor_y_bond(	instrument 					inst,
						instrument 					inst_fwd,
						date 						trade_d,
						out number hr,					      	
				      	number option(nullable) 	cpn_inv_r,							 
						disc_func option(nullable) 	fr,							
						day_count_method 			dc, 
						rate_type 					rt,
						number 						nominal,
						cpn_reinv_type				cr_type,					
						out number dp, 
						out number dp_fwd,
						out number ai, 
						out number ai_fwd, 
				      	out number reinv_cpns, 
				      	out number no_reinv_cpns,
						out number fund_cost,
						out number net_fund_cost,
						out number coc) 
	
	{		
		logical is_ilb = false; 

		..instr_type t = inst.instr_type(); 
		if( t == ..instr_type.IL_REAL_FIXED_CPN_BOND || t == ..instr_type.IL_REAL_ZERO_CPN_BOND)
			is_ilb = true;
		else if( t != ..instr_type.FIXED_CPN_BOND && t != ..instr_type.BILL && t != ..instr_type.ZERO_CPN_BOND)
			QL_FAIL('invalid instrument type');
		
		QL_FAIL_COND(inst.instr_type() != inst_fwd.instr_type(),'invalid instrument type (fwd instrument is a different instrument type)');
		QL_FAIL_COND(null(hr) && null(fr),"invalid horizon rate or fit_result");		
		QL_FAIL_COND(inst_fwd.settle_date() >= inst.maturity(),"invalid forward settlement");	
	
		if(nominal>100){
			dp 		= round(inst.dirty_price()*nominal/100, 2);
			ai 		= round(inst.accrued()*nominal/100, 2);	
			ai_fwd 	= round(inst_fwd.accrued()*nominal/100, 2);
		}
		else{
			dp 		= inst.dirty_price()*nominal/100;
			ai 		= inst.accrued()*nominal/100;
			ai_fwd 	= inst_fwd.accrued()*nominal/100 ;
		}
		
		if(null(hr)){
			hr = CORE_INT.rate(trade_d,  inst.settle_date(), inst_fwd.settle_date(), fr, dc, rt);
			QL_REQUIRE(!null(hr),"invalid horizon rate");
		}
	
		if(cr_type == cpn_reinv_type.FREE)				
			cpn_inv_r = hr;	
		else if(cr_type == cpn_reinv_type.FIXED)
			QL_REQUIRE(!null(cpn_inv_r),"invalid cpn reinvest rate");					
		else if((cr_type == cpn_reinv_type.FWD || cr_type == cpn_reinv_type.SPOT) && null(fr))
			QL_FAIL("invalid disc_func");	
		
		reinv_cpn_bond(inst, inst_fwd, trade_d,  cpn_inv_r, fr,dc,rt, is_ilb, cr_type, reinv_cpns, no_reinv_cpns) ;	
		reinv_cpns 		*= nominal;
		no_reinv_cpns 	*= nominal;	
	
		//dp_fwd 			= fv(dp, hr, rt, inst.settle_date(), inst_fwd.settle_date(), dc) ;	
		dp_fwd 			= future_value(dp, hr, inst.settle_date(), inst_fwd.settle_date(), dc, rt) ;
		fund_cost 		= -dp_fwd + dp;
		dp_fwd 			-= reinv_cpns;	
		if(nominal>100) dp_fwd = round(dp_fwd, 2);
	
		net_fund_cost 	= fund_cost + (reinv_cpns-no_reinv_cpns);
		number accr_chg = ai_fwd - ai;
		coc 			= accr_chg + fund_cost + reinv_cpns;
	}
	//-------------------------------------------------------
	//  hor_y_bond
	//-------------------------------------------------------
	number hor_y_bond(	instrument 			inst,
						date 				settle_fwd,
						date 				trade_d,
						out number hr,					      	
				      	number option(nullable) 	cpn_inv_r,							 
						disc_func option(nullable) fr,							
						day_count_method 	dc, 
						rate_type 			rt,
						cpn_reinv_type		cr_type) 
	{
		try {
			date settle_spot = inst.settle_date();
			if(settle_fwd < settle_spot)
				return null<number>;
			else if(settle_fwd == settle_spot)
				return inst.yield();
		
			instrument inst_fwd = inst.set_settle_date(settle_fwd);
			number dp,dp_fwd,ai,ai_fwd,reinv_cpns,no_reinv_cpns,fund_cost, net_fund_cost,coc;
			hor_y_bond(inst,inst_fwd,trade_d,hr,cpn_inv_r,fr,dc,rt,100,cr_type,					
						dp,dp_fwd,ai,ai_fwd,reinv_cpns,no_reinv_cpns,fund_cost,net_fund_cost,coc) ;	
	
			return inst_fwd.price_to_yield (dp_fwd);
		}
		catch {
			CORE_INT.write_warning_message(err.message());
			return null<number>;
		}
	}
	//-------------------------------------------------------
	//  pl_bond
	//-------------------------------------------------------
	number pl_bond(	instrument 					inst,
					instrument 					inst_fwd,
					date 						trade_d,
					number  					r,
			      	number option(nullable) 	cpn_inv_r,
					disc_func option(nullable) 	fr,	
					day_count_method 			dc, 
					rate_type 					rt,
					number 						nominal,
					cpn_reinv_type				cr_type,
					logical						buy_bond,
					out number cash_inst, 
					out number cash_inst_fwd,
					out number fund_cost) 
	
	{		
		logical is_ilb = false;

		..instr_type t = inst.instr_type(); 
		if( t == ..instr_type.IL_REAL_FIXED_CPN_BOND || t == ..instr_type.IL_REAL_ZERO_CPN_BOND)
			is_ilb = true;
		else if( t != ..instr_type.FIXED_CPN_BOND && t != ..instr_type.BILL && t != ..instr_type.ZERO_CPN_BOND)
			QL_FAIL('invalid instrument type');
		
	
		QL_FAIL_COND(inst.instr_type() != inst_fwd.instr_type(),'invalid instrument type (fwd instrument is a different instrument type)');

		date settle_fwd = inst_fwd.settle_date();
		date mat 		= inst.maturity();
		number dp, dp_fwd;

		if(nominal>100){
			dp 		= round(inst.dirty_price()*nominal/100, 2);
			dp_fwd 	= settle_fwd < mat ? round(inst_fwd.dirty_price()*nominal/100, 2): 0;
			//ai = round(inst.accrued()*nominal/100, 2);	
			//ai_fwd = settle_fwd < mat ? round(inst_fwd.accrued()*nominal/100, 2): 0;
		}
		else{
			dp 		= inst.dirty_price()*nominal/100;
			dp_fwd 	= settle_fwd < mat ? inst_fwd.dirty_price()*nominal/100: 0;
			//ai = inst.accrued()*nominal/100;
			//ai_fwd = settle_fwd < mat ? inst_fwd.accrued()*nominal/100: 0;
		}	
	
		if(cr_type == cpn_reinv_type.FREE)				
			cpn_inv_r = r;	
		else if(cr_type == cpn_reinv_type.FIXED)
			QL_REQUIRE(!null(cpn_inv_r),"invalid cpn reinvest rate");		
		else if((cr_type == cpn_reinv_type.FWD || cr_type == cpn_reinv_type.SPOT) && null(fr))
			QL_FAIL("invalid disc_func");		
	
		number reinv_cpns, no_reinv_cpns;
		reinv_cpn_bond(inst, inst_fwd, trade_d,  cpn_inv_r, fr, dc, rt, is_ilb, cr_type, reinv_cpns, no_reinv_cpns) ;	
		reinv_cpns 		*= nominal;
		no_reinv_cpns 	*= nominal;
		//neg sign is a cost
		//fund_cost 		= -fv(dp, r, rt, inst.settle_date(), inst_fwd.settle_date(), dc) + dp ;
		fund_cost 		= -future_value(dp, r, inst.settle_date(), inst_fwd.settle_date(), dc, rt) + dp ;
		if(!buy_bond) fund_cost *= -1; 	
		cash_inst 		= buy_bond ? -dp : dp;
		cash_inst_fwd 	= buy_bond ? dp_fwd+reinv_cpns : -(dp_fwd+reinv_cpns);											
		number pl 		= cash_inst + cash_inst_fwd + fund_cost; 					
		return pl;
	}

	
	//-------------------------------------------------------
	//  portf_yld_cf_approx
	//-------------------------------------------------------
	number portf_yld_cf_approx(	vector(number) y,																									
								vector(number) pvbp,
								vector(number) nom)
								
	{
		try{
			number n = 0;
			number d = 0;
			for (integer i = 0; i < v_size(y); i++ ) {
				n += y[i]*pvbp[i]*nom[i];
				d += pvbp[i]*nom[i];
			}
			return n/d;	
		}
		catch {
			CORE_INT.write_warning_message(err.message());
			return null<number>;
		}
	}
	//-------------------------------------------------------
	//  portf_yld_avg
	//-------------------------------------------------------
	number portf_yld_avg(	vector(number) y,																									
							vector(number) dp,
							vector(number) nom)
							
	{
		try{
			number n = 0;
			number d = 0;
			for (integer i = 0; i < v_size(y); i++ ) {
				n += y[i]*dp[i]*nom[i];
				d += dp[i]*nom[i];
			}
			return n/d;
		}
		catch {
			CORE_INT.write_warning_message(err.message());
			return null<number>;
		}	
	}
	
	//----------------------------------------------------------------------
	// swap_pmt_dates                                                        
	//---------------------------------------------------------------------- 
	void swap_pmt_dates(	date			issue, 
							date option(nullable)	first,
							date option(nullable)	last_reg,
							date			mat, 						
							number 	option(nullable)	roll_day, 
							integer			freq, 							
							bd_convention	bd,
							calendar		cal,
							logical 		eom,										
							out vector(date)   dates)
	{
		if(null(roll_day)) roll_day = 0;
		date out_first,out_last_reg;
		integer out_roll_day;
		logical out_is_reg_first, out_is_reg_last, out_is_eom_roll;
	    ..swap_pmt_dates(issue,first,last_reg,mat,integer(roll_day) , freq,bd,cal,eom,	out_first,out_last_reg, 
						out_roll_day, out_is_reg_first, out_is_reg_last, out_is_eom_roll,dates);
						    
	    return;
	}
	
	//----------------------------------------------------------------------
	// swap_flt_dates                                                         
	//---------------------------------------------------------------------- 
	void swap_flt_dates(date					issue, 
						date option(nullable)	first,
						date option(nullable)	last_reg,
						date					mat, 
						number option(nullable) roll_day, 
						integer					freq, 
						bd_convention			bd,
						calendar				cal,
						logical 				eom,
						bd_convention			idx_bd,
						calendar				idx_cal,
						logical 				idx_eom,
						logical 				ip_idx_stub,
						out vector(date)   pmt_dates_beg,
						out vector(date)   pmt_dates_end,		
						out vector(date)   idx_dates_end)
	{    
		date out_first, out_last_reg;
		integer out_roll_day;
		logical is_reg_first, is_reg_last, eom_roll;
		vector(date) dates, idx_dates, fix_dates;
		if(null(roll_day)) roll_day = 0;
		integer reset_days = 2;
		
		..swap_pmt_dates_flt(	issue,first,last_reg,mat, integer(roll_day),freq,bd,cal,eom,idx_bd,idx_cal,idx_eom,reset_days,ip_idx_stub,out_first,out_last_reg, 
						out_roll_day, is_reg_first, is_reg_last, eom_roll, dates, idx_dates, fix_dates);
		vector(date) tmp_dates_beg;
		dates_1v_to_2v( dates,pmt_dates_beg,pmt_dates_end);
		dates_1v_to_2v( idx_dates,tmp_dates_beg,idx_dates_end);    
	    return;
	}
	//----------------------------------------------------------------------
	// swap_flt_dates                                                         
	//---------------------------------------------------------------------- 
	void swap_flt_dates(date					issue, 
						date option(nullable)	first,
						date option(nullable)	last_reg,
						date					mat, 
						number option(nullable) roll_day, 
						integer					freq, 
						bd_convention			bd,
						calendar				cal,
						logical 				eom,
						bd_convention			idx_bd,
						calendar				idx_cal,
						logical 				idx_eom,
						logical 				ip_idx_stub,
						out date  option(nullable)			out_first,
						out date  option(nullable)			out_last_reg,
						out number  option(nullable)		out_roll_day,
						out logical option(nullable) 		is_reg_first,
						out logical option(nullable) 		is_reg_last,
						out logical  option(nullable)		eom_roll,
						out vector(date)   	pmt_dates_beg,
						out vector(date)   	pmt_dates_end,		
						out vector(date)   	idx_dates_end)
	{    
		vector(date) dates, idx_dates;
		integer out_roll_day_i;
		//logical is_reg_first_i, is_reg_last_i, eom_roll_i;
		vector(date) fix_dates;
		integer reset_days = 2;
	
		..swap_pmt_dates_flt(	issue,first,last_reg,mat, integer(roll_day),freq,bd,cal,eom,idx_bd,idx_cal,idx_eom,reset_days,ip_idx_stub,out_first,out_last_reg, 
							out_roll_day_i, is_reg_first, is_reg_last, eom_roll, dates, idx_dates, fix_dates);
	
		out_roll_day 	= out_roll_day_i;
		//is_reg_first 	= is_reg_first_i != 0;
		//is_reg_last 	= is_reg_last_i != 0;
		//eom_roll 		= eom_roll_i != 0;
	
		vector(date) tmp_dates_beg;
		dates_1v_to_2v( dates,pmt_dates_beg,pmt_dates_end);
		dates_1v_to_2v( idx_dates,tmp_dates_beg,idx_dates_end);    
	    return;
	}
	
	//-------------------------------------------------------
	// par_fixflt_swap_rate
	//-------------------------------------------------------
	number par_fixflt_swap_rate(date	 			d,
								calendar 			cal ,		
								date 				settle ,
								bd_convention 		bd ,	
								date 				issue ,
								date 				maturity ,
								logical 			eom,
								integer 			flt_freq,
								integer 			fix_freq,								
								day_count_method 	flt_dc,
								day_count_method 	fix_dc,
								disc_func			df_disc,
								fwd_func			f_flt_idx,
								integer 			flt_reset_days,
								number option(nullable) curr_fix)
	
	{		
		vector(date) flt_dates;
		CORE_INT.swap_pmt_dates(issue ,null<date>,null<date>, maturity,day(issue), flt_freq,bd, cal, eom,flt_dates);
		
		vector(date) fix_dates;
		CORE_INT.swap_pmt_dates(issue ,null<date>,null<date>, maturity,day(issue), fix_freq,bd, cal, eom,fix_dates);
		
		date flt_date_fix = cal.move_bus_days(flt_dates[0], -flt_reset_days) ; 
		if(flt_date_fix > d) curr_fix = null<number>; 
		
		return swap_imp_fix_rate(d,settle,df_disc,f_flt_idx,/*flt_freq,*/0,curr_fix, flt_dates,flt_dc,df_disc,fix_dates,fix_dc);	
	}
	//-------------------------------------------------------
	// par_fixflt_swap_rate
	//-------------------------------------------------------
	number par_fixflt_swap_rate(date	 			d,
								instrument 			swap,
								disc_func			df_disc,
								fwd_func			f_flt_idx,
								integer 			flt_reset_days,
								number option(nullable) curr_fix)
	{
		calendar cal 		= swap.calendar();		
		date settle 		= swap.settle_date();
		bd_convention bd 	= swap.pmt_bus_day_d(BD_MOD_FOLLOWING);	
		logical	eom			= swap.eom();
		date issue			= swap.issue_date();
		date maturity 		= swap.maturity();
		integer flt_freq	= swap.swap_coupon_freq_leg2();
		integer fix_freq	= swap.swap_coupon_freq_leg1();	
		day_count_method flt_dc = swap.swap_dc_method_leg2() ;		
		day_count_method fix_dc = swap.swap_dc_method_leg1();		
		return par_fixflt_swap_rate(d,cal,settle,bd,issue,maturity,eom,flt_freq,fix_freq,flt_dc,fix_dc,df_disc,f_flt_idx, flt_reset_days, curr_fix);	
	}
	//-------------------------------------------------------
	// par_fixflt_swap_rate
	//-------------------------------------------------------
	curve par_fixflt_swap_rate(date	 			d,
								curve 				swap_crv,
								disc_func			df_disc,
								fwd_func			f_flt_idx,
								integer 			flt_reset_days,
								number option(nullable) first_fix)
	{
		vector(instrument) vi = swap_crv.instruments();
		number size = v_size(vi);
		
		for (integer i = 0 ; i < size ; i++) {
			if(!vi[i].is_swap_fixflt())
				return null<curve>	;
			number y = par_fixflt_swap_rate(d,vi[i],df_disc,f_flt_idx, flt_reset_days,first_fix);
			swap_fixflt c = swap_fixflt(vi[i]);//cast
			c.add_quote(y,quote_style.YIELD);
			vi[i] = c;//CORE_INT.create_par_swap(vi[i],y, name);						
		}
		return curve(vi,swap_crv.name()); 	
	}

	
	
	

	/*-----------------------------------------------------------------------
	  Name :  	hedge_weights
	  buck_risk: 	bucket risk vector (for bp_shift = x), the risk in portfolio, per hedging instrument, that we would like to hedge
					expressed in currency
	  pvbp_m: 		sensitivity matrix, risk of a specific hedging instrument (for bp_shift = x) vs all hedging instruments i.e. buckets, for a certain nominal
					rows = instrument
					cols = bucket instruments
	  result: 		a weight vector for each hedging instrument
	  ----------------------------------------------------------------------*/
	vector(number) hedge_weights_alt(vector(number) buck_risk,
								 matrix(number) pvbp_m)
	{
		integer s = v_size(buck_risk);
		QL_FAIL_COND(s == 0,"invalid bucket risk vector");

		integer cols = n_cols(pvbp_m);
		integer rows = n_rows(pvbp_m);
		QL_REQUIRE(s == cols && s == rows,"invalid matrix size");

		matrix(number) pvbp_inv_m = invert(pvbp_m);
		vector(number) a[s] ;

		for(integer j =0 ;j<s;j++) {
			a[j] = 0;
			for(integer k = 0 ;k < s; k++) {

				QL_REQUIRE(!null(buck_risk[k]),"invalid bucket risk (null value in vector)");
				QL_REQUIRE(!null(pvbp_inv_m[k,j]),"invalid sensitivity element (null value)");
				a[j] += buck_risk[k] * pvbp_inv_m[k,j];
			}
		}

		return -a;
		
	}

	vector(number) hedge_weights(vector(number) buck_risk,
									matrix(number) pvbp_m)
	{
		integer s = v_size(buck_risk);
		QL_FAIL_COND(s == 0,"invalid bucket risk vector");

		integer cols = n_cols(pvbp_m);
		integer rows = n_rows(pvbp_m);
		QL_REQUIRE(s == cols && s == rows,"invalid matrix size");

		sparse_matrix pvbp_m_sp = sparse_matrix (s);

	
		for(integer i =0 ;i<s;i++) {			
			for(integer j = 0 ;j < s; j++) {		
				QL_REQUIRE(!null(pvbp_m[j,i]),"invalid sensitivity element (null value)");
				pvbp_m_sp.add(i, j, pvbp_m[j,i]);
			}
		}

		return pvbp_m_sp.solve(- buck_risk) ;
		
	}
}

module CORE_INT_FX
{
	
	/*-----------------------------------------------------------------------
	  Name :  	fx_cross_quote_init
	  Notes :   direct_cross has the perspective that curr1 is domestic i.e. if true we have a FORDOM quote
	 			direct1 and direct2 takes the perspective that the common currency is foreign i.e. the cross currencies are domestic
	  Changes:
	  			Date    		Who     Detail
	  			8-Dec-2013 		MN      Original code 
	  ----------------------------------------------------------------------*/	
	/*ex1 USDSEK vs USDNOK
	  ex2 USDNOK vs USDSEK
	  ...
	  ex8 USDUSD vs USDSEK ->kastar fel
	  ex9 SEKUSD vs USDNOK
	
	  ex1		ex2			ex3		ex4			ex5		ex6			ex7			ex8		ex9			
	  USD		USD			SEK*	SEK*		USD		USD			USD			USD		SEK*			
	  USD		USD			NOK		NOK			USD		USD			SEK			USD		USD			
	  SEK*		SEK*		USD		USD			NOK*	NOK*		NOK*		USD		USD			
	  NOK		NOK			USD		USD			SEK		SEK			USD			SEK		NOK
	  SEK		NOK			SEK		NOK			SEK		NOK			SEK			SEK		SEK
	  true		true		false	false		true	true		true		true	false
	  true		true		false	false		true	true		false		true	true
	  false		true		false	true		true	false		true		->fel	false
	  NOK		SEK			NOK		SEK			NOK		SEK			NOK					NOK
	  SEKNOK(id)NOKSEK(d)	SEKNOK(id)NOKSEK(d)	SEKNOK(d) NOKSEK(id) SEKNOK(d)			SEKNOK(id)
	*/

	logical fx_cross_quote_init(	string 		base1, 			
									string 		base2, 			
									string 		price1, 		
									string 		price2, 		
									string 		base_cross, 	
									out logical option(nullable) 	direct1,		
									out logical option(nullable) 	direct2, 		
									out logical option(nullable) 	direct_cross,	
									out string 	p,
									out string 	err)				
																
	{
		if(equal_casei(price1, base1)){
			err = "invalid input (base_ccy1 = price_ccy1)";
			return false;	
		}
	
		if(equal_casei(price2, base2)){
			err = "invalid input (base_ccy2 = price_ccy2)";
			return false;	
		}
	
		if(equal_casei(base1, base2)){ 				//1,2,5,6,8:true
			direct1 = direct2 = true;
			if(equal_casei(base_cross, price2)){	//2,5,8:true
				/*if(equal_casei(base1, price1)){
					err = "invalid input (base_ccy1 = price_ccy1)"
					return false;	
				}*/
				direct_cross = true;			
				p = price1;							//2:SEK, 5:NOK		
			}
			else if(equal_casei(base_cross, price1)){//1,6:true					
				if(equal_casei(base1, price2)){
					err = "invalid input (base_ccy1 = price_ccy2)";
					return false;	
				}
				direct_cross = false;			
				p = price2;							//1:NOK, 6:SEK
			}		
			else{
				err = "invalid base currency in cross";
				return false;	
			}	
		}
		else if(equal_casei(price1, price2)){		//3,4:true
			direct1 = direct2 = false;
			if(equal_casei(base_cross, base2)){		//4:true		
				/*if(equal_casei(price1, base1)){
					err = "invalid input (base_ccy1 = price_ccy1)"
					return false;	
				}*/
				direct_cross = true;					
				p = base1;							//4:SEK
			}	
			else if(equal_casei(base_cross, base1)){//3:true
				if(equal_casei(price1, base2)){
					err = "invalid input (base_ccy2 = price_ccy1)";
					return false;	
				}
				direct_cross = false;	
				p = base2;							//3:NOK
			}			
			else{
				err = "invalid base currency in cross";
				return false;	
			}	
		}
		else if(equal_casei(price1, base2)){		//9:true
			direct1 = false;
			direct2 = true;
			if(equal_casei(base_cross, price2)){					
				/*if(equal_casei(price1, base1)){
					err = "invalid input (base_ccy1 = price_ccy1)"
					return false;	
				}*/
				direct_cross = true;
				p = base1;
			}		
			else if(equal_casei(base_cross, base1)){//9:true
				if(equal_casei(price1, price2)){
					err = "invalid input (price_ccy1 = price_ccy2)";
					return false;	
				}
				direct_cross = false;
				p = price2;
			}			
			else{
				err = "invalid base currency in cross";
				return false;	
			}	
		}
		else if(equal_casei(price2, base1)){		//7:true
			direct1 = true;
			direct2 = false;
			if(equal_casei(base_cross, price1)){	
				/*if(equal_casei(price2, base2)){
					err = "invalid input (base_ccy2 = price_ccy2)";
					return false;	
				}*/
				direct_cross = false;
				p = base2;							
			}			
			else if(equal_casei(base_cross, base2)){//7:true
				if(equal_casei(price2, price1))
					return false;
				direct_cross = true;
				p = price1;							//7:NOK
			}	
			else{
				err = "invalid base currency in cross";
				return false;	
			}	
	
		}
		else{
			err = "invalid input";
			return false;	
		}	
	
		return true;	
	}
	/*-----------------------------------------------------------------------
	  Name :  	fx_cross_mid
	  Notes :  	direct = domestic currency per unit of the foreign currency (base currency on bottom and the term currency on top)
	 					USDSEK --> SEK = domestic, USD = foreign --> direct quotation        
	  Changes:
	  			Date    		Who     Detail
	  			8-Dec-2013 		MN      Original code (moved from fiws)
	  ----------------------------------------------------------------------*/	
	number fx_cross_mid(number 	ccy1_mid, 				 
						logical ccy1_direct,
						number  ccy1_scale, 
						number 	ccy2_mid, 						 
						logical	ccy2_direct, 		 
						number 	ccy2_scale, 
						logical cross_direct, 
						number 	cross_scale,
						number 	cross_qu_dec)
	{    				    			
		ccy1_mid /= ccy1_scale;
		ccy2_mid /= ccy2_scale;
	
	    number cross;
		if(ccy1_direct && ccy2_direct) 
		    cross = ccy1_mid/ccy2_mid ;
		else if(!ccy1_direct && !ccy2_direct)  
		    cross = ccy2_mid/ccy1_mid ;	    
		else if(ccy1_direct)  
		    cross = ccy1_mid*ccy2_mid ;	  
		else 
		    cross = 1.0/(ccy1_mid*ccy2_mid) ;
		   
		if(!cross_direct)     
		    cross = 1.0/cross;		
	
		cross *= cross_scale;
		cross = round (cross, integer(cross_qu_dec));
		return cross;	
	}
	/*-----------------------------------------------------------------------
	  Name :  	fx_cross_quote_mid
	  Notes :      
	  Changes:
	  			Date    		Who     Detail
	  			8-Dec-2013 		MN      Original code 
				25-Apr-2018		MN		Adapt this legacy func to new fxspot (fx_spot --> fxspot)
	  ----------------------------------------------------------------------*/	
	number fx_cross_quote_mid(	fx_spot		ccy1, 
								fx_spot		ccy2,
								string 		base_ccy_cross,
								logical 	unit_quote_cross,
								out fx_spot option(nullable) fx_cross)//this instrument will be null if the cross does not exist in the database
	{	
		//should be able to handle an outright
		
		QL_FAIL_COND(ccy1.trade_date() != ccy2.trade_date(),"trade date mismatch");
		QL_FAIL_COND(ccy1.spot_date() != ccy2.spot_date(),"settlement date mismatch");

		string base1	= ccy1.base_ccy();
		string price1	= ccy1.price_ccy();
		number q1 		= ccy1.quote(true, false);
	
		string base2	= ccy2.base_ccy();
		string price2	= ccy2.price_ccy();
		number q2 		= ccy2.quote(true, false);
			
		logical direct1,direct2, direct_cross;
		string price_ccy_cross, err_msg;		
		if(!fx_cross_quote_init(base1, base2, price1, price2, base_ccy_cross, direct1,direct2, 
								direct_cross, price_ccy_cross, err_msg))
			QL_FAIL(err_msg);
	
		number scale_cross = 1;
		number qu_dec_cross = 15;
		try {	//if the cross is in the database use the info on it

			//fx_cross = create_fx_spot_inst(	base_ccy_cross,price_ccy_cross,ccy1.trade_date(), 
											//ccy1.quote_side(), false);
			
			error_info error = error_info();
			fx_cross = fx_spot(	base_ccy_cross,price_ccy_cross, ccy1.trade_date() ,ccy1.quote_side(),error);
			QL_FAIL_COND(error.is_error(),"");

			//instrument_name ccy_cross_str = fx_spot_db_name(base_ccy_cross,price_ccy_cross);	
			//fx_cross = new fx_spot(ccy_cross_str, ccy1.inst().trade_date(), ccy1.inst().quote_side(), true);
			if(fx_cross.scale() != -1)
				scale_cross = fx_cross.scale();
			if(fx_cross.quote_dec() != -1)
				qu_dec_cross = fx_cross.quote_dec();
			else if(fx_cross.pip_decimals() != -1)
				qu_dec_cross = fx_cross.pip_decimals();	
		}
		catch {}
	
		number fx_mid = fx_cross_mid(q1, direct1,1, q2, direct2, 1, direct_cross, scale_cross , qu_dec_cross);
		if(!null(fx_cross))
			fx_cross.set_quote(fx_mid, false);

		if(unit_quote_cross) 
			return fx_mid/scale_cross ;
		else
			return fx_mid;	
		
	}

	/*-----------------------------------------------------------------------
	  Name :  	fx_cross_quote_mid
	  Notes :      
	  Changes:
	  			Date    		Who     Detail
	  			15-Aug-2016 	MN      Original code 
				25-Apr-2018		MN		Adapt this legacy func to new fx_spot (fx_spot --> fx_spot)
	  ----------------------------------------------------------------------*/	
	number fx_cross_quote_mid(	fx_spot		ccy1, 
								fx_spot		ccy2,
								logical 	ccy2_is_base,
								logical 	unit_quote_cross,
								out fx_spot option(nullable) fx_cross)//this instrument will be null if the cross does not exist in the database
	{
		string base1	= ccy1.base_ccy();
		string price1	= ccy1.price_ccy();	
		string base2	= ccy2.base_ccy();
		string price2	= ccy2.price_ccy();	
		string 	base_ccy_cross;
		
		if(equal_casei(base1, base2))
			base_ccy_cross = ccy2_is_base ? price2: price1;
		else if(equal_casei(base1, price2))
			base_ccy_cross = ccy2_is_base ? base2: price1;
		else if(equal_casei(base2, price1))
			base_ccy_cross = ccy2_is_base ? price2 : base1;
		else if(equal_casei(price1, price2))
			base_ccy_cross = ccy2_is_base ? base2 : base1;
		else
			QL_FAIL("invalid input");

		return fx_cross_quote_mid(ccy1,ccy2,base_ccy_cross,unit_quote_cross,fx_cross);
	}
	/*-----------------------------------------------------------------------
	  Name :  	fx_cross_quote_mid
	  Notes :      
	  Changes:
	  			Date    		Who     Detail
	  			8-Dec-2013 		MN      Original code 
				25-Apr-2018		MN		Adapt this legacy func to new fx_spot (fx_spot --> fx_spot)
	  ----------------------------------------------------------------------*/	
	number fx_cross_quote_mid(	instrument 	i1, 
								instrument 	i2,
								string 		base_ccy_cross,
								logical 	unit_quote_cross,
								out fx_spot option(nullable) fx_cross)//this instrument will be null if the cross does not exist in the database
	{
		fx_spot ccy1 = fx_spot(i1);
		fx_spot ccy2 = fx_spot(i2);	
		return fx_cross_quote_mid(ccy1, ccy2,base_ccy_cross,unit_quote_cross,fx_cross);
	}
	/*-----------------------------------------------------------------------
	  Name :  	fx_cross_quote_mid
	  Notes :      
	  Changes:
	  			Date    		Who     Detail
	  			15-Aug-2016 	MN      
				25-Apr-2018		MN		Adapt this legacy func to new fx_spot (fx_spot --> fx_spot) 
	  ----------------------------------------------------------------------*/	
	number fx_cross_quote_mid(	instrument 	i1, 
								instrument 	i2,
								logical 	i2_is_base,
								logical 	unit_quote_cross,
								out fx_spot option(nullable) fx_cross)//this instrument will be null if the cross does not exist in the database
	{
		fx_spot ccy1 	= fx_spot(i1);
		fx_spot ccy2 	= fx_spot(i2);	
		return fx_cross_quote_mid(ccy1, ccy2,i2_is_base,unit_quote_cross,fx_cross);
	}

	/*-----------------------------------------------------------------------
	  Name :  	fx_cross
	  Notes :   direct = domestic currency per unit of the foreign currency
			 			 USDSEK --> SEK = domestic, USD = foreign --> direct quotation    
	  Changes:
	  			Date    		Who     Detail
	  			8-Dec-2013 		MN      Original code (moved from fiws)
	  ----------------------------------------------------------------------*/	
	void fx_cross(	number 		bid1, 
					number 		ask1, 
					logical 	direct1,
					number  	scale1, 
					number 		bid2, 
					number 		ask2, 		 
					logical		direct2, 		 
					number 		scale2, 
					logical 	direct_cross, 
					number 		scale_cross,
					number 		qu_dec_cross, 
					logical 	fix_spread_cross, 
					number option(nullable) 	bid_spread_cross, 
					number option(nullable)		ask_spread_cross, 
					out number bid_cross,
					out number ask_cross)
	{    
					    
		bid1 /= scale1;
		ask1 /= scale1;
		bid2 /= scale2;
		ask2 /= scale2;
	    
		if(direct1 && direct2) {
		    bid_cross = bid1/ask2 ;
		    ask_cross = ask1/bid2 ;
		}
		else if(!direct1 && !direct2)  {
		    bid_cross = bid2/ask1 ;
		    ask_cross = ask2/bid1 ;
		}
		else if(direct1)  {
		    bid_cross = bid1*bid2 ;
		    ask_cross = ask1*ask2 ;
		}
		else {
		    bid_cross = 1.0/(ask1*ask2) ;
		    ask_cross = 1.0/(bid1*bid2) ;
		} 
		if(!direct_cross) {         
		    number tmp = bid_cross;	    
		    bid_cross = 1.0/ask_cross;
		    ask_cross = 1.0/tmp ;
		}
	
		bid_cross *= scale_cross;
		ask_cross *= scale_cross;			
	
		if (fix_spread_cross) {
		    number mid = (bid_cross + ask_cross)/2.0;
		    bid_cross = mid - bid_spread_cross;
		    ask_cross = mid + ask_spread_cross;
			bid_cross = round (bid_cross, integer(qu_dec_cross));
			ask_cross = round (ask_cross, integer(qu_dec_cross));
		}
		else {
			bid_cross = round (bid_cross, integer(qu_dec_cross));
			ask_cross = round (ask_cross, integer(qu_dec_cross));
		}
		return ;
			
	}
	
	/*-----------------------------------------------------------------------
	  Name :  	fx_cross_quote
	  Notes :      
	  Changes:
	  			Date    		Who     Detail
	  			10-Mar-2014 	MN      Original code 
				25-Apr-2018		MN		Adapt this legacy func to new fx_spot (fx_spot --> fx_spot) 
	  ----------------------------------------------------------------------*/	
	void _fx_cross_quote(	fx_spot		ccy1_bid,
							fx_spot		ccy1_ask, 
							fx_spot		ccy2_bid,
							fx_spot		ccy2_ask,
							string 		base_ccy_cross,
							logical 	unit_quote_cross,
							logical 	fix_spread_cross, 
							number option(nullable) 	bid_spread_cross, 
							number option(nullable)		ask_spread_cross,
							out number bid_cross,
							out number	ask_cross,
							out string price_ccy_cross,
							out fx_spot option(nullable) fx_cross_bid,//fx_cross_bid and fx_cross_ask will be null if the cross does not exist in the database
							out fx_spot option(nullable) fx_cross_ask)

	{	
		//should be able to handle an outright
		date trade = ccy1_bid.trade_date();
		QL_FAIL_COND(trade  != ccy2_bid.trade_date(),"trade date mismatch");
		date settle = ccy1_bid.spot_date();
		QL_FAIL_COND(settle  != ccy2_bid.spot_date(),"settlement date mismatch");
		QL_FAIL_COND(settle  != ccy1_ask.spot_date(),"settlement date mismatch");
		QL_FAIL_COND(settle  != ccy2_ask.spot_date(),"settlement date mismatch");
		QL_FAIL_COND(ccy1_bid.scale() != ccy1_ask.scale(),"incompatible bid - ask instrument (instrument 1 scale)");
		QL_FAIL_COND(ccy2_bid.scale() != ccy2_ask.scale(),"incompatible bid - ask instrument (instrument 2 scale)");
		QL_FAIL_COND(ccy1_bid.pip_decimals() != ccy1_ask.pip_decimals(),"incompatible bid - ask instrument (instrument 1 decimals)");
		QL_FAIL_COND(ccy2_bid.pip_decimals() != ccy2_ask.pip_decimals(),"incompatible bid - ask instrument (instrument 2 decimals)");

		string base1	= ccy1_bid.base_ccy();
		string price1	= ccy1_bid.price_ccy();
		number bid1 	= ccy1_bid.quote(true, false);
		number ask1 	= ccy1_ask.quote(true, false);

		string base2	= ccy2_bid.base_ccy();
		string price2	= ccy2_bid.price_ccy();
		number bid2 	= ccy2_bid.quote(true, false);
		number ask2 	= ccy2_ask.quote(true, false);
		
		logical direct1,direct2, direct_cross;
		string err_msg;		
		if(!fx_cross_quote_init(base1, base2, price1, price2, base_ccy_cross, direct1,direct2, direct_cross, price_ccy_cross, err_msg))
			throw(E_INVALID_ARG, err_msg);
	
		number scale_cross = 1;
		number qu_dec_cross = 15;
		try {	//if the cross is in the database use the info on it	
			error_info error = error_info();	
			fx_cross_bid = fx_spot(	base_ccy_cross,price_ccy_cross,ccy1_bid.trade_date(), 
									ccy1_bid.quote_side(),error);
			QL_FAIL_COND(error.is_error(),"");
			fx_cross_ask = fx_spot(	base_ccy_cross,price_ccy_cross,ccy1_ask.trade_date(), 
									ccy1_ask.quote_side(), error);
			
			QL_FAIL_COND(error.is_error(),"");

			if(fx_cross_bid.scale() != -1)
				scale_cross = fx_cross_bid.scale();
			if(fx_cross_bid.quote_dec() != -1)
				qu_dec_cross = fx_cross_bid.quote_dec();
			else if(fx_cross_bid.pip_decimals() != -1 )
				qu_dec_cross = fx_cross_bid.pip_decimals();	
		}
		catch {}
			
		fx_cross(	bid1,ask1,direct1,1,bid2,ask2,direct2,1, 
					direct_cross,scale_cross,qu_dec_cross, 
					fix_spread_cross,bid_spread_cross,ask_spread_cross, 
					bid_cross,ask_cross);

		if(!null(fx_cross_bid))
			fx_cross_bid.set_quote(bid_cross, null<date>,false);

		if(!null(fx_cross_ask))
			fx_cross_ask.set_quote(ask_cross, null<date>,false);

		if(unit_quote_cross) {
			bid_cross /= scale_cross;
			ask_cross /= scale_cross;
		}
	}

	/*-----------------------------------------------------------------------
	  Name :  	fx_cross_quote
	  Notes :      
	  Changes:
	  			Date    		Who     Detail
	  			10-Mar-2014 	MN      Original code 
				25-Apr-2018		MN		Adapt this legacy func to new fx_spot (fx_spot --> fx_spot) 
	  ----------------------------------------------------------------------*/	
	void _fx_cross_quote(	instrument	ccy1_bid,
							instrument	ccy1_ask, 
							instrument	ccy2_bid,
							instrument	ccy2_ask,
							string 		base_ccy_cross,
							logical 	unit_quote_cross,
							logical 	fix_spread_cross, 
							number option(nullable) 	bid_spread_cross, 
							number option(nullable)		ask_spread_cross,
							out number bid_cross,
							out number	ask_cross,
							out string  option(nullable) price_ccy_cross,
							out fx_spot option(nullable) fx_cross_bid,
							out fx_spot option(nullable) fx_cross_ask)
	{
		fx_spot	ccy1_bid_ = fx_spot(ccy1_bid);
		fx_spot	ccy1_ask_ = fx_spot(ccy1_ask);
		fx_spot	ccy2_bid_ = fx_spot(ccy2_bid);
		fx_spot	ccy2_ask_ = fx_spot(ccy2_ask);
			
		_fx_cross_quote(	ccy1_bid_,ccy1_ask_,ccy2_bid_,ccy2_ask_,base_ccy_cross,unit_quote_cross,
						fix_spread_cross,bid_spread_cross,ask_spread_cross,bid_cross,ask_cross,price_ccy_cross,
						fx_cross_bid,fx_cross_ask);
	}
}

/*-----------------------------------------------------------------------
  Name :  	fx_spot_quote
  Notes :   handles a possible reverse of bid ask sides   
  Changes:
  			Date    		Who     Detail
  			10-Mar-2014 	MN      Original code 
			25-Apr-2018		MN		Adapt this legacy func to new fx_spot (fx_spot --> fx_spot) 
  ----------------------------------------------------------------------*/	
void fx_spot_quote(	fx_spot fx_bid, 
					fx_spot fx_ask, 
					string base_ccy, 
					logical unit_quote,
					out number bid, 
					out number ask)

{
	try {
		QL_FAIL_COND(fx_ask.base_ccy() != fx_bid.base_ccy(),"invalid fx spot (different base currencies)");
		QL_FAIL_COND(fx_ask.price_ccy() != fx_bid.price_ccy(),"invalid fx spot (different price currencies)");

		if(base_ccy == fx_bid.base_ccy()){
			bid = fx_bid.quote(unit_quote, false);
			ask = fx_ask.quote(unit_quote, false);			
		}
		else if(base_ccy == fx_bid.price_ccy()){//reverse bid and ask		
			ask = fx_bid.quote(unit_quote, false);
			bid = fx_ask.quote(unit_quote, false);
		}
		else			
			QL_FAIL("incompatible base currency");			
		
	}
	catch{
		CORE_INT.write_warning_message(err.message());		
		bid = ask = null<number>;
	}	
}

/*-----------------------------------------------------------------------
  Name :  	fx_cross_quote
  Notes :      
  Changes:
  			Date    		Who     Detail
  			8-Dec-2013 		MN      Original code 
			25-Apr-2018		MN		Adapt this legacy func to new fx_spot (fx_spot --> fx_spot) 
  ----------------------------------------------------------------------*/	
void fx_cross_quote(	fx_spot 				i1_bid,
						fx_spot				i1_ask,  
						fx_spot				i2_bid,
						fx_spot				i2_ask,
						string 				base_ccy_cross,
						logical				unit_quote_cross,
						logical 			fix_spread_cross, 
						number option(nullable) 		bid_spread_cross, 
						number option(nullable)			ask_spread_cross,
						out number 	bid_cross,
						out number		ask_cross,
						out string  option(nullable)	price_ccy_cross,
						out fx_spot option(nullable) 	fx_cross_bid,
						out fx_spot option(nullable) 	fx_cross_ask)
{
	try{
		CORE_INT_FX._fx_cross_quote(i1_bid,i1_ask,i2_bid, i2_ask,
								base_ccy_cross,unit_quote_cross,
								fix_spread_cross,bid_spread_cross,ask_spread_cross,
								bid_cross,ask_cross,price_ccy_cross,fx_cross_bid,fx_cross_ask);
		
		return ;	
	}
	catch {
		CORE_INT.write_warning_message(err.message());
		bid_cross = null<number>;
		ask_cross = null<number>;	
	}
}

/*-----------------------------------------------------------------------
  Name :  	fx_cross_quote
  Notes :      
  Changes:
  			Date    		Who     Detail
  			8-Dec-2013 		MN      Original code 
			25-Apr-2018		MN		Adapt this legacy func to new fx_spot (fx_spot --> fx_spot) 
  ----------------------------------------------------------------------*/	
void fx_cross_quote(	instrument 			i1_bid,
						instrument 			i1_ask,  
						instrument 			i2_bid,
						instrument 			i2_ask,
						string 				base_ccy_cross,
						logical				unit_quote_cross,
						logical 			fix_spread_cross, 
						number option(nullable) 		bid_spread_cross, 
						number option(nullable)			ask_spread_cross,
						out number 	bid_cross,
						out number		ask_cross,
						out string 	price_ccy_cross,
						out fx_spot option(nullable) 	fx_cross_bid,
						out fx_spot option(nullable) 	fx_cross_ask)
{
	try{
		CORE_INT_FX._fx_cross_quote(i1_bid,i1_ask,i2_bid, i2_ask,
								base_ccy_cross,unit_quote_cross,
								fix_spread_cross,bid_spread_cross,ask_spread_cross,
								bid_cross,ask_cross,price_ccy_cross,fx_cross_bid,fx_cross_ask);
		
		return ;	
	}
	catch {
		CORE_INT.write_warning_message(err.message());
		bid_cross = null<number>;
		ask_cross = null<number>;	
	}
}

/*-----------------------------------------------------------------------
  Name :  	fx_cross_quote_v
  Notes :      
  Changes:
  			Date    		Who     Detail
  			11-Mar-2014 	MN      Original code 
			25-Apr-2018		MN		Adapt this legacy func to new fx_spot (fx_spot --> fx_spot) 
  ----------------------------------------------------------------------*/	
void fx_cross_quote_v(		date 				d,
							vector(string) 		currencies,
							string				base_ccy,							
							string option(nullable)	link_ccy,
							quote_side			qs_bid,
							quote_side			qs_ask,	
							logical				invert,						
							out vector(string) option(nullable) cross_name,
							out vector(number)   bid,
							out vector(number)   ask,
							out vector(number)   sprd)

{
	try{
		resize(bid,0);
		resize(ask,0);
		resize(cross_name,0);
		resize(sprd,0);
		

		if(v_size(currencies) <= 0)
			return;
			
		base_ccy = str_to_upper(base_ccy);
		link_ccy = null(link_ccy) ? base_ccy : str_to_upper(link_ccy);		

		vector(fx_spot) id_cross_bid, id_cross_ask;
		number prim_idx = -1;

		for(integer i=0; i< v_size(currencies); i++) {
			currencies[i] = str_to_upper(currencies[i]);
	
			if( currencies[i] == base_ccy){
				push_back(id_cross_bid,null<fx_spot>);
				push_back(id_cross_ask,null<fx_spot>);
				//if(link_ccy == base_ccy)
				//	prim_idx = i;
			}	
			else if(currencies[i] == link_ccy){
				push_back(id_cross_bid,null<fx_spot>);
				push_back(id_cross_ask,null<fx_spot>);				
				//if(link_ccy != base_ccy)
					prim_idx = i;
			}	
			else  {
				error_info error = error_info();
				string fx_name = strcat([link_ccy, currencies[i]]);				
				fx_spot fx_bid = fx_spot(link_ccy, currencies[i],d, qs_bid, error);
				//fx_spot fx_bid = CORE_INT.create_fx_spot_inst_flex(link_ccy, currencies[i],d, qs_bid, false);
				QL_FAIL_COND(error.is_error(),strcat(fx_name, "-cross not available"));

				if(!null(fx_bid)){					
					push_back(id_cross_bid,fx_bid);
					fx_spot fx_ask = fx_spot(fx_bid.name(), d, qs_ask); 
					push_back(id_cross_ask,fx_ask);						
				}
				else {
					push_back(id_cross_bid,null<fx_spot>);
					push_back(id_cross_ask,null<fx_spot>);
				}
			}					
		}
	
		if(link_ccy == base_ccy) {
				
			for(integer i=0; i< v_size(id_cross_bid); i++) {
				
				if(null(id_cross_bid[i]) || null(id_cross_ask[i])) {
					push_back(bid, null<number>);
					push_back(ask, null<number>);					
					push_back(sprd, null<number>);
					push_back(cross_name,null<string>);	
				}
				else {
					string price_ccy = id_cross_bid[i].base_ccy() == base_ccy? id_cross_bid[i].price_ccy() : id_cross_bid[i].base_ccy();
					string base_ccy_cross = invert ? price_ccy : base_ccy;
					string price_ccy_cross = invert ? base_ccy : price_ccy;
					
					number b,a;
					fx_spot_quote(id_cross_bid[i], id_cross_ask[i],base_ccy_cross,false,b,a);//handles a possible reverse of bid ask sides		
					push_back(bid, b);
					push_back(ask, a);					
					push_back(sprd, id_cross_bid[i].pips(b,a));
					string fx_name = strcat([base_ccy_cross,price_ccy_cross]);	
					//string name = fx_spot_db_name(base_ccy_cross,price_ccy_cross, alt_name);
					push_back(cross_name,fx_name);
				}
			}
	
		}
		else {
			error_info error = error_info();
			string fx_name = strcat([link_ccy, base_ccy]);							
			fx_spot fx_prim_bid = fx_spot(link_ccy, base_ccy,d, qs_bid, error);
			QL_FAIL_COND(error.is_error(),strcat(fx_name,"-cross not available"));
			//fx_spot fx_prim_bid = CORE_INT.create_fx_spot_inst_flex(link_ccy, base_ccy,d, qs_bid, false);

			fx_spot fx_prim_ask;
			number prim_bid,prim_ask, prim_sprd;
			string prim_name; 
			if(!null(fx_prim_bid)){																			
				fx_prim_ask 			= fx_spot(fx_prim_bid.name(),d, qs_ask, error); 	
				//fx_prim_ask 			= new fx_spot(fx_prim_bid.name(), d, qs_ask, false);						
				string 	base_ccy_cross 	= invert ? link_ccy : base_ccy;
				string 	price_ccy_cross = invert ? base_ccy : link_ccy;							
				fx_spot_quote(fx_prim_bid, fx_prim_ask,base_ccy_cross,false,prim_bid,prim_ask);//handles a possible reverse of bid ask sides									
				prim_sprd 				= fx_prim_bid.pips(prim_bid,prim_ask);
				prim_name 				= strcat([base_ccy_cross,price_ccy_cross]);				
				//prim_name 				= fx_spot_db_name(base_ccy_cross,price_ccy_cross, alt_name);					
			}
			else {
				string alt_name;
				string err = strcat([fx_spot_db_name(base_ccy, link_ccy,alt_name)," and ", fx_spot_db_name(link_ccy, base_ccy,alt_name)," not in database"]);
				throw(E_INVALID_ARG, err);
			}
		
			for(integer i=0; i< v_size(id_cross_bid); i++) {
					
				if(null(id_cross_bid[i]) || null(id_cross_ask[i])) {
					if(i == prim_idx){
						push_back(bid, prim_bid);
						push_back(ask, prim_ask);					
						push_back(sprd, prim_sprd);
						push_back(cross_name,prim_name);	
					}
					else {
						push_back(bid, null<number>);
						push_back(ask, null<number>);					
						push_back(sprd, null<number>);
						push_back(cross_name,null<string>);	
					}
				}
				else {			
					string 	base_ccy_cross;		
					if(invert) {			
						base_ccy_cross = base_ccy == fx_prim_bid.base_ccy() ? fx_prim_bid.price_ccy() : fx_prim_bid.base_ccy(); 
						base_ccy_cross = base_ccy_cross == id_cross_bid[i].base_ccy() ? id_cross_bid[i].price_ccy() : id_cross_bid[i].base_ccy(); 					 				
					}
					else 
						base_ccy_cross = base_ccy;
								
					fx_spot fx_cross_bid,fx_cross_ask;
					string price_ccy_cross;
					number  bid_cross,ask_cross;
					fx_cross_quote(	fx_prim_bid,fx_prim_ask,id_cross_bid[i],id_cross_ask[i],
									base_ccy_cross,false,false,null<number>, null<number>,						
									bid_cross,ask_cross, price_ccy_cross,fx_cross_bid,fx_cross_ask);
			
					if(!null(fx_cross_bid) && !null(fx_cross_ask)){	
						bid_cross = fx_cross_bid.quote(false,false);
						ask_cross = fx_cross_ask.quote(false,false);						
						push_back(sprd, fx_cross_bid.pips(bid_cross,ask_cross));						
					}
					else 			
						push_back(sprd, ask_cross-bid_cross);					
					
					push_back(bid, bid_cross);
					push_back(ask, ask_cross);	
					string fx_name = strcat([base_ccy_cross,price_ccy_cross]);
					//string name = fx_spot_db_name(base_ccy_cross,price_ccy_cross, alt_name);
					push_back(cross_name,fx_name);
				}
			}						
		}
		return ;
	}
	catch {
		//CORE_INT.write_warning_message(err.message());		
		resize(bid,0);
		resize(ask,0);
		resize(cross_name,0);
		resize(sprd,0);	
		throw(E_INVALID_ARG, err.message());
	}
}



module CORE_INT
{
	

	//-------------------------------------------------------
	// bump_spread
	//-------------------------------------------------------
	logical bump_spread(number shift,
						vector(date) mat_v,
						out vector(number) bump_v,
						vector(integer) incl_v, integer i)
	{
		
		integer s_incl = v_size(incl_v);
		integer s_bump = v_size(bump_v);

		bump_v = 0;
		
		if(i > incl_v[s_incl-1]) {			
			return false;
		}
		
		if(s_incl == s_bump) {	//no ip		
			bump_v[i] = shift;
			return true;
		}
		
		integer incl_cnt_i = 0;
		
		if(i > incl_v[0]) {
			incl_cnt_i = 1;		
			for(integer k=1;k<s_incl;k++){
				if(i > incl_v[k] )
					incl_cnt_i++;
			}
		}
		

		//for(integer i=0;i<s_bump;i++){
			//if(incl_cnt == s_incl)
				//break;
		if(i != incl_v[incl_cnt_i])
			return false;
			
		//if(i == incl_v[incl_cnt_i]) {

			date curr_mat = mat_v[i];
			
			if(incl_cnt_i == s_incl-1){//last
				for(integer k=i;k<s_bump;k++)
					bump_v[k] = shift;

				if(incl_cnt_i > 0) {
					integer idx_prev 	= incl_v[incl_cnt_i-1];
					date prev_mat 		= mat_v[idx_prev];
					integer d_prev 		= prev_mat - curr_mat;
					interpolation ip 	= interpolation(ip_linear(), [d_prev,0], [0,shift]);
					for(integer k=idx_prev+1;k<i;k++) {
						integer d = mat_v[k] - curr_mat;
						bump_v[k] = ip.eval(d);
					}
				}
				else {
					for(integer k=0;k<i;k++)
						bump_v[k] = shift;
				}
				
			}
			else if(incl_cnt_i == 0){//first incl
				for(integer k=0;k<=i;k++)
					bump_v[k] = shift;

				
				integer idx_next 	= incl_v[incl_cnt_i+1];
				date next_mat 		= mat_v[idx_next];
				integer d_next 		= next_mat - curr_mat;
				interpolation ip 	= interpolation(ip_linear(), [0,d_next], [shift,0]);
				for(integer k=i+1;k<idx_next;k++) {
					integer d = mat_v[k] - curr_mat;
					bump_v[k] = ip.eval(d);
				}
			}
		 
			else {
				integer idx_next 	= incl_v[incl_cnt_i+1];
				date next_mat 		= mat_v[idx_next];
				integer d_next 		= next_mat - curr_mat;
				integer idx_prev 	= incl_v[incl_cnt_i-1];
				date prev_mat 		= mat_v[idx_prev];
				integer d_prev 		= prev_mat - curr_mat;
				interpolation ip 	= interpolation(ip_linear(), [d_prev,0,d_next], [0,shift,0]);
				for(integer k=idx_prev+1;k<idx_next;k++) {
					integer d = mat_v[k] - curr_mat;
					bump_v[k] = ip.eval(d);
				}
			}
		//}
		return true;
	}
	//-------------------------------------------------------
	// instr_bumper_df_vect
	//-------------------------------------------------------
	vector(disc_func) instr_bumper_df_vect(	curve 			crv,
											disc_z_model 	disc_model,
											number 			y_shift,
											out vector(integer) incl,//input and output, size is equal to number of included instruments
											out disc_func	df_base,
											out disc_func	df_all_shift,
											out vector(instrument) bump_instr,//size is equal to number of included instruments 
											out vector(disc_func) df_v)//size is equal to number of instruments in the curve (=curve_)
	{
		df_base = df_all_shift = null<disc_func>;
		resize(bump_instr,0);

		vector(instrument) v_i 	= crv.instruments();
		integer size_all 		= v_size(v_i);
		QL_REQUIRE(size_all > 0,"invalid instrument vector");
		integer size_incl ;

		//check and maybe update the incl vector
		if(null(incl) || v_size(incl) == 0) {
			incl 		= index_vector(size_all);
			size_incl 	= size_all;
		}
		else {
			size_incl 	= v_size(incl);
			QL_REQUIRE(size_incl <= size_all,"invalid size of include vector");
			
			for(integer i=0;i<size_incl;i++){
				QL_REQUIRE(incl[i] >= 0 && incl[i] < size_all,"invalid index in include vector");
				if(i>0){
					QL_REQUIRE(incl[i] > incl[i-1],"invalid index order in include vector");
				}
			}
		}
		
		//logical no_shift_idx 	= false;//first_instr_is_fixed();	
		//vector(logical) incl_in_bump[size_all];
		//incl_in_bump			= true;
		
		vector(number) y_adj[size_all];
		y_adj 				= y_shift;			
		vector(disc_func) df_incl_v[0];
		resize(df_v,size_all);
		
		df_base 			= disc_model.df(crv, null<vector(number)>);
		curve c				= y_shift_curve(crv,y_adj);
		df_all_shift		= disc_model.df(c, null<vector(number)>);
		vector(date) mat_v 	=  v_i.maturity();
		
		for(integer i=0;i<size_all;i++){
			logical is_shifted = bump_spread(y_shift, mat_v, y_adj, incl, i);
			if(is_shifted) {
				push_back(bump_instr,v_i[i]);		
				c				= y_shift_curve(crv,y_adj);
				df_v[i] 		= disc_model.df(c, null<vector(number)>);
				push_back(df_incl_v,df_v[i]);
			}
			else {
				df_v[i] 		= df_base;
			}
			
		}
		QL_REQUIRE(v_size(bump_instr) == size_incl,"fatal inconsistency in calculations");
		return df_incl_v;
	}

	//-------------------------------------------------------
	// sens_matrix_instr_disc
	//-------------------------------------------------------
	matrix(number) sens_matrix_instr_disc(	number 				nominal,
											vector(instrument) 	instr_v,	//note: a swap here must be updated with fixings and disc_funcs
											fwd_func option(nullable) f,	//for swaps with flt-leg
											number 				y_shift,
											disc_func			df_base,
											disc_func			df_all_shift,													
											vector(disc_func) 	df_v)
	{
		QL_REQUIRE(nominal>0,"nominal for the sensitivity matrix must be positive");
		logical mm_borrow = true;
		logical swap_payfix = true;
		integer rc_size 	= v_size(instr_v);	
		error_info err 		= new error_info(true,true);
		integer cols		= v_size(df_v);

		QL_REQUIRE(rc_size == cols, "rows != columns");//check
		matrix(number) sens_m[rc_size,cols];
		vector(number) pv_diff_pert[cols];
		sens_m = 0;
		
		error_info ee = new error_info(true,true);
		pv_diff_pert 	= 0;
		integer j 		= 0;
		for(integer i=0;i<rc_size;i++) {
			
			instr_type it = instr_v[i].instr_type(ee);
			
			switch(it){
			case instr_type.DEPOSIT:
			case instr_type.DEPOSIT_INDEX:
			{
				number pv_diff_all;	
				deposit depo = deposit(instr_v[i],err);
				CORE_INT_DEPO.risk_ladder_disc(depo, mm_borrow, nominal,df_v, df_base, df_all_shift, pv_diff_pert, pv_diff_all,y_shift,err ) ;
				replace_row (sens_m, pv_diff_pert, j);
				break;
			}
			case instr_type.FRA:
			{
				QL_FAIL("work in progress");
				//number pv_diff_all;			
				//fra f = fra(instr_v[i],err);
				//CORE_INT.risk_ladder_disc(f, mm_borrow, nominal,df_v, df_base, df_all_shift, pv_diff_pert, pv_diff_all,err ) ;
				replace_row (sens_m, pv_diff_pert, j);
				break;
			}
			case instr_type.SWAP_FIXFLT:
			{
				number pv_diff_all;			
				swap_fixflt sw = swap_fixflt(instr_v[i],err);
				CORE_INT_FIXFLT.risk_ladder_disc_fixflt(sw, swap_payfix, nominal,f,df_v, df_base, df_all_shift, pv_diff_pert, pv_diff_all,y_shift,err ) ;
				replace_row (sens_m, pv_diff_pert, j);
				break;
			}
			case instr_type.SWAP_FIXOIS://FIX_ME:  risk_ladder_disc_fixois or risk_ladder_fixois
			{				
				number pv_diff_all;			
				swap_fixois sw = swap_fixois(instr_v[i],err);
				CORE_INT_FIXOIS.risk_ladder_disc_fixois(sw, swap_payfix, nominal,/*f,*/df_v, df_base, df_all_shift, pv_diff_pert, pv_diff_all,y_shift,err ) ;
				replace_row (sens_m, pv_diff_pert, j);
				break;
			}
			case instr_type.IR_FUTURE:
			{						
				ir_future fut = ir_future(instr_v[i],err);
				if(fut.is_daily_mtm(err))
					continue;
				
				QL_FAIL("work in progress");
				//number pv_diff_all;
				
				//logical is_instr = fut.style(err) == irfut_calc_method.IR_FUTURE;
				//CORE_INT.risk_ladder_disc(f, !is_instr, nominal,df_v, df_base, df_all_shift, pv_diff_pert, pv_diff_all,err ) ;
				replace_row (sens_m, pv_diff_pert, j);
				break;
			}
			default:			
				QL_FAIL('invalid instrument type');
			}
			j++;
		}

		return sens_m;
	}

	//-------------------------------------------------------
	// sens_matrix_instr_fwd
	//-------------------------------------------------------
	matrix(number) sens_matrix_instr_fwd(number 			nominal,										 
										vector(instrument) 	instr_v,//note: a swap here must be updated with fixings and disc_funcs
										disc_func			df_disc,
										fwd_func			fwd_base,
										fwd_func			fwd_all,
										vector(fwd_func) 	fwd_v,
										number 				y_shift)//i.e. rates up --> pv down
	{
		QL_REQUIRE(nominal>0,"nominal for the sensitivity matrix must be positive");
		logical mm_borrow = true;
		logical swap_payfix = mm_borrow;
		integer rc_size 	= v_size(instr_v);
		error_info err 		= new error_info(true,true);
		integer cols		= v_size(fwd_v);		
		
		QL_REQUIRE(rc_size == cols, "rows != columns");//check
		matrix(number) sens_m[rc_size,cols];
		vector(number) pv_diff_pert[cols];

		error_info ee = new error_info(true,true);
		pv_diff_pert 	= 0;
		integer j 		= 0;
		for(integer i=0;i<rc_size;i++) {

			instr_type it = instr_v[i].instr_type(ee);
			
			switch(it){
			case instr_type.DEPOSIT:
			case instr_type.DEPOSIT_INDEX:
			{
				number pv_diff_all;	
				deposit depo = deposit(instr_v[i],err);
				CORE_INT_DEPO.risk_ladder_fwd(depo, mm_borrow, nominal,df_disc,fwd_v, fwd_base, fwd_all, pv_diff_pert, pv_diff_all,y_shift,err ) ;
				replace_row (sens_m, pv_diff_pert, j);
				break;
			}
			case instr_type.FRA:
			{
				number pv_diff_all;			
				fra fr = fra(instr_v[i],err);
				CORE_INT_FRA.risk_ladder_fwd(fr, mm_borrow, nominal,fwd_v, fwd_base, fwd_all, pv_diff_pert, pv_diff_all,y_shift,err ) ;
				replace_row (sens_m, pv_diff_pert, j);
				break;
			}
			case instr_type.SWAP_FIXFLT:
			{
				number pv_diff_all;			
				swap_fixflt sw = swap_fixflt(instr_v[i],err);
				CORE_INT_FIXFLT.risk_ladder_fwd_fixflt(sw, swap_payfix, nominal,fwd_v, fwd_base, fwd_all, pv_diff_pert, pv_diff_all,y_shift,err ) ;
				replace_row (sens_m, pv_diff_pert, j);
				break;
			}
			case instr_type.SWAP_FIXOIS:
			{
				number pv_diff_all;			
				swap_fixois sw = swap_fixois(instr_v[i],err);
				CORE_INT_FIXOIS.risk_ladder_fwd_fixois(sw, swap_payfix, nominal,fwd_v, fwd_base, fwd_all, pv_diff_pert, pv_diff_all,y_shift,err ) ;
				replace_row (sens_m, pv_diff_pert, j);
				break;
			}
			case instr_type.IR_FUTURE:
			{
				number pv_diff_all;			
				ir_future fut = ir_future(instr_v[i],err);
				logical is_instr = fut.style(err) == irfut_calc_method.IR_FUTURE;
				CORE_INT_IRFUT.risk_ladder_fwd(fut, !is_instr, nominal,fwd_v, fwd_base, fwd_all, pv_diff_pert, pv_diff_all,y_shift,err ) ;
				replace_row (sens_m, pv_diff_pert, j);
				break;
			}
			default:			
				QL_FAIL('invalid instrument type');
			}
			
			j++;
		}

		return sens_m;
	}

	/*-----------------------------------------------------------------------
	  cf_present_value
	  ----------------------------------------------------------------------*/
	number cf_present_value(vector(date)		cash_flow_date,
							vector(number) 		fixed_cash_flow,
							date 				d,
							date option(nullable) pv_d,
							date option(nullable) cutoff_d,
							disc_func 			disc_func)
	{
		integer size = v_size(cash_flow_date);
		if(size==0)
			return null<number>;
		QL_REQUIRE(size == v_size(fixed_cash_flow),"invalid vector size");
		
		number dfs ;
		if(null(pv_d) || pv_d <= d){
			dfs = 1;
			pv_d = d;
		}
		else {
			dfs = disc_func.disc((pv_d-d)/365.0);
		}
		
		if(null(cutoff_d) || cutoff_d < d )
			cutoff_d = d;
		
		number sum = 0;

		for(integer i=0;i<size;i++){
			date cfd = cash_flow_date[i];
			if(cfd < cutoff_d)
				continue;
			number t = (cfd-d)/365.0;
			number df = disc_func.disc(t)/dfs;
			sum += fixed_cash_flow[i] * df;
		}
		return sum;
	}

	/*-----------------------------------------------------------------------
	  cf_flt_present_value
	  ----------------------------------------------------------------------*/
	number cf_flt_present_value(vector(date)		cash_flow_reset_date,
								vector(date)		cash_flow_start_date,
								vector(date)		cash_flow_mat_date,								
								vector(number) 		float_ir_cash_flow,//must be flat i.e. without a fixed spread
								vector(number) 		float_nom_cash_flow,
								day_count_method 	dc,
								date 				d,
								date option(nullable) pv_d,
								date option(nullable) cutoff_d,
								disc_func 			disc_func,
								fwd_func 			fwd_func)
	{
		integer size = v_size(cash_flow_mat_date);
		if(size==0)
			return null<number>;
		QL_REQUIRE(size == v_size(cash_flow_reset_date),"invalid vector size");
		QL_REQUIRE(size == v_size(cash_flow_start_date),"invalid vector size");
		QL_REQUIRE(size == v_size(float_ir_cash_flow),"invalid vector size");
		QL_REQUIRE(size == v_size(float_nom_cash_flow),"invalid vector size");
		
		number dfs ;
		if(null(pv_d) || pv_d <= d){
			dfs = 1;
			pv_d = d;
		}
		else {
			dfs = disc_func.disc((pv_d-d)/365.0);
		}
		
		if(null(cutoff_d) || cutoff_d < d )
			cutoff_d = d;
		
		number sum = 0;

		for(integer i=0;i<size;i++){
			date cfd = cash_flow_mat_date[i];
			if(cash_flow_mat_date[i] < cutoff_d)
				continue;

			number t = (cfd-d)/365.0;
			number df = disc_func.disc(t)/dfs;
				
			if(cash_flow_reset_date[i] <= d) {
				
				sum += float_ir_cash_flow[i] * df;
			}
			else {
				number tt = (cash_flow_start_date[i]-d)/365.0;
				number per = period(cash_flow_start_date[i],cash_flow_mat_date[i],dc);
				number r = fwd_func.fwd(tt);
				number cf = r * per * float_nom_cash_flow[i];
				sum += cf * df;
			}
		}
		return sum;
	}
}