/*	
	Core fx functions
	Developer: Algorithmica Research, Magnus Nyström	

	ToDo: need a global mid-code	

	Comments:
	an fx spot instrument in the database should have a name equal to "<basecurr><pricecurr>" eg. USDSEK 
	
	tags needed in the database:
	
	FX SPOT
	
		ccy_dec 			= decimals that defines a pip
		ccy_q_dec			= decimals in quotation	
		ccy_price 			= price currency						
		ccy_reverse_ric		= indicates reverse quotation (OBS bid-ask fields in realtime link must be changed accordingly) 			
		ccy_scale 			= number of units of base currency

		***ccy_price_calendar 	= calendar for price currency		-> REMOVE
		***ccy_base			= base currency 						-> REMOVE
		***ccy_spot_code	= spot code for currency pair 			-> REMOVE
	
	FX SWAPS, FX_FWD
		ccy_spot_inst_name	= name of associated spot instrument	-> REMOVE

	- 	the calendar in the general tab should be both base and price currency (fx_spot_rule = FXS_STD otherwise mismatches will occur)
		eg. the built-in member function settle_date() will be incorrect if only one calendar is defined 

	Note: US public holidays are normally now disregarded in value dates of cross transactions. Standard nowadays (after the Euro)?
		
*/
option(null: hard);	
//enum fx_spot_rule option (category: 'Curve and Instrument/FX') {FXS_STD, FXS_STD_USD_ADJ, FXS_LA};


module CORE_INT_COMPAT_FX
{


	
	enum fx_inst_type {FXI_SPOT, FXI_SWAP, FXI_FWD};


	vector(integer ) v_find_fx(vector(fx_inst_type) v, fx_inst_type v_ref)
	{
		vector(integer) v_out[0];
	
		for(integer i;i<v_size(v);i++) {		
			if (v[i] == v_ref)	
				push_back(v_out, i);				
		}	
		return v_out;
	}
	
	vector(fx_inst_type) sub_vector_fx(vector(fx_inst_type) v, vector(integer) index)
	{
		vector(fx_inst_type) v_out[0];
	
		for(integer i;i<v_size(index);i++) 		
			push_back(v_out, v[index[i]]);				
		
		return v_out;
	}

	/*-----------------------------------------------------------------------
	  Name :  	class _fx
	  Notes :   a root container for fx (abstract)    
	  Changes:
	  			Date    		Who     Detail
	  			8-Dec-2013 		MN      Original code 
	  ----------------------------------------------------------------------*/	

	class _fx 
	{
	
		protected:
			_fx();
			_fx(date,number ,number ,number option(nullable),string,string,string,calendar,calendar,calendar option(nullable), 
				fx_spot_rule.fx_spot_rule,number option(nullable), string );
			_fx(instrument, logical allow_defaults = false);
			_fx(instrument_name, date t = today(), quote_side option(nullable) qs = null<string>, logical allow_defaults = false);
			_fx(_fx) ;
	
			void					init(instrument, logical allow_defaults = false);
			void					init(instrument_name,date,quote_side option(nullable),logical allow_defaults = false);
			void 					init(_fx ) ;
			
			logical					is_ok(out string option(nullable) err);
			//logical				check();		
			number 					spot_adjustment(number,number option(nullable));	
			
		public:		
			virtual _fx				clone() 											= 0;	
			virtual fx_inst_type 	type() 												= 0;
			virtual string			name();
			string					spot_name();		
			virtual number			quote(logical, logical);
			virtual number			quote() 											= 0;		
			number					spot_quote(logical, logical );
			number					spot_quote();
			virtual void			set_quote(number option(nullable), logical unit_quote = false );
			void					set_spot_quote(number option(nullable), logical unit_quote = false);
	
			virtual number 			swap();
			virtual number 			swap(logical, logical);
			virtual number 			swap(logical, number option(nullable), logical)	;
			virtual number 			swap(disc_func,disc_func,logical u = false,number option(nullable) sa = 0, logical nr = false) ;
	
			virtual number			outright( )	;
			virtual number			outright(logical,logical )	;
			virtual number			outright(logical,number option(nullable),logical) ;
			virtual number 			outright(disc_func,disc_func,logical u = false,number option(nullable) sa = 0, logical nr = false);
	
			virtual date			maturity() 											= 0;
			date					maturity(date, string);
			virtual date			settle_date() 										= 0;
			instrument 				spot_inst() ;		
			virtual instrument 		inst() 												= 0;
			date					trade_date();
			number					scale();
			number					dec();
			virtual number 			qu_dec();
			virtual number 			spot_qu_dec();
			string 					base_ccy();
			string  				price_ccy();
			logical 				reverse_ric();
			string 					spot_code();
			calendar  				cal_base();
			calendar  				cal_price();
			calendar  				cal_ny();
			fx_spot_rule.fx_spot_rule 			rule();			
			date					spot_settle();		
			virtual logical			is_post_spot()										= 0;	
	
		private:	
			//void		set_default();
			void		initialize(logical allow_defaults = false);
			void		initialize_no_inst();
			date		trade_date_;	
			instrument 	fx_spot_;
			number 		scale_;
			number 		dec_;
			number 		qu_dec_;
			string 		base_ccy_;
			string  	price_ccy_;
			logical 	reverse_ric_;
			string 		spot_code_;
			calendar  	cal_base_;
			calendar  	cal_price_;
			calendar	cal_ny_;
			fx_spot_rule.fx_spot_rule rule_;
			number		q_;
			string		name_;
			logical 	ok_;
			string 		err_;
	};
	
	_fx._fx(date		d,
			number 		scale,
			number 		dec,
			number option(nullable)	qu_dec,
			string 		base_ccy,
			string  	price_ccy,
			string 		spot_code,
			calendar  	cal_base,
			calendar  	cal_price,
			calendar  	option(nullable) cal_ny,
			fx_spot_rule.fx_spot_rule rule,
			number option(nullable) q,
			string		name) 
			:trade_date_(d), fx_spot_(null<instrument>), scale_(scale),dec_(dec),qu_dec_(qu_dec),base_ccy_(str_to_upper(base_ccy)), 
			price_ccy_(str_to_upper(price_ccy)), reverse_ric_(false), spot_code_(spot_code),cal_base_(cal_base), cal_price_(cal_price), 
			cal_ny_(cal_ny), rule_(rule), q_(q), name_(name), ok_(false), err_("")
	{
		initialize_no_inst() ;	
	}
	
	_fx._fx(): ok_(false)
	{
		//set_default();
	}
	
	_fx._fx(instrument fx,
			logical allow_defaults):fx_spot_(fx), ok_(false)
	{
		initialize(allow_defaults);
	}
	
	_fx._fx(instrument_name 			fx_name, 
			date 						trade_date,
			quote_side option(nullable) quote_side,
			logical 					allow_defaults): ok_(false)
	{
		fx_spot_ = instrument(fx_name, trade_date, quote_side);	
		initialize(allow_defaults);
	}
	
	_fx._fx(_fx f) 
			:trade_date_(f.trade_date_), fx_spot_(f.fx_spot_), scale_(f.scale_),dec_(f.dec_),qu_dec_(f.qu_dec_),
			base_ccy_(f.base_ccy_), price_ccy_(f.price_ccy_), reverse_ric_(f.reverse_ric_), 
			spot_code_(f.spot_code_),cal_base_(f.cal_base_), cal_price_(f.cal_price_), cal_ny_(f.cal_ny_), 
			rule_(f.rule_), q_(f.q_), name_(f.name_),ok_(f.ok_), err_(f.err_)
	{}
	
	void _fx.init(	instrument 	fx, 
					logical 	allow_defaults) 
	{
		fx_spot_ = fx;
		initialize(allow_defaults);
	}
	
	void _fx.init(	instrument_name 			fx_name,
					date 						d,
					quote_side option(nullable) quote_side,
					logical 					allow_defaults)
	{
		fx_spot_ = instrument(fx_name, d, quote_side);	
		initialize(allow_defaults);
	}
	
	void _fx.init(_fx f) 		
	{
		if(this == f)
			return;
		trade_date_	= f.trade_date_;
		fx_spot_ 	= f.fx_spot_;
		scale_ 		= f.scale_;
		dec_ 		= f.dec_;
		qu_dec_ 	= f.qu_dec_;
		base_ccy_ 	= f.base_ccy_; 
		price_ccy_ 	= f.price_ccy_; 
		reverse_ric_ = f.reverse_ric_;
		spot_code_ 	= f.spot_code_;
		cal_base_ 	= f.cal_base_; 
		cal_price_ 	= f.cal_price_; 
		cal_ny_ 	= f.cal_ny_; 
		rule_ 		= f.rule_; 
		q_			= f.q_;
		name_		= f.name_;
		ok_ 		= f.ok_; 
		err_ 		= f.err_;
	}
	/*
	void _fx.set_default()
	{
		scale_ 		= 1; 
		dec_		= 4; 
		qu_dec_		= 15; 
		base_ccy_	= ''; 
		price_ccy_	= ''; 
		reverse_ric_= false; 
		spot_code_	= 'BD2'; 
		cal_base_ 	= calendar(); 
		cal_price_	= calendar(); 
		ok_			= false;
	}
	*/
	void _fx.initialize_no_inst() 
	{				
		if(null(qu_dec_))
			qu_dec_	= dec_;					
		QL_FAIL_COND(equal_casei(price_ccy_,base_ccy_),strcat(["invalid fx instrument (price currency = base currency)"," [",name(),"]"]));
		QL_FAIL_COND(null(cal_ny_) && rule_ == fx_spot_rule.STD_USD_ADJ,strcat(["invalid fx instrument (usd calendar is missing)"," [",name(),"]"]));
			
		err_ = '';
		ok_ = true;		
	}
	
	
	void _fx.initialize(logical allow_defaults) 
	{
		if(null(fx_spot_)) {
			err_ = strcat(["invalid fx spot instrument"," [",name(),"]"]);
			QL_FAIL(err_);
		}		
		
		trade_date_ = fx_spot_.trade_date();
		name_		= fx_spot_.name();
	
		if(!fx_spot_.is_fx_spot()){
			err_ = strcat(["invalid type of instrument"," [",name(),"]"]);
			QL_FAIL(err_);
		}
	
		try{
			base_ccy_ 	= fx_spot_.fx_base_ccy(); 
			price_ccy_ 	= fx_spot_.fx_price_ccy(); 
			dec_ 		= fx_spot_.fx_pip_decimals(); 
			qu_dec_ 	= fx_spot_.fx_quote_dec();
			if(null(qu_dec_) || qu_dec_ == -1)
				qu_dec_	= dec_;		
			scale_ 		= fx_spot_.fx_scale(); 
			reverse_ric_= fx_spot_.fx_reverse_ric(); 
			spot_code_ 	= fx_spot_.settle_code(); 
			cal_base_ 	= fx_spot_.fx_base_ccy_calendar();
			cal_price_ 	= fx_spot_.fx_price_ccy_calendar(); 
			cal_ny_ 	= fx_spot_.fx_usd_calendar();
			rule_ 		= fx_spot_.fx_rule();
	
			QL_FAIL_COND(null(base_ccy_),strcat(["invalid fx instrument (base currency)"," [",name(),"]"]));	 		
			QL_FAIL_COND(null(price_ccy_),strcat(["invalid fx instrument (ccy_price tag)"," [",name(),"]"]));	
			QL_FAIL_COND(equal_casei(price_ccy_,base_ccy_),strcat(["invalid fx instrument (price currency = base currency)"," [",name(),"]"]));		
			QL_FAIL_COND(null(dec_),strcat(["invalid fx instrument (ccy_dec tag)"," [",name(),"]"]));	
			QL_FAIL_COND(null(qu_dec_),strcat(["invalid fx instrument (ccy_qu_dec tag)"," [",name(),"]"]));		
			QL_FAIL_COND(null(scale_),strcat(["invalid fx instrument (ccy_scale tag)"," [",name(),"]"]));								
			QL_FAIL_COND(null(reverse_ric_),strcat(["invalid fx instrument (ccy_reverse_ric tag)"," [",name(),"]"]));				
			QL_FAIL_COND(null(spot_code_),strcat(["invalid fx instrument (settle_code)"," [",name(),"]"]));					
			QL_FAIL_COND(null(cal_base_),strcat(["invalid fx instrument (base currency calendar)"," [",name(),"]"]));	
			QL_FAIL_COND(null(cal_price_),strcat(["invalid fx instrument (price currency calendar)"," [",name(),"]"]));
			QL_FAIL_COND(null(rule_),strcat(["invalid fx instrument (fx rule)"," [",name(),"]"]));
			QL_FAIL_COND(null(cal_ny_) && rule_ == fx_spot_rule.STD_USD_ADJ,strcat(["invalid fx instrument (usd calendar)"," [",name(),"]"]));		
			err_ = '';
			ok_ = true;
		}
		catch {		
			err_ = err.message();		
			ok_ = false; 
		}
		if(!ok_) {
			if(!allow_defaults)
				throw(E_INVALID_ARG, err_);
			else {
				if(null(scale_))		scale_ 		= 1; 
				if(null(dec_))			dec_		= 4; 
				if(null(qu_dec_))		qu_dec_		= 15; 	
				if(null(base_ccy_))		QL_FAIL(err_); 
				if(null(price_ccy_))	QL_FAIL(err_); 		 
				if(null(reverse_ric_))	reverse_ric_= false; 
				if(null(spot_code_))	spot_code_	= 'BD2'; 
				if(null(cal_base_))		cal_base_ 	= calendar();//empty 
				if(null(cal_price_))	cal_price_	= calendar(); //empty
				if(null(cal_ny_))		cal_ny_		= calendar(); //empty
				if(null(rule_))			rule_		= fx_spot_rule.STD; 
			}
		}
	}
	
	logical _fx.is_ok(out string option(nullable) err)
	{
		err = ok_ ? "" : err_;
		return ok_;
	}
	
	number 		_fx.swap() 														{ QL_FAIL(strcat(["invalid call"," [",name(),"]"])); }
	number 		_fx.swap(logical l, logical r) 									{ QL_FAIL(strcat(["invalid call"," [",name(),"]"])); }
	number 		_fx.swap(logical l, number option(nullable) n, logical r) 		{ QL_FAIL(strcat(["invalid call"," [",name(),"]"])); }
	number 		_fx.swap(disc_func b,disc_func p,logical u,number option(nullable) sa,logical nr) 			
																				{ QL_FAIL(strcat(["invalid call"," [",name(),"]"])); }
	number 		_fx.outright() 													{ QL_FAIL(strcat(["invalid call"," [",name(),"]"])); }
	number 		_fx.outright(logical l,logical r) 								{ QL_FAIL(strcat(["invalid call"," [",name(),"]"])); }
	number 		_fx.outright(logical l, number option(nullable) n, logical r) 	{ QL_FAIL(strcat(["invalid call"," [",name(),"]"])); }
	number 		_fx.outright(disc_func b,disc_func p,logical u,number option(nullable) sa, logical nr) 		
																				{ QL_FAIL(strcat(["invalid call"," [",name(),"]"])); }
	
	string 		_fx.name()			{ return name_;}
	string 		_fx.spot_name()		{ return name_;}
	date 		_fx.trade_date()	{ return trade_date_;}
	instrument 	_fx.spot_inst() 	{ return fx_spot_;}
	number 		_fx.scale()			{ return scale_;}
	number 		_fx.dec()			{ return dec_;}
	number 		_fx.qu_dec()		{ return qu_dec_;}
	number 		_fx.spot_qu_dec()	{ return qu_dec_;}
	string 		_fx.base_ccy()		{ return base_ccy_;}
	string  	_fx.price_ccy()		{ return price_ccy_;}
	logical 	_fx.reverse_ric()	{ return reverse_ric_;}
	string 		_fx.spot_code()		{ return spot_code_;}
	calendar  	_fx.cal_base()		{ return cal_base_;}
	calendar  	_fx.cal_price()		{ return cal_price_;}
	calendar  	_fx.cal_ny()		{ return cal_ny_;}
	fx_spot_rule.fx_spot_rule _fx.rule()			{ return rule_;}
	number 		_fx.spot_quote(logical unit_quote, logical no_rnd ) { return _fx.quote(unit_quote, no_rnd); }
	number 		_fx.spot_quote() 									{ return _fx.quote(false, false); }
	void 		_fx.set_spot_quote(number option(nullable)	q, logical unit_quote )	{ _fx.set_quote(q, unit_quote ); }
	
	date _fx.spot_settle() 
	{
		return CORE_INT_FX.fx_spot_date_s(trade_date_, cal_base_, base_ccy_, cal_price_, price_ccy_, rule_, spot_code_,cal_ny_);	
	}
	
	date _fx.maturity(date sd, string mc)
	{
		return fx_maturity(sd,cal_base_, base_ccy_, cal_price_, price_ccy_, rule_,mc, cal_ny_);
	}
	
	void _fx.set_quote(	number option(nullable)	q, 
						logical unit_quote)
	{
		if(reverse_ric_) {
			q = unit_quote ? q : q / scale_;
			q = 1.0/q;			
		}
		else if(unit_quote)
			q *= scale_;		
		
		if(null(fx_spot_))
			q_ = q;
		else
			fx_spot_ = fx_spot_.set_quote(q);	
	}
	
	number _fx.quote(	logical unit_quote,
						logical no_rnd)
	{	
		number q = null(fx_spot_) ? q_ : fx_spot_.quote_orig();
		if(null(q)) return q;
		//note! even if it is a mid we do a rounding below (possible loss of precision)
		if(reverse_ric_) {
			q = 1.0/q;//if reverse it is assumed the ric is per 1 unit of price currency (not sure about this!!) 
			q = unit_quote ? q : q * scale_;
			return no_rnd ? q: round(q, integer(qu_dec_));				
		}
		else {
			q = no_rnd ? q: round(q, integer(qu_dec_));	
			return unit_quote ? q/scale_ : q;
		}	
	}
	
	number _fx.spot_adjustment(	number 					fx_spot,
								number option(nullable) spot_adj)
	{
		if(null(spot_adj) || spot_adj == 0)
			return fx_spot;
		fx_spot 		/= scale();
		number adj_prec = floor(log(scale())/log(10.0)+0.5)	;
		fx_spot 		-= (spot_adj / pow(10.0, dec()+ adj_prec)) ;
		return fx_spot * scale();	
	}
	/*-----------------------------------------------------------------------
	  Name :  	class fx_spot
	  Notes :   a container for an fx spot instrument and static data    
	  Changes:
	  			Date    		Who     Detail
	  			8-Dec-2013 		MN      Original code 
	  ----------------------------------------------------------------------*/	
	class fx_spot : public _fx 
	{
		public:
			fx_spot(date,number,number,number,string,string,string,calendar,calendar,calendar option(nullable), fx_spot_rule.fx_spot_rule,number option(nullable), string);
			fx_spot(instrument, logical allow_defaults = false);
			fx_spot(instrument_name, date t = today(), quote_side option(nullable) qs = null<string>, logical allow_defaults = false);
			fx_spot(fx_spot) ;
	
			virtual _fx				clone();
			virtual fx_inst_type 	type();			
			virtual number			quote(string,logical u = false,logical n = false  );			
			virtual number			quote(logical unit_quote,logical no_rnd);	
			virtual number			quote();	
			number					quote_inv(logical ,logical  );	
			number					quote_inv( );		
	 
			virtual date			maturity() ;
			virtual date			settle_date() ;
			virtual instrument 		inst();
			
			virtual number			convert(number,string); 
			number					points(number,number);
			virtual logical			is_post_spot();
	
		protected:
			logical	check(out string option(nullable));
			fx_spot create_inv_fx_spot(quote_side,logical );
	};
	
	fx_spot.fx_spot(date		d,
					number 		scale,
					number 		dec,
					number 		qu_dec,
					string 		base_ccy,
					string  	price_ccy,
					string 		spot_code,
					calendar  	cal_base,
					calendar  	cal_price,
					calendar option(nullable) cal_ny,				
					fx_spot_rule.fx_spot_rule rule,
					number option(nullable) q,
					string		name) 
					: _fx(d,scale,dec,qu_dec,base_ccy,price_ccy, spot_code,cal_base,cal_price,cal_ny,rule,q, name)
	{	
		string err_msg;
		if(!check(err_msg))
			throw(E_INVALID_ARG, err_msg);
	}
	
	fx_spot.fx_spot(instrument i_fx_spot,
					logical allow_defaults): _fx(i_fx_spot, allow_defaults)
	{
		string err_msg;
		QL_FAIL_COND(!check(err_msg), err_msg);
	}
	
	fx_spot.fx_spot(instrument_name 			fx_spot_name, 
					date 						trade_date,
					quote_side option(nullable) quote_side,
					logical 					allow_defaults)
					:_fx(fx_spot_name, trade_date, quote_side, allow_defaults)
	{
		string err_msg;
		QL_FAIL_COND(!check(err_msg),err_msg);
	}
	
	fx_spot.fx_spot(fx_spot f) : _fx(f) {}
	
	
	_fx		fx_spot.clone() 													{ return new fx_spot(this);}
	date 	fx_spot.maturity() 													{ return settle_date(); }
	date 	fx_spot.settle_date() 												{ return spot_settle(); }
	fx_inst_type 	fx_spot.type()												{ return FXI_SPOT;}
	number 	fx_spot.quote(logical unit_quote,logical no_rnd )					{ return spot_quote(unit_quote, no_rnd);}
	number 	fx_spot.quote()														{ return spot_quote(false, false); }
	number 	fx_spot.quote_inv()													{ return quote_inv(false, false ); }
	instrument 	fx_spot.inst()													{ return spot_inst();}
	number 	fx_spot.points(	number b,number a)									{ return (a-b)*pow(10.0, dec());}
	logical	fx_spot.is_post_spot()												{ QL_FAIL(strcat(["invalid call"," [",name(),"]"])); }
	
	logical fx_spot.check(	out string option(nullable) err_msg)
	{
		err_msg = '';	
		return true;
	}
	
	
	fx_spot fx_spot.create_inv_fx_spot(quote_side qs,
										logical allow_defaults)
	{
		instrument_name sec;
		instrument_name prim = fx_spot_db_name(price_ccy(), base_ccy(), sec);	
		fx_spot fx;
	
		try {	
			fx = new fx_spot(prim, trade_date(), qs, allow_defaults);
		}
		catch {}
	
		if(null(fx)) {
			try {	
				fx = new fx_spot(sec, trade_date(), qs, allow_defaults);			
				if(!null(fx) && fx.base_ccy() == price_ccy() && fx.price_ccy() == base_ccy())
					return fx;
				else
					return null<fx_spot>; 		
			}
			catch {}
		}
		return fx;
	}
	
	number fx_spot.quote(	string base_c,
							logical unit_quote,
							logical no_rnd )
	{
		if(base_c == base_ccy())
			return quote(unit_quote, no_rnd);
		else if(base_c == price_ccy())
			return quote_inv(unit_quote, no_rnd);
		else
			return null<number>;
	}
	 
	number fx_spot.quote_inv(logical unit_quote,
							logical no_rnd )
	{		
		number q_unit = quote(true,no_rnd);
		if(null(q_unit))
			return null<number>;
	
		number q_inv_unit = 1.0/q_unit;	//this conversion will not handle bid and ask quotes since they will not be reversed
										//hence this function is safe only for mid 
		number q_inv = q_inv_unit;
		number qu_dec = 15;
		try {//if the inverted pair is in the database use the info on it otherwise no scaling and no decimals	
			
			fx_spot ccy_inv = create_inv_fx_spot(quote_side_default(), true);
		
			//instrument_name ccy_inv_str = fx_spot_db_name(price_ccy(), base_ccy());		
			//fx_spot ccy_inv = new fx_spot(ccy_inv_str, trade_date(), __get_def_quote_side(), true);
			if(!null(ccy_inv.qu_dec()))
				qu_dec = ccy_inv.qu_dec();
			else if(!null(ccy_inv.dec()))
				qu_dec = ccy_inv.dec();		
			number scale = 	null(ccy_inv.scale()) ? 1 : ccy_inv.scale();
			q_inv = unit_quote ? q_inv_unit : q_inv_unit * scale ;			
		}
		catch {}
					
		return no_rnd ? q_inv : round(q_inv, integer(qu_dec));	//note! even if it is a mid we do a rounding (possible loss of precision)	
	}
	
	number fx_spot.convert(	number amount,
							string amount_ccy)
	{
		number q;
		if(amount_ccy == price_ccy()) 
			q = quote_inv(true, false);
		else if(amount_ccy == base_ccy()) 
			q = quote(true, false);
		else
			QL_FAIL(strcat(["incompatible currency"," [",name(),"]"]));
	
		return q * amount;	
	}



//module CORE_INT_COMPAT
//{	
	fx_spot create_fx_spot_inst(string 		base_c,
								string 		price_c,
								date 		d,
								quote_side option(nullable)	qs,
								logical 	allow_defaults)
	{
		instrument_name alt_name;
		instrument_name name = fx_spot_db_name(base_c, price_c, alt_name);	
		fx_spot fx;
	
		try {	
			fx = new fx_spot(name, d, qs, allow_defaults);
		}
		catch {}
	
		if(null(fx)) {
			try {	
				fx = new fx_spot(alt_name, d, qs, allow_defaults);
				if(fx.base_ccy() == base_c && fx.price_ccy() == price_c)
					return fx;
				else
					return null<fx_spot>; 			
			}
			catch {}
		}
		return fx;
	}

	
	/*fx_spot create_fx_spot_inst_flex(	string 		base_c, 
										string 		price_c,
										date 		d,
										quote_side option(nullable)	qs,
										logical 	allow_defaults)
	
	{	
		instrument_name alt_name;
		instrument_name base_price_name = fx_spot_db_name(base_c,price_c,alt_name);
		instrument_name price_base_name = fx_spot_db_name(price_c,base_c,alt_name);		
		logical base_price_ok = false; 
		logical price_base_ok = false;
		fx_spot fx_bp, fx_pb;
	
		try {
				
			fx_bp = new fx_spot(base_price_name, d, qs, allow_defaults);
			QL_FAIL_COND(null(fx_bp),"");
			if(fx_bp.reverse_ric()) {
				try {
					fx_pb = new fx_spot(price_base_name, d, qs, allow_defaults);
					QL_FAIL_COND(null(fx_pb),"");
					price_base_ok = true;
				}
				catch {
					base_price_ok = true;
				}
			}
			else			
				base_price_ok = true;
		}
		catch {}
	
		if(!base_price_ok) {
			try {
				fx_pb = new fx_spot(price_base_name, d, qs, allow_defaults);
				QL_FAIL_COND(null(fx_pb),"");
				price_base_ok = true;
			}
			catch {}
		}
	
		if(base_price_ok)
			return fx_bp;
		else if(price_base_ok)
			return fx_pb;
		else {
			try {
				fx_bp = new fx_spot(alt_name, d, qs, allow_defaults);
				return fx_bp;			
			}
			catch {
				return null<fx_spot>;
			}
		}
		
	}
*/
	




	/*-----------------------------------------------------------------------
	  Name :  	class fx_swap
	  Notes :   a container for an fx swap instrument and static data    
	  Changes:
	  			Date    		Who     Detail
	  			11-Dec-2013 		MN      Original code 
	  ----------------------------------------------------------------------*/	
	class fx_swap : public _fx 
	{
		public:
			fx_swap(date,date,number option(nullable), number option(nullable),string,fx_spot, 
					number option(nullable) spot_adj = null<number>, logical mp = false);
			fx_swap(instrument, number option(nullable) spot_adj = null<number>, logical mp = false);
			fx_swap(instrument, instrument, number option(nullable) spot_adj = null<number>, logical mp = false);
			fx_swap(instrument, fx_spot, number option(nullable) spot_adj = null<number>, logical mp = false);
			fx_swap(instrument_name, date t = today(), quote_side option(nullable) qs = null<string>, 
					quote_side option(nullable) qss = null<string>, number option(nullable) spot_adj = null<number>,logical mp = false);		
			fx_swap(fx_swap) ;
	
			virtual _fx				clone();
			virtual fx_inst_type 	type();	
			virtual string 			name();	
			virtual number			quote();	
			virtual number			quote(logical,logical );
			virtual void			set_quote(number option(nullable), logical unit_quote = false);	
	
			virtual number			swap();
			virtual number			swap(logical, logical);		
			virtual number 			swap(disc_func,disc_func,logical u = false,number option(nullable) sa = 0,logical nr = false);
	
			virtual number 			outright();
			virtual number 			outright(logical, logical );
			virtual number			outright(logical, number option(nullable), logical);
			instrument				synt_depo(day_count_method,number,day_count_method,logical,number option(nullable));
			instrument				synt_depo(day_count_method,disc_func,logical,number option(nullable));
			number 					rate_base(	day_count_method,rate_type,number,day_count_method,rate_type,number option(nullable));
			number 					rate_price(	day_count_method,rate_type,number,day_count_method,rate_type,number option(nullable));
			void					rate_pvbp(	number,number option(nullable),logical,number,day_count_method,	
												number,day_count_method,number,number option(nullable),out number option(nullable),out number option(nullable) );
			void 					tail(number,number option(nullable),logical,disc_func,disc_func,logical ,number option(nullable),
										logical,out number option(nullable),out number option(nullable));
			void					cash_flow(	number,number option(nullable),logical,disc_func option(nullable),				
												number option(nullable),out vector(number) option(nullable),out vector(number) option(nullable));
			virtual instrument 		inst() ;
			virtual date			maturity() ;
			virtual date			settle_date() ;
			void					init_fx_spot(fx_spot);
			virtual number 			qu_dec();
			virtual logical			is_post_spot();	
		
		protected:
			
			logical	check(out string option(nullable));
			logical	check_comp(out string option(nullable));
	
		private:
			date		settle_;
			date 		maturity_;
			instrument  fx_swap_;
			logical 	matched_prin_;
			number 		spot_adj_;
			number 		q_;
			number 		qu_dec_;
			string		name_;
			logical		is_post_spot_;
	};
	
	fx_swap.fx_swap(date 					settle,
					date 					maturity,
					number option(nullable)	q,
					number option(nullable) qu_dec,
					string					name,
					fx_spot 				i_fx_spot,
					number option(nullable) spot_adj,
					logical 				matched_prin)
					: _fx(dynamic_cast<_fx>(i_fx_spot)), settle_(settle), maturity_(maturity),fx_swap_(null<instrument>), 
					  matched_prin_(matched_prin), spot_adj_(spot_adj), q_(q), qu_dec_(qu_dec), name_(name)
	
	{	
		is_post_spot_ = settle_ == i_fx_spot.spot_settle();//fwd swaps??
		string err_msg;
		QL_FAIL_COND(!check(err_msg),err_msg);			
	}
	
	fx_swap.fx_swap(instrument 				i_fx_swap,
					number option(nullable) spot_adj,
					logical 				matched_prin)
					:settle_(i_fx_swap.settle_date()), maturity_(i_fx_swap.maturity()), fx_swap_(i_fx_swap), 
					 matched_prin_(matched_prin), spot_adj_(spot_adj), qu_dec_(i_fx_swap.fx_quote_dec()),name_(i_fx_swap.name())
	{
		fx_spot fx 	= CORE_INT_COMPAT_FX.create_fx_spot_inst(	fx_swap_.fx_base_ccy(),fx_swap_.fx_price_ccy(),
											i_fx_swap.trade_date(),i_fx_swap.quote_side(),false);	
		_fx.init(dynamic_cast<_fx>(fx));
	
		is_post_spot_ = equal_casei (fx_swap_.settle_code(), spot_code());//fwd swaps??
		 
		string err_msg;
		QL_FAIL_COND(!check(err_msg), err_msg);		
	}
	fx_swap.fx_swap(instrument 				i_fx_swap, 
					instrument 				i_fx_spot,
					number option(nullable) spot_adj,
					logical 				matched_prin)
					: _fx(i_fx_spot), settle_(i_fx_swap.settle_date()), maturity_(i_fx_swap.maturity()), fx_swap_(i_fx_swap), 
					  matched_prin_(matched_prin), spot_adj_(spot_adj), qu_dec_(i_fx_swap.fx_quote_dec()),name_(i_fx_swap.name())
	{
		is_post_spot_ = equal_casei (fx_swap_.settle_code(), spot_code());//fwd swaps??
		string err_msg;
		QL_FAIL_COND(!check(err_msg),err_msg);		
	}
	
	fx_swap.fx_swap(instrument 				i_fx_swap, 
					fx_spot 				i_fx_spot,
					number option(nullable) spot_adj,
					logical 				matched_prin)
					: _fx(dynamic_cast<_fx>(i_fx_spot)), settle_(i_fx_swap.settle_date()), maturity_(i_fx_swap.maturity()), 
					  fx_swap_(i_fx_swap), matched_prin_(matched_prin), spot_adj_(spot_adj),qu_dec_(i_fx_swap.fx_quote_dec()),
					  name_(i_fx_swap.name())
	
	{	
		//_fx.init(dynamic_cast<_fx>(i_fx_spot));
		is_post_spot_ = equal_casei (fx_swap_.settle_code(), spot_code());//fwd swaps??
		string err_msg;
		QL_FAIL_COND(!check(err_msg),err_msg);		
	}
	
	fx_swap.fx_swap(instrument_name 			fx_swap_name, 
					date 						trade_date,
					quote_side option(nullable) quote_side_swap,
					quote_side option(nullable) quote_side_spot,
					number option(nullable) 	spot_adj,
					logical 					matched_prin): matched_prin_(matched_prin), spot_adj_(spot_adj)				
	{
		fx_swap_ = instrument(fx_swap_name, trade_date, quote_side_swap);
		QL_FAIL_COND(null(fx_swap_),strcat(["invalid fx swap"," [",fx_swap_.name(),"]"]));
		
		settle_ 	= fx_swap_.settle_date();
		maturity_ 	= fx_swap_.maturity(); 
		qu_dec_		= fx_swap_.fx_quote_dec();
		name_		= fx_swap_.name();
	
		fx_spot fx 	= CORE_INT_COMPAT_FX.create_fx_spot_inst(	fx_swap_.fx_base_ccy(),fx_swap_.fx_price_ccy(),
											trade_date,quote_side_spot,false);	
		_fx.init(dynamic_cast<_fx>(fx));
		//_fx.init(fx_spot);	
		is_post_spot_ = equal_casei (fx_swap_.settle_code(), spot_code());//fwd swaps??
		string err_msg;	
		QL_FAIL_COND(!check(err_msg),err_msg);		
	}
	
	fx_swap.fx_swap(fx_swap f) : _fx(f), settle_(f.settle_), maturity_(f.maturity_), fx_swap_(f.fx_swap_), 
								matched_prin_(f.matched_prin_), spot_adj_(f.spot_adj_), q_(f.q_), qu_dec_(f.qu_dec_),
								name_(f.name_), is_post_spot_(f.is_post_spot_)
	{}
	
	logical fx_swap.check(	out string  option(nullable) err_msg)
	{
		err_msg = '';
		
		if(!_fx.is_ok(err_msg))
			return false;
	
		if(!null(fx_swap_) && !fx_swap_.is_fx_swap()){
			err_msg = strcat(["invalid type of instrument (fx swap)"," [",name(),"]"]);
			return false;
		}
		
		/*if(fx_swap_.maturity() < spot_settle() && null(spot_adj_)){
			err_msg = strcat(["maturity before spot requires a spot adjustment"," [",name(),"]"]);
			return false;
		}*/
		if(!check_comp(err_msg))
			return false;
	
		return true;
	}
	
	logical fx_swap.check_comp(	out string option(nullable) err_msg)
	{
		if(null(fx_swap_))
			return true;
	
		err_msg = '';
		
		if(trade_date() != fx_swap_.trade_date()){
			err_msg = strcat(["invalid instrument (trade_date mismatch in fx_spot and fx_swap)"," [",name(),"]"]);
			return false;
		}
		string base = fx_swap_.fx_base_ccy(); 	
		if(!null(base) && base_ccy() != base){
			err_msg = strcat(["different base currency in fx swap vs fx spot"," [",name(),"]"]);
			return false;
		}
	
		string price = fx_swap_.fx_price_ccy(); 
		if(!null(price) &&  price_ccy() != price){
			err_msg = strcat(["different price currency in fx swap vs fx spot"," [",name(),"]"]);
			return false;
		}
	
		return true;
	}
	
	void fx_swap.init_fx_spot(fx_spot i_fx_spot)							
	
	{
		//_fx s = new fx_spot(i_fx_spot);
		//_fx.init(s);
		_fx.init(dynamic_cast<_fx>(i_fx_spot));
		string err_msg;
		QL_FAIL_COND(!check_comp(err_msg), err_msg);		
	}
	
	_fx				fx_swap.clone() 									{ return new fx_swap(this);}
	fx_inst_type 	fx_swap.type()										{ return FXI_SWAP;}
	string 			fx_swap.name()										{ return name_;}
	date 			fx_swap.maturity() 									{ return maturity_; }
	date 			fx_swap.settle_date() 								{ return settle_; }
	instrument 		fx_swap.inst() 										{ return fx_swap_;}
	number 			fx_swap.qu_dec()									{ return null(qu_dec_) ? 15: qu_dec_;}
	number 			fx_swap.swap() 										{ return fx_swap.quote();}
	number 			fx_swap.swap(logical unit_quote, logical no_rnd)	{ return fx_swap.quote(unit_quote, no_rnd);}
	number 			fx_swap.outright()									{ return fx_swap.outright(false, spot_adj_, false);}
	number 			fx_swap.outright(logical unit_quote,logical no_rnd)	{ return fx_swap.outright(unit_quote, spot_adj_, no_rnd);}
	logical			fx_swap.is_post_spot()								{ return is_post_spot_;}
	void fx_swap.set_quote(	number 	option(nullable) q,
							logical unit_quote)
	{	
		if(unit_quote){
			number adj_prec = floor(log(scale())/log(10.0)+0.5)	;
			q *= pow(10.0, dec()+ adj_prec);
		}
	
		if(null(fx_swap_))
			q_ = q;
		else 
			fx_swap_ = fx_swap_.set_quote(q);		
	}
	
	number fx_swap.quote()			
	{ 
		number q = null(fx_swap_) ? q_: fx_swap_.quote_orig(); 
		if(null(q)) return q;
		return round(q, integer(qu_dec()));
	}
	
	number fx_swap.quote(	logical unit_quote,
							logical no_rnd)
	{
		number q = null(fx_swap_) ? q_: fx_swap_.quote_orig(); 
		q = no_rnd ? q : round(q, integer(qu_dec()));
	
		if(unit_quote){
			number adj_prec = floor(log(scale())/log(10.0)+0.5)	;		
			return q / pow(10.0, dec()+ adj_prec);
		}
		else
			return q;	
	}
	
	number fx_swap.swap(	disc_func 				df_base,
							disc_func 				df_price,
							logical 				unit_quote,
							number option(nullable) spot_adj,
							logical 				no_rnd)
	{
		number	fx_ref	= spot_quote(false,no_rnd);
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
	
		date mat		= maturity();	
		date spot 		= spot_settle();
	
		if(mat < spot) {//only O/N if spot is t+2
			if(null(spot_adj))
				spot_adj = spot_adj_;
			QL_FAIL_COND(null(spot_adj),strcat(["maturity before spot requires a spot adjustment"," [",name(),"]"]));					
			//if O/N swap --> spotadj = T/N points
			fx_ref = spot_adjustment(fx_ref,spot_adj);
		}
	
		number q = CORE_INT_FX.fx_df_to_swap(trade_date(),settle_date(),mat,spot,fx_ref,df_base,df_price,!reverse_ric(),dec(),scale());
		q = no_rnd ? q : round(q, integer(qu_dec()));
		if(unit_quote){
			number adj_prec = floor(log(scale())/log(10.0)+0.5)	;		
			return q / pow(10.0, dec()+ adj_prec);
		}
		else
			return q;		
	}
	
	number fx_swap.outright(logical 				unit_quote,
							number option(nullable) spot_adj,
							logical 				no_rnd)
	{	
		number	fx_ref	= spot_quote(false,no_rnd);
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
	
		date mat		= maturity();
		//date settle 	= settle_date();
		number fx_sw 	= quote(false,no_rnd);
		QL_FAIL_COND(null(fx_sw),strcat(["invalid swap quote"," [",name(),"]"]));	
		date spot 		= spot_settle();
			
		if(mat < spot) {//only O/N if spot is t+2
			if(null(spot_adj))
				spot_adj = spot_adj_;
			QL_FAIL_COND(null(spot_adj),strcat(["maturity before spot requires a spot adjustment"," [",name(),"]"]));					
	
			//if O/N swap --> spotadj = T/N points
			fx_ref = spot_adjustment(fx_ref,spot_adj);
		}
		number fx_out = CORE_INT_FX.fx_swap_to_f(mat, spot,fx_ref, fx_sw, dec(),scale()) ;
		fx_out = no_rnd ? fx_out : round(fx_out, integer(spot_qu_dec() + qu_dec()));
		return unit_quote ? fx_out / scale() : fx_out;	
	}
	
	instrument fx_swap.synt_depo(	day_count_method 	dc_synt,
									disc_func			df,	
									logical 			synt_is_base,
									number option(nullable) spot_adj)
	{	
		number	fx_ref	= spot_quote();
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
		date mat		= maturity();
		date settle 	= settle_date();
		number fx_sw 	= quote();
		QL_FAIL_COND(null(fx_sw),strcat(["invalid swap quote"," [",name(),"]"]));	
		date spot 		= spot_settle();
	
		if(mat < spot) {
			if(null(spot_adj))
				spot_adj = spot_adj_;
			QL_FAIL_COND(null(spot_adj),strcat(["maturity before spot requires a spot adjustment"," [",name(),"]"]));			
			fx_ref = spot_adjustment(fx_ref,spot_adj);
		}
	
		/*instr_def i_d = instr_def(synt_depo_cn);
		if(null(i_d))
			throw(E_INVALID_ARG, strcat(["invalid deposit class name"," [",name(),"]"]));	
	*/
	
		//day_count_method dc	= i_d.accr_dc_method();
		day_count_method dc_in 	= DC_ACT_360;
		number rate_in 			= df.simple_rate(trade_date(),settle,mat, dc_in) ;
		string ne 				= strcat(["synt depo (", name(),")"]);
		
	
		//Note! cashflows of the created depo is always nom=100, .face_amount() will return the last argument in create_depo(.)	
		if(synt_is_base){
			number fv_base,fv_price; 
			string ccy = fx_swap_.fx_base_ccy();
			number r_base = CORE_INT_FX.fx_swap_to_rate(settle, mat, spot ,fx_ref, fx_sw, null<number>,rate_in,dc_synt,dc_in, 	
													RT_SIMPLE,RT_SIMPLE,!reverse_ric(),dec(),scale(),CORE_INT_FX.FXP_BASE_R,fv_base,fv_price);	
			instr_def i_d = create_depo_instr_def(dc_synt, BD_MOD_FOLLOWING, true, "bd2", calendar(),ccy);
	
			instrument depo = deposit(i_d, ne ,trade_date(),settle, settle, mat, r_base, r_base, (matched_prin_ ? 100 : fv_base * 100));
			return depo;//create_depo_instr(i_d, ne ,trade_date(),settle, settle, mat, r_base, r_base, (matched_prin_ ? 100 : fv_base * 100));
				
			//return create_depo(synt_depo_cn, ne ,trade_date(),settle,mat,r_base, settle, (matched_prin_ ? 100 : fv_base * 100));
			
		}
		else {
			number fv_base,fv_price; 
			string ccy = fx_swap_.fx_price_ccy();
			number r_price = CORE_INT_FX.fx_swap_to_rate(settle, mat, spot ,fx_ref, fx_sw, rate_in,null<number>,dc_in,dc_synt, 	
													RT_SIMPLE,RT_SIMPLE,!reverse_ric(),dec(),scale(),CORE_INT_FX.FXP_PRICE_R,fv_base,fv_price);	
			instr_def i_d = create_depo_instr_def(dc_synt, BD_MOD_FOLLOWING, true, "bd2", calendar(),ccy);
	
			instrument depo = deposit(i_d, ne ,trade_date(),settle, settle, mat, r_price, r_price, fv_price*100);
			return depo;//create_depo_instr(i_d, ne ,trade_date(),settle, settle, mat, r_price, r_price, fv_price*100);	
	
			//return create_depo(synt_depo_cn, ne ,trade_date(),settle,mat,r_price, settle, fv_price*100);
		}		
	}
	
	instrument fx_swap.synt_depo(	/*instr_class_name  	synt_depo_cn,
									instr_class_name 	depo_cn_in,*/
									day_count_method 	dc_synt,
									number				rate_in,								
									day_count_method 	dc_in,	
									logical 			synt_is_base,
									number option(nullable) spot_adj)
	{	
		number	fx_ref	= spot_quote();
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
		date mat		= maturity();
		date settle 	= settle_date();
		number fx_sw 	= quote();
		QL_FAIL_COND(null(fx_sw),strcat(["invalid swap quote"," [",name(),"]"]));	
		date spot 		= spot_settle();
	
		if(mat < spot) {
			if(null(spot_adj))
				spot_adj = spot_adj_;
			QL_FAIL_COND(null(spot_adj),strcat(["maturity before spot requires a spot adjustment"," [",name(),"]"]));			
			fx_ref = spot_adjustment(fx_ref,spot_adj);
		}
	
		/*instr_def i_d_synt = instr_def(synt_depo_cn);
		if(null(i_d_synt))
			throw(E_INVALID_ARG, strcat(["invalid deposit class name (synt)"," [",name(),"]"]));	
	
		instr_def i_d_in = instr_def(depo_cn_in);
		if(null(i_d_in))
			throw(E_INVALID_ARG, strcat(["invalid deposit class name (in)"," [",name(),"]"]));*/
	
		//day_count_method dc_synt 	= i_d_synt.accr_dc_method();
		//day_count_method dc_in 		= i_d_in.accr_dc_method();
		string ne 					= strcat(["synt depo (", name(),")"]);
	
		//Note! cashflows of the created depo is always nom=100, .face_amount() will return the last argument in create_depo(.)	
		if(synt_is_base){
				
			number fv_base,fv_price;
			string ccy = fx_swap_.fx_base_ccy(); 
			number r_base = CORE_INT_FX.fx_swap_to_rate(settle, mat, spot ,fx_ref, fx_sw, null<number>,rate_in,dc_synt,dc_in, 	
													RT_SIMPLE,RT_SIMPLE,!reverse_ric(),dec(),scale(),CORE_INT_FX.FXP_BASE_R,fv_base,fv_price);	
			instr_def i_d = create_depo_instr_def(dc_synt, BD_MOD_FOLLOWING, true, "bd2", calendar(),ccy);
	
			instrument depo = deposit(i_d, ne ,trade_date(),settle, settle, mat, r_base, r_base, (matched_prin_ ? 100 : fv_base * 100));
			return depo;//create_depo_instr(i_d, ne ,trade_date(),settle, settle, mat, r_base, r_base, (matched_prin_ ? 100 : fv_base * 100));
				
			//return create_depo(synt_depo_cn, ne ,trade_date(),settle,mat,r_base, settle, (matched_prin_ ? 100 : fv_base * 100));
			
		}
		else {
			number fv_base,fv_price; 
			string ccy = fx_swap_.fx_price_ccy();
			number r_price = CORE_INT_FX.fx_swap_to_rate(settle, mat, spot ,fx_ref, fx_sw, rate_in,null<number>,dc_in,dc_synt, 	
													RT_SIMPLE,RT_SIMPLE,!reverse_ric(),dec(),scale(),CORE_INT_FX.FXP_PRICE_R,fv_base,fv_price);	
			instr_def i_d = create_depo_instr_def(dc_synt, BD_MOD_FOLLOWING, true, "bd2", calendar(),ccy);
	
			instrument depo = deposit(i_d, ne ,trade_date(),settle, settle, mat, r_price, r_price ,fv_price*100);
			return depo;//create_depo_instr(i_d, ne ,trade_date(),settle, settle, mat, r_price, r_price ,fv_price*100);		
			//return create_depo(synt_depo_cn, ne ,trade_date(),settle,mat,r_price, settle, fv_price*100);
		}		
	}
	
	number fx_swap.rate_base(	day_count_method 	dc_base,
								rate_type			rt_base,								
								number				rate_price,
								day_count_method 	dc_price,
								rate_type			rt_price,
								number option(nullable) spot_adj)
	{	
		number	fx_ref	= spot_quote();
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
		date mat		= maturity();
		date settle 	= settle_date();
		number fx_sw 	= quote();
		QL_FAIL_COND(null(fx_sw),strcat(["invalid swap quote"," [",name(),"]"]));	
		date spot 		= spot_settle();
	
		if(mat < spot) {
			if(null(spot_adj))
				spot_adj = spot_adj_;
			QL_FAIL_COND(null(spot_adj),strcat(["maturity before spot requires a spot adjustment"," [",name(),"]"]));			
			fx_ref = spot_adjustment(fx_ref,spot_adj);
		}	
		
		number fv_base,fv_price; 
		return CORE_INT_FX.fx_swap_to_rate(settle, mat, spot ,fx_ref, fx_sw, null<number>,rate_price,dc_base,dc_price, 	
										rt_base,rt_price,!reverse_ric(),dec(),scale(),CORE_INT_FX.FXP_BASE_R,fv_base,fv_price);			
	}
	
	number fx_swap.rate_price(	day_count_method 	dc_price,
								rate_type			rt_price,								
								number				rate_base,
								day_count_method 	dc_base,
								rate_type			rt_base,
								number option(nullable) spot_adj)
	{	
		number	fx_ref	= spot_quote();
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
		date mat		= maturity();
		date settle 	= settle_date();
		number fx_sw 	= quote();
		QL_FAIL_COND(null(fx_sw),strcat(["invalid swap quote"," [",name(),"]"]));	
		date spot 		= spot_settle();
	
		if(mat < spot) {
			if(null(spot_adj))
				spot_adj = spot_adj_;
			QL_FAIL_COND(null(spot_adj),strcat(["maturity before spot requires a spot adjustment"," [",name(),"]"]));			
			fx_ref = spot_adjustment(fx_ref,spot_adj);
		}	
			
		number fv_base,fv_price; 
		return CORE_INT_FX.fx_swap_to_rate(settle, mat, spot ,fx_ref, fx_sw, rate_base,null<number>,dc_base,dc_price, 	
										rt_base,rt_price,!reverse_ric(),dec(),scale(),CORE_INT_FX.FXP_PRICE_R,fv_base,fv_price);		
			
			
	}
	void fx_swap.rate_pvbp(	number 				nom_base,
							number option(nullable) nom_base_out,
							logical 			buy_base,
							number 				r_price,
							day_count_method 	dc_price,	
							number 				r_base,
							day_count_method 	dc_base,
							number 				rate_shift,
							number option(nullable) spot_adj,
							out number  option(nullable)	pvbp_base,
							out number  option(nullable)	pvbp_price)
	{		
		number	fx_ref	= spot_quote();
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
		date mat		= maturity();
		date settle 	= settle_date();
		number fx_sw 	= quote();
		QL_FAIL_COND(null(fx_sw),strcat(["invalid swap quote"," [",name(),"]"]));	
		date spot 		= spot_settle();
	
		if(matched_prin_){
			if(!null(nom_base_out) && nom_base_out != nom_base)
				QL_FAIL(strcat(["nom_base_out must be equal to nom_base for matched principal swaps"," [",name(),"]"]));
			else
				nom_base_out = nom_base;
		}
		else {	
			if(null(nom_base_out))
				//nom_base_out = CORE_INT.fv(nom_base,r_base,RT_SIMPLE,settle,mat,dc_base) ;
				nom_base_out = future_value(nom_base,r_base,settle,mat,dc_base,RT_SIMPLE) ;
		}
	
		if(mat < spot) {
			if(null(spot_adj))
				spot_adj = spot_adj_;
			QL_FAIL_COND(null(spot_adj),strcat(["maturity before spot requires a spot adjustment"," [",name(),"]"]));			
			fx_ref = spot_adjustment(fx_ref,spot_adj);
		}
	
		vector(number)	cf_base; 
		vector(number)	cf_price;
		CORE_INT_FX.fx_swap_to_cf(	mat,spot,nom_base,nom_base_out,buy_base,fx_ref, fx_sw, 								 
								!reverse_ric(),dec(),scale(),cf_base,cf_price);
	
		number 	spot_cf;
		pvbp_base = CORE_INT.pv_diff_end(cf_base[1],r_base,rate_shift,RT_SIMPLE,settle,mat,dc_base,spot_cf);
		
		pvbp_price = CORE_INT.pv_diff_end(cf_price[1],r_price,rate_shift,RT_SIMPLE,settle,mat,dc_price,spot_cf);
	
		return ;
	}
	
	void fx_swap.cash_flow(	number 					nom_base,
							number option(nullable) nom_base_out,
							logical 				buy_base,
							disc_func option(nullable)	df_base,
							//disc_func 				df_price,					
							number option(nullable) spot_adj,					
							out vector(number) option(nullable)	cf_base, 
							out vector(number) option(nullable)	cf_price)
	{			
		number	fx_ref	= spot_quote();
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
		date mat		= maturity();
		date settle 	= settle_date();
		number fx_sw 	= quote();
		QL_FAIL_COND(null(fx_sw),strcat(["invalid swap quote"," [",name(),"]"]));	
		date spot 		= spot_settle();
	
		if(matched_prin_){
			if(!null(nom_base_out) && nom_base_out != nom_base)
				QL_FAIL(strcat(["nom_base_out must be equal to nom_base for matched principal swaps"," [",name(),"]"]));
			else
				nom_base_out = nom_base;
		}
		else {	
			if(null(nom_base_out)){
				number fv = df_base.disc((settle-trade_date())/365) / df_base.disc((mat-trade_date())/365) ;
				nom_base_out = nom_base * fv;
			}
		}
	
		if(mat < spot) {
			if(null(spot_adj))
				spot_adj = spot_adj_;
			QL_FAIL_COND(null(spot_adj),strcat(["maturity before spot requires a spot adjustment"," [",name(),"]"]));			
			fx_ref = spot_adjustment(fx_ref,spot_adj);
		}
		
		CORE_INT_FX.fx_swap_to_cf(	mat,spot,nom_base,nom_base_out,buy_base,fx_ref, fx_sw, 								 
								!reverse_ric(),dec(),scale(),cf_base,cf_price);				
	}
	
	void fx_swap.tail(	number 					nom_base,
						number option(nullable) nom_base_out,
						logical 				buy_base,
						disc_func 				df_base,
						disc_func 				df_price,
						logical 				unit_quote,
						number option(nullable) spot_adj,
						logical					hdg_is_price,
						out number option(nullable)	tail_base,
						out number option(nullable)	tail_price)
	{			
		number	fx_ref	= spot_quote();
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
		date mat		= maturity();
		date settle 	= settle_date();
		number fx_sw 	= quote();
		QL_FAIL_COND(null(fx_sw),strcat(["invalid swap quote"," [",name(),"]"]));	
		date spot 		= spot_settle();
	
		if(matched_prin_){
			if(!null(nom_base_out) && nom_base_out != nom_base)
				QL_FAIL(strcat(["nom_base_out must be equal to nom_base for matched principal swaps"," [",name(),"]"]));
			else
				nom_base_out = nom_base;
		}
		else {	
			if(null(nom_base_out)){
				number fv = df_base.disc((settle-trade_date())/365) / df_base.disc((mat-trade_date())/365) ;
				nom_base_out = nom_base * fv;
			}
		}
	
		if(mat < spot) {
			if(null(spot_adj))
				spot_adj = spot_adj_;
			QL_FAIL_COND(null(spot_adj),strcat(["maturity before spot requires a spot adjustment"," [",name(),"]"]));			
			fx_ref = spot_adjustment(fx_ref,spot_adj);
		}
		
		vector(number)	cf_base; 
		vector(number)	cf_price;
		CORE_INT_FX.fx_swap_to_cf(	mat,spot,nom_base,nom_base_out,buy_base,fx_ref, fx_sw, 								 
								!reverse_ric(),dec(),scale(),cf_base,cf_price);	
			
		if(hdg_is_price) {
			number pv_price = df_price.disc((mat-trade_date())/365)/df_price.disc((settle -trade_date())/365)*cf_price[1]; 
			tail_price = cf_price[0] + pv_price;
			tail_base = -tail_price / fx_ref;
		}
		else {
			number pv_base = df_base.disc((mat-trade_date())/365)/df_base.disc((settle -trade_date())/365)*cf_base[1]; 
			tail_base = cf_base [0] + pv_base ;
			tail_price = -tail_base * fx_ref;
		}
	}
	/*-----------------------------------------------------------------------
	  Name :  	class fx_fwd
	  Notes :   a container for an fx swap instrument and static data    
	  Changes:
	  			Date    		Who     Detail
	  			11-Dec-2013 		MN      Original code 
	  ----------------------------------------------------------------------*/	
	class fx_fwd : public _fx 
	{
		public:
			fx_fwd (date, number option(nullable),number option(nullable),string, fx_spot);
			fx_fwd (instrument);
			fx_fwd (instrument, instrument);
			fx_fwd (instrument, fx_spot);
			fx_fwd (instrument_name, date t = today(), quote_side option(nullable) qs = null<string>, 
					quote_side option(nullable) qss = null<string>);	
			fx_fwd (fx_fwd ) ;
	
			virtual _fx				clone();
			virtual fx_inst_type 	type();
			virtual string 			name();
			virtual number			quote(logical,logical);
			virtual number			quote();
			virtual void			set_quote(number option(nullable), logical unit_quote = false);
	
			virtual number 			swap();
			virtual number 			swap(logical, logical);
			virtual number 			swap(logical, number option(nullable), logical);
			
			virtual number			outright();	
			virtual number			outright(logical, logical);
			virtual number 			outright(disc_func,disc_func,logical u = false,number option(nullable) sa = 0, logical nr = false);	
			
			instrument				synt_depo(day_count_method,number,day_count_method,logical,number option(nullable) sa = 0);
			instrument				synt_depo(day_count_method,disc_func,logical,number option(nullable) sa = 0);
			/*void					rate_pvbp(	number,number option(nullable),logical,number,day_count_method,	
												number,day_count_method,number,out number,out number );*/
			virtual instrument 		inst() ;
			virtual date			maturity() ;
			virtual date			settle_date() ;
			void					init_fx_spot(fx_spot);
			virtual number 			qu_dec();
			virtual logical			is_post_spot();	
	
		protected:
			
			logical	check(out string option(nullable));
			logical	check_comp(out string option(nullable));
	
		private:
			
			date 		maturity_;
			instrument  fx_fwd_;
			number 		q_;
			number		qu_dec_;
			string		name_;
			logical 	is_post_spot_;
	};
	
	fx_fwd.fx_fwd(	date 					maturity, 
					number option(nullable) q,
					number option(nullable) qu_dec,
					string					name,
					fx_spot 				i_fx_spot)
					: _fx(dynamic_cast<_fx>(i_fx_spot)), maturity_(maturity), fx_fwd_(null<instrument>), q_(q), qu_dec_(qu_dec), name_(name)
	{
		is_post_spot_ = maturity > spot_settle();
		string err_msg;
		QL_FAIL_COND(!check(err_msg),err_msg);		
	}
	
	fx_fwd.fx_fwd(instrument i_fx_fwd): maturity_(i_fx_fwd.maturity()), fx_fwd_(i_fx_fwd),
					  					qu_dec_(i_fx_fwd.fx_quote_dec()),name_(i_fx_fwd.name()) 
	{
		fx_spot fx 	= CORE_INT_COMPAT_FX.create_fx_spot_inst(	fx_fwd_.fx_base_ccy(),fx_fwd_.fx_price_ccy(),
											i_fx_fwd.trade_date(),i_fx_fwd.quote_side(),false);	
		_fx.init(dynamic_cast<_fx>(fx));
	
		is_post_spot_ = equal_casei (fx_fwd_.settle_code(), spot_code());
	
		string err_msg;
		QL_FAIL_COND(!check(err_msg),err_msg);		
	}
	
	
	fx_fwd.fx_fwd(	instrument 	i_fx_fwd, 
					instrument 	i_fx_spot)
					: _fx(i_fx_spot), maturity_(i_fx_fwd.maturity()), fx_fwd_(i_fx_fwd),
					  qu_dec_(i_fx_fwd.fx_quote_dec()),name_(i_fx_fwd.name()) 
	{
		is_post_spot_ = equal_casei (fx_fwd_.settle_code(), spot_code());
		string err_msg;
		QL_FAIL_COND(!check(err_msg),err_msg);		
	}
	
	fx_fwd.fx_fwd(	instrument 	i_fx_fwd, 
					fx_spot 	i_fx_spot)
					: _fx(dynamic_cast<_fx>(i_fx_spot)), maturity_(i_fx_fwd.maturity()), fx_fwd_(i_fx_fwd),
					  qu_dec_(i_fx_fwd.fx_quote_dec()),name_(i_fx_fwd.name()) 
	{	
		is_post_spot_ = equal_casei (fx_fwd_.settle_code(), spot_code());	
		string err_msg;
		QL_FAIL_COND(!check(err_msg),err_msg);		
	}
	
	fx_fwd.fx_fwd(	instrument_name 			fx_fwd_name, 
					date 						trade_date,
					quote_side option(nullable) quote_side_fwd,
					quote_side option(nullable) quote_side_spot)		
	{
		fx_fwd_ = instrument(fx_fwd_name, trade_date, quote_side_fwd);
		QL_FAIL_COND(null(fx_fwd_),strcat(["invalid fx fwd"," [",fx_fwd_.name(),"]"]));
	
		maturity_ 	= fx_fwd_.maturity();
		qu_dec_ 	= fx_fwd_.fx_quote_dec();
		name_ 		= fx_fwd_.name();
	
		fx_spot fx 	= CORE_INT_COMPAT_FX.create_fx_spot_inst(	fx_fwd_.fx_base_ccy(),fx_fwd_.fx_price_ccy(),
											trade_date,quote_side_spot,false);	
		_fx.init(dynamic_cast<_fx>(fx));
		
		is_post_spot_ = equal_casei (fx_fwd_.settle_code(), spot_code());
		string err_msg;	
		QL_FAIL_COND(!check(err_msg),err_msg);		
	}
	
	fx_fwd.fx_fwd(fx_fwd f) : _fx(f), maturity_(f.maturity_), fx_fwd_(f.fx_fwd_), q_(f.q_), 
								qu_dec_(f.qu_dec_), name_(f.name_), is_post_spot_(f.is_post_spot_)
	{}
	
	void fx_fwd.init_fx_spot(fx_spot i_fx_spot)							
	
	{
		_fx.init(dynamic_cast<_fx>(i_fx_spot));
		string err_msg;
		QL_FAIL_COND(!check_comp(err_msg),err_msg);		
	}
	
	logical fx_fwd.check(out string  option(nullable) err_msg)
	{
		err_msg = '';
			
		if(!_fx.is_ok(err_msg))
			return false;
	
		if(!null(fx_fwd_) && !fx_fwd_.is_fx_fwd()){
			err_msg = strcat(["invalid type of instrument (fx fwd)"," [",name(),"]"]);
			return false;
		}
			
		if(!check_comp(err_msg))
			return false;
	
		return true;
	}
	
	logical fx_fwd.check_comp(out string option(nullable) err_msg)
	{
		err_msg = '';
			
		if(maturity_ == spot_settle()){
			err_msg = strcat(["invalid instrument (fwd maturity equals spot settlement)"," [",name(),"]"]);
			return false;
		}
	
		if(null(fx_fwd_)) //is ok but prevent the rest of the checks
			return true;
	
		if(trade_date() != fx_fwd_.trade_date()){
			err_msg = strcat(["invalid instrument (trade_date mismatch in fx_spot and fx_fwd)"," [",name(),"]"]);
			return false;
		}	
			
		string base = fx_fwd_.fx_base_ccy(); //not a tag anymore	
		if(!null(base) && base_ccy() != base){
			err_msg = strcat(["different base currency in fx fwd vs fx spot"," [",name(),"]"]);
			return false;
		}
	
		string price = fx_fwd_.fx_price_ccy(); 
		if(!null(price) &&  price_ccy() != price){
			err_msg = strcat(["different price currency in fx fwd vs fx spot"," [",name(),"]"]);
			return false;
		}
	
		return true;
	}
	
	_fx				fx_fwd.clone()							{ return new fx_fwd(this);}
	fx_inst_type 	fx_fwd.type()							{ return FXI_FWD;}
	string 			fx_fwd.name() 							{ return name_; }
	date 			fx_fwd.maturity() 						{ return maturity_; }
	date 			fx_fwd.settle_date() 					{ return maturity(); }
	instrument 		fx_fwd.inst() 							{ return fx_fwd_;}
	number 			fx_fwd.qu_dec()							{ return null(qu_dec_) ? 15: qu_dec_;}
	number 			fx_fwd.outright()						{ return fx_fwd.quote();	}
	number 			fx_fwd.outright(logical unit_quote,	logical no_rnd)	{ return fx_fwd.quote(unit_quote, no_rnd);	}
	number 			fx_fwd.swap()										{ return fx_fwd.swap(false, 0, false);}
	number 			fx_fwd.swap(logical unit_quote, logical no_rnd)		{ return fx_fwd.swap(unit_quote,0, no_rnd);}
	logical			fx_fwd.is_post_spot()					{ return is_post_spot_;}	
	
	void fx_fwd.set_quote(	number option(nullable)	q,
							logical unit_quote)
	{
		if(unit_quote)
			q *= scale();	
	
		if(null(fx_fwd_))
			q_ = q;
		else
			fx_fwd_ = fx_fwd_.set_quote(q);
	}
	
	number fx_fwd.quote()							
	{
		number q =  null(fx_fwd_) ? q_ : fx_fwd_.quote_orig();
		return round(q, integer(qu_dec()));	
	}
	
	number fx_fwd.quote(logical unit_quote,
						logical no_rnd)
	{
		number q =  null(fx_fwd_) ? q_ : fx_fwd_.quote_orig();
		q = no_rnd ? q : round(q, integer(qu_dec()));	
		return unit_quote ? q / scale() : q;	
	}
	
	number fx_fwd.outright(	disc_func 	df_base,
							disc_func 	df_price,
							logical 	unit_quote,
							number option(nullable) spot_adj,
							logical 	no_rnd)	
	{ 
		number	fx_ref	= spot_quote(false,no_rnd);
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
		date mat		= maturity();	
		date spot 		= spot_settle();
		if(mat < spot) {
			// assume all swaps before spot is 1 day swaps. In this case we can determine how to handle a spot adjustment 				
			// 1) if we calculate an o/n outright (i.e. mat=tradedate) we can use a spot adjustment
			// 2) if we calculate an t/n outright we do not need a spot adjustment
			// in both 1) and 2) we have mat < spot
			//date mat_next = cal().move_bus_days (mat, 1);
			date mat_next = _fx.maturity(mat, "BD1");
			if(mat_next < spot && !null(spot_adj)) {
				fx_ref 	= spot_adjustment(fx_ref,spot_adj);			
				spot 	= mat_next;		
			}
			//if(mat_next < spot && null(spot_adj))	
			//--> this outright is calculated relative to spot eg. an outright with mat = trade_date is independent of the t/n swap points										   											
		}
			
		number q = CORE_INT_FX.fx_df_to_f(trade_date(),mat,spot,fx_ref,df_base,df_price,!reverse_ric());
		q = no_rnd ? q : round(q, integer(qu_dec()));	
		return unit_quote ? q / scale() : q;
	}
	
	
	number fx_fwd.swap(	logical 				unit_quote, 
						number option(nullable) spot_adj,
						logical 				no_rnd)
	{		
		number	fx_ref	= spot_quote(false,no_rnd);
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
		date mat		= maturity();
		number fx_f 	= quote(false,no_rnd);
		QL_FAIL_COND(null(fx_f),strcat(["invalid outright quote"," [",name(),"]"]));	
		date spot 		= spot_settle();
		number adj_prec = floor(log(scale())/log(10.0)+0.5)	;
			
		if(mat < spot) {
			// assume all swaps before spot is 1 day swaps. In this case we can determine when a spot adjustment 
			// is necessary. 		
			// 1) if we calculate an o/n swap we need a spot adjustment
			// 2) if we calculate an t/n swap we do not need a spot adjustment
			// in both 1) and 2) we have mat < spot
			//date mat_next = cal().move_bus_days (mat, 1);
			date mat_next = _fx.maturity(mat, "BD1");
	
			QL_FAIL_COND(mat_next < spot && null(spot_adj),strcat(["swap maturity before spot requires a spot adjustment"," [",name(),"]"]));			
			fx_ref 	= spot_adjustment(fx_ref,spot_adj);	
		}		
		number s = CORE_INT_FX.fx_f_to_swap(mat, spot, fx_ref, fx_f, dec(),scale());
		if(!no_rnd) {
			integer dec = integer(qu_dec() - spot_qu_dec());
			if(dec < 0) dec = 0;
			s = round(s, dec);
		}
		return unit_quote ? s / pow(10.0, dec()+ adj_prec) : s;	
	}
	
	instrument fx_fwd.synt_depo(	day_count_method 	dc_synt,
									disc_func			df,	
									logical 			synt_is_base,
									number option(nullable) spot_adj)
	{	
		number	fx_ref	= spot_quote();
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
		date mat		= maturity();	
		number fx_f 	= quote();
		QL_FAIL_COND(null(fx_f),strcat(["invalid outright quote"," [",name(),"]"]));	
		date spot 		= spot_settle();
		
		/*instr_def i_d= instr_def(synt_depo_cn);
		if(null(i_d))
			throw(E_INVALID_ARG, strcat(["invalid deposit class name"," [",name(),"]"]));*/					
	
		if(mat < spot) {
			// assume all depos before spot is a 1 day depo. In this case we can determine when a spot adjustment 
			// is necessary. 		
			// 1) if we calculate an o/n depo we need a spot adjustment
			// 2) if we calculate an t/n depo we do not need a spot adjustment
			// in both 1) and 2) we have mat < spot
			// if spot_adj is null and we have an o/n we will calculate a depo between t and spot
			//date mat_next = cal().move_bus_days (mat, 1);
			date mat_next = _fx.maturity(mat, "BD1");
			if(mat_next < spot && !null(spot_adj)) {//o/n														   												
				fx_ref 	= spot_adjustment(fx_ref,spot_adj);	
				spot 	= mat_next;	
			}
		}
	
		//day_count_method dc	= i_d.accr_dc_method();
		day_count_method dc_in	= DC_ACT_360;
		number rate_in 		= df.simple_rate(trade_date(),spot,mat, dc_in) ;
		string ne 			= strcat(["synt depo (", name(),")"]);
	
		if(synt_is_base){
				
			number fv_base,fv_price; 
			string ccy = fx_fwd_.fx_base_ccy();
			number r_base = CORE_INT_FX.fx_f_to_rate(mat, spot ,fx_ref, fx_f, null<number>,rate_in,dc_synt,dc_in, 	
													RT_SIMPLE,RT_SIMPLE,!reverse_ric(),CORE_INT_FX.FXP_BASE_R,fv_base,fv_price);
			if(mat < spot) 
				CORE_INT.flip (spot, mat);	
	
			instr_def i_d = create_depo_instr_def(dc_synt, BD_MOD_FOLLOWING, true, "bd2", calendar(),ccy);
			instrument depo = deposit(i_d, ne ,trade_date(),spot, spot, mat, r_base, r_base, fv_base * 100);
			return depo;//create_depo_instr(i_d, ne ,trade_date(),spot, spot, mat, r_base, r_base, fv_base * 100);
				
			//return create_depo(synt_depo_cn, ne ,trade_date(),spot,mat,r_base, spot, fv_base * 100);
			
		}
		else {
			number fv_base,fv_price;
			string ccy = fx_fwd_.fx_price_ccy(); 
			number r_price = CORE_INT_FX.fx_f_to_rate(mat, spot ,fx_ref, fx_f, rate_in,null<number>,dc_in,dc_synt, 	
													RT_SIMPLE,RT_SIMPLE,!reverse_ric(),CORE_INT_FX.FXP_PRICE_R,fv_base,fv_price);
			if(mat < spot) 
				CORE_INT.flip (spot, mat);
			instr_def i_d = create_depo_instr_def(dc_synt, BD_MOD_FOLLOWING, true, "bd2", calendar(),ccy);
	
			instrument depo = deposit(i_d, ne ,trade_date(),spot, spot, mat, r_price, r_price, fv_price * 100);
			return depo;//create_depo_instr(i_d, ne ,trade_date(),spot, spot, mat, r_price, r_price, fv_price * 100);
			//return create_depo(synt_depo_cn, ne ,trade_date(),spot,mat,r_price, spot, fv_price*100);
		}		
	}
	
	
	instrument fx_fwd.synt_depo(	day_count_method 	dc_synt,
									number				rate_in,								
									day_count_method 	dc_in,										
									logical 			synt_is_base,
									number option(nullable) spot_adj)
	{	
		number	fx_ref	= spot_quote();
		QL_FAIL_COND(null(fx_ref),strcat(["invalid spot quote"," [",name(),"]"]));	
		date mat		= maturity();	
		number fx_f 	= quote();
		QL_FAIL_COND(null(fx_f),strcat(["invalid outright quote"," [",name(),"]"]));	
		date spot 		= spot_settle();
	
		/*instr_def i_d_synt = instr_def(synt_depo_cn);
		if(null(i_d_synt))
			throw(E_INVALID_ARG, strcat(["invalid deposit class name (synt)"," [",name(),"]"]));	
	
		instr_def i_d_in = instr_def(depo_cn_in);
		if(null(i_d_in))
			throw(E_INVALID_ARG, strcat(["invalid deposit class name (in)"," [",name(),"]"]));	
	*/
		if(mat < spot) {
			// assume all depos before spot is a 1 day depo. In this case we can determine when a spot adjustment 
			// is necessary. 		
			// 1) if we calculate an o/n depo we need a spot adjustment
			// 2) if we calculate an t/n depo we do not need a spot adjustment
			// in both 1) and 2) we have mat < spot
			// if spot_adj is null and we have an o/n we will calculate a depo between t and spot
			//date mat_next = cal().move_bus_days (mat, 1);
			date mat_next = _fx.maturity(mat, "BD1");
			if(mat_next < spot && !null(spot_adj)) {//o/n														   												
				fx_ref 	= spot_adjustment(fx_ref,spot_adj);	
				spot 	= mat_next;	
			}
		}
		//day_count_method dc_synt 	= i_d_synt.accr_dc_method();
		//day_count_method dc_in 		= i_d_in.accr_dc_method();
		string ne 					= strcat(["synt depo (", name(),")"]);
		
		//Note! cashflows of the created depo is always nom=100, .face_amount() will return the last argument in create_depo(.)	
		if(synt_is_base){			
			number fv_base,fv_price; 
			string ccy = fx_fwd_.fx_base_ccy();
			number r_base = CORE_INT_FX.fx_f_to_rate(mat, spot ,fx_ref, fx_f, null<number>,rate_in,dc_synt,dc_in, 	
													RT_SIMPLE,RT_SIMPLE,!reverse_ric(),CORE_INT_FX.FXP_BASE_R,fv_base,fv_price);
			if(mat < spot) 
				CORE_INT.flip (spot, mat);
			instr_def i_d = create_depo_instr_def(dc_synt, BD_MOD_FOLLOWING, true, "bd2", calendar(),ccy);	
	
			instrument depo = deposit(i_d, ne ,trade_date(),spot, spot, mat, r_base, r_base,fv_base * 100);			
			return depo;//create_depo_instr(i_d, ne ,trade_date(),spot, spot, mat, r_base, r_base,fv_base * 100);
			//return create_depo(synt_depo_cn, ne ,trade_date(),spot,mat,r_base, spot, fv_base * 100);
			
		}
		else {
			number fv_base,fv_price; 
			string ccy = fx_fwd_.fx_price_ccy();
			number r_price = CORE_INT_FX.fx_f_to_rate(mat, spot ,fx_ref, fx_f, rate_in,null<number>,dc_in,dc_synt, 	
													RT_SIMPLE,RT_SIMPLE,!reverse_ric(),CORE_INT_FX.FXP_PRICE_R,fv_base,fv_price);
			if(mat < spot) 
				CORE_INT.flip (spot, mat);
			instr_def i_d = create_depo_instr_def(dc_synt, BD_MOD_FOLLOWING, true, "bd2", calendar(),ccy);
	
			instrument depo = deposit(i_d, ne ,trade_date(),spot, spot, mat, r_price, r_price, fv_price * 100);
			return depo;//create_depo_instr(i_d, ne ,trade_date(),spot, spot, mat, r_price, r_price, fv_price * 100);
			//return create_depo(synt_depo_cn, ne ,trade_date(),spot,mat,r_price, spot, fv_price*100);
		}		
	}
}

/*-----------------------------------------------------------------------
  Name :  	fx_outright_quote
  Notes :   handles a possible reverse of bid ask sides, //OBS spot_adj  
  Changes:
  			Date    		Who     Detail
  			12-Mar-2014 	MN      Original code 
  ----------------------------------------------------------------------*/	
/*void fx_outright_quote(	fx_spot fx_spot_bid, 
						fx_spot fx_spot_ask, 
						fx_swap fx_bid,
						fx_swap fx_ask,  					
						logical unit_quote,						
						out number option(nullable) bid, 
						out number option(nullable) ask)

{
	try {
		QL_FAIL_COND(fx_ask.base_ccy() != fx_bid.base_ccy(),"invalid fx swaps (different base currencies)");
		QL_FAIL_COND(fx_ask.price_ccy() != fx_bid.price_ccy(),"invalid fx swaps (different price currencies)");
		QL_FAIL_COND(fx_ask.maturity() != fx_bid.maturity(),"invalid fx swaps (different fx swap maturities)");
		QL_FAIL_COND(fx_ask.settle_date() != fx_bid.settle_date(),"invalid fx swaps (different fx swap settlement dates)");
		QL_FAIL_COND(fx_spot_bid.base_ccy() != fx_bid.base_ccy(),"invalid fx spot (different base currencies)");
		QL_FAIL_COND(fx_spot_bid.price_ccy() != fx_bid.price_ccy(),"invalid fx spot (different price currencies)");
		QL_FAIL_COND(fx_spot_ask.base_ccy() != fx_bid.base_ccy(),"invalid fx spot (different base currencies)");
		QL_FAIL_COND(fx_spot_ask.price_ccy() != fx_bid.price_ccy(),"invalid fx spot (different price currencies)");				
		QL_FAIL_COND(fx_spot_ask.spot_settle() != fx_spot_bid.spot_settle(),"invalid fx spot (different spot settlement)");
		QL_FAIL_COND(fx_ask.maturity() != fx_bid.maturity(),"invalid fx swap (different maturities)");

		fx_swap fx_bid_flip, fx_ask_flip;
		CORE_INT.fx_swap_flip(	fx_spot_bid,fx_spot_ask,fx_bid,fx_ask,  					
									unit_quote,fx_bid_flip,fx_ask_flip);

		bid = fx_bid_flip.outright(unit_quote,false);	//OBS spot_adj		
		ask = fx_ask_flip.outright(unit_quote,false);	//OBS spot_adj		
	}
	catch{
		CORE_INT.write_warning_message(true,err.message());		
		bid = ask = null<number>;
	}	
}*/


