option(null: hard);	// Throw error if a null value is used in a non-nullable context


module CORE_INT_COMPAT_FX
{
	/*-----------------------------------------------------------------------
	  Name :  	fx_swap_flip
	  Notes :   handles a possible reverse of bid ask sides   
	  Changes:
	  			Date    		Who     Detail
	  			12-Mar-2014 	MN      Original code 
	  ----------------------------------------------------------------------*/	
	void fx_swap_flip(	fx_spot fx_spot_bid, 
						fx_spot fx_spot_ask, 
						fx_swap fx_bid_in,
						fx_swap fx_ask_in,  					
						logical unit_quote,
						out fx_swap option(nullable) fx_bid, 
						out fx_swap option(nullable) fx_ask)
	{														
		number b = fx_bid_in.quote();
		number a = fx_ask_in.quote(); 
		
		if(null(a) && null(b)) {
			fx_bid = new fx_swap(fx_bid_in);
			fx_ask = new fx_swap(fx_ask_in);
			
			fx_bid.init_fx_spot(fx_spot_bid);
			fx_ask.init_fx_spot(fx_spot_ask);
			return;
		}

		if(fx_bid_in.maturity() <= fx_spot_bid.spot_settle()) {//only O/N and T/N if spot is t+2

			fx_bid = new fx_swap(fx_ask_in);
			fx_ask = new fx_swap(fx_bid_in);

			if(b > a) {
				fx_bid.set_quote(-a);
				fx_ask.set_quote(-b);
			}

			fx_bid.init_fx_spot(fx_spot_bid);
			fx_ask.init_fx_spot(fx_spot_ask);
									
		}
		else {
			fx_bid = new fx_swap(fx_bid_in);
			fx_ask = new fx_swap(fx_ask_in);

			if(b > a) {
				fx_bid.set_quote(-b);
				fx_ask.set_quote(-a);
			}
			fx_bid.init_fx_spot(fx_spot_bid);
			fx_ask.init_fx_spot(fx_spot_ask);			 
			
		}			
	}

	/*-----------------------------------------------------------------------
	  Name :  	fx_outright_flip
	  Notes :   handles a possible reverse of bid ask sides   
	  Changes:
	  			Date    		Who     Detail
	  			12-Mar-2014 	MN      Original code 
	  ----------------------------------------------------------------------*/	
	void fx_outright_flip(	fx_spot fx_spot_bid, 
							fx_spot fx_spot_ask, 
							fx_fwd 	fx_bid_in,
							fx_fwd 	fx_ask_in,  					
							logical unit_quote,
							out fx_fwd option(nullable)	fx_bid, 
							out fx_fwd option(nullable)	fx_ask)
	{	
		
		if(fx_bid_in.maturity() < fx_spot_bid.spot_settle()) {
			fx_bid = new fx_fwd(fx_ask_in);
			fx_ask = new fx_fwd(fx_bid_in);
			
			fx_bid.init_fx_spot(fx_spot_ask);
			fx_ask.init_fx_spot(fx_spot_bid);
									
		}
		else {
			fx_bid = new fx_fwd(fx_bid_in);
			fx_ask = new fx_fwd(fx_ask_in);
			
			fx_bid.init_fx_spot(fx_spot_bid);
			fx_ask.init_fx_spot(fx_spot_ask);			 
			
		}			
	}

	vector(_fx)	clone_vector_deep(vector(_fx) option(nullable) vi)
	{
		if(null(vi))
			return null<vector(_fx)>;
	
		integer s = v_size(vi);
		if(s == 0)
			return null<vector(_fx)>;
	
		vector(_fx) vi_new[s];
		for(integer i = 0; i < s; i++) 
			vi_new[i] = vi[i].clone();
		return vi_new;								
	}




	/*-----------------------------------------------------------------------
	  Name :  	class fx_curve
	  Notes :   This curve can contain a mix of either fx swaps or fx fwds with one restriction that before spot it must be 
				either fx swaps or fx fwds. The purpose of this class is to produce for example fx implied disc_funcs, synthetic deposits,
				interpolated fwd fx quotes (interpolation is done via disc_funcs i.e. no interpolation directly on fx fwd quotes) , ...       
	  Changes:
	  			Date    		Who     Detail
	  			11-Dec-2013 	MN      Original code (from fiws but redesigned)
	  ----------------------------------------------------------------------*/	
	class fx_curve 
	{
	   	public:
			/**/fx_curve(curve, instrument option(nullable));
			/**/fx_curve(curve, fx_spot option(nullable) );
			/**/fx_curve(vector(_fx));								
			/**/fx_curve(fx_curve);
	
			/**/void 			swaps_pre_spot(	fx_spot,fx_spot,vector(_fx),vector(_fx),logical,fx_inst_type option(nullable),
											logical,logical,logical,out vector(date) option(nullable),out vector(number) option(nullable));
			/**/void 			swaps_pre_spot(logical,fx_inst_type option(nullable),logical ,logical,out vector(date) option(nullable),out vector(number) option(nullable));
			/**/void 			swaps_post_spot(logical,fx_inst_type option(nullable),logical ,logical,out vector(date) option(nullable),out vector(number) option(nullable));
			/**/void 			swaps(logical,fx_inst_type option(nullable),logical ,logical,logical,out vector(date) option(nullable),out vector(number) option(nullable));
	
			/**/void 			outrights_pre_spot(fx_spot,fx_spot,vector(_fx),vector(_fx),logical,fx_inst_type option(nullable),
												logical,logical,logical,out vector(date) option(nullable),out vector(number) option(nullable));
			/**/void 			outrights_pre_spot(logical,fx_inst_type option(nullable),logical ,logical,out vector(date) option(nullable),out vector(number) option(nullable));
			/**/void 			outrights_post_spot(logical,fx_inst_type option(nullable),logical ,logical,out vector(date) option(nullable),out vector(number) option(nullable));
			/**/void			outrights(logical,fx_inst_type option(nullable),logical ,logical,logical,out vector(date) option(nullable),out vector(number) option(nullable));
	
			/**/void			quotes_df_pre_spot(disc_func,disc_func,logical,fx_inst_type option(nullable),logical,logical,out vector(date) option(nullable),out vector(number) option(nullable));
			/**/void			quotes_df_post_spot(disc_func,disc_func,logical,fx_inst_type option(nullable),logical,logical,
												out vector(date) option(nullable),out vector(number) option(nullable));
	
			/**/void			quotes_df(disc_func,disc_func,logical,fx_inst_type option(nullable),logical,logical,out vector(date) option(nullable), out vector(number) option(nullable));
			/**/vector(number)	quotes_df(disc_func,disc_func, logical);
	
			/**/vector(number)	quotes(fx_inst_type option(nullable) ,logical);
			void			set_quotes(vector(number), fx_inst_type option(nullable), logical);
			/**/void			set_spot_quote(number);
	
			/**/vector(date) 	maturity(fx_inst_type option(nullable) type, logical all);
			/**/vector(date) 	settle_date(fx_inst_type option(nullable) type, logical all);
	
			/**/disc_func		disc_base(disc_func,interpolator,rate_type,logical nr = false);
			/**/disc_func		disc_price(disc_func,interpolator,rate_type,logical nr = false);			
			/**/curve			synt_depo_base(	/*instr_class_name,*/disc_func,interpolator,rate_type,day_count_method ,
											out vector(date) option(nullable), logical nr = false, date option(nullable) max = null<date>);
			/**/curve			synt_depo_price(/*instr_class_name,*/disc_func,interpolator,rate_type,day_count_method ,
											out vector(date) option(nullable), logical nr = false, date option(nullable) max = null<date>);
			/**/number			ip_outright(disc_func,disc_func,date, logical);
			fx_spot			fx_spot_inst(logical ref = true);
	
			/**/vector(_fx)			fx_inst_pre_spot(fx_inst_type option(nullable), logical, logical ref ) ;
			/**/vector(_fx)			fx_inst_post_spot(fx_inst_type option(nullable), logical, logical ref); 
			/**/vector(_fx)			fx_inst(fx_inst_type option(nullable), logical, logical ref );
			/**/vector(_fx)			fx_inst(logical); 
			vector(instrument)	inst_pre_spot(fx_inst_type option(nullable),logical ) ;
			vector(instrument)	inst_post_spot(fx_inst_type option(nullable),logical ); 
			vector(instrument)	inst(fx_inst_type option(nullable), logical ); 
			/**/vector(logical)		inst_in_curve(fx_inst_type option(nullable));		
	
		protected:
			
			/**/vector(_fx) 	s_pre(logical);
			/**/vector(_fx) 	f_pre(logical);
			/**/vector(_fx) 	post(logical);
			/**/vector(fx_inst_type) type_post() ;
			/**/void			check_first_inst(instrument);
			/**/void 			init_type(vector(instrument));
			/**/void 			init_type(vector(_fx));
			/**/void 			init_int(vector(_fx), vector(date), vector(logical));
			/**/void			init(vector(instrument));
			/**/void			init(vector(_fx));
			/**/logical			check(out string option(nullable));
			/**/void 			check_ccy(vector(_fx), string, string);
			/**/_fx 			create_fx_inst(instrument , fx_inst_type , fx_spot );
	
			vector(_fx) 		create_fx_inst(vector(instrument) , vector(fx_inst_type) , fx_spot );
	
			/**/void 			outrights_pre_spot(logical,fx_inst_type option(nullable),logical,logical,out vector(date) option(nullable),out vector(number) option(nullable), vector(_fx) option(nullable));
			/**/void 			swaps_pre_spot(logical,fx_inst_type option(nullable),logical, logical,out vector(date) option(nullable),out vector(number) option(nullable), vector(_fx) option(nullable));		
			/**/disc_func		disc(disc_func,logical,interpolator,rate_type,logical,out vector(date) option(nullable),out vector(number) option(nullable));
			/**/curve			synt_depo(/*instr_class_name,*/disc_func,logical ,interpolator,rate_type,day_count_method,logical,date option(nullable), out vector(date) option(nullable));
	
		private:					
			fx_spot				fx_spot_;
			vector(_fx)			s_pre_;
			vector(_fx)			f_pre_;
			vector(_fx)			post_;	
			vector(logical)		s_pre_l_;
			vector(logical)		f_pre_l_;
			vector(logical)		post_l_;			
			vector(fx_inst_type) type_;
			logical 			pre_is_swap_;
															
	};
	
	vector(fx_inst_type) fx_curve.type_post() {
		integer s_pre = pre_is_swap_? v_size(s_pre_) : v_size(f_pre_);
		return ..sub_vector(type_, s_pre, v_size(type_) - s_pre);
	}
	
	
	fx_curve.fx_curve(	curve 						fx_c,
						instrument option(nullable) fx_s)
	{
		vector(instrument) 	fx_i = fx_c.instruments();
		QL_FAIL_COND(v_size(fx_i) == 0,"invalid fx curve");
	
		if(null(fx_s)){
			check_first_inst(fx_i[0]);			
			_fx fx		= fx_i[0].is_fx_fwd() ? new fx_fwd(fx_i[0]) : new fx_swap(fx_i[0]);
			fx_spot_	= new fx_spot(fx.spot_inst());
		}
		else
			fx_spot_ = new fx_spot(fx_s);
	
		init(fx_i);
	}	
	
	fx_curve.fx_curve(	curve 						fx_c,
						fx_spot option(nullable) 	fx_s)
	{
		vector(instrument) 	fx_i = fx_c.instruments();
		QL_FAIL_COND(v_size(fx_i) == 0,"invalid fx curve");
	
		if(null(fx_s)){	
			check_first_inst(fx_i[0]);		
			_fx fx		= fx_i[0].is_fx_fwd() ? new fx_fwd(fx_i[0]) : new fx_swap(fx_i[0]);
			fx_spot_	= new fx_spot(fx.spot_inst());
		}
		else
			fx_spot_ = new fx_spot(fx_s);
	
		init(fx_i);
	}
	
	fx_curve.fx_curve(vector(_fx) fx_i_in)
	{
		vector(_fx) fx_i = CORE_INT_COMPAT_FX.clone_vector_deep(fx_i_in);
			
		QL_FAIL_COND(v_size(fx_i) == 0,"invalid fx curve");	
	
		if(fx_i[0].type() != FXI_FWD && fx_i[0].type() != FXI_SWAP){
			string s = strcat(["invalid instrument in fx curve (must be fx swap or fx fwd) [",fx_i[0].inst().name(),"]"]);
			QL_FAIL(s);			
		}		
	
		vector(date) mat = fx_i.maturity();
		sort (fx_i, mat);
	
		fx_spot_ = new fx_spot(fx_i[0].spot_inst());
	
		init(fx_i);
	}
	
	fx_curve.fx_curve(fx_curve f):  fx_spot_(f.fx_spot_), s_pre_( CORE_INT_COMPAT_FX.clone_vector_deep(f.s_pre_)), 
									f_pre_( CORE_INT_COMPAT_FX.clone_vector_deep(f.f_pre_)), 
									post_( CORE_INT_COMPAT_FX.clone_vector_deep(f.post_)),
									s_pre_l_(!null(f.s_pre_l_) ? clone_vector(f.s_pre_l_) : null<vector(logical)>), 
									f_pre_l_(!null(f.f_pre_l_) ? clone_vector(f.f_pre_l_): null<vector(logical)>), 
									post_l_( !null(f.post_l_) ? clone_vector(f.post_l_): null<vector(logical)>), 
									type_(   !null(f.type_) ? clone_vector(f.type_): null<vector(fx_inst_type)>), 
									pre_is_swap_(f.pre_is_swap_)
	
	{}
	
	void fx_curve.check_first_inst(instrument fx_i)
	{
		if(!fx_i.is_fx_fwd() && !fx_i.is_fx_swap()){					
			string s = strcat(["invalid instrument in fx curve (must be fx swap or fx fwd) [",fx_i.name(),"]"]);
			QL_FAIL(s);			
		}	
	}
	
	void fx_curve.init_type(vector(instrument) fx_i)
	{
		integer size = v_size(fx_i);
		resize(type_,size);
		for(integer i = 0; i < size;i++){
			if(fx_i[i].is_fx_fwd())
				type_[i] = FXI_FWD;
			else if(fx_i[i].is_fx_swap())
				type_[i] = FXI_SWAP;
			else{			
				string s = strcat(["invalid instrument in fx curve (must be fx swap or fx fwd) [",fx_i[i].name(),"]"]);
				QL_FAIL(s);			
			}
		}
	}
	
	void fx_curve.init_type(vector(_fx) fx_i)
	{
		type_ = fx_i.type();
	
		for(integer i = 0; i < v_size(fx_i);i++){
			if(type_[i] != FXI_FWD && type_[i] != FXI_SWAP) {				
				string s = strcat(["invalid instrument in fx curve (must be fx swap or fx fwd) [",fx_i[i].inst().name(),"]"]);
				QL_FAIL(s);			
			}
		}
	}
	
	_fx fx_curve.create_fx_inst(instrument i, fx_inst_type t, fx_spot s)
	{
		if(t == FXI_FWD)
			return new fx_fwd(i,s);
		else if(t == FXI_SWAP)
			return new fx_swap(i,s);
		else
			return null<_fx>;		
	}
	
	vector(_fx) fx_curve.create_fx_inst(vector(instrument) i, vector(fx_inst_type) t, fx_spot s)
	{
		integer size = v_size(i);
		vector(_fx) f;
	
		for(integer j=0;j<size;j++){
			if(t[j] == FXI_FWD)
				push_back(f, new fx_fwd(i[j],s));
			else if(t[j] == FXI_SWAP)
				push_back(f, new fx_swap(i[j],s));
			else
				push_back(f, null<_fx>);
		}
		return f;		
	}
	
	void fx_curve.init_int(vector(_fx) fx, vector(date) mat, vector(logical) post_spot)
	{
		date spot_settle = fx_spot_.spot_settle();
	
		resize(s_pre_,0);
		resize(f_pre_,0);
		resize(post_,0);
		resize(s_pre_l_,0);
		resize(f_pre_l_,0);
		resize(post_l_,0);
	
		for(integer i = 0; i < v_size(fx);i++){
			
			if(type_[i] == FXI_FWD){						
				if(mat[i] < spot_settle) {
					push_back(f_pre_, fx[i]);
					logical incl = false;
					if(!post_spot[i]) {
						if(i == 0)
							incl = true;
						else if(i == 1 && mat[i] > mat[i-1])
							incl = true;
						else if(i == 1 && mat[i] == mat[i-1]){
							QL_FAIL_COND(v_size(f_pre_) != 2,"mixing fx_swap and fx fwd instruments before spot is not supported");
							incl = true;
							f_pre_l_[0] = false;
							f_pre_[0].set_quote(null<number>);
						}
					}
					push_back(f_pre_l_, incl);
					if(!incl)
						f_pre_[v_size(f_pre_)-1].set_quote(null<number>);				
				}	
				else if(mat[i] > spot_settle) {
					push_back(post_, fx[i]);
					logical incl = post_spot[i]  && (i == 0 || (i > 0 && mat[i] > mat[i-1]));					
					push_back(post_l_, incl);
					if(!incl)
						post_[v_size(post_)-1].set_quote(null<number>);
				}
				else {//spot
					push_back(post_, fx[i]);
					push_back(post_l_, false);
					post_[v_size(post_)-1].set_quote(null<number>);
				}	
			}
			else {//fx_swap
				
				if(mat[i] <= spot_settle)	{
					push_back(s_pre_, fx[i]);
					logical incl = false;
					if(!post_spot[i]) {
						if(i == 0)
							incl = true;
						else if(i==1 && mat[i] > mat[i-1])
							incl = true;
						else if(i==1 && mat[i] == mat[i-1]){
							QL_FAIL_COND(v_size(s_pre_) != 2,"mixing fx_swap and fx fwd instruments before spot is not supported");
							incl = true;
							s_pre_l_[0] = false;
							s_pre_[0].set_quote(null<number>);
						}					
					}
					push_back(s_pre_l_, incl);
					if(!incl)
						s_pre_[v_size(s_pre_)-1].set_quote(null<number>);
				}
				else {
					push_back(post_,fx[i]);
					logical incl = post_spot[i]  && (i == 0 || (i > 0 && mat[i] > mat[i-1]));
					push_back(post_l_, incl);
					if(!incl)
						post_[v_size(post_)-1].set_quote(null<number>);
				}
			}		
		} 
		integer size_p_s = v_size(s_pre_);
		integer size_p_f = v_size(f_pre_);
		QL_FAIL_COND(size_p_s > 0 && size_p_f > 0,"mixing fx_swap and fx fwd instruments before spot is not supported");
		pre_is_swap_ = size_p_f == 0;//note: pre_is_swap_ is true if size_p_s >= 0
	}
	
	void fx_curve.init(vector(instrument) fx_i)
	{	
		QL_FAIL_COND(null(fx_spot_),"invalid fx spot");
	
		init_type(fx_i);
		
		vector(_fx) fx 				= create_fx_inst(fx_i, type_, fx_spot_);
		vector(date) mat 			= fx_i.maturity(); 
		vector(logical) post_spot 	= equal_casei(fx_i.settle_code(), fx_spot_.spot_code()); 
	
		init_int(fx, mat, post_spot);
		string err_msg;
		check(err_msg);	
	}
	
	void fx_curve.init(vector(_fx) fx)
	{
		QL_FAIL_COND(null(fx_spot_),"invalid fx spot");
	
		init_type(fx);
	
		vector(date) mat 			= fx.maturity(); 
		vector(logical) post_spot 	= fx.is_post_spot(); 
	
		init_int(fx, mat, post_spot);
	
		for(integer i = 0; i < v_size(fx);i++){
			if(fx[i].spot_inst().id() != fx_spot_.spot_inst().id()){
				string s = strcat(["invalid fx spot instrument [",fx[i].inst().name(),"]"]);
				QL_FAIL(s);	
			}
		}
	
		string err_msg;
		check(err_msg);
	}
	
	logical fx_curve.check(out string  option(nullable) err_msg)
	{
		err_msg = '';
		
		integer size_p_s = v_size(s_pre_);
		integer size_p_f = v_size(f_pre_);
		
		vector(_fx)	fx_s_pre_tmp = s_pre(false);	
		integer size_p_s_e = v_size(fx_s_pre_tmp);
	
		QL_FAIL_COND(size_p_s_e > 0 && fx_s_pre_tmp[size_p_s_e - 1].maturity() != fx_spot_.spot_settle(),"missing fx swap (before spot)");
		
		for(integer i = 1; i < size_p_s_e; i++) {
			if(fx_s_pre_tmp[i].maturity() <= fx_s_pre_tmp[i - 1].maturity())			
				QL_FAIL(strcat(["duplicate (or unordered) maturities (fx swaps pre spot) [",fx_s_pre_tmp[i].name(),"]"]));
			if(fx_s_pre_tmp[i].settle_date() <= fx_s_pre_tmp[i - 1].settle_date())			
				QL_FAIL(strcat(["duplicate (or unordered) settlement dates (fx swaps pre spot) [",fx_s_pre_tmp[i].name(),"]"]));
			if(fx_s_pre_tmp[i].settle_date() != fx_s_pre_tmp[i - 1].maturity())			
				QL_FAIL(strcat(["non consecutive fx swap (fx swaps pre spot) [",fx_s_pre_tmp[i].name(),"]"]));
		}
	
		vector(_fx)	fx_f_pre_tmp = f_pre(false);	
		integer size_p_f_e = v_size(fx_f_pre_tmp);
	
		for(integer i = 1; i < size_p_f_e; i++) {
			if(fx_f_pre_tmp[i].maturity() <= fx_f_pre_tmp[i - 1].maturity())			
				QL_FAIL(strcat(["duplicate maturities (fx fwd pre spot) [",fx_f_pre_tmp[i].name(),"]"]));
		}	
	
		integer size = v_size(post_);
	
		vector(_fx)	fx_post_tmp = post(false);		
		integer size_e = v_size(fx_post_tmp);
		
		for(integer i=1; i<size_e; i++) {
			if(fx_post_tmp[i].maturity() <= fx_post_tmp[i - 1].maturity())
				QL_FAIL(strcat(["duplicate (or unordered) maturities (after spot) [",fx_post_tmp[i].name(),"]"]));		
		}
		date settle = null<date>;
		for(integer i=0; i<size_e; i++) {
			if(fx_post_tmp[i].type() != FXI_SWAP)
				continue;
			if(null(settle))
				settle = fx_post_tmp[i].settle_date();
			else {		
				if(fx_post_tmp[i].settle_date() != settle)
					QL_FAIL(strcat(["invalid settlement date (fx swap after spot) [",fx_post_tmp[i].name(),"]"]));
			}
		}
	
		string base_c = fx_spot_.base_ccy();
		string price_c = fx_spot_.price_ccy();
		check_ccy(fx_s_pre_tmp, base_c,price_c);
		check_ccy(fx_f_pre_tmp, base_c,price_c);
		check_ccy(fx_post_tmp, base_c,price_c);
			
		return true;
	}
	
	void fx_curve.check_ccy(vector(_fx) inst, string base_c, string price_c)
	{
		for(integer i=0; i < v_size(inst); i++) {
			QL_FAIL_COND(base_c!= inst[i].base_ccy(),strcat(["invalid base currency [",inst[i].name(),"]"]));
			QL_FAIL_COND(price_c!= inst[i].price_ccy(),strcat(["invalid price currency [",inst[i].name(),"]"]));
		}
	}
	
	vector(_fx) fx_curve.s_pre(logical all)
	{
		if(all) 
			return s_pre_;
		vector(integer) i = ..v_find(s_pre_l_, [true]);
		return ..sub_vector(s_pre_, i);
	}
	
	vector(_fx) fx_curve.f_pre(logical all)
	{
		if(all) 
			return f_pre_;
		vector(integer) i = ..v_find(f_pre_l_, [true]);
		return ..sub_vector(f_pre_, i);
	}
	vector(_fx) fx_curve.post(logical all)
	{
		if(all) 
			return post_;
		vector(integer) i = ..v_find(post_l_, [true]);
		return ..sub_vector(post_, i);
	}
	
	
	
	fx_spot 			fx_curve.fx_spot_inst(logical ref)												{ return ref ? fx_spot_: dynamic_cast<fx_spot>(fx_spot_.clone());}
	vector(number) 		fx_curve.quotes(fx_inst_type option(nullable) type, logical all)				{ return fx_inst(type,all,true).quote() ;}
	vector(instrument)	fx_curve.inst_pre_spot(fx_inst_type option(nullable) type, logical all) 		{ return fx_inst_pre_spot(type,all,false).inst();}
	vector(instrument)	fx_curve.inst_post_spot(fx_inst_type option(nullable) type, logical all) 		{ return fx_inst_post_spot(type,all,false).inst();}
	vector(instrument)	fx_curve.inst(fx_inst_type option(nullable) type, logical all) 					{ return fx_inst(type,all,false).inst();}
	void 				fx_curve.set_quotes(vector(number) quotes, 
											fx_inst_type option(nullable) type,logical all) 			{ fx_inst(type, all, true).set_quote(quotes) ;}
	
	vector(date)  fx_curve.maturity(fx_inst_type option(nullable) t, logical all)				
	{ 	
		if(!all)
			return fx_inst(t,all,true).maturity();
	
		//if all we would like to nullify maturity for not included instruments
	
		vector(_fx)	fx 			= fx_inst(t,true,true);
		vector(logical) fx_l 	= inst_in_curve(t);
		vector(date) m 			= fx.maturity();
	
		for(integer i = 0; i<v_size(fx) ; i++){
			if(!fx_l[i])
				m[i] = null<date>;		
		} 
		
		return m;
	
	}
	
	vector(date)  fx_curve.settle_date(fx_inst_type option(nullable) t, logical all)				
	{ 	
		if(!all)
			return fx_inst(t,all,true).settle_date();
	
		//if all we would like to nullify maturity for not included instruments
	
		vector(_fx)	fx 			= fx_inst(t,true,true);
		vector(logical) fx_l 	= inst_in_curve(t);
		vector(date) s 			= fx.settle_date();
	
		for(integer i = 0; i<v_size(fx) ; i++){
			if(!fx_l[i])
				s[i] = null<date>;		
		} 
		
		return s;
	
	}
	
	
	
	vector(logical) fx_curve.inst_in_curve(fx_inst_type option(nullable) t)
	{
		vector(logical) fx_l = pre_is_swap_ ? s_pre_l_ : f_pre_l_;
		fx_l = concat(fx_l, post_l_);
	
		if(null(t))
			return fx_l;
		else {	
			vector(integer) i = CORE_INT_COMPAT_FX.v_find_fx(type_, t);
			return ..sub_vector(fx_l, i);	
		}
	}
	
	void fx_curve.set_spot_quote(number quote) 		
	{ 
		fx_inst(null<fx_inst_type>,false,true).set_spot_quote(quote) ;
		fx_spot_.set_spot_quote(quote) ;
	}
	
	vector(_fx)	fx_curve.fx_inst_pre_spot(	fx_inst_type option(nullable) 	t, 
											logical 						all,
											logical 						ref) 	
	{ 	
		vector(_fx) fx;
	
		if(null(t))			
			fx = pre_is_swap_ ? s_pre(all) : f_pre(all);		
		else if(t == FXI_FWD)
			fx = f_pre(all);
		else if(t == FXI_SWAP)
			fx = s_pre(all);
		else
			return null<vector(_fx)>;
	
		return ref ? fx : CORE_INT_COMPAT_FX.clone_vector_deep(fx);
	}
	
	vector(_fx)	fx_curve.fx_inst_post_spot(	fx_inst_type option(nullable) 	t, 
											logical 						all,
											logical 						ref) 	
	{ 
		vector(_fx) fx;
	
		if(null(t))
			fx = post(all);
		else {			
			if(all) {
				vector(integer) i = CORE_INT_COMPAT_FX.v_find_fx(type_post(), t);
				fx = ..sub_vector(post_, i);
			}
			else {
				vector(integer) i 				= ..v_find(post_l_, [true]);
				vector(_fx) post_sub 			= ..sub_vector(post_, i);
				vector(fx_inst_type) type_sub 	= CORE_INT_COMPAT_FX.sub_vector_fx(type_post(), i);
				i 								= CORE_INT_COMPAT_FX.v_find_fx(type_sub, t);
				fx = ..sub_vector(post_sub, i);
			}			
		}
	
		return ref ? fx : CORE_INT_COMPAT_FX.clone_vector_deep(fx);
	}
	
	vector(_fx)	fx_curve.fx_inst(logical ref)
	{
		if(ref) {
			vector(_fx) fx 	= pre_is_swap_ ? s_pre(true) : f_pre(true);
			fx 				= !null(fx) ? clone_vector(fx) : null<vector(_fx)>;
			vector(_fx) fxp = post(true);
	 		fxp 			= !null(fxp) ? clone_vector(fxp) : null<vector(_fx)>;
			return concat(fx, fxp);
		}
		else {
			vector(_fx) fx = pre_is_swap_ ? CORE_INT_COMPAT_FX.clone_vector_deep(s_pre(true)) : CORE_INT_COMPAT_FX.clone_vector_deep(f_pre(true));
			return concat(fx, CORE_INT_COMPAT_FX.clone_vector_deep(post(true)));
		}
	}
	
	vector(_fx)	fx_curve.fx_inst(	fx_inst_type option(nullable) 	type, 
									logical 						all,
									logical 						ref) 	
	{ 
		vector(_fx)	fx_pre = fx_inst_pre_spot(type, all,ref); 
		vector(_fx)	fx_post = fx_inst_post_spot(type, all,ref);
		return concat(fx_pre,fx_post);
	}
	
	void fx_curve.swaps_pre_spot(	logical 			unit_quote,
									fx_inst_type option(nullable) type,
									logical 			no_rnd,
									logical				all,
									out vector(date)  option(nullable)	mat,
									out vector(number) option(nullable) 	quote)
	{	
		swaps_pre_spot(	unit_quote,type,no_rnd,all,mat,quote, null<vector(_fx)>);
	}
	
	void fx_curve.swaps_pre_spot(	fx_spot 			fx_spot_bid,
									fx_spot 			fx_spot_ask,
									vector(_fx) 		fx_f_pre_bid,
									vector(_fx) 		fx_f_pre_ask,
									logical 			unit_quote,
									fx_inst_type option(nullable) type,
									logical 			no_rnd,
									logical 			calc_bid,
									logical				all,
									out vector(date)  option(nullable)	mat,
									out vector(number) option(nullable) 	quote)
	{	
		vector(_fx) fx_f_pre;
	
		if(null(type) ||  (!null(type) && type == FXI_FWD)){
			for(integer i=0; i< v_size(fx_f_pre_bid); i++) {
				fx_fwd fx_bid_flip, fx_ask_flip;
				fx_fwd f_b = dynamic_cast<fx_fwd>(fx_f_pre_bid[i]);
				fx_fwd f_a = dynamic_cast<fx_fwd>(fx_f_pre_ask[i]);
				CORE_INT_COMPAT_FX.fx_outright_flip(	fx_spot_bid,fx_spot_ask,f_b,f_a,  					
											unit_quote,fx_bid_flip,fx_ask_flip);
				if(calc_bid)
					push_back(fx_f_pre,fx_bid_flip);
				else
					push_back(fx_f_pre,fx_ask_flip);
			}
		}
		swaps_pre_spot(	unit_quote,type,no_rnd,all,mat,quote, fx_f_pre);
	}
	
	void fx_curve.swaps_pre_spot(	logical 						unit_quote,
									fx_inst_type option(nullable) 	type,
									logical 						no_rnd,
									logical							all,
									out vector(date)  option(nullable)				mat,
									out vector(number)  option(nullable)				quote,
									vector(_fx)	option(nullable) 	fx_f_pr)
	{	
		logical incl_fx_swaps = null(type) ||  (!null(type) && type == FXI_SWAP);
		logical incl_fx_fwds = null(type) ||  (!null(type) && type == FXI_FWD);
	
		if(incl_fx_fwds) {
			if(null(fx_f_pr))
				fx_f_pr = f_pre(all);
			else if(v_size(f_pre(all)) != v_size(fx_f_pr))
				QL_FAIL("size mismatch (internal error)");	
		}
	
		resize(mat, 0);
		resize(quote, 0);
		
		integer size_p_f = v_size(fx_f_pr);
	
		vector(_fx)	fx_s_pr = s_pre(all);
		integer size_p_s = v_size(fx_s_pr);
	
		if(size_p_f > 0 && incl_fx_fwds) {
	
			number spot_adj = null<number>;
			for(integer i = size_p_f-1; i >= 0; i--){
				number f = fx_f_pr[i].swap(unit_quote,spot_adj,no_rnd);
				QL_FAIL_COND(null(f),strcat(["invalid fx swap (from fwd outright) [",fx_f_pr[i].name(),"]"]));			
				push_back(quote,f);	
				push_back(mat,fx_f_pr[i].maturity());
				spot_adj = null(spot_adj) ? f : spot_adj + f;	//OBS				
			}
		}
		else if(size_p_s > 0 && incl_fx_swaps) {
			/*if(fx_s_pre[size_p_s-1].maturity() != spot_settle)
				throw(E_INVALID_ARG, "missing fx swap (before spot)"); -> already checked*/
					
			for(integer i = size_p_s-1; i >= 0; i--){
				
				number f = fx_s_pr[i].swap(unit_quote, no_rnd);
				QL_FAIL_COND(null(f),strcat(["invalid fx swap [",fx_s_pr[i].name(),"]"]));	
							
				push_back(quote,f);	
				push_back(mat,fx_s_pr[i].settle_date());
				//spot_adj = null(spot_adj) ? fx_s_pre_[i].quote(false,no_rnd) : spot_adj + fx_s_pre_[i].quote(false,no_rnd);	//OBS		
			}			
		}
	
		sort(quote, mat);	
		sort(mat, mat);		
	}
	
	void fx_curve.swaps_post_spot(	logical 			unit_quote,
									fx_inst_type option(nullable) type,
									logical 			no_rnd,
									logical				all,
									out vector(date) option(nullable) 	mat,
									out vector(number) option(nullable) 	quote)
	{	
		logical incl_fx_swaps = null(type) ||  (!null(type) && type == FXI_SWAP);
		logical incl_fx_fwds = null(type) ||  (!null(type) && type == FXI_FWD);
	
		resize(mat, 0);
		resize(quote, 0);
	
		vector(_fx)	fx = post(all);
	
		for(integer i = 0; i < v_size(fx); i++){
			
			if(!incl_fx_fwds && fx[i].type() == FXI_FWD)
				continue;
			
			if(!incl_fx_swaps && fx[i].type() == FXI_SWAP)
				continue;
	
			number f = fx[i].swap(unit_quote,no_rnd);
			QL_FAIL_COND(null(f),strcat(["invalid fx swap [",fx[i].name(),"]"]));	
				
			push_back(quote,f);	
			push_back(mat,fx[i].maturity());				
		}		
	}
	
	void fx_curve.swaps(logical 			unit_quote,
						fx_inst_type option(nullable) type,
						logical 			no_rnd,					
						logical 			incl_spot,
						logical				all,
						out vector(date)  option(nullable)	mat,
						out vector(number)  option(nullable)	quote)
	{	
		swaps_pre_spot(	unit_quote,type,no_rnd,all,mat,quote);
	
		if(incl_spot) {
			push_back(quote, fx_spot_.spot_quote(unit_quote, no_rnd));	
			push_back(mat, fx_spot_.spot_settle());
		}
	
		vector(date) 	mat_post;
		vector(number) 	quote_post;
		swaps_post_spot(unit_quote,type,no_rnd,all,mat_post,quote_post);
	
		mat		= concat(mat, mat_post);
		quote	= concat(quote, quote_post);		
	}
	
	void fx_curve.outrights_pre_spot(	logical 			unit_quote,
										fx_inst_type option(nullable) type,
										logical 			no_rnd,
										logical				all,
										out vector(date)  option(nullable)	mat_fwd,
										out vector(number) option(nullable) 	quote_fwd)
	{	
		outrights_pre_spot(	unit_quote,type,no_rnd,all,mat_fwd,quote_fwd, null<vector(_fx)>);
	}
	
	void fx_curve.outrights_pre_spot(	fx_spot 			fx_spot_bid,
										fx_spot 			fx_spot_ask,
										vector(_fx) 		fx_s_pre_bid,
										vector(_fx) 		fx_s_pre_ask,
										logical 			unit_quote,
										fx_inst_type option(nullable) type,
										logical 			no_rnd,
										logical 			calc_bid,
										logical				all,
										out vector(date)  option(nullable)	mat_fwd,
										out vector(number)  option(nullable)	quote_fwd)
	{	
		vector(_fx) fx_s_pre;
	
		if(null(type) ||  (!null(type) && type == FXI_SWAP)){
			for(integer i=0; i< v_size(fx_s_pre_bid); i++) {
				fx_swap fx_bid_flip, fx_ask_flip;
				fx_swap s_b = dynamic_cast<fx_swap>(fx_s_pre_bid[i]);
				fx_swap s_a = dynamic_cast<fx_swap>(fx_s_pre_ask[i]);
				CORE_INT_COMPAT_FX.fx_swap_flip(	fx_spot_bid,fx_spot_ask,s_b,s_a,  					
											unit_quote,fx_bid_flip,fx_ask_flip);
				if(calc_bid)
					push_back(fx_s_pre,fx_bid_flip);
				else
					push_back(fx_s_pre,fx_ask_flip);
			}
		}
		outrights_pre_spot(	unit_quote,type,no_rnd,all,mat_fwd,quote_fwd, fx_s_pre);
	}
	
	
	
	void fx_curve.outrights_pre_spot(	logical 			unit_quote,
										fx_inst_type option(nullable) type,
										logical 			no_rnd,
										logical				all,
										out vector(date) option(nullable) 	mat_fwd,
										out vector(number) option(nullable) 	quote_fwd,
										vector(_fx)	option(nullable) fx_s_pr)
	{	
		logical incl_fx_swaps = null(type) ||  (!null(type) && type == FXI_SWAP);
		logical incl_fx_fwds = null(type) ||  (!null(type) && type == FXI_FWD);
	
		if(incl_fx_swaps) {
			if(null(fx_s_pr))
				fx_s_pr = s_pre(all);
			else if(v_size(s_pre(all)) != v_size(fx_s_pr))
				QL_FAIL("size mismatch (internal error)");	
		}
	
		resize(mat_fwd, 0);
		resize(quote_fwd, 0);
		
		vector(_fx)	fx_f_pr = f_pre(all);
		integer size_p_f = v_size(fx_f_pr);
		integer size_p_s = v_size(fx_s_pr);
	
		if(size_p_f > 0 && incl_fx_fwds) {
			
			for(integer i = size_p_f-1; i >= 0; i--){
				number f = fx_f_pr[i].outright(unit_quote,no_rnd);
				QL_FAIL_COND(null(f),strcat(["invalid outright [",fx_f_pr[i].name(),"]"]));			
				push_back(quote_fwd,f);	
				push_back(mat_fwd,fx_f_pr[i].maturity());				
			}
		}
		else if(size_p_s > 0 && incl_fx_swaps) {
			/*if(fx_s_pre[size_p_s-1].maturity() != spot_settle)
				throw(E_INVALID_ARG, "missing fx swap (before spot)"); -> already checked*/
			
			number spot_adj = null<number>;
			for(integer i = size_p_s-1; i >= 0; i--){
				
				number f = fx_s_pr[i].outright(unit_quote, spot_adj,no_rnd);
				QL_FAIL_COND(null(f),strcat(["invalid outright (from fx swap) [",fx_s_pr[i].name(),"]"]));
							
				push_back(quote_fwd,f);	
				push_back(mat_fwd,fx_s_pr[i].settle_date());
				spot_adj = null(spot_adj) ? fx_s_pr[i].quote(false,no_rnd) : spot_adj + fx_s_pr[i].quote(false,no_rnd);			
			}			
		}
	
		sort(quote_fwd, mat_fwd);	
		sort(mat_fwd, mat_fwd);	
	}
	
	
	void fx_curve.outrights_post_spot(	logical 			unit_quote,
										fx_inst_type option(nullable) type,
										logical 			no_rnd,
										logical				all,
										out vector(date)  option(nullable)	mat_fwd,
										out vector(number) option(nullable) 	quote_fwd)
	{	
		logical incl_fx_swaps = null(type) ||  (!null(type) && type == FXI_SWAP);
		logical incl_fx_fwds = null(type) ||  (!null(type) && type == FXI_FWD);
	
		resize(mat_fwd, 0);
		resize(quote_fwd, 0);
		vector(_fx)	fx = post(all);
	
		for(integer i = 0; i < v_size(fx); i++){
			
			if(!incl_fx_fwds && fx[i].type() == FXI_FWD)
				continue;
			
			if(!incl_fx_swaps && fx[i].type() == FXI_SWAP)
				continue;
	
			number f = fx[i].outright(unit_quote,no_rnd);
			QL_FAIL_COND(null(f),strcat(["invalid outright [",fx[i].name(),"]"]));
				
			push_back(quote_fwd,f);	
			push_back(mat_fwd,fx[i].maturity());				
		}		
	}
	
	void fx_curve.outrights(logical 			unit_quote,
							fx_inst_type option(nullable) type,
							logical 			no_rnd,
							logical 			incl_spot,
							logical				all,
							out vector(date) option(nullable) 	mat_fwd,
							out vector(number) option(nullable) 	quote_fwd)
	{	
		outrights_pre_spot(	unit_quote,type,no_rnd,all,mat_fwd,quote_fwd);
	
		if(incl_spot) {
			push_back(quote_fwd, fx_spot_.spot_quote(unit_quote, no_rnd));	
			push_back(mat_fwd, fx_spot_.spot_settle());
		}
	
		vector(date) 	mat_fwd_post;
		vector(number) 	quote_fwd_post;
		outrights_post_spot(unit_quote,type,no_rnd,all,mat_fwd_post,quote_fwd_post);
	
		mat_fwd 	= concat(mat_fwd, mat_fwd_post);
		quote_fwd 	= concat(quote_fwd, quote_fwd_post);			
	}
	
	void fx_curve.quotes_df_pre_spot(	disc_func 			df_base,
										disc_func 			df_price,
										logical 			unit_quote,
										fx_inst_type option(nullable) type,
										logical 			no_rnd,
										logical				all,
										out vector(date) option(nullable) 	mat,
										out vector(number) option(nullable) 	quote)
	{	
		logical incl_fx_swaps = null(type) ||  (!null(type) && type == FXI_SWAP);
		logical incl_fx_fwds = null(type) ||  (!null(type) && type == FXI_FWD);
		
		resize(mat, 0);
		resize(quote, 0);
		
		vector(_fx)	fx_f_pr = f_pre(all);
		integer size_p_f = v_size(fx_f_pr);
		vector(_fx)	fx_s_pr = s_pre(all);
		integer size_p_s = v_size(fx_s_pr);
	
		if(size_p_f > 0 && incl_fx_fwds) {
			
			for(integer i = size_p_f-1; i >= 0; i--){
				number f = fx_f_pr[i].outright(df_base,df_price,unit_quote,null<number>,no_rnd);
				QL_FAIL_COND(null(f),strcat(["invalid outright (from disc_func) [",fx_f_pr[i].name(),"]"]));		
				push_back(quote,f);	
				push_back(mat,fx_f_pr[i].maturity());								
			}
		}
		else if(size_p_s > 0 && incl_fx_swaps) {
			
			number spot_adj = null<number>;
			for(integer i = size_p_s-1; i >= 0; i--){
				
				number f = fx_s_pr[i].swap(df_base,df_price,unit_quote, spot_adj,no_rnd);
				QL_FAIL_COND(null(f),strcat(["invalid fx swap (from disc_func) [",fx_s_pr[i].name(),"]"]));
							
				push_back(quote,f);	
				push_back(mat,fx_s_pr[i].settle_date());
				spot_adj = null(spot_adj) ? f : spot_adj + f;			
			}			
		}
	
		sort(quote, mat);	
		sort(mat, mat);	
	}
	
	void fx_curve.quotes_df_post_spot(	disc_func 			df_base,
										disc_func 			df_price,
										logical 			unit_quote,
										fx_inst_type option(nullable) type,
										logical 			no_rnd,
										logical				all,
										out vector(date) option(nullable) 	mat,
										out vector(number) option(nullable) 	quote)
	{	
		logical incl_fx_swaps = null(type) ||  (!null(type) && type == FXI_SWAP);
		logical incl_fx_fwds = null(type) ||  (!null(type) && type == FXI_FWD);
	
		resize(mat, 0);
		resize(quote, 0);
		vector(_fx)	fx = post(all);
	
		for(integer i = 0; i < v_size(fx); i++){
			
			number f;
			if(incl_fx_fwds && fx[i].type() == FXI_FWD) {
				f = fx[i].outright(df_base,df_price,unit_quote,null<number>,no_rnd);			
			}				
			else if(incl_fx_swaps && fx[i].type() == FXI_SWAP) {
				f = fx[i].swap(df_base,df_price,unit_quote,null<number>,no_rnd);
			}	
								
			push_back(quote,f);	
			push_back(mat,fx[i].maturity());				
		}		
	}
	
	vector(number) fx_curve.quotes_df(	disc_func df_base,
										disc_func df_price,
										logical all)
	{	
		vector(date) 	mat;
		vector(number) 	quote;
		fx_curve.quotes_df(df_base,df_price,false,null<fx_inst_type>,false,all,mat,quote);
		return quote;	
	}
	
	void fx_curve.quotes_df(disc_func 			df_base,
							disc_func 			df_price,
							logical 			unit_quote,
							fx_inst_type option(nullable) type,
							logical 			no_rnd,	
							logical				all,					
							out vector(date) option(nullable) 	mat,
							out vector(number) option(nullable) 	quote)
	{	
		quotes_df_pre_spot(df_base,df_price,unit_quote,type,no_rnd,all,mat,quote);	
	
		vector(date) 	mat_post;
		vector(number) 	quote_post;
		quotes_df_post_spot(df_base,df_price,unit_quote,type,no_rnd,all,mat_post,quote_post);
	
		mat		= concat(mat, mat_post);
		quote	= concat(quote, quote_post);		
	}
	
	disc_func fx_curve.disc(	disc_func 			df,
								logical 			df_is_price,
								interpolator 		ip,
								rate_type			ip_rt,
								logical 			no_rnd,							
								out vector(date) option(nullable) 	mat_fwd,
								out vector(number) option(nullable) 	quote_fwd)
	{						
		outrights(true, null<fx_inst_type>,no_rnd,true, false,mat_fwd, quote_fwd);
	
		number size = v_size(quote_fwd);
		vector(number)	v_df;
		vector(number)	v_per;
		
		date d 				= fx_spot_.trade_date();
		date spot_settle	= fx_spot_.spot_settle();
	
		number dff;
		if(mat_fwd[0] > d){
			number p 	= (mat_fwd[0] - d)/365;
			dff 		= df.disc(p);//approx
			push_back(v_df, dff);
			push_back(v_per, p);	
		}
		else 
			dff = 1;
		
		for(integer i = 1; i < size; i++){
	
			number fx_start = quote_fwd[i-1];
			number fx_end = quote_fwd[i];
			
			if(mat_fwd[i] <= spot_settle) 
				CORE_INT.flip (fx_start, fx_end) ;
	
			if(df_is_price) 
				dff *= CORE_INT_FX.fx_f_to_df(	d,mat_fwd[i-1], mat_fwd[i],spot_settle,fx_start, fx_end,null<disc_func>,df,  
											!fx_spot_.reverse_ric(),CORE_INT_FX.FXP_BASE_R);
			else
				dff *= CORE_INT_FX.fx_f_to_df(	d,mat_fwd[i-1], mat_fwd[i],spot_settle,fx_start, fx_end,df,null<disc_func>,  
											!fx_spot_.reverse_ric(),CORE_INT_FX.FXP_PRICE_R);
			if(!null(dff)){
				push_back(v_df, dff);
				push_back(v_per, (mat_fwd[i] - d)/365);	
			}
		}		
		
		return disc_func_interp(v_per, v_df, ip, ip_rt);	
	}
	
	disc_func fx_curve.disc_base(	disc_func 	df_price,										
									interpolator ip,
									rate_type	ip_rt,
									logical 	no_rnd)
	{	
		vector(date) 	mat_fwd;
		vector(number) 	quote_fwd;			
		return disc(df_price,true,ip,ip_rt, no_rnd,mat_fwd, quote_fwd)	;	
	}
	
	disc_func fx_curve.disc_price(	disc_func 	df_base,										
									interpolator ip,
									rate_type	ip_rt,
									logical 	no_rnd)
	{	
		vector(date) 	mat_fwd;
		vector(number) 	quote_fwd;			
		return disc(df_base,false,ip,ip_rt, no_rnd,mat_fwd, quote_fwd)	;	
	}
	
	number fx_curve.ip_outright(disc_func 	df_base,
								disc_func 	df_price,
								date 		mat,
								logical 	unit_quote)
	{	
		logical no_rnd = false;
		date d 				= fx_spot_.trade_date();
		date spot_settle	= fx_spot_.spot_settle();
		number fx_sp		= fx_spot_.spot_quote(unit_quote, no_rnd);									
		return CORE_INT_FX.fx_df_to_f(d, mat, spot_settle, fx_sp, df_base, df_price,!fx_spot_.reverse_ric());	
	}
	
	curve fx_curve.synt_depo(	/*instr_class_name  	depo_cn,*/
								disc_func 			df,	
								logical 			df_is_price,
								interpolator 		ip,
								rate_type			ip_rt,
								day_count_method 	dc,
								logical				no_rnd,
								date option(nullable) max,
								out vector(date) option(nullable)	excl_mat)
	{				
		//vector(instrument) vi 	= fx_swaps_.instruments();
		//number size 			= v_size(vi);
		//vector(date) mat		= vi.maturity();
		//vector(date) settle 	= vi.settle_date();
	
		resize(excl_mat,0);
		vector(date) 	mat_fwd;
		vector(number) 	quote_fwd;
		disc_func df_calc = disc(df,df_is_price,ip,ip_rt, no_rnd,mat_fwd, quote_fwd);
		
		/*instr_def i_d = instr_def(depo_cn);
		if(null(i_d))
			throw(E_INVALID_ARG, "invalid deposit class name");	
	
		string class_curr = i_d.currency();
		if(class_curr != fx_spot_.base_ccy() && class_curr != fx_spot_.price_ccy())
			throw(E_INVALID_ARG, "invalid deposit class name (currency)");	
	
		day_count_method dc = i_d.accr_dc_method();*/
		date d 				= fx_spot_.trade_date();
		date spot_settle 	= fx_spot_.spot_settle();
		string synt_ccy 	= df_is_price ? fx_spot_.base_ccy(): fx_spot_.price_ccy();
		vector(instrument) vi_d;
	
		//before spot	
		vector(_fx)	fx_f_pr = f_pre(false);
		integer size_p_f 	= v_size(fx_f_pr);
		vector(_fx)	fx_s_pr = s_pre(false);
		integer size_p_s 	= v_size(fx_s_pr);
	
		instr_def i_d = create_depo_instr_def(dc, BD_MOD_FOLLOWING, true, "bd2", calendar(),synt_ccy);
	
		if(size_p_f > 0) {
			//the depos created here should be mkt maturitities eg. if we have [fwdt=t] and a [fwdt=t+1] and we have spot=t+2
			//--> the depos created should be an O/N and a T/N
			// for this to work we need to look at the fwds in reverse order		
			//first fwd used is the last one i.e. the one just before spot 
			date m = spot_settle;
			date s = fx_f_pr[size_p_f-1].maturity();
			number rate = df_calc.simple_rate(d, s, m, dc) ;
			//instrument depo = create_depo(depo_cn,fx_f_pre_[size_p_f-1].inst().name(),d,s, m, rate, s);
			instrument depo = deposit(i_d,fx_f_pr[size_p_f-1].inst().name(),d,s, s, m, rate, rate);
			//instrument depo = create_depo_instr(i_d,fx_f_pr[size_p_f-1].inst().name(),d,s, s, m, rate, rate);
	
			if(!null(depo))
				push_back(vi_d, depo);
			else
				push_back(excl_mat, s);
			for(integer i = size_p_f-2; i >=0 ; i--){//we loop backwards
				m = fx_f_pr[i+1].maturity();
				s = fx_f_pr[i].maturity();
				rate = df_calc.simple_rate(d, s, m, dc) ;
				
				instrument depo = deposit(i_d, fx_f_pr[i].inst().name(),d,s, s, m, rate, rate);
				//instrument depo = create_depo_instr(i_d, fx_f_pr[i].inst().name(),d,s, s, m, rate, rate);
				//instrument depo = create_depo(depo_cn,fx_f_pre_[i].inst().name(),d,s, m, rate, s);
				
				if(!null(depo))
					push_back(vi_d, depo);
				else
					push_back(excl_mat, s);
			}		
		}
		else if(size_p_s > 0) {
			for(integer i = 0; i < size_p_s; i++){
				date m = fx_s_pr[i].maturity();
				date s = fx_s_pr[i].settle_date();
				number rate = df_calc.simple_rate(d, s, m, dc) ;
	
				//instrument depo = create_depo_instr(i_d,fx_s_pr[i].inst().name(),d,s, s, m, rate, rate);
				instrument depo = deposit(i_d,fx_s_pr[i].inst().name(),d,s, s, m, rate, rate);
				//instrument depo = create_depo(depo_cn,fx_s_pre_[i].inst().name(),d,s, m, rate, s);
				if(!null(depo))
					push_back(vi_d, depo);
				else
					push_back(excl_mat, m);
			}
		}
	
		//after spot
		vector(_fx)	fx = post(false);
		for(integer i = 0; i < v_size(fx); i++){
			number rate;
			try {
				rate = df_calc.simple_rate(d, spot_settle, fx[i].maturity(), dc) ;
			}
			catch {
				rate = null<number>;
			}
			if(!null(max) && fx[i].maturity() > max && null(rate))
				continue;
	
			//instrument depo = create_depo_instr(i_d,fx[i].inst().name(),d,spot_settle, spot_settle, fx[i].maturity(), rate, rate);
			instrument depo = deposit(i_d,fx[i].inst().name(),d,spot_settle, spot_settle, fx[i].maturity(), rate, rate);
	
			//instrument depo = create_depo(depo_cn,fx_post_[i].inst().name(),d,spot_settle, fx_post_[i].maturity(), rate, spot_settle);
			if(!null(depo))
				push_back(vi_d, depo);
			else
				push_back(excl_mat, fx[i].maturity());
		}
		
		return curve(vi_d);	
	}
	
	curve fx_curve.synt_depo_base(	/*instr_class_name  	depo_cn_base,*/
									disc_func 			df_price,
									interpolator 		ip,
									rate_type			ip_rt,
									day_count_method 	dc,
									out vector(date) option(nullable)	excl_mat,
									logical				no_rnd,
									date option(nullable) max)
	{	
		return synt_depo(df_price,true,ip,ip_rt,dc, no_rnd,max,excl_mat);					
	}
	
	curve fx_curve.synt_depo_price(	/*instr_class_name  	depo_cn_price,*/
									disc_func 			df_base,
									interpolator 		ip,
									rate_type			ip_rt,
									day_count_method 	dc,
									out vector(date) option(nullable)	excl_mat,
									logical				no_rnd,
									date option(nullable) max)
	{	
		return synt_depo(df_base,false,ip,ip_rt,dc,no_rnd,max,excl_mat);					
	}
}


module CORE_INT_COMPAT_FX
{
	/*-----------------------------------------------------------------------
	  Name :  	fx_curve_swaps
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			13-Mar-2014 	MN      Original code 
	  ----------------------------------------------------------------------*/
	void fx_curve_quotes_df(fx_curve			fx_c_bid,
							fx_curve			fx_c_ask,
							disc_func			df_base_bid,
							disc_func			df_base_ask,
							disc_func			df_price_bid,
							disc_func			df_price_ask,
							logical 			unit_quote,
							fx_inst_type option(nullable) type,
							logical 			no_rnd,
							logical 			incl_spot,
							out vector(date) option(nullable) 	mat,
							out vector(number) option(nullable) 	quote_bid,
							out vector(number) option(nullable) 	quote_ask)
	
	{
		fx_spot fx_s_bid = fx_c_bid.fx_spot_inst();
		fx_spot fx_s_ask = fx_c_ask.fx_spot_inst();
		
		fx_c_bid.quotes_df_pre_spot(df_base_ask,df_price_bid,unit_quote,type,no_rnd,false,mat,quote_bid);
		fx_c_ask.quotes_df_pre_spot(df_base_bid,df_price_ask,unit_quote,type,no_rnd,false,mat,quote_ask);
		
		if(incl_spot) {		
			push_back(quote_bid, fx_s_bid.spot_quote(unit_quote, no_rnd));	
			push_back(quote_ask, fx_s_ask.spot_quote(unit_quote, no_rnd));
			push_back(mat, fx_s_bid.spot_settle());
		}
		
		vector(date) 	mat_post;
		vector(number) 	quote_bid_post,quote_ask_post;
	
		fx_c_bid.quotes_df_post_spot(df_base_ask,df_price_bid,unit_quote,type,no_rnd,false,mat_post,quote_bid_post);
		fx_c_ask.quotes_df_post_spot(df_base_bid,df_price_ask,unit_quote,type,no_rnd,false,mat_post,quote_ask_post);
		
		mat			= concat(mat, mat_post);
		quote_bid 	= concat(quote_bid, quote_bid_post);
		quote_ask 	= concat(quote_ask, quote_ask_post);	
	}
	
	/*-----------------------------------------------------------------------
	  Name :  	fx_curve_swaps
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			13-Mar-2014 	MN      Original code 
	  ----------------------------------------------------------------------*/
	void fx_curve_swaps(fx_curve			fx_c_bid,
						fx_curve			fx_c_ask,
						logical 			unit_quote,
						fx_inst_type option(nullable) type,
						logical 			no_rnd,
						logical 			incl_spot,
						out vector(date) option(nullable) 	mat,
						out vector(number) option(nullable) 	quote_bid,
						out vector(number)  option(nullable)	quote_ask)
	
	{
		fx_spot fx_s_bid = fx_c_bid.fx_spot_inst();
		fx_spot fx_s_ask = fx_c_ask.fx_spot_inst();
	
		vector(_fx) fx_f_pre_bid = fx_c_bid.fx_inst_pre_spot(FXI_FWD,false,true);
		vector(_fx) fx_f_pre_ask = fx_c_ask.fx_inst_pre_spot(FXI_FWD,false,true);
		
		fx_c_bid.swaps_pre_spot(fx_s_bid,fx_s_ask,fx_f_pre_bid,fx_f_pre_ask,unit_quote,type,no_rnd,true,false,mat,quote_bid);
		fx_c_ask.swaps_pre_spot(fx_s_bid,fx_s_ask,fx_f_pre_bid,fx_f_pre_ask,unit_quote,type,no_rnd,false,false,mat,quote_ask);
	
		if(incl_spot) {		
			push_back(quote_bid, fx_s_bid.spot_quote(unit_quote, no_rnd));	
			push_back(quote_ask, fx_s_ask.spot_quote(unit_quote, no_rnd));
			push_back(mat, fx_s_bid.spot_settle());
		}
		
		vector(date) 	mat_post;
		vector(number) 	quote_bid_post,quote_ask_post;
		fx_c_bid.swaps_post_spot(unit_quote,type,no_rnd,false,mat_post,quote_bid_post);
		fx_c_ask.swaps_post_spot(unit_quote,type,no_rnd,false,mat_post,quote_ask_post);
	
		mat			= concat(mat, mat_post);
		quote_bid 	= concat(quote_bid, quote_bid_post);
		quote_ask 	= concat(quote_ask, quote_ask_post);	
	}
	
	/*-----------------------------------------------------------------------
	  Name :  	fx_curve_outrights
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			13-Mar-2014 	MN      Original code 
	  ----------------------------------------------------------------------*/
	void fx_curve_outrights(fx_curve			fx_c_bid,
							fx_curve			fx_c_ask,
							logical 			unit_quote,
							fx_inst_type option(nullable) type,
							logical 			no_rnd,
							logical 			incl_spot,
							out vector(date) option(nullable) 	mat_fwd,
							out vector(number) option(nullable) 	quote_fwd_bid,
							out vector(number) option(nullable) 	quote_fwd_ask)
	
	{
		fx_spot fx_s_bid = fx_c_bid.fx_spot_inst();
		fx_spot fx_s_ask = fx_c_ask.fx_spot_inst();
	
		vector(_fx) fx_s_pre_bid = fx_c_bid.fx_inst_pre_spot(FXI_SWAP,false,true);
		vector(_fx) fx_s_pre_ask = fx_c_ask.fx_inst_pre_spot(FXI_SWAP,false,true);
		
		fx_c_bid.outrights_pre_spot(fx_s_bid,fx_s_ask,fx_s_pre_bid,fx_s_pre_ask,unit_quote,type,no_rnd,true,false,mat_fwd,quote_fwd_bid);
		fx_c_ask.outrights_pre_spot(fx_s_bid,fx_s_ask,fx_s_pre_bid,fx_s_pre_ask,unit_quote,type,no_rnd,false,false,mat_fwd,quote_fwd_ask);
	
		if(incl_spot) {		
			push_back(quote_fwd_bid, fx_s_bid.spot_quote(unit_quote, no_rnd));	
			push_back(quote_fwd_ask, fx_s_ask.spot_quote(unit_quote, no_rnd));
			push_back(mat_fwd, fx_s_bid.spot_settle());
		}
		
		vector(date) 	mat_fwd_post;
		vector(number) 	quote_fwd_bid_post,quote_fwd_ask_post;
		fx_c_bid.outrights_post_spot(unit_quote,type,no_rnd,false,mat_fwd_post,quote_fwd_bid_post);
		fx_c_ask.outrights_post_spot(unit_quote,type,no_rnd,false,mat_fwd_post,quote_fwd_ask_post);
	
		mat_fwd 		= concat(mat_fwd, mat_fwd_post);
		quote_fwd_bid 	= concat(quote_fwd_bid, quote_fwd_bid_post);
		quote_fwd_ask 	= concat(quote_fwd_ask, quote_fwd_ask_post);	
	}
	
	/*-----------------------------------------------------------------------
	  Name :  	fx_curve_disc_base
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			17-Mar-2014 	MN      Original code 
	  ----------------------------------------------------------------------*/
	void fx_curve_disc_base(fx_curve		fx_c_bid,
							fx_curve		fx_c_ask,						
							disc_func		df_price_bid,
							disc_func		df_price_ask,
							interpolator 	ip,
							rate_type		ip_rt,
							logical			no_rnd,
							out disc_func option(nullable) 	df_bid,
							out disc_func option(nullable) 	df_ask)
	
	{		
		df_ask = fx_c_bid.disc_base(df_price_bid,ip, ip_rt,no_rnd);
		df_bid = fx_c_ask.disc_base(df_price_ask,ip, ip_rt,no_rnd);		
	}
	/*-----------------------------------------------------------------------
	  Name :  	fx_curve_disc_base
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			17-Mar-2014 	MN      Original code 
	  ----------------------------------------------------------------------*/
	disc_func fx_curve_disc_base(fx_curve	fx_c,
							disc_func		df_price,
							interpolator 	ip,
							rate_type		ip_rt,
							logical			no_rnd)
	
	{		
		return fx_c.disc_base(df_price,ip, ip_rt,no_rnd);	
	}
	/*-----------------------------------------------------------------------
	  Name :  	fx_curve_disc_base
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			17-Mar-2014 	MN      Original code 
	  ----------------------------------------------------------------------*/
	void fx_curve_disc_price(fx_curve		fx_c_bid,
							fx_curve		fx_c_ask,						
							disc_func		df_base_bid,
							disc_func		df_base_ask,
							interpolator 	ip,
							rate_type		ip_rt,
							logical			no_rnd,
							out disc_func option(nullable) 	df_bid,
							out disc_func option(nullable) 	df_ask)
	
	{		
		df_ask = fx_c_ask.disc_price(df_base_bid,ip, ip_rt,no_rnd);
		df_bid = fx_c_bid.disc_price(df_base_ask,ip, ip_rt,no_rnd);		
	}
	/*-----------------------------------------------------------------------
	  Name :  	fx_curve_disc_base
	  Notes :         
	  Changes:
	  			Date    		Who     Detail
	  			17-Mar-2014 	MN      Original code 
	  ----------------------------------------------------------------------*/
	disc_func fx_curve_disc_price(fx_curve	fx_c,						
							disc_func		df_base,
							interpolator 	ip,
							rate_type		ip_rt,
							logical			no_rnd)
	
	{		
		return fx_c.disc_price(df_base,ip, ip_rt,no_rnd);	
	}
}

