//************SMITH-WILSON LIAB CURVE ****
/*
1)
2) add max_iter i orig
3) alfa_end?
4) break of repricing
*/

module CORE_INT 
{	
	number wilson_func(	number ufrc, 
						number alfa, 
						number t, 
						number u)
	{
		number e = exp(-ufrc*(t+u));	
		return e*(alfa*min(t, u) - 0.5*( exp(alfa*(min(t,u)-max(t,u)))  - exp(-alfa*(t+u)) ) );	//RT update 20220504
	}

	number sm_price(vector(number) 	df, 
					vector(number) 	df_t, 
					number 			t, 
					number 			alfa, 
					number 			ufr)
	{		
		number ufrc 		= log(1+ufr);
		vector(number) my 	= exp(-ufrc*df_t);
		integer n 			= v_size(df_t);
			
		matrix(number) w[n,n];
		
		for (integer i = 0; i<n; i++) {
			for (integer j = 0; j<n; j++)
				w[i,j] = wilson_func(ufrc, alfa, df_t[i], df_t[j]);
			
		}
		
		matrix(number) w_inv 	= invert(w);			
		vector(number) g 		= w_inv*(df-my);		
		return exp(-ufrc*t) + g*wilson_func(ufrc, alfa, t, df_t);	
	}
				
	// sm_price function split into two parts for optimization purposes.			
	vector(number) sm_price_part1(	vector(number) 	df, 
									vector(number) 	df_t, 
									number 			alfa, 
									number 			ufr)
	{
		
		number ufrc 		= log(1+ufr);
		vector(number) my 	= exp(-ufrc*df_t);
		integer n 			= v_size(df_t);
			
		matrix(number) w[n,n];
		
		vector(number) v0 	= alfa * df_t;
		vector(number) v1 	= exp(v0);
		
		number mini, maxi, mini0;	
				
		if(alfa >= 0)
		{
			//v0 = alfa * df_t;
			//v1 = exp(v0);
			
			for (integer i = 0; i<n; i++)
			{
				for (integer j = 0; j<n; j++)
				{
					mini0 	= min(v0[i], v0[j]);					
					mini  	= min(v1[i], v1[j]);
					maxi  	= max(v1[i], v1[j]);		
					w[i,j] 	= my[i] * my[j] * (mini0 - 0.5 * ( mini - 1/mini) / maxi); //RT update 20220504
				}
			}				
		}
		else
		{
			//v0 = alfa * df_t;
			//v1 = exp(v0);
			
			for (integer i = 0; i<n; i++)
			{
				for (integer j = 0; j<n; j++)
				{
					mini0 	= max(v0[i], v0[j]);					
					mini  	= max(v1[i], v1[j]);
					maxi  	= min(v1[i], v1[j]);		
					w[i,j] 	= my[i] * my[j] * (mini0 - 0.5 * ( mini - 1/mini) / maxi); //RT update 20220504
				}
			}	
		}
			
		matrix(number) w_inv 	= invert(w);	
		vector(number) g 		= w_inv*(df-my);			
		return g;	
	}	
			
	number sm_price_part2(	vector(number) 	g, 
							vector(number) 	df_t, 
							number 			t, 
							number 			alfa, 
							number 			ufr)
	{		
		number ufrc = log(1+ufr);
		return exp(-ufrc*t) + g*wilson_func(ufrc, alfa, t, df_t);
	}
}


class disc_func_sw : public disc_func
    //option(category: "Yield Curve/Fitting")
{
public:
	

	virtual number	disc(number);
	disc_func_sw(vector(number),vector(number) ,number,number, out logical option(nullable) );
protected:
	
	number 			ufr_;
	number 			alfa_;
	vector(number) 	df_t_;
	vector(number) 	g_;	
	
};

disc_func_sw.disc_func_sw(	vector(number) 	df, 
							vector(number) 	df_t, 
							number 			alfa, 
							number 			ufr,
							out logical option(nullable) is_null): ufr_(ufr), alfa_(alfa), df_t_(df_t)
{
	
	try {				
		g_ 		= CORE_INT.sm_price_part1(df,df_t,alfa,ufr);
		is_null = null(g_);
		return;
	}
	catch {
		g_ 		= null<vector(number)>;
		is_null = true;
	}
					
}

number disc_func_sw.disc(number t)
{
	return CORE_INT.sm_price_part2(g_, df_t_,t,alfa_,ufr_);
}

disc_func_sw disc_func_sw(	vector(number) 	df, 
							vector(number) 	df_t, 
							number 			alfa, 
							number 			ufr)
{
	logical is_null;
	disc_func_sw d = new disc_func_sw(df,df_t,alfa,ufr, is_null);
	return is_null? null<disc_func_sw>: d;	
}


module CORE_INT {

	class forward_rate_sw
	{
	public:
		forward_rate_sw(vector(number),vector(number),number,number,number);
		number f(number alfa);

		vector(number) 	df_;
		vector(number) 	df_t_;
		number 			T_;
		number 			ufr_;
		number 			tol_;		
		//vector(number) 	tmp_;					
	};
	forward_rate_sw.forward_rate_sw(vector(number) 	df,
									vector(number) 	df_t,
									number 			T,
									number 			ufr,
									number 			tol)
									: df_(df), df_t_(df_t), T_(T), ufr_(ufr), tol_(tol)
	{}

	number forward_rate_sw.f(number alfa)
	{
		number dt 		= 1;			
		vector(number) tmp_ = sm_price_part1(df_,df_t_,alfa, ufr_);		
		number d_t  	= sm_price_part2(tmp_, df_t_, T_, alfa, ufr_);
		number d_t2 	= sm_price_part2(tmp_, df_t_, T_+dt, alfa, ufr_);		
		number sm_fwd 	= (d_t/d_t2 - 1)/dt;
		
		return sm_fwd - (ufr_-tol_);  // Wants to find sm_fwd which is equal to ufr-tol	
	}
	
	number alfa_orig(	vector(number) 	df, 
						vector(number) 	df_t, 
						number 			T2, 
						number 			ufr, 
						number 			ufr_tol, 
						number 			alfa_start, 
						number 			alfa_step)
	{					
		forward_rate_sw fw_obj 	= new forward_rate_sw(df,df_t,T2,ufr,0);		
		number alfa0 			= alfa_start;
	
		//add MAX_ITER
		while (abs(fw_obj.f(alfa0)) > ufr_tol) {
			alfa0 += alfa_step;			
		}		
	
		return alfa0;
	
	}
		
	number alfa_brent(	vector(number) 	df, 
						vector(number) 	df_t, 
						number 			T2, 
						number 			ufr, 
						number 			ufr_tol, 
						number 			alfa_start, 
						number 			alfa_end)
	{
					
		forward_rate_sw fw_obj = new forward_rate_sw(df,df_t,T2,ufr,ufr_tol);	
		//fw_obj.tmp_ 	= one_vector(v_size(df_t), 0); 	
		number br_tol 	= ufr_tol/100.0;			
		number x0 		= alfa_start;
		number x1 		= alfa_end;		
		number f0 		= fw_obj.f(x0); 
		number f1 		= fw_obj.f(x1);
	
		number res;		
		number iter 	= 0;
		number MAX_ITER = 10;
		
		if(f0 * f1 < 0)
			res = zero_brent(fw_obj, &forward_rate_sw.f, x0, x1, br_tol);
		else
		{
			while(f0 * f1 >= 0 && iter < MAX_ITER)
			{
				x0 *= 0.5;
				x1 *= 1.5;	
				
				f0 = fw_obj.f(x0); 
				f1 = fw_obj.f(x1);
				
				iter++;
			}
			
			if(iter == MAX_ITER || f0 * f1 >= 0 )
				throw(E_CALC, "brent solver: no solution found");
			else
				res = zero_brent(fw_obj, &forward_rate_sw.f, x0, x1, br_tol);
		
		}
		
		return res;
	}

	number alfa_opt(vector(number) 	df, 
					vector(number) 	df_t, 
					number 			T2, 
					number 			ufr, 
					number 			ufr_tol, 
					number 			alfa_start, 
					number 			alfa_step,
					number 			alfa_end)
	{				
		try {
			return alfa_brent(df, df_t, T2,ufr,ufr_tol,alfa_start,alfa_end);
		}
		catch{	//Default to step method if fail
			return alfa_orig(df,df_t,T2,ufr,ufr_tol,alfa_start,alfa_step);
		}
	}
	/*
		knot = [1, 5, 10, 150] //yrs
		m = [0.25, 1, 1, 12]
	*/
	/*vector(number) sw_points(vector(number) knot, vector(number) months)
	{
		integer size = v_size(knot);
		if(size == 0 || size  != v_size(months) )
			throw(E_INVALID_ARG, "Invalid vector input") ;
		if(knot[0] <= 0 )
			throw(E_INVALID_ARG, "Invalid vector input (first element must not be zero") ;

		vector(number) t_v;
		number t;
		for(integer i=0;i<size;i++){
			if(i>0 && knot[i] <= knot[i-1])
 				throw(E_INVALID_ARG, "Invalid vector input (knot)") ;
			if(months[i] <= 0)
 				throw(E_INVALID_ARG, "Invalid vector input (months)") ;

			t = i==0 ? 0 : knot[i-1];					
			t += months[i]/12.0;
			while (t <= knot[i]) {
				push_back(t_v,t);
				if(t == knot[i]) break; 
				t  += months[i]/12.0;
			}			 		
		}

		return t_v;		
	}*/	
}


disc_func smith_wilson(	curve 		c, 
						number 		last_liq_t,//yrs 
						number 		ufr_t, //yrs
						number 		ufr, 
						interpolator interp, 
						number 		ufr_tol , 
						number 		alfa_start , 
						string 		solver , 
						number 		alfa_step , 
						out number alfa_star)
option(category : "Yield Curve/Solvency Liability Curve")
{	
	if(c.empty())
		throw(E_INVALID_ARG, "Invalid curve") ;

	vector(number) df_t = (c.instruments().maturity_adj() - c.trade_date())/365;
	df_t 				= sub_vector(df_t,last_liq_t+0.1);
	vector(number) df   = bootstrap(c).disc(df_t);//breaks repricing because not all cashflows in the vector

	if(equal_casei(solver, "step"))
		alfa_star = CORE_INT.alfa_orig(df,df_t,ufr_t, ufr, ufr_tol, alfa_start, alfa_step);
	else if (equal_casei(solver, "brent")){
		number alfa_end = 0.4;
		alfa_star = CORE_INT.alfa_opt(df,df_t,ufr_t, ufr, ufr_tol, alfa_start, alfa_step, alfa_end);
	}
	else if (equal_casei(solver, "brent_no_fallback")){
		number alfa_end = 0.4;		
		alfa_star = CORE_INT.alfa_brent(df, df_t, ufr_t,ufr,ufr_tol,alfa_start,alfa_end);
	}
	else if (equal_casei(solver, "fix_alfa") || equal_casei(solver, "fix_alpha"))
		alfa_star = alfa_start; //Use input alfa without solver
	else
		throw(E_INVALID_ARG, "Invalid solver! Use 'STEP', 'BRENT', or 'FIX_ALPHA'") ;
	

	return disc_func_sw(df,df_t, alfa_star, ufr);

	/*vector(number) i 		= (index_vector(150*12)+1)/12;
	vector(number) df_sm 	= CORE_INT.sm_price(df,df_t,i,alfa_star,ufr);

	disc_func disc_f = disc_func_interp(i,df_sm,interp, RT_EFFECTIVE);
	logical no_extrap = false;
	number extrap_t = 150;
	if(no_extrap) 
		return CORE_INT.no_extrap(disc_f, extrap_t);	
	else
		return disc_f;*/
	
}

disc_func smith_wilson(	vector(number) 	df_t,
						vector(number) 	df, 
						number 			ufr_t, 
						number 			ufr,
						interpolator 	interp, 
						number 			ufr_tol, 
						number 			alfa_start, 
						string 			solver, 
						number 			alfa_step )
{
	if(v_size(df_t) == 0 || v_size(df_t) != v_size(df) )
		throw(E_INVALID_ARG, "Invalid vector input") ;

	//optimize for alfa parameter using chosen solver
	number alfa_star;
	
	if(equal_casei(solver, "step"))
		alfa_star = CORE_INT.alfa_orig(df,df_t,ufr_t, ufr, ufr_tol, alfa_start, alfa_step);	
	else if (equal_casei(solver, "brent"))	{
		number alfa_end = 0.4;	
		alfa_star = CORE_INT.alfa_opt(df,df_t,ufr_t, ufr, ufr_tol, alfa_start, alfa_step, alfa_end);	
	}
	else if (equal_casei(solver, "brent_no_fallback")){
		number alfa_end = 0.4;
		alfa_star = CORE_INT.alfa_brent(df,df_t, ufr_t,ufr,ufr_tol,alfa_start,alfa_end);
	}
	else if (equal_casei(solver, "fix_alfa") || equal_casei(solver, "fix_alpha"))
		alfa_star = alfa_start; //Use input alfa without solver	
	else	
		throw(E_INVALID_ARG, "Invalid solver! Use 'STEP', 'BRENT', or 'FIX_ALPHA'") ;
	
	return disc_func_sw(df,df_t, alfa_star, ufr);
	
	/*vector(number) i 		= (index_vector(150*12)+1)/12;		
	vector(number) g     	= CORE_INT.sm_price_part1(df, periods, alfa_star, ufr);
	vector(number) df_sm 	= CORE_INT.sm_price_part2(g, periods, i, alfa_star, ufr);
	
	disc_func disc_f = disc_func_interp(i,df_sm,interp, RT_EFFECTIVE);

	logical no_extrap = false;
	number extrap_t = 150;
	if(no_extrap) 
		return CORE_INT.no_extrap(disc_f, extrap_t);	
	else
		return disc_f;	*/
}


//************SWEDEN LIAB CURVE **********

module CORE_INT {

	class p_swe_liab 
	{
	public:
		p_swe_liab(number ,number ,number ,integer ,integer);
		number f(number x);

		number par_;	//the i:th par rate
		number df_sum_; //running sum of all known dfs
		number df_last_;//last known df
		integer t_;		// the time to maturity for the i:th par rate
		integer empty_p_;		
	};

	p_swe_liab.p_swe_liab(	number par,
							number df_sum,
							number df_last,
							integer t,
							integer empty_p)
							:par_(par),df_sum_(df_sum), df_last_(df_last),t_(t),empty_p_(empty_p)
	{}

	number p_swe_liab.f(number x)
	{
		number tmp = 0;
		for(integer k = 1; k < empty_p_;k++)
			tmp += (1/((df_last_/x)^(1/empty_p_))^k);
			
		return x - (1-par_*(df_sum_ + df_last_*tmp))/(1 + par_);
	}
	
	vector(number) solve_z_swe_liab(vector(number) par, vector(integer) t)
	{
	
		integer n1 = v_size(par);
		integer n2 = v_size(t);
		if(n1!=n2 || t[0] != 1)
			throw(E_INVALID_ARG, "Not the same size of par rates and times, or first time is not t = 1!");

		vector(number) df[t[n2-1]]; //make df:s for all whole year maturities up and including the last maturity in t
		df[0] 			= 1.0/(1.0+par[0]);
		number df_sum 	= df[0];
		number x1 		= 0.99;
		number x2 		= 0.95;
		number tol 		= 0.000001;
		integer off 	= 0;

		for(integer i = 1; i< n2; i++){
			
			if(t[i] == t[i-1]+1){
				df[i+off ] 	= (1-par[i]*df_sum)/(1 + par[i]);
				df_sum 		+= df[i+off ];
			}
			else {
				integer empty_p 	= t[i] - t[i-1];
				p_swe_liab p 		= new p_swe_liab(par[i], df_sum, df[i+off -1], t[i], empty_p);				
				number df_next 		= zero_secant(p, &p_swe_liab.f,x1,x2,tol);

				for(integer k = 1; k < empty_p; k++){
					df[i+k-1+off ] = df[i-1+off ]/(((df[i-1+off ]/df_next)^(1/empty_p))^k);
					df_sum 			+= df[i+k-1+off ];			
				}

				df[i+empty_p-1+off ] 	= df_next;
				df_sum 					+= df_next;
				off 					+= empty_p-1;
			}
		}
		return df;
	}

	number w_exp_swe_liab(	number t,
							number T1, 
							number T2)
	{
		if(t<0 || T1 > T2)
			throw(E_INVALID_ARG, "Invalid input. t<0 or T2<T1");
		
		if(t <= T1)
			return 0;
		if(t > T1 && t <= T2)
			return (t-T1)/(T2-T1+1);
		// t > T2
		return 1; 	
	}

	number fwd_t_interpol_swe_liab(	integer t,
									number 	df_t, 
									number 	df_t_1, 
									integer T1, 
									integer T2, 
									number 	ufr)
	{
		number w = w_exp_swe_liab(t,T1,T2);
		number f = df_t_1/df_t - 1;
		return (1-w)*f + w*ufr;
	}

	number z_t_calc_swe_liab(	integer t,
								number df_t, 
								number df_t_1,
								number df_updated_1, 
								integer T1, 
								integer T2, 
								number ufr)
	{
		number f = fwd_t_interpol_swe_liab(t,df_t,df_t_1,T1,T2,ufr);
		return ((1+f)/df_updated_1)^(1/t) - 1;
	}

	number z_t_expol_swe_liab(	integer t,
								number df_updated_1, 
								number ufr)
	{		
		return ((1+ufr)/df_updated_1)^(1/t) - 1;
	}
}


disc_func swe_solvency_liab_curve(	vector(integer) periods, 
									vector(number) 	par_rate, 
									integer 		last_liq_t,//yrs 
									integer 		ufr_t, //yrs
									number 			ufr, 
									interpolator 	interp, 
									rate_type 		rt = RT_EFFECTIVE)
 option(category : "Yield Curve/Solvency Liability Curve")
{
	integer n1 = v_size(par_rate);
	integer n2 = v_size(periods);
	if(n1!=n2 || periods[0] != 1)
		throw(E_INVALID_ARG, "Not the same size of par rates and periods, or first time is not 1 year!");

	integer max_t 			= periods[n2-1];
	//vector(integer) t_short = index_vector(max_t)+1;
	vector(integer) t_tot 	= index_vector(150) + 1;
	vector(number) df 		= CORE_INT.solve_z_swe_liab(par_rate, periods);	
	integer T1 				= last_liq_t;
	integer T2 				= ufr_t;
	vector(number) df_new[v_size(t_tot)];
	df_new[0] 				= df[0];			

	for(integer i = 1; i<t_tot[v_size(t_tot)-1];i++){

		integer t = i + 1; //we solve to next t but for i:th pos in df and z.
		number z_t = 0;
		if(t<=T2 && t<= max_t) //use original df:s for current fwd but updated df for previous z(t-1)
			z_t = CORE_INT.z_t_calc_swe_liab(t,df[i],df[i-1],df_new[i-1],T1,T2,ufr);		
		else if(t<=T2 && t > max_t) //end of market df:s use last known forward as market quote
			z_t = CORE_INT.z_t_calc_swe_liab(t,df[max_t - 1],df[max_t - 2],df_new[i-1],T1,T2,ufr);		
		else	// use only ufr and trailing df:s
			z_t = CORE_INT.z_t_expol_swe_liab(t, df_new[i-1],ufr);
		
		number df_t = 1/(1+z_t)^t;
		df_new[i] 	= df_t;
	}

	disc_func df_ret = disc_func_interp(t_tot, df_new, interp, rt);
	
	return df_ret;
}

disc_func swe_solvency_liab_curve(	curve 		c, //swaps only?
									number 		credit_adj, 
									integer 	last_liq_t, //yrs
									integer 	ufr_t, 		//yrs
									number 		ufr, 
									interpolator interp, 
									rate_type 	rt = RT_EFFECTIVE)
 option(category : "Yield Curve/Solvency Liability Curve")
{
	//first cut the curve short so that only swaps up and to the last liq point are included
	date trade_date 		= c.trade_date();
	c 						= c.chop_long(add_years(trade_date+20,ufr_t));
	vector(number) par_rate = c.instruments().yield(); //i.e. swap par rates
	vector(integer) periods = integer(round((c.instruments().maturity() - trade_date)/365,0));

	// adjust for credit spread
	par_rate = max(par_rate - credit_adj,0);

	integer n1 = v_size(par_rate);
	integer n2 = v_size(periods);
	if(n1!=n2 || periods[0] != 1)
		throw(E_INVALID_ARG, "Not the same size of par rates and periods, or first time is not 1 year!");

	integer max_t 			= periods[n2-1];
	vector(integer) t_short = index_vector(max_t)+1;
	vector(integer) t_tot 	= index_vector(150) + 1;
	vector(number) df 		= CORE_INT.solve_z_swe_liab(par_rate, periods);
	integer T1 				= last_liq_t;
	integer T2 				= ufr_t;
	vector(number) df_new[v_size(t_tot)];
	df_new[0] 				= df[0];
	
	for(integer i = 1; i<t_tot[v_size(t_tot)-1];i++){
		integer t 	= i + 1; //we solve to next t but for i:th pos in df and z.
		number z_t 	= 0;
		if(t<=T2 && t<= max_t) //use original df:s for current fwd but updated df for previous z(t-1)
			z_t = CORE_INT.z_t_calc_swe_liab(t,df[i],df[i-1],df_new[i-1],T1,T2,ufr);		
		else if(t<=T2 && t > max_t) //end of market df:s use last known forward as market quote
			z_t = CORE_INT.z_t_calc_swe_liab(t,df[max_t - 1],df[max_t - 2],df_new[i-1],T1,T2,ufr);		
		else	// use only ufr and trailing df:s
			z_t = CORE_INT.z_t_expol_swe_liab(t, df_new[i-1],ufr);
		
		number df_t = 1/(1+z_t)^t;
		df_new[i] 	= df_t;
	}

	disc_func df_ret = disc_func_interp(t_tot, df_new, interp, rt);
	
	return df_ret;
}


