option(null: hard);	

class enum knot_scheme option (category: 'Yield Curve/005 Fitting') 
{
	SS_SQR_T option(str: "shortstd"),
	SS_EQ_TIME option(str: "eqtime"),
	/*,SS_EQ_FLOW*/
	SS_EQ_MAT option(str: "eqmat")
	/*,SS_MCCU*/
};

vector(string) knot_scheme_list() option (category: 'Yield Curve/005 Fitting')
{
	return [string(knot_scheme.SS_SQR_T),string(knot_scheme.SS_EQ_TIME), string(knot_scheme.SS_EQ_MAT)];
}

knot_scheme knot_scheme_map(string ks)
option (category: 'Yield Curve/005 Fitting')
{
	switch(str_to_lower(ks)){
	case "shortstd":
		return knot_scheme.SS_SQR_T;
    case "eqtime":
		return knot_scheme.SS_EQ_TIME;
    case "eqmat":
		return knot_scheme.SS_EQ_MAT;
	/*case "eqflow":
		return SS_EQ_FLOW;*/
	/*case "mccull":
		return SS_MCCU;*/
    default:
		throw(E_INVALID_ARG,throw_err("invalid spline knot scheme"));
	}
}

knot_scheme get_scheme(string ks)
option (category: 'Obsolete Functions/Yield Curve/Fitting')
{
    return knot_scheme_map(ks);
}

module CORE_INT
{

	void do_eq_mat(out vector(number)	knots,
			      integer				n_knots,
			      vector(number)		t,
			      integer		 		n_t)
	{
	    if (n_t <= 0)
			throw(E_INVALID_ARG,"ouch");
	
	    vector(number)	n[n_t];
	    number	a = floor((n_knots - 2) / n_t);
	    number	b = floor((n_knots - 2) - n_t * a);
	
	    for (integer i = 0 ; i < n_t ; i++)
			n[i] = a;
	
	    if (b != 0) {
			number	f = n_t / (b + 1);	
			for (integer i = 1 ; i <= b ; i++)
				n[integer(floor(f * i))] += 1;
	    }
	
	    integer	k  = 0;
	    number	t0 = 0;
	
	    knots[k++] = 0;
	    for (integer i = 0 ; i < n_t ; i++) {
			number	dt = (t[i] - t0) / (n[i] + 1);
		
			for (integer j = 0 ; j < n[i] ; j++)
			    knots[k++] = t0 + (j + 1) * dt;
		
			t0 = t[i];
	    }
	    knots[k++] = t[n_t - 1];
	
	    if (k != n_knots)
			throw(E_INVALID_ARG,throw_err("ouch"));
	}
	
	
	
	vector(number) spline_knots(knot_scheme		scheme,						 
								integer			n_mat,
								vector(number)	mat,
								integer			n_knots)
	{
		if (n_knots <= 1)
			throw(E_INVALID_ARG,throw_err("ouch"));
	
	    number	max_t	= floor(mat[n_mat - 1] + 1.25);    
		vector(number) knots[n_knots];
	   
	    if(scheme == knot_scheme.SS_SQR_T){
		    number	T = max_t / sqr(n_knots - 1);
	
		    for (integer i = 0 ; i < n_knots ; i++)
				knots[i] = T * sqr(i);
		}
		else if(scheme == knot_scheme.SS_EQ_TIME){
		    number	T = max_t / (n_knots - 1);
	
		    for (integer i = 0 ; i < n_knots ; i++)
				knots[i] = T * i;
		}
		else if(scheme == knot_scheme.SS_EQ_MAT){
		    vector(number)	T[n_mat];
	
		    for (integer i = 0 ; i < n_mat ; i++)
				T[i] = mat[i];
	
			sort (T, T);
		  
		    do_eq_mat(knots, n_knots, T, n_mat);
		}
		return knots;
	}
	
	integer get_n_knots(number option(nullable) N,
	       				number		 			n_instr)
	{
	    if (null(N) || N == -1)
			return integer(floor(1 + sqrt(n_instr)));
	    else
			return integer(N);
	}
}

vector(number) spline_knots(curve 			c,
							date 			trade, 
							knot_scheme 	scheme,
							number option(nullable)	N)//if null or -1 use MCCULL
option (category: 'Yield Curve/005 Fitting')
{	    
	/*if(scheme == SS_EQ_FLOW){
		

	}
	else */
	vector(number) mat 	= (c.instruments().maturity() - trade) / 365.0;					
	integer n_mat		= v_size(mat);
	integer n_k 		= CORE_INT.get_n_knots(N, n_mat);
   
    if (n_k < 2)
		throw(E_INVALID_ARG,throw_err("not enough knots"));

    return CORE_INT.spline_knots(scheme, n_mat, mat, n_k);
}

