/*	
	Core zc functions
	Developer: Algorithmica Research, Magnus Nyström	

			
*/
/*
include freq
*/
option(null: hard);	

module CORE_INT
{

	class fwd_func_ext: public fwd_func
	    /*option(category: "Yield Curve")*/
	{
	public:
		fwd_func_ext(fwd_func option(fixedref),date, date);
	
		fwd_func 		ff();
	    override number	fwd(number);
		date			start();
		date			end();	
	
	private:
	
		fwd_func option(fixedref)	ff_;
		date						start_;
		date						end_;    
	};
	
	
	fwd_func_ext.fwd_func_ext(	fwd_func option(fixedref)	fwd,
				  				date						start,
								date						end)
							    : ff_(fwd),start_(start), end_(end)
	{}
	
	fwd_func fwd_func_ext.ff()			{ return ff_; }
	number 	fwd_func_ext.fwd(number t) 	{ return fwd_func.fwd(t); }
	date 	fwd_func_ext.start() 		{ return start_; }
	date 	fwd_func_ext.end() 			{ return end_; }
	
	fwd_func fwd_func_ext(	fwd_func	ff,
			       			date		start,
							date		end)
	{
		return new fwd_func_ext(ff, start, end);
	}
	
	
	class disc_func_ext: public disc_func
	    /*option(category: "Yield Curve")*/
	{
	public:
		disc_func_ext(disc_func option(fixedref),date option(nullable), date option(nullable));
	
		disc_func		df();
	    override number	disc(number);
		number			disc(date);
		date			start();
		date			end();	
	
	private:
	
		disc_func option(fixedref)	df_;
		date						start_;
		date						end_;    
	};
	
	
	disc_func_ext.disc_func_ext(disc_func option(fixedref)	df,
				  				date option(nullable)		start,
								date option(nullable)		end)
							 	: df_(df),start_(start), end_(end)
	{}
	
	disc_func 	disc_func_ext.df() 				{ return df_; }
	number 		disc_func_ext.disc(number t) 	{ return disc_func.disc(t); }
	number 		disc_func_ext.disc(date t) 		{ return null(start_) ? null<number> : disc_func.disc((t-start_)/365.0); }
	date 		disc_func_ext.start() 			{ return start_; }
	date 		disc_func_ext.end() 			{ return end_; }
	
	disc_func disc_func_ext(disc_func				df,
			       			date option(nullable)	start,
							date option(nullable)	end)
	{
		return new disc_func_ext(df, start, end);
	}
}

module CORE_INT
{
	//---defaults---
	std_weights def_w_scheme 		= std_weights.WS_PVBP; 
	number 		def_norm 			= 2;
	number 		def_spline_degree 	= 3;
	logical		def_mc_force_pos	= true;
	rate_type	def_boot_rate_type	= rate_type.RT_CONT;
	number		def_boot_left_der	= 0;
	number		def_boot_right_der	= null<number>;
	logical		def_boot_on_interp	= false;
	logical 	def_boot_apply_hyman= false; 
	logical 	def_boot_extrap		= false;
	smoothness_measure def_smooth	= smoothness_measure.SMOOTH_C2;
	number		def_fit_spread		= 1;
	number 		def_w_factor		= 0.1;
	vector(number) 	def_knots		= [0,1,5,10,20];
	vector(number) 	def_sigma_x		= [1,10,30];
	vector(number) 	def_sigma_y		= [0,0,0];	
	number 		def_tens_epsilon	= 0.3;
	number		def_tens_left_der	= 0;//null<number> = natural boundary condition
	number		def_tens_right_der	= 0;//null<number> = natural boundary condition
	number 		def_tens_lambda		= 0.1;

	number 		def_liq_t			= 10;
	number 		def_ufr_t			= 20;
	number 		def_ufr				= 0.039; 
	interpolator def_interp			= ip_linear(); 
	number 		def_ufr_tol			= 0.0001; 
	number 		def_a_start			= 0.05; 
	string 		def_solver			= "brent_no_fallback"; 
	number 		def_a_step			= 0.0001;
	//--------------

	vector(string) smith_w_ceiops_solver_list_ext()
	{
		return ["step","brent","brent_no_fallback","fix_alpha", "default"];
	}

	vector(string) smooth_m_list_ext()
	{
		return ["smooth_d1","smooth_d2","smooth_c1","smooth_c2", "default"];
	}

	smoothness_measure smooth_m_map_ext(string option(nullable) s) 
	{
		if(null(s)) return null<smoothness_measure>;

		switch(str_to_lower(s))
		{
		case "smooth_d1":		
			return SMOOTH_D1;
			break ;		
		case "smooth_d2":		
			return SMOOTH_D2;
			break ;	
		case "smooth_c1":		
			return SMOOTH_C1;
			break ;	
		case "smooth_c2":		
			return SMOOTH_C2;
			break ;
		case "default":
		default:
			return null<smoothness_measure>;			
		}
	}

	vector(string) knot_scheme_list_ext() 
	{ 
		return concat(knot_scheme_list(),["user", "default"]);
	}


	vector(string) norm_list_ext() 
	{
		return ["L1", "L2", "default"];
	}

	number norm_map_ext(string option(nullable) w) 
	{
		if(null(w)) return null<number>;

		switch(str_to_lower(w))
		{
		case "l1":		
			return 1;	
		case "l2":		
			return  2;
		case "default":
			return null<number>;
		default:
			return null<number>;	
		}
	}

	vector(string) weight_scheme_list_ext()
	{ 
		return ["equal", "pvbp","pvbp_squared", "default"];
	}

	std_weights weight_scheme_map_ext(string option(nullable) w) 
	{
		if(null(w))
			return null<std_weights>;

		switch(str_to_lower(w))
		{
		case "equal":		
			return WS_EQUAL;	
		case "pvbp":		
			return  WS_PVBP;
		case "pvbp_squared":		
			return  WS_PVBP2;
		case "default":
			return null<std_weights>;
		default:
			return null<std_weights>;	
		}
	}	

	vector(number) spread_v(curve c, vector(number) spread)
	{
		vector(instrument) v_i = c.instruments();
		integer size = v_size(v_i);
		integer size_s = v_size(spread);
		
		vector(number) res[size];
		if(size_s == 1) {
			for(integer i=0; i<size;i++) {
				number pvbp = v_i[i].pvbp(100,null<number>);
				res[i] = pvbp * spread[0];
			}
		}
		else {
			if(size != size_s)
				return null<vector(number)>;
			for(integer i=0; i<size;i++) {
				number pvbp = v_i[i].pvbp(100,null<number>);
				res[i] = pvbp * spread[i];
			}
		}
		return res;
	}

	class enum ts_model_type {
		TS_NONE  		option(str: "none"),
		TS_NS  			option(str: "ns_bf"),
		TS_NS_SVEN  	option(str: "ns_svensson_bf"),
		TS_NS_BLISS		option(str: "ns_bliss_bf"),
		TS_NS_LAG  		option(str: "ns_laguerre_bf"),
		TS_HERM_D  		option(str: "hermite_disc_bf"),
		TS_HERM_S  		option(str: "hermite_spot_bf"),
		TS_HERM_F  		option(str: "hermite_fwd_bf"),
		TS_BSPL_D  		option(str: "bspline_disc_bf"),
		TS_BSPL_S  		option(str: "bspline_spot_bf"),
		TS_BSPL_F  		option(str: "bspline_fwd_bf"),
		TS_CSPL_D  		option(str: "cspline_disc_bf"),
		TS_CSPL_S  		option(str: "cspline_spot_bf"),
		TS_CSPL_F  		option(str: "cspline_fwd_bf"),
		TS_TANG  		option(str: "tanggaard_bf"),
		TS_TENS_W_BF  	option(str: "tension_w_bf"),
		TS_TENS_Y_BF  	option(str: "tension_y_bf"),
		TS_TENS_Z_BF  	option(str: "tension_z_bf"),
		TS_LIN 			option(str: "linear_ef"),
		TS_SPL  		option(str: "spline_ef"),
		TS_SPL_NAT  	option(str: "spline_nat_ef"),
		TS_SPL_FIN  	option(str: "spline_fin_ef"),
		TS_STEP			option(str: "step_ef"),
		TS_HAGAN  		option(str: "hagan_west_ef"),
		TS_HERM_AK  	option(str: "hermite_akima_ef"),
		TS_HERM_CR  	option(str: "hermite_cr_ef"),
		TS_HERM_FD  	option(str: "hermite_findiff_ef"),
		TS_HERM_FB  	option(str: "hermite_fb_ef"),
		TS_HERM_K  		option(str: "hermite_kruger_ef"),
		TS_HERM_M  		option(str: "hermite_mono_ef"),
		TS_HERM_P  		option(str: "hermite_para_ef"),
		TS_MS  			option(str: "max_smooth_ef"),
		TS_MS_S  		option(str: "max_smooth_sprd_ef"),
		TS_TENS_W  		option(str: "tension_w_ef"),
		TS_TENS_Y  		option(str: "tension_y_ef"),
		TS_TENS_Z  		option(str: "tension_z_ef"),
		TS_TENS_W_S  	option(str: "tension_w_sprd_ef"),
		TS_TENS_Y_S  	option(str: "tension_y_sprd_ef"),
		TS_TENS_Z_S  	option(str: "tension_z_sprd_ef"),
		TS_SW_CEI  		option(str: "smith_wilson_ceiops_ef")
		/*models not included for now are: cir, cir2, gamtaux, gamtaux_ext, ls, vasicek, vasicek_fong, ccw, exp_spline, flat_rate*/

	};

	logical model_is_fit_hermite(ts_model_type mt)
	{
		return mt ==  ts_model_type.TS_HERM_D || mt ==  ts_model_type.TS_HERM_S || mt ==  ts_model_type.TS_HERM_F;
	}
	logical model_is_tension(ts_model_type mt)
	{
		return   mt ==  ts_model_type.TS_TENS_W_BF || mt ==  ts_model_type.TS_TENS_Y_BF || mt ==  ts_model_type.TS_TENS_Z_BF
				|| mt ==  ts_model_type.TS_TENS_W_S  || mt ==  ts_model_type.TS_TENS_Y_S || mt ==  ts_model_type.TS_TENS_Z_S
				|| mt ==  ts_model_type.TS_TENS_W  || mt ==  ts_model_type.TS_TENS_Y || mt ==  ts_model_type.TS_TENS_Z;
	}
	
	vector(string) ts_model_list()
	{
		vector(ts_model_type) e;
		expand_enum(e) ;
		return string(e);
	}

	
	

	ts_model_type ts_model_type_map(string mn) 
	{
		ts_model_type e;
		str_to_enum(mn,e);
		return e;	
	}

	class enum ts_spline_type
	{
		ST_DISC,
		ST_SPOT,
		ST_FWD
	};
			
	class enum fit_func_type
	{
		FT_NONE,
		FT_FIT,
		FT_BOOT,
		FT_SPEC
	};

	class fit_model_params
	{
	public:
		fit_model_params(integer,ts_model_type);
		fit_model_params(vector(number),vector(number),vector(number),vector(logical),vector(number),ts_model_type);		
		void	init_params(vector(number),vector(number),vector(number),vector(logical),vector(number));
		void	init_params(vector(number),vector(number),vector(number),vector(string),vector(number));
		integer	param_size();
		void set_fixed(vector(string));
		
	//protected:
		vector(number) 		guess_;
		vector(number)  	min_;
		vector(number)  	max_;
		vector(logical)  	fixed_;
		vector(number) 		fixed_val_;
		ts_model_type		model_;
	protected:
		void check();
	};

	fit_model_params.fit_model_params(integer size, ts_model_type m): model_(m)
	{
		QL_REQUIRE(size > 0,"invalid size");
		check();
		resize(guess_,size);			
		resize(min_,size);						
		resize(max_,size);						
		resize(fixed_,size);
		resize(fixed_val_,size);

		guess_ 	= null<number>;
		min_ 	= null<number>;	
		max_ 	= null<number>;
		fixed_ 	= false;
		fixed_val_= null<number>;
	}
	
	fit_model_params.fit_model_params(	vector(number) guess,vector(number) min,vector(number) max,
										vector(logical) fixed,vector(number) fixed_val, ts_model_type m)
									: guess_(clone_vector(guess)),min_(clone_vector(min)),
									  max_(..clone_vector(max)), fixed_(..clone_vector(fixed)), fixed_val_(..clone_vector(fixed_val)),model_(m)
	{
		check();
		integer size = param_size();
		QL_REQUIRE(size > 0,"invalid vector size");
		QL_REQUIRE(size == v_size(min_),"invalid vector size (min)");
		QL_REQUIRE(size == v_size(max_) ,"invalid vector size (max)");
		QL_REQUIRE(size == v_size(fixed_),"invalid vector size (fixed)");
		QL_REQUIRE(size == v_size(fixed_val_),"invalid vector size (fixed values)");
	}

	void fit_model_params.check()
	{
		logical is_ok = model_ ==  ts_model_type.TS_NS || model_ == ts_model_type.TS_NS_SVEN
						|| model_ == ts_model_type.TS_NS_LAG || model_ == ts_model_type.TS_NS_BLISS;
		QL_REQUIRE(is_ok,"invalid model");
	}
	
	void fit_model_params.init_params(	vector(number) guess,vector(number) min,vector(number) max,
										vector(logical) fixed,vector(number) fixed_val)								
	{
		guess_ 		= clone_vector(guess);
		min_ 		= clone_vector(min);
		max_ 		= clone_vector(max);
		fixed_ 		= ..clone_vector(fixed);
		fixed_val_ 	= ..clone_vector(fixed_val);

		integer size = param_size();
		QL_REQUIRE(size > 0,"invalid vector size");
		QL_REQUIRE(size == v_size(min_),"invalid vector size (min)");
		QL_REQUIRE(size == v_size(max_) ,"invalid vector size (max)");
		QL_REQUIRE(size == v_size(fixed_),"invalid vector size (fixed)");
		QL_REQUIRE(size == v_size(fixed_val_),"invalid vector size (fixed values)");
	}

	void fit_model_params.init_params(	vector(number) guess,vector(number) min,vector(number) max,
										vector(string) fixed,vector(number) fixed_val)								
	{
		guess_ 	= clone_vector(guess);
		min_ 	= clone_vector(min);
		max_ 	= clone_vector(max);
		fixed_val_ 	= clone_vector(fixed_val);

		integer size = param_size();
		QL_REQUIRE(size > 0,"invalid vector size");
		QL_REQUIRE(size == v_size(min_),"invalid vector size (min)");
		QL_REQUIRE(size == v_size(max_) ,"invalid vector size (max)");
		QL_REQUIRE(size == v_size(fixed_val_),"invalid vector size (fixed values)");

		set_fixed(fixed);
	}

	void fit_model_params.set_fixed(vector(string) f)
	{
		integer size = param_size();
		QL_REQUIRE(size == v_size(f),"invalid vector size");
		for(integer i=0;i<size;i++) {			
			if(!null(f[i]) && !null(fixed_val_[i]) && equal_casei(f[i], "true")) 
				fixed_[i] = true;			
			else 
				fixed_[i] = false;					
		}
		
	}
	
	fit_model_params fit_model_params(integer size,ts_model_type m)
	{
		return new fit_model_params(size,m);
	}
	

	integer fit_model_params.param_size() { return v_size(guess_);}

	class enum ts_fit_type
	{
		FIT_BEST_FIT,
		FT_EXACT_FIT,
		FT_EXACT_FIT_SPRD
	};

	//--------------------------
	// class ts_model <abstract>
	//--------------------------
	class ts_model //option (category: 'Yield Curve/005 Fitting')
	{
	public:
		ts_model(ts_model_type, fit_func_type,logical,logical u=false);	
	
		ts_fit_type				fit_t();
		string 					model_name();
		fit_func_type 			fit_func_t() ;
		virtual disc_func 		fit(curve) = 0 ;
		ts_model_type			model_t();
	protected:		
		ts_model_type	model_;
		fit_func_type	ft_;
		logical 		best_fit_;
		logical			using_sprd_;  
	};

	ts_model.ts_model(	ts_model_type 	mt,
						fit_func_type 	ft,
						logical 		best_fit,
						logical			using_sprd ) 
						: model_(mt), ft_(ft), best_fit_(best_fit), using_sprd_(using_sprd) {}	

	string 			ts_model.model_name() 	{ return string(model_);}		
	fit_func_type 	ts_model.fit_func_t() 	{ return ft_;}

	ts_fit_type		ts_model.fit_t()		
	{ 
		if(best_fit_) 
			return ts_fit_type.FIT_BEST_FIT;
		else 
			return using_sprd_ ? ts_fit_type.FT_EXACT_FIT_SPRD : ts_fit_type.FT_EXACT_FIT;
		
	}

	ts_model_type ts_model.model_t() { return model_;}
		
	//------------------------------
	// class ts_smith_wilson_ceiops
	//------------------------------
	class ts_smith_wilson_ceiops : public ts_model
	{
	public:
		ts_smith_wilson_ceiops();
		ts_smith_wilson_ceiops(	number 		option(nullable) , 
								number 		option(nullable),
								number 		option(nullable), 
								interpolator option(nullable), 
								number 		option(nullable), 
								number 		option(nullable), 
								string 		option(nullable) , 
								number 		option(nullable));

		void 				init_parm(	number 		option(nullable) , 
										number 		option(nullable),
										number 		option(nullable), 
										interpolator option(nullable), 
										number 		option(nullable), 
										number 		option(nullable), 
										string 		option(nullable) , 
										number 		option(nullable));
		override disc_func 	fit(curve) ;							

	protected:
										
		number 		liq_t_;//yrs 
		number 		ufr_t_; //yrs
		number 		ufr_; 
		interpolator interp_; 
		number 		ufr_tol_; 
		number 		a_start_; 
		string 		solver_; 
		number 		a_step_;
	};

	ts_smith_wilson_ceiops.ts_smith_wilson_ceiops() : ts_model(ts_model_type.TS_SW_CEI,fit_func_type.FT_SPEC,false), 
														liq_t_(def_liq_t),
														ufr_t_(def_ufr_t), 
														ufr_(def_ufr), 
														interp_(def_interp), 
														ufr_tol_(def_ufr_tol), 
														a_start_(def_a_start), 
														solver_(def_solver), 
														a_step_(def_a_step){}

	ts_smith_wilson_ceiops.ts_smith_wilson_ceiops(	number 		option(nullable) liq_t, 
													number 		option(nullable) ufr_t,
													number 		option(nullable) ufr, 
													interpolator option(nullable) interp, 
													number 		option(nullable) ufr_tol, 
													number 		option(nullable) a_start, 
													string 		option(nullable) solver, 
													number 		option(nullable) a_step) 
													: ts_model(ts_model_type.TS_SW_CEI,fit_func_type.FT_SPEC,false), 
													liq_t_(null(liq_t)? def_liq_t : liq_t), 
												  	ufr_t_(null(ufr_t)? def_ufr_t : ufr_t), 
													ufr_(null(ufr) ? def_ufr: ufr), 
													interp_(null(interp) ? def_interp: interp), 
													ufr_tol_(null(ufr_tol) ? def_ufr_tol: ufr_tol), 
													a_start_(null(a_start) ? def_a_start: a_start), 
													solver_(null(solver) ? def_solver: solver), 
													a_step_(null(a_step) ? def_a_step: a_step)
													{}

	void ts_smith_wilson_ceiops.init_parm(	number 		option(nullable) liq_t, 
											number 		option(nullable) ufr_t,
											number 		option(nullable) ufr, 
											interpolator option(nullable) interp, 
											number 		option(nullable) ufr_tol, 
											number 		option(nullable) a_start, 
											string 		option(nullable) solver, 
											number 		option(nullable) a_step)
	{
		liq_t_ = null(liq_t) ? def_liq_t : liq_t ;
		ufr_t_ = null(ufr_t)? def_ufr_t : ufr_t; 
		ufr_ = null(ufr) ? def_ufr: ufr; 
		interp_ = null(interp) ? def_interp: interp; 
		ufr_tol_ = null(ufr_tol) ? def_ufr_tol: ufr_tol; 
		a_start_ = null(a_start) ? def_a_start: a_start; 
		solver_ = null(solver) ? def_solver: solver; 
		a_step_ = null(a_step) ? def_a_step: a_step;
	}

	disc_func ts_smith_wilson_ceiops.fit(curve c)
	{
		number alfa_star;
		return smith_wilson(c,liq_t_,ufr_t_,ufr_,interp_,ufr_tol_,a_start_,solver_,a_step_,alfa_star);
	}

	//------------------------------
	// class ts_tanggaard
	//------------------------------
	class ts_tanggaard : public ts_model
	{
	public:
		ts_tanggaard();
		ts_tanggaard(smoothness_measure option(nullable),number option(nullable), std_weights option(nullable));

		void 				init_parm(smoothness_measure option(nullable),number option(nullable), std_weights option(nullable));
		override disc_func 	fit(curve) ;							

	protected:
										
		smoothness_measure 	smooth_;
		number				wf_;	
		std_weights 		w_;
	};

	ts_tanggaard.ts_tanggaard() : ts_model(ts_model_type.TS_TANG,fit_func_type.FT_SPEC,true), smooth_(def_smooth), wf_(def_w_factor), w_(def_w_scheme){}

	ts_tanggaard.ts_tanggaard(	smoothness_measure 	option(nullable) smooth,
								number 				option(nullable) weight_factor, 
								std_weights 		option(nullable) weights) 
								: ts_model(ts_model_type.TS_TANG,fit_func_type.FT_SPEC,true), smooth_(null(smooth) ?def_smooth :smooth), 
								  wf_(null(weight_factor)? def_w_factor : weight_factor), w_(null(weights) ? def_w_scheme:weights){}

	void ts_tanggaard.init_parm(smoothness_measure 	option(nullable) smooth,
								number 				option(nullable) weight_factor, 
								std_weights 		option(nullable) weights)
	{
		smooth_ = null(smooth) ? def_smooth :smooth;
		wf_ 	= null(weight_factor)? def_w_factor : weight_factor;
		w_ 		= null(weights) ? def_w_scheme:weights;
	}

	disc_func ts_tanggaard.fit(curve c)
	{
		return tanggaard(c,smooth_,wf_, w_);
	}

	//------------------------------
	// class ts_max_smooth
	//------------------------------
	class ts_max_smooth : public ts_model
	{

	public:
		ts_max_smooth();
		ts_max_smooth(smoothness_measure option(nullable) );

		void 					init_parm(smoothness_measure option(nullable) );
		override disc_func 		fit(curve) ;							

	protected:
										
		smoothness_measure smooth_;	
	};

	ts_max_smooth.ts_max_smooth() : ts_model(ts_model_type.TS_MS,fit_func_type.FT_SPEC,false), smooth_(def_smooth){}

	ts_max_smooth.ts_max_smooth(smoothness_measure option(nullable)  smooth) 
								: ts_model(ts_model_type.TS_MS,fit_func_type.FT_SPEC,false), smooth_(null(smooth) ?def_smooth :smooth){}

	void ts_max_smooth.init_parm(smoothness_measure option(nullable)  smooth)
	{
		smooth_ = null(smooth) ?def_smooth :smooth;
	}

	disc_func ts_max_smooth.fit(curve c)
	{
		return max_smooth(c,smooth_);
	}

	//------------------------------
	// class ts_max_smooth_sprd
	//------------------------------
	class ts_max_smooth_sprd : public ts_model
	{

	public:
		ts_max_smooth_sprd();
		ts_max_smooth_sprd(smoothness_measure option(nullable), vector(number) option(nullable));

		void 					init_parm(smoothness_measure option(nullable), vector(number) option(nullable));
		override disc_func 		fit(curve) ;							

	protected:
										
		smoothness_measure smooth_;	
		vector(number) spread_;		
	};

	ts_max_smooth_sprd.ts_max_smooth_sprd() : ts_model(ts_model_type.TS_MS_S,fit_func_type.FT_SPEC,false,true), smooth_(def_smooth), spread_([def_fit_spread]){}

	ts_max_smooth_sprd.ts_max_smooth_sprd(	smoothness_measure option(nullable) smooth, 
											vector(number) option(nullable) spread) 
											: ts_model(ts_model_type.TS_MS_S,fit_func_type.FT_SPEC,false,true), smooth_(null(smooth) ?def_smooth :smooth), 
											  spread_(null(spread)?[def_fit_spread]:clone_vector(spread)){}
	
	void ts_max_smooth_sprd.init_parm(	smoothness_measure 	option(nullable) smooth, 
										vector(number) 		option(nullable) spread)
	{
		smooth_ = null(smooth) ?def_smooth :smooth;
		spread_ = null(spread)?[def_fit_spread]:clone_vector(spread);
	}

	disc_func ts_max_smooth_sprd.fit(curve c)
	{
		vector(number) v_s 	= spread_v(c,spread_);
		return max_smooth_spreads(c, v_s, smooth_);

	}
	//------------------------------
	// class ts_tension_y
	//------------------------------
	class ts_tension_y : public ts_model
	{

	public:
		ts_tension_y();
		ts_tension_y(vector(number), vector(number), number option(nullable), number option(nullable));

		void 				init_parm(vector(number), vector(number), number option(nullable), number option(nullable));
		override disc_func 	fit(curve) ;							

	protected:
										
		interpolation 	sigma_;
		number 			left_der_;
		number 			right_der_;
	};

	ts_tension_y.ts_tension_y() : ts_model(ts_model_type.TS_TENS_Y,fit_func_type.FT_SPEC,false), left_der_(def_tens_left_der), right_der_(def_tens_right_der)
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), def_sigma_x, def_sigma_y);
	}
	ts_tension_y.ts_tension_y(	vector(number) sigma_x, 
								vector(number) sigma_y, 
								number option(nullable) left_der, 
								number option(nullable) right_der) 
							: ts_model(ts_model_type.TS_TENS_Y,fit_func_type.FT_SPEC,false), left_der_(left_der), right_der_(right_der)
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), sigma_x, sigma_y);
	}

	void ts_tension_y.init_parm(vector(number) 	sigma_x, 
								vector(number) 	sigma_y, 
								number 	option(nullable) left_der, 
								number 	option(nullable) right_der)
	{
		left_der_ 	= left_der;
		right_der_ 	= right_der;
		sigma_ 		= interpolation(ip_linear(0,0), sigma_x, sigma_y);	
	}

	disc_func ts_tension_y.fit(curve c)
	{
		return tension_spline_y(c,sigma_, left_der_,right_der_);
	}

	//------------------------------
	// class ts_tension_y_sprd
	//------------------------------
	class ts_tension_y_sprd : public ts_model
	{

	public:
		ts_tension_y_sprd();
		ts_tension_y_sprd(vector(number), vector(number), number option(nullable), number option(nullable), vector(number) option(nullable));

		void 					init_parm(vector(number), vector(number), number option(nullable), number option(nullable), vector(number) option(nullable));
		override disc_func 		fit(curve) ;						

	protected:
										
		interpolation 	sigma_;
		number 			left_der_;
		number 			right_der_;
		vector(number) 	spread_;	
	};

	ts_tension_y_sprd.ts_tension_y_sprd() : ts_model(ts_model_type.TS_TENS_Y_S,fit_func_type.FT_SPEC,false,true), left_der_(def_tens_left_der), 
											right_der_(def_tens_right_der), spread_([def_fit_spread])
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), def_sigma_x, def_sigma_y);
	}

	ts_tension_y_sprd.ts_tension_y_sprd(vector(number) 	sigma_x, 
										vector(number) 	sigma_y, 
										number option(nullable)	left_der, 
										number option(nullable)	right_der,
										vector(number) 	option(nullable) spread) 
										: ts_model(ts_model_type.TS_TENS_Y_S,fit_func_type.FT_SPEC,false,true), left_der_(left_der), 
										right_der_(right_der), spread_(null(spread) ? [def_fit_spread]: clone_vector(spread))
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), sigma_x, sigma_y);
	}

	void ts_tension_y_sprd.init_parm(	vector(number) 	sigma_x, 
										vector(number) 	sigma_y, 
										number option(nullable)	left_der, 
										number option(nullable)	right_der,
										vector(number) 	option(nullable) spread)
	{
		left_der_ 	= left_der;
		right_der_ 	= right_der;
		sigma_ 		= interpolation(ip_linear(0,0), sigma_x, sigma_y);	
		spread_ 	= null(spread) ? [def_fit_spread]: clone_vector(spread);
	}

	disc_func ts_tension_y_sprd.fit(curve c)
	{
		vector(number) v_s 	= spread_v(c,spread_);
		return tension_spline_y_spreads(c,v_s,sigma_, left_der_,right_der_);
	}

	//------------------------------
	// class ts_tension_y_bf
	//------------------------------
	class ts_tension_y_bf : public ts_model
	{

	public:
		ts_tension_y_bf();
		ts_tension_y_bf(vector(number), vector(number), std_weights option(nullable),number option(nullable), 
						number option(nullable), number option(nullable));

		void 				init_parm(vector(number), vector(number), std_weights option(nullable),
										number option(nullable), number option(nullable), number option(nullable));
		override disc_func 	fit(curve) ;							

	protected:
		std_weights		w_;									
		interpolation 	sigma_;
		number 			lambda_;
		number 			left_der_;
		number 			right_der_;
	};

	ts_tension_y_bf.ts_tension_y_bf() : ts_model(ts_model_type.TS_TENS_Y_BF,fit_func_type.FT_SPEC,true), w_(def_w_scheme), lambda_(def_tens_lambda), 
										left_der_(def_tens_left_der), right_der_(def_tens_right_der)
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), def_sigma_x, def_sigma_y);
	}

	ts_tension_y_bf.ts_tension_y_bf(vector(number) sigma_x, 
									vector(number) sigma_y, 
									std_weights	 option(nullable) w, 
									number  option(nullable) lambda, 
									number  option(nullable) left_der, 
									number  option(nullable) right_der) 
									: ts_model(ts_model_type.TS_TENS_Y_BF,fit_func_type.FT_SPEC,true), w_(null(w)?def_w_scheme:w), 
										lambda_(null(lambda)?def_tens_lambda:lambda), left_der_(left_der), right_der_(right_der)
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), sigma_x, sigma_y);
	}

	void ts_tension_y_bf.init_parm(	vector(number) 	sigma_x, 
									vector(number) 	sigma_y,
									std_weights	 option(nullable) w, 
									number 	option(nullable) lambda, 
									number 	option(nullable) left_der, 
									number 	 option(nullable) right_der)
	{
		w_			= null(w)?def_w_scheme:w;
		lambda_ 	= null(lambda)?def_tens_lambda:lambda;
		left_der_ 	= left_der;
		right_der_ 	= right_der;
		sigma_ 		= interpolation(ip_linear(0,0), sigma_x, sigma_y);	
	}

	disc_func ts_tension_y_bf.fit(curve c)
	{		
		return tension_spline_y_l2(c,w_,sigma_, lambda_,left_der_,right_der_);
	}

	//------------------------------
	// class ts_tension_w_bf
	//------------------------------
	class ts_tension_w_bf : public ts_model
	{

	public:
		ts_tension_w_bf();
		ts_tension_w_bf(vector(number), vector(number), std_weights option(nullable),number option(nullable),
						number option(nullable), number option(nullable));

		void 				init_parm(	vector(number), vector(number), std_weights option(nullable),number option(nullable),
										number option(nullable), number option(nullable));
		override disc_func 	fit(curve) ;							

	protected:
										
		std_weights		w_;									
		interpolation 	sigma_;
		number 			lambda_;
		number 			epsilon_;
		number 			left_der_;		
	};

	ts_tension_w_bf.ts_tension_w_bf() : ts_model(ts_model_type.TS_TENS_W_BF,fit_func_type.FT_SPEC,true), w_(def_w_scheme), lambda_(def_tens_lambda), 
										epsilon_(def_tens_epsilon), left_der_(def_tens_left_der)
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), def_sigma_x, def_sigma_y);
	}

	ts_tension_w_bf.ts_tension_w_bf(vector(number) 	sigma_x, 
									vector(number) 	sigma_y, 
									std_weights	 option(nullable) w, 
									number  option(nullable) lambda, 
									number  option(nullable) epsilon, 
									number  option(nullable) left_der) 
									: ts_model(ts_model_type.TS_TENS_W_BF,fit_func_type.FT_SPEC,true), w_(null(w)?def_w_scheme:w), 
										lambda_(null(lambda)?def_tens_lambda:lambda), 
										epsilon_(null(epsilon)?def_tens_epsilon:epsilon), left_der_(left_der)
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), sigma_x, sigma_y);
	}
	void ts_tension_w_bf.init_parm(vector(number) 	sigma_x, 
									vector(number) 	sigma_y, 
									std_weights	option(nullable) w, 
									number 	option(nullable) lambda, 
									number 	option(nullable) epsilon, 
									number 	option(nullable) left_der)
	{
		w_			= null(w)?def_w_scheme:w;
		lambda_ 	= null(lambda)?def_tens_lambda:lambda;
		epsilon_ 	= null(epsilon)?def_tens_epsilon:epsilon;
		left_der_ 	= left_der;
		sigma_ 		= interpolation(ip_linear(0,0), sigma_x, sigma_y);	
	}

	disc_func ts_tension_w_bf.fit(curve c)
	{
		return tension_spline_w_l2(c, w_, sigma_, lambda_, epsilon_,left_der_);
	}

	//------------------------------
	// class ts_tension_z_bf
	//------------------------------
	class ts_tension_z_bf : public ts_model
	{

	public:
		ts_tension_z_bf();
		ts_tension_z_bf(vector(number), vector(number),std_weights option(nullable),number option(nullable));

		void 					init_parm(vector(number), vector(number),std_weights option(nullable),number option(nullable));
		override disc_func 		fit(curve) ;							
	protected:							
		std_weights		w_;									
		interpolation 	sigma_;
		number 			lambda_;
		
	};

	ts_tension_z_bf.ts_tension_z_bf() : ts_model(ts_model_type.TS_TENS_Z_BF, fit_func_type.FT_SPEC,true),w_(def_w_scheme), lambda_(def_tens_lambda)
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), def_sigma_x, def_sigma_y);
	}

	ts_tension_z_bf.ts_tension_z_bf(vector(number) sigma_x, 
									vector(number) sigma_y,
									std_weights	option(nullable) w, 
									number option(nullable) lambda) 
									: ts_model(ts_model_type.TS_TENS_Z_BF, fit_func_type.FT_SPEC,true),w_(null(w)?def_w_scheme:w), 
										lambda_(null(lambda)?def_tens_lambda:lambda)
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), sigma_x, sigma_y);
	}

	void ts_tension_z_bf.init_parm(	vector(number) sigma_x, 
									vector(number) sigma_y, 
									std_weights	option(nullable) w, 
									number 	option(nullable) lambda)
	{
		w_		= null(w)?def_w_scheme:w;
		lambda_ = null(lambda)?def_tens_lambda:lambda;
		sigma_ 	= interpolation(ip_linear(0,0), sigma_x, sigma_y);	
	}

	disc_func ts_tension_z_bf.fit(curve c)
	{
		return tension_spline_z_l2(c,w_,sigma_,lambda_);
	}

	//------------------------------
	// class ts_tension_w
	//------------------------------
	class ts_tension_w : public ts_model
	{

	public:
		ts_tension_w();
		ts_tension_w(vector(number), vector(number), number option(nullable), number option(nullable));

		void 					init_parm(vector(number), vector(number), number option(nullable), number option(nullable));
		override disc_func 		fit(curve) ;							

	protected:
										
		interpolation 	sigma_;
		number 			epsilon_;
		number 			left_der_;		
	};

	ts_tension_w.ts_tension_w() : ts_model(ts_model_type.TS_TENS_W,fit_func_type.FT_SPEC,false), epsilon_(def_tens_epsilon), left_der_(def_tens_left_der)
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), def_sigma_x, def_sigma_y);
	}

	ts_tension_w.ts_tension_w(	vector(number) 	sigma_x, 
								vector(number) 	sigma_y, 
								number option(nullable) epsilon, 
								number option(nullable) left_der) 
								: ts_model(ts_model_type.TS_TENS_W,fit_func_type.FT_SPEC,false), epsilon_(null(epsilon)?def_tens_epsilon:epsilon), left_der_(left_der)
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), sigma_x, sigma_y);
	}
	void ts_tension_w.init_parm(vector(number) 	sigma_x, 
								vector(number) 	sigma_y, 
								number option(nullable) epsilon, 
								number 	option(nullable) left_der)
	{
		epsilon_ 	= null(epsilon)?def_tens_epsilon:epsilon;
		left_der_ 	= left_der;
		sigma_ 		= interpolation(ip_linear(0,0), sigma_x, sigma_y);	
	}

	disc_func ts_tension_w.fit(curve c)
	{
		return tension_spline_w(c,sigma_, epsilon_,left_der_);
	}

	//------------------------------
	// class ts_tension_w_sprd
	//------------------------------
	class ts_tension_w_sprd : public ts_model
	{

	public:
		ts_tension_w_sprd();
		ts_tension_w_sprd(vector(number), vector(number), number option(nullable), number option(nullable), vector(number) option(nullable));

		void 					init_parm(vector(number), vector(number), number option(nullable), number option(nullable), vector(number) option(nullable));
		override disc_func 		fit(curve) ;							

	protected:
										
		interpolation 	sigma_;
		number 			epsilon_;
		number 			left_der_;
		vector(number)	spread_;
	};

	ts_tension_w_sprd.ts_tension_w_sprd() : ts_model(ts_model_type.TS_TENS_W_S,fit_func_type.FT_SPEC,false,true), epsilon_(def_tens_epsilon), 
											left_der_(def_tens_left_der),  spread_([def_fit_spread])
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), def_sigma_x, def_sigma_y);
	}

	ts_tension_w_sprd.ts_tension_w_sprd(vector(number) 	sigma_x, 
										vector(number) 	sigma_y, 
										number 	 option(nullable) epsilon, 
										number  option(nullable) left_der,
										vector(number)  option(nullable) spread) 
										: ts_model(ts_model_type.TS_TENS_W_S,fit_func_type.FT_SPEC,false,true), epsilon_(null(epsilon)?def_tens_epsilon:epsilon), 
										left_der_(left_der), spread_(null(spread) ? [def_fit_spread] : clone_vector(spread))
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), sigma_x, sigma_y);
	}

	void ts_tension_w_sprd.init_parm(	vector(number) 	sigma_x, 
										vector(number) 	sigma_y, 
										number 	 option(nullable) epsilon, 
										number 	 option(nullable) left_der,
										vector(number)  option(nullable) spread)
	{
		epsilon_ 	= null(epsilon)?def_tens_epsilon:epsilon;
		left_der_ 	= left_der;
		sigma_ 		= interpolation(ip_linear(0,0), sigma_x, sigma_y);	
		spread_ 	= null(spread) ? [def_fit_spread] : clone_vector(spread);
	}

	disc_func ts_tension_w_sprd.fit(curve c)
	{
		vector(number) v_s 	= spread_v(c,spread_);
		return tension_spline_w_spreads(c,v_s, sigma_, epsilon_,left_der_);
	}


	//------------------------------
	// class ts_tension_z
	//------------------------------
	class ts_tension_z : public ts_model
	{

	public:
		ts_tension_z();
		ts_tension_z(vector(number), vector(number));

		void 					init_parm(vector(number), vector(number));
		override disc_func 		fit(curve) ;							
	protected:							
		interpolation 	sigma_;
		
	};

	ts_tension_z.ts_tension_z() : ts_model(ts_model_type.TS_TENS_Z, fit_func_type.FT_SPEC,false)
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), def_sigma_x, def_sigma_y);
	}

	ts_tension_z.ts_tension_z(vector(number) sigma_x, vector(number) sigma_y) 
							: ts_model(ts_model_type.TS_TENS_Z, fit_func_type.FT_SPEC,false)
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), sigma_x, sigma_y);
	}

	void ts_tension_z.init_parm(vector(number) sigma_x, vector(number) sigma_y)
	{
		sigma_ 	= interpolation(ip_linear(0,0), sigma_x, sigma_y);	
	}

	disc_func ts_tension_z.fit(curve c)
	{
		return tension_spline_z(c,sigma_);
	}

	//------------------------------
	// class ts_tension_z_sprd
	//------------------------------
	class ts_tension_z_sprd : public ts_model
	{

	public:
		ts_tension_z_sprd();
		ts_tension_z_sprd(vector(number), vector(number),vector(number) option(nullable));

		void 					init_parm(vector(number), vector(number),vector(number)  option(nullable));
		override disc_func 		fit(curve) ;							
	protected:							
		interpolation 	sigma_;
		vector(number)	spread_;
	};

	ts_tension_z_sprd.ts_tension_z_sprd() : ts_model(ts_model_type.TS_TENS_Z_S,fit_func_type.FT_SPEC,false,true), spread_([def_fit_spread])
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), def_sigma_x, def_sigma_y);
	}

	ts_tension_z_sprd.ts_tension_z_sprd(vector(number) sigma_x, 
										vector(number) sigma_y,
										vector(number)  option(nullable) spread) 
										: ts_model(ts_model_type.TS_TENS_Z_S,fit_func_type.FT_SPEC,false,true), spread_(null(spread) ? [def_fit_spread] : clone_vector(spread))
	{ 
		sigma_ 	= interpolation(ip_linear(0,0), sigma_x, sigma_y);
	}

	void ts_tension_z_sprd.init_parm(	vector(number) sigma_x, 
										vector(number) sigma_y,
										vector(number)  option(nullable) spread)
	{
		sigma_ 		= interpolation(ip_linear(0,0), sigma_x, sigma_y);	
		spread_		= null(spread) ? [def_fit_spread] : clone_vector(spread);
	}

	disc_func ts_tension_z_sprd.fit(curve c)
	{
		vector(number) v_s 	= spread_v(c,spread_);
		return tension_spline_z_spreads(c,v_s, sigma_);
	}

	//------------------------------
	// class ts_fit_model <abstract>
	//------------------------------
	class ts_fit_model : public ts_model
	{

	public:
		void 					set_fit_parm(string, string, logical);
		void 					set_fit_parm(std_weights option(nullable) , number option(nullable), logical );
		void					set_model_parm(	vector(number) option(nullable) 	guess,
												vector(number) option(nullable)  	min,
												vector(number) option(nullable)  	max,
												vector(logical) option(nullable)  	fixed,
												vector(number) option(nullable)  	fixed_val);
		void					set_model_parm(fit_model_params);

		virtual integer			n_parameters() = 0;
		fit_result 				fit_fr(curve) ;
		override disc_func 		fit(curve) ;		
		vector(string) 			parameter_names();						
		logical					is_init();
		std_weights 			w_scheme();
		
		//returns err
		virtual void 			init_knots(vector(number));	
		virtual void 			init_knots( curve, date, ..knot_scheme, number option(nullable));	
		virtual void 			init_knots( curve option(nullable), date option(nullable), string, vector(number) option(nullable), number option(nullable));
		virtual void			init_degree(integer);

	protected:
		ts_fit_model(ts_model_type);								
		void 					init_parm();
		void 					check_knots(vector(number));		
		virtual void 			set_knots(vector(number) option(nullable));
		virtual void 			set_knots(curve option(nullable), date option(nullable), string, vector(number) option(nullable), number option(nullable));	
		virtual void			set_knots(curve,date,..knot_scheme ,number option(nullable) );

		std_weights 	w_;
		number			norm_;
		logical			use_dhsa_;
		vector(number)  guess_;
		vector(number)  min_;
		vector(number)  max_;
		vector(logical) fixed_;
		vector(number)  fixed_val_;
		model			m_;	
		vector(number) 	knots_;	
		number 			degree_;	
	};

	ts_fit_model.ts_fit_model(ts_model_type mt) : ts_model(mt,fit_func_type.FT_FIT,true), w_(def_w_scheme), norm_(def_norm),use_dhsa_(false), m_(null<model>)
	{}
	
	logical			ts_fit_model.is_init() 	{ return !null(m_);}

	void ts_fit_model.set_fit_parm(string w, string norm, logical use_dhsa)
	{
		w_ = weight_scheme_map_ext(w);
		if(null(w_)) w_ = def_w_scheme;
		norm_ = norm_map_ext(norm);
		if(null(norm_)) norm_ = def_norm;
		use_dhsa_ = use_dhsa;
	}
	
	void ts_fit_model.set_fit_parm(std_weights option(nullable)  w, number option(nullable)  norm, logical use_dhsa)
	{
		w_ = null(w) ? def_w_scheme : w;
		norm_ = null(norm) ? def_norm: norm;
		use_dhsa_ = use_dhsa; 
	}

	void ts_fit_model.init_parm()  
	{ 		
		integer size 	= n_parameters();
		resize(guess_,size);
		guess_			= null<number>;
		resize(min_,size);
		min_			= null<number>;
		resize(max_,size);
		max_			= null<number>;
		resize(fixed_,size);
		fixed_ 			= false;
		resize(fixed_val_,size);
		fixed_val_		= null<number>;
	}

	std_weights ts_fit_model.w_scheme() { return w_;}
	
	vector(string) ts_fit_model.parameter_names() 	
	{ 
		QL_REQUIRE(is_init(), "model not initialized");
		return m_.parameter_names();
	}

	fit_result ts_fit_model.fit_fr(curve c)
	{
		QL_REQUIRE(is_init(), "model not initialized");

		return use_dhsa_ ? ..fit_dhsa(c, m_, w_, norm_) : ..fit(c, m_, w_, norm_);
	}
	
	disc_func ts_fit_model.fit(curve c)
	{
		return fit_fr(c);
	}

	void ts_fit_model.set_model_parm(	vector(number) option(nullable) 	guess,
										vector(number) option(nullable)  	min,
										vector(number) option(nullable)  	max,
										vector(logical) option(nullable)  	fixed,
										vector(number) option(nullable)  	fixed_val)
	{
		integer size = n_parameters();			

		if (v_size(guess) == size)
			guess_ 	= clone_vector(guess);

		if (v_size(min) == size) 
			min_	= clone_vector(min);

		if (v_size(max) == size) 
			max_	= clone_vector(max);

		if (v_size(fixed) == size) 
			fixed_	= ..clone_vector(fixed);												

		if (v_size(fixed_val) == size) 
			fixed_val_	= ..clone_vector(fixed_val);	

		for(integer i=0;i<size;i++) {
			if(!null(fixed_val_[i]) && !null(fixed_[i]) && fixed_[i])					
				fixed_[i] = true; 
			else
				fixed_[i] = false;
		}

		vector(string) params = parameter_names();
		m_ = m_.bound_params(params, guess_, min_, max_);
	
		for(integer i=0;i<size;i++) {			
			if(fixed_[i]) 
				m_ = m_.fix_param (params[i], fixed_val_[i]);							
		}								
	}
	void ts_fit_model.set_model_parm(fit_model_params p)
	{
		integer size = n_parameters();
		QL_REQUIRE(size == p.param_size() , "invalid number of parameters");

		QL_REQUIRE(model_ == p.model_,"invalid params (wrong model)");
		
		guess_ 	= p.guess_;
		min_ 	= p.min_;
		max_ 	= p.max_;
		fixed_ 	= p.fixed_;
		fixed_val_ 	= p.fixed_val_;

		vector(string) params = parameter_names();
		m_ = m_.bound_params(params, guess_, min_, max_);
	
		for(integer i=0;i<size;i++) {			
			if(fixed_[i]) 
				m_ = m_.fix_param (params[i], fixed_val_[i]);							
		}
	}

	void  ts_fit_model.check_knots(vector(number) knots)
	{
		integer no_knots = v_size(knots);

		QL_REQUIRE(no_knots > 1 , "invalid number of knots");
						
		for (integer i=0; i<no_knots;i++) {
			if(null(knots[i]) || knots[i] < 0)
				QL_FAIL("invalid knot");
			if(i>0) {
				QL_REQUIRE(knots[i] > knots[i-1], "invalid knot (unordered)");
			}
		}	
	}
	void  ts_fit_model.set_knots(vector(number) option(nullable) knots)
	{		
		QL_REQUIRE(!null(knots),"invalid knots");
		check_knots(knots);
		knots_	= clone_vector(knots);	
	}

	void ts_fit_model.set_knots(curve option(nullable) 	c, 
								date option(nullable) 	t, 
								string 					ks, 
								vector(number) option(nullable) knots, 
								number option(nullable) no_knots)////if null or -1 use MCCULL
	{
		if(str_to_lower(ks) != "default")
			set_knots(def_knots);
		else if(str_to_lower(ks) != "user") {												
			QL_REQUIRE(!null(c),"invalid curve");
			QL_REQUIRE(!null(t),"invalid date");	
			set_knots(c, t, get_scheme(ks), no_knots);						
		}
		else
			set_knots(knots);
	}

	void ts_fit_model.set_knots( 	curve 			c, 
									date 			t, 
									..knot_scheme 	ks, 
									number option(nullable) no_knots)
	{
		vector(number) knots = ..spline_knots(c, t, ks, no_knots);
		set_knots(knots);
	}

	void 	ts_fit_model.init_knots(vector(number) )										{ return;/*throw(E_INVALID_ARG,"invalid function call");*/}
	void 	ts_fit_model.init_knots(curve, date, ..knot_scheme, number option(nullable) )	{ return;/*throw(E_INVALID_ARG,"invalid function call");*/}
	void 	ts_fit_model.init_knots(curve option(nullable),date option(nullable),string,vector(number) option(nullable), 
									number option(nullable) ) 								{ return;/*throw(E_INVALID_ARG,"invalid function call");*/}
	void 	ts_fit_model.init_degree(integer) 												{ return;/*throw(E_INVALID_ARG,"invalid function call");*/}

	//------------------
	// class ts_ns_orig
	//------------------
	class ts_ns_orig : public ts_fit_model
	{
	public:
		ts_ns_orig();						
		override integer		n_parameters();					
	};	
	
	ts_ns_orig.ts_ns_orig() : ts_fit_model(ts_model_type.TS_NS)  
	{ 
		m_ = ..ns();
		init_parm();
	}
	
	integer ts_ns_orig.n_parameters() 				
	{ 
		QL_REQUIRE(is_init(),"model not initialized");
		return m_.n_parameters();
	}

	//------------------
	// class ts_ns_bliss
	//------------------
	class ts_ns_bliss: public ts_fit_model
	{
	public:
		ts_ns_bliss();						
		override integer		n_parameters();					
	};	
	
	ts_ns_bliss.ts_ns_bliss() : ts_fit_model(ts_model_type.TS_NS_BLISS)  
	{ 
		m_ = ..ns_bliss();
		init_parm();
	}
	
	integer ts_ns_bliss.n_parameters() 				
	{ 
		QL_REQUIRE(is_init(),"model not initialized");
		return m_.n_parameters();
	}
	
	//------------------
	// class ts_ns_svensson
	//------------------
	class ts_ns_svensson: public ts_fit_model
	{
	public:
		ts_ns_svensson();						
		override integer		n_parameters();					
	};	
	
	ts_ns_svensson.ts_ns_svensson() : ts_fit_model(ts_model_type.TS_NS_SVEN)  
	{ 
		m_ = ..ns_svensson();
		init_parm();
	}
	
	integer ts_ns_svensson.n_parameters() 				
	{ 
		QL_REQUIRE(is_init(),"model not initialized");
		return m_.n_parameters();
	}

	//------------------
	// class ts_ns_laguerre
	//------------------
	class ts_ns_laguerre: public ts_fit_model
	{
	public:
		ts_ns_laguerre();						
		override integer		n_parameters();					
	};	
	
	ts_ns_laguerre.ts_ns_laguerre() : ts_fit_model(ts_model_type.TS_NS_LAG)  
	{ 
		m_ = ..ns_laguerre();
		init_parm();
	}
	
	integer ts_ns_laguerre.n_parameters() 				
	{ 
		QL_REQUIRE(is_init(),"model not initialized");
		return m_.n_parameters();
	}

	//------------------
	// class ts_hermite
	//------------------
	class ts_hermite : public ts_fit_model
	{
	public:
		ts_hermite(ts_spline_type,vector(number) option(nullable) k=null<vector(number)>);

		override void 			init_knots(vector(number));	
		override void 			init_knots( curve, date, ..knot_scheme, number option(nullable));	
		override void 			init_knots( curve option(nullable), date option(nullable), string, vector(number) option(nullable), number option(nullable));
		override integer		n_parameters();		
		override void 			set_knots(vector(number) option(nullable));
		override void 			set_knots(curve option(nullable), date option(nullable), string, vector(number) option(nullable), number option(nullable));	
		override void			set_knots(curve,date,..knot_scheme ,number option(nullable) );

	private:
		void set_model();
		ts_spline_type t_;				
	};	
		
	ts_hermite.ts_hermite(ts_spline_type t, vector(number) option(nullable) knots) 
						: ts_fit_model(t == ts_spline_type.ST_DISC ? ts_model_type.TS_HERM_D : (t == ts_spline_type.ST_SPOT ? ts_model_type.TS_HERM_S :ts_model_type.TS_HERM_F )), t_(t)  
	{	
		if(null(knots)) 
			knots = def_knots;
		set_knots(knots);		
		set_model();
	}	

	void ts_hermite.set_model()
	{
		switch(t_){
		case ts_spline_type.ST_DISC:
			m_ 	= ..hermite_disc(knots_); 
			break;
		case ts_spline_type.ST_SPOT:
			m_ 	= ..hermite_spot(knots_); 
			break;
		default:
			m_ 	= ..hermite_fwd(knots_); 
			break;
		}
		init_parm();
	}

	void ts_hermite.init_knots(vector(number) knots)
	{
		set_knots(knots);
		set_model();
	}

	void ts_hermite.init_knots( curve c, date t, ..knot_scheme ks, number option(nullable) no_knots)
	{
		set_knots(c, t, ks, no_knots);		
		set_model();
	}

	void ts_hermite.init_knots(curve option(nullable) c, 
							date option(nullable) t, 
							string ks, 
							vector(number) option(nullable) knots, 
							number option(nullable) no_knots)////if null or -1 use MCCULL
	{
		set_knots(c, t, ks, knots, no_knots);
		set_model();
	}

	integer ts_hermite.n_parameters() 				
	{ 
		QL_REQUIRE(is_init(),"model not initialized");
		return m_.n_parameters();
	}
	
	void  ts_hermite.set_knots(vector(number) option(nullable) knots)
	{
		ts_fit_model.set_knots(knots);
	}

	void ts_hermite.set_knots(curve option(nullable) 	c, 
							date option(nullable) 	t, 
							string 					ks, 
							vector(number) option(nullable) knots, 
							number option(nullable) no_knots)////if null or -1 use MCCULL
	{
		ts_fit_model.set_knots(c, t,ks,knots,no_knots);
	}

	void ts_hermite.set_knots( curve 			c, 
							date 			t, 
							..knot_scheme 	ks, 
							number option(nullable) no_knots)
	{
		ts_fit_model.set_knots(c, t,ks,no_knots);
	}


	//------------------
	// class ts_bspline
	//------------------
	class ts_bspline : public ts_fit_model
	{
	public:
		ts_bspline(ts_spline_type,number option(nullable) d = null<number>, vector(number) option(nullable) k = null<vector(number)>);

		override void 			init_knots(vector(number));	
		override void 			init_knots( curve, date, ..knot_scheme, number option(nullable));	
		override void 			init_knots( curve option(nullable), date option(nullable), string, vector(number) option(nullable), number option(nullable));
		override integer		n_parameters();		
		override void 			set_knots(vector(number) option(nullable));
		override void 			set_knots(curve option(nullable), date option(nullable), string, vector(number) option(nullable), number option(nullable));	
		override void			set_knots(curve,date,..knot_scheme ,number option(nullable) );
		override void			init_degree(integer);
	private:
		void set_model();
		ts_spline_type t_;				
	};	
		
	ts_bspline.ts_bspline(ts_spline_type t, number option(nullable) degree, vector(number) option(nullable) knots) 
					: ts_fit_model(t == ts_spline_type.ST_DISC ? ts_model_type.TS_BSPL_D : (t == ts_spline_type.ST_SPOT ? ts_model_type.TS_BSPL_S :ts_model_type.TS_BSPL_F )), t_(t)  
	{	
		degree_ = integer(null(degree) ? def_spline_degree: degree);	
		QL_REQUIRE(degree_ > 2,"invalid spline degree");
		if(null(knots)) 
			knots = def_knots;
		set_knots(knots);		
		set_model();
	}	

	void ts_bspline.set_model()
	{
		switch(t_){
		case ts_spline_type.ST_DISC:
			m_ 	= ..bspline_disc(knots_); 
			break;
		case ts_spline_type.ST_SPOT:
			m_ 	= ..bspline_spot(knots_); 
			break;
		default:
			m_ 	= ..bspline_fwd(knots_); 
			break;
		}
		init_parm();
	}

	void ts_bspline.init_knots(vector(number) knots)
	{
		set_knots(knots);
		set_model();
	}

	void ts_bspline.init_knots( curve c, date t, ..knot_scheme ks, number option(nullable) no_knots)
	{
		set_knots(c, t, ks, no_knots);		
		set_model();
	}

	void ts_bspline.init_knots(curve option(nullable) c, 
							date option(nullable) t, 
							string ks, 
							vector(number) option(nullable) knots, 
							number option(nullable) no_knots)////if null or -1 use MCCULL
	{
		set_knots(c, t, ks, knots, no_knots);
		set_model();
	}

	integer ts_bspline.n_parameters() 				
	{ 
		QL_REQUIRE(is_init(),"model not initialized");
		return m_.n_parameters();
	}

	void  ts_bspline.set_knots(vector(number) option(nullable) knots)
	{
		ts_fit_model.set_knots(knots);
	}

	void ts_bspline.set_knots(curve option(nullable) 	c, 
							date option(nullable) 	t, 
							string 					ks, 
							vector(number) option(nullable) knots, 
							number option(nullable) no_knots)////if null or -1 use MCCULL
	{
		ts_fit_model.set_knots(c, t,ks,knots,no_knots);
	}

	void ts_bspline.set_knots( curve 			c, 
							date 			t, 
							..knot_scheme 	ks, 
							number option(nullable) no_knots)
	{
		ts_fit_model.set_knots(c, t,ks,no_knots);
	}

	void	ts_bspline.init_degree(integer d)
	{
		QL_REQUIRE(d > 2,"invalid spline degree");
		degree_ = d;
	}

	//------------------
	// class ts_cspline
	//------------------
	class ts_cspline : public ts_fit_model
	{
	public:
		ts_cspline(ts_spline_type,number option(nullable) d = null<number> , vector(number) option(nullable) k=null<vector(number)>);

		override void 			init_knots(vector(number));	
		override void 			init_knots( curve, date, ..knot_scheme, number option(nullable));	
		override void 			init_knots( curve option(nullable), date option(nullable), string, vector(number) option(nullable), number option(nullable));
		override integer		n_parameters();		
		override void 			set_knots(vector(number) option(nullable));
		override void 			set_knots(curve option(nullable), date option(nullable), string, vector(number) option(nullable), number option(nullable));	
		override void			set_knots(curve,date,..knot_scheme ,number option(nullable) );
		override void			init_degree(integer);
	private:
		void set_model();
		ts_spline_type t_;				
	};	
		
	ts_cspline.ts_cspline(ts_spline_type t, number option(nullable) degree, vector(number) option(nullable) knots) 
					: ts_fit_model(t == ts_spline_type.ST_DISC ? ts_model_type.TS_CSPL_D : (t == ts_spline_type.ST_SPOT ? ts_model_type.TS_CSPL_S :ts_model_type.TS_CSPL_F )), t_(t)  
	{	
		degree_ = integer(null(degree) ? def_spline_degree: degree);
		QL_REQUIRE(degree_ > 2,"invalid spline degree");
		if(null(knots)) 
			knots = def_knots;
		set_knots(knots);		
		set_model();
	}	

	void ts_cspline.set_model()
	{
		switch(t_){
		case ts_spline_type.ST_DISC:
			m_ 	= ..cspline_disc(knots_); 
			break;
		case ts_spline_type.ST_SPOT:
			m_ 	= ..cspline_spot(knots_); 
			break;
		default:
			m_ 	= ..cspline_fwd(knots_); 
			break;
		}
		init_parm();
	}

	void ts_cspline.init_knots(vector(number) knots)
	{
		set_knots(knots);
		set_model();
	}

	void ts_cspline.init_knots( curve c, date t, ..knot_scheme ks, number option(nullable) no_knots)
	{
		set_knots(c, t, ks, no_knots);		
		set_model();
	}

	void ts_cspline.init_knots(curve option(nullable) c, 
							date option(nullable) t, 
							string ks, 
							vector(number) option(nullable) knots, 
							number option(nullable) no_knots)////if null or -1 use MCCULL
	{
		set_knots(c, t, ks, knots, no_knots);
		set_model();
	}

	integer ts_cspline.n_parameters() 				
	{ 
		QL_REQUIRE(is_init(),"model not initialized");
		return m_.n_parameters();
	}

	void  ts_cspline.set_knots(vector(number) option(nullable) knots)
	{
		ts_fit_model.set_knots(knots);
	}

	void ts_cspline.set_knots(curve option(nullable) 	c, 
							date option(nullable) 	t, 
							string 					ks, 
							vector(number) option(nullable) knots, 
							number option(nullable) no_knots)////if null or -1 use MCCULL
	{
		ts_fit_model.set_knots(c, t,ks,knots,no_knots);
	}

	void ts_cspline.set_knots( curve 			c, 
							date 			t, 
							..knot_scheme 	ks, 
							number option(nullable) no_knots)
	{
		ts_fit_model.set_knots(c, t,ks,no_knots);
	}

	void	ts_cspline.init_degree(integer d)
	{
		QL_REQUIRE(d > 2,"invalid spline degree");	
		degree_ = d;
	}

	//-------------------------------
	// class ts_boot_model <abstract>
	//-------------------------------
	class ts_boot_model : public ts_model
	{

	public:
		
		void 					set_parm(rate_type option(nullable), logical option(nullable), calendar option(nullable));
		/*void 					set_fit_parm(std_weights , number);
		void					set_model_parm(	vector(number) option(nullable) 	guess,
												vector(number) option(nullable)  	min,
												vector(number) option(nullable)  	max,
												vector(logical) option(nullable)  	fixed);*/	
		override disc_func 		fit(curve) ;									
		virtual logical			is_init() = 0;
		virtual void 			set_flags(logical ,logical ) ;
		virtual void 			init_ip(number option(nullable),number option(nullable),logical, logical);
		virtual void 			init_ip(number option(nullable),logical, logical);
		virtual void 			init_ip(logical option(nullable), logical);
		virtual void 			init_ip(number option(nullable),number option(nullable), logical);
		virtual void 			init_ip(logical);
	protected:
		ts_boot_model(ts_model_type);								

		interpolator 	ip_;
		logical 		on_interp_;	
		rate_type		rate_type_;		
		calendar		cal_;
		integer			flags_;
		number			left_der_;
		number			right_der_;
	};
	
	ts_boot_model.ts_boot_model(ts_model_type mt) : ts_model(mt,fit_func_type.FT_BOOT,false), ip_(null<interpolator>), on_interp_(def_boot_on_interp), 
													rate_type_(def_boot_rate_type), cal_(calendar()),  
													left_der_(def_boot_left_der),right_der_(def_boot_right_der)
	{ 
		if(model_ == ts_model_type.TS_HAGAN && on_interp_)
			QL_FAIL("o/n interpolation not compatible with the Hagan West model");
		set_flags(def_boot_apply_hyman, def_boot_extrap);
	}

	void ts_boot_model.set_parm(rate_type option(nullable) rt, 
								logical option(nullable) on_interp, 
								calendar option(nullable) cal)
	{		
		rate_type_ 	= null(rt)?def_boot_rate_type:rt;
		on_interp_ 	= null(on_interp)?def_boot_on_interp:on_interp;
		cal_		= cal;
		if(model_ == ts_model_type.TS_HAGAN && on_interp_)
			QL_FAIL("o/n interpolation not compatible with the Hagan West model");
	}

	disc_func ts_boot_model.fit(curve c)
	{
		QL_REQUIRE(is_init(),"model not initialized");
		date td = c.instruments()[0].trade_date();
		date max = CORE_INT.max_curve_date(c);
		return on_interp_ ? bootstrap_on_extrap_flat(c,td,max, cal_): bootstrap(c,ip_, rate_type_);
	}	

	void ts_boot_model.set_flags(logical apply_hyman, logical extrap) 				
	{ 
		if(apply_hyman) 
			flags_ = !extrap ? IP_HYMAN | IP_NO_XPOL_R : IP_HYMAN;		
		else if(!extrap)
			flags_ = IP_NO_XPOL_R;	
	}

	void ts_boot_model.init_ip( logical) 															{ QL_FAIL("invalid function call");}
	void ts_boot_model.init_ip(logical option(nullable), logical) 									{ QL_FAIL("invalid function call");}
	void ts_boot_model.init_ip(number option(nullable),number option(nullable),logical, logical) 	{ QL_FAIL("invalid function call");}
	void ts_boot_model.init_ip(number option(nullable),number option(nullable),logical) 			{ QL_FAIL("invalid function call");}
	void ts_boot_model.init_ip(number option(nullable),logical, logical) 							{ QL_FAIL("invalid function call");}

	//------------------
	// class ts_boot_hagan_west
	//------------------
	class ts_boot_hagan_west: public ts_boot_model
	{
	public:
		ts_boot_hagan_west();						
		ts_boot_hagan_west(logical option(nullable), logical );
		override void 		init_ip(logical option(nullable), logical);
		override void 		set_flags(logical ,logical );
		override logical	is_init() ;					
	protected:
		logical force_pos_;
		
	};	
	
	ts_boot_hagan_west.ts_boot_hagan_west() : ts_boot_model(ts_model_type.TS_HAGAN), force_pos_(def_mc_force_pos)  
	{
		ip_ = ip_hagan_west(force_pos_,flags_);
	}

	ts_boot_hagan_west.ts_boot_hagan_west(	logical option(nullable) force_pos, 
											logical extrap) : ts_boot_model(ts_model_type.TS_HAGAN), force_pos_(null(force_pos)?def_mc_force_pos:force_pos)  
	{ 		
		init_ip(force_pos, extrap); 
	}
	
	logical	ts_boot_hagan_west.is_init() { return !null(ip_);}

	void ts_boot_hagan_west.set_flags(logical apply_hyman, logical extrap) //apply_hyman not applicable				
	{ 
		if(!extrap)
			flags_ = IP_NO_XPOL_R;	
	}

	void ts_boot_hagan_west.init_ip(logical option(nullable) force_pos, logical extrap) 				
	{ 
		force_pos_ = null(force_pos)?def_mc_force_pos:force_pos;
		set_flags(false, extrap);
		ip_ = ip_hagan_west(force_pos_,flags_);
	}

	//------------------
	// class ts_boot_step
	//------------------
	class ts_boot_step: public ts_boot_model
	{
	public:
		ts_boot_step();						
		ts_boot_step(logical );
		override void 		init_ip(logical);
		override void 		set_flags(logical ,logical );
		override logical	is_init() ;					
	protected:
		
	};	
	
	ts_boot_step.ts_boot_step() : ts_boot_model(ts_model_type.TS_STEP) 
	{
		ip_ = ip_step(flags_);
	}

	ts_boot_step.ts_boot_step(logical extrap) : ts_boot_model(ts_model_type.TS_STEP) 
	{ 
		init_ip(extrap); 
	}
	
	logical	ts_boot_step.is_init() { return !null(ip_);}

	void ts_boot_step.set_flags(logical apply_hyman, logical extrap) //apply_hyman not applicable				
	{ 
		if(!extrap)
			flags_ = IP_NO_XPOL_R;	
	}

	void ts_boot_step.init_ip(logical extrap) 				
	{ 
		set_flags(false, extrap);
		ip_ = ip_step(flags_);
	}
	//------------------
	// class ts_boot_linear
	//------------------
	class ts_boot_linear: public ts_boot_model
	{
	public:
		ts_boot_linear();						
		ts_boot_linear(number option(nullable),number option(nullable),logical );
		override void 		init_ip(number option(nullable),number option(nullable), logical);
		override logical	is_init() ;						
	protected:
				
	};	
	
	ts_boot_linear.ts_boot_linear() : ts_boot_model(ts_model_type.TS_LIN)  
	{
		ip_ = ip_linear(left_der_,right_der_,flags_);
	}

	ts_boot_linear.ts_boot_linear(	number option(nullable) left_der,
									number option(nullable) right_der,
									//logical apply_hyman, 
									logical extrap) : ts_boot_model(ts_model_type.TS_LIN)  
	{ 
		init_ip(left_der,right_der,extrap); 
	}
		
	logical	ts_boot_linear.is_init() { return !null(ip_);}

	void ts_boot_linear.init_ip(number option(nullable) left_der,
								number option(nullable) right_der,
								//logical apply_hyman, 
								logical extrap) 				
	{ 
		right_der_ 	= right_der;
		left_der_ 	= left_der;
		set_flags(false, extrap);
		ip_ 		= ip_linear(left_der_,right_der_,flags_);
	}

	//------------------
	// class ts_boot_spline
	//------------------
	class ts_boot_spline: public ts_boot_model
	{
	public:
		ts_boot_spline();						
		ts_boot_spline(number option(nullable),number option(nullable),logical, logical );
		override void 		init_ip(number option(nullable),number option(nullable),logical, logical);
		override logical	is_init() ;						
	protected:
				
	};	
	
	ts_boot_spline.ts_boot_spline() : ts_boot_model(ts_model_type.TS_SPL)  
	{
		ip_ = ip_spline(left_der_,right_der_,flags_);
	}

	ts_boot_spline.ts_boot_spline(	number option(nullable) left_der,
									number option(nullable) right_der,
									logical apply_hyman, 
									logical extrap) : ts_boot_model(ts_model_type.TS_SPL)  
	{ 
		init_ip(left_der,right_der,apply_hyman,extrap); 
	}
		
	logical	ts_boot_spline.is_init() { return !null(ip_);}

	void ts_boot_spline.init_ip(number option(nullable) left_der,
								number option(nullable) right_der,
								logical apply_hyman, 
								logical extrap) 				
	{ 
		right_der_ 	= right_der;
		left_der_ 	= left_der;
		set_flags(apply_hyman, extrap);
		ip_ 		= ip_spline(left_der_,right_der_,flags_);
	}

	//------------------
	// class ts_boot_spline_nat
	// right_der = null<number>
	//------------------
	class ts_boot_spline_nat: public ts_boot_model
	{
	public:
		ts_boot_spline_nat();						
		ts_boot_spline_nat(number option(nullable),logical, logical );
		override void 		init_ip(number option(nullable),logical, logical);
		override logical	is_init() ;						
	protected:
				
	};	
	
	ts_boot_spline_nat.ts_boot_spline_nat() : ts_boot_model(ts_model_type.TS_SPL_NAT)  
	{
		ip_ = ip_spline(left_der_,null<number>,flags_);
	}

	ts_boot_spline_nat.ts_boot_spline_nat(	number option(nullable) left_der,									
											logical apply_hyman, 
											logical extrap) : ts_boot_model(ts_model_type.TS_SPL_NAT)  
	{ 
		init_ip(left_der,apply_hyman,extrap); 
	}
		
	logical	ts_boot_spline_nat.is_init() { return !null(ip_);}

	void ts_boot_spline_nat.init_ip(number option(nullable) left_der,								
									logical apply_hyman, 
									logical extrap) 				
	{ 
		left_der_ 	= left_der;
		set_flags(apply_hyman, extrap);
		ip_ 		= ip_spline(left_der_,null<number>,flags_);
	}

	//------------------
	// class ts_boot_spline_fin
	// right_der = 0
	//------------------
	class ts_boot_spline_fin: public ts_boot_model
	{
	public:
		ts_boot_spline_fin();						
		ts_boot_spline_fin(number option(nullable),logical, logical );
		override void 		init_ip(number option(nullable),logical, logical);
		override logical	is_init() ;						
	protected:
				
	};	
	
	ts_boot_spline_fin.ts_boot_spline_fin() : ts_boot_model(ts_model_type.TS_SPL_FIN)  
	{
		ip_ = ip_spline(left_der_,0,flags_);
	}

	ts_boot_spline_fin.ts_boot_spline_fin(	number option(nullable) left_der,									
											logical apply_hyman, 
											logical extrap) : ts_boot_model(ts_model_type.TS_SPL_FIN)  
	{ 
		init_ip(left_der,apply_hyman,extrap); 
	}
		
	logical	ts_boot_spline_fin.is_init() { return !null(ip_);}

	void ts_boot_spline_fin.init_ip(number option(nullable) left_der,								
									logical apply_hyman, 
									logical extrap) 				
	{ 
		left_der_ 	= left_der;
		set_flags(apply_hyman, extrap);
		ip_ 		= ip_spline(left_der_,0,flags_);
	}

	//------------------
	// class ts_boot_hermite_ak
	//------------------
	class ts_boot_hermite_ak: public ts_boot_model
	{
	public:
		ts_boot_hermite_ak();						
		ts_boot_hermite_ak(number option(nullable),number option(nullable),logical, logical );
		override void 		init_ip(number option(nullable),number option(nullable),logical, logical);
		override logical	is_init() ;						
	protected:
				
	};	
	
	ts_boot_hermite_ak.ts_boot_hermite_ak() : ts_boot_model(ts_model_type.TS_HERM_AK)    
	{
		ip_ = ip_hermite_akima(left_der_,right_der_,flags_);
	}

	ts_boot_hermite_ak.ts_boot_hermite_ak(	number option(nullable) left_der,
											number option(nullable) right_der,
											logical apply_hyman, 
											logical extrap) : ts_boot_model(ts_model_type.TS_HERM_AK)  
	{ 
		init_ip(left_der,right_der,apply_hyman,extrap); 
	}
		
	logical	ts_boot_hermite_ak.is_init() { return !null(ip_);}

	void ts_boot_hermite_ak.init_ip(number option(nullable) left_der,
								number option(nullable) right_der,
								logical apply_hyman, 
								logical extrap) 				
	{ 
		right_der_ 	= right_der;
		left_der_ 	= left_der;
		set_flags(apply_hyman, extrap);
		ip_ 		= ip_hermite_akima(left_der_,right_der_,flags_);
	}

	//------------------
	// class ts_boot_hermite_cr
	//------------------
	class ts_boot_hermite_cr: public ts_boot_model
	{
	public:
		ts_boot_hermite_cr();						
		ts_boot_hermite_cr(number option(nullable),number option(nullable),logical, logical );
		override void 		init_ip(number option(nullable),number option(nullable),logical, logical);
		override logical	is_init() ;						
	protected:
				
	};	
	
	ts_boot_hermite_cr.ts_boot_hermite_cr() : ts_boot_model(ts_model_type.TS_HERM_CR)      
	{
		ip_ = ip_hermite_catmull_rom(left_der_,right_der_,flags_);
	}

	ts_boot_hermite_cr.ts_boot_hermite_cr(	number option(nullable) left_der,
									number option(nullable) right_der,
									logical apply_hyman, 
									logical extrap) : ts_boot_model(ts_model_type.TS_HERM_CR)  
	{ 
		init_ip(left_der,right_der,apply_hyman,extrap); 
	}
		
	logical	ts_boot_hermite_cr.is_init() { return !null(ip_);}

	void ts_boot_hermite_cr.init_ip(number option(nullable) left_der,
								number option(nullable) right_der,
								logical apply_hyman, 
								logical extrap) 				
	{ 
		right_der_ 	= right_der;
		left_der_ 	= left_der;
		set_flags(apply_hyman, extrap);
		ip_ 		= ip_hermite_catmull_rom(left_der_,right_der_,flags_);
	}

	//------------------
	// class ts_boot_hermite_fd
	//------------------
	class ts_boot_hermite_fd: public ts_boot_model
	{
	public:
		ts_boot_hermite_fd();						
		ts_boot_hermite_fd(number option(nullable),number option(nullable),logical, logical );
		override void 		init_ip(number option(nullable),number option(nullable),logical, logical);
		override logical	is_init() ;						
	protected:
				
	};	
	
	ts_boot_hermite_fd.ts_boot_hermite_fd() : ts_boot_model(ts_model_type.TS_HERM_FD)      
	{
		ip_ = ip_hermite_finite_diff(left_der_,right_der_,flags_);
	}

	ts_boot_hermite_fd.ts_boot_hermite_fd(	number option(nullable) left_der,
									number option(nullable) right_der,
									logical apply_hyman, 
									logical extrap) : ts_boot_model(ts_model_type.TS_HERM_FD)  
	{ 
		init_ip(left_der,right_der,apply_hyman,extrap); 
	}
		
	logical	ts_boot_hermite_fd.is_init() { return !null(ip_);}

	void ts_boot_hermite_fd.init_ip(number option(nullable) left_der,
								number option(nullable) right_der,
								logical apply_hyman, 
								logical extrap) 				
	{ 
		right_der_ 	= right_der;
		left_der_ 	= left_der;
		set_flags(apply_hyman, extrap);
		ip_ 		= ip_hermite_finite_diff(left_der_,right_der_,flags_);
	}

	//------------------
	// class ts_boot_hermite_fb
	//------------------
	class ts_boot_hermite_fb: public ts_boot_model
	{
	public:
		ts_boot_hermite_fb();						
		ts_boot_hermite_fb(number option(nullable),number option(nullable),logical, logical );
		override void 		init_ip(number option(nullable),number option(nullable),logical, logical);
		override logical	is_init() ;						
	protected:
				
	};	
	
	ts_boot_hermite_fb.ts_boot_hermite_fb() : ts_boot_model(ts_model_type.TS_HERM_FB)     
	{
		ip_ = ip_hermite_fritsch_butland(left_der_,right_der_,flags_);
	}

	ts_boot_hermite_fb.ts_boot_hermite_fb(	number option(nullable) left_der,
									number option(nullable) right_der,
									logical apply_hyman, 
									logical extrap) : ts_boot_model(ts_model_type.TS_HERM_FB)  
	{ 
		init_ip(left_der,right_der,apply_hyman,extrap); 
	}
		
	logical	ts_boot_hermite_fb.is_init() { return !null(ip_);}

	void ts_boot_hermite_fb.init_ip(number option(nullable) left_der,
								number option(nullable) right_der,
								logical apply_hyman, 
								logical extrap) 				
	{ 
		right_der_ 	= right_der;
		left_der_ 	= left_der;
		set_flags(apply_hyman, extrap);
		ip_ 		= ip_hermite_fritsch_butland(left_der_,right_der_,flags_);
	}

	//------------------
	// class ts_boot_hermite_krug
	//------------------
	class ts_boot_hermite_krug: public ts_boot_model
	{
	public:
		ts_boot_hermite_krug();						
		ts_boot_hermite_krug(number option(nullable),number option(nullable),logical, logical );
		override void 		init_ip(number option(nullable),number option(nullable),logical, logical);
		override logical	is_init() ;						
	protected:
				
	};	
	
	ts_boot_hermite_krug.ts_boot_hermite_krug() : ts_boot_model(ts_model_type.TS_HERM_K)     
	{
		ip_ = ip_hermite_kruger(left_der_,right_der_,flags_);
	}

	ts_boot_hermite_krug.ts_boot_hermite_krug(	number option(nullable) left_der,
									number option(nullable) right_der,
									logical apply_hyman, 
									logical extrap) : ts_boot_model(ts_model_type.TS_HERM_K)  
	{ 
		init_ip(left_der,right_der,apply_hyman,extrap); 
	}
		
	logical	ts_boot_hermite_krug.is_init() { return !null(ip_);}

	void ts_boot_hermite_krug.init_ip(number option(nullable) left_der,
								number option(nullable) right_der,
								logical apply_hyman, 
								logical extrap) 				
	{ 
		right_der_ 	= right_der;
		left_der_ 	= left_der;
		set_flags(apply_hyman, extrap);
		ip_ 		= ip_hermite_kruger(left_der_,right_der_,flags_);
	}

	//------------------
	// class ts_boot_hermite_mono
	//------------------
	class ts_boot_hermite_mono: public ts_boot_model
	{
	public:
		ts_boot_hermite_mono();						
		ts_boot_hermite_mono(number option(nullable),number option(nullable),logical, logical );
		override void 		init_ip(number option(nullable),number option(nullable),logical, logical);
		override logical	is_init() ;						
	protected:
				
	};	
	
	ts_boot_hermite_mono.ts_boot_hermite_mono() : ts_boot_model(ts_model_type.TS_HERM_M)     
	{
		ip_ = ip_hermite_monotone(left_der_,right_der_,flags_);
	}

	ts_boot_hermite_mono.ts_boot_hermite_mono(	number option(nullable) left_der,
									number option(nullable) right_der,
									logical apply_hyman, 
									logical extrap) : ts_boot_model(ts_model_type.TS_HERM_M)  
	{ 
		init_ip(left_der,right_der,apply_hyman,extrap); 
	}
		
	logical	ts_boot_hermite_mono.is_init() { return !null(ip_);}

	void ts_boot_hermite_mono.init_ip(number option(nullable) left_der,
								number option(nullable) right_der,
								logical apply_hyman, 
								logical extrap) 				
	{ 
		right_der_ 	= right_der;
		left_der_ 	= left_der;
		set_flags(apply_hyman, extrap);
		ip_ 		= ip_hermite_monotone(left_der_,right_der_,flags_);
	}

	//------------------
	// class ts_boot_hermite_para
	//------------------
	class ts_boot_hermite_para: public ts_boot_model
	{
	public:
		ts_boot_hermite_para();						
		ts_boot_hermite_para(number option(nullable),number option(nullable),logical, logical );
		override void 		init_ip(number option(nullable),number option(nullable),logical, logical);
		override logical	is_init() ;						
	protected:
				
	};	
	
	ts_boot_hermite_para.ts_boot_hermite_para() : ts_boot_model(ts_model_type.TS_HERM_P)    
	{
		ip_ = ip_hermite_parabolic(left_der_,right_der_,flags_);
	}

	ts_boot_hermite_para.ts_boot_hermite_para(	number option(nullable) left_der,
									number option(nullable) right_der,
									logical apply_hyman, 
									logical extrap) : ts_boot_model(ts_model_type.TS_HERM_P)  
	{ 
		init_ip(left_der,right_der,apply_hyman,extrap); 
	}
		
	logical	ts_boot_hermite_para.is_init() { return !null(ip_);}

	void ts_boot_hermite_para.init_ip(number option(nullable) left_der,
								number option(nullable) right_der,
								logical apply_hyman, 
								logical extrap) 				
	{ 
		right_der_ 	= right_der;
		left_der_ 	= left_der;
		set_flags(apply_hyman, extrap);
		ip_ 		= ip_hermite_parabolic(left_der_,right_der_,flags_);
	}
//*****************************************************//

	ts_model create_ts_model(ts_model_type m) 
	{
		switch(m)
		{
		case ts_model_type.TS_NS_LAG:	
			return new ts_ns_laguerre() ;		
		case ts_model_type.TS_NS_BLISS:
			return new ts_ns_bliss() ;		
		case ts_model_type.TS_NS_SVEN:
			return new ts_ns_svensson() ;
		case ts_model_type.TS_NS:
			return new ts_ns_orig() ;
		case ts_model_type.TS_HERM_D:
			return new ts_hermite(ts_spline_type.ST_DISC) ;
		case ts_model_type.TS_HERM_S:
			return new ts_hermite(ts_spline_type.ST_SPOT) ;
		case ts_model_type.TS_HERM_F:
			return new ts_hermite(ts_spline_type.ST_FWD) ;		
		case ts_model_type.TS_BSPL_D:
			return new ts_bspline(ts_spline_type.ST_DISC) ;
		case ts_model_type.TS_BSPL_S:	
			return new ts_bspline(ts_spline_type.ST_SPOT) ;
		case ts_model_type.TS_BSPL_F:	
			return new ts_bspline(ts_spline_type.ST_FWD) ;
		case ts_model_type.TS_CSPL_D:	
			return new ts_cspline(ts_spline_type.ST_DISC) ;
		case ts_model_type.TS_CSPL_S:	
			return new ts_cspline(ts_spline_type.ST_SPOT) ;
		case ts_model_type.TS_CSPL_F:	
			return new ts_cspline(ts_spline_type.ST_FWD) ;
		case ts_model_type.TS_TANG:				 	
			return new ts_tanggaard() ;
		case ts_model_type.TS_TENS_W_BF:	
			return new ts_tension_w_bf() ;
		case ts_model_type.TS_TENS_Y_BF:	
			return new ts_tension_y_bf() ;
		case ts_model_type.TS_TENS_Z_BF:	
			return new ts_tension_z_bf() ;
		case ts_model_type.TS_LIN:	
			return new ts_boot_linear() ;
		case ts_model_type.TS_SPL:	
			return new ts_boot_spline() ;
		case ts_model_type.TS_SPL_NAT:	
			return new ts_boot_spline_nat() ;
		case ts_model_type.TS_SPL_FIN:	
			return new ts_boot_spline_fin() ;
		case ts_model_type.TS_STEP:	
			return new ts_boot_step() ;
		case ts_model_type.TS_HAGAN:	
			return new ts_boot_hagan_west() ;
		case ts_model_type.TS_HERM_AK:
			return new ts_boot_hermite_ak() ;
		case ts_model_type.TS_HERM_CR:			
			return new ts_boot_hermite_cr() ;
		case ts_model_type.TS_HERM_FD:
			return new ts_boot_hermite_fd() ;
		case ts_model_type.TS_HERM_FB:	
			return new ts_boot_hermite_fb() ;
		case ts_model_type.TS_HERM_K:	
			return new ts_boot_hermite_krug() ;
		case ts_model_type.TS_HERM_M:	
			return new ts_boot_hermite_mono() ;
		case ts_model_type.TS_HERM_P:	
			return new ts_boot_hermite_para() ;
		case ts_model_type.TS_MS:	
			return new ts_max_smooth() ;
		case ts_model_type.TS_MS_S:	
			return new ts_max_smooth_sprd() ;
		case ts_model_type.TS_TENS_W:	
			return new ts_tension_w() ;
		case ts_model_type.TS_TENS_Y:	
			return new ts_tension_y() ;
		case ts_model_type.TS_TENS_Z:	
			return new ts_tension_z() ;
		case ts_model_type.TS_TENS_W_S:	
			return new ts_tension_w_sprd() ;
		case ts_model_type.TS_TENS_Y_S:	
			return new ts_tension_y_sprd() ;
		case ts_model_type.TS_TENS_Z_S:	
			return new ts_tension_z_sprd() ;
		case ts_model_type.TS_SW_CEI:	
			return new ts_smith_wilson_ceiops() ;
		default:
			QL_FAIL("invalid model"); 
		}
	}
	ts_model ts_model_map(string model_name) 
	{
		return create_ts_model(ts_model_type_map(model_name));
	}	


	//------------------
	// wmse
	//------------------
	number wmse(vector(number)  v_obs,
				vector(number)  v_est,
				vector(number) weight)//weights sum to 1
	{
		vector(number) res = v_est - v_obs;
		res = res .* res;
		res = weight .* res;
		return v_sum(res);
	}
	//------------------
	// wmae
	//------------------
	number wmae(vector(number)  v_obs,
				vector(number)  v_est,
				vector(number) weight)//weights sum to 1
	{
		vector(number) res = v_est - v_obs;
		res = abs(res);
		res = weight .* res;
		return v_sum(res);
	}
	//------------------
	// mse
	//------------------
	number mse(	vector(number)  v_obs,
				vector(number)  v_est)
	{
		vector(number) res = v_est - v_obs;
		res = res .* res;
		return v_sum(res)/v_size(v_obs);
	}
	//------------------
	// mae
	//------------------
	number mae(	vector(number)  v_obs,
				vector(number)  v_est)
	{
		vector(number) res = v_est - v_obs;
		res = abs(res);
		return v_sum(res)/v_size(v_obs);
	}
	//------------------
	// rmse
	//------------------
	number rmse(vector(number)  v_obs,
				vector(number)  v_est)
	{
		return sqrt(mse(v_obs,v_est));
	}
	//------------------
	// maxae
	//------------------
	number maxae(vector(number)  v_obs,
				vector(number)  v_est, 
				out integer idx)
	{
		vector(number) res = v_est - v_obs;
		res = abs(res);
		number x = v_max(res);
		vector(integer) idx_v = ..v_find(res,[x]);
		idx = idx_v[0];
		return x;
	}
	//------------------
	// sae
	//------------------
	number sae(	vector(number)  v_obs,
				vector(number)  v_est)
	{
		vector(number) res = v_est - v_obs;
		res = abs(res);
		return v_sum(res);
	}
	//------------------
	// sse
	//------------------
	number sse(	vector(number)  v_obs,
				vector(number)  v_est)
	{
		vector(number) res = v_est - v_obs;
		vector(number) ress = res .* res;
		return v_sum(ress);
	}
	//------------------
	// hit
	//------------------
	number hit(	vector(number)  v_obs,
				vector(number)  v_est,
				number l)
	{
		vector(number) res = v_est - v_obs;
		res = abs(res);
		vector(integer) v = v_find_le(res, l);
		return v_size(v);
	}

	/* old version output:
	-The fitted norm, or an error value if unsuccessful.
	-Weighted mean absolute deviation (the l1 norm).
	-Weighted mean squared error (the l2 norm).
	-Sum of absolute deviations (the unweighted l1 norm).
	-Sum of squared errors (the unweighted l2 norm).
	-Max error (the l¥ norm).
	-Hit count, i.e. number of instruments priced correctly (within 0.05).
	-Weighted sum of relative errors.
*/
	//------------------
	// ts_fit_results_calc
	//------------------
	void ts_fit_results_calc(	vector(string) name, 
								vector(logical) incl, 
								vector(number)  model_clean, 
								vector(number)  clean, 
								vector(number)  model_yld, 
								vector(number)  yld,
								vector(number)  weight,
								out vector(number) res,
								out vector(string) aux)
	{
		vector(integer) i_v;

		for(integer i=0;i<v_size(incl);i++) {
			if(!incl[i])
				push_back(i_v,i);
		}
		
		vector(number)  model_yield; 
		vector(number)  yield;

		if(v_size(i_v) > 0) {
			model_clean = v_exclude(model_clean, i_v);
			clean 		= v_exclude(clean, i_v);
			model_yield	= v_exclude(model_yld, i_v);
			yield		= v_exclude(yld, i_v);
			weight 		= v_exclude(weight, i_v);
			name		= v_exclude(name, i_v);		
		}
		else {
			model_yield	= clone_vector(model_yld);
			yield		= clone_vector(yld);
		}
		//*10000 --> to get all diff-results in bp
		model_yield	*= 10000;
		yield		*= 10000;

		integer idx_y, idx_p;

		number count 		= v_size(model_clean);   	
		number abs_max_bp 	= maxae(yield,model_yield, idx_y);
		number abs_max_p 	= maxae(clean,model_clean, idx_p);
		number s_sq 		= sse(clean,model_clean);
		number s_sq_y 		= sse(yield,model_yield);
		number s_abs 		= sae(clean,model_clean);
		number s_abs_y 		= sae(yield,model_yield);
		number hit_count 	= hit(yield,model_yield,1.5);
		number hit_rate 	= hit_count/count*100.0;
		number mw_sq 		= wmse(clean,model_clean, weight);
		number mw_sq_y 		= wmse(yield,model_yield, weight);
		number mw_abs 		= wmae(clean,model_clean, weight);
		number mw_abs_y 	= wmae(yield,model_yield, weight);
		number norm 		= sqrt(s_sq);
		number norm_y 		= sqrt(s_sq_y);

		res	= [ abs_max_bp, abs_max_p, s_abs,s_abs_y, s_sq, s_sq_y, mw_abs, mw_abs_y, mw_sq,mw_sq_y, 
				hit_count, hit_rate, norm, norm_y, count];
		aux = [ name[idx_y], name[idx_p]];
		
		return;
	}
	//------------------
	// ts_fit_results
	//------------------
	void ts_fit_results(ts_model 		m,
						vector(string) 	name, 
						disc_func 		df, 
						fit_result option(nullable) fr, 
						std_weights option(nullable) w_scheme,
						vector(logical) incl,
						vector(number) 	model_clean, 				
						vector(number) 	clean,
						vector(number) 	model_yield, 				
						vector(number) 	yield,
						vector(number) 	weight,
						number			min_days,
						number 			max_days,
						out vector(string) param_names,
						out vector(number) param_vals,
						out vector(string) label,
						out vector(string) res,
						out vector(string) spread,
						out vector(number) spread_bp)
	{
		logical is_bf			= m.fit_t() == ts_fit_type.FIT_BEST_FIT;
		logical is_tension 		= model_is_tension(m.model_t());
		logical is_tension_bf 	= is_tension && is_bf;
		logical is_tang_bf 		= m.model_t() == ts_model_type.TS_TANG;

		if(!null(fr)){
			model m 		= fr.model();
			param_names 	= m.parameter_names();			
			param_vals		= fr.parameters();
		}
		else {
			resize(param_names,0);			
			resize(param_vals,0);
		}
		integer size = v_size(param_names);
		resize(label,size>14?size:14);
		resize(res,size>14?size:14);
		
		vector(number) calc_res	;
		vector(string) calc_aux	;
		ts_fit_results_calc(name,incl, model_clean, clean, model_yield, yield, weight, calc_res, calc_aux);

		logical weights_ok = is_bf && !null(w_scheme); 
		 
		string norm;
		if(is_tension_bf || is_tang_bf)
			norm ="L2";
		else	
			norm = !null(fr) ? strcat("L",str(fr.norm(),0)) : "n/a";

		integer n 		= 0;
		label[n] 	="Model name";
		res[n] 		= strcat([string(m.model_t())," <",(is_bf?"fit with residuals":"exact fit"),">"]);
		label[++n] 	="Optimization Norm";
		res[n] 		= norm;
		label[++n] 	="Weight Scheme";
		res[n] 		= weights_ok ? string(w_scheme) : "n/a";
		label[++n] 	="# Instruments";
		res[n] 		= str(calc_res[14],0);

		label[++n] 	= "Max Err (bp / price)";
		res[n] 		= strcat([str(calc_res[0],2),"  [",calc_aux[0],"]"]);
		res[n] 		= strcat([res[n]," / ", str(calc_res[1],3),"  [",calc_aux[1],"]"]);
		
		label[++n] 	="Sum Abs. Err (bp / price)";
		res[n] 		= strcat([str(calc_res[3],3)," / ",str(calc_res[2],3)]);
		label[++n] 	="Sum Sqr. Err (bp / price)";
		res[n] 		= strcat([str(calc_res[5],5)," / ",str(calc_res[4],5)]);

		label[++n] 	="Weighted Mean Abs. Err (bp / price)";
		res[n] 		= !weights_ok?"n/a":strcat([str(calc_res[7],3)," / ",str(calc_res[6],3)]);
		label[++n] 	="Weighted Mean Sqr. Err (bp / price)";
		res[n] 		= !weights_ok?"n/a":strcat([str(calc_res[9],5)," / ",str(calc_res[8],5)]);

		label[++n] 	="Norm Err (bp / price)";
		res[n] 		= !is_bf?"n/a":strcat([str(calc_res[13],3)," / ",str(calc_res[12],3)]);

		label[++n] 	="Hit count (<= 1.5bp)";
		res[n] 		= str(calc_res[10],0);

		label[++n] 	="Hit rate (%)";
		res[n] 		= str(calc_res[11],0);

		label[++n] 	="Min time-to-maturity";
		res[n] 		= str(min_days/365.0,3);

		label[++n] 	="Max time-to-maturity";
		res[n] 		= str(max_days/365.0,3);

		resize(spread,5);
		resize(spread_bp,5);
		number r30, r10, r5, r2,r3m;

		try {
			r30 = df.zero_rate_spot(30,RT_CONT);
		}
		catch {
			r30 = null<number>;
		}

		try {
			r10 = df.zero_rate_spot(10,RT_CONT);
		}
		catch {
			r10 = null<number>;
		}

		try {
			r5 = df.zero_rate_spot(5,RT_CONT);
		}
		catch {
			r5 = null<number>;
		}

		try {
			r2 = df.zero_rate_spot(2,RT_CONT);
		}
		catch {
			r2 = null<number>;
		}

		try {
			r3m = df.zero_rate_spot(0.25,RT_CONT);
		}
		catch {
			r3m = null<number>;
		}

		try {
			n = 0;
			spread[n] ="2y-3m";
			spread_bp[n] = (r2 -r3m)*10000;
			spread[++n] ="10y-3m";
			spread_bp[n] = (r10-r3m)*10000;
			spread[++n] ="10y-2y";
			spread_bp[n] = (r10-r2)*10000;
			spread[++n] ="10y-5y";
			spread_bp[n] = (r10-r5)*10000;
			spread[++n] ="30y-10y";
			spread_bp[n] = (r30 - r10)*10000;
		}
		catch {}

	}

	//--------------------------
	// class ts_fit
	//--------------------------
	class ts_fit //option (category: 'Yield Curve/005 Fitting')
	{
	public:
		ts_fit(string);	
		logical 	is_fit();
		logical 	is_boot();
		logical 	is_spec();
		logical 	is_bf();
		logical 	is_tension_bf();
		ts_fit_type ft();
		fit_result 	fit_res(curve);
		disc_func	disc_f(curve);

		void set_fit_model_parm_ns(fit_model_params);
		void set_fit_model_parm_ns_sven(fit_model_params);
		void set_fit_model_parm_ns_lag(fit_model_params);
		void set_fit_model_parm_ns_bliss(fit_model_params);
		void set_fit_model_parm_hermite(curve option(nullable), date option(nullable), string,
										vector(number) option(nullable) , number option(nullable));

		std_weights fit_w_scheme();
		void fit_ext(vector(instrument),vector(logical),logical,logical,
						out fit_result fr,
						out disc_func df);
		
	protected:
		void init();
		void init_fit(ts_fit_model);
		void init_boot(ts_boot_model);
		void init_spec();
		ts_model_type mt_;
		ts_model m_;
	
		
	};

	//--------------------------
	// constructor
	//--------------------------
	ts_fit.ts_fit(string model_name)
	{
		mt_ 	= ts_model_type_map(model_name) ;
		m_ 		= create_ts_model(mt_);		
	}

	logical 		ts_fit.is_fit() 		{ return m_.fit_func_t() == fit_func_type.FT_FIT; }
	logical 		ts_fit.is_boot() 		{ return m_.fit_func_t() == fit_func_type.FT_BOOT; }
	logical 		ts_fit.is_spec() 		{ return m_.fit_func_t() == fit_func_type.FT_SPEC; }
	logical 		ts_fit.is_bf() 			{ return ft() == ts_fit_type.FIT_BEST_FIT; }
	ts_fit_type 	ts_fit.ft()				{ return m_.fit_t(); }
	logical 		ts_fit.is_tension_bf() 	{ return is_bf() && model_is_tension(mt_);}

	//--------------------------
	// fit_ext
	//--------------------------
	void ts_fit.fit_ext(vector(instrument) instr,
						vector(logical) incl,
						logical  		remove_err,
						logical 		remove_nullq,
						out fit_result 	fr,
						out disc_func 	df)
	{
		
		integer size 	= v_size(instr);
		vector(instrument) instr_incl[0];
				
		for(integer i=0;i<size;i++) {	
			if(incl[i]) {
				push_back(instr_incl, instr[i] );
			}
		}

		vector(instrument) err_instr;
				
		curve c = curve(instr_incl,"",remove_err, remove_nullq, true,true,true);
		if(v_size(c.instruments()) != v_size(instr_incl)){
			instr_incl			= c.instruments();
			vector(integer) vf 	= ..v_find(instr.name(),instr_incl.name());
			incl 				= false;
			for(integer i=0;i<v_size(vf);i++) {
				incl[vf[i]] = true;
			}
		}
						
		if(is_fit()){
			fr = fit_res(c);
			df = fr;
		}
		else {
			fr = null<fit_result>;
			df = disc_f(c);
		}

		vector(string)  instr_name 		= instr.name();
		//vector(number)  instr_quote 	= instr.quote();
		vector(number)  model_clean 	= instr.clean_price(df,false,true);
		vector(number)  model_yield 	= instr.yield(df,false);
		vector(number)  yield			= instr.yield();
		vector(number)  clean 			= instr.clean_price(null<number>,false,true);
		vector(number)  model_vs_mkt 	= model_yield-yield;			
		vector(number)  weight[size];
		weight 							= null<number>;
		
		number sw = 0;

		if(is_bf()) {
			for(integer i=0;i<size;i++) {		
		
				if(incl[i]){					
					number w;
					switch(fit_w_scheme()){
					case WS_EQUAL:
						w = 1.0;
						break;	
					case WS_PVBP:
						w = 1.0/instr[i].pvbp(10000,0.0001);
						break;
					case WS_PVBP2:
						w = 1.0/sqr(instr[i].pvbp(10000,0.0001));
						break;
					default:
					}					
					weight[i] = w;	
					sw += w;	
				}	
			}	
		}
		date 		t = instr_incl[0].trade_date() ;
		date min_date = v_min(instr_incl.maturity());
		date max_date = v_max(instr_incl.maturity());				
		number min_d = null(min_date) ? null<number>:min_date-t;
		number max_d = null(max_date) ? null<number>:max_date-t;


		vector(string) param_names;
		vector(number) param_vals;
		vector(string) label;
		vector(string) res;
		vector(string) spread;
		vector(number) spread_bp;
					
		ts_fit_results(	m_, instr.name(), df, fr, fit_w_scheme(), incl, model_clean, clean, 
						model_yield, yield, weight, min_d, max_d,param_names,param_vals, label, res, spread, spread_bp );

		for(integer i=0; i < v_size(err_instr);i++) {
			if(null(err_instr[i]))
				push_back(instr_name, "removed: [null instrument]");
			else
				push_back(instr_name, strcat(["removed: [",err_instr[i].name(),"]"]));
		}

		model_yield 	*=100;
		yield 			*=100;
		model_vs_mkt 	*=10000;
	}

	//--------------------------
	// fit_w_scheme
	//--------------------------
	std_weights ts_fit.fit_w_scheme()
	{
		ts_fit_model mm = dynamic_cast<ts_fit_model>(m_);
		return mm.w_scheme();
	}
	
	//--------------------------
	// init
	//--------------------------
	void ts_fit.init( )
	{
		if(is_fit()){
			ts_fit_model mm = dynamic_cast<ts_fit_model>(m_);
			init_fit(mm);
		}
		else if(is_boot()){
			ts_boot_model mm = dynamic_cast<ts_boot_model>(m_);
			init_boot(mm);
		}
		else if(is_spec()){
			init_spec();
		}
		else {
			QL_FAIL("model not specified");
		}
		
	}
		
	//--------------------------
	// init_fit
	//--------------------------
	void ts_fit.init_fit(ts_fit_model mm )
	{
		switch(mt_) {
		case ts_model_type.TS_NS:
			//if(!null(p_))
				//mm.set_model_parm(p_);	
			break;	
		case ts_model_type.TS_NS_SVEN:
			//if(!null(p_))
				//mm.set_model_parm(p_);
			break;	
		case ts_model_type.TS_NS_LAG:
			//if(!null(p_))
				//mm.set_model_parm(p_);
			break;
		case ts_model_type.TS_NS_BLISS:
			//if(!null(p_))
				//mm.set_model_parm(p_);
			break;	
		case ts_model_type.TS_HERM_D:
		case ts_model_type.TS_HERM_S:
		case ts_model_type.TS_HERM_F:
			//mm.init_knots(c_b, t, gl_h_knot_scheme, gl_h_knots,gl_h_no_knots);			
			break;
		case ts_model_type.TS_BSPL_D:			
		case ts_model_type.TS_BSPL_S:	
		case ts_model_type.TS_BSPL_F:				
		case ts_model_type.TS_CSPL_D:				
		case ts_model_type.TS_CSPL_S:				
		case ts_model_type.TS_CSPL_F:
			//mm.init_knots(c_b, t, gl_bc_knot_scheme, gl_bc_knots,gl_bc_no_knots);	
			//mm.init_degree(integer(gl_bc_spline_degree));		
			break;
		}
		
	}

	//--------------------------
	// init_boot
	//--------------------------
	void ts_fit.init_boot(ts_boot_model mm )
	{
		//mm.set_parm(gl_boot_rate_type, gl_boot_on_interp, calendar());

		switch(mt_) {								
		case ts_model_type.TS_HAGAN:
			//mm.init_ip(gl_boot_force_pos,gl_boot_extrap) ;	
			break;
		case ts_model_type.TS_STEP:	
			//mm.init_ip(gl_boot_extrap) ;
			break;
		case ts_model_type.TS_SPL_NAT:
		case ts_model_type.TS_SPL_FIN:
			//mm.init_ip(gl_boot_left_der,gl_boot_apply_hyman,gl_boot_extrap) ;
			break;
		case ts_model_type.TS_LIN:
			//mm.init_ip(gl_boot_left_der,gl_boot_right_der,gl_boot_extrap) ;
			break;		
		default:
			//mm.init_ip(gl_boot_left_der,gl_boot_right_der,gl_boot_apply_hyman,gl_boot_extrap) ;
			break;
		}
	}

	//--------------------------
	// init_spec
	//--------------------------
	void ts_fit.init_spec( )
	{
		switch(mt_) {						
		case ts_model_type.TS_MS:{
			//ts_max_smooth mm = dynamic_cast<ts_max_smooth>(m_);
			//mm.init_parm(gl_ms_smooth);								
			break;}
		case ts_model_type.TS_MS_S:{
			//ts_max_smooth_sprd mm = dynamic_cast<ts_max_smooth_sprd>(m_);
			//mm.init_parm(gl_ms_smooth,[gl_ms_spread]);								
			break;}			
		case ts_model_type.TS_TANG:{
			//ts_tanggaard mm = dynamic_cast<ts_tanggaard>(m_);
			//mm.init_parm(gl_tang_smooth, gl_tang_weight_factor, gl_tang_w_scheme);
			//w_scheme = gl_tang_w_scheme;					
			break;}
		case ts_model_type.TS_TENS_W_BF:{
			//ts_tension_w_bf mm = dynamic_cast<ts_tension_w_bf>(m_);
			//mm.init_parm([gl_tens_min_sigma_x, gl_tens_max_sigma_x],[gl_tens_min_sigma_y, gl_tens_max_sigma_y], gl_tens_w_scheme, gl_tens_lambda, gl_tens_epsilon, gl_tens_left_der);
			//w_scheme = gl_tens_w_scheme;
			break;}
		case ts_model_type.TS_TENS_Y_BF:{
			//ts_tension_y_bf mm = dynamic_cast<ts_tension_y_bf>(m_);
			//mm.init_parm([gl_tens_min_sigma_x, gl_tens_max_sigma_x],[gl_tens_min_sigma_y, gl_tens_max_sigma_y],gl_tens_w_scheme, gl_tens_lambda, gl_tens_left_der,gl_tens_right_der);
			//w_scheme = gl_tens_w_scheme;
			break;}
		case ts_model_type.TS_TENS_Z_BF:{	
			//ts_tension_z_bf mm = dynamic_cast<ts_tension_z_bf>(m_);
			//mm.init_parm([gl_tens_min_sigma_x, gl_tens_max_sigma_x],[gl_tens_min_sigma_y, gl_tens_max_sigma_y], gl_tens_w_scheme, gl_tens_lambda);
			//w_scheme = gl_tens_w_scheme;
			break;}
		case ts_model_type.TS_TENS_W:	{
			//ts_tension_w mm = dynamic_cast<ts_tension_w>(m_);
			//mm.init_parm([gl_tens_min_sigma_x, gl_tens_max_sigma_x],[gl_tens_min_sigma_y, gl_tens_max_sigma_y], gl_tens_epsilon, gl_tens_left_der);				
			break;}
		case ts_model_type.TS_TENS_Y:	{
			//ts_tension_y mm = dynamic_cast<ts_tension_y>(m_);				
			//mm.init_parm([gl_tens_min_sigma_x, gl_tens_max_sigma_x],[gl_tens_min_sigma_y, gl_tens_max_sigma_y], gl_tens_left_der, gl_tens_right_der);				
			break;}
		case ts_model_type.TS_TENS_Z:	{
			//ts_tension_z mm = dynamic_cast<ts_tension_z>(m_);
			//mm.init_parm([gl_tens_min_sigma_x, gl_tens_max_sigma_x],[gl_tens_min_sigma_y, gl_tens_max_sigma_y]);				
			break;}
		case ts_model_type.TS_TENS_W_S:{	
			//ts_tension_w_sprd mm = dynamic_cast<ts_tension_w_sprd>(m_);
			//mm.init_parm([gl_tens_min_sigma_x, gl_tens_max_sigma_x],[gl_tens_min_sigma_y, gl_tens_max_sigma_y], gl_tens_epsilon, gl_tens_left_der,[gl_tens_spread]);				
			break;}
		case ts_model_type.TS_TENS_Y_S:{	
			//ts_tension_y_sprd mm = dynamic_cast<ts_tension_y_sprd>(m_);
			//mm.init_parm([gl_tens_min_sigma_x, gl_tens_max_sigma_x],[gl_tens_min_sigma_y, gl_tens_max_sigma_y], gl_tens_left_der, gl_tens_right_der,[gl_tens_spread]);				
			break;}
		case ts_model_type.TS_TENS_Z_S:{	
			//ts_tension_z_sprd mm = dynamic_cast<ts_tension_z_sprd>(m_);
			//mm.init_parm([gl_tens_min_sigma_x, gl_tens_max_sigma_x],[gl_tens_min_sigma_y, gl_tens_max_sigma_y],[gl_tens_spread]);				
			break;}
		case ts_model_type.TS_SW_CEI:{
			//ts_smith_wilson_ceiops mm = dynamic_cast<ts_smith_wilson_ceiops>(m_);
			//mm.init_parm(gl_sw_c_liq_t,gl_sw_c_ufr_t,gl_sw_c_ufr,gl_sw_c_interp,gl_sw_c_ufr_tol,gl_sw_c_a_start,gl_sw_c_solver,gl_sw_c_a_step);								
			break;}
		}
	}
	
	

	//--------------------------
	// fr
	//--------------------------
	fit_result ts_fit.fit_res(curve c)
	{
		if(m_.fit_func_t() == fit_func_type.FT_FIT){
			ts_fit_model mm = dynamic_cast<ts_fit_model>(m_);
			return mm.fit_fr(c) ;
		}
		else {
			QL_FAIL("not applicable for this model");
			//return null<fit_result>;
		}
	}
	//--------------------------
	// df
	//--------------------------
	disc_func ts_fit.disc_f(curve c)
	{
		return m_.fit(c);
	}

	//--------------------------
	// set_fit_model_parm_ns
	//--------------------------
	void ts_fit.set_fit_model_parm_ns(fit_model_params p)
	{
		QL_REQUIRE(mt_ == ts_model_type.TS_NS,"not applicable (wrong model)");
		ts_fit_model mm = dynamic_cast<ts_fit_model>(m_);
		mm.set_model_parm(p);
	}

	//--------------------------
	// set_fit_model_parm_ns_sven
	//--------------------------
	void ts_fit.set_fit_model_parm_ns_sven(fit_model_params p)
	{
		QL_REQUIRE(mt_ == ts_model_type.TS_NS_SVEN,"not applicable (wrong model)");
		ts_fit_model mm = dynamic_cast<ts_fit_model>(m_);
		mm.set_model_parm(p);
	}

	//--------------------------
	// set_fit_model_parm_ns_lag
	//--------------------------
	void ts_fit.set_fit_model_parm_ns_lag(fit_model_params p)
	{
		QL_REQUIRE(mt_ == ts_model_type.TS_NS_LAG,"not applicable (wrong model)");
		ts_fit_model mm = dynamic_cast<ts_fit_model>(m_);
		mm.set_model_parm(p);
	}

	//--------------------------
	// set_fit_model_parm_ns_bliss
	//--------------------------
	void ts_fit.set_fit_model_parm_ns_bliss(fit_model_params p)
	{
		QL_REQUIRE(mt_ == ts_model_type.TS_NS_BLISS,"not applicable (wrong model)");
		ts_fit_model mm = dynamic_cast<ts_fit_model>(m_);
		mm.set_model_parm(p);
	}

	//--------------------------
	// set_fit_model_parm_hermite
	//--------------------------
	void ts_fit.set_fit_model_parm_hermite(curve option(nullable) c, date option(nullable) t, string ks,
										   vector(number) option(nullable) knots, number option(nullable) no_knots)
	{
		QL_REQUIRE(model_is_fit_hermite(mt_),"not applicable (wrong model)");
		ts_hermite mm = dynamic_cast<ts_hermite>(m_);
		mm.set_knots(c, t, ks, knots,no_knots);
	}
}